//******************************************************************************
// file: AltiConfiguration.cc
// desc: configuration for the ALTI module
// auth: 28-NOV-2017 P. Kuzmanovic
//*****************************************************************************

#include "ALTI/AltiConfiguration.h"
#include "ALTI/AltiCommon.h"
#include "RCDUtilities/RCDUtilities.h"

#include <fstream>
#include <sstream>
#include <algorithm>

#include <iostream>
#include <iomanip>
#include <string>

using namespace LVL1;
using namespace std;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

AltiConfiguration::AltiConfiguration() {

    // default values

    clk_config                   = true;
    clk_select                   = AltiModule::PRIMARY_FXO;
    clk_jc_select                = AltiModule::PRIMARY_FXO;
    clk_jc_file                  = "";
    clk_phase_shift              = 0;
 
    sig_config                   = true;
    sig_fp_bgo2enable            = true;
    sig_fp_bgo3enable            = true;
    unsigned int i, j;
    for (j = 0; j < AltiModule::SIGNAL_DESTINATION_NUMBER; j++) {
        for (i = 0; i < AltiModule::SIGNAL_NUMBER_ROUTED; i++) {
            sig_swx[i][j] = AltiModule::SIGNAL_SOURCE::CTP_IN;
        }
        sig_swx_ttyp[j] = AltiModule::SIGNAL_SOURCE::CTP_IN;
    }
   
    sig_eqz_ctp_in_enable = false;
    sig_eqz_ctp_in_mode = AltiModule::SHORT_CABLE;
    sig_eqz_alti_in_enable = false;
    sig_eqz_alti_in_mode = AltiModule::SHORT_CABLE;
    sig_io_sync_enable.resize(AltiModule::ASYNC_INPUT_SIGNAL_NUMBER);
    sig_io_sync_phase.resize(AltiModule::ASYNC_INPUT_SIGNAL_NUMBER);
    sig_io_shape_enable.resize(AltiModule::ASYNC_INPUT_SIGNAL_NUMBER);
    for (i = 0; i < AltiModule::ASYNC_INPUT_SIGNAL_NUMBER; i++) {
        sig_io_sync_enable[i]    = true;
        sig_io_sync_phase[i]     = 0; // auto
        sig_io_shape_enable[i]   = false;
    }

    bsy_config                   = true;
    bsy_in_select.resize(AltiModule::BUSY_INPUT_NUMBER);
    for (i = 0; i < AltiModule::BUSY_INPUT_NUMBER; i++) {
        bsy_in_select[i]         = false;
    }
    bsy_out_select.resize(AltiModule::BUSY_OUTPUT_NUMBER);
    for (i = 0; i < AltiModule::BUSY_OUTPUT_NUMBER; i++) {
        bsy_out_select[i]        = AltiModule::BUSY_INACTIVE;
    }
    bsy_data_vme                 = false;
    bsy_masking_l1a_enable       = false;
    bsy_level                    = AltiModule::NIM;

    crq_config                   = true;
    crq_in_select.resize(3);
    for (i = 0; i < 3; i++) {
        crq_in_select[i]         = AltiModule::CALREQ_FROM_FRONT_PANEL;
    }
    crq_out_select[0].resize(3);
    crq_out_select[1].resize(3);
    for (i = 0; i < 3; i++) {
        crq_out_select[0][i]     = AltiModule::CALREQ_INACTIVE;
        crq_out_select[1][i]     = AltiModule::CALREQ_INACTIVE;
    }
    crq_data_vme                 = 0;

    mem_config                   = true;
    mem_pg_enable                = false;
    mem_pg_config.resize(AltiModule::SIGNAL_PG_NUMBER);
    for (i = 0; i < AltiModule::SIGNAL_PG_NUMBER; i++) {
        mem_pg_config[i]         = false;
    }
    mem_pg_repeat                = false;
    mem_pg_start_addr            = 0x00000000;
    mem_pg_stop_addr             = 0x00000000;
    mem_pg_file                  = "";
    mem_snap_masked.resize(AltiModule::SIGNAL_NUMBER);
    for (i = 0; i < AltiModule::SIGNAL_NUMBER; i++) {
        mem_snap_masked[i]       = false;
    }

    ttc_config                   = true;
    for (i = 0; i < AltiModule::TRANSMITTER_NUMBER; i++) {
        ttc_tx_enable[i]         = false;
    };
    ttc_l1a_source.resize(4);
    ttc_l1a_source[0] = true; // L1A
    for (i = 1; i < 4; i++) ttc_l1a_source[i] = false; // TTR1-3
    ttc_bgo_inhibit_width.resize(4);
    ttc_bgo_inhibit_delay.resize(4);
    for (i = 0; i <= 3; i++) {
        ttc_bgo_inhibit_width[i] = 0;
        ttc_bgo_inhibit_delay[i] = 0;
    }
    ttc_ttyp_delay               = 2;
    ttc_ttyp_addr                = 0x000;
    ttc_ttyp_subaddr             = 0x00;
    ttc_ttyp_frame_type          = AltiModule::TTC_FRAME_TYPE::LONG;
    ttc_ttyp_address_space       = AltiModule::TTC_ADDRESS_SPACE::EXTERNAL;
    ttc_bgo_mode.resize(4);
    for (i = 0; i <= 3; i++) {
        ttc_bgo_mode[i] = AltiModule::TTC_BGO_MODE::ASYNCHRONOUS_VME_ON_TRIGGER;
    }
    ttc_bgo_busy_gate.resize(4);
    for (i = 0; i <= 3; i++) {
        ttc_bgo_busy_gate[i] = false;
    }

    ttc_fifo_retransmit.resize(AltiModule::TTC_FIFO_NUMBER);
    for (i = 0; i < AltiModule::TTC_FIFO_NUMBER; i++) {
        ttc_fifo_retransmit[i] = false;
    }

    cnt_config                  = true;
    for (i = 0; i < AltiModule::CNT_NUMBER; i++) {
        cnt_enable[i] = true;
    };
    cnt_ttyp_lut_file           = "0";

    ctp_config                  = true;

    setDefaultFileNames();
}

//------------------------------------------------------------------------------

AltiConfiguration::AltiConfiguration(const std::string &pfx) : AltiConfiguration() {

    // add file prefix
    addFileNamePrefix(pfx);
}

//------------------------------------------------------------------------------

AltiConfiguration::~AltiConfiguration() {

    // nothing to be done
}


//------------------------------------------------------------------------------

void AltiConfiguration::setDefaultFileNames() {

    mem_pg_file                  = "ALTI/data/pg_alti.dat";
    cnt_ttyp_lut_file            = "data/AltiModule_cnt_ttyp_lut.dat";
}

//------------------------------------------------------------------------------

void AltiConfiguration::clearFileNames() {

    mem_pg_file                 = "";
    cnt_ttyp_lut_file           = "";
}

//------------------------------------------------------------------------------

void AltiConfiguration::addFileNamePrefix(const std::string &pfx) {

    mem_pg_file                 = pfx + mem_pg_file;
    cnt_ttyp_lut_file           = pfx + cnt_ttyp_lut_file;
}

//------------------------------------------------------------------------------

void AltiConfiguration::config(const bool cfg) {

    clk_config  = cfg;
    sig_config  = cfg;
    mem_config  = cfg;
    bsy_config  = cfg;
    crq_config  = cfg;
    ttc_config  = cfg;
    cnt_config  = cfg;
    ctp_config  = cfg;
}

//------------------------------------------------------------------------------

void AltiConfiguration::print(std::ostream &s) const {

    unsigned int i;
    char os[STRING_LENGTH];

    std::vector<std::string> arg;
    //int num;

    s << std::endl;
    std::sprintf(os, "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n"); s << os;
    std::sprintf(os, "AltiConfiguration:\n"); s << os;
    std::sprintf(os, "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n"); s << os;

    if (clk_config) {
        std::sprintf(os, "CLK source                           = %s\n", AltiModule::CLK_NAME[clk_select].c_str()); s << os;
        std::sprintf(os, "CLK jitter cleaner source            = %s\n", AltiModule::CLK_NAME[clk_jc_select].c_str()); s << os;
	std::sprintf(os, "CLK jitter cleaner configuration     = %s\n", clk_jc_file == "" ? "DEFAULT" : clk_jc_file.c_str()); s << os;
	std::sprintf(os, "CLK output phase shift               = %f ns\n", clk_phase_shift*0.015); s << os;
        s << "------------------------------------------------------------------------------------------------------" << std::endl;
    }

    if (sig_config) {
        unsigned int i, ttyp = 0;
        AltiModule::SIGNAL signal;
        std::sprintf(os, "SIG BGO front panel output selection = %s, %s\n", sig_fp_bgo2enable ? "BGO2" : "BGO0", sig_fp_bgo3enable ? "BGO3" : "BGO1"); s << os;
        for (i = 0; i < AltiModule::SIGNAL_NUMBER_ROUTED; i++) {
            signal = (AltiModule::SIGNAL) i;
            if ((AltiModule::SIGNAL_NAME[signal]).rfind("TTYP", 0) == 0) {
                if (ttyp++ == 0) {
                    std::sprintf(os, "SIG %-5s   switch    configuration  = %s, %s, %s, %s\n", "TTYP", AltiModule::SIGNAL_SOURCE_NAME[sig_swx_ttyp[0]].c_str(), AltiModule::SIGNAL_SOURCE_NAME[sig_swx_ttyp[1]].c_str(), AltiModule::SIGNAL_SOURCE_NAME[sig_swx_ttyp[2]].c_str(), AltiModule::SIGNAL_SOURCE_NAME[sig_swx_ttyp[3]].c_str()); s << os;
                }
            }
            else {
                std::sprintf(os, "SIG %-5s   switch    configuration  = %s, %s, %s, %s\n", AltiModule::SIGNAL_NAME[signal].c_str(), AltiModule::SIGNAL_SOURCE_NAME[sig_swx[i][0]].c_str(), AltiModule::SIGNAL_SOURCE_NAME[sig_swx[i][1]].c_str(), AltiModule::SIGNAL_SOURCE_NAME[sig_swx[i][2]].c_str(), AltiModule::SIGNAL_SOURCE_NAME[sig_swx[i][3]].c_str()); s << os;
            }
        }
      
	std::sprintf(os, "SIG %-7s equalizer %s, configuration  = %s\n", AltiModule::SIGNAL_SOURCE_NAME[AltiModule::CTP_IN].c_str(), sig_eqz_ctp_in_enable ? "ENABLED" : "DISABLED", AltiModule::EQUALIZER_CONFIG_NAME[sig_eqz_ctp_in_mode].c_str()); s << os;
	std::sprintf(os, "SIG %-7s equalizer %s, configuration  = %s\n", AltiModule::SIGNAL_SOURCE_NAME[AltiModule::ALTI_IN].c_str(), sig_eqz_alti_in_enable ? "ENABLED" : "DISABLED", AltiModule::EQUALIZER_CONFIG_NAME[sig_eqz_alti_in_mode].c_str()); s << os;

        bool sync_enable[8];
        unsigned int phase[8];
        bool shape_enable[8];

        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_ORB];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::SWX_ORB];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_ORB];
        std::sprintf(os, "SIG IO sync  ORB      from switch    = %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape ORB      from switch    = %s\n", bool2string(shape_enable[0]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_L1A];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::SWX_L1A];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_L1A];
        std::sprintf(os, "SIG IO sync  L1A      from switch    = %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape L1A      from switch    = %s\n", bool2string(shape_enable[0]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO0];
        sync_enable[1] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO1];
        sync_enable[2] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO2];
        sync_enable[3] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO3];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO0];
        phase[1] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO1];
        phase[2] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO2];
        phase[3] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO3];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO0];
        shape_enable[1] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO1];
        shape_enable[2] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO2];
        shape_enable[3] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO3];
        std::sprintf(os, "SIG IO sync  BGO0..3  from switch    = %s,%s, %s,%s, %s,%s, %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str(), bool2string(sync_enable[1]).c_str(), phase2string(phase[1]).c_str(), bool2string(sync_enable[2]).c_str(), phase2string(phase[2]).c_str(), bool2string(sync_enable[3]).c_str(), phase2string(phase[3]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape BGO0..3  from switch    = %s,%s,%s,%s\n", bool2string(shape_enable[0]).c_str(), bool2string(shape_enable[1]).c_str(), bool2string(shape_enable[2]).c_str(), bool2string(shape_enable[3]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP0];
        sync_enable[1] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP1];
        sync_enable[2] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP2];
        sync_enable[3] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP3];
        sync_enable[4] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP4];
        sync_enable[5] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP5];
        sync_enable[6] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP6];
        sync_enable[7] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP7];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP0];
        phase[1] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP1];
        phase[2] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP2];
        phase[3] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP3];
        phase[4] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP4];
        phase[5] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP5];
        phase[6] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP6];
        phase[7] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP7];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP0];
        shape_enable[1] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP1];
        shape_enable[2] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP2];
        shape_enable[3] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP3];
        shape_enable[4] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP4];
        shape_enable[5] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP5];
        shape_enable[6] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP6];
        shape_enable[7] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP7];
        std::sprintf(os, "SIG IO sync  TTYP0..3 from switch    = %s,%s, %s,%s, %s,%s, %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str(), bool2string(sync_enable[1]).c_str(), phase2string(phase[1]).c_str(), bool2string(sync_enable[2]).c_str(), phase2string(phase[2]).c_str(), bool2string(sync_enable[3]).c_str(), phase2string(phase[3]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape TTYP0..3 from switch    = %s,%s,%s,%s\n", bool2string(shape_enable[0]).c_str(), bool2string(shape_enable[1]).c_str(), bool2string(shape_enable[2]).c_str(), bool2string(shape_enable[3]).c_str()); s << os;
        std::sprintf(os, "SIG IO sync  TTYP4..7 from switch    = %s,%s, %s,%s, %s,%s, %s,%s\n", bool2string(sync_enable[4]).c_str(), phase2string(phase[4]).c_str(), bool2string(sync_enable[5]).c_str(), phase2string(phase[5]).c_str(), bool2string(sync_enable[6]).c_str(), phase2string(phase[6]).c_str(), bool2string(sync_enable[7]).c_str(), phase2string(phase[7]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape TTYP4..7 from switch    = %s,%s,%s,%s\n", bool2string(shape_enable[4]).c_str(), bool2string(shape_enable[5]).c_str(), bool2string(shape_enable[6]).c_str(), bool2string(shape_enable[7]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTR1];
        sync_enable[1] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTR2];
        sync_enable[2] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTR3];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTR1];
        phase[1] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTR2];
        phase[2] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTR3];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTR1];
        shape_enable[1] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTR2];
        shape_enable[2] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTR3];
        std::sprintf(os, "SIG IO sync  TTR1..3  from switch    = %s,%s, %s,%s, %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str(), bool2string(sync_enable[1]).c_str(), phase2string(phase[1]).c_str(), bool2string(sync_enable[2]).c_str(), phase2string(phase[2]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape TTR1..3  from switch    = %s,%s,%s\n", bool2string(shape_enable[0]).c_str(), bool2string(shape_enable[1]).c_str(), bool2string(shape_enable[2]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::LEMO_BUSY];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::LEMO_BUSY];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::LEMO_BUSY];
        std::sprintf(os, "SIG IO sync  BUSY     from LEMO      = %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape BUSY     from LEMO      = %s\n", bool2string(shape_enable[0]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::RJ45_CALREQ0];
        sync_enable[1] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::RJ45_CALREQ1];
        sync_enable[2] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::RJ45_CALREQ2];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::RJ45_CALREQ0];
        phase[1] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::RJ45_CALREQ1];
        phase[2] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::RJ45_CALREQ2];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::RJ45_CALREQ0];
        shape_enable[1] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::RJ45_CALREQ1];
        shape_enable[2] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::RJ45_CALREQ2];
        std::sprintf(os, "SIG IO sync  CALREQ0..2 from RJ45    = %s,%s, %s,%s, %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str(), bool2string(sync_enable[1]).c_str(), phase2string(phase[1]).c_str(), bool2string(sync_enable[2]).c_str(), phase2string(phase[2]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape CALREQ0..2 from RJ45    = %s,%s,%s\n", bool2string(shape_enable[0]).c_str(), bool2string(shape_enable[1]).c_str(), bool2string(shape_enable[2]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::CTP_BUSY];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::CTP_BUSY];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::CTP_BUSY];
        std::sprintf(os, "SIG IO sync  BUSY     from CTP       = %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape BUSY     from CTP       = %s\n", bool2string(shape_enable[0]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::CTP_CALREQ0];
        sync_enable[1] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::CTP_CALREQ1];
        sync_enable[2] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::CTP_CALREQ2];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::CTP_CALREQ0];
        phase[1] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::CTP_CALREQ1];
        phase[2] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::CTP_CALREQ2];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::CTP_CALREQ0];
        shape_enable[1] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::CTP_CALREQ1];
        shape_enable[2] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::CTP_CALREQ2];
        std::sprintf(os, "SIG IO sync  CALREQ0..2 from CTP     = %s,%s, %s,%s, %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str(), bool2string(sync_enable[1]).c_str(), phase2string(phase[1]).c_str(), bool2string(sync_enable[2]).c_str(), phase2string(phase[2]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape CALREQ0..2 from CTP     = %s,%s,%s\n", bool2string(shape_enable[0]).c_str(), bool2string(shape_enable[1]).c_str(), bool2string(shape_enable[2]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::ALTI_BUSY];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::ALTI_BUSY];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::ALTI_BUSY];
        std::sprintf(os, "SIG IO sync  BUSY     from ALTI      = %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape BUSY     from ALTI      = %s\n", bool2string(shape_enable[0]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::ALTI_CALREQ0];
        sync_enable[1] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::ALTI_CALREQ1];
        sync_enable[2] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::ALTI_CALREQ2];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::ALTI_CALREQ0];
        phase[1] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::ALTI_CALREQ1];
        phase[2] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::ALTI_CALREQ2];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::ALTI_CALREQ0];
        shape_enable[1] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::ALTI_CALREQ1];
        shape_enable[2] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::ALTI_CALREQ2];
        std::sprintf(os, "SIG IO sync  CALREQ0..2 from ALTI    = %s,%s, %s,%s, %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str(), bool2string(sync_enable[1]).c_str(), phase2string(phase[1]).c_str(), bool2string(sync_enable[2]).c_str(), phase2string(phase[2]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape CALREQ0..2 from ALTI    = %s,%s,%s\n", bool2string(shape_enable[0]).c_str(), bool2string(shape_enable[1]).c_str(), bool2string(shape_enable[2]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::CDR_LOCK];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::CDR_LOCK];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::CDR_LOCK];
        std::sprintf(os, "SIG IO sync  LOCK     from CDR       = %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape LOCK     from CDR       = %s\n", bool2string(shape_enable[0]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::CDR_L1A];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::CDR_L1A];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::CDR_L1A];
        std::sprintf(os, "SIG IO sync  L1A      from CDR       = %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape L1A      from CDR       = %s\n", bool2string(shape_enable[0]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::CDR_BRCST];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::CDR_BRCST];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::CDR_BRCST];
        std::sprintf(os, "SIG IO sync  BRCST    from CDR       = %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape BRCST    from CDR       = %s\n", bool2string(shape_enable[0]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::CDR_INDIV];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::CDR_INDIV];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::CDR_INDIV];
        std::sprintf(os, "SIG IO sync  INDIV    from CDR       = %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape INDIV    from CDR       = %s\n", bool2string(shape_enable[0]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::CDR_BCR];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::CDR_BCR];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::CDR_BCR];
        std::sprintf(os, "SIG IO sync  BCR      from CDR       = %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape BCR      from CDR       = %s\n", bool2string(shape_enable[0]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::CDR_ECR];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::CDR_ECR];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::CDR_ECR];
        std::sprintf(os, "SIG IO sync  ECR      from CDR       = %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape ECR      from CDR       = %s\n", bool2string(shape_enable[0]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::CDR_SERR];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::CDR_SERR];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::CDR_SERR];
        std::sprintf(os, "SIG IO sync  SERR     from CDR       = %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape SERR     from CDR       = %s\n", bool2string(shape_enable[0]).c_str()); s << os;
        sync_enable[0] = sig_io_sync_enable[AltiModule::ASYNC_INPUT_SIGNAL::CDR_DERR];
        phase[0] = sig_io_sync_phase[AltiModule::ASYNC_INPUT_SIGNAL::CDR_DERR];
        shape_enable[0] = sig_io_shape_enable[AltiModule::ASYNC_INPUT_SIGNAL::CDR_DERR];
        std::sprintf(os, "SIG IO sync  DERR     from CDR       = %s,%s\n", bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str()); s << os;
        std::sprintf(os, "SIG IO shape DERR     from CDR       = %s\n", bool2string(shape_enable[0]).c_str()); s << os;

        s << "------------------------------------------------------------------------------------------------------" << std::endl;
    }

    if (bsy_config) {
        std::sprintf(os, "BSY from front panel                 = %s\n", bsy_in_select[0] ? "ENABLED" : "DISABLED"); s << os;
        std::sprintf(os, "BSY from CTP                         = %s\n", bsy_in_select[1] ? "ENABLED" : "DISABLED"); s << os;
        std::sprintf(os, "BSY from ALTI                        = %s\n", bsy_in_select[2] ? "ENABLED" : "DISABLED"); s << os;
        std::sprintf(os, "BSY from pattern generation          = %s\n", bsy_in_select[3] ? "ENABLED" : "DISABLED"); s << os;
        std::sprintf(os, "BSY from VME register                = %s\n", bsy_in_select[4] ? "ENABLED" : "DISABLED"); s << os;
        std::sprintf(os, "BSY from ECR                         = %s\n", bsy_in_select[5] ? "ENABLED" : "DISABLED"); s << os;
        /*num = */AltiCommon::splitString(AltiModule::BUSY_SOURCE_NAME[bsy_out_select[0]], "_", arg);
        std::sprintf(os, "BSY to CTP                           = %s\n", arg[1].c_str()); s << os;
        /*num = */AltiCommon::splitString(AltiModule::BUSY_SOURCE_NAME[bsy_out_select[1]], "_", arg);
        std::sprintf(os, "BSY to ALTI                          = %s\n", arg[1].c_str()); s << os;
        std::sprintf(os, "BSY const level VME register         = %d\n", bsy_data_vme); s << os;
        std::sprintf(os, "BSY masking of L1A                   = %s\n", bsy_masking_l1a_enable ? "ENABLED" : "DISABLED"); s << os;
	std::sprintf(os, "BSY Front-Panel level                = %s\n", AltiModule::BUSY_LEVEL_NAME[bsy_level].c_str()); s << os;
        s << "------------------------------------------------------------------------------------------------------" << std::endl;
    }

    if (crq_config) {
        std::sprintf(os, "CRQ source                           = %s,%s,%s\n", AltiModule::CALREQ_INPUT_NAME[crq_in_select[2]].c_str(), AltiModule::CALREQ_INPUT_NAME[crq_in_select[1]].c_str(), AltiModule::CALREQ_INPUT_NAME[crq_in_select[0]].c_str()); s << os;
        std::sprintf(os, "CRQ to CTP                           = %s,%s,%s\n", AltiModule::CALREQ_SOURCE_NAME[crq_out_select[0][2]].c_str(), AltiModule::CALREQ_SOURCE_NAME[crq_out_select[0][1]].c_str(), AltiModule::CALREQ_SOURCE_NAME[crq_out_select[0][0]].c_str()); s << os;
        std::sprintf(os, "CRQ to ALTI                          = %s,%s,%s\n", AltiModule::CALREQ_SOURCE_NAME[crq_out_select[1][2]].c_str(), AltiModule::CALREQ_SOURCE_NAME[crq_out_select[1][1]].c_str(), AltiModule::CALREQ_SOURCE_NAME[crq_out_select[1][0]].c_str()); s << os;
        std::sprintf(os, "CRQ const level VME register         = %d\n", crq_data_vme); s << os;
        s << "------------------------------------------------------------------------------------------------------" << std::endl;
    }

    if (mem_config) {
      
        std::sprintf(os, "MEM pattern generation start address = 0x%08x\n", mem_pg_start_addr); s << os;
        std::sprintf(os, "MEM pattern generation stop  address = 0x%08x\n", mem_pg_stop_addr); s << os;
        std::sprintf(os, "MEM pattern generation file          = %s\n", mem_pg_file.c_str()); s << os;
        std::sprintf(os, "MEM pattern generation               = %s\n", mem_pg_enable ? "ENABLED" : "DISABLED"); s << os;
        std::sprintf(os, "MEM pattern generation ORB           = %s\n", mem_pg_config[AltiModule::PG_ORB] ? "ENABLED" : "DISABLED"); s << os;
        std::sprintf(os, "MEM pattern generation L1A           = %s\n", mem_pg_config[AltiModule::PG_L1A] ? "ENABLED" : "DISABLED"); s << os;
        std::sprintf(os, "MEM pattern generation TTR1..3       = %s,%s,%s\n", mem_pg_config[AltiModule::PG_TTR1] ? "ENABLED" : "DISABLED", mem_pg_config[AltiModule::PG_TTR2] ? "ENABLED" : "DISABLED", mem_pg_config[AltiModule::PG_TTR3] ? "ENABLED" : "DISABLED"); s << os;
        std::sprintf(os, "MEM pattern generation TTYP          = %s\n", mem_pg_config[AltiModule::PG_TTYP] ? "ENABLED" : "DISABLED"); s << os;
        std::sprintf(os, "MEM pattern generation BGO0..3       = %s,%s,%s,%s\n", mem_pg_config[AltiModule::PG_BGO0] ? "ENABLED" : "DISABLED", mem_pg_config[AltiModule::PG_BGO1] ? "ENABLED" : "DISABLED", mem_pg_config[AltiModule::PG_BGO2] ? "ENABLED" : "DISABLED", mem_pg_config[AltiModule::PG_BGO3] ? "ENABLED" : "DISABLED"); s << os;
        std::sprintf(os, "MEM pattern generation repeat        = %s\n", mem_pg_repeat ? "REPEATED" : "ONESHOT"); s << os;
        std::sprintf(os, "MEM snapshot mask ORB                = %s\n", mem_snap_masked[AltiModule::ORB] ? "MASKED" : "NOT_MASKED"); s << os;
        std::sprintf(os, "MEM snapshot mask L1A                = %s\n", mem_snap_masked[AltiModule::L1A] ? "MASKED" : "NOT_MASKED"); s << os;
        std::sprintf(os, "MEM snapshot mask TTR1..3            = %s,%s,%s\n", mem_snap_masked[AltiModule::TTR1] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::TTR2] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::TTR3] ? "MASKED" : "NOT_MASKED"); s << os;
        std::sprintf(os, "MEM snapshot mask TTYP0..3           = %s,%s,%s,%s\n", mem_snap_masked[AltiModule::TTYP0] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::TTYP1] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::TTYP2] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::TTYP3] ? "MASKED" : "NOT_MASKED"); s << os;
        std::sprintf(os, "MEM snapshot mask TTYP4..7           = %s,%s,%s,%s\n", mem_snap_masked[AltiModule::TTYP4] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::TTYP5] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::TTYP6] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::TTYP7] ? "MASKED" : "NOT_MASKED"); s << os;
        std::sprintf(os, "MEM snapshot mask BGO0..3            = %s,%s,%s,%s\n", mem_snap_masked[AltiModule::BGO0] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::BGO1] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::BGO2] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::BGO3] ? "MASKED" : "NOT_MASKED"); s << os;
        std::sprintf(os, "MEM snapshot mask BUSY               = %s\n", mem_snap_masked[AltiModule::BUSY] ? "MASKED" : "NOT_MASKED"); s << os;
        std::sprintf(os, "MEM snapshot mask CALREQ0..2         = %s,%s,%s\n", mem_snap_masked[AltiModule::CALREQ0] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::CALREQ1] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::CALREQ2] ? "MASKED" : "NOT_MASKED"); s << os;
        s << "------------------------------------------------------------------------------------------------------" << std::endl;
    }

    if (ttc_config) {
        for (i = 0; i < AltiModule::TRANSMITTER_NUMBER; i++) {
            std::sprintf(os, "TTC TX[%02d]                           = %s\n", i, ttc_tx_enable[i] ? "ENABLED" : "DISABLED"); s << os;
        }
        std::sprintf(os, "TTC encoder L1 source from L1A             = %s\n", ttc_l1a_source[0] ? "ENABLED" : "DISABLED"); s << os;
	std::sprintf(os, "TTC encoder L1 source from TTR1..3         = %s,%s,%s\n", ttc_l1a_source[1] ? "ENABLED" : "DISABLED", ttc_l1a_source[2] ? "ENABLED" : "DISABLED", ttc_l1a_source[3] ? "ENABLED" : "DISABLED"); s << os;
        for (i = 0; i <= 3; i++) {
            std::sprintf(os, "TTC encoder BGO%d inhibit parameters  = width %4dBCs, delay %4dBCs\n", i, ttc_bgo_inhibit_width[i], ttc_bgo_inhibit_delay[i]); s << os;
        }
        std::sprintf(os, "TTC encoder TTYP delay               =                      %4dBCs\n", ttc_ttyp_delay); s << os;
        std::sprintf(os, "TTC encoder TTYP word settings       = address 0x%04x, subaddress 0x%02x, %s, %s\n", ttc_ttyp_addr, ttc_ttyp_subaddr, AltiModule::TTC_FRAME_TYPE_NAME[ttc_ttyp_frame_type].c_str(), AltiModule::TTC_ADDRESS_SPACE_NAME[ttc_ttyp_address_space].c_str()); s << os;
        for (i = 0; i <= 3; i++) {
            std::sprintf(os, "TTC encoder BGO%d mode                = %s\n", i, AltiModule::TTC_BGO_MODE_NAME[ttc_bgo_mode[i]].c_str()); s << os;
        }
	for (i = 0; i <= 3; i++) {
	  std::sprintf(os, "TTC encoder BGO%d busy gate             = %s\n", i, ttc_bgo_busy_gate[i] ? "ENABLED" : "DISABLED"); s << os;
        }
        for (i = 0; i < AltiModule::TTC_FIFO_NUMBER; i++) {
            //std::sprintf(os, "TTC encoder \"%-9s\" retransmit   = %s\n", AltiModule::TTC_FIFO_NAME[(AltiModule::TTC_FIFO) i].c_str(), ttc_fifo_retransmit[i] ? "ENABLED" : "DISABLED"); s << os;
            std::string fifoName = AltiModule::TTC_FIFO_NAME[(AltiModule::TTC_FIFO) i];
            fifoName.replace(fifoName.find("_"), 1, " ");
            std::sprintf(os, "TTC encoder %-9s retransmit     = %s\n", fifoName.c_str(), ttc_fifo_retransmit[i] ? "ENABLED" : "DISABLED"); s << os;
        }
        s << "------------------------------------------------------------------------------------------------------" << std::endl;
    }

    if (cnt_config) {
        std::sprintf(os, "CNT BC      counter                  = %s\n", cnt_enable[AltiModule::CNT_BC] ? "ENABLED" : "DISABLED"); s << os;
        std::sprintf(os, "CNT ORB     counter                  = %s\n", cnt_enable[AltiModule::CNT_ORB] ? "ENABLED" : "DISABLED"); s << os;
        std::sprintf(os, "CNT L1A     counter                  = %s\n", cnt_enable[AltiModule::CNT_L1A] ? "ENABLED" : "DISABLED"); s << os;
        std::sprintf(os, "CNT TTR1..3 counter                  = %s,%s,%s\n", cnt_enable[AltiModule::CNT_TTR1] ? "ENABLED" : "DISABLED", cnt_enable[AltiModule::CNT_TTR2] ? "ENABLED" : "DISABLED", cnt_enable[AltiModule::CNT_TTR3] ? "ENABLED" : "DISABLED"); s << os;
        std::sprintf(os, "CNT BGO0..3 counter                  = %s,%s,%s,%s\n", cnt_enable[AltiModule::CNT_BGO0] ? "ENABLED" : "DISABLED", cnt_enable[AltiModule::CNT_BGO1] ? "ENABLED" : "DISABLED", cnt_enable[AltiModule::CNT_BGO2] ? "ENABLED" : "DISABLED", cnt_enable[AltiModule::CNT_BGO3] ? "ENABLED" : "DISABLED"); s << os;
        std::sprintf(os, "CNT trigger type LUT file            = %s\n", cnt_ttyp_lut_file.c_str()); s << os;
        s << "------------------------------------------------------------------------------------------------------" << std::endl;
    }

    if (ctp_config) {
        std::sprintf(os, "CTP ECR generation                    = %s\n", ctp_ecr_generation.print().c_str()); s << os;
    }
}

//------------------------------------------------------------------------------

int AltiConfiguration::read(const std::string &fn) {

    std::ostringstream buf;
    std::string key, key1, val;
    std::vector<std::string> arg;
    unsigned int i, j;
    int num;

    // open input file
    std::ifstream s(fn.c_str());
    if (!s) {
        CERR("cannot open input file \"%s\"", fn.c_str());
        return (AltiModule::FAILURE);
    }

    key = "CLK_CONFIG";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if (val == "CONFIG") clk_config = true;
    else if (val == "NOCONFIG") clk_config = false;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
    key = "CLK_SOURCE";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if (val == AltiModule::CLK_NAME[AltiModule::PRIMARY_FXO]) clk_select = AltiModule::PRIMARY_FXO;
    else if (val == AltiModule::CLK_NAME[AltiModule::SECONDARY_EXT]) clk_select = AltiModule::SECONDARY_EXT;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
    key = "CLK_JC_SOURCE";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if (val == AltiModule::CLK_NAME[AltiModule::PRIMARY_FXO]) clk_jc_select = AltiModule::PRIMARY_FXO;
    else if (val == AltiModule::CLK_NAME[AltiModule::SECONDARY_EXT]) clk_jc_select = AltiModule::SECONDARY_EXT;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
    key = "CLK_JC_FILE"; 
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    clk_jc_file = val;
    key = "CLK_PHASE_SHIFT"; 
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
      CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
      return (AltiModule::FAILURE);
    }
    if (abs(std::stoi(val)) <= 1073741823) clk_phase_shift = std::stoi(val);
    else { 
      CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
      return (AltiModule::FAILURE);
    }
    
    key = "SIG_CONFIG";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if (val == "CONFIG") sig_config = true;
    else if (val == "NOCONFIG") sig_config = false;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
    key = "SIG_BGO_FP";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if ((num = AltiCommon::splitString(val, ",", arg)) != 2) {
        CERR("error in fixed format: for key \"%s\" expected 2 values, got %d value%s", key.c_str(), num, (num > 1) ? "s" : "");
        return (AltiModule::FAILURE);
    }
    for (i = 0; i < ((unsigned int) num); i++) arg[i] = AltiCommon::trimString(arg[i]);
    if (arg[0] == "BGO2") sig_fp_bgo2enable = true;
    else if (arg[0] == "BGO0") sig_fp_bgo2enable = false;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
    if (arg[1] == "BGO3") sig_fp_bgo3enable = true;
    else if (arg[1] == "BGO1") sig_fp_bgo3enable = false;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
    for (i = 0; i < AltiModule::SIGNAL_NUMBER_ROUTED; i++) {
        // needs to be changed, key must be ostringstream?
        switch (i) {
            case 0:
                key = "SIG_SWX_BC";
                break;
            case 1:
                key = "SIG_SWX_ORB";
                break;
            case 2:
                key = "SIG_SWX_L1A";
                break;
            case 3:
                key = "SIG_SWX_TTR1";
                break;
            case 4:
                key = "SIG_SWX_TTR2";
                break;
            case 5:
                key = "SIG_SWX_TTR3";
                break;
            case 6:
                key = "SIG_SWX_TTYP";
                break;
            case 7:
                //key = "SIG_SWX_TTYP1";
                //break;
            case 8:
                //key = "SIG_SWX_TTYP2";
                //break;
            case 9:
                //key = "SIG_SWX_TTYP3";
                //break;
            case 10:
                //key = "SIG_SWX_TTYP4";
                //break;
            case 11:
                //key = "SIG_SWX_TTYP5";
                //break;
            case 12:
                //key = "SIG_SWX_TTYP6";
                //break;
            case 13:
                //key = "SIG_SWX_TTYP7";
                //break;
                continue;
            case 14:
                key = "SIG_SWX_BGO0";
                break;
            case 15:
                key = "SIG_SWX_BGO1";
                break;
            case 16:
                key = "SIG_SWX_BGO2";
                break;
            case 17:
                key = "SIG_SWX_BGO3";
                break;
            default:
                break;
        }
        if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
            CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
            return (AltiModule::FAILURE);
        }
        if ((num = AltiCommon::splitString(val, ",", arg)) != 4) {
            CERR("error in fixed format: for key \"%s\" expected 4 values, got %d value%s", key.c_str(), num, (num > 1) ? "s" : "");
            return (AltiModule::FAILURE);
        }
        for (j = 0; j < 4; j++) {
            arg[j] = AltiCommon::trimString(arg[j]);
            if (key == "SIG_SWX_TTYP") {
                if (arg[j] == AltiModule::SIGNAL_SOURCE_NAME[AltiModule::CTP_IN]) sig_swx_ttyp[j] = sig_swx[i][j] = sig_swx[i + 1][j] =  sig_swx[i + 2][j] = sig_swx[i + 3][j] = sig_swx[i + 4][j] = sig_swx[i + 5][j] = sig_swx[i + 6][j] = sig_swx[i + 7][j] = AltiModule::CTP_IN;
                else if (arg[j] == AltiModule::SIGNAL_SOURCE_NAME[AltiModule::ALTI_IN]) sig_swx_ttyp[j] = sig_swx[i][j] = sig_swx[i + 1][j] =  sig_swx[i + 2][j] = sig_swx[i + 3][j] = sig_swx[i + 4][j] = sig_swx[i + 5][j] = sig_swx[i + 6][j] = sig_swx[i + 7][j] = AltiModule::ALTI_IN;
                else if (arg[j] == AltiModule::SIGNAL_SOURCE_NAME[AltiModule::NIM_IN]) sig_swx_ttyp[j] = sig_swx[i][j] = sig_swx[i + 1][j] =  sig_swx[i + 2][j] = sig_swx[i + 3][j] = sig_swx[i + 4][j] = sig_swx[i + 5][j] = sig_swx[i + 6][j] = sig_swx[i + 7][j] = AltiModule::NIM_IN;
                else if (arg[j] == AltiModule::SIGNAL_SOURCE_NAME[AltiModule::FROM_FPGA]) sig_swx_ttyp[j] = sig_swx[i][j] = sig_swx[i + 1][j] =  sig_swx[i + 2][j] = sig_swx[i + 3][j] = sig_swx[i + 4][j] = sig_swx[i + 5][j] = sig_swx[i + 6][j] = sig_swx[i + 7][j] = AltiModule::FROM_FPGA;
                else {
                    CERR("value \"%s\" for the key \"%s\":%s not allowed", arg[j].c_str(), key.c_str(), AltiModule::SIGNAL_DESTINATION_NAME[j].c_str());
                   return (AltiModule::FAILURE);
                }
            }
            else {
                if (arg[j] == AltiModule::SIGNAL_SOURCE_NAME[AltiModule::CTP_IN]) sig_swx[i][j] = AltiModule::CTP_IN;
                else if (arg[j] == AltiModule::SIGNAL_SOURCE_NAME[AltiModule::ALTI_IN]) sig_swx[i][j] = AltiModule::ALTI_IN;
                else if (arg[j] == AltiModule::SIGNAL_SOURCE_NAME[AltiModule::NIM_IN]) sig_swx[i][j] = AltiModule::NIM_IN;
                else if (arg[j] == AltiModule::SIGNAL_SOURCE_NAME[AltiModule::FROM_FPGA]) sig_swx[i][j] = AltiModule::FROM_FPGA;
                else {
                    CERR("value \"%s\" for the key \"%s\":%s not allowed", arg[j].c_str(), key.c_str(), AltiModule::SIGNAL_DESTINATION_NAME[j].c_str());
                   return (AltiModule::FAILURE);
                }
            }
        }
    }
    
    key = "SIG_EQZ_CTP_IN";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
      CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
      return (AltiModule::FAILURE);
    }
    if ((num = AltiCommon::splitString(val, ",", arg)) != 2) {
      CERR("error in fixed format: for key \"%s\" expected 2 values, got %d value%s", key.c_str(), num, (num > 1) ? "s" : "");
      return (AltiModule::FAILURE);
    }
    arg[0] =  AltiCommon::trimString(arg[0]);
    arg[1] =  AltiCommon::trimString(arg[1]);
    if (arg[0] == "ENABLED") sig_eqz_ctp_in_enable = true;
    else if (arg[0] == "DISABLED") sig_eqz_ctp_in_enable = false;
    else {
      CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
      return (AltiModule::FAILURE);
    }
    if (arg[1] == AltiModule::EQUALIZER_CONFIG_NAME[AltiModule::SHORT_CABLE]) sig_eqz_ctp_in_mode = AltiModule::SHORT_CABLE;
    else if (arg[1] == AltiModule::EQUALIZER_CONFIG_NAME[AltiModule::LONG_CABLE]) sig_eqz_ctp_in_mode = AltiModule::LONG_CABLE;
    else {
      CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
      return (AltiModule::FAILURE);
    }
    
    key = "SIG_EQZ_ALTI_IN";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
      CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
      return (AltiModule::FAILURE);
    }
    if ((num = AltiCommon::splitString(val, ",", arg)) != 2) {
      CERR("error in fixed format: for key \"%s\" expected 2 values, got %d value%s", key.c_str(), num, (num > 1) ? "s" : "");
      return (AltiModule::FAILURE);
    }
    arg[0] =  AltiCommon::trimString(arg[0]);
    arg[1] =  AltiCommon::trimString(arg[1]);
    if (arg[0] == "ENABLED") sig_eqz_alti_in_enable = true;
    else if (arg[0] == "DISABLED") sig_eqz_alti_in_enable = false;
    else {
      CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
      return (AltiModule::FAILURE);
    }
    if (arg[1] == AltiModule::EQUALIZER_CONFIG_NAME[AltiModule::SHORT_CABLE]) sig_eqz_alti_in_mode = AltiModule::SHORT_CABLE;
    else if (arg[1] == AltiModule::EQUALIZER_CONFIG_NAME[AltiModule::LONG_CABLE]) sig_eqz_alti_in_mode = AltiModule::LONG_CABLE;
    else {
      CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
      return (AltiModule::FAILURE);
    }
    
    AltiModule::ASYNC_INPUT_SIGNAL asyncIn;
    unsigned int num_signals;
    for (i = 0; i < AltiModule::ASYNC_INPUT_SIGNAL_NUMBER; i++) {
        asyncIn = (AltiModule::ASYNC_INPUT_SIGNAL) i;
        // LEMO sync inputs: removed!
        if ((asyncIn >= AltiModule::LEMO_ORB) && (asyncIn <= AltiModule::LEMO_BGO3)) continue;
        // merging
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO1) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO3)) continue;
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP1) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP3)) continue;
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP5) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP7)) continue;
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTR2) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTR3)) continue;
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::LEMO_TTR2) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::LEMO_TTR3)) continue;
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::LEMO_BGO3) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::LEMO_BGO3)) continue;
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::RJ45_CALREQ1) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::RJ45_CALREQ2)) continue;
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::CTP_CALREQ1) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::CTP_CALREQ2)) continue;
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::ALTI_CALREQ1) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::ALTI_CALREQ2)) continue;

        if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO0) {
            key = "SIG_IO_SYNC_SWX_BGO";
            key1 = "SIG_IO_SHAPE_SWX_BGO";
            num_signals = 4;
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP0) {
            key = "SIG_IO_SYNC_SWX_TTYP0..3";
            key1 = "SIG_IO_SHAPE_SWX_TTYP0..3";
            num_signals = 4;
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP4) {
            key = "SIG_IO_SYNC_SWX_TTYP4..7";
            key1 = "SIG_IO_SHAPE_SWX_TTYP4..7";
            num_signals = 4;
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTR1) {
            key = "SIG_IO_SYNC_SWX_TTR";
            key1 = "SIG_IO_SHAPE_SWX_TTR";
            num_signals = 3;
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::LEMO_TTR1) {
            key = "SIG_IO_SYNC_LEMO_TTR";
            key1 = "SIG_IO_SHAPE_LEMO_TTR";
            num_signals = 3;
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::LEMO_BGO2) {
            key = "SIG_IO_SYNC_LEMO_BGO";
            key1 = "SIG_IO_SHAPE_LEMO_BGO";
            num_signals = 2;
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::RJ45_CALREQ0) {
            key = "SIG_IO_SYNC_RJ45_CALREQ";
            key1 = "SIG_IO_SHAPE_RJ45_CALREQ";
            num_signals = 3;
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::CTP_CALREQ0) {
            key = "SIG_IO_SYNC_CTP_CALREQ";
            key1 = "SIG_IO_SHAPE_CTP_CALREQ";
            num_signals = 3;
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::ALTI_CALREQ0) {
            key = "SIG_IO_SYNC_ALTI_CALREQ";
            key1 = "SIG_IO_SHAPE_ALTI_CALREQ";
            num_signals = 3;
        }
        else {
            key = std::string("SIG_IO_SYNC_") + AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str();
            key1 = std::string("SIG_IO_SHAPE_") + AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str();
            num_signals = 1;
        }

        if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
            CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
            return (AltiModule::FAILURE);
        }
        if ((num = AltiCommon::splitString(val, ",", arg)) != ((int) (2*num_signals))) {
            CERR("error in fixed format: for key \"%s\" expected %d values, got %d value%s", key.c_str(), 2*num_signals, num, (num > 1) ? "s" : "");
            return (AltiModule::FAILURE);
        }
        for (j = 0; j < num_signals; j++) {
            arg[2*j] = AltiCommon::trimString(arg[2*j]);
            arg[2*j + 1] = AltiCommon::trimString(arg[2*j + 1]);
            if (arg[2*j] == "ENABLED") sig_io_sync_enable[i + j] = true;
            else if (arg[2*j] == "DISABLED") sig_io_sync_enable[i + j] = false;
            else {
                CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
                return (AltiModule::FAILURE);
            }
            if (arg[2*j + 1] == "000DEG") sig_io_sync_phase[i + j] = 1;
            else if (arg[2*j + 1] == "090DEG") sig_io_sync_phase[i + j] = 2;
            else if (arg[2*j + 1] == "180DEG") sig_io_sync_phase[i + j] = 3;
            else if (arg[2*j + 1] == "270DEG") sig_io_sync_phase[i + j] = 4;
            else {
                CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
                return (AltiModule::FAILURE);
            }
        }

        if (readNextRecord(s, key1, val) != AltiModule::SUCCESS) {
            CERR("error reading key \"%s\" from file \"%s\"", key1.c_str(), fn.c_str());
            return (AltiModule::FAILURE);
        }
        if ((num = AltiCommon::splitString(val, ",", arg)) != ((int) num_signals)) {
            CERR("error in fixed format: for key \"%s\" expected %d values, got %d value%s", key1.c_str(), num_signals, num, (num > 1) ? "s" : "");
            return (AltiModule::FAILURE);
        }
        for (j = 0; j < num_signals; j++) {
            arg[j] = AltiCommon::trimString(arg[j]);
            if (arg[j] == "ENABLED") sig_io_shape_enable[i + j] = true;
            else if (arg[j] == "DISABLED") sig_io_shape_enable[i + j] = false;
            else {
                CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key1.c_str());
                return (AltiModule::FAILURE);
            }
        }
    }

    key = "BSY_CONFIG";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if (val == "CONFIG") bsy_config = true;
    else if (val == "NOCONFIG") bsy_config = false;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
    for (i = 0; i < AltiModule::BUSY_INPUT_NUMBER; i++) {
        switch (i) {
            case 0:
                key = "BSY_FROM_FP";
                break;
            case 1:
                key = "BSY_FROM_CTP";
                break;
            case 2:
                key = "BSY_FROM_ALTI";
                break;
            case 3:
                key = "BSY_FROM_PG";
                break;
            case 4:
                key = "BSY_FROM_VME";
                break;
            case 5:
		key = "BSY_FROM_ECR";
		break;
            default:
                break;
        }
        if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
            CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
            return (AltiModule::FAILURE);
        }
        if (val == "ENABLED") bsy_in_select[i] = true;
        else if (val == "DISABLED") bsy_in_select[i] = false;
        else {
            CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
            return (AltiModule::FAILURE);
        }
    }
    for (i = 0; i < AltiModule::BUSY_OUTPUT_NUMBER; i++) {
        switch (i) {
            case 0:
                key = "BSY_TO_CTP";
                break;
            case 1:
                key = "BSY_TO_ALTI";
                break;
            default:
                break;
        }
        if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
            CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
            return (AltiModule::FAILURE);
        }
        if (val == "INACTIVE") bsy_out_select[i] = AltiModule::BUSY_INACTIVE;
        else if (val == "LOCAL") bsy_out_select[i] = AltiModule::BUSY_LOCAL;
        else if (val == "CTP") bsy_out_select[i] = AltiModule::BUSY_CTP;
        else if (val == "ALTI") bsy_out_select[i] = AltiModule::BUSY_ALTI;
        else {
            CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
            return (AltiModule::FAILURE);
        }
    }
    key = "BSY_DATA_VME";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if (val == "0") bsy_data_vme = false;
    else if (val == "1") bsy_data_vme = true;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
    key = "BSY_MASKING_L1A";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if (val == "ENABLED") bsy_masking_l1a_enable = true;
    else if (val == "DISABLED") bsy_masking_l1a_enable = false;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
    key = "BSY_LEVEL";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
      CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
      return (AltiModule::FAILURE);
    }
    if (val == "NIM") bsy_level = AltiModule::NIM;
    else if (val == "TTL") bsy_level = AltiModule::TTL;
    else {
      CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }

    key = "CRQ_CONFIG";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if (val == "CONFIG") crq_config = true;
    else if (val == "NOCONFIG") crq_config = false;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
    key = "CRQ_SOURCE";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if ((num = AltiCommon::splitString(val, ",", arg)) != 3) {
        CERR("error in fixed format: for key \"%s\" expected %d values, got %d value%s", key.c_str(), 3, num, (num > 1) ? "s" : "");
        return (AltiModule::FAILURE);
    }
    for (i = 0; i < 3; i++) {
        arg[i] = AltiCommon::trimString(arg[i]);
        if (arg[i] == AltiModule::CALREQ_INPUT_NAME[AltiModule::CALREQ_FROM_RJ45]) crq_in_select[2 - i] = AltiModule::CALREQ_FROM_RJ45;
        else if (arg[i] == AltiModule::CALREQ_INPUT_NAME[AltiModule::CALREQ_FROM_FRONT_PANEL]) crq_in_select[2 - i] = AltiModule::CALREQ_FROM_FRONT_PANEL;
        else if (arg[i] == AltiModule::CALREQ_INPUT_NAME[AltiModule::CALREQ_FROM_CTP]) crq_in_select[2 - i] = AltiModule::CALREQ_FROM_CTP;
        else if (arg[i] == AltiModule::CALREQ_INPUT_NAME[AltiModule::CALREQ_FROM_ALTI]) crq_in_select[2 - i] = AltiModule::CALREQ_FROM_ALTI;
        else if (arg[i] == AltiModule::CALREQ_INPUT_NAME[AltiModule::CALREQ_FROM_PG]) crq_in_select[2 - i] = AltiModule::CALREQ_FROM_PG;
        else if (arg[i] == AltiModule::CALREQ_INPUT_NAME[AltiModule::CALREQ_FROM_VME]) crq_in_select[2 - i] = AltiModule::CALREQ_FROM_VME;
        else {
            CERR("value \"%s\" for the key \"%s\" not allowed", arg[i].c_str(), key.c_str());
            return (AltiModule::FAILURE);
        }
    }
    for (i = 0; i < AltiModule::CALREQ_OUTPUT_NUMBER; i++) {
        switch (i) {
            case 0:
                key = "CRQ_TO_CTP";
                break;
            case 1:
                key = "CRQ_TO_ALTI";
                break;
            default:
                break;
        }
        if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
            CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
            return (AltiModule::FAILURE);
        }
        if ((num = AltiCommon::splitString(val, ",", arg)) != 3) {
            CERR("error in fixed format: for key \"%s\" expected %d values, got %d value%s", key.c_str(), 3, num, (num > 1) ? "s" : "");
            return (AltiModule::FAILURE);
        }
        for (j = 0; j < 3; j++) {
            arg[j] = AltiCommon::trimString(arg[j]);
            if (arg[j] == AltiModule::CALREQ_SOURCE_NAME[AltiModule::CALREQ_INACTIVE]) crq_out_select[i][2 - j] = AltiModule::CALREQ_INACTIVE;
            else if (arg[j] == AltiModule::CALREQ_SOURCE_NAME[AltiModule::CALREQ_LOCAL]) crq_out_select[i][2 - j] = AltiModule::CALREQ_LOCAL;
            else if (arg[j] == AltiModule::CALREQ_SOURCE_NAME[AltiModule::CALREQ_CTP]) crq_out_select[i][2 - j] = AltiModule::CALREQ_CTP;
            else if (arg[j] == AltiModule::CALREQ_SOURCE_NAME[AltiModule::CALREQ_ALTI]) crq_out_select[i][2 - j] = AltiModule::CALREQ_ALTI;
            else {
                CERR("value \"%s\" for the key \"%s\" not allowed", arg[j].c_str(), key.c_str());
                return (AltiModule::FAILURE);
            }
        }
    }
    key = "CRQ_DATA_VME";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if ((val[0] >= '0') && (val[0] <= '7')) crq_data_vme = (val[0] - '0');
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }

    key = "MEM_CONFIG";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if (val == "CONFIG") mem_config = true;
    else if (val == "NOCONFIG") mem_config = false;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
   
    key = "MEM_PG_START";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    mem_pg_start_addr = std::stoi(val, nullptr, 0);
    key = "MEM_PG_STOP";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    mem_pg_stop_addr = std::stoi(val, nullptr, 0);
    key = "MEM_PG_FILE";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    mem_pg_file = val;
    key = "MEM_PG_ENABLE";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if (val == "ENABLED") mem_pg_enable = true;
    else if (val == "DISABLED") mem_pg_enable = false;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
    AltiModule::SIGNAL_PG signal_pg;
    for (i = 0; i < AltiModule::SIGNAL_PG_NUMBER; i++) {
        signal_pg = (AltiModule::SIGNAL_PG) i;
        // merging
        if ((signal_pg >= AltiModule::PG_TTR2) && (signal_pg <= AltiModule::PG_TTR3)) continue;
        if ((signal_pg >= AltiModule::PG_BGO1) && (signal_pg <= AltiModule::PG_BGO3)) continue;

        if (signal_pg == AltiModule::PG_TTR1) {
            key = "MEM_PG_ENABLE_TTR";
            num_signals = 3;
        }
        else if (signal_pg == AltiModule::PG_BGO0) {
            key = "MEM_PG_ENABLE_BGO";
            num_signals = 4;
        }
        else {
            key = "MEM_PG_ENABLE_" + AltiModule::SIGNAL_PG_NAME[signal_pg];
            num_signals = 1;
        }

        if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
            CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
            return (AltiModule::FAILURE);
        }
        if ((num = AltiCommon::splitString(val, ",", arg)) != ((int) num_signals)) {
            CERR("error in fixed format: for key \"%s\" expected %d values, got %d value%s", key.c_str(), num_signals, num, (num > 1) ? "s" : "");
            return (AltiModule::FAILURE);
        }
        for (j = 0; j < num_signals; j++) {
            arg[j] = AltiCommon::trimString(arg[j]);
            if (arg[j] == "ENABLED") mem_pg_config[i + j] = true;
            else if (arg[j] == "DISABLED") mem_pg_config[i + j] = false;
            else {
                CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
                return (AltiModule::FAILURE);
            }
        }
    }
    key = "MEM_PG_REPEAT";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if (val == "REPEATED") mem_pg_repeat = true;
    else if (val == "ONESHOT") mem_pg_repeat = false;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
    AltiModule::SIGNAL signal;
    for (i = 1; i < AltiModule::SIGNAL_NUMBER; i++) {
        signal = (AltiModule::SIGNAL) i;
        // merging
        if ((signal >= AltiModule::TTR2) && (signal <= AltiModule::TTR3)) continue;
        if ((signal >= AltiModule::TTYP1) && (signal <= AltiModule::TTYP3)) continue;
        if ((signal >= AltiModule::TTYP5) && (signal <= AltiModule::TTYP7)) continue;
        if ((signal >= AltiModule::BGO1) && (signal <= AltiModule::BGO3)) continue;
        if ((signal >= AltiModule::CALREQ1) && (signal <= AltiModule::CALREQ2)) continue;

        if (signal == AltiModule::TTR1) {
            key = "MEM_SNAP_MASK_TTR";
            num_signals = 3;
        }
        else if (signal == AltiModule::TTYP0) {
            key = "MEM_SNAP_MASK_TTYP0..3";
            num_signals = 4;
        }
        else if (signal == AltiModule::TTYP4) {
            key = "MEM_SNAP_MASK_TTYP4..7";
            num_signals = 4;
        }
        else if (signal == AltiModule::BGO0) {
            key = "MEM_SNAP_MASK_BGO";
            num_signals = 4;
        }
        else if (signal == AltiModule::CALREQ0) {
            key = "MEM_SNAP_MASK_CALREQ";
            num_signals = 3;
        }
        else {
            key = "MEM_SNAP_MASK_" + AltiModule::SIGNAL_NAME[signal];
            num_signals = 1;
        }

        if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
            CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
            return (AltiModule::FAILURE);
        }
        if ((num = AltiCommon::splitString(val, ",", arg)) != ((int) num_signals)) {
            CERR("error in fixed format: for key \"%s\" expected %d values, got %d value%s", key.c_str(), num_signals, num, (num > 1) ? "s" : "");
            return (AltiModule::FAILURE);
        }
        for (j = 0; j < num_signals; j++) {
            arg[j] = AltiCommon::trimString(arg[j]);
            if (arg[j] == "MASKED") mem_snap_masked[i + j] = true;
            else if (arg[j] == "NOT_MASKED") mem_snap_masked[i + j] = false;
            else {
                CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
                return (AltiModule::FAILURE);
            }
        }
    }

    key = "TTC_CONFIG";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if (val == "CONFIG") ttc_config = true;
    else if (val == "NOCONFIG") ttc_config = false;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
    key = "TTC_TX_ENABLE";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if ((num = AltiCommon::splitString(val, ",", arg)) != AltiModule::TRANSMITTER_NUMBER) {
        CERR("error in fixed format: for key \"%s\" expected %d values, got %d value%s", key.c_str(), AltiModule::TRANSMITTER_NUMBER, num, (num > 1) ? "s" : "");
        return (AltiModule::FAILURE);
    }
    for (i = 0; i < ((unsigned int) num); i++) {
        arg[i] = AltiCommon::trimString(arg[i]);
        if (arg[i] == "ENABLED") ttc_tx_enable[i] = true;
        else if (arg[i] == "DISABLED") ttc_tx_enable[i] = false;
        else {
            CERR("value \"%s\" for the key \"%s\" not allowed", arg[i].c_str(), key.c_str());
            return (AltiModule::FAILURE);
        }
    }
    key = "TTC_L1A_SOURCE_FROM_L1A"; 
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
      CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
      return (AltiModule::FAILURE);
    }
    if (val == "ENABLED") ttc_l1a_source[0] = true;
    else if (val == "DISABLED") ttc_l1a_source[0] = false;
    else {
      CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
      return (AltiModule::FAILURE);
    }
    key = "TTC_L1A_SOURCE_FROM_TTR1..3";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
      CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
      return (AltiModule::FAILURE);
    }
    if ((num = AltiCommon::splitString(val, ",", arg)) != 3) {
      CERR("error in fixed format: for key \"%s\" expected %d values, got %d value%s", key.c_str(), 3, num, (num > 1) ? "s" : "");
      return (AltiModule::FAILURE);
    }
    for (i = 0; i < ((unsigned int) num); i++) {
      arg[i] = AltiCommon::trimString(arg[i]);
      if (arg[i] == "ENABLED") ttc_l1a_source[i+1] = true;
      else if (arg[i] == "DISABLED") ttc_l1a_source[i+1] = false;
      else {
	CERR("value \"%s\" for the key \"%s\" not allowed", arg[i].c_str(), key.c_str());
	return (AltiModule::FAILURE);
      }
    }
/*
    key = "TTC_TX_DELAY";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if ((num = AltiCommon::splitString(val, ",", arg)) != AltiModule::TRANSMITTER_NUMBER) {
        CERR("error in fixed format: for key \"%s\" expected %d values, got %d value%s", key.c_str(), AltiModule::TRANSMITTER_NUMBER, num, (num > 1) ? "s" : "");
        return (AltiModule::FAILURE);
    }
    for (i = 0; i < ((unsigned int) num); i++) {
        arg[i] = AltiCommon::trimString(arg[i]);
        int delay = std::stoi(arg[i], nullptr, 0);
        if (delay < 0x00 || delay > 0x1f) {
            CERR("transmitter delay must be in range 0x00 to 0x1f (5 bits)", "");
            return (AltiModule::FAILURE);
        }
        ttc_tx_delay[i] = delay;
    }
*/
    for (i = 0; i <= 3; i++) {
        switch (i) {
            case 0:
                key = "TTC_BGO0_INHIBIT";
                break;
            case 1:
                key = "TTC_BGO1_INHIBIT";
                break;
            case 2:
                key = "TTC_BGO2_INHIBIT";
                break;
            case 3:
                key = "TTC_BGO3_INHIBIT";
                break;
            default:
                break;
        }
        if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
            CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
            return (AltiModule::FAILURE);
        }
        if ((num = AltiCommon::splitString(val, ",", arg)) != 2) {
            CERR("error in fixed format: for key \"%s\" expected 2 values, got %d value%s", key.c_str(), num, (num > 1) ? "s" : "");
            return (AltiModule::FAILURE);
        }
        for (j = 0; j < ((unsigned int) num); j++) arg[j] = AltiCommon::trimString(arg[j]);

        int width = std::stoi(arg[0]);
        //std::printf("width: %d\n", width);
        if ((width < 0) || (width > ((int) ALTI_TTC_TRIGGERTYPEWORDPARAMS_BITSTRING::MASK_DELAYINBCS.number()[0]))) {
            CERR("Inhibit width must be in range 0 to %d (12 bits)", ALTI_BGO_INHIBITGENERATORPARAMS_BITSTRING::MASK_DELAYINBCS.number()[0]);
            return (AltiModule::FAILURE);
        }
        ttc_bgo_inhibit_width[i] = width;

        int delay = std::stoi(arg[1]);
        //std::printf("delay: %d\n", delay);
        if ((delay < 0) || (delay > ((int) ALTI_TTC_TRIGGERTYPEWORDPARAMS_BITSTRING::MASK_DELAYINBCS.number()[0]))) {
            CERR("Inhibit delay must be in range 0 to %d (12 bits)", ALTI_BGO_INHIBITGENERATORPARAMS_BITSTRING::MASK_DELAYINBCS.number()[0]);
            return (AltiModule::FAILURE);
        }
        ttc_bgo_inhibit_delay[i] = delay;
    }
  
    key = "TTC_TTYP_DELAY";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    int ttyp_delay = std::stoi(val);
    //std::printf("ttyp_delay: %d\n", ttyp_delay);
    if ((ttyp_delay < 0) || (ttyp_delay > ((int) ALTI_TTC_TRIGGERTYPEWORDPARAMS_BITSTRING::MASK_DELAYINBCS.number()[0]))) {
        CERR("Trigger type word delay must be in range 0 to %d (12 bits)", ALTI_TTC_TRIGGERTYPEWORDPARAMS_BITSTRING::MASK_DELAYINBCS.number()[0]);
        return (AltiModule::FAILURE);
    }
    ttc_ttyp_delay = ttyp_delay;
    key = "TTC_TTYP_ADDR";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    int ttyp_addr = std::stoi(val, nullptr, 0);
    //std::printf("ttyp_addr: 0x%04x\n", ttyp_addr);
    if (ttyp_addr < 0x0000 || ttyp_addr > 0x3fff) {
        CERR("Trigger type address must be in range 0x0000 to 0x3fff (14 bits)", "");
        return (AltiModule::FAILURE);
    }
    ttc_ttyp_addr = ttyp_addr;
    key = "TTC_TTYP_SUBADDR";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    int ttyp_subaddr = std::stoi(val, nullptr, 0);
    //std::printf("ttyp_subaddr: 0x%02x\n", ttyp_subaddr);
    if (ttyp_subaddr < 0x00 || ttyp_subaddr > 0xff) {
        CERR("Trigger type subadress must be in range 0x00 to 0xff (8 bits)", "");
        return (AltiModule::FAILURE);
    }
    ttc_ttyp_subaddr = ttyp_subaddr;
    key = "TTC_TTYP_FRAME";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if (val == AltiModule::TTC_FRAME_TYPE_NAME[AltiModule::TTC_FRAME_TYPE::SHORT].c_str()) ttc_ttyp_frame_type = AltiModule::TTC_FRAME_TYPE::SHORT;
    else if (val == AltiModule::TTC_FRAME_TYPE_NAME[AltiModule::TTC_FRAME_TYPE::LONG].c_str()) ttc_ttyp_frame_type = AltiModule::TTC_FRAME_TYPE::LONG;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
    key = "TTC_TTYP_SPACE";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if (val == AltiModule::TTC_ADDRESS_SPACE_NAME[AltiModule::TTC_ADDRESS_SPACE::INTERNAL].c_str()) ttc_ttyp_address_space = AltiModule::TTC_ADDRESS_SPACE::INTERNAL;
    else if (val == AltiModule::TTC_ADDRESS_SPACE_NAME[AltiModule::TTC_ADDRESS_SPACE::EXTERNAL].c_str()) ttc_ttyp_address_space = AltiModule::TTC_ADDRESS_SPACE::EXTERNAL;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
    for (i = 0; i <= 3; i++) {
        switch (i) {
            case 0:
                key = "TTC_BGO0_MODE";
                break;
            case 1:
                key = "TTC_BGO1_MODE";
                break;
            case 2:
                key = "TTC_BGO2_MODE";
                break;
            case 3:
                key = "TTC_BGO3_MODE";
                break;
            default:
                break;
        }
        if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
            CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
            return (AltiModule::FAILURE);
        }
        if (val == AltiModule::TTC_BGO_MODE_NAME[AltiModule::TTC_BGO_MODE::SYNCHRONOUS_SINGLE_BGO_SIGNAL]) ttc_bgo_mode[i] = AltiModule::TTC_BGO_MODE::SYNCHRONOUS_SINGLE_BGO_SIGNAL;
        else if (val == AltiModule::TTC_BGO_MODE_NAME[AltiModule::TTC_BGO_MODE::SYNCHRONOUS_REPETITIVE]) ttc_bgo_mode[i] = AltiModule::TTC_BGO_MODE::SYNCHRONOUS_REPETITIVE;
        else if (val == AltiModule::TTC_BGO_MODE_NAME[AltiModule::TTC_BGO_MODE::ASYNCHRONOUS_BGO_SIGNAL]) ttc_bgo_mode[i] = AltiModule::TTC_BGO_MODE::ASYNCHRONOUS_BGO_SIGNAL;
        else if (val == AltiModule::TTC_BGO_MODE_NAME[AltiModule::TTC_BGO_MODE::ASYNCHRONOUS_VME_ON_TRIGGER]) ttc_bgo_mode[i] = AltiModule::TTC_BGO_MODE::ASYNCHRONOUS_VME_ON_TRIGGER;
        else if (val == AltiModule::TTC_BGO_MODE_NAME[AltiModule::TTC_BGO_MODE::ASYNCHRONOUS_VME_WHEN_NOT_EMPTY]) ttc_bgo_mode[i] = AltiModule::TTC_BGO_MODE::ASYNCHRONOUS_VME_WHEN_NOT_EMPTY;
        else {
            CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
            return (AltiModule::FAILURE);
        }
    }
    for (i = 0; i <= 3; i++) {
      switch (i) {
      case 0:
	key = "TTC_BGO0_BUSY_GATE";
	break;
      case 1:
	key = "TTC_BGO1_BUSY_GATE";
	break;
      case 2:
	key = "TTC_BGO2_BUSY_GATE";
	break;
      case 3:
	key = "TTC_BGO3_BUSY_GATE";
	break;
      default:
	break;
      }
      if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
	CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
	return (AltiModule::FAILURE);
      }
      if (val == "ENABLED") ttc_bgo_busy_gate[i] = true;
      else if (val == "DISABLED") ttc_bgo_busy_gate[i] = false;
      else {
	CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
	return (AltiModule::FAILURE);
      }
    }
    
    for (i = 0; i < AltiModule::TTC_FIFO_NUMBER; i++) {
        key = "TTC_" + AltiModule::TTC_FIFO_NAME[(AltiModule::TTC_FIFO) i] + "_RETRANS";
        if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
            CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
            return (AltiModule::FAILURE);
        }
        if (val == "ENABLED") ttc_fifo_retransmit[i] = true;
        else if (val == "DISABLED") ttc_fifo_retransmit[i] = false;
        else {
            CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
            return (AltiModule::FAILURE);
        }
    }

    key = "CNT_CONFIG";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if (val == "CONFIG") cnt_config = true;
    else if (val == "NOCONFIG") cnt_config = false;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
    AltiModule::CNT_TYPE cnt;
    for (i = 0; i < AltiModule::CNT_NUMBER; i++) {
        cnt = (AltiModule::CNT_TYPE) i;
        // merging
        if ((cnt >= AltiModule::CNT_TTR2) && (cnt <= AltiModule::CNT_TTR3)) continue;
        if ((cnt >= AltiModule::CNT_BGO1) && (cnt <= AltiModule::CNT_BGO3)) continue;
        if (cnt == AltiModule::CNT_TTR1) {
            key = "CNT_TTR";
            num_signals = 3;
        }
        else if (cnt == AltiModule::CNT_BGO0) {
            key = "CNT_BGO";
            num_signals = 4;
        }
        else {
            key = AltiModule::CNT_NAME[i];
            num_signals = 1;
        }

        if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
            CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
            return (AltiModule::FAILURE);
        }
        if ((num = AltiCommon::splitString(val, ",", arg)) != ((int) num_signals)) {
            CERR("error in fixed format: for key \"%s\" expected %d values, got %d value%s", key.c_str(), num_signals, num, (num > 1) ? "s" : "");
            return (AltiModule::FAILURE);
        }
        for (j = 0; j < num_signals; j++) {
            arg[j] = AltiCommon::trimString(arg[j]);
            if (arg[j] == "ENABLED") cnt_enable[i + j] = true;
            else if (arg[j] == "DISABLED") cnt_enable[i + j] = false;
            else {
                CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
                return (AltiModule::FAILURE);
            }
        }
    }
    key = "CNT_TTYP_LUT_FILE";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    cnt_ttyp_lut_file = val;

    key = "CTP_CONFIG";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if (val == "CONFIG") ctp_config = true;
    else if (val == "NOCONFIG") ctp_config = false;
    else {
        CERR("value \"%s\" for the key \"%s\" not allowed", val.c_str(), key.c_str());
        return (AltiModule::FAILURE);
    }
    key = "CTP_ECR_GENERATION";
    if (readNextRecord(s, key, val) != AltiModule::SUCCESS) {
        CERR("error reading key \"%s\" from file \"%s\"", key.c_str(), fn.c_str());
        return (AltiModule::FAILURE);
    }
    if (ctp_ecr_generation.read(val) != AltiModule::SUCCESS) {
        CERR("error in fixed format: for key \"%s\", cannot read ECR generation (\"%s\")", key.c_str(), val.c_str());
        return (AltiModule::FAILURE);
    }

    // close input file
    s.close();

    // print
    COUT("read AltiConfiguration from file \"%s\"", fn.c_str());

    return (AltiModule::SUCCESS);
}

//------------------------------------------------------------------------------

int AltiConfiguration::write(const std::string& fn) const {

    std::ostringstream buf;
    std::string key, key1;
    char os[STRING_LENGTH];
    unsigned int i, j;
  
    std::vector<std::string> arg;
  
    // open output file
    std::ofstream s(fn.c_str());
    if (!s) {
        CERR("cannot open output file \"%s\"", fn.c_str());
        return (AltiModule::FAILURE);
    }

    std::sprintf(os, "###############################################################################################################################\n"); s << os;

    std::sprintf(os, "# CONFIG|NOCONFIG\n"); s << os;
    key = "CLK_CONFIG";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), clk_config ? "CONFIG" : "NOCONFIG"); s << os;
    std::sprintf(os, "# PRIMARY_FXO|SECONDARY_EXT\n"); s << os;
    key = "CLK_SOURCE";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), AltiModule::CLK_NAME[clk_select].c_str()); s << os;
    std::sprintf(os, "# PRIMARY_FXO|SECONDARY_EXT\n"); s << os;
    key = "CLK_JC_SOURCE";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), AltiModule::CLK_NAME[clk_jc_select].c_str()); s << os;
    key = "CLK_JC_FILE";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), clk_jc_file.c_str()); s << os;
    key = "CLK_PHASE_SHIFT";
    std::sprintf(os, "# -1073741823 .. 1073741823 [step: 15 ps]\n"); s << os;
    std::sprintf(os, "%-27s= %d\n", key.c_str(), clk_phase_shift); s << os;
    
    std::sprintf(os, "###############################################################################################################################\n"); s << os;

    std::sprintf(os, "# CONFIG|NOCONFIG\n"); s << os;
    key = "SIG_CONFIG";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), sig_config ? "CONFIG" : "NOCONFIG"); s << os;
    std::sprintf(os, "# BGO0|BGO2, BGO1|BGO3\n"); s << os;
    key = "SIG_BGO_FP";
    std::sprintf(os, "%-27s= %s, %s\n", key.c_str(), sig_fp_bgo2enable ? "BGO2" : "BGO0", sig_fp_bgo3enable ? "BGO3" : "BGO1"); s << os;
    std::sprintf(os, "# <CTP_OUT>, <ALTI_OUT>, <NIM_OUT>, <TO_FPGA>\n"); s << os;
    std::sprintf(os, "# <CTP_OUT>  = CTP_IN|ALTI_IN|NIM_IN|FROM_FPGA\n"); s << os;
    std::sprintf(os, "# <ALTI_OUT> = CTP_IN|ALTI_IN|NIM_IN|FROM_FPGA\n"); s << os;
    std::sprintf(os, "# <NIM_OUT>  = CTP_IN|ALTI_IN|NIM_IN|FROM_FPGA\n"); s << os;
    std::sprintf(os, "# <TO_FPGA>  = CTP_IN|ALTI_IN|NIM_IN|FROM_FPGA\n"); s << os;
    for (i = 0; i < AltiModule::SIGNAL_NUMBER_ROUTED; i++) {
        // needs to be changed, key must be ostringstream?
        switch (i) {
            case 0:
                key = "SIG_SWX_BC";
                break;
            case 1:
                key = "SIG_SWX_ORB";
                break;
            case 2:
                key = "SIG_SWX_L1A";
                break;
            case 3:
                key = "SIG_SWX_TTR1";
                break;
            case 4:
                key = "SIG_SWX_TTR2";
                break;
            case 5:
                key = "SIG_SWX_TTR3";
                break;
            case 6:
                key = "SIG_SWX_TTYP";
                break;
            case 7:
                //key = "SIG_SWX_TTYP1";
                //break;
            case 8:
                //key = "SIG_SWX_TTYP2";
                //break;
            case 9:
                //key = "SIG_SWX_TTYP3";
                //break;
            case 10:
                //key = "SIG_SWX_TTYP4";
                //break;
            case 11:
                //key = "SIG_SWX_TTYP5";
                //break;
            case 12:
                //key = "SIG_SWX_TTYP6";
                //break;
            case 13:
                //key = "SIG_SWX_TTYP7";
                //break;
                continue;
            case 14:
                key = "SIG_SWX_BGO0";
                break;
            case 15:
                key = "SIG_SWX_BGO1";
                break;
            case 16:
                key = "SIG_SWX_BGO2";
                break;
            case 17:
                key = "SIG_SWX_BGO3";
                break;
            default:
                break;
        }
        if (key == "SIG_SWX_TTYP") {
            std::sprintf(os, "%-27s= %s, %s, %s, %s\n", key.c_str(), AltiModule::SIGNAL_SOURCE_NAME[sig_swx_ttyp[0]].c_str(), AltiModule::SIGNAL_SOURCE_NAME[sig_swx_ttyp[1]].c_str(), AltiModule::SIGNAL_SOURCE_NAME[sig_swx_ttyp[2]].c_str(), AltiModule::SIGNAL_SOURCE_NAME[sig_swx_ttyp[3]].c_str()); s << os;
        }
        else {
            std::sprintf(os, "%-27s= %s, %s, %s, %s\n", key.c_str(), AltiModule::SIGNAL_SOURCE_NAME[sig_swx[i][0]].c_str(), AltiModule::SIGNAL_SOURCE_NAME[sig_swx[i][1]].c_str(), AltiModule::SIGNAL_SOURCE_NAME[sig_swx[i][2]].c_str(), AltiModule::SIGNAL_SOURCE_NAME[sig_swx[i][3]].c_str()); s << os;
        }
    }
   
    std::sprintf(os, "# SHORT_CABLE|LONG_CABLE\n"); s << os;
    key = "SIG_EQZ_CTP_IN";
    std::sprintf(os, "%-27s= %s, %s\n", key.c_str(), sig_eqz_ctp_in_enable ? "ENABLED":"DISABLED", AltiModule::EQUALIZER_CONFIG_NAME[sig_eqz_ctp_in_mode].c_str()); s << os;
    key = "SIG_EQZ_ALTI_IN";
    std::sprintf(os, "%-27s= %s, %s\n", key.c_str(), sig_eqz_alti_in_enable ? "ENABLED":"DISABLED", AltiModule::EQUALIZER_CONFIG_NAME[sig_eqz_alti_in_mode].c_str()); s << os;
    
    std::sprintf(os, "# SYNC:  ENABLED|DISABLED,000DEG|090DEG|180DEG|270DEG\n"); s << os;
    std::sprintf(os, "# SHAPE: ENABLED|DISABLED\n"); s << os;
    AltiModule::ASYNC_INPUT_SIGNAL asyncIn;
    bool sync_enable[4];
    unsigned int phase[4];
    bool shape_enable[4];
    for (i = 0; i < AltiModule::ASYNC_INPUT_SIGNAL_NUMBER; i++) {
        asyncIn = (AltiModule::ASYNC_INPUT_SIGNAL) i;
        // LEMO sync inputs: removed!
        if ((asyncIn >= AltiModule::LEMO_ORB) && (asyncIn <= AltiModule::LEMO_BGO3)) continue;
        // merging
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO1) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO3)) continue;
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP1) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP3)) continue;
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP5) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP7)) continue;
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTR2) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTR3)) continue;
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::LEMO_TTR2) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::LEMO_TTR3)) continue;
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::LEMO_BGO3) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::LEMO_BGO3)) continue;
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::RJ45_CALREQ1) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::RJ45_CALREQ2)) continue;
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::CTP_CALREQ1) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::CTP_CALREQ2)) continue;
        if ((asyncIn >= AltiModule::ASYNC_INPUT_SIGNAL::ALTI_CALREQ1) && (asyncIn <= AltiModule::ASYNC_INPUT_SIGNAL::ALTI_CALREQ2)) continue;

        if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO0) {
            std::sprintf(os, "# BGO0..3\n"); s << os;
            key = "SIG_IO_SYNC_SWX_BGO";
            key1 = "SIG_IO_SHAPE_SWX_BGO";
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP0) {
            std::sprintf(os, "# TTYP0..3\n"); s << os;
            key = "SIG_IO_SYNC_SWX_TTYP0..3";
            key1 = "SIG_IO_SHAPE_SWX_TTYP0..3";
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP4) {
            std::sprintf(os, "# TTYP4..7\n"); s << os;
            key = "SIG_IO_SYNC_SWX_TTYP4..7";
            key1 = "SIG_IO_SHAPE_SWX_TTYP4..7";
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTR1) {
            std::sprintf(os, "# TTR1..3\n"); s << os;
            key = "SIG_IO_SYNC_SWX_TTR";
            key1 = "SIG_IO_SHAPE_SWX_TTR";
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::LEMO_TTR1) {
            std::sprintf(os, "# TTR1..3\n"); s << os;
            key = "SIG_IO_SYNC_LEMO_TTR";
            key1 = "SIG_IO_SHAPE_LEMO_TTR";
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::LEMO_BGO2) {
            std::sprintf(os, "# BGO2..3\n"); s << os;
            key = "SIG_IO_SYNC_LEMO_BGO";
            key1 = "SIG_IO_SHAPE_LEMO_BGO";
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::RJ45_CALREQ0) {
            std::sprintf(os, "# CALREQ0..2\n"); s << os;
            key = "SIG_IO_SYNC_RJ45_CALREQ";
            key1 = "SIG_IO_SHAPE_RJ45_CALREQ";
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::CTP_CALREQ0) {
            std::sprintf(os, "# CALREQ0..2\n"); s << os;
            key = "SIG_IO_SYNC_CTP_CALREQ";
            key1 = "SIG_IO_SHAPE_CTP_CALREQ";
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::ALTI_CALREQ0) {
            std::sprintf(os, "# CALREQ0..2\n"); s << os;
            key = "SIG_IO_SYNC_ALTI_CALREQ";
            key1 = "SIG_IO_SHAPE_ALTI_CALREQ";
        }
        else {
            key = std::string("SIG_IO_SYNC_") + AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str();
            key1 = std::string("SIG_IO_SHAPE_") + AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str();
        }

        if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::SWX_BGO0) {
            for (j = 0; j < 4; j++) {
                sync_enable[j] = sig_io_sync_enable[asyncIn + j];
                phase[j] = sig_io_sync_phase[asyncIn + j];
                shape_enable[j] = sig_io_shape_enable[asyncIn + j];
            }
            std::sprintf(os, "%-27s= %s,%s, %s,%s, %s,%s, %s,%s\n", key.c_str(), bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str(), bool2string(sync_enable[1]).c_str(), phase2string(phase[1]).c_str(), bool2string(sync_enable[2]).c_str(), phase2string(phase[2]).c_str(), bool2string(sync_enable[3]).c_str(), phase2string(phase[3]).c_str()); s << os;
            std::sprintf(os, "%-27s= %s,%s,%s,%s\n", key1.c_str(), bool2string(shape_enable[0]).c_str(), bool2string(shape_enable[1]).c_str(), bool2string(shape_enable[2]).c_str(), bool2string(shape_enable[3]).c_str()); s << os;
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP0) {
            for (j = 0; j < 4; j++) {
                sync_enable[j] = sig_io_sync_enable[asyncIn + j];
                phase[j] = sig_io_sync_phase[asyncIn + j];
                shape_enable[j] = sig_io_shape_enable[asyncIn + j];
            }
            std::sprintf(os, "%-27s= %s,%s, %s,%s, %s,%s, %s,%s\n", key.c_str(), bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str(), bool2string(sync_enable[1]).c_str(), phase2string(phase[1]).c_str(), bool2string(sync_enable[2]).c_str(), phase2string(phase[2]).c_str(), bool2string(sync_enable[3]).c_str(), phase2string(phase[3]).c_str()); s << os;
            std::sprintf(os, "%-27s= %s,%s,%s,%s\n", key1.c_str(), bool2string(shape_enable[0]).c_str(), bool2string(shape_enable[1]).c_str(), bool2string(shape_enable[2]).c_str(), bool2string(shape_enable[3]).c_str()); s << os;
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTYP4) {
            for (j = 0; j < 4; j++) {
                sync_enable[j] = sig_io_sync_enable[asyncIn + j];
                phase[j] = sig_io_sync_phase[asyncIn + j];
                shape_enable[j] = sig_io_shape_enable[asyncIn + j];
            }
            std::sprintf(os, "%-27s= %s,%s, %s,%s, %s,%s, %s,%s\n", key.c_str(), bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str(), bool2string(sync_enable[1]).c_str(), phase2string(phase[1]).c_str(), bool2string(sync_enable[2]).c_str(), phase2string(phase[2]).c_str(), bool2string(sync_enable[3]).c_str(), phase2string(phase[3]).c_str()); s << os;
            std::sprintf(os, "%-27s= %s,%s,%s,%s\n", key1.c_str(), bool2string(shape_enable[0]).c_str(), bool2string(shape_enable[1]).c_str(), bool2string(shape_enable[2]).c_str(), bool2string(shape_enable[3]).c_str()); s << os;
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::SWX_TTR1) {
            for (j = 0; j < 3; j++) {
                sync_enable[j] = sig_io_sync_enable[asyncIn + j];
                phase[j] = sig_io_sync_phase[asyncIn + j];
                shape_enable[j] = sig_io_shape_enable[asyncIn + j];
            }
            std::sprintf(os, "%-27s= %s,%s, %s,%s, %s,%s\n", key.c_str(), bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str(), bool2string(sync_enable[1]).c_str(), phase2string(phase[1]).c_str(), bool2string(sync_enable[2]).c_str(), phase2string(phase[2]).c_str()); s << os;
            std::sprintf(os, "%-27s= %s,%s,%s\n", key1.c_str(), bool2string(shape_enable[0]).c_str(), bool2string(shape_enable[1]).c_str(), bool2string(shape_enable[2]).c_str()); s << os;
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::LEMO_TTR1) {
            for (j = 0; j < 3; j++) {
                sync_enable[j] = sig_io_sync_enable[asyncIn + j];
                phase[j] = sig_io_sync_phase[asyncIn + j];
                shape_enable[j] = sig_io_shape_enable[asyncIn + j];
            }
            std::sprintf(os, "%-27s= %s,%s, %s,%s, %s,%s\n", key.c_str(), bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str(), bool2string(sync_enable[1]).c_str(), phase2string(phase[1]).c_str(), bool2string(sync_enable[2]).c_str(), phase2string(phase[2]).c_str()); s << os;
            std::sprintf(os, "%-27s= %s,%s,%s\n", key1.c_str(), bool2string(shape_enable[0]).c_str(), bool2string(shape_enable[1]).c_str(), bool2string(shape_enable[2]).c_str()); s << os;
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::LEMO_BGO2) {
            for (j = 0; j < 2; j++) {
                sync_enable[j] = sig_io_sync_enable[asyncIn + j];
                phase[j] = sig_io_sync_phase[asyncIn + j];
                shape_enable[j] = sig_io_shape_enable[asyncIn + j];
            }
            std::sprintf(os, "%-27s= %s,%s, %s,%s\n", key.c_str(), bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str(), bool2string(sync_enable[1]).c_str(), phase2string(phase[1]).c_str()); s << os;
            std::sprintf(os, "%-27s= %s,%s\n", key1.c_str(), bool2string(shape_enable[0]).c_str(), bool2string(shape_enable[1]).c_str()); s << os;
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::RJ45_CALREQ0) {
            for (j = 0; j < 3; j++) {
                sync_enable[j] = sig_io_sync_enable[asyncIn + j];
                phase[j] = sig_io_sync_phase[asyncIn + j];
                shape_enable[j] = sig_io_shape_enable[asyncIn + j];
            }
            std::sprintf(os, "%-27s= %s,%s, %s,%s, %s,%s\n", key.c_str(), bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str(), bool2string(sync_enable[1]).c_str(), phase2string(phase[1]).c_str(), bool2string(sync_enable[2]).c_str(), phase2string(phase[2]).c_str()); s << os;
            std::sprintf(os, "%-27s= %s,%s,%s\n", key1.c_str(), bool2string(shape_enable[0]).c_str(), bool2string(shape_enable[1]).c_str(), bool2string(shape_enable[2]).c_str()); s << os;
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::CTP_CALREQ0) {
            for (j = 0; j < 3; j++) {
                sync_enable[j] = sig_io_sync_enable[asyncIn + j];
                phase[j] = sig_io_sync_phase[asyncIn + j];
                shape_enable[j] = sig_io_shape_enable[asyncIn + j];
            }
            std::sprintf(os, "%-27s= %s,%s, %s,%s, %s,%s\n", key.c_str(), bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str(), bool2string(sync_enable[1]).c_str(), phase2string(phase[1]).c_str(), bool2string(sync_enable[2]).c_str(), phase2string(phase[2]).c_str()); s << os;
            std::sprintf(os, "%-27s= %s,%s,%s\n", key1.c_str(), bool2string(shape_enable[0]).c_str(), bool2string(shape_enable[1]).c_str(), bool2string(shape_enable[2]).c_str()); s << os;
        }
        else if (asyncIn == AltiModule::ASYNC_INPUT_SIGNAL::ALTI_CALREQ0) {
            for (j = 0; j < 3; j++) {
                sync_enable[j] = sig_io_sync_enable[asyncIn + j];
                phase[j] = sig_io_sync_phase[asyncIn + j];
                shape_enable[j] = sig_io_shape_enable[asyncIn + j];
            }
            std::sprintf(os, "%-27s= %s,%s, %s,%s, %s,%s\n", key.c_str(), bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str(), bool2string(sync_enable[1]).c_str(), phase2string(phase[1]).c_str(), bool2string(sync_enable[2]).c_str(), phase2string(phase[2]).c_str()); s << os;
            std::sprintf(os, "%-27s= %s,%s,%s\n", key1.c_str(), bool2string(shape_enable[0]).c_str(), bool2string(shape_enable[1]).c_str(), bool2string(shape_enable[2]).c_str()); s << os;
        }
        else {
            sync_enable[0] = sig_io_sync_enable[asyncIn];
            phase[0] = sig_io_sync_phase[asyncIn];
            shape_enable[0] = sig_io_shape_enable[asyncIn];
            std::sprintf(os, "%-27s= %s,%s\n", key.c_str(), bool2string(sync_enable[0]).c_str(), phase2string(phase[0]).c_str()); s << os;
            std::sprintf(os, "%-27s= %s\n", key1.c_str(), bool2string(shape_enable[0]).c_str()); s << os;
        }
    }

    std::sprintf(os, "###############################################################################################################################\n"); s << os;

    std::sprintf(os, "# CONFIG|NOCONFIG\n"); s << os;
    key = "BSY_CONFIG";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), bsy_config ? "CONFIG" : "NOCONFIG"); s << os;
    std::sprintf(os, "# ENABLED|DISABLED\n"); s << os;
    key = "BSY_FROM_FP"; 
    std::sprintf(os, "%-27s= %s\n", key.c_str(), bsy_in_select[0] ? "ENABLED" : "DISABLED"); s << os;
    key = "BSY_FROM_CTP"; 
    std::sprintf(os, "%-27s= %s\n", key.c_str(), bsy_in_select[1] ? "ENABLED" : "DISABLED"); s << os;
    key = "BSY_FROM_ALTI"; 
    std::sprintf(os, "%-27s= %s\n", key.c_str(), bsy_in_select[2] ? "ENABLED" : "DISABLED"); s << os;
    key = "BSY_FROM_PG"; 
    std::sprintf(os, "%-27s= %s\n", key.c_str(), bsy_in_select[3] ? "ENABLED" : "DISABLED"); s << os;
    key = "BSY_FROM_VME";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), bsy_in_select[4] ? "ENABLED" : "DISABLED"); s << os;
    key = "BSY_FROM_ECR";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), bsy_in_select[5] ? "ENABLED" : "DISABLED"); s << os;
    std::sprintf(os, "# INACTIVE|LOCAL|CTP|ALTI\n"); s << os;
    key = "BSY_TO_CTP"; 
    /*num = */AltiCommon::splitString(AltiModule::BUSY_SOURCE_NAME[bsy_out_select[0]], "_", arg);
    std::sprintf(os, "%-27s= %s\n", key.c_str(), arg[1].c_str()); s << os;
    key = "BSY_TO_ALTI"; 
    /*num = */AltiCommon::splitString(AltiModule::BUSY_SOURCE_NAME[bsy_out_select[1]], "_", arg);
    std::sprintf(os, "%-27s= %s\n", key.c_str(), arg[1].c_str()); s << os;
    std::sprintf(os, "# 0|1\n"); s << os;
    key = "BSY_DATA_VME"; 
    std::sprintf(os, "%-27s= %s\n", key.c_str(), bsy_data_vme ? "1" : "0"); s << os;
    std::sprintf(os, "# ENABLED|DISABLED\n"); s << os;
    key = "BSY_MASKING_L1A"; 
    std::sprintf(os, "%-27s= %s\n", key.c_str(), bsy_masking_l1a_enable ? "ENABLED" : "DISABLED"); s << os;
    key = "BSY_LEVEL"; 
    std::sprintf(os, "%-27s= %s\n", key.c_str(), AltiModule::BUSY_LEVEL_NAME[bsy_level].c_str()); s << os;

    std::sprintf(os, "###############################################################################################################################\n"); s << os;

    std::sprintf(os, "# CONFIG|NOCONFIG\n"); s << os;
    key = "CRQ_CONFIG";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), crq_config ? "CONFIG" : "NOCONFIG"); s << os;
    std::sprintf(os, "# CALREQ2..0\n"); s << os;
    std::sprintf(os, "# CALREQ_FROM_RJ45|CALREQ_FROM_FRONT_PANEL|CALREQ_FROM_CTP|CALREQ_FROM_ALTI|CALREQ_FROM_PG|CALREQ_FROM_VME\n"); s << os;
    key = "CRQ_SOURCE"; 
    std::sprintf(os, "%-27s= %s,%s,%s\n", key.c_str(), AltiModule::CALREQ_INPUT_NAME[crq_in_select[2]].c_str(), AltiModule::CALREQ_INPUT_NAME[crq_in_select[1]].c_str(), AltiModule::CALREQ_INPUT_NAME[crq_in_select[0]].c_str()); s << os;
    std::sprintf(os, "# CALREQ2..0\n"); s << os;
    std::sprintf(os, "# CALREQ_INACTIVE|CALREQ_LOCAL|CALREQ_CTP|CALREQ_ALTI\n"); s << os;
    key = "CRQ_TO_CTP"; 
    std::sprintf(os, "%-27s= %s,%s,%s\n", key.c_str(), AltiModule::CALREQ_SOURCE_NAME[crq_out_select[0][2]].c_str(), AltiModule::CALREQ_SOURCE_NAME[crq_out_select[0][1]].c_str(), AltiModule::CALREQ_SOURCE_NAME[crq_out_select[0][0]].c_str()); s << os;
    key = "CRQ_TO_ALTI"; 
    std::sprintf(os, "%-27s= %s,%s,%s\n", key.c_str(), AltiModule::CALREQ_SOURCE_NAME[crq_out_select[1][2]].c_str(), AltiModule::CALREQ_SOURCE_NAME[crq_out_select[1][1]].c_str(), AltiModule::CALREQ_SOURCE_NAME[crq_out_select[1][0]].c_str()); s << os;
    std::sprintf(os, "# 0..7\n"); s << os;
    key = "CRQ_DATA_VME"; 
    std::sprintf(os, "%-27s= %d\n", key.c_str(), crq_data_vme); s << os;

    std::sprintf(os, "###############################################################################################################################\n"); s << os;

    std::sprintf(os, "# CONFIG|NOCONFIG\n"); s << os;
    key = "MEM_CONFIG";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), mem_config ? "CONFIG" : "NOCONFIG"); s << os;
    std::sprintf(os, "# 0x00000000..0x000fffff\n"); s << os;
    key = "MEM_PG_START";
    std::sprintf(os, "%-27s= 0x%08x\n", key.c_str(), mem_pg_start_addr); s << os;
    key = "MEM_PG_STOP";
    std::sprintf(os, "%-27s= 0x%08x\n", key.c_str(), mem_pg_stop_addr); s << os;
    std::sprintf(os, "# ALTI/data/pg_alti.dat\n"); s << os;
    key = "MEM_PG_FILE";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), mem_pg_file.c_str()); s << os;
    std::sprintf(os, "# ENABLED|DISABLED\n"); s << os;
    key = "MEM_PG_ENABLE";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), mem_pg_enable ? "ENABLED" : "DISABLED"); s << os;
    key = "MEM_PG_ENABLE_ORB";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), mem_pg_config[AltiModule::PG_ORB] ? "ENABLED" : "DISABLED"); s << os;
    key = "MEM_PG_ENABLE_L1A";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), mem_pg_config[AltiModule::PG_L1A] ? "ENABLED" : "DISABLED"); s << os;
    std::sprintf(os, "# TTR1..3\n"); s << os;
    key = "MEM_PG_ENABLE_TTR";
    std::sprintf(os, "%-27s= %s,%s,%s\n", key.c_str(), mem_pg_config[AltiModule::PG_TTR1] ? "ENABLED" : "DISABLED", mem_pg_config[AltiModule::PG_TTR2] ? "ENABLED" : "DISABLED", mem_pg_config[AltiModule::PG_TTR3] ? "ENABLED" : "DISABLED"); s << os;
    std::sprintf(os, "# TTYP\n"); s << os;
    key = "MEM_PG_ENABLE_TTYP";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), mem_pg_config[AltiModule::PG_TTYP] ? "ENABLED" : "DISABLED"); s << os;
    std::sprintf(os, "# BGO0..3\n"); s << os;
    key = "MEM_PG_ENABLE_BGO";
    std::sprintf(os, "%-27s= %s,%s,%s,%s\n", key.c_str(), mem_pg_config[AltiModule::PG_BGO0] ? "ENABLED" : "DISABLED", mem_pg_config[AltiModule::PG_BGO1] ? "ENABLED" : "DISABLED", mem_pg_config[AltiModule::PG_BGO2] ? "ENABLED" : "DISABLED", mem_pg_config[AltiModule::PG_BGO3] ? "ENABLED" : "DISABLED"); s << os;
    std::sprintf(os, "# ONESHOT|REPEATED\n"); s << os;
    key = "MEM_PG_REPEAT";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), mem_pg_repeat ? "REPEATED" : "ONESHOT"); s << os;
    std::sprintf(os, "# MASKED|NOT_MASKED\n"); s << os;
    key = "MEM_SNAP_MASK_ORB";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), mem_snap_masked[AltiModule::ORB] ? "MASKED" : "NOT_MASKED"); s << os;
    key = "MEM_SNAP_MASK_L1A";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), mem_snap_masked[AltiModule::L1A] ? "MASKED" : "NOT_MASKED"); s << os;
    std::sprintf(os, "# TTR1..3\n"); s << os;
    key = "MEM_SNAP_MASK_TTR";
    std::sprintf(os, "%-27s= %s,%s,%s\n", key.c_str(), mem_snap_masked[AltiModule::TTR1] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::TTR2] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::TTR3] ? "MASKED" : "NOT_MASKED"); s << os;
    std::sprintf(os, "# TTYP0..3\n"); s << os;
    key = "MEM_SNAP_MASK_TTYP0..3";
    std::sprintf(os, "%-27s= %s,%s,%s,%s\n", key.c_str(), mem_snap_masked[AltiModule::TTYP0] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::TTYP1] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::TTYP2] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::TTYP3] ? "MASKED" : "NOT_MASKED"); s << os;
    std::sprintf(os, "# TTYP4..7\n"); s << os;
    key = "MEM_SNAP_MASK_TTYP4..7";
    std::sprintf(os, "%-27s= %s,%s,%s,%s\n", key.c_str(), mem_snap_masked[AltiModule::TTYP4] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::TTYP5] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::TTYP6] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::TTYP7] ? "MASKED" : "NOT_MASKED"); s << os;
    std::sprintf(os, "# BGO0..3\n"); s << os;
    key = "MEM_SNAP_MASK_BGO";
    std::sprintf(os, "%-27s= %s,%s,%s,%s\n", key.c_str(), mem_snap_masked[AltiModule::BGO0] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::BGO1] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::BGO2] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::BGO3] ? "MASKED" : "NOT_MASKED"); s << os;
    key = "MEM_SNAP_MASK_BUSY";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), mem_snap_masked[AltiModule::BUSY] ? "MASKED" : "NOT_MASKED"); s << os;
    std::sprintf(os, "# CALREQ0..2\n"); s << os;
    key = "MEM_SNAP_MASK_CALREQ";
    std::sprintf(os, "%-27s= %s,%s,%s\n", key.c_str(), mem_snap_masked[AltiModule::CALREQ0] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::CALREQ1] ? "MASKED" : "NOT_MASKED", mem_snap_masked[AltiModule::CALREQ2] ? "MASKED" : "NOT_MASKED"); s << os;

    std::sprintf(os, "###############################################################################################################################\n"); s << os;

    std::sprintf(os, "# CONFIG|NOCONFIG\n"); s << os;
    key = "TTC_CONFIG";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), ttc_config ? "CONFIG" : "NOCONFIG"); s << os;
    std::sprintf(os, "# TX0..11\n"); s << os;
    std::sprintf(os, "# ENABLED|DISABLED\n"); s << os;
    key = "TTC_TX_ENABLE";
    std::sprintf(os, "%-27s= ", key.c_str()); s << os;
    for (i = 0; i < AltiModule::TRANSMITTER_NUMBER; i++) {
        if (i < AltiModule::TRANSMITTER_NUMBER - 1) {
            std::sprintf(os, "%s", ttc_tx_enable[i] ? "ENABLED," : "DISABLED,"); s << os;
        }
        else {
            std::sprintf(os, "%s\n", ttc_tx_enable[i] ? "ENABLED" : "DISABLED"); s << os;
        }
    }
    key = "TTC_L1A_SOURCE_FROM_L1A";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), ttc_l1a_source[0] ? "ENABLED" : "DISABLED"); s << os;
    key = "TTC_L1A_SOURCE_FROM_TTR1..3";
    std::sprintf(os, "%-27s= %s,%s,%s\n", key.c_str(), ttc_l1a_source[1] ? "ENABLED" : "DISABLED", ttc_l1a_source[2] ? "ENABLED" : "DISABLED", ttc_l1a_source[3] ? "ENABLED" : "DISABLED"); s << os;
    
/*
    std::sprintf(os, "# 0x00..0x1f [step: 3.12ns]\n"); s << os;
    key = "TTC_TX_DELAY";
    std::sprintf(os, "%-27s= ", key.c_str()); s << os;
    for (i = 0; i < AltiModule::TRANSMITTER_NUMBER; i++) {
        if (i < AltiModule::TRANSMITTER_NUMBER - 1) {
            std::sprintf(os, "0x%02x,", ttc_tx_delay[i]); s << os;
        }
        else {
            std::sprintf(os, "0x%02x\n", ttc_tx_delay[i]); s << os;
        }
    }
*/
    std::sprintf(os, "# <WIDTH>,<DELAY>\n"); s << os;
    std::sprintf(os, "# <WIDTH> = 0..4095 [step: BC]\n"); s << os;
    std::sprintf(os, "# <DELAY> = 0..4095 [step: BC]\n"); s << os;
    for (i = 0; i <= 3; i++) {
        switch (i) {
            case 0:
                key = "TTC_BGO0_INHIBIT";
                break;
            case 1:
                key = "TTC_BGO1_INHIBIT";
                break;
            case 2:
                key = "TTC_BGO2_INHIBIT";
                break;
            case 3:
                key = "TTC_BGO3_INHIBIT";
                break;
            default:
                break;
        }
        std::sprintf(os, "%-27s= ", key.c_str()); s << os;
        std::sprintf(os, "%04d,%04d\n", ttc_bgo_inhibit_width[i], ttc_bgo_inhibit_delay[i]); s << os;
    }
  
    std::sprintf(os, "# 0..4095 [step: BC]\n"); s << os;
    key = "TTC_TTYP_DELAY";
    std::sprintf(os, "%-27s= ", key.c_str()); s << os;
    std::sprintf(os, "%04d\n", ttc_ttyp_delay); s << os;
    std::sprintf(os, "# 0x0000..0x3fff\n"); s << os;
    key = "TTC_TTYP_ADDR";
    std::sprintf(os, "%-27s= ", key.c_str()); s << os;
    std::sprintf(os, "0x%04x\n", ttc_ttyp_addr); s << os;
    std::sprintf(os, "# 0x00..0xff\n"); s << os;
    key = "TTC_TTYP_SUBADDR";
    std::sprintf(os, "%-27s= ", key.c_str()); s << os;
    std::sprintf(os, "0x%02x\n", ttc_ttyp_subaddr); s << os;
    std::sprintf(os, "# SHORT|LONG\n"); s << os;
    key = "TTC_TTYP_FRAME";
    std::sprintf(os, "%-27s= ", key.c_str()); s << os;
    std::sprintf(os, "%s\n", AltiModule::TTC_FRAME_TYPE_NAME[ttc_ttyp_frame_type].c_str()); s << os;
    std::sprintf(os, "# INTERNAL|EXTERNAL\n"); s << os;
    key = "TTC_TTYP_SPACE";
    std::sprintf(os, "%-27s= ", key.c_str()); s << os;
    std::sprintf(os, "%s\n", AltiModule::TTC_ADDRESS_SPACE_NAME[ttc_ttyp_address_space].c_str()); s << os;
    std::sprintf(os, "# SYNCHRONOUS_SINGLE_BGO_SIGNAL|SYNCHRONOUS_REPETITIVE|ASYNCHRONOUS_BGO_SIGNAL|ASYNCHRONOUS_VME_ON_TRIGGER|ASYNCHRONOUS_VME_WHEN_NOT_EMPTY\n"); s << os;
    for (i = 0; i <= 3; i++) {
        switch (i) {
            case 0:
                key = "TTC_BGO0_MODE";
                break;
            case 1:
                key = "TTC_BGO1_MODE";
                break;
            case 2:
                key = "TTC_BGO2_MODE";
                break;
            case 3:
                key = "TTC_BGO3_MODE";
                break;
            default:
                break;
        }
        std::sprintf(os, "%-27s= ", key.c_str()); s << os;
        std::sprintf(os, "%s\n", AltiModule::TTC_BGO_MODE_NAME[ttc_bgo_mode[i]].c_str()); s << os;
    }
    std::sprintf(os, "# ENABLED|DISABLED\n"); s << os;
    for (i = 0; i <= 3; i++) {
      key = "TTC_BGO" + to_string(i) + "_BUSY_GATE";
      std::sprintf(os, "%-27s= ", key.c_str()); s << os;
      std::sprintf(os, "%s\n", ttc_bgo_busy_gate[i] ? "ENABLED" : "DISABLED"); s << os;
    }
    std::sprintf(os, "# ENABLED|DISABLED\n"); s << os;
    for (i = 0; i < AltiModule::TTC_FIFO_NUMBER; i++) {
        key = "TTC_" + AltiModule::TTC_FIFO_NAME[(AltiModule::TTC_FIFO) i] + "_RETRANS";
        std::sprintf(os, "%-27s= ", key.c_str()); s << os;
        std::sprintf(os, "%s\n", ttc_fifo_retransmit[i] ? "ENABLED" : "DISABLED"); s << os;
    }

    std::sprintf(os, "###############################################################################################################################\n"); s << os;

    std::sprintf(os, "# CONFIG|NOCONFIG\n"); s << os;
    key = "CNT_CONFIG";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), cnt_config ? "CONFIG" : "NOCONFIG"); s << os;

    std::sprintf(os, "# ENABLED|DISABLED\n"); s << os;
    AltiModule::CNT_TYPE cnt;
    for (i = 0; i < AltiModule::CNT_NUMBER; i++) {
        cnt = (AltiModule::CNT_TYPE) i;
        // merging
        if ((cnt >= AltiModule::CNT_TTR2) && (cnt <= AltiModule::CNT_TTR3)) continue;
        if ((cnt >= AltiModule::CNT_BGO1) && (cnt <= AltiModule::CNT_BGO3)) continue;
        if (cnt == AltiModule::CNT_TTR1) {
            std::sprintf(os, "# TTR1..3\n"); s << os;
            key = "CNT_TTR";
            std::sprintf(os, "%-27s= %s,%s,%s\n", key.c_str(), cnt_enable[i] ? "ENABLED" : "DISABLED", cnt_enable[i + 1] ? "ENABLED" : "DISABLED", cnt_enable[i + 2] ? "ENABLED" : "DISABLED"); s << os;
        }
        else if (cnt == AltiModule::CNT_BGO0) {
            std::sprintf(os, "# BGO0..3\n"); s << os;
            key = "CNT_BGO";
            std::sprintf(os, "%-27s= %s,%s,%s,%s\n", key.c_str(), cnt_enable[i] ? "ENABLED" : "DISABLED", cnt_enable[i + 1] ? "ENABLED" : "DISABLED", cnt_enable[i + 2] ? "ENABLED" : "DISABLED", cnt_enable[i + 3] ? "ENABLED" : "DISABLED"); s << os;
        }
        else {
            key = AltiModule::CNT_NAME[i];
            std::sprintf(os, "%-27s= %s\n", key.c_str(), cnt_enable[i] ? "ENABLED" : "DISABLED"); s << os;
        }
    }
    std::sprintf(os, "# ALTI/data/AltiModule_cnt_ttyp_lut.dat\n"); s << os;
    key = "CNT_TTYP_LUT_FILE";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), cnt_ttyp_lut_file.c_str()); s << os;

    std::sprintf(os, "###############################################################################################################################\n"); s << os;

    std::sprintf(os, "# CONFIG|NOCONFIG\n"); s << os;
    key = "CTP_CONFIG";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), ctp_config ? "CONFIG" : "NOCONFIG"); s << os;
    std::sprintf(os, "# VME|INTERNAL,LEN=<LEN>,FRQ=<FRQ>,BEF=<BEF>,AFT=<AFT>,OFF=<OFF>\n"); s << os;
    std::sprintf(os, "# <LEN> = 0..3          [step: BC, pulse length           = <LEN> + 1]\n"); s << os;
    std::sprintf(os, "# <FRQ> = 0..4294967295 [step: BC, frequency rate divider = <FRQ> + 1]\n"); s << os;
    std::sprintf(os, "# <BEF> = 0..16777215   [step: BC, length of the busy before ECR = <BEF>]\n"); s << os;
    std::sprintf(os, "# <AFT> = 0..16777215   [step: BC, length of the busy after  ECR = <AFT>]\n"); s << os;
    std::sprintf(os, "# <OFF> = 0..4095       [step: BC, orbit offset in order to remove the ECR busy = <OFF>]\n"); s << os;
    key = "CTP_ECR_GENERATION";
    std::sprintf(os, "%-27s= %s\n", key.c_str(), ctp_ecr_generation.print().c_str()); s << os;

    std::sprintf(os, "###############################################################################################################################\n"); s << os;

    // close output file
    s.close();

    // print
    COUT("wrote AltiConfiguration to file \"%s\"", fn.c_str());

    return (0);
}


//------------------------------------------------------------------------------

void AltiConfiguration::readSWXIndividualToBundleTTYP() {
    unsigned int i, j, ttyp;
    AltiModule::SIGNAL_SOURCE src = AltiModule::SIGNAL_SOURCE::INVALID_CONFIG;
    bool mismatch;
    for (j = 0; j < AltiModule::SIGNAL_DESTINATION_NUMBER; j++) {
        ttyp = 0;
        mismatch = false;
        for (i = 0; i < AltiModule::SIGNAL_NUMBER_ROUTED; i++) {
            if ((AltiModule::SIGNAL_NAME[i]).rfind("TTYP", 0) == 0) {
                if (ttyp++ == 0) {
                    src = sig_swx[i][j];
                }
                else {
                    if (sig_swx[i][j] != src) {
                        mismatch = true;
                        break;
                    }
                }
            }
        }
        if (mismatch) {
            sig_swx_ttyp[j] = AltiModule::SIGNAL_SOURCE::INVALID_CONFIG;
        }
        else {
            sig_swx_ttyp[j] = src;
        }
    }
}

//------------------------------------------------------------------------------

int AltiConfiguration::readNextRecord(std::istream &s, std::string &key, std::string &val) {

    char os[STRING_LENGTH];
    std::string line, nkey;

    val.clear();

    while (true) {
        if (!s.getline(os, STRING_LENGTH)) return (AltiModule::FAILURE);
        line = os;
        if (!line.empty() && !(line.substr(0, 1) == "#") && !(line.substr(0, 2) == "//")) break;
    }

    std::string::size_type pos = line.find_first_of("=", 0);
    nkey = AltiCommon::trimString(line.substr(0, pos));

    if (nkey != key) {
        CERR("error in fixed format: expected key \"%s\" got key \"%s\"", key.c_str(), nkey.c_str());
        return (AltiModule::FAILURE);
    }

    val = AltiCommon::trimString(line.substr(pos + 1));

    return (AltiModule::SUCCESS);
}//------------------------------------------------------------------------------

std::string AltiConfiguration::bool2string(const bool enabled) {
    return (enabled ? "ENABLED" : "DISABLED");
}

//------------------------------------------------------------------------------

std::string AltiConfiguration::phase2string(const unsigned phase) {
    return ((phase == 0) ? "AUTO" : ((phase == 1) ? "000DEG" : ((phase == 2) ? "090DEG" : ((phase == 3) ? "180DEG" : (phase == 4) ? "270DEG" : ""))));
}
