//******************************************************************************
// file: AltiCounter.cc
// desc: library for ALTI counter (scalars)
// auth: 04-DEC-2017 P. Kuzmanovic
//******************************************************************************

#include "ALTI/AltiCommon.h"
#include "ALTI/AltiCounter.h"
#include "RCDUtilities/RCDUtilities.h"

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <cmath>

using namespace LVL1;

//------------------------------------------------------------------------------

const std::string AltiCounter::Error::TYPE_NAME[TYPE_NUMBER] = {
    "LUT", "TBP", "TAP", "TAV", "L1A"
};

const unsigned int AltiCounter::Error::TYPE_MASK[TYPE_NUMBER] = {
    0x00000001, 0x00000002, 0x00000004, 0x00000008, 0x00000010
};

//------------------------------------------------------------------------------

AltiCounter::Error::Error(const int num) {

    // reset to num
    reset(num);
}

//------------------------------------------------------------------------------

AltiCounter::Error::~Error() {

    // nothing to be done
}

//------------------------------------------------------------------------------

void AltiCounter::Error::reset(const int num) {

    /*for (unsigned int typ = 0; typ < TYPE_NUMBER; typ++) {
        m_error[typ] = num;
        m_bit_error[typ] = num;
    }*/
    m_error = num;
    m_bit_error = num;
}

//------------------------------------------------------------------------------

int64_t AltiCounter::Error::error() const {

    int64_t sum(0);

    //for (unsigned int typ = 0; typ < TYPE_NUMBER; typ++) sum += m_error[typ];
    sum += m_error;

    return (sum);
}

//------------------------------------------------------------------------------

void AltiCounter::Error::print(const Error &msk, const int num) const {

    double perc, ebit;

    /*for (unsigned int t = 0; t < TYPE_NUMBER; t++) {
        if (msk[(TYPE) t] != 0) {
            perc = (num) ? (((double) m_error[t])/((double) num)/double(AltiModule::CNT_SIZE[t])*100.0) : (0.0);
            ebit = (m_error[t]) ? (((double) m_bit_error[t])/((double) m_error[t])) : (0.0);
            std::printf("   number of %-20s : %16" PRId64 " => %11.3f%% and %7.3f bit/err\n", (TYPE_NAME[t] + " errors").c_str(), m_error[t], perc, ebit);
        }
    }*/
    if (msk[(TYPE) 0] != 0) {
        perc = (num) ? (((double) m_error)/((double) num)/double(AltiModule::CNT_SIZE)*100.0) : (0.0);
        ebit = (m_error) ? (((double) m_bit_error)/((double) m_error)) : (0.0);
        std::printf("   number of %-20s : %16" PRId64 " => %11.3f%% and %7.3f bit/err\n", (/*TYPE_NAME[t]*/std::string("CNT") + " errors").c_str(), m_error, perc, ebit);
    }
}

//------------------------------------------------------------------------------

std::string AltiCounter::Error::printTypeMask(const unsigned int mask) {

    std::ostringstream buf;
    bool flag(false);

    /*for (unsigned int i = 0; i < TYPE_NUMBER; i++) {
        if (mask & TYPE_MASK[i]) {
            buf << (flag ? "|" : "") << TYPE_NAME[i];
            flag = true;
        }
    }*/
    if (mask & TYPE_MASK[0]) {
        buf << (flag ? "|" : "") << TYPE_NAME[0];
        flag = true;
    }

    return (buf.str());
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

/*const unsigned int AltiCounter::DSIZE[AltiModule::CNT_NUMBER] = {
    AltiModule::CAM_INPUT_NUMBER,
    AltiModule::ITEM_NUMBER,
    AltiModule::ITEM_NUMBER,
    AltiModule::ITEM_NUMBER
};*/
const unsigned int AltiCounter::DSIZE = AltiModule::CNT_SIZE;

/*const unsigned int AltiCounter::LSIZE[AltiModule::CNT_NUMBER] = {
    AltiModule::CAM_INPUT_NUMBER,
    AltiModule::ITEM_NUMBER,
    AltiModule::ITEM_NUMBER,
    AltiModule::ITEM_NUMBER + AltiModule::PARTITION_NUMBER
};*/
const unsigned int AltiCounter::LSIZE = AltiModule::CNT_SIZE;

const std::string AltiCounter::CNT_NAME[AltiModule::CNT_SIZE] = {
    "TTYP_BUCKET0",
    "TTYP_BUCKET1",
    "TTYP_BUCKET2",
    "TTYP_BUCKET3",
    "TTYP_BUCKET4",
    "TTYP_BUCKET5",
    "TTYP_BUCKET6",
    "TTYP_BUCKET7",
    "ORB",
    "L1A",
    "TTR1",
    "TTR2",
    "TTR3",
    "BGO0",
    "BGO1",
    "BGO2",
    "BGO3"
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

AltiCounter::AltiCounter(/*const AltiModule::CNT_TYPE typ*/) : m_segment(0)/*, m_type(typ)*/, m_turn(0) {

    // create data array
    m_data = new unsigned int[size()];
}

//------------------------------------------------------------------------------

AltiCounter::AltiCounter(RCD::CMEMSegment *seg/*, const AltiModule::CNT_TYPE typ*/) : m_segment(seg)/*, m_type(typ)*/, m_turn(0) {

    // get virtual address
    m_data = (unsigned int *) m_segment->VirtualAddress();
}

//------------------------------------------------------------------------------

AltiCounter::~AltiCounter() {

    // delete data array
    if (!m_segment) delete [] m_data;
}

//------------------------------------------------------------------------------

unsigned int AltiCounter::overflow() const {

    unsigned int ovf(0), i;

    for (i = 0; i < size(); i++) if (overflow(i)) ovf++;

    return (ovf);
}

//------------------------------------------------------------------------------

void AltiCounter::dump() const {

    unsigned int i;
   
    std::cout << ">> ------------------------------------------------------------------------------------------------------------------" << std::endl;
    std::cout << ">> ALTI: counters" << std::endl;

    for (i = 0; i < DSIZE; i++) {
        std::printf("%-14s:  %s0x%08x", ("\"" + CNT_NAME[i] + "\"").c_str(), overflow(i) ? "O" : " ", counter(i));
        std::cout << std::endl;
    }
    std::printf("TURN          :  %s0x%08x (= %10d)\n", turnOverflow() ? "O" : " ", m_turn, m_turn);
    std::printf("              :   overflow    = %10d\n", overflow());
    std::cout << ">> ------------------------------------------------------------------------------------------------------------------" << std::endl;
}

//------------------------------------------------------------------------------

void AltiCounter::dump(const AltiCounter &cnt, const Error &msk) const {

    bool err(msk[static_cast<AltiCounter::Error::TYPE>(/*m_type*/0)]);
    unsigned int i;
    char s[AltiCommon::STRING_LENGTH];

    std::cout << ">> ------------------------------------------------------------------------------------------------------------------" << std::endl;
    std::cout << ">> ALTI:" << std::endl;

    for (i = 0; i < DSIZE; i++) {
        if ((i%8) == 0) {
            if (size() > 1) std::sprintf(s, "%3d", i);
            std::printf("Counter  %s: ", (size() > 1) ? s : "   ");
        }
        std::printf(" %s%s0x%08x", (err && (counter(i) != cnt.counter(i))) ? "E" : " ", overflow(i) ? "O" : " ", counter(i));
        if ((i%8) == 7) std::cout << std::endl;
    }
    if ((i%8) != 0) std::cout << std::endl;

    std::printf("TURN        : %s%s0x%08x (= %s%10d)\n", (err && ((turn() != cnt.turn()))) ? "E" : " ", turnOverflow() ? "O" : " ", m_turn, (err && ((turn() != cnt.turn()))) ? "E" : " ", m_turn);
    std::printf("            :   overflow    = %s%10d\n", (overflow() != cnt.overflow()) ? "E" : " ", overflow());
    std::cout << ">> ------------------------------------------------------------------------------------------------------------------" << std::endl;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

// double has at least 15 decimal digits precision: 4 before decimal point + 11 after decimal point
const double AltiCounterRate::EPSILON = 0.00000000001;

//------------------------------------------------------------------------------

AltiCounterRate::AltiCounterRate(/*const AltiModule::CNT_TYPE t*/)/* : m_type(t)*/ {

    m_name = AltiModule::CNT_NAME[/*m_type*/0];
    m_size = AltiModule::CNT_SIZE/*[m_type]*/;
    m_data = new double[m_size];
    m_turn = 0.0;

    // reset state
    resetState();

    // reset data
    for (unsigned int i = 0; i < m_size; i++) m_data[i] = 0.0;
}

//------------------------------------------------------------------------------

AltiCounterRate::AltiCounterRate(const AltiCounter &cnt) {

    m_name = AltiModule::CNT_NAME[static_cast<int>(/*cnt.type()*/0)];
    //m_type = cnt.type();
    m_size = cnt.size();
    m_data = new double[m_size];
    m_turn = static_cast<double>(cnt.turn());

    // reset state
    resetState();

    // set data and update state
    for (unsigned int i = 0; i < m_size; i++) {
        m_data[i] = (m_turn <= 0.0) ? 0.0 : (static_cast<double>(cnt.counter(i))/m_turn);
        if (std::abs(m_data[i] - round(m_data[i])) > EPSILON) m_fraction++;
    }
}

//------------------------------------------------------------------------------

AltiCounterRate::AltiCounterRate(const AltiCounterRate &cnt) {

    m_name = cnt.name();
    //m_type = cnt.type();
    m_size = cnt.size();
    m_data = new double[m_size];
    m_turn = cnt.turn();

    // reset state with state from counter
    m_fraction = cnt.fraction();
    m_error = -1;
    m_infinity = cnt.infinity();

    // set data
    for (unsigned int i = 0; i < m_size; i++) m_data[i] = cnt.counter(i);
}

//------------------------------------------------------------------------------

AltiCounterRate::~AltiCounterRate() {

    if (m_data) delete [] m_data;
}

//------------------------------------------------------------------------------

AltiCounterRate &AltiCounterRate::operator=(const AltiCounter &cnt) {

    m_name = AltiModule::CNT_NAME[static_cast<int>(/*cnt.type()*/0)];
    if (m_size != cnt.size()) {
        m_size = cnt.size();
        if (m_data) delete [] m_data;
        m_data = new double[m_size];
    }
    m_turn = static_cast<double>(cnt.turn());

    // reset state
    resetState();

    // copy data and update state
    for (unsigned int i = 0; i < m_size; i++) {
        m_data[i] = (m_turn <= 0.0) ? 0.0 : (static_cast<double>(cnt.counter(i))/m_turn);
        if (std::abs(m_data[i] - round(m_data[i])) > EPSILON) m_fraction++;
    }

    return (*this);
}

//------------------------------------------------------------------------------

AltiCounterRate &AltiCounterRate::operator=(const AltiCounterRate &cnt) {

    m_name = cnt.name();
    if (m_size != cnt.size()) {
        m_size = cnt.size();
        if (m_data) delete [] m_data;
        m_data = new double[m_size];
    }
    m_turn = cnt.turn();

    // reset state with state from counter
    m_fraction = cnt.fraction();
    m_error = -1;
    m_infinity = cnt.infinity();

    // copy data
    for (unsigned int i = 0; i < m_size; i++) {
        m_data[i] = cnt.counter(i);
    }

    return (*this);
}

//------------------------------------------------------------------------------

AltiCounterRate AltiCounterRate::operator/(const AltiCounterRate &cnt) const {

    if (m_size != cnt.m_size) {
        CERR("wrong size for division %d != %d expected", cnt.m_size, m_size);
        return (*this);
    }

    AltiCounterRate frq(*this);

    // reset state
    frq.resetState();

    // copy data and update state
    for (unsigned int i = 0; i < m_size; i++) {
        if (m_data[i] <= 0.0) {
            frq.m_data[i] = 0.0;
            if (cnt.m_data[i] <= 0.0) frq.m_infinity++;
        }
        else if (cnt.m_data[i] <= 0) {
            frq.m_data[i] = 0.0;
            frq.m_infinity++;
        }
        else {
            frq.m_data[i] = (m_data[i]/cnt.m_data[i]);
            if (std::abs(frq.m_data[i] - round(frq.m_data[i])) > EPSILON) frq.m_fraction++;
        }
    }

    return (frq);
}

//------------------------------------------------------------------------------

int AltiCounterRate::setFraction() {

    // reset state
    resetState();

    // fill data and update state
    for (unsigned int i = 0; i < AltiCounter::DSIZE/*[m_type]*/; i++) {
        if (std::abs(m_data[i] - round(m_data[i])) > EPSILON) m_fraction++;
    }

    return (AltiModule::SUCCESS);
}

//------------------------------------------------------------------------------

int AltiCounterRate::scale(const double scl) {

    // reset state
    resetState();

    for (unsigned int i = 0; i < m_size; i++) {
        m_data[i] = m_data[i]*scl;
        if (std::abs(m_data[i] - round(m_data[i])) > EPSILON) m_fraction++;
    }

    return (AltiModule::SUCCESS);
}

//------------------------------------------------------------------------------

unsigned int AltiCounterRate::size(const unsigned int n)  {

    m_size = n;

    if (m_data) delete [] m_data;
    m_data = new double[m_size];
    for (unsigned int i = 0; i < m_size; i++) m_data[i] = 0.0;

    m_turn = 0.0;
    m_fraction = -1;
    m_error = -1;
    m_infinity = -1;

    return (m_size);
}


//------------------------------------------------------------------------------

void AltiCounterRate::dump() const {

    unsigned int i;
    
    std::cout << ">> ------------------------------------------------------------------------------------------------------------------------------" << std::endl;
    std::cout << ">> ALTI: rates in kHz" << std::endl;

    for (i = 0; i < AltiCounter::DSIZE; i++) {
     
        std::printf("%-14s: %12.6f", ("\"" + AltiCounter::CNT_NAME[i] + "\"").c_str(), m_data[i]*11.2455);
         std::cout << std::endl;
    }
 
    std::printf("TURN          :  %12.3f\n", m_turn);
    if ((m_fraction > -1) || (m_infinity > -1)) std::cout << "            " << "  : ";
    if (m_fraction > -1) std::cout << ((m_fraction) ? " FRACTION = " : " fraction = ") << m_fraction;
    if (m_infinity > -1) std::cout << ((m_infinity) ? ", INFINITY = " : ", infinity = ") << m_infinity;
    if ((m_fraction > -1) || (m_infinity > -1)) std::cout << std::endl;
    std::cout << ">> ------------------------------------------------------------------------------------------------------------------------------" << std::endl;
}

//------------------------------------------------------------------------------

void AltiCounterRate::dump(const AltiCounterRate &cnt) const {

    unsigned int i;
    char s[AltiCommon::STRING_LENGTH];

    std::cout << ">> --------------------------------------------------------------------------------------------------------------------------------------" << std::endl;
    std::cout << ">> ALTI: rate in kHz" << std::endl;

    for (i = 0; i < AltiCounter::DSIZE; i++) {
        if ((i%8) == 0) {
            (size() > 1) ? std::sprintf(s, "%3d", i) : std::sprintf(s, "   ");
            std::printf("CounterRate  %s: ", s);
        }
        std::printf(" %s%12.6f", (std::abs(m_data[i] - cnt[i]) > EPSILON) ? "E" : " ", m_data[i]*11.2455);
        if ((i%8) == 7) std::cout << std::endl;
    }
    if ((i%8) != 0) std::cout << std::endl;

    std::cout << "                          -------------------------------------------------------------------------------------------------------" << std::endl;
    std::printf("TURN          %12.3f\n", m_turn);
    if ((m_fraction > -1) || (m_error > -1) || (m_infinity > -1)) std::printf("                : ");
    if (m_fraction > -1) std::cout << ((m_fraction) ? " FRACTION = "  : " fraction = ") << m_fraction;
    if (m_error > -1)    std::cout << ((m_error)    ? ", ERROR = "    : ", error = ") << m_error;
    if (m_infinity > -1) std::cout << ((m_infinity) ? ", INFINITY = " : ", infinity = ") << m_infinity;
    if ((m_fraction > -1) || (m_error > -1) || (m_infinity > -1)) std::cout << std::endl;
    std::cout << ">> --------------------------------------------------------------------------------------------------------------------------------------" << std::endl;
}

//------------------------------------------------------------------------------

void AltiCounterRate::dump(const AltiCounterRate &cnt, const AltiCounter::Error &msk) const {

    unsigned int i/*, j*/;
    bool err;
    char s[AltiCommon::STRING_LENGTH];

    std::cout << ">> --------------------------------------------------------------------------------------------------------------------------------------" << std::endl;
    std::cout << ">> ALTI: rate in kHz" << std::endl;

    err = (msk[static_cast<AltiCounter::Error::TYPE>(0)]);
    for (i = 0; i < AltiCounter::DSIZE; i++) {
        if ((i%8) == 0) {
            (size() > 1) ? std::sprintf(s, "%3d", i) : std::sprintf(s, "   ");
            std::printf("CounterRate  %s: ", s);
        }
        std::printf(" %s%12.6f", (err && (std::abs(m_data[i] - cnt[i]) > EPSILON)) ? "E" : " ", m_data[i]*11.2455);
        if ((i%8) == 7) std::cout << std::endl;
    }
    if ((i%8) != 0) std::cout << std::endl;

    std::cout << "                          -------------------------------------------------------------------------------------------------------" << std::endl;
    std::printf("TURN          %12.3f\n", m_turn);
    if ((m_fraction > -1) || (m_error > -1) || (m_infinity > -1)) std::printf("                : ");
    if (m_fraction > -1) std::cout << ((m_fraction) ? " FRACTION = "  : " fraction = ") << m_fraction;
    if (m_error > -1)    std::cout << ((m_error)    ? ", ERROR = "    : ", error = ") << m_error;
    if (m_infinity > -1) std::cout << ((m_infinity) ? ", INFINITY = " : ", infinity = ") << m_infinity;
    if ((m_fraction > -1) || (m_error > -1) || (m_infinity > -1)) std::cout << std::endl;
    std::cout << ">> --------------------------------------------------------------------------------------------------------------------------------------" << std::endl;
}

//------------------------------------------------------------------------------

void AltiCounterRate::dump(const AltiCounterRate &cnt, const AltiCounterRate &sig, const AltiCounter::Error &msk) const {

    unsigned int i/*, j*/;
    bool err;
    char s[AltiCommon::STRING_LENGTH];

    std::cout << ">> --------------------------------------------------------------------------------------------------------------------------------------" << std::endl;
    std::cout << ">> ALTI: rate in kHz" << std::endl;

    err = (msk[static_cast<AltiCounter::Error::TYPE>(0)]);
    for (i = 0; i < AltiCounter::DSIZE; i++) {
        if ((i%8) == 0) {
            (size() > 1) ? std::sprintf(s, "%3d", i) : std::sprintf(s, "   ");
            std::printf("CounterRate  %s: ", s);
        }
        std::printf(" %s%12.6f", (err && (std::abs(m_data[i] - cnt[i]) > std::sqrt(sig[i]*sig[i] + EPSILON*EPSILON))) ? "E" : " ", m_data[i]*11.2455);
        if ((i%8) == 7) std::cout << std::endl;
    }
    if ((i%8) != 0) std::cout << std::endl;

    std::cout << "                          -------------------------------------------------------------------------------------------------------" << std::endl;
    std::cout << "TURN        " << "    : ";
    std::printf("  %12.3f\n", m_turn);
    if ((m_fraction > -1) || (m_error > -1) || (m_infinity > -1)) std::cout << "            " << "    : ";
    if (m_fraction > -1) std::cout << ((m_fraction) ? " FRACTION = "  : " fraction = ") << m_fraction;
    if (m_error > -1)    std::cout << ((m_error)    ? ", ERROR = "    : ", error = ") << m_error;
    if (m_infinity > -1) std::cout << ((m_infinity) ? ", INFINITY = " : ", infinity = ") << m_infinity;
    if ((m_fraction > -1) || (m_error > -1) || (m_infinity > -1)) std::cout << std::endl;
    std::cout << ">> --------------------------------------------------------------------------------------------------------------------------------------" << std::endl;
}

//------------------------------------------------------------------------------

void AltiCounterRate::dumpSigma(const AltiCounterRate &sig) const {

    unsigned int i, j;
    char s[AltiCommon::STRING_LENGTH];

    std::cout << ">> ------------------------------------------------------------------------------------------------------------------------------" << std::endl;
    std::cout << ">> ALTI:" << std::endl;

    // DATA: LUT, TBP, TAP, or TAV
    for (i = 0; i < AltiCounter::DSIZE/*[m_type]*/; i++) {
        if ((i%8) == 0) {
            (size() > 1) ? std::sprintf(s," %3d", i) : std::sprintf(s, "   ");
            //std::printf("CounterRate [\"%s\"] %s: ", m_name.c_str(), s);
            std::printf("CounterRate  %s: ", s);
        }
        std::printf(" %12.6f", m_data[i]);
        if ((i%8) == 7) {
            std::cout << std::endl;
            //std::printf("SigmaRate   [\"%s\"] %3d: ", m_name.c_str(), i - 7);
            std::printf("SigmaRate    %3d: ", i - 7);
            for (j = i - 7; j <= i; j++) {
                std::sprintf(s, "+-%.6f", sig[j]);
                std::printf(" %12s", s);
            }
            std::cout << std::endl;
            std::cout << "                          -------------------------------------------------------------------------------------------------------" << std::endl;
        }
    }
    if ((i%8) != 0) {
        std::cout << std::endl;
        //std::printf("SigmaRate   [\"%s\"] %3d: ", m_name.c_str(), i - i%8);
        std::printf("SigmaRate    %3d: ", i - i%8);
        for (j = i - i%8; j < i; j++) {
            std::sprintf(s, "+-%.6f", sig[j]);
            std::printf(" %12s", s);
        }
        std::cout << std::endl;
        std::cout << "                          -------------------------------------------------------------------------------------------------------" << std::endl;
    }

    // L1A
    /*if (m_type == AltiModule::CNT_TAV) {
        for (i = AltiCounter::DSIZE[m_type], j = 0; i < AltiCounter::LSIZE[m_type]; i++, j++) {
            if ((j%8) == 0) std::printf("CounterRate [\"L1A\"]    : ");
            std::printf(" %12.6f", m_data[i]);
            if ((j%8) == 7) std::cout << std::endl;
        }
        if ((j%8) != 0) std::cout << std::endl;
        std::printf("SigmaRate   [\"L1A\"]    : ");
        for (i = AltiCounter::DSIZE[m_type]; i < AltiCounter::LSIZE[m_type]; i++) {
            std::sprintf(s, "+-%.6f", sig[i]);
            std::printf(" %12s", s);
        }
        std::cout << std::endl;
        std::cout << "                          -------------------------------------------------------------------------------------------------------" << std::endl;
    }*/

    //std::printf("TURN        [\"%s\"]    :  %12.3f\n", m_name.c_str(), m_turn);
    std::printf("TURN            :  %12.3f\n", m_turn);
    //if ((m_fraction > -1) || (m_infinity > -1)) std::cout << "            [\"" << m_name << "\"]    : ";
    if ((m_fraction > -1) || (m_infinity > -1)) std::cout << "            " << "    : ";
    if (m_fraction > -1) std::cout << ((m_fraction) ? " FRACTION = " : " fraction = ") << m_fraction;
    if (m_infinity > -1) std::cout << ((m_infinity) ? ", INFINITY = " : ", infinity = ") << m_infinity;
    if ((m_fraction > -1) || (m_infinity > -1)) std::cout << std::endl;
    std::cout << ">> ------------------------------------------------------------------------------------------------------------------------------" << std::endl;
}

//------------------------------------------------------------------------------

int AltiCounterRate::check(const AltiCounterRate &cnt) {

    // reset error
    m_error = 0;

    // for all DATA
    for (unsigned int i = 0; i < m_size; i++) {
        if (std::abs(m_data[i] - cnt[i]) > EPSILON) m_error++;
    }

    return (m_error);
}

//------------------------------------------------------------------------------

int AltiCounterRate::check(const AltiCounterRate &cnt, const AltiCounter::Error &msk, AltiCounter::Error &err) {

    // reset error
    m_error = 0;

    // DATA: LUT, TBP, TAP, or TAV
    if (msk[static_cast<AltiCounter::Error::TYPE>(/*m_type*/0)]) {
        for (unsigned int i = 0; i < AltiCounter::DSIZE/*[m_type]*/; i++) {
            if (std::abs(m_data[i] - cnt[i]) > EPSILON) {
                err.m_error/*[m_type]*/++;
                err.m_bit_error/*[m_type]*/ += checkWord(m_data[i], cnt[i]);
                m_error++;
            }
        }
    }

    // L1A
    /*if ((m_type == AltiModule::CNT_TAV) && (msk[AltiCounter::Error::L1A])) {
        for (unsigned int i = AltiCounter::DSIZE[m_type], j = 0; i < AltiCounter::LSIZE[m_type]; i++, j++) {
            if (std::abs(m_data[i] - cnt[i]) > EPSILON) {
                err.m_error[AltiCounter::Error::L1A]++;
                err.m_bit_error[AltiCounter::Error::L1A] += checkWord(m_data[i], cnt[i]);
                m_error++;
            }
        }
    }*/

    return (m_error);
}

//------------------------------------------------------------------------------

int AltiCounterRate::check(const AltiCounterRate &cnt, const AltiCounterRate &sig, const AltiCounter::Error &msk, AltiCounter::Error &err) {

    // reset error
    m_error = 0;

    // DATA: LUT, TBP, TAP, or TAV
    if (msk[static_cast<AltiCounter::Error::TYPE>(/*m_type*/0)]) {
        for (unsigned int i = 0; i < AltiCounter::DSIZE/*[m_type]*/; i++) {
            CDBG("CNT[%s%03d]: data = %15.8f, cnt = %12.8f, tol = %12.8f, dev = %12.8f%%", (AltiCounter::Error::TYPE_NAME[m_type].c_str()) "CNT", i, m_data[i], cnt[i], sig[i], std::abs(m_data[i] - cnt[i])/std::sqrt(sig[i]*sig[i] + EPSILON*EPSILON)*100.0);
            if (std::abs(m_data[i] - cnt[i]) > std::sqrt(sig[i]*sig[i] + EPSILON*EPSILON)) {
                err.m_error/*[m_type]*/++;
                err.m_bit_error/*[m_type]*/ += checkWord(m_data[i], cnt[i]);
                m_error++;
            }
        }
    }

    // L1A
    /*if ((m_type == AltiModule::CNT_TAV) && (msk[AltiCounter::Error::L1A])) {
        for (unsigned int i = AltiCounter::DSIZE[m_type], j = 0; i < AltiCounter::LSIZE[m_type]; i++, j++) {
            if (std::abs(m_data[i] - cnt[i]) > std::sqrt(sig[i]*sig[i] + EPSILON*EPSILON)) {
                err.m_error[AltiCounter::Error::L1A]++;
                err.m_bit_error[AltiCounter::Error::L1A] += checkWord(m_data[i], cnt[i]);
                m_error++;
            }
        }
    }*/

    return (m_error);
}

//------------------------------------------------------------------------------

int AltiCounterRate::checkWord(const unsigned int data, const unsigned int datb) const {

    unsigned int i;
    int nbit(0);

    // count number of bit errors
    for (i = 0; i < WORD_SIZE; i++) {
        if ((data & (1U << i)) != (datb & (1U << i))) nbit++;
    }

    // return bit error
    return (nbit);
}

//------------------------------------------------------------------------------

void AltiCounterRate::resetState() {

    // reset state
    m_fraction = 0;
    m_error = -1;
    m_infinity = -1;
}
