#include <AccessManager/client/ClientConfig.h>

#include <ers/ers.h>
#include <stdlib.h>
#include <fstream>
#include <iostream>

#ifndef CONFIGURATION_SOURCE_FILE
#define CONFIGURATION_SOURCE_FILE "/sw/tdaq/AccessManager/cfg/client.cfg"
#endif

namespace daq
{

namespace am
{

// the environment variable that stores the AM server host
const std::string ClientConfig::PARAMETER_AM_SERVER_HOST("TDAQ_AM_SERVER_HOST");

// the environment variable that stores the AM server port
const std::string ClientConfig::PARAMETER_AM_SERVER_PORT("TDAQ_AM_SERVER_PORT");

// the environment variable that stores the flag enabled/disabled of AM authorization
const std::string ClientConfig::PARAMETER_AM_AUTHORIZATION("TDAQ_AM_AUTHORIZATION");
const std::string ClientConfig::VALUE_AM_AUTHORIZATION_ENABLED("on");
const std::string ClientConfig::VALUE_AM_AUTHORIZATION_DISABLED("off");
const std::string ClientConfig::VALUE_AM_AUTHORIZATION_DEFAULT(ClientConfig::VALUE_AM_AUTHORIZATION_DISABLED);

// the timeout variables in milliseconds
const std::string ClientConfig::PARAMETER_AM_CLIENT_CONNECT_TIMEOUT("TDAQ_AM_CLIENT_CONNECT_TIMEOUT");
const std::string ClientConfig::PARAMETER_AM_CLIENT_RECV_TIMEOUT("TDAQ_AM_CLIENT_RECV_TIMEOUT");

const std::string ClientConfig::PARAMETER_AM_CLIENT_USE_SECONDARY_SERVERS("TDAQ_AM_CLIENT_USE_SECONDARY_SERVERS");
const std::string ClientConfig::VALUE_AM_CLIENT_DONT_USE_SECONDARY_SERVERS("false");
const std::string ClientConfig::VALUE_AM_CLIENT_USE_SECONDARY_SERVERS("true");
const std::string ClientConfig::VALUE_AM_CLIENT_USE_SECONDARY_SERVERS_DEFAULT(ClientConfig::VALUE_AM_CLIENT_DONT_USE_SECONDARY_SERVERS);

ClientConfig::ClientConfig(){
	this->parameters[PARAMETER_AM_SERVER_HOST] 					= "";
	this->parameters[PARAMETER_AM_SERVER_PORT] 					= "";
	this->parameters[PARAMETER_AM_AUTHORIZATION] 				= "";
	this->parameters[PARAMETER_AM_CLIENT_CONNECT_TIMEOUT]		= "";
	this->parameters[PARAMETER_AM_CLIENT_RECV_TIMEOUT]			= "";
	this->parameters[PARAMETER_AM_CLIENT_USE_SECONDARY_SERVERS]	= "";
	refresh();
}

ClientConfig::~ClientConfig(){

}

void ClientConfig::refresh(){

	// first try to read the variables from the configuration file
	std::string line;
	std::ifstream cfgFile(CONFIGURATION_SOURCE_FILE);
	if (cfgFile.is_open()){
		ERS_DEBUG(1, "Read configuration from file [" << CONFIGURATION_SOURCE_FILE << "]");
		while (! (cfgFile.eof())) {
			std::getline(cfgFile, line);

			// Trim Leading Spaces
			std::string::size_type pos = line.find_first_not_of(" \t"); // Find the first character position after excluding leading blank spaces
 			if(std::string::npos != pos) {
				line = line.substr(pos);
			}
			// Trim trailing Spaces
			pos = line.find_last_not_of(" \t"); // Find the first character position from reverse after excluding trailing white spaces
			if(std::string::npos != pos) {
				line = line.substr( 0, pos + 1 );
			}

			// check if the line is a comment
			if (line.find('#', 0) == 0) {
				ERS_DEBUG(3, "Ignore comment from input file: " << line);
				continue;
			}

			// the parameters in the file should be in this format 'VARIABLE_NAME=VALUE'
			pos = line.find("=");
			if (std::string::npos != pos) {
				// get the key
				std::string key = line.substr(0, pos);
				// get the value
				std::string value = line.substr(pos + 1);

				//check if already added
				std::map<std::string,std::string>::iterator pit = parameters.find(key);
				if( pit != parameters.end() ) {
					if ( pit->second.size() != 0) {
						ERS_DEBUG(2, "parameter[" << key << "] old value [" << pit->second << "] new value from file [" << value << "]");
					}
				} else {
					ERS_DEBUG(2, "new parameter[" << key << "] with value from file [" << value << "]");
				}

				// set the new value for the key
				parameters[key] = value;
			} else {
				ERS_DEBUG(3, "Following line is not in the format 'VARIABLE_NAME=variable value':" << line);
			}
	    }
	    cfgFile.close();
	 } else {
		 ERS_LOG("Unable to open the configuration file " << CONFIGURATION_SOURCE_FILE);
	 }

	// now read the parameters from the environment
	char *envValue;
	for(std::map<std::string, std::string>::iterator pit = parameters.begin(); pit != parameters.end(); pit++){
		if (pit->second.size() > 0) {
			ERS_DEBUG(3, "Variable " << (pit->first) << " will not be read from environment because has already the value [" << (pit->second) << "]");
			continue;
		}

		envValue = getenv((pit->first).c_str());
		if (envValue == NULL){
			ERS_DEBUG(2, "Environment variable [" << (pit->first) << "] not found!");
		} else {
			ERS_DEBUG(2, "Environment variable [" << (pit->first) << "] found!"
						<< " Value [" << envValue << "]");
			parameters[(pit->first)] = envValue;
		}
	}

}

const std::string ClientConfig::getConstantValueFromMap(const std::string& key) const{
	for(std::map<std::string, std::string>::const_iterator pit = parameters.begin(); pit != parameters.end(); pit++){
		if (pit->first == key){
			ERS_DEBUG(2, "Parameter [" << (key) << "] found with value [" << pit->second << "]");
			return pit->second;
		}
	}
	ERS_DEBUG(1, "Parameter [" << (key) << "] not found!");
	return "";
}


// get the AM server host name
const std::string ClientConfig::getAMServerHosts() const{
	return getConstantValueFromMap(PARAMETER_AM_SERVER_HOST);
}

// METHODS TO ITERATE THROUGH THE AM SERVER HOSTS IF THERE ARE MORE THAN ONE
// get the first AM server host name;
// the others are returned by the getNextAMSErverHost method
const std::string ClientConfig::getFirstAMServerHost(){
	this->amServerHostsAvailable = getAMServerHosts();
	return getNextAMServerHost();
}

// get the next AM server host name;
// returns an empty string if there are no more AM servers in the list
const std::string ClientConfig::getNextAMServerHost(){
	std::string currentAMServer = "";
	if (this->amServerHostsAvailable.size() > 0){
		// find the next separator or end of string
		std::string::size_type loc = amServerHostsAvailable.find(',', 0);
		if (loc != std::string::npos) {
			currentAMServer = amServerHostsAvailable.substr(0, loc);
			amServerHostsAvailable = amServerHostsAvailable.substr(loc + 1);
		} else {
		   // found only one secondary server
		   currentAMServer = amServerHostsAvailable;
		   amServerHostsAvailable = "";
		}

		// Trim Both leading and trailing spaces
		std::string::size_type startpos = currentAMServer.find_first_not_of(" \t"); // Find the first character position after excluding leading blank spaces
		std::string::size_type endpos = currentAMServer.find_last_not_of(" \t"); // Find the first character position from reverse af
		// if all spaces or empty return an empty string
		if(( std::string::npos == startpos ) || ( std::string::npos == endpos)){
			currentAMServer = "";
		} else {
			currentAMServer = currentAMServer.substr( startpos, endpos - startpos + 1);
		}

		ERS_DEBUG(1, "Next AM server from the client configuration is: " << currentAMServer);
	}
	return currentAMServer;
}

// get the AM server port number
unsigned long ClientConfig::getAMServerPort() const{
	return atol(getConstantValueFromMap(PARAMETER_AM_SERVER_PORT).c_str());
}

// get the AM server host name
bool ClientConfig::isAMAuthorizationEnabled() const {
	std::string amEnabledValue = getConstantValueFromMap(PARAMETER_AM_AUTHORIZATION);

	if (amEnabledValue.length() > 0){
		// convert it to lower case
		for(std::string::iterator sit = amEnabledValue.begin(); sit != amEnabledValue.end(); sit++){
			*sit = std::tolower(*sit);
		}
	} else{
		amEnabledValue = VALUE_AM_AUTHORIZATION_DEFAULT;
		ERS_DEBUG(1, "Parameter [" << PARAMETER_AM_AUTHORIZATION << "] defaults to "
		 			 << "[" << VALUE_AM_AUTHORIZATION_DEFAULT << "]");
	}

	//check its lowercase value
	return (amEnabledValue == VALUE_AM_AUTHORIZATION_ENABLED);
}

unsigned long ClientConfig::getAMClientConnectTimeout() const{
	return atol(getConstantValueFromMap(PARAMETER_AM_CLIENT_CONNECT_TIMEOUT).c_str());
}

unsigned long ClientConfig::getAMClientReceiveTimeout() const{
	return atol(getConstantValueFromMap(PARAMETER_AM_CLIENT_RECV_TIMEOUT).c_str());
}

// check if the AM client APi should use the secondary AM servers
bool ClientConfig::useSecondaryAMServers() const {
	std::string amSecAMSValue = getConstantValueFromMap(PARAMETER_AM_CLIENT_USE_SECONDARY_SERVERS);

	if (amSecAMSValue.length() > 0){
		// convert it to lower case
		for(std::string::iterator sit = amSecAMSValue.begin(); sit != amSecAMSValue.end(); sit++){
			*sit = std::tolower(*sit);
		}
	} else{
		amSecAMSValue = VALUE_AM_CLIENT_USE_SECONDARY_SERVERS_DEFAULT;
		ERS_DEBUG(1, "Parameter [" << PARAMETER_AM_CLIENT_USE_SECONDARY_SERVERS << "] defaults to "
		 			 << "[" << VALUE_AM_CLIENT_USE_SECONDARY_SERVERS_DEFAULT << "]");
	}

	//check its lowercase value
	return (amSecAMSValue == VALUE_AM_CLIENT_USE_SECONDARY_SERVERS);
}


}

}
