package autoprescaleeditor;

import java.util.List;
import triggerdb.Entities.SMT.SuperMasterTable;
import triggerdb.Entities.L1.L1Master;
import triggerdb.Entities.L1.L1Menu;
import triggerdb.Entities.L1.L1Item;
import triggerdb.Entities.L1.L1Prescale;
import triggerdb.Entities.L1Links.L1TM_PS;
import triggerdb.Connections.InitInfo;
import triggerdb.Connections.ConnectionManager;
import triggertool.L1Records.CutToValueConverter;

import java.util.HashMap;

import java.sql.SQLException;

import Igui.IguiLogger;
import java.sql.PreparedStatement;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Map;


public class L1PrescaleSaver {

    /**
     * Request validity states
     */
    @SuppressWarnings("PackageVisibleInnerClass")
    enum RequestState
    { VALID, VALID_HIGH_RATE, CANCELED_REQUEST, INVALID_SIZE, EMPTY_REQUEST, FORBIDDEN_ITEM }
    
    
    private enum CorrectionMode
    { ASIS, ROUNDUP, MAXROUNDUP, DISABLE }
    
    
    /**
     * list of items that are allowed to change prescales, should be moved to OKS
     */
    final List<String> items_allowed_to_change;

    // confirmatio dialog required
    private boolean confirm = false;
    
    /**
     * Connection information and manager (manager is singleton)
     */
    private static InitInfo initinfo = new InitInfo();
    private static ConnectionManager dbconn_mgr = ConnectionManager.getInstance();
    private IguiInterface iguiInterface = null;

    /**
     * Current prescale set 
     */    
    private L1Prescale l1ps = null;
    public L1Prescale getL1ps() { return l1ps; }

    /**
     * Current L1 menu 
     */    
    private L1Menu current_l1_menu = null;
    private int loaded_smk = 0;

    /**
     * Map from CTP_ID (0..255) to Item name 
     */    
    private Map<Integer,String> item_name_by_ctp_id = new HashMap<Integer,String>();
    public Map<Integer, String> getItem_name_by_ctp_id() {
        return Collections.unmodifiableMap(item_name_by_ctp_id);
    }

    private void abortOnException(Exception ex)
    {
        IguiLogger.error("Aborting on exception",ex);
        ex.printStackTrace();
        System.exit(-1);
    }

    
    public void setIguiInterface(IguiInterface iguiint)
    { iguiInterface = iguiint; }

    
    /**
     * constructor
     **/
    public L1PrescaleSaver(final String triggerdb,
                           final List<String> items,
                           boolean confirm)
    {
        this.items_allowed_to_change = items;
        this.confirm = confirm;
        
        // initialize the connection information
        try {
            initinfo.setConnectionFromCommandLine(triggerdb);
            initinfo.setSystemUsername("AutoPrescaleCTP");
        }
        catch (Exception ex) { abortOnException(ex); }
        IguiLogger.info("The connection parameters are:\n" + initinfo.toString());
    }


    /**
     * configureWithSMK, reads the menu
     **/
    public void configureWithSMK(final int smk) throws SQLException
    {
        // during tests only
        if ( smk==0 ) { return; }
        if( smk != this.loaded_smk ) {
            IguiLogger.info("Reading L1 menu for SMK " + smk);
            this.readL1Menu(smk);
            this.loaded_smk = smk;
        } else {
            IguiLogger.info("L1 menu for SMK " + smk + " already read.");
        }
    }

    
    public L1Prescale readL1PSS(final int l1_psk) throws SQLException
    {
        IguiLogger.info("Read L1 pss for L1 PSK="+l1_psk);
        boolean open_here = connectToDB();
        L1Prescale tmpl1ps = new L1Prescale(l1_psk);
        if(open_here) { disconnectFromDB(); }
        return tmpl1ps;
    }    


    public int saveNewL1PrescaleSet() throws SQLException
    {
        // upload and attach to menu
        boolean open_here = connectToDB();
        int new_l1_psk = upload_pss(this.l1ps);
        if(new_l1_psk>0) { attach_psk_to_l1menu(new_l1_psk); }
        if(open_here) { disconnectFromDB(); }

        return new_l1_psk>=0?new_l1_psk:-1;
    }


    private int upload_pss(L1Prescale l1ps) throws SQLException
    {
        IguiLogger.info("Saving new L1 pss");
        boolean open_here = connectToDB();
        String current_name = l1ps.get_name();
        if( ! current_name.startsWith("AUTO_") ) {
            l1ps.set_name("AUTO_" + current_name);
        }
        IguiLogger.info("Saving L1 PSS with comment\n:" + l1ps.get_comment());
        String comment = l1ps.get_comment();
        final int newl1psk = l1ps.save();
        l1ps.set_comment(comment);
        saveComment(newl1psk, comment);
        if(open_here) { disconnectFromDB(); }
        IguiLogger.info(" -> L1 PSK = " + newl1psk);
        return newl1psk;
    }

    
    private void attach_psk_to_l1menu(final int l1_psk) throws SQLException
    {
        // make a new link from current l1 menu to l1 prescale set l1_psk
        IguiLogger.info("Attaching L1 pss "+l1_psk + " to current menu (SMK="+this.loaded_smk+", L1M_ID=" + this.current_l1_menu.get_id()+")");
        boolean open_here = connectToDB();
        L1TM_PS newlink = new L1TM_PS();
        newlink.set_menu_id(this.current_l1_menu.get_id());
        newlink.set_prescale_set_id(l1_psk);
        newlink.set_used(true); // true == hide
        int id = newlink.save();
        IguiLogger.info("Menu -> PSS link ID=" + id);
        if(open_here) { disconnectFromDB(); }
    }

    
    /**
     * saves a new L1 prescale set after checking that the changes
     * with respect to the current set are allowed
     *
     * @param final int current_l1_psk: current L1 prescale key
     * @param final Vector<L1PSNumber> new_pss: current L1 prescales 
     *
     * @returns new l1 PSK, or -1 in case of error
     **/
    RequestState validateNewPrescales(final int current_l1_psk, 
                                      final L1Prescale new_pss) throws SQLException
    {
        IguiLogger.info("Validate new prescales request compared to L1 pss " +  current_l1_psk );

        this.l1ps = this.readL1PSS(current_l1_psk);

        return validation(new_pss); // validation modifies this.l1ps
    }

    
    /**
     * Validates a new L1 prescale set with respect to the current set
     * are allowed. Modifies the current prescale set in memory,
     * rather then creating a new one.
     *
     * @param final Vector<L1PSNumber> new_pss: current L1 prescales 
     *
     * @returns boolean true if the changes are valid
     **/
    private RequestState validation(final L1Prescale new_pss)
    {
        IguiLogger.info("Checking request ...");

        final List<String> forbidden_changes = new ArrayList<>();
        final List<String> itemnames = new ArrayList<>();
        final List<Integer> ctp_ids = new ArrayList<>();
        final L1Prescale prescale_suggestions = new L1Prescale();

        for(int ctp_id = 0; ctp_id < L1Prescale.LENGTH; ctp_id++) {
            Integer current_ps = this.l1ps.get_val(ctp_id+1);
            Integer new_ps = new_pss.get_val(ctp_id+1);
	    //            Integer new_ps = new_pss.get_val(ctp_id);

            String item_name = item_name_by_ctp_id.get(ctp_id);

	    // Skip unassigned item names
	    if(item_name==null)
		{
		    continue;
		}

	    if(item_name.equals("L1_J30_EMPTY")) {
		IguiLogger.info("buggy"+item_name + ": " + current_ps + " to new " + new_ps);
	    }
	    
            if(current_ps.equals(new_ps)) {
                continue;
            } // no change for this item

	    //            String item_name = item_name_by_ctp_id.get(ctp_id+1);
            //String item_name = item_name_by_ctp_id.get(ctp_id);

            IguiLogger.info(item_name + ": " + current_ps + " to new " + new_ps);
            //System.out.println(item_name + ": " + current_ps + " to new " + new_ps);

	    // This protection isn't needed anymore, though we might want to think about it
            //if( ! items_allowed_to_change.contains(item_name) ) { 
                // item is not allowed to change (the CTP should make sure this does not happen)
                //forbidden_changes.add(item_name);
                //continue;
            //}

            itemnames.add(item_name);
            ctp_ids.add(ctp_id);
            prescale_suggestions.set_val(ctp_id+1,new_ps);

            //ctp_ids.add(ctp_id);
            //prescale_suggestions.set_val(ctp_id,new_ps);
        }

        if(itemnames.isEmpty()) {
            return RequestState.EMPTY_REQUEST;
        }

        if(!forbidden_changes.isEmpty()) {
            return RequestState.FORBIDDEN_ITEM;
        }

        IguiLogger.info("Request change of prescales: " + prescale_suggestions.toString() );
	// Circumvent rounding for now
        // CorrectionMode mode = CorrectionMode.MAXROUNDUP;
	CorrectionMode mode = CorrectionMode.ASIS;
        L1Prescale new_prescales = new L1Prescale();

        switch(mode) {
        case ASIS:
            //new_prescales=(L1Prescale) prescale_suggestions.clone();
	    new_prescales=prescale_suggestions;
            break;
        case ROUNDUP:
            for(int ctpid=0; ctpid< L1Prescale.LENGTH; ++ctpid) {
                //new_prescales.set_val(ctpid,roundup(prescale_suggestions.get_val(ctpid), 5));
                new_prescales.set_val(ctpid+1,roundup(prescale_suggestions.get_val(ctpid+1), 5));
            }
            break;
        case MAXROUNDUP:
            Integer max = null;
            for(int ctpid=0; ctpid<L1Prescale.LENGTH;++ctpid) {
                //Integer roundedps = roundup(prescale_suggestions.get_val(ctpid), 5);
                Integer roundedps = roundup(prescale_suggestions.get_val(ctpid+1), 5);
                if(max==null) {
                    max = roundedps;
                } else {
                    if(roundedps>max) { max = roundedps; }
                }
            }
            for(int ctpid=0; ctpid< L1Prescale.LENGTH; ++ctpid) {
                new_prescales.set_val(ctpid+1, max);
                //new_prescales.set_val(ctpid, max);
            }
            break;
        case DISABLE:
            for(int ctpid=0; ctpid< L1Prescale.LENGTH; ++ctpid) {
                int ps = prescale_suggestions.get_val(ctpid+1);
                new_prescales.set_val(ctpid+1, ps>0?-1*ps:ps);
                //int ps = prescale_suggestions.get_val(ctpid);
                //new_prescales.set_val(ctpid, ps>0?-1*ps:ps);
            }
            break;
        }
        
        if(this.confirm) {
            boolean answer_yes = true;
            IguiLogger.info("Request user confirmation for TriggerDB upload.");
            String windowTitle = "CTP auto prescale request";
            String question = "Noise burst of calorimeter cosmic items " + itemnames.toString() + ". Do you want to suppress these hot triggers ?";
            boolean standalone = (iguiInterface==null);
            if(standalone) {
               UploadConfirmationDialog dialog = UploadConfirmationDialog.showDialog(null, itemnames, prescale_suggestions);
               answer_yes = dialog.confirmed();
            } else {
               answer_yes = iguiInterface.userDialog(question, windowTitle);
            }
            if(!answer_yes) {
                IguiLogger.info("User requests cancelation.");
                return RequestState.CANCELED_REQUEST; // user cancel
            }
        }

        boolean high_rate_warning = false;

        RequestState request = RequestState.VALID;
        IguiLogger.info("Upload of following prescales to TriggerDB: " + new_prescales.toString() );
        StringBuilder comment = new StringBuilder("generated by auto prescaler\n");
        comment.append("Compared to key ").append(this.l1ps.get_id()).append(" (").append(this.l1ps.get_name()).append(")\nprescaled items:");
        //for(int i=0; i<L1Prescale.LENGTH; i++) {
        for(int i=0; i<ctp_ids.size(); i++) {
            Integer ctp_id = ctp_ids.get(i);
            Integer new_ps = new_prescales.get_val(ctp_id + 1);
            Integer old_ps = this.l1ps.get_val(ctp_id + 1);
            this.l1ps.set_val(ctp_id+1, new_ps);
            if(new_ps>300000) {
                request = RequestState.VALID_HIGH_RATE;
            }
            comment.append("\n").append(itemnames.get(i)).append(" (").append(CutToValueConverter.calculatePrescaleFromCut(old_ps)).append("->").append(CutToValueConverter.calculatePrescaleFromCut(new_ps)).append("), ");
        }
        
        this.l1ps.set_comment(comment.toString());
        this.l1ps.set_id(-1);
        return request;
    }

    
    /**
     * returns a prescale value that is greater or equal the requested one and a power of "step" 
     */
    private Integer roundup(final Integer exact_ps,
                               final int step)
    {
        int finalps = 1;
        while(finalps < exact_ps) { finalps *= step; }
        return finalps;
    }


    /*************************
     * Reading the L1 menu from the trigger DB. Retains the menu
     * object and fills a map ctp -> item_name
     *
     * @param: smk the key describing the menu
     *************************/
    private void readL1Menu(final int smk) throws SQLException
    {
        boolean open_here = connectToDB();
        SuperMasterTable smt = new SuperMasterTable(smk);
        L1Master l1m = new L1Master((int)smt.get_l1_master_table_id());
        this.current_l1_menu = new L1Menu(l1m.get_menu_id());
        List<L1Item> items = this.current_l1_menu.getItems();
        if(open_here) { disconnectFromDB(); }
        item_name_by_ctp_id.clear();
        for(L1Item it: items) {
            item_name_by_ctp_id.put(it.get_ctp_id(), it.get_name());
        }
    }

    
    
    private void saveComment(final int l1psk, String comment)
    {
        
        System.out.println("Entered saveComment '" + comment + "'");
        if (comment==null || l1psk<=0) { return; }
        String query = "UPDATE L1_PRESCALE_SET SET L1PS_COMMENT=? WHERE L1PS_ID=?";
        query = L1PrescaleSaver.dbconn_mgr.fix_schema_name(query);
        try {
            PreparedStatement ps = L1PrescaleSaver.dbconn_mgr.getConnection().prepareStatement(query);
            ps.setString(1, comment.isEmpty()?"~":comment);
            ps.setInt(2, l1psk);
            System.out.println("SaveComment Query '" + query + "'");
            ps.executeUpdate();
            ps.close();
        } catch (SQLException e) {
            e.printStackTrace();
        }
    }
    
    

    private boolean connectToDB()
    {
        if (L1PrescaleSaver.dbconn_mgr.isConnectionOpen()) { return false; }
        try {
            IguiLogger.info("Connecting to db");            
            L1PrescaleSaver.dbconn_mgr.connect(L1PrescaleSaver.initinfo);
        }
        catch (SQLException ex) {
            abortOnException(ex);
        }
        return true;
    }


    private void disconnectFromDB()
    {
        if (!L1PrescaleSaver.dbconn_mgr.isConnectionOpen()) { return; }
        try {
            L1PrescaleSaver.dbconn_mgr.disconnect();
        }
        catch (SQLException ex) {
            abortOnException(ex);
        }
    }


    static public L1Prescale asL1PSNumberVector(final Integer[] new_pss)
    {
        L1Prescale pss = new L1Prescale();
        int ctpid=0;
        for( Integer ps: new_pss) { 
            pss.set_val(ctpid++,ps); 
            if(ctpid==L1Prescale.LENGTH) break;
        }
        return pss;
    }


}
