#include <AutoPrescaler/rulechecker/CombinedRule.h>
#include <AutoPrescaler/rulechecker/RuleChecker.h>
#include <AutoPrescaler/rulechecker/DataAbstraction.h>

//------------------------------------------------------------------------------
using namespace L1CT;
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
int CombinedRule::powerOf(float targetPrescale)
{
  int prescale = 1;
  while(prescale < targetPrescale)
    prescale*=m_roundingFactor;

  return prescale;
}


bool CombinedRule::request()
{

  // We need to consider all possibilities for state transitions here
  bool makeRequest = false;
  
  // Temporary book-keeping for the current state
  bool tempHot = false;
  bool tempBeam = false;
  bool tempCold = false;
  bool tempNotBeam = false;

  // Begin by extracting information from the rate history
  // Is the entire history above/below the maxRate?
  bool allAbove = true;
  // entire rate history with original prescale below max rate
  bool allBelow = true;
  // entire rate history with prescale of one below max rate
  bool allBelowU = true;
  
  float largestRateP = 0;
  float largestRateO = 0;
  float largestRateU = 0;

  // Extract current prescale
  float currentPS = rulechecker().dataModel().abstraction().getL1Prescale(m_itemName);
  float targetPrescale = currentPS;
  int requestedPS = int(targetPrescale+0.5);

  ERS_DEBUG(3,"Preparing for action: ");
ERS_DEBUG(3,"Current status:");
  if(m_isHot)
    ERS_DEBUG(3,"We are already in hot item mode");  
  if(m_isBeam)
    ERS_DEBUG(3,"We are already in beam mode");    
  if(!m_isBeam && !m_isHot)
    ERS_DEBUG(3,"We are in a neutral state");

  // Look through the prescaled rates: are we hot?
  for( auto rate : m_ratesP )
    {
      if(rate.first > largestRateP)
	largestRateP = rate.first;
      
      if(rate.first <= m_maxRate)
        allAbove = false;
    }

  // Look through the rates, using the original prescale: would we still be hot?
  for( auto rate : m_ratesO )
    {
      if(rate.first > largestRateO)
	largestRateO = rate.first;

      if(rate.first > m_maxRate)
	allBelow = false;
    }

  for(auto rate: m_ratesU)
    {
      if(rate.first > largestRateU)
	largestRateU = rate.first;
      
      if(rate.first > m_maxRate)
	allBelowU = false;
    }

  ERS_DEBUG(3,"Max allowed rate is: " << m_maxRate);
  ERS_DEBUG(3,"Target rate is: " << m_targetRate);
  ERS_DEBUG(3,"Largest rate is (prescaled): " << largestRateP);
  ERS_DEBUG(3,"Largest rate is (original prescale): " << largestRateO);
  ERS_DEBUG(3,"Largest rate is (un-prescaled): " << largestRateU);

  if(allAbove)
    ERS_DEBUG(3,"We are going hotter");
  if(allBelow)
    ERS_DEBUG(3,"We are cold");
  if(allBelowU)
    ERS_DEBUG(3,"We can unprescale the item entirely");
  
  bool applyHotPrescale = false;
  // Cold -> Hot
  if(!m_isHot && m_timeElapsed && allAbove)
    applyHotPrescale = true;

  // Hot -> Hotter (only if a full integration loop has gone by)
  // Need a finer-tuned logic here...
  boost::posix_time::time_duration diff = boost::posix_time::microsec_clock::local_time() - m_lastHot;
  float timeElapsed = diff.total_microseconds()/1000000.;

  if(m_isHot && m_timeElapsed && allAbove && timeElapsed > m_integrationTime)
    applyHotPrescale = true;
  
  // Any transition that requires pre-scaling
  if(applyHotPrescale)
    {
      ERS_LOG("Going hotter:");

      // Now, we look for the target rate, keeping in mind the safety factor
      targetPrescale *= (largestRateP / float(m_targetRate));
      ERS_LOG("Target pre-scale is: " << targetPrescale);

      // Round to nearest power of 2, but can eventually use something much smarter...
      requestedPS = powerOf(targetPrescale);

      // This is useful to check we've been hot long enough
      m_lastHot = boost::posix_time::microsec_clock::local_time();      
      tempHot = true;
      makeRequest = true;
      //m_newPS = false;
    }
  
  // Hot -> Cold
  if(m_isHot && allBelow)
    {
      // We are not in beamMode: relax entirely by reverting pre-scale - simple as that
      if(!m_isBeam)
	{
	  ERS_LOG("Going cold, back to original PS: " << m_originalPS);
	  
	  requestedPS = int(m_originalPS+0.5);
	  tempCold = true;
	  makeRequest = true;
	}
      // We are in a beamMode
      else
	{
	  // We need to check that we can handle a PS of 1
	  // Otherwise, we stay in hot item mode!
	  if(allBelowU)
	    {
	      ERS_LOG("Going cold, back to beamMode PS of 1");

	      requestedPS = 1;
	      tempCold = true;
	      makeRequest = true;
	    }
	}
    }

  // Hot -> Still Hot: nothing happens - we carry on living happily

  // Now handle the beamMode cases (!!!)
  std::string beamWord = rulechecker().dataModel().abstraction().getBeamMode();

  bool goodMode = false;
  ERS_DEBUG(3,"Current Beam Mode: " << beamWord);

  for(unsigned int i=0; i< m_beamModes.size(); i++)
    {
      if(beamWord == m_beamModes.at(i))
	{
	  goodMode = true;
	  ERS_DEBUG(3,"Beam Mode matches requirements");
	}
    }

  // Entering beamMode
  if(!m_isBeam && goodMode)
    {
      // We unprescale the item - but we need to be careful
      // If the item is currently hot, there's nothing we can do.  Just wait
      // for the next iteration
      bool currentlyHot = m_isHot || tempHot;
      
      // If we are not hot, we can un-prescale the item, but...
      if(!currentlyHot)
	{
	  // Can we handle the rate?
	  if(largestRateU > m_maxRate)
	    {
	      // If we can't handle the rate, we design a safe 'beamPrescale'
	      // though we don't declare ourselves hot (that could lead to disasters)
	      // This safe beamPrescale will remain there for the entire period

	      // Now, we look for the target rate, keeping in mind the safety factor
	      targetPrescale = (largestRateU / float(m_targetRate));
	      ERS_LOG("Beam Mode: Unprescale with: " << targetPrescale );	      

	      // Round to nearest power of 2, but can eventually use something much smarter...
	      requestedPS = powerOf(targetPrescale);
	      makeRequest = true;
	      tempBeam = true;
	    }
	  else
	    {
	      // Phew, finally a simple case.  Unprescaaaale!
	      ERS_LOG("Beam Mode: Unprescale fully");
	      requestedPS = 1;
	      tempBeam = true;
	      makeRequest = true;
	    }
	}
    }
  
  // Leaving beamMode
  if(m_isBeam && !goodMode)
    {
      // Return to original prescale
      // No real need for safety here
      ERS_LOG("Leaving Beam Mode: Reverting to original prescale of: " << m_originalPS);

      requestedPS = int(m_originalPS+0.5);
      tempCold = true;
      tempNotBeam = true;
      makeRequest = true;
    }
  
  // Request menu update, no matter what (this is a bit much, but it works fine)
  if(makeRequest)
    {
      ERS_LOG("Making a request for: " << requestedPS);
      rulechecker().requestDraft(m_itemName, (unsigned int)(requestedPS));
    } else
    ERS_DEBUG(3,"No requests made..."); 
  ERS_DEBUG(3,"Updating current state...");
  // We are done: trigger state transition
  if(tempHot) m_isHot = true;
  if(tempCold) m_isHot = false;

  if(tempBeam) m_isBeam = true;
  if(tempNotBeam) m_isBeam = false;

  return true;

}

//------------------------------------------------------------------------------
bool CombinedRule::update()
{
  ERS_DEBUG(3,"Entering CombinedRule::update(): ");

  // Check if the monitored item is present: probably want a smarter protection
  if(!rulechecker().dataModel().abstraction().checkItem(m_itemName))
    {
      ERS_DEBUG(3,"Item is not in the abstraction: " << m_itemName);
      return false;
    }

  // If item is disabled, we just idle
  if(!rulechecker().dataModel().abstraction().isEnabled(m_itemName))
    {
      ERS_DEBUG(3,"Item is disabled: " << m_itemName);
      reset();
      return false;
    }
  ERS_DEBUG(3,"Updating for item: " << m_itemName);

  float currentPS = rulechecker().dataModel().abstraction().getL1Prescale(m_itemName);

  // Store the starting prescale
  if(!m_initialized)
    {
      m_originalPS = currentPS;
      ERS_DEBUG(3,"Storing original pre-scale value: " << m_originalPS);
      m_initialized = true;
    }  

  // Some ugly placeholder logic here
  float currentRateP = rulechecker().dataModel().abstraction().getL1Rate(m_itemName, L1CT::DataAbstraction::L1Item::Type::TAP);
  float currentRateU = rulechecker().dataModel().abstraction().getL1Rate(m_itemName, L1CT::DataAbstraction::L1Item::Type::TBP);
  ERS_DEBUG(3,"Current Rate after Prescale: " << currentRateP);
  ERS_DEBUG(3,"Current Rate before Prescale: " << currentRateU);

  // Book-keeping via FIFO
  boost::posix_time::ptime thisTime = boost::posix_time::microsec_clock::local_time();
  std::pair<float, boost::posix_time::ptime> currentEntryU(currentRateU, thisTime);
  std::pair<float, boost::posix_time::ptime> currentEntryO(currentRateU/m_originalPS, thisTime);
  std::pair<float, boost::posix_time::ptime> currentEntryP(currentRateP, thisTime);
  std::pair<int, boost::posix_time::ptime> currentEntryPS(int(currentPS+0.5), thisTime);

  m_ratesU.push_back(currentEntryU);
  m_ratesO.push_back(currentEntryO);
  m_ratesP.push_back(currentEntryP);
  m_prescales.push_back(currentEntryPS);

  ERS_DEBUG(3,"Filling with 3 rate entries, U, O, P: " << currentEntryU.first << ", " << currentEntryO.first << ", " << currentEntryP.first);
  ERS_DEBUG(3,"Filling with PS: " << currentEntryPS.first);
  ERS_DEBUG(3,"Number of entries in prescale array: " << m_prescales.size());

  // remove entries where time elapsed was too long ago
  if(m_prescales.size() > 0)
    {
      while(1)
	{
	  boost::posix_time::time_duration diff = m_ratesU.back().second - m_ratesU.front().second;
	  float timeElapsed = diff.total_seconds();
	  ERS_DEBUG(3,"Time delta is: " << timeElapsed);
	  if(timeElapsed > m_integrationTime)
	    {
	      // First time we remove an entry, we are assured that the
	      // minimum integration time has been fulfilled
	      ERS_DEBUG(3,"First entry is removed: ");
	      m_timeElapsed = true;
	      m_ratesU.pop_front();
	      m_ratesP.pop_front();
	      m_ratesO.pop_front();
	      m_prescales.pop_front();
	    }
	  else
	    break;
	}
    }

  /// Detect whether we've got a pre-scale change in there
  // Wipe the store if that's the case - should make things cleaner now
  ERS_DEBUG(3,"Number of entries in prescale array after timeWindow: " << m_prescales.size());  

  if(m_prescales.size() > 1)
    {

      int lastPS = m_prescales.back().first;
      ERS_DEBUG(3,"Prescale value to be matched: " << lastPS);
      while(1)
	{
	  ERS_DEBUG(3,"Size before removal: " << m_prescales.size());

	  // If we only have one entry, it makes no sense to try to wipe the array
	  if(m_prescales.size() < 2)
	    break;
	  ERS_DEBUG(3,"Checking for mismatch... ");
	  
	  // Clear entries that had been recorded with a different pre-scale
	  if(m_prescales.front().first != lastPS)
	    {
	      ERS_DEBUG(3,"Mismatch: " << lastPS << " " << m_prescales.front().first);

	      m_ratesU.pop_front();
	      m_ratesP.pop_front();
	      m_ratesO.pop_front();
	      m_prescales.pop_front();
	      m_timeElapsed = false;
	      ERS_DEBUG(3,"Size after removal: " << m_prescales.size());
	      
	    } else {
	    ERS_DEBUG(3,"No mismatch... ");
	    break;
	  }
	}
    }
  // We're done updating the internal stores
  
  return true;
}
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
std::vector<std::string> CombinedRule::needsItems()
{
  std::vector<std::string> theItems;
  theItems.push_back(m_itemName); 
  
  return theItems;
}
//------------------------------------------------------------------------------


