//------------------------------------------------------------------------------
#include <AutoPrescaler/rulechecker/RuleChecker.h>
//------------------------------------------------------------------------------
#include <AutoPrescaler/ipc/PrescalerIpcCommander.h>
//------------------------------------------------------------------------------
#include "rc/RCStateInfo.h"
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
using namespace L1CT;
//------------------------------------------------------------------------------


RuleChecker::RuleChecker()
{
  ERS_INFO("Constructing Rulechecker");

  // Subscribe to DataModel
  m_dataModel.add(*this);
  ERS_INFO("Subscribed RuleChecker to DataModel");

  // ensure MenuDraft knows we exist
  m_menuDraft.setRuleChecker(*this);
  
  // Empty Initialization
  initialize();
}

void RuleChecker::addRule(CombinedRule& rulein)
{
  ERS_INFO("Adding rule to RuleChecker");
  m_rules.push_back(rulein);
  for(auto item: rulein.needsItems())
    {
      ERS_LOG("Item needed: " << item);
      m_dataModel.abstraction().addL1Item(item);
    }
}


bool RuleChecker::checkPSKHistory()
{
  // Check if the histories are aligned - if not, call a global reset
  std::vector<CtpControllerL1PSKey> publishedPSKHistory = dataModel().abstraction().online().triggerPSKHistory().items;
  
  // if published history is shorter: surely wrong
  if(publishedPSKHistory.size() < m_lastPSKHistory.size())
    return false;
  
  for(unsigned int i=0; i<m_lastPSKHistory.size(); i++)
    {
      if(publishedPSKHistory.at(i).L1PrescaleKey != m_lastPSKHistory.at(i).L1PrescaleKey )
	return false;
    }
  return true;
}

void RuleChecker::globalReset()
{
  // Reset everything
  m_menuDraft.reset();
  for(unsigned int i=0; i<m_rules.size(); i++)
    {
      m_rules[i].reset();
    }
}


bool RuleChecker::update(ISCallbackInfo * isc)
{
  // Read IS observables
  ERS_DEBUG(3,"RuleChecker::update() -- entered");

  // We only want to start putting in work if this is a rate update
  // Otherwise, we'll start doing some weird stuff...
  const std::string name  = isc->name();
  if(name != "L1CT.CTPCORE.Instantaneous.TriggerRates" && name !="LHC.BeamMode")
    {
      ERS_DEBUG(3,"RuleChecker::update() -- rates or beamMode were not updated - carry on.");
      return true;
    }      

  // Additionally, we don't want to do any work if we are not in the RUNNING state
  RCStateInfo rcStateInfo;
  auto & dict = ISServer::ATLAS.dictionary();
  dict.getValue("RunCtrl.RootController", rcStateInfo);
  if (rcStateInfo.state != "RUNNING") {
    ERS_DEBUG(3,"RuleChecker::update() -- we are not running - carry on.");    
    return true;
  }


  if(!checkPSKHistory())
    {
      ERS_DEBUG(3,"RuleChecker::update() -- calling a reset");
      globalReset();
    }  

  // Hope this is a copy :D
  m_lastPSKHistory = dataModel().abstraction().online().triggerPSKHistory().items;
  
  // Based on the published prescale key, define whether to wipe out the draft
  // completely or just push out an update
  ERS_DEBUG(3,"RuleChecker::update() -- clearing draft");
  m_menuDraft.reset();
  
  // Will need some prescale key book-keeping in RuleChecker, at the very least...

  // Update conditions and actions - do it through the rules for the moment
  ERS_DEBUG(3,"RuleChecker::update() -- rules size " << m_rules.size());
  for(unsigned int i=0; i<m_rules.size(); i++)
    {
      // Update the conditions, the actions, the state, and call the necessary action
      ERS_DEBUG(3,"RuleChecker::update() -- updating rule " << i);
      if(m_rules[i].update())
	{
	  ERS_DEBUG(3,"RuleChecker::update() -- successful rule update: calling request");
	  m_rules[i].request();
	} else {
	ERS_DEBUG(3,"RuleChecker::update() -- rule failed...");
      }
    }
  ERS_DEBUG(3,"RuleChecker::update() -- checking menu draft");

  // Check the menu draft for any diff
  MenuDraft::Draft_t draftDifference = m_menuDraft.check();

  // If needed, request a PS-change
  if(draftDifference.size() > 0)
    return requestPS(draftDifference);
  
  return true;
}

bool RuleChecker::requestPS(MenuDraft::Draft_t diffRequest)
{
  ERS_DEBUG(3,"RuleChecker::requestPS() -- entered");
  PrescalerIpcCommander prescaler;
  return prescaler.requestPrescale(dataModel().abstraction().getL1PSK(), diffRequest);

}
