"""Module includes classes and methods for histogram definitions.
"""

import functools
from typing import Optional


class HistoDef:
    """Definition of a histogram.

    Attributes
    ----------
    name : str
        Histogram name, this is a BST internal short name like "posX".
    title : str or None
        Title to use for newly-made histograms, may be None.
    is_bcid : bool
        True if histogram is per-BCID histogram.
    is_rs_res : bool
        True if histogram should be taken from runsummary folder when --useRSRes
        option is used
    """

    def __init__(self, name: str, title: Optional[str] = None,
                 is_bcid: bool = False, is_rs_res: bool = False):
        self.name = name
        self.title = title
        self.is_bcid = is_bcid
        self.is_rs_res = is_rs_res


def make_hdefs_vtx(do_bcid: bool):
    """Create the set of histogram definitions for vertex-based method.

    Parameters
    ----------
    do_bcid : bool
        If True then return BCID histograms also

    Returns
    -------
    hdefs : `list` [ `HistoDef` ]
        List of HistoDef objects.
    """

    hdefs = []
    for axis in "XYZ":
        hdefs += [
            HistoDef(f"pos{axis}", title=f"Acc. Vertex {axis}"),
            HistoDef(f"delta{axis}", title=f"Acc. Vertex {axis} - Nominal"),
            HistoDef(f"pos{axis}_vs_ntrks", title=f"Vertex {axis} vs NTrks"),
            HistoDef(f"split{axis}_vs_ntrks", title=f"Vertex D{axis} vs NTrks", is_rs_res=True),
        ]

    for axis in "XY":
        hdefs += [
            HistoDef(f"tilt{axis}", title=f"Acc. Vertex {axis} vs Vertex Z")
        ]

    if do_bcid:
        for axis in "XYZ":
            hdefs += [
                HistoDef(f"delta{axis}_bcid", title=f"Vertex {axis} - Nominal vs BCID", is_bcid=True),
                HistoDef(f"pos{axis}_bcid", title=f"Vertex {axis} vs BCID", is_bcid=True),
            ]

    return hdefs


def make_hdefs_trk(do_bcid: bool):
    """Create the set of histogram definitions for track-based method.

    Parameters
    ----------
    doBCID : bool
        If True then return BCID histograms also

    Returns
    -------
    hdefs : `list` [ `HistoDef` ]
        List of HistoDef objects.
    """
    hdefs = [
        HistoDef("beam_ls", title="Beam LS matrix"),
        HistoDef("llpoly", title="TRack LL Polynomial coefficients"),
    ]
    if do_bcid:
        hdefs += [
            HistoDef("beam_ls_bcid", title="Per-BCID Beam LS matrices", is_bcid=True)
        ]
    return hdefs


@functools.lru_cache(maxsize=3)
def names_vtx():
    """Complete set of histogram names used by vertex method"""
    return set(hdef.name for hdef in make_hdefs_vtx(do_bcid=True))


@functools.lru_cache(maxsize=3)
def names_trk():
    """Complete set of histogram names used by track method"""
    return set(hdef.name for hdef in make_hdefs_trk(do_bcid=True))
