"""Module includes classes and methods for histogram publishing in OH server.
"""

import logging
import os
import time

from ipc import IPCPartition
from oh import OHRootProvider

# use module name as logger name (remove top-level package name)
_LOG = logging.getLogger(__name__.partition('.')[2])


class OHPublisher:
    """Helper class for publishing histograms in OH server.

    Parameters
    ----------
    partition : str
        Parition name, if empty $TDAQ_PARTITION is used.
    server : str
        OH server name.
    provider : str
        Provider name.
    folder : str
        Folder name.
    """
    def __init__(self, partition, server, provider, folder):
        self._partition = partition
        if not self._partition:
            self._partition = os.environ.get("TDAQ_PARTITION")
        self._server = server
        self._provider = provider
        self._folder = folder

        self._ohprovider = None

    @property
    def do_publish(self):
        return self._partition and self._server and self._provider and self._folder

    @property
    def provider(self):
        if self.do_publish and self._ohprovider is None:
            _LOG.info("will publish histograms: parition=%r server=%r provider=%r",
                      self._partition, self._server, self._provider)
            part = IPCPartition(self._partition)

            # try to register OH provider, there may be few BSTs running
            # at the same time trying to do the same, if there is a conflict
            # retry for max. of 3 seconds
            try_until = time.time() + 3
            while self._ohprovider is None:
                try:
                    self._ohprovider = OHRootProvider(part, self._server,
                                                      self._provider, None)
                except Exception as exc:
                    # ERS catches C++ exceptions (ers::Issue) and translates
                    # them into Python exceptions but it's not possible to
                    # use Python exception type directly, have to work based
                    # on class name
                    if exc.__class__.__name__ == "ProviderAlreadyExist" and time.time() < try_until:
                        # wait a bit longer
                        _LOG.info("Provider name already in use, will wait until released")
                        time.sleep(0.1)
                    else:
                        raise

        return self._ohprovider

    def finish(self):
        """Finish publishing, deactivate provider.
        """
        self._ohprovider = None

    def publish(self, histo, name, tag):
        """Publish one histogram.

        Parameter
        ---------
        histo : TH1
            ROOT histogram instance
        name : `str`
            Histogram name
        tag : `int`
            Tag/LB number
        """

        if not self.provider:
            # no publishing
            return

        path = self._folder.rstrip('/')
        path += "/"
        path += name
        _LOG.debug("publishing histogram with name %r", path)
        self.provider.publish(histo, path, tag)
