"""Methods for track-based fitting.
"""

import array
import ctypes
import logging

import math
try:
    import numpy as np
    from scipy.optimize import minimize
except ImportError:
    np = minimize = None

from .freaking_ROOT import ROOT
from . import model


# use module name as logger name (remove top-level package name)
_LOG = logging.getLogger(__name__.partition(".")[2])


class LSMatrix:
    """Representation of the matrix data produced by HLT for least-squares
    fitting of the beamspot.

    Attributes
    ----------
    matrix_elements : `array.array`
        Array of 16 elements representing 4x4 matrix for X, Y, Tx, Ty
    rhs : `array.array`
        Array of 4 elements representing 4x1 right hand side matrix
    n_tracks : `int`
        Number of tracks
    sum_d0_sq : `float`
        Sum of track d_0**2.
    sum_z0 : `float`
        Sum of track z0
    sum_z0_sq : `float
        Sum of track z0**2
    """
    def __init__(self, matrix_elements, rhs, sum_d0_sq, n_tracks, sum_z0, sum_z0_sq):
        self.matrix_elements = matrix_elements
        self.rhs = rhs
        self.sum_d0_sq = sum_d0_sq
        self.n_tracks = n_tracks
        self.sum_z0 = sum_z0
        self.sum_z0_sq = sum_z0_sq

    @classmethod
    def from_array(cls, data):
        """Unpack data from an HLT-produced packed data array (which is
        usually published as a histogram ).

        Parameters
        ----------
        data : `list` of `float`, or `array.array`
            Packed data array, 18 numbers.
        """
        A_x_x, A_x_y, A_y_y, A_x_tx, A_y_tx, A_tx_tx, A_x_ty, A_y_ty, A_tx_ty, A_ty_ty = data[:10]
        matrix_elements = array.array("d", [
            A_x_x, A_x_y, A_x_tx, A_x_ty,
            A_x_y, A_y_y, A_y_tx, A_y_ty,
            A_x_tx, A_y_tx, A_tx_tx, A_tx_ty,
            A_x_ty, A_y_ty, A_tx_ty, A_ty_ty
        ])
        self = cls(
            matrix_elements=matrix_elements,
            rhs=array.array("d", data[10:14]),
            sum_d0_sq=data[14],
            sum_z0=data[15],
            sum_z0_sq=data[16],
            n_tracks=int(data[17])
        )
        return self

    @classmethod
    def from_histo(cls, histo):
        """Unpack data from an HLT-produced histogram.

        Parameters
        ----------
        histo : `ROOT.TH1`
            Histogram with packed data.
        """
        data = [histo.GetBinContent(i) for i in range(1, 19)]
        return cls.from_array(data)


class LSMatrixBCID:
    """Representation of the matrix data produced by HLT for least-squares
    fitting of the beamspot for one BCID.

    Attributes
    ----------
    matrix_elements : `array.array`
        Array of 4 elements representing 2x2 matrix for X, Y
    rhs : `array.array`
        Array of 2 elements representing 2x1 right hand side matrix
    n_tracks : `int`
        Number of tracks
    sum_z0 : `float`
        Sum of track z0
    sum_z0_sq : `float
        Sum of track z0**2
    """
    def __init__(self, matrix_elements, rhs, n_tracks, sum_z0, sum_z0_sq):
        self.matrix_elements = matrix_elements
        self.rhs = rhs
        self.n_tracks = n_tracks
        self.sum_z0 = sum_z0
        self.sum_z0_sq = sum_z0_sq

    @classmethod
    def from_array(cls, data):
        """Unpack data from an HLT-produced packed data array (which is
        usually published as a histogram ).

        Parameters
        ----------
        data : `list` of `float`, or `array.array`
            Packed data array, 8 numbers.
        """
        A_x_x, A_x_y, A_y_y = data[:3]
        self = cls(
            matrix_elements=array.array("d", [A_x_x, A_x_y, A_x_y, A_y_y]),
            rhs=array.array("d", data[3:5]),
            sum_z0=data[5],
            sum_z0_sq=data[6],
            n_tracks=int(data[7])
        )
        return self

    @classmethod
    def from_histo(cls, histo, bcid):
        """Unpack data from an HLT-produced histogram.

        Parameters
        ----------
        histo : `ROOT.TH1`
            Histogram with packed data.
        bcid : `int`
            BCID number.
        """
        data = [histo.GetBinContent(1 + bcid * 8 + i) for i in range(8)]
        return cls.from_array(data)


class PolyLL:
    """Representation of polynomial LL approximation.

    Parameters
    ----------
    coeff_sq : list of float
        Coefficients of polinomial approximating exponent, 90 numbers,
        order is the same as powers generated by `_make_ll_powers`.
    coeff_ln : list of float
        Coefficients of polinomial approximating logarithm, 6 numbers.
    n_tracks : `int`
        Number of tracks
    sum_z0 : `float`
        Sum of track z0
    sum_z0_sq : `float
        Sum of track z0**2
    sum_beam_size : `float`
        Beam size used for Taylor expansion multiplied by number of tracks.
    """
    def __init__(self, coeff_sq, coeff_ln, n_tracks, sum_z0, sum_z0_sq, sum_beam_size):
        self._coeff_sq = coeff_sq
        self._coeff_ln = coeff_ln
        self.n_tracks = n_tracks
        self.sum_z0 = sum_z0
        self.sum_z0_sq = sum_z0_sq
        self.sum_beam_size = sum_beam_size

    @classmethod
    def from_array(cls, data):
        """Unpack data from an HLT-produced packed data array (which is
        usually published as a histogram ).

        Parameters
        ----------
        data : `list` of `float`, or `array.array`
            Packed data array, 100 numbers.
        """
        self = cls(
            coeff_sq=data[:90],
            coeff_ln=data[90:96],
            sum_z0=data[96],
            sum_z0_sq=data[97],
            n_tracks=int(data[98]),
            sum_beam_size=data[99]
        )
        return self

    @classmethod
    def from_histo(cls, histo):
        """Unpack data from an HLT-produced histogram.

        Parameters
        ----------
        histo : `ROOT.TH1`
            Histogram with packed data.
        """
        nbins = histo.GetNbinsX()
        if nbins != 100:
            raise ValueError(f"Unexpected number of bins: {nbins}")
            return
        data = [histo.GetBinContent(i) for i in range(1, 1 + nbins)]
        return cls.from_array(data)

    def __call__(self, x):
        """Calculate LL value (negative).

        Parameters
        ----------
        x : tuple of float
            Parameters, 6-tuple (B_x, B_y, width_x, width_y, tilt_x, tilt_y)

        Returns
        -------
        ll : float
            Negative LL value.
        """
        ll = []
        for coefficients in (self._coeff_sq, self._coeff_ln):
            for coeff, powers in zip(coefficients, _ll_powers):
                monom = coeff
                for var, pow in zip(x, powers):
                    monom *= var**pow
                ll += [monom]
        ll.sort(key=abs)
        return -sum(ll)

    def chi2(self, x):
        """Calculate chi2 value.

        Parameters
        ----------
        x : tuple of float
            Parameters, 6-tuple (B_x, B_y, width_x, width_y, tilt_x, tilt_y)

        Returns
        -------
        chi2 : float
            Chi**2 value.
        """
        ll = []
        # chi2 is a (negative) 2*LL without log part
        for coefficients in (self._coeff_sq, ):
            for coeff, powers in zip(coefficients, _ll_powers):
                monom = coeff
                for var, pow in zip(x, powers):
                    monom *= var**pow
                ll += [monom]
        ll.sort(key=abs)
        return -2 * sum(ll)

    @staticmethod
    def _monom_deriv(x, powers, i):
        """Calculate derivative of a monomial.

        Parameters
        ----------
        x : tuple of float
            Parameters, 6-tuple.
        powers : tuple of int
            Exponents for each parameter, 6 numbers.
        i : int
            Index of a parameter for differentiation

        Returns
        -------
        diff : float
            Derivative of negative LL value.
        """
        monom = 1
        for k in range(len(x)):
            pow = powers[k]
            if k == i:
                monom *= pow
                if pow > 0:
                    pow -= 1
            monom *= x[k]**pow
        return monom

    @staticmethod
    def _monom_deriv2(x, powers, i, j):
        """Calculate second derivative of a monomial.

        Parameters
        ----------
        x : tuple of float
            Parameters, 6-tuple.
        powers : tuple of int
            Exponents for each parameter, 6 numbers.
        i, j : int
            Indices of parameters for differentiation

        Returns
        -------
        diff : float
            Second derivative of negative LL value.
        """
        monom = 1
        for k in range(len(x)):
            pow = powers[k]
            if k == i:
                monom *= pow
                if pow > 0:
                    pow -= 1
            if k == j:
                monom *= pow
                if pow > 0:
                    pow -= 1
            monom *= x[k]**pow
        return monom

    def jacobian(self, x):
        """Calculate Jacobian of LL value (negative).

        Parameters
        ----------
        x : tuple of float
            Parameters, 6-tuple (B_x, B_y, width_x, width_y, tilt_x, tilt_y)

        Returns
        -------
        ll : `array.array`
            Jacobian of negative LL.
        """
        jacs = []
        for i in range(len(x)):
            jac = []
            for coefficients in (self._coeff_sq, self._coeff_ln):
                for coeff, powers in zip(coefficients, _ll_powers):
                    monom = coeff * self._monom_deriv(x, powers, i)
                    jac += [monom]
            jac.sort(key=abs)
            jacs.append(-sum(jac))
        return array.array("d", jacs)

    def hessian(self, x):
        """Calculate Hessian matrix of LL value (negative).

        Parameters
        ----------
        x : tuple of float
            Parameters, 6-tuple (B_x, B_y, width_x, width_y, tilt_x, tilt_y)

        Returns
        -------
        ll : `array.array`
            Hessian of negative LL (6x6 matrix, 36 elements)
        """
        npar = len(x)
        hess = array.array("d", [0.] * npar**2)
        for i in range(len(x)):
            for j in range(len(x)):
                h = []
                for coefficients in (self._coeff_sq, self._coeff_ln):
                    for coeff, powers in zip(coefficients, _ll_powers):
                        monom = coeff * self._monom_deriv2(x, powers, i, j)
                        h += [monom]
                h.sort(key=abs)
                hess[i * npar + j] = -sum(h)
        return hess

    def minuit_fcn(self, npar, grad, fval, xval, iflag):
        """Minimization function for use with Minuit

        Parameters
        ----------
        npar : ctypes.c_int
            Number of parameters
        grad : array of float
            Gradients, to be filled if iflag == 2.
        fval : ctypes.c_double
            Returned value
        xval :
            Parameters
        iflag : int
        """
        par = [xval[i] for i in range(npar.value)]
        if iflag == 2:
            jac = self.jacobian(par)
            for i in range(npar.value):
                grad[i] = jac[i]

        fval.value = self(par)


class MinuitFitter:
    """Class that uses ROOT to fit track data.
    """
    def __init__(self):
        pass

    def fit_total_data(self, data):
        """Beamspot fit of the least squares BCID-blind data.

        Parameters
        ----------
        data : `LSMatrix`
            Data for the fit.

        Returns
        -------
        snap : `model.BSSnapshot`
            Beamspot with fitted values.
        """
        status = model.STATUS_OK

        A = ROOT.TMatrixDSym(4, data.matrix_elements)
        B = ROOT.TMatrixD(4, 1, data.rhs)

        d0_2 = data.sum_d0_sq
        nTracks = data.n_tracks
        z0 = data.sum_z0 / nTracks
        z0_rms = math.sqrt(data.sum_z0_sq / nTracks - z0 * z0)

        _LOG.debug("# of tracks = %d", nTracks)

        cov = ROOT.TMatrixDSym(A)
        cov.Invert()
        _LOG.debug("covariance matrix: %s", cov)

        sol = ROOT.TMatrixD(4, 1)
        sol.Mult(cov, B)
        _LOG.debug("solution: %-10.8f %-10.8f %-10.8f %-10.8f",
                   sol[0, 0], sol[1, 0], sol[2, 0], sol[3, 0])

        err = [math.sqrt(cov[i, i]) for i in range(4)]
        _LOG.debug("errors: %s", err)

        _LOG.debug("z0: %f rms: %f", z0, z0_rms)

        chi2m = ROOT.TMatrixD(1, 1)
        chi2m.TMult(sol, B)
        chi2 = d0_2 - chi2m[0, 0]
        _LOG.debug("chi2: %f", chi2)

        try:
            mean_z, mean_z_err, sigma_z, sigma_z_err = \
                _gaussian_estimate(data.n_tracks, data.sum_z0, data.sum_z0_sq)
        except Exception:
            status = model.STATUS_NOT_CONVERGED
            mean_z, mean_z_err, sigma_z, sigma_z_err = 0., 0., 0., 0.

        snap = model.BSSnapshot(
            status=status,
            posX=sol[0, 0],
            posY=sol[1, 0],
            tiltX=sol[2, 0],
            tiltY=sol[3, 0],
            posXErr=err[0],
            posYErr=err[1],
            tiltXErr=err[2],
            tiltYErr=err[3],
            posZ=mean_z,
            posZErr=mean_z_err,
            sigmaZ=sigma_z,
            sigmaZErr=sigma_z_err,
        )
        return snap

    def fit_bcid_data(self, data, bcid):
        """
        Parameters
        ----------
        data : `LSMatrixBCID`
            Data for the fit.
        bcid : `int`
            Bunch crossing ID

        Returns
        -------
        snap : `model.BSSnapshot`
            Beamspot with fitted values.
        """
        if data.n_tracks == 0:
            _LOG.debug("No data for BCID=%d", bcid)
            return model.BSSnapshot()

        status = model.STATUS_OK

        A = ROOT.TMatrixDSym(2, array.array("d", data.matrix_elements))
        B = ROOT.TMatrixD(2, 1, array.array("d", data.rhs))

        A_inv = ROOT.TMatrixDSym(A)
        A_inv.Invert()
        try:
            sol = ROOT.TMatrixD(2, 1)
            sol.Mult(A_inv, B)
            cov = A_inv
            err = [math.sqrt(cov[i, i]) for i in range(2)]
        except Exception:
            status = model.STATUS_NOT_CONVERGED
            sol = [0., 0.]
            err = [0., 0.]

        try:
            mean_z, mean_z_err, sigma_z, sigma_z_err = \
                _gaussian_estimate(data.n_tracks, data.sum_z0, data.sum_z0_sq)
        except Exception:
            status = model.STATUS_NOT_CONVERGED
            mean_z, mean_z_err, sigma_z, sigma_z_err = 0., 0., 0., 0.

        _LOG.debug("BCID %d fit X: %s += %s", bcid, sol[0, 0], err[0])
        _LOG.debug("BCID %d fit Y: %s += %s", bcid, sol[1, 0], err[1])
        _LOG.debug("BCID %d fit Z: %s += %s, sigma: %s +- %s", bcid, mean_z, mean_z_err, sigma_z, sigma_z_err)

        snap = model.BSSnapshot(
            status=status,
            bcid=bcid,
            posX=sol[0, 0],
            posY=sol[1, 0],
            posXErr=err[0],
            posYErr=err[1],
            posZ=mean_z,
            posZErr=mean_z_err,
            sigmaZ=sigma_z,
            sigmaZErr=sigma_z_err,
        )
        return snap

    def ll_fit_data(self, ll_poly):
        """Estimate beamspot from log likelihood fit.

        Parameters
        ----------
        ll_poly : `PolyLL`
            Representation of log likelihood function.

        Returns
        -------
        snap : `model.BSSnapshot`
            Beamspot with fitted values.
        """
        n_tracks = ll_poly.n_tracks
        if n_tracks == 0:
            _LOG.debug("No tracks in LL histogram")
            return model.BSSnapshot()

        status = model.STATUS_OK

        beam_size = ll_poly.sum_beam_size / n_tracks
        _LOG.debug("LL: n_tracks: %d beam_size: %g", n_tracks, beam_size)

        minuit = ROOT.TMinuit(6)
        fcn = ll_poly.minuit_fcn  # stupid PyROOT needs local reference to a method
        minuit.SetFCN(fcn)
        minuit.SetPrintLevel(-1)

        minuit.DefineParameter(0, "B_x", 0., .1, -5., 5.)
        minuit.DefineParameter(1, "B_y", 0., .1, -5., 5.)
        minuit.DefineParameter(2, "tilt_x", 0., 1e-5, -0.001, 0.001)
        minuit.DefineParameter(3, "tilt_y", 0., 1e-5, -0.001, 0.001)
        minuit.DefineParameter(4, "omega_x", 1e-4, 1e-5, -beam_size**2, 0.1)
        minuit.DefineParameter(5, "omega_y", 1e-4, 1e-5, -beam_size**2, 0.1)

        # ERR should be set to 0.5 for log likelihood
        minuit.SetErrorDef(0.5)
        # force gradients from FCN
        ierflg = ctypes.c_int(0)
        arglist = array.array("d", [1.])
        arglist[0] = 1
        minuit.mnexcm("SET GRAD", arglist, 1, ierflg)
        # do minimization
        stat = minuit.Migrad()
        _LOG.debug("fit status: %s", stat)

        value, value_err = ctypes.c_double(0.), ctypes.c_double(0.)
        params, errors = [], []
        for i in range(6):
            minuit.GetParameter(i, value, value_err)
            params.append(value.value)
            errors.append(value_err.value)

        _LOG.debug("      fun: %s", ll_poly(params))
        _LOG.debug("      jac: %s", ll_poly.jacobian(params))
        _LOG.debug("        x: %s", params)
        _LOG.debug("      err: %s", errors)

        mean_x, mean_y, tilt_x, tilt_y = params[:4]
        sigma_x, sigma_y = [math.sqrt(beam_size**2 + params[i]) for i in (4, 5)]
        hess_array = ll_poly.hessian(params)
        hessian = ROOT.TMatrixDSym(len(params), hess_array)
        hess_inv = hessian.Invert()
        mean_x_err, mean_y_err, tilt_x_err, tilt_y_err = [math.sqrt(hess_inv(i, i)) for i in range(4)]
        sigma_x_err, sigma_y_err = [math.sqrt(hess_inv(i, i)) / math.sqrt(beam_size**2 + params[i]) / 2
                                    for i in (4, 5)]
        _LOG.debug("mean_x: %g +- %g", mean_x, mean_x_err)
        _LOG.debug("mean_y: %g +- %g", mean_y, mean_y_err)
        _LOG.debug("tilt_x: %g +- %g", tilt_x, tilt_x_err)
        _LOG.debug("tilt_y: %g +- %g", tilt_y, tilt_y_err)
        _LOG.debug("width_x: %g +- %g", sigma_x, sigma_x_err)
        _LOG.debug("width_y: %g +- %g", sigma_y, sigma_y_err)
        _LOG.debug("Chi^2: %f", ll_poly.chi2(params))

        try:
            mean_z, mean_z_err, sigma_z, sigma_z_err = \
                _gaussian_estimate(ll_poly.n_tracks, ll_poly.sum_z0, ll_poly.sum_z0_sq)
        except Exception:
            status = model.STATUS_NOT_CONVERGED
            mean_z, mean_z_err, sigma_z, sigma_z_err = 0., 0., 0., 0.

        snap = model.BSSnapshot(
            status=status,
            posX=mean_x,
            posXErr=mean_x_err,
            posY=mean_y,
            posYErr=mean_y_err,
            posZ=mean_z,
            posZErr=mean_z_err,
            sigmaX=sigma_x,
            sigmaXErr=sigma_x_err,
            sigmaY=sigma_y,
            sigmaYErr=sigma_y_err,
            sigmaZ=sigma_z,
            sigmaZErr=sigma_z_err,
            tiltX=tilt_x,
            tiltXErr=tilt_x_err,
            tiltY=tilt_y,
            tiltYErr=tilt_y_err,
        )
        return snap


class ScipyFitter:
    """Class that uses scipy to fit track data.
    """
    def __init__(self):
        pass

    def fit_total_data(self, data):
        """Beamspot fit of the least squares BCID-blind data.

        Parameters
        ----------
        data : `LSMatrix`
            Data for the fit.

        Returns
        -------
        snap : `model.BSSnapshot`
            Beamspot with fitted values.
        """
        status = model.STATUS_OK

        B = np.array(data.rhs)
        d0_2 = data.sum_d0_sq
        nTracks = data.n_tracks
        z0 = data.sum_z0 / nTracks
        z0_rms = math.sqrt(data.sum_z0_sq / nTracks - z0 * z0)

        _LOG.debug("# of tracks = %d", nTracks)
        A = np.array(data.matrix_elements)
        A = A.reshape((4, 4))

        cov = np.linalg.inv(A)
        _LOG.debug("covariance matrix: %s", cov)

        sol = np.linalg.solve(A, B)
        _LOG.debug("solution: %s", sol)

        err = np.array([math.sqrt(cov[i, i]) for i in range(4)])
        _LOG.debug("errors: %s", err)

        _LOG.debug("z0: %f rms: %f", z0, z0_rms)

        chi2 = d0_2 - np.dot(sol, B)
        _LOG.debug("chi2: %f", chi2)

        try:
            mean_z, mean_z_err, sigma_z, sigma_z_err = \
                _gaussian_estimate(data.n_tracks, data.sum_z0, data.sum_z0_sq)
        except Exception:
            status = model.STATUS_NOT_CONVERGED
            mean_z, mean_z_err, sigma_z, sigma_z_err = 0., 0., 0., 0.

        snap = model.BSSnapshot(
            status=status,
            posX=sol[0],
            posY=sol[1],
            tiltX=sol[2],
            tiltY=sol[3],
            posXErr=err[0],
            posYErr=err[1],
            tiltXErr=err[2],
            tiltYErr=err[3],
            posZ=mean_z,
            posZErr=mean_z_err,
            sigmaZ=sigma_z,
            sigmaZErr=sigma_z_err,
        )
        return snap

    def fit_bcid_data(self, data, bcid):
        """
        Parameters
        ----------
        data : `LSMatrixBCID`
            Data for the fit.
        bcid : `int`
            Bunch crossing ID

        Returns
        -------
        snap : `model.BSSnapshot`
            Beamspot with fitted values.
        """
        if data.n_tracks == 0:
            return model.BSSnapshot()

        status = model.STATUS_OK

        A = np.array(data.matrix_elements)
        A = A.reshape((2, 2))
        B = np.array(data.rhs)

        try:
            sol = np.linalg.solve(A, B)
            cov = np.linalg.inv(A)
            err = np.array([math.sqrt(cov[i, i]) for i in range(2)])
        except Exception:
            status = model.STATUS_NOT_CONVERGED
            sol = [0., 0.]
            err = [0., 0.]

        try:
            mean_z, mean_z_err, sigma_z, sigma_z_err = \
                _gaussian_estimate(data.n_tracks, data.sum_z0, data.sum_z0_sq)
        except Exception:
            status = model.STATUS_NOT_CONVERGED
            mean_z, mean_z_err, sigma_z, sigma_z_err = 0., 0., 0., 0.

        snap = model.BSSnapshot(
            status=status,
            bcid=bcid,
            posX=sol[0],
            posY=sol[1],
            posXErr=err[0],
            posYErr=err[1],
            posZ=mean_z,
            posZErr=mean_z_err,
            sigmaZ=sigma_z,
            sigmaZErr=sigma_z_err,
        )
        return snap

    def ll_fit_data(self, ll_poly):
        """Estimate beamspot from log likelihood fit.

        Parameters
        ----------
        ll_poly : `PolyLL`
            Representation of log likelihood function.

        Returns
        -------
        snap : `model.BSSnapshot`
            Beamspot with fitted values.
        """
        n_tracks = ll_poly.n_tracks
        if n_tracks == 0:
            _LOG.debug("No tracks in LL histogram")
            return model.BSSnapshot()

        status = model.STATUS_OK

        beam_size = ll_poly.sum_beam_size / n_tracks
        _LOG.debug("LL: n_tracks: %d beam_size: %g", n_tracks, beam_size)

        x0 = np.array([0., 0., 0., 0., 0., 0.])
        optr = minimize(
            ll_poly, x0, jac=ll_poly.jacobian, method="L-BFGS-B",
            bounds=[
                (-5, 5), (-5, 5),
                (-0.001, 0.001), (-0.001, 0.001),
                (-beam_size**2, 0.1), (-beam_size**2, 0.1)
            ],
            options={"gtol": .01, "ftol": 1e-30})

        _LOG.debug("optr: %s", optr)

        mean_x, mean_y, tilt_x, tilt_y = optr.x[:4]
        sigma_x, sigma_y = [math.sqrt(beam_size**2 + optr.x[i]) for i in (4, 5)]
        hessian_array = ll_poly.hessian(optr.x)
        hessian = np.array(hessian_array)
        hessian = hessian.reshape((len(optr.x), len(optr.x)))
        hess_inv = np.linalg.inv(hessian)
        mean_x_err, mean_y_err, tilt_x_err, tilt_y_err = [math.sqrt(hess_inv[i, i]) for i in range(4)]
        sigma_x_err, sigma_y_err = [math.sqrt(hess_inv[i, i]) / math.sqrt(beam_size**2 + optr.x[i]) / 2
                                    for i in (4, 5)]

        _LOG.debug("mean_x: %g +- %g", mean_x, mean_x_err)
        _LOG.debug("mean_y: %g +- %g", mean_y, mean_y_err)
        _LOG.debug("tilt_x: %g +- %g", tilt_x, tilt_x_err)
        _LOG.debug("tilt_y: %g +- %g", tilt_y, tilt_y_err)
        _LOG.debug("width_x: %g +- %g", sigma_x, sigma_x_err)
        _LOG.debug("width_y: %g +- %g", sigma_y, sigma_y_err)
        _LOG.debug("Chi^2: %f", ll_poly.chi2(optr.x))

        try:
            mean_z, mean_z_err, sigma_z, sigma_z_err = _gaussian_estimate(
                ll_poly.n_tracks, ll_poly.sum_z0, ll_poly.sum_z0_sq
            )
        except Exception:
            status = model.STATUS_NOT_CONVERGED
            mean_z, mean_z_err, sigma_z, sigma_z_err = 0., 0., 0., 0.

        snap = model.BSSnapshot(
            status=status,
            posX=mean_x,
            posXErr=mean_x_err,
            posY=mean_y,
            posYErr=mean_y_err,
            posZ=mean_z,
            posZErr=mean_z_err,
            sigmaX=sigma_x,
            sigmaXErr=sigma_x_err,
            sigmaY=sigma_y,
            sigmaYErr=sigma_y_err,
            sigmaZ=sigma_z,
            sigmaZErr=sigma_z_err,
            tiltX=tilt_x,
            tiltXErr=tilt_x_err,
            tiltY=tilt_y,
            tiltYErr=tilt_y_err,
        )
        return snap


def _make_ll_powers():
    """Make a list of exponents for the variables in monomials.

    The order is exactly the same as in TrigT2BeamSpot/T2TrackBSLLPoly class.
    Total number of monomials is 90.

    Yields
    ------
    p_Bx : int
        Exponent for B_x
    p_By : int
        Exponent for B_y
    p_ty : int
        Exponent for T_x (tilt)
    p_ty : int
        Exponent for T_y (tilt)
    p_omx : int
        Exponent for omega_x (sigma_x**2 - beam_size**2), beam_size is
        a constant
    p_omy : int
        Exponent for omega_x (sigma_y**2 - beam_size**2), beam_size is
        a constant
    """
    for p_Bx in range(3):
        for p_By in range(3):
            for p_tx in range(3):
                for p_ty in range(3):
                    if p_Bx + p_By + p_tx + p_ty > 2:
                        continue
                    for p_omx in range(3):
                        for p_omy in range(3):
                            if p_omx + p_omy > 2:
                                continue
                            yield (p_Bx, p_By, p_tx, p_ty, p_omx, p_omy)


_ll_powers = list(_make_ll_powers())


def _gaussian_estimate(n, sum_v, sum_v_sq):
    """Estimate parameters of gaussian distribution.

    Parameters
    ----------
    n : int
        Number of tracks
    sum_v : float
        Sum of observed values
    sum_v_sq : float
        Sum of observed values**2

    Returns
    -------
    mean : float
        Estimate of mean
    mean_err : float
        Error of mean estimate
    sigma : float
        Estimate of sigma
    sigma_err : float
        Error of sigma estimate
    """
    if n == 1:
        return sum, 0., 0., 0.

    mean = sum_v / n
    # unbiased estimate for sample variance
    s_2 = sum_v_sq / (n - 1) - sum_v**2 / n / (n - 1)
    mean_err = math.sqrt(s_2 / n)
    sigma = math.sqrt(s_2)
    # variance of sample variance
    # https://en.wikipedia.org/wiki/Variance#Distribution_of_the_sample_variance
    var_s_2 = 2 * s_2**2 / (n - 1)
    sigma_err = math.sqrt(var_s_2 / s_2) / 2
    _LOG.debug("standard mean: %g +- %g", mean, mean_err)
    _LOG.debug("standard sigma: %g +- %g", sigma, sigma_err)

    return mean, mean_err, sigma, sigma_err
