''' Library for write only (publish) python wrappers for the CTP.
    CTPTranslator writes in the normal way,
    DummyCTPTranslator prints to stdout instead of publishing
    ConservativeCTPTranslator sets status to 5 then publishes
'''

from contextlib import contextmanager
import copy
import logging
import os
import fcntl
from typing import Optional

from .. import model, os_utils
from . import base
import ers


# use module name as logger name (remove top-level package name)
_LOG = logging.getLogger(__name__.partition('.')[2])

_LOCKFILE = "/tmp/beamspot_ctp_serialize"


@contextmanager
def ctp_serialize():
    try:
        fd = os.open(_LOCKFILE, os.O_RDONLY | os.O_CREAT)
        _LOG.info("locking file %s", _LOCKFILE)
        fcntl.flock(fd, fcntl.LOCK_EX)
        yield
        _LOG.info("unlocking file %s", _LOCKFILE)
        fcntl.flock(fd, fcntl.LOCK_UN)
        os.close(fd)
    except OSError as exc:
        _LOG.error("Failed to open lock file %s: %s", _LOCKFILE, exc)
        yield


class CTPError(ers.Issue):
    """ERS Issue class for CTP-related errors"""

    def __init__(self, message, cause=None, **kw):
        ers.Issue.__init__(self, 'Failure executing command: ' + message,
                           kw,
                           cause if isinstance(cause, ers.Issue) else None)


class CTPTranslator(base.Translator):
    '''Class capable of executing a write to the CTP.
        It first writes a restricted set of fields
        to the beam spot Information Service then
        sends a special command to the CTP.
        Only works online.
    '''

    def __init__(self, **kwargs):
        pass

    def read(self) -> Optional[model.BeamSpotRepresentation]:
        # dosctring is inherited from base class
        raise NotImplementedError("CTPTranslator cannot read, it is write only!")

    def _processCommand(self, cmd):
        """Execute command.
        """
        _LOG.info("%s is executing %s", self.__class__.__name__, cmd)
        out, err, rc = os_utils.quickOSCallReturnCode(cmd)
        if rc != 0:
            _LOG.error("Failure executing command: %s", cmd)
            _LOG.error("  standard output: \n<<<\n%s\n>>>", out)
            _LOG.error("  standard error: \n<<<\n%s\n>>>", err)
            issue = CTPError(err.strip(), command=cmd)
            ers.error(issue)
            raise issue

    def write(self, rep: model.BeamSpotRepresentation):
        # dosctring is inherited from base class
        if isinstance(rep, model.BeamSpotRepresentation):
            if len(rep) != 1:
                raise ValueError('You must have exactly one snapshot to update the CTP')
            rep = rep[0]

        _LOG.info("Updating IS with this beam spot")
        rep.prettyLog(_LOG)
        # order specified in ../src/BeamSpotParameters.cxx

        args = [
            'status', 'posX', 'posY', 'posZ', 'sigmaX', 'sigmaY', 'sigmaZ',
            'tiltX', 'tiltY', 'sigmaXY', 'posXErr', 'posYErr', 'posZErr',
            'sigmaXErr', 'sigmaYErr', 'sigmaZErr', 'tiltXErr', 'tiltYErr', 'sigmaXYErr']

        bsISWriter = 'beamSpotISWriterApp'
        ctpUpdateCommand = 'ctpUpdateCommand'
        (out, _) = os_utils.quickOSCall(["which", bsISWriter], mustReturnZero=True)
        _LOG.info('Executable for publishing to IS: %s', out)
        (out, _) = os_utils.quickOSCall(["which", ctpUpdateCommand], mustReturnZero=True)
        _LOG.info('Executable for updating the CTP: %s', out)

        cmd = [bsISWriter]
        cmd += [str(getattr(rep, arg)) for arg in args]

        _LOG.debug("ISWriterApp wants values in this order: %s", ' '.join(args))
        with ctp_serialize():
            self._processCommand(cmd)

            _LOG.info("Telling the CTP to update!")
            self._processCommand([ctpUpdateCommand, '-i', 'beamspot'])


class DummyCTPTranslator(CTPTranslator):
    '''Debug-only CTP translator safe for dummies.
        this Translator prints to the screen only.
        Good for testing online or offline
    '''

    def _processCommand(self, cmd):
        _LOG.info("I would have processed this command, but instead"
                  " I'll just print it.")
        _LOG.info('$> %s', cmd)
        _LOG.info('')


class ConservativeCTPTranslator(CTPTranslator):
    ''' CTP translator writes an invalid (status 5) beam spot to the CTP
        Otherwise like CTPTranslator.
    '''

    def write(self, rep: model.BeamSpotRepresentation):
        # dosctring is inherited from base class

        # make deep copy to avoid modifying the original
        safeRep = copy.deepcopy(rep)
        safeStatus = 5
        for snap in safeRep:
            snap.status = safeStatus

        _LOG.info("Changing your beamspot status flag to %d", safeStatus)
        CTPTranslator.write(self, safeRep)
