'''Library for write only translators to Information Services'''

import logging
import os

from .. import os_utils
from . import base
import ers
from ispy import IS


_PYWRAPPER_LOADED = True
try:
    from libISWriterPyWrapper import writePylist2IS  # pylint: disable=no-name-in-module
except ImportError:
    _PYWRAPPER_LOADED = False


# use module name as logger name (remove top-level package name)
_LOG = logging.getLogger(__name__.partition('.')[2])


class ISError(ers.Issue):
    """ERS Issue class for IS errors"""

    def __init__(self, message, partition, is_object, cause=None):
        ers.Issue.__init__(self, message,
                           {'partition': partition, 'is_object': is_object},
                           cause if isinstance(cause, ers.Issue) else None)


monitoringISName = 'LiveMon'                    # pylint: disable=invalid-name
monitoringISNameByBunch = 'ByBunchMon'          # pylint: disable=invalid-name
monitoringISServerName = 'BeamSpot'             # pylint: disable=invalid-name
MAX_BUNCHES = 3564


class ISTranslator(base.Translator):
    """Implementation for write-only translator for data in IS.
    """

    def __init__(self, isObject="HLTParameters", isByBunchObject="ByBunchMon",
                 isServerName="BeamSpot",
                 doBCIDs=False, usePywrapper=False, isPartition=None, **kwargs):
        self.isObject = isObject
        self.isByBunchObject = isByBunchObject
        self.isServerName = isServerName
        self.isPartition = isPartition or os.environ['TDAQ_PARTITION']
        self.doBCIDs = doBCIDs

        self.usePywrapper = usePywrapper
        if self.usePywrapper and not _PYWRAPPER_LOADED:
            self.usePywrapper = False
            _LOG.warning("libISWriterPyWrapper could not be loaded."
                         " Reverting to ISWriterApp shell interface.")

    def read(self):
        # docstring is inherited from base class
        raise NotImplementedError("%s cannot read, it is write only!" % self.__class__.__name__)

    def write_ispy(self, rep):
        """Translate representation to external format.

        This method uses ispy wrapper.

        Parameters
        ----------
        rep : BeamSpotRepresentation
            In-memory beam spot representation.
        """

        args = ['status',
                'posX',
                'posY',
                'posZ',
                'sigmaX',
                'sigmaY',
                'sigmaZ',
                'tiltX',
                'tiltY',
                'sigmaXY',
                'posXErr',
                'posYErr',
                'posZErr',
                'sigmaXErr',
                'sigmaYErr',
                'sigmaZErr',
                'tiltXErr',
                'tiltYErr',
                'sigmaXYErr']
        # default values to use for bunches that are missing
        default_vals = [4] + [-999.] * (len(args) - 1)

        if not self.doBCIDs:

            # bcid-average case
            try:
                dstInfo = self.isServerName + '.' + self.isObject
                _LOG.info("Trying to write beam spot parameters to IS object \"%s!%s\"",
                          self.isPartition, dstInfo)
                isobj = IS.BeamSpotParameters(self.isPartition, dstInfo)
            except Exception as exc:
                ers.error(ISError('Error instantiating IS object',
                                  partition=self.isPartition,
                                  is_object=dstInfo,
                                  cause=exc))
                return

            try:
                snap = rep[0]
                _LOG.info("BeamSpotParameters: %s", snap)
                # fill all data members
                for arg in args:
                    setattr(isobj, arg, getattr(snap, arg))

                # publish
                isobj.checkin()
            except Exception as exc:
                ers.error(ISError('Error publishing IS object',
                                  partition=self.isPartition,
                                  is_object=dstInfo,
                                  cause=exc))
                return

        else:

            # per-bcid case
            try:
                dstInfo = self.isServerName + '.' + self.isByBunchObject
                _LOG.info("Trying to write per-bunch parameters to IS object \"%s\"", dstInfo)
                isobj = IS.BeamSpotParametersByBunch(self.isPartition, dstInfo)
            except Exception as exc:
                ers.error(ISError('Error instantiating IS object',
                                  partition=self.isPartition,
                                  is_object=dstInfo,
                                  cause=exc))
                return

            try:
                # index snapshots by bcid to make filling the gaps easier
                repIndex = dict((snap.bcid, snap) for snap in rep)

                isobj['paramsByBunch'].resize(MAX_BUNCHES)
                for bcid in range(0, MAX_BUNCHES):
                    snap = repIndex.get(bcid)
                    if snap:  # have info for this bunch
                        data = dict((arg, getattr(snap, arg)) for arg in args)
                    else:  # use default values
                        data = dict(zip(args, default_vals))
                    isobj['paramsByBunch'][bcid].set(data)

                # publish
                isobj.checkin()
            except Exception as exc:
                ers.error(ISError('Error publishing IS object',
                                  partition=self.isPartition,
                                  is_object=dstInfo,
                                  cause=exc))
                return

    def write_pywrapper(self, rep):
        """Translate representation to external format.

        This method uses writePylist2IS method from python wrapper.

        Parameters
        ----------
        rep : BeamSpotRepresentation
            In-memory beam spot representation.
        """
        args = ['status',
                'posX',
                'posY',
                'posZ',
                'sigmaX',
                'sigmaY',
                'sigmaZ',
                'tiltX',
                'tiltY',
                'sigmaXY',
                'posXErr',
                'posYErr',
                'posZErr',
                'sigmaXErr',
                'sigmaYErr',
                'sigmaZErr',
                'tiltXErr',
                'tiltYErr',
                'sigmaXYErr']
        # default values to use for bunches that are missing
        default_vals = [4] + [-999.] * (len(args) - 1)

        _LOG.debug("ISWriter wants values in this order: %s", ' '.join(args))

        if not self.doBCIDs:
            snap = rep[0]
            writePylist2IS([getattr(snap, arg) for arg in args],
                           self.isServerName + "." + self.isObject,
                           '')  # by bunch name is empty
            return

        # index snapshots by bcid to make filling the gaps easier
        repIndex = dict((snap.bcid, snap) for snap in rep)

        parslist = []
        for bcid in range(0, MAX_BUNCHES):
            snap = repIndex.get(bcid)
            if snap:  # have info for this bunch
                parslist += [getattr(snap, arg) for arg in args]
            else:  # use default values
                parslist += default_vals

        _LOG.info("%s writing...", self.__class__.__name__)
        writePylist2IS(parslist, "", self.isServerName + "." + self.isByBunchObject)

    def write(self, rep):
        # docstring is inherited from base class
        if len(rep) < 1:
            _LOG.info("No snapshots to upload!")
            return

        _LOG.info("Using ispy to write to IS.")
        self.write_ispy(rep)
        return

        _LOG.info("Updating IS with this beam spot")
        rep[0].prettyLog(_LOG)

        if self.usePywrapper:
            _LOG.info("Using the boost::python C++ wrapper to call ISWriter.")
            self.write_pywrapper(rep)
            return
        _LOG.info("Using the ISWriterApp shell interface.")

        args = ['status',
                'posX',
                'posY',
                'posZ',
                'sigmaX',
                'sigmaY',
                'sigmaZ',
                'tiltX',
                'tiltY',
                'sigmaXY',
                'posXErr',
                'posYErr',
                'posZErr',
                'sigmaXErr',
                'sigmaYErr',
                'sigmaZErr',
                'tiltXErr',
                'tiltYErr',
                'sigmaXYErr']
        # default values to use for bunches that are missing
        default_vals = [4] + [-999.] * (len(args) - 1)

        _LOG.debug("ISWriterApp wants values in this order: %s", ' '.join(args))

        cmd = ['beamSpotISWriterApp']
        if self.isObject != '':
            if self.doBCIDs:
                cmd += ['--by-bunch-name', self.isByBunchObject]
            else:
                cmd += ['--name', self.isObject]

        if not self.doBCIDs:
            snap = rep[0]
            cmd += [str(getattr(snap, arg)) for arg in args]
            os_utils.quickOSCall(cmd, mustReturnZero=True)
            return

        # index snapshots by bcid to make filling the gaps easier
        repIndex = dict((snap.bcid, snap) for snap in rep)

        for bcid in range(0, MAX_BUNCHES):
            snap = repIndex.get(bcid)
            if snap:  # have info for this bunch
                cmd += [str(getattr(snap, arg)) for arg in args]
            else:  # use default values
                cmd += [str(val) for val in default_vals]

        _LOG.info("%s writing...", self.__class__.__name__)
        os_utils.quickOSCall(cmd, mustReturnZero=True)


class ISMonTranslator(ISTranslator):
    """Translator which writes to monitoring IS server.
    """
    isObject = monitoringISName
    isByBunchObject = monitoringISNameByBunch
    isServerName = monitoringISServerName

    def __init__(self, **kwargs):
        kwargs = kwargs.copy()
        kwargs.update(usePywrapper=True)  # try to use the new boost::python wrapper
        ISTranslator.__init__(self,
                              self.isObject,
                              self.isByBunchObject,
                              self.isServerName,
                              **kwargs)


class ISCTPTranslator(ISTranslator):
    """Use this one to write HLTParameters.
    """
    isObject = ''  # use ISWriter default

    def __init__(self, **kwargs):
        kwargs = kwargs.copy()
        kwargs.update(doBCID=False)
        kwargs.update(usePywrapper=False)  # don't use the new boost:python wrapper
        ISTranslator.__init__(self, **kwargs)


class ISTagTranslator(ISTranslator):
    """Use this one to write IS object with name taken from a tag.
    """

    def __init__(self, tag, **kwargs):
        kwargs = kwargs.copy()
        # tag can be [partition!][server.]object
        part, sep, obj = tag.partition('!')
        if not obj:
            obj, part = part, None
        srv, sep, obj = obj.partition('.')
        if not obj:
            obj, srv = srv, "BeamSpot"
        kwargs.update(isObject=obj)
        kwargs.update(isServerName=srv)
        kwargs.update(isPartition=part)
        kwargs.update(doBCID=False)
        kwargs.update(usePywrapper=True)  # try to use the new boost::python wrapper
        ISTranslator.__init__(self, **kwargs)
