#!/usr/bin/env tdaq_python

from __future__ import print_function

import argparse
import sys

from coldpie import cool


def _fmt(x):
    if type(x) is float:
        return "{:.6g}".format(x)
    return str(x)


class CSVPrinter:

    def header(self, header):
        print(",".join(header))

    def row(self, values):
        print(",".join(values))

    def close(self):
        pass


class TablePrinter:

    def __init__(self):
        self._table = None

    def header(self, header):
        import prettytable
        self._table = prettytable.PrettyTable()
        for col in header:
            self._table.add_column(col, [], align="r")

    def row(self, values):
        self._table.add_row(values)

    def close(self):
        print(self._table)


def main():

    parser = argparse.ArgumentParser(description="Dump BeamSpot conditions from COOL")
    parser.add_argument("conn_str",
                        help="COOL connection string, e.g. COOLOFL_INDET/CONDBR2")
    parser.add_argument("folder",
                        help="COOL folder name, e.g. /Indet/Beampos")
    parser.add_argument("tag",
                        help="COOL tag name, e.g. IndetBeampos-RUN2-ES1-UPD2")
    parser.add_argument("-c", "--channel", type=int, default=0, metavar="NUMBER",
                        help="COOL channel number, def: $(default)s, use -1 to dump all channels.")
    parser.add_argument("-s", "--start", type=int, default=0, metavar="NUMBER",
                        help="Starting run number, default is all runs.")
    parser.add_argument("-e", "--end", type=int, default=1000000, metavar="NUMBER",
                        help="Ending run number, default is all runs.")
    parser.add_argument("-a", "--auto-fields", action="store_true", default=False,
                        help="Dump all fields instead of pre-defined ones, can be used on any folder.")
    parser.add_argument("-t", "--table", action="store_true", default=False,
                        help="Dump in table format, default is CSV.")
    args = parser.parse_args()

    printer = TablePrinter() if args.table else CSVPrinter()

    db_svc = cool.DatabaseSvcFactory.databaseService()
    db = db_svc.openDatabase(args.conn_str, True)
    folder = db.getFolder(args.folder)
    folder.setPrefetchAll(False)

    showChannel = args.channel < 0

    if args.auto_fields:
        fields = [field.name() for field in folder.payloadSpecification()]
    else:
        fields = ['posX', 'posY', 'posZ', 'sigmaX', 'sigmaY', 'sigmaZ', 'tiltX', 'tiltY']
        fields = ['status'] + fields + [f + 'Err' for f in fields]

    header = ['run', 'LB', 'toRun', 'toLB'] + fields
    if showChannel:
        header.insert(0, "channel")
    printer.header(header)

    since = args.start << 32
    until = args.end << 32
    if args.channel >= 0:
        channels = cool.ChannelSelection(args.channel)
    else:
        channels = cool.ChannelSelection.all()
    itr = folder.browseObjects(since, until, channels, args.tag)
    for obj in itr:
        since = divmod(obj.since(), 1 << 32)
        until = divmod(obj.until(), 1 << 32)
        payload = obj.payload()
        values = [since[0], since[1], until[0], until[1]] + \
            [payload[f] for f in fields]
        if showChannel:
            values.insert(0, obj.channelId())
        printer.row([_fmt(v) for v in values])

    printer.close()


if __name__ == "__main__":
    sys.exit(main())
