#!/usr/bin/env tdaq_python

"""Unit test for BeamSpotUtils.translators module
"""

import contextlib
import os
import tempfile
import unittest

from BeamSpotUtils import model, translators


@contextlib.contextmanager
def makeTmpFile():
    """Context manager for generating temporary file name.

    Temporary file is deleted on exiting context.
    """
    fd, tmpname = tempfile.mkstemp()
    os.close(fd)
    yield tmpname
    try:
        os.remove(tmpname)
    except OSError:
        pass


class _WriteError(Exception):
    pass


class _ReadError(Exception):
    pass


class _TestTran(translators.Translator):
    """Simple implementation of translator for testing"""

    def __init__(self):
        self.data = None

    def read(self):
        if self.data is None:
            return model.BeamSpotRepresentation()
        return self.data

    def write(self, rep):
        self.data = rep


class _TestWriteOnlyTran(translators.Translator):
    """Simple implementation of translator for testing"""

    def __init__(self, doException=False):
        self.doException = doException
        self.data = None

    def read(self):
        if self.doException:
            raise _ReadError("read error")
        else:
            return None

    def write(self, rep):
        self.data = rep


class _TestReadOnlyTran(translators.Translator):
    """Simple implementation of translator for testing"""

    def __init__(self, data=model.BeamSpotRepresentation()):
        self.data = data

    def read(self):
        return self.data

    def write(self, rep):
        raise _WriteError("write is not supported")


class TestTranslators(unittest.TestCase):
    """Test case for BeamSpotUtils.rootil module.
    """

    def setUp(self):
        os.environ['TDAQ_PARTITION'] = ""

    def tearDown(self):
        pass

    def test_multi(self):
        """Test for MultiTranslator class"""

        tr1 = _TestTran()
        tr_ro = _TestReadOnlyTran()
        tr_wo = _TestWriteOnlyTran()
        tr_wo_ex = _TestWriteOnlyTran(True)

        # instantiate empty translator and fill it
        tr = translators.MultiTranslator()
        self.assertEqual(len(tr._translators), 0)
        tr.append(tr1)
        tr.append(tr_ro)
        tr.append(tr_wo)
        self.assertEqual(len(tr._translators), 3)

        # instantiate with a list
        tr = translators.MultiTranslator([tr1, tr_ro, tr_wo])
        self.assertEqual(len(tr._translators), 3)

        # test write() method
        tr = translators.MultiTranslator([tr1, tr_wo])
        rep = model.BeamSpotRepresentation()
        tr.write(rep)
        self.assertIs(tr1.data, rep)
        self.assertIs(tr_wo.data, rep)

        # test write() method with one guy raising exception
        tr = translators.MultiTranslator([tr1, tr_ro, tr_wo])
        rep = model.BeamSpotRepresentation()
        with self.assertRaises(_WriteError):
            tr.write(rep)
        self.assertIs(tr1.data, rep)
        self.assertIsNot(tr_wo.data, rep)

        # test write() method
        rep = model.BeamSpotRepresentation()
        tr1.write(rep)
        tr = translators.MultiTranslator([tr1])
        self.assertIs(tr.read(), tr1.data)
        tr = translators.MultiTranslator([tr_ro, tr1])
        self.assertIs(tr.read(), tr_ro.data)
        self.assertIsNot(tr.read(), tr1.data)

        # test read() with one guy returning None
        tr = translators.MultiTranslator([tr_wo, tr1])
        self.assertIs(tr.read(), tr1.data)

        # test read() with one guy raising exception
        tr = translators.MultiTranslator([tr_wo_ex, tr1])
        self.assertIs(tr.read(), tr1.data)

    def test_factory(self):
        """Test for make_translator method"""

        # file name *.csv should return CSVTranslator
        tran = translators.make_translator("test.csv")
        self.assertIsNotNone(tran)
        self.assertIsInstance(tran, translators.CSVTranslator)

        # file name *.db,*.sql should return SQLiteTranslator
        tran = translators.make_translator("test.db")
        self.assertIsNotNone(tran)
        self.assertIsInstance(tran, translators.SQLiteTranslator)

        # Any other file name (with dot in it) should give Pickle
        tran = translators.make_translator("test.anything")
        self.assertIsNotNone(tran)
        self.assertIsInstance(tran, translators.PickleTranslator)

        # file name with tag
        tran = translators.make_translator("test.csv:sometag")
        self.assertIsNotNone(tran)
        self.assertIsInstance(tran, translators.CSVTranslator)

        # file name with keyword tag
        tran = translators.make_translator("test.csv", tag="gat")
        self.assertIsNotNone(tran)
        self.assertIsInstance(tran, translators.CSVTranslator)

        # test few known names that do not try to access database immediately
        tran = translators.make_translator('isctp')
        self.assertIsNotNone(tran)
        self.assertIsInstance(tran, translators.ISCTPTranslator)

        tran = translators.make_translator('ctp')
        self.assertIsNotNone(tran)
        self.assertIsInstance(tran, translators.CTPTranslator)

        tran = translators.make_translator('dummyctp')
        self.assertIsNotNone(tran)
        self.assertIsInstance(tran, translators.DummyCTPTranslator)

        tran = translators.make_translator('conctp')
        self.assertIsNotNone(tran)
        self.assertIsInstance(tran, translators.ConservativeCTPTranslator)

        tran = translators.make_translator('stdout')
        self.assertIsNotNone(tran)
        self.assertIsInstance(tran, translators.StdOutTranslator)

        tran = translators.make_translator('massi')
        self.assertIsNotNone(tran)
        self.assertIsInstance(tran, translators.MassiTranslator)

        # sqlite needs a file name
        tran = translators.make_translator('sqlite:test.xyz')
        self.assertIsNotNone(tran)
        self.assertIsInstance(tran, translators.SQLiteTranslator)
        with self.assertRaises(TypeError):
            tran = translators.make_translator('sqlite')

        tran = translators.make_translator('csv:test.xyz')
        self.assertIsNotNone(tran)
        self.assertIsInstance(tran, translators.CSVTranslator)
        with self.assertRaises(TypeError):
            tran = translators.make_translator('csv')

        tran = translators.make_translator('pickle:test.xyz')
        self.assertIsNotNone(tran)
        self.assertIsInstance(tran, translators.PickleTranslator)
        with self.assertRaises(TypeError):
            tran = translators.make_translator('pickle')

        tran = translators.make_translator('file : test.csv')
        self.assertIsNotNone(tran)
        self.assertIsInstance(tran, translators.CSVTranslator)
        with self.assertRaises(TypeError):
            tran = translators.make_translator('file')

        # these should not try to open database connection immediately
        # so it should be safe to test them
        for desc in ("off", "onhlt", "onbunch", "onmon"):
            tran = translators.make_translator(desc)
            self.assertIsNotNone(tran)
            self.assertIsInstance(tran, translators.DBTranslator)

        # unknown
        with self.assertRaises(TypeError):
            tran = translators.make_translator('not-a-thing')

        # description with a tag
        tran = translators.make_translator('stdout:tag-123')
        self.assertIsNotNone(tran)
        self.assertIsInstance(tran, translators.StdOutTranslator)

        # few of them together
        tran = translators.make_translator('stdout,csv:test.dat, sqlite:test.sqlite , onbunch ')
        self.assertIsNotNone(tran)
        self.assertIsInstance(tran, translators.MultiTranslator)
        self.assertEqual(len(tran._translators), 4)
        self.assertIsInstance(tran._translators[0], translators.StdOutTranslator)
        self.assertIsInstance(tran._translators[1], translators.CSVTranslator)
        self.assertIsInstance(tran._translators[2], translators.SQLiteTranslator)
        self.assertIsInstance(tran._translators[3], translators.DBTranslator)

    def test_db_tran(self):
        """Test for DBTranslator class"""

        # make an object to store in a database
        snap = model.BSSnapshot(run=358333, lbn=95)
        bsrepr = model.BeamSpotRepresentation([snap])

        with makeTmpFile() as filename:
            tran = translators.make_translator('sqlite:' + filename)
            tran.write(bsrepr)

            tran = translators.make_translator('sqlite:' + filename, run=358333, lbn=95)
            self.assertIsNotNone(tran.read())


if __name__ == "__main__":
    unittest.main()
