package chip.executor;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.factory.annotation.Autowired;

import rc.RunParamsNamed;
import TRIGGER.CommandExecutionFailed;
import TRIGGER.TriggerCommander;
import chip.CEPService;
import chip.CHIPException;
import chip.CHIPException.ExecutorTriggerException;
import daq.eformat.DetectorMask;

// TODO: all the commands to the MasterTrigger are synchronized because of an issue
// with concurrent calls to the same CORBA servant. The synchronization should be
// removed once the issue is understood and fixed

public class ExecutorTrigger implements Executor {

    @Autowired
    CEPService cep;

    private static final Log log = LogFactory.getLog(ExecutorTrigger.class);

    private volatile String partName = null;

    /**
     * Commander used to send the hold and resume trigger commands to the master trigger
     */
    private volatile TriggerCommander trgCommander;

    public ExecutorTrigger() {
    }

    @Override
    public Category getCategory() {
        return Executor.Category.Trigger;
    }

    public boolean isInitialized() {
        if(this.trgCommander == null) {
            return false;
        }

        return true;
    }

    public void init(final String partitionName, final String controller) throws ExecutorTriggerException {
        this.partName = partitionName;

        try {
            if(this.trgCommander == null) {
                if(controller != null) {
                    this.trgCommander = new TriggerCommander(this.partName, controller);
                }
            }
        }
        catch(final Exception e) {
            throw new CHIPException.ExecutorTriggerException("Init trigger commander", e);
        }
    }

    // This returns the last ECR
    public synchronized long hold() throws ExecutorTriggerException {
        final long ecr = this.hold(new DetectorMask());
        return ecr;
    }

    // This returns the last ECR
    public synchronized long hold(final DetectorMask dmask) throws ExecutorTriggerException {
        final long ecr = this.hold(dmask, false);
        return ecr;
    }
    
    // This can return the ECR or the extended L1 ID
    public synchronized long hold(final DetectorMask dmask, final boolean extendedL1ID) throws ExecutorTriggerException {
        long value = 0;

        // hold trigger
        try {
            if(dmask.get().isEmpty() == true) {
                value = this.trgCommander.hold("", extendedL1ID);
                ExecutorTrigger.log.debug("Holding the trigger, " + ((extendedL1ID == false) ? "ecr=" : "extended L1ID=") + String.valueOf(value));
            } else {
                value = this.trgCommander.hold(dmask.toString(), extendedL1ID);
                ExecutorTrigger.log.debug("Holding the trigger for " + StringUtils.join(dmask.getSubdetectors().iterator(), ", ") + ", "
                                          + ((extendedL1ID == false) ? "ecr=" : "extended L1ID=")
                                          + String.valueOf(value));
            }
        }
        catch(final CommandExecutionFailed e) {
            throw new CHIPException.ExecutorTriggerException("Hold trigger: " + e.reason, e);
        }
        catch(final Exception e) {
            throw new CHIPException.ExecutorTriggerException("Hold trigger", e);
        }

        return value;
    }

    public synchronized void resume() throws ExecutorTriggerException {
        if(this.trgCommander == null) {
            ExecutorTrigger.log.error("EcecutorTrigger not yet initialized!");
        }

        // resume trigger
        try {
            this.trgCommander.resume();
        }
        catch(final CommandExecutionFailed e) {
            throw new CHIPException.ExecutorTriggerException("Resume trigger: " + e.reason, e);
        }
        catch(final Exception e) {
            throw new CHIPException.ExecutorTriggerException("Resume trigger", e);
        }
    }

    public synchronized void resume(final DetectorMask dm) throws ExecutorTriggerException {
        if(this.trgCommander == null) {
            ExecutorTrigger.log.error("EcecutorTrigger not yet initialized!");
        }

        // resume trigger
        try {
            if(dm.get().isEmpty() == false) {
                this.trgCommander.resume(dm.toString());
            } else {
                this.trgCommander.resume();
            }
        }
        catch(final CommandExecutionFailed e) {
            throw new CHIPException.ExecutorTriggerException("Resume trigger: " + e.reason, e);
        }
        catch(final Exception e) {
            throw new CHIPException.ExecutorTriggerException("Resume trigger", e);
        }
    }
    
    public synchronized void setHLTPrescales(final int l1Ps, final int hltPs) throws ExecutorTriggerException {
        if(this.trgCommander == null) {
            ExecutorTrigger.log.error("EcecutorTrigger not yet initialized!");
        }

        // resume trigger
        try {
            this.trgCommander.setPrescales(l1Ps, hltPs);
        }
        catch(final CommandExecutionFailed e) {
            throw new CHIPException.ExecutorTriggerException("Set prescales: " + e.reason, e);
        }
        catch(final Exception e) {
            throw new CHIPException.ExecutorTriggerException("Set prescales", e);
        }
    }

    public synchronized void increaseLumiBlock() throws ExecutorTriggerException {
        if(this.trgCommander == null) {
            ExecutorTrigger.log.error("EcecutorTrigger not yet initialized!");
        }

        try {
            // read run number from IS
            int runNumber = 0;
            final ipc.Partition part = new ipc.Partition(this.partName);
            final RunParamsNamed runParams = new RunParamsNamed(part, "RunParams.RunParams");
            runParams.checkout();
            runNumber = runParams.run_number;
            // increase lumi block
            this.trgCommander.increaseLumiBlock(runNumber);

        }
        catch(final CommandExecutionFailed e) {
            throw new CHIPException.ExecutorTriggerException("Increase luminosity block: " + e.reason, e);
        }
        catch(final Exception e) {
            throw new CHIPException.ExecutorTriggerException("Increase luminosity block", e);
        }
    }
}
