package chip.subscriber.auto;

import ipc.Partition;
import is.InfoNotCompatibleException;
import is.RepositoryNotFoundException;

import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.factory.annotation.Autowired;

import rc.HoldTriggerInfoNamed;
import rc.HoldTriggerInfoNamed.CausedBy;
import rc.HoldTriggerInfoNamed.Reason;
import chip.CEPService;
import chip.CHIPException.DAQConfigurationError;
import chip.CHIPException.ExecutorIGUIException;
import chip.CHIPException.ExecutorRCException;
import chip.CHIPException.ExecutorTriggerException;
import chip.CHIPException.WaitTimeoutException;
import chip.DAQConfigReader;
import chip.dao.EsperDAO;
import chip.event.config.RCApplicationConfig;
import chip.event.en.FSM;
import chip.event.issue.Issue;
import chip.event.issue.auto.ClockSwitch;
import chip.event.issue.auto.ClockSwitch.STATUS;
import chip.executor.ExecutorIGUI;
import chip.executor.ExecutorRC;
import chip.executor.ExecutorTrigger;
import chip.msg.Common;
import chip.subscriber.DeferredActionExecutor;
import daq.EsperUtils.ISReader;
import daq.EsperUtils.TypedObject;
import daq.rc.Command.FSMCommands;
import daq.rc.RCException.CORBAException;


public class ClockSwitchExecutor extends DeferredActionExecutor {
    @Autowired
    CEPService cep;

    private static final Log log = LogFactory.getLog(ClockSwitchExecutor.class);
    private final static String L1_CALO_CONTROLLER = "l1calo-processor";
    private final static String L1_CALO_RESYNC_CMD = "l1caloResync";

    private enum EXIT_REASON {
        FAILURE,
        SUCCESS
    }

    private final ExecutorRC execRC_partition;
    private final ExecutorRC execRC_initialL1CT;
    private final ExecutorIGUI execIGUI;
    private final ExecutorTrigger execTrigger;
    private final DAQConfigReader daqReader;

    public ClockSwitchExecutor(final ExecutorRC execRC_partition,
                               final ExecutorRC execRC_initialL1CT,
                               final ExecutorIGUI execIGUI,
                               final ExecutorTrigger execTrigger,
                               final EsperDAO dao,
                               final DAQConfigReader daqReader)
    {
        this.execRC_partition = execRC_partition;
        this.execRC_initialL1CT = execRC_initialL1CT;
        this.execIGUI = execIGUI;
        this.execTrigger = execTrigger;
        this.daqReader = daqReader;
    }

    @Override
    public void doUpdate(final Issue issue) {
        final ClockSwitch clockSwitch = (ClockSwitch) issue;

        switch(clockSwitch.getType()) {
            case SELECT_INTERNAL:
                this.selectInternalClock(clockSwitch);
                break;
            case SELECT_LHC:
                this.selectLhcClock(clockSwitch);
                break;
            case SELECT_INTERNAL_AND_DO_UNCONFIG:
                this.selectInternalAndUnconfig(clockSwitch);
                break;
            default:
                break;
        }
    }

    // This is executed in ExpertSystemServant.transitionDone
    // It is executed when the NONE state is reached
    // It must be executed synchronously because the action must be completed
    // before the transition is done 
    public void setAutomaticClock() {
        final rc.AutomaticClockNamed clk = new rc.AutomaticClockNamed(new ipc.Partition(this.daqReader.getPartitionName()),
                                                                      "RunParams.AutomaticClock");
        clk.value = true;

        try {
            clk.checkin();
        }
        catch(final RepositoryNotFoundException | InfoNotCompatibleException ex) {
            ers.Logger.error(new Common("Not able to set clock switch to automatic mode: " + ex, ex));
        }
    }

    private synchronized void selectInternalAndUnconfig(final ClockSwitch clockSwitch) {
        ers.Logger.info(new Common("The clock has to be switched back to INTERNAL since the beam mode changed to "
                                   + clockSwitch.getBeamMode() + ". Afterwards ATLAS will be unconfigured."));

        final EXIT_REASON exit = this.selectInternalClock(clockSwitch);

        if(exit == EXIT_REASON.FAILURE) {
            // Logging of problem done in selectInternalClock method
            return;
        }

        try {
            final String rootName = this.daqReader.getRootController().UID();
            this.execRC_partition.makeTransition(rootName, FSMCommands.UNCONFIG);
        }
        catch(final ExecutorRCException | DAQConfigurationError e) {
            final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                      STATUS.FAILED,
                                                      clockSwitch.getBeamMode(),
                                                      clockSwitch.getState(),
                                                      clockSwitch.getId());
            event.setError("Not able to send unconfig to the Root Controller", e);
            this.cep.injectEvent(event);
            return;
        }

        final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                  STATUS.DONE,
                                                  clockSwitch.getBeamMode(),
                                                  clockSwitch.getState(),
                                                  clockSwitch.getId());
        this.cep.injectEvent(event);
    }

    private synchronized void selectLhcClock(final ClockSwitch clockSwitch) {
        final Boolean rcStateIsRunning = (clockSwitch.getState() == FSM.RUNNING) ? Boolean.TRUE : Boolean.FALSE;

        TypedObject externalClockChoiceTypedObj = null;
        Exception cause = null;
        try {
            externalClockChoiceTypedObj = ISReader.getInfoByIndex("initial", "LHC.ExternalClockChoice", 0);
        }
        catch(final Exception ex) {
            cause = ex;
        }

        if(externalClockChoiceTypedObj == null) {
            final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                      STATUS.FAILED,
                                                      clockSwitch.getBeamMode(),
                                                      clockSwitch.getState(),
                                                      clockSwitch.getId());

            final String errMsg = "Could not retrieve new keys for clock from IS! Check if IS object LHC.ExternalClockCoice exists";
            if(cause == null) {
                event.setError(errMsg);
            } else {
                event.setError(errMsg, cause);
            }
            this.cep.injectEvent(event);

            return;
        }

        final String externalClockChoice = externalClockChoiceTypedObj.getString();

        if(externalClockChoice.isEmpty()) {
            final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                      STATUS.FAILED,
                                                      clockSwitch.getBeamMode(),
                                                      clockSwitch.getState(),
                                                      clockSwitch.getId());
            event.setError("Could not retrieve new keys for clock from IS! Check if IS object LHC.ExternalClockCoice exists");
            this.cep.injectEvent(event);
            return;
        }

        if(rcStateIsRunning.booleanValue() == true) {
            try {
                this.execTrigger.hold();
                this.publishHoldInfo(clockSwitch);

                ers.Logger.info(new Common("Sending resynch command to L1Calo before changing clock to external."));

                this.sendResync_pre(clockSwitch);

                ers.Logger.info(new Common("1 second before changing clock to external."));

                try {
                    Thread.sleep(1000);
                }
                catch(final InterruptedException e) {
                    final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                              STATUS.PROBLEM,
                                                              clockSwitch.getBeamMode(),
                                                              clockSwitch.getState(),
                                                              clockSwitch.getId());
                    event.setError("Problem while sleeping", e);
                    this.cep.injectEvent(event);
                }
            }
            catch(final ExecutorTriggerException e) {
                final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                          STATUS.FAILED,
                                                          clockSwitch.getBeamMode(),
                                                          clockSwitch.getState(),
                                                          clockSwitch.getId());
                event.setError("Failed to hold the trigger", e);
                this.cep.injectEvent(event);
                return;
            }
        }

        boolean failed = false;
        try {
            this.execRC_initialL1CT.sendUserCommand("RF2TTCApp", "SELECT", externalClockChoice.split("\\s+"), Boolean.TRUE, 60);
        }
        catch(final ExecutorRCException e) {
            // Send special resynch to L1Calo because of the failure
            this.sendResync_failure(clockSwitch);

            final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                      STATUS.FAILED,
                                                      clockSwitch.getBeamMode(),
                                                      clockSwitch.getState(),
                                                      clockSwitch.getId());
            event.setError("Failed to talk to RF2TTCApp in initialL1CT partition! Could not switch clock to " + externalClockChoice, e);
            this.cep.injectEvent(event);

            failed = true;
        }
        finally {
            if(rcStateIsRunning.booleanValue() == true) {
                if(failed == false) {
                    try {
                        Thread.sleep(3000);
                    }
                    catch(final InterruptedException e) {
                        final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                                  STATUS.PROBLEM,
                                                                  clockSwitch.getBeamMode(),
                                                                  clockSwitch.getState(),
                                                                  clockSwitch.getId());
                        event.setError("Problem while sleeping", e);
                        this.cep.injectEvent(event);
                    }

                    this.sendResync(clockSwitch);
                }

                try {
                    this.execTrigger.resume();
                }
                catch(final ExecutorTriggerException e) {
                    if(failed == false) {
                        final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                                  STATUS.FAILED,
                                                                  clockSwitch.getBeamMode(),
                                                                  clockSwitch.getState(),
                                                                  clockSwitch.getId());
                        event.setError("Failed to resume the trigger", e);
                        this.cep.injectEvent(event);

                        failed = true;
                    } else {
                        ClockSwitchExecutor.log.error("Failed to resume the trigger", e);
                    }

                    this.sendErrorToIGUI("Failed to resume the trigger after a clock switch; a manual intervention is needed");
                }
            }
        }

        if(failed == false) {
            final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                      STATUS.DONE,
                                                      clockSwitch.getBeamMode(),
                                                      clockSwitch.getState(),
                                                      clockSwitch.getId());
            this.cep.injectEvent(event);
        }
    }

    private synchronized EXIT_REASON selectInternalClock(final ClockSwitch clockSwitch) {
        final Boolean rcIsRunning = (clockSwitch.getState() == FSM.RUNNING) ? Boolean.TRUE : Boolean.FALSE;
        if(rcIsRunning.booleanValue()) {
            try {
                this.execTrigger.hold();
                this.publishHoldInfo(clockSwitch);

                ers.Logger.info(new Common("Sending resynch command to L1Calo before changing clock to internal."));

                this.sendResync_pre(clockSwitch);

                ers.Logger.info(new Common("1 second before changing clock to internal"));

                try {
                    Thread.sleep(1000);
                }
                catch(final InterruptedException e) {
                    final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                              STATUS.PROBLEM,
                                                              clockSwitch.getBeamMode(),
                                                              clockSwitch.getState(),
                                                              clockSwitch.getId());
                    event.setError("Problem while sleeping", e);
                    this.cep.injectEvent(event);
                }
            }
            catch(final ExecutorTriggerException e) {
                final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                          STATUS.FAILED,
                                                          clockSwitch.getBeamMode(),
                                                          clockSwitch.getState(),
                                                          clockSwitch.getId());
                event.setError("Failed to hold the trigger.", e);
                this.cep.injectEvent(event);
                return EXIT_REASON.FAILURE;
            }
        }

        boolean failed = false;
        try {
            this.execRC_initialL1CT.sendUserCommand("RF2TTCApp", "SELECT", new String[] {"BCREF", "INT"}, Boolean.TRUE, 60);
        }
        catch(final ExecutorRCException e) {
            // Send command to L1Calo because of the failure
            this.sendResync_failure(clockSwitch);

            final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                      STATUS.FAILED,
                                                      clockSwitch.getBeamMode(),
                                                      clockSwitch.getState(),
                                                      clockSwitch.getId());
            event.setError("Failed to talk to RF2TTCApp in initialL1CT partition! Could not switch clock to BCREF", e);
            this.cep.injectEvent(event);

            failed = true;
        }
        finally {
            if(rcIsRunning.booleanValue()) {
                if(failed == false) {
                    try {
                        Thread.sleep(3000);
                    }
                    catch(final InterruptedException e) {
                        final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                                  STATUS.PROBLEM,
                                                                  clockSwitch.getBeamMode(),
                                                                  clockSwitch.getState(),
                                                                  clockSwitch.getId());
                        event.setError("Problem while sleeping", e);
                        this.cep.injectEvent(event);
                    }

                    this.sendResync(clockSwitch);
                }

                try {
                    this.execTrigger.resume();
                }
                catch(final ExecutorTriggerException e) {
                    if(failed == false) {
                        final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                                  STATUS.FAILED,
                                                                  clockSwitch.getBeamMode(),
                                                                  clockSwitch.getState(),
                                                                  clockSwitch.getId());
                        event.setError("Failed to resume the trigger", e);
                        this.cep.injectEvent(event);

                        failed = true;
                    } else {
                        ClockSwitchExecutor.log.error("Failed to resume the trigger", e);
                    }

                    this.sendErrorToIGUI("Failed to resume the trigger after a clock switch; a manual intervention is needed");
                }
            }
        }

        if(failed == false) {
            final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                      STATUS.DONE,
                                                      clockSwitch.getBeamMode(),
                                                      clockSwitch.getState(),
                                                      clockSwitch.getId());
            this.cep.injectEvent(event);
        }

        return((failed == false) ? EXIT_REASON.SUCCESS : EXIT_REASON.FAILURE);
    }

    private void sendResync_pre(final ClockSwitch clockSwitch) {
        final Map<String, RCApplicationConfig> allApps = this.daqReader.getRCApplications();
        if(allApps.containsKey(ClockSwitchExecutor.L1_CALO_CONTROLLER) == true) {
            try {
                this.execRC_partition.sendUserBroadcast(ClockSwitchExecutor.L1_CALO_CONTROLLER,
                                                        ClockSwitchExecutor.L1_CALO_RESYNC_CMD,
                                                        new String[] {"0"},
                                                        Boolean.TRUE,
                                                        60);
            }
            catch(final ExecutorRCException ex) {
                final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                          STATUS.PROBLEM,
                                                          clockSwitch.getBeamMode(),
                                                          clockSwitch.getState(),
                                                          clockSwitch.getId());
                event.setError("Problem sending the resync \"0 \" command to L1Calo", ex);
                this.cep.injectEvent(event);
            }
        } else {
            ClockSwitchExecutor.log.info("Not sending resynch \"0\" command to " + ClockSwitchExecutor.L1_CALO_CONTROLLER
                                         + " because it is not part of the partition");
        }
    }

    private void sendResync_failure(final ClockSwitch clockSwitch) {
        final Map<String, RCApplicationConfig> allApps = this.daqReader.getRCApplications();
        if(allApps.containsKey(ClockSwitchExecutor.L1_CALO_CONTROLLER) == true) {
            try {
                this.execRC_partition.sendUserBroadcast(ClockSwitchExecutor.L1_CALO_CONTROLLER,
                                                        ClockSwitchExecutor.L1_CALO_RESYNC_CMD,
                                                        new String[] {"-1"},
                                                        Boolean.TRUE,
                                                        60);
            }
            catch(final ExecutorRCException ex) {
                final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                          STATUS.PROBLEM,
                                                          clockSwitch.getBeamMode(),
                                                          clockSwitch.getState(),
                                                          clockSwitch.getId());
                event.setError("Problem sending the resync \"-1\" command to L1Calo", ex);
                this.cep.injectEvent(event);
            }
        } else {
            ClockSwitchExecutor.log.info("Not sending resynch \"-1\" command to " + ClockSwitchExecutor.L1_CALO_CONTROLLER
                                         + " because it is not part of the partition");
        }
    }

    private void sendResync(final ClockSwitch clockSwitch) {
        final Map<String, RCApplicationConfig> allApps = this.daqReader.getRCApplications();
        if(allApps.containsKey(ClockSwitchExecutor.L1_CALO_CONTROLLER) == true) {
            short step = 0;

            try {
                step = 1;
                this.execRC_partition.sendUserBroadcast(ClockSwitchExecutor.L1_CALO_CONTROLLER,
                                                        ClockSwitchExecutor.L1_CALO_RESYNC_CMD,
                                                        new String[] {"1"},
                                                        Boolean.TRUE,
                                                        60);

                step = 2;
                this.execRC_partition.sendUserBroadcast(ClockSwitchExecutor.L1_CALO_CONTROLLER,
                                                        ClockSwitchExecutor.L1_CALO_RESYNC_CMD,
                                                        new String[] {"2"},
                                                        Boolean.TRUE,
                                                        60);

                step = 3;
                this.execRC_partition.sendUserBroadcast(ClockSwitchExecutor.L1_CALO_CONTROLLER,
                                                        ClockSwitchExecutor.L1_CALO_RESYNC_CMD,
                                                        new String[] {"3"},
                                                        Boolean.TRUE,
                                                        60);
            }
            catch(final ExecutorRCException ex) {
                // Send the resynch failure command to L1Calo when any of the previous command failed and:
                // - the failure is not due to a communication issue (most likely the application is not running)
                // - the failure is due to the elapsed timeout unless the last command has been sent
                if((CORBAException.class.isInstance(ex.getCause()) == false)
                   && ((WaitTimeoutException.class.isInstance(ex.getCause()) == false) || (step != 3)))
                {
                    this.sendResync_failure(clockSwitch);
                }

                final ClockSwitch event = new ClockSwitch(clockSwitch.getType(),
                                                          STATUS.PROBLEM,
                                                          clockSwitch.getBeamMode(),
                                                          clockSwitch.getState(),
                                                          clockSwitch.getId());
                event.setError("Problem sending the resync \"" + step + "\" command to L1Calo", ex);
                this.cep.injectEvent(event);
            }
        } else {
            ClockSwitchExecutor.log.info("Not sending resynch commands to " + ClockSwitchExecutor.L1_CALO_CONTROLLER
                                         + " because it is not part of the partition");
        }
    }

    private void publishHoldInfo(final ClockSwitch clock) {
        try {
            final HoldTriggerInfoNamed info = new HoldTriggerInfoNamed(new Partition(this.daqReader.getPartitionName()),
                                                                       "RunParams.HoldTrigger");
            info.causedBy = CausedBy.Other;
            info.causedByString = CausedBy.Other.name();
            info.reason = Reason.ClockSwitch;
            info.reasonString = Reason.ClockSwitch.toString();
            info.checkin();
        }
        catch(final Exception e) {
            final ClockSwitch event =
                                    new ClockSwitch(clock.getType(), STATUS.PROBLEM, clock.getBeamMode(), clock.getState(), clock.getId());
            event.setError("Failed to publish HoldTriggerInfo to IS", e);
            this.cep.injectEvent(event);
        }
    }

    private void sendErrorToIGUI(final String msg) {
        try {
            this.execIGUI.showError(msg);
        }
        catch(final ExecutorIGUIException ex) {
            ClockSwitchExecutor.log.error("Failed to send error message to the IGUI", ex);
        }
    }
}
