package chip.subscriber.recovery;

import ipc.Partition;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;

import rc.HoldTriggerInfoNamed;
import rc.HoldTriggerInfoNamed.Reason;
import chip.CEPService;
import chip.CHIPException.ExecutorIGUIException;
import chip.CHIPException.ExecutorRCException;
import chip.CHIPException.ExecutorTriggerException;
import chip.CHIPException.WaitTimeoutException;
import chip.DAQConfigReader;
import chip.dao.EsperDAO;
import chip.event.issue.HoldsTrigger;
import chip.event.issue.Issue;
import chip.event.issue.recovery.Resynch;
import chip.event.issue.recovery.Resynch.STATUS;
import chip.executor.ExecutorIGUI;
import chip.executor.ExecutorRC;
import chip.executor.ExecutorTrigger;
import chip.subscriber.DeferredActionExecutor;
import daq.eformat.DetectorMask;


public class ResynchExecutor extends DeferredActionExecutor {

    @Autowired
    CEPService cep;

    private final ExecutorRC execRC;
    private final ExecutorIGUI execIGUI;
    private final ExecutorTrigger execTrigger;
    private final DAQConfigReader daqReader;

    @Value("#{configuration['chip.resynch.timeout']}")
    Integer resynchTimeout;

    public ResynchExecutor(final ExecutorRC execRC,
                           final ExecutorIGUI execIGUI,
                           final ExecutorTrigger execTrigger,
                           final EsperDAO dao,
                           final DAQConfigReader daqReader)
    {
        this.execRC = execRC;
        this.execIGUI = execIGUI;
        this.execTrigger = execTrigger;
        this.daqReader = daqReader;
    }

    @Override
    public void doUpdate(final Issue issue) {
        final Resynch resynch = (Resynch) issue;

        switch(resynch.getType()) {
            case RESYNCH:
                this.doResynch(resynch);
                break;
            default:
                break;
        }
    }

    synchronized private void doResynch(final Resynch resynch) {
        final String controller = resynch.getController();

        final DetectorMask dmask = this.daqReader.getDetMaskForApplication(controller);

        // hold trigger
        long el1id = 0;
        try {
            el1id = this.execTrigger.hold(dmask, true);
        }
        catch(final ExecutorTriggerException e) {
            final Resynch event = new Resynch(controller, resynch.getType(), STATUS.FAILED, resynch.getId());
            final String msg = "Failed to hold the trigger";
            event.setError(msg, e);
            this.cep.injectEvent(event);

            try {
                this.execIGUI.showError("The resynch for controller " + controller + " cannot be started: " + msg);
            }
            catch(final ExecutorIGUIException ex) {
                ers.Logger.error(ex);
            }

            return;
        }

        // publish name of detector which causes dead-time
        final HoldTriggerInfoNamed.CausedBy detIdEnum = HoldsTrigger.getCausedBy(dmask);
        if(detIdEnum.equals(HoldTriggerInfoNamed.CausedBy.Other) == true) {
            final Resynch event = new Resynch(controller, resynch.getType(), STATUS.PROBLEM, resynch.getId());
            event.setError("Failed to determine the detector of " + controller + ". Associated detector mask is " + dmask.toString());
            this.cep.injectEvent(event);
        }

        try {
            final HoldTriggerInfoNamed info = new HoldTriggerInfoNamed(new Partition(this.daqReader.getPartitionName()),
                                                                       "RunParams.HoldTrigger");
            info.causedBy = detIdEnum;
            info.causedByString = detIdEnum.name();
            info.reason = Reason.Resync;
            info.reasonString = Reason.Resync.toString();
            info.checkin();
        }
        catch(final Exception e) {
            final Resynch event = new Resynch(controller, resynch.getType(), STATUS.PROBLEM, resynch.getId());
            event.setError("Failed to publish the name of the detector to IS", e);
            this.cep.injectEvent(event);
        }

        // send RESYNCH to controller
        try {
            this.execRC.sendResynch(controller, el1id, new String[] {}, true, this.resynchTimeout.intValue());
        }
        catch(final ExecutorRCException e) {
            final StringBuilder msg = new StringBuilder();
            msg.append("Failed to perform resynch of ");
            msg.append(controller);

            boolean triggerResumed = false;
            final Throwable cause = e.getCause();
            if((WaitTimeoutException.class.isInstance(cause) == true) || (InterruptedException.class.isInstance(cause) == true)) {
                // The resync was executed by the controller but the command timed-out: do not resume the trigger
                msg.append(" (timeout elapsed)");
            } else {
                // The resync could not just be executed: resume the trigger
                try {
                    this.execTrigger.resume(dmask);
                    triggerResumed = true;
                }
                catch(final ExecutorTriggerException ex) {
                    final Resynch event = new Resynch(controller, resynch.getType(), STATUS.PROBLEM, resynch.getId());
                    event.setError("Not able to resume the trigger after a failure in executing s resynch command (timeout)", ex);
                    this.cep.injectEvent(event);
                }
            }

            final Resynch event = new Resynch(controller, resynch.getType(), STATUS.FAILED, resynch.getId());
            event.setError(msg.toString(), e);
            this.cep.injectEvent(event);

            if(triggerResumed == false) {
                msg.append(". Please, note that the trigger is not resumed and a manual intervention is needed if appropriate");
            }

            try {
                this.execIGUI.showError(msg.toString());
            }
            catch(final ExecutorIGUIException ex) {
                ers.Logger.error(ex);
            }

            return;
        }

        // increase luminosity block
        try {
            this.execTrigger.increaseLumiBlock();
        }
        catch(final ExecutorTriggerException e) {
            final Resynch event = new Resynch(controller, resynch.getType(), STATUS.PROBLEM, resynch.getId());
            event.setError("Failed to increase luminosity block", e);
            this.cep.injectEvent(event);
        }

        // resume trigger
        try {
            this.execTrigger.resume(dmask);
        }
        catch(final ExecutorTriggerException e) {
            final Resynch event = new Resynch(controller, resynch.getType(), STATUS.PROBLEM, resynch.getId());
            event.setError("Not able to resume trigger at end of recovery", e);
            this.cep.injectEvent(event);
        }

        final Resynch event = new Resynch(controller, resynch.getType(), STATUS.DONE, resynch.getId());
        this.cep.injectEvent(event);
    }
}
