//
// @author Rainer Bartoldus, bartoldu@slac.stanford.edu
// @version $Id$
//

#include <string>

#include "CoolUtils/CoolReader.h"

#include "CoolUtils/CoolReadable.h"

#include "ers/ers.h"

#include <CoralBase/AttributeList.h>
#include <CoralBase/Attribute.h>
#include <CoolKernel/RecordSpecification.h>
#include <CoolKernel/Record.h>
#include <CoolKernel/StorageType.h>
#include <CoolKernel/Exception.h>
#include <CoolKernel/IFolder.h>
#include <CoolKernel/FolderVersioning.h>
#include <CoolKernel/IDatabase.h>
#include <CoolKernel/IDatabaseSvc.h>
#include <CoolKernel/IObject.h>
#include <CoolKernel/ValidityKey.h>
#include <CoolKernel/pointers.h>
#include <CoolApplication/DatabaseSvcFactory.h>

// Custom ERS exceptions

namespace daq 
{
  ERS_DECLARE_ISSUE(	coolutils,
			BadCoolRead,
			"Reading from COOL failed because: " << explanation,
                        ( (const char*) explanation )
                        )
}

// Constructors

daq::coolutils::CoolReader::CoolReader( const std::string& connectionName,
                                        const std::string& folderName,
                                        const std::string& tagName )
  : m_connectionName( connectionName )
  , m_folderName( folderName )
  , m_tagName( tagName )
{
  ERS_INFO( "Instantiated a CoolReader for"
            << " connection \"" << m_connectionName << "\""
            << ", folder \"" << m_folderName << "\""
            << ", and tag \"" << m_tagName << "\"" );
}

// Destructor

daq::coolutils::CoolReader::~CoolReader()
{
}

// Methods

bool
daq::coolutils::CoolReader::read( CoolReadable& object,
                                  unsigned int runNumber,
                                  unsigned int lumiBlockNumber,
                                  unsigned int channel ) const
{
  ERS_INFO( "Reading from COOL"
            << " database \"" << m_connectionName << "\""
            << " and folder \"" << m_folderName << "\""
            << "\" with tag \"" << m_tagName << "\"" );

  ERS_INFO( "Selecting IOV that contains run number: " << runNumber
            << ", lumi block number: " << lumiBlockNumber );

  ERS_INFO( "Using channel: " << channel );

  try
    {
      // Open or create the database
      cool::IDatabasePtr database;

      try 
        {
          ERS_INFO( "Trying to open database \"" << m_connectionName << "\"" );

          const bool read_only = true;
          database = cool::DatabaseSvcFactory::databaseService().openDatabase( m_connectionName, read_only );
        }
      catch ( cool::DatabaseDoesNotExist& e )
        {
          daq::coolutils::BadCoolRead i( ERS_HERE, "Database does not exist" );
          ers::error(i);
          return false;
        }

      if ( ! database )
        {
          daq::coolutils::BadCoolRead i( ERS_HERE, "Failed to open database" );
          ers::error(i);
          return false;
        }

      ERS_DEBUG(0, "Successfully opened database" );

      // Find the folder
      ERS_INFO( "Looking for folder \"" << m_folderName << "\"" );

      if ( ! database->existsFolder( m_folderName ) )
        {
          daq::coolutils::BadCoolRead i( ERS_HERE, "Folder does not exist" );
          ers::error(i);
          return false;
        }

      // Fetch the folder
      cool::IFolderPtr folder = database->getFolder( m_folderName );

      if ( ! folder )
        {
          daq::coolutils::BadCoolRead i( ERS_HERE, "Failed to get folder" );
          ers::error(i);
          return false;
        }

      ERS_DEBUG(0, "Retrieved folder \"" << m_folderName << "\"" );

      const cool::ChannelId channelId = channel;

      // Index IOV by run number and lumi block
      const uint64_t a = runNumber;
      const cool::ValidityKey since = ((a<<32) | lumiBlockNumber);
      const cool::ValidityKey until = cool::ValidityKeyMax;

      ERS_INFO( "Computed validity keys: since=" << since << ", until=" << until );

      try
        {
          // Fetch the object for the given time
          ERS_INFO( "Looking for object with tag \"" << m_tagName << "\"" );

          cool::IObjectPtr foundObject = folder->findObject( since, channelId, m_tagName );

          if ( ! foundObject )
            {
              daq::coolutils::BadCoolRead i( ERS_HERE, "Failed to find object" );
              ers::error(i);
              return false;
            }

          ERS_INFO( "Found object:" << foundObject->payload() );

          // Save the object smart pointer
          object.set( foundObject );
          //          ERS_INFO( "Found object:" << object.object()->payload() );
        }
      catch ( cool::TagNotFound& e )
        {
          daq::coolutils::BadCoolRead i( ERS_HERE, "Failed to find tag" );
          ers::error(i);
          return false;
        }
      catch ( cool::ObjectNotFound& e )
        {
          daq::coolutils::BadCoolRead i( ERS_HERE, "Failed to find object" );
          ers::error(i);
          return false;
        }

    }
  catch ( std::exception& e )
    {
      // COOL, CORAL and POOL exceptions inherit from std exceptions,
      // so catching std::exception will catch all errors from COOL,
      // CORAL and POOL.
      daq::coolutils::BadCoolRead i( ERS_HERE, e.what() );
      ers::error(i);
      // FIXME: close database?
      return false;
    }

  return true;
}
