//
// @author Rainer Bartoldus, bartoldu@slac.stanford.edu
// @version $Id$
//

#include <string>

// This class
#include "CoolUtils/CoolWriter.h"

#include "ers/ers.h"

#include "CoolUtils/CoolWritable.h"

#include <CoralBase/AttributeList.h>
#include <CoralBase/Attribute.h>
#include <CoolKernel/RecordSpecification.h>
#include <CoolKernel/Record.h>
#include <CoolKernel/StorageType.h>
#include <CoolKernel/Exception.h>
#include <CoolKernel/IFolder.h>
#include <CoolKernel/FolderSpecification.h>
#include <CoolKernel/FolderVersioning.h>
#include <CoolKernel/IDatabase.h>
#include <CoolKernel/IDatabaseSvc.h>
#include <CoolKernel/IObject.h>
#include <CoolKernel/ValidityKey.h>
#include <CoolKernel/pointers.h>
#include <CoolApplication/DatabaseSvcFactory.h>


// Folder specification

//const std::string folderName = "/Indet/Onl/Beampos";

// This is beamspot-specific, but it is only used to create new folders in tests with sqlite,
// so there is no big worry about it being used to create other types of folders.
const std::string folderDesc = "<timeStamp>run-lumi</timeStamp><addrHeader><address_header service_type=\"71\" clid=\"40774348\" /></addrHeader><typeName>AthenaAttributeList</typeName><key>/Indet/Beampos</key>";
//
// Retrieve this information with:
// 
// $ echo "listinfo /Indet/Onl/Beampos" | AtlCoolConsole.py COOLONL_INDET/COMP200
// ['ATLF', 'ATLAS_COOLPROD', 'atlas_dd']
// __Warning: use COOL_DISABLE_CORALCONNECTIONPOOLCLEANUP__
// Connected to 'oracle://ATLAS_COOLPROD;schema=ATLAS_COOLONL_INDET;dbname=COMP200'
// Welcome to AtlCoolConsole. Type 'help' for instructions.
// >>> Specification for multi-version folder /Indet/Onl/Beampos
//  status (Int32)
//  posX (Float)
//  posY (Float)
//  posZ (Float)
//  sigmaX (Float)
//  sigmaY (Float)
//  sigmaZ (Float)
//  tiltX (Float)
//  tiltY (Float)
//  sigmaXY (Float)
//  posXErr (Float)
//  posYErr (Float)
//  posZErr (Float)
//  sigmaXErr (Float)
//  sigmaYErr (Float)
//  sigmaZErr (Float)
//  tiltXErr (Float)
//  tiltYErr (Float)
//  sigmaXYErr (Float)
// Description: <timeStamp>run-lumi</timeStamp><addrHeader><address_header service_type="71" clid="40774348" /></addrHeader><typeName>AthenaAttributeList</typeName><key>/Indet/Beampos</key>
// 

// Custom ERS exceptions

namespace daq 
{
  ERS_DECLARE_ISSUE(	coolutils,
			BadCoolWrite,
			"Writing to COOL failed because: " << explanation,
                        ( (const char*) explanation )
                        )
}

// Constructors

daq::coolutils::CoolWriter::CoolWriter( const std::string& connectionName,
                                        const std::string& folderName,
                                        const std::string& tagName )
  : m_connectionName( connectionName )
  , m_folderName( folderName )
  , m_tagName( tagName )
{
  ERS_INFO( "Instantiated a CoolWriter for"
            << " connection \"" << m_connectionName << "\""
            << ", folder \"" << m_folderName << "\""
            << ", and tag \"" << m_tagName << "\"" );
}

// Destructor

daq::coolutils::CoolWriter::~CoolWriter()
{
}

// Methods
bool
daq::coolutils::CoolWriter::write( unsigned int runNumber,
                                   unsigned int lumiBlockNumber,
                                   const CoolWritable& object,
                                   unsigned int channel ) const
{
    return _write(runNumber, lumiBlockNumber, object, channel, false);
}

bool
daq::coolutils::CoolWriter::overwrite( unsigned int runNumber,
                                       unsigned int lumiBlockNumber,
                                       const CoolWritable& object,
                                       unsigned int channel) const
{
    return _write(runNumber, lumiBlockNumber, object, channel, true);
}

bool
daq::coolutils::CoolWriter::_write( unsigned int runNumber,
                                    unsigned int lumiBlockNumber,
                                    const CoolWritable& object,
                                    unsigned int channel,
                                    bool overwrite ) const
{
  ERS_INFO( "Writing to COOL"
            << " database \"" << m_connectionName << "\""
            << " the folder \"" << m_folderName << "\""
            << "\" with tag \"" << m_tagName << "\"" );

  ERS_INFO( "Start of IOV is run number: " << runNumber
            << ", lumi block number: " << lumiBlockNumber );

  ERS_INFO( "Using channel: " << channel );

  try
    {
      cool::RecordSpecification recordSpec;
      object.setRecordSpec( recordSpec );  // Derived class method

      // Open or create the database
      cool::IDatabasePtr database;

      try 
        {
          ERS_INFO( "Trying to open database \"" << m_connectionName << "\"" );

          database = cool::DatabaseSvcFactory::databaseService().openDatabase( m_connectionName, false );
          // false means !read_only    
        }
      catch ( cool::DatabaseDoesNotExist& e )
        {
          ERS_INFO( "Database \"" << m_connectionName << "\" does not exist. Trying to create it" );

          database = cool::DatabaseSvcFactory::databaseService().createDatabase( m_connectionName );
        }

      if ( ! database )
        {
          daq::coolutils::BadCoolWrite i( ERS_HERE, "Failed to open database" );
          ers::error(i);
          return false;
        }

      ERS_DEBUG(0, "Successfully opened database" );

      // Fetch or create the folder
      ERS_INFO( "Looking for folder \"" << m_folderName << "\"" );

      if ( ! database->existsFolder( m_folderName ) )
        {
          ERS_INFO( "Folder \"" << m_folderName << "\" does not exist. Trying to create it" );

          // The folder has to be multi-version for this writer to be useful
          cool::FolderSpecification folderSpec( cool::FolderVersioning::MULTI_VERSION, recordSpec );
          cool::IFolderPtr folder =
            database->createFolder( m_folderName, folderSpec, folderDesc, true );

          ERS_INFO( "Successfully created folder" );
        }

      // Fetch the folder
      cool::IFolderPtr folder = database->getFolder( m_folderName );

      if ( ! folder )
        {
          daq::coolutils::BadCoolWrite i( ERS_HERE, "Failed to get folder" );
          ers::error(i);
          return false;
        }

      ERS_DEBUG(0, "Retrieved folder \"" << m_folderName << "\"" );

      const cool::ChannelId channelId = channel;

      // Index IOV by run number and lumi block
      const uint64_t a = runNumber;
      const cool::ValidityKey since = ((a<<32) | lumiBlockNumber);
      const cool::ValidityKey until = cool::ValidityKeyMax;

      ERS_INFO( "Computed validity keys: since=" << since << ", until=" << until );

      // Check if the folder tag is locked
      bool relock = false;

      try
        {
          if ( folder->tagLockStatus( m_tagName ) == cool::HvsTagLock::LOCKED )
            {
              ERS_INFO( "Tag \"" << m_tagName << "\" is locked" );

              // Check that the IOV we are closing (if any) is open-ended
              cool::IObjectPtr iov = folder->findObject( since, channelId, m_tagName );

              if ( iov && iov->until() != cool::ValidityKeyMax )
                {
                  ERS_INFO( "Found an existing IOV overlapping with this one that is not open-ended" );

                  daq::coolutils::BadCoolWrite i( ERS_HERE, "Illegal insert over locked IOV" );
                  ers::error(i);
                  return false;
                }

              // Check that the IOV is different from (older) than the one we are appending
              if ( iov && iov->since() == since )
                {
                  if (!overwrite)
                    {
                      ERS_INFO( "Found an existing open-ended IOV identical to this one" );
                      daq::coolutils::BadCoolWrite i( ERS_HERE, "Illegal insert over locked, identical IOV" );
                      ers::error(i);
                      return false;
                    }
                  else
                    {

                      ERS_INFO( "Found an existing open-ended IOV identical to this one, but ovewrite flag is set" );
                    }
                }

              relock = true;

              ERS_INFO( "Temporarily unlocking tag \"" << m_tagName << "\" in order to append IOV" );

              folder->setTagLockStatus( m_tagName, cool::HvsTagLock::UNLOCKED );
            }
          else
            {
              ERS_INFO( "Tag \"" << m_tagName << "\" is not locked" );
            }
        }
      catch ( cool::TagNotFound& e )
        {
          ERS_INFO( "Tag \"" << m_tagName << "\" was not found. Will try to create it (leaving it unlocked)" );
          // The new tag will be left unlocked. This is what makes most sense when writing more than one channel.
        }
      catch ( cool::ObjectNotFound& e )
        {
          ERS_INFO( "Tag \"" << m_tagName << "\" is locked but no object was found at start of IOV" );
          // This doesn't look safe. I believe it is permissible by the offline UPD1 policy but it doesn't seem like
          // proceeding is what is intended here.  One situation in which this can arise is if there exists an object
          // for one channel but not for the given one. We prefer to require that all valid channels be initialized
          // before the tag is locked.
          daq::coolutils::BadCoolWrite i( ERS_HERE, "Illegal insert over unitialized IOV" );
          ers::error(i);
          return false;
        }

      // Store Object
      cool::Record record( recordSpec );
      object.fillRecord( record );   // Method implemented in the derived class

      const bool userTagOnly = true;
      folder->storeObject( since, until, record, channelId, m_tagName, userTagOnly );

      ERS_INFO( "Stored object: " << record.attributeList() );

      // Relock if necessary
      if ( relock )
        {
          ERS_INFO( "Relocking tag \"" << m_tagName << "\"" );

          folder->setTagLockStatus( m_tagName, cool::HvsTagLock::LOCKED );
        }

      database->closeDatabase();
    }
  catch ( std::exception& e )
    {
      // COOL, CORAL and POOL exceptions inherit from std exceptions,
      // so catching std::exception will catch all errors from COOL,
      // CORAL and POOL.
      daq::coolutils::BadCoolWrite i( ERS_HERE, e.what() );
      ers::error(i);
      // FIXME: close database?
      return false;
    }

  return true;
}
