//--------------------------------------------------------------------------
// File and Version Information:
// 	$Id$
//
// Description:
//	Test suite case for the UnitTestCoolWriter.
//
//------------------------------------------------------------------------

//---------------
// C++ Headers --
//---------------

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include <CoolApplication/DatabaseSvcFactory.h>
#include <CoolKernel/FolderSpecification.h>
#include <CoolKernel/IDatabase.h>
#include <CoolKernel/IDatabaseSvc.h>
#include <CoolKernel/IFolder.h>
#include <CoolKernel/Record.h>
#include <CoolKernel/RecordSpecification.h>
#include "CoolUtils/CoolWritable.h"
#include "CoolUtils/CoolWriter.h"

using namespace daq::coolutils;

#define BOOST_TEST_MODULE UnitTestCoolWriter
#define BOOST_TEST_DYN_LINK
#include <boost/test/unit_test.hpp>

/**
 * Simple test suite for class CoolWriter.
 * See http://www.boost.org/doc/libs/1_59_0/libs/test/doc/html/index.html
 */

namespace {

const std::string folderDesc = "<timeStamp>run-lumi</timeStamp><addrHeader><address_header service_type=\"71\" clid=\"40774348\" /></addrHeader><typeName>AthenaAttributeList</typeName><key>/Indet/Beampos</key>";

// implementation of writable interface just for this test
class TestWritable: public CoolWritable {
public:
    // Constructor
    TestWritable(int a, float b) : _a(a), _b(b) {}

    // Destructor
    virtual ~TestWritable() {}

    // These two methods have to be implemented by the derived class:

    void fillRecord(cool::Record& record) const override {
        record["valueA"].setValue<cool::Int32>(this->_a);
        record["valueB"].setValue<cool::Float>(this->_b);
    }

    void setRecordSpec(cool::RecordSpecification& recordSpec) const override {
        recordSpec.extend("valueA", cool::StorageType::Int32);
        recordSpec.extend("valueB", cool::StorageType::Float);
    }
private:
    int _a;
    float _b;
};


}

// ==============================================================

// Fixture that creates new empty sqlite database
class MakeDb {
public:
    MakeDb(bool lockTag=true) {
        // make temporary file name
        file = tmpnam(nullptr);
        conn = "sqlite://;schema=" + file + ";dbname=CONDBR2";
        BOOST_TEST_MESSAGE("using file " + file);
        folder = "/FOLDER";
        tag = "TAG-UPD1";

        auto db = cool::DatabaseSvcFactory::databaseService().createDatabase(conn);

        TestWritable tw(0, 0.);
        cool::RecordSpecification recordSpec;
        tw.setRecordSpec(recordSpec);
        cool::FolderSpecification folderSpec(cool::FolderVersioning::MULTI_VERSION, recordSpec);
        auto fldr = db->createFolder(folder, folderSpec, folderDesc, true);
        BOOST_TEST_MESSAGE("created database and folder");

        cool::Record record(recordSpec);
        tw.fillRecord(record);
        fldr->storeObject(cool::ValidityKeyMin, cool::ValidityKeyMax, record, 0, tag, true);
        BOOST_TEST_MESSAGE("stored default record");
        if (lockTag) {
            fldr->setTagLockStatus(tag, cool::HvsTagLock::LOCKED);
            BOOST_TEST_MESSAGE("locked tag");
        }
    }

    ~MakeDb() {
        unlink(file.c_str());
        BOOST_TEST_MESSAGE("deleted file " + file);
    }

    std::string file;
    std::string conn;
    std::string folder;
    std::string tag;
};

// ==============================================================

BOOST_AUTO_TEST_CASE(test_no_lock)
{
    MakeDb fixture(false);

    CoolWriter writer(fixture.conn, fixture.folder, fixture.tag);

    // we should be able to write any interval overlaps
    TestWritable tw(1, 2.);

    BOOST_CHECK(writer.write(100, 0, tw));
    BOOST_CHECK(writer.write(100, 10, tw));
    BOOST_CHECK(writer.write(101, 10, tw));
    BOOST_CHECK(writer.write(100001, 100000, tw));
    BOOST_CHECK(writer.write(100001, 100000, tw));
    BOOST_CHECK(writer.write(100001, 0, tw));
    BOOST_CHECK(writer.write(1, 0, tw));
    BOOST_CHECK(writer.write(0, 0, tw));
}

// ==============================================================

BOOST_AUTO_TEST_CASE(test_lock)
{
    MakeDb fixture;

    CoolWriter writer(fixture.conn, fixture.folder, fixture.tag);

    // only moving forward is allowed
    TestWritable tw(1, 2.);

    BOOST_CHECK(writer.write(100, 0, tw));
    BOOST_CHECK(writer.write(100, 10, tw));
    BOOST_CHECK(writer.write(101, 10, tw));
    BOOST_CHECK(writer.write(100001, 100000, tw));
    BOOST_CHECK(!writer.write(100001, 100000, tw));
    BOOST_CHECK(!writer.write(100001, 0, tw));
    BOOST_CHECK(!writer.write(1, 0, tw));
    BOOST_CHECK(!writer.write(0, 0, tw));
    BOOST_CHECK(writer.write(100002, 0, tw));
}

// ==============================================================

BOOST_AUTO_TEST_CASE(test_lock_owerwrite)
{
    MakeDb fixture;

    CoolWriter writer(fixture.conn, fixture.folder, fixture.tag);

    // moving forward or overwriting of the same interval is allowed
    TestWritable tw(1, 2.);

    BOOST_CHECK(writer.overwrite(100, 0, tw));
    BOOST_CHECK(writer.overwrite(100, 10, tw));
    BOOST_CHECK(writer.overwrite(101, 10, tw));
    BOOST_CHECK(writer.overwrite(100001, 100000, tw));
    BOOST_CHECK(writer.overwrite(100001, 100000, tw));
    BOOST_CHECK(!writer.overwrite(100001, 0, tw));
    BOOST_CHECK(!writer.overwrite(1, 0, tw));
    BOOST_CHECK(!writer.overwrite(0, 0, tw));
    BOOST_CHECK(writer.overwrite(100002, 0, tw));
}
