/*
 * RemoteES.cxx
 *
 *  Created on: May 28, 2013
 *      Author: avolio
 */

#include "ExpertSystemInterface/RemoteES.h"
#include "ExpertSystemInterface/Exceptions.h"

#include <ipc/exceptions.h>
#include <dal/Computer.h>
#include <dal/ComputerProgram.h>

#include <map>
#include <memory>

namespace daq { namespace es {

RemoteES::RemoteES(const std::string& partitionName, const std::string& esName)
    : ExpertSystemInterface(), m_partition(partitionName), m_es_name(esName)
{
}

// throws daq::es::ESInterface::CannotNotify
void RemoteES::applicationUpdate(const std::string& appName, const ProcessInfo& info) {
    try {
        ::es::ExpertSystem_var es = m_partition.lookup< ::es::ExpertSystem>(m_es_name);

        ::es::ProcessInformation procInfo;
        RemoteES::fillInfo(info, procInfo);

        es->applicationProcessUpdate(appName.c_str(), procInfo);
    }
    catch(daq::ipc::Exception& ex) {
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, "some error occurred looking for it in the IPC", ex);
    }
    catch(CORBA::Exception& ex) {
        const std::string reason = std::string("a communication problem (\"") + ex._name() + "\") occurred";
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, reason);
    }
}

// throws daq::es::ESInterface::CannotNotify
void RemoteES::applicationUpdate(const std::string& appName, const StateInfo& info) {
    try {
        ::es::ExpertSystem_var es = m_partition.lookup< ::es::ExpertSystem>(m_es_name);

        ::es::StateInformation stateInfo;
        RemoteES::fillInfo(info, stateInfo);

        es->applicationStateUpdate(appName.c_str(), stateInfo);
    }
    catch(daq::ipc::Exception& ex) {
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, "some error occurred looking for it in the IPC", ex);
    }
    catch(CORBA::Exception& ex) {
        const std::string reason = std::string("a communication problem (\"") + ex._name() + "\") occurred";
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, reason);
    }
}

// throws daq::es::ESInterface::CannotNotify
void RemoteES::applicationUpdate(const std::map<std::string, StateInfo>& infoList) {
    try {
        ::es::ExpertSystem_var es = m_partition.lookup< ::es::ExpertSystem>(m_es_name);

        ::es::StateInformation_List allInfo;
        allInfo.length(infoList.size());

        unsigned int idx = 0;
        for(const auto& entry : infoList) {
            allInfo[idx].applicationName = CORBA::string_dup(entry.first.c_str());
            RemoteES::fillInfo(entry.second, allInfo[idx].info);
            ++idx;
        }

        es->applicationsStateUpdate(allInfo);
    }
    catch(daq::ipc::Exception& ex) {
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, "some error occurred looking for it in the IPC", ex);
    }
    catch(CORBA::Exception& ex) {
        const std::string reason = std::string("a communication problem (\"") + ex._name() + "\") occurred";
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, reason);
    }
}

// throws daq::es::ESInterface::CannotNotify
void RemoteES::applicationUpdate(const std::map<std::string, ProcessInfo>& infoList) {
    try {
        ::es::ExpertSystem_var es = m_partition.lookup< ::es::ExpertSystem>(m_es_name);

        ::es::ProcessInformation_List allInfo;
        allInfo.length(infoList.size());

        unsigned int idx = 0;
        for(const auto& entry : infoList) {
            allInfo[idx].applicationName = CORBA::string_dup(entry.first.c_str());
            RemoteES::fillInfo(entry.second, allInfo[idx].info);
            ++idx;
        }

        es->applicationsProcessUpdate(allInfo);
    }
    catch(daq::ipc::Exception& ex) {
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, "some error occurred looking for it in the IPC", ex);
    }
    catch(CORBA::Exception& ex) {
        const std::string reason = std::string("a communication problem (\"") + ex._name() + "\") occurred";
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, reason);
    }
}

// throws daq::es::ESInterface::CannotNotify
void RemoteES::applicationUpdate(const std::string& appName, const TestInfo& info) {
    try {
        ::es::ExpertSystem_var es = m_partition.lookup< ::es::ExpertSystem>(m_es_name);

        ::es::TestStatusInformation testInfo;

        // The global result
        testInfo.globalResult = convert(info.m_global_result);

        // Set the maximum capacity so that reallocations do not happen
        const auto components_size = info.m_components.size();
        testInfo.components.length(components_size);

        // Now loop over all the components
        for(unsigned int c_idx = 0; c_idx < components_size; ++c_idx) {
            testInfo.components[c_idx].componentName = CORBA::string_dup(info.m_components[c_idx]->m_component_name.c_str());
            testInfo.components[c_idx].result = convert(info.m_components[c_idx]->m_test_result);

            const auto numActions = info.m_components[c_idx]->m_actions.size();
            testInfo.components[c_idx].actions.length(numActions);

            for(std::size_t idx = 0; idx < numActions; ++idx) {
                testInfo.components[c_idx].actions[idx].name = CORBA::string_dup(info.m_components[c_idx]->m_actions[idx].command.c_str());
                testInfo.components[c_idx].actions[idx].parameters = CORBA::string_dup(info.m_components[c_idx]->m_actions[idx].parameters_json.c_str());
                testInfo.components[c_idx].actions[idx].diagnosis = CORBA::string_dup(info.m_components[c_idx]->m_actions[idx].description.c_str());
                testInfo.components[c_idx].actions[idx].timeout = info.m_components[c_idx]->m_actions[idx].timeout;
            }
        }

        // Notify the Expert System
        es->applicationTestUpdate(appName.c_str(), testInfo);
    }
    catch(daq::ipc::Exception& ex) {
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, "some error occurred looking for it in the IPC", ex);
    }
    catch(daq::es::ESInterface::Exception& ex) {
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, "unexpected error", ex);
    }
    catch(CORBA::Exception& ex) {
        const std::string reason = std::string("a communication problem (\"") + ex._name() + "\") occurred";
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, reason);
    }
}

// throws daq::es::ESInterface::CannotNotify
void RemoteES::dbReloadInit() {
    try {
        ::es::ExpertSystem_var es = m_partition.lookup< ::es::ExpertSystem>(m_es_name);
        es->dbReloadInit();
    }
    catch(daq::ipc::Exception& ex) {
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, "some error occurred looking for it in the IPC", ex);
    }
    catch(CORBA::Exception& ex) {
        const std::string reason = std::string("a communication problem (\"") + ex._name() + "\") occurred";
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, reason);
    }
}

// throws daq::es::ESInterface::CannotNotify
void RemoteES::dbReloadDone() {
    try {
        ::es::ExpertSystem_var es = m_partition.lookup< ::es::ExpertSystem>(m_es_name);
        es->dbReloadDone();
    }
    catch(daq::ipc::Exception& ex) {
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, "some error occurred looking for it in the IPC", ex);
    }
    catch(CORBA::Exception& ex) {
        const std::string reason = std::string("a communication problem (\"") + ex._name() + "\") occurred";
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, reason);
    }
}

// throws daq::es::ESInterface::CannotNotify
void RemoteES::reset() {
    try {
        ::es::ExpertSystem_var es = m_partition.lookup< ::es::ExpertSystem>(m_es_name);
        es->reset();
    }
    catch(daq::ipc::Exception& ex) {
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, "some error occurred looking for it in the IPC", ex);
    }
    catch(CORBA::Exception& ex) {
        const std::string reason = std::string("a communication problem (\"") + ex._name() + "\") occurred";
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, reason);
    }
}

// throws daq::es::ESInterface::CannotNotify
void RemoteES::transitionDone(const std::string& transitionName, const std::string& newState) {
    try {
        ::es::ExpertSystem_var es = m_partition.lookup< ::es::ExpertSystem>(m_es_name);
        es->transitionDone(transitionName.c_str(), newState.c_str());
    }
    catch(daq::ipc::Exception& ex) {
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, "some error occurred looking for it in the IPC", ex);
    }
    catch(CORBA::Exception& ex) {
        const std::string reason = std::string("a communication problem (\"") + ex._name() + "\") occurred";
        throw daq::es::ESInterface::CannotNotify(ERS_HERE, m_es_name, reason);
    }
}

std::string RemoteES::centralESName() {
    return std::string(::es::centralES);
}

::es::TestResult RemoteES::convert(const daq::tmgr::TestResult& ret) const {
    static const std::map<daq::tmgr::TestResult, ::es::TestResult> m = {{daq::tmgr::TmPass, ::es::TestResult::PASSED},
                                                                        {daq::tmgr::TmUndef, ::es::TestResult::UNDEFINED},
                                                                        {daq::tmgr::TmFail, ::es::TestResult::FAILED},
                                                                        {daq::tmgr::TmUnresolved, ::es::TestResult::UNRESOLVED},
                                                                        {daq::tmgr::TmUntested, ::es::TestResult::UNTESTED},
                                                                        {daq::tmgr::TmUnsupported, ::es::TestResult::UNSUPPORTED}};

    const auto& it = m.find(ret);
    if(it != m.end()) {
        return it->second;
    }

    throw daq::es::ESInterface::Exception(ERS_HERE, std::string("The test result \"") + std::to_string(ret) + "\" is unknown");
}

void RemoteES::fillInfo(const ProcessInfo& from, ::es::ProcessInformation& to) {
    to.status = CORBA::string_dup(from.m_status.c_str());
    to.runningHost = CORBA::string_dup(from.m_running_host.c_str());
    to.membership = from.m_membership;
    to.isRestarting = from.m_restarting;
    to.isNotResponding = from.m_not_responding;
    to.exitCode = from.m_exit_code;
    to.exitSignal = from.m_exit_signal;
    to.infoTime = from.m_info_time;

    to.parentInfo.isExiting = from.m_parent_info.m_exiting;
    to.parentInfo.isTransitioning = from.m_parent_info.m_transitioning;
    to.parentInfo.isInitializing = from.m_parent_info.m_initializing;
    to.parentInfo.fsmState = CORBA::string_dup(from.m_parent_info.m_fsm_state.c_str());

    to.parentInfo.currentTransition.uid = CORBA::string_dup(from.m_parent_info.m_current_transition.m_uid.c_str());
    to.parentInfo.currentTransition.command = CORBA::string_dup(from.m_parent_info.m_current_transition.m_command.c_str());
    to.parentInfo.currentTransition.sourceState = CORBA::string_dup(from.m_parent_info.m_current_transition.m_source_state.c_str());
    to.parentInfo.currentTransition.destinationState = CORBA::string_dup(from.m_parent_info.m_current_transition.m_destination_state.c_str());

    to.parentInfo.lastTransition.uid = CORBA::string_dup(from.m_parent_info.m_last_transition.m_uid.c_str());
    to.parentInfo.lastTransition.command = CORBA::string_dup(from.m_parent_info.m_last_transition.m_command.c_str());
    to.parentInfo.lastTransition.sourceState = CORBA::string_dup(from.m_parent_info.m_last_transition.m_source_state.c_str());
    to.parentInfo.lastTransition.destinationState = CORBA::string_dup(from.m_parent_info.m_last_transition.m_destination_state.c_str());
}

void RemoteES::fillInfo(const StateInfo& from, ::es::StateInformation& to) {
    to.isBusy = from.m_busy;
    to.fsmState = CORBA::string_dup(from.m_fsm_state.c_str());

    to.lastTransition.uid = CORBA::string_dup(from.m_last_transition.m_uid.c_str());
    to.lastTransition.command = CORBA::string_dup(from.m_last_transition.m_command.c_str());
    to.lastTransition.sourceState = CORBA::string_dup(from.m_last_transition.m_source_state.c_str());
    to.lastTransition.destinationState = CORBA::string_dup(from.m_last_transition.m_destination_state.c_str());

    to.currentTransition.uid = CORBA::string_dup(from.m_current_transition.m_uid.c_str());
    to.currentTransition.command = CORBA::string_dup(from.m_current_transition.m_command.c_str());
    to.currentTransition.sourceState = CORBA::string_dup(from.m_current_transition.m_source_state.c_str());
    to.currentTransition.destinationState = CORBA::string_dup(from.m_current_transition.m_destination_state.c_str());

    to.lastCommand.uid = CORBA::string_dup(from.m_last_command.m_uid.c_str());
    to.lastCommand.name = CORBA::string_dup(from.m_last_command.m_name.c_str());
    to.lastCommand.arguments.length(from.m_last_command.m_arguments.size());
    for(unsigned long i = 0; i < to.lastCommand.arguments.length(); ++i) {
        to.lastCommand.arguments[i] = CORBA::string_dup(from.m_last_command.m_arguments[i].c_str());
    }

    to.isInternalError = from.m_internal_error;
    to.lastTransitionTime = from.m_last_transition_time;
    to.infoTime = from.m_info_time;

    to.parentInfo.isExiting = from.m_parent_info.m_exiting;
    to.parentInfo.isTransitioning = from.m_parent_info.m_transitioning;
    to.parentInfo.isInitializing = from.m_parent_info.m_initializing;
    to.parentInfo.fsmState = CORBA::string_dup(from.m_parent_info.m_fsm_state.c_str());

    to.parentInfo.currentTransition.uid = CORBA::string_dup(from.m_parent_info.m_current_transition.m_uid.c_str());
    to.parentInfo.currentTransition.command = CORBA::string_dup(from.m_parent_info.m_current_transition.m_command.c_str());
    to.parentInfo.currentTransition.sourceState = CORBA::string_dup(from.m_parent_info.m_current_transition.m_source_state.c_str());
    to.parentInfo.currentTransition.destinationState = CORBA::string_dup(from.m_parent_info.m_current_transition.m_destination_state.c_str());

    to.parentInfo.lastTransition.uid = CORBA::string_dup(from.m_parent_info.m_last_transition.m_uid.c_str());
    to.parentInfo.lastTransition.command = CORBA::string_dup(from.m_parent_info.m_last_transition.m_command.c_str());
    to.parentInfo.lastTransition.sourceState = CORBA::string_dup(from.m_parent_info.m_last_transition.m_source_state.c_str());
    to.parentInfo.lastTransition.destinationState = CORBA::string_dup(from.m_parent_info.m_last_transition.m_destination_state.c_str());
}

}}

