/*
 * ProxyObjectBase.hpp
 *
 *  Created on: Oct 25, 2018
 *      Author: kolos
 */

#ifndef IPCGATEWAYPROXY_PROXYOBJECTBASE_HPP_
#define IPCGATEWAYPROXY_PROXYOBJECTBASE_HPP_

#include <string>

#include <ipc/object.h>

/*
 * This is an utility structure for converting protected destructor of class T
 * to the public one of class ProtectedBase<T>
 */
template<class I>
struct ReferenceHelper {
    template<class T>
    struct ProtectedBase: public T {
    };

    typedef ProtectedBase<
    			typename std::remove_pointer<typename I::_ptr_type>::type
			> Type;

    static Type * cast(typename I::_ptr_type ptr) {
    	return reinterpret_cast<Type*>(ptr);
    }

    static Type * narrow(CORBA::Object_ptr ref) {
    	return cast(I::_narrow(ref));
    }
};

/*
 * This is the base class for implementing proxy object for a specific IDL interface.
 * A specific proxy implementation has to use the classes generated by the IDL-to-C++
 * compiler from the actual IDL interface as template parameters.
 * This class holds a reference to a target object and forwards all incoming requests
 * to it.
 */
template<class I, class POA_I, template<class > class POA_I_tie>
class ProxyObjectBase:
        public IPCObject<POA_I, ipc::multi_thread, ipc::persistent>,
        public POA_I_tie<typename ReferenceHelper<I>::Type>
{

public:
    /*
     * Creates a new proxy object with the given \p name and \p ref as
     * target object reference.
     */
    ProxyObjectBase(const std::string & name, CORBA::Object_ptr ref) :
            POA_I_tie<typename ReferenceHelper<I>::Type>(
            		ReferenceHelper<I>::narrow(ref), false),
            m_name(name)
    {
        m_target = this->_tied_object();
        if (CORBA::is_nil(m_target)) {
            throw CORBA::INV_OBJREF(0, CORBA::COMPLETED_NO);
        }
        m_self = this->_this();
        ERS_DEBUG(1, "New proxy object with the '"
                << m_name << "' id has been created");
    }

    /*
     * Returns the proxy object id, which should be the same as the target object id.
     */
    std::string unique_id() const {
        return m_name;
    }

    /*
     * Returns the unique type id for this specific proxy object type
     */
    static std::string getTypeId() {
        static const std::string type(I::_PD_repoId);
        return type;
    }

    /*
     * Service function which is implicitly used for CORBA object activation
     */
    PortableServer::POA_ptr _default_POA() {
        return IPCObject<POA_I, ipc::multi_thread, ipc::persistent>::_default_POA();
    }

    /*
     * If \p ref is a reference to this proxy object the function has no effect.
     * Otherwise it will replace the target reference with \p ref.
     *
     */
    bool updateRemoteReference(CORBA::Object_ptr ref) {
        typename I::_var_type typed_ref = I::_narrow(ref);
        if (CORBA::is_nil(typed_ref)) {
            ERS_DEBUG(1, "The new reference is nil");
            return false;
        }
        if (m_self->_is_equivalent(typed_ref)) {
            ERS_DEBUG(1, "The new reference corresponds to this proxy object");
            return false;
        }
        if (!m_target->_is_equivalent(typed_ref)) {
            this->_tied_object(ReferenceHelper<I>::cast(typed_ref), false);
            m_target = typed_ref;
            ERS_DEBUG(1, "Remote object reference has been updated");
            return true;
        }
        ERS_DEBUG(1, "The actual object reference is the same as the new one");
        return false;
    }

    /*
     * Returns CORBA object reference for this proxy object
     */
    CORBA::Object_var getSelfReference() {
        return I::_duplicate(m_self);
    }

    /*
     * Returns CORBA object reference for the target object
     */
    CORBA::Object_var getTargetReference() {
        return I::_duplicate(m_target);
    }

    /*
     * Returns stringified CORBA object reference for the target of this proxy object
     */
    std::string getTargetIOR() {
        return IPCCore::objectToString(m_target);
    }

    /*
     * Returns true if the target object exists, false otherwise.
     *
     */
    bool isTargetExist() {
        try {
            return !CORBA::is_nil(m_target) && !m_target->_non_existent();
        }
        catch (CORBA::SystemException & ex) {
            ERS_DEBUG(2, "Remote object '" << IPCCore::objectToString(m_target)
                << "' does not exist: " << daq::ipc::CorbaSystemException(ERS_HERE, &ex));
            return false;
        }
    }

protected:
    typename I::_var_type m_target;
    typename I::_var_type m_self;
    std::string m_name;
};

#endif /* IPCGATEWAYPROXY_PROXYOBJECTBASE_HPP_ */
