package Igui;

import java.io.IOException;
import java.io.StringReader;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.springframework.context.ApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import Igui.IguiException.ElogException;
import elisa_client.model.InputMessage;
import elisa_client.model.MessageType;
import elisa_client.model.OptionMetadata;
import elisa_client.service.ElisaClient;


class ElogInterface {
    private final static MessageType MESSAGE_TYPE = MessageType.ONLINE;
    private final Map<String, String[]> optionMap = new HashMap<String, String[]>();
    private final ElisaClient elisaClient;
    private final List<String> systemAffected;

    static {
        System.setProperty("elisaRest.URL", (IguiConstants.ELOG_SERVER_URL != null) ? IguiConstants.ELOG_SERVER_URL : "");
    }

    ElogInterface() throws ElogException {
        final ApplicationContext appContext = new ClassPathXmlApplicationContext("java/src/main/resources/applicationContext.xml");
        this.elisaClient = appContext.getBean("elisaClient", ElisaClient.class);
        if(this.elisaClient == null) {
            throw new IguiException.ElogException("Got a null handler to connect to the e-log");
        }

        try {
            this.systemAffected = this.elisaClient.getSystemsAffected(null, null, null, null).getList();

            final List<OptionMetadata> omdList = this.elisaClient.getMessageTypeOptions(null, null, null, ElogInterface.MESSAGE_TYPE).getList();
            for(final OptionMetadata omd : omdList) {
                final String optionName = omd.getName();
                final String vals = omd.getPossibleValues();
                if(vals != null) {
                    this.optionMap.put(optionName, vals.split(","));
                } else {
                    this.optionMap.put(optionName, new String[] {});
                }
            }
        }
        catch(final Exception ex) {
            throw new IguiException.ElogException("Failed to retrieve the e-log configuration: " + ex.getMessage(), ex);
        }
    }

    public List<String> getSystemAffected() {
        return Collections.unmodifiableList(this.systemAffected);
    }

    public String getMessageType() {
        return ElogInterface.MESSAGE_TYPE.toString();
    }

    public String[] getAttributeValues(final String optionName) {
        return this.optionMap.get(optionName);
    }

    public boolean isAttributeValid(final String attributeName) {
        return this.optionMap.containsKey(attributeName);
    }

    public Set<String> getValidAttributes() {
        return Collections.unmodifiableSet(this.optionMap.keySet());
    }

    public void sendMessage(final InputMessage message, final String user, final String pwd) throws ElogException {
        try {
            this.elisaClient.insertMessage(user, pwd, null, message);
        }
        catch(final Exception ex) {
            throw new IguiException.ElogException("Failed to post the e-log entry:\n" + this.parseErrorMessage(ex.getMessage()), ex);
        }
    }

    public static boolean isElogEnvReady() {
        if(IguiConstants.ELOG_SERVER_URL != null) {
            return true;
        }

        return false;
    }

    private String parseErrorMessage(final String msg) {
        String returnMsg;

        try {
            final Document xmlDoc = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(new InputSource(new StringReader(msg.trim())));
            final XPath xpath = XPathFactory.newInstance().newXPath();
            final String status = xpath.evaluate("error_report/error/entry[@key='status']" + "/text()", xmlDoc);
            final String statusCode = xpath.evaluate("error_report/error/entry[@key='code']" + "/text()", xmlDoc);
            final String message = xpath.evaluate("error_report/error/entry[@key='message']" + "/text()", xmlDoc);
            returnMsg = message + " (status = " + status + ", code = " + statusCode + ")";
        }
        catch(final SAXException | IOException | ParserConfigurationException | XPathExpressionException ex) {
            // If the parsing fails, just return the initial string
            IguiLogger.warning("Failed to parse Elog XML error message: " + ex , ex);
            returnMsg = msg;
        }
            
        return returnMsg;
    }
    
//    public static void main(final String[] args) {
//        try {
//            final ElogInterface elogInterface = new ElogInterface();
//            final List<String> sysEff = elogInterface.getSystemAffected();
//            for(final String s : sysEff) {
//                System.err.println(s);
//            }
//        }
//        catch(final Exception ex) {
//            ex.printStackTrace();
//        }
//    }
}
