
//-----------------------
// This Class's Header --
//-----------------------
#include "ISDynAnyNGCont.h"

//-----------------
// C/C++ Headers --
//-----------------

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "ContainerUtils.h"
#include "is/infoany.h"
#include "is/infodynany.h"
#include "MonInfoGatherer/Config.h"
#include "MonInfoGatherer/Exceptions.h"
#include "MonInfoGatherer/InfoPublisher.h"
#include "ers/ers.h"
#include "../algos/ISSum.h"

using namespace MIG;

ISDynAnyNGCont::ISDynAnyNGCont(const std::string &name,
                               const ISType& isType,
                               std::unique_ptr<ISSum> alg,
                               InfoPublisher& publisher,
                               const ConfigParameters& parameters)
  : m_myName(name),
    m_isType(isType),
    m_alg(std::move(alg)),
    m_publisher(publisher),
    m_publishDelay(parameters.ISDynAnyPublishDelay()),
    m_missingCycles(parameters.uintValue("ISDynAnyCyclesMissing", 3)),
    m_publishSum(parameters.boolValue("PublishSum", true)),
    m_publishAvg(parameters.boolValue("PublishAverage", false)),
    m_suffixSum(parameters.value("SuffixSum")),
    m_suffixAvg(parameters.value("SuffixAverage", "_avg"))
{
}

ISDynAnyNGCont::~ISDynAnyNGCont()
{
}

void ISDynAnyNGCont::operate(ISInfoAny &info, const std::string& providerName)
{
    // get provider index
    unsigned provider, nProviders;
    bool newProvider;
    std::tie(provider, nProviders, newProvider) = m_providerMap.index(providerName);

    // lock thyself
    std::lock_guard<std::mutex> lock(m_updateMutex);

    // make sure enough space in array
    if (provider >= m_contributions.size()) {
        m_contributions.resize(provider+1);
    }
    if (m_contributions[provider].info == nullptr) {
        m_contributions[provider].info.reset(new ISInfoDynAny(IPCPartition(), m_isType));
    }

    // just remember it
    info >>= *m_contributions[provider].info;

    // count updates
    for (auto&& contribution: m_contributions) {
        ++ contribution.nUpdatesSinceLastUpdate;
    }
    m_contributions[provider].nUpdatesSinceLastUpdate = 0;
    ++ m_contributions[provider].nUpdatesSincePublish;

    ++m_numUpdatesSincePublish;
    m_lastUpdate = Clock::instance().now();

    // See if we have full set of providers updated since last _publish,
    // and publish if yes, but bootstrap needs some care.
    if (not newProvider) {
        bool all_providers = true;
        for (auto&& contribution: m_contributions) {
            if (contribution.info != nullptr and contribution.nUpdatesSincePublish == 0) {
                // expected a contribution but have not received it
                all_providers = false;
                break;
            }
        }
        if (all_providers) {
            // looks like all contributions were updated
            _publish(lock);
        }
    }
}

// method called periodically to do timeout-based stuff
void ISDynAnyNGCont::execute()
{
    // lock thyself
    std::lock_guard<std::mutex> lock(m_updateMutex);

    // if time since last publication is above threshold then publish
    if (m_numUpdatesSincePublish > 0) {
        auto now = Clock::instance().now();
        if (now - m_lastUpdate >= m_publishDelay) {
            _publish(lock);
        } else {
            // check also degenrate case when publishers go crazy and
            // start publishing more frequently than m_publishDelay
            int np = numUpdatingProviders();
            if (m_numUpdatesSincePublish > np*2) {
                ERS_LOG("Warning: detected frequent publishing: #updates=" << m_numUpdatesSincePublish
                        << " #providers=" << np);
                _publish(lock);
            }
        }
    }
}

void ISDynAnyNGCont::_publish(std::lock_guard<std::mutex> const& /*ensure_lock*/)
{
    unsigned numProviders = std::count_if(
        m_contributions.cbegin(), m_contributions.cend(),
        [](auto& c) {return c.info != nullptr; });

    if (numProviders > 1) {
        // if some provider is missing for few cycles presume it's dead
        for (auto&& contribution: m_contributions) {
            if (contribution.info == nullptr) continue;
            if (contribution.nUpdatesSinceLastUpdate >= m_missingCycles * (numProviders - 1)) {
                contribution.info.reset(nullptr);
            }
        }
    }

    // Sum all contributions.
    ISInfoDynAny sum;
    Annotations annotations;
    int count = 0;
    for (auto&& contribution: m_contributions) {
        if (contribution.info == nullptr) continue;
        try {
            // try to sum all of them but skip failing ones
            if (count == 0) {
                // make a copy
                sum = *contribution.info;
            } else {
                // this can throw, will be handled by caller
                m_alg->operate(*contribution.info, sum);
            }
            annotations.merge(contribution.info->annotations());
            ++count;
        } catch (const std::exception& ex) {
            ERS_LOG("Warning: Caught exception while merging ISInfoDynAny " << m_myName
                    << " exception was " << ex.what());
        }
    }

    // reset it even if publisher raises an exception
    m_numUpdatesSincePublish = 0;
    for (auto&& contribution: m_contributions) {
        if (contribution.nUpdatesSincePublish > 0) {
            // to avoid some aliasing behavior we reset this counter for all
            // contributions that were updated in this cycle.
            contribution.nUpdatesSinceLastUpdate = 0;
        }
        contribution.nUpdatesSincePublish = 0;
    }

    if (count > 0) {

        // update annotations
        annotations.update(sum.annotations(), count);
        try {
            //publish info
            _publish(sum, m_myName, count);
        } catch (const std::exception& ex) {
            ERS_LOG("Warning: Caught exception while merging ISInfoDynAny " << m_myName
                    << " exception was " << ex.what());
        }
    }
}

// Publish result of a summing
void ISDynAnyNGCont::_publish(ISInfoDynAny& info, std::string const& name, int nUpdates)
{
    // publish sum first
    if (m_publishSum) {
        m_publisher.publish(info, name + m_suffixSum, nUpdates);
    }

    if (m_publishAvg) {
        // scale every numeric (or numeric vector) attribute
        if (nUpdates > 0) {
            double factor = 1.0 / nUpdates;
            detail::scaleInfo(info, factor);
        }
        // publish scaled info
        m_publisher.publish(info, name + m_suffixAvg, nUpdates);
    }
}
