"""Module which defines Olc2Hlt class and related methods.
"""

from __future__ import absolute_import, print_function, division

#--------------------------------
#  Imports of standard modules --
#--------------------------------
import logging
import subprocess
import six
if six.PY2:
    import Queue as queue
else:
    import queue
import threading
import time

#-----------------------------
# Imports for other modules --
#-----------------------------
import ispy

#----------------------------------
# Local non-exported definitions --
#----------------------------------
_LOG = logging.getLogger("olc2hlt")

#------------------------
# Exported definitions --
#------------------------


class Olc2Hlt(object):
    """Main application class of olc2hlt application.

    Parameters
    ----------
    inputQueue : `Queue.Queue`
        Queue for input data (source IS objects).
    dstPart : str
        Name of the partition containg IS server for output data.
    dstInfo : str
        Name of the output IS information object.
    decider :
        Object responsible for update decisions
    noCTPUpdate : bool
        If true then pass -N option to ctpUpdateCommand
    receiver : ISReceiver
        Instance of ISReceiver class
    """

    ISObjectClass = ispy.ISObject  # can be changed for testing
    process_exec = subprocess.check_output  # can be changed for testing

    def __init__(self, inputQueue, dstPart, dstInfo, decider, noCTPUpdate, receiver):

        self.inputQueue = inputQueue
        self.dstPart = dstPart
        self.dstInfo = dstInfo
        self.decider = decider
        self.noCTPUpdate = noCTPUpdate
        self.receiver = receiver

        self.wakeupInterval = 10  # in seconds
        self.lastWakeup = 0

        self.thread = None

    def run(self):
        """
        This is main method of the whole application, it runs the whole shebang
        and never returns.
        """

        # tell receiver to start getting data from IS
        self.receiver.subscribe()

        # run (forever)
        self._loop()

        # tell receiver to stop receiving data
        self.receiver.unsubscribe()

    def start(self):
        """
        Start event loop in a separate thread and return.
        """

        # tell receiver to start getting data from IS
        self.receiver.subscribe()

        # start a thread
        self.thread = threading.Thread(target=self._loop)
        self.thread.start()
        _LOG.info('Started background thread')

    def stop(self):
        """
        Stop a thread with event loop.
        """

        if not self.thread:
            _LOG.warning('Background thread is not running')
            return

        # tell receiver to stop receiving data
        self.receiver.unsubscribe()

        # put special token into queue to signal that no more items is expected
        item = None
        _LOG.info('Adding stop flag to input queue')
        self.inputQueue.put(item, True)

        # wait for the thread to finish
        _LOG.info('Waiting the background thread to stop')
        self.thread.join()
        self.thread = None
        _LOG.info('Background thread finished')

    def _loop(self):

        # find out what was the last data that we sent to HLT
        oldData = self._restoreData()

        # loop and wait for data
        while True:

            try:
                item = self.inputQueue.get(block=True, timeout=self.wakeupInterval)
            except queue.Empty:
                # on timeout just call wakeup
                self._wakeups()
                continue

            # someone could push None into queue to force clean stop
            if item is None:
                self.inputQueue.task_done()
                break

            # item is a tuple of IS reason and IS data
            reason, data = item

            # if IS info or server is gone do not do anyhting
            if reason == ispy.Reason.Deleted:
                _LOG.warning('Source IS info object has disappeared, will wait until it returns')
                self.inputQueue.task_done()
                continue

            _LOG.info('Received new LB: run=%s LBN=%s lumi=%s', data.RunNumber, data.LBN, data.LBAvInstLumi)

            # see if data has changed
            if self.decider.decide(oldData, data):
                # update HLT and remember it
                if self._sendToHLT(data):
                    oldData = data

            self.inputQueue.task_done()

            # do wakeups if necessary
            self._wakeups()

    def _restoreData(self):
        '''
        Try to restore data most recently sent to HLT. There could be several
        options for that:
        1. look at the conditions data that CTP handler stored most recently,
           for that we need to know what kind of intervals CTP handler creates
        2. look at IS data in the destination IS info object (if it exists),
           not 100% reliable because CTP can refuse to use that data
        3. store most recent data in some external source (file) and recover
           from there

        It's actually not clear yet if this needs to be done at all, so for now
        we do not restore anything which means that we are forcing update at
        the start.
        '''

        return None

    def _wakeups(self):
        '''
        Do some periodic work.
        '''

        now = time.time()
        if now - self.lastWakeup > self.wakeupInterval - 0.01:

            # currently there is nothing to do

            # wakeups can take some time, do not count that
            self.lastWakeup = time.time()

    def _sendToHLT(self, data):
        '''
        Stores the data in destination IS server and notifies CTP.
        '''

        try:
            isobj = self.ISObjectClass(self.dstPart, self.dstInfo, 'OnlPrefLumi')
        except UserWarning as exc:
            _LOG.warning('Problem with server for publishing to IS: %s', exc)
            return False

        try:

            # fill all data members
            isobj.RunNumber = data.RunNumber
            isobj.LBN = data.LBN
            isobj.StartTime = data.StartTime
            isobj.EndTime = data.EndTime
            isobj.Channel = data.Channel
            isobj.AlgorithmID = data.AlgorithmID
            isobj.LBAvInstLumi = data.LBAvInstLumi
            isobj.LBAvEvtsPerBX = data.LBAvEvtsPerBX
            isobj.LumiType = data.LumiType
            isobj.Valid = data.Valid
            isobj.BunchInstLumi = data.BunchInstLumi

            # publish
            isobj.checkin()

            _LOG.info('Publishing OnlPrefLumi to IS: RunNumber=%s LBN=%s Channel=%s '
                      'LBAvInstLumi=%s LBAvEvtsPerBX=%s Valid=%s',
                      isobj.RunNumber, isobj.LBN, isobj.Channel, isobj.LBAvInstLumi,
                      isobj.LBAvEvtsPerBX, isobj.Valid)

        except Exception as exc:
            _LOG.warning('Problem publishing status info to IS: %s', exc)
            return False

        # notify CTP, just run an external application which does whole job
        cmd = ['ctpUpdateCommand', '-i', 'luminosity']
        if self.noCTPUpdate:
            cmd += ['-N']
        try:
            _LOG.debug('CTP update command: %s', cmd)
            out = self.process_exec(cmd, stderr=subprocess.STDOUT)
            _LOG.info('CTP updated successfully')
            _LOG.debug('CTP update output: %s', out)
        except subprocess.CalledProcessError as exc:
            _LOG.error('CTP update failed: Command %s returned error code %s', cmd, exc.returncode)
            _LOG.error('Command output:\n\t%s', exc.output.strip().replace('\n', '\n\t'))
            return False

        return True
