#!/usr/bin/env tdaq_python

"""Unit test for StableBeamsFlag class.
"""

from __future__ import absolute_import, print_function, division

from contextlib import contextmanager
import six
if six.PY2:
    import Queue as queue
else:
    import queue
import unittest

from ispy import Reason
from pyolc2hlt import Decider, ISReceiver, Olc2Hlt, StableBeamsFlag, StableDecider, ThresholdDecider
from _mocks import MockISDynAny, MockISInfoReceiver, MockOnlPrefLumi, MockISObjectStableBeams


@contextmanager
def runOlc2Hlt(olc2hlt):
    """Stop Olc2Hlt even if test generates exception"""
    olc2hlt.start()
    try:
        yield olc2hlt
    except Exception:
        raise
    finally:
        olc2hlt.stop()


class MockExec(object):

    def __init__(self):
        self.args = None

    def __call__(self, args, **kw):
        self.args = args
        return "OK"


class MockISObject(object):

    checked_in = []

    def __init__(self, partName, infoName, infoType):
        self.partName = partName
        self.infoName = infoName
        self.infoType = infoType

    def checkin(self):
        MockISObject.checked_in.append(self)


class TestCase(unittest.TestCase):

    def setUp(self):

        Olc2Hlt.ISObjectClass = MockISObject
        Olc2Hlt.process_exec = MockExec()
        StableBeamsFlag.ISObjectClass = MockISObjectStableBeams
        ISReceiver.ISInfoReceiverClass = MockISInfoReceiver
        ISReceiver.ISInfoDynAnyClass = MockISDynAny

        lumiChangePercent = 10.
        lumiChangePercentUnstable = 25.
        self.unstableDeadBand = 0.01
        self.minLBLength = 10
        sbeamFlag = StableBeamsFlag("initial", "LHC.StableBeamsFlag")
        stableDecider = StableDecider(lumiChangePercent)
        unstableDecider = ThresholdDecider(self.unstableDeadBand, lumiChangePercentUnstable)
        self.decider = Decider(stableDecider, unstableDecider, self.minLBLength, sbeamFlag)

        self.infoName = "OLC2COOL.OnlPrefLumi"
        self.run = 222222
        self.lbn = 0
        self.time = 1500000000 * 1000000000
        self.queue = queue.Queue()

    def _makeISInfo(self, lumi, lb_len=60, channel=0, algoID=0, lumiType=0, valid=0):

        self.lbn += 1
        endTime = self.time + lb_len * 1000000000
        per_bunch_lumi = lumi / 5.
        info = MockOnlPrefLumi(
            RunNumber=self.run,
            LBN=self.lbn,
            StartTime=self.time,
            EndTime=endTime,
            Channel=channel,
            AlgorithmID=algoID,
            LBAvInstLumi=lumi,
            LBAvEvtsPerBX=per_bunch_lumi,
            LumiType=lumiType,
            Valid=valid,
            BunchInstLumi=[0., per_bunch_lumi, 0, per_bunch_lumi, 0., per_bunch_lumi,
                           0., per_bunch_lumi, 0., per_bunch_lumi],
        )
        self.time = endTime
        return info

    def _publish(self, data):

        for reason, isinfo in data:
            self.mock_is_recv.receive(self.infoName, isinfo, isinfo.EndTime + 1000000, reason=reason)
        # synchronize with processing
        self.queue.join()

    def test01(self):

        dstPart, dstInfo = "", "OLC2HLT.OnlPrefLumi"
        noCTPUpdate = False
        recv = ISReceiver("OLC", self.infoName, self.queue)
        olc2hlt = Olc2Hlt(self.queue, dstPart, dstInfo, self.decider, noCTPUpdate, recv)

        with runOlc2Hlt(olc2hlt):

            # recv.receiver is only defined after subscribe()
            self.mock_is_recv = recv.receiver

            #  tuples (stableBeams, reason, lumiValue, expectQueueSize, expectValue, expectLBN, kw)
            updates = [
                (False, Reason.Subscribed, 0.001, 1, 0.001, 1, {}),

                # 0 is consistent with 0.001
                (False, Reason.Updated, 0.0, 1, 0.001, 1, {}),

                # LB is too short, so no update
                (False, Reason.Updated, 0.1, 1, 0.001, 1, dict(lb_len=self.minLBLength / 2)),

                # jump above thresold, still non-stable beams
                (False, Reason.Updated, 0.1, 2, 0.1, 4, {}),

                # relative change 24%, no update
                (False, Reason.Updated, 0.124, 2, 0.1, 4, {}),

                # relative change 26%, will update
                (False, Reason.Updated, 0.126, 3, 0.126, 6, {}),

                # drop to almost zero
                (False, Reason.Updated, 0.001, 4, 0.001, 7, {}),

                # jump again
                (False, Reason.Updated, 100., 5, 100., 8, {}),

                # go to stable beams
                # small change but should trigger update
                (True, Reason.Updated, 101., 6, 101., 9, {}),

                # small change, no update
                (True, Reason.Updated, 110., 6, 101., 9, {}),

                # >10% change
                (True, Reason.Updated, 112., 7, 112., 11, {}),

                # <10% drop, no update
                (True, Reason.Updated, 101., 7, 112., 11, {}),

                # >10% drop, should update
                (True, Reason.Updated, 100., 8, 100., 13, {}),

                # Drop to 0
                (True, Reason.Updated, 0., 9, 0., 14, {}),

                # Any change from 0 should update
                (True, Reason.Updated, 1e-10, 10, 1e-10, 15, {}),

                # back to non-stable, below threshold
                (False, Reason.Updated, 0.001, 10, 1e-10, 15, {}),

                # Info object is deleted, no updates
                (False, Reason.Deleted, 100., 10, 1e-10, 15, {}),

                # Info object is re-created
                (False, Reason.Created, 100., 11, 100., 18, {}),
            ]

            self.assertEqual(len(MockISObject.checked_in), 0)

            for stableBeams, reason, lumiValue, expectQueueSize, expectValue, expectLBN, kw in updates:
                MockISObjectStableBeams.stableBeams = stableBeams
                self._publish([(reason, self._makeISInfo(lumiValue, **kw))])
                self.assertEqual(len(MockISObject.checked_in), expectQueueSize)
                info = MockISObject.checked_in[-1]
                self.assertEqual(info.LBN, expectLBN)
                self.assertEqual(info.LBAvInstLumi, expectValue)

            # all command lines should be identical, and we have at least one called
            self.assertEqual(Olc2Hlt.process_exec.args, ['ctpUpdateCommand', '-i', 'luminosity'])


if __name__ == "__main__":
    unittest.main()
