#!/usr/bin/env tdaq_python
"""
 Modify an existing OKS database.

 Instructions:

    pm_set.py [-r|--reload] [<backend>:]<oks_database> [<variable_file>]

 <backend> should be either "oksconfig" (for files) or "rdbconfig" (for RDB
 servers) if provided. If you omit it, "oksconfig" is assumed.

 <variable_file> can be unset, in which case pm_set.py will read from the
 standard input. All usual techniques streaming input there will work as
 expected. A few examples:

    $ echo "uid@class.relation = xyz" | pm_set.py database.data.xml

    $ pm_set.py database.data.xml <<EOF
    uid@class.relation = xyz
    EOF

 <variable_file> should be in the following syntax:

 Empty lines or lines starting with # are ignored

 To change the attribute of a single object:

    uid@class.attribute = value

 To change the attribute of all objects of a certain class:

    @class.attribute = value

 To set a relation for one or all objects of a certain class:

    uid@class.relation = uid@class
    @class.relation    = uid@class
    
 To add an attribute or a relation to an object or all objects of a class:

    uid@class.attribute += [ value ]
    uid@class.attribute += [ value1, value2 ]
    uid@class.relation  += [ uid@class ]
    uid@class.relation  += [ uid1@class, uid2@class ]
    @class.relation     += [ uid@class ]
 etc.

 To change the attribute or relation of a set of objects matching
 a regular expression:

    ~regexp@class.attribute = value

 To create a new object:

    uid@class

 Examples:

    # Change all objects of this type

    @L2SVConfiguration.l2puTimeout_ms = 3000
    @PTTemplateApplication.Instances  = 8

    # Change specific objects only

    L2PU->ROS@InterfaceMap.Protocol = 'udp'
    ROS->L2PU@InterfaceMap.Protocol = 'udp'

    HLT_COMMON_PATCH_AREA@Variable.Value = [ '/home/rhauser/work', '/sw/atlas/AtlasPoint1/14.0.10.1' ]

    # Set relationsihps
    @L2PUTemplateApplication.ProcessEnvironment += [ HLT_COMMON_PATCH_AREA@Variable ]

    # Create new object, if it does not exist
    Preload@Variable

    # Modify it
    Preload@Variable.Name  = 'HLT_PRELOAD'
    Preload@Variable.Value = 'libGaudiKernel.so:libGaudiSvc.so'

    @L2PUTemplateApplication.ProcessEnvironment += [ Preload@Variable ]

"""
from __future__ import print_function

import ipc
import config
import re

dbg = False

def modify(db, parameters, includes, new_db):
    """
    Modify the OKS database 'db' with the paramaters from the opened file 
    'parameters'. Addes the files in 'includes'. If new_db is true, create
    a new database from scratch.
    """
    if db[:9] not in ('oksconfig', 'rdbconfig'): db = 'oksconfig:' + db

    if new_db:
        conf = config.Configuration()
        conf.create_db(db.split(':',1)[1], includes, 'oksconfig')
    else:
        conf = config.Configuration(db)
        for inc in includes:
            conf.add_include(inc)

    modified = []

    for line in parameters.readlines():
        line = line.strip()

        if dbg: 
            print('line =',line)

        if len(line) == 0 or line[0] == '#':
            continue

        rhash = line.rfind('#')
        if rhash > 0:
            line = line[:rhash]

        add    = False
        remove = False
        
        if line.find('+=') != -1:
            (left, right) = line.split('+=')
            add = True
        elif line.find('-=') != -1:
            (left, right) = line.split('-=')
            remove = True
        elif line.find('=') != -1:
            (left, right) = line.split('=',1)
        else:
            # Creation of new objects
            left = line.strip()
            right = ''

        if dbg:
            print('left =',left, '\nright =',right)

        (uid,classattr) = left.split('@')

        rs = right.strip()
        rs = rs.lstrip('[')
        rs = rs.rstrip(']')
        rlist  = rs.split(',')
        if dbg:
            print('rs     =',rs)
            print('rlist  =',rlist)

        if len(right) > 0:
            (class_name, attr) = classattr.split('.')
            attr  = attr.strip()
            right = right.strip()

        else:
            # create new object
            try:
               conf.get_obj(classattr, uid)
            except:
               conf.create_obj(classattr, uid)
            continue
        
        if dbg:
            print(uid,class_name,attr)


        if len(uid) == 0:
            objects = conf.get_dals(class_name)
        elif uid.startswith('~'):
            objects = [ k for k in conf.get_dals(class_name) if re.match(uid[1:], k.id) ]
        else:
            objects = [ conf.get_dal(class_name, uid) ]
            
        # 
        if dbg:
            print('objects =',objects)
            print('add     =',add)
            print('remove  =',remove)
        
        for obj in objects:
            if add:
                x = getattr(obj, attr)
                to_add = []
                for r in rlist:
                    t =  eval(re.sub('([^\s]+)@([^\s]+)','conf.get_dal("\\2","\\1")',r))
                    to_add.append(t)
                for i in to_add:
                    if i in x:
                        x.remove(i)
                x += to_add
                setattr(obj, attr, x)
            elif remove:
                try:
                    x = getattr(obj, attr)
                    for r in rlist:
                        x.remove(eval(re.sub('([^\s]+)@([^\s]+)','conf.get_dal("\\2","\\1")',r)))
                    setattr(obj, attr, x)
                except:
                    pass
            else:
                to_set = []
                for r in rlist:
                    t =  eval(re.sub('([^\s]+)@([^\s]+)','conf.get_dal("\\2","\\1")',r))
                    to_set.append(t)
                try:
                    setattr(obj, attr, to_set)
                except ValueError as a:
                    setattr(obj, attr, to_set[0])
            # print obj
            conf.update_dal(obj)

    conf.commit()
                
if __name__ == '__main__':
  import sys
  from EventApps import myopt

  option = {}
  option['reload'] = {'short': 'r', 'arg': False, 'default': None,
                      'description': 'issues a full database reload after a commit'}

  option['extra-includes'] = { 'short': 'I', 'arg': True, 'default': [],
                               'description': 'Extra OKS include file' }

  option['new'] = { 'short': 'n', 'arg': False, 'default': None,
                    'description': 'create a new database from scratch (only useful with appropriate input to -I)' }

  parser = myopt.Parser(extra_args=True, post_options=__doc__)
  for (k,v) in option.items():
    parser.add_option(k, v['short'], v['description'], v['arg'], v['default'])

  # process arguments 
  (kwargs, extra) = parser.parse(sys.argv[1:], prefix='"%s" options:' % sys.argv[0])

  if len(extra) == 0 or len(extra) > 2:
    print(parser.usage())
    sys.exit(1)

  if len(extra) == 1: input1 = sys.stdin
  else: input1 = open(extra[1], 'rt')

  # print 'extra[0]                 ', extra[0]
  # print 'input1                   ',input1
  # print "kwargs['extra-includes'] ",kwargs['extra-includes']
  # print "kwargs['new']            ",kwargs['new']

  modify(extra[0], input1, kwargs['extra-includes'], kwargs['new'])

  # if the user asks, we issue an db reload, if the server is an RDB server.
  if kwargs['reload']:
    import re, subprocess, logging
    r = re.search(r'rdbconfig:(?P<server>.+)@(?P<partition>.+)', extra[0])
    if not r:
      logging.error('Can only reload RDB servers and "%s" does not look like one. Ignoring user request.' % extra[0])
    else:
      cmd = ['rdb_admin', '--partition', r.group('partition'),
             '--database', r.group('server'), '--reload', 'partition']
      status = subprocess.check_call(cmd)
