package pmgClient;


/**
 * The class for the representation of Process Manager Handles
 * <p>
 * Handles are in the following URL format: <b>pmg://server_name:port/partition/application/id</b>  <br>
 * The ":port" part of the URL may be absent, in case where the default port is being used ( @see constructor of {@link PmgClient} )  .
 * 
 */
public final class Handle {
    private final String server;							// server host the process
    private final int port;                                 // IP port of CORBA sever
    private final String partitionName;                     // Partition of the process
    private final String appName;                           // Application name as defined by the client
    private final long  uniqueId;                           // Application unique identity (generated by the server)
    
    public static final String PREFIX = "pmg://";			// The handle prefix
    public static final String SEPARATOR = "/";				// The handle section separator
    public static final String PORT_SEPARATOR = ":";		// The separator which identifies the port number
    public static final int DEFAULT_PORT_MARK = 0;			// The default port number
    
    
    /**
     * Constructs a Handle object from the given parameters. 
     * <p>
     * This constructor is the simplest and should be preferred, if this demands no extra work.
     * In practice though, it will be rarely (if ever) used.
     * 
     * @param server Server's name
     * @param port Server's port, use Handle.DEFAULT_PORT_MARK for the default one
     * @param partitionName Name of the partition
     * @param appName Name of the application
     * @param uniqueId Unique ID of the specific application (given by the server who launched that process)
     */
	public Handle(String server, int port, String partitionName, String appName, long uniqueId) {
		this.server = server;
		this.port = port;
		this.partitionName = partitionName;
		this.appName = appName;
		this.uniqueId = uniqueId;
	}
	
	/**
	 * Constructs a Handle object from a string with the expected structure.
	 * <p>
	 * The structure must be in the following format (without the spaces):<br>
	 * 		PREFIX server_name PORT_SEPARATOR port SEPARATOR partition SEPARATOR applicationName SEPARATOR uniqueId <br>
	 * e.g. pmg://server_name:portNUM/partition/application/idNUM <br>
	 * "PORT_SEPARATOR port" are optional, in case the default port is used
	 * 
	 * @param handle String representation of the Handle in the format of a url as described above
	 * @throws InvalidHandleStringException If the given string is not in on of the expected formats
	 */
	public Handle(final String handle) throws InvalidHandleStringException {
		
		if( handle == null || handle.equals("")) {
			throw new InvalidHandleStringException("Handle string cannot be empty");
		}
		
		String noPrefix;
		try {
			noPrefix = handle.substring(PREFIX.length());
		} catch(IndexOutOfBoundsException e) {
			throw new InvalidHandleStringException("Handle is invalid (check prefix)", e);
		}
		
		//handle can contain ':' in the app name, e.g. pmg://pc-tbed-pub-05.cern.ch/part_hlt_lpapaevg/HLT:pc-tbed-pub-05/1
		String hostnameAndPort = noPrefix.substring(0, noPrefix.indexOf(SEPARATOR));	// get the "server_name:port"  OR  "server_name"  part
		int portSep = hostnameAndPort.indexOf(PORT_SEPARATOR);		
		
		String[] tokens;
		
		if( portSep != -1 ) {	// if port not omitted. format pmg://server_name:port/partition/application/idNUM
			this.server = noPrefix.substring(0, portSep);
			tokens = (noPrefix.substring(this.server.length()+1)).split(SEPARATOR);
			try {
				this.port = Integer.parseInt(tokens[0]);
			} catch(NumberFormatException ex ){
				throw new InvalidHandleStringException("PORT must be an integer", ex);
			}
		}
		else {					// else format pmg://server_name/partition/application/idNUM
			this.port = DEFAULT_PORT_MARK;
			tokens = noPrefix.split(SEPARATOR);
			this.server = tokens[0];
		}
		
		this.partitionName = tokens[1];
		this.appName = tokens[2];
		try {
			this.uniqueId = Integer.parseInt(tokens[3]);
		} catch(NumberFormatException ex ) {
			throw new InvalidHandleStringException("UniqueID must be an integer", ex);
		}
	}

	
	/**
	 * Returns a String representation of the Handle ( prefix included, e.g. "pmg://...").
	 */
	@Override
	public String toString(){
		return PREFIX + fullPath();
	}

    /**
     * It gets the path of the Handle as a String (no prefix included, useful for files).
     *
     * @return The path of the handle as a String.
     */
    public String fullPath() {
    	return this.server + (DEFAULT_PORT_MARK==this.port ? "" : Handle.PORT_SEPARATOR + this.port) + Handle.SEPARATOR
    		+ this.partitionName + Handle.SEPARATOR + this.appName + Handle.SEPARATOR + this.uniqueId;    	
    }
	
	/**
	 * 
	 * @return The String representation of the server for this Handle object.
	 */
	public String getServer() {
		return this.server;
	}


	/**
	 * 
	 * @return The port number of this Handle.
	 */
	public int getPort() {
		return this.port;
	}


	/**
	 * 
	 * @return The partition of this Handle.
	 */
	public String getPartitionName() {
		return this.partitionName;
	}


	/**
	 * 
	 * @return The application name of this Handle as a String.
	 */
	public String getAppName() {
		return this.appName;
	}


	/**
	 * 
	 * @return The unique ID of this Handle that was generated by the server.
	 */
	public long getUniqueId() {
		return this.uniqueId;
	}
	

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((this.appName == null) ? 0 : this.appName.hashCode());
		result = prime * result + ((this.partitionName == null) ? 0 : this.partitionName.hashCode());
		result = prime * result + this.port;
		result = prime * result + ((this.server == null) ? 0 : this.server.hashCode());
		result = prime * result + (int) (this.uniqueId ^ (this.uniqueId >>> 32));
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Handle other = (Handle) obj;
		if (this.appName == null) {
			if (other.appName != null)
				return false;
		} else if (!this.appName.equals(other.appName))
			return false;
		if (this.partitionName == null) {
			if (other.partitionName != null)
				return false;
		} else if (!this.partitionName.equals(other.partitionName))
			return false;
		if (this.port != other.port)
			return false;
		if (this.server == null) {
			if (other.server != null)
				return false;
		} else if (!this.server.equals(other.server))
			return false;
		if (this.uniqueId != other.uniqueId)
			return false;
		return true;
	}
}
