//   eff_daq_splunk.cc
//
//   Application to provide daq efficiency related data in std::out to evade the intermediate txt files problems
//
//   Author: Adrian.Chitan@cern.ch, March 2016
//   Modifications:
//  

#include <signal.h>
#include <fstream>
#include <vector>
#include <iostream>
#include <sstream>

#include <boost/program_options.hpp>

#include <CoralBase/Attribute.h>
#include <CoolKernel/Exception.h>
#include <CoolKernel/ValidityKey.h>
#include <CoolKernel/IDatabaseSvc.h>
#include <CoolKernel/IDatabase.h>
#include <CoolKernel/IFolder.h>
#include <CoolKernel/IRecordSpecification.h>
#include <CoolKernel/Record.h>
#include <CoolApplication/DatabaseSvcFactory.h>
#include <CoolKernel/IObjectIterator.h>
#include <CoolKernel/IObject.h>
#include <CoolKernel/ChannelSelection.h>

#include "owl/time.h"
#include "ers/ers.h"

#include "RunControl/Common/Exceptions.h"

namespace daq {
	ERS_DECLARE_ISSUE(      rc,
		BadCoolDb,
		"Retrieving data from COOL failed because " << explanation,
		((const char*) explanation)
		)
}

static     cool::IDatabasePtr db;
static     cool::IDatabasePtr dbd;
static     cool::IDatabasePtr dbt;
static     cool::IDatabasePtr dbm;

static     std::string FillState_Name = "/LHC/DCS/FILLSTATE";
static     std::string L1Counters_Name = "/TRIGGER/LUMI/LVL1COUNTERS";
static     std::string L1Menu_Name = "/TRIGGER/LVL1/Menu";
static     std::string LB_Name = "/TRIGGER/LUMI/LBLB";
static     std::string LBT_Name = "/TRIGGER/LUMI/LBTIME";
static     std::string DataTakingMode_Name = "/TDAQ/RunCtrl/DataTakingMode";
static     std::string BusyConf_Name = "/TRIGGER/LVL1/BUSYCONF";
static     std::string BusyRate_Name = "/TRIGGER/LVL1/BUSYRATES_RUN2";
static     std::string StopReason_Name = "/TDAQ/RunCtrl/StopReason";
static     std::string HoldTrigger_Name = "/TDAQ/RunCtrl/HoldTriggerReason";
static	   std::string t0Project_Name = "/TDAQ/RunCtrl/SOR";
static     std::string TotalInt_Name = "/LHC/DCS/BPTX/TOTALINT";

static     cool::IFolderPtr FillState_Folder;
static     cool::IFolderPtr L1Counters_Folder;
static     cool::IFolderPtr L1Menu_Folder;
static     cool::IFolderPtr LB_Folder;
static     cool::IFolderPtr LBT_Folder;
static     cool::IFolderPtr DataTakingMode_Folder;
static     cool::IFolderPtr BusyConf_Folder;
static     cool::IFolderPtr BusyRate_Folder;
static     cool::IFolderPtr StopReason_Folder;
static     cool::IFolderPtr HoldTrigger_Folder;
static	   cool::IFolderPtr t0Project_Folder;
static     cool::IFolderPtr TotalInt_Folder;

static     std::map<std::string, std::string> ctpMap;
static     std::map<std::string, double> busyMap;
static	   std::map<std::string, double> stopSystemsMap;
static	   std::map<std::string, double> stopReasonMap;
static	   std::map<std::string, double> holdTriggerSystemsMap;
static	   std::map<std::string, double> holdTriggerReasonMap;

static     std::string triggerItem = "L1_EM30";
static	   std::string connName = "oracle://ATLR";
static	   std::string timestampPath = "";//"/tmp/";
static     int atlasRun = 2;
static     int noBeam = 0;
static     int useTotalBusy = 1;
static	   int getTotalBeam = 0;
//static	   int useNanosec = 0;

static std::vector<long long> beams_start_vector;
static std::vector<long long> beams_end_vector;
static std::vector<long long> start_of_run_vector;
static std::vector<long long> end_of_run_vector;
static std::vector<int> beamBuffer;

extern "C" void signalhandler(int)
{
	signal(SIGINT,  SIG_DFL);
	signal(SIGTERM, SIG_DFL);
	//std::cout << "Terminating eff_run_efficiency ..." << std::endl;
}

// throws daq::rc::BadCoolDb
void openDB(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

	try {
		db = db_Svc.openDatabase(db_Id, false); // false = !read_only
	}
	catch (std::exception &ex) {
		//std::cout << "Failure opening DB database - " << ex.what() << std::endl;
		daq::rc::BadCoolDb i(ERS_HERE, ex.what());
		throw i;
	}
}

// throws daq::rc::BadCoolDb
void openDBD(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

	try {
		dbd = db_Svc.openDatabase(db_Id, false); // false = !read_only
	} catch (std::exception &ex) {
		//std::cout << "Failure opening DCS database - " << ex.what() << std::endl;
		daq::rc::BadCoolDb i (ERS_HERE, ex.what());
		throw i;
	}
}

// throws daq::rc::BadCoolDb
void openDBT(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

	try {
		dbt = db_Svc.openDatabase(db_Id, false); // false = !read_only
	}
	catch (std::exception &ex) {
		//std::cout << "Failure opening DBT database - " << ex.what() << std::endl;
		daq::rc::BadCoolDb i(ERS_HERE, ex.what());
		throw i;
	}
}

// throws daq::rc::BadCoolDb
void openDBM(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

	try {
		dbm = db_Svc.openDatabase(db_Id, false); // false = !read_only
	} catch (std::exception &ex) {
		//std::cout << "Failure opening DBM database - " << ex.what() << std::endl;
		daq::rc::BadCoolDb i(ERS_HERE, ex.what());
		throw i;
	}
}

std::string get_date(int seconds) {
	OWLTime owl_date = OWLTime(seconds);
	long year = owl_date.year();
	// unsigned short month = owl_date.month()+1;
	unsigned short month = owl_date.month();
	unsigned short day = owl_date.day();
	std::string date_string = std::string(owl_date.c_str());
	size_t lastIndex = date_string.size();
	std::string time_string = date_string.substr(lastIndex-8, lastIndex);
	std::ostringstream stm;
	stm << year;
	stm << "-";
	if (month < 10) {
		stm << "0";
	}
	stm << month << "-";
	if (day < 10) {
		stm << "0";
	}
	stm << day;
	stm << " ";
	stm << time_string;
	return stm.str();
}

std::string get_date_OWL(int seconds) {
	OWLTime owl_date = OWLTime(seconds);
	long year = owl_date.year();
	// unsigned short month = owl_date.month()+1;
	unsigned short month = owl_date.month();
	unsigned short day = owl_date.day();
	std::string date_string = std::string(owl_date.c_str());
	size_t lastIndex = date_string.size();
	std::string time_string = date_string.substr(lastIndex-8, lastIndex);
	std::ostringstream stm;
	if (day < 10) {
		stm << "0";
	}
	stm << day << "/";
	if (month < 10) {
		stm << "0";
	}
	stm << month << "/";
	stm << year;	

	stm << " ";
	stm << time_string;
	return stm.str();
}

int get_ctpid(cool::ValidityKey sKey, cool::ValidityKey uKey, std::string itemName, cool::IFolderPtr & lvl1menu_folder) {
	int itemNo = -1;
	cool::IObjectIteratorPtr objects = lvl1menu_folder->browseObjects(sKey, uKey, cool::ChannelSelection::all());
	while ( objects->goToNext() ) {
		const cool::IObject& obj = objects->currentRef();
		std::string item_name = obj.payloadValue("ItemName");
		if (item_name.compare(itemName) == 0) {
			itemNo = obj.channelId();
			break;
		}
	}
	return itemNo;
}

double eff_fraction(cool::ValidityKey key, int ctpid, cool::IFolderPtr & l1counters_folder) {
	// for the moment overflow flags are not used
	cool::IObjectPtr obj_l1counters = l1counters_folder->findObject(key, ctpid);
	std::string l1Accept = obj_l1counters->payloadValue("L1Accept");
	std::string afterPrescale = obj_l1counters->payloadValue("AfterPrescale");
	long long l1_accept = atoll(l1Accept.c_str());
	long long after_prescale = atoll(afterPrescale.c_str());
	double fraction = 1;
	if (after_prescale > 0) {
		fraction = (double) (l1_accept) / (double) after_prescale;
	}
	return fraction;
}

long long l1_accepted(cool::ValidityKey key, int ctpid, cool::IFolderPtr & l1counters_folder) {
	// for the moment overflow flags are not used
	cool::IObjectPtr obj_l1counters = l1counters_folder->findObject(key, ctpid);
	std::string l1Accept = obj_l1counters->payloadValue("L1Accept");
	long long l1_accept = atoll(l1Accept.c_str());
	return l1_accept;
} 

float busy_run2(cool::ValidityKey key, std::string identifier, cool::IFolderPtr & the_busy_folder ) {

	// Find the busy values for a specific detector in this lumiblock:

	cool::IObjectPtr objects = the_busy_folder->findObject(key, 0);
	// const cool::IObject& obj = objects->currentRef();
	std::string this_busy_string= objects->payloadValue(identifier.c_str());
	float this_busy = atof(this_busy_string.c_str());
	return this_busy;
}

double busy(long long startTime, long long endTime, std::string identifier, cool::IFolderPtr & the_busy_folder ) {

	// Find the busy values for a specific detector in this lumiblock:

	// Note: this also includes records that are partialy outside this lb

	cool::IObjectIteratorPtr objects = the_busy_folder->browseObjects( cool::ValidityKey(startTime),cool::ValidityKey(endTime), 0 );
	double totaltime=0;
	double totalbusy=0;

	while ( objects->goToNext() ) {
		const cool::IObject& obj = objects->currentRef();

		std::string this_busy_string= obj.payloadValue(identifier.c_str());
		double this_busy = atof(this_busy_string.c_str());

		double this_starttime = obj.since();
		double this_stoptime = obj.until();
		totalbusy+=this_busy*(this_stoptime-this_starttime);
		totaltime+=(this_stoptime-this_starttime);
	}
	if (totaltime>0) {
		totalbusy = 0.01* totalbusy/totaltime;
	} else {
		totalbusy = 0; // if no info, let's just say this detector was not busy
	}
	return totalbusy;
}

void prepareCtpMap(int runNumber) {
	// Populate ctp map for the start run (no change for the next run ...)
	ctpMap.clear();
	cool::UInt63 a = runNumber;
	cool::UInt63 z = runNumber + 1;
	cool::ValidityKey sKey = (a <<32) + 1;
	cool::ValidityKey uKey = z <<32;
	cool::IObjectIteratorPtr objs = BusyConf_Folder->browseObjects(sKey, uKey, cool::ChannelSelection::all());
	while ( objs->goToNext() ) {
		const cool::IObject& obj = objs->currentRef();
		std::string identifier = obj.payloadValue("identifier");
		std::string description = obj.payloadValue("description");
		std::string busyEnabled = obj.payloadValue("busyEnabled");
		//std::cout << identifier << "," << description << std::endl;
		if (busyEnabled.compare("TRUE") == 0) {
			ctpMap.insert(std::pair<std::string, std::string>(identifier, description) );
		}
	}
}

std::string get_stop_reason(long long starttime, long long endtime) {
	// starttime in msec
	long long start_time_nsec = ((long long)(starttime))*1000000;
	long long end_time_nsec = ((long long)(endtime))*1000000;
	cool::IObjectIteratorPtr sr_objects = StopReason_Folder->browseObjects(cool::ValidityKey(start_time_nsec), cool::ValidityKey(end_time_nsec), 0);
	std::ostringstream strm;
	while ( sr_objects->goToNext() ) {
		const cool::IObject& obj = sr_objects->currentRef();
		std::string subSystems = obj.payloadValue("SubSystems");
		std::string reason = obj.payloadValue("Reason");
		strm << subSystems << "/" << reason << ";";
	}
	return strm.str();
}

int get_fill_number(long long start_search_nsec, long long end_search_nsec, cool::IFolderPtr & FillState_Folder) {
	int fill_number = 0;
	cool::IObjectIteratorPtr objects = FillState_Folder->browseObjects( cool::ValidityKey(start_search_nsec),cool::ValidityKey(end_search_nsec), cool::ChannelSelection::all() );
	while ( objects->goToNext() ) {
		const cool::IObject& obj = objects->currentRef();
		std::string fillNumber = obj.payloadValue("FillNumber");
		fill_number = atoi(fillNumber.c_str());
	}
	return fill_number;
}

/*int get_stable_beams(long long start_search_nsec, long long end_search_nsec, cool::IFolderPtr& TotalInt_Folder) {
	int stable_beams = 0;
	cool::IObjectIteratorPtr objects = TotalInt_Folder->browseObjects( cool::ValidityKey(start_search_nsec),cool::ValidityKey(end_search_nsec), cool::ChannelSelection::all() );
	while ( objects->goToNext() ) {
		const cool::IObject& obj = objects->currentRef();
		std::string stableBeams1 = obj.payloadValue("Beam1_BeamPresent");
		std::string stableBeams2 = obj.payloadValue("Beam2_BeamPresent");
		if ((stableBeams1.compare("TRUE") == 0) && (stableBeams2.compare("TRUE") == 0)) {
			stable_beams = 1;
		}
	}
	return stable_beams;
}*/

int get_stable_beams(long long start_search_nsec, long long end_search_nsec, cool::IFolderPtr & FillState_Folder) {
	int stable_beams = 0;
	cool::IObjectIteratorPtr objects = FillState_Folder->browseObjects(cool::ValidityKey(start_search_nsec), cool::ValidityKey(end_search_nsec), cool::ChannelSelection::all());
	while (objects->goToNext()) {
		const cool::IObject& obj = objects->currentRef();
		std::string stableBeams = obj.payloadValue("StableBeams");
		if (stableBeams.compare("TRUE") == 0) {
			stable_beams = 1;
		}
	}
	return stable_beams;
}

long long get_beam_dump_time(long long start_search_nsec, long long end_search_nsec, cool::IFolderPtr& TotalInt_Folder) {
	long long beams_dumped = 0;
	cool::IObjectIteratorPtr objects = TotalInt_Folder->browseObjects(cool::ValidityKey(start_search_nsec), cool::ValidityKey(end_search_nsec), cool::ChannelSelection::all());
	while (objects->goToNext()) {
		const cool::IObject& obj = objects->currentRef();
		std::string stableBeams1 = obj.payloadValue("Beam1_BeamPresent");
		std::string stableBeams2 = obj.payloadValue("Beam2_BeamPresent");
		if ((stableBeams1.compare("FALSE") == 0) || (stableBeams2.compare("FALSE") == 0)) {
			beams_dumped = obj.since();
		}
	}
	return beams_dumped;
}

std::pair<std::string,std::string> get_beam_presence(long long start_search_nsec, long long end_search_nsec, cool::IFolderPtr& TotalInt_Folder) {
	std::pair<std::string,std::string> beams_presence;
	cool::IObjectIteratorPtr objects = TotalInt_Folder->browseObjects(cool::ValidityKey(start_search_nsec), cool::ValidityKey(end_search_nsec), cool::ChannelSelection::all());
	objects->goToNext();
	const cool::IObject& obj = objects->currentRef();
	std::string stableBeams1 = obj.payloadValue("Beam1_BeamPresent");
	std::string stableBeams2 = obj.payloadValue("Beam2_BeamPresent");
	beams_presence.first = stableBeams1;
	beams_presence.second = stableBeams2;
	
	return beams_presence;
}

int get_ready_for_physics(cool::ValidityKey key, cool::IFolderPtr & datatakingmode_folder) {
	int flag = 1;
	cool::IObjectPtr obj = datatakingmode_folder->findObject(key, 0);
	std::string readyForPhysics = obj->payloadValue("ReadyForPhysics");
	flag = atoi(readyForPhysics.c_str());
	return flag;
}

std::string get_t0project(cool::ValidityKey key, cool::IFolderPtr &t0project_folder) {
	std::string t0p = "";
	cool::IObjectPtr obj = t0project_folder->findObject(key, 0);
	t0p = obj->payloadValue("T0ProjectTag");
	if (t0p.compare("") == 0)
		t0p = "None";
	return t0p;
}


void openDatabasesAndFolders() {
	// open cool databases
	std::string coolDb;
	std::string coolDbTrig;
	std::string coolDbDCS;
	std::string coolDbMon;
	if (atlasRun == 1) {
		coolDbDCS = connName+";schema=ATLAS_COOLOFL_DCS;dbname=COMP200";
	} else if (atlasRun == 2) {
		coolDbDCS = connName+";schema=ATLAS_COOLOFL_DCS;dbname=CONDBR2";
	}
	cool::DatabaseId dbIdDCS = coolDbDCS;
	cool::IDatabaseSvc& dbSvcDCS = cool::DatabaseSvcFactory::databaseService();
	try {
		openDBD(dbIdDCS, dbSvcDCS);
	}
	catch(daq::rc::BadCoolDb& ex) {
		ers::fatal(ex);
		exit(0);
	}

	if (atlasRun == 1) {
		coolDb = connName+";schema=ATLAS_COOLONL_TDAQ;dbname=COMP200";
	} else if (atlasRun == 2) {
		coolDb = connName+";schema=ATLAS_COOLONL_TDAQ;dbname=CONDBR2";		
	}
	cool::DatabaseId dbId = coolDb;
	// Open or Create DB
	cool::IDatabaseSvc& dbSvc = cool::DatabaseSvcFactory::databaseService();
	try {
		openDB(dbId, dbSvc);
	} catch(daq::rc::BadCoolDb& ex) {
		ers::fatal(ex);
		exit(0);
	}

	if (atlasRun == 1) {
		coolDbTrig = connName+";schema=ATLAS_COOLONL_TRIGGER;dbname=COMP200";
	} else if (atlasRun == 2) {
		coolDbTrig = connName+";schema=ATLAS_COOLONL_TRIGGER;dbname=CONDBR2";
	}
	cool::DatabaseId dbIdTrig = coolDbTrig;
	cool::IDatabaseSvc& dbSvcTrig = cool::DatabaseSvcFactory::databaseService();

	try {
		openDBT(dbIdTrig, dbSvcTrig);
	}
	catch(daq::rc::BadCoolDb& ex) {
		ers::fatal(ex);
		exit(0);
	}

	if (atlasRun == 1) {
		coolDbMon = "oracle://ATLAS_COOLPROD;schema=ATLAS_COOLONL_TRIGGER;dbname=MONP200";
	} else if (atlasRun == 2) {
		coolDbMon = connName+";schema=ATLAS_COOLONL_TRIGGER;dbname=MONP200";
	}
	cool::DatabaseId dbIdMon = coolDbMon;
	cool::IDatabaseSvc& dbSvcMon = cool::DatabaseSvcFactory::databaseService();
	try {
		openDBM(dbIdMon, dbSvcMon);
	}
	catch(daq::rc::BadCoolDb& ex) {
		ers::fatal(ex);
		exit(0);
	}

	// check folders
	if (noBeam == 0) {
		try {
			if (!dbd->existsFolder(FillState_Name)) {
				//std::cout << "DB Folder FillState does not exist???" << std::endl;
			}
		} catch (std::exception &e) {
			//std::cout << "DB Folder FillState does not exist. " << e.what() << std::endl;
		}
	}
	if (!dbt->existsFolder(LB_Name)) {
		//std::cout << "DB Folder LB does not exist???" << std::endl;
	}
	if (!dbt->existsFolder(LBT_Name)) {
		//std::cout << "DB Folder LBT does not exist???" << std::endl;
	}
	if (!db->existsFolder(DataTakingMode_Name)) {
		//std::cout << "DB Folder DataTakingMode does not exist???" << std::endl;
	}
	try {
		if (!db->existsFolder(StopReason_Name)) {
			//std::cout << "DB Folder StopReason does not exist???" << std::endl;
		}
	} catch (std::exception &e) {
		//std::cout << "DB Folder StopReason does not exist. " << e.what() << std::endl;
	}
	try {
		if (!db->existsFolder(HoldTrigger_Name)) {
			//std::cout << "DB Folder HoldTrigger does not exist???" << std::endl;
		}
	} catch (std::exception &e) {
		//std::cout << "DB Folder HoldTrigger does not exist. " << e.what() << std::endl;
	}
	try {
		if (!dbm->existsFolder(BusyRate_Name)) {
			//std::cout << "DB Folder BusyRate does not exist???" << std::endl;
		}
	} catch (std::exception &e) {
		//std::cout << "DB Folder BusyRate does not exist. " << e.what() << std::endl;
	}
	try {
		if (!dbm->existsFolder(BusyConf_Name)) {
			//std::cout << "DB Folder BusyConf does not exist???" << std::endl;
		}
	} catch (std::exception &e) {
		//std::cout << "DB Folder BusyConf does not exist. " << e.what() << std::endl;
	}
	if (!dbt->existsFolder(L1Counters_Name)) {
		//std::cout << "DB Folder L1Counters does not exist???" << std::endl;
	}
	if (!dbt->existsFolder(L1Menu_Name)) {
		//std::cout << "DB Folder Menu does not exist???" << std::endl;
	}

	// open folders
	if (noBeam ==0) {
		try {
			FillState_Folder = dbd->getFolder(FillState_Name);
			TotalInt_Folder = dbd->getFolder(TotalInt_Name);
		}
		catch (std::exception &e) {
			//std::cout << "Cannot open FillState folder " << e.what() << std::endl;
			dbd->closeDatabase();
			//std::cout << "DBD closed." << std::endl;

			exit(0);
		}
	}
	LB_Folder = dbt->getFolder(LB_Name);
	LBT_Folder = dbt->getFolder(LBT_Name);
	DataTakingMode_Folder = db->getFolder(DataTakingMode_Name);
	try {
		StopReason_Folder = db->getFolder(StopReason_Name);
	} catch (std::exception &e) {
		//std::cout << "Cannot open StopReason folder " << e.what() << std::endl;
		db->closeDatabase();
		//std::cout << "DB closed." << std::endl;
		exit(0);
	}
	try {
		t0Project_Folder = db->getFolder(t0Project_Name);
	}
	catch (std::exception &e) {
		dbt->closeDatabase();
		exit(0);
	}
	try {
		HoldTrigger_Folder = db->getFolder(HoldTrigger_Name);
	} catch (std::exception &e) {
		//std::cout << "Cannot open HoldTrigger folder " << e.what() << std::endl;
		db->closeDatabase();
		//std::cout << "DB closed." << std::endl;
		exit(0);
	}
	try {
		BusyRate_Folder = dbm->getFolder(BusyRate_Name);
	}
	catch (std::exception &e) {
		//std::cout << "Cannot open BusyRate folder " << e.what() << std::endl;
		dbm->closeDatabase();
		//std::cout << "DBM closed." << std::endl;
		exit(0);
	}
	try {
		BusyConf_Folder = dbm->getFolder(BusyConf_Name);
	}
	catch (std::exception &e) {
		//std::cout << "Cannot open BusyConf folder " << e.what() << std::endl;
		dbm->closeDatabase();
		//std::cout << "DBM closed." << std::endl;
		exit(0);
	}
	L1Counters_Folder = dbt->getFolder(L1Counters_Name);
	L1Menu_Folder = dbt->getFolder(L1Menu_Name);
}

int get_beams_data(long long startTime, long long endTime, cool::IFolderPtr & folder) {
	beams_start_vector.clear();
	beams_end_vector.clear();
	cool::IObjectIteratorPtr objects = folder->browseObjects(cool::ValidityKey(startTime), cool::ValidityKey(endTime), cool::ChannelSelection::all());
	std::string old_beam = "FALSE";   bool beamStarted = false;
	while (objects->goToNext()) {
		const cool::IObject& obj = objects->currentRef();
		std::string beam = obj.payloadValue("StableBeams");
		if (((beam.compare("TRUE") == 0) && (old_beam.compare("FALSE") == 0)) ||
			((beam.compare("TRUE") == 0) && (old_beam.compare("NULL") == 0))) {
			// start of a new beam interval
			long long start_beam = (long long)(obj.since());
			if (start_beam < startTime) {
				start_beam = startTime;
			}
			beams_start_vector.push_back(start_beam / 1000000);
			//std::cout << start_beam / 10000000 << " ";
			old_beam = beam;
			beamStarted = true;
		}
		else if ((beam.compare("FALSE") == 0) && (old_beam.compare("TRUE") == 0)) {
			// end of a beam interval
			long long end_beam = (long long)(obj.since());
			beams_end_vector.push_back(end_beam / 1000000);
			//std::cout << end_beam / 10000000 << std::endl;
			beamStarted = false;
			old_beam = beam;
		}
		else {
			old_beam = beam;
		}
	}
	if (beamStarted) {
		beams_end_vector.push_back(endTime / 1000000);
		//std::cout << endTime / 10000000 << std::endl;
	}
	if (beams_start_vector.size())
		return 1;
	else
		return 0;
}

void get_beams_data_beamsON(long long startTime, long long endTime, cool::IFolderPtr & folder) {
	beams_start_vector.clear();
	beams_end_vector.clear();
	cool::IObjectIteratorPtr objects = folder->browseObjects( cool::ValidityKey(startTime),cool::ValidityKey(endTime), cool::ChannelSelection::all());
	std::string old_beam = "RAMP DOWN";   bool beamStarted = false; //"NO BEAM"
	while ( objects->goToNext() ) {
		const cool::IObject& obj = objects->currentRef();
		std::string beam = obj.payloadValue("BeamMode");
		if((beam.compare("RAMP DOWN") != 0) && (old_beam.compare("RAMP DOWN") == 0)) {
			// start of a new beam interval
			long long start_beam = (long long) (obj.since());
			if (start_beam < startTime) {
				start_beam = startTime;
			}
			beams_start_vector.push_back(start_beam/1000000);
			old_beam = beam;
			beamStarted = true;
		} else if ((beam.compare("RAMP DOWN") == 0) && (old_beam.compare("RAMP DOWN") != 0)) {
			// end of a beam interval
			long long end_beam  = (long long) (obj.since());
			beams_end_vector.push_back(end_beam/1000000);
			beamStarted = false;
			old_beam = beam;
		} else {
			old_beam = beam;
		}
	}
	if (beamStarted) {
		beams_end_vector.push_back(endTime/1000000);
	}
}

int setRunVectors (long long starttime, long long endtime) {
	start_of_run_vector.clear();
	end_of_run_vector.clear();

	// find list of runs and lumi blocks
	try {
		long long start_time_nsec = ((long long)(starttime))*1000;
		long long end_time_nsec = ((long long)(endtime))*1000;
		bool firstRun = true;
		bool newRun = false;
		int current_run_number = 0;
		long long start_of_first_LB_ms = 0;
		long long end_of_last_LB_ms = 0;
		cool::IObjectIteratorPtr rl_objects = LBT_Folder->browseObjects(cool::ValidityKey(start_time_nsec), cool::ValidityKey(end_time_nsec), 0);
		while ( rl_objects->goToNext() ) {
			const cool::IObject& obj = rl_objects->currentRef();
			std::string runNumber = obj.payloadValue("Run");
			std::string lumiBlock = obj.payloadValue("LumiBlock");
			int run_nb = atoi(runNumber.c_str());
			if(run_nb < current_run_number)
				continue;
			if (run_nb != current_run_number) {
				newRun = true;
				current_run_number = run_nb;
			}
			cool::UInt63 rn = run_nb;
			cool::UInt63 lb = atoi(lumiBlock.c_str());
			cool::ValidityKey key1 = (rn<<32) + lb;
			cool::IObjectPtr obj_lb = LB_Folder->findObject(key1, 0);
			std::string st = obj_lb->payloadValue("StartTime");
			std::string et = obj_lb->payloadValue("EndTime");
			long long sT = atoll(st.c_str());
			long long eT = atoll(et.c_str());
			if (newRun) {
				// save data for previous run
				if (firstRun) {
					firstRun = false;
				} else {
					start_of_run_vector.push_back(start_of_first_LB_ms);
					end_of_run_vector.push_back(end_of_last_LB_ms);
				}
				start_of_first_LB_ms = sT/1000000;
				newRun = false;
			}
			end_of_last_LB_ms = eT/1000000;
		}
		// save data for last run
		start_of_run_vector.push_back(start_of_first_LB_ms);
		end_of_run_vector.push_back(end_of_last_LB_ms);
	} catch (std::exception &e) {
		//std::cout << OWLTime() << " Warning: Read failed in setRunVectors: " << e.what() << std::endl;
	}
	if(start_of_run_vector.size())
		return 1;
	else
		return 0;
}

int main(int argc, char ** argv)
{

	// parse commandline parameters

	std::string startTime, endTime, interval;
	bool timeInterval = false;
	boost::program_options::options_description desc("This program retrieves info from COOL conditions database, evaluates the data taking efficiency and the inneficiency sources for selected time intervals and stores the results in /tmp/run.log file.");
	desc.add_options()
		("help,h", "help message")
		("startTime,s", boost::program_options::value<std::string>(), "Start time (epoch time in ns)")
		("endTime,e", boost::program_options::value<std::string>(), "End time (epoch time in ns)")
		("triggerItem,c", boost::program_options::value<std::string>(&triggerItem)->default_value(triggerItem), "Trigger item (default L1_EM30))")
		("atlasRun,r", boost::program_options::value<int>(&atlasRun)->default_value(atlasRun), "ATLAS Run: 1 - Run 1; 2 - Run 2 (default 2)")
		("noBeam,b", boost::program_options::value<int>(&noBeam)->default_value(noBeam), "No Beam: 0 - with beam; 1 - no beam (default 0)")
		("useTotalBusy,t", boost::program_options::value<int>(&useTotalBusy)->default_value(useTotalBusy), "Use Total Busy: 0 - use LVL1 counters; 1 - use total busy (default 1)")
		("getTotalBeam,k", boost::program_options::value<int>(&getTotalBeam)->default_value(getTotalBeam), "Get Total Beam: 0 - use SB beams; 1 - use RAMP DOWN beams (default 0)")
		("connectionName,d", boost::program_options::value<std::string>(&connName)->default_value(connName), "Connection name (default oracle://ATLR)")
		("timestampPath,p", boost::program_options::value<std::string>(&timestampPath)->default_value(timestampPath), "Timestamp file path (default: not writing the timestamp file)")
		;

	boost::program_options::variables_map vm;
	boost::program_options::store(boost::program_options::parse_command_line(argc,argv,desc), vm);
	boost::program_options::notify(vm);

	if(vm.count("help")) {
		std::cout << desc << std::endl;
		return EXIT_SUCCESS;
	}
	if ((atlasRun < 1) || (atlasRun > 2)) {
		std::cout << "Command line error: atlasRun should be 1 or 2" << std::endl;
		return EXIT_FAILURE;
	}
	if ((useTotalBusy < 0) || (useTotalBusy > 1)) {
		std::cout << "Command line error: useTotalBusy should be 0 or 1" << std::endl;
		return EXIT_FAILURE;
	}
	if ( (vm.count("interval")) && ( (vm.count("startTime")) || (vm.count("endTime")) ) ) {
		std::cout << desc << std::endl;
		std::cout << "Command line error: time interval used with an explicit start time or end time" << std::endl;
		return EXIT_FAILURE;
	}
	else if ( (vm.count("startTime")) && (vm.count("endTime")) ) {
		startTime = vm["startTime"].as<std::string>();
		endTime = vm["endTime"].as<std::string>();
	}
	else if ( vm.count("interval") ) {
		interval = vm["interval"].as<std::string>();
		timeInterval = true;
	}
	else {
		std::cout << desc << std::endl;
		daq::rc::CmdLineError issue(ERS_HERE,"");
		ers::fatal(issue);
		return EXIT_FAILURE;
	}

	//Install signal handler
	signal(SIGINT,  signalhandler);
	signal(SIGTERM, signalhandler);

	openDatabasesAndFolders();

	//FOR SPLUNK WE NEED TO PUT EVERYTHING IN STD.OUT
	std::string dummyLBsString;
	std::string beamTimesString;

	OWLTime now = OWLTime();
	long long nowtime = now.total_mksec_utc();
	long long starttime = 0;
	long long endtime = 0;
	if (!timeInterval) {
		if (timestampPath.size() > 0) {
			timestampPath += "lastTime.data";
			std::ofstream lastPeriod(timestampPath.c_str());
			lastPeriod << endTime;
			lastPeriod.close();
		}
		//OWLTime start_time(startTime.c_str());
		//starttime = start_time.total_mksec_utc();
		starttime = atoll(startTime.c_str())/1000;
		//OWLTime end_time(endTime.c_str());
		//endtime  = end_time.total_mksec_utc();
		endtime = atoll(endTime.c_str())/1000;
		if (starttime > nowtime) {
			std::cout << " Start time greater than current time" << std::endl;
			return EXIT_FAILURE;
		}
		if (endtime > nowtime) {
			std::cout << " End time greater than current time" << std::endl;
			return EXIT_FAILURE;
		}
	}
	else {
		long long total_duration_mksec = 0;
		if (interval.compare("day") == 0) {
			total_duration_mksec = ((long long) 86400) * ((long long) 1000000);
		}
		else if (interval.compare("week") == 0) {
			total_duration_mksec = ((long long) 604800) * ((long long) 1000000);
		}
		else if (interval.compare("month") == 0) {
			total_duration_mksec = ((long long) 2592000) * ((long long) 1000000);
		}
		else {
			std::cout << desc << std::endl;
			std::cout << "Wrong time interval option!" << std::endl;
			return EXIT_FAILURE;
		}
		endtime  = nowtime;
		// take into account total duration
		starttime = endtime - total_duration_mksec;
		endTime = std::string(now.c_str());
		OWLTime start = OWLTime(starttime/1000000);
		startTime = std::string(start.c_str());
	}

	std::vector<int> index;
	std::vector<int> fill;
	std::vector<int> stable_beams;
	std::vector<int> ready_for_physics;
	std::vector<int> run_number;
	std::vector<int> lumi_block;
	std::vector<std::string> tier0_project;
	std::vector<std::string> start_time;
	std::vector<std::string> end_time;
	std::vector<std::string> b1p;
	std::vector<std::string> b2p;
	std::vector<double> lb_duration;
	std::vector<double> run_eff;
	std::vector<int> l1_triggers;
	std::vector<double> LUCID_busy;
	std::vector<double> Pixel_busy;
	std::vector<double> SCT_busy;
	std::vector<double> TRT_busy;
	std::vector<double> L1Calo_busy;
	std::vector<double> BCM_busy;
	std::vector<double> ZDC_busy;
	std::vector<double> LAr_HF_busy;
	std::vector<double> LAr_EMEC_busy;
	std::vector<double> LAr_EMB_busy;
	std::vector<double> MDT_B_busy;
	std::vector<double> MDT_EC_busy;
	std::vector<double> Tile_EB_busy;
	std::vector<double> Tile_LB_busy;
	std::vector<double> CSC_busy;
	std::vector<double> TGC_busy;
	std::vector<double> RPC_busy;
	std::vector<double> MUCTPI_busy;
	std::vector<double> IBL_DBM_busy;
	std::vector<double> ALFA_busy;
	std::vector<double> Micromegas_busy;
	std::vector<double> VME_busy;
	std::vector<double> SMPL_busy;
	std::vector<double> CPX0_busy;
	std::vector<double> CTP_busy;
	std::vector<double> ECR_busy;
	std::vector<double> AFP_busy;
	std::vector<double> DAQ_busy;
	std::vector<double> DAQslink_busy;
	std::vector<int> warm_stop;

	//HT
	std::vector<double> Other_holdTrigger;
	std::vector<double> PIX_holdTrigger;
	std::vector<double> SCT_holdTrigger;
	std::vector<double> TRT_holdTrigger;
	std::vector<double> LAR_holdTrigger;
	std::vector<double> TIL_holdTrigger;
	std::vector<double> RPC_holdTrigger;
	std::vector<double> MDT_holdTrigger;
	std::vector<double> TGC_holdTrigger;
	std::vector<double> CSC_holdTrigger;
	std::vector<double> BCM_holdTrigger;
	std::vector<double> LUCID_holdTrigger;
	std::vector<double> ZDC_holdTrigger;
	std::vector<double> ALFA_holdTrigger;
	std::vector<double> L1CALO_holdTrigger;
	std::vector<double> L1CT_holdTrigger;
	std::vector<double> HLT_holdTrigger;
	std::vector<double> DAQ_holdTrigger;
	std::vector<double> IBL_holdTrigger;
	std::vector<double> DBM_holdTrigger;
	std::vector<double> MMEGA_holdTrigger;
	std::vector<double> L1TOPO_holdTrigger;
	std::vector<double> FTK_holdTrigger;
	std::vector<double> AFP_holdTrigger;
	std::vector<double> Unknown_holdTriggerReason;
	std::vector<double> TTCRestart_holdTriggerReason;
	std::vector<double> Resync_holdTriggerReason;
	std::vector<double> StoplessRecovery_holdTriggerReason;
	std::vector<double> ManualRecovery_holdTriggerReason;
	std::vector<double> Test_holdTriggerReason;
	std::vector<double> ClockSwitch_holdTriggerReason;
	std::vector<double> CustomRecovery_holdTriggerReason;

	// find list of runs and lumi blocks
	//long long start_time_nsec = ((long long)(starttime - 120000000))*1000;
	long long start_time_nsec = ((long long)(starttime))*1000;
	long long end_time_nsec = ((long long)(endtime))*1000;
	int current_run_number = 0;
	int ctpId = 227; //default value, to be checked
	int ind = 0;

	//set and save inter-run periods
	int runsExist = setRunVectors(starttime, endtime);
	std::vector<long long>::iterator its;
	std::vector<long long>::iterator ite;
	std::vector<long long> pausePeriods;
	std::vector<int> runsVector;

	if(start_of_run_vector.size()) {
		its = start_of_run_vector.begin();
		ite = end_of_run_vector.begin();
		if(start_of_run_vector.size() != 1) {
			++its;
			for(; its != start_of_run_vector.end(); ++its) {
				pausePeriods.push_back(*ite);
				pausePeriods.push_back(*its);
				++ite;
			}
		}
		else if(starttime != *its && endtime != *ite) {
			pausePeriods.push_back(starttime/1000);
			pausePeriods.push_back(*its);
			pausePeriods.push_back(*ite);
			pausePeriods.push_back(endtime/1000);
		}
		else if(starttime != *its) {
			pausePeriods.push_back(starttime/1000);
			pausePeriods.push_back(*its);
		}
		else if(endtime != *ite) {
			pausePeriods.push_back(*ite);
			pausePeriods.push_back(endtime/1000);
		}
	}

	if(pausePeriods.size()) {
		for(int i = 0; i < pausePeriods.size(); i++) {
			dummyLBsString += get_date_OWL(pausePeriods.at(i)/1000);
			if(i%2) {
				dummyLBsString += "\n";
			}
			else {
				dummyLBsString += "\t";
			}
		}
	}

	//get and save beam periods
	int beamsExist = -1;
	if(getTotalBeam)
		get_beams_data_beamsON(starttime*1000, endtime*1000, TotalInt_Folder);
	else
		beamsExist = get_beams_data(starttime*1000, endtime*1000, FillState_Folder);
	if(beams_end_vector.size()) {
		ite = beams_end_vector.begin();
		for(its = beams_start_vector.begin(); its != beams_start_vector.end(); ++its) {
			beamBuffer.push_back(get_fill_number(*its*1000000, *ite*1000000, FillState_Folder));
			beamTimesString += get_date_OWL(*its/1000) + "\t" + get_date_OWL(*ite/1000) + "\n";
			++ite;
		}
	}
	beamTimesString += "endbeam\n";
	if(beamBuffer.size()) {
		for(std::vector<int>::iterator itInt = beamBuffer.begin(); itInt != beamBuffer.end(); ++itInt) {
			beamTimesString += std::to_string(*itInt) + "\n";
		}
	}

	double total_lb_time_beam = 0.;
	double total_lb_time_run = 0.;
	bool data_taking_started = false;
	try {
		cool::IObjectIteratorPtr rl_objects = LBT_Folder->browseObjects(cool::ValidityKey(start_time_nsec), cool::ValidityKey(end_time_nsec), 0);
		while ( rl_objects->goToNext() ) {
			const cool::IObject& obj = rl_objects->currentRef();
			std::string runNumber = obj.payloadValue("Run");
			std::string lumiBlock = obj.payloadValue("LumiBlock");
			int run_nb = atoi(runNumber.c_str());
			int lumi_block_nb = atoi(lumiBlock.c_str());
			if (run_nb != current_run_number) {
				//newRunWitness = 1;
				//std::cout << " New RN " << run_nb << " old RN was " << current_run_number << std::endl;
				if(run_nb < current_run_number) {
					continue;
				}
				prepareCtpMap(run_nb);
				current_run_number = run_nb;
				runsVector.push_back(run_nb);
				cool::UInt63 a = run_nb;
				cool::UInt63 z = run_nb+1;
				cool::ValidityKey sKey = a <<32;
				cool::ValidityKey uKey = (z <<32) - 1;
				if(useTotalBusy == 0) {
					ctpId = get_ctpid(sKey, uKey, triggerItem, L1Menu_Folder);
				}
				if (ctpId == -1) {
					//std::cout << "For Run " << runNumber << " Trigger Item not found. Default value (227) will be used." << std::endl;
					ctpId = 227;
				}
			}
			cool::UInt63 rn = run_nb;
			cool::UInt63 lb = atoi(lumiBlock.c_str());
			cool::ValidityKey key1 = (rn <<32) + lb;
			cool::IObjectPtr obj_lb = LB_Folder->findObject(key1, 0);
			std::string st = obj_lb->payloadValue("StartTime");
			std::string et = obj_lb->payloadValue("EndTime");
			long long sT = atoll(st.c_str());
			long long eT = atoll(et.c_str());
			int fill_number = 0;
			int stable_beams_flag = 0;
			double start_time_sec = (double) sT/1000000000;
			double end_time_sec = (double) eT/1000000000;
			std::string start_string = get_date(start_time_sec);
			std::string end_string = get_date(end_time_sec);
			double lb_duration_msec = (eT - sT)/1000000;
			double efficiency = 1.0;
			try {
				if (useTotalBusy == 0) {
					efficiency = eff_fraction(key1, ctpId, L1Counters_Folder);
				} else {
					float lb_busy = busy_run2(key1, "ctpcore_moni0_rate", BusyRate_Folder );
					efficiency = 1.0 - lb_busy;
				}
			} catch (std::exception &ex) {
				//std::cout << "Failure in eff evaluation - " << ex.what() << std::endl;
			}
			if (noBeam == 0) {
				fill_number = get_fill_number(sT, eT, FillState_Folder);
				stable_beams_flag = get_stable_beams(sT, eT, FillState_Folder);
				//stable_beams_flag = get_stable_beams(sT, eT, TotalInt_Folder);
			}

			total_lb_time_beam += lb_duration_msec*stable_beams_flag;
			total_lb_time_run += lb_duration_msec*efficiency*stable_beams_flag;

			int ready_flag = get_ready_for_physics(key1, DataTakingMode_Folder);
			std::string tier0Project = get_t0project(key1, t0Project_Folder);
			long long l1accept = 0;
			//WATCH OUT!!!
			try {
				l1accept = l1_accepted(key1, ctpId, L1Counters_Folder);
			} catch (std::exception &ex) {
				//std::cout << "Failure in l1_accepted - " << ex.what() << std::endl;
			}

			// initialize the busy map
			busyMap["DAQslink"] = 0;
			busyMap["DAQ"] = 0;
			busyMap["LUCID"] = 0;
			busyMap["Pixel"] = 0;
			busyMap["SCT"] = 0;
			busyMap["TRT"] = 0;
			busyMap["L1Calo"] = 0;
			busyMap["BCM"] = 0;
			busyMap["ZDC"] = 0;
			busyMap["LAr H/F"] = 0;
			busyMap["LAr EMEC"] = 0;
			busyMap["LAr EMB"] = 0;
			busyMap["MDT B"] = 0;
			busyMap["MDT EC"] = 0;
			busyMap["Tile EB"] = 0;
			busyMap["Tile LB"] = 0;
			busyMap["CSC"] = 0;
			busyMap["TGC"] = 0;
			busyMap["RPC"] = 0;
			busyMap["MUCTPI"] = 0;
			busyMap["IBL/DBM"] = 0;
			busyMap["ALFA"] = 0;
			busyMap["Micromegas"] = 0;
			busyMap["VME"] = 0;
			busyMap["SMPL"] = 0;
			busyMap["CPX0"] = 0;
			busyMap["CTP"] = 0;
			busyMap["ECR"] = 0;
			busyMap["AFP"] = 0;

			// initialize the hold trigger systems map
			holdTriggerSystemsMap["Other"] = 0;
			holdTriggerSystemsMap["PIX"] = 0;
			holdTriggerSystemsMap["SCT"] = 0;
			holdTriggerSystemsMap["TRT"] = 0;
			holdTriggerSystemsMap["LAR"] = 0;
			holdTriggerSystemsMap["TIL"] = 0;
			holdTriggerSystemsMap["RPC"] = 0;
			holdTriggerSystemsMap["MDT"] = 0;
			holdTriggerSystemsMap["TGC"] = 0;
			holdTriggerSystemsMap["CSC"] = 0;
			holdTriggerSystemsMap["BCM"] = 0;
			holdTriggerSystemsMap["LUCID"] = 0;
			holdTriggerSystemsMap["ZDC"] = 0;
			holdTriggerSystemsMap["ALFA"] = 0;
			holdTriggerSystemsMap["L1CALO"] = 0;
			holdTriggerSystemsMap["L1CT"] = 0;
			holdTriggerSystemsMap["HLT"] = 0;
			holdTriggerSystemsMap["DAQ"] = 0;
			holdTriggerSystemsMap["IBL"] = 0;
			holdTriggerSystemsMap["DBM"] = 0;
			holdTriggerSystemsMap["MMEGA"] = 0;
			holdTriggerSystemsMap["L1TOPO"] = 0;
			holdTriggerSystemsMap["FTK"] = 0;
			holdTriggerSystemsMap["AFP"] = 0;

			// initialize the hold trigger reason map
			holdTriggerReasonMap["Unknown"] = 0;
			holdTriggerReasonMap["TTCRestart"] = 0;
			holdTriggerReasonMap["Resync"] = 0;
			holdTriggerReasonMap["StoplessRecovery"] = 0;
			holdTriggerReasonMap["ManualRecovery"] = 0;
			holdTriggerReasonMap["Test"] = 0;
			holdTriggerReasonMap["ClockSwitch"] = 0;
			holdTriggerReasonMap["CustomRecovery"] = 0;

			long long lumi_block_duration_ns = eT - sT;
			double lumi_block_duration_msec = (double)lumi_block_duration_ns/1000000;
			double lumi_block_duration_sec = lumi_block_duration_msec/1000;

			// evaluate dead time for each sub-detector
			std::map<std::string,std::string>::iterator iter;
			for ( iter=ctpMap.begin() ; iter != ctpMap.end(); iter++ ) {
				float det_busy = 0;
				try {
					det_busy = busy_run2(key1, (*iter).first, BusyRate_Folder );
				} catch (std::exception &ex) {
					//std::cout << "Failure in busy_run2 - " << ex.what() << std::endl;
				}
				busyMap[(*iter).second] = lumi_block_duration_sec*det_busy;
			}
			float det_busy = 0;
			try {
				det_busy = busy_run2(key1, "ctpmi_vme_rate", BusyRate_Folder );
			} catch (std::exception &ex) {
				//std::cout << "Failure in busy_run2 - " << ex.what() << std::endl;
			}
			busyMap["VME"] = lumi_block_duration_sec*det_busy;
			det_busy = 0;
			try {
				det_busy = busy_run2(key1, "ctpcore_moni3_rate", BusyRate_Folder );
			} catch (std::exception &ex) {
				//std::cout << "Failure in busy_run2 - " << ex.what() << std::endl;
			}
			busyMap["SMPL"] = lumi_block_duration_sec*det_busy;
			det_busy = 0;
			try {
				det_busy = busy_run2(key1, "ctpcore_moni1_rate", BusyRate_Folder );
			} catch (std::exception &ex) {
				//std::cout << "Failure in busy_run2 - " << ex.what() << std::endl;
			}
			busyMap["CPX0"] = lumi_block_duration_sec*det_busy;
			det_busy = 0;
			try {
				det_busy = busy_run2(key1, "ctpcore_rate", BusyRate_Folder );
			} catch (std::exception &ex) {
				//std::cout << "Failure in busy_run2 - " << ex.what() << std::endl;
			}
			busyMap["CTP"] = lumi_block_duration_sec*det_busy;
			det_busy = 0;
			try {
				det_busy = busy_run2(key1, "ctpmi_ecr_rate", BusyRate_Folder );
			} catch (std::exception &ex) {
				//std::cout << "Failure in busy_run2 - " << ex.what() << std::endl;
			}
			busyMap["ECR"] = lumi_block_duration_sec*det_busy;
			det_busy = 0;
			try {
				det_busy = busy_run2(key1, "ctpcore_roi_fifo_rate", BusyRate_Folder);
			}
			catch (std::exception &ex) {
			}
			busyMap["DAQ"] = lumi_block_duration_sec*det_busy;
			det_busy = 0;
			try {
				det_busy = busy_run2(key1, "ctpcore_roi_slink_rate", BusyRate_Folder);
			}
			catch (std::exception &ex) {
			}
			busyMap["DAQslink"] = lumi_block_duration_sec*det_busy;

			//routine for warm stop
			long long warm_stop_seconds = 0;
			if (stable_beams_flag && ready_flag) {
				data_taking_started = true;
			}
			else if (data_taking_started && stable_beams_flag && !ready_flag) {
				//std::cout << (get_beam_dump_time(sT, eT, TotalInt_Folder) - sT)/1000000000 << std::endl;
				warm_stop_seconds = (get_beam_dump_time(sT, eT, TotalInt_Folder) - sT) / 1000000000;
				data_taking_started = false;
			}
			else
				warm_stop_seconds = 0;
			std::pair<std::string, std::string> bp = get_beam_presence(sT, eT, TotalInt_Folder);
			b1p.push_back(bp.first);
			b2p.push_back(bp.second);
			
			// update data vectors
			index.push_back(ind);
			fill.push_back(fill_number); 
			stable_beams.push_back(stable_beams_flag);
			ready_for_physics.push_back(ready_flag);
			run_number.push_back(run_nb);
			lumi_block.push_back(lumi_block_nb);
			tier0_project.push_back(tier0Project);
			start_time.push_back(start_string);
			end_time.push_back(end_string);
			lb_duration.push_back(lb_duration_msec);
			run_eff.push_back(efficiency);
			l1_triggers.push_back(l1accept);
			LUCID_busy.push_back(busyMap["LUCID"]);
			Pixel_busy.push_back(busyMap["Pixel"]);
			SCT_busy.push_back(busyMap["SCT"]);
			TRT_busy.push_back(busyMap["TRT"]);
			L1Calo_busy.push_back(busyMap["L1Calo"]);
			BCM_busy.push_back(busyMap["BCM"]);
			ZDC_busy.push_back(busyMap["ZDC"]);
			LAr_HF_busy.push_back(busyMap["LAr H/F"]);
			LAr_EMEC_busy.push_back(busyMap["LAr EMEC"]);
			LAr_EMB_busy.push_back(busyMap["LAr EMB"]);
			MDT_B_busy.push_back(busyMap["MDT B"]);
			MDT_EC_busy.push_back(busyMap["MDT EC"]);
			Tile_EB_busy.push_back(busyMap["Tile EB"]);
			Tile_LB_busy.push_back(busyMap["Tile LB"]);
			CSC_busy.push_back(busyMap["CSC"]);
			TGC_busy.push_back(busyMap["TGC"]);
			RPC_busy.push_back(busyMap["RPC"]);
			MUCTPI_busy.push_back(busyMap["MUCTPI"]);
			IBL_DBM_busy.push_back(busyMap["IBL/DBM"]);
			ALFA_busy.push_back(busyMap["ALFA"]);
			Micromegas_busy.push_back(busyMap["Micromegas"]);
			VME_busy.push_back(busyMap["VME"]);
			SMPL_busy.push_back(busyMap["SMPL"]);
			CPX0_busy.push_back(busyMap["CPX0"]);
			CTP_busy.push_back(busyMap["CTP"]);
			ECR_busy.push_back(busyMap["ECR"]);
			AFP_busy.push_back(busyMap["AFP"]);
			DAQ_busy.push_back(busyMap["DAQ"]);
			DAQslink_busy.push_back(busyMap["DAQslink"]);

			//get HT info
			cool::IObjectIteratorPtr ht_objects = HoldTrigger_Folder->browseObjects(cool::ValidityKey(sT), cool::ValidityKey(eT), 0);
			//try {
			while ( ht_objects->goToNext() ) {
				const cool::IObject& ht_obj = ht_objects->currentRef();
				double ht_starttime = 0.;
				double ht_stoptime = 0.;
				if(ht_obj.since() < sT)
					ht_starttime = sT;
				else
					ht_starttime = ht_obj.since();

				if(ht_obj.until() >= eT)
					ht_stoptime = eT;
				else
					ht_stoptime = ht_obj.until();
				double ht_duration_msec = (ht_stoptime - ht_starttime)/1000000;
				std::string subSystems = ht_obj.payloadValue("SubSystems");
				std::string reason = ht_obj.payloadValue("Reason");
				if (!(subSystems.compare("None") == 0)) {
					// use milisec here
					holdTriggerSystemsMap[subSystems] = holdTriggerSystemsMap[subSystems] + ht_duration_msec;
					holdTriggerReasonMap[reason] = holdTriggerReasonMap[reason] + ht_duration_msec;
				}
			}
			//} catch (std::exception &ex) {
			//	std::cout << "No HT info - " << ex.what() << std::endl;
			//}

			Other_holdTrigger.push_back(holdTriggerSystemsMap["Other"]/1000);
			PIX_holdTrigger.push_back(holdTriggerSystemsMap["PIX"]/1000);
			SCT_holdTrigger.push_back(holdTriggerSystemsMap["SCT"]/1000);
			TRT_holdTrigger.push_back(holdTriggerSystemsMap["TRT"]/1000);
			LAR_holdTrigger.push_back(holdTriggerSystemsMap["LAR"]/1000);
			TIL_holdTrigger.push_back(holdTriggerSystemsMap["TIL"]/1000);
			RPC_holdTrigger.push_back(holdTriggerSystemsMap["RPC"]/1000);
			MDT_holdTrigger.push_back(holdTriggerSystemsMap["MDT"]/1000);
			TGC_holdTrigger.push_back(holdTriggerSystemsMap["TGC"]/1000);
			CSC_holdTrigger.push_back(holdTriggerSystemsMap["CSC"]/1000);
			BCM_holdTrigger.push_back(holdTriggerSystemsMap["BCM"]/1000);
			LUCID_holdTrigger.push_back(holdTriggerSystemsMap["LUCID"]/1000);
			ZDC_holdTrigger.push_back(holdTriggerSystemsMap["ZDC"]/1000);
			ALFA_holdTrigger.push_back(holdTriggerSystemsMap["ALFA"]/1000);
			L1CALO_holdTrigger.push_back(holdTriggerSystemsMap["L1CALO"]/1000);
			L1CT_holdTrigger.push_back(holdTriggerSystemsMap["L1CT"]/1000);
			HLT_holdTrigger.push_back(holdTriggerSystemsMap["HLT"]/1000);
			DAQ_holdTrigger.push_back(holdTriggerSystemsMap["DAQ"]/1000);
			IBL_holdTrigger.push_back(holdTriggerSystemsMap["IBL"]/1000);
			DBM_holdTrigger.push_back(holdTriggerSystemsMap["DBM"]/1000);
			MMEGA_holdTrigger.push_back(holdTriggerSystemsMap["MMEGA"]/1000);
			L1TOPO_holdTrigger.push_back(holdTriggerSystemsMap["L1TOPO"]/1000);
			FTK_holdTrigger.push_back(holdTriggerSystemsMap["FTK"]/1000);
			AFP_holdTrigger.push_back(holdTriggerSystemsMap["AFP"]/1000);
			Unknown_holdTriggerReason.push_back(holdTriggerReasonMap["Unknown"]/1000);
			TTCRestart_holdTriggerReason.push_back(holdTriggerReasonMap["TTCRestart"]/1000);
			Resync_holdTriggerReason.push_back(holdTriggerReasonMap["Resync"]/1000);
			StoplessRecovery_holdTriggerReason.push_back(holdTriggerReasonMap["StoplessRecovery"]/1000);
			ManualRecovery_holdTriggerReason.push_back(holdTriggerReasonMap["ManualRecovery"]/1000);
			Test_holdTriggerReason.push_back(holdTriggerReasonMap["Test"]/1000);
			ClockSwitch_holdTriggerReason.push_back(holdTriggerReasonMap["ClockSwitch"]/1000);
			CustomRecovery_holdTriggerReason.push_back(holdTriggerReasonMap["CustomRecovery"]/1000);
			warm_stop.push_back(warm_stop_seconds);

			ind = ind + 1;
		}//END OF PER LB WHILE
	} catch (std::exception &e) {
		//std::cout << OWLTime() << " Warning: Read failed in MAIN: " << e.what() << std::endl;
	}

	std::vector<double> Other_stopped;
	std::vector<double> PIX_stopped;
	std::vector<double> SCT_stopped;
	std::vector<double> TRT_stopped;
	std::vector<double> LAR_stopped;
	std::vector<double> TIL_stopped;
	std::vector<double> RPC_stopped;
	std::vector<double> MDT_stopped;
	std::vector<double> TGC_stopped;
	std::vector<double> CSC_stopped;
	std::vector<double> BCM_stopped;
	std::vector<double> LUCID_stopped;
	std::vector<double> ZDC_stopped;
	std::vector<double> ALFA_stopped;
	std::vector<double> L1CALO_stopped;
	std::vector<double> L1CT_stopped;
	std::vector<double> HLT_stopped;
	std::vector<double> DAQ_stopped;
	std::vector<double> IBL_stopped;
	std::vector<double> DBM_stopped;
	std::vector<double> MMEGA_stopped;
	std::vector<double> L1TOPO_stopped;
	std::vector<double> Solenoid_stopped;
	std::vector<double> Toroid_stopped;
	std::vector<double> Cooling_stopped;
	std::vector<double> LHC_stopped;
	std::vector<double> FTK_stopped;
	std::vector<double> AFP_stopped;
	std::vector<double> Unknown_reason;
	std::vector<double> SystemBusy_reason;
	std::vector<double> SystemNeedsReconfiguration_reason;
	std::vector<double> HardwareFailure_reason;
	std::vector<double> BeamDump_reason;
	std::vector<double> HandShake_reason;
	std::vector<double> TestRun_reason;

	// initialize stop system map
	stopSystemsMap["Other"] = 0;
	stopSystemsMap["PIX"] = 0;
	stopSystemsMap["SCT"] = 0;
	stopSystemsMap["TRT"] = 0;
	stopSystemsMap["LAR"] = 0;
	stopSystemsMap["TIL"] = 0;
	stopSystemsMap["RPC"] = 0;
	stopSystemsMap["MDT"] = 0;
	stopSystemsMap["TGC"] = 0;
	stopSystemsMap["CSC"] = 0;
	stopSystemsMap["BCM"] = 0;
	stopSystemsMap["LUCID"] = 0;
	stopSystemsMap["ZDC"] = 0;
	stopSystemsMap["ALFA"] = 0;
	stopSystemsMap["L1CALO"] = 0;
	stopSystemsMap["L1CT"] = 0;
	stopSystemsMap["HLT"] = 0;
	stopSystemsMap["DAQ"] = 0;
	stopSystemsMap["IBL"] = 0;
	stopSystemsMap["DBM"] = 0;
	stopSystemsMap["MMEGA"] = 0;
	stopSystemsMap["L1TOPO"] = 0;
	stopSystemsMap["Solenoid"] = 0;
	stopSystemsMap["Toroid"] = 0;
	stopSystemsMap["Cooling"] = 0;
	stopSystemsMap["LHC"] = 0;
	stopSystemsMap["FTK"] = 0;
	stopSystemsMap["AFP"] = 0;

	// initialize the stop reason map
	stopReasonMap["Unknown"] = 0;
	stopReasonMap["SystemBusy"] = 0;
	stopReasonMap["SystemNeedsReconfiguration"] = 0;
	stopReasonMap["HardwareFailure"] = 0;
	stopReasonMap["BeamDump"] = 0;
	stopReasonMap["HandShake"] = 0;
	stopReasonMap["TestRun"] = 0;

	if(pausePeriods.size()) {
		for(int i = 0; i < pausePeriods.size(); i++) {
			if(!(i%2))
				continue;
			std::string stopReason = get_stop_reason(pausePeriods.at(i-1),pausePeriods.at(i));
			if (stopReason.size() > 3) {
				size_t indexStops = stopReason.find_first_of("/");
				std::string stop_systems = stopReason.substr(0, indexStops);
				size_t lastIndex = stopReason.size();
				std::string stop_reason = stopReason.substr(indexStops+1, lastIndex-indexStops);
				size_t firstIndex = 0;
				indexStops = 0;
				while (indexStops != std::string::npos) {
					indexStops = stop_systems.find(", ", firstIndex);
					if (indexStops != std::string::npos) {
						std::string system = stop_systems.substr(firstIndex, indexStops-firstIndex);
						stopSystemsMap[system] = stopSystemsMap[system] + (pausePeriods.at(i)-pausePeriods.at(i-1));
						firstIndex = indexStops+2;
					}
				}
				indexStops = 0;
				firstIndex = 0;
				while (indexStops != std::string::npos) {
					indexStops = stop_reason.find(", ", firstIndex);
					if (indexStops != std::string::npos) {
						std::string reason = stop_reason.substr(firstIndex, indexStops-firstIndex);
						stopReasonMap[reason] = stopReasonMap[reason] + (pausePeriods.at(i)-pausePeriods.at(i-1));
						firstIndex = indexStops+2;
					}
				}
			}
			Other_stopped.push_back(stopSystemsMap["Other"]);
			PIX_stopped.push_back(stopSystemsMap["PIX"]);
			SCT_stopped.push_back(stopSystemsMap["SCT"]);
			TRT_stopped.push_back(stopSystemsMap["TRT"]);
			LAR_stopped.push_back(stopSystemsMap["LAR"]);
			TIL_stopped.push_back(stopSystemsMap["TIL"]);
			RPC_stopped.push_back(stopSystemsMap["RPC"]);
			MDT_stopped.push_back(stopSystemsMap["MDT"]);
			TGC_stopped.push_back(stopSystemsMap["TGC"]);
			CSC_stopped.push_back(stopSystemsMap["CSC"]);
			BCM_stopped.push_back(stopSystemsMap["BCM"]);
			LUCID_stopped.push_back(stopSystemsMap["LUCID"]);
			ZDC_stopped.push_back(stopSystemsMap["ZDC"]);
			ALFA_stopped.push_back(stopSystemsMap["ALFA"]);
			L1CALO_stopped.push_back(stopSystemsMap["L1CALO"]);
			L1CT_stopped.push_back(stopSystemsMap["L1CT"]);
			HLT_stopped.push_back(stopSystemsMap["HLT"]);
			DAQ_stopped.push_back(stopSystemsMap["DAQ"]);
			IBL_stopped.push_back(stopSystemsMap["IBL"]);
			DBM_stopped.push_back(stopSystemsMap["DBM"]);
			MMEGA_stopped.push_back(stopSystemsMap["MMEGA"]);
			L1TOPO_stopped.push_back(stopSystemsMap["L1TOPO"]);
			Solenoid_stopped.push_back(stopSystemsMap["Solenoid"]);
			Toroid_stopped.push_back(stopSystemsMap["Toroid"]);
			Cooling_stopped.push_back(stopSystemsMap["Cooling"]);
			LHC_stopped.push_back(stopSystemsMap["LHC"]);
			FTK_stopped.push_back(stopSystemsMap["FTK"]);
			AFP_stopped.push_back(stopSystemsMap["AFP"]);
			Unknown_reason.push_back(stopReasonMap["Unknown"]);
			SystemBusy_reason.push_back(stopReasonMap["SystemBusy"]);
			SystemNeedsReconfiguration_reason.push_back(stopReasonMap["SystemNeedsReconfiguration"]);
			HardwareFailure_reason.push_back(stopReasonMap["HardwareFailure"]);
			BeamDump_reason.push_back(stopReasonMap["BeamDump"]);
			HandShake_reason.push_back(stopReasonMap["HandShake"]);
			TestRun_reason.push_back(stopReasonMap["TestRun"]);

		}
	}

	dummyLBsString += "end\n";
	if(runsVector.size()) {
		for(int iRuns = 0; iRuns < (runsVector.size()-1); iRuns++) {
			dummyLBsString += std::to_string(runsVector.at(iRuns)) + "\n";
		}
	}
	dummyLBsString += "endrun\n";
	if(Other_stopped.size()) {
		for(int i = 0; i < Other_stopped.size(); i++) {
			dummyLBsString += std::to_string(Other_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(PIX_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(SCT_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(TRT_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(LAR_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(TIL_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(RPC_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(MDT_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(TGC_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(CSC_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(BCM_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(LUCID_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(ZDC_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(ALFA_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(L1CALO_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(L1CT_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(HLT_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(DAQ_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(IBL_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(DBM_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(MMEGA_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(L1TOPO_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(Solenoid_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(Toroid_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(Cooling_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(LHC_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(FTK_stopped[i]/1000.) + "\n";
			dummyLBsString += std::to_string(AFP_stopped[i] / 1000.) + "\n";
			dummyLBsString += std::to_string(Unknown_reason[i]/1000.) + "\n";
			dummyLBsString += std::to_string(SystemBusy_reason[i]/1000.) + "\n";
			dummyLBsString += std::to_string(SystemNeedsReconfiguration_reason[i]/1000.) + "\n";
			dummyLBsString += std::to_string(HardwareFailure_reason[i]/1000.) + "\n";
			dummyLBsString += std::to_string(BeamDump_reason[i]/1000.) + "\n";
			dummyLBsString += std::to_string(HandShake_reason[i]/1000.) + "\n";
			dummyLBsString += std::to_string(TestRun_reason[i]/1000.) + "\n";
		}
	}

	int n = 0;
	n = index.size();
	// prepare csv file
	std::string csv_outString;
	csv_outString += "Fill,StableBeams,ReadyForPhysics,RunNumber,LumiBlock,T0Project,StartDate,EndDate,Duration,Eff,BCM,CSC,CTPLUCID,L1Calo,LArEMB,LArEMEC,LArHF,MDTB,MDTEC,MUCTPI,Pixel,RPC,SCT,TGC,TRT,TileEB,TileLB,CTPMI_VME,ZDC,IBLDBM,ALFA,Micromegas,SMPL,CPX0,CTP,ECR,AFP,Other_HT,PIX_HT,SCT_HT,TRT_HT,LAR_HT,TIL_HT,RPC_HT,MDT_HT,TGC_HT,CSC_HT,BCM_HT,LUCID_HT,ZDC_HT,ALFA_HT,L1CALO_HT,L1CT_HT,HLT_HT,DAQ_HT,IBL_HT,DBM_HT,MMEGA_HT,L1TOPO_HT,FTK_HT,AFP_HT,Unknown,TTCRestart,Resync,StoplessRecovery,ManualRecovery,Test,ClockSwitch,CustomRecovery,WS,B1Present,B2Present,DAQ,DAQslink,\n";
	for (int i=0; i<n; i++) {
		csv_outString += std::to_string(fill[i]) + "," + std::to_string(stable_beams[i]) + "," + std::to_string(ready_for_physics[i]) + ",";
		csv_outString += std::to_string(run_number[i]) + "," + std::to_string(lumi_block[i]) + "," + tier0_project[i] + "," + start_time[i] + "," + end_time[i] + "," + std::to_string(lb_duration[i]) + "," + std::to_string(run_eff[i]) + ",";
		csv_outString += std::to_string(BCM_busy[i]) + ",";
		csv_outString += std::to_string(CSC_busy[i]) + ",";
		csv_outString += std::to_string(LUCID_busy[i]) + ",";
		csv_outString += std::to_string(L1Calo_busy[i]) + ",";
		csv_outString += std::to_string(LAr_EMB_busy[i]) + ",";
		csv_outString += std::to_string(LAr_EMEC_busy[i]) + ",";
		csv_outString += std::to_string(LAr_HF_busy[i]) + ",";
		csv_outString += std::to_string(MDT_B_busy[i]) + ",";
		csv_outString += std::to_string(MDT_EC_busy[i]) + ",";
		csv_outString += std::to_string(MUCTPI_busy[i]) + ",";
		csv_outString += std::to_string(Pixel_busy[i]) + ",";
		csv_outString += std::to_string(RPC_busy[i]) + ",";
		csv_outString += std::to_string(SCT_busy[i]) + ",";
		csv_outString += std::to_string(TGC_busy[i]) + ",";
		csv_outString += std::to_string(TRT_busy[i]) + ",";
		csv_outString += std::to_string(Tile_EB_busy[i]) + ",";
		csv_outString += std::to_string(Tile_LB_busy[i]) + ",";
		csv_outString += std::to_string(VME_busy[i]) + ",";
		csv_outString += std::to_string(ZDC_busy[i]) + ",";
		csv_outString += std::to_string(IBL_DBM_busy[i]) + ",";
		csv_outString += std::to_string(ALFA_busy[i]) + ",";
		csv_outString += std::to_string(Micromegas_busy[i]) + ",";
		csv_outString += std::to_string(SMPL_busy[i]) + ",";
		csv_outString += std::to_string(CPX0_busy[i]) + ",";
		csv_outString += std::to_string(CTP_busy[i]) + ",";
		csv_outString += std::to_string(ECR_busy[i]) + ",";
		csv_outString += std::to_string(AFP_busy[i]) + ",";
		csv_outString += std::to_string(Other_holdTrigger[i]) + ",";
		csv_outString += std::to_string(PIX_holdTrigger[i]) + ",";
		csv_outString += std::to_string(SCT_holdTrigger[i]) + ",";
		csv_outString += std::to_string(TRT_holdTrigger[i]) + ",";
		csv_outString += std::to_string(LAR_holdTrigger[i]) + ",";
		csv_outString += std::to_string(TIL_holdTrigger[i]) + ",";
		csv_outString += std::to_string(RPC_holdTrigger[i]) + ",";
		csv_outString += std::to_string(MDT_holdTrigger[i]) + ",";
		csv_outString += std::to_string(TGC_holdTrigger[i]) + ",";
		csv_outString += std::to_string(CSC_holdTrigger[i]) + ",";
		csv_outString += std::to_string(BCM_holdTrigger[i]) + ",";
		csv_outString += std::to_string(LUCID_holdTrigger[i]) + ",";
		csv_outString += std::to_string(ZDC_holdTrigger[i]) + ",";
		csv_outString += std::to_string(ALFA_holdTrigger[i]) + ",";
		csv_outString += std::to_string(L1CALO_holdTrigger[i]) + ",";
		csv_outString += std::to_string(L1CT_holdTrigger[i]) + ",";
		csv_outString += std::to_string(HLT_holdTrigger[i]) + ",";
		csv_outString += std::to_string(DAQ_holdTrigger[i]) + ",";
		csv_outString += std::to_string(IBL_holdTrigger[i]) + ",";
		csv_outString += std::to_string(DBM_holdTrigger[i]) + ",";
		csv_outString += std::to_string(MMEGA_holdTrigger[i]) + ",";
		csv_outString += std::to_string(L1TOPO_holdTrigger[i]) + ",";
		csv_outString += std::to_string(FTK_holdTrigger[i]) + ",";
		csv_outString += std::to_string(AFP_holdTrigger[i]) + ",";
		csv_outString += std::to_string(Unknown_holdTriggerReason[i]) + ",";
		csv_outString += std::to_string(TTCRestart_holdTriggerReason[i]) + ",";
		csv_outString += std::to_string(Resync_holdTriggerReason[i]) + ",";
		csv_outString += std::to_string(StoplessRecovery_holdTriggerReason[i]) + ",";
		csv_outString += std::to_string(ManualRecovery_holdTriggerReason[i]) + ",";
		csv_outString += std::to_string(Test_holdTriggerReason[i]) + ",";
		csv_outString += std::to_string(ClockSwitch_holdTriggerReason[i]) + ",";
		csv_outString += std::to_string(CustomRecovery_holdTriggerReason[i]) + ",";
		csv_outString += std::to_string(warm_stop[i]) + ",";
		csv_outString += b1p[i] + ",";
		csv_outString += b2p[i] + ",";
		csv_outString += std::to_string(DAQ_busy[i]) + ",";
		csv_outString += std::to_string(DAQslink_busy[i]) + ",";
		csv_outString += "\n";

	}

	//EFF_CSV_SPLUNK PART
	std::vector<std::string> titleNames;
	std::string bufferStr;
	std::size_t indexCSV;
	std::size_t indexCSVOld;
	std::string titleLine = csv_outString.substr(0,csv_outString.find("\n",0));
	csv_outString.erase(0,csv_outString.find("\n",0));
	indexCSV = 0;
	indexCSVOld = -1;
	while(titleLine.find(",",indexCSV+1) != std::string::npos) {
		indexCSV = titleLine.find(",",indexCSV+1);
		bufferStr = titleLine.substr(indexCSVOld+1,indexCSV-1-indexCSVOld);
		indexCSVOld = indexCSV;
		bufferStr += "=";
		titleNames.push_back(bufferStr);
		bufferStr.clear();
	}

	std::string rowIn;
	indexCSV = 0;
	indexCSVOld = -1;
	size_t indexLine = 0;
	int i = 0;
	std::string output;
	while(csv_outString.find("\n",indexLine+1) != std::string::npos) {
		titleLine = csv_outString.substr(indexLine+1,csv_outString.find("\n",indexLine+1)-indexLine);
		indexLine = csv_outString.find("\n",indexLine+1);
		rowIn.clear();
		indexCSV = 0;
		indexCSVOld = -1;
		i = 0;
		while(titleLine.find(",",indexCSV+1) != std::string::npos) {
			indexCSV = titleLine.find(",",indexCSV+1);
			bufferStr = titleLine.substr(indexCSVOld+1,indexCSV-1-indexCSVOld);
			indexCSVOld = indexCSV;
			bufferStr += ",";
			rowIn += titleNames.at(i);
			rowIn += bufferStr;
			i++;
		}
		output += rowIn + "\n";
	}
	
	//PAUSE STUFF
	std::vector<long long> pauseStarts;
	std::vector<long long> pauseEnds;
	std::vector<std::string> stopVector;
	std::vector<std::string> runNo;
	std::string bufferRow;
	bool getRuns = false, getStop = false;
	OWLTime *bufferOwl;
	indexLine = -1;
	indexCSV = 0;
	indexCSVOld = 0;
	
	while(dummyLBsString.find("\n", indexLine+1) != std::string::npos) {
		bufferRow = dummyLBsString.substr(indexLine+1,dummyLBsString.find("\n",indexLine+1)-indexLine);
		indexLine = dummyLBsString.find("\n",indexLine+1);
		indexCSV = bufferRow.find("\t");
		if(indexCSV == std::string::npos) {
			if(getStop) {
				bufferRow.pop_back();
				stopVector.push_back(bufferRow);
			}
			if(bufferRow.compare("endrun\n") == 0) {
				getRuns = false;
				getStop = true;
			}
			if(getRuns) {
				bufferRow.pop_back();
				runNo.push_back(bufferRow);
			}
			if(bufferRow.compare("end\n") == 0) {
				getRuns = true;
			}
		}
		else {
			bufferStr = bufferRow.substr(0,indexCSV);
			bufferOwl = new OWLTime(bufferStr.c_str());
			pauseStarts.push_back(bufferOwl->total_mksec_utc());
			bufferStr = bufferRow.substr(indexCSV+1,19);
			bufferOwl = new OWLTime(bufferStr.c_str());
			pauseEnds.push_back(bufferOwl->total_mksec_utc());
		}
	}
	
	std::vector<long long> beamStarts;
	std::vector<long long> beamEnds;
	std::vector<std::string> fillNo;
	bool getFills = false;
	indexLine = -1;
	indexCSV = 0;
	indexCSVOld = -1;
	while(beamTimesString.find("\n", indexLine+1) != std::string::npos) {
		bufferRow = beamTimesString.substr(indexLine+1,beamTimesString.find("\n",indexLine+1)-indexLine);
		indexLine = beamTimesString.find("\n",indexLine+1);
		indexCSV = bufferRow.find("\t");
		if(indexCSV == std::string::npos) {
			if(getFills) {
				bufferRow.pop_back();
				fillNo.push_back(bufferRow);
			}
			if(bufferRow.compare("endbeam\n") ==0) {
				getFills = true;
			}
		}
		else {
			bufferStr = bufferRow.substr(0,indexCSV);
			bufferOwl = new OWLTime(bufferStr.c_str());
			beamStarts.push_back(bufferOwl->total_mksec_utc());
			bufferStr = bufferRow.substr(indexCSV+1,19);
			bufferOwl = new OWLTime(bufferStr.c_str());
			beamEnds.push_back(bufferOwl->total_mksec_utc());
		}
	}
	int rowsNo = 0;
	int contor = 0;
	std::string overflows;
	if(runNo.size()) {
		for(int iPauses = 0; iPauses < pauseStarts.size(); iPauses++) {
			for(int iBeams = 0; iBeams < beamStarts.size(); iBeams++) {
				if(beamStarts.at(iBeams) >= pauseStarts.at(iPauses) && beamEnds.at(iBeams) <= pauseEnds.at(iPauses)) {
					overflows += get_date(beamStarts.at(iBeams)/1000000) + "\t" + get_date(beamEnds.at(iBeams)/1000000) + "\t" + std::to_string((beamEnds.at(iBeams)-beamStarts.at(iBeams))/1000);
					overflows += "\t" + fillNo.at(iBeams); 
					overflows += "\t" + runNo.at(iPauses) + "\t";
					for(int inputs = 0; inputs < 35; inputs++) {
						overflows += stopVector.at(contor*35+inputs) + "\t";
					}
					overflows += "\n";
					rowsNo++;
				}
				else if(beamStarts.at(iBeams) >= pauseStarts.at(iPauses) && beamStarts.at(iBeams) < pauseEnds.at(iPauses)) {
					overflows += get_date(beamEnds.at(iBeams)/1000000) + "\t" + get_date(pauseEnds.at(iPauses)/1000000) + "\t" + std::to_string((pauseEnds.at(iPauses)-beamEnds.at(iBeams))/1000);
					overflows += "\t" + fillNo.at(iBeams);
					overflows += "\t" + runNo.at(iPauses) + "\t";
					for(int inputs = 0; inputs < 35; inputs++) {
						overflows += stopVector.at(contor*35+inputs) + "\t";
					}
					overflows += "\n";
					rowsNo++;
				}
				else if(beamEnds.at(iBeams) > pauseStarts.at(iPauses) && beamEnds.at(iBeams) <= pauseEnds.at(iPauses)) {
					overflows += get_date(pauseStarts.at(iPauses)/1000000) + "\t" + get_date(beamEnds.at(iBeams)/1000000) + "\t" + std::to_string((beamEnds.at(iBeams)-pauseStarts.at(iPauses))/1000);
					overflows += "\t" + fillNo.at(iBeams);
					overflows += "\t" + runNo.at(iPauses) + "\t";
					for(int inputs = 0; inputs < 35; inputs++) {
						overflows += stopVector.at(contor*35+inputs) + "\t";
					}
					overflows += "\n";
					rowsNo++;
				}
				else if(pauseStarts.at(iPauses) >= beamStarts.at(iBeams) && pauseEnds.at(iPauses) <= beamEnds.at(iBeams)) {
					overflows += get_date(pauseStarts.at(iPauses)/1000000) + "\t" + get_date(pauseEnds.at(iPauses)/1000000) + "\t" + std::to_string((pauseEnds.at(iPauses)-pauseStarts.at(iPauses))/1000);
					overflows += "\t" + fillNo.at(iBeams);
					overflows += "\t" + runNo.at(iPauses) + "\t";
					for(int inputs = 0; inputs < 35; inputs++) {
						overflows += stopVector.at(contor*35+inputs) + "\t";
					}
					overflows += "\n";
					rowsNo++;
				}
				else if(pauseStarts.at(iPauses) >= beamStarts.at(iBeams) && pauseStarts.at(iPauses) < beamEnds.at(iBeams)) {
					overflows += get_date(pauseStarts.at(iPauses)/1000000) + "\t" + get_date(beamEnds.at(iBeams)/1000000) + "\t" + std::to_string((beamEnds.at(iBeams)-pauseStarts.at(iPauses))/1000);
					overflows += "\t" + fillNo.at(iBeams);
					overflows += "\t" + runNo.at(iPauses) + "\t";
					for(int inputs = 0; inputs < 35; inputs++) {
						overflows += stopVector.at(contor*35+inputs) + "\t";
					}
					overflows += "\n";
					rowsNo++;
				}
				else if(pauseEnds.at(iPauses) > beamStarts.at(iBeams) && pauseEnds.at(iPauses) <= beamEnds.at(iBeams)) {
					overflows += get_date(beamStarts.at(iBeams)/1000000) + "\t" + get_date(pauseEnds.at(iPauses)/1000000) + "\t" + std::to_string((pauseEnds.at(iPauses)-beamStarts.at(iBeams))/1000);
					overflows += "\t" + fillNo.at(iBeams);
					overflows += "\t" + runNo.at(iPauses) + "\t";
					for(int inputs = 0; inputs < 35; inputs++) {
						overflows += stopVector.at(contor*35+inputs) + "\t";
					}
					overflows += "\n";
					rowsNo++;
				}
			}
			contor++;
		}
	}
	rowIn.clear();
	std::vector<std::string> valuesForDummies;
	indexLine = -1;
	while(overflows.find("\n", indexLine+1) != std::string::npos) {
		bufferRow = overflows.substr(indexLine+1,overflows.find("\n",indexLine+1)-indexLine);
		indexLine = overflows.find("\n",indexLine+1);
		indexCSV = 0;
		indexCSVOld = -1;
		while(bufferRow.find("\t",indexCSVOld+1) != std::string::npos) {
			indexCSV = bufferRow.find("\t",indexCSVOld+1);
			valuesForDummies.push_back(bufferRow.substr(indexCSVOld+1,indexCSV-1-indexCSVOld));
			//std::cout << bufferRow.substr(indexCSVOld + 1, indexCSV - 1 - indexCSVOld) << std::endl;
			indexCSVOld = indexCSV;
		}
	}

	titleNames.push_back("Other_S=");
	titleNames.push_back("PIX_S=");
	titleNames.push_back("SCT_S=");
	titleNames.push_back("TRT_S=");
	titleNames.push_back("LAR_S=");
	titleNames.push_back("TIL_S=");
	titleNames.push_back("RPC_S=");
	titleNames.push_back("MDT_S=");
	titleNames.push_back("TGC_S=");
	titleNames.push_back("CSC_S=");
	titleNames.push_back("BCM_S=");
	titleNames.push_back("LUCID_S=");
	titleNames.push_back("ZDC_S=");
	titleNames.push_back("ALFA_S=");
	titleNames.push_back("L1CALO_S=");
	titleNames.push_back("L1CT_S=");
	titleNames.push_back("HLT_S=");
	titleNames.push_back("DAQ_S=");
	titleNames.push_back("IBL_S=");
	titleNames.push_back("DBM_S=");
	titleNames.push_back("MMEGA_S=");
	titleNames.push_back("L1TOPO_S=");
	titleNames.push_back("Solenoid_S=");
	titleNames.push_back("Toroid_S=");
	titleNames.push_back("Cooling_S=");
	titleNames.push_back("LHC_S=");
	titleNames.push_back("FTK_S=");
	titleNames.push_back("AFP_S=");
	titleNames.push_back("Unknown_S=");
	titleNames.push_back("SystemBusy_S=");
	titleNames.push_back("SystemNeedsReconfiguration_S=");
	titleNames.push_back("HardwareFailure_S=");
	titleNames.push_back("BeamDump_S=");
	titleNames.push_back("HandShake_S=");
	titleNames.push_back("TestRun_S=");
	std::string t0_for_stops;
	for(int j = 0; j < rowsNo; j++) {
		rowIn.clear();
		for(int ja = 0; ja < titleNames.size(); ja++) {
			rowIn += titleNames.at(ja);
			if(titleNames.at(ja).compare("StableBeams=") == 0)
				rowIn += "1";
			else if(titleNames.at(ja).compare("StartDate=") == 0)
				rowIn += valuesForDummies.at(40*j);
			else if(titleNames.at(ja).compare("EndDate=") == 0)
				rowIn += valuesForDummies.at(40*j+1);
			else if(titleNames.at(ja).compare("Fill=") == 0)
				rowIn += valuesForDummies.at(40*j+3);
			else if(titleNames.at(ja).compare("Duration=") == 0)
				rowIn += valuesForDummies.at(40*j+2);
			else if (titleNames.at(ja).compare("RunNumber=") == 0) {
				rowIn += valuesForDummies.at(40 * j + 4);
				cool::UInt63 run_number_stops = atoi(valuesForDummies.at(40 * j + 4).c_str());
				cool::ValidityKey key_stops = (run_number_stops << 32) + 1;
				t0_for_stops = get_t0project(key_stops, t0Project_Folder);
			}
			else if (titleNames.at(ja).compare("T0Project=") == 0)
				rowIn += t0_for_stops;
			else if(titleNames.at(ja).compare("Other_S=") == 0)
				rowIn += valuesForDummies.at(40*j+5);
			else if(titleNames.at(ja).compare("PIX_S=") == 0)
				rowIn += valuesForDummies.at(40*j+6);
			else if(titleNames.at(ja).compare("SCT_S=") == 0)
				rowIn += valuesForDummies.at(40*j+7);
			else if(titleNames.at(ja).compare("TRT_S=") == 0)
				rowIn += valuesForDummies.at(40*j+8);
			else if(titleNames.at(ja).compare("LAR_S=") == 0)
				rowIn += valuesForDummies.at(40*j+9);
			else if(titleNames.at(ja).compare("TIL_S=") == 0)
				rowIn += valuesForDummies.at(40*j+10);
			else if(titleNames.at(ja).compare("RPC_S=") == 0)
				rowIn += valuesForDummies.at(40*j+11);
			else if(titleNames.at(ja).compare("MDT_S=") == 0)
				rowIn += valuesForDummies.at(40*j+12);
			else if(titleNames.at(ja).compare("TGC_S=") == 0)
				rowIn += valuesForDummies.at(40*j+13);
			else if(titleNames.at(ja).compare("CSC_S=") == 0)
				rowIn += valuesForDummies.at(40*j+14);
			else if(titleNames.at(ja).compare("BCM_S=") == 0)
				rowIn += valuesForDummies.at(40*j+15);
			else if(titleNames.at(ja).compare("LUCID_S=") == 0)
				rowIn += valuesForDummies.at(40*j+16);
			else if(titleNames.at(ja).compare("ZDC_S=") == 0)
				rowIn += valuesForDummies.at(40*j+17);
			else if(titleNames.at(ja).compare("ALFA_S=") == 0)
				rowIn += valuesForDummies.at(40*j+18);
			else if(titleNames.at(ja).compare("L1CALO_S=") == 0)
				rowIn += valuesForDummies.at(40*j+19);
			else if(titleNames.at(ja).compare("L1CT_S=") == 0)
				rowIn += valuesForDummies.at(40*j+20);
			else if(titleNames.at(ja).compare("HLT_S=") == 0)
				rowIn += valuesForDummies.at(40*j+21);
			else if(titleNames.at(ja).compare("DAQ_S=") == 0)
				rowIn += valuesForDummies.at(40*j+22);
			else if(titleNames.at(ja).compare("IBL_S=") == 0)
				rowIn += valuesForDummies.at(40*j+23);
			else if(titleNames.at(ja).compare("DBM_S=") == 0)
				rowIn += valuesForDummies.at(40*j+24);
			else if(titleNames.at(ja).compare("MMEGA_S=") == 0)
				rowIn += valuesForDummies.at(40*j+25);
			else if(titleNames.at(ja).compare("L1TOPO_S=") == 0)
				rowIn += valuesForDummies.at(40*j+26);
			else if(titleNames.at(ja).compare("Solenoid_S=") == 0)
				rowIn += valuesForDummies.at(40*j+27);
			else if(titleNames.at(ja).compare("Toroid_S=") == 0)
				rowIn += valuesForDummies.at(40*j+28);
			else if(titleNames.at(ja).compare("Cooling_S=") == 0)
				rowIn += valuesForDummies.at(40*j+29);
			else if(titleNames.at(ja).compare("LHC_S=") == 0)
				rowIn += valuesForDummies.at(40*j+30);
			else if(titleNames.at(ja).compare("FTK_S=") == 0)
				rowIn += valuesForDummies.at(40*j+31);
			else if (titleNames.at(ja).compare("AFP_S=") == 0)
				rowIn += valuesForDummies.at(40*j+32);
			else if(titleNames.at(ja).compare("Unknown_S=") == 0)
				rowIn += valuesForDummies.at(40*j+33);
			else if(titleNames.at(ja).compare("SystemBusy_S=") == 0)
				rowIn += valuesForDummies.at(40*j+34);
			else if(titleNames.at(ja).compare("SystemNeedsReconfiguration_S=") == 0)
				rowIn += valuesForDummies.at(40*j+35);
			else if(titleNames.at(ja).compare("HardwareFailure_S=") == 0)
				rowIn += valuesForDummies.at(40*j+36);
			else if(titleNames.at(ja).compare("BeamDump_S=") == 0)
				rowIn += valuesForDummies.at(40*j+37);
			else if(titleNames.at(ja).compare("HandShake_S=") == 0)
				rowIn += valuesForDummies.at(40*j+38);
			else if(titleNames.at(ja).compare("TestRun_S=") == 0)
				rowIn += valuesForDummies.at(40*j+39);
			else
				rowIn += "0";
			rowIn += ",";
		}
		output += rowIn + "\n";
	}

	db->closeDatabase();
	dbd->closeDatabase();
	dbt->closeDatabase();
	dbm->closeDatabase();
	//std::cout << "DB closed." << std::endl;

	std::cout << output;



	return EXIT_SUCCESS;
}
