// eff_make_combined_graphs.cc
//
// Application to produce a root file from efficiency data files
//
// Author: Mihai.Caprini@cern.ch, May 2010   
//   Modifications:
//
//   Author  Date            Reason
//   MC      05/07/2010      Add 3 histograms and 4 axis
//   MC      16/07/2010      Use /tmp/runeff.data and /tmp/runtime.data 
//                            as default files
//   MC      11/10/10        Review axis
//   MC      25/11/10        Use new data format (eff without WStop)
//   MC      10/03/11        Add options -y, -m, -d (start date)
//   MC      14/03/11        Updates for 2011
//   MC      06/04/12        Updates for 2012
//   MC      18/01/13        Updates for 2013
//   MC      06/06/15        Updates for 2015
//

#include <signal.h>
#include <fstream>
#include <vector>
#include <stdlib.h>

#include <cmdl/cmdargs.h>
#include "owl/time.h"
#include "TFile.h"
#include "TDatime.h"
#include "TH1F.h"
#include "TH2F.h"
#include "TGaxis.h"
#include "TGraph.h"
#include "TMultiGraph.h"
#include "TStyle.h"
#include "TCanvas.h"
#include "TPad.h"
#include "TLatex.h"
const char * effFileName;
const char * timeFileName;

extern "C" void signalhandler(int)
{
   signal(SIGINT,  SIG_DFL);
   signal(SIGTERM, SIG_DFL);
}


int main(int argc, char ** argv)
{
   // Declare arguments
   CmdArgStr       eff_file_name('f', "eff-file-name", "eff-file-name", "efficiency file name (default: /tmp/runeff.data).");
   CmdArgStr       time_file_name('p', "time-file-name", "time-file-name", "time file name (default: /tmp/runtime.data).");
   CmdArgInt       year('y', "year", "year", "year (default: 2015).");
   CmdArgInt       month('m', "month", "month", "month (default: 04).");
   CmdArgInt       day('d', "day", "day", "day (default: 01).");
   CmdLine cmd(*argv, &eff_file_name, &time_file_name, &year, &month, &day, NULL);
   CmdArgvIter arg_iter(--argc, ++argv);

   cmd.description("Utility to produce a root file from efficiency data files");

  // Initialize arguments to appropiate default values
  eff_file_name = "/tmp/runeff.data";
  time_file_name = "/tmp/runtime.data";
  year = 2015;
  month = 6;
  day = 1;

  // Parse arguments
  cmd.parse(arg_iter);

  effFileName = eff_file_name;
  timeFileName = time_file_name;

  
  // Install signal handler
  signal(SIGINT,  signalhandler);
  signal(SIGTERM, signalhandler);

  TFile myFile("/tmp/eff_combined.root", "RECREATE");
  TDatime T1(year,month,day,00,00,00);
  int X1 = T1.Convert();
  std::vector<int> seconds;
  std::vector<double> beam;
  std::vector<double> runEff;
  std::vector<double> runEffReady;
  std::vector<int> seconds1;
  std::vector<double> beamTime;
  std::vector<double> runTime;
  std::vector<double> runTimeReady;
  //read data
  std::filebuf fb;
  fb.open (effFileName, std::ios::in);
  if (fb.is_open()) {
    std::istream is(&fb);
    std::string line;
    while (!(line.compare("enddata") == 0)) {
      getline(is, line);
      if (!(line.compare("enddata") == 0)) {
        if (!(line.compare("enddata") == 0)) {
          // read time
          std::string time = line.substr(0, 13);
          long long time_msec = atoll(time.c_str());
          long time_sec = time_msec/1000;
          int short_sec = time_sec - X1;
          seconds.push_back(short_sec);
          // read values
          size_t beamEndIndex = line.find_first_of(" ", 14);
          std::string beamValue = line.substr(14, beamEndIndex-14);
          double val = atof(beamValue.c_str());
          beam.push_back(val);
          size_t runEffEndIndex = line.find_first_of(" ", beamEndIndex+1);
          std::string runEffValue = line.substr(beamEndIndex+1, runEffEndIndex-beamEndIndex-1);
          val = atof(runEffValue.c_str());
          runEff.push_back(val);
          size_t runEffReadyEndIndex = line.find_first_of(" ", runEffEndIndex+1);
          std::string runEffReadyValue = line.substr(runEffEndIndex+1, runEffReadyEndIndex-runEffEndIndex-1);
          val = atof(runEffReadyValue.c_str());
          runEffReady.push_back(val);
        }
      }
    }
  } else {
    std::cout << "Warning: cannot open the first data file" << std::endl;
  }
  fb.close();
  fb.open (timeFileName, std::ios::in);
  if (fb.is_open()) {
    std::istream is(&fb);
    std::string line;
    while (!(line.compare("enddata") == 0)) {
      getline(is, line);
      if (!(line.compare("enddata") == 0)) {
        if (!(line.compare("enddata") == 0)) {
          // read time
          std::string time = line.substr(0, 13);
          long long time_msec = atoll(time.c_str());
          long time_sec = time_msec/1000;
          int short_sec = time_sec - X1;
          seconds1.push_back(short_sec);
          // read values
          size_t beamTimeEndIndex = line.find_first_of(" ", 14);
          std::string beamTimeValue = line.substr(14, beamTimeEndIndex);
          double val = atof(beamTimeValue.c_str());
          beamTime.push_back(val/3600);
          size_t runTimeEndIndex = line.find_first_of(" ", beamTimeEndIndex+1);
          std::string runTimeValue = line.substr(beamTimeEndIndex+1, runTimeEndIndex-beamTimeEndIndex-1);
          val = atof(runTimeValue.c_str());
          runTime.push_back(val/3600);
          size_t runTimeReadyEndIndex = line.find_first_of(" ", runTimeEndIndex+1);
          std::string runTimeReadyValue = line.substr(runTimeEndIndex+1, runTimeReadyEndIndex-runTimeEndIndex-1);
          val = atof(runTimeReadyValue.c_str());
          runTimeReady.push_back(val/3600);
        }
      }
    }
  } else {
    std::cout << "Warning: cannot open the second data file" << std::endl;
  }
  fb.close();

  int n = seconds.size();
/*
  for (int i=0; i<n; i++) {
    std::cout << seconds[i] << " " << beam[i] << " " << runEff[i] << " " << runEffReady[i] << std::endl;
  }
*/
  double beamTimeMax = 0;
  for (int i=0; i<n; i++) {
    if (beamTimeMax < beam[i]) {
      beamTimeMax = beam[i];
    }
  }
  double timeMax = (double)seconds[n-1];
  TGraph *beamGraph = new TGraph (n);
  for (int i=0; i<n; i++) {
    beamGraph->SetPoint(i, (double)seconds[i], 0.7+0.3*beam[i]/beamTimeMax);
  }
  TGraph *beamGraph1 = new TGraph (n);
  for (int i=0; i<n; i++) {
    beamGraph1->SetPoint(i, (double)seconds[i], beam[i]/beamTimeMax);
  }
  TGraph *runEffGraph = new TGraph (n);
  for (int i=0; i<n; i++) {
    runEffGraph->SetPoint(i, (double)seconds[i], runEff[i]);
  }
  TGraph *runEffReadyGraph = new TGraph (n);
  for (int i=0; i<n; i++) {
    runEffReadyGraph->SetPoint(i, (double)seconds[i], runEffReady[i]);
  }
  beamGraph->SetName("BeamGraph");
  beamGraph->GetXaxis()->SetTimeDisplay(1);
  beamGraph->GetXaxis()->SetTimeOffset(X1);
  beamGraph->GetXaxis()->SetTimeFormat("%d/%m");
  beamGraph->GetXaxis()->SetTitle("Date in 2015");
  beamGraph->GetYaxis()->SetTitle("Run Efficiency");
  beamGraph1->SetName("BeamGraph1");
  beamGraph1->GetXaxis()->SetTimeDisplay(1);
  beamGraph1->GetXaxis()->SetTimeOffset(X1);
  beamGraph1->GetXaxis()->SetTimeFormat("%d/%m");
  beamGraph1->GetXaxis()->SetTitle("Date in 2015");
  beamGraph1->GetYaxis()->SetTitle("Run Efficiency");
  runEffGraph->SetName("RunEffGraph");
  runEffGraph->GetXaxis()->SetTimeDisplay(1);
  runEffGraph->GetXaxis()->SetTimeOffset(X1);
  runEffGraph->GetXaxis()->SetTimeFormat("%d/%m");
  runEffGraph->GetXaxis()->SetTitle("Date in 2015");
  runEffGraph->GetYaxis()->SetTitle("Run Efficiency");
  runEffReadyGraph->SetName("RunEffReadyGraph");
  runEffReadyGraph->GetXaxis()->SetTimeDisplay(1);
  runEffReadyGraph->GetXaxis()->SetTimeOffset(X1);
  runEffReadyGraph->GetXaxis()->SetTimeFormat("%d/%m");
  runEffReadyGraph->GetXaxis()->SetTitle("Date in 2015");
  runEffReadyGraph->GetYaxis()->SetTitle("Run Efficiency");

  TH2F * ht0 = new TH2F("ht0"," ",3000,0.,timeMax+86400,10,0,1.30);
  ht0->GetXaxis()->SetTimeOffset(X1);
  ht0->GetXaxis()->SetTimeFormat("%d/%m");
  ht0->GetXaxis()->SetTitle("Date in 2015");
  ht0->GetYaxis()->SetTitle("Run Efficiency");
  ht0->GetXaxis()->SetTitleOffset(1.2);
  ht0->GetXaxis()->SetTimeDisplay(1);

  TH2F * ht = new TH2F("ht"," ",3000,0.,timeMax+86400,10,0.70,1.05);
  ht->GetXaxis()->SetTimeOffset(X1);
  ht->GetXaxis()->SetTimeFormat("%d/%m");
  ht->GetXaxis()->SetTitle("Date in 2015");
  ht->GetYaxis()->SetTitle("Run Efficiency");
  ht->GetXaxis()->SetTitleOffset(1.2);
  ht->GetXaxis()->SetTimeDisplay(1);

  TH2F * ht2 = new TH2F("ht2"," ",3000,0.,timeMax+86400,10,0.65,1.05);
  ht2->GetXaxis()->SetTimeOffset(X1);
  ht2->GetXaxis()->SetTimeFormat("%d/%m");
  ht2->GetXaxis()->SetTitle("Date in 2015");
  ht2->GetYaxis()->SetTitle("Run Efficiency");
  ht2->GetXaxis()->SetTitleOffset(1.2);
  ht2->GetXaxis()->SetTimeDisplay(1);
  ht2->GetXaxis()->SetNdivisions(8);

  TH2F * ht3 = new TH2F("ht3"," ",3000,0.,timeMax+86400,10,0.40,1.10);
  ht3->GetXaxis()->SetTimeOffset(X1);
  ht3->GetXaxis()->SetTimeFormat("%d/%m");
  ht3->GetXaxis()->SetTitle("Date in 2015");
  ht3->GetYaxis()->SetTitle("Run Efficiency");
  ht3->GetXaxis()->SetTitleOffset(1.2);
  ht3->GetXaxis()->SetTimeDisplay(1);
  ht3->GetXaxis()->SetNdivisions(8);

  n = seconds1.size();
  seconds1.push_back(seconds1[n-1]+1000);
  beamTime.push_back(beamTime[n-1]);
  runTime.push_back(runTime[n-1]);
  runTimeReady.push_back(runTimeReady[n-1]);
  seconds1.push_back(seconds1[n-1]+1010);
  beamTime.push_back(0);
  runTime.push_back(0);
  runTimeReady.push_back(0);
  n = n+2;
/*
  for (int i=0; i<n; i++) {
    std::cout << seconds1[i] << " " << beamTime[i] << " " << runTime[i] << " " << runTimeReady[i] << std::endl;
  }
*/
  TGraph *beamTimeGraph = new TGraph (n);
  for (int i=0; i<n; i++) {
    beamTimeGraph->SetPoint(i, (double)seconds1[i], beamTime[i]);
  }
  TGraph *runTimeGraph = new TGraph (n);
  for (int i=0; i<n; i++) {
    runTimeGraph->SetPoint(i, (double)seconds1[i], runTime[i]);
  }
  TGraph *runTimeReadyGraph = new TGraph (n);
  for (int i=0; i<n; i++) {
    runTimeReadyGraph->SetPoint(i, (double)seconds1[i], runTimeReady[i]);
  }
  beamTimeGraph->SetName("BeamTimeGraph");
  beamTimeGraph->GetXaxis()->SetTimeDisplay(1);
  beamTimeGraph->GetXaxis()->SetTimeOffset(X1);
  beamTimeGraph->GetXaxis()->SetTimeFormat("%d/%m");
  beamTimeGraph->GetXaxis()->SetTitle("Date in 2015");
  beamTimeGraph->GetYaxis()->SetTitle("Time (hours)");
  runTimeGraph->SetName("RunTimeGraph");
  runTimeGraph->GetXaxis()->SetTimeDisplay(1);
  runTimeGraph->GetXaxis()->SetTimeOffset(X1);
  runTimeGraph->GetXaxis()->SetTimeFormat("%d/%m");
  runTimeGraph->GetXaxis()->SetTitle("Date in 2015");
  runTimeGraph->GetYaxis()->SetTitle("Time (hours)");
  runTimeReadyGraph->SetName("RunTimeReadyGraph");
  runTimeReadyGraph->GetXaxis()->SetTimeDisplay(1);
  runTimeReadyGraph->GetXaxis()->SetTimeOffset(X1);
  runTimeReadyGraph->GetXaxis()->SetTimeFormat("%d/%m");
  runTimeReadyGraph->GetXaxis()->SetTitle("Date in 2015");
  runTimeReadyGraph->GetYaxis()->SetTitle("Time (hours)");

//  TGaxis *axis_hours = new TGaxis(timeMax+3*86400,0.7,timeMax+3*86400,1,0,beamTimeMax,50100,"+L");
  TGaxis *axis_hours = new TGaxis(timeMax+86400,0.7,timeMax+86400,1.05,0,0.35*beamTimeMax/0.3,50100,"+L");
  axis_hours->SetTitle("Beam Time (hours)");
  axis_hours->SetLabelSize(0.05);
  axis_hours->SetLabelFont(42);
  axis_hours->SetTitleOffset(0.8);
  axis_hours->SetTitleSize(0.05);
  axis_hours->SetTitleFont(42);
  axis_hours->SetLineColor(kRed);
  axis_hours->SetLabelColor(kRed);
  axis_hours->SetTitleColor(kRed);
  axis_hours->SetNdivisions(1005);
  axis_hours->SetName("axis_hours");
  TGaxis *axis_hours1 = new TGaxis(timeMax+86400,0,timeMax+86400,1.05,0,beamTimeMax,50100,"+L");
  axis_hours1->SetTitle("Beam Time (hours)");
  axis_hours1->SetLabelSize(0.05);
  axis_hours1->SetLabelFont(42);
  axis_hours1->SetTitleOffset(0.8);
  axis_hours1->SetTitleSize(0.05);
  axis_hours1->SetTitleFont(42);
  axis_hours1->SetLineColor(kRed);
  axis_hours1->SetLabelColor(kRed);
  axis_hours1->SetTitleColor(kRed);
  axis_hours1->SetNdivisions(1005);
  axis_hours1->SetName("axis_hours1");

//  double beamPercentageMax = (beamTimeMax/168)*100;
//  TGaxis *axis_percent = new TGaxis(timeMax+3*86400,0.7,timeMax+3*86400,1,0,beamPercentageMax,50100,"+L");
  TGaxis *axis_percent = new TGaxis(timeMax+86400,0.7,timeMax+86400,1.05,0,0.35*beamTimeMax/0.3,50100,"+L");
  axis_percent->SetTitle("Beam Availability (%)");
  axis_percent->SetLabelSize(0.05);
  axis_percent->SetLabelFont(42);
  axis_percent->SetTitleOffset(0.8);
  axis_percent->SetTitleSize(0.05);
  axis_percent->SetTitleFont(42);
  axis_percent->SetLineColor(kRed);
  axis_percent->SetLabelColor(kRed);
  axis_percent->SetTitleColor(kRed);
  axis_percent->SetNdivisions(1005);
  axis_percent->SetName("axis_percent");

//  TGaxis *axis2 = new TGaxis(timeMax+3*86400,0.65,timeMax+3*86400,1,0,1,50100,"+L");
  TGaxis *axis2 = new TGaxis(timeMax+86400,0.65,timeMax+86400,1,0,1,50100,"+L");
  axis2->SetTitle("Beam");
  axis2->SetLabelSize(0.05);
  axis2->SetLabelFont(42);
  axis2->SetTitleOffset(0.5);
  axis2->SetTitleSize(0.05);
  axis2->SetTitleFont(42);
  axis2->SetLineColor(kGray+2);
  axis2->SetNdivisions(1);
  axis2->SetName("axis2");

//  TGaxis *axis3 = new TGaxis(timeMax+3*86400,0.4,timeMax+3*86400,1,0,1,50100,"+L");
  TGaxis *axis3 = new TGaxis(timeMax+86400,0.4,timeMax+86400,1,0,1,50100,"+L");
  axis3->SetTitle("Beam");
  axis3->SetLabelSize(0.05);
  axis3->SetLabelFont(42);
  axis3->SetTitleOffset(0.5);
  axis3->SetTitleSize(0.05);
  axis3->SetTitleFont(42);
  axis3->SetLineColor(kGray+2);
  axis3->SetNdivisions(1);
  axis3->SetName("axis3");

  ht0->Write();
  ht->Write();
  ht2->Write();
  ht3->Write();
  beamGraph->Write();
  beamGraph1->Write();
  runEffGraph->Write();
  runEffReadyGraph->Write();
  beamTimeGraph->Write();
  runTimeGraph->Write();
  runTimeReadyGraph->Write();
  axis_hours->Write();
  axis_hours1->Write();
  axis_percent->Write();
  axis2->Write();
  axis3->Write();

  myFile.Close();
  return EXIT_SUCCESS;
}
