// eff_make_csv_file.cc
//
// Application to produce a csv file with run efficiency data
//
// Author: Mihai.Caprini@cern.ch, October 2010
//   Modifications:
//
//   Author  Date            Reason
//    MC     25/10/2010      add delay between StableBeams flag and
//                             ReadyForPhysics flag
//    MC     01/11/2010      add delay from the WarmStop to the end of
//                             StableBeams
//    MC     05/11/2010      add busy evaluation
//    MC     10/11/2010      add efficiency evaluation eliminating the
//                             slot after the WarmStop
//    MC     15/11/2010      review delay evaluation
//    MC     18/11/2010      add run stopped information
//    MC     23/11/2010      use new data format (including eff no WarmStop)
//    MC     09/02/2011      use Stop Reason
//    MC     30/03/2011      add new Stop Reasons
//    MC     10/06/2011      add Simple and Complex dead time
//    MC     23/08/2011      review delay evaluation
//    MC     06/09/2011      add Derand-2 dead time
//    MC     11/06/2012      correct delay evaluation for RfP flag changes
//                             during a run
//    MC     26/06/2012      add hold trigger info
//    MC     18/02/2015      replace cmdline by boost
//    MC     18/02/2015      Use new cond db for Run 2 (CONDBR2)
//    MC     18/02/2015      Add option:  -r (atlasRun)
//    MC     02/06/2015      Use BUSYRATES_RUN2 folder
//

#include <signal.h>
#include <fstream>
#include <vector>
#include <iostream>
#include <sstream>

#include <boost/program_options.hpp>

#include <CoralBase/Attribute.h>
#include <CoolKernel/Exception.h>
#include <CoolKernel/ValidityKey.h>
#include <CoolKernel/IDatabaseSvc.h>
#include <CoolKernel/IDatabase.h>
#include <CoolKernel/IFolder.h>
#include <CoolKernel/IRecordSpecification.h>
#include <CoolKernel/Record.h>
#include <CoolApplication/DatabaseSvcFactory.h>
#include <CoolKernel/IObjectIterator.h>
#include <CoolKernel/IObject.h>
#include <CoolKernel/ChannelSelection.h>

#include <cmdl/cmdargs.h>
#include "owl/time.h"
#include "ers/ers.h"

namespace daq {
ERS_DECLARE_ISSUE(      rc,
                        BadCoolDb,
                        "Retrieving data from COOL failed because " << explanation,
                        ((const char*) explanation)
                )
}

static     cool::IDatabasePtr db;
static     cool::IDatabasePtr dbd;
static     cool::IDatabasePtr dbt;
static     cool::IDatabasePtr dbm;

static     std::string FillState_Name = "/LHC/DCS/FILLSTATE";
static     std::string LB_Name = "/TRIGGER/LUMI/LBLB";
static     std::string LBT_Name = "/TRIGGER/LUMI/LBTIME";
static     std::string DataTakingMode_Name = "/TDAQ/RunCtrl/DataTakingMode";
static     std::string BusyConf_Name = "/TRIGGER/LVL1/BUSYCONF";
static     std::string BusyRate_Name = "/TRIGGER/LVL1/BUSYRATES_RUN2";
static     std::string StopReason_Name = "/TDAQ/RunCtrl/StopReason";
static     std::string HoldTrigger_Name = "/TDAQ/RunCtrl/HoldTriggerReason";

static     cool::IFolderPtr FillState_Folder;
static     cool::IFolderPtr LB_Folder;
static     cool::IFolderPtr LBT_Folder;
static     cool::IFolderPtr DataTakingMode_Folder;
static     cool::IFolderPtr BusyConf_Folder;
static     cool::IFolderPtr BusyRate_Folder;
static     cool::IFolderPtr StopReason_Folder;
static     cool::IFolderPtr HoldTrigger_Folder;

static     std::map<std::string, std::string> ctpMap;

// const char * eff_file_name;

extern "C" void signalhandler(int)
{
   signal(SIGINT,  SIG_DFL);
   signal(SIGTERM, SIG_DFL);
}

// throws daq::rc::BadCoolDb
void openDB(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

  try {
    db = db_Svc.openDatabase(db_Id, false); // false = !read_only
  }
  catch (std::exception &ex) {
    std::cout << "Failure opening DB database - " << ex.what() << std::endl;
    daq::rc::BadCoolDb i(ERS_HERE, ex.what());
    throw i;
  }
}

// throws daq::rc::BadCoolDb
void openDBD(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

  try {
    dbd = db_Svc.openDatabase(db_Id, false); // false = !read_only
  } catch (std::exception &ex) {
    std::cout << "Failure opening DCS database - " << ex.what() << std::endl;
    daq::rc::BadCoolDb i (ERS_HERE, ex.what());
    throw i;
  }
}

// throws daq::rc::BadCoolDb
void openDBT(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

  try {
    dbt = db_Svc.openDatabase(db_Id, false); // false = !read_only
  }
  catch (std::exception &ex) {
    std::cout << "Failure opening DBT database - " << ex.what() << std::endl;
    daq::rc::BadCoolDb i(ERS_HERE, ex.what());
    throw i;
  }
}

// throws daq::rc::BadCoolDb
void openDBM(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

  try {
    dbm = db_Svc.openDatabase(db_Id, false); // false = !read_only
  } catch (std::exception &ex) {
    std::cout << "Failure opening DBM database - " << ex.what() << std::endl;
    daq::rc::BadCoolDb i(ERS_HERE, ex.what());
    throw i;
  }
}

std::string get_date(int seconds) {
   OWLTime owl_date = OWLTime(seconds);
   long year = owl_date.year();
   unsigned short month = owl_date.month();
   unsigned short day = owl_date.day();
   std::string date_string = std::string(owl_date.c_str());
   size_t lastIndex = date_string.size();
   std::string time_string = date_string.substr(lastIndex-8, lastIndex);
   std::ostringstream stm;
   stm << year;
   stm << "-";
   if (month < 10) {
     stm << "0";
   }
   stm << month << "-";
   if (day < 10) {
     stm << "0";
   }
   stm << day;
   stm << " ";
   stm << time_string;
   return stm.str();
}

int ready_for_physics(cool::ValidityKey key, cool::IFolderPtr & datatakingmode_folder) {
  int flag = 1;
  cool::IObjectPtr obj = datatakingmode_folder->findObject(key, 0);
  std::string readyForPhysics = obj->payloadValue("ReadyForPhysics");
  flag = atoi(readyForPhysics.c_str());
  return flag;
}

double busy(long long startTime, long long endTime, std::string identifier, cool::IFolderPtr & the_busy_folder ) {

  // Find the busy values for a specific detector in this lumiblock:

  // Note: this also includes records that are partialy outside this lb

  cool::IObjectIteratorPtr objects = the_busy_folder->browseObjects( cool::ValidityKey(startTime),cool::ValidityKey(endTime), 0 );
  double totaltime=0;
  double totalbusy=0;

  while ( objects->goToNext() ) {
    const cool::IObject& obj = objects->currentRef();

    std::string this_busy_string= obj.payloadValue(identifier.c_str());
    double this_busy = atof(this_busy_string.c_str());

    double this_starttime = obj.since();
    double this_stoptime = obj.until();
    totalbusy+=this_busy*(this_stoptime-this_starttime);
    totaltime+=(this_stoptime-this_starttime);
  }
  if (totaltime>0) {
    totalbusy = 0.01* totalbusy/totaltime;
  } else {
    totalbusy = 0; // if no info, let's just say this detector was not busy
  }
  return totalbusy;
}

float busy_run2(cool::ValidityKey key, std::string identifier, cool::IFolderPtr & the_busy_folder ) {

  // Find the busy values for a specific detector in this lumiblock:

  cool::IObjectPtr objects = the_busy_folder->findObject(key, 0);
  std::string this_busy_string= objects->payloadValue(identifier.c_str());
  float this_busy = atof(this_busy_string.c_str());
  // std::cout << "for " << identifier << " busy = " << this_busy << std::endl;
  return this_busy;
}

void prepareCtpMap(int runNumber) {
  // Populate ctp map for the start run (no change for the next run ...)
  ctpMap.clear();
  cool::UInt63 a = runNumber;
  cool::UInt63 z = runNumber + 1;
  cool::ValidityKey sKey = (a <<32) + 1;
  cool::ValidityKey uKey = z <<32;
  cool::IObjectIteratorPtr objs = BusyConf_Folder->browseObjects(sKey, uKey, cool::ChannelSelection::all());
  while ( objs->goToNext() ) {
    const cool::IObject& obj = objs->currentRef();
    std::string identifier = obj.payloadValue("identifier");
    std::string description = obj.payloadValue("description");
    std::string busyEnabled = obj.payloadValue("busyEnabled");
    if (busyEnabled.compare("TRUE") == 0) {
      ctpMap.insert(std::pair<std::string, std::string>(identifier, description) );
    }
  }
  std::cout << "For run number = " << runNumber << ":" << std::endl;
/*
  std::map<std::string,std::string>::iterator iter;
  for ( iter=ctpMap.begin() ; iter != ctpMap.end(); iter++ ) {
    std::cout << "    " << (*iter).first << "  " << (*iter).second << std::endl;
  }
*/
}

std::string get_stop_reason(long long starttime, long long endtime) {
  // starttime in msec
  long long start_time_nsec = ((long long)(starttime))*1000000;
  long long end_time_nsec = ((long long)(endtime))*1000000;
  cool::IObjectIteratorPtr sr_objects = StopReason_Folder->browseObjects(cool::ValidityKey(start_time_nsec), cool::ValidityKey(end_time_nsec), 0);
  std::ostringstream strm;
  while ( sr_objects->goToNext() ) {
    const cool::IObject& obj = sr_objects->currentRef();
    std::string subSystems = obj.payloadValue("SubSystems");
    std::string reason = obj.payloadValue("Reason");
    strm << subSystems << "/" << reason << ";";
  }
  return strm.str();
}

int main(int argc, char ** argv)
{

   std::string effFileName = "/tmp/runeff.data";
   int atlasRun = 2;
   boost::program_options::options_description desc("Utility to produce a csv file with info on deadtime sources");
   desc.add_options()
    ("help,h", "help message")
    ("effFileName,f", boost::program_options::value<std::string>(&effFileName)->default_value(effFileName), "Efficiency file name (default: /tmp/runeff.data).")
    ("atlasRun,r", boost::program_options::value<int>(&atlasRun)->default_value(atlasRun), "ATLAS Run: 1 - Run 1; 2 - Run 2 (default 2)")
    ;
   boost::program_options::variables_map vm;
   boost::program_options::store(boost::program_options::parse_command_line(argc,argv,desc), vm);
   boost::program_options::notify(vm);
   if (vm.count("help")) {
     std::cout << desc << std::endl;
     return EXIT_SUCCESS;
   }
   if ((atlasRun < 1) || (atlasRun > 2)) {
     std::cout << "Command line error: atlasRun should be 1 or 2" << std::endl;
     return EXIT_FAILURE;
   }
 /*
   // Declare arguments
   CmdArgStr       eff_file_name('f', "eff-file-name", "eff-file-name", "efficiency file name (default: /tmp/runeff.data).");
   CmdLine cmd(*argv, &eff_file_name, NULL);
   CmdArgvIter arg_iter(--argc, ++argv);

   cmd.description("Utility to produce a csv file with info on deadtime sources");

  // Initialize arguments to appropiate default values
  eff_file_name = "/tmp/runeff.data";

  // Parse arguments
  cmd.parse(arg_iter);

  effFileName = eff_file_name;
*/

  char * eff_file_name = new char [effFileName.length()+1];
  std::strcpy (eff_file_name, effFileName.c_str());
  // eff_file_name = effFileName; 

  // Install signal handler
  signal(SIGINT,  signalhandler);
  signal(SIGTERM, signalhandler);

  // open cool databases
  std::string coolDb;
  std::string coolDbTrig;
  std::string coolDbDCS;
  std::string coolDbMon;
  if (atlasRun == 1) {
    coolDbDCS = "oracle://ATLAS_COOLPROD;schema=ATLAS_COOLOFL_DCS;dbname=COMP200;user=ATLAS_COOL_READER;password=COOLRED4PRO";
  } else if (atlasRun == 2) {
    coolDbDCS = "oracle://ATLAS_COOLPROD;schema=ATLAS_COOLOFL_DCS;dbname=CONDBR2;user=ATLAS_COOL_READER;password=COOLRED4PRO";
  } 
  cool::DatabaseId dbIdDCS = coolDbDCS;
  cool::IDatabaseSvc& dbSvcDCS = cool::DatabaseSvcFactory::databaseService();
  try {
    openDBD(dbIdDCS, dbSvcDCS);
  }
  catch(daq::rc::BadCoolDb& ex) {
      ers::fatal(ex);
      exit(0);
  }
  if (atlasRun == 1) {
    coolDbTrig = "oracle://ATLAS_COOLPROD;schema=ATLAS_COOLONL_TRIGGER;dbname=COMP200";
  } else if (atlasRun == 2) {
    coolDbTrig = "oracle://ATLAS_COOLPROD;schema=ATLAS_COOLONL_TRIGGER;dbname=CONDBR2";
  }
  cool::DatabaseId dbIdTrig = coolDbTrig;
  cool::IDatabaseSvc& dbSvcTrig = cool::DatabaseSvcFactory::databaseService();
  try {
    openDBT(dbIdTrig, dbSvcTrig);
  }
  catch(daq::rc::BadCoolDb& ex) {
      ers::fatal(ex);
      exit(0);
  }
  if (atlasRun == 1) {
    coolDb = "oracle://ATLAS_COOLPROD;schema=ATLAS_COOLONL_TDAQ;dbname=COMP200;user=ATLAS_COOL_READER;password=COOLRED4PRO";
  } else if (atlasRun == 2) {
    coolDb = "oracle://ATLAS_COOLPROD;schema=ATLAS_COOLONL_TDAQ;dbname=CONDBR2;user=ATLAS_COOL_READER;password=COOLRED4PRO";
  }
  cool::DatabaseId dbId = coolDb;
  // Open or Create DB
  cool::IDatabaseSvc& dbSvc = cool::DatabaseSvcFactory::databaseService();
  try {
    openDB(dbId, dbSvc);
  }
  catch(daq::rc::BadCoolDb& ex) {
      ers::fatal(ex);
      exit(0);
  }
  if (atlasRun == 1) {
    coolDbMon = "oracle://ATLAS_COOLPROD;schema=ATLAS_COOLONL_TRIGGER;dbname=MONP200";
  } else if (atlasRun == 2) {
    coolDbMon = "oracle://ATLAS_COOLPROD;schema=ATLAS_COOLONL_TRIGGER;dbname=MONP200";
  }
  cool::DatabaseId dbIdMon = coolDbMon;
  cool::IDatabaseSvc& dbSvcMon = cool::DatabaseSvcFactory::databaseService();
  try {
    openDBM(dbIdMon, dbSvcMon);
  }
  catch(daq::rc::BadCoolDb& ex) {
      ers::fatal(ex);
      exit(0);
  }

  // check folders  
  try {
    if (!dbd->existsFolder(FillState_Name)) {
      std::cout << "DB Folder FillState does not exist???" << std::endl;
    }
  } catch (std::exception &e) {
    std::cout << "DB Folder FillState does not exist. " << e.what() << std::endl;
  }
    if (!dbt->existsFolder(LB_Name)) {
    std::cout << "DB Folder LB does not exist???" << std::endl;
  }
  if (!dbt->existsFolder(LBT_Name)) {
    std::cout << "DB Folder LBT does not exist???" << std::endl;
  }
  if (!db->existsFolder(DataTakingMode_Name)) {
    std::cout << "DB Folder DataTakingMode does not exist???" << std::endl;
  }
  try {
    if (!db->existsFolder(StopReason_Name)) {
        std::cout << "DB Folder StopReason does not exist???" << std::endl;
     }
  } catch (std::exception &e) {
     std::cout << "DB Folder StopReason does not exist. " << e.what() << std::endl;
  }
  try {
    if (!db->existsFolder(HoldTrigger_Name)) {
        std::cout << "DB Folder HoldTrigger does not exist???" << std::endl;
     }
  } catch (std::exception &e) {
     std::cout << "DB Folder HoldTrigger does not exist. " << e.what() << std::endl;
  }
  try {
     if (!dbm->existsFolder(BusyRate_Name)) {
        std::cout << "DB Folder BusyRate does not exist???" << std::endl;
     }
  } catch (std::exception &e) {
     std::cout << "DB Folder BusyRate does not exist. " << e.what() << std::endl;
  }
  try {
     if (!dbm->existsFolder(BusyConf_Name)) {
        std::cout << "DB Folder BusyConf does not exist???" << std::endl;
     }
  } catch (std::exception &e) {
     std::cout << "DB Folder BusyConf does not exist. " << e.what() << std::endl;
  }


  // open folders
  try {
    FillState_Folder = dbd->getFolder(FillState_Name);
  }
  catch (std::exception &e) {
     std::cout << "Cannot open FillState folder " << e.what() << std::endl;
     dbd->closeDatabase();
     std::cout << "DBD closed." << std::endl;

     exit(0);
  }
  LB_Folder = dbt->getFolder(LB_Name);
  LBT_Folder = dbt->getFolder(LBT_Name);
  DataTakingMode_Folder = db->getFolder(DataTakingMode_Name);
  try {
    StopReason_Folder = db->getFolder(StopReason_Name);
  } catch (std::exception &e) {
    std::cout << "Cannot open StopReason folder " << e.what() << std::endl;
    db->closeDatabase();
    std::cout << "DB closed." << std::endl;
    exit(0);
  }
  try {
    HoldTrigger_Folder = db->getFolder(HoldTrigger_Name);
  } catch (std::exception &e) {
    std::cout << "Cannot open HoldTrigger folder " << e.what() << std::endl;
    db->closeDatabase();
    std::cout << "DB closed." << std::endl;
    exit(0);
  }
  try {
    BusyRate_Folder = dbm->getFolder(BusyRate_Name);
  }
  catch (std::exception &e) {
     std::cout << "Cannot open BusyRate folder " << e.what() << std::endl;
     dbm->closeDatabase();
     std::cout << "DBM closed." << std::endl;

     exit(0);
  }

  try {
    BusyConf_Folder = dbm->getFolder(BusyConf_Name);
  }
  catch (std::exception &e) {
     std::cout << "Cannot open BusyConf folder " << e.what() << std::endl;
     dbm->closeDatabase();
     std::cout << "DBM closed." << std::endl;

     exit(0);
  }

  std::vector<int> index;
  std::vector<int> fill;
  std::vector<std::string> startTime;
  std::vector<std::string> endTime;
  std::vector<double> beam;
  std::vector<double> runEff;
  std::vector<double> runEffReady;
  std::vector<double> runEffReadyWithoutWarmStop;
  std::vector<double> delay;
  std::vector<double> end_delay;
  std::vector<double> runStopped;
  std::vector<double> CTP_LUCID_busy;
  std::vector<double> Pixel_busy;
  std::vector<double> SCT_busy;
  std::vector<double> TRT_busy;
  std::vector<double> L1Calo_busy;
  std::vector<double> BCM_busy;
  std::vector<double> ZDC_busy;
  std::vector<double> LAr_HF_busy;
  std::vector<double> LAr_EMEC_busy;
  std::vector<double> LAr_EMB_busy;
  std::vector<double> MDT_B_busy;
  std::vector<double> MDT_EC_busy;
  std::vector<double> Tile_EB_busy;
  std::vector<double> Tile_LB_busy;
  std::vector<double> CSC_busy;
  std::vector<double> TGC_busy;
  std::vector<double> RPC_busy;
  std::vector<double> MUCTPI_busy;
  std::vector<double> VME_busy;
  std::vector<double> SMPL_busy;
  std::vector<double> CPX0_busy;
  std::vector<double> Derand2_busy;
  std::vector<double> Other_stopped;
  std::vector<double> PIX_stopped;
  std::vector<double> SCT_stopped;
  std::vector<double> TRT_stopped;
  std::vector<double> LAR_stopped;
  std::vector<double> TIL_stopped;
  std::vector<double> RPC_stopped;
  std::vector<double> MDT_stopped;
  std::vector<double> TGC_stopped;
  std::vector<double> CSC_stopped;
  std::vector<double> BCM_stopped;
  std::vector<double> LUCID_stopped;
  std::vector<double> ZDC_stopped;
  std::vector<double> ALFA_stopped;
  std::vector<double> L1CALO_stopped;
  std::vector<double> L1CT_stopped;
  std::vector<double> HLT_stopped;
  std::vector<double> DAQ_stopped;
  std::vector<double> Solenoid_stopped;
  std::vector<double> Toroid_stopped;
  std::vector<double> Cooling_stopped;
  std::vector<double> LHC_stopped;
  std::vector<double> Unknown_reason;
  std::vector<double> SystemBusy_reason;
  std::vector<double> SystemNeedsReconfiguration_reason;
  std::vector<double> HardwareFailure_reason;
  std::vector<double> BeamDump_reason;
  std::vector<double> HandShake_reason;
  std::vector<double> TestRun_reason;
  std::vector<double> Other_holdTrigger;
  std::vector<double> PIX_holdTrigger;
  std::vector<double> SCT_holdTrigger;
  std::vector<double> TRT_holdTrigger;
  std::vector<double> LAR_holdTrigger;
  std::vector<double> TIL_holdTrigger;
  std::vector<double> RPC_holdTrigger;
  std::vector<double> MDT_holdTrigger;
  std::vector<double> TGC_holdTrigger;
  std::vector<double> CSC_holdTrigger;
  std::vector<double> BCM_holdTrigger;
  std::vector<double> LUCID_holdTrigger;
  std::vector<double> ZDC_holdTrigger;
  std::vector<double> ALFA_holdTrigger;
  std::vector<double> L1CALO_holdTrigger;
  std::vector<double> L1CT_holdTrigger;
  std::vector<double> HLT_holdTrigger;
  std::vector<double> DAQ_holdTrigger;
  std::vector<double> Unknown_holdTriggerReason;
  std::vector<double> TTCRestart_holdTriggerReason;
  std::vector<double> Resync_holdTriggerReason;
  std::vector<double> StoplessRecovery_holdTriggerReason;
  std::vector<double> ManualRecovery_holdTriggerReason;
  std::vector<double> Test_holdTriggerReason;

  int nr_lines = 0;
  std::vector<std::string> buffer;
  //read old csv data
  std::filebuf fb;
  fb.open ("/tmp/runeff_csv_old.data", std::ios::in);
  if (fb.is_open()) {
    std::istream is(&fb);
    std::string line;
    getline(is, line);
    while (!is.eof()) {
      getline(is, line);
      buffer.push_back(line);
      nr_lines = nr_lines +1;
    }
    std::cout << " nr_lines = " << nr_lines << std::endl;
/*
    for (int i = 0; i < nr_lines; i++) {
      std::cout << buffer[i] << std::endl;
    }
*/
  } else {
    std::cout << "Warning: cannot open the old csv data file" << std::endl;
  }
  fb.close();

  //read data
  fb.open (eff_file_name, std::ios::in);
  if (fb.is_open()) {
    std::istream is(&fb);
    std::string line;
    std::string line1;
    int n = 0;
    if (nr_lines > 1) {
      n = nr_lines - 1;
    }
    long start_time_sec = 0;
    long end_time_sec = 0;
    int current_run_number = 0;
    while (!(line1.compare("enddata") == 0)) {
      n = n + 1; 
      // read start time
      getline(is, line);
      getline(is, line1);
      if (!(line1.compare("enddata") == 0)) {
        std::string start_time = line.substr(0, 10);
        start_time_sec = atol(start_time.c_str());
        // get start time
        std::string start_string = get_date(start_time_sec);
        startTime.push_back(start_string);
        // get fill number
        long long start_search_nsec = ((long long)(start_time_sec-1))*1000000000;
        long long end_search_nsec = ((long long)(start_time_sec+1))*1000000000;
        int fill_number = 1000;
        cool::IObjectIteratorPtr objects = FillState_Folder->browseObjects( cool::ValidityKey(start_search_nsec),cool::ValidityKey(end_search_nsec), cool::ChannelSelection::all() );
        while ( objects->goToNext() ) {
          const cool::IObject& obj = objects->currentRef();
          std::string fillNumber = obj.payloadValue("FillNumber");
          fill_number = atoi(fillNumber.c_str());
        }
        fill.push_back(fill_number);
        // read values
        size_t beamEndIndex = line1.find_first_of(" ", 14);
        size_t runEffEndIndex = line1.find_first_of(" ", beamEndIndex+1);
        std::string runEffValue = line1.substr(beamEndIndex+1, runEffEndIndex-beamEndIndex-1);
        double run_eff = atof(runEffValue.c_str());
        runEff.push_back(run_eff);
        size_t runEffReadyEndIndex = line1.find_first_of(" ", runEffEndIndex+1);
        std::string runEffReadyValue = line1.substr(runEffEndIndex+1, runEffReadyEndIndex-runEffEndIndex-1);
        double run_eff_ready = atof(runEffReadyValue.c_str());
        runEffReady.push_back(run_eff_ready);
        size_t lastIndex = line1.size();
        std::string runEffReadyNoWarmStopValue = line1.substr(runEffReadyEndIndex+1, lastIndex-runEffReadyEndIndex-1);
        double run_eff_ready_noWS = atof(runEffReadyNoWarmStopValue.c_str());
        runEffReadyWithoutWarmStop.push_back(run_eff_ready_noWS);
        // read end time
        getline(is, line);
        if (!(line.compare("enddata") == 0)) {
          getline(is, line);
          std::string end_time = line.substr(0, 10);
          end_time_sec = atol(end_time.c_str());
          std::string end_string = get_date(end_time_sec);
          endTime.push_back(end_string);
          long duration_sec = end_time_sec - start_time_sec;
          double duration_hours = ((double)duration_sec)/3600;
          beam.push_back(duration_hours);
          index.push_back(n);
     
          // evaluate delay and end delay
          // evaluate busy sources
          // find list of runs and lumi blocks
          std::cout << "Fill = " << fill_number << std::endl;
          long long start_time_nsec = ((long long)(start_time_sec))*1000000000;
          long long end_time_nsec = ((long long)(end_time_sec))*1000000000;
          cool::IObjectIteratorPtr rl_objects = LBT_Folder->browseObjects(cool::ValidityKey(start_time_nsec), cool::ValidityKey(end_time_nsec), 0);

          // initialize the busy map
          std::map<std::string, double> busyMap;
          busyMap["CTP(LUCID)"] = 0;
          busyMap["Pixel    "] = 0;
          busyMap["SCT"] = 0;
          busyMap["TRT"] = 0;
          busyMap["L1Calo"] = 0;
          busyMap["BCM"] = 0;
          busyMap["ZDC"] = 0;
          busyMap["LAr H/F"] = 0;
          busyMap["LAr EMEC"] = 0;
          busyMap["LAr EMB"] = 0;
          busyMap["MDT B"] = 0;
          busyMap["MDT EC"] = 0;
          busyMap["Tile EB"] = 0;
          busyMap["Tile LB"] = 0;
          busyMap["CSC"] = 0;
          busyMap["TGC"] = 0;
          busyMap["RPC"] = 0;
          busyMap["MUCTPI"] = 0;
          busyMap["VME"] = 0;
          busyMap["SMPL"] = 0;
          busyMap["CPX0"] = 0;
          busyMap["Derand2"] = 0;
    
          // initialize the stop systems map
          std::map<std::string, double> stopSystemsMap;
          stopSystemsMap["Other"] = 0;
          stopSystemsMap["PIX"] = 0;
          stopSystemsMap["SCT"] = 0;
          stopSystemsMap["TRT"] = 0;
          stopSystemsMap["LAR"] = 0;
          stopSystemsMap["TIL"] = 0;
          stopSystemsMap["RPC"] = 0;
          stopSystemsMap["MDT"] = 0;
          stopSystemsMap["TGC"] = 0;
          stopSystemsMap["CSC"] = 0;
          stopSystemsMap["BCM"] = 0;
          stopSystemsMap["LUCID"] = 0;
          stopSystemsMap["ZDC"] = 0;
          stopSystemsMap["ALFA"] = 0;
          stopSystemsMap["L1CALO"] = 0;
          stopSystemsMap["L1CT"] = 0;
          stopSystemsMap["HLT"] = 0;
          stopSystemsMap["DAQ"] = 0;
          stopSystemsMap["Solenoid"] = 0;
          stopSystemsMap["Toroid"] = 0;
          stopSystemsMap["Cooling"] = 0;
          stopSystemsMap["LHC"] = 0;

          // initialize the stop reason map
          std::map<std::string, double> stopReasonMap;
          stopReasonMap["Unknown"] = 0;
          stopReasonMap["SystemBusy"] = 0;
          stopReasonMap["SystemNeedsReconfiguration"] = 0;
          stopReasonMap["HardwareFailure"] = 0;
          stopReasonMap["BeamDump"] = 0;
          stopReasonMap["HandShake"] = 0;
          stopReasonMap["TestRun"] = 0;

          // initialize the hold trigger systems map
          std::map<std::string, double> holdTriggerSystemsMap;
          holdTriggerSystemsMap["Other"] = 0;
          holdTriggerSystemsMap["PIX"] = 0;
          holdTriggerSystemsMap["SCT"] = 0;
          holdTriggerSystemsMap["TRT"] = 0;
          holdTriggerSystemsMap["LAR"] = 0;
          holdTriggerSystemsMap["TIL"] = 0;
          holdTriggerSystemsMap["RPC"] = 0;
          holdTriggerSystemsMap["MDT"] = 0;
          holdTriggerSystemsMap["TGC"] = 0;
          holdTriggerSystemsMap["CSC"] = 0;
          holdTriggerSystemsMap["BCM"] = 0;
          holdTriggerSystemsMap["LUCID"] = 0;
          holdTriggerSystemsMap["ZDC"] = 0;
          holdTriggerSystemsMap["ALFA"] = 0;
          holdTriggerSystemsMap["L1CALO"] = 0;
          holdTriggerSystemsMap["L1CT"] = 0;
          holdTriggerSystemsMap["HLT"] = 0;
          holdTriggerSystemsMap["DAQ"] = 0;

          // initialize the hold trigger reason map
          std::map<std::string, double> holdTriggerReasonMap;
          holdTriggerReasonMap["Unknown"] = 0;
          holdTriggerReasonMap["TTCRestart"] = 0;
          holdTriggerReasonMap["Resync"] = 0;
          holdTriggerReasonMap["StoplessRecovery"] = 0;
          holdTriggerReasonMap["ManualRecovery"] = 0;
          holdTriggerReasonMap["Test"] = 0;

          long long start_of_first_LB_with_ready_sec = 0;
          long long end_of_last_LB_with_ready_sec = 0;
          long long start_of_first_LB_with_ready_nsec = 0;
          long long end_of_last_LB_with_ready_nsec = 0;
          bool firstRun = true;
          bool newRun = false;
          bool readyForPhysicsSet = false;
          double end_of_last_run_sec = (double) start_time_sec;
          double start_of_new_run_sec = (double) start_time_sec;
          double runStopped_sec = 0;
          double delay_sec = 0;
          double current_run_delay_sec = 0;
          int current_ready_flag = 0;
          int old_ready_flag = 0;
          while ( rl_objects->goToNext() ) {
            const cool::IObject& obj = rl_objects->currentRef();
            std::string runNumber = obj.payloadValue("Run");
            std::string lumiBlock = obj.payloadValue("LumiBlock");
            int run_nb = atoi(runNumber.c_str());
            if (run_nb != current_run_number) {
              std::cout << " New RN " << run_nb << " old RN was " << current_run_number << std::endl;
              newRun = true;
              current_run_number = run_nb;
              end_of_last_LB_with_ready_sec = 0;
              end_of_last_LB_with_ready_nsec = 0;
            }
            cool::UInt63 rn = run_nb;
            cool::UInt63 lb = atoi(lumiBlock.c_str());
            cool::ValidityKey key1 = (rn<<32) + lb;
            cool::IObjectPtr obj_lb = LB_Folder->findObject(key1, 0);
            std::string st = obj_lb->payloadValue("StartTime");
            std::string et = obj_lb->payloadValue("EndTime");
            long long sT = atoll(st.c_str());
            long long eT = atoll(et.c_str());
            if (newRun) {
              readyForPhysicsSet = false;
              if (firstRun == true) {
                prepareCtpMap(run_nb);
                firstRun = false;
              } else {
              // check delay for previous run
                if (current_run_delay_sec == 0) {
                  // no ready flag during previous run, add run duration to delay
                  delay_sec = delay_sec + end_of_last_run_sec - start_of_new_run_sec;
                  std::cout << " from previous run: delay = " << delay_sec << std::endl;
                }
              }
              current_run_delay_sec = 0;
              start_of_new_run_sec = (double)sT/1000000000;
              // check if lumiblock start was before start of fill
              if (start_of_new_run_sec < (double) start_time_sec ) {
                start_of_new_run_sec = (double) start_time_sec;
              }
              std::cout << " run stopped sec = " << (start_of_new_run_sec - end_of_last_run_sec) << std::endl;
              if ((start_of_new_run_sec - end_of_last_run_sec) > 0) {
                runStopped_sec = runStopped_sec + (start_of_new_run_sec - end_of_last_run_sec);
                std::string stopReason = get_stop_reason(end_of_last_run_sec*1000, start_of_new_run_sec*1000);
//                stopReason = "TRT, LAR, LUCID, /SystemBusy, Unknown, ";
                if (stopReason.size() > 3) {
                  size_t index = stopReason.find_first_of("/");
                  std::string stop_systems = stopReason.substr(0, index);
                  size_t lastIndex = stopReason.size();
                  std::string stop_reason = stopReason.substr(index+1, lastIndex-index);
                  size_t firstIndex = 0;
                  index = 0;
                  while (index != std::string::npos) {
                    index = stop_systems.find(", ", firstIndex);
                    if (index != std::string::npos) {
                      std::string system = stop_systems.substr(firstIndex, index-firstIndex);
                      stopSystemsMap[system] = stopSystemsMap[system] + (start_of_new_run_sec - end_of_last_run_sec);
                      firstIndex = index+2;
                    }
                  }
                  index = 0;
                  firstIndex = 0;
                  while (index != std::string::npos) {
                    index = stop_reason.find(", ", firstIndex);
                    if (index != std::string::npos) {
                      std::string reason = stop_reason.substr(firstIndex, index-firstIndex);
                      stopReasonMap[reason] = stopReasonMap[reason] + (start_of_new_run_sec - end_of_last_run_sec);
                      firstIndex = index+2;
                    }
                  }
                }
              }
              newRun = false;
            }
            end_of_last_run_sec = (double)eT/1000000000;
            long long lumi_block_duration_ns = eT - sT;
            double lumi_block_duration_msec = (double)lumi_block_duration_ns/1000000;
            double lumi_block_duration_sec = lumi_block_duration_msec/1000;
            current_ready_flag = ready_for_physics(key1, DataTakingMode_Folder);
            // std::cout << "ready_flag = " << current_ready_flag << " at lb = " << lb << std::endl;
            if ((current_ready_flag == 1) && (old_ready_flag == 0) && (readyForPhysicsSet == false)) {
              // end of start delay
              readyForPhysicsSet = true;
              start_of_first_LB_with_ready_sec = sT/1000000000;
              start_of_first_LB_with_ready_nsec = sT;
              current_run_delay_sec = start_of_first_LB_with_ready_sec - start_of_new_run_sec;
              delay_sec = delay_sec + current_run_delay_sec;
              std::cout << " new delay = " << current_run_delay_sec << std::endl;
              std::cout << " new total delay = " << delay_sec << std::endl; 
            }
            if (current_ready_flag == 1) { 
              end_of_last_LB_with_ready_sec = eT/1000000000;
              end_of_last_LB_with_ready_nsec = eT;

              // evaluate dead time for each sub-detector
              std::map<std::string,std::string>::iterator iter;
              for ( iter=ctpMap.begin() ; iter != ctpMap.end(); iter++ ) {
                // double det_busy = busy(sT,eT, (*iter).first, BusyRate_Folder );
                float det_busy = 0;
                try {
                  det_busy = busy_run2(key1, (*iter).first, BusyRate_Folder );
                 } catch (std::exception &ex) {
                   std::cout << "Failure in busy_run2 - " << ex.what() << std::endl;
                 }  
                 busyMap[(*iter).second] = busyMap.find((*iter).second)->second + lumi_block_duration_sec*det_busy;
                 // std::cout << "for lb = " << lb << " and for " << (*iter).second << " busy is " << busyMap[(*iter).second] << std::endl;
               }
               float det_busy = 0;
               try {
                 det_busy = busy_run2(key1, "ctpmi_vme_rate", BusyRate_Folder );
               } catch (std::exception &ex) {
                 std::cout << "Failure in busy_run2 - " << ex.what() << std::endl;
               }      
               busyMap["VME"] = busyMap.find("VME")->second + lumi_block_duration_sec*det_busy;
               det_busy = 0;
               try {
                 det_busy = busy_run2(key1, "ctpcore_moni3_rate", BusyRate_Folder );
               } catch (std::exception &ex) {
                 std::cout << "Failure in busy_run2 - " << ex.what() << std::endl;
               }
               busyMap["SMPL"] = busyMap.find("SMPL")->second + lumi_block_duration_sec*det_busy;
               det_busy = 0;
               try {
                 det_busy = busy_run2(key1, "ctpcore_moni1_rate", BusyRate_Folder );
               } catch (std::exception &ex) {
                 std::cout << "Failure in busy_run2 - " << ex.what() << std::endl;
               }
               busyMap["CPX0"] = busyMap.find("CPX0")->second + lumi_block_duration_sec*det_busy;
               det_busy = 0;
               try {
                 det_busy = busy_run2(key1, "ctpcore_rate", BusyRate_Folder );
               } catch (std::exception &ex) {
                 std::cout << "Failure in busy_run2 - " << ex.what() << std::endl;
               }          
               busyMap["Derand2"] = busyMap.find("Derand2")->second + lumi_block_duration_sec*det_busy;
/*  code for Run 1:
              double det_busy = busy(sT,eT, "ctpmi_vme_rate", BusyRate_Folder );
              busyMap["VME"] = busyMap.find("VME")->second + lumi_block_duration_sec*det_busy;
              // std::cout << " for lb = " << lb << "  det_busy = " << det_busy << "  busy duration = " << lumi_block_duration_sec*det_busy << " and total vme busy = " << busyMap["VME"] << std::endl;
              det_busy = busy(sT,eT, "ctpcore_moni2_rate", BusyRate_Folder );
              busyMap["SMPL"] = busyMap.find("SMPL")->second + lumi_block_duration_sec*det_busy;
              det_busy = busy(sT,eT, "ctpcore_moni3_rate", BusyRate_Folder );
              busyMap["CPX0"] = busyMap.find("CPX0")->second + lumi_block_duration_sec*det_busy;
              det_busy = busy(sT,eT, "ctpcore_mon_rate", BusyRate_Folder );
              busyMap["Derand2"] = busyMap.find("Derand2")->second + lumi_block_duration_sec*det_busy;
            }
*/
            }
            old_ready_flag = current_ready_flag;
          }
          // evaluate the hold trigger info
          cool::IObjectIteratorPtr ht_objects = HoldTrigger_Folder->browseObjects(cool::ValidityKey(start_of_first_LB_with_ready_nsec), cool::ValidityKey(end_of_last_LB_with_ready_nsec), 0);
              
          while ( ht_objects->goToNext() ) {
            const cool::IObject& ht_obj = ht_objects->currentRef();
            double ht_starttime = ht_obj.since();
            double ht_stoptime = ht_obj.until();
            double ht_duration_msec = (ht_stoptime - ht_starttime)/1000000;
            std::string subSystems = ht_obj.payloadValue("SubSystems");
            std::string reason = ht_obj.payloadValue("Reason");
            if (!(subSystems.compare("None") == 0)) {
              // use milisec here
              holdTriggerSystemsMap[subSystems] = holdTriggerSystemsMap[subSystems] + ht_duration_msec;
              // std::cout << " holdTrigger[" << subSystems << "] = " << ht_duration_msec/1000 << std::endl;
              holdTriggerReasonMap[reason] = holdTriggerReasonMap[reason] + ht_duration_msec;
            }
          }  

/*
          std::map<std::string,double>::iterator it;
          for ( it=busyMap.begin() ; it != busyMap.end(); it++ ) {
            std::cout << "---->    " << (*it).first << "  " << (*it).second << std::endl;
          }
*/
          double delay_minutes = ((double) (end_time_sec - start_time_sec))/60;
          if (delay_sec > 0) {
            delay_minutes = ((double) (delay_sec))/60;
          }
          double end_delay_minutes = 0;
          if (end_of_last_LB_with_ready_sec > 0) {
            end_delay_minutes = ((double)((long long)(end_time_sec) - end_of_last_LB_with_ready_sec))/60;
            if (end_delay_minutes < 0) {
              end_delay_minutes = 0;
            }
          }
          delay.push_back(delay_minutes);
          end_delay.push_back(end_delay_minutes);
          double runStopped_minutes = runStopped_sec/60;
          runStopped.push_back(runStopped_minutes);
          CTP_LUCID_busy.push_back(busyMap["CTP(LUCID)"]);
          Pixel_busy.push_back(busyMap["Pixel    "]);
          SCT_busy.push_back(busyMap["SCT"]);
          TRT_busy.push_back(busyMap["TRT"]);
          L1Calo_busy.push_back(busyMap["L1Calo"]);
          BCM_busy.push_back(busyMap["BCM"]);
          ZDC_busy.push_back(busyMap["ZDC"]);
          LAr_HF_busy.push_back(busyMap["LAr H/F"]);
          LAr_EMEC_busy.push_back(busyMap["LAr EMEC"]);
          LAr_EMB_busy.push_back(busyMap["LAr EMB"]);
          MDT_B_busy.push_back(busyMap["MDT B"]);
          MDT_EC_busy.push_back(busyMap["MDT EC"]);
          Tile_EB_busy.push_back(busyMap["Tile EB"]);
          Tile_LB_busy.push_back(busyMap["Tile LB"]);
          CSC_busy.push_back(busyMap["CSC"]);
          TGC_busy.push_back(busyMap["TGC"]);
          RPC_busy.push_back(busyMap["RPC"]);
          MUCTPI_busy.push_back(busyMap["MUCTPI"]);
          VME_busy.push_back(busyMap["VME"]);
          SMPL_busy.push_back(busyMap["SMPL"]);
          CPX0_busy.push_back(busyMap["CPX0"]);
          Derand2_busy.push_back(busyMap["Derand2"]);
          Other_stopped.push_back(stopSystemsMap["Other"]);
          PIX_stopped.push_back(stopSystemsMap["PIX"]);
          SCT_stopped.push_back(stopSystemsMap["SCT"]);
          TRT_stopped.push_back(stopSystemsMap["TRT"]);
          LAR_stopped.push_back(stopSystemsMap["LAR"]);
          TIL_stopped.push_back(stopSystemsMap["TIL"]);
          RPC_stopped.push_back(stopSystemsMap["RPC"]);
          MDT_stopped.push_back(stopSystemsMap["MDT"]);
          TGC_stopped.push_back(stopSystemsMap["TGC"]);
          CSC_stopped.push_back(stopSystemsMap["CSC"]);
          BCM_stopped.push_back(stopSystemsMap["BCM"]);
          LUCID_stopped.push_back(stopSystemsMap["LUCID"]);
          ZDC_stopped.push_back(stopSystemsMap["ZDC"]);
          ALFA_stopped.push_back(stopSystemsMap["ALFA"]);
          L1CALO_stopped.push_back(stopSystemsMap["L1CALO"]);
          L1CT_stopped.push_back(stopSystemsMap["L1CT"]);
          HLT_stopped.push_back(stopSystemsMap["HLT"]);
          DAQ_stopped.push_back(stopSystemsMap["DAQ"]);
          Solenoid_stopped.push_back(stopSystemsMap["Solenoid"]);
          Toroid_stopped.push_back(stopSystemsMap["Toroid"]);
          Cooling_stopped.push_back(stopSystemsMap["Cooling"]);
          LHC_stopped.push_back(stopSystemsMap["LHC"]);
          Unknown_reason.push_back(stopReasonMap["Unknown"]);
          SystemBusy_reason.push_back(stopReasonMap["SystemBusy"]);
          SystemNeedsReconfiguration_reason.push_back(stopReasonMap["SystemNeedsReconfiguration"]);
          HardwareFailure_reason.push_back(stopReasonMap["HardwareFailure"]);
          BeamDump_reason.push_back(stopReasonMap["BeamDump"]);
          HandShake_reason.push_back(stopReasonMap["HandShake"]);
          TestRun_reason.push_back(stopReasonMap["TestRun"]);
          // use sec now
          Other_holdTrigger.push_back(holdTriggerSystemsMap["Other"]/1000);
          PIX_holdTrigger.push_back(holdTriggerSystemsMap["PIX"]/1000);
          SCT_holdTrigger.push_back(holdTriggerSystemsMap["SCT"]/1000);
          TRT_holdTrigger.push_back(holdTriggerSystemsMap["TRT"]/1000);
          LAR_holdTrigger.push_back(holdTriggerSystemsMap["LAR"]/1000);
          TIL_holdTrigger.push_back(holdTriggerSystemsMap["TIL"]/1000);
          RPC_holdTrigger.push_back(holdTriggerSystemsMap["RPC"]/1000);
          MDT_holdTrigger.push_back(holdTriggerSystemsMap["MDT"]/1000);
          TGC_holdTrigger.push_back(holdTriggerSystemsMap["TGC"]/1000);
          CSC_holdTrigger.push_back(holdTriggerSystemsMap["CSC"]/1000);
          BCM_holdTrigger.push_back(holdTriggerSystemsMap["BCM"]/1000);
          LUCID_holdTrigger.push_back(holdTriggerSystemsMap["LUCID"]/1000);
          ZDC_holdTrigger.push_back(holdTriggerSystemsMap["ZDC"]/1000);
          ALFA_holdTrigger.push_back(holdTriggerSystemsMap["ALFA"]/1000);
          L1CALO_holdTrigger.push_back(holdTriggerSystemsMap["L1CALO"]/1000);
          L1CT_holdTrigger.push_back(holdTriggerSystemsMap["L1CT"]/1000);
          HLT_holdTrigger.push_back(holdTriggerSystemsMap["HLT"]/1000);
          DAQ_holdTrigger.push_back(holdTriggerSystemsMap["DAQ"]/1000);
          Unknown_holdTriggerReason.push_back(holdTriggerReasonMap["Unknown"]/1000);
          TTCRestart_holdTriggerReason.push_back(holdTriggerReasonMap["TTCRestart"]/1000);
          Resync_holdTriggerReason.push_back(holdTriggerReasonMap["Resync"]/1000);
          StoplessRecovery_holdTriggerReason.push_back(holdTriggerReasonMap["StoplessRecovery"]/1000);
          ManualRecovery_holdTriggerReason.push_back(holdTriggerReasonMap["ManualRecovery"]/1000);
          Test_holdTriggerReason.push_back(holdTriggerReasonMap["Test"]/1000);

        }
      }
    }
  } else {
    std::cout << "Warning: cannot open the first data file" << std::endl;
  }
  fb.close();

  int n = index.size();
  // prepare csv file
  std::ostream * csv_out = new std::ofstream ("/tmp/runeff_csv.data", std::ios::out | std::ios::trunc);
  if (!(*csv_out)) {
    std::cout << "Can't create the csv data file " << std::endl;
  }
  (*csv_out) << "Index,Fill,StartDate,EndDate,Duration,Delay,EndDelay,RunStopped,Eff,Eff-Phys,Eff-Phys-noWarmStop,BCM,CSC,CTPLUCID,L1Calo,LArEMB,LArEMEC,LArHF,MDTB,MDTEC,MUCTPI,Pixel,RPC,SCT,TGC,TRT,TileEB,TileLB,CTPMI_VME,ZDC,SMPL,CPX0,Other_S,PIX_S,SCT_S,TRT_S,LAR_S,TIL_S,RPC_S,MDT_S,TGC_S,CSC_S,BCM_S,LUCID_S,ZDC_S,ALPA_S,L1CALO_S,L1CT_S,HLT_S,DAQ_S,Solenoid_S,Toroid_S,Cooling_S,LHC_S,Unknown,SystemBusy,SystemNeedsReconfiguration,HardwareFailure,BeamDump,HandShake,TestRun,Derand2,Other_HT,PIX_HT,SCT_HT,TRT_HT,LAR_HT,TIL_HT,RPC_HT,MDT_HT,TGC_HT,CSC_HT,BCM_HT,LUCID_HT,ZDC_HT,ALPA_HT,L1CALO_HT,L1CT_HT,HLT_HT,DAQ_HT,UnknownHT,TTCRestart,Resync,StoplessRecovery,ManualRecovery,Test" << std::endl;
//  (*csv_out).precision(3);
  for (int i = 0; i < nr_lines-1; i++) {
    (*csv_out) << buffer[i] << std::endl;
  }
  for (int i=0; i<n; i++) {
    if (index[i] < 10) {
       (*csv_out) << "00";
    } else if (index[i] < 100) {
       (*csv_out) << "0";
    }
    (*csv_out).precision(3);
    (*csv_out) << index[i] << "," << fill[i] << "," << startTime[i] << "," << endTime[i] << "," << beam[i] << "," << delay[i] << "," << end_delay[i] << "," << runStopped[i] << "," << runEff[i]*100 << "," << runEffReady[i]*100 << "," << runEffReadyWithoutWarmStop[i]*100 << ",";
    (*csv_out).precision(6);
    (*csv_out) << BCM_busy[i] << ","; 
    (*csv_out) << CSC_busy[i] << ",";
    (*csv_out) << CTP_LUCID_busy[i] << ",";
    (*csv_out) << L1Calo_busy[i] << ",";
    (*csv_out) << LAr_EMB_busy[i] << ",";
    (*csv_out) << LAr_EMEC_busy[i] << ",";
    (*csv_out) << LAr_HF_busy[i] << ",";
    (*csv_out) << MDT_B_busy[i] << ",";
    (*csv_out) << MDT_EC_busy[i] << ",";
    (*csv_out) << MUCTPI_busy[i] << ",";
    (*csv_out) << Pixel_busy[i] << ",";
    (*csv_out) << RPC_busy[i] << ",";
    (*csv_out) << SCT_busy[i] << ",";
    (*csv_out) << TGC_busy[i] << ",";
    (*csv_out) << TRT_busy[i] << ",";
    (*csv_out) << Tile_EB_busy[i] << ",";
    (*csv_out) << Tile_LB_busy[i] << ",";
    (*csv_out) << VME_busy[i] << ",";
    (*csv_out) << ZDC_busy[i] << ",";
    (*csv_out) << SMPL_busy[i] << ",";
    (*csv_out) << CPX0_busy[i] << ",";
    (*csv_out) << Other_stopped[i] << ",";
    (*csv_out) << PIX_stopped[i] << ",";
    (*csv_out) << SCT_stopped[i] << ",";
    (*csv_out) << TRT_stopped[i] << ",";
    (*csv_out) << LAR_stopped[i] << ",";
    (*csv_out) << TIL_stopped[i] << ",";
    (*csv_out) << RPC_stopped[i] << ",";
    (*csv_out) << MDT_stopped[i] << ",";
    (*csv_out) << TGC_stopped[i] << ",";
    (*csv_out) << CSC_stopped[i] << ",";
    (*csv_out) << BCM_stopped[i] << ",";
    (*csv_out) << LUCID_stopped[i] << ",";
    (*csv_out) << ZDC_stopped[i] << ",";
    (*csv_out) << ALFA_stopped[i] << ",";
    (*csv_out) << L1CALO_stopped[i] << ",";
    (*csv_out) << L1CT_stopped[i] << ",";
    (*csv_out) << HLT_stopped[i] << ",";
    (*csv_out) << DAQ_stopped[i] << ",";
    (*csv_out) << Solenoid_stopped[i] << ",";
    (*csv_out) << Toroid_stopped[i] << ",";
    (*csv_out) << Cooling_stopped[i] << ",";
    (*csv_out) << LHC_stopped[i] << ",";
    (*csv_out) << Unknown_reason[i] << ",";
    (*csv_out) << SystemBusy_reason[i] << ",";
    (*csv_out) << SystemNeedsReconfiguration_reason[i] << ",";
    (*csv_out) << HardwareFailure_reason[i] << ",";
    (*csv_out) << BeamDump_reason[i] << ",";
    (*csv_out) << HandShake_reason[i] << ",";
    (*csv_out) << TestRun_reason[i] << ",";
    (*csv_out) << Derand2_busy[i] << ",";
    (*csv_out) << Other_holdTrigger[i] << ",";
    (*csv_out) << PIX_holdTrigger[i] << ",";
    (*csv_out) << SCT_holdTrigger[i] << ",";
    (*csv_out) << TRT_holdTrigger[i] << ",";
    (*csv_out) << LAR_holdTrigger[i] << ",";
    (*csv_out) << TIL_holdTrigger[i] << ",";
    (*csv_out) << RPC_holdTrigger[i] << ",";
    (*csv_out) << MDT_holdTrigger[i] << ",";
    (*csv_out) << TGC_holdTrigger[i] << ",";
    (*csv_out) << CSC_holdTrigger[i] << ",";
    (*csv_out) << BCM_holdTrigger[i] << ",";
    (*csv_out) << LUCID_holdTrigger[i] << ",";
    (*csv_out) << ZDC_holdTrigger[i] << ",";
    (*csv_out) << ALFA_holdTrigger[i] << ",";
    (*csv_out) << L1CALO_holdTrigger[i] << ",";
    (*csv_out) << L1CT_holdTrigger[i] << ",";
    (*csv_out) << HLT_holdTrigger[i] << ",";
    (*csv_out) << DAQ_holdTrigger[i] << ",";
    (*csv_out) << Unknown_holdTriggerReason[i] << ",";
    (*csv_out) << TTCRestart_holdTriggerReason[i] << ",";
    (*csv_out) << Resync_holdTriggerReason[i] << ",";
    (*csv_out) << StoplessRecovery_holdTriggerReason[i] << ",";
    (*csv_out) << ManualRecovery_holdTriggerReason[i] << ",";
    (*csv_out) << Test_holdTriggerReason[i];


    (*csv_out) << std::endl;
  }
  dbd->closeDatabase();
  dbt->closeDatabase();
  dbm->closeDatabase();
  std::cout << "DB closed." << std::endl;

  return EXIT_SUCCESS;
}
