//   eff_run_efficiency.cc
//
//   Application to evaluate run efficiency
//   
//   Author: Adrian.Chitan@cern.ch, January 2015
//

//   13 Feb 2015
//   Introduced the atlasRun command line option to use the Run 2 Cool DB (CONDBR2)

#include <iostream>
#include <sstream>
#include <stdlib.h> // for getenv
#include <signal.h>
#include <fstream>
#include <vector>

#include <boost/program_options.hpp>

#include <CoralBase/Attribute.h>
#include <CoolKernel/Exception.h>
#include <CoolKernel/ValidityKey.h>
#include <CoolKernel/IDatabaseSvc.h>
#include <CoolKernel/IDatabase.h>
#include <CoolKernel/IFolder.h>
#include <CoolKernel/IRecordSpecification.h>
#include <CoolKernel/Record.h>
#include <CoolApplication/DatabaseSvcFactory.h>
#include <CoolKernel/IObjectIterator.h>
#include <CoolKernel/IObject.h>
#include <CoolKernel/ChannelSelection.h>

#include "ers/ers.h"
#include "owl/time.h"

#include "pmg/pmg_initSync.h"

#include "RunControl/Common/Exceptions.h"

namespace daq {
	ERS_DECLARE_ISSUE(rc,
		BadCoolDb,
		"Retrieving data from COOL failed because " << explanation,
		((const char*)explanation)
		)
}

static     cool::IDatabasePtr db;
static     cool::IDatabasePtr dbt;
static     cool::IDatabasePtr dbd;
static     cool::IDatabasePtr dbm;

static     std::string CTP_Name = "/TRIGGER/LVL1/CTPCORELBDATA";
static     std::string LB_Name = "/TRIGGER/LUMI/LBLB";
static     std::string LBT_Name = "/TRIGGER/LUMI/LBTIME";
static     std::string FillState_Name = "/LHC/DCS/FILLSTATE";
static     std::string TotalInt_Name = "/LHC/DCS/BPTX/TOTALINT";
static     std::string L1Counters_Name = "/TRIGGER/LUMI/LVL1COUNTERS";
static     std::string L1Menu_Name = "/TRIGGER/LVL1/Menu";
static     std::string DataTakingMode_Name = "/TDAQ/RunCtrl/DataTakingMode";
static	   std::string busyConf_Name = "/TRIGGER/LVL1/BUSYCONF";
static	   std::string busyRates_Name = "/TRIGGER/LVL1/BUSYRATES_RUN2";

static     cool::IFolderPtr CTP_Folder;
static     cool::IFolderPtr FillState_Folder;
static     cool::IFolderPtr TotalInt_Folder;
static     cool::IFolderPtr LB_Folder;
static     cool::IFolderPtr LBT_Folder;
static     cool::IFolderPtr L1Counters_Folder;
static     cool::IFolderPtr L1Menu_Folder;
static     cool::IFolderPtr DataTakingMode_Folder;
static	   cool::IFolderPtr busyConf_Folder;
static	   cool::IFolderPtr busyRates_Folder;

//static     std::map<std::string, std::string> ctpMap;

static     int runLimit = 5;
static     int writeBeamFile = 1;
static     int display = 0;
static     std::string triggerItem = "L1_EM30";
static     int atlasRun = 2;
static     int useTotalBusy = 1;

static     int totalRunTime = 0;
static     int totalGoodRunTime = 0;
static     int totalBusyTime = 0;
static     int realRunTime = 0;

static     int integralBeamTime = 0;
static     int integralBeamRunningTime = 0;
static     int integralBeamRunningReadyTime = 0;
static     int integralDeadTime = 0;
static     int integralReadyDeadTime = 0;
static     int integralRunningTime = 0;
static     int integralRunningReadyTime = 0;
static     double integralRunEff = 0;
static     double integralRunEffReady = 0;
static	   long long totalTimeEvaluated = 0;

static     std::vector<long long> start_run_vector;
static     std::vector<long long> end_run_vector;

static     std::vector<long long> run_start_vector;
static     std::vector<long long> run_end_vector;

static     std::vector<long long> runs_start_vector;
static     std::vector<long long> runs_end_vector;

static     std::vector<std::string> runeff_data;
static     std::vector<std::string> runtime_data;
static     std::vector<std::string> ctpMap;

int ctp_busy;
int pixel_busy;
int sct_busy;
int trt_busy;
int l1calo_busy;
int bcm_busy;
int zdc_busy;
int larhf_busy;
int laremec_busy;
int laremb_busy;
int mdtb_busy;
int mdtec_busy;
int tileeb_busy;
int tilelb_busy;
int csc_busy;
int tgc_busy;
int rpc_busy;
int muctpi_busy;
int vme_busy;
int smpl_busy;
int cpx0_busy;
int derand2_busy;

extern "C" void signalhandler(int)
{
	signal(SIGINT, SIG_DFL);
	signal(SIGTERM, SIG_DFL);
	std::cout << "Terminating eff_run_efficiency ..." << std::endl;
}

// throws daq::rc::BadCoolDb
void openDB(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

	try {
		db = db_Svc.openDatabase(db_Id, false); // false = !read_only
	}
	catch (std::exception &ex) {
		std::cout << "Failure opening DB database - " << ex.what() << std::endl;
		daq::rc::BadCoolDb i(ERS_HERE, ex.what());
		throw i;
	}
}

// throws daq::rc::BadCoolDb
void openDBT(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

	try {
		dbt = db_Svc.openDatabase(db_Id, false); // false = !read_only
	}
	catch (std::exception &ex) {
		std::cout << "Failure opening DBT database - " << ex.what() << std::endl;
		daq::rc::BadCoolDb i(ERS_HERE, ex.what());
		throw i;
	}
}

// throws daq::rc::BadCoolDb
void openDBD(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

	try {
		dbd = db_Svc.openDatabase(db_Id, false); // false = !read_only
	}
	catch (std::exception &ex) {
		std::cout << "Failure opening DCS database - " << ex.what() << std::endl;
		daq::rc::BadCoolDb i(ERS_HERE, ex.what());
		throw i;
	}
}

// throws daq::rc::BadCoolDb
void openDBM(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

	try {
		dbm = db_Svc.openDatabase(db_Id, false); // false = !read_only
	}
	catch (std::exception &ex) {
		std::cout << "Failure opening DBM database - " << ex.what() << std::endl;
		daq::rc::BadCoolDb i(ERS_HERE, ex.what());
		throw i;
	}
}

int valInBeam(long long startBlock, long long endBlock, long long startBeam, long long endBeam) {
	int result = 0;
	if ((startBlock <= startBeam) && (endBlock <= startBeam)) {
		// block before beam
		return result;
	}
	if ((startBlock <= startBeam) && (endBlock > startBeam) && (endBlock <= endBeam)) {
		result = endBlock - startBeam;
		return result;
	}
	if ((startBlock <= startBeam) && (endBlock >= endBeam)) {
		result = endBeam - startBeam;
		return result;
	}
	if ((startBlock >= startBeam) && (endBlock <= endBeam)) {
		result = endBlock - startBlock;
		return result;
	}
	if ((startBlock > startBeam) && (startBlock < endBeam) && (endBlock >= endBeam)) {
		result = endBeam - startBlock;
		return result;
	}
	if ((startBlock >= endBeam) && (endBlock > endBeam)) {
		// block after beam
		return result;
	}
	return result;
}

void print_vectors(std::vector<long long> start_vector, std::vector<long long> end_vector, std::string file_name) {
	std::ostream * m_lout = new std::ofstream(file_name.c_str(), std::ios::out | std::ios::trunc);
	if (!(*m_lout)) {
		std::cout << "Can't create the " << file_name << " log file " << std::endl;
		return;
	}
	std::ostringstream strm;
	if (start_vector.size() > 0) {
		for (unsigned int i = 0; i < start_vector.size(); i++) {
			long long start_beam_sec = start_vector[i] / 1000;
			OWLTime start_owl = OWLTime(start_beam_sec);
			std::string start_string = std::string(start_owl.c_str());
			if (start_string.size() == 15) {
				strm << "  " << start_string;
			}
			else if (start_string.size() == 16) {
				strm << " " << start_string;
			}
			else {
				strm << start_string;
			}
			long long end_beam_sec = end_vector[i] / 1000;
			OWLTime end_owl = OWLTime(end_beam_sec);
			std::string end_string = std::string(end_owl.c_str());
			strm << "  " << end_string << std::endl;
		}
		(*m_lout) << strm.str();
	}
	(*m_lout) << "endbeam" << std::endl;
	delete m_lout;
}

void set_main_vectors(std::vector<long long> start_vector, std::vector<long long> end_vector) {
	start_run_vector.clear();
	end_run_vector.clear();
	long long bLimit = ((long long)runLimit) * ((long long)1000);
	// exclude intervals before 30/03/2010 08:00:00
	long long timeLimit = ((long long)1269928800) * ((long long)1000);
	if (start_vector.size() > 0) {
		for (unsigned int i = 0; i < start_vector.size(); i++) {
			if ((end_vector[i] - start_vector[i]) < bLimit) {
				continue;
			}
			if (end_vector[i] < timeLimit) {
				continue;
			}
			start_run_vector.push_back(start_vector[i]);
			end_run_vector.push_back(end_vector[i]);
		}
	}
}

void set_run_vectors(long long startTime, long long endTime) {
	run_start_vector.clear();
	run_end_vector.clear();
	if (start_run_vector.size() > 0) {
		for (unsigned int i = 0; i < start_run_vector.size(); i++) {
			if ((startTime < start_run_vector[i]) &&
				(end_run_vector[i] < endTime)) {
					run_start_vector.push_back(start_run_vector[i]);
					run_end_vector.push_back(end_run_vector[i]);
			}
			else if ((start_run_vector[i] <= startTime) &&
				(startTime < end_run_vector[i]) &&
				(end_run_vector[i] <= endTime)) {
					run_start_vector.push_back(startTime);
					run_end_vector.push_back(end_run_vector[i]);
			}
			else if ((start_run_vector[i] < startTime) &&
				(endTime < end_run_vector[i])) {
					run_start_vector.push_back(startTime);
					run_end_vector.push_back(endTime);
			}
			else if ((startTime <= start_run_vector[i]) &&
				(start_run_vector[i] < endTime) &&
				(endTime <= end_run_vector[i])) {
					run_start_vector.push_back(start_run_vector[i]);
					run_end_vector.push_back(endTime);
			}
		}
	}
}

int get_total_run_time() {
	// get total time from run_start_vector and run_end_vector
	int total_run_time = 0;
	int valrun = 0;
	if (run_start_vector.size() > 0) {
		for (unsigned int i = 0; i < run_start_vector.size(); i++) {
			valrun = run_end_vector[i] / 1000 - run_start_vector[i] / 1000;
			total_run_time += valrun;
		}
	}
	return total_run_time;
}

int get_ctpid(cool::ValidityKey sKey, cool::ValidityKey uKey, std::string itemName, cool::IFolderPtr & lvl1menu_folder) {
	int itemNo = -1;
	cool::IObjectIteratorPtr objects = lvl1menu_folder->browseObjects(sKey, uKey, cool::ChannelSelection::all());
	while (objects->goToNext()) {
		const cool::IObject& obj = objects->currentRef();
		std::string item_name = obj.payloadValue("ItemName");
		//itemName = item_name; //to not care about the trigger used
		if (item_name.compare(itemName) == 0) {
			itemNo = obj.channelId();
			break;
		}
	}
	return itemNo;
}

float busy_run2(cool::ValidityKey key, std::string identifier, cool::IFolderPtr & the_busy_folder ) {

  // Find the busy values for a specific detector in this lumiblock:

  cool::IObjectPtr objects = the_busy_folder->findObject(key, 0);
  // const cool::IObject& obj = objects->currentRef();
  std::string this_busy_string= objects->payloadValue(identifier.c_str());
  float this_busy = atof(this_busy_string.c_str());
  // std::cout << "for " << identifier << " busy = " << this_busy << std::endl;
  return this_busy;
}

double veto_fraction(cool::ValidityKey key, int ctpid, cool::IFolderPtr & l1counters_folder) {
	// for the moment overflow flags are not used
	cool::IObjectPtr obj_l1counters = l1counters_folder->findObject(key, ctpid);
	std::string l1Accept = obj_l1counters->payloadValue("L1Accept");
	std::string afterPrescale = obj_l1counters->payloadValue("AfterPrescale");
	long long l1_accept = atoll(l1Accept.c_str());
	long long after_prescale = atoll(afterPrescale.c_str());
        std::cout << " AP = " << after_prescale << " L1A = " << l1_accept << std::endl;
	double fraction = 0;
	if (after_prescale > 0) {
		fraction = (double)(after_prescale - l1_accept) / (double)after_prescale;
	}

	return fraction;
}

void openDatabasesAndFolders() {
	std::string coolDb;
	std::string coolDbTrig;
	std::string coolDbDCS;

	if (atlasRun == 2)
		coolDb = "oracle://ATLR;schema=ATLAS_COOLONL_TDAQ;dbname=CONDBR2";
	else
		coolDb = "oracle://ATLR;schema=ATLAS_COOLONL_TDAQ;dbname=COMP200";

	cool::DatabaseId dbId = coolDb;
	// Open or Create DB
	cool::IDatabaseSvc& dbSvc = cool::DatabaseSvcFactory::databaseService();
	ERS_DEBUG(3, "Retrieved DB service.");

	try {
		openDB(dbId, dbSvc);
	}
	catch (daq::rc::BadCoolDb& ex) {
		ers::fatal(ex);
		exit(0);
	}

	if (atlasRun == 2)
		coolDbTrig = "oracle://ATLR;schema=ATLAS_COOLONL_TRIGGER;dbname=CONDBR2";
	else
		coolDbTrig = "oracle://ATLR;schema=ATLAS_COOLONL_TRIGGER;dbname=COMP200";

	cool::DatabaseId dbIdTrig = coolDbTrig;
	cool::IDatabaseSvc& dbSvcTrig = cool::DatabaseSvcFactory::databaseService();
	try {
		openDBT(dbIdTrig, dbSvcTrig);
	}
	catch (daq::rc::BadCoolDb& ex) {
		ers::fatal(ex);
		exit(0);
	}

	if (atlasRun == 2)
		coolDbDCS = "oracle://ATLR;schema=ATLAS_COOLOFL_DCS;dbname=CONDBR2";
	else
		coolDbDCS = "oracle://ATLR;schema=ATLAS_COOLOFL_DCS;dbname=COMP200";

	cool::DatabaseId dbIdDCS = coolDbDCS;
	cool::IDatabaseSvc& dbSvcDCS = cool::DatabaseSvcFactory::databaseService();
	try {
		openDBD(dbIdDCS, dbSvcDCS);
	}
	catch (daq::rc::BadCoolDb& ex) {
		ers::fatal(ex);
		exit(0);
	}

	std::string coolDbMon = "oracle://ATLR;schema=ATLAS_COOLONL_TRIGGER;dbname=MONP200";
	cool::DatabaseId dbIdMon = coolDbMon;
	cool::IDatabaseSvc& dbSvcMon = cool::DatabaseSvcFactory::databaseService();
	try {
		openDBM(dbIdMon, dbSvcMon);
	}
	catch (daq::rc::BadCoolDb& ex) {
		ers::fatal(ex);
		exit(0);
	}

	// check folders
	try {
		if (!dbt->existsFolder(CTP_Name)) {
			std::cout << "DB Folder CTP does not exist???" << std::endl;
		}
	}
	catch (std::exception &e) {
		std::cout << "DB Folder CTP does not exist. " << e.what() << std::endl;
	}
	try {
		if (!dbd->existsFolder(TotalInt_Name)) {
			std::cout << "DB Folder TotalInt does not exist???" << std::endl;
		}
	}
	catch (std::exception &e) {
		std::cout << "DB Folder TotalInt does not exist. " << e.what() << std::endl;
	}
	if (!dbt->existsFolder(LB_Name)) {
		std::cout << "DB Folder LB does not exist???" << std::endl;
	}
	if (!dbt->existsFolder(LBT_Name)) {
		std::cout << "DB Folder LBT does not exist???" << std::endl;
	}
	if (!dbt->existsFolder(L1Counters_Name)) {
		std::cout << "DB Folder L1Counters does not exist???" << std::endl;
	}
	if (!dbt->existsFolder(L1Menu_Name)) {
		std::cout << "DB Folder Menu does not exist???" << std::endl;
	}
	if (!db->existsFolder(DataTakingMode_Name)) {
		std::cout << "DB Folder DataTakingMode does not exist???" << std::endl;
	}
	//open folders

	try {
		CTP_Folder = dbt->getFolder(CTP_Name);
	}
	catch (std::exception &e) {
		std::cout << "Cannot open CTP folder " << e.what() << std::endl;
		dbt->closeDatabase();
		std::cout << "DBT closed." << std::endl;

		exit(0);
	}

	LB_Folder = dbt->getFolder(LB_Name);
	LBT_Folder = dbt->getFolder(LBT_Name);
	L1Counters_Folder = dbt->getFolder(L1Counters_Name);
	L1Menu_Folder = dbt->getFolder(L1Menu_Name);
	DataTakingMode_Folder = db->getFolder(DataTakingMode_Name);
	busyConf_Folder = dbm->getFolder(busyConf_Name);
	busyRates_Folder = dbm->getFolder(busyRates_Name);

	try {
		TotalInt_Folder = dbd->getFolder(TotalInt_Name);
	}
	catch (std::exception &e) {
		std::cout << "Cannot open TotalInt folder " << e.what() << std::endl;
		dbd->closeDatabase();
		std::cout << "DBD closed." << std::endl;

		exit(0);
	}
}

void setRunVectors(long long starttime, long long endtime) {
	runs_start_vector.clear();
	runs_end_vector.clear();

	// find list of runs and lumi blocks
	try {
		long long start_time_nsec = ((long long)(starttime)) * 1000;
		long long end_time_nsec = ((long long)(endtime)) * 1000;
		bool firstRun = true;
		bool newRun = false;
		int current_run_number = 0;
		long long start_of_first_LB_ms = 0;
		long long end_of_last_LB_ms = 0;
		cool::IObjectIteratorPtr rl_objects = LBT_Folder->browseObjects(cool::ValidityKey(start_time_nsec), cool::ValidityKey(end_time_nsec), 0);
		while (rl_objects->goToNext()) {
			const cool::IObject& obj = rl_objects->currentRef();
			std::string runNumber = obj.payloadValue("Run");
			std::string lumiBlock = obj.payloadValue("LumiBlock");
			int run_nb = atoi(runNumber.c_str());
			if (run_nb != current_run_number) {
				newRun = true;
				current_run_number = run_nb;
			}
			cool::UInt63 rn = run_nb;
			cool::UInt63 lb = atoi(lumiBlock.c_str());
			cool::ValidityKey key1 = (rn << 32) + lb;
			cool::IObjectPtr obj_lb = LB_Folder->findObject(key1, 0);
			std::string st = obj_lb->payloadValue("StartTime");
			std::string et = obj_lb->payloadValue("EndTime");
			long long sT = atoll(st.c_str());
			long long eT = atoll(et.c_str());
			if (newRun) {
				// save data for previous run
				if (firstRun) {
					firstRun = false;
				}
				else {
					runs_start_vector.push_back(start_of_first_LB_ms);
					runs_end_vector.push_back(end_of_last_LB_ms);
				}
				start_of_first_LB_ms = sT / 1000000;
				newRun = false;
			}
			end_of_last_LB_ms = eT / 1000000;
		}
		// save data for last run
		runs_start_vector.push_back(start_of_first_LB_ms);
		runs_end_vector.push_back(end_of_last_LB_ms);
	}
	catch (std::exception &e) {
		std::cout << OWLTime() << " Warning: Read failed in setRunVectors: " << e.what() << std::endl;
	}
}

double busy(long long startTime, long long endTime, std::string identifier, cool::IFolderPtr & the_busy_folder ) {

	// Find the busy values for a specific detector in this lumiblock:

	// Note: this also includes records that are partialy outside this lb

	cool::IObjectIteratorPtr objects = the_busy_folder->browseObjects( cool::ValidityKey(startTime),cool::ValidityKey(endTime), 0 );
	double totaltime=0.;
	double totalbusy=0.;

	//std::cout << "objects : " << objects << std::endl;

	while ( objects->goToNext() ) {
		std::cout << "IN THE BUSY LOOP" << std::endl;
		const cool::IObject& obj = objects->currentRef();
		std::string this_busy_string= obj.payloadValue(identifier.c_str());
		//std::cout << "busyString: " << this_busy_string << std::endl;
		double this_busy = atof(this_busy_string.c_str());
		//std::cout << "this_busy: " << this_busy << std::endl;

		double this_starttime = obj.since();
		double this_stoptime = obj.until();
		//std::cout << "stoptime: " << this_stoptime << std::endl;
		totalbusy += this_busy*(this_stoptime-this_starttime);
		totaltime += (this_stoptime-this_starttime);
	}
	if (totaltime>0) {
		totalbusy = totalbusy/totaltime;//*0.01
	} else {
		totalbusy = 0; // if no info, let's just say this detector was not busy
	}
	return totalbusy;
}

void prepareCtpMap(int runNumber) {
	// Populate ctp map for the start run (no change for the next run ...)
	ctpMap.clear();
	cool::UInt63 a = runNumber;
	cool::UInt63 z = runNumber + 1;
	cool::ValidityKey sKey = (a <<32) + 1;
	cool::ValidityKey uKey = z <<32;
	cool::IObjectIteratorPtr objs = busyConf_Folder->browseObjects(sKey, uKey, cool::ChannelSelection::all());
	while ( objs->goToNext() ) {
		const cool::IObject& obj = objs->currentRef();
		std::string identifier = obj.payloadValue("identifier");
		//std::string description = obj.payloadValue("description");
		std::string busyEnabled = obj.payloadValue("busyEnabled");
		if (busyEnabled.compare("TRUE") == 0) {
			ctpMap.push_back(identifier);
		}
	}
	std::cout << "For run number = " << runNumber << ":" << std::endl;
}

double getEfficiency(long long starttime, long long endtime) {
	totalRunTime = 0;
	totalGoodRunTime = 0;
	totalBusyTime = 0;
	realRunTime = 0;

	int totalGoodRunTime_msec = 0;

	// set beam vectors, having already start_run_vector and end_run_vector
	set_run_vectors(starttime / 1000, endtime / 1000);

	totalRunTime = get_total_run_time();

	if (totalRunTime > 0) {
		// find list of runs and lumi blocks
		int no_data_time_in_beam_msec = 0;
		try {
			long long start_time_nsec = ((long long)(starttime - 120000000)) * 1000;
			long long end_time_nsec = ((long long)(endtime)) * 1000;
			cool::IObjectIteratorPtr rl_objects = LBT_Folder->browseObjects(cool::ValidityKey(start_time_nsec), cool::ValidityKey(end_time_nsec), 0);
			int current_run_number = 0;
			int ctpId = 227; //default value, to be checked
			while (rl_objects->goToNext()) {
				const cool::IObject& obj = rl_objects->currentRef();
				//try {
				std::string runNumber = obj.payloadValue("Run");
				std::string lumiBlock = obj.payloadValue("LumiBlock");
				int run_nb = atoi(runNumber.c_str());
				if (run_nb != current_run_number) {
				    // prepareCtpMap(run_nb);
				    current_run_number = run_nb;
				    cool::UInt63 a = run_nb;
				    cool::UInt63 z = run_nb + 1;
				    cool::ValidityKey sKey = a << 32;
				    cool::ValidityKey uKey = (z << 32) - 1;
				    ctpId = get_ctpid(sKey, uKey, triggerItem, L1Menu_Folder);
				    if (ctpId == -1) {
					std::cout << "For Run " << runNumber << " Trigger Item not found. Default value (227) will be used." << std::endl;
					ctpId = 227;
				    }
                                }   
				cool::UInt63 rn = run_nb;
				cool::UInt63 lb = atoi(lumiBlock.c_str());
                                std::cout << "LB = " << lb << std::endl;
				cool::ValidityKey key1 = (rn << 32) + lb;
				cool::IObjectPtr obj_lb = LB_Folder->findObject(key1, 0);
				std::string st = obj_lb->payloadValue("StartTime");
				std::string et = obj_lb->payloadValue("EndTime");
				long long sT = atoll(st.c_str());
				//std::cout << sT << std::endl;
				long long eT = atoll(et.c_str());
				double vetofraction = 0.0;
				try {
                                  if (useTotalBusy == 0) {
				    vetofraction = veto_fraction(key1, ctpId, L1Counters_Folder);
                                  } else {
                                    vetofraction = (double) busy_run2(key1, "ctpcore_moni0_rate", busyRates_Folder );
                                  }
                                  std::cout << "for lb = " << lb << "   vetofraction = " << vetofraction << std::endl;
				} catch(std::exception &ex) {
                                  std::cout << "Failure in veto fraction evaluation - " << ex.what() << std::endl;
				  // std::cout << "veto fraction error!!! with key " << key1 << " and ctpId " << ctpId << std::endl;
				  vetofraction = 1.0;
				}
				if (run_start_vector.size() > 0) {
					for (unsigned int i = 0; i < run_start_vector.size(); i++) {
						int data_lumi_block_in_beam_msec = valInBeam(sT / 1000000, eT / 1000000, run_start_vector[i], run_end_vector[i]);
						if (data_lumi_block_in_beam_msec > 0) {
							totalGoodRunTime_msec += data_lumi_block_in_beam_msec;
							int no_data_lumi_block_in_beam_msec = (int)(((double)data_lumi_block_in_beam_msec)*vetofraction);
							no_data_time_in_beam_msec += no_data_lumi_block_in_beam_msec;
						}
					}
				}
				totalBusyTime = no_data_time_in_beam_msec / 1000;
				// totalBusyTime += totalRunTime*det_busy;
				totalGoodRunTime = totalGoodRunTime_msec / 1000;
				realRunTime = totalGoodRunTime - no_data_time_in_beam_msec / 1000;
			}
		}
		catch (std::exception &e) {
			std::cout << OWLTime() << " Warning: Read failed: " << e.what() << std::endl;
		}

	}
	// evaluate efficiency in beam
	double beamEff = 0.0;
	if (totalRunTime > 0) {
		beamEff = (double)realRunTime / (double)totalRunTime;//totalGoodRunTime;
	}

	return beamEff;
}

//----------------------- MAIN ----------------------------//
int main(int argc, char ** argv)
{

	// parse commandline parameters

	std::string startTime, endTime, interval;
	bool timeInterval = false;
	int duration = 60;
	int period = 60;
	int evaluationMode = 0;
	boost::program_options::options_description desc("This program retrieves info from COOL conditions database, evaluates the data taking efficiency for selected time intervals and stores the results in /tmp/eff.log file.");
	desc.add_options()
		("help,h", "help message")
		("startTime,s", boost::program_options::value<std::string>(), "Start time (\"dd/mm/yy hh:mm:ss\")")
		("endTime,e", boost::program_options::value<std::string>(), "End time (\"dd/mm/yy hh:mm:ss\")")
		("interval,i", boost::program_options::value<std::string>(), "Time interval (day, week, month) - to be used only if no start time or end time is specified")
		("duration,d", boost::program_options::value<int>(&duration)->default_value(duration), "Duration for evaluation in minutes (default 60)")
		("period,p", boost::program_options::value<int>(&period)->default_value(period), "Period for evaluation in minutes (default 60)")
		("runLimit,b", boost::program_options::value<int>(&runLimit)->default_value(runLimit), "Run session limit option in seconds (default 5 s)")
		("display,x", boost::program_options::value<int>(&display)->default_value(display), "Display options: 0 - no graphical display, 1 - open graphical display (default 0)")
		("triggerItem,c", boost::program_options::value<std::string>(&triggerItem)->default_value(triggerItem), "Trigger item (default L1_EM30))")
		("evaluationMode,a", boost::program_options::value<int>(&evaluationMode)->default_value(evaluationMode), "Evaluation mode option: 0 - evaluate efficiency periodically, 1 - evaluate efficiency for each run interval, 2 - evaluate efficiency and running time for each run interval and prepare data for root graphs, 3 - evaluate efficiency periodically for ready and not ready and prepare data for root graphs, 4 - evaluate efficiency and running time for each run interval and update old data for root graphs, 5 - evaluate efficiency periodically for ready and not ready and update old data for root graphs (default 0)")
		("atlasRun,r", boost::program_options::value<int>(&atlasRun)->default_value(atlasRun), "ATLAS Run: 1 - Run 1 [before 1.1.2015], 2 - Run 2 [after 1.1.2015]")
                ("useTotalBusy,t", boost::program_options::value<int>(&useTotalBusy)->default_value(useTotalBusy), "Use Total Busy: 0 - use LVL1 counters; 1 - use total busy (default 1)")
		;

	boost::program_options::variables_map vm;
	boost::program_options::store(boost::program_options::parse_command_line(argc, argv, desc), vm);
	boost::program_options::notify(vm);

	//	std::cout << "atlasRun: " << atlasRun << std::endl;

	if (vm.count("help")) {
		std::cout << desc << std::endl;
		return EXIT_SUCCESS;
	}
	if ((useTotalBusy < 0) || (useTotalBusy > 1)) {
          std::cout << "Command line error: useTotalBusy should be 0 or 1" << std::endl;
          return EXIT_FAILURE;
        }   
	if ((vm.count("interval")) && ((vm.count("startTime")) || (vm.count("endTime")))) {
		std::cout << desc << std::endl;
		std::cout << "Command line error: time interval used with an explicit start time or end time" << std::endl;
		return EXIT_FAILURE;
	}
	else if ((vm.count("startTime")) && (vm.count("endTime"))) {
		startTime = vm["startTime"].as<std::string>();
		endTime = vm["endTime"].as<std::string>();
	}
	else if (vm.count("interval")) {
		interval = vm["interval"].as<std::string>();
		timeInterval = true;
	}
	else {
		std::cout << desc << std::endl;
		daq::rc::CmdLineError issue(ERS_HERE, "");
		ers::fatal(issue);
		return EXIT_FAILURE;
	}

	std::string runs("/tmp/eff.log");
	std::ostream * m_lout = new std::ofstream(runs.c_str(), std::ios::out | std::ios::trunc);
	if (!(*m_lout)) {
		std::cout << "Can't create the efficiency log file " << std::endl;
	}

	//openDatabasesAndFolders();

	long long interval_mksec = ((long long)60) * ((long long)1000000);
	interval_mksec = period * interval_mksec;
	long long duration_mksec = ((long long)60) * ((long long)1000000);
	duration_mksec = duration * duration_mksec;
	OWLTime now = OWLTime();
	long long nowtime = now.total_mksec_utc();
	long long starttime = 0;
	long long endtime = 0;
	if (!timeInterval) {
		OWLTime start_time(startTime.c_str());
		starttime = start_time.total_mksec_utc();
		OWLTime end_time(endTime.c_str());
		endtime = end_time.total_mksec_utc();
		if (starttime > nowtime) {
			std::cout << " Start time greater than current time" << std::endl;
			return EXIT_FAILURE;
		}
		if (endtime > nowtime) {
			std::cout << " End time greater than current time" << std::endl;
			return EXIT_FAILURE;
		}
		if ((starttime < (long long)1420070400000000 && atlasRun == 2) || (starttime >= (long long)1420070400000000 && atlasRun == 1)) {
			std::cout << desc << std::endl;
			std::cout << "Command line error: run 2 should be used only after 1.1.2015 while run 1 is for periods before 1.1.2015" << std::endl;
			return EXIT_FAILURE;
		}
	}
	else {
		long long total_duration_mksec = 0;
		if (interval.compare("day") == 0) {
			total_duration_mksec = ((long long)86400) * ((long long)1000000);
		}
		else if (interval.compare("week") == 0) {
			total_duration_mksec = ((long long)604800) * ((long long)1000000);
		}
		else if (interval.compare("month") == 0) {
			total_duration_mksec = ((long long)2592000) * ((long long)1000000);
		}
		else {
			std::cout << desc << std::endl;
			std::cout << "Wrong time interval option!" << std::endl;
			return EXIT_FAILURE;
		}
		if (atlasRun != 2) {
			std::cout << desc << std::endl;
			std::cout << "Command line error: run 2 should be used only after 1.1.2015 while run 1 is for periods before 1.1.2015" << std::endl;
			return EXIT_FAILURE;
		}
		endtime = nowtime;
		// take into account total duration
		starttime = endtime - total_duration_mksec;
		endTime = std::string(now.c_str());
		OWLTime start = OWLTime(starttime / 1000000);
		startTime = std::string(start.c_str());
	}
	// now take into account the duration and the period
	starttime = starttime + interval_mksec - duration_mksec;

	openDatabasesAndFolders();

	// for update mode
	bool updateLastFill = false;
	long long starttimeG = starttime + duration_mksec - interval_mksec;
	long long endtimeG = endtime;
	totalTimeEvaluated = (endtime-starttime)/1000000;

	if (evaluationMode == 4) {
		// read old run eff data
		std::filebuf fb;
		fb.open("/tmp/runeff_old.data", std::ios::in);
		if (fb.is_open()) {
			std::istream is(&fb);
			std::string line;
			while (!(line.compare("enddata") == 0)) {
				getline(is, line);
				if (!(line.compare("enddata") == 0)) {
					runeff_data.push_back(line);
				}
			}
		}
		else {
			std::cout << "Warning: cannot open the old runeff data file" << std::endl;
		}
		fb.close();
		// read old run time data
		fb.open("/tmp/runtime_old.data", std::ios::in);
		if (fb.is_open()) {
			std::istream is(&fb);
			std::string line;
			while (!(line.compare("enddata") == 0)) {
				getline(is, line);
				if (!(line.compare("enddata") == 0)) {
					runtime_data.push_back(line);
				}
			}
		}
		else {
			std::cout << "Warning: cannot open the old runtime data file" << std::endl;
		}

		fb.close();
		int n = runeff_data.size();
		std::string eff_data = runeff_data[n-1];
		n = runtime_data.size();
		std::string time_data = runtime_data[n-1];
		std::string end_time_old = time_data.substr(0, 10);
		long end_time_old_sec = atol(end_time_old.c_str());
		// check if the new start evaluation time is during a beam interval
		if (end_time_old_sec == starttime / 1000000) {
			updateLastFill = true;
			std::string start_time_old = eff_data.substr(0, 10);
			long start_time_old_sec = atol(start_time_old.c_str());
			// consider starttime 10 sec before start of last beam interval
			starttime = ((long long)(start_time_old_sec - 10)) * 1000000;
		}
		//    std::cout << "updateLastFill = " << updateLastFill << std::endl;
	}
	if (evaluationMode == 5) {
		// read old run eff data
		std::filebuf fb;
		fb.open("/tmp/runeff_old.data", std::ios::in);
		if (fb.is_open()) {
			std::istream is(&fb);
			std::string line;
			while (!(line.compare("enddata") == 0)) {
				getline(is, line);
				if (!(line.compare("enddata") == 0)) {
					runeff_data.push_back(line);
				}
			}
		}
		else {
			std::cout << "Warning: cannot open the old runeff data file" << std::endl;
		}
		fb.close();
	}

	// read run data
	setRunVectors(starttimeG, endtimeG);
	set_main_vectors(runs_start_vector, runs_end_vector);
	if (writeBeamFile == 1) {
		print_vectors(start_run_vector, end_run_vector, "/tmp/runs.log");
	}

	if (start_run_vector.size() > 0) {
		for (unsigned int i = 0; i < start_run_vector.size(); i++) {
			(*m_lout) << start_run_vector[i] << " " << end_run_vector[i] << std::endl;
		}
	}
	(*m_lout) << "endbeam" << std::endl;

	// read run data

	if ((evaluationMode == 2) || (evaluationMode == 4)) {
		// prepare data files
		std::ostream * eff_out = new std::ofstream("/tmp/runeff.data", std::ios::out | std::ios::trunc);
		if (!(*eff_out)) {
			std::cout << "Can't create the efficiency data file " << std::endl;
		}
		std::ostream * runtime_out = new std::ofstream("/tmp/runtime.data", std::ios::out | std::ios::trunc);
		if (!(*runtime_out)) {
			std::cout << "Can't create the runtime data file " << std::endl;
		}
		int globalBeamTime = 0;
		int globalRunningTime = 0;
		int globalRunningReadyTime = 0;
		if (evaluationMode == 4) {
			if (updateLastFill == true) {
				// remove data for last beam interval (interval to be updated)
				runeff_data.pop_back();
				runeff_data.pop_back();
				runeff_data.pop_back();
				runeff_data.pop_back();
				runeff_data.pop_back();
				runtime_data.pop_back();
				runtime_data.pop_back();
				runtime_data.pop_back();
			}
			//else {
			// remove last entries, not useful
			//runeff_data.pop_back();
			//runtime_data.pop_back();
			//}
			// insert old data
			unsigned int n = runeff_data.size();
			if (n > 0) {
				for (unsigned int i = 0; i < n; i++) {
					(*eff_out) << runeff_data[i] << std::endl;
				}
			}
			n = runtime_data.size();
			if (n > 0) {
				for (unsigned int i = 0; i < n; i++) {
					(*runtime_out) << runtime_data[i] << std::endl;
				}
			}
			// update time info
			std::string time_data = runtime_data[n - 1];
			size_t beamTimeEndIndex = time_data.find_first_of(" ", 14);
			std::string beamTimeValue = time_data.substr(14, beamTimeEndIndex - 14);
			globalBeamTime = atoi(beamTimeValue.c_str());
			size_t runTimeEndIndex = time_data.find_first_of(" ", beamTimeEndIndex + 1);
			std::string runTimeValue = time_data.substr(beamTimeEndIndex + 1, runTimeEndIndex - beamTimeEndIndex - 1);
			globalRunningTime = atoi(runTimeValue.c_str());
			size_t runTimeReadyEndIndex = time_data.find_first_of(" ", runTimeEndIndex + 1);
			std::string runTimeReadyValue = time_data.substr(runTimeEndIndex + 1, runTimeReadyEndIndex - runTimeEndIndex - 1);
			globalRunningReadyTime = atoi(runTimeReadyValue.c_str());
		}

		int globalBeamRunningTime = 0;
		int globalBeamRunningReadyTime = 0;
		//    int globalBeamRunningReadyNoWarmStopTime = 0;
		int globalNoDataInBeamTime = 0;
		int globalNoDataInBeamReadyTime = 0;
		//    int globalNoDataInBeamReadyNoWarmStopTime = 0;
		double runEff = 0;
		double runEffReady = 0;
		double noBeam = 0.0;
		double noData = 0.0;
		double maxBeam = 1.0;
		if (start_run_vector.size() > 0) {
			for (unsigned int i = 0; i < start_run_vector.size(); i++) {
				int initialBeamTime = globalBeamTime;
				int initialRunningTime = globalRunningTime;
				int initialRunningReadyTime = globalRunningReadyTime;
				starttime = start_run_vector[i] * 1000;
				endtime = end_run_vector[i] * 1000;
				runEff = getEfficiency(starttime, endtime);
				globalBeamTime = globalBeamTime + totalRunTime;
				globalBeamRunningTime = globalBeamRunningTime + totalGoodRunTime;
				globalNoDataInBeamTime = globalNoDataInBeamTime + totalBusyTime;
				globalRunningTime = globalRunningTime + realRunTime;
				OWLTime current = OWLTime(end_run_vector[i] / 1000);
				std::cout << current.c_str() << "  " << runEff << std::endl;
				(*m_lout) << start_run_vector[i] << " " << noBeam << std::endl;
				(*m_lout) << start_run_vector[i] + 1000 << " " << runEff << std::endl;
				(*m_lout) << end_run_vector[i] - 1000 << " " << runEff << std::endl;
				(*m_lout) << end_run_vector[i] << " " << noBeam << std::endl;
				(*eff_out) << start_run_vector[i] << " " << noBeam << " " << noData << " " << noData << std::endl;
				(*eff_out) << start_run_vector[i] + 1000 << " " << maxBeam << " " << runEff << " " << runEffReady << std::endl;
				(*eff_out) << end_run_vector[i] - 1000 << " " << maxBeam << " " << runEff << " " << runEffReady << std::endl;
				(*eff_out) << end_run_vector[i] << " " << noBeam << " " << noData << " " << noData << std::endl;
				(*runtime_out) << start_run_vector[i] << " " << initialBeamTime << " " << initialRunningTime << " " << initialRunningReadyTime << std::endl;
				(*runtime_out) << end_run_vector[i] << " " << globalBeamTime << " " << globalRunningTime << " " << globalRunningReadyTime << std::endl;
			}
		}
		(*eff_out) << endtimeG / 1000 << " " << noBeam << " " << noData << " " << noData << std::endl;
		(*runtime_out) << endtimeG / 1000 << " " << globalBeamTime << " " << globalRunningTime << " " << globalRunningReadyTime << std::endl;
		(*eff_out) << "enddata" << std::endl;
		(*runtime_out) << "enddata" << std::endl;
		integralBeamTime = globalBeamTime;
		integralBeamRunningTime = globalBeamRunningTime;
		integralDeadTime = globalNoDataInBeamTime;
		integralRunningTime = globalRunningTime;
		integralRunEff = (double)globalRunningTime / (double)globalBeamTime;
		integralBeamRunningReadyTime = globalBeamRunningReadyTime;
		integralReadyDeadTime = globalNoDataInBeamReadyTime;
		integralRunningReadyTime = globalRunningReadyTime;
		integralRunEffReady = (double)globalRunningReadyTime / (double)globalBeamTime;
		delete eff_out;
		delete runtime_out;
	}
	if (evaluationMode == 1) {
		if (start_run_vector.size() > 0) {
			for (unsigned int i = 0; i < start_run_vector.size(); i++) {
				starttime = start_run_vector[i] * 1000;
				endtime = end_run_vector[i] * 1000;
				double beamEff = getEfficiency(starttime, endtime);
				OWLTime current = OWLTime(end_run_vector[i] / 1000);
				std::cout << current.c_str() << "  " << beamEff << std::endl;
				double noBeam = 0.0;
				(*m_lout) << start_run_vector[i] << " " << noBeam << std::endl;
				(*m_lout) << start_run_vector[i] + 1000 << " " << beamEff << std::endl;
				(*m_lout) << end_run_vector[i] - 1000 << " " << beamEff << std::endl;
				(*m_lout) << end_run_vector[i] << " " << noBeam << std::endl;
			}
		}
	}
	if (evaluationMode == 0) {
                std::cout << "AAAAA" << std::endl;
		long long currenttime = starttimeG + interval_mksec;
		while (currenttime <= endtimeG) {
			OWLTime current = OWLTime(currenttime / 1000000);
			endtime = currenttime;
			starttime = endtime - duration_mksec;
			double beamEff = getEfficiency(starttime, endtime);
			std::cout << current.c_str() << "  " << beamEff << std::endl;
			int ended = 0;
			int started = 0;
			double noBeam = 0.0;
			//cool::IObjectIteratorPtr objects = FillState_Folder->browseObjects( cool::ValidityKey(currenttime*1000),cool::ValidityKey(currenttime*1000+1000000000), cool::ChannelSelection::all());
			//while(objects->goToNext())
			//{
			//	const cool::IObject& obj = objects->currentRef();
			//	fn = obj.payloadValue("FillNumber");
			//}
			if ((start_run_vector.size() > 0) && (beamEff > noBeam)) {
				for (unsigned int i = 0; i < start_run_vector.size(); i++) {
					if ((start_run_vector[i] >= starttime / 1000) && (start_run_vector[i] < endtime / 1000) && (end_run_vector[i] > endtime / 1000)) {
						(*m_lout) << start_run_vector[i] << " " << noBeam << std::endl;
						(*m_lout) << start_run_vector[i] + 1000 << " " << beamEff << std::endl;
						(*m_lout) << endtime / 1000 << " " << beamEff << std::endl;
						started = 1;
					}
					if ((start_run_vector[i] >= starttime / 1000) && (end_run_vector[i] <= endtime / 1000)) {
						(*m_lout) << start_run_vector[i] << " " << noBeam << std::endl;
						(*m_lout) << start_run_vector[i] + 1000 << " " << beamEff << std::endl;
						(*m_lout) << end_run_vector[i] - 1000 << " " << beamEff << std::endl;
						(*m_lout) << end_run_vector[i] << " " << noBeam << std::endl;
						ended = 1;
					}
					if ((start_run_vector[i] < starttime / 1000) && (end_run_vector[i] > starttime / 1000) && (end_run_vector[i] <= endtime / 1000)) {
						(*m_lout) << starttime / 1000 + 1000 << " " << beamEff << std::endl;
						(*m_lout) << end_run_vector[i] - 1000 << " " << beamEff << std::endl;
						(*m_lout) << end_run_vector[i] << " " << noBeam << std::endl;
						ended = 1;
					}
				}
			}
			if ((ended == 0) && (started == 0)) {
				(*m_lout) << starttime / 1000 + 1000 << " " << beamEff << std::endl;
				(*m_lout) << endtime / 1000 << " " << beamEff << std::endl;
			}
			currenttime = currenttime + interval_mksec;
		}
	}
	if ((evaluationMode == 3) || (evaluationMode == 5)) {
		// prepare data files
		std::ostream * eff_out = new std::ofstream("/tmp/runeff.data", std::ios::out | std::ios::trunc);
		if (!(*eff_out)) {
			std::cout << "Can't create the efficiency data file " << std::endl;
		}
		if (evaluationMode == 5) {
			// remove last entriy, not useful
			//runeff_data.pop_back();
			// insert old data
			unsigned int n = runeff_data.size();
			if (n > 0) {
				for (unsigned int i = 0; i < n; i++) {
					(*eff_out) << runeff_data[i] << std::endl;
				}
			}
		}

		int globalBeamTime = 0;
		int globalRunningTime = 0;
		int globalRunningReadyTime = 0;
		int globalBeamRunningTime = 0;
		int globalBeamRunningReadyTime = 0;
		int globalNoDataInBeamTime = 0;
		int globalNoDataInBeamReadyTime = 0;
		double runEff = 0;
		double runEffReady = 0;
		double noBeam = 0.0;
		double noData = 0.0;
		double beamTime = 0;
		long long currenttime = starttimeG + interval_mksec;
		if (evaluationMode == 3) {
			(*eff_out) << starttimeG / 1000 - 1000 << " " << noBeam << " " << noData << " " << noData << std::endl;
		}
		while (currenttime <= endtimeG) {
			OWLTime current = OWLTime(currenttime / 1000000);
			endtime = currenttime;
			starttime = endtime - duration_mksec;
			runEff = getEfficiency(starttime, endtime);
			globalBeamTime += totalRunTime;
			globalBeamRunningTime = globalBeamRunningTime + totalGoodRunTime;
			globalNoDataInBeamTime = globalNoDataInBeamTime + totalBusyTime;
			globalRunningTime = globalRunningTime + realRunTime;
			beamTime = (double)totalRunTime / 3600;
			std::cout << current.c_str() << "  " << runEff << std::endl;
			int ended = 0;
			int started = 0;
			if ((start_run_vector.size() > 0) && (runEff > noBeam)) {
				for (unsigned int i = 0; i < start_run_vector.size(); i++) {
					if ((start_run_vector[i] >= starttime / 1000) && (start_run_vector[i] < endtime / 1000) && (end_run_vector[i] > endtime / 1000)) {
						(*m_lout) << start_run_vector[i] << " " << noBeam << std::endl;
						(*m_lout) << start_run_vector[i] + 1000 << " " << runEff << std::endl;
						(*m_lout) << endtime / 1000 << " " << runEff << std::endl;
						started = 1;
					}
					if ((start_run_vector[i] >= starttime / 1000) && (end_run_vector[i] <= endtime / 1000)) {
						(*m_lout) << start_run_vector[i] << " " << noBeam << std::endl;
						(*m_lout) << start_run_vector[i] + 1000 << " " << runEff << std::endl;
						(*m_lout) << end_run_vector[i] - 1000 << " " << runEff << std::endl;
						(*m_lout) << end_run_vector[i] << " " << noBeam << std::endl;
						ended = 1;
					}
					if ((start_run_vector[i] < starttime / 1000) && (end_run_vector[i] > starttime / 1000) && (end_run_vector[i] <= endtime / 1000)) {
						(*m_lout) << starttime / 1000 + 1000 << " " << runEff << std::endl;
						(*m_lout) << end_run_vector[i] - 1000 << " " << runEff << std::endl;
						(*m_lout) << end_run_vector[i] << " " << noBeam << std::endl;
						ended = 1;
					}
				}
			}
			if ((ended == 0) && (started == 0)) {
				(*m_lout) << starttime / 1000 + 1000 << " " << runEff << std::endl;
				(*m_lout) << endtime / 1000 << " " << runEff << std::endl;
			}
			(*eff_out) << starttime / 1000 << " " << beamTime << " " << runEff << " " << runEffReady << std::endl;
			(*eff_out) << endtime / 1000 << " " << beamTime << " " << runEff << " " << runEffReady << std::endl;
			currenttime = currenttime + interval_mksec;
		}
		(*eff_out) << endtime / 1000 + 1000 << " " << noBeam << " " << noData << " " << noData << std::endl;
		(*eff_out) << "enddata" << std::endl;
		integralBeamTime = globalBeamTime;
		integralBeamRunningTime = globalBeamRunningTime;
		integralDeadTime = globalNoDataInBeamTime;
		integralRunningTime = globalRunningTime;
		integralRunEff = (double)globalRunningTime / (double)globalBeamTime;
		integralBeamRunningReadyTime = globalBeamRunningReadyTime;
		integralReadyDeadTime = globalNoDataInBeamReadyTime;
		integralRunningReadyTime = globalRunningReadyTime;
		integralRunEffReady = (double)globalRunningReadyTime / (double)globalBeamTime;
		delete eff_out;
	}
	(*m_lout) << "end" << std::endl;

	std::cout << std::endl;
	double beamEff = 0;
	if ((evaluationMode == 0) || (evaluationMode == 1)) {
                std::cout << "BBBBBB" << std::endl;
		beamEff = getEfficiency(starttimeG, endtimeG);

		if (runs_start_vector.size() > 0) {
			for (unsigned int i = 0; i < runs_start_vector.size(); i++) {
				(*m_lout) << runs_start_vector[i] << " " << runs_end_vector[i] << std::endl;
			}
		}
	}
	(*m_lout) << "finalend" << std::endl;

	// print summary data
	if ((evaluationMode == 0) || (evaluationMode == 1)) {
		std::cout << "Total evaluated time (hrs)	= " << (double)totalTimeEvaluated/3600. << std::endl;
		std::cout << "Total run time (sec)        = " << totalRunTime << std::endl;
		std::cout << "Busy time (sec)              = " << totalBusyTime << std::endl;
		std::cout << "Real running time (sec) = " << realRunTime << std::endl;
		if (totalRunTime > 0) {
			std::cout << "Data taking efficiency       = " << beamEff << std::endl;
		}
	}
	if ((evaluationMode == 2) || (evaluationMode == 3) || (evaluationMode == 4) || (evaluationMode == 5)) {
		std::cout << "Total run time (sec)                     = " << integralBeamTime << std::endl;
		std::cout << "Busy time (sec)                           = " << integralDeadTime << std::endl;
		std::cout << "Real running time (sec)              = " << integralRunningTime << std::endl;
		double integralBeamTimeHours = (double)integralBeamTime / 3600;
		double integralRunningTimeHours = (double)integralRunningTime / 3600;
		double integralRunningReadyTimeHours = (double)integralRunningReadyTime / 3600;
		// prepare summary file
		std::ostream * summary_out = new std::ofstream("/tmp/summary.data", std::ios::out | std::ios::trunc);
		if (!(*summary_out)) {
			std::cout << "Can't create the summary data file " << std::endl;
		}
		if (integralBeamTimeHours >= 1000) {
			(*summary_out).precision(5);
		}
		else {
			(*summary_out).precision(4);
		}
		(*summary_out) << integralBeamTimeHours << std::endl;
		if (integralRunningTimeHours >= 1000) {
			(*summary_out).precision(5);
		}
		else {
			(*summary_out).precision(4);
		}
		(*summary_out) << integralRunningTimeHours << std::endl;
		if (integralRunningReadyTimeHours >= 1000) {
			(*summary_out).precision(5);
		}
		else {
			(*summary_out).precision(4);
		}
		(*summary_out) << integralRunningReadyTimeHours << std::endl;
		double integralRunEffOut = integralRunEff * 100;
		double integralRunEffReadyOut = integralRunEffReady * 100;
		(*summary_out).precision(3);
		(*summary_out) << integralRunEffOut << std::endl;
		(*summary_out) << integralRunEffReadyOut << std::endl;
		delete summary_out;
	}

	std::cout << std::endl;

	(*m_lout) << "Run efficiency from " << startTime << " to " << endTime << std::endl;

	if ((evaluationMode == 0) || (evaluationMode == 1)) {
		(*m_lout) << "Total running time (sec) = " << totalRunTime << std::endl;
		(*m_lout) << "Busy time during run (sec) = " << totalBusyTime << std::endl;
		if (totalRunTime > 0) {
			(*m_lout) << "(RunningTime - BusyTime)/RunningTime = " << beamEff << std::endl;
		}
		else {
			(*m_lout) << "                           " << std::endl;
		}
	}

	if (((evaluationMode == 2) || (evaluationMode == 3) || (evaluationMode == 4) || (evaluationMode == 5))) {
		(*m_lout) << "Total running time (sec) = " << integralBeamTime << std::endl;
		(*m_lout) << "Busy time during run (sec) = " << integralDeadTime << std::endl;
		if (integralBeamTime > 0) {
			(*m_lout) << "(RunningTime - BusyTime)/RunningTime = " << integralRunEff << std::endl;
		}
		else {
			(*m_lout) << "                           " << std::endl;
		}
	}
	delete m_lout;
	// Close the DB
	db->closeDatabase();
	dbt->closeDatabase();
	dbd->closeDatabase();
	dbm->closeDatabase();
	std::cout << "DB closed." << std::endl;

	// Install signal handler
	signal(SIGINT, signalhandler);
	signal(SIGTERM, signalhandler);

	ERS_DEBUG(3, "Installed signal handlers.");

	if (display == 1) {
		system("eff_display -o 4");
	}
	return EXIT_SUCCESS;
}

