//   get_efficiency.cc
//
//   Application to evaluate data taking efficiency
//   
//   Author: Mihai.Caprini@cern.ch, February 2010
//   Modifications:
//
//   Author  Date            Reason
//   MC      11/03/2010      Add option:  -t (beam type)
//   MC      17/03/2010      Add options: -w (write beam file)
//                                        -x (start graphical interface)
//   MC      23/03/2010      Add option:  -n (no ramps in beam time)
//   MC      12/04/2010      Add options: -c (trigger item)
//                                        -l (busy limit)
//                                        -f (ready for physics)
//   MC      15/04/2010      Check only lumi blocks in the analysis interval
//   MC      20/04/2010      Add options: -o (display mode)
//                                        -v (no beam value)
//   MC      03/05/2010      Add option:  -g (debug mode)
//                           Replace L1_MBTS_1_col by L1_MBTS_2
//   MC      10/05/2010      Use veto_fraction instead of busy_fraction
//   MC      13/05/2010      Add option:  -a (evaluation mode)
//   MC      31/05/2010      For evaluation mode add option 2 
//   MC      09/06/2010      For evaluation mode add option 3
//   MC      05/07/2010      Exclude intervals before 30/03/2010 08:00:00
//   MC      14/07/2010      Use evaluation mode for summary data
//   MC      02/09/2010      For evaluation mode add option 4 (update mode)
//   MC      15/11/2010      Replace L1_MBTS_2 by L1_LUCID_A_C
//   MC      23/11/2010      Add option:  -k (excludeWarmStop)
//   MC      25/01/2011      Use /TRIGGER/LUMI/LBTIME folder in getEfficiency
//                           Remove SOR, EOR, EC, BusyRate and BusyConf folders
//   MC      31/01/2011      Add setRunVectors, remove checkRunInRNDB
//   MC      15/02/2011      Go back to L1_MBTS_2 as default value
//   MC	     22/06/2011      For evaluation mode add option 5 (update mode)
//   MC      17/07/2011      Replace L1_MBTS_2 by L1_EM14
//   MC      08/09/2011      Replace L1_EM14 by L1_EM12
//   MC      14/11/2011      Replace L1_EM12 by L1_EM5
//   MC      03/04/2012      Replace L1_EM5 by L1_EM30
//   MC      13/02/2015      Use new cond db for Run 2 (CONDBR2)
//   MC      13/02/2015      Add option:  -r (atlasRun)
//   MC      20/02/2015      Remove option: -m (runningOffline)
//   MC      02/06/2015      Add option: -t useTotalBusy
//

#include <iostream>
#include <sstream>
#include <stdlib.h> // for getenv
#include <signal.h>
#include <fstream>
#include <vector>

#include <boost/program_options.hpp>

#include <CoralBase/Attribute.h>
#include <CoolKernel/Exception.h>
#include <CoolKernel/ValidityKey.h>
#include <CoolKernel/IDatabaseSvc.h>
#include <CoolKernel/IDatabase.h>
#include <CoolKernel/IFolder.h>
#include <CoolKernel/IRecordSpecification.h>
#include <CoolKernel/Record.h>
#include <CoolApplication/DatabaseSvcFactory.h>
#include <CoolKernel/IObjectIterator.h>
#include <CoolKernel/IObject.h>
#include <CoolKernel/ChannelSelection.h>

#include "ers/ers.h"
#include "owl/time.h"

#include "pmg/pmg_initSync.h"

#include "RunControl/Common/Exceptions.h"

namespace daq {
ERS_DECLARE_ISSUE(	rc,
			BadCoolDb,
			"Retrieving data from COOL failed because " << explanation,
                        ((const char*) explanation)
		)
}

static     cool::IDatabasePtr db;
static     cool::IDatabasePtr dbt;
static     cool::IDatabasePtr dbd;
static     cool::IDatabasePtr dbm;

static     std::string CTP_Name = "/TRIGGER/LVL1/CTPCORELBDATA";
static     std::string LB_Name = "/TRIGGER/LUMI/LBLB";
static     std::string LBT_Name = "/TRIGGER/LUMI/LBTIME";
static     std::string FillState_Name = "/LHC/DCS/FILLSTATE";
static     std::string TotalInt_Name = "/LHC/DCS/BPTX/TOTALINT";
static     std::string L1Counters_Name = "/TRIGGER/LUMI/LVL1COUNTERS";
static     std::string L1Menu_Name = "/TRIGGER/LVL1/Menu";
static     std::string DataTakingMode_Name = "/TDAQ/RunCtrl/DataTakingMode";
static     std::string BusyRate_Name = "/TRIGGER/LVL1/BUSYRATES_RUN2";

static     cool::IFolderPtr CTP_Folder;
static     cool::IFolderPtr FillState_Folder;
static     cool::IFolderPtr TotalInt_Folder;
static     cool::IFolderPtr LB_Folder;
static     cool::IFolderPtr LBT_Folder;
static     cool::IFolderPtr L1Counters_Folder;
static     cool::IFolderPtr L1Menu_Folder;
static     cool::IFolderPtr DataTakingMode_Folder;
static     cool::IFolderPtr BusyRate_Folder;

static     std::map<std::string, std::string> ctpMap;

static     int atlasRun = 2;
static     int beamType = 3;
static     int beamLimit = 5;
static     int writeBeamFile = 0;
static     int display = 0;
static     int noRamp = 0;
static     double busyLimit = 0.03;
static     std::string triggerItem = "L1_EM30";
static     int readyForPhysics = 0;
static     int displayMode = 0;
static     int noBeamValue = 0;
static     int debugMode = 0;
static     int useTotalBusy = 1;

static     int totalBeamTime = 0;
static     int totalInBeamTime = 0;
static     int totalNoDataInBeamTime = 0;
static     int realInBeamTime = 0;

static     int integralBeamTime = 0;
static     int integralBeamRunningTime = 0;
static     int integralBeamRunningReadyTime = 0;
static     int integralDeadTime = 0;
static     int integralReadyDeadTime = 0;
static     int integralRunningTime = 0;
static     int integralRunningReadyTime = 0;
static     int integralBeamNoWarmStopTime = 0;
static     int integralRunningReadyNoWarmStopTime = 0;
static     double integralRunEff = 0;
static     double integralRunEffReady = 0;
static     double integralRunEffReadyNoWarmStop = 0;

static     std::vector<long long> start_beam_vector;
static     std::vector<long long> end_beam_vector;
static     std::vector<long long> start_of_run_vector;
static     std::vector<long long> end_of_run_vector;

static     std::vector<long long> beam_start_vector;
static     std::vector<long long> beam_end_vector;

static     std::vector<long long> beam1_start_vector;
static     std::vector<long long> beam2_start_vector;
static     std::vector<long long> beam1_end_vector;
static     std::vector<long long> beam2_end_vector;
static     std::vector<long long> beams_start_vector;
static     std::vector<long long> beams_end_vector;
static     std::vector<long long> beam1and2_start_vector;
static     std::vector<long long> beam1and2_end_vector;
static     std::vector<long long> beam1or2_start_vector;
static     std::vector<long long> beam1or2_end_vector;
static     std::vector<long long> beam1and2andStable_start_vector;
static     std::vector<long long> beam1and2andStable_end_vector;
static     std::vector<long long> beam1or2noRamp_start_vector;
static     std::vector<long long> beam1or2noRamp_end_vector;
static     std::vector<long long> beam1and2noRamp_start_vector;
static     std::vector<long long> beam1and2noRamp_end_vector;
static     std::vector<long long> ramp_start_vector;
static     std::vector<long long> ramp_end_vector;

static     std::vector<std::string> runeff_data;
static     std::vector<std::string> runtime_data;

extern "C" void signalhandler(int)
{
   signal(SIGINT,  SIG_DFL);
   signal(SIGTERM, SIG_DFL);
   std::cout << "Terminating rc_get_efficiency ..." << std::endl;
}

// throws daq::rc::BadCoolDb
void openDB(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

  try {
    db = db_Svc.openDatabase(db_Id, false); // false = !read_only
  }
  catch (std::exception &ex) {
    std::cout << "Failure opening DB database - " << ex.what() << std::endl;
    daq::rc::BadCoolDb i(ERS_HERE, ex.what());
    throw i;
  }
}

// throws daq::rc::BadCoolDb
void openDBT(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

  try {
    dbt = db_Svc.openDatabase(db_Id, false); // false = !read_only
  }
  catch (std::exception &ex) {
    std::cout << "Failure opening DBT database - " << ex.what() << std::endl;
    daq::rc::BadCoolDb i(ERS_HERE, ex.what());
    throw i;
  }
}

// throws daq::rc::BadCoolDb
void openDBD(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

  try {
    dbd = db_Svc.openDatabase(db_Id, false); // false = !read_only
  } catch (std::exception &ex) {
    std::cout << "Failure opening DCS database - " << ex.what() << std::endl;
    daq::rc::BadCoolDb i(ERS_HERE, ex.what());
    throw i;
  }
}

// throws daq::rc::BadCoolDb
void openDBM(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

  try {
     dbm = db_Svc.openDatabase(db_Id, false); // false = !read_only
  } catch (std::exception &ex) {
     std::cout << "Failure opening DBM database - " << ex.what() << std::endl;
     daq::rc::BadCoolDb i(ERS_HERE, ex.what());
     throw i;
  }
}


int valInBeam(long long startBlock, long long endBlock, long long startBeam, long long endBeam) {
  int result = 0;
  if ((startBlock <= startBeam) && (endBlock <= startBeam)) {
    // block before beam
    return result;
  }
  if ((startBlock <= startBeam) && (endBlock > startBeam) && (endBlock <= endBeam)) {
    result = endBlock - startBeam;
    return result;
  }
  if ((startBlock <= startBeam) && (endBlock >= endBeam)) {
    result = endBeam - startBeam;
    return result;
  }
  if ((startBlock >= startBeam) && (endBlock <= endBeam)) {
    result = endBlock - startBlock;
    return result;
  }
  if ((startBlock > startBeam) && (startBlock < endBeam) && (endBlock >= endBeam)) {
    result = endBeam - startBlock;
    return result;
  }
  if ((startBlock >= endBeam) && (endBlock > endBeam)) {
    // block after beam
    return result;
  }
  return result;
}

void get_beam1_data(long long startTime, long long endTime, cool::IFolderPtr & folder) {
  beam1_start_vector.clear();
  beam1_end_vector.clear();
  cool::IObjectIteratorPtr objects = folder->browseObjects( cool::ValidityKey(startTime),cool::ValidityKey(endTime), cool::ChannelSelection::all());
  std::string old_beam = "FALSE";
  bool beamStarted = false;
  while ( objects->goToNext() ) {
    const cool::IObject& obj = objects->currentRef();
    std::string beam = obj.payloadValue("Beam1_BeamPresent");
    if (((beam.compare("TRUE") == 0) && (old_beam.compare("FALSE") == 0)) ||
        ((beam.compare("TRUE") == 0) && (old_beam.compare("NULL") == 0))) {
      // start of a new beam interval
      long long start_beam = (long long) (obj.since());
      if (start_beam < startTime) {
        start_beam = startTime;
      }
      beam1_start_vector.push_back(start_beam/1000000);
      old_beam = beam;
      beamStarted = true;
    } else if ((beam.compare("FALSE") == 0) && (old_beam.compare("TRUE") == 0)) {
      // end of a beam interval
      long long end_beam  = (long long) (obj.since());
      beam1_end_vector.push_back(end_beam/1000000);
      beamStarted = false;
      old_beam = beam;
    } else {
      old_beam = beam;
    }
  }
  if (beamStarted) {
    beam1_end_vector.push_back(endTime/1000000);
  }
}

void get_beam2_data(long long startTime, long long endTime, cool::IFolderPtr & folder) {
  beam2_start_vector.clear();
  beam2_end_vector.clear();
  cool::IObjectIteratorPtr objects = folder->browseObjects( cool::ValidityKey(startTime),cool::ValidityKey(endTime), cool::ChannelSelection::all());
  std::string old_beam = "FALSE";
  bool beamStarted = false;
  while ( objects->goToNext() ) {
    const cool::IObject& obj = objects->currentRef();
    std::string beam = obj.payloadValue("Beam2_BeamPresent");
    if (((beam.compare("TRUE") == 0) && (old_beam.compare("FALSE") == 0)) ||
        ((beam.compare("TRUE") == 0) && (old_beam.compare("NULL") == 0))) {
      // start of a new beam interval
      long long start_beam = (long long) (obj.since());
      if (start_beam < startTime) {
        start_beam = startTime;
      }
      beam2_start_vector.push_back(start_beam/1000000);
      old_beam = beam;
      beamStarted = true;
    } else if ((beam.compare("FALSE") == 0) && (old_beam.compare("TRUE") == 0)) {
      // end of a beam interval
      long long end_beam  = (long long) (obj.since());
      beam2_end_vector.push_back(end_beam/1000000);
      beamStarted = false;
      old_beam = beam;
    } else {
      old_beam = beam;
    }
  }
  if (beamStarted) {
    beam2_end_vector.push_back(endTime/1000000);
  }
}

void get_beams_data(long long startTime, long long endTime, cool::IFolderPtr & folder) {
  beams_start_vector.clear();
  beams_end_vector.clear();
  cool::IObjectIteratorPtr objects = folder->browseObjects( cool::ValidityKey(startTime),cool::ValidityKey(endTime), cool::ChannelSelection::all());
  std::string old_beam = "FALSE";   bool beamStarted = false;
  while ( objects->goToNext() ) {
    const cool::IObject& obj = objects->currentRef();
    std::string beam = obj.payloadValue("StableBeams");
    if (((beam.compare("TRUE") == 0) && (old_beam.compare("FALSE") == 0)) ||
        ((beam.compare("TRUE") == 0) && (old_beam.compare("NULL") == 0))) {
      // start of a new beam interval
      long long start_beam = (long long) (obj.since());
      if (start_beam < startTime) {
        start_beam = startTime;
      }
      beams_start_vector.push_back(start_beam/1000000);
      old_beam = beam;
      beamStarted = true;
    } else if ((beam.compare("FALSE") == 0) && (old_beam.compare("TRUE") == 0)) {
      // end of a beam interval
      long long end_beam  = (long long) (obj.since());
      beams_end_vector.push_back(end_beam/1000000);
      beamStarted = false;
      old_beam = beam;
    } else {
      old_beam = beam;
    }
  }
  if (beamStarted) {
    beams_end_vector.push_back(endTime/1000000);
  }
}

void get_ramp_data(long long startTime, long long endTime, cool::IFolderPtr & folder) {
  // in fact the vectors contain NO RAMP UP intervals 
  ramp_start_vector.clear();
  ramp_end_vector.clear();
  cool::IObjectIteratorPtr objects = folder->browseObjects( cool::ValidityKey(startTime),cool::ValidityKey(endTime), cool::ChannelSelection::all());
  std::string old_beam_mode = "RAMP";   
  bool rampStarted = false;
  while ( objects->goToNext() ) {
    const cool::IObject& obj = objects->currentRef();
    std::string beam_mode = obj.payloadValue("BeamMode");
    // use the same conditions as Run Control to hold triggers
    // other conditions could be derived from the LHC state machine:
    //    INJECTION PROBE BEAM (?) -> PREPARE RAMP -> RAMP
    //    PREPARE RAMP -> BEAM DUMP
    //    RAMP -> BEAM DUMP
    //    RAMP -> FLAT TOP
    if (((beam_mode.compare("PREPARE RAMP") != 0) && (beam_mode.compare("RAMP") != 0)) && ((old_beam_mode.compare("PREPARE RAMP") == 0) || (old_beam_mode.compare("RAMP") == 0))) {
      // start of a new no ramp interval
      long long start_ramp = (long long) (obj.since());
      if (start_ramp < startTime) {
        start_ramp = startTime;
      }
      ramp_start_vector.push_back(start_ramp/1000000);
      old_beam_mode = beam_mode;
      rampStarted = true;
    } else if (((beam_mode.compare("PREPARE RAMP") == 0) || (beam_mode.compare("RAMP") == 0)) && ((old_beam_mode.compare("PREPARE RAMP") != 0) && (old_beam_mode.compare("RAMP") != 0))) {
      // end of a no ramp interval
      long long end_ramp  = (long long) (obj.since());
      ramp_end_vector.push_back(end_ramp/1000000);
      rampStarted = false;
      old_beam_mode = beam_mode;
    } else {
      old_beam_mode = beam_mode;
    }
  }
  if (rampStarted) {
    ramp_end_vector.push_back(endTime/1000000);
  }
}

void get_beam_data_from_file(long long starttime, long long endtime) {
  std::filebuf fb;
  fb.open ("beam.log",std::ios::in);
  if (fb.is_open()) {
    std::istream is(&fb);
    std::string str;
    long long bLimit = ((long long) beamLimit) * ((long long) 1000000);
    while (!(str.compare("endbeam") == 0)) {
      getline(is, str);
      if (!(str.compare("endbeam") == 0)) {
        OWLTime startBeam(str.substr(0, 17).c_str());
        long long startbeam = startBeam.total_mksec_utc();
        OWLTime endBeam(str.substr(19, 17).c_str());
        long long endbeam = endBeam.total_mksec_utc();
        if ((endbeam - startbeam) < bLimit) {
           continue;
        }
        if ((startbeam >= starttime) && (endbeam <= endtime)) {
          start_beam_vector.push_back(startbeam/1000);
          end_beam_vector.push_back(endbeam/1000);
        }
        if ((startbeam < starttime) && (endbeam > starttime) && (endbeam <= endtime)) {
          start_beam_vector.push_back(starttime/1000);
          end_beam_vector.push_back(endbeam/1000);
        }
        if ((startbeam < starttime) && (endbeam > endtime)) {
          start_beam_vector.push_back(starttime/1000);
          end_beam_vector.push_back(endtime/1000);
        }
        if ((startbeam >= starttime) && (startbeam < endtime) && (endbeam > endtime)) {
          start_beam_vector.push_back(startbeam/1000);
          end_beam_vector.push_back(endtime/1000);
        }
      }
    }
    fb.close();
  } else {
    std::cout << OWLTime() << " Warning: cannot open the beam data file" << std::endl;
  }
}

void print_vectors(std::vector<long long> start_vector, std::vector<long long> end_vector, std::string file_name) {
  std::ostream * m_lout = new std::ofstream (file_name.c_str(),std::ios::out | std::ios::trunc);
  if (!(*m_lout)) {
    std::cout << "Can't create the " << file_name << " log file " << std::endl;
    return;
  }
  std::ostringstream strm;
  if (start_vector.size() > 0) {
    for (unsigned int i = 0; i < start_vector.size(); i++ ) {
      long long start_beam_sec = start_vector[i]/1000;
      OWLTime start_owl = OWLTime(start_beam_sec);
      std::string start_string = std::string(start_owl.c_str());
      if (start_string.size() == 15) {
        strm << "  " << start_string;
      } else if (start_string.size() == 16) {
        strm << " " << start_string;
      } else {
        strm << start_string;
      }
      long long end_beam_sec = end_vector[i]/1000;
      OWLTime end_owl = OWLTime(end_beam_sec);
      std::string end_string = std::string(end_owl.c_str());
      strm << "  " << end_string << std::endl;
    }
    (*m_lout) << strm.str();
  }
  (*m_lout) << "endbeam" << std::endl;
  delete m_lout;
}

void set_main_vectors(std::vector<long long> start_vector, std::vector<long long> end_vector) {
  start_beam_vector.clear();
  end_beam_vector.clear(); 
  long long bLimit = ((long long) beamLimit) * ((long long) 1000);
  // exclude intervals before 30/03/2010 08:00:00
  long long timeLimit = ((long long)1269928800) * ((long long) 1000);
  if (start_vector.size() > 0) {
    for (unsigned int i = 0; i < start_vector.size(); i++ ) {
      if ((end_vector[i] - start_vector[i])  < bLimit) {
         continue;
      }
      if (end_vector[i] < timeLimit) {
         continue;
      }
      start_beam_vector.push_back(start_vector[i]);
      end_beam_vector.push_back(end_vector[i]);
    }
  }
}

void set_beam_vectors(long long startTime, long long endTime) {
  beam_start_vector.clear();
  beam_end_vector.clear();
  if (start_beam_vector.size() > 0) {
    for (unsigned int i = 0; i < start_beam_vector.size(); i++) {
      if ((startTime < start_beam_vector[i]) &&
          (end_beam_vector[i] < endTime)) {
        beam_start_vector.push_back(start_beam_vector[i]);
        beam_end_vector.push_back(end_beam_vector[i]);
      } else if ((start_beam_vector[i] <= startTime) &&
                 (startTime < end_beam_vector[i]) &&
                 (end_beam_vector[i] <= endTime)) {
        beam_start_vector.push_back(startTime);
        beam_end_vector.push_back(end_beam_vector[i]);
      } else if ((start_beam_vector[i] < startTime) &&
                 (endTime < end_beam_vector[i])) {
        beam_start_vector.push_back(startTime);
        beam_end_vector.push_back(endTime);
      } else if ((startTime <= start_beam_vector[i]) &&
                 (start_beam_vector[i] < endTime) &&
                 (endTime <= end_beam_vector[i])) {
        beam_start_vector.push_back(start_beam_vector[i]);
        beam_end_vector.push_back(endTime);
      }
    }
  }
}

int get_total_beam_time() {
  // get total time from beam_start_vector and beam_end_vector
  int total_beam_time = 0;
  int valbeam = 0;
  if (beam_start_vector.size() > 0) {
    for (unsigned int i = 0; i < beam_start_vector.size(); i++ ) {
      valbeam = beam_end_vector[i]/1000 - beam_start_vector[i]/1000;
      total_beam_time = total_beam_time + valbeam;
    }
  }
  return total_beam_time;
}

int get_ctpid(cool::ValidityKey sKey, cool::ValidityKey uKey, std::string itemName, cool::IFolderPtr & lvl1menu_folder) {
  int itemNo = -1;
  cool::IObjectIteratorPtr objects = lvl1menu_folder->browseObjects(sKey, uKey, cool::ChannelSelection::all());
  while ( objects->goToNext() ) {
    const cool::IObject& obj = objects->currentRef();
    std::string item_name = obj.payloadValue("ItemName");
    if (item_name.compare(itemName) == 0) {
      itemNo = obj.channelId();
      break;
    }
  }
  return itemNo;
}

int ready_for_physics(cool::ValidityKey key, cool::IFolderPtr & datatakingmode_folder) {
  int flag = 1;
  cool::IObjectPtr obj = datatakingmode_folder->findObject(key, 0);
  std::string readyForPhysics = obj->payloadValue("ReadyForPhysics");
  flag = atoi(readyForPhysics.c_str());
  return flag;
}

double veto_fraction(cool::ValidityKey key, int ctpid, cool::IFolderPtr & l1counters_folder) {
  // for the moment overflow flags are not used
  cool::IObjectPtr obj_l1counters = l1counters_folder->findObject(key, ctpid);
  std::string l1Accept = obj_l1counters->payloadValue("L1Accept");
  std::string afterPrescale = obj_l1counters->payloadValue("AfterPrescale");
  long long l1_accept = atoll(l1Accept.c_str());
  long long after_prescale = atoll(afterPrescale.c_str());
  double fraction = 0;
  if (debugMode == 1) {
    std::cout << "AP = " << after_prescale << " L1A = " << l1_accept << std::endl;
  }
  if (after_prescale > 0) {
    fraction = (double) (after_prescale - l1_accept) / (double) after_prescale;
  }
/*
  if (fraction > 0.1) {
    std::cout << "AP = " << after_prescale << " L1A = " << l1_accept << std::endl;
  }
*/
  return fraction;
}

float busy_run2(cool::ValidityKey key, std::string identifier, cool::IFolderPtr & the_busy_folder ) {

  // Find the busy values for a specific detector in this lumiblock:
  
  cool::IObjectPtr objects = the_busy_folder->findObject(key, 0);
  std::string this_busy_string= objects->payloadValue(identifier.c_str());
  float this_busy = atof(this_busy_string.c_str());
  return this_busy;
}

void openDatabasesAndFolders() {
  std::string coolDb;
  std::string coolDbTrig;
  std::string coolDbDCS;
  if (atlasRun == 2) {
    coolDb = "oracle://ATLR;schema=ATLAS_COOLONL_TDAQ;dbname=CONDBR2";
  } else if (atlasRun == 1) {
    coolDb = "oracle://ATLR;schema=ATLAS_COOLONL_TDAQ;dbname=COMP200";
  }
  cool::DatabaseId dbId = coolDb;
  // Open or Create DB
  cool::IDatabaseSvc& dbSvc = cool::DatabaseSvcFactory::databaseService();
  ERS_DEBUG(3, "Retrieved DB service.");

  try {
    openDB(dbId, dbSvc);
  }
  catch(daq::rc::BadCoolDb& ex) {
      ers::fatal(ex);
      exit(0);
  }

  if (atlasRun == 2) {
    coolDbTrig = "oracle://ATLR;schema=ATLAS_COOLONL_TRIGGER;dbname=CONDBR2";
  } else if (atlasRun == 1) {
    coolDbTrig = "oracle://ATLR;schema=ATLAS_COOLONL_TRIGGER;dbname=COMP200";
  }
  cool::DatabaseId dbIdTrig = coolDbTrig;
  cool::IDatabaseSvc& dbSvcTrig = cool::DatabaseSvcFactory::databaseService();
  try {
    openDBT(dbIdTrig, dbSvcTrig);
  }
  catch(daq::rc::BadCoolDb& ex) {
      ers::fatal(ex);
      exit(0);
  }

  if (atlasRun == 2) {
    coolDbDCS = "oracle://ATLR;schema=ATLAS_COOLOFL_DCS;dbname=CONDBR2;user=ATLAS_COOL_READER;password=COOLRED4PRO";
  } else if (atlasRun == 1) {
    coolDbDCS = "oracle://ATLR;schema=ATLAS_COOLOFL_DCS;dbname=COMP200";
  }
  cool::DatabaseId dbIdDCS = coolDbDCS;
  cool::IDatabaseSvc& dbSvcDCS = cool::DatabaseSvcFactory::databaseService();
  try {
    openDBD(dbIdDCS, dbSvcDCS);
  }
  catch(daq::rc::BadCoolDb& ex) {
      ers::fatal(ex);
      exit(0);
  }

  std::string coolDbMon = "oracle://ATLR;schema=ATLAS_COOLONL_TRIGGER;dbname=MONP200";
  cool::DatabaseId dbIdMon = coolDbMon;
  cool::IDatabaseSvc& dbSvcMon = cool::DatabaseSvcFactory::databaseService();
  try {
    openDBM(dbIdMon, dbSvcMon);
  }
    catch (daq::rc::BadCoolDb& ex) {
       ers::fatal(ex);
       exit(0);
  }

  // check folders
  try {
    if (!dbt->existsFolder(CTP_Name)) {
      std::cout << "DB Folder CTP does not exist???" << std::endl;
    }
  }
  catch (std::exception &e) {
     std::cout << "DB Folder CTP does not exist. " << e.what() << std::endl;
  }
  try {
     if (!dbd->existsFolder(FillState_Name)) {
        std::cout << "DB Folder FillState does not exist???" << std::endl;
     }
  } catch (std::exception &e) {
     std::cout << "DB Folder FillState does not exist. " << e.what() << std::endl;
  }
  try {
     if (!dbd->existsFolder(TotalInt_Name)) {
        std::cout << "DB Folder TotalInt does not exist???" << std::endl;
     }
  } catch (std::exception &e) {
     std::cout << "DB Folder TotalInt does not exist. " << e.what() << std::endl;
  }
  if (!dbt->existsFolder(LB_Name)) {
    std::cout << "DB Folder LB does not exist???" << std::endl;
  }
  if (!dbt->existsFolder(LBT_Name)) {
    std::cout << "DB Folder LBT does not exist???" << std::endl;
  }
  if (!dbt->existsFolder(L1Counters_Name)) {
    std::cout << "DB Folder L1Counters does not exist???" << std::endl;
  }
  if (!dbt->existsFolder(L1Menu_Name)) {
    std::cout << "DB Folder Menu does not exist???" << std::endl;
  }
  if (!db->existsFolder(DataTakingMode_Name)) {
    std::cout << "DB Folder DataTakingMode does not exist???" << std::endl;
  }
  //open folders

  try {
    CTP_Folder = dbt->getFolder(CTP_Name);
  }
  catch (std::exception &e) {
     std::cout << "Cannot open CTP folder " << e.what() << std::endl;
     dbt->closeDatabase();
     std::cout << "DBT closed." << std::endl;

     exit(0);
  }

  LB_Folder = dbt->getFolder(LB_Name);
  LBT_Folder = dbt->getFolder(LBT_Name);
  L1Counters_Folder = dbt->getFolder(L1Counters_Name);
  L1Menu_Folder = dbt->getFolder(L1Menu_Name);
  DataTakingMode_Folder = db->getFolder(DataTakingMode_Name);
  try {
    BusyRate_Folder = dbm->getFolder(BusyRate_Name);
  }
  catch (std::exception &e) {
     std::cout << "Cannot open BusyRate folder " << e.what() << std::endl;
     dbm->closeDatabase();
     std::cout << "DBM closed." << std::endl;

     exit(0);
  }
  try {
    FillState_Folder = dbd->getFolder(FillState_Name);
  }
  catch (std::exception &e) {
     std::cout << "Cannot open FillState folder " << e.what() << std::endl;
     dbd->closeDatabase();
     std::cout << "DBD closed." << std::endl;

     exit(0);
  }

  try {
    TotalInt_Folder = dbd->getFolder(TotalInt_Name);
  }
  catch (std::exception &e) {
     std::cout << "Cannot open TotalInt folder " << e.what() << std::endl;
     dbd->closeDatabase();
     std::cout << "DBD closed." << std::endl;

     exit(0);
  }
}

void set_beam1and2_vectors() {
  beam1and2_start_vector.clear();
  beam1and2_end_vector.clear();
  // set vectors
  if (beam1_start_vector.size() > 0) {
    for (unsigned int i = 0; i < beam1_start_vector.size(); i++ ) {
      if (beam2_start_vector.size() > 0) {
        for (unsigned int j = 0; j < beam2_start_vector.size(); j++ ) {
          if ((beam1_start_vector[i] <= beam2_start_vector[j]) &&
              (beam1_end_vector[i] > beam2_start_vector[j]) &&
              (beam1_end_vector[i] <= beam2_end_vector[j])) {
            beam1and2_start_vector.push_back(beam2_start_vector[j]);
            beam1and2_end_vector.push_back(beam1_end_vector[i]);
          } else {
            if ((beam1_start_vector[i] <= beam2_start_vector[j]) &&
                     (beam2_end_vector[j] <= beam1_end_vector[i])) {
              beam1and2_start_vector.push_back(beam2_start_vector[j]);
              beam1and2_end_vector.push_back(beam2_end_vector[j]);
            } else {
              if ((beam2_start_vector[j] <= beam1_start_vector[i]) &&
                   (beam1_start_vector[i] < beam2_end_vector[j]) &&
                   (beam2_end_vector[j] <= beam1_end_vector[i])) {
                 beam1and2_start_vector.push_back(beam1_start_vector[i]);
                 beam1and2_end_vector.push_back(beam2_end_vector[j]);
              } else {
                if ((beam2_start_vector[j] < beam1_start_vector[i]) &&
                    (beam1_end_vector[i] < beam2_end_vector[j])) {
                  beam1and2_start_vector.push_back(beam1_start_vector[i]);
                  beam1and2_end_vector.push_back(beam1_end_vector[i]);
                }
              }
            }
          }
        }
      }
    }
  }
}

bool checkInBeam(long long time) {
  if (beam1_start_vector.size() > 0) {
    for (unsigned int i = 0; i < beam1_start_vector.size(); i++) {
      if ((beam1_start_vector[i] <= time) &&
          (time <= beam1_end_vector[i])) {
        return true;
      }
    }
  }
  if (beam2_start_vector.size() > 0) {
    for (unsigned int i = 0; i < beam2_start_vector.size(); i++) {
      if ((beam2_start_vector[i] <= time) &&
          (time <= beam2_end_vector[i])) {
        return true;
      }
    }
  }
  return false;
}

void set_beam1or2_vectors() {
  beam1or2_start_vector.clear();
  beam1or2_end_vector.clear();
  long long starttime;
  long long endtime;
  int beam1_max = beam1_start_vector.size();
  int beam2_max = beam2_start_vector.size();
  if ((beam1_max > 0) && (beam2_max > 0)) {
    if (beam1_start_vector[0] <= beam2_start_vector[0]) {
      starttime = beam1_start_vector[0];
    } else {
      starttime = beam2_start_vector[0];
    }
    if (beam1_end_vector[beam1_max-1] <= beam2_end_vector[beam2_max-1]) {
      endtime = beam2_end_vector[beam2_max-1] + 1500;
    } else {
      endtime = beam1_end_vector[beam1_max-1] + 1500;
    }
  } else {
    if (beam1_max > 0) {
      starttime = beam1_start_vector[0];
      endtime = beam1_end_vector[beam1_max-1] + 1500;
    } else {
      if (beam2_max > 0) {
        starttime = beam2_start_vector[0];
        endtime = beam2_end_vector[beam2_max-1] + 1500;
      } else {
//        std::cout << "No beam 1 or beam 2 data" << std::endl;
        return;
      }
    }
  }
  long long currenttime = starttime;
  bool oldInBeam = false;
  int i = 0;
  while (currenttime < endtime) {
    bool inBeam = checkInBeam(currenttime);
    if ((inBeam == true) && (oldInBeam == false)) {
      beam1or2_start_vector.push_back(currenttime);
      oldInBeam = inBeam;
    } else {
      if ((inBeam == false) && (oldInBeam == true)) {
        beam1or2_end_vector.push_back(currenttime - 1000);
        oldInBeam = inBeam;
        i = i + 1;
      } else {
        oldInBeam = inBeam;
      }
    }
    // check every second
    currenttime = currenttime + 1000;
  }
}

void set_beam1and2andStable_vectors() {
  beam1and2andStable_start_vector.clear();
  beam1and2andStable_end_vector.clear();
  // set vectors
  if (beams_start_vector.size() > 0) {
    for (unsigned int i = 0; i < beams_start_vector.size(); i++ ) {
      if (beam1and2_start_vector.size() > 0) {
        for (unsigned int j = 0; j < beam1and2_start_vector.size(); j++ ) {
          if ((beams_start_vector[i] <= beam1and2_start_vector[j]) &&
              (beams_end_vector[i] > beam1and2_start_vector[j]) &&
              (beams_end_vector[i] <= beam1and2_end_vector[j])) {
            beam1and2andStable_start_vector.push_back(beam1and2_start_vector[j]);
            beam1and2andStable_end_vector.push_back(beams_end_vector[i]);
          } else {
            if ((beams_start_vector[i] <= beam1and2_start_vector[j]) &&
                (beam1and2_end_vector[j] <= beams_end_vector[i])) {
              beam1and2andStable_start_vector.push_back(beam1and2_start_vector[j]);
              beam1and2andStable_end_vector.push_back(beam1and2_end_vector[j]);
            } else {
              if ((beam1and2_start_vector[j] <= beams_start_vector[i]) &&
                  (beams_start_vector[i] < beam1and2_end_vector[j]) &&
                  (beam1and2_end_vector[j] <= beams_end_vector[i])) {
                 beam1and2andStable_start_vector.push_back(beams_start_vector[i]);
                 beam1and2andStable_end_vector.push_back(beam1and2_end_vector[j]);
              } else {
                if ((beam1and2_start_vector[j] < beams_start_vector[i]) &&
                    (beams_end_vector[i] < beam1and2_end_vector[j])) {
                  beam1and2andStable_start_vector.push_back(beams_start_vector[i]);
                  beam1and2andStable_end_vector.push_back(beams_end_vector[i]);
                }
              }
            }
          }
        }
      }
    }
  }
}

void set_beam1or2noRamp_vectors() {
  beam1or2noRamp_start_vector.clear();
  beam1or2noRamp_end_vector.clear();
  // set vectors
  if (beam1or2_start_vector.size() > 0) {
    for (unsigned int i = 0; i < beam1or2_start_vector.size(); i++ ) {
      if (ramp_start_vector.size() > 0) {
        for (unsigned int j = 0; j < ramp_start_vector.size(); j++ ) {
          if ((beam1or2_start_vector[i] <= ramp_start_vector[j]) &&
              (beam1or2_end_vector[i] > ramp_start_vector[j]) &&
              (beam1or2_end_vector[i] <= ramp_end_vector[j])) {
            beam1or2noRamp_start_vector.push_back(ramp_start_vector[j]);
            beam1or2noRamp_end_vector.push_back(beam1or2_end_vector[i]);
          } else {
            if ((beam1or2_start_vector[i] <= ramp_start_vector[j]) &&
                (ramp_end_vector[j] <= beam1or2_end_vector[i])) {
              beam1or2noRamp_start_vector.push_back(ramp_start_vector[j]);
              beam1or2noRamp_end_vector.push_back(ramp_end_vector[j]);
            } else {
              if ((ramp_start_vector[j] <= beam1or2_start_vector[i]) &&
                  (beam1or2_start_vector[i] < ramp_end_vector[j]) &&
                  (ramp_end_vector[j] <= beam1or2_end_vector[i])) {
                 beam1or2noRamp_start_vector.push_back(beam1or2_start_vector[i]);
                 beam1or2noRamp_end_vector.push_back(ramp_end_vector[j]);
              } else {
                if ((ramp_start_vector[j] < beam1or2_start_vector[i]) &&
                    (beam1or2_end_vector[i] < ramp_end_vector[j])) {
                  beam1or2noRamp_start_vector.push_back(beam1or2_start_vector[i]);
                  beam1or2noRamp_end_vector.push_back(beam1or2_end_vector[i]);
                }
              }
            }
          }
        }
      }
    }
  }
}

void set_beam1and2noRamp_vectors() {
  beam1and2noRamp_start_vector.clear();
  beam1and2noRamp_end_vector.clear();
  // set vectors
  if (beam1and2_start_vector.size() > 0) {
    for (unsigned int i = 0; i < beam1and2_start_vector.size(); i++ ) {
      if (ramp_start_vector.size() > 0) {
        for (unsigned int j = 0; j < ramp_start_vector.size(); j++ ) {
          if ((beam1and2_start_vector[i] <= ramp_start_vector[j]) &&
              (beam1and2_end_vector[i] > ramp_start_vector[j]) &&
              (beam1and2_end_vector[i] <= ramp_end_vector[j])) {
            beam1and2noRamp_start_vector.push_back(ramp_start_vector[j]);
            beam1and2noRamp_end_vector.push_back(beam1and2_end_vector[i]);
          } else {
            if ((beam1and2_start_vector[i] <= ramp_start_vector[j]) &&
                (ramp_end_vector[j] <= beam1and2_end_vector[i])) {
              beam1and2noRamp_start_vector.push_back(ramp_start_vector[j]);
              beam1and2noRamp_end_vector.push_back(ramp_end_vector[j]);
            } else {
              if ((ramp_start_vector[j] <= beam1and2_start_vector[i]) &&
                  (beam1and2_start_vector[i] < ramp_end_vector[j]) &&
                  (ramp_end_vector[j] <= beam1and2_end_vector[i])) {
                 beam1and2noRamp_start_vector.push_back(beam1and2_start_vector[i]);
                 beam1and2noRamp_end_vector.push_back(ramp_end_vector[j]);
              } else {
                if ((ramp_start_vector[j] < beam1and2_start_vector[i]) &&
                    (beam1and2_end_vector[i] < ramp_end_vector[j])) {
                  beam1and2noRamp_start_vector.push_back(beam1and2_start_vector[i]);
                  beam1and2noRamp_end_vector.push_back(beam1and2_end_vector[i]);
                }
              }
            }
          }
        }
      }
    }
  }
}

void setRunVectors (long long starttime, long long endtime) {
  start_of_run_vector.clear();
  end_of_run_vector.clear();

  // find list of runs and lumi blocks
  try {
    long long start_time_nsec = ((long long)(starttime))*1000;
    long long end_time_nsec = ((long long)(endtime))*1000;
    bool firstRun = true;
    bool newRun = false;
    int current_run_number = 0;
    long long start_of_first_LB_ms = 0;
    long long end_of_last_LB_ms = 0;
    cool::IObjectIteratorPtr rl_objects = LBT_Folder->browseObjects(cool::ValidityKey(start_time_nsec), cool::ValidityKey(end_time_nsec), 0);
    while ( rl_objects->goToNext() ) {
      const cool::IObject& obj = rl_objects->currentRef();
      std::string runNumber = obj.payloadValue("Run");
      std::string lumiBlock = obj.payloadValue("LumiBlock");
      int run_nb = atoi(runNumber.c_str());
      if (run_nb != current_run_number) {
//        std::cout << " New RN " << run_nb << " old RN was " << current_run_number << std::endl;
        newRun = true;
        current_run_number = run_nb;
      }
      cool::UInt63 rn = run_nb;
      cool::UInt63 lb = atoi(lumiBlock.c_str());
      cool::ValidityKey key1 = (rn<<32) + lb;
      cool::IObjectPtr obj_lb = LB_Folder->findObject(key1, 0);
      std::string st = obj_lb->payloadValue("StartTime");
      std::string et = obj_lb->payloadValue("EndTime");
      long long sT = atoll(st.c_str());
      long long eT = atoll(et.c_str());
      if (newRun) {
        // save data for previous run
        if (firstRun) {
          firstRun = false;
        } else {
//          std::cout << "RN = " << run_nb << "  LB = " << lb << std::endl;
          start_of_run_vector.push_back(start_of_first_LB_ms);
          end_of_run_vector.push_back(end_of_last_LB_ms);
        }
        start_of_first_LB_ms = sT/1000000;
        newRun = false;
      }
      end_of_last_LB_ms = eT/1000000;
    }
    // save data for last run
    start_of_run_vector.push_back(start_of_first_LB_ms);
    end_of_run_vector.push_back(end_of_last_LB_ms);
  } catch (std::exception &e) {
      std::cout << OWLTime() << " Warning: Read failed in setRunVectors: " << e.what() << std::endl;
  }
}

double getEfficiency (long long starttime, long long endtime) {
  totalBeamTime = 0;
  totalInBeamTime = 0;
  totalNoDataInBeamTime = 0;
  realInBeamTime = 0;

  int totalInBeamTime_msec = 0;

  // set beam vectors, having already start_beam_vector and end_beam_vector
  set_beam_vectors(starttime/1000, endtime/1000);

  totalBeamTime = get_total_beam_time();

  if (totalBeamTime > 0) {
    // find list of runs and lumi blocks
    int no_data_time_in_beam_msec = 0;
    try {
      long long start_time_nsec = ((long long)(starttime - 120000000))*1000;
      long long end_time_nsec = ((long long)(endtime))*1000;
      cool::IObjectIteratorPtr rl_objects = LBT_Folder->browseObjects(cool::ValidityKey(start_time_nsec), cool::ValidityKey(end_time_nsec), 0);
      int current_run_number = 0;
      int ctpId = 227; //default value, to be checked
      while ( rl_objects->goToNext() ) {
        const cool::IObject& obj = rl_objects->currentRef();
        std::string runNumber = obj.payloadValue("Run");
        std::string lumiBlock = obj.payloadValue("LumiBlock");
        int run_nb = atoi(runNumber.c_str());
        if (run_nb != current_run_number) {
//          std::cout << " New RN " << run_nb << " old RN was " << current_run_number << std::endl;
          current_run_number = run_nb;
          cool::UInt63 a = run_nb;
          cool::UInt63 z = run_nb+1;
          cool::ValidityKey sKey = a <<32;
          cool::ValidityKey uKey = (z <<32) - 1;
          if (useTotalBusy == 0) {
            ctpId = get_ctpid(sKey, uKey, triggerItem, L1Menu_Folder);
            if (ctpId == -1) {
              std::cout << "For Run " << runNumber << " Trigger Item not found. Default value (227) will be used." << std::endl;
              ctpId = 227;
            }
          } 
        }
        cool::UInt63 rn = run_nb;
        cool::UInt63 lb = atoi(lumiBlock.c_str());
//        std::cout << "RN = " << run_nb << "  LB = " << lb << std::endl;
        cool::ValidityKey key1 = (rn<<32) + lb;
        cool::IObjectPtr obj_lb = LB_Folder->findObject(key1, 0);
        std::string st = obj_lb->payloadValue("StartTime");
        std::string et = obj_lb->payloadValue("EndTime");
        long long sT = atoll(st.c_str());
        long long eT = atoll(et.c_str());
        int readyForPhysicsFlag = 1;
        if (((beamType == 0) || (beamType == 3)) && (readyForPhysics == 1)) {
          readyForPhysicsFlag = ready_for_physics(key1, DataTakingMode_Folder);
        }
        double vetofraction = 0.0;
        try {
          if (useTotalBusy == 0) {
            vetofraction = veto_fraction(key1, ctpId, L1Counters_Folder);
          } else {
            vetofraction = (double) busy_run2(key1, "ctpcore_moni0_rate", BusyRate_Folder );
          }
        } catch(std::exception &ex) {
          std::cout << "Failure in veto fraction evaluation - " << ex.what() << std::endl;
          vetofraction = 1.0;
        }
        if (readyForPhysicsFlag == 0) {
          vetofraction = 1;
        }
        if (beam_start_vector.size() > 0) {
          for (unsigned int i = 0; i < beam_start_vector.size(); i++ ) {
            int data_lumi_block_in_beam_msec = valInBeam(sT/1000000, eT/1000000, beam_start_vector[i], beam_end_vector[i]);
//            std::cout << "data_lumi_block_in_beam_msec = " << data_lumi_block_in_beam_msec << " vf = " << vetofraction << std::endl;
            if (data_lumi_block_in_beam_msec > 0) {
//              totalInBeamTime = totalInBeamTime + data_lumi_block_in_beam_msec/1000;
              totalInBeamTime_msec = totalInBeamTime_msec + data_lumi_block_in_beam_msec;
              int no_data_lumi_block_in_beam_msec = (int) (((double) data_lumi_block_in_beam_msec) * vetofraction);
              no_data_time_in_beam_msec = no_data_time_in_beam_msec + no_data_lumi_block_in_beam_msec;
            }
          }
        }     
        totalNoDataInBeamTime = no_data_time_in_beam_msec/1000;
        totalInBeamTime = totalInBeamTime_msec/1000;
        realInBeamTime = totalInBeamTime - no_data_time_in_beam_msec/1000;
      }
    } catch (std::exception &e) {
      std::cout << OWLTime() << " Warning: Read failed: " << e.what() << std::endl;
    }
  }
  // evaluate efficiency in beam
  double beamEff = 0.0;
  if (totalBeamTime > 0) {
//    std::cout << "realInBeamTime = " << realInBeamTime << "  totalBeamTime = " << totalBeamTime << std::endl;
    beamEff = (double)realInBeamTime/(double)totalBeamTime;
  } else {
    if (noBeamValue == 0) {
      beamEff = 0.0;
    } else {
      beamEff = -0.2;
    }
  }
  return beamEff;
}

long long end_delay (long long starttime, long long endtime) {
  long long start_time_nsec = starttime*1000;
  long long end_time_nsec = endtime*1000;
  int current_run_number = 0;
  cool::IObjectIteratorPtr rl_objects = LBT_Folder->browseObjects(cool::ValidityKey(start_time_nsec), cool::ValidityKey(end_time_nsec), 0);
  long long end_of_last_LB_with_ready_mksec = 0;
  while ( rl_objects->goToNext() ) {
    const cool::IObject& obj = rl_objects->currentRef();
    std::string runNumber = obj.payloadValue("Run");
    std::string lumiBlock = obj.payloadValue("LumiBlock");
    int run_nb = atoi(runNumber.c_str());
    if (run_nb != current_run_number) {
//      std::cout << " New RN " << run_nb << " old RN was " << current_run_number << std::endl;
      current_run_number = run_nb;
      end_of_last_LB_with_ready_mksec = 0;
    }
    cool::UInt63 rn = run_nb;
    cool::UInt63 lb = atoi(lumiBlock.c_str());
    cool::ValidityKey key1 = (rn<<32) + lb;
    cool::IObjectPtr obj_lb = LB_Folder->findObject(key1, 0);
    std::string st = obj_lb->payloadValue("StartTime");
    std::string et = obj_lb->payloadValue("EndTime");
    long long eT = atoll(et.c_str());
    if (ready_for_physics(key1, DataTakingMode_Folder) == 1) {
      end_of_last_LB_with_ready_mksec = eT/1000;
    }
  }
  long long end_delay_mksec = 0;
  if (end_of_last_LB_with_ready_mksec > 0) {
    end_delay_mksec = endtime - end_of_last_LB_with_ready_mksec;
  }
  if (end_delay_mksec < 0) {
    end_delay_mksec = 0;
  }
  return end_delay_mksec;
}



//----------------------- MAIN ----------------------------//
int main(int argc, char ** argv)
{

  // parse commandline parameters
  
  std::string startTime, endTime, interval;
  bool timeInterval = false;
  int duration = 60;
  int period = 60;
  int evaluationMode = 0;
  int excludeWarmStop = 0;
  boost::program_options::options_description desc("This program retrieves info from COOL conditions database, evaluates the data taking efficiency for selected time intervals and stores the results in /tmp/eff.log file.");
  desc.add_options()
    ("help,h", "help message")
    ("startTime,s", boost::program_options::value<std::string>(), "Start time (\"dd/mm/yy hh:mm:ss\")")
    ("endTime,e", boost::program_options::value<std::string>(), "End time (\"dd/mm/yy hh:mm:ss\")")
    ("interval,i", boost::program_options::value<std::string>(), "Time interval (day, week, month) - to be used only if no start time or end time is specified")
    ("duration,d", boost::program_options::value<int>(&duration)->default_value(duration), "Duration for evaluation in minutes (default 60)")
    ("period,p", boost::program_options::value<int>(&period)->default_value(period), "Period for evaluation in minutes (default 60)")
    ("atlasRun,r", boost::program_options::value<int>(&atlasRun)->default_value(atlasRun), "ATLAS Run: 1 - Run 1; 2 - Run 2 (default 2)")
    ("beamType,r", boost::program_options::value<int>(&beamType)->default_value(beamType), "Beam type: 0 - stable beams; 1 - two beams present; 2 - one beam present; 3 - two beams present and stable; 4 - beam data from file (default 3)")
    ("beamLimit,b", boost::program_options::value<int>(&beamLimit)->default_value(beamLimit), "Beam limit option in seconds (default 5 s)")
    ("writeBeamFile,w", boost::program_options::value<int>(&writeBeamFile)->default_value(writeBeamFile), "Write beam file: 0 - no beam file; 1 - write beam file (default 0)")
    ("display,x", boost::program_options::value<int>(&display)->default_value(display), "Display options: 0 - no graphical display, 1 - open graphical display (default 0)")
    ("noRamp,n",boost::program_options::value<int>(&noRamp)->default_value(noRamp), "No ramp option: 0 - ramps not excluded from beam time, 1 - ramps excluded from beam time (default 0)")
    ("busyLimit,l", boost::program_options::value<double>(&busyLimit)->default_value(busyLimit), "Busy limit (default 0.03)")
    ("triggerItem,c", boost::program_options::value<std::string>(&triggerItem)->default_value(triggerItem), "Trigger item (default L1_EM30))")
    ("readyForPhysics,f", boost::program_options::value<int>(&readyForPhysics)->default_value(readyForPhysics), "Ready for physics option: 0 - do not use ready for physics flag, 1 - use ready for physics flag (default 0)")
    ("displayMode,o", boost::program_options::value<int>(&displayMode)->default_value(displayMode), "Display mode options: 0 - no values ouside beams, 1 - one point per interval (default 0)")
    ("noBeamValue,v", boost::program_options::value<int>(&noBeamValue)->default_value(noBeamValue), "No beam value option: 0 - 0 if no beam, 1 - -0.2 if no beam (default 0)")
    ("debugMode,g", boost::program_options::value<int>(&debugMode)->default_value(debugMode), "Debug mode option: 0 - no debug info, 1 - with debug info (default 0)")
    ("evaluationMode,a", boost::program_options::value<int>(&evaluationMode)->default_value(evaluationMode), "Evaluation mode option: 0 - evaluate efficiency periodically, 1 - evaluate efficiency for each beam interval, 2 - evaluate efficiency and running time for each beam interval and prepare data for root graphs, 3 - evaluate efficiency periodically for ready and not ready and prepare data for root graphs, 4 - evaluate efficiency and running time for each beam interval and update old data for root graphs, 5 - evaluate efficiency periodically for ready and not ready and update old data for root graphs (default 0)")
    ("excludeWarmStop,k", boost::program_options::value<int>(&excludeWarmStop)->default_value(excludeWarmStop), "Exclude the interval after WarmStop option: 0 - evaluate efficiency including the interval after WarmStop, 1 - evaluate efficiency exluding the interval after WarmStop (default 0)")
    ("useTotalBusy,t", boost::program_options::value<int>(&useTotalBusy)->default_value(useTotalBusy), "Use Total Busy: 0 - use LVL1 counters; 1 - use total busy (default 1)")
    ;

  boost::program_options::variables_map vm;
  boost::program_options::store(boost::program_options::parse_command_line(argc,argv,desc), vm);
  boost::program_options::notify(vm);

  if(vm.count("help")) {
    std::cout << desc << std::endl;
    return EXIT_SUCCESS;
  }
  if ((useTotalBusy < 0) || (useTotalBusy > 1)) {
    std::cout << "Command line error: useTotalBusy should be 0 or 1" << std::endl;
    return EXIT_FAILURE;
  }
  if ((atlasRun < 1) || (atlasRun > 2)) {
    std::cout << "Command line error: atlasRun should be 1 or 2" << std::endl;
    return EXIT_FAILURE;
  }
  if ( (vm.count("interval")) && ( (vm.count("startTime")) || (vm.count("endTime")) ) ) {
   std::cout << desc << std::endl;
   std::cout << "Command line error: time interval used with an explicit start time or end time" << std::endl;
   return EXIT_FAILURE;
  }
  else if ( (vm.count("startTime")) && (vm.count("endTime")) ) {
    startTime = vm["startTime"].as<std::string>();
    endTime = vm["endTime"].as<std::string>();
  }
  else if ( vm.count("interval") ) {
    interval = vm["interval"].as<std::string>();
    timeInterval = true;
  } 
  else {
    std::cout << desc << std::endl;
    daq::rc::CmdLineError issue(ERS_HERE,"");
    ers::fatal(issue);
    return EXIT_FAILURE;
  }

  std::string runs("/tmp/eff.log");
  std::ostream * m_lout = new std::ofstream (runs.c_str(),std::ios::out | std::ios::trunc);        
  if (!(*m_lout)) {
    std::cout << "Can't create the efficiency log file " << std::endl;
  }

  openDatabasesAndFolders();

  // Install signal handler
  signal(SIGINT,  signalhandler);
  signal(SIGTERM, signalhandler);
  ERS_DEBUG(3, "Installed signal handlers.");

  long long interval_mksec = ((long long) 60) * ((long long) 1000000);
  interval_mksec = period * interval_mksec;
  long long duration_mksec = ((long long) 60) * ((long long) 1000000);
  duration_mksec = duration * duration_mksec;
  OWLTime now = OWLTime();
  long long nowtime = now.total_mksec_utc();
  long long starttime = 0;
  long long endtime = 0;
  if (!timeInterval) {
    OWLTime start_time(startTime.c_str());
    starttime = start_time.total_mksec_utc();
    OWLTime end_time(endTime.c_str());
    endtime  = end_time.total_mksec_utc();
    if (starttime > nowtime) {
      std::cout << " Start time greater than current time" << std::endl;
      return EXIT_FAILURE;
    }
    if (endtime > nowtime) {
      std::cout << " End time greater than current time" << std::endl;
      return EXIT_FAILURE;
    }
  }
  else {
    long long total_duration_mksec = 0;
    if (interval.compare("day") == 0) {
      total_duration_mksec = ((long long) 86400) * ((long long) 1000000);
    }
    else if (interval.compare("week") == 0) {
      total_duration_mksec = ((long long) 604800) * ((long long) 1000000);
    }
    else if (interval.compare("month") == 0) {
      total_duration_mksec = ((long long) 2592000) * ((long long) 1000000);
    } 
    else {
      std::cout << desc << std::endl;
      std::cout << "Wrong time interval option!" << std::endl;
      return EXIT_FAILURE;
    }
    endtime  = nowtime;
    // take into account total duration
    starttime = endtime - total_duration_mksec;
    endTime = std::string(now.c_str());
    OWLTime start = OWLTime(starttime/1000000);
    startTime = std::string(start.c_str());
  }
  // now take into account the duration and the period
  starttime = starttime + interval_mksec - duration_mksec;

  // for update mode
  bool updateLastFill = false;
  if (evaluationMode == 4) {
    // read old run eff data
    std::filebuf fb;
    fb.open ("/tmp/runeff_old.data",std::ios::in);
    if (fb.is_open()) {
      std::istream is(&fb);
      std::string line;
      while (!(line.compare("enddata") == 0)) {
        getline(is, line);
        if (!(line.compare("enddata") == 0)) {
          runeff_data.push_back(line);
        }
      }
      fb.close();
    } else {
      std::cout << "Warning: cannot open the old runeff data file" << std::endl;
      db->closeDatabase();
      dbt->closeDatabase();
      dbd->closeDatabase();
      std::cout << "DB closed." << std::endl;
      return EXIT_SUCCESS;
    }
    // read old run time data
    fb.open ("/tmp/runtime_old.data",std::ios::in);
    if (fb.is_open()) {
      std::istream is(&fb);
      std::string line;
      while (!(line.compare("enddata") == 0)) {
        getline(is, line);
        if (!(line.compare("enddata") == 0)) {
          runtime_data.push_back(line);
        }
      }
      fb.close();
    } else {
      std::cout << "Warning: cannot open the old runtime data file" << std::endl;    
      db->closeDatabase();
      dbt->closeDatabase();
      dbd->closeDatabase();
      std::cout << "DB closed." << std::endl;
      return EXIT_SUCCESS;
    }
    int n = runeff_data.size();
    std::string eff_data = runeff_data[n-5];
    n = runtime_data.size();
    std::string time_data = runtime_data[n-2];
    std::string end_time_old = time_data.substr(0, 10);
    long end_time_old_sec = atol(end_time_old.c_str());
    // check if the new start evaluation time is during a beam interval
    if (end_time_old_sec == starttime/1000000) {
      updateLastFill = true;
      std::string start_time_old = eff_data.substr(0, 10);
      long start_time_old_sec = atol(start_time_old.c_str());
      // consider starttime 10 sec before start of last beam interval
      starttime = ((long long)(start_time_old_sec-10))*1000000;
    }
//    std::cout << "updateLastFill = " << updateLastFill << std::endl;
  } 
  if (evaluationMode == 5) {
    // read old run eff data
    std::filebuf fb;
    fb.open ("/tmp/runeff_old.data",std::ios::in);
    if (fb.is_open()) {
      std::istream is(&fb);
      std::string line;
      while (!(line.compare("enddata") == 0)) {
        getline(is, line);
        if (!(line.compare("enddata") == 0)) {
          runeff_data.push_back(line);
        }
      }
      fb.close();
    } else {
      std::cout << "Warning: cannot open the old runeff data file" << std::endl;
      db->closeDatabase();
      dbt->closeDatabase();
      dbd->closeDatabase();
      std::cout << "DB closed." << std::endl;
      return EXIT_SUCCESS;
    }
  }

  // read beam data
  if (beamType == 0) {
    // stable beams
    get_beams_data (starttime*1000, endtime*1000, FillState_Folder);
    set_main_vectors(beams_start_vector, beams_end_vector);
    if (writeBeamFile == 1) {
      print_vectors(start_beam_vector, end_beam_vector, "/tmp/beams.log");
    }
  }
  if (beamType == 1) {
    // two beams present
    get_beam1_data (starttime*1000, endtime*1000, TotalInt_Folder);
    get_beam2_data (starttime*1000, endtime*1000, TotalInt_Folder);
    set_beam1and2_vectors();
    if (noRamp == 0) {
      set_main_vectors(beam1and2_start_vector, beam1and2_end_vector);
    }
    if (noRamp == 1) {
      get_ramp_data (starttime*1000, endtime*1000, FillState_Folder);
      set_beam1and2noRamp_vectors();
      set_main_vectors(beam1and2noRamp_start_vector, beam1and2noRamp_end_vector);
    }
    if (writeBeamFile == 1) {
      print_vectors(beam1_start_vector, beam1_end_vector, "/tmp/beam1.log");
      print_vectors(beam2_start_vector, beam2_end_vector, "/tmp/beam2.log");
      print_vectors(beam1and2_start_vector, beam1and2_end_vector, "/tmp/beam1and2.log");
      if (noRamp == 1) {
        print_vectors(ramp_start_vector, ramp_end_vector, "/tmp/ramp.log");
        print_vectors(start_beam_vector, end_beam_vector, "/tmp/beam1and2noRamp.log");
      }
    }
  }
  if (beamType == 2) {
    // one beam present
    get_beam1_data (starttime*1000, endtime*1000, TotalInt_Folder);
    get_beam2_data (starttime*1000, endtime*1000, TotalInt_Folder);
    set_beam1or2_vectors();
    if (noRamp == 0) {
      set_main_vectors(beam1or2_start_vector, beam1or2_end_vector);
    }
    if (noRamp == 1) {
      get_ramp_data (starttime*1000, endtime*1000, FillState_Folder);
      set_beam1or2noRamp_vectors();
      set_main_vectors(beam1or2noRamp_start_vector, beam1or2noRamp_end_vector);
    }
    if (writeBeamFile == 1) {
      print_vectors(beam1_start_vector, beam1_end_vector, "/tmp/beam1.log");
      print_vectors(beam2_start_vector, beam2_end_vector, "/tmp/beam2.log");
      print_vectors(beam1or2_start_vector, beam1or2_end_vector, "/tmp/beam1or2.log");  
      if (noRamp == 1) {
        print_vectors(ramp_start_vector, ramp_end_vector, "/tmp/ramp.log");
        print_vectors(start_beam_vector, end_beam_vector, "/tmp/beam1or2noRamp.log");
      }
    }
  }
  if (beamType == 3) {
    // two beams present and stable
    get_beam1_data (starttime*1000, endtime*1000, TotalInt_Folder);
    get_beam2_data (starttime*1000, endtime*1000, TotalInt_Folder);
    get_beams_data (starttime*1000, endtime*1000, FillState_Folder);
    set_beam1and2_vectors();
    set_beam1and2andStable_vectors();
    set_main_vectors(beam1and2andStable_start_vector, beam1and2andStable_end_vector);
    if (writeBeamFile == 1) {
      print_vectors(beam1_start_vector, beam1_end_vector, "/tmp/beam1.log");
      print_vectors(beam2_start_vector, beam2_end_vector, "/tmp/beam2.log");
      print_vectors(beams_start_vector, beams_end_vector, "/tmp/beams.log");
      print_vectors(start_beam_vector, end_beam_vector, "/tmp/beam1and2andStable.log");
    }
  }
  if (beamType == 4) {
    // beam data from file
    get_beam_data_from_file (starttime, endtime);
    if (writeBeamFile == 1) {
      print_vectors(start_beam_vector, end_beam_vector, "/tmp/beamfromfile.log");
    }
  }
  if (start_beam_vector.size() > 0) {
    for (unsigned int i = 0; i < start_beam_vector.size(); i++ ) {
      (*m_lout) << start_beam_vector[i] << " " << end_beam_vector[i] << std::endl;
    }
  }
  (*m_lout) << "endbeam" << std::endl;

  long long starttimeG = starttime + duration_mksec - interval_mksec;
  long long endtimeG = endtime;

  // read run data
  if ((evaluationMode == 2) || (evaluationMode == 4)) {
    // prepare data files
    std::ostream * eff_out = new std::ofstream ("/tmp/runeff.data", std::ios::out | std::ios::trunc);
    if (!(*eff_out)) {
      std::cout << "Can't create the efficiency data file " << std::endl;
    }
    std::ostream * runtime_out = new std::ofstream ("/tmp/runtime.data", std::ios::out | std::ios::trunc);
    if (!(*runtime_out)) {
      std::cout << "Can't create the runtime data file " << std::endl;
    }
    int initialReadyForPhysics = readyForPhysics;
    int globalBeamTime = 0;
    int globalRunningTime = 0;
    int globalRunningReadyTime = 0;
    int globalBeamNoWarmStopTime = 0;
    int globalRunningReadyNoWarmStopTime = 0;
    if (evaluationMode == 4) {
      if (updateLastFill == true) {
        // remove data for last beam interval (interval to be updated)
        runeff_data.pop_back();
        runeff_data.pop_back();
        runeff_data.pop_back();
        runeff_data.pop_back();
        runeff_data.pop_back();
        runtime_data.pop_back();
        runtime_data.pop_back();
        runtime_data.pop_back();
      } else {
        // remove last entries, not useful
        runeff_data.pop_back();
        runtime_data.pop_back();
      }
      // insert old data
      unsigned int n = runeff_data.size();
      if (n > 0) {
        for (unsigned int i = 0; i < n; i++ ) {
          (*eff_out) << runeff_data[i] << std::endl;
        }
      }
      n = runtime_data.size();
      if (n > 0) {
        for (unsigned int i = 0; i < n; i++ ) {
          (*runtime_out) << runtime_data[i] << std::endl;
        }
      }
      // update time info
      std::string time_data = runtime_data[n-1];
      size_t beamTimeEndIndex = time_data.find_first_of(" ", 14);
      std::string beamTimeValue = time_data.substr(14, beamTimeEndIndex-14);
      globalBeamTime = atoi(beamTimeValue.c_str());
      size_t runTimeEndIndex = time_data.find_first_of(" ", beamTimeEndIndex+1);
      std::string runTimeValue = time_data.substr(beamTimeEndIndex+1, runTimeEndIndex-beamTimeEndIndex-1);
      globalRunningTime = atoi(runTimeValue.c_str());
      size_t runTimeReadyEndIndex = time_data.find_first_of(" ", runTimeEndIndex+1);
      std::string runTimeReadyValue = time_data.substr(runTimeEndIndex+1, runTimeReadyEndIndex-runTimeEndIndex-1);
      globalRunningReadyTime = atoi(runTimeReadyValue.c_str());
      if (excludeWarmStop == 1) {
        size_t beamTimeNoWarmStopEndIndex = time_data.find_first_of(" ", runTimeReadyEndIndex+1);
        if (beamTimeNoWarmStopEndIndex == std::string::npos) {
          std::cout << "No data about No WarmStop available in runtime.data" << std::endl;
        } else {
          std::string beamNoWarmStopTimeValue = time_data.substr(runTimeReadyEndIndex+1, beamTimeNoWarmStopEndIndex-runTimeReadyEndIndex-1);
          globalBeamNoWarmStopTime = atoi(beamNoWarmStopTimeValue.c_str());
          size_t lastIndex = time_data.size();
          std::string runTimeReadyNoWarmStopValue = time_data.substr(beamTimeNoWarmStopEndIndex+1, lastIndex-beamTimeNoWarmStopEndIndex-1);
          globalRunningReadyNoWarmStopTime = atoi(runTimeReadyNoWarmStopValue.c_str());
        }
      }
    }      

    int globalBeamRunningTime = 0;
    int globalBeamRunningReadyTime = 0;
//    int globalBeamRunningReadyNoWarmStopTime = 0;
    int globalNoDataInBeamTime = 0;
    int globalNoDataInBeamReadyTime = 0;
//    int globalNoDataInBeamReadyNoWarmStopTime = 0;
    double runEff = 0;
    double runEffReady = 0;
    double runEffReadyNoWarmStop = 0;
    double noBeam = 0.0;
    double noData = 0.0;
    double maxBeam = 1.0;
    if (noBeamValue == 1) {
      noBeam = -0.2;
    }
    if (start_beam_vector.size() > 0) {
      for (unsigned int i = 0; i < start_beam_vector.size(); i++ ) {
        int initialBeamTime = globalBeamTime;
        int initialRunningTime = globalRunningTime;
        int initialRunningReadyTime = globalRunningReadyTime;
        int initialBeamNoWarmStopTime = globalBeamNoWarmStopTime;
        int initialRunningReadyNoWarmStopTime = globalRunningReadyNoWarmStopTime;
        starttime = start_beam_vector[i]*1000;
        endtime = end_beam_vector[i]*1000;
        // evaluate efficiency without RfP flag
        readyForPhysics = 0;
        runEff = getEfficiency (starttime, endtime);
        globalBeamTime = globalBeamTime + totalBeamTime;
        globalBeamRunningTime = globalBeamRunningTime + totalInBeamTime;
        globalNoDataInBeamTime = globalNoDataInBeamTime + totalNoDataInBeamTime;
        globalRunningTime = globalRunningTime + realInBeamTime;
        OWLTime current = OWLTime(end_beam_vector[i]/1000);
        if (initialReadyForPhysics == 0) {
          std::cout << current.c_str() << "  " << runEff << std::endl;
          (*m_lout) << start_beam_vector[i] << " " << noBeam << std::endl;
          (*m_lout) << start_beam_vector[i] + 1000 << " " << runEff << std::endl;        
          (*m_lout) << end_beam_vector[i] - 1000 << " " << runEff << std::endl;
          (*m_lout) << end_beam_vector[i] << " " << noBeam << std::endl;
        }
        // evaluate efficiency with RfP flag
        readyForPhysics = 1;
        runEffReady = getEfficiency (starttime, endtime);
        globalBeamRunningReadyTime = globalBeamRunningReadyTime + totalInBeamTime;
        globalNoDataInBeamReadyTime = globalNoDataInBeamReadyTime + totalNoDataInBeamTime;
        globalRunningReadyTime = globalRunningReadyTime + realInBeamTime;
        if (initialReadyForPhysics == 1) {
          std::cout << current.c_str() << "  " << runEffReady << std::endl;
          (*m_lout) << start_beam_vector[i] << " " << noBeam << std::endl;
          (*m_lout) << start_beam_vector[i] + 1000 << " " << runEffReady << std::endl;
          (*m_lout) << end_beam_vector[i] - 1000 << " " << runEffReady << std::endl;
          (*m_lout) << end_beam_vector[i] << " " << noBeam << std::endl;
        }
        if (excludeWarmStop == 0) {
          (*eff_out) << start_beam_vector[i] << " " << noBeam << " " << noData << " " << noData << std::endl;
          (*eff_out) << start_beam_vector[i] + 1000 << " " << maxBeam << " " << runEff << " " << runEffReady << std::endl;
          (*eff_out) << end_beam_vector[i] - 1000 << " " << maxBeam << " " << runEff << " " << runEffReady << std::endl;
          (*eff_out) << end_beam_vector[i] << " " << noBeam << " " << noData << " " << noData << std::endl;
          (*runtime_out) << start_beam_vector[i] << " " << initialBeamTime << " " << initialRunningTime << " " << initialRunningReadyTime << std::endl;
          (*runtime_out) << end_beam_vector[i] << " " << globalBeamTime << " " << globalRunningTime << " " << globalRunningReadyTime << std::endl;
        } else {
          // evaluate efficiency with RfP flag and excluding WarmStop
          long long endDelay = end_delay(starttime, endtime);
          if (endDelay == 0) {
            globalRunningReadyNoWarmStopTime = globalRunningReadyTime;
            runEffReadyNoWarmStop = runEffReady;
          } else {
            endtime = endtime - endDelay;
            runEffReadyNoWarmStop = getEfficiency (starttime, endtime);
            globalRunningReadyNoWarmStopTime = globalRunningReadyNoWarmStopTime + realInBeamTime;
          }
          globalBeamNoWarmStopTime = globalBeamNoWarmStopTime + totalBeamTime; 
          (*eff_out) << start_beam_vector[i] << " " << noBeam << " " << noData << " " << noData << " " << noData << std::endl;
          (*eff_out) << start_beam_vector[i] + 1000 << " " << maxBeam << " " << runEff << " " << runEffReady << " " << runEffReadyNoWarmStop << std::endl;
          (*eff_out) << end_beam_vector[i] - 1000 << " " << maxBeam << " " << runEff << " " << runEffReady << " " << runEffReadyNoWarmStop << std::endl;
          (*eff_out) << end_beam_vector[i] << " " << noBeam << " " << noData << " " << noData << " " << noData << std::endl;
          (*runtime_out) << start_beam_vector[i] << " " << initialBeamTime << " " << initialRunningTime << " " << initialRunningReadyTime << " " << initialBeamNoWarmStopTime << " " << initialRunningReadyNoWarmStopTime << std::endl;
          (*runtime_out) << end_beam_vector[i] << " " << globalBeamTime << " " << globalRunningTime << " " << globalRunningReadyTime << " " << globalBeamNoWarmStopTime << " " << globalRunningReadyNoWarmStopTime << std::endl;
        }
      }
    }
    if (excludeWarmStop == 0) {
      (*eff_out) << endtimeG/1000 << " " << noBeam << " " << noData << " " << noData << std::endl;
      (*runtime_out) << endtimeG/1000 <<" " << globalBeamTime << " " << globalRunningTime << " " << globalRunningReadyTime << std::endl;
    } else {
      (*eff_out) << endtimeG/1000 << " " << noBeam << " " << noData << " " << noData << " " << noData << std::endl;
      (*runtime_out) << endtimeG/1000 <<" " << globalBeamTime << " " << globalRunningTime << " " << globalRunningReadyTime << " " << globalBeamNoWarmStopTime << " " << globalRunningReadyNoWarmStopTime << std::endl;
    }
    (*eff_out) << "enddata" << std::endl;
    (*runtime_out) << "enddata" << std::endl;
    integralBeamTime = globalBeamTime;
    integralBeamRunningTime = globalBeamRunningTime;
    integralDeadTime = globalNoDataInBeamTime;
    integralRunningTime = globalRunningTime;
    integralRunEff = (double) globalRunningTime / (double) globalBeamTime;
    integralBeamRunningReadyTime = globalBeamRunningReadyTime;
    integralReadyDeadTime = globalNoDataInBeamReadyTime;
    integralRunningReadyTime = globalRunningReadyTime;
    integralRunEffReady = (double) globalRunningReadyTime / (double) globalBeamTime;
    if (excludeWarmStop == 1) {
      integralBeamNoWarmStopTime = globalBeamNoWarmStopTime;
      integralRunningReadyNoWarmStopTime = globalRunningReadyNoWarmStopTime;
      integralRunEffReadyNoWarmStop = (double) globalRunningReadyNoWarmStopTime / (double) globalBeamNoWarmStopTime;
    }
    readyForPhysics = initialReadyForPhysics;
    delete eff_out;
    delete runtime_out;
  }
  if (evaluationMode == 1) {
    if (start_beam_vector.size() > 0) {
      for (unsigned int i = 0; i < start_beam_vector.size(); i++ ) {
        starttime = start_beam_vector[i]*1000;
        endtime = end_beam_vector[i]*1000;
        double beamEff = getEfficiency (starttime, endtime);
        OWLTime current = OWLTime(end_beam_vector[i]/1000);
        std::cout << current.c_str() << "  " << beamEff << std::endl;
        double noBeam = 0.0;
        if (noBeamValue == 1) {
          noBeam = -0.2;
        }
        (*m_lout) << start_beam_vector[i] << " " << noBeam << std::endl;
        (*m_lout) << start_beam_vector[i] + 1000 << " " << beamEff << std::endl;
        (*m_lout) << end_beam_vector[i] - 1000 << " " << beamEff << std::endl;
        (*m_lout) << end_beam_vector[i] << " " << noBeam << std::endl;
      }
    }
  }
  if (evaluationMode == 0) {
    long long currenttime = starttimeG + interval_mksec;
    while (currenttime <= endtimeG) {
      OWLTime current = OWLTime(currenttime/1000000);
      endtime  = currenttime;
      starttime = endtime - duration_mksec;
      double beamEff = getEfficiency (starttime, endtime);
      std::cout << current.c_str() << "  " << beamEff << std::endl;
      if (displayMode == 1) {
        (*m_lout) << currenttime/1000 << " " << beamEff << std::endl;
      } else {
        int ended = 0;
        int started = 0;
        double noBeam = 0.0;
        if (noBeamValue == 1) {
          noBeam = -0.2;
        }
        if ((start_beam_vector.size() > 0) && (beamEff > noBeam)) {
          for (unsigned int i = 0; i < start_beam_vector.size(); i++ ) {
            if ((start_beam_vector[i] >= starttime/1000) && (start_beam_vector[i] < endtime/1000) && (end_beam_vector[i] > endtime/1000)) {
              (*m_lout) << start_beam_vector[i] << " " << noBeam << std::endl;
              (*m_lout) << start_beam_vector[i] + 1000 << " " << beamEff << std::endl;
              (*m_lout) << endtime/1000 << " " << beamEff << std::endl;
              started = 1;
            }
            if ((start_beam_vector[i] >= starttime/1000) && (end_beam_vector[i] <= endtime/1000)) {
              (*m_lout) << start_beam_vector[i] << " " << noBeam << std::endl;
              (*m_lout) << start_beam_vector[i] + 1000 << " " << beamEff << std::endl;
              (*m_lout) << end_beam_vector[i] - 1000 << " " << beamEff << std::endl;
              (*m_lout) << end_beam_vector[i] << " " << noBeam << std::endl;
              ended = 1;
            }
            if ((start_beam_vector[i] < starttime/1000) && (end_beam_vector[i] > starttime/1000) && (end_beam_vector[i] <= endtime/1000)) {
              (*m_lout) << starttime/1000 + 1000 << " " << beamEff << std::endl;
              (*m_lout) << end_beam_vector[i] - 1000 << " " << beamEff << std::endl;
              (*m_lout) << end_beam_vector[i] << " " << noBeam << std::endl;
              ended = 1;
            }
          }
        }
        if ((ended == 0) && (started == 0)) {
          (*m_lout) << starttime/1000 + 1000 << " " << beamEff << std::endl;
          (*m_lout) << endtime/1000 << " " << beamEff << std::endl;
        }
      }
      currenttime = currenttime + interval_mksec;
    }
  }
  if ((evaluationMode == 3) || (evaluationMode == 5)) {
    // prepare data files
    std::ostream * eff_out = new std::ofstream ("/tmp/runeff.data", std::ios::out | std::ios::trunc);
    if (!(*eff_out)) {
      std::cout << "Can't create the efficiency data file " << std::endl;
    }
    if (evaluationMode == 5) {
      // remove last entriy, not useful
      runeff_data.pop_back();
      // insert old data
      unsigned int n = runeff_data.size();
      if (n > 0) {
        for (unsigned int i = 0; i < n; i++ ) {
          (*eff_out) << runeff_data[i] << std::endl;
        }
      }
    }

    int initialReadyForPhysics = readyForPhysics;
    int globalBeamTime = 0;
    int globalRunningTime = 0;
    int globalRunningReadyTime = 0;
    int globalBeamRunningTime = 0;
    int globalBeamRunningReadyTime = 0;
    int globalNoDataInBeamTime = 0;
    int globalNoDataInBeamReadyTime = 0;
    double runEff = 0;
    double runEffReady = 0;
    double noBeam = 0.0;
    double noData = 0.0;
    if (noBeamValue == 1) {
      noBeam = -0.2;
    }
    double beamTime = 0;
    long long currenttime = starttimeG + interval_mksec;
    if (evaluationMode == 3) {
      (*eff_out) << starttimeG/1000-1000 << " " << noBeam << " " << noData << " " << noData << std::endl;
    }
    while (currenttime <= endtimeG) {
      OWLTime current = OWLTime(currenttime/1000000);
      endtime  = currenttime;
      starttime = endtime - duration_mksec;
      readyForPhysics = 0;
      runEff = getEfficiency (starttime, endtime);
      globalBeamTime = globalBeamTime + totalBeamTime;
      globalBeamRunningTime = globalBeamRunningTime + totalInBeamTime;
      globalNoDataInBeamTime = globalNoDataInBeamTime + totalNoDataInBeamTime;
      globalRunningTime = globalRunningTime + realInBeamTime;
      beamTime = (double)totalBeamTime / 3600;
      if (initialReadyForPhysics == 0) {
        std::cout << current.c_str() << "  " << runEff << std::endl;
        if (displayMode == 1) {
          (*m_lout) << currenttime/1000 << " " << runEff << std::endl;
        } else {
          int ended = 0;
          int started = 0;
          if (noBeamValue == 1) {
            noBeam = -0.2;
          }
          if ((start_beam_vector.size() > 0) && (runEff > noBeam)) {
            for (unsigned int i = 0; i < start_beam_vector.size(); i++ ) {
              if ((start_beam_vector[i] >= starttime/1000) && (start_beam_vector[i] < endtime/1000) && (end_beam_vector[i] > endtime/1000)) {
                (*m_lout) << start_beam_vector[i] << " " << noBeam << std::endl;
                (*m_lout) << start_beam_vector[i] + 1000 << " " << runEff << std::endl;
                (*m_lout) << endtime/1000 << " " << runEff << std::endl;
                started = 1;
              }
              if ((start_beam_vector[i] >= starttime/1000) && (end_beam_vector[i] <= endtime/1000)) {
                (*m_lout) << start_beam_vector[i] << " " << noBeam << std::endl;
                (*m_lout) << start_beam_vector[i] + 1000 << " " << runEff << std::endl;
                (*m_lout) << end_beam_vector[i] - 1000 << " " << runEff << std::endl;
                (*m_lout) << end_beam_vector[i] << " " << noBeam << std::endl;
                ended = 1;
              }
              if ((start_beam_vector[i] < starttime/1000) && (end_beam_vector[i] > starttime/1000) && (end_beam_vector[i] <= endtime/1000)) {
                (*m_lout) << starttime/1000 + 1000 << " " << runEff << std::endl;
                (*m_lout) << end_beam_vector[i] - 1000 << " " << runEff << std::endl;
                (*m_lout) << end_beam_vector[i] << " " << noBeam << std::endl;
                ended = 1;
              }
            }
          }
          if ((ended == 0) && (started == 0)) {
            (*m_lout) << starttime/1000 + 1000 << " " << runEff << std::endl;
            (*m_lout) << endtime/1000 << " " << runEff << std::endl;
          }
        }
      }
      readyForPhysics = 1;
      runEffReady = getEfficiency (starttime, endtime);
      globalBeamRunningReadyTime = globalBeamRunningReadyTime + totalInBeamTime;
      globalNoDataInBeamReadyTime = globalNoDataInBeamReadyTime + totalNoDataInBeamTime;
      globalRunningReadyTime = globalRunningReadyTime + realInBeamTime;
      if (initialReadyForPhysics == 1) {
        std::cout << current.c_str() << "  " << runEffReady << std::endl;
        if (displayMode == 1) {
          (*m_lout) << currenttime/1000 << " " << runEffReady << std::endl;
        } else {
          int ended = 0;
          int started = 0;
          if (noBeamValue == 1) {
            noBeam = -0.2;
          }
          if ((start_beam_vector.size() > 0) && (runEffReady > noBeam)) {
            for (unsigned int i = 0; i < start_beam_vector.size(); i++ ) {
              if ((start_beam_vector[i] >= starttime/1000) && (start_beam_vector[i] < endtime/1000) && (end_beam_vector[i] > endtime/1000)) {
                (*m_lout) << start_beam_vector[i] << " " << noBeam << std::endl;
                (*m_lout) << start_beam_vector[i] + 1000 << " " << runEffReady << std::endl;
                (*m_lout) << endtime/1000 << " " << runEffReady << std::endl;
                started = 1;
              }
              if ((start_beam_vector[i] >= starttime/1000) && (end_beam_vector[i] <= endtime/1000)) {
                (*m_lout) << start_beam_vector[i] << " " << noBeam << std::endl;
                (*m_lout) << start_beam_vector[i] + 1000 << " " << runEffReady << std::endl;
                (*m_lout) << end_beam_vector[i] - 1000 << " " << runEff << std::endl;
                (*m_lout) << end_beam_vector[i] << " " << noBeam << std::endl;
                ended = 1;
              }
              if ((start_beam_vector[i] < starttime/1000) && (end_beam_vector[i] > starttime/1000) && (end_beam_vector[i] <= endtime/1000)) {
                (*m_lout) << starttime/1000 + 1000 << " " << runEffReady << std::endl;
                (*m_lout) << end_beam_vector[i] - 1000 << " " << runEffReady << std::endl;
                (*m_lout) << end_beam_vector[i] << " " << noBeam << std::endl;
                ended = 1;
              }
            }
          }
          if ((ended == 0) && (started == 0)) {
            (*m_lout) << starttime/1000 + 1000 << " " << runEffReady << std::endl;
            (*m_lout) << endtime/1000 << " " << runEffReady << std::endl;
          }
        }
      }
      (*eff_out) << starttime/1000 << " " << beamTime << " " << runEff << " " << runEffReady << std::endl;
      (*eff_out) << endtime/1000 << " " << beamTime << " " << runEff << " " << runEffReady << std::endl;
      currenttime = currenttime + interval_mksec;
    }
    (*eff_out) << endtime/1000+1000 << " " << noBeam << " " << noData << " " << noData << std::endl; 
    (*eff_out) << "enddata" << std::endl;
    integralBeamTime = globalBeamTime;
    integralBeamRunningTime = globalBeamRunningTime;
    integralDeadTime = globalNoDataInBeamTime;
    integralRunningTime = globalRunningTime;
    integralRunEff = (double) globalRunningTime / (double) globalBeamTime;
    integralBeamRunningReadyTime = globalBeamRunningReadyTime;
    integralReadyDeadTime = globalNoDataInBeamReadyTime;
    integralRunningReadyTime = globalRunningReadyTime;
    integralRunEffReady = (double) globalRunningReadyTime / (double) globalBeamTime;
    readyForPhysics = initialReadyForPhysics;
    delete eff_out;
  }
  (*m_lout) << "end" << std::endl; 

  std::cout << std::endl;
  double beamEff = 0;
  if ((evaluationMode == 0) || (evaluationMode == 1)) {
    beamEff = getEfficiency (starttimeG, endtimeG);
    setRunVectors (starttimeG, endtimeG);

    if (start_of_run_vector.size() > 0) {
      for (unsigned int i = 0; i < start_of_run_vector.size(); i++ ) {
        (*m_lout) << start_of_run_vector[i] << " " << end_of_run_vector[i] << std::endl;
      }
    }
  }
  (*m_lout) << "finalend" << std::endl;

  // print summary data

  if ((evaluationMode == 0) || (evaluationMode == 1)) {
    std::cout << "Total beam time (sec)        = " << totalBeamTime << std::endl;
    std::cout << "Beam running time (sec)      = " << totalInBeamTime << std::endl;
    std::cout << "Dead time (sec)              = " << totalNoDataInBeamTime << std::endl;
    std::cout << "Real beam running time (sec) = " << realInBeamTime << std::endl;
    if (totalBeamTime > 0) {
      std::cout << "Data taking efficiency       = " << beamEff << std::endl;
    }
  }
  if ((evaluationMode == 2) || (evaluationMode == 3) || (evaluationMode == 4) || (evaluationMode == 5)) {
    std::cout << "Total beam time (sec)                     = " << integralBeamTime << std::endl;
    std::cout << "Beam running time (sec)                   = " << integralBeamRunningTime << std::endl;
    std::cout << "Dead time (sec)                           = " << integralDeadTime << std::endl;
    std::cout << "Ready beam running time (sec)             = " << integralBeamRunningReadyTime << std::endl;
    std::cout << "Ready dead time (sec)                     = " << integralReadyDeadTime << std::endl;
    std::cout << "Real beam running time (sec)              = " << integralRunningTime << std::endl;
    std::cout << "Real ready beam running time (sec)        = " << integralRunningReadyTime << std::endl; 
    double integralBeamTimeHours = (double) integralBeamTime / 3600;
    double integralRunningTimeHours = (double) integralRunningTime / 3600;
    double integralRunningReadyTimeHours = (double) integralRunningReadyTime / 3600;
    double integralBeamNoWarmStopTimeHours = (double) integralBeamNoWarmStopTime / 3600;
    double integralRunningReadyNoWarmStopTimeHours = (double) integralRunningReadyNoWarmStopTime / 3600;
    std::cout << "Stable beams (hours)                      = " << integralBeamTimeHours << std::endl;
    std::cout << "Running and not busy (hours)              = " << integralRunningTimeHours << std::endl;
    std::cout << "Running with RfP flag (hours)             = " << integralRunningReadyTimeHours << std::endl;
    if (integralBeamTime > 0) {
      std::cout << "Data taking efficiency (without RfP flag) = " << integralRunEff << std::endl;
      std::cout << "Data taking efficiency (with RfP flag)    = " << integralRunEffReady << std::endl;
      if (excludeWarmStop == 1) {
        std::cout << "Stable beams until Warm Stop (hours)      = " << integralBeamNoWarmStopTimeHours << std::endl;
        std::cout << "Data taking efficiency (with RfP flag) excluding WarmStop = " << integralRunEffReadyNoWarmStop << std::endl; 
      }
    }
    // prepare summary file
    std::ostream * summary_out = new std::ofstream ("/tmp/summary.data", std::ios::out | std::ios::trunc);
    if (!(*summary_out)) {
      std::cout << "Can't create the summary data file " << std::endl;
    }
    if (integralBeamTimeHours >= 1000) {
      (*summary_out).precision(5);
    } else {
      (*summary_out).precision(4);
    }
    (*summary_out) << integralBeamTimeHours << std::endl;
    if (integralRunningTimeHours >= 1000) {
      (*summary_out).precision(5);
    } else {
      (*summary_out).precision(4);
    }
    (*summary_out) << integralRunningTimeHours << std::endl;
    if (integralRunningReadyTimeHours >= 1000) {
      (*summary_out).precision(5);
    } else {
      (*summary_out).precision(4);
    }
    (*summary_out) << integralRunningReadyTimeHours << std::endl;
    double integralRunEffOut = integralRunEff*100;
    double integralRunEffReadyOut = integralRunEffReady*100;
    (*summary_out).precision(3);
    (*summary_out) << integralRunEffOut << std::endl;
    (*summary_out) << integralRunEffReadyOut << std::endl;
    if (excludeWarmStop == 1) {
      if (integralBeamNoWarmStopTimeHours >= 1000) {
        (*summary_out).precision(5);
      } else {
        (*summary_out).precision(4);
      }
      (*summary_out) << integralBeamNoWarmStopTimeHours << std::endl;
      if (integralRunningReadyNoWarmStopTimeHours >= 1000) {
        (*summary_out).precision(5);
      } else {
        (*summary_out).precision(4);
      }
      (*summary_out) << integralRunningReadyNoWarmStopTimeHours << std::endl;
      double integralRunEffReadyNoWarmStopOut = integralRunEffReadyNoWarmStop*100;
      (*summary_out).precision(3);
      (*summary_out) << integralRunEffReadyNoWarmStopOut << std::endl;
    }
    delete summary_out;
  }

  std::cout << std::endl;

  (*m_lout) << "Run efficiency from " << startTime << " to " << endTime << std::endl;

  if (beamType == 0) {
    (*m_lout) << "Beam type : stable beams" << std::endl;
  } 
  if (beamType == 1) {
    (*m_lout) << "Beam type : two beams present" << std::endl;
  }
  if (beamType == 2) {
    (*m_lout) << "Beam type : at least one beam present" << std::endl;
  } 
  if (beamType == 3) {
    (*m_lout) << "Beam type : two beams present and stable" << std::endl;
  }
  if (beamType == 4) {
    (*m_lout) << "Beam data from file" << std::endl;
  }
  if ((evaluationMode == 0) || (evaluationMode == 1)) {
    (*m_lout) << "Total beam time (sec) = " << totalBeamTime << std::endl;
    (*m_lout) << "Running time in bean (sec) = " << totalInBeamTime << std::endl;
    (*m_lout) << "Dead time in beam (sec) = " << totalNoDataInBeamTime << std::endl;
    if (totalBeamTime > 0) {
      (*m_lout) << "(RunningBeamTime - DeadTime)/BeamTime = " << beamEff << std::endl;
    } else {
      (*m_lout) << "                           " << std::endl;
    }
  }

  if (((evaluationMode == 2) || (evaluationMode == 3) || (evaluationMode == 4) || (evaluationMode == 5)) && (readyForPhysics == 0)) {
    (*m_lout) << "Total beam time (sec) = " << integralBeamTime << std::endl;
    (*m_lout) << "Running time in bean (sec) = " << integralBeamRunningTime << std::endl;
    (*m_lout) << "Dead time in beam (sec) = " << integralDeadTime << std::endl;
    if (integralBeamTime > 0) {
      (*m_lout) << "(RunningBeamTime - DeadTime)/BeamTime = " << integralRunEff << std::endl;
    } else {
      (*m_lout) << "                           " << std::endl;
    }
  }
  if (((evaluationMode == 2) || (evaluationMode == 3) || (evaluationMode == 4) || (evaluationMode == 5)) && (readyForPhysics == 1)) {
    (*m_lout) << "Total beam time (sec) = " << integralBeamTime << std::endl;
    (*m_lout) << "Running time in bean (sec) = " << integralBeamRunningReadyTime << std::endl;
    (*m_lout) << "Dead time in beam (sec) = " << integralReadyDeadTime << std::endl;
    if (integralBeamTime > 0) {
      (*m_lout) << "(RunningBeamTime - DeadTime)/BeamTime = " << integralRunEffReady << std::endl;
    } else {
      (*m_lout) << "                           " << std::endl;
    }
  }
  delete m_lout;
  // Close the DB
  db->closeDatabase();
  dbt->closeDatabase();
  dbd->closeDatabase();
  std::cout << "DB closed." << std::endl;

  if (display == 1) {
    if ((evaluationMode == 0) || (evaluationMode == 1)) { 
      system("eff_display -o 2");
    } else {
      system("eff_display -o 1");
    }
  }
  return EXIT_SUCCESS;
}

