//   get_running_time.cc
//
//   Application to evaluate run efficiency
//   Run and beam informations are read from COOL database
//   
//   Author: Mihai.Caprini@cern.ch, March 2009
//   Modifications:
//
//   Author  Date            Reason
//
//   MC      15/04/2009      Add beam information from a file
//   MC      22/04/2009      New options for command line 
//   MC      04/05/2009      Count correctly the running time during the beam
//   MC      16/05/2009      Skip runs with a very long active time
//   MC      05/06/2009      Process correctly a run overlapping several
//                           beam intervals
//   MC      17/09/2009      Add options: -r (check lumi block trigger rate)
//                                        -v (display low rate lumi blocks) 
//                                        -o (flat or lumi block rate display)
//                           Add mean start up time
//   MC      24/09/2009      Replace the use of TotalTime by (EOR - SOR)
//   MC      01/10/2009      Start search from SOR_Params folder (to find the
//                           runs without EOR_Params)
//   MC      04/11/2009      For mean start up time consider the run started 
//                           just before the beam
//   MC      10/11/2009      Check online if the run is ongoing
//   MC      08/12/2009      Add dead time from /TRIGGER/LVL1/BUSYRATES folder
//   MC      09/12/209       Add options: -b (beam duration low limit)
//   MC      26/01/2010      Check in the Run Number DB if the run is ongoing
//   MC      27/01/2010      For runs without EOR, get duration from RN DB
//   MC      28/01/2010      Add busy configuration from /TRIGGER/LVL1/BUSYCONF
//   MC      29/10/2010      Add detector busy from /TRIGGER/LVL1/BUSYRATES
//   MC      29/01/2010      Remove DB name and folder from command line
//   MC      04/02/2010      Read beam stable flag from /LHC/DCS/FILLSTATE
//   MC      08/02/2010      Add options: -f (read beam data from file)
//                                        -m (use offline or online databases)
//   MC      10/02/2010      Replace 1 hour by 15 minutes as minimum run time
//   MC      14/02/2010      Examine only run numbers > 140000 (18-Nov-2009)
//                           Add options: -l (lumi block busy limit)
//                                        -d (detector busy limit) 
//   MC      22/02/2010      Read beam data from /LHC/DCS/BPTX/TOTALINT
//                           Add option:  -t (beam type)
//   MC      10/03/2010      Remove option: -f (read beam data from file)
//                           For -t option add values: 
//                                        3 - beams present and stable
//                                        4 - beam data from file
//                           Use only vectors, no intermediate files
//   MC      11/03/2010      Add option: -w (write beam file)
//                           Remove check online if the run is ongoing
//   MC      17/03/2010      Add option: -x (start graphical interface)
//   MC      23/03/2010      Add option: -n (no ramps in beam time)
//   MC      12/04/2010      Add options: -c (trigger item)
//                                        -l (busy limit)
//                                        -f (ready for physics)
//   MC      15/04/2010      Partition name is optional (default ATLAS)
//   MC      03/05/2010      Replace L1_MBTS_1_col by L1_MBTS_2
//   MC      10/05/2010      Use veto_fraction instead of busy_fraction
//   MC      15/02/2011      Use the same algorithm as in get_efficiency
//                           Find first run from COOL data (LBT_Folder)
//   MC      01/08/2011      Replace L1_MBTS_2 by L1_EM14
//   MC      03/04/2012      Replace L1_EM14 by L1_EM30
//   MC      10/02/2015      Comment owleor (not used)
//   MC      18/02/2015      Use new cond db for Run 2 (CONDBR2)
//   MC      18/02/2015      Add option:  -r (atlasRun)
//   MC      20/02/2015      Remove option: -m (runningOffline)
//

#include <iostream>
#include <sstream>
#include <stdlib.h> // for getenv
#include <signal.h>
#include <fstream>
#include <vector>

#include <boost/program_options.hpp>

#include <CoralBase/Attribute.h>
#include <CoolKernel/Exception.h>
#include <CoolKernel/ValidityKey.h>
#include <CoolKernel/IDatabaseSvc.h>
#include <CoolKernel/IDatabase.h>
#include <CoolKernel/IFolder.h>
#include <CoolKernel/IRecordSpecification.h>
#include <CoolKernel/Record.h>
#include <CoolApplication/DatabaseSvcFactory.h>
#include <CoolKernel/IObjectIterator.h>
#include <CoolKernel/IObject.h>
#include <CoolKernel/ChannelSelection.h>

#include "ers/ers.h"
#include "owl/time.h"

#include "pmg/pmg_initSync.h"

#include "RunControl/Common/Exceptions.h"

namespace daq {
ERS_DECLARE_ISSUE(	rc,
			BadCoolDb,
			"Retrieving data from COOL failed because " << explanation,
                        ((const char*) explanation)
		)
}

static     cool::IDatabasePtr db;
static     cool::IDatabasePtr dbt;
static     cool::IDatabasePtr dbm;
static     cool::IDatabasePtr dbd;

static     std::string EOR_Name = "/TDAQ/RunCtrl/EOR_Params";
static     std::string SOR_Name = "/TDAQ/RunCtrl/SOR_Params";
static     std::string EC_Name = "/TDAQ/RunCtrl/EventCounters";
static     std::string CTP_Name = "/TRIGGER/LVL1/CTPCORELBDATA";
static     std::string LB_Name = "/TRIGGER/LUMI/LBLB";
static     std::string LBT_Name = "/TRIGGER/LUMI/LBTIME";
static     std::string BusyRate_Name = "/TRIGGER/LVL1/BUSYRATES";
static     std::string BusyConf_Name = "/TRIGGER/LVL1/BUSYCONF";
static     std::string FillState_Name = "/LHC/DCS/FILLSTATE";
static     std::string TotalInt_Name = "/LHC/DCS/BPTX/TOTALINT";
static     std::string L1Counters_Name = "/TRIGGER/LUMI/LVL1COUNTERS";
static     std::string L1Menu_Name = "/TRIGGER/LVL1/Menu";
static     std::string DataTakingMode_Name = "/TDAQ/RunCtrl/DataTakingMode";

static     std::map<std::string, std::string> ctpMap;

static     cool::IFolderPtr EOR_Folder;
static     cool::IFolderPtr SOR_Folder;
static     cool::IFolderPtr EC_Folder;
static     cool::IFolderPtr CTP_Folder;
static     cool::IFolderPtr BusyRate_Folder;
static     cool::IFolderPtr BusyConf_Folder;
static     cool::IFolderPtr FillState_Folder;
static     cool::IFolderPtr TotalInt_Folder;
static     cool::IFolderPtr LB_Folder;
static     cool::IFolderPtr LBT_Folder;
static     cool::IFolderPtr L1Counters_Folder;
static     cool::IFolderPtr L1Menu_Folder;
static     cool::IFolderPtr DataTakingMode_Folder;

static     std::string message = "";
static     bool no_busy_info = false;

static     int atlasRun = 2;
static     int beamType = 3;
static     int beamLimit = 5;
static     int writeBeamFile = 0;
static     int display = 0;
static     int noRamp = 0;
static     int readyForPhysics = 0;

static     std::vector<long long> start_beam_vector;
static     std::vector<long long> end_beam_vector;
static     std::vector<int> delay_vector;
static     std::vector<int> run_number_vector;


static     std::vector<long long> beam1_start_vector;
static     std::vector<long long> beam2_start_vector;
static     std::vector<long long> beam1_end_vector;
static     std::vector<long long> beam2_end_vector;
static     std::vector<long long> beams_start_vector;
static     std::vector<long long> beams_end_vector;
static     std::vector<long long> beam1and2_start_vector;
static     std::vector<long long> beam1and2_end_vector;
static     std::vector<long long> beam1or2_start_vector;
static     std::vector<long long> beam1or2_end_vector;
static     std::vector<long long> beam1and2andStable_start_vector;
static     std::vector<long long> beam1and2andStable_end_vector;
static     std::vector<long long> beam1or2noRamp_start_vector;
static     std::vector<long long> beam1or2noRamp_end_vector;
static     std::vector<long long> beam1and2noRamp_start_vector;
static     std::vector<long long> beam1and2noRamp_end_vector;
static     std::vector<long long> ramp_start_vector;
static     std::vector<long long> ramp_end_vector;

extern "C" void signalhandler(int)
{
   signal(SIGINT,  SIG_DFL);
   signal(SIGTERM, SIG_DFL);
   std::cout << "Terminating rc_get_running_time ..." << std::endl;
}

// throws daq::rc::BadCoolDb
void openDB(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

  try {
    db = db_Svc.openDatabase(db_Id, false); // false = !read_only
  }
  catch (std::exception &ex) {
    std::cout << "Failure opening DB database - " << ex.what() << std::endl;
    daq::rc::BadCoolDb i(ERS_HERE, ex.what());
    throw i;
  }
}

// throws daq::rc::BadCoolDb
void openDBT(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

  try {
    dbt = db_Svc.openDatabase(db_Id, false); // false = !read_only
  }
  catch (std::exception &ex) {
    std::cout << "Failure opening DBT database - " << ex.what() << std::endl;
    daq::rc::BadCoolDb i(ERS_HERE, ex.what());
    throw i;
  }
}

// throws daq::rc::BadCoolDb
void openDBM(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

  try {
    dbm = db_Svc.openDatabase(db_Id, false); // false = !read_only
  } catch (std::exception &ex) {
    std::cout << "Failure opening DBM database - " << ex.what() << std::endl;
    daq::rc::BadCoolDb i(ERS_HERE, ex.what());
    throw i;
  }
}

// throws daq::rc::BadCoolDb
void openDBD(const cool::DatabaseId& db_Id, const cool::IDatabaseSvc& db_Svc) {

  try {
    dbd = db_Svc.openDatabase(db_Id, false); // false = !read_only
  } catch (std::exception &ex) {
    std::cout << "Failure opening DCS database - " << ex.what() << std::endl;
    daq::rc::BadCoolDb i(ERS_HERE, ex.what());
    throw i;
  }
}


int valInBeam(long long startBlock, long long endBlock, long long startBeam, long long endBeam) {
  int result = 0;
  if ((startBlock <= startBeam) && (endBlock <= startBeam)) {
    // block before beam
    return result;
  }
  if ((startBlock <= startBeam) && (endBlock > startBeam) && (endBlock <= endBeam)) {
    result = endBlock - startBeam;
    return result;
  }
  if ((startBlock <= startBeam) && (endBlock >= endBeam)) {
    result = endBeam - startBeam;
    return result;
  }
  if ((startBlock >= startBeam) && (endBlock <= endBeam)) {
    result = endBlock - startBlock;
    return result;
  }
  if ((startBlock > startBeam) && (startBlock < endBeam) && (endBlock >= endBeam)) {
    result = endBeam - startBlock;
    return result;
  }
  if ((startBlock >= endBeam) && (endBlock > endBeam)) {
    // block after beam
    return result;
  }
  return result;
}

void get_ctp_map(int runNumber, cool::IFolderPtr & the_conf_folder) {

  // Populate ctp map for this run
  
  cool::UInt63 a = runNumber;
  cool::UInt63 z = runNumber+1;
  cool::ValidityKey sKey = (a <<32) + 1;
  cool::ValidityKey uKey = z <<32;;

  cool::IObjectIteratorPtr objects = the_conf_folder->browseObjects(sKey, uKey, cool::ChannelSelection::all());
  while ( objects->goToNext() ) {
    const cool::IObject& obj = objects->currentRef();
    std::string identifier = obj.payloadValue("identifier");
    std::string description = obj.payloadValue("description");
    std::string busyEnabled = obj.payloadValue("busyEnabled");
    if (busyEnabled.compare("TRUE") == 0) {
      ctpMap.insert(std::pair<std::string, std::string>(identifier, description) );
    }
  }
}

void get_beam1_data(long long startTime, long long endTime, cool::IFolderPtr & folder) {
  beam1_start_vector.clear();
  beam1_end_vector.clear();
  cool::IObjectIteratorPtr objects = folder->browseObjects( cool::ValidityKey(startTime),cool::ValidityKey(endTime), cool::ChannelSelection::all());
  std::string old_beam = "FALSE";
  bool beamStarted = false;
  while ( objects->goToNext() ) {
    const cool::IObject& obj = objects->currentRef();
    std::string beam = obj.payloadValue("Beam1_BeamPresent");
    if (((beam.compare("TRUE") == 0) && (old_beam.compare("FALSE") == 0)) ||
        ((beam.compare("TRUE") == 0) && (old_beam.compare("NULL") == 0))) {
      // start of a new beam interval
      long long start_beam = (long long) (obj.since());
      if (start_beam < startTime) {
        start_beam = startTime;
      }
      beam1_start_vector.push_back(start_beam/1000000);
      old_beam = beam;
      beamStarted = true;
    } else if ((beam.compare("FALSE") == 0) && (old_beam.compare("TRUE") == 0)) {
      // end of a beam interval
      long long end_beam  = (long long) (obj.since());
      beam1_end_vector.push_back(end_beam/1000000);
      beamStarted = false;
      old_beam = beam;
    } else {
      old_beam = beam;
    }
  }
  if (beamStarted) {
    beam1_end_vector.push_back(endTime/1000000);
  }
}

void get_beam2_data(long long startTime, long long endTime, cool::IFolderPtr & folder) {
  beam2_start_vector.clear();
  beam2_end_vector.clear();
  cool::IObjectIteratorPtr objects = folder->browseObjects( cool::ValidityKey(startTime),cool::ValidityKey(endTime), cool::ChannelSelection::all());
  std::string old_beam = "FALSE";
  bool beamStarted = false;
  while ( objects->goToNext() ) {
    const cool::IObject& obj = objects->currentRef();
    std::string beam = obj.payloadValue("Beam2_BeamPresent");
    if (((beam.compare("TRUE") == 0) && (old_beam.compare("FALSE") == 0)) ||
        ((beam.compare("TRUE") == 0) && (old_beam.compare("NULL") == 0))) {
      // start of a new beam interval
      long long start_beam = (long long) (obj.since());
      if (start_beam < startTime) {
        start_beam = startTime;
      }
      beam2_start_vector.push_back(start_beam/1000000);
      old_beam = beam;
      beamStarted = true;
    } else if ((beam.compare("FALSE") == 0) && (old_beam.compare("TRUE") == 0)) {
      // end of a beam interval
      long long end_beam  = (long long) (obj.since());
      beam2_end_vector.push_back(end_beam/1000000);
      beamStarted = false;
      old_beam = beam;
    } else {
      old_beam = beam;
    }
  }
  if (beamStarted) {
    beam2_end_vector.push_back(endTime/1000000);
  }
}

void get_beams_data(long long startTime, long long endTime, cool::IFolderPtr & folder) {
  beams_start_vector.clear();
  beams_end_vector.clear();
  cool::IObjectIteratorPtr objects = folder->browseObjects( cool::ValidityKey(startTime),cool::ValidityKey(endTime), cool::ChannelSelection::all());
  std::string old_beam = "FALSE";   bool beamStarted = false;
  while ( objects->goToNext() ) {
    const cool::IObject& obj = objects->currentRef();
    std::string beam = obj.payloadValue("StableBeams");
    if (((beam.compare("TRUE") == 0) && (old_beam.compare("FALSE") == 0)) ||
        ((beam.compare("TRUE") == 0) && (old_beam.compare("NULL") == 0))) {
      // start of a new beam interval
      long long start_beam = (long long) (obj.since());
      if (start_beam < startTime) {
        start_beam = startTime;
      }
      beams_start_vector.push_back(start_beam/1000000);
      old_beam = beam;
      beamStarted = true;
    } else if ((beam.compare("FALSE") == 0) && (old_beam.compare("TRUE") == 0)) {
      // end of a beam interval
      long long end_beam  = (long long) (obj.since());
      beams_end_vector.push_back(end_beam/1000000);
      beamStarted = false;
      old_beam = beam;
    } else {
      old_beam = beam;
    }
  }
  if (beamStarted) {
    beams_end_vector.push_back(endTime/1000000);
  }
}

void get_ramp_data(long long startTime, long long endTime, cool::IFolderPtr & folder) {
  // in fact the vectors contain NO RAMP UP intervals
  ramp_start_vector.clear();
  ramp_end_vector.clear();
  cool::IObjectIteratorPtr objects = folder->browseObjects( cool::ValidityKey(startTime),cool::ValidityKey(endTime), cool::ChannelSelection::all());
  std::string old_beam_mode = "RAMP";
  bool rampStarted = false;
  while ( objects->goToNext() ) {
    const cool::IObject& obj = objects->currentRef();
    std::string beam_mode = obj.payloadValue("BeamMode");
    // use the same conditions as Run Control to hold triggers
    // other conditions could be derived from the LHC state machine:
    //    INJECTION PROBE BEAM (?) -> PREPARE RAMP -> RAMP
    //    PREPARE RAMP -> BEAM DUMP
    //    RAMP -> BEAM DUMP
    //    RAMP -> FLAT TOP
    if (((beam_mode.compare("PREPARE RAMP") != 0) && (beam_mode.compare("RAMP") != 0)) && ((old_beam_mode.compare("PREPARE RAMP") == 0) || (old_beam_mode.compare("RAMP") == 0))) {
      // start of a new no ramp interval
      long long start_ramp = (long long) (obj.since());
      if (start_ramp < startTime) {
        start_ramp = startTime;
      }
      ramp_start_vector.push_back(start_ramp/1000000);
      old_beam_mode = beam_mode;
      rampStarted = true;
    } else if (((beam_mode.compare("PREPARE RAMP") == 0) || (beam_mode.compare("RAMP") == 0)) && ((old_beam_mode.compare("PREPARE RAMP") != 0) && (old_beam_mode.compare("RAMP") != 0))) {
      // end of a no ramp interval
      long long end_ramp  = (long long) (obj.since());
      ramp_end_vector.push_back(end_ramp/1000000);
      rampStarted = false;
      old_beam_mode = beam_mode;
    } else {
      old_beam_mode = beam_mode;
    }
  }
  if (rampStarted) {
    ramp_end_vector.push_back(endTime/1000000);
  }
}

void get_beam_data_from_file(long long starttime, long long endtime) {
  std::filebuf fb;
  fb.open ("beam.log",std::ios::in);
  if (fb.is_open()) {
    std::istream is(&fb);
    std::string str;
    long long bLimit = ((long long) beamLimit) * ((long long) 1000000);
    while (!(str.compare("endbeam") == 0)) {
      getline(is, str);
      if (!(str.compare("endbeam") == 0)) {
        OWLTime startBeam(str.substr(0, 17).c_str());
        long long startbeam = startBeam.total_mksec_utc();
        OWLTime endBeam(str.substr(19, 17).c_str());
        long long endbeam = endBeam.total_mksec_utc();
        if ((endbeam - startbeam) < bLimit) {
//           std::cout << "short beam" << std::endl;
           continue;
        }
        if ((startbeam >= starttime) && (endbeam <= endtime)) {
          start_beam_vector.push_back(startbeam/1000);
          end_beam_vector.push_back(endbeam/1000);
        }
        if ((startbeam < starttime) && (endbeam > starttime) && (endbeam <= endtime)) {
          start_beam_vector.push_back(starttime/1000);
          end_beam_vector.push_back(endbeam/1000);
        }
        if ((startbeam < starttime) && (endbeam > endtime)) {
          start_beam_vector.push_back(starttime/1000);
          end_beam_vector.push_back(endtime/1000);
        }
        if ((startbeam >= starttime) && (startbeam < endtime) && (endbeam > endtime)) {
          start_beam_vector.push_back(startbeam/1000);
          end_beam_vector.push_back(endtime/1000);
        }
      }
    }
  } else {
    std::cout << "Warning: cannot open the beam data file" << std::endl;
  }
  fb.close();
}

void print_vectors(std::vector<long long> start_vector, std::vector<long long> end_vector, std::string file_name) {
  std::ostream * m_lout = new std::ofstream (file_name.c_str(),std::ios::out | std::ios::trunc);
  if (!(*m_lout)) {
    std::cout << "Can't create the " << file_name << " log file " << std::endl;
    return;
  }
  std::ostringstream strm;
  if (start_vector.size() > 0) {
    for (unsigned int i = 0; i < start_vector.size(); i++ ) {
      long long start_beam_sec = start_vector[i]/1000;
      OWLTime start_owl = OWLTime(start_beam_sec);
      std::string start_string = std::string(start_owl.c_str());
      if (start_string.size() == 15) {
        strm << "  " << start_string;
      } else if (start_string.size() == 16) {
        strm << " " << start_string;
      } else {
        strm << start_string;
      }
      long long end_beam_sec = end_vector[i]/1000;
      OWLTime end_owl = OWLTime(end_beam_sec);
      std::string end_string = std::string(end_owl.c_str());
      strm << "  " << end_string << std::endl;
    }
    (*m_lout) << strm.str();
  }
  (*m_lout) << "endbeam" << std::endl;
  delete m_lout;
}

void set_main_vectors(std::vector<long long> start_vector, std::vector<long long> end_vector) {
  start_beam_vector.clear();
  end_beam_vector.clear();
  long long bLimit = ((long long) beamLimit) * ((long long) 1000);
  if (start_vector.size() > 0) {
    for (unsigned int i = 0; i < start_vector.size(); i++ ) {
      if ((end_vector[i] - start_vector[i])  < bLimit) {
//        std::cout << "short beam" << std::endl;
         continue;
      }
      start_beam_vector.push_back(start_vector[i]);
      end_beam_vector.push_back(end_vector[i]);
    }
  }
}

int get_total_beam_time() {
  int total_beam_time = 0;
  int valbeam = 0;
  if (start_beam_vector.size() > 0) {
    for (unsigned int i = 0; i < start_beam_vector.size(); i++ ) {
      valbeam = end_beam_vector[i]/1000 - start_beam_vector[i]/1000;
      total_beam_time = total_beam_time + valbeam;
      delay_vector.push_back(valbeam);
      run_number_vector.push_back(0);
    }
  }
  return total_beam_time;
}

int get_ctpid(cool::ValidityKey sKey, cool::ValidityKey uKey, std::string itemName, cool::IFolderPtr & lvl1menu_folder) {
  int itemNo = -1;
  cool::IObjectIteratorPtr objects = lvl1menu_folder->browseObjects(sKey, uKey, cool::ChannelSelection::all());
  while ( objects->goToNext() ) {
    const cool::IObject& obj = objects->currentRef();
    std::string item_name = obj.payloadValue("ItemName");
    if (item_name.compare(itemName) == 0) {
      itemNo = obj.channelId();
      break;
    }
  }
  return itemNo;
}

int ready_for_physics(cool::ValidityKey key, cool::IFolderPtr & datatakingmode_folder) {
  int flag = 1;
  cool::IObjectPtr obj = datatakingmode_folder->findObject(key, 0);
  std::string readyForPhysics = obj->payloadValue("ReadyForPhysics");
  flag = atoi(readyForPhysics.c_str());
  return flag;
}

double veto_fraction(cool::ValidityKey key, int ctpid, cool::IFolderPtr & l1counters_folder) {
  // for the moment overflow flags are not used
  cool::IObjectPtr obj_l1counters = l1counters_folder->findObject(key, ctpid);
  std::string l1Accept = obj_l1counters->payloadValue("L1Accept");
  std::string afterPrescale = obj_l1counters->payloadValue("AfterPrescale");
  long long l1_accept = atoll(l1Accept.c_str());
  long long after_prescale = atoll(afterPrescale.c_str());
  double fraction = 0;
  if (after_prescale > 0) {
    fraction = (double) (after_prescale - l1_accept) / (double) after_prescale;
  } 
  return fraction;
}
   
double fraction(long long startTime, long long endTime, cool::IFolderPtr & the_busy_folder ) {

  // Find all busy fraction values in this lumiblock:

  //Note: this also includes records that are partialy outside this lb

  message = "";
  no_busy_info = false;

  cool::IObjectIteratorPtr objects = the_busy_folder->browseObjects( cool::ValidityKey(startTime),cool::ValidityKey(endTime), 0 );
  double totaltime=0;
  double totalfraction=0;

  while ( objects->goToNext() ) {
    const cool::IObject& obj = objects->currentRef();
    
    std::string this_fraction_string= obj.payloadValue("ctpcore_moni0_rate");
    double this_fraction = atof(this_fraction_string.c_str());

    double this_starttime = obj.since();
    double this_stoptime = obj.until();
    totalfraction+=this_fraction*(this_stoptime-this_starttime);
    totaltime+=(this_stoptime-this_starttime);
//    std::cout << "This record: tstart: " <<  this_starttime ;
//    std::cout << " tstop: " <<  this_stoptime ;
//    std::cout << " fraction: " << this_fraction <<std::endl  ;

  }

  if(totaltime>0){
    totalfraction=0.01* totalfraction/totaltime;
  }else{
//    std::cout << " No busy fraction info available ... " << std::endl;
    message = " (No busy fraction info available ...)";
    no_busy_info = true;
    totalfraction = 1; // if no info, let's just say we are busy
//    totalfraction = 0; // temporary, in 2010 there is no fraction info
  }
//  std::cout << std::endl << " final fraction for this lb: " << totalfraction <<std::endl;

  return totalfraction;

}

double busy(long long startTime, long long endTime, std::string identifier, cool::IFolderPtr & the_busy_folder ) {

  // Find the busy values for a specific detector in this lumiblock:

  //Note: this also includes records that are partialy outside this lb

  cool::IObjectIteratorPtr objects = the_busy_folder->browseObjects( cool::ValidityKey(startTime),cool::ValidityKey(endTime), 0 );
  double totaltime=0;
  double totalbusy=0;

  while ( objects->goToNext() ) {
    const cool::IObject& obj = objects->currentRef();

    std::string this_busy_string= obj.payloadValue(identifier.c_str());
    double this_busy = atof(this_busy_string.c_str());

    double this_starttime = obj.since();
    double this_stoptime = obj.until();
    totalbusy+=this_busy*(this_stoptime-this_starttime);
    totaltime+=(this_stoptime-this_starttime);
  }

  if (totaltime>0) {
    totalbusy = 0.01* totalbusy/totaltime;
  } else {
//    std::cout << " No detector busy info available ... " << std::endl;
    totalbusy = 0; // if no info, let's just say this detector was not busy
  }
//  std::cout << std::endl << " final busy for this lb: " << totalbusy <<std::endl;

  return totalbusy;

}

long long checkRunInRNDB(std::string partitionName, int runNumber) {
  // returns -1 if it is not a run for this partition
  //          0 if it is an ongoing run for this partition
  //          run duration in microsec
  long long runDuration = -1;
  std::ostringstream stm;
  stm << runNumber;
  std::string run_number = stm.str();
  std::string command = "rn_ls -c \"oracle://atonr/rn_r\" -w atlas_run_number -p " + partitionName + " -n " + run_number + " -m " + run_number + " > /tmp/rnls.out";
  command = "rn_ls -c \"oracle://atlr/rn_r\" -w atlas_run_number -p " + partitionName + " -n " + run_number + " -m " + run_number + " > /tmp/rnls.out";
  system(command.c_str());
  std::filebuf fb;
  fb.open ("/tmp/rnls.out",std::ios::in);
  if (fb.is_open()) {
    std::istream is(&fb);
    std::string str;
    while (! is.eof()) {
      getline(is, str);
      if (str.find(partitionName) != std::string::npos) {
        std::string duration = str.substr(44, 8);
        std::cout << "duration string = " << duration << std::endl;
        if (duration.compare("        ") == 0) {
          runDuration = 0;
        } else {
          int hours = atoi((duration.substr(0, 2)).c_str());
          int minutes = atoi((duration.substr(3, 2)).c_str());
          int seconds = atoi((duration.substr(6, 2)).c_str());
          std::cout << "hours = " << hours << "  minutes = " << minutes << "  seconds = " << seconds << std::endl;
          int duration_sec = hours*3600 + minutes*60 + seconds;
          runDuration = duration_sec*1000000;
        }
      }
    }
  }
  fb.close();
  return runDuration;
}

void openDatabasesAndFolders() {
  std::string coolDb;
  std::string coolDbTrig;
  std::string coolDbMon;
  std::string coolDbDCS;
  if (atlasRun == 2) {
    coolDb = "oracle://ATLR;schema=ATLAS_COOLONL_TDAQ;dbname=CONDBR2";
  } else if (atlasRun == 1) {
    coolDb = "oracle://ATLR;schema=ATLAS_COOLONL_TDAQ;dbname=COMP200";
  }
  cool::DatabaseId dbId = coolDb;
  // Open or Create DB
  cool::IDatabaseSvc& dbSvc = cool::DatabaseSvcFactory::databaseService();
  ERS_DEBUG(3, "Retrieved DB service.");

  try {
    openDB(dbId, dbSvc);
  }
  catch(daq::rc::BadCoolDb& ex) {
      ers::fatal(ex);
      exit(0);
  }

  if (atlasRun == 2) {
    coolDbTrig = "oracle://ATLR;schema=ATLAS_COOLONL_TRIGGER;dbname=CONDBR2";
  } else if (atlasRun == 1) {
    coolDbTrig = "oracle://ATLR;schema=ATLAS_COOLONL_TRIGGER;dbname=COMP200";
  }
  cool::DatabaseId dbIdTrig = coolDbTrig;
  cool::IDatabaseSvc& dbSvcTrig = cool::DatabaseSvcFactory::databaseService();
  try {
    openDBT(dbIdTrig, dbSvcTrig);
  }
  catch(daq::rc::BadCoolDb& ex) {
      ers::fatal(ex);
      exit(0);
  }

  if (atlasRun == 2) {
    coolDbMon = "oracle://ATLR;schema=ATLAS_COOLONL_TRIGGER;dbname=CONDBR2";
  } else if (atlasRun == 1) {
    coolDbMon = "oracle://ATLR;schema=ATLAS_COOLONL_TRIGGER;dbname=MONP200";
  }
  cool::DatabaseId dbIdMon = coolDbMon;
  cool::IDatabaseSvc& dbSvcMon = cool::DatabaseSvcFactory::databaseService();
  try {
    openDBM(dbIdMon, dbSvcMon);
  }
  catch(daq::rc::BadCoolDb& ex) {
      ers::fatal(ex);
      exit(0);
  }

  if (atlasRun == 2) {
    coolDbDCS = "oracle://ATLR;schema=ATLAS_COOLOFL_DCS;dbname=CONDBR2";
  } else if (atlasRun == 1) {
    coolDbDCS = "oracle://ATLR;schema=ATLAS_COOLOFL_DCS;dbname=COMP200";
  }
  cool::DatabaseId dbIdDCS = coolDbDCS;
  cool::IDatabaseSvc& dbSvcDCS = cool::DatabaseSvcFactory::databaseService();
  try {
    openDBD(dbIdDCS, dbSvcDCS);
  }
  catch(daq::rc::BadCoolDb& ex) {
      ers::fatal(ex);
      exit(0);
  }

  // check folders
  if (!db->existsFolder(EOR_Name)) {
    std::cout << "DB Folder EOR does not exist???" << std::endl;
  }
  if (!db->existsFolder(SOR_Name)) {
    std::cout << "DB Folder SOR does not exist???" << std::endl;
  }
  try {
    if (!dbt->existsFolder(CTP_Name)) {
      std::cout << "DB Folder CTP does not exist???" << std::endl;
    }
  }
  catch (std::exception &e) {
     std::cout << "DB Folder CTP does not exist. " << e.what() << std::endl;
  }
  try {
     if (!dbm->existsFolder(BusyRate_Name)) {
        std::cout << "DB Folder BusyRate does not exist???" << std::endl;
     }
  } catch (std::exception &e) {
     std::cout << "DB Folder BusyRate does not exist. " << e.what() << std::endl;
  }
  try {
     if (!dbm->existsFolder(BusyConf_Name)) {
        std::cout << "DB Folder BusyConf does not exist???" << std::endl;
     }
  } catch (std::exception &e) {
     std::cout << "DB Folder BusyConf does not exist. " << e.what() << std::endl;
  }
  try {
     if (!dbd->existsFolder(FillState_Name)) {
        std::cout << "DB Folder FillState does not exist???" << std::endl;
     }
  } catch (std::exception &e) {
     std::cout << "DB Folder FillState does not exist. " << e.what() << std::endl;
  }
  try {
     if (!dbd->existsFolder(TotalInt_Name)) {
        std::cout << "DB Folder TotalInt does not exist???" << std::endl;
     }
  } catch (std::exception &e) {
     std::cout << "DB Folder TotalInt does not exist. " << e.what() << std::endl;
  }
  if (!dbt->existsFolder(LB_Name)) {
    std::cout << "DB Folder LB does not exist???" << std::endl;
  }
  if (!dbt->existsFolder(LBT_Name)) {
    std::cout << "DB Folder LBT does not exist???" << std::endl;
  }
  if (!dbt->existsFolder(L1Counters_Name)) {
    std::cout << "DB Folder L1Counters does not exist???" << std::endl;
  }
  if (!dbt->existsFolder(L1Menu_Name)) {
    std::cout << "DB Folder Menu does not exist???" << std::endl;
  }
    if (!db->existsFolder(DataTakingMode_Name)) {
    std::cout << "DB Folder Menu does not exist???" << std::endl;
  }

  //open folders

  EOR_Folder = db->getFolder(EOR_Name);
  SOR_Folder = db->getFolder(SOR_Name);
  EC_Folder = db->getFolder(EC_Name);

  try {
    CTP_Folder = dbt->getFolder(CTP_Name);
  }
  catch (std::exception &e) {
     std::cout << "Cannot open CTP folder " << e.what() << std::endl;
     db->closeDatabase();
     std::cout << "DBT closed." << std::endl;

     exit(0);
  }

  try {
    BusyRate_Folder = dbm->getFolder(BusyRate_Name);
  }
  catch (std::exception &e) {
     std::cout << "Cannot open BusyRate folder " << e.what() << std::endl;
     dbm->closeDatabase();
     std::cout << "DBM closed." << std::endl;

     exit(0);
  }

  try {
    BusyConf_Folder = dbm->getFolder(BusyConf_Name);
  }
  catch (std::exception &e) {
     std::cout << "Cannot open BusyConf folder " << e.what() << std::endl;
     dbm->closeDatabase();
     std::cout << "DBM closed." << std::endl;

     exit(0);
  }

  LB_Folder = dbt->getFolder(LB_Name);
  LBT_Folder = dbt->getFolder(LBT_Name);
  L1Counters_Folder = dbt->getFolder(L1Counters_Name);
  L1Menu_Folder = dbt->getFolder(L1Menu_Name);
  DataTakingMode_Folder = db->getFolder(DataTakingMode_Name);

  try {
    FillState_Folder = dbd->getFolder(FillState_Name);
  }
  catch (std::exception &e) {
     std::cout << "Cannot open FillState folder " << e.what() << std::endl;
     dbd->closeDatabase();
     std::cout << "DBD closed." << std::endl;

     exit(0);
  }

  try {
    TotalInt_Folder = dbd->getFolder(TotalInt_Name);
  }
  catch (std::exception &e) {
     std::cout << "Cannot open TotalInt folder " << e.what() << std::endl;
     dbd->closeDatabase();
     std::cout << "DBD closed." << std::endl;

     exit(0);
  }
}

void set_beam1and2_vectors() {
  beam1and2_start_vector.clear();
  beam1and2_end_vector.clear();
  // set vectors
  if (beam1_start_vector.size() > 0) {
    for (unsigned int i = 0; i < beam1_start_vector.size(); i++ ) {
      if (beam2_start_vector.size() > 0) {
        for (unsigned int j = 0; j < beam2_start_vector.size(); j++ ) {
          if ((beam1_start_vector[i] <= beam2_start_vector[j]) &&
              (beam1_end_vector[i] > beam2_start_vector[j]) &&
              (beam1_end_vector[i] <= beam2_end_vector[j])) {
            beam1and2_start_vector.push_back(beam2_start_vector[j]);
            beam1and2_end_vector.push_back(beam1_end_vector[i]);
          } else { 
            if ((beam1_start_vector[i] <= beam2_start_vector[j]) &&
                     (beam2_end_vector[j] <= beam1_end_vector[i])) {
              beam1and2_start_vector.push_back(beam2_start_vector[j]);
              beam1and2_end_vector.push_back(beam2_end_vector[j]);
            } else {
              if ((beam2_start_vector[j] <= beam1_start_vector[i]) &&
                   (beam1_start_vector[i] < beam2_end_vector[j]) &&
                   (beam2_end_vector[j] <= beam1_end_vector[i])) {
                 beam1and2_start_vector.push_back(beam1_start_vector[i]);
                 beam1and2_end_vector.push_back(beam2_end_vector[j]);
              } else {
                if ((beam2_start_vector[j] < beam1_start_vector[i]) &&
                    (beam1_end_vector[i] < beam2_end_vector[j])) {
                  beam1and2_start_vector.push_back(beam1_start_vector[i]);
                  beam1and2_end_vector.push_back(beam1_end_vector[i]);
                }
              }
            }
          }
        }
      }
    }
  }
}

bool checkInBeam(long long time) {
  if (beam1_start_vector.size() > 0) {
    for (unsigned int i = 0; i < beam1_start_vector.size(); i++) {
      if ((beam1_start_vector[i] <= time) && 
          (time <= beam1_end_vector[i])) {
        return true;
      }
    }
  }
  if (beam2_start_vector.size() > 0) {
    for (unsigned int i = 0; i < beam2_start_vector.size(); i++) {
      if ((beam2_start_vector[i] <= time) &&
          (time <= beam2_end_vector[i])) {
        return true;
      }
    }
  }
  return false;
}
  
void set_beam1or2_vectors() {
  beam1or2_start_vector.clear();
  beam1or2_end_vector.clear();
  long long starttime;
  long long endtime;
  int beam1_max = beam1_start_vector.size();
  int beam2_max = beam2_start_vector.size();
  if ((beam1_max > 0) && (beam2_max > 0)) {
    if (beam1_start_vector[0] <= beam2_start_vector[0]) {
      starttime = beam1_start_vector[0];
    } else {
      starttime = beam2_start_vector[0];
    }
    if (beam1_end_vector[beam1_max-1] <= beam2_end_vector[beam2_max-1]) {
      endtime = beam2_end_vector[beam2_max-1] + 1500;
    } else {
      endtime = beam1_end_vector[beam1_max-1] + 1500;
    }
  } else {
    if (beam1_max > 0) {
      starttime = beam1_start_vector[0];
      endtime = beam1_end_vector[beam1_max-1] + 1500;
    } else {
      if (beam2_max > 0) {
        starttime = beam2_start_vector[0];
        endtime = beam2_end_vector[beam2_max-1] + 1500;
      } else {
//        std::cout << "No beam 1 or beam 2 data" << std::endl;
        return;
      }
    }
  }
  long long currenttime = starttime;
  bool oldInBeam = false;
  int i = 0;
  while (currenttime < endtime) {
    bool inBeam = checkInBeam(currenttime);
    if ((inBeam == true) && (oldInBeam == false)) {
      beam1or2_start_vector.push_back(currenttime);
      oldInBeam = inBeam;
    } else {
      if ((inBeam == false) && (oldInBeam == true)) {
        beam1or2_end_vector.push_back(currenttime - 1000);
        oldInBeam = inBeam;
        i = i + 1;
      } else {
        oldInBeam = inBeam;
      }
    }
    // check every second
    currenttime = currenttime + 1000;
  }
} 

void set_beam1and2andStable_vectors() {
  beam1and2andStable_start_vector.clear();
  beam1and2andStable_end_vector.clear();
  // set vectors
  if (beams_start_vector.size() > 0) {
    for (unsigned int i = 0; i < beams_start_vector.size(); i++ ) {
      if (beam1and2_start_vector.size() > 0) {
        for (unsigned int j = 0; j < beam1and2_start_vector.size(); j++ ) {
          if ((beams_start_vector[i] <= beam1and2_start_vector[j]) &&
              (beams_end_vector[i] > beam1and2_start_vector[j]) &&
              (beams_end_vector[i] <= beam1and2_end_vector[j])) {
            beam1and2andStable_start_vector.push_back(beam1and2_start_vector[j]);
            beam1and2andStable_end_vector.push_back(beams_end_vector[i]);
          } else {
            if ((beams_start_vector[i] <= beam1and2_start_vector[j]) &&
                (beam1and2_end_vector[j] <= beams_end_vector[i])) {
              beam1and2andStable_start_vector.push_back(beam1and2_start_vector[j]);
              beam1and2andStable_end_vector.push_back(beam1and2_end_vector[j]);
            } else {
              if ((beam1and2_start_vector[j] <= beams_start_vector[i]) &&
                  (beams_start_vector[i] < beam1and2_end_vector[j]) &&
                  (beam1and2_end_vector[j] <= beams_end_vector[i])) {
                 beam1and2andStable_start_vector.push_back(beams_start_vector[i]);
                 beam1and2andStable_end_vector.push_back(beam1and2_end_vector[j]);
              } else {
                if ((beam1and2_start_vector[j] < beams_start_vector[i]) &&
                    (beams_end_vector[i] < beam1and2_end_vector[j])) {
                  beam1and2andStable_start_vector.push_back(beams_start_vector[i]);
                  beam1and2andStable_end_vector.push_back(beams_end_vector[i]);
                }
              }
            }
          }
        }
      }
    }
  }
}

void set_beam1or2noRamp_vectors() {
  beam1or2noRamp_start_vector.clear();
  beam1or2noRamp_end_vector.clear();
  // set vectors
  if (beam1or2_start_vector.size() > 0) {
    for (unsigned int i = 0; i < beam1or2_start_vector.size(); i++ ) {
      if (ramp_start_vector.size() > 0) {
        for (unsigned int j = 0; j < ramp_start_vector.size(); j++ ) {
          if ((beam1or2_start_vector[i] <= ramp_start_vector[j]) &&
              (beam1or2_end_vector[i] > ramp_start_vector[j]) &&
              (beam1or2_end_vector[i] <= ramp_end_vector[j])) {
            beam1or2noRamp_start_vector.push_back(ramp_start_vector[j]);
            beam1or2noRamp_end_vector.push_back(beam1or2_end_vector[i]);
          } else {
            if ((beam1or2_start_vector[i] <= ramp_start_vector[j]) &&
                (ramp_end_vector[j] <= beam1or2_end_vector[i])) {
              beam1or2noRamp_start_vector.push_back(ramp_start_vector[j]);
              beam1or2noRamp_end_vector.push_back(ramp_end_vector[j]);
            } else {
              if ((ramp_start_vector[j] <= beam1or2_start_vector[i]) &&
                  (beam1or2_start_vector[i] < ramp_end_vector[j]) &&
                  (ramp_end_vector[j] <= beam1or2_end_vector[i])) {
                 beam1or2noRamp_start_vector.push_back(beam1or2_start_vector[i]);
                 beam1or2noRamp_end_vector.push_back(ramp_end_vector[j]);
              } else {
                if ((ramp_start_vector[j] < beam1or2_start_vector[i]) &&
                    (beam1or2_end_vector[i] < ramp_end_vector[j])) {
                  beam1or2noRamp_start_vector.push_back(beam1or2_start_vector[i]);
                  beam1or2noRamp_end_vector.push_back(beam1or2_end_vector[i]);
                }
              }
            }
          }
        }
      }
    }
  }
}

void set_beam1and2noRamp_vectors() {
  beam1and2noRamp_start_vector.clear();
  beam1and2noRamp_end_vector.clear();
  // set vectors
  if (beam1and2_start_vector.size() > 0) {
    for (unsigned int i = 0; i < beam1and2_start_vector.size(); i++ ) {
      if (ramp_start_vector.size() > 0) {
        for (unsigned int j = 0; j < ramp_start_vector.size(); j++ ) {
          if ((beam1and2_start_vector[i] <= ramp_start_vector[j]) &&
              (beam1and2_end_vector[i] > ramp_start_vector[j]) &&
              (beam1and2_end_vector[i] <= ramp_end_vector[j])) {
            beam1and2noRamp_start_vector.push_back(ramp_start_vector[j]);
            beam1and2noRamp_end_vector.push_back(beam1and2_end_vector[i]);
          } else {
            if ((beam1and2_start_vector[i] <= ramp_start_vector[j]) &&
                (ramp_end_vector[j] <= beam1and2_end_vector[i])) {
              beam1and2noRamp_start_vector.push_back(ramp_start_vector[j]);
              beam1and2noRamp_end_vector.push_back(ramp_end_vector[j]);
            } else {
              if ((ramp_start_vector[j] <= beam1and2_start_vector[i]) &&
                  (beam1and2_start_vector[i] < ramp_end_vector[j]) &&
                  (ramp_end_vector[j] <= beam1and2_end_vector[i])) {
                 beam1and2noRamp_start_vector.push_back(beam1and2_start_vector[i]);
                 beam1and2noRamp_end_vector.push_back(ramp_end_vector[j]);
              } else {
                if ((ramp_start_vector[j] < beam1and2_start_vector[i]) &&
                    (beam1and2_end_vector[i] < ramp_end_vector[j])) {
                  beam1and2noRamp_start_vector.push_back(beam1and2_start_vector[i]);
                  beam1and2noRamp_end_vector.push_back(beam1and2_end_vector[i]);                }
              }
            }
          }
        }
      }
    }
  }
}

//--------------------- MAIN ----------------------------//
int main(int argc, char ** argv)
{
  // parse commandline parameters
  std::string startTime, endTime, interval;
  std::string partitionName = "ATLAS";
  bool timeInterval = false;
  double rateLevel = 0.1;
  double lumiBlockBusyLimit = 0.5;
  double detBusyLimit = 0.3;
  std::string triggerItem = "L1_EM30";
  int verbosity = 0;
  int displayOption = 0;
  boost::program_options::options_description desc("This program retrieves run and beam info from COOL for a selected time interval and stores global run efficiency data in /tmp/runs.log file.");
  desc.add_options()
    ("help,h", "help message")
    ("startTime,s", boost::program_options::value<std::string>(), "Start time (\"dd/mm/yy hh:mm:ss\")")
    ("endTime,e", boost::program_options::value<std::string>(), "End time (\"dd/mm/yy hh:mm:ss\")")
    ("interval,i", boost::program_options::value<std::string>(), "Time interval (day, week, month) - to be used only if no start time or end time is specified")
    ("partitionName,p", boost::program_options::value<std::string>(&partitionName)->default_value(partitionName), "Partition name (default ATLAS)")
    ("rateLevel,q", boost::program_options::value<double>(&rateLevel)->default_value(rateLevel), "Level 1 rate level (default 0.1 Hz)")
    ("verbosity,v", boost::program_options::value<int>(&verbosity)->default_value(verbosity), "Verbosity: 0 - no lumi block info, 1 - lumi blocks with low rate (default 0)") 
    ("displayOption,o", boost::program_options::value<int>(&displayOption)->default_value(displayOption), "Display option: 0 - flat run; 1 - lumi block rate display (default 0)")
    ("atlasRun,r", boost::program_options::value<int>(&atlasRun)->default_value(atlasRun), "ATLAS Run: 1 - Run 1; 2 - Run 2 (default 2)")
    ("beamType,t", boost::program_options::value<int>(&beamType)->default_value(beamType), "Beam type: 0 - stable beams; 1 - two beams present; 2 - one beam present; 3 - two beams present and stable; 4 - beam data from file  (default 3)")
    ("beamLimit,b", boost::program_options::value<int>(&beamLimit)->default_value(beamLimit), "Beam limit option in seconds (default 5 s)")
    ("lumiBlockBusyLimit,l", boost::program_options::value<double>(&lumiBlockBusyLimit)->default_value(lumiBlockBusyLimit), "Lumi block busy fraction limit (default 0.5)")
    ("detBusyLimit,d", boost::program_options::value<double>(&detBusyLimit)->default_value(detBusyLimit), "Sub-detector busy limit (default 0.3)")
    ("writeBeamFile,w", boost::program_options::value<int>(&writeBeamFile)->default_value(writeBeamFile), "Write beam file: 0 - no beam file; 1 - write beam file (default 0)")
    ("display,x", boost::program_options::value<int>(&display)->default_value(display), "Display options: 0 - no graphical display, 1 - open graphical display (default 0)")
    ("noRamp,n",boost::program_options::value<int>(&noRamp)->default_value(noRamp), "No ramp option: 0 - ramps not excluded from beam time, 1 - ramps excluded from beam time (default 0)")
    ("triggerItem,c", boost::program_options::value<std::string>(&triggerItem)->default_value(triggerItem), "Trigger item (default L1_EM30)")
    ("readyForPhysics,f", boost::program_options::value<int>(&readyForPhysics)->default_value(readyForPhysics), "Ready for physics option: 0 - do not use ready for physics flag, 1 - use ready for physics flag (default 0)")
    ;

  boost::program_options::variables_map vm;
  boost::program_options::store(boost::program_options::parse_command_line(argc,argv,desc), vm);
  boost::program_options::notify(vm);

  if(vm.count("help")) {
    std::cout << desc << std::endl;
    return EXIT_SUCCESS;
  }
  if ((atlasRun < 1) || (atlasRun > 2)) {
    std::cout << "Command line error: atlasRun should be 1 or 2" << std::endl;
    return EXIT_FAILURE;
  }
  if ( (vm.count("interval")) && ( (vm.count("startTime")) || (vm.count("endTime")) ) ) {
   std::cout << desc << std::endl;
   std::cout << "Command line error: time interval used with an explicit start time or end time" << std::endl;
   return EXIT_FAILURE;
  }
  else if ( (vm.count("startTime")) && (vm.count("endTime")) ) {
    startTime = vm["startTime"].as<std::string>();
    endTime = vm["endTime"].as<std::string>();
  }
  else if ( vm.count("interval") ) {
    interval = vm["interval"].as<std::string>();
    timeInterval = true;
  } 
  else {
    std::cout << desc << std::endl;
    daq::rc::CmdLineError issue(ERS_HERE,"");
    ers::fatal(issue);
    return EXIT_FAILURE;
  }

  std::string runs("/tmp/runs.log");
  std::ostream * m_lout = new std::ofstream (runs.c_str(),std::ios::out | std::ios::trunc);        
  if (!(*m_lout)) {
    std::cout << "Can't create the runs log file " << std::endl;
    exit(0);
  }

  if (displayOption == 0) {
    (*m_lout) << "option0" << std::endl;
  }
  if (displayOption == 1) {
    (*m_lout) << "option1" << std::endl;
  }

  openDatabasesAndFolders();

  // Install signal handler
  signal(SIGINT,  signalhandler);
  signal(SIGTERM, signalhandler);
  ERS_DEBUG(3, "Installed signal handlers.");

  OWLTime now = OWLTime();
  long long nowtime = now.total_mksec_utc();
  long long starttime = 0;
  long long endtime = 0;
  if (!timeInterval) {
    OWLTime start_time(startTime.c_str());
    starttime = start_time.total_mksec_utc();
    OWLTime end_time(endTime.c_str());
    endtime  = end_time.total_mksec_utc();
    if (starttime > nowtime) {
      std::cout << " Start time greater than current time" << std::endl;
      return EXIT_FAILURE;
    }
    if (endtime > nowtime) {
      std::cout << " End time greater than current time" << std::endl;
      return EXIT_FAILURE;
    }
  }
  else {
    long long duration_mksec = 0;
    if (interval.compare("day") == 0) {
      duration_mksec = ((long long) 86400) * ((long long) 1000000);
    }
    else if (interval.compare("week") == 0) {
      duration_mksec = ((long long) 604800) * ((long long) 1000000);
    }
    else if (interval.compare("month") == 0) {
      duration_mksec = ((long long) 2592000) * ((long long) 1000000);
    } 
    else {
      std::cout << desc << std::endl;
      std::cout << "Wrong time interval option!" << std::endl;
      return EXIT_FAILURE;
    }
    endtime  = nowtime;
    starttime = endtime - duration_mksec;
    endTime = std::string(now.c_str());
    OWLTime start = OWLTime(starttime/1000000);
    startTime = std::string(start.c_str());
  }
  long long totaltime = endtime - starttime;
  int total_time_sec = totaltime/1000000;

  std::vector<long long> start_of_run_vector;
  std::vector<long long> end_of_run_vector;

  int totalBeamTime = 0;

  // read beam data
  if (beamType == 0) {
    // stable beams
    get_beams_data (starttime*1000, endtime*1000, FillState_Folder);
    set_main_vectors(beams_start_vector, beams_end_vector);
    if (writeBeamFile == 1) {
      print_vectors(start_beam_vector, end_beam_vector, "/tmp/beams.log");
    }
  }
  if (beamType == 1) {
    // two beams present
    get_beam1_data (starttime*1000, endtime*1000, TotalInt_Folder);
    get_beam2_data (starttime*1000, endtime*1000, TotalInt_Folder);
    set_beam1and2_vectors();
    if (noRamp == 0) {
      set_main_vectors(beam1and2_start_vector, beam1and2_end_vector);
    }
    if (noRamp == 1) {
      get_ramp_data (starttime*1000, endtime*1000, FillState_Folder);
      set_beam1and2noRamp_vectors();
      set_main_vectors(beam1and2noRamp_start_vector, beam1and2noRamp_end_vector);
    }
    if (writeBeamFile == 1) {
      print_vectors(beam1_start_vector, beam1_end_vector, "/tmp/beam1.log");
      print_vectors(beam2_start_vector, beam2_end_vector, "/tmp/beam2.log");
      print_vectors(beam1and2_start_vector, beam1and2_end_vector, "/tmp/beam1and2.log");
      if (noRamp == 1) {
        print_vectors(ramp_start_vector, ramp_end_vector, "/tmp/ramp.log");
        print_vectors(start_beam_vector, end_beam_vector, "/tmp/beam1and2noRamp.log");
      }
    }
  }
  if (beamType == 2) {
    // one beam present
    get_beam1_data (starttime*1000, endtime*1000, TotalInt_Folder);
    get_beam2_data (starttime*1000, endtime*1000, TotalInt_Folder);
    set_beam1or2_vectors();
    if (noRamp == 0) {
      set_main_vectors(beam1or2_start_vector, beam1or2_end_vector);
    }
    if (noRamp == 1) {
      get_ramp_data (starttime*1000, endtime*1000, FillState_Folder);
      set_beam1or2noRamp_vectors();
      set_main_vectors(beam1or2noRamp_start_vector, beam1or2noRamp_end_vector);
    }
    if (writeBeamFile == 1) {
      print_vectors(beam1_start_vector, beam1_end_vector, "/tmp/beam1.log");
      print_vectors(beam2_start_vector, beam2_end_vector, "/tmp/beam2.log");
      print_vectors(beam1or2_start_vector, beam1or2_end_vector, "/tmp/beam1or2.log");
      if (noRamp == 1) {
        print_vectors(ramp_start_vector, ramp_end_vector, "/tmp/ramp.log");
        print_vectors(start_beam_vector, end_beam_vector, "/tmp/beam1or2noRamp.log");
      }
    }
  }
  if (beamType == 3) {
    // two beams present and stable
    get_beam1_data (starttime*1000, endtime*1000, TotalInt_Folder);
    get_beam2_data (starttime*1000, endtime*1000, TotalInt_Folder);
    get_beams_data (starttime*1000, endtime*1000, FillState_Folder);
    set_beam1and2_vectors();
    set_beam1and2andStable_vectors();
    set_main_vectors(beam1and2andStable_start_vector, beam1and2andStable_end_vector);
    if (writeBeamFile == 1) {
      print_vectors(beam1_start_vector, beam1_end_vector, "/tmp/beam1.log");
      print_vectors(beam2_start_vector, beam2_end_vector, "/tmp/beam2.log");
      print_vectors(beams_start_vector, beams_end_vector, "/tmp/beams.log");
      print_vectors(start_beam_vector, end_beam_vector, "/tmp/beam1and2andStable.log");
    }
  }
  if (beamType == 4) {
    // beam data from file
    get_beam_data_from_file (starttime, endtime);
    if (writeBeamFile == 1) {
      print_vectors(start_beam_vector, end_beam_vector, "/tmp/beamfromfile.log");
    }
  }
  totalBeamTime = get_total_beam_time();
  if (start_beam_vector.size() > 0) {
    for (unsigned int i = 0; i < start_beam_vector.size(); i++ ) {
      (*m_lout) << start_beam_vector[i] << " " << end_beam_vector[i] << std::endl;
    }
  }
  (*m_lout) << "endbeam" << std::endl;

  // read run data

  int totalTime = 0;
  int totalInBeamTime = 0;
  int totalInBeamTime_msec = 0;
  int no_data_time_in_beam_msec = 0;
  int totalRunNumber = 0;
    // find first run
  int firstRunNumber = 173500; // default value, runs after Jan 2011
  try {
    long long start_time_nsec = ((long long)(starttime))*1000;
    long long end_time_nsec = ((long long)(endtime))*1000;
    cool::IObjectIteratorPtr rl_objects = LBT_Folder->browseObjects(cool::ValidityKey(start_time_nsec), cool::ValidityKey(end_time_nsec), 0);
    rl_objects->goToNext();
    const cool::IObject& obj = rl_objects->currentRef();
    std::string runNumber = obj.payloadValue("Run");
    firstRunNumber = atoi(runNumber.c_str());
  } catch (std::exception &e) {
    std::cout << "Warning: Retrieving firstRunNumber failed! -> " << e.what() << std::endl;
  }
  try {
    // consider only runs with RN > 140000 (18-Nov-2009)
    // cool::UInt63 aa = 140000;
    cool::UInt63 aa = firstRunNumber;
    cool::ValidityKey initialKey = aa <<32;

    // start looking for runs in SOR_Folder
    cool::IObjectIteratorPtr objects = SOR_Folder->browseObjects( initialKey, cool::ValidityKeyMax, 0 );

    cool::ValidityKey key;
    cool::ValidityKey key1;
    cool::IObjectPtr obj_eor;
    cool::IObjectPtr obj_ec;
    cool::IObjectPtr obj_lb;
    while ( objects->goToNext() ) {
      const cool::IObject& obj = objects->currentRef();
      // read start of run time
      std::string sor = obj.payloadValue("SORTime").substr(0,16);
      long long sortime = atoll(sor.c_str());
      time_t sor_tt = time_t(sortime/1000000);
      OWLTime owlsor = OWLTime(sor_tt);
      // read run number
      std::string runNumber = obj.payloadValue("RunNumber");
      int run_number = atoi(runNumber.c_str());
      // prepare the since and until value for the search in EOR_Folder
      cool::UInt63 a = run_number;
      cool::UInt63 z = run_number+1;
      cool::ValidityKey sKey = a <<32;
      cool::ValidityKey uKey = (z <<32) - 1;
      // initial value for end of run (to be read later)
      long long eortime = 0;
      // check only for runs with start time three days before start of search
      // interval
      long long pre_check = ((long long)(72000))*((long long)(3600000));
      long long check_time = starttime - pre_check;
//      if ((sortime > (starttime - pre_check)) && (sortime < endtime)) {
      if ((sortime > check_time) && (sortime < endtime)) {
        // try to find end of run time
        cool::IObjectIteratorPtr objects_eor = EOR_Folder->browseObjects(sKey, uKey, 0 );
        bool found = false;
        bool noEOR = false;
        std::string activeTime = "";
        while ( objects_eor->goToNext() ) {
          found = true;
          const cool::IObject& obj_eor = objects_eor->currentRef();
          std::string eor = obj_eor.payloadValue("EORTime").substr(0,16); 
          activeTime = obj_eor.payloadValue("TotalTime");
          eortime = atoll(eor.c_str());     
          key = obj.since();
        }
        bool currentRun = false;
        if (found == false) {
          noEOR = true;
          // consider that the run just started, to be checked later ??
          long long run_duration = checkRunInRNDB(partitionName, run_number);
          if (run_duration == 0) {
            currentRun = true;
            eortime = endtime;
            key = obj.since();
          } else if (run_duration > 0) {
            eortime = sortime + run_duration;
            key = obj.since();
          } else {
            // run was not found as a run for this partition in RN DB
            // skip this run
            continue;
          }
        }
        // OWLTime owleor = OWLTime(eor_tt);
        if (((sortime > check_time) && (sortime < endtime)) &&
            (eortime > starttime))  {
          try {
            bool goodRun = false;
            int recorded_events = 0;
            int active_time = 0;
            if (!noEOR) {
              obj_ec = EC_Folder->findObject(key, 0);
              std::string partition = obj_ec->payloadValue("PartitionName");
              active_time = atoi(activeTime.c_str());
              if (partition.compare(partitionName) == 0) {
                std::string recordedEvents = obj_ec->payloadValue("RecordedEvents");
                recorded_events = atoi(recordedEvents.c_str());
                goodRun = true;
              }
            }
            if (noEOR || goodRun) {
              long long run_time = (eortime - sortime)/1000000;
              if (currentRun) {
                active_time = run_time;
              }
//              std::cout << "RN = " << runNumber << " SOR = " << owlsor.c_str() << " EOR = " << owleor.c_str() << " Active time = " << activeTime << "  Real time = " << run_time << std::endl;
              long long sor_ms = 0;
              long long eor_ms = 0;
              int val = 0;
              bool toSkip = false;
              bool noEvents = false;
              if (recorded_events == 0) {
                noEvents = true;
// for test                toSkip = true;
              }
              if ((sortime >= starttime) && (eortime <= endtime)) {
                val = (eortime - sortime) / 1000000;
                if (currentRun) {
                  std::cout << "RN = " << runNumber << " SOR = " << owlsor.c_str() << " Run time = " << run_time << " sec  => ongoing" << std::endl;
                } else {
                  std::cout << "RN = " << runNumber << " SOR = " << owlsor.c_str() << " Run time = " << run_time << " sec" << std::endl;
                }
                sor_ms = sortime/1000;
                eor_ms = eortime/1000;
              }
              if ((sortime < starttime) && (eortime <= endtime)) {
                val = (eortime - starttime) / 1000000;
                std::cout << "RN = " << runNumber << " SOR = " << owlsor.c_str() << " Run time = " << run_time << " sec" << "  Counted time = " << val << " sec" << std::endl;
                sor_ms = starttime/1000;
                eor_ms = eortime/1000;
              }
              if ((sortime >= starttime) && (eortime > endtime)) {
                val = (endtime - sortime) / 1000000;
                std::cout << "RN = " << runNumber << " SOR = " << owlsor.c_str() << " Run time = " << run_time << " sec" << "  Counted time = " << val << " sec" << std::endl;
                sor_ms = sortime/1000;
                eor_ms = endtime/1000;
              } 
              if ((sortime < starttime) && (eortime > endtime)) {
                val = (endtime - starttime) / 1000000;
                std::cout << "RN = " << runNumber << " SOR = " << owlsor.c_str() << " Run time = " << run_time << " sec" << "  Counted time = " << val << " sec" << std::endl;
                sor_ms = starttime/1000;
                eor_ms = endtime/1000;
              }
              if (noEOR) {
                std::cout << "            no EOR record!" << std::endl;
              }
              if (noEvents) {
                std::cout << "            no recorded events ???" << std::endl;
              }
              if (!toSkip) {
                if (displayOption == 0) {
                  (*m_lout) << sor_ms << " " << eor_ms << std::endl;
                }
                if (displayOption == 1) {
                  start_of_run_vector.push_back(sor_ms);
                  end_of_run_vector.push_back(eor_ms);
                  (*m_lout) << sor_ms << " 000000.000" << std::endl;
                }
                totalTime = totalTime + val;
                totalRunNumber++;
                ctpMap.clear();
                get_ctp_map(run_number, BusyConf_Folder );
                std::map<std::string,std::string>::iterator it;
                int ctpId = get_ctpid(sKey, uKey, triggerItem, L1Menu_Folder);
                if (ctpId == -1) {
                  std::cout << "Trigger item not found. Default value (227) will be used." << std::endl;
                  ctpId = 227;
                }
                std::cout << "Triger Item: " << ctpId << std::endl;
//                int no_data_time_in_beam_msec = 0;
                int no_data_time_msec = 0;
                if (rateLevel > 0) {
                  // use the same since and until keys as for EOR_Folder
                  cool::IObjectIteratorPtr objects_ctp = CTP_Folder->browseObjects(sKey, uKey, 0 );
                  int lumiBlocks = 0;
                  int lowRateLumiBlocks = 0;
                  while ( objects_ctp->goToNext() ) {
                    const cool::IObject& obj_ctp = objects_ctp->currentRef();
                    lumiBlocks = lumiBlocks + 1;
                    key1 = obj_ctp.since();
                    obj_lb = LB_Folder->findObject(key1, 0);
                    std::string st = obj_lb->payloadValue("StartTime");
                    std::string et = obj_lb->payloadValue("EndTime");
                    long long sT = atoll(st.c_str());
                    long long eT = atoll(et.c_str());
                    long long eT_msec = eT/1000000;
                    long long lb = eT - sT;
                    long long lb_msec = lb/1000000;
                    std::string lbLVL1 = obj_ctp.payloadValue("GlobalL1AcceptCounter");
                    long long lb_LVL1 = atoll(lbLVL1.c_str());
                    long long lb_LVL1_long = lb_LVL1*1000;               
                    double lb_rate = (double)lb_LVL1_long/ (double)lb_msec;
                    int readyForPhysicsFlag = 1;
                    if (((beamType == 0) || (beamType == 3)) && (readyForPhysics == 1)) {
                      readyForPhysicsFlag = ready_for_physics(key1, DataTakingMode_Folder);
                    }
                    double vetofraction = veto_fraction(key1, ctpId, L1Counters_Folder);
                    double busyfraction = fraction(sT, eT, BusyRate_Folder );
                    if (busyfraction > lumiBlockBusyLimit) {
                      long long start_lb_sec = sT/1000000000;
                      OWLTime start_lb_owl = OWLTime(start_lb_sec);
                      std::cout << " for Lumi Block " << lumiBlocks << " (" << start_lb_owl << ") high busy fraction = " << busyfraction << message << " and veto fraction = " << vetofraction << std::endl;
                      if (!no_busy_info) {
                        std::map<std::string,std::string>::iterator iter;
                        for ( iter=ctpMap.begin() ; iter != ctpMap.end(); iter++ ) {
                          double det_busy = busy(sT,eT, (*iter).first, BusyRate_Folder );
                          if (det_busy > detBusyLimit) {
                            std::cout << "    High busy for " << (*iter).second << " : " << det_busy << std::endl;
                          } 
                        }
                        double det_busy = busy(sT,eT, "ctpmi_vme_rate", BusyRate_Folder );
                        if (det_busy > detBusyLimit) {
                          std::cout << "    High busy for VME is " << det_busy << std::endl;
                        }
                      }
                    } 
                    if ((displayOption == 1) && (eT_msec > sor_ms) && (eT_msec < eor_ms)) {
                      if (lb_rate < 0.001) {
                        (*m_lout) << et.substr(0,13) << " 000000.000" << std::endl;
                      }
                      if ((lb_rate >= 0.001) && (lb_rate < 10)) {
                        (*m_lout) << et.substr(0,13) << " 00000" << lb_rate << std::endl;
                      }
                      if ((lb_rate >= 10) && (lb_rate < 100)) {
                        (*m_lout) << et.substr(0,13) << " 0000" << lb_rate << std::endl;
                      }
                      if ((lb_rate >= 100) && (lb_rate < 1000)) {
                        (*m_lout) << et.substr(0,13) << " 000" << lb_rate << std::endl;
                      }
                      if ((lb_rate >= 1000) && (lb_rate < 10000)) {
                        (*m_lout) << et.substr(0,13) << " 00" << lb_rate << std::endl;
                      }
                      if ((lb_rate >= 10000) && (lb_rate < 100000)) {
                        (*m_lout) << et.substr(0,13) << " 0" << lb_rate << std::endl;
                      }
                    }
                    if (lb_rate < rateLevel) {
                      if (verbosity == 1) {
                        std::cout << "   LB " << lumiBlocks << ": LVL1 accepts = " << lb_LVL1 << ";  LB duration = " << lb_msec << " ms; " << "  LB rate = " << lb_rate << std::endl;
                      }
                      lowRateLumiBlocks = lowRateLumiBlocks + 1;
                    }
                    if (start_beam_vector.size() > 0) {
                      for (unsigned int i = 0; i < start_beam_vector.size(); i++ ) {
                        int data_lumi_block_in_beam_msec = valInBeam(sT/1000000, eT/1000000, start_beam_vector[i], end_beam_vector[i]);
                        if (data_lumi_block_in_beam_msec > 0) {
                          if (readyForPhysicsFlag == 0) {
                            vetofraction = 1;
                          }
                          totalInBeamTime_msec = totalInBeamTime_msec + data_lumi_block_in_beam_msec;
                          int no_data_lumi_block_in_beam_msec = (int) (((double) data_lumi_block_in_beam_msec) * vetofraction);
                          no_data_time_in_beam_msec = no_data_time_in_beam_msec + no_data_lumi_block_in_beam_msec;
                        }
                      }
                    }
                    totalInBeamTime = totalInBeamTime_msec/1000 - no_data_time_in_beam_msec/1000;
                  }
                  std::cout << "  Total number of LB = " << lumiBlocks << ";  Low rate LB = " << lowRateLumiBlocks << std::endl;
                }
                if (displayOption == 1 ) {
                  (*m_lout) << eor_ms << " 000000.000" << std::endl;
                }
                if (no_data_time_msec > 0) {
                  totalTime = totalTime - no_data_time_msec/1000;
                }
                if (start_beam_vector.size() > 0) {
                  // use same criteria as DQ, for collisions duration > 15 min
                  if (active_time > 900) {
                    for (unsigned int i = 0; i < start_beam_vector.size(); i++ ) {
                      if (((sor_ms < start_beam_vector[i]) && (eor_ms > start_beam_vector[i])) && ((eor_ms - start_beam_vector[i])/1000 > 3600)) {
                        delay_vector[i] = 0;
                        run_number_vector[i] = run_number;
                      }
                      if ((sor_ms >= start_beam_vector[i]) && (sor_ms < end_beam_vector[i])) {
                        if ((sor_ms - start_beam_vector[i])/1000 < delay_vector[i]) {
                          delay_vector[i] = (sor_ms - start_beam_vector[i])/1000;
                          run_number_vector[i] = run_number;
                        }
                      }
                    }
                  }
                }
              }
            }
          } catch (std::exception &e) {
            std::cout << "Warning: Retrieving run infos failed for RN = " << obj.payloadValue("RunNumber") << "  -> " << e.what() << std::endl;
          }
        }
      }
    }
  } catch (std::exception &e) {
     std::cout << "Read failed: " << e.what() << std::endl;
  }

  (*m_lout) << "end" << std::endl;
  if (displayOption == 1) {
    if (start_of_run_vector.size() > 0) {
      for (unsigned int i = 0; i < start_of_run_vector.size(); i++ ) {
        (*m_lout) << start_of_run_vector[i] << " " << end_of_run_vector[i] << std::endl;
      }
    }
    (*m_lout) << "finalend" << std::endl;
    std::cout << std::endl;
  } 

  // print summary data
  double meanTime = 0;
  if (totalRunNumber > 0) { 
     meanTime = (double)(double)totalTime/(double)totalRunNumber;
  }
  std::cout << "Total time (sec)            = " << total_time_sec << std::endl;
  std::cout << "Total beam time (sec)       = " << totalBeamTime << std::endl;
  std::cout << "Running time (sec)          = " << totalTime << std::endl;
  std::cout << "Beam running time (sec)     = " << totalInBeamTime << std::endl;
  std::cout << "Total run number            = " << totalRunNumber << std::endl;
  std::cout << "Mean run time (sec)         = " << meanTime << std::endl;
  double fraction = (double)totalTime/(double)total_time_sec;
  std::cout << "RunningTime/TotalTime       = " << fraction << std::endl;
  double fractionBeam = 0;
  if (totalBeamTime > 0) {
    fractionBeam = (double)totalInBeamTime/(double)totalBeamTime;
    std::cout << "BeamRunningTime/BeamTime    = " << fractionBeam << std::endl;
  }
  std::cout << std::endl;

  (*m_lout) << partitionName << " partition" << std::endl;
  (*m_lout) << "Run efficiency from " << startTime << " to " << endTime << std::endl;
  (*m_lout) << "Total time (sec) = " << total_time_sec << std::endl;
  (*m_lout) << "Total beam time (sec) = " << totalBeamTime << std::endl;
  (*m_lout) << "Running time (sec) = " << totalTime << std::endl;
  (*m_lout) << "Total run number = " << totalRunNumber << std::endl;
  (*m_lout) << "Mean run time (sec) = " << meanTime << std::endl;
  (*m_lout) << "RunningTime/TotalTime = " << fraction << std::endl;
  if (totalBeamTime > 0) {
    (*m_lout) << "BeamRunningTime/BeamTime = " << fractionBeam << std::endl;
  } else {
    (*m_lout) << "                           " << std::endl;
  }
  if (start_beam_vector.size() > 0) {
    int totalDelay = 0;
    for (unsigned int i = 0; i < start_beam_vector.size(); i++ ) {
      totalDelay = totalDelay + delay_vector[i];
      // get fill number
      long long start_time_sec = start_beam_vector[i]/1000;
      long long start_search_nsec = ((long long)(start_time_sec-1))*1000000000;
      long long end_search_nsec = ((long long)(start_time_sec+1))*1000000000;
      int fill_number = 1000;
      cool::IObjectIteratorPtr objects = FillState_Folder->browseObjects( cool::ValidityKey(start_search_nsec),cool::ValidityKey(end_search_nsec), cool::ChannelSelection::all() );
      while ( objects->goToNext() ) {
        const cool::IObject& obj = objects->currentRef();
        std::string fillNumber = obj.payloadValue("FillNumber");
        fill_number = atoi(fillNumber.c_str());
      }
//      std::cout << "Beam " << i << ":   RN = " << run_number_vector[i] <<"   delay = " << delay_vector[i] << std::endl;
      std::cout << "Fill " << fill_number << ":   RN = " << run_number_vector[i] <<"   delay = " << delay_vector[i] << std::endl;
    }
    double meanDelay = totalDelay/start_beam_vector.size();
    std::cout << "Mean delay = " << meanDelay << " sec" << std::endl;
  }
  delete m_lout;
  // Close the DB
  db->closeDatabase();
  dbt->closeDatabase();
  dbm->closeDatabase();
  dbd->closeDatabase();
  std::cout << "DB closed." << std::endl;

  if (display == 1) {
    system("run_eff_display");
  }
  return EXIT_SUCCESS;
}

