#ifndef RM_Client_H_INCLUDED
#define RM_Client_H_INCLUDED

#include "ipc/object.h"
#include "ipc/partition.h"
#include <ResourceManager/idlhelper.h>
#include <ResourceManager/exceptions.h>
#include <time.h>
#include <boost/thread/mutex.hpp>
#include <stdint.h>

#include <AccessManager/xacml/impl/RMResource.h>
#include <AccessManager/client/RequestorInfo.h>
#include <AccessManager/client/ServerInterrogator.h>
#include <AccessManager/util/ErsIssues.h>

#include "config/Configuration.h"
#include "config/ConfigObject.h"

namespace daq   {

namespace rmgr  {

//working class
enum RequestType {FreePartition, FreeComputer, FreeResource, FreePartitionResource, FreeComputerResource, FreeApplicationResources};

class ClientRequest
{
public:
RequestType type;
const char* param1;
const char* param2;
const char* param3;
bool _log_this_call;
ClientRequest(RequestType tp, const char* par1, bool log_this_call) : type(tp), param1(par1), param2(0), param3(0), _log_this_call(log_this_call)  {}
ClientRequest( RequestType tp, const char* par1, const char* par2, bool log_this_call) : type(tp), param1(par1), param2(par2), param3(0), _log_this_call(log_this_call) {}
ClientRequest(RequestType tp, const char* par1, const char* par2, const char* par3, bool log_this_call) : type(tp), param1(par1), param2(par2), param3(par3), _log_this_call(log_this_call) {}
};

class RM_ClientImpl;

/*! \brief This class provides a public API for all the Resource manager (RM) functionality.
    \ingroup public
    \sa
 */
class RM_Client
{
protected:
  RM_ClientImpl* impl;
///  ::rmgr::ResMgr_var RMDB;
///  ::IPCPartition m_defpart;
///  std::string m_rmDbName;

///  void lookupRMDDB() throw ( CommunicationException );
public:

//constructors

  /*! Creates the object which represents the default client to be interacted with Resource Manager server with name RM_Server
  */
  RM_Client();

  /*! Creates the object which represents the client to be interacted with Resource Manager server with name sid
  *   \param sid	 Resource Manager server name
  */
  RM_Client(const char *sid);

  ~RM_Client();

//loading of configuration data into RM server

/*!   Loads on the client side all configuration data that associated with RM resources (i.e. RM resources and binaries that accosiated with any RM resource) into the Resource Manager server. It updates according data if configuration was loaded earlier.
 *   \param  dbname   Configuration DB name. The name of DB that is used to create ::Configuration
 *   \throw    daq::rmgr::ConfigurationException   Can not load requested configuration data.
 *   \throw    daq::rmgr::CommunicationException    IPC or system CORBA exception catched and  according messages transferred into this exception.
 */
void rmConfigurationUpdate ( const std::string& dbname );

/*!    Loads on the client side all configuration data that associated with RM resources (i.e. RM resources and binaries that accosiated with any RM resource) into the Resource Manager server. It updates according data if configuration was loaded earlier.
 *   \param config    Pointer to DAQ DB configuration that should be loaded into the RM.
 *   \throw    daq::rmgr::ConfigurationException   Can not load requested configuration data.
 *   \throw    daq::rmgr::CommunicationException    IPC or system CORBA exception catched and  according messages transferred into this exception.
 */
void rmConfigurationUpdate ( Configuration * conf );

// resource allocation methods

/*!   Requests resources in partition for software object to run on the requested computer.
  *   \param  partition   Partition name.
  *   \param  swobjectid   Software object identifier.
  *   \param  computerName   Computer name. Computer on which application to be started.
  *  \param clientName     Client name. This name is stored in the Resource Manager server as the owner of according resources.
  *   \return      Resource Manager handle identifier. If this value is equaled to zero - means there are no resources associated in the configuration with software object (nothing to be allocated). If this value more than zero - means resources were granted and this handle value can  be used to free these resources when according process will be stopped.
  *   \throw    daq::rmgr::UnavailableResources    There are no enough resources for the application.
  *   \throw    daq::rmgr::CommunicationException    IPC or system CORBA exception catched and  according messages transferred into this exception.
*/
 long requestResources (
                        const std::string& partition,
                        const std::string& swobjectid, //id in confdb
                        const std::string& computerName, // name in confdb
                        const std::string& clientName = "NO");

/*!   Requests resources for process. Requested resources are those that associated with software object.
  *   \param  partition   Partition name.
  *   \param  swobjectid   Software object identifier.
  *   \param  computerName   Computer name.  Computer on which process is running.
  *   \param  pid PID of process that needs resource
  *  \param clientName     Client name. This name is stored in the Resource Manager server as the owner of according resources.
  *   \return      Resource Manager handle identifier. If this value is equaled to zero - means there are no resources that were associated in the configuration with software object (nothing to be allocated). If this value more than zero - means resources were granted and this handle value can  be used to free these resources when according process will be stopped.
  *   \throw    daq::rmgr::UnavailableResources    There are no enough resources for the application.
  *   \throw    daq::rmgr::CommunicationException    IPC or system CORBA exception catched and  according messages transferred into this exception.
*/
 long requestResources (
                        const std::string& partition,
                        const std::string& swobjectid, //id in confdb
                        const std::string& computerName, // name in confdb
                        unsigned long mypid,
                        const std::string& clientName = "NO");

/*!   Requests resource for process. This method will get process id automatically from the system and this value is used in the Resource Manager server during request processing.
  *   \param  partition   Partition name.
  *   \param  resource RM resource unique ID.
  *   \param  computer Computer name.
  *  \param clientName     Client name. This name is stored in the Resource Manager server as the owner of according resources.
  *   \return      Resource Manager handle identifier. If this value is equaled to zero - means there are no resources that were associated in the configuration with software object (nothing to be allocated). If this value more than zero - means resources were granted and this handle value can  be used to free these resources when according process will be stopped.
  *   \throw    daq::rmgr::UnavailableResources    There are no enough resources for the process.
 *   \throw    daq::rmgr::CommunicationException    IPC or system CORBA exception catched and  according messages transferred into this exception.
*/
 long requestResourceForMyProcess(
                        const std::string& partition,
                        const std::string& resource,
                        const std::string& computer,
                        const std::string& clientName = "NO");

//request hardware resources (not Computer)

/*!   Requests hardware resources (not Computer class) in partition for software object to run on the requested computer. Requested resources are those that associated with software object.
  *   \param  partition   Partition name.
  *   \param  swobjectid  Software object identifier.
  *   \param  computer    Computer name. Computer on which application to be started.
  *    \param  pid        PID of process that needs resource
  *   \param  hwClass     Hawdware class
  *   \param  hwId        Hardware UID
  *  \param client        Client name. This name is stored in the Resource Manager server as the owner of according resources.
  *   \return      Resource Manager handle identifier. If this value is equaled to zero - means there are no resources associated in the configuration with software object (nothing to be allocated). If this value more than zero - means resources were granted and this handle value can  be used to free these resources when according process will be stopped.
  *   \throw    daq::rmgr::UnavailableResources    There are no enough resources for the application.
  *   \throw    daq::rmgr::CommunicationException    IPC or system CORBA exception catched and  according messages transferred into this exception.
*/
 long requestHWResources(
                        const std::string& partition,
                        const std::string& swId,
                        const std::string& computer,
                        unsigned long pid,
                        const std::string& hwClass,
                        const std::string& hwId,
                        const std::string& client = "NO");

/*!   Requests hardware resources (not Computer class) for process.
  *   \param  partition   Partition name.
  *   \param  resource    Resource  identifier.
  *   \param  computer    Computer name.  Computer on which process is running.
  *   \param  hwId        Hardware object indentifier
  *   \param  pid         PID of process that needs resource
  *  \param client        Client name. This name is stored in the Resource Manager server as the owner of according resources.
  *   \return      Resource Manager handle identifier. If this value is equaled to zero - means there are no resources that were associated in the configuration with software object (nothing to be allocated). If this value more than zero - means resources were granted and this handle value can  be used to free these resources when according process will be stopped.
  *   \throw    daq::rmgr::UnavailableResources    There are no enough resources for the application.
  *   \throw    daq::rmgr::CommunicationException    IPC or system CORBA exception catched and  according messages transferred into this exception.
*/
 long requestHWResourceForProcess(
                        const std::string& partition,
                        const std::string& resource,
                        const std::string& computer,
                        const std::string& hwId,
                        unsigned long mypid,
                        const std::string& client = "NO");

//free resources methods

/*! Frees resources that were granted with the RM handle.
  *   \param handleId  RM handle identifier.
  *   \param log_this_call true if log message should be added about this call
  *   \throw   daq::rmgr::HandleIdNotFound  There is no handle with requested identifier in the RM server. 
  *   \throw    daq::rmgr::CommunicationException    IPC or system CORBA exception catched and  according messages transferred into this exception.
*/
 void freeResources (long handleId, bool log_this_call=true);


/*!   Releases all resources granted in the partition that were granted for software object ID and computer name.
  *   \param  p   Name of the partition. Value "*" means in all partitions. 
  *   \param  swobjid   Configuration UID of the software object. Value "*" means for all software objects.
  *   \param  computerName   Computer name. Do not used in case of swobjid and partition equals to "*".  This parameter should be set in case partition and swobjid equals to "*". TI means that someone needs to free all the resources in all the partitions for a particular machine.      
  *   \throw    daq::rmgr::PartitionNotFound    Partition was not found in the RM server that means was not registered.
  *   \throw    daq::rmgr::ObjectNotFound    Software object or computer with requested UID was not found in the RM.
  *   \throw    daq::rmgr::CommunicationException    IPC or system CORBA exception catched and  according messages transferred into this exception.
  *   \throw    daq::rmgr::AcceessDenied  Access manager did not allow to free resources
*/
 //////void	freeAllResources(const std::string&  p = "*", const std::string& swobjid = "*", const std::string& computerName = "*");

//TODO comments

/*!   Free all resources in partition.
 *   \param  p partition name
*/
void freePartition(const std::string&  p );

/*!   Free all resources on commputer for any partition 
 *   \param  computer  computer id
 *   \param log_this_call true if log message should be added about this call
*/
void freeComputer(const std::string& computer, bool log_this_call=true);

/*!   Free one resource in all  partitions.
 *   \param  resource resource id
 *   \param log_this_call true if log message should be added about this call
*/
void freeResource(const std::string& resource, bool log_this_call=true);

/*!   Free one resource in one partition
 *   \param  p partition
 *   \param  resource resource id
 *   \param log_this_call true if log message should be added about this call
*/
void freePartitionResource(const std::string&  p, const std::string& resource, bool log_this_call=true);

/*!   Free one resource in one partition for one computer
 *   \param  resource resource id
 *   \param  computer  computer id
 *   \param  p partition
 *   \param log_this_call true if log message should be added about this call
*/
void freeComputerResource(const std::string& resource, const std::string& computer, const std::string&  p, bool log_this_call=true);

/*!   Free application resources
 *   \param  p partition
 *   \param  swObj software object id
 *   \param  computer  computer id
 *   \param log_this_call true if log message should be added about this call
*/
void freeApplicationResources(const std::string& p, const std::string& swObj, const std::string& computer, bool log_this_call=true);

/*!   Free ALL allocated resources of the process. Resource that was granted for process with the same partition, resource and computer UID will be released in the RM and can be used for other clients. Also all resources  that where allocated with handle handleId also will be free. 
  *   \param handleId  RM handle identifier.
  *   \param  partition   Partition unique ID.
  *    \param  computer Computer unique ID.
  *  \param pid Child process id.
  *   \param log_this_call true if log message should be added about this call
  *   \throw    daq::rmgr::PartitionNotFound    Partition was not found in the RM server that means was not registered.
  *   \throw    daq::rmgr::ObjectNotFound    Software object or computer with requested UID was not found in the RM.
  *    \throw     daq::rmgr::HandleIdNotFound  Handle object with handleId was not found 
  *    \throw     daq::rmgr::ProcessNotFound   Process with pid for which resources were allocated in the partition was not found 
  *   \throw    daq::rmgr::CommunicationException    IPC or system CORBA exception catched and  according messages transferred into this exception.
*/
void freeAllProcessResources( long handleId, const std::string& partition,
                           const std::string& computer, unsigned long pid, bool log_this_call=true );

/*!   Free resources of the process. Resource that was granted for process with the same partition, resource and computer UID will be released in the RM and can be used for other clients.
  *   \param  partition   Partition unique ID.
  *    \param  computer Computer unique ID.
 *  \param pid Child process id.
   \param log_this_call true if log message should be added about this call
*/
void freeProcessResources(
                        const std::string& partition,
                        const std::string& computer,
                        unsigned long pid, bool log_this_call=true);


//different info

/*! Gets the list of the partitions for which any resource was allocated.
  * \return list of partitions name in pointer to daq::rmgr::Rmstr_lstHelper.
  * \throw    daq::rmgr::CommunicationException    IPC or system CORBA exception catched and  according messages transferred into this exception.
*/
daq::rmgr::Rmstr_lstHelper*  getPartitionsList();

/*!   Gets those client names that were used for allocation $rname resource in partition $p.
  *    \param rname   Name of the resource.
  *    \param p   Name of the partition.
  *   \return      List of client names in pointer to daq::rmgr::Rmstr_lstHelper.
*/
daq::rmgr::Rmstr_lstHelper* getResourceOwners(const std::string& rname, const std::string& p);

/*! Gets information about resources that were granted with requested handle indentifier.
  *   \param  handleId   Handle identifier.
  *   \return    Pointer to daq::rmgr::Handle_InfoHelper that keep data about resources that were granted with requested handle indentifier.
  *   \throw   daq::rmgr::HandleIdNotFound    Handle with requested identifier was not found in the RM server.
  *   \throw    daq::rmgr::CommunicationException    IPC or system CORBA exception catched and  according messages transferred into this exception.
*/
daq::rmgr::Handle_InfoHelper* getHandleInfo(long handleId);


/*! Gets information about all allocated resources in the partition.
  *   \param  partition   Name of the partition.
  *    \return    Pointer to daq::rmgr::Allocated_InfoHelper with information about allocted resources.
  *   \throw    daq::rmgr::PartitionNotFound    Partition was not found in the RM server that means was not registered.
  *   \throw    daq::rmgr::ResourceNotFound    Resource with requested name was not found in the RM server.
  *   \throw    daq::rmgr::CommunicationException    IPC or system CORBA exception catched and  according messages transferred into this exception.
*/
daq::rmgr::Allocated_InfoHelper* getPartitionInfo( const std::string& partition );

/*! Gets information about resource in the partition.
  *   \param  partition   Name of the partition.
  *   \param  rname   Name of the resource.
  *    \return    Pointer to daq::rmgr::Allocated_InfoHelper with information about allocted
 resources.
  *   \throw    daq::rmgr::PartitionNotFound    Partition was not found in the RM server that means was not registered.
  *   \throw    daq::rmgr::ResourceNotFound    Resource with requested name was not found in the RM server.
  *   \throw    daq::rmgr::CommunicationException    IPC or system CORBA exception catched and  according messages transferred into this exception.
*/
daq::rmgr::Allocated_InfoHelper* getResourceInfo( const std::string& partition, const std::string& rname);

/*! Gets information about resource.
  *   \param  resource Resource ID.
  *    \return    Pointer to daq::rmgr::Allocated_InfoHelper with information about allocted
 resources.
  *   \throw    daq::rmgr::CommunicationException    IPC or system CORBA exception catched and  according messages transferred into this exception.
*/
daq::rmgr::Allocated_InfoHelper* getFullResInfo(const std::string& resource);

/*! Gets information about all resources that were allocated for computer.
  *   \param  computer Computer name.
  *    \return    Pointer to daq::rmgr::Allocated_InfoHelper with information about allocted
 resources.
  *   \throw    daq::rmgr::CommunicationException    IPC or system CORBA exception catched and  according messages transferred into this exception.
*/
daq::rmgr::Allocated_InfoHelper* getComputerInfo(const std::string& computer);

/*! Gets information about resource that was allocated for computer in the partition.
  *   \param  computer Computer name.
  *   \param  partition   Name of the partition.
  *   \param   resource Resource ID.
  *    \return    Pointer to daq::rmgr::Allocated_InfoHelper with information about allocted resources.
  *   \throw    daq::rmgr::CommunicationException    IPC or system CORBA exception catched and  according messages transferred into this exception.
*/
daq::rmgr::Allocated_InfoHelper* getComputerResourceInfo(const std::string& computer, const std::string& partition, const std::string& resource);

bool checkAccesPermission();

//ONLY FOR TESTS
void rmConfigurationUpdate4MyTest ( daq::rmgr::Config_update & idl_conf );

/*! Gets information about configuration resources and software objects that were loaded into server. 
  *    \return    Pointer to daq::rmgr::Config_updateHelper with according information
*/
daq::rmgr::Config_updateHelper* getConfigInfo();

private:
/*!   Generate call stamp structure: in uint64_t (8 bytes total, to be unique for any client). To be used in case of interruptions in request resources. Structure of stamp:
  *     pid - 2 bytes
  *     ip4 - 1 byte
  *     ip3 - 1 byte
  *     ip2 - 1 byte
  *     ip1 - 1 byte
  *     call id (generated in RM_Client) - 2 bytes
*/
  uint64_t generateCallStamp();
  uint16_t getStampId();
  long getSleepTime( long startTime );

Configuration* getConfiguration( const std::string& dbname );
void setRMData(::Configuration * conf, daq::rmgr::Config_update & rm_conf);
long performRequest(ClientRequest& req); 
};

class RM_ClientImpl {
protected:
  ::rmgr::ResMgr_var RMDB;
  ::IPCPartition m_defpart;
  std::string m_rmDbName;

  /**
   * \throw daq::rmgr::CommunicationException
   */
  void lookupRMDDB();
public:
 RM_ClientImpl();
 RM_ClientImpl(const char *sid);
 void rmConfigurationUpdate ( const std::string& dbname );
 void rmConfigurationUpdate ( Configuration * conf );
 long requestResources (
                        const std::string& partition,
                        const std::string& swobjectid, //id in confdb
                        const std::string& computerName, // name in confdb
                        const std::string& clientName = "NO");
 long requestResources (
                        const std::string& partition,
                        const std::string& swobjectid, //id in confdb
                        const std::string& computerName, 
                        unsigned long mypid,
                        const std::string& clientName = "NO");
 long requestResourceForMyProcess(
                        const std::string& partition,
                        const std::string& resource,
                        const std::string& computer,
                       const std::string& clientName = "NO");
 long requestHWResources(
                        const std::string& partition,
                        const std::string& swId,
                        const std::string& computer,
                        unsigned long pid,
                        const std::string& hwClass,
                        const std::string& hwId,
                        const std::string& clientName);
 long requestHWResourceForProcess(
                        const std::string& partition,
                        const std::string& resource,
                        const std::string& computer,
                        const std::string& hwId,
                        unsigned long pid,
                        const std::string& client = "NO");
 void freeResources (long handleId, bool log_this_call=true);
 void freePartition(const std::string&  p);
 void freeComputer(const std::string& computer, bool log_this_call=true);
 void freeResource(const std::string& resource, bool log_this_call=true);
 void freePartitionResource(const std::string&  p, const std::string& resource, bool log_this_call=true);
 void freeComputerResource(const std::string& resource, const std::string& computer, const std::string&  p, bool log_this_call=true);
 void freeApplicationResources(const std::string& p, const std::string& swObj, const std::string& computer, bool log_this_call=true);
 void freeAllProcessResources( long handleId, const std::string& partition,
                           const std::string& computer, unsigned long pid, bool log_this_call=true );
 void freeProcessResources(
                        const std::string& partition,
                        const std::string& computer,
                        unsigned long pid, bool log_this_call=true);
 daq::rmgr::Rmstr_lstHelper*  getPartitionsList();
 daq::rmgr::Rmstr_lstHelper* getResourceOwners(const std::string& rname, const std::string& p);
 daq::rmgr::Handle_InfoHelper* getHandleInfo(long handleId);
 daq::rmgr::Allocated_InfoHelper* getPartitionInfo( const std::string& partition );
 daq::rmgr::Allocated_InfoHelper* getResourceInfo( const std::string& partition, const std::string& rname);
 daq::rmgr::Allocated_InfoHelper* getFullResInfo(const std::string& resource);
 daq::rmgr::Allocated_InfoHelper* getComputerInfo(const std::string& computer);
 daq::rmgr::Allocated_InfoHelper* getComputerResourceInfo(const std::string& computer, const std::string& partition, const std::string& resource);
 bool checkAccesPermission();
 void rmConfigurationUpdate4MyTest ( daq::rmgr::Config_update & idl_conf );
 daq::rmgr::Config_updateHelper* getConfigInfo();

//private:
  uint64_t generateCallStamp();
  uint16_t getStampId();
  long getSleepTime( long startTime );

Configuration* getConfiguration( const std::string& dbname );
void setRMData(::Configuration * conf, daq::rmgr::Config_update & rm_conf);
long performRequest(ClientRequest& req);
private:
daq::rmgr::CallerInfo getCallerInfo();

};

} // namespace rm
} // namespace daq

#endif				      /* RM_Client_H_INCLUDED */ 
