/*
 * Application.h
 *
 *  Created on: Oct 31, 2012
 *      Author: avolio
 */

/**
 * @file RunControl/Controller/Application.h
 */

#ifndef DAQ_RC_APPLICATION_H_
#define DAQ_RC_APPLICATION_H_

#include "RunControl/Controller/ApplicationSupervisorInfo.h"
#include "RunControl/Controller/ApplicationSelfInfo.h"
#include "RunControl/Common/Clocks.h"

#include <tmgr/tmresult.h>

#include <string>
#include <vector>
#include <set>
#include <map>
#include <memory>
#include <utility>
#include <shared_mutex>
#include <iosfwd>

class Configuration;
namespace daq { namespace core {
    class BaseApplication;
    class Partition;
}}

namespace daq { namespace rc {

class TransitionCmd;
class RunControlBasicCommand;
enum class FSM_STATE : unsigned int;

/**
 * @brief Enumeration for all the available application types
 */
enum class ApplicationType : unsigned int {
	STANDARD = 1,                     //!< STANDARD Item identifying standard applications
	INFRASTRUCTURE = 2,               //!< INFRASTRUCTURE Item identifying infrastructure applications
	ONLINE_SEGMENT_INFRASTRUCTURE = 3,//!< ONLINE_SEGMENT_INFRASTRUCTURE Item identifying infrastructure applications of the Online segment
	RUN_CONTROL = 4,                  //!< RUN_CONTROL Item identifying RC applications
	CHILD_CONTROLLER = 5              //!< CHILD_CONTROLLER Item identifying child controllers
};

/**
 * @brief Functor implementing hash function for ApplicationType.
 */
struct ApplicationType_hasher {
    /**
     * @brief Hash function for ApplicationType
     *
     * @param appType The application type
     * @return Hash value
     */
	std::size_t operator()(const ApplicationType& appType) const;
};

/**
 * @brief Class representing and describing an application managed by the run controller
 *
 * The application's description is made up of two parts:
 * @li A read-only static part including parameters taken from the configuration database (and used to ask the
 *     PMG system to start the corresponding process);
 * @li A dynamic part including information generated by this parent controller (daq::rc::ApplicationSVInfo) and
 *     information provided by the application itself (daq::rc::ApplicationSelfInfo).
 *
 * All the methods of this class are thread-safe and work reliable if out-of-order updates are not a concern (that
 * is always the case when concurrent changes are done on different attributes).
 *
 * When out-of-order updates are a concern, the "batch" @em update methods should be used.
 *
 * @note Derived class are asked to properly override daq::rc::Application::type and to set properties (overwriting default values)
 *       in their <b>constructors</b> calling any of the following methods:
 *       @li void addStartParameters(const std::string& params);
 *       @li void addRestartParameters(const std::string& params);
 *       @li void addEnvironment(const std::string& name, const std::string& value);
 *       @li void addBackupHost(const std::string& hostName);
 *       @li void setExitTimeout(unsigned int timeout);
 *       @li void setStartAt(const std::string& startAt);
 *       @li void setStopAt(const std::string& stopAt);
 *
 * @see daq::rc::ApplicationList
 * @see daq::rc::ApplicationSVInfo
 * @see daq::rc::ApplicationSelfInfo
 */
class Application {
    public:
        /**
         * @brief String for "UserDefined" start-at
         */
        static const std::string START_AT_USERDEFINED;

	public:
        /**
         * @brief Constructor
         *
         * @param db The configuration database
         * @param p The partition object
         * @param appConfig The daq::core::BaseApplication object describing the application
         * @param isParentRoot @em true if the parent of this application is the RootController
         *
         * @throws daq::rc::ConfigurationIssue The application's description cannot be created because some error occurred
         *                                     retrieving information from the configuration database
         */
		Application(Configuration& db,
		            const daq::core::Partition& p,
		            const daq::core::BaseApplication& appConfig,
		            bool isParentRoot = false);

		/**
		 * @brief Copy constructor: deleted
		 */
		Application(const Application&) = delete;

		/**
		 * @brief Move constructor: deleted
		 */
		Application(Application&&) = delete;

		/**
		 * @brief Copy assignment: deleted
		 */
		Application& operator=(const Application&) = delete;

		/**
		 * @brief Move assignment operator: deleted
		 */
        Application& operator=(Application&&) = delete;

        /**
         * @brief Destructor
         */
		virtual ~Application() noexcept;

		/**
		 * @brief It returns the type of this application
		 *
		 * Derived class are asked to override this method.
		 * Default value is daq::rc::ApplicationType::STANDARD.
		 *
		 * @return The type of this application
		 */
		virtual ApplicationType type() const noexcept;

		// Static information (i.e., from the database)
		/**
		 * @brief It returns the name of the partition
		 *
		 * @return The name of the partition
		 */
		std::string getPartition() const;

		/**
		 * @brief It returns the name of the primary host the application should be executed on
		 *
		 * @return The name of the primary host the application should be executed on
		 */
		std::string getHost() const;

		/**
		 * @brief It returns the set of applications this application start-depends on
		 *
		 * @return The set of applications this application start-depends on
		 */
		const std::set<std::string>& getInitDep() const;

		/**
		 * @brief It returns the application's standard input
		 *
		 * @return The application's standard input
		 */
		std::string getInputDevice() const;

		/**
		 * @brief It returns the path where the appliction's standard out and err should be written
		 *
		 * @return The path where the appliction's standard out and err should be written
		 */
		std::string getLogPath() const;

		/**
		 * @brief It returns the application's init timeout (in seconds)
		 *
		 * @return The application's init timeout (in seconds)
		 */
		unsigned int getInitTimeout() const;

		/**
		 * @brief It returns whether the application's standard out and err should be redirected to a file or not
		 *
		 * @return @em true if the application's standard out and err should be redirected to a file
		 */
		bool isLogging() const;

		/**
		 * @brief It returns the name of this application
		 *
		 * @return The name of this application
		 */
		std::string getName() const;

		/**
		 * @brief It returns whether the application can be restarted in the FSM_STATE::RUNNING state or not
		 *
		 * @return @em true if the application can be restarted in the FSM_STATE::RUNNING state
		 */
		bool isRestartableRunning() const;

		/**
		 * @brief It returns the name of the segment this application belongs to
         *
		 * @return The name of the segment this application belongs to
		 */
		std::string getSegment() const;

		/**
		 * @brief It returns the set of applications this application stop-depends on
		 *
		 * @return The set of applications this application stop-depends on
		 */
		const std::set<std::string>& getShutdownDep() const;

		/**
		 * @brief It returns the application's SW object
		 *
		 * This is a particular value allowing the PMG to ask the RM for resources
		 *
		 * @return The application's SW object
		 */
		std::string getSwObject() const;

		/**
		 * @brief It returns the application's working directory
		 *
		 * @return The application's working directory
		 */
		std::string getWorkingDir() const;

		/**
		 * @brief It returns a list of paths where the process's executable may be found
		 *
		 * @return A list of paths where the process's executable may be found
		 */
		const std::vector<std::string>& getBinPath() const;

		/**
		 * @brief It returns the application's environment
		 *
		 * @return The application's environment
		 */
		const std::map<std::string, std::string>& getEnvironment() const;

		/**
		 * @brief It returns the command line parameters the application should be started with
		 *
		 * @return The application's command line parameters
		 */
		std::string getParameters() const;

		/**
		 * @brief It returns the command line parameters to be used when the application is restarted
		 *
		 * @return The command line parameters to be used when the application is restarted
		 */
		std::string getRestartParameters() const;

		/**
		 * @brief It returns the application's exit timeout (in seconds)
		 *
		 * @return The application's exit timeout (in seconds)
		 */
		unsigned int getExitTimeout() const;

		/**
		 * @brief It returns a string representing the FSM transition the application should be started at
		 *
		 * The returned string can be compared to the ones defined the daq::rc::FSMCommands with the only
		 * exception being Application::START_AT_USERDEFINED.
		 *
		 * @return A string representing the FSM transition the application should be started at
		 */
		std::string getStartAt() const;

		/**
		 * @brief It returns a string representing the FSM transition the application should be stopped at
		 *
		 * The returned string can be compared to the ones defined the daq::rc::FSMCommands.
		 *
		 * @return A string representing the FSM transition the application should be stopped at
		 */
		std::string getStopAt() const;

		/**
		 * @brief It returns the set of hosts to be used as back-up
		 *
		 * @return The set of hosts to be used as back-up
		 */
		const std::set<std::string>& getBackupHosts() const;

		/**
		 * @brief It returns @em true if the application is allowed to exit spontaneously
		 *
		 * @return @em true if the application is allowed to exit spontaneously
		 */
		bool isSpontaneousExitAllowed() const;

		/**
		 * @brief It properly calculates the application's start dependencies
		 *
		 * @param appConf The daq::core::BaseApplication object describing this application
		 * @param allApps All the applications belonging to this segment and the child ones
		 */
		void unfoldStartDependencies(const daq::core::BaseApplication& appConf, const std::vector<const daq::core::BaseApplication*>& allApps);

		/**
		 * @brief It properly calculates the application's stop dependencies
		 *
		 * @param appConf The daq::core::BaseApplication object describing this application
		 * @param allApps All the applications belonging to this segment and the child ones
		 */
		void unfoldShutdownDependencies(const daq::core::BaseApplication& appConf, const std::vector<const daq::core::BaseApplication*>& allApps);

		// Dynamic part of the information
		/**
		 * @brief It returns a copy of the current application's daq::rc::ApplicationSVInfo status
		 *
		 * @return A copy of the current application's daq::rc::ApplicationSVInfo status
		 */
		ApplicationSVInfo getApplicationSVInfo() const;

		/**
		 * @brief It sets this application's membership
		 *
		 * @param membership The new application's membership
		 * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSVInfo status and whose
		 *         second element is @em true only if the application's status has been really modified
		 */
		std::pair<ApplicationSVInfo, bool> setMembership(bool membership);

		/**
		 * @brief It sets the application's restarting flag
		 *
		 * @param restarting @em The application's restarting flag
		 * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSVInfo status and whose
         *         second element is @em true only if the application's status has been really modified
		 */
		std::pair<ApplicationSVInfo, bool> setRestarting(bool restarting);

		/**
		 * @brief It sets the host where the application is currently being executed
		 *
		 * @param runningHost The host where the application is currently being executed
		 * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSVInfo status and whose
         *         second element is @em true only if the application's status has been really modified
		 */
		std::pair<ApplicationSVInfo, bool> setRunningHost(const std::string& runningHost);

		/**
		 * @brief It sets the new application's status
		 *
		 * @param status The new application's status
		 * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSVInfo status and whose
         *         second element is @em true only if the application's status has been really modified
		 */
		std::pair<ApplicationSVInfo, bool> setStatus(APPLICATION_STATUS status);

		/**
		 * @brief It sets the application's "not responding" flag
		 *
		 * @param notResponding The new application's "not responding" flag
		 * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSVInfo status and whose
         *         second element is @em true only if the application's status has been really modified
		 */
		std::pair<ApplicationSVInfo, bool> setNotResponding(bool notResponding);

		/**
		 * @brief It sets the current PMG handle for the corresponding process
		 *
		 * @param pmgHandle The current PMG handle for the corresponding process
		 * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSVInfo status and whose
         *         second element is @em true only if the application's status has been really modified
		 */
		std::pair<ApplicationSVInfo, bool> setHandle(const std::string& pmgHandle);

		/**
		 * @brief It sets the current ID (the last part of the PMH handle) for this application
		 *
		 * @param id The current ID (the last part of the PMH handle) for this application
		 * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSVInfo status and whose
         *         second element is @em true only if the application's status has been really modified
		 */
		std::pair<ApplicationSVInfo, bool> setLastID(unsigned int id);

		/**
		 * @brief It sets the exit code for this application
		 *
		 * @param exCode The exit code for this application
		 * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSVInfo status and whose
         *         second element is @em true only if the application's status has been really modified
		 */
		std::pair<ApplicationSVInfo, bool> setExitCode(int exCode);

		/**
		 * @brief It sets the exit signal for this application
		 *
		 * @param signal The exit signal for this application
		 * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSVInfo status and whose
		 *         second element is @em true only if the application's status has been really modified
		 */
        std::pair<ApplicationSVInfo, bool> setExitSignal(int signal);

        /**
         * @brief It sets whether the PMG lookup should be performed for this application regardless its status
         *
         * This is useful when no PMG is found running on some host. In this case the PMG may be re-asked at the next application's start.
         *
         * @param force @em true if the PMG lookup should be forced
         * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSVInfo status and whose
         *         second element is @em true only if the application's status has been really modified
         */
		std::pair<ApplicationSVInfo, bool> setForceLookup(bool force);

		/**
		 * @brief It sets the status of tests performed for this application
		 *
		 * @param status The new test's status
		 * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSVInfo status and whose
         *         second element is @em true only if the application's status has been really modified
		 */
		std::pair<ApplicationSVInfo, bool> setTestStatus(daq::tmgr::TestResult status);

		/**
		 * @brief It atomically updates this application's daq::rc::ApplicationSVInfo status
		 *
		 * @param newInfo The new application's status
		 * @param checkTime Whether to check for out-of-order updates or not
		 * @param timeWindow The minimum difference in time (in milliseconds) in order to accept an out-of-order update
		 */
		void update(const std::shared_ptr<ApplicationSVInfo>& newInfo, bool checkTime = false, unsigned int timeWindow = 1000);

		/**
		 * @brief It returns a copy of the current application's daq::rc::ApplicationSelfInfo status
		 *
		 * @return A copy of the current application's daq::rc::ApplicationSelfInfo status
		 */
		ApplicationSelfInfo getApplicationSelfInfo() const;

		/**
		 * @brief It sets this application's FSM state
		 *
		 * @param state The application's FSM state
		 * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSelfInfo status and whose
         *         second element is @em true only if the application's status has been really modified
		 */
		std::pair<ApplicationSelfInfo, bool> setFSMState(FSM_STATE state);

		/**
		 * @brief It sets whether the application is performing an FSM transition or not
		 *
		 * @param isTransitioning @em true if the application is performing an FSM transition
		 * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSelfInfo status and whose
         *         second element is @em true only if the application's status has been really modified
		 */
		std::pair<ApplicationSelfInfo, bool> setTransitioning(bool isTransitioning);

		/**
		 * @brief It sets this application's internal error
		 *
		 * @param errorReasons Reasons why the application is in error (an empty list clears the error state)
		 * @param badChildren List of children causing the error state
		 * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSelfInfo status and whose
         *         second element is @em true only if the application's status has been really modified
		 */
        std::pair<ApplicationSelfInfo, bool> setInternalError(const std::vector<std::string>& errorReasons,
                                                              const std::vector<std::string>& badChildren);

        /**
         * @brief It sets the last transition command executed by this application
         *
         * @param cmd The last transition command executed by this application (@em nullptr to clear the value)
         * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSelfInfo status and whose
         *         second element is @em true only if the application's status has been really modified
         */
		std::pair<ApplicationSelfInfo, bool> setLastTransitionCommand(const std::shared_ptr<const TransitionCmd>& cmd);

		 /**
		  * @brief It sets the current transition command being executed by this application
		  *
		  * @param cmd The current transition command being executed by this application (@em nullptr to clear the value)
		  * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSelfInfo status and whose
		  *         second element is @em true only if the application's status has been really modified
		  */
		std::pair<ApplicationSelfInfo, bool> setCurrentTransitionCommand(const std::shared_ptr<const TransitionCmd>& cmd);

		/**
		 * @brief It sets the time (in milliseconds) needed to perform the last transition
		 *
		 * @param time The the time (in milliseconds) needed to perform the last transition
		 * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSelfInfo status and whose
         *         second element is @em true only if the application's status has been really modified
		 */
		std::pair<ApplicationSelfInfo, bool> setLastTransitionTime(infotime_t  time);

		/**
		 * @brief It sets the last command executed by this application
		 *
		 * @param cmd The last command executed by this application (@em nullptr to clear the value)
		 * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSelfInfo status and whose
		 *         second element is @em true only if the application's status has been really modified
		 */
		std::pair<ApplicationSelfInfo, bool> setLastCommand(const std::shared_ptr<const RunControlBasicCommand>& cmd);

		/**
		 * @brief It sets this application's busy state
		 *
		 * @param isBusy @em true if the application is busy
		 * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSelfInfo status and whose
         *         second element is @em true only if the application's status has been really modified
		 */
		std::pair<ApplicationSelfInfo, bool> setBusy(bool isBusy);

		/**
		 * @brief It sets the "fully initialized" status of this application
		 *
		 * "Fully initialized" means that the application has completed a (re)start procedure (i.e., reached the proper state)
		 *
		 * @param done @em true if the application is fully initialized
		 * @return A pair whose first element is a copy of the current application's daq::rc::ApplicationSelfInfo status and whose
         *         second element is @em true only if the application's status has been really modified
		 */
        std::pair<ApplicationSelfInfo, bool> setFullyInitialized(bool done);

        /**
         * @brief It atomically updates this application's daq::rc::ApplicationSelfInfo status
         *
         * @param newInfo The new application's status
         * @param checkTime Whether to check for out-of-order updates or not
         * @param timeWindow The minimum difference in time (in milliseconds) in order to accept an out-of-order update
         */
		void update(const std::shared_ptr<ApplicationSelfInfo>& newInfo, bool checkTime = false, unsigned int timeWindow = 1000);

		/**
		 * @brief It updates this application's dynamic information with the one coming from the @em app application
		 *
		 * This method is meant to be used during a database reload, where the description of the application in the
		 * configuration database may have been changed but, of course, the application's dynamic information cannot
		 * get lost. In that case this application represents the new application's description, while @em app is the
		 * old instance with the dynamic information to be kept.
		 *
		 * @param app The application whose dynamic information is used to update this application
		 */
		void update(const Application& app);

		/**
		 * @brief Functor implementing the comparator function for Application::ApplicationSet
		 */
		struct Comparator {
		    /**
		     * @brief Comparator function for Application::ApplicationSet
		     *
		     * @param e1 First Application
		     * @param e2 Second Application
		     * @return <em>e1->getName() < e2->getName()</em>
		     */
		    bool operator()(const std::shared_ptr<const Application>& e1, const std::shared_ptr<const Application>& e2) const {
		        return (e1->getName() < e2->getName());
		    }
		};

		/**
		 * @brief Typedef for a set of daq::rc::Application's shared pointer ordered by application's name
		 */
		typedef std::set<std::shared_ptr<Application>, Application::Comparator> ApplicationSet;

	protected:
		/**
		 * @brief It adds the @em params parameters to the list of start parameters
		 *
		 * @param params Additional parameters
		 */
		void addStartParameters(const std::string& params);

		/**
		 * @brief It adds the @em params parameters to the list of re-start parameters
		 *
		 * @param params Additional parameters
		 */
		void addRestartParameters(const std::string& params);

		/**
		 * @brief It adds a variable to this application's environment
		 *
		 * @param name The name of the environment variable
		 * @param value The value of the environment variable
		 */
		void addEnvironment(const std::string& name, const std::string& value);

		/**
		 * @brief It sets the exit timeout (in seconds)
		 *
		 * @param timeout The timeout's value (in seconds)
		 */
		void setExitTimeout(unsigned int timeout);

		/**
		 * @brief It sets when this application should be started
		 *
		 * @em startAt should be one of the strings defined in daq::rc::FSMCommands or Application::START_AT_USERDEFINED.
		 *
		 * @param startAt String representing the FSM transition the application should be started at (or Application::START_AT_USERDEFINED)
		 */
		void setStartAt(const std::string& startAt);

		/**
		 * @brief It sets when this application should be stopped
		 *
		 * @em stopAt should be one of the strings defined in daq::rc::FSMCommands.
		 *
		 * @param stopAt String representing the FSM transition the application should be started at
		 */
		void setStopAt(const std::string& stopAt);

	private:
        class DynamicInfo {
            public:
		        DynamicInfo() = default;

		        std::shared_ptr<ApplicationSVInfo> m_app_sv_info;
                std::shared_mutex m_app_sv_mutex;

                std::shared_ptr<ApplicationSelfInfo> m_app_self_info;
                std::shared_mutex m_app_self_mutex;
        };

        bool m_spontaneous_exit_allowed;
		const std::string m_partition;
		const std::string m_name;
		const std::string m_host;
		const std::string m_segment;
		std::string m_parameters;
		std::string m_restart_parameters;
		bool m_logging;
		std::string m_log_path;
		std::string m_input_device;
		unsigned int m_init_timeout;
		unsigned int m_exit_timeout;
		std::string m_working_dir;
		std::string m_start_at;
		std::string m_stop_at;
		bool m_restartable_running;
		std::set<std::string> m_init_dep;
		std::set<std::string> m_shutdown_dep;
		std::string m_sw_object;
		std::vector<std::string> m_bin_path;
		std::map<std::string, std::string> m_env;
		std::set<std::string> m_backup_hosts;

		std::shared_ptr<DynamicInfo> m_dyn_info;
};

}}

#endif /* DAQ_RC_APPLICATION_H_ */
