/*
 * RunControllerFSMOperations.cxx
 *
 *  Created on: Jan 9, 2013
 *      Author: avolio
 */

#include "RunControl/Controller/RunControllerFSMOperations.h"
#include "RunControl/Controller/DVSController.h"
#include "RunControl/Controller/ApplicationController.h"
#include "RunControl/Controller/ApplicationList.h"
#include "RunControl/Controller/Application.h"
#include "RunControl/Controller/ApplicationSupervisorInfo.h"
#include "RunControl/Controller/ApplicationSelfInfo.h"
#include "RunControl/Controller/InformationPublisher.h"
#include "RunControl/FSM/FSMStates.h"
#include "RunControl/FSM/FSMCommands.h"
#include "RunControl/Common/Exceptions.h"
#include "RunControl/Common/ThreadPool.h"
#include "RunControl/Common/RunControlCommands.h"
#include "RunControl/Common/OnlineServices.h"
#include "RunControl/Common/RunControlBasicCommand.h"

#include <ers/Assertion.h>
#include <ers/Issue.h>
#include <ers/LocalContext.h>
#include <ers/ers.h>

#include <boost/scope_exit.hpp>

#include <chrono>
#include <thread>
#include <map>
#include <string>
#include <utility>
#include <algorithm>
#include <random>
#include <future>
#include <iterator>
#include <ostream>
#include <set>
#include <vector>


namespace daq { namespace rc {

RunControllerFSMOperations::RunControllerFSMOperations(const std::shared_ptr<ApplicationController>& appCtrl,
                                                       const std::shared_ptr<DVSController>& dvsCtrl,
                                                       const std::shared_ptr<ApplicationList>& appList,
                                                       const std::shared_ptr<InformationPublisher>& infoPublisher)
    :   RunControlTransitionActions(), m_is_interrupted(false), m_app_ctrl(appCtrl), m_dvs_ctrl(dvsCtrl), m_app_list(appList),
        m_info_publisher(infoPublisher), m_cmd_sender(OnlineServices::instance().getIPCPartition(), OnlineServices::instance().applicationName()),
        m_tp(new ThreadPool(5))
{
}

RunControllerFSMOperations::~RunControllerFSMOperations() noexcept
{
}

void RunControllerFSMOperations::enterFSM(const TransitionCmd& cmd) {
    performTransition(cmd, false, false);
}

void RunControllerFSMOperations::initialize(const TransitionCmd& cmd) {
    performTransition(cmd, false, false);
}

void RunControllerFSMOperations::configure(const TransitionCmd& cmd) {
    performTransition(cmd);
}

void RunControllerFSMOperations::connect(const TransitionCmd& cmd) {
    performTransition(cmd);
}

void RunControllerFSMOperations::prepareForRun(const TransitionCmd& cmd) {
    performTransition(cmd);
}

void RunControllerFSMOperations::stopROIB(const TransitionCmd& cmd) {
    performTransition(cmd);
}

void RunControllerFSMOperations::stopDC(const TransitionCmd& cmd) {
    performTransition(cmd);
}

void RunControllerFSMOperations::stopHLT(const TransitionCmd& cmd) {
    performTransition(cmd);
}

void RunControllerFSMOperations::stopRecording(const TransitionCmd& cmd) {
    performTransition(cmd);
}

void RunControllerFSMOperations::stopGathering(const TransitionCmd& cmd) {
    performTransition(cmd);
}

void RunControllerFSMOperations::stopArchiving(const TransitionCmd& cmd) {
    performTransition(cmd);
}

void RunControllerFSMOperations::disconnect(const TransitionCmd& cmd) {
    performTransition(cmd);
}

void RunControllerFSMOperations::unconfigure(const TransitionCmd& cmd) {
    performTransition(cmd);
}

void RunControllerFSMOperations::subTransition(const SubTransitionCmd& cmd) {
    performTransition(cmd);
}

void RunControllerFSMOperations::shutdown(const TransitionCmd& cmd) {
    performTransition(cmd, false, false);
}

void RunControllerFSMOperations::userBroadcast(const UserBroadcastCmd& cmd) {
    // Self-transition: source and destination states are the same
    FSM_STATE currState = FSMStates::stringToState(cmd.fsmSourceState());
    switch(currState) {
        case FSM_STATE::NONE:
        {
            ERS_LOG("The UserBroadcast command is not propagated to any child because the current FSM state is NONE");
            break;
        }
        default:
        {
            performTransition(cmd);
            break;
        }
    }
}

void RunControllerFSMOperations::resynch(const ResynchCmd& cmd) {
    performTransition(cmd);
}

void RunControllerFSMOperations::reloadDB(const TransitionCmd& cmd) {
    // Reload supported only in NONE
    ERS_ASSERT(cmd.fsmSourceState() == FSMStates::NONE_STATE);

    try {
        // Here basically we reset the converters
        OnlineServices::instance().reloadConfiguration(false);

        // Build the new application list
        ApplicationList newAppList;
        newAppList.load(OnlineServices::instance().getConfiguration(), OnlineServices::instance().getPartition(), OnlineServices::instance().segmentName());

        ERS_DEBUG(0, "The new application list has been created");

        // Get all the applications in the current application list
        const auto& oldApps = m_app_list->getAllApplications();

        // Divide the applications into three categories:
        // Applications removed: they are not part any more of the application list
        // Applications added: these are new applications
        // Applications kept: applications still present after the database reload
        Application::ApplicationSet removedApps;
        Application::ApplicationSet addedApps;
        Application::ApplicationSet keptApps;

        auto appChanged = [](const std::shared_ptr<Application>& oldApp_, const std::shared_ptr<Application>& newApp_) -> bool
                            {
                                return (oldApp_->getEnvironment() != newApp_->getEnvironment() ||
                                        oldApp_->getParameters() != newApp_->getParameters() ||
                                        oldApp_->getRestartParameters() != newApp_->getRestartParameters() ||
                                        oldApp_->getHost() != newApp_->getHost() ||
                                        oldApp_->getBackupHosts() != newApp_->getBackupHosts());
                            };

        // Loop over the current list in order to identify applications been removed or kept
        for(const auto& oldApp_ : oldApps) {
            const auto& newApp_ = newAppList.getApplication(oldApp_->getName());
            if(newApp_.get() != nullptr) {
                keptApps.insert(oldApp_);

                // Update the dynamic part of the info: starting from now the Applications
                // in the new list share the dynamic information with the Applications in the
                // old list: in such a way no information is lost when the lists are swapped
                // and there is no additional need of synchronization
                newApp_->update(*oldApp_);

                // Restarting the Online segment applications is incredibly dangerous...
                if((newApp_->type() == ApplicationType::ONLINE_SEGMENT_INFRASTRUCTURE) && (appChanged(oldApp_, newApp_) == true)) {
                    const std::string& msg = std::string("this is an online segment infrastructure application whose description has been modified ") +
                                            "with the database reload; the partition needs to be restarted if the changes have to be applied";
                    ers::warning(daq::rc::ApplicationGenericIssue(ERS_HERE, newApp_->getName(), msg));
                }
             } else {
                removedApps.insert(oldApp_);
            }
        }

#ifndef ERS_NO_DEBUG
        auto printApps = [](const Application::ApplicationSet& apps) -> std::string
                           {
                                std::string s;
                                for(const auto& app_ : apps) {
                                    s += app_->getName() + " ";
                                }
                                return s;
                           };
#endif

        ERS_DEBUG(1, "Removed applications:\n" << printApps(removedApps));
        ERS_DEBUG(1, "Kept applications:\n" << printApps(keptApps));

        // Now determine the new applications
        const auto& newApps = newAppList.getAllApplications();
        std::set_difference(newApps.begin(), newApps.end(),
                            keptApps.begin(), keptApps.end(),
                            std::inserter(addedApps, addedApps.begin()),
                            addedApps.key_comp());

        ERS_DEBUG(1, "New applications:\n" << printApps(addedApps));

        // 1. Stop applications: they are the ones been removed
        {
            if(removedApps.empty() == false) {
                std::vector<std::string> av;
                for(const auto& app_ : removedApps) {
                    av.push_back(app_->getName());
                }

                StartStopAppCmd stopCmd(StartStopAppCmd::is_stop_app_cmd, av);
                stopCmd.currentState(FSMStates::stringToState(cmd.fsmSourceState()));
                stopCmd.destinationState(FSMStates::stringToState(cmd.fsmDestinationState()));
                stopCmd.transitionCommand(FSMCommands::stringToCommand(cmd.fsmCommand()));

                const auto& reallyStoppedApps = m_app_ctrl->stopApps(stopCmd);

                // Remove the information of removed applications from IS
                for(const auto& app_ : reallyStoppedApps) {
                    m_info_publisher->removeProcessInfo(app_->getName());
                }

                if(removedApps != reallyStoppedApps) {
                    // TODO: this is probably too much
                    throw daq::rc::DBReloadFailure(ERS_HERE, "some no more existing applications could not be properly stopped");
                }

                ERS_DEBUG(0, "Removed applications have been stopped");
            }
        }

        // Applications to be started: they are the one that have been added
        //
        // NOTE: swap the application list before, otherwise new applications will be unknown!!!
        m_app_list->swap(newAppList);

        // Reload the DVS core, so that new applications can be properly tested
        m_dvs_ctrl->reload(addedApps, removedApps);

        {
            std::vector<std::string> av;
            for(const auto& app_ : addedApps) {
                // This works only because the reload is done in NONE (and then only by the Root)
                // If this assumption is no more valid, then the applications should be started
                // taking into account their types (i.e., first the OS, then infrastructure, etc)
                // and the state in which the reload is done (in NONE the only transition done is INIT_FSM).
                if(app_->getStartAt() == FSMCommands::INIT_FSM_CMD) {
                    av.push_back(app_->getName());
                }
            }

            if(av.empty() == false) {
                StartStopAppCmd startCmd(StartStopAppCmd::is_start_app_cmd, av);
                startCmd.currentState(FSMStates::stringToState(cmd.fsmSourceState()));
                startCmd.destinationState(FSMStates::stringToState(cmd.fsmDestinationState()));
                startCmd.transitionCommand(FSMCommands::stringToCommand(cmd.fsmCommand()));

                m_app_ctrl->startApps(startCmd);

                ERS_DEBUG(0, "New added applications have been started");
            }
        }
    }
    catch(daq::rc::DBReloadFailure&) {
        throw;
    }
    catch(ers::Issue& ex) {
        throw daq::rc::DBReloadFailure(ERS_HERE, ex.message(), ex);
    }
}


void RunControllerFSMOperations::performTransition(const TransitionCmd& cmd, bool toApplications, bool toControllers) {
    if(cmd.isRestart() == false) {
        std::vector<std::future<void>> tasks;
        BOOST_SCOPE_EXIT(&tasks)
        {
            for(auto& t : tasks) {
                try {
                    t.get();
                }
                catch(std::exception&) {
                }
            }
        }
        BOOST_SCOPE_EXIT_END

        try {
            /// 1 --> Start applications
            interrupted(cmd);
            ERS_DEBUG(0, "Starting applications for transition " << cmd.fsmCommand());

            // Remember that this returns only when all the applications to be started
            // have been actually started (or put out of membership) - and it returns
            // all of them (regardless the membership status)
            Application::ApplicationSet&& startedApps = m_app_ctrl->startApps(cmd);

            ERS_DEBUG(1, "Applications for transition " << cmd.fsmCommand() << " are started");

            /// 2 --> Transition command propagation
            interrupted(cmd);
            ERS_DEBUG(0, "Now propagating the transition command " << cmd.fsmCommand());

            std::vector<std::shared_ptr<Application>> apps;
            if(toControllers == true) {
                const auto& cc = m_app_list->getApplications(ApplicationType::CHILD_CONTROLLER);
                apps.insert(apps.end(), cc.begin(), cc.end());
            }

            if(toApplications == true) {
                const auto& rc = m_app_list->getApplications(ApplicationType::RUN_CONTROL);
                apps.insert(apps.end(), rc.begin(), rc.end());
            }

            // Do not propagate the transition to the just started applications
            // (they reach the proper state by themselves)
            apps.erase(std::remove_if(apps.begin(), apps.end(), [&startedApps](const std::shared_ptr<Application>& app_) -> bool
                                                                              {
                                                                                    return startedApps.find(app_) != startedApps.end();
                                                                              }
                                     ),
                       apps.end());

            // Propagate the command in a random fashion (avoid saturation of single
            // node where several applications are running)
            std::shuffle(apps.begin(), apps.end(), std::mt19937(std::random_device()()));

            // Keep track of the previous transition command
            std::map<std::string, std::shared_ptr<const TransitionCmd>> lastTrCmdMap;

            // In this loop the transition command is sent to all the applications with membership IN
            // Applications out of membership are removed from the list
            auto dispatch = [&lastTrCmdMap, &tasks, &cmd, this](const std::shared_ptr<Application>& app_) -> bool
                            {
                                interrupted(cmd);

                                // Decide whether to propagate or not the command to the application
                                // The command must not be propagated if:
                                // - the application is out of membership
                                // - the application has already executed or is being executed the same command (for instance
                                //   a controller dying in the middle of a transition does not know the applications its previous
                                //   instance sent a command to).
                                const bool membership = app_->getApplicationSVInfo().isMembership();

                                const ApplicationSelfInfo& appInfo = app_->getApplicationSelfInfo();
                                const auto& appCurrTrCmd = appInfo.getCurrentTransitionCommand();
                                const auto& cmdToCmpr = (appCurrTrCmd.get() != nullptr) ? appCurrTrCmd : appInfo.getLastTransitionCommand();
                                const bool alreadyDone = (cmdToCmpr.get() == nullptr) ? false : ((*cmdToCmpr == cmd) && (cmdToCmpr->uid() == cmd.uid()));

                                const bool propagate = ((membership == true) && (alreadyDone == false));

                                if(propagate == true) {
                                    // Take note of the last transition done before sending the new transition command
                                    lastTrCmdMap.insert(std::make_pair(app_->getName(), appInfo.getLastTransitionCommand()));

                                    auto&& fut = m_tp->submit<void>([this, app_, cmd]()
                                                                    {
                                                                        short nRetries = 2;

                                                                        // The membership is checked again in order to avoid some command to
                                                                        // be re-sent even if the application has been put out of membership
                                                                        while((app_->getApplicationSVInfo().isMembership() == true) && (nRetries-- > 0)) {
                                                                            try {
                                                                                 interrupted(cmd);

                                                                                 m_cmd_sender.makeTransition(app_->getName(), cmd);

                                                                                 ERS_DEBUG(1, "Transition command " << cmd.fsmCommand() <<
                                                                                              " sent to application " << app_->getName());

                                                                                 break;
                                                                            }
                                                                            catch(daq::rc::InterruptedException&) {
                                                                                // Just stop doing it...
                                                                                break;
                                                                            }
                                                                            catch(ers::Issue& ex) {
                                                                                daq::rc::TransitionPropagationFailure issue(ERS_HERE,
                                                                                                                            cmd.fsmCommand(),
                                                                                                                            app_->getName(),
                                                                                                                            ex);

                                                                                if(nRetries == 0) {
                                                                                    ers::error(issue);
                                                                                } else {
                                                                                    issue.wrap_message("", ". Another attempt will be done.");
                                                                                    ers::warning(issue);

                                                                                    std::this_thread::sleep_for(std::chrono::milliseconds(50));
                                                                                }
                                                                            }
                                                                        }
                                                                    });

                                    tasks.push_back(std::move(fut));
                                } else {
                                    if(membership == false) {
                                        ERS_DEBUG(1, "Transition command " << cmd.fsmCommand() << " not sent to application " << app_->getName() << " because out of membership");
                                    } else if(alreadyDone) {
                                        ERS_LOG("Transition command " << cmd.fsmCommand() << " not sent to application " <<
                                                app_->getName() << " because already executed or being executed");
                                    }
                                }

                                return !propagate;
                            };

            apps.erase(std::remove_if(apps.begin(), apps.end(), dispatch), apps.end());

            /// 3 --> Waiting for the started application to reach the right state
            ERS_DEBUG(1, "Waiting for started application to reach the right state...");

            while(startedApps.empty() == false) {
                interrupted(cmd);

                for(auto it = startedApps.begin(); it != startedApps.end();) {
                    const ApplicationSVInfo& appSVInfo = (*it)->getApplicationSVInfo();
                    const ApplicationSelfInfo& appSelfInfo = (*it)->getApplicationSelfInfo();
                    bool isAppOK = false;

                    ERS_DEBUG(3, "Application " << (*it)->getName() << " is in the FSM state " << FSMStates::stateToString(appSelfInfo.getFSMState()) <<
                                 " with membership " << appSVInfo.isMembership() << " and its transitioning state is " << appSelfInfo.isTransitioning());

                    if(appSVInfo.isMembership() == false) {
                        // If the app is out of membership, then just forget about it
                        isAppOK = true;
                    } else {
                        ApplicationType appType = (*it)->type();
                        switch(appType) {
                            case ApplicationType::CHILD_CONTROLLER:
                            case ApplicationType::RUN_CONTROL: {
                                // RC apps and controllers are requested to be in the right state, not transitioning,
                                // not in error and with their test status equal to PASSED
                                if((FSMStates::stateToString(appSelfInfo.getFSMState()) == cmd.fsmDestinationState()) &&
                                   (appSelfInfo.isTransitioning() == false) &&
                                   (appSelfInfo.isInternalError() == false) &&
                                   (appSVInfo.isRestarting() == false))
                                {
                                    isAppOK = true;
                                }

                                break;
                            }
                            default: {
                                isAppOK = true;
                                break;
                            }
                        }
                    }

                    if(isAppOK == true) {
                        ERS_DEBUG(2, "Start procedure for application " << (*it)->getName() << " is done");
                        startedApps.erase(it++);
                    } else {
                        ++it;
                    }
                }

                std::this_thread::sleep_for(std::chrono::milliseconds(1));
            }

            /// 4 --> Waiting for the applications that received the transition command to reach the right state
            ERS_DEBUG(1, "Waiting for applications that received the transition command to reach the right state...");

            auto done = [&lastTrCmdMap, &cmd](const std::shared_ptr<Application>& app_) -> bool
                        {
                            const ApplicationSelfInfo& appSelfInfo = app_->getApplicationSelfInfo();
                            const ApplicationSVInfo& appSVInfo = app_->getApplicationSVInfo();

                            ERS_DEBUG(3, "Application " << app_->getName() << " with membership " << appSVInfo.isMembership() <<
                                         " is in the FSM state " << FSMStates::stateToString(appSelfInfo.getFSMState()) <<
                                         " and its transitioning state is " << appSelfInfo.isTransitioning());

                            // For a transition between two different states, the transition is considered as done if the application:
                            // - is OUT of membership
                            // OR
                            // - has reached the destination state;
                            // - is no more transitioning;
                            // - is not in error;
                            // - is not in the restarting phase;
                            //
                            // For a self transition (i.e., a transition that does not change the FSM state), the transition is considered
                            // as done if the application:
                            // - is OUT of membership
                            // OR
                            // - is no more transitioning;
                            // - is not in error;
                            // - is not in the restarting phase;
                            // - the current last executed transition command is not null (it can be null for the first received command)
                            //   and equal to the current transition command;
                            // - the previous last executed command:
                            //      - null (this is the very first transition command received by the application);
                            //      - different (semantically) from the current last received command or with a different id (this removes the possibility
                            //        to have a false positive if the same command is issued two times sequentially).

                            const TransitionCmd* const oldLastTrCmd = (lastTrCmdMap.find(app_->getName())->second).get();
                            const TransitionCmd* const newLastTrCmd = appSelfInfo.getLastTransitionCommand().get();

                            const FSM_STATE trSrc = FSMStates::stringToState(cmd.fsmSourceState());
                            const FSM_STATE trDst = FSMStates::stringToState(cmd.fsmDestinationState());

                            // At a first glance it may look that the check on the UID may be sufficient
                            // But in that case we are more exposed to clashes in the UID: two commands may actually be
                            // semantically different but with the same UID (although the probability to have
                            // two identical UIDs during the life of the applications is really null)
                            const bool transitionDone = (appSVInfo.isMembership() == false) ||
                                                        ((appSelfInfo.isTransitioning() == false) && (appSVInfo.isRestarting() == false) && (appSelfInfo.isInternalError() == false) && (appSelfInfo.getFSMState() == trDst) &&
                                                        ((trDst != trSrc) || ((newLastTrCmd != nullptr) && (*newLastTrCmd == cmd) &&
                                                        ((oldLastTrCmd == nullptr) ? true : ((!(*oldLastTrCmd == *newLastTrCmd)) || (oldLastTrCmd->uid() != newLastTrCmd->uid()))))));

                            if(transitionDone == true) {
                                ERS_DEBUG(2, "Transition done for application " << app_->getName());
                            }

                            return transitionDone;
                        };

            while(apps.empty() == false) {
                interrupted(cmd);
                apps.erase(std::remove_if(apps.begin(), apps.end(), done), apps.end());
                std::this_thread::sleep_for(std::chrono::milliseconds(1));
            }

            ERS_DEBUG(0, "The transition " << cmd.fsmCommand() << " has been propagated and all the relevant applications have reached the right state");

            /// 5 --> Stop applications
            interrupted(cmd);
            ERS_DEBUG(0, "Stopping applications for transition " << cmd.fsmCommand());

            m_app_ctrl->stopApps(cmd);

            ERS_DEBUG(1, "Applications for transition " << cmd.fsmCommand() << " are stopped");
        }
        catch(daq::rc::InterruptedException& ex) {
            throw daq::rc::TransitionInterrupted(ERS_HERE, cmd.fsmCommand(), ex);
        }
    } else {
        ERS_LOG("No actions taken for transition \"" << cmd.fsmCommand() << "\" because the controller is restarting");
    }
}

void RunControllerFSMOperations::interrupt(bool shouldInterrupt) {
    m_is_interrupted = shouldInterrupt;
}

void RunControllerFSMOperations::interrupted(const TransitionCmd& cmd) const {
    if(m_is_interrupted == true) {
        throw daq::rc::TransitionInterrupted(ERS_HERE, cmd.fsmCommand());
    }
}

}}
