/*
 *  ThreadPool.h
 *  Implementation of a simple thread pool class.
 *  Created by Kazarov A. on 12.11.2012.
 *  Copyright 2012 CERN. All rights reserved.
 *
 */

#ifndef TM_TPOOL_H
#define TM_TPOOL_H

#include <stdlib.h>
#include <thread>
#include <atomic>
#include <mutex>
#include <condition_variable>
#include <vector>
#include <functional>

#include <tbb/concurrent_queue.h>

#include <ers/ers.h>

namespace daq {
namespace tm {

class ThreadPool ;

// needed to construct threads
class WorkerThread {
public:
    WorkerThread(ThreadPool &s) : pool(s) { ERS_DEBUG(2, "Worker thread created" ) ; }
    void operator()();
    ThreadPool &pool;
};


       /*! 	\brief Thread pool implementaion, used by TM.
	 	\ingroup public
	 	\sa
       */
class ThreadPool {
friend class WorkerThread ;

public:
    ThreadPool(size_t pool_size = 24) ;
    ~ThreadPool() ;

    size_t get_tasks_count() const { return tasks_queue.unsafe_size() ; }

    template<typename F>
    void schedule(F) ; // schedule async task execution, just adds the task F to the queue

    template<typename... Args>
    void schedule(Args&&...) ;

private:
 
    // need to keep track of threads so we can join them
    std::vector< std::thread > worker_threads ;
  
    // the task queue
    tbb::concurrent_queue< std::function<void() > > tasks_queue ;
 
    // synchronization
    std::mutex queue_mutex;
    std::condition_variable condition;  
    std::atomic<bool> stop;
};

// add new work item to the queue
// F: anything std::function<void()> can be constructed from, a plenty of options
template<typename F>
void ThreadPool::schedule(F f)
{
    tasks_queue.push(std::move(f)) ; // force to use push(&&) semantic

ERS_DEBUG(3, "Job scheduled") ;

    // wake up one thread
    std::unique_lock<std::mutex> lock(queue_mutex) ; // to avoid notify to be sent before a worker thread enter in wait() 
    condition.notify_one(); // TODO: optimze: notify only if queue was empty before the push here
ERS_DEBUG(3, "Thread awaked") ;
}

// emplaced construction of a callable
template<typename... Args>
void ThreadPool::schedule(Args&&... args)
{
    tasks_queue.emplace(args...);

    // wake up one thread
    std::unique_lock<std::mutex> lock(queue_mutex) ; // to avoid notify to be sent before a worker thread enter in wait() 
    condition.notify_one(); // TODO: optimze: notify only if queue was empty before the push here
}
}} // daq::tm

#endif
