/**
 * CHANGELOG
 * May 2017 - Daniel Zheng - Created PredAdapter.cpp to combine HLT and L1 Predadapters
 * May 2018 - Andy Aukerman - Cleanup, small adapation to allow multiparamater fits, fixed m_triggerDB and checkIn functions
 * August 2019 - Nick Felice - Functions for new IS data format, global offset, reformatting & organization of code
 *
 * The code is, for the most part, organized into 4 sections:
 * 1. Setup (configurations, PredAdapter::Run(), etc.)
 * 2. Callbacks (callback functions for each IS type)
 * 3. Calculations (basic formatting and prediction calculations)
 * 4. Publishing (publish results into IS)
 */

#include "PredAdapter.h"

#include "TRP/Utils.h"
#include <ipc/core.h>
#include <iostream>
#include "ers/ers.h"
#include <time.h>
#include <regex.h>
#include "dqmf/is/Tag.h"
#include "is/inforeceiver.h"
#include "ers/ers.h"
#include "is/serveriterator.h"
#include <errno.h>
#include <stdlib.h>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>
#include <sys/stat.h>

ERS_DECLARE_ISSUE(predadapter, Issue, "conversion issue", )

namespace pt = boost::property_tree;

namespace {
    bool fileExists(const std::string& name) {
        struct stat buffer;   
        return (stat (name.c_str(), &buffer) == 0); 
    }
}


/**
 * Constructor
 * Creates the new object, and sets a large amount of parameters
 */
PredAdapter::PredAdapter(std::string partition_name, std::string server_name, std::string object_name,
                        std::string triggerDBAlias, std::string L1_name_out, std::string hlt_name_out,
                        float bunch_factor,
                        std::string trigger_groups,
                        float flagOk, float flagWarning, float flagError, bool debug) :
    m_partitionName(partition_name),
    m_objectName(object_name),
    m_serverName(server_name),
    m_l1OutputPartition(L1_name_out), 
    m_hltOutputPartition(hlt_name_out),
    m_triggerDB(triggerDBAlias),
    m_trigger_groups(trigger_groups),
    m_bunchFactor(bunch_factor),
    m_flagOk(flagOk),
    m_flagWarning(flagWarning),
    m_flagError(flagError),
    m_debug(debug)
{

    // Begin Constructor Code
    std::string alertprefix = "Xmon()::PredAdapter---> ";

    
    std::cout << alertprefix << "IRH the debug set is " << m_debug
              << "; number of bunch crossings for prediction file is "<< bunch_factor
              << "; flag set is: "<<m_flagOk<<", "<<m_flagWarning<<", "<<m_flagError<<std::endl;

    // LVL1 Y-Labels
    y_L1.resize(pYLabels_L1_size);
    y_L1.at(PS_L1)        = "PS" ;
    y_L1.at(TBP_L1)       = "TBP";
    y_L1.at(TAP_L1)       = "TAP";
    y_L1.at(PBP_L1)       = "PBP";
    y_L1.at(PAP_L1)       = "PAP";
    y_L1.at(DIFF_L1)      = "DIFF";
    y_L1.at(NSIGMA_L1)    = "NSIGMA";
    y_L1.at(FLAG_L1)      = "FLAG";
    y_L1.at(RATIO_L1)     = "RATIO";
    y_L1.at(SRATIO_L1)    = "SRATIO";
    y_L1.at(P1_L1)        = "Par1";
    y_L1.at(P0_L1)        = "Par0";
    y_L1.at(OTFPRED_L1)   = "OTFPRED";

    // HLT Y-Labels
    y_HLT.resize(pYLabels_HLT_size);
    y_HLT.at(OUTPUT_HLT)    = "output" ;
    y_HLT.at(PS_HLT)        = "ps" ;
    y_HLT.at(PRED_HLT)      = "pred";
    y_HLT.at(DIFF_HLT)      = "diff";
    y_HLT.at(NSIGMA_HLT)    = "nsigma";
    y_HLT.at(FLAG_HLT)      = "flag";
    y_HLT.at(RATIO_HLT)     = "ratio";
    y_HLT.at(SRATIO_HLT)    = "sratio";
    y_HLT.at(P1_HLT)        = "Par1";
    y_HLT.at(P0_HLT)        = "Par0";
    y_HLT.at(OTFPRED_HLT)   = "otfpred";

}

/**
 * @brief Online Luminosity Calculator configuration
 * 
 * @param lumiISServer 
 * @return * Online 
 */
void PredAdapter::setLumiPartition(std::string olcPartition, std::string lumiISServer) {
    ERS_DEBUG(0, "In Configuration");
    m_lumiPartition = olcPartition;
    m_olcIsServer = lumiISServer;
}

/**
 * PredConfigure
 * Prepares adapter for downloading Trigger DB info and
 * subscribing to IS
 */
bool
PredAdapter::PredConfigure(const std::string & predictionFile)
{
    std::string alertprefix = "Xmon()::PredConfigure---> ";

    //set vector of trigger items
    //Flag::-1: can't make predict, -2: rate=0, no make flag,
    //3: OMG fatal!, 2: error, 1: warning, 0: Okay
    std::string     name;
    std::string     function;
    std::string     equation;
    std::string     ratio_equation;
    double          pred_parameters[10];
    double          ratio_pred_parameters[10];

    std::cout << alertprefix<< "Reading prediction file = " << predictionFile << std::endl;

    if ( !fileExists(predictionFile)) {
        std::cout << alertprefix << "The prediction file does not exist!" << std::endl;
        std::cout << alertprefix << "Aborting PredAdapter Configuration" << std::endl;
        return failure;
    }

    pt::ptree pt;
    pt::read_json(predictionFile, pt);

    for (pt::ptree::value_type& trigger: pt.get_child("LVL1_ratios")) {
        trigger_ratio ratio;
        ratio.nominator = trigger.second.get<std::string>("first");
        ratio.denominator = trigger.second.get<std::string>("second");
        ratio_equation = trigger.second.get<std::string>("function");
        std::cout << alertprefix << "Ratio for " << ratio.nominator << "/" << ratio.denominator << std::endl;
        for (pt::ptree::value_type& par: trigger.second.get_child("par")) {
            ratio_pred_parameters[par.second.get<int>("id")] = par.second.get<float>("value");
            std::cout << alertprefix << "Par " << par.second.get<int>("id") << ": " << ratio_pred_parameters[par.second.get<int>("id")]<< std::endl;
        }
        ratio.CalcRatio();
        ratio.setFormula(ratio_equation, ratio_pred_parameters);
        trig_ratios.push_back(ratio);
    }
    std::cout << alertprefix << "LVL1 Ratios configuration successful." << std::endl;
    for(pt::ptree::value_type& trigger: pt.get_child("LVL1_triggers")) {
        trigger_item  trig_tmp;
        trig_tmp.isHLT(false);
        std::cout << alertprefix << "NEW LVL1 Trigger" << std::endl;
        trig_tmp.L1_name = trigger.second.get<std::string>("yaxis");
        std::cout << alertprefix << "Name: " << trig_tmp.L1_name << std::endl;
        trig_tmp.rms  = trigger.second.get<float>("rms");
        std::cout << alertprefix << "RMS: " << trig_tmp.rms << std::endl;
        equation = trigger.second.get<std::string>("function");
        std::cout << alertprefix << "Function: " << equation << std::endl;
        for(pt::ptree::value_type& par: trigger.second.get_child("par")) {
            pred_parameters[par.second.get<int>("id")] = par.second.get<float>("value");
            std::cout << alertprefix << "Par: " << pred_parameters[par.second.get<int>("id")]<< std::endl;
        }
        trig_tmp.setFormula(equation, pred_parameters);
        l1trigs.push_back(trig_tmp);
    }
    
    std::cout << alertprefix << "LVL1 Triggers configuration successful." << std::endl;
    for(pt::ptree::value_type& trigger: pt.get_child("HLT_triggers")) {
        trigger_item  trig_tmp;

        trig_tmp.isHLT(true);
        std::cout << alertprefix << "NEW HLT Trigger" << std::endl;
        trig_tmp.HLT_name = trigger.second.get<std::string>("yaxis");
        std::cout << alertprefix << "Name: " << trig_tmp.HLT_name << std::endl;
        trig_tmp.rms  = trigger.second.get<float>("rms");
        std::cout << alertprefix << "RMS: " << trig_tmp.rms << std::endl;
        equation = trigger.second.get<std::string>("function");
        std::cout << alertprefix << "Function: " << equation << std::endl;
        for(pt::ptree::value_type& par: trigger.second.get_child("par")) {
            pred_parameters[par.second.get<int>("id")] = par.second.get<float>("value");
            std::cout << alertprefix << "Par: " << pred_parameters[par.second.get<int>("id")]<< std::endl;
        }
        trig_tmp.setFormula(equation, pred_parameters);
        hlttrigs.push_back(trig_tmp);
    }
    std::cout << alertprefix << "HLT Triggers configuration successful." << std::endl;
    return success;
}

/**
 * Set isHI flag 
 * Will report rates = f(lumi), instead of pp method of rates = lumi*f(mu)
 */
void PredAdapter::SetHI(bool config_hi_flag) {
    m_isHIrun = config_hi_flag;
    if (m_isHIrun) {
        ERS_INFO( "Configured Xmon predadapter for HI Collisions." );
    }
    else {
        ERS_INFO( "Configured Xmon predadapter for pp Collisions." );
    }
}

/**
 * Set newISFormat flag
 * Will determine how data is collected & published
 */
void PredAdapter::SetIS(bool config_IS_flag) {
    m_newISFormat = config_IS_flag;
    if (m_newISFormat) {
        ERS_LOG( "Configured Xmon predadapter for new IS data format." );
    }
    else {
        ERS_LOG( "Configured Xmon predadapter for old IS data format." );
    }
}

void PredAdapter::SetTest(bool isTechRun) {
    m_isTechRun = isTechRun;
}

void PredAdapter::setDBKeysForTesting(unsigned int smk, unsigned int l1psk, unsigned int hltpsk) {
    m_smk = smk;
    m_l1k = l1psk;
    m_hltk = hltpsk;
}

void PredAdapter::updateHLTPrescales(const confadapter_imp::TriggerMenuHelpers::Mapping & hlt_ps)
{
    // udpate the configured HLT triggers with new prescales
    std::string alertprefix = "Updating prescales---> ";
    for (auto & trigger : hlttrigs) {
        if(!trigger.status) {
            continue;
        }
        //Match PS to HLT Trigger
        const std::vector<std::string> & vps = hlt_ps.at(trigger.HLT_name);
        for (auto hltps : hlt_ps) {
            if( vps[0].find("NA") == std::string::npos) {
                trigger.HLT_PS = std::stoi(vps[0]);
            }
        }
        std::cout << alertprefix << " Configured trigger " << trigger.HLT_name << " has HLT PS updated to " << trigger.HLT_PS << std::endl;
    }
}



/**
 * readTriggerDBInfo (for HLT)
 * Gets trigger keys and sets some global vars
 */
bool
PredAdapter::readTriggerDBInfo()
{
    std::string alertprefix = "Xmon()::readTriggerDBInfo---> ";

    // Try to get trigger keys if not externally provided
    if(m_smk==0) {
        try {
            trp_utils::GetTriggerKeys(m_partitionName, m_smk, m_hltk, m_l1k );
        }
        catch (std::exception & ex) {
            std::cout << alertprefix << "Exception on getTriggerKeys: " << ex.what() << std::endl;
            return failure;
        }
    }
    if(!m_smk) {
        std::cout << alertprefix << "Supermasterkey is 0. No menu available" << std::endl;
        return failure;
    }

    //Sets global vars to results from trputils::GetTriggerKeys
    std::cout << alertprefix << "Trigger db " << m_triggerDB << " with smk " << m_smk << ", hlt psk " << m_hltk << " and l1 psk " <<m_l1k << std::endl;

    confadapter_imp::TriggerMenuHelpers menuHelper(m_partitionName, m_triggerDB);

    //Try to get the configuration of menu of the 3 keys
    try {
        menuHelper.getConfiguration(m_smk, m_hltk, 0);
    }
    catch (std::exception &ex) {
        ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
        return failure;
    }
    ERS_LOG("Succeeded loading the trigger configuration");

    //Declare Mapping variables l1_hlt and hlt_ef, and get them from menu
    const auto & map_L1toHLT = menuHelper.getMultiMapping("HLT_L1");
    std::cout << alertprefix << "Got L1:HLT map of size " << map_L1toHLT.size() << std::endl;
    if (map_L1toHLT.size()==0) {
        ers::warning(predadapter::Issue(ERS_HERE, "Failed to retrieve L1:HLT map"));
        return failure;
    }
    menuHelper.printMultiMapping(map_L1toHLT, "L1 -> HLT name mapping: ");
    std::unordered_map<std::string,std::string> hlt2L1Map;
    for(auto & l1hlt : map_L1toHLT) {
        hlt2L1Map[l1hlt.second] = l1hlt.first;
    }

    // udpate the configured HLT triggers with new prescales
    for (auto & trigger : hlttrigs) {
        bool inMenu = hlt2L1Map.count(trigger.HLT_name) > 0; // is the configured trigger in the menu

        // ignore non-existing triggers
        if(!inMenu) {
            trigger.status = 0;
            std::cout << alertprefix << " Configured trigger " << trigger.HLT_name << (inMenu ? "" : " (not part of the menu)") << std::endl;
            continue;
        }
        // Match L1 to HLT Trigger
        if (m_isTechRun) {
            trigger.L1_name = "DUMMY_NAME";
            trigger.status = 1;
        }
        else {
            for (auto & it_L1mHLT : map_L1toHLT) {
                const std::string l1Name = it_L1mHLT.first;
                const std::string hltName = it_L1mHLT.second;
                if (hltName != trigger.HLT_name) {
                    continue;
                }
                trigger.status = 1;
                if (l1Name.find(",") == std::string::npos) { // no multi-item seed
                    trigger.L1_name = l1Name;
                }
                else if (l1Name == "L1_MU20,L1_MU21") { // seeded by L1_MU20, MU21
                    trigger.L1_name = "L1_MU20";
                }
                else {
                    trigger.status = 0; // disable multi-item seeded chains
                }
                break;
            }
        }
    }

    //Declare Mapping Variables for the prescale values, and get them from menu
    const auto & map_HLTtoPS = menuHelper.getMapping("hlt_ps");
    std::cout << alertprefix << "Got HLT:PS map of size " << map_HLTtoPS.size() << std::endl;
    if (map_HLTtoPS.size()==0) {
        ers::warning(predadapter::Issue(ERS_HERE, "Failed to retrieve HLT:PS map"));
        return failure;
    }
    menuHelper.printMapping(map_HLTtoPS, "HLT -> PS name mapping: ");
    updateHLTPrescales(map_HLTtoPS);

    for (auto & trigger : hlttrigs) {
        if(!trigger.status) {
            continue;
        }
        std::cout << alertprefix << " Configured trigger " << trigger.HLT_name 
                << " is seeded by " << trigger.L1_name << " with L1 PS " << trigger.L1_PS
                << " and HLT PS " << trigger.HLT_PS << std::endl;
    }
    return success;
}

/*
 * Retrieves reference histogram data from a root file
 * */
bool PredAdapter::HistConfigure(std::string histPath) {
    
    std::string alertprefix = "Xmon()::HistConfigure---> ";
    
    if(! fileExists(histPath)) {
        std::cout << alertprefix << "Histogram file " << histPath << " does not exist" << std::endl;
        return failure;
    }

    TFile *f = new TFile(histPath.c_str());
    if (f->IsZombie()) {
        std::cout << alertprefix << "Error opening reference histogram file " << histPath << std::endl;
        return failure;
    }

    TTree *T = (TTree*)f->Get("T");
    {
        for (auto x : l1trigs) x.hist_config = true;
        for (auto x : hlttrigs) x.hist_config = true;
        std::cout << alertprefix << "Histogram file exists!!!!!" << std::endl;
    }

    TH1F *trigger = 0;
    TH1F *previous = 0;

    for(unsigned int i = 0; i < l1trigs.size(); i++) {

        std::string temp = l1trigs[i].L1_name;
        size_t pos = temp.find("-");
        while (pos != std::string::npos) {
            temp.replace(pos, 1, "_");
            pos = temp.find("-", pos + 1);
        }
        const char * name = temp.c_str();

        trigger = NULL;

        T->SetBranchAddress(name, &trigger);
        T->GetEntry(0);
        if (name != l1trigs[0].L1_name && trigger->GetMean() == previous->GetMean() && trigger->GetStdDev() == previous->GetStdDev()) {
            std::cout << alertprefix << "No reference histogram for " << name << std::endl;
            continue;
        }
        l1trigs[i].reference_hist = trigger;
        previous = trigger;
        l1trigs[i].hist_config = true;
        std::cout << alertprefix << "Configured reference histogram for " << name << std::endl;
    }
    return success;
}

/**
 * check HLT dependencies on L1
 */
void PredAdapter::CheckHLT() {
    for (unsigned int i = 0; i < hlttrigs.size(); i++) {
        if (hlttrigs[i].status==0) {
            continue;
        }
        for (unsigned int j = 0; j < l1trigs.size(); j++) {
            if (hlttrigs[i].L1_name == l1trigs[j].L1_name) {
                break;
            }
            if (j == l1trigs.size() - 1) {
                trigger_item trig_tmp;
                trig_tmp.isHLT(false);
                trig_tmp.L1_name = (hlttrigs[i].L1_name);
                extral1trigs.push_back(trig_tmp);
            }
        }
    }
}

/**
 * Global offset: get list of trigger groups and determine which triggers live in each group
 */
void PredAdapter::ParseTriggers() {
    std::string alertprefix = "Xmon()::ParseTriggers---> ";

    std::vector<std::string> group_array;
    const std::string delimiter = ",";
    size_t pos = 0;
    
    while ((pos = m_trigger_groups.find(delimiter)) != std::string::npos) {
        std::string token = m_trigger_groups.substr(0, pos);
        group_array.push_back(token);
        m_trigger_groups.erase(0, pos + delimiter.length());
    }
    group_array.push_back(m_trigger_groups);
    
    std::cout << alertprefix << "Initializing " << group_array.size() << " groups of triggers:" << std::endl;
    for (unsigned int i = 0; i < group_array.size(); i++) {
        trigger_list  trig_list_tmp;
        trig_list_tmp.incl_trigs.clear();
        trig_list_tmp.grp_name = group_array[i];
        trig_list_tmp.GroupTriggers(l1trigs);
        trig_list_tmp.SetBool(i);
        trig_groups.push_back(trig_list_tmp);
        std::cout << alertprefix << trig_groups[i].grp_name << " contains " << trig_groups[i].incl_trigs.size() << " triggers: " << std::endl;
        for (unsigned int j = 0; j < trig_groups[i].incl_trigs.size(); j++) {
            std::cout << alertprefix << "          " << trig_groups[i].incl_trigs[j].L1_name << std::endl;
        }
    }
}

/**
 * Run
 * Sets up conditions for run and subscribes to IS & OLC servers
 */

bool PredAdapter::SubscribeToIS() {
    ERS_LOG( "subscribing to IS information on partition " << m_partitionName );

    IPCPartition partition(m_partitionName);
    m_receiver = std::make_unique<ISInfoReceiver>(partition);
    ERS_DEBUG( 0, "Subscribing to IS Servers");
    if (!m_newISFormat){
        try {
            try {
                m_receiver->subscribe(m_serverName+".L1_Rate", (&PredAdapter::callback_l1rates), this);
            }
            catch (daq::is::AlreadySubscribed &ex) {
                ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
            }
            catch (daq::is::InvalidCriteria &ex) {
                ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
            }
        
            try {
                m_receiver->subscribe(m_serverName+".HLT_Rate", (&PredAdapter::callback_hltrates), this);
            }
            catch (daq::is::AlreadySubscribed  &ex) {
                ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
            }
            catch (daq::is::InvalidCriteria    &ex) {
                ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
            }
        }
        catch (daq::is::RepositoryNotFound &ex) {
            ers::warning(predadapter::Issue(ERS_HERE, std::string(ex.what())));
        }
    }
    //subscriptions for new IS data format, need to subscribe to each trigger individually
    else {
        try {
            // HLT subscription
            for (const auto & trigger: hlttrigs) {
                try {
                    m_receiver->subscribe(m_serverName + '.' + trigger.HLT_name, (&PredAdapter::newcallback_hltrates), this);
                }
                catch (daq::is::AlreadySubscribed  &ex) {
                    ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
                }
                catch(daq::is::InvalidCriteria& ex) {
                    ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
                }
            }

            // L1 subscription
            for (const auto & trigger: l1trigs) {
                try {
                    m_receiver->subscribe(m_serverName + '.' + trigger.L1_name, (&PredAdapter::newcallback_l1rates), this);
                }
                catch (daq::is::AlreadySubscribed &ex) {
                    ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
                }
                catch (daq::is::InvalidCriteria &ex) {
                    ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
                }
            }
        
            // L1 subscription
            for (const auto & trigger: extral1trigs) {
                try {
                    m_receiver->subscribe(m_serverName + '.' + trigger.L1_name, (&PredAdapter::newcallback_l1rates_simple), this);
                }
                catch (daq::is::AlreadySubscribed  &ex) {
                    ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
                }
                catch (daq::is::InvalidCriteria &ex) {
                    ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
                }
            }
        }
        catch (daq::is::RepositoryNotFound &ex) {
            ers::warning(predadapter::Issue(ERS_HERE, "Cannot subscribe to rate updates. " + std::string(ex.what())));
            return failure;
        }
    }

    try {
        try {
            m_receiver->subscribe("RunParams.TrigConfL1PsKey", (&PredAdapter::callback_l1k), this);
        }
        catch (daq::is::AlreadySubscribed  &ex) {
            ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
        }
        catch (daq::is::InvalidCriteria    &ex) {
            ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
        }

        try {
            m_receiver->subscribe("RunParams.TrigConfHltPsKey", (&PredAdapter::callback_hltk), this);
        }
        catch (daq::is::AlreadySubscribed &ex) {
            ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
        }
        catch (daq::is::InvalidCriteria &ex) {
            ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
        }
    }
    catch (daq::is::RepositoryNotFound &ex) {
        ers::error(predadapter::Issue(ERS_HERE, "Cannot subscribe to PSK updates. " + std::string(ex.what())));
        return failure;
    } 

    IPCPartition olcPartition(m_lumiPartition);
    m_olcReceiver = std::make_unique<ISInfoReceiver>(olcPartition);
    
    ERS_LOG("Subscribing to OLC Servers");
    try {
        m_olcReceiver->subscribe(m_olcIsServer+"."+m_objectName, (&PredAdapter::callback_Mu), this);
    }
    catch (daq::is::RepositoryNotFound& ex) {
        ers::info(predadapter::Issue(ERS_HERE, "No luminosiy information available. Will not be able to provide luminosity-base rate predictions"));
        ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
    }
    catch (daq::is::AlreadySubscribed &ex) {
        ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
    }
    catch (daq::is::InvalidCriteria &ex) {
        ers::warning(predadapter::Issue(ERS_HERE, ex.what()));
    }
    return success;
}

void PredAdapter::Run() {
    std::string alertprefix = "Xmon()::Run---> ";
    if (m_debug) {
        std::cout << alertprefix<< "Run before receiver!!!"<<std::endl;
    }
    OWLSemaphore s;
    m_semaphore = &s;
    m_semaphore->wait();
    m_semaphore = nullptr;
    if (m_debug) {
        std::cout << alertprefix<< "Run after receiver!!!"<<std::endl;
    }
}

//=====================================================================================
// Callback Functions
//      Functions that are called when IS data objects are updated
//=====================================================================================


// --------------------------------------------------------------------------------
// L1 KEY
// --------------------------------------------------------------------------------
void PredAdapter::callback_l1k(ISCallbackInfo *isc) {
    ERS_LOG("Callback for L1 psk");
    if (isc->reason() == ISInfo::Deleted) {
        return;
    }
    TrigConfL1PsKey l1_k;
    isc->value(l1_k);
    m_l1k = l1_k.L1PrescaleKey;
    ERS_LOG("New L1 psk (currently not used): " << m_l1k);
}

// --------------------------------------------------------------------------------
// HLT KEY
// --------------------------------------------------------------------------------
void PredAdapter::callback_hltk(ISCallbackInfo *isc) {
    ERS_LOG("Callback for HLT psk");
    if (isc->reason() == ISInfo::Deleted) {
        return;
    }
    TrigConfHltPsKey hlt_k;
    isc->value(hlt_k);
    m_hltk = hlt_k.HltPrescaleKey;
    ERS_LOG("New HLT psk: " << m_hltk);
    confadapter_imp::TriggerMenuHelpers menuHelper(m_partitionName, m_triggerDB);
    try {
        menuHelper.getConfiguration(0, m_hltk, 0);
    }
    catch (std::exception &ex) { 
        ers::error(predadapter::Issue(ERS_HERE, ex.what()));
    }

    const auto & map_HLTtoPS = menuHelper.getMapping("hlt_ps");

    updateHLTPrescales(map_HLTtoPS);

    // Loop for each element in the hlt vector (hlttrigs) formed in the PredConfigure
    for (auto & chain : hlttrigs) {
        if(!chain.status) {
            continue;
        }
        std::string ps = map_HLTtoPS.at(chain.HLT_name)[0];
        chain.HLT_PS = std::stoi(ps);
    }
}

//---------------------------------------------------------------------------------
// L1 Prescales that we need but aren't in l1trigs
// --------------------------------------------------------------------------------
void PredAdapter::newcallback_l1rates_simple(ISCallbackInfo *isc) {
    if (isc->reason() == ISInfo::Deleted) return;
    if (m_debug) std::cout << "\n\n>>>>> Callback for " << isc->name() << std::endl;

    if(!TRIG_FORMATTED) setIndices();
    
    unsigned int index = m_extral1map[isc->name()];
    isc->value(extral1trigs[index].ISC_L1);

    if (extral1trigs[index].status) {    
        extral1trigs[index].PS  = extral1trigs[index].ISC_L1.PS;
    
        if (extral1trigs[index].PS >= 1) {
            extral1trigs[index].checkPS = 1;
        }
        else {
            extral1trigs[index].checkPS = 0;
        }

        extral1trigs[index].initialized = true;
    }
}

// --------------------------------------------------------------------------------
// L1 Rates (new)
// --------------------------------------------------------------------------------
void PredAdapter::newcallback_l1rates(ISCallbackInfo *isc) {
    if (isc->reason() == ISInfo::Deleted) return;
    if (m_debug) std::cout << "\n\n>>>>> Callback for " << isc->name() << std::endl;
    
    if (!TRIG_FORMATTED) setIndices();
 
    unsigned int index = m_l1map[isc->name()];
    isc->value(l1trigs[index].ISC_L1);

    if (l1trigs[index].status) {
        if (m_debug) std::cout << ">>>>> Status is true for " << isc->name() << std::endl;
        l1trigs[index].rate         = l1trigs[index].ISC_L1.TAP;
        l1trigs[index].rate_unpresc = l1trigs[index].ISC_L1.TBP;
        l1trigs[index].PS           = l1trigs[index].ISC_L1.PS;

        if (l1trigs[index].PS >= 1) {
            l1trigs[index].checkPS = 1;
        }
        else {
            l1trigs[index].checkPS = 0;
        }
    
        l1trigs[index].initialized = true;
    }                
    
    if (l1trigs[index].initialized && LUM_INITIALIZED) {
        if (m_debug) std::cout << ">>>>> Calculating prediction for " << isc->name() << std::endl;
        calculatePred_new(index);
        
        // l1trigs[index].L1_IS_Pred.TBP     = l1trigs[index].rate_unpresc;
        // l1trigs[index].L1_IS_Pred.TAP     = l1trigs[index].rate;
        // l1trigs[index].L1_IS_Pred.PS      = l1trigs[index].PS;
        l1trigs[index].L1_IS_Pred.PBP     = l1trigs[index].pred_unpresc;
        l1trigs[index].L1_IS_Pred.PAP     = l1trigs[index].pred;
        // l1trigs[index].L1_IS_Pred.NSIGMA  = l1trigs[index].nSigma;
        // l1trigs[index].L1_IS_Pred.RATIO   = l1trigs[index].ratio;
        // l1trigs[index].L1_IS_Pred.SRATIO  = l1trigs[index].sratio;
        // l1trigs[index].L1_IS_Pred.DIFF    = l1trigs[index].diff;
        // l1trigs[index].L1_IS_Pred.FLAG    = l1trigs[index].flag;
        // l1trigs[index].L1_IS_Pred.Par0    = l1trigs[index].p0;
        // l1trigs[index].L1_IS_Pred.Par1    = l1trigs[index].p1;
        l1trigs[index].L1_IS_Pred.OTFPRED = l1trigs[index].otfpred;

        L1_ISPublish(l1trigs[index].L1_name, l1trigs[index].L1_IS_Pred);
    }

    if (l1trigs[index].initialized && l1trigs[index].in_List) {
        trig_groups[l1trigs[index].list_Index].CalcAvg();
        GO_ISPublish(trig_groups[l1trigs[index].list_Index].grp_name, trig_groups[l1trigs[index].list_Index].group_stats);
    }
}

// --------------------------------------------------------------------------------
// HLT Rates (new)
// --------------------------------------------------------------------------------
void PredAdapter::newcallback_hltrates(ISCallbackInfo *isc) {
    std::string alertprefix = "Xmon()::callback_hltrates---> ";
    if (isc->reason() == ISInfo::Deleted) return;
    if (m_debug) std::cout << "\n\n>>>>> Callback for " << isc->name() << std::endl;

    if (!TRIG_FORMATTED) setIndices();
    
    unsigned int index = m_hltmap[isc->name()];
    isc->value(hlttrigs[index].ISC_HLT);

    if (hlttrigs[index].status) {
        if (m_debug) std::cout << ">>>>> Status is true for " << isc->name() << std::endl;
        hlttrigs[index].rate  = hlttrigs[index].ISC_HLT.Output;
        
        if (hlttrigs[index].L1_name == l1trigs[hlttrigs[index].L1_index].L1_name) {
            hlttrigs[index].L1_PS = l1trigs[hlttrigs[index].L1_index].PS;
        }
        else if (hlttrigs[index].L1_name == extral1trigs[hlttrigs[index].L1_index].L1_name) {
            hlttrigs[index].L1_PS = extral1trigs[hlttrigs[index].L1_index].PS;
        }
        else if (m_isTechRun) {
            hlttrigs[index].L1_PS = 1;
        }
        else {
            std::cout << "Error retrieving L1 prescale for HLT trigger " << hlttrigs[index].HLT_name << std::endl;
            std::cout << "Deactivating trigger" << std::endl;
            hlttrigs[index].status = false;
        }

        hlttrigs[index].initialized = true;
    }

    if (hlttrigs[index].initialized && LUM_INITIALIZED) {
        if (m_debug) std::cout << ">>>>> Calculating prediction for " << isc->name() << std::endl;
        calculate_HLT_new(index);

        // hlttrigs[index].HLT_IS_Pred.output  = hlttrigs[index].rate;
        // hlttrigs[index].HLT_IS_Pred.ps      = hlttrigs[index].HLT_PS;
        hlttrigs[index].HLT_IS_Pred.pred    = hlttrigs[index].pred;
        // hlttrigs[index].HLT_IS_Pred.diff    = hlttrigs[index].diff;
        // hlttrigs[index].HLT_IS_Pred.nsigma  = hlttrigs[index].nSigma;
        // hlttrigs[index].HLT_IS_Pred.ratio   = hlttrigs[index].ratio;
        // hlttrigs[index].HLT_IS_Pred.sratio  = hlttrigs[index].sratio;
        // hlttrigs[index].HLT_IS_Pred.flag    = hlttrigs[index].flag;
        // hlttrigs[index].HLT_IS_Pred.Par0    = hlttrigs[index].p0;
        // hlttrigs[index].HLT_IS_Pred.Par1    = hlttrigs[index].p1;
        hlttrigs[index].HLT_IS_Pred.otfpred = hlttrigs[index].otfpred;

        HLT_ISPublish(hlttrigs[index].HLT_name, hlttrigs[index].HLT_IS_Pred);
    }
}

// --------------------------------------------------------------------------------
// L1 Rates
// --------------------------------------------------------------------------------
void PredAdapter::callback_l1rates(ISCallbackInfo *isc) {
    std::string alertprefix = "Xmon()::callback_l1rates---> ";
    if (isc->reason() == ISInfo::Deleted) return;
    if (m_debug) std::cout << "\n\n>>>>> Callback for " << isc->name() << std::endl;

    // >>>> *************************************** - Weird section that can't happen during callback
    if (!L1_ISCALLBACK) {
        isc->value(ISC_L1TimePoint);
        L1_RunNumber = ISC_L1TimePoint.RunNumber;
        L1_LumiBlock = ISC_L1TimePoint.LumiBlock;
        L1_TimeStamp = ISC_L1TimePoint.TimeStamp;
        if (m_debug) {
            std::cout << alertprefix<< "Run Number : " << L1_RunNumber << std::endl;
            std::cout << alertprefix<< "Lumiblock  : " << L1_LumiBlock << std::endl;
            std::cout << alertprefix<< "TimeStamp  : " << L1_TimeStamp << std::endl;
        }
    }
    // <<<< *************************************** - Not Sure why this stuff is important yet

    // >>>> *************************************** - BEGIN L1 Formatting
    // Find the index of TBP, TAP, PS
    if (!L1_FORMATTED) {
        // Find Indexes
        YLabels_size = ISC_L1TimePoint.YLabels.size();
        for (unsigned int k = 0; k < YLabels_size; k++) {
            if      (ISC_L1TimePoint.YLabels[k] == "TBP") TBP_index = k;
            else if (ISC_L1TimePoint.YLabels[k] == "TAP") TAP_index = k;
            else if (ISC_L1TimePoint.YLabels[k] == "PS")  PS_index  = k;
        }
        std::cout << alertprefix <<"TBP:TAP:PS index =  "<<TBP_index<<" : "<<TAP_index<<" : "<< PS_index << std::endl;

        // Set number of Triggers
        l1rates_IS_trigN = ISC_L1TimePoint.XLabels.size();

        // 2015/12/17 TMH -- Check for xxx item at the beginning of the list
        // 2017 ???
        std::string trigname_0th = ISC_L1TimePoint.XLabels[0];
        std::size_t found_xxx    = trigname_0th.find("xxx");
        if ((int)found_xxx > -1) {
            // Found xxx, so must remove the first one
            ISC_L1TimePoint.XLabels.erase( ISC_L1TimePoint.XLabels.begin() );
            l1rates_IS_trigN --;
        }

        x_L1.resize(l1trigs.size());
        // Set defaults
        for (unsigned int i = 0; i <l1trigs.size(); i++) {
            x_L1[i] = "null";
            l1trigs[i].xIndex = -1;
            l1trigs[i].status = false;
        }

        // Match IS to Trig Vector
        for (unsigned int i = 0; i < l1rates_IS_trigN; i++) {
            for (unsigned int j = 0; j <l1trigs.size(); j++ ) {
                if (ISC_L1TimePoint.XLabels[i] ==l1trigs[j].L1_name) {
                    std::cout << alertprefix << "Putting " << l1trigs[j].L1_name << " into x_L1[" << j << "]" << std::endl;
                    x_L1[j] =l1trigs[j].L1_name;
                    l1trigs[j].xIndex = i;
                    l1trigs[j].status = true;
                }

            }

        }
        Pred_TimePoint_L1.format(x_L1, y_L1);
        L1_FORMATTED = true;


        // section from callback_l1rates from hlt
        for (unsigned int i = 0; i < hlttrigs.size(); i++ ) {
            bool L1MATCHEXISTS = false;
            for (unsigned int j = 0; j < l1rates_IS_trigN; j++) {
                if (hlttrigs[i].L1_name == ISC_L1TimePoint.XLabels[j]) {
                    hlttrigs[i].L1_index = j;
                    L1MATCHEXISTS = true;
                    break;
                }
            }
            if (!L1MATCHEXISTS)  {
                hlttrigs[i].status = 0;
                std::cout << alertprefix << "Deactivating HLTTrig: " << hlttrigs[i].HLT_name << " - L1Trig: " << hlttrigs[i].L1_name << " -> not found" << std::endl;
            }
        }

        std::cout << alertprefix << ">>>>> L1 Formatting Complete!!! <<<<<< " << std::endl;
    }
    // <<<< *************************************** - END L1 Formating
    // dump of section from hlt callback_l1rates
    for (unsigned int i = 0; i < hlttrigs.size(); i++) {
        if (hlttrigs[i].status) {
            hlttrigs[i].L1_PS = ISC_L1TimePoint.Data[(hlttrigs[i].L1_index * YLabels_size) + PS_index];
        }

        if (m_debug) {
            std::cout << alertprefix << i << ". HLT Trigger : " << hlttrigs[i].HLT_name << std::endl;
            std::cout << alertprefix<< "L1 Trig : " << hlttrigs[i].L1_name << std::endl;
            std::cout << alertprefix<< "L1 PS   : " << hlttrigs[i].L1_PS << std::endl;
        }
    }

    // >>>> *************************************** - L1 main callback
    //don't rewrite to predTimePoint and vector while doing calculation and publish
    if (!L1_ISCALCULATING) {
        L1_ISCALLBACK = true;

        Pred_TimePoint_L1.RunNumber = L1_RunNumber;
        Pred_TimePoint_L1.LumiBlock = L1_LumiBlock;
        Pred_TimePoint_L1.TimeStamp = L1_TimeStamp;

        for (unsigned int j = 0; j <l1trigs.size(); j++ ) {
            if (l1trigs[j].status) {
                l1trigs[j].rate           = ISC_L1TimePoint.Data[l1trigs[j].xIndex*YLabels_size + TAP_index];
                l1trigs[j].rate_unpresc   = ISC_L1TimePoint.Data[l1trigs[j].xIndex*YLabels_size + TBP_index];
                l1trigs[j].PS             = ISC_L1TimePoint.Data[l1trigs[j].xIndex*YLabels_size + PS_index];

                if (m_debug) {
                    std::cout << alertprefix << "L1 Trig: "<<l1trigs[j].L1_name << std::endl;
                    std::cout << alertprefix << "    TAP: "<<l1trigs[j].rate << std::endl;
                    std::cout << alertprefix << "    TBP: "<<l1trigs[j].rate_unpresc << std::endl;
                    std::cout << alertprefix << "     PS: "<<l1trigs[j].PS<<std::endl;
                }

                if (l1trigs[j].PS >= 1)l1trigs[j].checkPS = 1;
                else l1trigs[j].checkPS = 0;
            }
        }

        L1_INITIALIZED = true;
        L1_ISCALLBACK = false;
    }

    if (m_debug) std::cout << alertprefix << " >>>>>> End of L1 Callback <<<<<< " << std::endl;

    if (L1_INITIALIZED && LUM_INITIALIZED) {
        calculatePred();
        publishIS(m_l1OutputPartition, Pred_TimePoint_L1);
    }
    // <<<< *************************************** - End of Callback
}

// --------------------------------------------------------------------------------
// HLT Rates
//      Section 1: Get new isc value
//      Section 2: Format HLT; get output indices from timepoint array_L1;
//                 resize XLabels; get xIndex; set status
//      Section 3: Get rates
// --------------------------------------------------------------------------------
void PredAdapter::callback_hltrates(ISCallbackInfo * isc) {
    std::string alertprefix = "Xmon()::callback_hltrates---> ";
    if (isc->reason() == ISInfo::Deleted) return;
    if (m_debug) std::cout << "\n\n>>>>> Callback for "<< isc->name() << " <<<<<" << std::endl;

    // >>>> *************************************** - Weird section that can't happen during callback
    if (!HLT_isCallBack) {
        isc->value(ISC_HLTTimePoint);
        HLT_RunNumber = ISC_HLTTimePoint.RunNumber;
        HLT_LumiBlock = ISC_HLTTimePoint.LumiBlock;
        HLT_TimeStamp = ISC_HLTTimePoint.TimeStamp;
        if (m_debug) {
            std::cout << alertprefix << "The HLT RunNumber is  : " << HLT_RunNumber << std::endl;
            std::cout << alertprefix << "The HLT LumiBlock is  : " << HLT_LumiBlock << std::endl;
            std::cout << alertprefix << "The HLT TimeStamp is  : " << HLT_TimeStamp << std::endl;
        }
    }
    // <<<< *************************************** - Not Sure why this stuff is important yet

    // >>>> ******************************* Format/Initialize
    // Find the index of output in IS rate object
    if (!HLT_INITIALIZED) {
        size_yHLT = ISC_HLTTimePoint.YLabels.size();
        for (unsigned int k = 0; k < size_yHLT; k++) {
            if (ISC_HLTTimePoint.YLabels[k] == "output") output_index = k;
        }
        std::cout << alertprefix << "HLT Output Index = " << output_index <<std::endl;

        // Set Number of Triggers
        hltrates_IS_trigN = ISC_HLTTimePoint.XLabels.size();
        x_HLT.resize(hlttrigs.size());
        for (unsigned int i = 0; i < hlttrigs.size(); i++) {
            x_HLT[i] = "null";
        }

        for (unsigned int i = 0; i < hltrates_IS_trigN; i++) {
            for (unsigned int j = 0; j < hlttrigs.size(); j++) {
                if (hlttrigs[j].status && ISC_HLTTimePoint.XLabels[i] == hlttrigs[j].HLT_name) {
                    std::cout << alertprefix << "Putting " << hlttrigs[j].HLT_name << " into x_HLT[" << j << "]" << std::endl;
                    x_HLT[j] = hlttrigs[j].HLT_name;
                    hlttrigs[j].xIndex = i;
                    break;
                }
            }
        }

        Pred_TimePoint_HLT.format(x_HLT, y_HLT);
        HLT_INITIALIZED = true;
        std::cout << alertprefix << ">>>>> HLT Initialization Successfull! <<<<<<" << std::endl;
    }
    // <<<< ******************************* End Format/Initialize


    // >>>> ******************************* Actual Callback - get Rate
    // Don't rewrite to Pred_TimePoint_HLT and vector while doing calculation and publish
    if (!HLT_isCalculating) {
        HLT_isCallBack = true;

        Pred_TimePoint_HLT.RunNumber = HLT_RunNumber;
        Pred_TimePoint_HLT.LumiBlock = HLT_LumiBlock;
        Pred_TimePoint_HLT.TimeStamp = HLT_TimeStamp;

        for (unsigned int j = 0; j < hlttrigs.size(); j++ ) {
            if (hlttrigs[j].status) {
                hlttrigs[j].rate = ISC_HLTTimePoint.Data[hlttrigs[j].xIndex * size_yHLT + output_index];

                if (m_debug) {
                    std::cout << alertprefix << "HLTtrig : " << hlttrigs[j].HLT_name << std::endl;
                    std::cout << alertprefix << "    TAP : " << hlttrigs[j].rate     << std::endl;
                }
            }
        }
        HLT_isCallBack  = false;
    }
    if (HLT_INITIALIZED && LUM_INITIALIZED && L1_INITIALIZED ) {
        calculate_HLT();

        publishIS(m_hltOutputPartition, Pred_TimePoint_HLT);
    }
    // ******************************* End Callback
    if (m_debug) std::cout << alertprefix << "End HLT Rate Callback <<<<<<< " << std::endl;
}

// --------------------------------------------------------------------------------
// Luminosity Data
// --------------------------------------------------------------------------------
void PredAdapter::callback_Mu(ISCallbackInfo * isc) {
    std::string alertprefix = "Xmon()::callback_MU---> ";
    if (isc->reason() == ISInfo::Deleted) return;
    if (m_debug) std::cout << "\n\n>>>>> Callback for "<< isc->name()<<std::endl;

    ISInfoAny OLC_LumiData;
    isc->value(OLC_LumiData);
    double my_lum[7];
    for (int i = 0; i<7; ++i) {
        OLC_LumiData >> my_lum[i];
    }

    //Mu is 5th and caliblumi is 7th
    // most likely return to these values for Run 3
    //myMu   = my_lum[4];
    //myLumi = my_lum[6];

    myMu   = my_lum[2];
    myLumi = my_lum[4];
    
    LUM_INITIALIZED = true;

    if (m_isHIrun) {
        myMu = myLumi;
        myLumi = 1.0;
    }

    if (m_debug) {
        std::cout << alertprefix << "myMu = " << myMu << std::endl;
        std::cout << alertprefix << "myLumi = " << myLumi << std::endl;
        std::cout << alertprefix << "End Callback Mu <<<<<" << std::endl;
    }
}

/**
 * setIndices()
 * for each trigger in hlttrigs:
 *      determine its l1 counterpart for L1_index
 *      determine its index in hlttrigs
 *      add trigname : index to m_hltmap
 * for each trigger in l1trigs:
 *      determine its index
 *      add trigname : index to l1map
 * ***Only used with new IS data format***
 */
void PredAdapter::setIndices() {
    std::string alertprefix = "Xmon()::setIndices---> ";
    //map HLT to L1
    for (unsigned int i = 0; i < hlttrigs.size(); i++) {
        bool L1MATCHEXISTS = false;
        for (unsigned int j = 0; j < l1trigs.size(); j++) {
            if (hlttrigs[i].L1_name == l1trigs[j].L1_name) {
                hlttrigs[i].L1_index = j;
                L1MATCHEXISTS = true;
                break;
            }
        }
        for (unsigned int j = 0; j < extral1trigs.size(); j++) {
            if (hlttrigs[i].L1_name == extral1trigs[j].L1_name) {
                hlttrigs[i].L1_index = j;
                L1MATCHEXISTS = true;
                break;
            }
        }
        if (!L1MATCHEXISTS) {
            hlttrigs[i].status = 0;
            std::cout << alertprefix << "Deactivating HLTTrig: " << hlttrigs[i].HLT_name << " - L1Trig: " << hlttrigs[i].L1_name << " -> not found" << std::endl;
        }
    }

    //set indices for triggers
    //create maps trigname : index
    for (unsigned int i = 0; i < hlttrigs.size(); i++) {
        hlttrigs[i].HLT_index = i;
        m_hltmap[m_serverName+'.'+hlttrigs[i].HLT_name] = i;
    }
    for (unsigned int i = 0; i < l1trigs.size(); i++) {
        l1trigs[i].L1_index = i;
        m_l1map[m_serverName+'.'+l1trigs[i].L1_name] = i;
        l1trigs[i].status = true;
    }
    for (unsigned int i = 0; i < extral1trigs.size(); i++) {
        extral1trigs[i].L1_index = i;
        m_extral1map[m_serverName+'.'+extral1trigs[i].L1_name] = i;
        extral1trigs[i].status = true;
    }
    
    TRIG_FORMATTED = true;
}

//=====================================================================================
//Calculations
//      Calculates values for online & offline trigger rate predictions,
//      as well as global offset
//=====================================================================================

/**
 * calculate_HLT_new(unsigned int)
 * handles prediction calculations for HLT
 * using new timepoint format
 */
void PredAdapter::calculate_HLT_new(unsigned int index) {
    std::string alertprefix = "Xmon()::calculate_HLT_new---> ";
    if (m_debug) {
        std::cout << alertprefix << "HLT Trigger: " << hlttrigs[index].HLT_name << std::endl;
        std::cout << alertprefix << "STATUS : " << hlttrigs[index].status << std::endl;
    }

    if (hlttrigs[index].status) {
        // Calculate
        hlttrigs[index].CalcPS();
        hlttrigs[index].PredictRate(myMu, myLumi, m_bunchFactor);
        hlttrigs[index].OTFSetPred(myMu, myLumi, m_bunchFactor);
        hlttrigs[index].OTFPredictRate(myMu, myLumi, m_bunchFactor);
        hlttrigs[index].CalcErrors();
        hlttrigs[index].SetFlags (m_flagOk, m_flagWarning, m_flagError);
        hlttrigs[index].FillHistogram ();

        // Display Debug
        if (m_debug) hlttrigs[index].DisplayDebug(alertprefix);
    }
}

/**
 * calculate_HLT()
 * handles prediction calculations for HLT
 */
void PredAdapter::calculate_HLT() {
    std::string alertprefix = "Xmon()::calculate_HLT---> ";
    HLT_isCalculating = true;
    if (m_debug) {
        std::cout << "\n\n>>>>> Calculating HLT Rate Predictions <<<<<<" << std::endl;
    }

    for (unsigned int i = 0; i < hlttrigs.size(); i++) {
        if (m_debug) {
            std::cout << alertprefix<< i << ". HLT Trigger: " << hlttrigs[i].HLT_name <<std::endl;
            std::cout << alertprefix << "STATUS : " << hlttrigs[i].status << std::endl;
        }
        if (hlttrigs[i].status) {
            // Calculate
            hlttrigs[i].CalcPS();
            hlttrigs[i].PredictRate(myMu, myLumi, m_bunchFactor);
            hlttrigs[i].OTFSetPred(myMu, myLumi, m_bunchFactor);
            hlttrigs[i].OTFPredictRate(myMu, myLumi, m_bunchFactor);
            hlttrigs[i].CalcErrors();
            hlttrigs[i].SetFlags (m_flagOk, m_flagWarning, m_flagError);
            hlttrigs[i].FillHistogram();
            // Display Debug
            if (m_debug) hlttrigs[i].DisplayDebug(alertprefix);
        }
    }

    for (unsigned int i = 0; i < hlttrigs.size(); i++) {
        if (hlttrigs[i].status) {
            Pred_TimePoint_HLT.set(i, OUTPUT_HLT,  hlttrigs[i].rate);
            Pred_TimePoint_HLT.set(i, PS_HLT,      hlttrigs[i].HLT_PS);
            Pred_TimePoint_HLT.set(i, PRED_HLT,    hlttrigs[i].pred);
            Pred_TimePoint_HLT.set(i, DIFF_HLT,    hlttrigs[i].diff);
            Pred_TimePoint_HLT.set(i, NSIGMA_HLT,  hlttrigs[i].nSigma);
            Pred_TimePoint_HLT.set(i, RATIO_HLT,   hlttrigs[i].ratio);
            Pred_TimePoint_HLT.set(i, SRATIO_HLT,  hlttrigs[i].sratio);
            Pred_TimePoint_HLT.set(i, FLAG_HLT,    hlttrigs[i].flag);
            Pred_TimePoint_HLT.set(i, P0_HLT,      hlttrigs[i].p0);
            Pred_TimePoint_HLT.set(i, P1_HLT,      hlttrigs[i].p1);
            Pred_TimePoint_HLT.set(i, OTFPRED_HLT, hlttrigs[i].otfpred);
        }
    }

    if (m_debug) std::cout << alertprefix << "Calculation Successfull!!! <<<<< " << std::endl;
    HLT_isCalculating = false;
}

/**
 * calculatePred_new()
 * handles rate predictions for L1 triggers
 * using the new IS data format
 */
void PredAdapter::calculatePred_new(unsigned int index) {
    std::string alertprefix = "Xmon()::calculatePred_new---> ";
    if (m_debug) {
        std::cout << alertprefix << "L1 Trigger: " << l1trigs[index].L1_name << std::endl;
        std::cout << alertprefix << "Index: " << index << std::endl;
        std::cout << alertprefix << "Status : " << l1trigs[index].status << std::endl;
    }

    if (l1trigs[index].status) {
        // Calculate
        l1trigs[index].PredictRate (myMu, myLumi, m_bunchFactor);
        l1trigs[index].OTFSetPred(myMu, myLumi, m_bunchFactor);
        l1trigs[index].OTFPredictRate(myMu, myLumi, m_bunchFactor);
        l1trigs[index].CalcErrors ();
        l1trigs[index].SetFlags (m_flagOk, m_flagWarning, m_flagError);
        if (l1trigs[index].in_List) {
            l1trigs[index].FillHistogram ();
        }

        // Display Debug
        if (m_debug)l1trigs[index].DisplayDebug (alertprefix);
    }
    if (m_debug) std::cout << "Calculation Successfull!! <<<<< " << std::endl;
}

/**
 * calculatePred()
 * handles prediction calculation for L1 triggers
 */
void PredAdapter::calculatePred() {
    std::string alertprefix = "Xmon()::calculatePred---> ";
    L1_ISCALCULATING = true;
    // could indicate HLT or L1 maybe pNameOut
    if (m_debug) std::cout << "\n\n>>>>> Calculating L1 Rate Predictions <<<<<" << std::endl;


    for (trigger_item & item : l1trigs) {
        if (m_debug) {
            std::cout << alertprefix<<"L1 Trigger: " << item.L1_name << std::endl;
            std::cout << alertprefix<< "Status : " << item.status << std::endl;
        }

        if (!item.status) {
            continue;
        }
        // Calculate
        item.PredictRate (myMu, myLumi, m_bunchFactor);
        item.OTFSetPred(myMu, myLumi, m_bunchFactor);
        item.OTFPredictRate(myMu, myLumi, m_bunchFactor);
        item.CalcErrors ();
        item.SetFlags (m_flagOk, m_flagWarning, m_flagError);
        item.FillHistogram ();

        // Display Debug
        if (m_debug) {
            item.DisplayDebug(alertprefix);
        }
    }

    size_t pIdx{0};
    for (const trigger_item & item : l1trigs) {
        if (item.status) {
            Pred_TimePoint_L1.set(pIdx, TBP_L1,     item.rate_unpresc);
            Pred_TimePoint_L1.set(pIdx, TAP_L1,     item.rate);
            Pred_TimePoint_L1.set(pIdx, PS_L1,      item.PS);
            Pred_TimePoint_L1.set(pIdx, PBP_L1,     item.pred_unpresc);
            Pred_TimePoint_L1.set(pIdx, PAP_L1,     item.pred);
            Pred_TimePoint_L1.set(pIdx, NSIGMA_L1,  item.nSigma);
            Pred_TimePoint_L1.set(pIdx, RATIO_L1,   item.ratio);
            Pred_TimePoint_L1.set(pIdx, SRATIO_L1,  item.sratio);
            Pred_TimePoint_L1.set(pIdx, DIFF_L1,    item.diff);
            Pred_TimePoint_L1.set(pIdx, FLAG_L1,    item.flag);
            Pred_TimePoint_L1.set(pIdx, P0_L1,      item.p0);
            Pred_TimePoint_L1.set(pIdx, P1_L1,      item.p1);
            Pred_TimePoint_L1.set(pIdx, OTFPRED_L1, item.otfpred);
        }
        ++pIdx;
    }
    if (m_debug) std::cout << "Calculation Successfull!! <<<<< " << std::endl;
    L1_ISCALCULATING = false;
}

/**
 * Stop()
 * unsubscribe from everything
 */
void PredAdapter::Stop() {
    std::string alertprefix = "Xmon()::Stop---> ";
    if(m_receiver) {
        std::cout << alertprefix << "Now unsubscribing from TRP server '" << m_serverName << "'..." << std::endl;
        try {
            m_receiver->unsubscribe(m_serverName, ".*");
        } 
        catch(const std::exception & exception) {
            std::cout << alertprefix << "Cannot unsubscribe from ISS_TRP: " << exception.what() << std::endl;
        }

        std::cout << alertprefix << "Now unsusbscribing from IS server holding 'RunParams'..." << std::endl;
        try {
            m_receiver->unsubscribe("RunParams", ".*");
        }
        catch(const std::exception &exception) {
            std::cout << alertprefix << "Cannot unsubscribe to PrescaleAP server, "  << exception.what() << std::endl;
        }
    }

    if(m_olcReceiver) {
        std::cout << alertprefix << "Now unsusbscribing from OLC server '" << m_olcIsServer << "'..." << std::endl;
        try {
            m_olcReceiver->unsubscribe(m_olcIsServer, ".*");
        }
        catch (const std::exception &ex) {
            std::cout << alertprefix << "Cannot unsubscribe to OLC server, " << ex.what() << std::endl;
        }
    }
}

//=====================================================================================
//Publishing Functions
//      Publishes predictions and global offset data to IS servers
//      New IS data format requires multiple functions due to new classes,
//      but they all have the same structure
//=====================================================================================
/**
 * L1_ISPublish(std::string, L1_Rate_Pred &)
 * publish L1 prediction info in new format
 */
void PredAdapter::L1_ISPublish(std::string trig_name, L1_Rate_Pred & timePoint_Pred) {
    try {
        IPCPartition p(m_partitionName);
        ISInfoDictionary dict(p);
        std::string rate_server = m_serverName + "." + trig_name + "_pred";
        dict.checkin(rate_server, timePoint_Pred, true);
    }
    catch (std::exception & ex) {
        ers::warning(predadapter::Issue(ERS_HERE, ex));
    }
}

/**
 * HLT_ISPublish(std::string, HLT_Rate_Pred &)
 * publish HLT prediction info in new format
 */
void PredAdapter::HLT_ISPublish(std::string trig_name, HLT_Rate_Pred &TimePoint_Pred) {
    try {
        IPCPartition p(m_partitionName);
        ISInfoDictionary dict(p);
        std::string rate_server = m_serverName + "." + trig_name + "_pred";
        dict.checkin(rate_server, TimePoint_Pred, true);
    }
    catch (std::exception & ex) {
        ers::warning(predadapter::Issue(ERS_HERE, ex));
    }
}

/**
 * GO_ISPublish(std::string, Global_Offset
 * publish global offset info
 */
void PredAdapter::GO_ISPublish(std::string grp_name, Global_Offset & data) {
    try {
        IPCPartition p(m_partitionName);
        ISInfoDictionary dict(p);
        std::string rate_server = m_serverName + ".Global_Offset_" + grp_name;
        dict.checkin(rate_server, data, true);
    }
    catch (std::exception & ex) {
        ers::warning(predadapter::Issue(ERS_HERE, ex));
    }
}

/**
 * publishIS()
 * Can publish to L1 and HLT given a pNameOut 
 */
void PredAdapter::publishIS(std::string pNameOut, TimePoint_IS & predTimePoint) {
    try {
        IPCPartition p(m_partitionName);
        ISInfoDictionary dict(p);
        std::string rate_server = m_serverName + "." + pNameOut;
        dict.checkin(rate_server, predTimePoint, true);
    }
    catch (std::exception & ex) {
        ers::warning(predadapter::Issue(ERS_HERE, ex));
    }
}

/**
 * signal handler
 */
void PredAdapter::signal_handler() {
    if (m_semaphore) {
        m_semaphore->post();
    }
}

