/**
 * CHANGELOG
 * May 2017 - Daniel Zheng - Created PredAdapter.h to combine HLT and L1 Predadapters
 * August 2019 - Nick Felice - Created global offset functionality
 */


#ifndef predadapter_h
#define predadapter_h



#include <ipc/core.h>
#include <ipc/partition.h>

#include "owl/semaphore.h"

#include <is/inforeceiver.h>

#include "TRP/TriggerMenuHelpers.h"
#include "TRP/TimePoint_IS.h"
#include "TRP/L1_Rate.h"
#include "TRP/HLT_Rate.h"
#include "TRP/L1_Rate_Pred.h"
#include "TRP/HLT_Rate_Pred.h"
#include "TRP/Global_Offset.h"

#include "TFormula.h"
#include "TH1F.h"
#include "TFile.h"
#include "TTree.h"

#include <vector>
#include <string>
#include <numeric>
#include <unordered_map>

class PredAdapter {
    public:
        /**
         * Declare all the Class Level Variables and Members and Structures
         */
        PredAdapter(std::string partition_name,
                    std::string server_name,
                    std::string object_name,
                    std::string alias,
                    std::string L1_name_out,
                    std::string hlt_name_out,
                    float bunch_factor,
                    std::string trigger_groups,
                    float flagOk,
                    float flagWarning,
                    float flagError,
                    bool debug);
        ~PredAdapter() = default;
        PredAdapter(const PredAdapter &)=delete; // prevent copying

        void setLumiPartition(std::string lumiPartition, std::string lumiISServer);
        void SetHI(bool);
        void SetIS(bool);
        void SetTest(bool);
        void setDBKeysForTesting(unsigned int smk, unsigned int l1psk, unsigned int hltpsk);
        bool PredConfigure(const std::string & predictionFile);
        bool HistConfigure(std::string histPath);
        bool readTriggerDBInfo();
        void CheckHLT();
        void ParseTriggers();
        bool SubscribeToIS();
        void Run();
        void Stop();
        void signal_handler();

    private:
        static const bool success = true;
        static const bool failure = false;

        // ------------------------------------------------
        std::string m_partitionName;
        std::string m_objectName;
        std::string m_serverName;
        std::string m_lumiPartition;
        std::string m_olcIsServer;
        std::string m_l1OutputPartition;
        std::string m_hltOutputPartition;
        std::string m_triggerDB;
        std::string m_trigger_groups;
        std::string m_isServerFolder{"RunParams"};

        float m_bunchFactor;

        // configuration of OK, warning and error level
        float m_flagOk;
        float m_flagWarning;
        float m_flagError;

        // configuration flags
        bool m_isHIrun{false};
        bool m_newISFormat{false};
        bool m_isTechRun{false};

        // ------------------------------------------------
        std::unique_ptr<ISInfoReceiver>  m_receiver;
        std::unique_ptr<ISInfoReceiver>  m_olcReceiver;
        OWLSemaphore* m_semaphore{nullptr};

        // ------------------------------------------------
        unsigned int m_smk{0};
        unsigned int m_l1k{0};
        unsigned int m_hltk{0};

        // ------------------------------------------------
        bool L1_INITIALIZED{false};
        bool HLT_INITIALIZED{false};
        bool LUM_INITIALIZED{false};
        bool L1_FORMATTED{false};
        bool L1_ISCALLBACK{false};    // combining L1_ISCALLBACK/CALCULATING & HLT_ISCALLBACK/CALCULATING/
        bool L1_ISCALCULATING{false};
        bool HLT_isCallBack{false};
        bool HLT_isCalculating{false};
        bool TRIG_FORMATTED{false};

        bool m_debug;

        // ------------------------------------------------
        std::vector<std::string>  hlt_name_ps;
        std::vector<int>          hlt_ps;
        unsigned int              hlt_ps_size;

        // ------------------------------------------------
        unsigned int size_yHLT{0};
        int output_index{0};

        // ------------------------------------------------
        void callback_l1k(ISCallbackInfo *isc);
        void callback_hltk(ISCallbackInfo *isc);
        void callback_Mu(ISCallbackInfo *isc);
        void callback_l1rates(ISCallbackInfo *isc);
        void callback_hltrates(ISCallbackInfo *isc);
        void newcallback_hltrates(ISCallbackInfo *isc);
        void newcallback_l1rates(ISCallbackInfo *isc);
        void newcallback_l1rates_simple(ISCallbackInfo *isc);

        void calculate_HLT();
        void calculatePred();
        void calculate_HLT_new(unsigned int);
        void calculatePred_new(unsigned int);
        void setIndices();
        int  getL1Prescale(unsigned int);

        void HLT_ISPublish(std::string, HLT_Rate_Pred&);
        void L1_ISPublish(std::string, L1_Rate_Pred&);
        void GO_ISPublish(std::string, Global_Offset&);
        void publishIS(std::string, TimePoint_IS&);

        // ------------------------------------------------
        float myMu{0.};
        float myLumi{0.};
        float pDeadtime{0.};

        // ------------------------------------------------
        TimePoint_IS ISC_L1TimePoint;
        TimePoint_IS ISC_HLTTimePoint;
        TimePoint_IS Pred_TimePoint_HLT;
        TimePoint_IS Pred_TimePoint_L1;

        // ------------------------------------------------
        unsigned int    HLT_RunNumber{0};
        unsigned int    L1_RunNumber{0};
        unsigned short  HLT_LumiBlock{0};
        unsigned short  L1_LumiBlock{0};
        OWLTime         HLT_TimeStamp;
        OWLTime         L1_TimeStamp;

        // ------------------------------------------------
        // The number of L1 triggers on IS
        unsigned int l1rates_IS_trigN{0};
        unsigned int hltrates_IS_trigN{0};

        // ------------------------------------------------
        unsigned int YLabels_size{0}; // Number of Labels form rate object
        unsigned int TBP_index{0};    // Position of TBP in rate object
        unsigned int TAP_index{0};    // Position of TAP in rate object
        unsigned int PS_index{0};     // The position of PS from rate object

        // ------------------------------------------------
        // For HLT
        std::vector<std::string> y_HLT;
        std::vector<std::string> x_HLT;

        // For L1
        std::vector<std::string> x_L1;
        std::vector<std::string> y_L1;


        //-------------------------------------------------
        // internal maps from trigger name to index
        std::unordered_map<std::string, unsigned int> m_hltmap;
        std::unordered_map<std::string, unsigned int> m_l1map;
        std::unordered_map<std::string, unsigned int> m_extral1map;

        // ***************************************************************************
        // Structure: Trigger Ratio 
        // ***************************************************************************

        struct trigger_ratio {
            std::string nominator{""};
            std::string denominator{""};
            unsigned int xIndex_first{0};
            unsigned int xIndex_second{0};
            float rate_first{0.};
            float rate_second{0.};
            float ratio_calc{0.};
            float ratio_pred{0.};
            bool calc_success{false};
            bool found_first{false};
            bool found_second{false};
            bool status{false};
            float p1{0.};
            float p0{0.};
            int counter{0};
            float otfpred{0};

            std::vector<float> X;
            std::vector<float> Y;

            // Equations - Root stuff, Pred Paramaters usually only 2 slots used
            TString    equation;
            TFormula*  predFormula;
            double     pred_parameters[10];

            void OTFSetPred(float Mu) {
                float x = 0;
                float y = 0;
                if (Mu > 15) { // at low mu we don't care
                    x = Mu;
                    y = ratio_calc;
                    X.push_back(x);
                    Y.push_back(y);
                }

                if (X.size() > 800) X.erase(X.begin());
                if (Y.size() > 800) Y.erase(Y.begin());


                if (counter % 40 == 0) {
                    double xsum  = 0;
                    double ysum  = 0;
                    double xxsum = 0;
                    double xysum = 0;
                    int N = 0;

                    for (unsigned int i = 1; i < X.size(); i++) {
                        N++;
                        xsum += X[i];
                        ysum += Y[i];
                        xxsum += (X[i]*X[i]);
                        xysum += (X[i]*Y[i]);
                    }

                    //Calculate the new best fit slope/intercept
                    if (xsum*xsum - N*xxsum != 0) {
                        p1 = (ysum*xsum  - N*xysum)     / (xsum*xsum - N*xxsum);
                        p0 = (xsum*xysum - ysum* xxsum) / (xsum*xsum - N*xxsum);
                    }
                }
                counter++;
            }
            // ------------------------------------------------
            // OTF Predict Rate
            // ------------------------------------------------
            void OTFPredictRate(float Mu) {
                otfpred = (p1*Mu + p0);

                if (otfpred > 50000) otfpred = 50000;
            }

            void CalcRatio() {
                if (rate_second == 0) {
                    calc_success = false;
                }
                else {
                    ratio_calc = rate_first/rate_second;
                    if (ratio_calc > 10000) {
                        ratio_calc = 10000;
                    }
                    calc_success = true;
                }
            }

            bool setFormula(std::string equation, double pred_parameters[10]) {
                const char * use_name = (nominator + "/" + denominator).c_str();
                predFormula = new TFormula(Form("predFormula_%s",use_name), equation.c_str()) ;
                if (predFormula) {
                    predFormula->SetParameters(pred_parameters);
                    return true ;
                } else {
                    std::cout << "(trigger_item::setFormula()) Error, could not create TFormula for " << nominator + "/" + denominator << std::endl;
                    return false ;
                }
            }

            // ------------------------------------------------
            // Predict Ratio - Performs Ratio Ratio Prediction
            // ------------------------------------------------
            void PredictRatio (float Mu) {
                ratio_pred = predFormula->Eval(Mu);
            }

            std::string GetName() {
                return (nominator + "/" + denominator);
            }
        };

        // ***************************************************************************
        // Structure: Trigger Item
        // ***************************************************************************
        struct trigger_item {

            // ============================================================
            // Variables
            // ============================================================
            bool                   is_HLT{false};     // tracks if hlt or l1
            float                  pred_unpresc{0.};  // added from L1
            float                  rate_unpresc{0.};  // added from L1

            std::string            L1_name;
            std::string            HLT_name;
            unsigned int           xIndex{0};
            unsigned int           L1_index{0};
            unsigned int           HLT_index{0};
            unsigned int           HLT_psIndex{0};
            int                    L1_PS{0};
            int                    HLT_PS{0};
            int                    PS{0};
            float                  pred{0.};
            float                  otfpred{0.};
            float                  rate{0.};
            float                  diff{0.};
            float                  rms{0.};
            float                  nSigma{0.};
            float                  ratio{0.};
            float                  sratio{0.};
            std::vector<float>     X{};
            std::vector<float>     Y{};
            float                  p1{0.};
            float                  p0{0.};
            int                    counter{0};
            bool                   in_List{false}; // tracks if the trigger is in a global offset trigger_list
            unsigned int           list_Index{0};

	        //Histogram stuff 
            TH1F* h1{nullptr};
            TH1F* reference_hist{nullptr};
            bool hist_config{false};
            bool histFull{false};
            Double_t ksTest{0.};
            Double_t relativeMean{0.};

            // Status
            bool status{false};

            //Initialization
            bool initialized{false};

            // Check PS 0 = prescale is less than 1, 1 = prescale is not less than 1
            bool checkPS{false};

            //Flag 0 = Okay, 1 = warning, 2 = error, 3 = OMG fatal!, -1 = Can't make prediction, -2 = rate is 0 can't get
            int flag{0};

            // Equations - Root stuff, Pred Paramaters usually only 2 slots used
            std::string		equation;
            double		    pred_parameters[10];
            TFormula*       predFormula{nullptr};

            //IS Info
            L1_Rate_Pred    L1_IS_Pred;
            HLT_Rate_Pred   HLT_IS_Pred;
            L1_Rate         ISC_L1;
            HLT_Rate        ISC_HLT;

            // ============================================================
            // Member Functions
            // ============================================================

            // ------------------------------------------------
            // Set Formula - Sets the predFormulat using ROOT
            // ------------------------------------------------
            bool setFormula(std::string equation, double pred_parameters[10]) {
                const char * use_name;
                if (is_HLT) use_name = HLT_name.c_str();  // for HLT
                else
                    use_name = L1_name.c_str();        // for L1
                predFormula = new TFormula(Form("predFormula_%s",use_name), equation.c_str()) ;
                if (predFormula) {
                    predFormula->SetParameters(pred_parameters) ;
                    return true ;
                }
                std::cout << "(trigger_item::setFormula()) Error, could not create TFormula for " << HLT_name << std::endl ;
                return false ;
            }

            // ------------------------------------------------
            // OTF Predictions
            // ------------------------------------------------
            void OTFSetPred(float Mu, float L, float Nb) {
                float x = 0;
                float y = 0;
                if (Nb*L != 0 && counter % 2 == 0 && rate > 1 && Mu > 15) {
                    x = Mu;
                    y = float(rate*PS)/float(Nb*L);
                    X.push_back(x);
                    Y.push_back(y);
                }
                if (X.size() > 2000) X.erase(X.begin());
                if (Y.size() > 2000) Y.erase(Y.begin());

                if (counter % 50 == 0) {
                    double xsum  = 0;
                    double ysum  = 0;
                    double xxsum = 0;
                    double xysum = 0;
                    int N = 0;

                    if (is_HLT){
                        for (unsigned int i = 0; i < X.size() || i < Y.size(); i++) {
                            N++;
                            xsum += X[i];
                            ysum += Y[i];
                            xxsum += (X[i]*X[i]);
                            xysum += (X[i]*Y[i]);
                        }
                    }else{
                        for (unsigned int i = 1; i < X.size(); i++) {
                            N++;
                            xsum += X[i];
                            ysum += Y[i];
                            xxsum += (X[i]*X[i]);
                            xysum += (X[i]*Y[i]);
                        }
                    }

                    N++; // Implied point 0,0

                    //Calculate the new best fit slope/intercept
                    if (xsum*xsum - N*xxsum != 0) {
                        p1 = (ysum*xsum  - N*xysum)     / (xsum*xsum - N*xxsum);
                        p0 = (xsum*xysum - ysum* xxsum) / (xsum*xsum - N*xxsum);
                    }
                }

                counter++;

            }


            // ------------------------------------------------
            // PS
            // ------------------------------------------------
            void CalcPS() {
                PS = abs(L1_PS * HLT_PS);
            }

            // ------------------------------------------------
            // OTF Predict Rate
            // ------------------------------------------------
            void OTFPredictRate(float Mu, float L, float Nb) {
                otfpred = (p1*Mu + p0)*Nb*L;
                if (is_HLT){
                    if (PS != 0 && HLT_PS > 0) otfpred = otfpred/(double)PS;
                    else if (L1_PS != 0) otfpred = otfpred/(double)abs(L1_PS);  // Not recorded but still has rate
                    else if (HLT_PS == -1) otfpred = 0;  // No rate
                    else otfpred = 0;
                }else{
                    if (checkPS)  { otfpred = otfpred/PS; }
                    else          { otfpred = otfpred;    }
                }

                if (otfpred > 50000) otfpred = 50000;

            }


            // ------------------------------------------------
            // Predict Rate
            // ------------------------------------------------
            void PredictRate(float Mu, float L, float Nb) {


                if (is_HLT){
                    pred = predFormula->Eval(Mu)*Nb*L;
                    if (PS != 0 && HLT_PS > 0) pred = pred/(double)PS;
                    else if (L1_PS != 0) pred = pred/(double)L1_PS;  // Not recorded but still has rate
                    else if (HLT_PS == -1) pred = 0;  // No rate
                    else pred = 0;
                }else{
                    pred_unpresc = predFormula->Eval(Mu)*Nb*L;
                    if (checkPS)  { pred = pred_unpresc/PS; }
                    else          { pred = pred_unpresc;    }
                }
            }

            // ------------------------------------------------
            // Predict Rate + Deviations
            // ------------------------------------------------
            void CalcErrors() {
                diff =  rate - pred;

                float pred_error;
                if (is_HLT)
                    pred_error = rms*pred;
                else
                    pred_error = rms*pred_unpresc;

                if (fabs(pred_error)>0) {
                    if (is_HLT)
                        nSigma = (pred - rate)/pred_error;
                    else
                        nSigma = (pred_unpresc - rate_unpresc)/pred_error;
                } else nSigma = 0;

                if (pred != 0) {
                    ratio = fabs(rate / pred);
                    if (pred < 0.1) ratio = 1;
                } else {
                    ratio = 0;
                }

                if (pred != 0 && rate != 0) {
                    sratio = exp(fabs(log(fabs(rate/pred)))) - 1;
                    if (sratio > 3) sratio = 3;
                    if (pred < 0.1) sratio = 0;
                } else {
                    sratio = 3;
                    if (pred < 0.1) sratio = 0;
                }
                if (rate < pred) sratio *= -1;

            }

    	    // ------------------------------------------------
    	    //Create Histogram
    	    // ------------------------------------------------
            void FillHistogram() {
                if (hist_config) {
                    int n = 500;
    	            int i = 0;
    	            bool histFull = false;
    	            float percentDiff[500];
    	            h1 = new TH1F("h1", "", 100, -0.5, 0.5);
    	            percentDiff[i] = (rate-pred)/rate;
    	            i++;
    	            h1->Reset();
    	            if (!histFull) {
    	                for (int j = 0; j < i; j++) {
    	                    h1->Fill(percentDiff[j], 1/i);
    	                }
    	            }   
    	            else {
    	                for (int j = 0; j < n; j++) {
		                    h1->Fill(percentDiff[j], 1/n);
		                }
    	            }
    	   	        if (i == n) {
		           	    i -= n;
	       	            histFull = true;
    	   	        } 
    	   	        ksTest = h1->KolmogorovTest(reference_hist, "OU");
                    relativeMean = h1->GetMean(1) - reference_hist->GetMean(1);
                }
    	    }
            
            // ------------------------------------------------
            // Set Flags
            // ------------------------------------------------
            void SetFlags (float flagOk, float flagWarning, float flagError) {
                if ((is_HLT && rate != 0) || (!(is_HLT) && rate_unpresc != 0)) {
                    float flagValue = fabs(sratio);
                    int flag_set = -1;
                    if (flagValue <= flagOk)                                         { flag_set = 0; }
                    else if ((flagValue > flagOk) && (flagValue <= flagWarning))     { flag_set = 1; }
                    else if ((flagValue > flagWarning) && (flagValue <= flagError))  { flag_set = 2; }
                    else if (flagValue >= flagError)                                 { flag_set = 3; }
                    flag = flag_set;
                }
                else {
                    flag = -2;
                }
            }  

            // ------------------------------------------------
            // Display Debug Info
            // ------------------------------------------------
            void DisplayDebug (std::string stralt) const {
                if (is_HLT){
                    std::cout << stralt << "HLT PS : " << HLT_PS << std::endl;
                    std::cout << stralt << "L1 PS : " << L1_PS << std::endl;
                    std::cout << stralt << "PS : "     << PS << std::endl;
                    std::cout << stralt << "HLT output : " << rate<< std::endl;
                    std::cout << stralt << "HLT pred : " << pred<< std::endl;
                    std::cout << stralt << "HLT RMS is : " << rms<< std::endl;
                    std::cout << stralt << "HLT nSigma : " << nSigma<< std::endl;
                    std::cout << stralt << "HLT Flag : " << flag<<std::endl;
                }else{   // L1 debug info
                    std::cout << stralt << "L1 PBP : " << pred_unpresc<<std::endl;
                    std::cout << stralt << "L1 TBP : " << rate_unpresc<<std::endl;
                    std::cout << stralt << "L1 RMS : " << rms<<std::endl;
                    std::cout << stralt << "L1 nSigma : " << nSigma<<std::endl;
                    std::cout << stralt << "L1 Ratio : "<< ratio<<std::endl;
                    std::cout << stralt << "L1 Flag : " << flag << std::endl;
                }
            }

            void PrintOTF () {
                printf("<trigger>\n");
                printf("\t<yaxis>%s</yaxis>\n", HLT_name.c_str());
                printf("\t<function>pol1</function>\n");
                printf("\t<rms>0</rms>\n");
                printf("\t<par id=\"0\">%f</par>\n", p0);
                printf("\t<par id=\"1\">%f</par>\n", p1);
                printf("</trigger>\n");
            }

            void isHLT(bool is_HLT) {
                this->is_HLT =is_HLT;
            }

        };

        // ***************************************************************************
        // Structure: Trigger List
        // ***************************************************************************
        struct trigger_list {
            
            float                     avg_ks_val;
            float                     avg_mean_diff;
            std::string               grp_name;
            std::vector<trigger_item> incl_trigs;
            std::vector<float>        mean_diff_arr;
            std::vector<float>        ks_test_arr;
            Global_Offset             group_stats;

            // ------------------------------------------------
            // Create a vector of correlated triggers
            // ------------------------------------------------
            void GroupTriggers(std::vector<trigger_item> trig_list) {
                for (unsigned int i = 0; i < trig_list.size(); i++) {
                    if (trig_list[i].L1_name.find(grp_name) != std::string::npos && trig_list[i].hist_config) {
                        incl_trigs.push_back(trig_list[i]);
                        group_stats.trigs.push_back(trig_list[i].L1_name);
                    }
                }
            }

            //-------------------------------------------------
            //Set bool for triggers in list
            //-------------------------------------------------
            void SetBool(unsigned int index) { 
                for (unsigned int i = 0; i < incl_trigs.size(); i++) {
                    incl_trigs[i].in_List = true;
                    incl_trigs[i].list_Index = index;
                }
            }

            // ------------------------------------------------
            // Calculate average values for stats in triggers
            // ------------------------------------------------
            void CalcAvg() {
                for (unsigned int i = 0; incl_trigs.size(); i++) {
                    mean_diff_arr.push_back(incl_trigs[i].relativeMean);
                    ks_test_arr.push_back(incl_trigs[i].ksTest);
                }
                avg_mean_diff = std::accumulate(mean_diff_arr.begin(), mean_diff_arr.end(), 0) / incl_trigs.size();
                avg_ks_val = std::accumulate(ks_test_arr.begin(), ks_test_arr.end(), 0) / incl_trigs.size();
                group_stats.meanerr = avg_mean_diff;
                group_stats.ks = avg_ks_val;
            }

        };

        std::vector<trigger_item>  hlttrigs;
        std::vector<trigger_item>  l1trigs;
        std::vector<trigger_item>  extral1trigs;
        std::vector<trigger_ratio> trig_ratios;
        std::vector<trigger_list>  trig_groups;   // for global offset

        void updateHLTPrescales(const confadapter_imp::TriggerMenuHelpers::Mapping & hlt_ps);

        enum TimePoint_ENUM_HLT {
            PS_HLT,
            OUTPUT_HLT,
            PRED_HLT,
            DIFF_HLT,
            NSIGMA_HLT,
            FLAG_HLT,
            RATIO_HLT,
            SRATIO_HLT,
            P1_HLT,
            P0_HLT,
            OTFPRED_HLT,
            pYLabels_HLT_size
        };

        enum TimePoint_ENUM_L1 {
            PS_L1,
            TBP_L1,
            PBP_L1,
            TAP_L1,
            PAP_L1,
            DIFF_L1,
            NSIGMA_L1,
            FLAG_L1,
            RATIO_L1,
            SRATIO_L1,
            P1_L1,
            P0_L1,
            OTFPRED_L1,
            pYLabels_L1_size
        };
};
#endif
