package triggerpanel;

import Igui.IguiLogger;
import static ers.Logger.error;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Vector;
import triggerdb.Connections.ConnectionManager;
import triggerdb.Connections.InitInfo;
import static triggerpanel.TPGlobal.OKS;

/**
 *
 * @author Joerg Stelzer
 */
final class DBKeysLoader {
    
    private Vector<Vector<KeyDescriptor>> L1Prescales = null;
    private Vector<KeyDescriptor> HLTPrescales = null;
    private Vector<KeyDescriptor> Menus        = null;
    private Vector<KeyDescriptor> Bunchgroups  = null;
    private PrescaleSetAliasHandle aliases = new PrescaleSetAliasHandle();
            
    // control the state of the object
    private int currentSMK = 0;
    private boolean l1psAreLoaded = false;
    private boolean hltpsAreLoaded = false;
    private boolean psAliasesAreLoaded = false;
    
    
    // InitInfo object for connected to DB
    private final InitInfo initInfo = new InitInfo();

    DBKeysLoader() {
        this.L1Prescales = new Vector<>();
        this.L1Prescales.add( new Vector<>() );
        this.L1Prescales.add( new Vector<>() );
        this.L1Prescales.add( new Vector<>() );
        this.L1Prescales.add( new Vector<>() );
        this.HLTPrescales = new Vector<>();
        
        OKS.bind(() -> {
            // it gets called during readOKS
            // set connection string
            SetConnectionAlias(OKS.getDBConnVar().value());
        });
    }

    // Setting the SMK, invalidating the prescale sets and aliases
    void SetSMK(final int smk) {
        if(smk == this.currentSMK) {
            return;
        }
        currentSMK = smk;
        l1psAreLoaded = false;
        hltpsAreLoaded = false;
        psAliasesAreLoaded = false;
    }
    
    // access functions
    Vector<KeyDescriptor> GetPrescales(final TP.LVL level) { 
        return level == TP.LVL.L1 ? L1Prescales.get(0) : HLTPrescales;
    }

    Vector<KeyDescriptor> GetL1Prescales(final int ctpPartition)  {
        return L1Prescales.get(ctpPartition);
    }
    
    Vector<KeyDescriptor> GetHLTPrescales() { return HLTPrescales; }

    Vector<KeyDescriptor> GetMenus()        { return Menus; }

    Vector<KeyDescriptor> GetBunchgroups()  { return Bunchgroups; }
    
    PrescaleSetAliasHandle GetAliases() throws SQLException {
        UpdateAliasComponents();
        return aliases;
    }
    
    /**
     * Access to the full configuration description
     *
     * @param key key for which the description should be retrieved
     * @param level level for which the key is valid (MENU, LVL1, HLT)
     * @return the full configuration description
     */
    KeyDescriptor getKeyDescriptor(final int key, final TP.LVL level) {
        if (key == -1) {
            return KeyDescriptor.NotInIS;
        }

        if (key == 0) {
            return KeyDescriptor.NULL; // 0 is an invalid key and used as such throughout the code, -1 means 'not present in IS'
        }

        Vector<KeyDescriptor> keylist = level == TP.LVL.MENU ? TPGlobal.getDbKeysLoader().GetMenus() : TPGlobal.getDbKeysLoader().GetPrescales(level);

        if (keylist == null) {
            throw new RuntimeException("No Vector<KeyDescriptor> for level " + level);
        }

        for (KeyDescriptor kd : keylist) {
            if (kd.key == key) {
                return kd;
            }
        }
        
        IguiLogger.warning("Did not find key " + key + " in " + level + " ( " + keylist.size() + " keys available)");
        return KeyDescriptor.NotValidForSMK;
    }
    
    void FlagL1PskAsNotLoaded() {
        l1psAreLoaded = false;
    }
    
    void FlagHLTPskAsNotLoaded() {
        hltpsAreLoaded = false;
    }

    
    /********************************************************************
     * 
     * Load data from the database
     * 
     ********************************************************************/
    
    void UpdateMenus() {
        Menus = LoadMenus();
    }

    void UpdateBunchgroups() {
        Bunchgroups = LoadBunchgroups();
    }

    void UpdatePrescales(final int smk) throws SQLException {
        IguiLogger.info("UpdatePrescales called for " + smk + "\n  with currentSMK " + currentSMK + " l1psAreLoaded " + l1psAreLoaded + " hltpsAreLoaded " + hltpsAreLoaded);
        
        
        if(smk == currentSMK && l1psAreLoaded && hltpsAreLoaded) {
            return;
        }
        if(smk != currentSMK) {
            l1psAreLoaded = false;
            hltpsAreLoaded = false;
            psAliasesAreLoaded = false;
            currentSMK = smk;
        }
        IguiLogger.info("Update L1 and HLT prescales from DB for SMK " + smk);
        boolean connectedHere = DbConnect();
        if( ! l1psAreLoaded ) {
            LoadL1PrescaleKeys ( smk  );
        }
        if( ! hltpsAreLoaded ) {
            LoadHLTPrescales( smk );
        }
        if ( connectedHere ) {
            //DbDisconnect();
        }
    }
    
    void UpdateL1Prescales(final int smk) throws SQLException {
        if(smk == currentSMK && l1psAreLoaded) {
            return;
        }
        if(smk != currentSMK) {
            l1psAreLoaded = false;
            hltpsAreLoaded = false;
            psAliasesAreLoaded = false;
            currentSMK = smk;
        }

        boolean connectedHere = DbConnect();
        if( ! l1psAreLoaded ) {
            LoadL1PrescaleKeys ( smk  );
        }
        if ( connectedHere ) {
            //DbDisconnect();
        }
    }

    void UpdateHLTPrescales(final int smk) throws SQLException {
        if(smk == currentSMK && hltpsAreLoaded) {
            return;
        }
        if(smk != currentSMK) {
            l1psAreLoaded = false;
            hltpsAreLoaded = false;
            currentSMK = smk;
        }

        boolean connectedHere = DbConnect();
        if( ! hltpsAreLoaded ) {
            LoadHLTPrescales( smk );
        }
        if ( connectedHere ) {
            //DbDisconnect();
        }
    }

    private void UpdateAliasComponents() throws SQLException {
        
        boolean connectedHere = DbConnect();
        this.aliases = new PrescaleAliasLoader().load(currentSMK);
        if ( connectedHere ) {
            //DbDisconnect();
        }
        psAliasesAreLoaded = true;
    }
    
    boolean DbConnect() {
        //IguiLogger.info("In DbConnect()");
        if ( IsConnected() ) {
            //IguiLogger.info("IsConnected");
            return false;
        }
        try {
            //IguiLogger.info("Connect to TriggerDB\n" + initInfo);
            ConnectionManager.getInstance().connect(initInfo);
            return true;
        } catch (SQLException ex) {
            IguiLogger.error("Could not establish connection to TriggerDB" + ex, ex);
            ex.printStackTrace();
            return false;
        }
    }

    void DbDisconnect() {
        if ( ! IsConnected() ) {
            return;
        }
        try {
            ConnectionManager.getInstance().getConnection().close();
            IguiLogger.info("Close connection to TriggerDB");
        } catch (SQLException ex) {
            IguiLogger.error("Could not close connection to TriggerDB" + ex, ex);
        }
    }

    private boolean IsConnected() {
        try {
            boolean hasOpenConnection = 
                (ConnectionManager.getInstance().isConnectionOpen()) &&
                (!ConnectionManager.getInstance().getConnection().isClosed());
            //IguiLogger.info("Connection to TriggerDB is " + (hasOpenConnection ? "open" : "closed") );
            return  hasOpenConnection;
        } catch (SQLException ex) {
            IguiLogger.error("Could not determine TriggerDB connection status" + ex, ex);
        }
        return false;
    }

    void SetConnectionAlias(final String connectionalias) {
        try {
            IguiLogger.info("TriggerDBConnection alias: " + connectionalias);
            String dblookup = System.getenv("CORAL_DBLOOKUP_PATH");
            if(dblookup==null) { dblookup="not defined"; }
            IguiLogger.info("TriggerDBConnection dblookup path: " + dblookup);
            String dbauth = System.getenv("CORAL_AUTH_PATH");
            if(dbauth==null) { dbauth="not defined"; }
            IguiLogger.info("TriggerDBConnection dbauth path: " + dbauth);
            initInfo.setConnectionFromCommandLine(connectionalias);
            // now this is a terrible hack, but at the moment I don't see another way.
            final String schema = initInfo.getSchemaName();
            if(schema.endsWith("_R")) { // remove the _R
                initInfo.setSchemaName(schema.substring(0,schema.length()-2));
            }
            initInfo.setgoodInfo(true);
            IguiLogger.info(""+initInfo);
            
        } catch (Exception ex) {
            TPL.error("Could not set connection in InitInfo");
            error(new TriggerPanelIssue("Could not set connection in InitInfo"));
            ex.printStackTrace();
        }
    }
    
    int GetL1MenuId(final int smk) {
        int menuid = 0;
        boolean connectedHere = DbConnect();
        String schemaPrefix = ConnectionManager.getInstance().getInitInfo().getSchemaName().toUpperCase() + ".";
        String query = "SELECT SMT_L1_MENU_ID " +
            "FROM " + schemaPrefix + "SUPER_MASTER_TABLE " +
            "WHERE SMT_ID = ?";
        IguiLogger.info(query);
        try (PreparedStatement ps = ConnectionManager.getInstance().getConnection().prepareStatement(query)) {
            ps.setInt(1, smk);
            ResultSet rset = ps.executeQuery();
            if (rset.next()) {
                menuid = rset.getInt(1);
            }
        } catch (SQLException ex) {
            error(new TriggerPanelIssue("SQLException when getting l1 menu id from smk", ex));            
            if ( connectedHere ) {
                //DbDisconnect();
            }
        }
        if ( connectedHere ) {
            //DbDisconnect();
        }
        IguiLogger.info("Got menu id " + menuid);
        return menuid;
    }
    
    private void LoadL1PrescaleKeys(final int smk) {
        // clear the old set
        for( Vector<KeyDescriptor> x : L1Prescales) {
            x.clear();
        }
        if(! IsConnected()) {
            TPL.error("No TriggerDB connection, will not load prescales");
            error(new TriggerPanelIssue("No TriggerDB connection, cannot load L1 prescales"));
            return;
        }
        TPL.info("Loading L1 prescale keys for smk " + smk);
        String schemaPrefix = ConnectionManager.getInstance().getInitInfo().getSchemaName().toUpperCase() + ".";
        String query = "SELECT L1PS_ID, L1PS_NAME, L1PS_VERSION, L1PS_COMMENT, L1PS_HIDDEN, L1PS_PARTITION " +
            "FROM " + schemaPrefix + "L1_PRESCALE_SET, " + schemaPrefix + "SUPER_MASTER_TABLE, " + schemaPrefix + "L1_MENU " +
            "WHERE SMT_ID=? " +
            "AND SMT_L1_MENU_ID = L1TM_ID " +
            "AND L1PS_L1_MENU_ID = L1TM_ID " +
            "ORDER BY L1PS_ID DESC";
        try (PreparedStatement ps = ConnectionManager.getInstance().getConnection().prepareStatement(query)) {
            ps.setInt(1, smk);
            ResultSet rset = ps.executeQuery();
            while (rset.next()) {
                final int key = rset.getInt(1);
                final String name = rset.getString(2);
                final int version = rset.getInt(3);
                final String comment = (rset.getString(4)!=null && !rset.getString(4).equals("~"))?rset.getString(4):"";
                final boolean hidden = (Integer.parseInt(rset.getString(5))==1);
                final int partition = rset.getInt(6);
                L1Prescales.get(partition).add(new KeyDescriptor(key, name, version, comment, hidden));
            }
        } catch (SQLException ex) {
            TPL.error("SQLException when loading L1 prescales: " + ex.getMessage());
            error(new TriggerPanelIssue("SQLException when loading L1 prescales", ex));            
        }
        l1psAreLoaded = true;
    }

    private void LoadHLTPrescales(final int smk) {
        IguiLogger.info("Loading HLT prescale keys for smk " + smk);
        HLTPrescales.clear();
        if(! IsConnected()) {
            TPL.error("No TriggerDB connection, will not load prescales");
            error(new TriggerPanelIssue("No TriggerDB connection, cannot load HLT prescales"));
            return;
        }
        String schemaPrefix = ConnectionManager.getInstance().getInitInfo().getSchemaName().toUpperCase() + ".";
        String query = "SELECT  HPS_ID, HPS_NAME, HPS_VERSION, HPS_COMMENT, HPS_HIDDEN " +
            "FROM " + schemaPrefix + "HLT_PRESCALE_SET, " + schemaPrefix + "SUPER_MASTER_TABLE, " + schemaPrefix + "HLT_MENU " +
            "WHERE SMT_ID=? " +
            "AND SMT_HLT_MENU_ID = HTM_ID " +
            "AND HPS_HLT_MENU_ID = HTM_ID " +
            "ORDER BY HPS_ID DESC";
        try (PreparedStatement ps = ConnectionManager.getInstance().getConnection().prepareStatement(query)) {
            ps.setInt(1, smk);
            try (ResultSet rset = ps.executeQuery()) {
                while (rset.next()) {
                    final int key = rset.getInt(1);
                    final String name = rset.getString(2);
                    final int version = rset.getInt(3);
                    final String comment = (rset.getString(4)!=null && !rset.getString(4).equals("~"))?rset.getString(4):"";
                    final boolean hidden = (Integer.parseInt(rset.getString(5))==1);
                    HLTPrescales.add(new KeyDescriptor(key, name, version, comment, hidden));
                }
            }
        } catch (SQLException ex) {
            TPL.error("SQLException when loading HLT prescales: " + ex.getMessage());
            error(new TriggerPanelIssue("SQLException when loading HLT prescales", ex));            
        }
        hltpsAreLoaded = true;
        IguiLogger.info("Loaded " + HLTPrescales.size() + " HLT prescale keys for smk " + smk);
    }

    private Vector<KeyDescriptor> LoadBunchgroups()  {
        DbConnect();
        if(! IsConnected() ) {
            TPL.error("No TriggerDB connection, will not load bunchgroups");
            error(new TriggerPanelIssue("No TriggerDB connection, will not load bunchgroups"));
            return new Vector<>();
        }
        IguiLogger.info("Loading bunch group keys");
        Vector<KeyDescriptor> bgs = new Vector<>();
        String schemaPrefix = ConnectionManager.getInstance().getInitInfo().getSchemaName().toUpperCase() + ".";
        String query = "SELECT L1BGS_ID, L1BGS_NAME " +
            "FROM " + schemaPrefix + "L1_BUNCH_GROUP_SET " +
            "ORDER BY L1BGS_ID DESC";
        try (PreparedStatement ps = ConnectionManager.getInstance().getConnection().prepareStatement(query); ResultSet rset = ps.executeQuery()) {
            while (rset.next()) {
                final int key = rset.getInt(1);
                final String name = rset.getString(2);
                bgs.add(new KeyDescriptor(key,name,0,"",false));
            }
        } catch(SQLException ex) {
            TPL.error("SQLException caught when loading bunchgroups");
            error(new TriggerPanelIssue("SQLException caught when loading bunchgroups", ex));            
        }
        //DbDisconnect();
        return bgs;
    }

    private Vector<KeyDescriptor> LoadMenus() {
        DbConnect();
        if(! IsConnected() ) {
            TPL.error("No TriggerDB connection, will not load menus");
            error(new TriggerPanelIssue("No TriggerDB connection, will not load menus"));            
            return new Vector<>();
        }
        TPL.info("Loading menu keys and description from TriggerDB");
        Vector<KeyDescriptor> menus = new Vector<>();
        String schemaPrefix = ConnectionManager.getInstance().getInitInfo().getSchemaName().toUpperCase() + ".";
        String query = "SELECT SMT_ID, SMT_NAME, SMT_VERSION, SMT_COMMENT, SMT_HIDDEN " +
            "FROM " + schemaPrefix + "SUPER_MASTER_TABLE " +
            "ORDER BY SMT_ID DESC";
        try (PreparedStatement ps = ConnectionManager.getInstance().getConnection().prepareStatement(query); ResultSet rset = ps.executeQuery()) {
            while (rset.next()) {
                final int key = rset.getInt(1);
                final String name = rset.getString(2);
                final int version = rset.getInt(3);
                final String comment = (rset.getString(4)!=null && !rset.getString(4).equals("~"))?rset.getString(4):"";
                final boolean hidden = (rset.getInt(5)==1);  // SMT_HIDDEN==1 means key is hidden
                menus.add(new KeyDescriptor(key, name, version, comment, hidden));
            }
        } catch (SQLException ex) {
            TPL.error("SQLException when loading SMK: " + ex.getMessage());
            error(new TriggerPanelIssue("SQLException when loading SMK: ", ex));            
        }
        //DbDisconnect();
        return menus;
    }

    
}
