package triggertool.XML;

import triggertool.XML.Writers.XMLWriter_HLTChain;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.TimeZone;
import java.util.logging.Level;
import java.util.logging.Logger;

import triggerdb.Connections.ConnectionManager;
import triggerdb.Entities.HLT.*;
import triggerdb.Entities.SMT.SuperMasterTable;
import triggertool.HLTInteractors.ChainIdComparator;

/**
 * Write out the HLT XML files from the database.  The constructor takes
 * two arguments, the id of the hlt master table and the id of the prescale set.
 * If the prescale set id is -1, prescales of 1 are written out.
 * 
 * @author Paul Bell
 * @author Simon Head
 * @author Mark Owen
 */
public final class HLT_DBtoXML {

    /** Message Log. */
    protected static final Logger logger = Logger.getLogger("TriggerTool");
    ///Java 1.5 needs this for some reason.
    private static final long serialVersionUID = 1L;
    ///Full path for the release xml file we want to create.
    private String release_filename = "";
    ///Full path for the setup we want to create.
    private String setup_filename = "";
    ///Full path to the menu xml file we want to create.
    private String menu_filename = "";
    ///Save button fills this with the date, time and user comments for XML.
    private String header = "";
    ///Set by constructor.  Must be valid.
    private int master_id = -1;
    ///Set by the constructor.  -1 implies write out prescale values of 1 only.
    private int prescale_id = -1;
    ///HLT Master Table object.
    private HLTMaster master = null;

    /**
     * Construct the object to write HLT xml files.  Takes two arguments,
     * the id of the HLT Master Table, and the ID of the Prescale Set.
     * 
     * @param id The master table id for the hlt master you want to save to XML
     * @param psid The prescale set id
     */
    public HLT_DBtoXML(final int id, final int psid) {
        master_id = id;
        prescale_id = psid;
    }

    /**
     * Call to write the XML, pass filenames.
     * We want to save xml to the specified paths.
     * 
     * @param smt
     * @param menu menu filename
     * @param release release filename
     * @param setup l2 setup filename
     * @throws java.sql.SQLException
     */
    public void direct_write(final SuperMasterTable smt, final String menu,
            final String release, final String setup) throws SQLException {
        release_filename = release;
        menu_filename = menu;
        setup_filename = setup;

        
        //generate some comments
        String DATE_FORMAT = "yyyy-MM-dd HH:mm:ss";
        java.text.SimpleDateFormat sdf = new java.text.SimpleDateFormat(DATE_FORMAT);

        Calendar cal = Calendar.getInstance(TimeZone.getDefault());
        sdf.setTimeZone(TimeZone.getDefault());

        String ver = triggertool.TriggerTool.TT_TAG;
        String usr = ConnectionManager.getInstance().getInitInfo().getUserName();
        header = "<?xml version=\"1.0\" ?> \n"
                + "<!--This file was generated by the TriggerTool (ver. " + ver + ") -->\n"
                + "<!--By " + usr + " on " + sdf.format(cal.getTime()) + "-->\n";

        long t0 = System.currentTimeMillis();
        //load the master
        master = new HLTMaster(master_id);
        long tMaster = System.currentTimeMillis();

        logger.log(Level.INFO,"\n Time to load HLTMaster :{0}" + "sec."
                + "\n About to write release file", (tMaster - t0) / 1e3);
        write_releases(smt, release_filename);
        long tRel = System.currentTimeMillis();

        logger.log(Level.INFO, "\n Time to write HLT Release :{0}sec.\n About to write HLT menu file", (tRel - tMaster) / 1e3);
        write_menu(smt, master.get_menu(), menu_filename);
        long tMenu = System.currentTimeMillis();

        logger.log(Level.INFO, "\n Time to write HLT Menu :{0}sec.\n About to write setup file", (tMenu - tRel) / 1e3);
        write_setup(setup_filename);
        long hltSetUp = System.currentTimeMillis();

        logger.log(Level.INFO, "\n Time to write HLT SetUp :{0}sec.\n\nFinished writing HLT XML file", (hltSetUp - tMenu) / 1e3);
    }

    /**
     * Write the release XML file.
     * Write the release info to the release file.  This just consists of the 
     * release name and version.
     * 
     * @param release HLT release table from the database.
     */
    private void write_releases(final SuperMasterTable smt, final String release_filename) throws SQLException {

        logger.log(Level.FINE, "Writing Releases XML filename: {0}", release_filename);

        try {
            BufferedWriter out = new BufferedWriter(new FileWriter(release_filename));
            out.write(header);
            for (HLTRelease release : smt.getReleases()) {
                out.write("<RELEASE name =\"" + release.get_name() + "\"/>");
            }
            out.close();
            logger.fine("Release file written ok");
        } catch (IOException e) {
            logger.log(Level.WARNING, "Failed to write HLT release file:\n {0}\n {1}", new Object[]{release_filename, e.getMessage()});
        }

        logger.fine("Done writing XML release");
    }

    /**
     * Write the HLT Menu file.
     * TODO Tiago-Performance: avoid writing out file line-by-line. This is very slow. 
     * @param smt the SuperMasterTable.
     * @param menu HLT menu in the database.
     * @param menu_filename  the name of the file to write to.
     */
    private void write_menu(final SuperMasterTable smt, final HLTTriggerMenu menu, final String menu_filename) throws SQLException {
        logger.log(Level.INFO, "Writing Menu HLT XML filename: {0} ", menu_filename);

        // Load Menu signatures, elements, algos before writing xml
        //menu.loadMenu();
        ArrayList<HLTTriggerChain> chains = menu.getChains();
        //
        // Write XML
        HLTPrescaleSet hps = null;
        if (prescale_id > 0) {
            hps = new HLTPrescaleSet(prescale_id);
        }

        //            out.write(header);
        StringBuilder xmlBuilder = new StringBuilder();
        xmlBuilder.append(header);
        //We use the supermaster comment here!            
        xmlBuilder.append("<HLT_MENU menu_name=\"")
                .append(menu.get_name())
                .append("\" prescale_set_name=\"")
                .append(hps != null ? hps.get_name() : "")
                .append("\" comment=\"")
                .append(smt.get_comment()
                        .replace("&", "&amp;")
                        .replace("\"", "&quot;")
                        .replace("<", "&lt;")
                        .replace(">", "&gt;")
                        .replace("'", ""))
                .append("\">\n");
        xmlBuilder.append("  <SEQUENCE_LIST>\n");
        //out.write(xmlBuilder.toString());
        //xmlBuilder = new StringBuilder();
        logger.log(Level.FINE, "Getting Menu -> Chain links for menu id: {0}", menu.get_id());
        //vector to remember which sequences we already wrote to xml
        ArrayList<String> writtenSeqs = new ArrayList<>();
        for (HLTTriggerChain chain : chains) {
            //get signature links
            for (HLTTriggerSignature signature : chain.getSignatures()) {
                //get TE links
                for (HLTTriggerElement el : signature.getElements()) {
                    if (writtenSeqs.contains(el.get_name())) {
                        continue; // don't write same sequence twice 	 
                    }
                    writtenSeqs.add(el.get_name());
                    String algo_name = "";
                    for (HLTComponent algo : el.getAlgorithms()) {
                        String name = algo.get_name();
//                        if (algo.get_name().indexOf("_") >= 0) {
//                            //name = (algo.get_name()).substring(0, algo.get_name().indexOf("_"));
//                        }
                        algo_name += name + "/" + algo.get_alias() + " ";
                    }
                    algo_name = algo_name.trim();
                    String input_names = "";
                    for (HLTTE_TE link : el.getInputElements()) {
                        input_names += link.get_element_inp_id() + " ";
                    }
                    
                    input_names = input_names.trim();
                    if (input_names.equals("~")) {
                        input_names = "";
                    }
                      
                    xmlBuilder.append("    <SEQUENCE algorithm=\"" + algo_name + "\" input=\"" + input_names + "\" output=\"" + el.get_name());
                    logger.log(Level.FINE, "Before write {0}", el.get_topo_start_from());
                    if(ConnectionManager.TopoStartFromColumnExists == 1 && el.get_topo_start_from() != null){
                        String topo = el.get_topo_start_from();
                        if(!topo.equals("~") ){
                             xmlBuilder.append("\" topo_start_from=\"" + topo);
                        }
                    }
                    xmlBuilder.append("\"/>\n");
                }
            }
        }

        xmlBuilder.append("  </SEQUENCE_LIST>\n");
        //out.write(xmlBuilder.toString());
        //xmlBuilder = new StringBuilder();
        logger.fine("Done writing sequence list");
        logger.fine("Starting chain list");

        xmlBuilder.append("  <CHAIN_LIST>\n");

        Collections.sort(chains, new ChainIdComparator());
        for (HLTTriggerChain chain : chains) {
            xmlBuilder.append(new XMLWriter_HLTChain().WriteChain(chain, hps.GetPrescalesForChain(chain.get_chain_counter())));           
        }
        
        xmlBuilder.append("  </CHAIN_LIST>\n");
        xmlBuilder.append("</HLT_MENU>\n");

        try {
            BufferedWriter out = new BufferedWriter(new FileWriter(menu_filename));
            out.write(xmlBuilder.toString());
            out.close();
        } catch (IOException e) {
            String msg = "Failed to write HLT menu file:\n " + menu_filename
                    + "\n " + e.getMessage();
            logger.severe(msg);
        }
        logger.log(Level.INFO, "Finished Writing Menu HLT XML filename: {0} ", menu_filename);
    }

    /**
     * Writes the EF and L2 setup files.
     * In the new schema, the menu part of the setup must be accessed via the algorithms!
     * @param level pass either "EF" or "L2"
     * @param filename The output filename
     */
    private void write_setup(final String filename) throws SQLException {
        logger.log(Level.FINE, "Writing HLT Setup: {0}", filename);

        try {
            BufferedWriter out = new BufferedWriter(new FileWriter(filename));
            out.write(header);

            //get menu id for accessing menu components
            int menu_id = master.get_menu_id();

            //get setup ids for accessing infrastructure components
            int id = master.get_setup_id();
            //No need to write anything for a menu only configuration.
            if(id>0){
                //The version which may appear in the xml file is not like a db version, so don't write it!
                out.write("<setup name=\"" + master.get_name() + "\">\n");

                //PART 1. Write out the i-setup part...
                HLTSetup setup = new HLTSetup(id);
                ArrayList<String> writtenComps = new ArrayList<>();
                for (HLTComponent comp : setup.getInfraStructureComponents()) {
                    logger.fine("Setup comp");
                    writtenComps.add(comp.get_alias() + comp.get_name());
                    WriteOutComponent(comp, out);
                }

                
                //...PART 2: Write out menu setup part, in same file
                /*HLTTriggerMenu menu = new HLTTriggerMenu(menu_id);
                for (HLTTriggerChain chain : menu.getChains()) {
                    logger.info("Chain "+ chain.get_name());
                    for (HLTTriggerSignature signature : chain.getSignatures()) {f
                        logger.info("Signature "+ signature.get_name());
                        for (HLTTriggerElement element : signature.getElements()) {
                            logger.info("Element "+element.get_name());
                            for (HLTComponent algo : element.getAlgorithms()) {
                                logger.info("Algo "+algo.get_name());
                                ///Only write each unique component once
                                if (writtenComps.contains(algo.get_alias() + algo.get_name())) {
                                    logger.info("MAO: Skipping duplicate component " + algo.get_alias() + algo.get_name());
                                    continue;
                                }
                                writtenComps.add(algo.get_alias() + algo.get_name());
                                logger.info("Menu comp");
                                WriteOutComponent(algo, out);
                            }
                        }
                    }
                }*/
                out.write("</setup>\n");
            }
            out.close();
        } catch (IOException e) {
            String wrgMsg = "Failed to write HLT setup file:\n " + filename
                    + "\n " + e.getMessage();
            logger.warning(wrgMsg);
        }
        logger.fine("Done writing HLT Setup");
    }//write_setup   

    /**
     * Helper function to write out HLT Component to xml file.
     * Added to make it easier to write out the children info.
     * @param comp The HLT component to write out.
     * @param out output file buffer to write to.
     */
    private void WriteOutComponent(final HLTComponent comp,
            final BufferedWriter out) throws IOException, SQLException {

        //logger.info("MAO: Entering WriteOutComponent for " + comp.get_name() + " " + comp.get_alias());
        out.write("  <component alias=\"" + comp.get_alias().replace("\"", "&quot;").replace("<", "&lt;").replace(">", "&gt;") + "\" "
                + "name=\"" + comp.get_name().replace("\"", "&quot;").replace("<", "&lt;").replace(">", "&gt;") + "\" ");
        //+"type=\"" + comp.get_type() + "\"" +
        if (comp.get_type().equals("TopAlg")) {
            out.write("topalg=\"1\" alg=\"1\" steeralg=\"0\" ");
        } else { //not topalg
            out.write("topalg=\"0\" ");
            if (comp.get_type().equals("steeralg")) {
                out.write("alg=\"1\" steeralg=\"1\" ");
            } else {
                out.write("alg=\"0\" steeralg=\"0\" ");
            }
        }
        if (comp.get_type().equals("Service")) {
            out.write("svc=\"1\" ");
        } else {
            out.write("svc=\"0\" ");
        }
        if (comp.get_type().equals("PublicTool")) {
            out.write("pubT=\"1\" ");
        } else {
            out.write("pubT=\"0\" ");
        }
        if (comp.get_type().equals("PrivateTool")) {
            out.write("privT=\"1\" ");
        } else {
            out.write("privT=\"0\" ");
        }
        if (comp.get_type().equals("Auditor")) {
            out.write("aud=\"1\"");
        } else {
            out.write("aud=\"0\"");
        }
        String pyName = comp.get_py_name();
        if(pyName.equals("~")) pyName = "";
        String pyPackage = comp.get_py_package();
        if(pyPackage.equals("~")) pyPackage = "";
        out.write(" py_name=\"" + pyName + "\"");
        out.write(" py_package=\"" + pyPackage + "\"");
        out.write(">\n");

        /// Write out parameters
        for (HLTParameter param : comp.getParameters()) {
            String value = param.get_value().replace("&", "&amp;").replace("\"", "&quot;").replace("<", "&lt;").replace(">", "&gt;");
            if (value.equals("~")) {
                value = "";
            }  // if we invoke replace straightaway, a null pointer exception will be thrown if no ~ is present.
            out.write("    <parameter name=\"" + param.get_name().replace("\"", "&quot;").replace("<", "&lt;").replace(">", "&gt;") + "\" "
                    + "value=\"" + value + "\" "
                    + "op=\"" + param.get_op().replace("\"", "&quot;").replace("<", "&lt;").replace(">", "&gt;") + "\""
                    + "/>\n");
        }

        /// Now write out any child components 
        //for (HLTComponent childcomp : comp.getAllChildComponents()) {
        //- this is one loop too many!
        // getAllChildComponents gets children, grandchildren etc - then we end up here again
        // with the child, and it gets it children, which are the grandchildren we have already!
        // only need to get children:
        for (HLTComponent childcomp : comp.getChildComponents()) {
            logger.fine("Child comp");
            //Recursive
            //logger.info("MAO: Found child component of " + comp.get_name() + " " + comp.get_alias() +", recalling WriteOutComponent");
            WriteOutComponent(childcomp, out);
        }
        out.write("  </component> <!-- "+comp.get_alias().replace("\"", "&quot;").replace("<", "&lt;").replace(">", "&gt;")+" -->\n");
    }
}
