<?php

/**
 *	ATLAS Point 1 roles retrieved from LDAP.
 *	
 * 	@author Marius Leahu
**/

if (version_compare(phpversion(), '5.0.0', '<'))
{
    die('You need at least PHP version 5.0.0!');
}

class LDAP_roles
{
    // keys for the roles array
    const ROLE_ASSIGNABLE = 'assignable';
    const NODE_ID = 'nodeid';    
    const PARENT_NODE_IDS_ARRAY = 'parentnodeids';
    const SENIORS_ARRAY = 'seniors';
    const JUNIORS_ARRAY = 'juniors';
    const MEMBERS_ASSIGNED_ARRAY = 'membersassigned';
    const MEMBERS_ENABLED_ARRAY = 'membersenabled';
    
    const LDAP_ROLE_PREFIX_ASSIGNED = 'RA-';
    const LDAP_ROLE_PREFIX_ENABLED = 'RE-';
    const LDAP_OBJ_ROLE = 'amRole';
    const LDAP_ATTR_ROLE_NAME = 'cn';
    const LDAP_ATTR_ROLE_ASSIGNABLE = 'amroleassignable';
    const LDAP_ATTR_MEMBERS = 'nisnetgrouptriple';
    const LDAP_ATTR_SENIORS = 'membernisnetgroup';
    
    private $ldap_host;
    private $ldap_basedn;
    private $roles = array();

    private $verboseMode;
    
    public function __construct($ldap_host, $ldap_basedn) {
	$this->ldap_host = $ldap_host;
	$this->ldap_basedn = $ldap_basedn;
	$this->verboseMode = false;
    }

    public function setVerboseMode($verboseMode){
	$this->verboseMode = $verboseMode;
    }

    private function print_msg($msg) {
	if ( $this->verboseMode ) echo $msg;
    }

    private function print_r_msg($array) {
	if ( $this->verboseMode ) print_r($array);
    }
    /**
     * retrieves the roles from LDAP
     * @param string $ldap_search_filter 
     * @return array with search results as returned by the ldap search function 
     */    
    private function searchLDAPforRoles($ldap_search_filter) {
	$con=ldap_connect($this->ldap_host);
	if ($con) { 
		ldap_set_option($con, LDAP_OPT_PROTOCOL_VERSION, 3);
		
		$ldapb=ldap_bind($con);     // this is an "anonymous" bind, typically read-only access
		if ($ldapb){
			$sr=ldap_search($con, $this->ldap_basedn, $ldap_search_filter);
			ldap_sort($con, $sr, self::LDAP_ATTR_ROLE_NAME);
			$info = ldap_get_entries($con, $sr);
			return $info;
		} else {
			echo "<h4>Unable to bind to LDAP server ". $this->ldap_host ." </h4>";
		}
		ldap_close($con);
	} else {
		echo "<h4>Unable to connect to LDAP server</h4>";
	}    
    }

    /**
     * get the roles in an array
     * @param int $role_prefix optionally parameter to get only the role with a given prefix
     * @return array with the following structure: for each rolename, there are:
     *		the boolean attribute $roles[rolename][assignable]
     *		the node id $roles[rolename][nodeid]
     *		the parent node id $roles[rolename][parentnodeid]
     *		the users member of this role $roles[rolename][users]
     *		the roles seniors and juniors: $roles[rolename][seniors], $roles[rolename][juniors]
     */
    public function getRoles($role_prefix="") {
	$pos = strpos($role_prefix, "!");
	if (  $pos !== false && $pos == 0) {
		$role_prefix = substr($role_prefix, 1);
	}
	$role_name_filter =  "(|". 
				"(". self::LDAP_ATTR_ROLE_NAME ."=". self::LDAP_ROLE_PREFIX_ASSIGNED . $role_prefix ."*)".
				"(". self::LDAP_ATTR_ROLE_NAME ."=". self::LDAP_ROLE_PREFIX_ENABLED . $role_prefix ."*)".
			     ")";
	if (  $pos !== false && $pos == 0) {
		$role_name_filter =  "(&".
					"(!". $role_name_filter .")".
					"(|" . 
					    "(" . self::LDAP_ATTR_ROLE_NAME ."=". self::LDAP_ROLE_PREFIX_ASSIGNED ."*)".
					    "(" . self::LDAP_ATTR_ROLE_NAME ."=". self::LDAP_ROLE_PREFIX_ENABLED ."*)".
					")".
				      ")";
	}
	$ldap_search_filter = "(&". $role_name_filter . "(objectClass=". self::LDAP_OBJ_ROLE . "))";
	$this->print_msg("LDAP search filter [".$ldap_search_filter ."]");

	$ldap_search_result=$this->searchLDAPforRoles($ldap_search_filter);
	$this->print_r_msg($ldap_search_result);

	$this->roles = array();
	// iterate first time to extract the role names from the search result
	for($i=0; $i<$ldap_search_result['count']; $i++) {
		// get the role name from LDAP search result
		$role_name = $ldap_search_result[$i][self::LDAP_ATTR_ROLE_NAME][0];

		$pos = strpos($role_name, self::LDAP_ROLE_PREFIX_ASSIGNED);
		// check if the role name has the ASSIGNED prefix at position 0
		if ($pos === false || $pos != 0 ) continue;

		//cut both RA and RE prefixes
		$role_name = substr($role_name, strlen(self::LDAP_ROLE_PREFIX_ASSIGNED));

		// initialize the attributes
		$this->roles[$role_name][self::ROLE_ASSIGNABLE] = "";
		$this->roles[$role_name][self::NODE_ID] = $i + 1;
		$this->roles[$role_name][self::PARENT_NODE_IDS_ARRAY] = array();
		$this->roles[$role_name][self::MEMBERS_ASSIGNED_ARRAY] = array();
		$this->roles[$role_name][self::MEMBERS_ENABLED_ARRAY] = array();
		$this->roles[$role_name][self::SENIORS_ARRAY] = array();
		$this->roles[$role_name][self::JUNIORS_ARRAY] = array();
	}
	
	$current_node_id = $ldap_search_result['count'] + 1;
	
	// set the value of the attributes
	for($i=0; $i<$ldap_search_result['count']; $i++) {
		$ldap_role = $ldap_search_result[$i];
		
		// get the role name from LDAP search result
		$role_name = $ldap_role[self::LDAP_ATTR_ROLE_NAME][0];
		$role_type_prefix = self::LDAP_ROLE_PREFIX_ASSIGNED;

		$pos = strpos($role_name, self::LDAP_ROLE_PREFIX_ENABLED);
		// check if the role name has the ENABLED prefix at position 0
		if ($pos !== false && $pos == 0 ) {
		    $role_type_prefix = self::LDAP_ROLE_PREFIX_ENABLED;
		}
		// cut the prefix
		$role_name = substr($role_name, strlen($role_type_prefix));

		// set the assignable attribute
		$this->roles[$role_name][self::ROLE_ASSIGNABLE] = false;
		if (array_key_exists(self::LDAP_ATTR_ROLE_ASSIGNABLE, $ldap_role)) {
			$this->roles[$role_name][self::ROLE_ASSIGNABLE] = $ldap_role[self::LDAP_ATTR_ROLE_ASSIGNABLE][0];
		} else {
			echo "ERROR: Attribute [". self::LDAP_ATTR_ROLE_ASSIGNABLE ."] not found for role [". $role_name ."]";
		}
		
		//add the users
		if (array_key_exists(self::LDAP_ATTR_MEMBERS, $ldap_role)) {
			for($j=0; $j<$ldap_role[self::LDAP_ATTR_MEMBERS]['count']; $j++) {
			    if ( $role_type_prefix == self::LDAP_ROLE_PREFIX_ASSIGNED)
				$this->roles[$role_name][self::MEMBERS_ASSIGNED_ARRAY][$j] = str_replace(array("(", ",", ")"), "", $ldap_role[self::LDAP_ATTR_MEMBERS][$j]);
			    if ( $role_type_prefix == self::LDAP_ROLE_PREFIX_ENABLED)
				$this->roles[$role_name][self::MEMBERS_ENABLED_ARRAY][$j] = str_replace(array("(", ",", ")"), "", $ldap_role[self::LDAP_ATTR_MEMBERS][$j]);
			}
		}
		
		// process the seniors and juniors only for the roles assigned
		if ( $role_type_prefix == self::LDAP_ROLE_PREFIX_ENABLED)  continue;
		
		//add the senior roles of this role, them this role as junior of the senior
		if (array_key_exists(self::LDAP_ATTR_SENIORS, $ldap_role)) {
			for($j=0; $j<$ldap_role[self::LDAP_ATTR_SENIORS]['count']; $j++) {
				// get the senior name
				$senior_name = $ldap_role[self::LDAP_ATTR_SENIORS][$j];
				$senior_name = substr($senior_name, strlen($role_type_prefix));
				
				// check if the senior is still part of the roles in the search result; if not, then skip
				if ( ! array_key_exists($senior_name, $this->roles) ) continue;
				
				// add it to the list of seniors
				$this->roles[$role_name][self::SENIORS_ARRAY][$j] = $senior_name;
				
				// add the junior relation
				$this->roles[$senior_name][self::JUNIORS_ARRAY][] = $role_name;
			}
		}
	}
	return $this->roles;
    }
}

?>
