#ifndef COCA_DB_BASE
#define COCA_DB_BASE
//----------------------------------------------------------------------
// File and Version Information:
//      $Id$
//----------------------------------------------------------------------

//-----------------
// C/C++ Headers --
//-----------------
#include <memory>
#include <string>
#include <RelationalAccess/ISchema.h>
#include <RelationalAccess/ISessionProxy.h>
#include <RelationalAccess/ITable.h>

//----------------------
// Base Class Headers --
//----------------------

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "coca/common/NextId.h"
#include "coca/common/Str2Id.h"
#include "coca/common/Transaction.h"

//------------------------------------
// Collaborating Class Declarations --
//------------------------------------
namespace daq {
namespace coca {
    class Transaction;
}
}

//              ---------------------
//              -- Class Interface --
//              ---------------------

namespace daq {
namespace coca {

/// @addtogroup coca_common

/** 
 *  @ingroup coca_common
 *  @brief Class implementing common operations on CoCa database.
 * 
 *  Despite its name this class is not supposed to be a base class 
 *  for anything, as it does not define any public interface. 
 *  Its purpose is to keep common parts of the implementation so
 *  that other database classes can be implemented using the code 
 *  from this class. It is OK to use this class as private base 
 *  class to express "implemented in terms of" relation, it can 
 *  also be used as a member of other classes.  
 */

class DBBase {
public:

    /**
     *  @brief Constructor takes CORAL connection string.
     *  
     *  Instantiates read-only instance.
     *
     *  @param[in] connStr  CORAL connection string.
     */
    explicit DBBase (const std::string& connStr);

    /**
     *  @brief Constructor takes CORAL connection string and boolean update flag.
     *
     *  @param[in] connStr  CORAL connection string.
     *  @param[in] update   If true then create modifiable instance.
     */
    DBBase (const std::string& connStr, bool update);

    // noncopyable
    DBBase(const DBBase&) = delete;
    DBBase& operator=(const DBBase&) = delete;

    // Destructor
    ~DBBase () throw ();

    /// Get the name of the archive database table
    const std::string& archiveTableName() const { return m_arName; }

    /// Get the name of the cache database table
    const std::string& cacheTableName() const { return m_caName; }
    
    /**
     * @brief Get the handle for archive database table
     * 
     *  @throw coral::Exception Thrown for databases-related conditions.
     */
    coral::ITable& archiveTable(Transaction& tran) const;

    /**
     *  @brief Get the handle for cache database table
     *  
     *  @throw coral::Exception Thrown for databases-related conditions.
     */
    coral::ITable& cacheTable(Transaction& tran) const;
    
    /// Access STR2ID mapping object
    Str2Id& s2i() { return m_s2i; }

    /**
     *  @brief Generate next ID for given sequence name and index (see NextId)
     * 
     * @param[in]  seq   Sequence name, non-empty string.
     * @param[in]  seqIdx  Sequence index, a number.
     * @param[in]  tran  Transaction context.
     * 
     *  @throw coral::Exception Thrown for databases-related conditions.
     */
    unsigned nextId (const std::string& seq, unsigned seqIdx, Transaction& tran) {
        return m_nextid.next(seq, seqIdx, tran);
    }
    
    /**
     *  @brief Find a string for an ID (wraps Str2Id::getStr).
     *  
     *  @param[in] id  ID for which we need to find a string.
     *  @param[in] tran  Transaction context (see Transaction).
     *  @return  String corresponding to given ID, empty string is returned for zero ID.
     *  
     *  @throw Issues::NoSuchRecord Thrown if there is no such record exist.
     *  @throw coral::Exception Thrown for databases-related conditions.
     */
    std::string getStr (unsigned id, Transaction& tran);

    /**
     *  @brief Find an ID for a string (wraps Str2Id::getId).
     *  
     *  @param[in] str  String for which we need to find ID.
     *  @param[in] tran  Transaction context (see Transaction).
     *  @return  ID corresponding to given string, zero returned for empty string.
     *  
     *  @throw Issues::NoSuchRecord Thrown if there is no such record exist.
     *  @throw coral::Exception Thrown for databases-related conditions.
     */
    unsigned getId (const std::string &str, Transaction& tran);
    
    /**
     *  @brief Create new ID for a string or return existing one (wraps Str2Id::insert).
     *  
     *  If string contains control characters they will be replaces with 
     *  underscores first (there is a potential for name clash). Empty 
     *  strings are not stored in database, instead we return zero ID.
     *  
     *  @param[in] str  String for which we need to find or create ID.
     *  @param[in] tran  Transaction context (see Transaction).
     *  @return  ID corresponding to given string, zero returned for empty string.
     *  
     *  @throw coral::Exception Thrown for databases-related conditions.
     */
    unsigned insStr (const std::string &str, Transaction& tran);

    /**
     *  @brief Returns CORAL session object.
     *  
     *  @throw coral::Exception Thrown for databases-related conditions.
     */
    std::shared_ptr<coral::ISessionProxy> session();

    /**
     *  Returns default CORAL connection string for CoCa database.
     */
    static std::string defConnStr();

private:
    
    // data members
    const std::string m_connStr;  ///< CORAL connection string
    NextId m_nextid;              ///< ID generator object
    Str2Id m_s2i;                 ///< String to ID mapping
    const std::string m_arName;   ///< Archive table name
    const std::string m_caName;   ///< Cache table name
    const bool m_update;          ///< If true then create update sessions

};

} // namespace coca
} // namespace daq

#endif // COCA_DB_BASE
