#ifndef COCA_FILESYSTEM_H
#define COCA_FILESYSTEM_H

//--------------------------------------------------------------------------
// File and Version Information:
// 	$Id$
//
// Description:
//	Class FileSystem.
//
//------------------------------------------------------------------------

//-----------------
// C/C++ Headers --
//-----------------
#include <string>

//----------------------
// Base Class Headers --
//----------------------

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "ers/ers.h"

//------------------------------------
// Collaborating Class Declarations --
//------------------------------------

//		---------------------
// 		-- Class Interface --
//		---------------------

namespace daq {
namespace coca {

    /**
     *  @class FileSystem::Exception
     *  @brief Base class for the exception-type issues generated by FileSystem methods.
     */
    ERS_DECLARE_ISSUE (FileSystem, Exception, , ) 

    /**
     *  @class FileSystem::MkdirFileExists
     *  @brief Exception generated while trying to make a directory when regular file
     *  name with the same pathname exists already.
     */
    ERS_DECLARE_ISSUE_BASE (FileSystem, MkdirFileExists, Exception,
     "cannot make directory " << name << ", there is a file with this name", , 
     ((std::string) name)) 

     /**
      *  @class FileSystem::MkdirFailed
      *  @brief Exception generated when mkdir() call fails.
      */
    ERS_DECLARE_ISSUE_BASE (FileSystem, MkdirFailed, Exception,
     "failed to create new directory " << name , , ((std::string) name))

     /**
      *  @class FileSystem::BasedirMismatch
      *  @brief Exception generated when pathname does not start with the basedir name.
      */
    ERS_DECLARE_ISSUE_BASE (FileSystem, BasedirMismatch, Exception,
     "Path name '" << pathname << "' does not start with '" << basedir << "'", , 
     ((std::string) basedir) ((std::string) pathname))

/// @addtogroup coca_common

/**
 * @ingroup coca_common
 *  @brief Helper functions for filesystem manipulations, these are additions
 *  to boost/filesystem and use classes from that package.
 *
 *  This software was developed for the ATLAS project.  If you use all or 
 *  part of it, please give an appropriate acknowledgment.
 *
 *  @version $Id$
 *
 *  @author Andy Salnikov
 */

namespace FileSystem  {

    /**
     * @brief Create directories.
     * 
     * This function creates directory and all its parent directories,
     * and in addition sets UNIX group for new directories and group bit
     * if necessary.
     * 
     * @param[in] path Path name for new directory.
     * @return False if directory exists already, true if new directory was created.
     * 
     * @throw MkdirFileExists Thrown if the path or some part of it already exists
     *                        but it is not a directory.
     * @throw MkdirFailed  Thrown if the directory creation has failed.
     */
    bool create_dirs(const std::string& path);

    
    /**
     *  @brief Join two components into a full path name. 
     *  
     *  In addition to simply joining two components into single path 
     *  it also removes extra slashes from resulting path name.
     *  
     * @param[in] path1 First component of the resulting path.
     * @param[in] path2 Second component of the resulting path.
     * @return Joined path name.
     */
    std::string join(const std::string& path1, const std::string& path2);
    
    /**
     *  @brief Splits pathname into directory and relative path. 
     *  
     *  If pathname is relative then current working directory is prepended to make
     *  it absolute. If baseDir is not empty then path must start with the same directory.
     *  If baseDir is empty then basename is used for relative path and directory, 
     *   otherwise baseDir is stripped from pathname and remaning part used for relative path.
     *  
     *  @param[in] path    Pathname which can be absolute or relative.
     *  @param[in] baseDir Base directory name, may be empty.
     *  @param[in] cwd     Current working directory, if empty use getcwd().
     *  @return Directory name and relative name.
     *  
     *  @throw BasedirMismatch Thrown if path does not start with baseDir.
     */
    std::pair<std::string, std::string> split(const std::string& path, const std::string& baseDir,
            const std::string& cwd = std::string());
    
} // namespace FileSystem

} // namespace coca
} // namespace daq

#endif // COCA_FILESYSTEM_H
