#ifndef COCA_SYSTEM_H
#define COCA_SYSTEM_H
//--------------------------------------------------------------------------
// File and Version Information:
// 	$Id$
//
// Description:
//	Class System.
//
//------------------------------------------------------------------------

//-----------------
// C/C++ Headers --
//-----------------
#include <string>
#include <iosfwd>
#include <sys/types.h>

//----------------------
// Base Class Headers --
//----------------------

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "coca_dal/UnixProperties.h"

//------------------------------------
// Collaborating Class Declarations --
//------------------------------------

//		---------------------
// 		-- Class Interface --
//		---------------------

namespace daq {
namespace coca {

    
    /**
     *  @class SystemIssues::Exception
     *  @brief Base class for exception generated by System package
     */
    ERS_DECLARE_ISSUE (SystemIssues, Exception, , )
    
    /**
     *  @class SystemIssues::GroupNotFound
     *  @brief Exception generated when specified group name does not exist
     */
    ERS_DECLARE_ISSUE_BASE (SystemIssues, GroupNotFound, Exception,
            "cannot get information for group: " << group, ,((std::string) group))

    /**
     *  @class SystemIssues::NotInitialized
     *  @brief Exception generated when singleton has not been initialized
     */
    ERS_DECLARE_ISSUE_BASE (SystemIssues, NotInitialized, Exception,
            "System singleton was not properly initialized", , )
    

/// @addtogroup coca_common

/**
 *  @ingroup coca_common
 *
 *  @brief Set of the system-level utilities.
 * 
 *  These utilities borrow the implementation from retired pfzlib/System
 *  package using slightly improved logic.
 *
 *  This software was developed for the ATLAS project.  If you use all or 
 *  part of it, please give an appropriate acknowledgment.
 *
 *  @version $Id$
 *
 *  @author Andy Salnikov
 */

class System  {
public:

    /**
     *  @brief Creates an instance of the System object and initializes it.
     *
     *  This method creates a singleton instance of this class. It must be called
     *  exactly once before instance() method is called.
     *   
     *  @param[in] properties  DAL configuration object.
     *   
     *  @throw SystemIssues::GroupNotFound Thrown if the name of the UNIX group cannot be resolved.
     */
    static void initialize(const coca_dal::UnixProperties& properties);

    /**
     *  @brief return a singleton instance of this class.
     *
     *  The initialize() method must be called once before calling this method,
     *  otherwise an exception will be generated.
     *
     *  @throw SystemIssues::NotInitialized
     */
    static System& instance();
    
    /// Get current umask.
    mode_t umask() const { return m_umask; }
    
    /// Return group name for new directories.
    const std::string& newDirGroup() const { return m_newDirGroup; }

    /// Return group ID for new directories.
    gid_t newDirGroupId() const { return m_newDirGroupId; }
    
    /// Returns true if new directories need setgid bit
    bool newDirHasSetGid() const { return m_newDirHasSetGid; }
    
    /// Dump configuration info to a stream
    void print(std::ostream& out) const;

private:

    /// Constructor gets parameters from DAL object.
    System(const coca_dal::UnixProperties& properties);

    mode_t m_umask;
    std::string m_newDirGroup;
    gid_t m_newDirGroupId;
    bool m_newDirHasSetGid;

};

} // namespace coca
} // namespace daq

#endif // COCA_SYSTEM_H
