#ifndef COCA_SERVERISSUES_H
#define COCA_SERVERISSUES_H

//--------------------------------------------------------------------------
// File and Version Information:
// 	$Id$
//
// Description:
//	Class Issues.
//
//------------------------------------------------------------------------

//-----------------
// C/C++ Headers --
//-----------------

//----------------------
// Base Class Headers --
//----------------------
#include "ers/ers.h"

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "coca/common/Issues.h"

//------------------------------------
// Collaborating Class Declarations --
//------------------------------------

//		---------------------
// 		-- Class Interface --
//		---------------------

// Void Structure, Just For SemiColon; to make doxygen happier
#define VSJFSC struct _VSJFSC_

namespace daq {
namespace coca {

/// @addtogroup coca_server

/**
 *  @ingroup coca_server
 *
 *  @namespace daq::coca::ConfigIssues
 *  @brief Collection of issue classes raised by Config class
 */

/**
 *  @ingroup coca_server
 *
 *  @namespace daq::coca::FreeSpaceIssues
 *  @brief Collection of issue classes raised by FreeSpace class
 */

/**
 *  @ingroup coca_server
 *
 *  @namespace daq::coca::GetFilesIssues
 *  @brief Collection of issue classes raised by GetFiles class
 */

/**
 *  @ingroup coca_server
 *
 *  @namespace daq::coca::ProcessIssues
 *  @brief Collection of issue classes raised by Process class
 */


/**
 *  @class ConfigIssues::Exception
 *  @brief Generic exception class, base for exception classes used by Config class.
 */
ERS_DECLARE_ISSUE (ConfigIssues, Exception, , ) VSJFSC;

/**
 *  @class ConfigIssues::TDAQPartitionNotDefined
 *  @brief Exception generated when TDAQ partition name is not set.
 */
ERS_DECLARE_ISSUE_BASE (ConfigIssues, TDAQPartitionNotDefined, Exception,
        "environment variable TDAQ_PARTITION must be defined", , ) VSJFSC;

/**
 *  @class ConfigIssues::ServerNotFound
 *  @brief Exception generated when CoCa server is not defined in OKS config.
 */
ERS_DECLARE_ISSUE_BASE (ConfigIssues, ServerNotFound, Exception,
        "CoCaServer " << name << " not found in configuration db",
        , ((std::string) name)) VSJFSC;

/**
 *  @class ConfigIssues::GroupNotFound
 *  @brief Exception generated when user group is not found.
 */
ERS_DECLARE_ISSUE_BASE (ConfigIssues, GroupNotFound, Exception,
        "Unix group " << name << " not found", , ((std::string) name)) VSJFSC;

/**
 *  @class ConfigIssues::DatasetNotAllowed
 *  @brief Exception generated when dataset name was not defined.
 */
ERS_DECLARE_ISSUE_BASE (ConfigIssues, DatasetNotAllowed, Exception,
        "dataset " << name << " not allowed on this CoCa server",
        , ((std::string) name)) VSJFSC;

/**
 *  @class ConfigIssues::PartitionNotFound
 *  @brief Exception generated partition name is not found in OKS.
 */
ERS_DECLARE_ISSUE_BASE (ConfigIssues, PartitionNotFound, Exception,
        "partition " << partition << " not found in configuration db",
        , ((std::string) partition)) VSJFSC;

/**
 *  @class ConfigIssues::ConfigError
 *  @brief Exception produced for low-level configuration issues.
 */
ERS_DECLARE_ISSUE_BASE (ConfigIssues, ConfigError, Exception,
        "exception was generated by configuration db", ,) VSJFSC;

/**
 *  @class FreeSpaceIssues::FileSystemError
 *  @brief Issue generated for the filesystem-related problems.
 */
ERS_DECLARE_ISSUE (FreeSpaceIssues, FileSystemError, message, ((std::string) message)) VSJFSC;


/**
 *  @class GetFilesIssues::Message
 *  @brief Generic message class.
 */
ERS_DECLARE_ISSUE (GetFilesIssues, Message, , ) VSJFSC;

/**
 *  @class GetFilesIssues::SpaceExhausted
 *  @brief Message generated when archiver had to close all archives to free space.
 */
ERS_DECLARE_ISSUE_BASE(GetFilesIssues, SpaceExhausted, Message,
            "space exhausted; all archives were closed", , ) VSJFSC;

/**
 *  @class GetFilesIssues::CopyFailed
 *  @brief Message generated when archiver fails to copy file from client machine.
 */
ERS_DECLARE_ISSUE_BASE(GetFilesIssues, CopyFailed, Message,
            "Failed to copy file from client: " << path, , ((std::string) path)) VSJFSC;

/**
 *  @class GetFilesIssues::CopyRetryExceeded
 *  @brief Message generated when archiver gives up copying after several re-tries.
 */
ERS_DECLARE_ISSUE_BASE(GetFilesIssues, CopyRetryExceeded, Message,
            "Copy retry count exceeded limit (" << maxRetries << ") for " << path
            << ", will ignore this file.", , ((std::string) path)((int) maxRetries)) VSJFSC;


/**
 *  @class ProcessIssues::Exception
 *  @brief Base class for all issues generated by Process class.
 */
ERS_DECLARE_ISSUE (ProcessIssues, Exception, , ) VSJFSC;

/**
 *  @class ProcessIssues::ForkError
 *  @brief Issue generated when call to fork() fails.
 */
ERS_DECLARE_ISSUE_BASE (ProcessIssues, ForkError, Exception,
 "cannot create new process", , ) VSJFSC;

/**
 *  @class ProcessIssues::ChildError
 *  @brief Base class for various issues related to child process.
 */
ERS_DECLARE_ISSUE_BASE (ProcessIssues, ChildError, Exception, , , ) VSJFSC;

/**
 *  @class ProcessIssues::BadExitCode
 *  @brief Issue generated when child process returns non-zero code.
 */
ERS_DECLARE_ISSUE_BASE (ProcessIssues, BadExitCode, ChildError,
 "process " << process << "[" << pid << "]"
 " returned a suspicious exit code: " << retcode,
 , ((std::string) process) ((pid_t) pid) ((int) retcode)) VSJFSC;

/**
 *  @class ProcessIssues::Killed
 *  @brief Issue generated when child process is terminated with a signal.
 */
ERS_DECLARE_ISSUE_BASE (ProcessIssues, Killed, ChildError,
 "process " << process << "[" << pid << "] killed with signal "
 << signum << " (" << ::strsignal (signum) << ")",
 , ((std::string) process) ((pid_t) pid) ((int) signum)) VSJFSC;

/**
 *  @class ProcessIssues::KilledTimeout
 *  @brief Issue generated when child process does not finish in time
 *         and has to be killed.
 */
ERS_DECLARE_ISSUE_BASE (ProcessIssues, KilledTimeout, ChildError,
 "process " << process << "[" << pid << "]"
 " killed due to expired timeout",
 , ((std::string) process) ((pid_t) pid)) VSJFSC;

/**
 *  @class ProcessIssues::SSHFailure
 *  @brief Issue generated for SSH-related problems.
 */
ERS_DECLARE_ISSUE_BASE (ProcessIssues, SSHFailure, ChildError,
"remote process " << process << "[" << pid
<< "] failed with an SSH-related error",
, ((std::string) process) ((pid_t) pid)) VSJFSC;

/**
 *  @class ProcessIssues::BadExitCodeRemote
 *  @brief Issue generated when remote process returns non-zero status.
 */
ERS_DECLARE_ISSUE_BASE (ProcessIssues, BadExitCodeRemote, BadExitCode,
 "(running on " << machine << ") ",
 ((std::string) process) ((pid_t) pid) ((int) retcode),
 ((std::string) machine)) VSJFSC;


} // namespace coca
} // namespace daq

#endif // COCA_SERVERISSUES_H
