//--------------------------------------------------------------------------
// File and Version Information:
// 	$Id$
//
// Description:
//	Class FileSystem...
//
// Author List:
//      Andy Salnikov
//
//------------------------------------------------------------------------

//-----------------------
// This Class's Header --
//-----------------------
#include "coca/common/FileSystem.h"

//-----------------
// C/C++ Headers --
//-----------------
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <boost/filesystem.hpp>

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "coca/common/System.h"

//-----------------------------------------------------------------------
// Local Macros, Typedefs, Structures, Unions and Forward Declarations --
//-----------------------------------------------------------------------

namespace fs = boost::filesystem;

namespace {

    ERS_DECLARE_ISSUE (warnings, ChgrpFailed, 
            "failed to change group on directory " << name << ": " << strerror(errno), 
            ((std::string) name)) 

    ERS_DECLARE_ISSUE (warnings, ChmodFailed, 
            "failed to change permissions on directory " << name << ": " << strerror(errno), 
            ((std::string) name)) 


    // remove extra slashes and dots from path name 
    std::string normPath (const fs::path& path)
    {
        std::string result;
        for (fs::path::const_iterator it = path.begin(); it != path.end(); ++it) {
            if (*it != ".") {
                if(result.empty()) {
#if BOOST_FILESYSTEM_VERSION > 2
                    result += it->string();
#else
                    result += *it;
#endif
                } else {
                    if(result != "/") result += '/';
#if BOOST_FILESYSTEM_VERSION > 2
                    result += it->string();
#else
                    result += *it;
#endif
                }
            }
        }
        return result;
    }
            
}

//		----------------------------------------
// 		-- Public Function Member Definitions --
//		----------------------------------------

namespace daq {
namespace coca {

/**
 * This function creates directory and all its parent directories,
 * and in addition sets UNIX group for new directories and group bit
 * if necessary.
 */
bool 
FileSystem::create_dirs(const std::string& path)
{
    const boost::filesystem::path ph = path;

    // it may exist already, throw it it is a regular file
    if (ph.empty() || exists(ph)) {
        if ( not ph.empty() and not is_directory(ph) ) {
            throw MkdirFileExists(ERS_HERE, ph.string());
        }
        return false;
    }

    // First create branch, by calling ourself recursively
    create_dirs(ph.parent_path().string());
    
    // Now that parent's path exists, create the directory
    try {
        boost::filesystem::create_directory(ph);
    } catch (const std::exception& ex) {
        throw MkdirFailed(ERS_HERE, ph.string(), ex);
    }
    
    // group
    gid_t group = System::instance().newDirGroupId();
    if ( group != (gid_t)-1 ) {
        if ( chown (path.c_str(), (uid_t)-1, group) < 0 ) {
            // it's not a total failure
            ers::warning(::warnings::ChgrpFailed(ERS_HERE, path));
        }
        ERS_DEBUG (0, "directory " << path << ": group set to " << group);
    }
    
    // permissions
    if (System::instance().newDirHasSetGid()) {
        
        struct stat buf;
        if (stat (path.c_str(), &buf) != 0) {
            // complain but do no stop
            ers::warning(::warnings::ChmodFailed(ERS_HERE, path));
        } else {
            mode_t mode = buf.st_mode & 07777;
            mode |= S_ISGID;
            if (chmod(path.c_str(), mode) != 0) {
                // complain but do no stop
                ers::warning(::warnings::ChmodFailed(ERS_HERE, path));
            }
        }

        ERS_DEBUG (0, "setgid bit enabled for directory " << path);
    }
    
    return true;
}

// join two components and remove extra slashes from file name
std::string 
FileSystem::join(const std::string& path1, const std::string& path2)
{
    boost::filesystem::path opath(path1);
    opath /= path2;
    return ::normPath(opath);
}


// Splits pathname into directory and relative path. 
std::pair<std::string, std::string> 
FileSystem::split(const std::string& apath, const std::string& baseDir, const std::string& cwd)
{
    fs::path path = ::normPath (fs::path(apath));
    fs::path rel_path;
    fs::path dirname;
  
    // If file name is relative prepend CWD.
    if (not path.is_complete()) {
        fs::path wd = cwd.empty() ? fs::current_path() : fs::path(cwd);
#if BOOST_FILESYSTEM_VERSION > 2
        path = fs::absolute(path, ::normPath(wd));
#else
        path = fs::complete(path, ::normPath(wd));
#endif
    }
  
    if (baseDir.empty()) {
  
        // when basedir is empty we use basename as rel_path
#if BOOST_FILESYSTEM_VERSION > 2
        return std::make_pair(path.parent_path().string(), path.filename().string());
#else
        return std::make_pair(path.parent_path().string(), path.filename());
#endif

    } else {
  
        // If basedir is not empty then file name must start with the same
        // directory as basedir, strip basedir from file name.
        
        fs::path base_dir = ::normPath(baseDir);
        
        // make sure that number of components in path name is higher than in base name
        int ncBaseDir = std::distance(base_dir.begin(), base_dir.end());
        if (ncBaseDir >= std::distance(path.begin(), path.end())) {
            throw BasedirMismatch(ERS_HERE, base_dir.string(), path.string());
        }

        // check that path starts with baseDir
        if (not std::equal(base_dir.begin(), base_dir.end(), path.begin())) {
            throw BasedirMismatch(ERS_HERE, base_dir.string(), path.string());
        }

        // add remaining components to relative path
        fs::path rel_path;
        fs::path::iterator it = path.begin();
        std::advance(it, ncBaseDir);
        for (; it != path.end(); ++ it) {
            rel_path /= *it;
        }
        
        return std::make_pair(base_dir.string(), rel_path.string());

   }
    
}

} // namespace coca
} // namespace daq
