//----------------------------------------------------------------------
// File and Version Information:
//      $Id$
//
// Description:
//      Class FreeSpace...
//
// Author List:
//      Andy Salnikov
//
//----------------------------------------------------------------------

//-----------------------
// This Class's Header --
//-----------------------
#include "coca/server/FreeSpace.h"

//-----------------
// C/C++ Headers --
//-----------------
#include <ctime>
#include <boost/filesystem.hpp>
#include <boost/thread/mutex.hpp>

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "coca/common/FileSystem.h"
#include "coca/common/SizeMB.h"
#include "coca/server/CachedFile.h"
#include "coca/server/DBServer.h"

//-----------------------------------------------------------------------
// Local Macros, Typedefs, Structures, Unions and Forward Declarations --
//-----------------------------------------------------------------------

using namespace daq::coca;
namespace fs = boost::filesystem;

namespace {

// Return total and free space in MB, throws on error
fs::space_info
getSpace(const std::string& dir)
{
    try {
        return fs::space(dir);
    } catch (const fs::filesystem_error& ex) {
        throw FreeSpaceIssues::FileSystemError(ERS_HERE, "failed check free space", ex);
    }

}


void
deleteOneFile(DBServer& db, const CachedFile& file)
{
    try {

        // Delete from disk
        const fs::path& serverPath = file.serverPath();
        if (fs::is_regular_file(serverPath)) {
            // check that disk size is consistent with database
            SizeMB::checkSizeMB(serverPath.string(), file.sizeMB());
            fs::remove(serverPath);
            ERS_LOG("file removed from cache: " << serverPath);
        } else {
            ERS_LOG("file not found any more in cache: " << serverPath);
        }

        // Delete from database
        db.removeFromCache(file.dataset(), file.relPath().string());
        ERS_LOG("file removed from database: " << file.relPath());

    } catch (const ers::Issue &ex) {
        ers::warning(ex);
    } catch (const fs::filesystem_error& ex) {
        ers::warning(FreeSpaceIssues::FileSystemError(ERS_HERE, "Boost filesystem library error", ex));
    }
}

// As we operate on global state (filesystem)
// we also need global mutex here to protect it.
boost::mutex g_mutex;

}

namespace daq {
namespace coca {

FreeSpace::FreeSpace(DBServer& db, unsigned freePercentGoal)
    : m_db(db)
    , m_freePercentGoal(freePercentGoal)
{
}

bool FreeSpace::makeAvailable(unsigned goalMB, const std::string& dir)
{
    // Directory must exist
    try {
        FileSystem::create_dirs(dir);
    } catch (const FileSystem::Exception& ex) {
        throw FreeSpaceIssues::FileSystemError(ERS_HERE, "failed to create directory", ex);
    }

    time_t now = time(0);

    // get current space and set goals
    auto space = getSpace(dir);

    uint64_t percentGoal = space.capacity * m_freePercentGoal / 100;
    uint64_t explicitGoal = uint64_t(goalMB) * 1024 * 1024;
    uint64_t largeGoal = std::max(percentGoal, explicitGoal);
    ERS_LOG("Disk space total=" << space.capacity/(1024*1024) << "MB, available="
                    << space.available/(1024*1024) << "MB, free goal="
                    << largeGoal/(1024*1024) << "MB, free minimum=" << goalMB << "MB");

    // First pass is not to delete anything but check if the files are on disk, if not there then
    // unregister it from database (in case someone cleaned cache manually)
    for (const auto& file: m_db.filesToDelete()) {
        try {
            if (fs::status(file.serverPath()).type() == fs::file_not_found) {
                // Delete from database
                ERS_LOG("file not found in cache, will remove from database: " << file.serverPath());
                m_db.removeFromCache(file.dataset(), file.relPath().string());
            }
        } catch (const ers::Issue &ex) {
            ers::warning(ex);
        } catch (const fs::filesystem_error& ex) {
            ers::warning(FreeSpaceIssues::FileSystemError(ERS_HERE, "Boost filesystem library error", ex));
        }
    }

    // First check, maybe we have enough already
    if (space.available >= largeGoal)
        return true;

    // protect the rest with global lock so that multiple threads
    // do not try to delete the same file
    boost::mutex::scoped_lock lock(g_mutex);

    // get the files that can be deleted, sort by cost in _descending_ order
    std::vector<CachedFile> files = m_db.filesToDelete();
    std::sort(files.begin(), files.end(), [now](const CachedFile& lhs, const CachedFile& rhs) {
        return rhs.cacheCost(now) < lhs.cacheCost(now);
    });

    // First try to delete only expired files to try to reach largeGoal
    for (const auto& file: files) {
        if (file.expired(now)) {

            ::deleteOneFile(m_db, file);

            if (::getSpace(dir).available >= largeGoal)
                return true;
        }
    }

    ERS_LOG("Cannot free " << largeGoal/(1024*1024) << "MB by deleting expired files only, "
            "will remove non-expired files");

    if (::getSpace(dir).available >= explicitGoal)
        return true;

    // still not enough space, try to remove non-expired files to reach explicit goal
    files = m_db.filesToDelete();
    std::sort(files.begin(), files.end(), [now](const CachedFile& lhs, const CachedFile& rhs) {
        return rhs.cacheCost(now) < lhs.cacheCost(now);
    });

    for (const auto& file: files) {

        ::deleteOneFile(m_db, file);

        if (::getSpace(dir).available >= explicitGoal)
            return true;
    }

    ERS_LOG("Cannot free explicitly requested " << explicitGoal/(1024*1024) << "MB of disk space");
    return false;
}

} // namespace coca
} // namespace daq
