"""Module for AppBase class and related methods.
"""

#--------------------------------
#  Imports of standard modules --
#--------------------------------
import logging

#-----------------------------
# Imports for other modules --
#-----------------------------
import urwid
from .ui import UIPopUpLauncher, UIPopUpMessageBox

#----------------------------------
# Local non-exported definitions --
#----------------------------------

# default palette for all apps
_palette = [
    ('body', 'black', 'light gray'),
    ('selected', 'white', 'dark green'),

    ('title', 'light cyan', 'black', 'bold,standout'),

    ('hint', 'white', 'black', 'bold,standout'),
    ('hint-high', 'yellow', 'black', 'bold,standout'),

    ('status', 'light green', 'black', 'bold,standout'),
    ('status-high', 'yellow', 'black', 'bold,standout'),

    ('list-colhead', 'white', 'black', 'bold'),
    ('list-item', 'black', 'light gray'),
    ('list-selected', 'white', 'dark gray', 'standout'),

    ('menu', 'white', 'black'),
    ('menu-focus', 'black', 'white', 'standout'),

    ('messagebox', 'black', 'white'),
    ('messagebox-focus', 'white', 'dark gray', 'standout'),

    ('messagebox-error', 'white', 'dark red'),
    ('messagebox-error-focus', 'dark red', 'white', 'standout'),

    ('editbox', 'black', 'white'),
    ('editbox-focus', 'white', 'dark gray', 'standout'),
    ('editbox-field', 'white', 'black'),

    ('progressbar', 'white', 'dark blue'),
    ('progressbar-incomplete', 'yellow', 'black'),
    ('progressbar-complete', 'black', 'yellow'),
]

_log = logging.getLogger(__name__)


class FailedExit(UIPopUpMessageBox):

    def __init__(self, message, callback):
        buttons = [('Exit', 'exit')]
        self.__super.__init__('Application failure', message, buttons, callback,
                              escape_value='exit', attr='messagebox-error')

#------------------------
# Exported definitions --
#------------------------


class AppBase(urwid.WidgetWrap):
    """Class representing top-level widget for the whole application.

    It implements general application behavior such as input handling,
    status line and hint display, panel stack, and common palette.
    Specific application sub-classes will have to re-implement few
    methods to add application panels.
    """

    _selectable = True

    palette = _palette

    #----------------
    #  Constructor --
    #----------------
    def __init__(self):
        """
        @param conn_str:  CORAL connection string for database
        @param readonly:  boolean, if True then use read-only mode for database
        """
        self._hint = urwid.Text("", wrap='clip')
        self._status = urwid.Text("", wrap='clip', align='right')
        footer = urwid.Columns([self._hint, urwid.AttrMap(self._status, 'status')])
        root = urwid.LineBox(urwid.SolidFill(), '')
        title = self.title()
        header = urwid.AttrMap(urwid.Text(title, wrap='clip'), 'title') if title else None
        self._frame = urwid.Frame(urwid.AttrMap(root, 'body'),
                                  header=header,
                                  footer=urwid.AttrMap(footer, 'hint'))
        self._pop_up_launcher = UIPopUpLauncher(self._frame)
        self.__super.__init__(self._pop_up_launcher)

        self._widgetStack = []

        # make a main loop and arrange it to call app.start() when alive
        self._main_loop = urwid.MainLoop(self, self.palette, pop_ups=True)
        self._main_loop.set_alarm_in(0, self.start)

    def title(self):
        """Return string displayed as application title at the top
        of the window.

        If None or empty string is returned then nothing is displayed.
        """
        return None

    def make_pop_up(self, popup):
        """Show pop-up widget.

        Panels should call this method if they need to display popup widgets.

        Parameters
        ----------
        pop_up : `UIPopUp`
        """
        self._pop_up_launcher.make_pop_up(popup)

    def run(self):
        """Starts the whole shebang.
        """
        self._main_loop.run()

    def start(self, loop, user_data):
        """Will be called once event loop starts (e.g. via timer or signal).

        This method has to be overriden by sub-class to instantiate whatever
        first panel it needs.
        """
        raise NotImplementedError()

    def makePanel(self, PanelClass, *args, **kwargs):
        """Create panel widget from given panel class and additional arguments.

        thsi sis the factory method that should be used by sub-classes to
        instantiate panel widgets.
        """
        widget = PanelClass(self, *args, **kwargs)
        self.setCentralWidget(widget)
        return widget

    def setCentralWidget(self, widget):
        """Display given widget in the central area with line box around it
        """
        root = urwid.LineBox(widget, widget.title())
        self._frame.contents['body'] = (urwid.AttrMap(root, 'body'), None)
        self._widgetStack.append(widget)
        self.setHints(widget.hints())
        self._status.set_text(widget.status())

    def goBack(self):
        """Return to previous panel.
        """
        if len(self._widgetStack) > 1:
            del self._widgetStack[-1]
            widget = self._widgetStack[-1]
            root = urwid.LineBox(widget, widget.title())
            self._frame.contents['body'] = (urwid.AttrMap(root, 'body'), None)
            self.setHints(widget.hints())
            self._status.set_text(widget.status())

    def setHints(self, widgetHints):
        """Sets hint string (e.g. "^Q: Exit")
        """
        message = []
        hints = self.hints() + widgetHints
        for key, action in hints:
            message += [' ', ('hint-high', key), ':' + action]

        self._hint.set_text(message)

    def hints(self):
        hints = [('^X', "Exit")]
        if len(self._widgetStack) > 1:
            hints += [('Esc/BS', "Back")]
        return hints

    def keypress(self, size, key):
        """Handle keyboard input.
        """
        _log.debug("received key: %s", key)

        # update window size for pup-ups
        self._pop_up_launcher.set_parent_size(size)

        child = self._frame.contents['body'][0]
        if child.selectable():
            _log.debug("forward key to child: %s", key)
            size = (size[0], size[1] - 2)
            try:
                key = child.keypress(size, key)
            except Exception as exc:
                msg = "Exception: " + str(exc)
                self.make_pop_up(FailedExit(msg, self._exitOnFail))
                return None
            # update status line
            self._status.set_text(self._widgetStack[-1].status())
        if key == 'ctrl x':
            raise urwid.ExitMainLoop()
        if key in ('esc', 'backspace'):
            # handle "go back" key
            self.goBack()
            key = None
        return key

    def _exitOnFail(self, widget, button):
        raise urwid.ExitMainLoop()
