"""Module containing CocaDb class and related methods.
"""

#--------------------------------
#  Imports of standard modules --
#--------------------------------
from __future__ import division
from collections import namedtuple
from contextlib import contextmanager
from datetime import datetime
import logging

#-----------------------------
# Imports for other modules --
#-----------------------------
from coldpie import coral

#----------------------------------
# Local non-exported definitions --
#----------------------------------
_log = logging.getLogger(__name__)


def _ts2dt(ts):
    " Convert Coral timestamp into datetime"
    return datetime(ts.year(), ts.month(), ts.day(),
                    ts.hour(), ts.minute(), ts.second(),
                    ts.nanosecond() // 1000)


def _row2nt(row, tupletype):
    """
    Covert Coral AttributeList (result row) into an named tuple.
    """

    kv = {}
    for attr in tupletype._fields:
        val = row[attr].data()
        if isinstance(val, coral.TimeStamp):
            val = _ts2dt(val)
        kv[attr] = val
    return tupletype(**kv)


@contextmanager
def read_transaction(session):
    readonly = True
    trans = session.transaction()
    trans.start(readonly)
    yield trans
    trans.commit()


@contextmanager
def update_transaction(session):
    readonly = False
    trans = session.transaction()
    trans.start(readonly)
    yield trans
    trans.commit()

#------------------------
# Exported definitions --
#------------------------


Server = namedtuple("Server", """server_ipc n_archives totsize_mb""")

Dataset = namedtuple("Dataset", """name n_archives totsize_mb""")

Archive = namedtuple("Archive", """archive name dataset dataset_name booked_at close_at
                                delete_at server_ipc server_ipc_name server_host dir_server
                                totsize_mb totsize_mb_cached closed_at on_cdr on_cdr_at
                                cdr_path stored stored_at stor_path removed removed_at
                                open_if_0""")

File = namedtuple('File', """dataset dataset_name rel_path size_mb client_host dir_client
                          prio try2keep_til registered_at archive archive_name
                          on_client in_cache in_cache_at removed removed_at""")


class CocaDb(object):
    """Interface for CoCa database.
    """

    #----------------
    #  Constructor --
    #----------------
    def __init__(self, conn_str, readonly):
        """
        @param conn_str:  CORAL connection string for coca database
        @param readonly:  boolean, if True then use read-only mode for database
        """

        conn_svc = coral.ConnectionService()
        mode = coral.access_ReadOnly if readonly else coral.access_Update
        logging.debug("start new database session: %s", conn_str)
        self.session = conn_svc.connect(conn_str, mode)
        self.schema = self.session.nominalSchema()

        self.s2iName = 'COCA_STR2ID'
        self.arName = 'COCA1_ARCHIVES'
        self.caName = 'COCA1_CACHE'

    def servers(self):
        """
        Returns the list of Server instances ordered by IPC server_ipc name.
        """

        with read_transaction(self.session):

            query = self.schema.newQuery()

            query.addToTableList(self.arName, "AR")
            query.addToTableList(self.s2iName, "S2I_IPC")
            query.addToOutputList('S2I_IPC.STR', 'server_ipc')
            query.addToOutputList('COUNT(*)', 'n_archives')
            query.addToOutputList('SUM(AR.TOTSIZE_MB)', 'totsize_mb')

            cdata = coral.AttributeList()
            cond = """AR.SERVER_IPC = S2I_IPC.ID"""
            query.setCondition(cond, cdata)

            query.addToOrderList('S2I_IPC.STR')
            query.groupBy('S2I_IPC.STR')

            res = [_row2nt(row, Server) for row in query.execute()]

        return res

    def datasets(self, server):
        """
        Returns list of Dataset instances ordered by name.

        @param server: server name or None, if None returns datasets for all servers.
        """

        with read_transaction(self.session):

            query = self.schema.newQuery()
            query.addToTableList(self.arName, "AR")
            query.addToTableList(self.s2iName, "S2I_DS")

            query.addToOutputList('S2I_DS.STR', 'name')
            query.addToOutputList('COUNT(*)', 'n_archives')
            query.addToOutputList('SUM(AR.TOTSIZE_MB)', 'totsize_mb')

            cdata = coral.AttributeList()
            cond = """AR.DATASET = S2I_DS.ID"""
            if server:
                query.addToTableList(self.s2iName, "S2I_IPC")
                cdata.extend('IPCNAME', 'string')
                cdata[0].setData(server)
                cond += """ AND AR.SERVER_IPC = S2I_IPC.ID AND S2I_IPC.STR = :IPCNAME"""
            query.setCondition(cond, cdata)

            query.addToOrderList('S2I_DS.STR')
            query.groupBy('S2I_DS.STR')

            res = [_row2nt(row, Dataset) for row in query.execute()]

        return res

    def archives_count(self, server, dataset, open=None, stored=None, removed=None):
        """
        @param server: server name or None
        @param dataset: dataset name or None
        @param open: True/False or None
        @param stored: True/False or None
        @param removed: True/False or None
        @return: tuple of archive count and archive size
        """
        _log.debug('archives_count: add server: %r %r %r', server, dataset, open)

        with read_transaction(self.session):

            query = self.schema.newQuery()
            query.addToTableList(self.arName, "AR")
            query.addToOutputList('COUNT(*)', 'n_archives')
            query.addToOutputList('SUM(AR.TOTSIZE_MB)', 'totsize_mb')

            cdata = coral.AttributeList()
            cond = []
            if server:
                query.addToTableList(self.s2iName, "S2I_IPC")
                idx = len(cdata)
                cdata.extend('IPCNAME', 'string')
                cdata[idx].setData(server)
                cond += ["AR.SERVER_IPC = S2I_IPC.ID", "S2I_IPC.STR = :IPCNAME"]
            if dataset:
                query.addToTableList(self.s2iName, "S2I_DS")
                idx = len(cdata)
                cdata.extend('DATASET', 'string')
                cdata[idx].setData(dataset)
                cond += ["AR.DATASET = S2I_DS.ID", "S2I_DS.STR = :DATASET"]
            if open is True:
                cond += ["AR.OPEN_IF_0 = 0"]
            if open is False:
                cond += ["AR.OPEN_IF_0 <> 0"]
            if stored is True:
                cond += ["AR.STORED <> 0"]
            if stored is False:
                cond += ["AR.STORED = 0"]
            if removed is True:
                cond += ["AR.REMOVED <> 0"]
            if removed is False:
                cond += ["AR.REMOVED = 0"]
            if cond:
                cond = " AND ".join(cond)
                _log.debug('archives_count: cond: %r', cond)
                query.setCondition(cond, cdata)

            counter = size = None
            for row in query.execute():

                counter = int(row[0].data())
                size = int(row[1].data() or 0)

        return counter, size

    def archives(self, server, dataset, open=None, stored=None, removed=None, maxRows=None, offset=None):
        """
        Returns list of Archive instances ordered by booking time.

        @param server: server name or None
        @param dataset: dataset name or None
        @param open: True/False or None
        @param stored: True/False or None
        @param removed: True/False or None
        @param maxRows: max. number of rows to return (used for paging)
        @param offset: starting position in number of rows (used for paging)
        """

        with read_transaction(self.session):

            query = self.schema.newQuery()

            query.addToTableList(self.arName, "AR")
            query.addToTableList(self.s2iName, "S2I_IPC")
            query.addToTableList(self.s2iName, "S2I_DS")
            query.addToTableList(self.s2iName, "S2I_DIR")
            query.addToTableList(self.s2iName, "S2I_HOST")

            ar_attributes = ('dataset', 'archive', 'name', 'booked_at', 'close_at', 'delete_at',
                             'server_ipc', 'totsize_mb', 'totsize_mb_cached', 'open_if_0',
                             'closed_at', 'on_cdr', 'on_cdr_at', 'cdr_path', 'stored',
                             'stored_at', 'stor_path', 'removed', 'removed_at')
            for attr in ar_attributes:
                query.addToOutputList('AR.' + attr.upper(), attr)
            query.addToOutputList('S2I_DS.STR', 'dataset_name')
            query.addToOutputList('S2I_IPC.STR', 'server_ipc_name')
            query.addToOutputList('S2I_HOST.STR', 'server_host')
            query.addToOutputList('S2I_DIR.STR', 'dir_server')

            cdata = coral.AttributeList()
            cond = ["AR.SERVER_IPC = S2I_IPC.ID",
                    "AR.DATASET = S2I_DS.ID",
                    "AR.DIR_SERVER = S2I_DIR.ID",
                    "AR.SERVER_HOST = S2I_HOST.ID"]
            if server:
                idx = len(cdata)
                cdata.extend('IPCNAME', 'string')
                cdata[idx].setData(server)
                cond += ["S2I_IPC.STR = :IPCNAME"]
            if dataset:
                idx = len(cdata)
                cdata.extend('DATASET', 'string')
                cdata[idx].setData(dataset)
                cond += ["S2I_DS.STR = :DATASET"]
            if open is True:
                cond += ["AR.OPEN_IF_0 = 0"]
            if open is False:
                cond += ["AR.OPEN_IF_0 <> 0"]
            if stored is True:
                cond += ["AR.STORED <> 0"]
            if stored is False:
                cond += ["AR.STORED = 0"]
            if removed is True:
                cond += ["AR.REMOVED <> 0"]
            if removed is False:
                cond += ["AR.REMOVED = 0"]

            if cond:
                cond = " AND ".join(cond)
            query.setCondition(cond, cdata)

            query.addToOrderList('AR.BOOKED_AT')

            if maxRows is not None:
                if offset is None:
                    offset = 0
                query.limitReturnedRows(maxRows, offset)

            query.setRowCacheSize(100000)
            query.setMemoryCacheSize(10)

            res = [_row2nt(row, Archive) for row in query.execute()]

        return res

    def files_in_archive(self, dataset, archive):
        """
        @param dataset:  dataset ID (number)
        @param archive:  archive ID (number)
        @return list of file names contained in archive
        """

        with read_transaction(self.session):

            query = self.schema.newQuery()
            query.addToTableList(self.caName, "CA")
            query.addToOutputList('CA.REL_PATH', 'rel_path')

            cdata = coral.AttributeList()
            cond = []

            idx = len(cdata)
            cdata.extend('DATASET', 'int')
            cdata[idx].setData(dataset)
            idx = len(cdata)
            cdata.extend('ARCHIVE', 'int')
            cdata[idx].setData(archive)
            cond += ["CA.DATASET = :DATASET", "CA.ARCHIVE = :ARCHIVE"]

            cond = " AND ".join(cond)
            query.setCondition(cond, cdata)

            query.addToOrderList('CA.REL_PATH')

            res = [row[0].data() for row in query.execute()]

        return res

    def files_count(self, server, dataset, on_client=None, in_cache=None):
        """
        @param server: server name or None
        @param dataset: dataset name or None
        @param on_client: True/False or None
        @param in_cache: True/False or None
        @return: tuple of file count and file size
        """

        with read_transaction(self.session):

            query = self.schema.newQuery()
            query.addToTableList(self.caName, "CA")
            query.addToOutputList('COUNT(*)', 'COUNTER')
            query.addToOutputList('SUM(CA.SIZE_MB)', 'SIZE')

            cdata = coral.AttributeList()
            cond = []
            if server or dataset:
                query.addToTableList(self.arName, "AR")
                cond += ["AR.DATASET = CA.DATASET", "AR.ARCHIVE = CA.ARCHIVE"]
            if server:
                query.addToTableList(self.s2iName, "S2I_IPC")
                idx = len(cdata)
                cdata.extend('IPCNAME', 'string')
                cdata[idx].setData(server)
                cond += ["AR.SERVER_IPC = S2I_IPC.ID", "S2I_IPC.STR = :IPCNAME"]
            if dataset:
                query.addToTableList(self.s2iName, "S2I_DS")
                idx = len(cdata)
                cdata.extend('DATASET', 'string')
                cdata[idx].setData(dataset)
                cond += ["AR.DATASET = S2I_DS.ID", "S2I_DS.STR = :DATASET"]
            if on_client is True:
                cond += ["CA.ON_CLIENT <> 0"]
            if on_client is False:
                cond += ["CA.ON_CLIENT = 0"]
            if in_cache is True:
                cond += ["CA.IN_CACHE <> 0"]
            if in_cache is False:
                cond += ["CA.IN_CACHE = 0"]
            if cond:
                cond = " AND ".join(cond)
                query.setCondition(cond, cdata)

            counter = size = None
            for row in query.execute():

                counter = int(row[0].data())
                size = int(row[1].data() or 0)

        return counter, size

    def files(self, server, dataset, on_client=None, in_cache=None, maxRows=None, offset=None):
        """
        @param server: server name or None
        @param dataset: dataset name or None
        @param on_client: True/False or None
        @param in_cache: True/False or None
        @param maxRows: max. number of rows to return (used for paging)
        @param offset: starting position in number of rows (used for paging)
        """

        with read_transaction(self.session):

            query = self.schema.newQuery()

            query.addToTableList(self.arName, "AR")
            query.addToTableList(self.caName, "CA")
            query.addToTableList(self.s2iName, "S2I_DS")
            query.addToTableList(self.s2iName, "S2I_DIR")
            query.addToTableList(self.s2iName, "S2I_HOST")

            ca_attributes = ('dataset', 'rel_path', 'size_mb', 'prio', 'try2keep_til',
                             'registered_at', 'archive', 'on_client', 'in_cache',
                             'in_cache_at', 'removed', 'removed_at')
            for attr in ca_attributes:
                query.addToOutputList('CA.' + attr.upper(), attr)
            query.addToOutputList('S2I_DS.STR', 'dataset_name')
            query.addToOutputList('S2I_HOST.STR', 'client_host')
            query.addToOutputList('S2I_DIR.STR', 'dir_client')
            query.addToOutputList('AR.NAME', 'archive_name')

            cdata = coral.AttributeList()
            cond = ["AR.DATASET = CA.DATASET",
                    "AR.ARCHIVE = CA.ARCHIVE",
                    "CA.DATASET = S2I_DS.ID",
                    "S2I_HOST.ID = CA.CLIENT_HOST",
                    "S2I_DIR.ID = CA.DIR_CLIENT"]

            if server:
                query.addToTableList(self.s2iName, "S2I_IPC")
                idx = len(cdata)
                cdata.extend('IPCNAME', 'string')
                cdata[idx].setData(server)
                cond += ["AR.SERVER_IPC = S2I_IPC.ID", "S2I_IPC.STR = :IPCNAME"]
            if dataset:
                idx = len(cdata)
                cdata.extend('DATASET', 'string')
                cdata[idx].setData(dataset)
                cond += ["S2I_DS.STR = :DATASET"]
            if on_client is True:
                cond += ["CA.ON_CLIENT <> 0"]
            if on_client is False:
                cond += ["CA.ON_CLIENT = 0"]
            if in_cache is True:
                cond += ["CA.IN_CACHE <> 0"]
            if in_cache is False:
                cond += ["CA.IN_CACHE = 0"]

            if cond:
                cond = " AND ".join(cond)
            query.setCondition(cond, cdata)

            query.addToOrderList('CA.REGISTERED_AT')

            if maxRows is not None:
                if offset is None:
                    offset = 0
                query.limitReturnedRows(maxRows, offset)

            query.setRowCacheSize(10000)
            query.setMemoryCacheSize(10)

            res = [_row2nt(row, File) for row in query.execute()]

        return res

    def archives_set_closed(self, archives):
        """
        @param archives: list of tuples (server_ipc, dataset, archive),
                         server_ipc and dataset are IDs (integers), archive
                         is archive number
        """

        setClause = "OPEN_IF_0 = ARCHIVE"
        extraCond = "OPEN_IF_0 = 0"
        return self._archives_update(archives, setClause, extraCond)

    def archives_set_stored(self, archives):
        """
        @param archives: list of tuples (server_ipc, dataset, archive),
                         server_ipc and dataset are IDs (integers), archive
                         is archive number
        """

        # This does not set STOR_PATH, we'll have separate method for that
        # Note: CURRENT_TIMESTAMP may not be very portable.
        setClause = "STORED = 1, STORED_AT = CURRENT_TIMESTAMP"
        extraCond = "STORED = 0"
        return self._archives_update(archives, setClause, extraCond)

    def archives_set_removed(self, archives):
        """
        @param archives: list of tuples (server_ipc, dataset, archive),
                         server_ipc and dataset are IDs (integers), archive
                         is archive number
        """

        # Note: CURRENT_TIMESTAMP may not be very portable.
        setClause = "REMOVED = 1, REMOVED_AT = CURRENT_TIMESTAMP"
        extraCond = "REMOVED = 0"
        return self._archives_update(archives, setClause, extraCond)

    def _archives_update(self, archives, setClause, extraCond=""):
        """
        @param archives: list of tuples (server_ipc, dataset, archive),
                         server_ipc and dataset are IDs (integers), archive
                         is archive number
        @param setClause: string, SET clause for CORAL update query
        @param extraCond: string, optional WHERE condition
        """

        with update_transaction(self.session):

            table = self.schema.tableHandle(self.arName)
            dataEditor = table.dataEditor()

            cond = "SERVER_IPC = :SERVER_IPC AND DATASET = :DATASET AND ARCHIVE = :ARCHIVE"
            if extraCond:
                cond += " AND " + extraCond

            n_updates = 0
            for ar in archives:

                cdata = coral.AttributeList()

                idx = len(cdata)
                cdata.extend('SERVER_IPC', 'int')
                cdata[idx].setData(ar[0])

                idx = len(cdata)
                cdata.extend('DATASET', 'int')
                cdata[idx].setData(ar[1])

                idx = len(cdata)
                cdata.extend('ARCHIVE', 'int')
                cdata[idx].setData(ar[2])

                n_updates += dataEditor.updateRows(setClause, cond, cdata)

        return n_updates

    def files_add_to_cache(self, files):
        """
        @param files: list of tuples (dataset, rel_path), dataset is ID
                    (integer), rel_path is a string
        """

        setClause = "IN_CACHE = 1, IN_CACHE_AT = CURRENT_TIMESTAMP"
        extraCond = "IN_CACHE = 0"
        return self._files_update(files, setClause, extraCond)

    def files_remove_from_client(self, files):
        """
        @param files: list of tuples (dataset, rel_path), dataset is ID
                    (integer), rel_path is a string
        """

        setClause = "ON_CLIENT = 0"
        extraCond = "ON_CLIENT = 1"
        return self._files_update(files, setClause, extraCond)

    def files_remove_from_cache(self, files):
        """
        @param files: list of tuples (dataset, rel_path), dataset is ID
                    (integer), rel_path is a string
        """

        setClause = "IN_CACHE = 0"
        extraCond = "IN_CACHE = 1"
        return self._files_update(files, setClause, extraCond)

    def _files_update(self, files, setClause, extraCond=""):
        """
        @param files: list of tuples (dataset, rel_path), dataset is ID
                    (integer), rel_path is a string
        @param setClause: string, SET clause for CORAL update query
        @param extraCond: string, optional WHERE condition
        """

        with update_transaction(self.session):

            table = self.schema.tableHandle(self.caName)
            dataEditor = table.dataEditor()

            cond = "DATASET = :DATASET AND REL_PATH = :REL_PATH"
            if extraCond:
                cond += " AND " + extraCond

            n_updates = 0
            for entry in files:

                cdata = coral.AttributeList()

                idx = len(cdata)
                cdata.extend('DATASET', 'int')
                cdata[idx].setData(entry[0])

                idx = len(cdata)
                cdata.extend('REL_PATH', 'string')
                cdata[idx].setData(entry[1])

                n_updates += dataEditor.updateRows(setClause, cond, cdata)

        return n_updates
