"""Module containing CocaServerList class and related methods.
"""

#--------------------------------
#  Imports of standard modules --
#--------------------------------
import logging

#-----------------------------
# Imports for other modules --
#-----------------------------
import urwid
from ..AppPanel import AppPanel
from ..ui import UIListBoxWithHeader, UIColumns, UISelectableText
from . import CocaDb

#----------------------------------
# Local non-exported definitions --
#----------------------------------

_log = logging.getLogger(__name__)

#------------------------
# Exported definitions --
#------------------------


class CocaServerList(UIListBoxWithHeader, AppPanel):
    """Widget class containing list of CoCa servers.

    Parameters
    ----------
    app : `CocaAdmin`
    db : `CocaDb`
    """

    _selectable = True

    signals = ['selected']

    #----------------
    #  Constructor --
    #----------------
    def __init__(self, app, db):
        self._app = app

        # read data from database
        servers = db.servers()
        self._numServers = len(servers)

        # add special entry at the top
        total_archives = sum(s.n_archives for s in servers)
        total_size = sum(s.totsize_mb for s in servers)
        servers = [CocaDb.Server(server_ipc=None, n_archives=total_archives, totsize_mb=total_size)] + \
            sorted(servers)

        col_width = max(len(s.server_ipc or "<EVERYTHING>") for s in servers)
        col_width = [col_width, 10, 10]
        header = UIColumns(['Server',
                            urwid.Text('N Archives', align='right'),
                            urwid.Text('Size, MB', align='right')],
                           col_width, 2)

        items = []
        for server in servers:
            name = server.server_ipc or "<EVERYTHING>"
            item = UIColumns([UISelectableText(name),
                              urwid.Text(str(int(server.n_archives)), align='right'),
                              urwid.Text(str(int(server.totsize_mb)), align='right')],
                             col_width, 2, 0)
            urwid.connect_signal(item, 'activated', self._itemActivated, user_args=[server.server_ipc])
            items.append(item)

        UIListBoxWithHeader.__init__(self, items, header=header)

    def title(self):
        return 'List of CoCa servers'

    def hints(self):
        return [('Enter', "Select")]

    def status(self):
        return "Found {} servers".format(self._numServers)

    def _itemActivated(self, server_ipc, item):
        _log.debug("emitting signal 'selected': %r", server_ipc)
        self._emit('selected', server_ipc)
