"""Module containing CoolFolderList class and related methods.
"""

#--------------------------------
#  Imports of standard modules --
#--------------------------------
import logging

#-----------------------------
# Imports for other modules --
#-----------------------------
import urwid
from ..AppPanel import AppPanel
from ..ui import UIListBoxWithHeader, UIColumns, UISelectableText
from . import CoolDb

#----------------------------------
# Local non-exported definitions --
#----------------------------------

_log = logging.getLogger(__name__)

#------------------------
# Exported definitions --
#------------------------


class CoolFolderList(urwid.WidgetWrap, AppPanel):
    """Widget class containing list COOL folders and foldersets.

    Parameters
    ----------
    app: `CoolTui`
        application class
    db: `CoolDb`
        instance of CoolDb class
    folderSet : `CoolDb.FolderSet`, optional
    """

    _selectable = True

    signals = ['folderSetSelected', 'folderSelected', 'showSchema']

    #----------------
    #  Constructor --
    #----------------
    def __init__(self, app, db, folderSet=None):
        self._app = app
        self._db = db
        self._folderSet = folderSet
        self._col_width = [1, -1, 2, -3]

        foldersets, folders = self._db.lsFolder(self._folderSet)

        self._total_count = len(foldersets) + len(folders)
        items = []
        for fs in foldersets:
            item = UIColumns([UISelectableText("+"),
                              UISelectableText(fs.name),
                              urwid.Text(""),
                              urwid.Text(fs.description),
                              ],
                             self._col_width, 2, 0)
            urwid.connect_signal(item, 'activated', self._foldersetActivated, user_args=[fs])
            item = urwid.AttrMap(item, "folderset-item", "folderset-selected")
            item.user_data = fs
            items += [item]
        for fs in folders:
            item = UIColumns([UISelectableText(""),
                              UISelectableText(fs.name),
                              urwid.Text("MV" if fs.multi_version else ""),
                              urwid.Text(fs.description),
                              ],
                             self._col_width, 2, 0)
            urwid.connect_signal(item, 'activated', self._folderActivated, user_args=[fs])
            item = urwid.AttrMap(item, "folder-item", "folder-selected")
            # remember folder
            item.user_data = fs
            items += [item]

        header = UIColumns([urwid.Text('+'),
                            urwid.Text('Name'),
                            urwid.Text('MV'),
                            urwid.Text('Description'),
                            ],
                           self._col_width, 2)

        self._walker = urwid.SimpleFocusListWalker(items)
        lb = UIListBoxWithHeader(self._walker, header=header)
        self.__super.__init__(lb)

    def title(self):
        return "Folders in database " + self._db.name

    def hints(self):
        return [('Enter', "Select"),
                ('^E', "Schema")]

    def keypress(self, size, key):

        # handle '^E'
        if key == 'ctrl e':
            self._show_schema()
            return

        # send it to base class
        key = self.__super.keypress(size, key)
        return key

    def _foldersetActivated(self, fs, item):
        _log.debug("emitting signal 'folderSetSelected': %r", fs)
        self._emit('folderSetSelected', self._db, fs)

    def _folderActivated(self, fs, item):
        _log.debug("emitting signal 'folderSelected': %r", fs)
        self._emit('folderSelected', self._db, fs)

    def _show_schema(self):
        # figure out which folder is selected
        item = self._walker.get_focus()[0]
        folder = item.user_data
        _log.debug("_show_schema folder: %r", folder)
        if isinstance(folder, CoolDb.Folder):
            _log.debug("emitting signal 'showSchema': %r", folder)
            self._emit('showSchema', self._db, folder)
