"""Module for CocaAdmin class and related methods.
"""

#--------------------------------
#  Imports of standard modules --
#--------------------------------
import logging

#-----------------------------
# Imports for other modules --
#-----------------------------
import urwid
from ..AppBase import AppBase
from .MdaDatasetList import MdaDatasetList
from .MdaDb import MdaDb
from .MdaFileList import MdaFileList
from .MdaHgroupList import MdaHgroupList
from .MdaHistoList import MdaHistoList
from .MdaLBSeqList import MdaLBSeqList
from .MdaMainMenu import MdaMainMenu
from .MdaPartitionList import MdaPartitionList
from .MdaRunList import MdaRunList

#----------------------------------
# Local non-exported definitions --
#----------------------------------
_log = logging.getLogger(__name__)

#------------------------
# Exported definitions --
#------------------------


class MdaAdmin(AppBase):
    """Class representing top-level widget for the whole application.

    Parameters
    ----------
    conn_str : `str`
        CORAL connection string for mda database
    readonly : `str`
        boolean, if True then use read-only mode for database
    """

    #----------------
    #  Constructor --
    #----------------
    def __init__(self, conn_str, readonly):
        AppBase.__init__(self)
        self.db = MdaDb(conn_str, readonly)

    def title(self):
        return ' = MDA database administration = '

    def start(self, loop, user_data):
        """Make initial panel
        """
        mdaMenu = self.makePanel(MdaMainMenu)
        urwid.connect_signal(mdaMenu, 'selected', self.mainMenuSelected)

    def mainMenuSelected(self, widget, mainMenuItem):
        _log.debug("selected item: %r", mainMenuItem)

        if mainMenuItem == 'partitions':
            widget = self.makePanel(MdaPartitionList, self.db)
            urwid.connect_signal(widget, 'selected', self.partitionSelected)
        elif mainMenuItem == 'datasets':
            widget = self.makePanel(MdaDatasetList, self.db)
            urwid.connect_signal(widget, 'selected', self.datasetSelected)
        elif mainMenuItem == 'files_not_archived':
            title = "Files not in CoCa"
            options = dict(archived=False)
            widget = self.makePanel(MdaFileList, self.db, None, title, options)
            urwid.connect_signal(widget, 'selected', self.fileSelected)
        elif mainMenuItem == 'lb_seq':
            widget = self.makePanel(MdaLBSeqList, self.db)

    def partitionSelected(self, widget, part_id, partition):
        _log.debug("selected item: %r", (part_id, partition))
        widget = self.makePanel(MdaRunList, self.db, part_id, partition)
        urwid.connect_signal(widget, 'selected', self.runLbSelected)

    def datasetSelected(self, widget, ds_id, ds_name):
        _log.debug("selected item: %r", (ds_id, ds_name))
        options = dict(ds_id=ds_id)
        title = "File list for dataset %r" % ds_name
        widget = self.makePanel(MdaFileList, self.db, None, title, options)
        urwid.connect_signal(widget, 'selected', self.fileSelected)

    def runLbSelected(self, widget, runlb_id, run, lb):
        _log.debug("selected item: %r %r %r", runlb_id, run, lb)
        if runlb_id is None:
            title = "File list for all runs"
        else:
            title = "File list for run {}".format(run)
            if lb is None:
                title += " (EoR)"
            else:
                title += " (LB={})".format(lb)
        widget = self.makePanel(MdaFileList, self.db, runlb_id, title)
        urwid.connect_signal(widget, 'selected', self.fileSelected)

    def fileSelected(self, widget, file_id, file_name):
        _log.debug("selected item: %r", file_id)
        widget = self.makePanel(MdaHgroupList, self.db, file_id)
        urwid.connect_signal(widget, 'selected', self.hgroupSelected)

    def hgroupSelected(self, widget, hgrp_id):
        _log.debug("selected item: %r", hgrp_id)
        widget = self.makePanel(MdaHistoList, self.db, hgrp_id)
        urwid.connect_signal(widget, 'selected', self.histoSelected)

    def histoSelected(self, widget, hist_id, lbsq_id):
        _log.debug("selected item: %r %r", hist_id, lbsq_id)
        if lbsq_id is None:
            _log.debug("LBSQ_ID is None, returning")
            return
        self.makePanel(MdaLBSeqList, self.db, lbsq_id)
