"""Module containing MdaLBSeqList class and related methods.
"""

#--------------------------------
#  Imports of standard modules --
#--------------------------------
import logging

#-----------------------------
# Imports for other modules --
#-----------------------------
import urwid
from ..AppPanel import AppPanel
from ..ui import UIListBoxWithHeader, UIColumns, UISelectableText
from ..walkers import PageWalker

#----------------------------------
# Local non-exported definitions --
#----------------------------------

_log = logging.getLogger(__name__)


def _formatRanges(min, max, total, ranges):

    dist = max + 1 - min
    if len(ranges) > 1 and (dist - total) < (dist // 10):
        res = str(min) + '-' + str(max)
        whole_range = set(range(min, max + 1))
        for min, max in ranges:
            whole_range -= set(range(min, max + 1))
        exclude = ' '.join(str(x) for x in sorted(whole_range))
        res += ' (except: ' + exclude + ')'
        return res
    else:
        rlist = []
        for min, max in ranges:
            if min == max:
                rlist += [str(min)]
            else:
                rlist += [str(min) + '-' + str(max)]
        rlist = ' '.join(rlist)
        return rlist

#------------------------
# Exported definitions --
#------------------------


class MdaLBSeqList(urwid.WidgetWrap, AppPanel):
    """Widget class containing list of CoCa datasets.

    Parameters
    ----------
    app : `MdaAdmin`
    db : `MdaDb`
    lbsq_id : `int`
        Id oof LBSQ to display, None to display all
    """

    _selectable = True

    signals = ['selected']

    #----------------
    #  Constructor --
    #----------------
    def __init__(self, app, db, lbsq_id=None):

        self._app = app
        self._db = db
        self._lbsq_id = lbsq_id
        self._col_width = [8, 6, 6, 6, -1]
        self._total_count = self._counter()
        self._listwalker = PageWalker(self._pager, self._total_count)

        header = UIColumns([urwid.Text('ID', align='right'),
                            urwid.Text('SeqLen', align='right'),
                            urwid.Text('MinLB', align='right'),
                            urwid.Text('MaxLB', align='right'),
                            urwid.Text('Lumi Blocks')], self._col_width, 1)

        lb = UIListBoxWithHeader(self._listwalker, header=header)
        self.__super.__init__(lb)

    def title(self):
        if self._lbsq_id is None:
            return "LB Sequences"
        else:
            return "LB Sequences for LBSQ_ID={}".format(self._lbsq_id)

    def status(self):
        status = "{} out of {} sequences".format(self._listwalker.focus + 1, self._total_count)
        return status

    def _counter(self):
        return self._db.lbseq_count()

    def _pager(self, offset, pageSize):
        items = self._db.lbseq(self._lbsq_id, maxRows=pageSize, offset=offset)
        return [self._widgetFactory(item) for item in items]

    def _widgetFactory(self, lbseq):

        ranges = _formatRanges(lbseq.lbsq_min_lb, lbseq.lbsq_max_lb, lbseq.lbsq_size, lbseq.ranges)

        item = UIColumns([UISelectableText(str(int(lbseq.lbsq_id)), align='right'),
                          urwid.Text(str(int(lbseq.lbsq_size)), align='right'),
                          urwid.Text(str(int(lbseq.lbsq_min_lb)), align='right'),
                          urwid.Text(str(int(lbseq.lbsq_max_lb)), align='right'),
                          urwid.Text(ranges)],
                         self._col_width, 1, 0)
        item = urwid.AttrMap(item, "list-item", "list-selected")
        return item
