"""Helper methods and classes for UI part.
"""
from __future__ import division

#--------------------------------
#  Imports of standard modules --
#--------------------------------
from datetime import datetime
import logging
import sys
import time

#-----------------------------
# Imports for other modules --
#-----------------------------
import urwid

#----------------------------------
# Local non-exported definitions --
#----------------------------------

_log = logging.getLogger(__name__)

if sys.version_info.major >= 3:
    long = int

#------------------------
# Exported definitions --
#------------------------


class UIListBox(urwid.ListBox):
    """
    This is an extension of the ListBox with few additional niceties.

    Constructor takes the same arguments as urwid.ListBox.
    """

    def __init__(self, *args):
        urwid.ListBox.__init__(self, *args)

        # remap home/end keys
        self._command_map = self._command_map.copy()
        self._command_map['home'] = 'cursor top'
        self._command_map['end'] = 'cursor bottom'

    def keypress(self, size, key):
        """
        Define reaction to some special keys like begin/end
        """

        # see if any child wants to handle
        key = urwid.ListBox.keypress(self, size, key)

        command = self._command_map[key]
        _log.debug("UIListBox: received key: %s -> %s", key, command)

        if command in ["cursor top", "cursor bottom"]:
            # only support it if walker has positions() method
            positions_fn = getattr(self.body, 'positions', None)
            if positions_fn is not None:
                try:
                    if command == "cursor top":
                        positions = positions_fn(False)
                        self.change_focus(size, positions[0], 0, 'below')
                        _log.debug("change focus to: %s", positions[0])
                    else:
                        positions = positions_fn(True)
                        self.change_focus(size, positions[0], 0, 'above')
                        _log.debug("change focus to: %s", positions[0])
                except Exception as exc:
                    _log.error('exception: %s', exc, exc_info=True)
                return

        return key


class UIListBoxWithHeader(urwid.Frame):
    """
    Combination of listbox and headers/footers.
    """

    def __init__(self, body, header=None, footer=None, focus_part='body'):
        """
        @param body:  walker instance passed to ListBox constructor or a list of
                      items, each item will be wrapped into AttrMap.

        Other parameters are passed to Frame constructor
        """

        if getattr(body, 'get_focus', None) is None:
            # wrap into SimpleFocusListWalker
            body = [urwid.AttrMap(item, "list-item", "list-selected") for item in body]
            body = urwid.SimpleFocusListWalker(body)

        listbox = UIListBox(body)
        if header:
            header = urwid.AttrMap(header, 'list-colhead')
        if footer:
            footer = urwid.AttrMap(footer, 'list-colhead')
        body = urwid.AttrMap(listbox, 'body')
        urwid.Frame.__init__(self, body, header=header, footer=footer, focus_part=focus_part)


class UISelectableText(urwid.Text):

    _selectable = True

    def __init__(self, markup, align='left', wrap='space', layout=None):
        """
        @param markup: Text Markup: content of text widget
        @param align (text alignment mode) - typically 'left', 'center' or 'right'
        @param wrap (text wrapping mode) - typically 'space', 'any' or 'clip'
        @param layout (text layout instance) - defaults to a shared StandardTextLayout instance
        @param on_press - shorthand for connect_signal() function call for a single callback
        @param user_data - user_data for on_press
        """
        urwid.Text.__init__(self, markup, align, wrap, layout)

    def keypress(self, size, key):
        return key


class UIColumns(urwid.Columns):
    """
    Convenience wrapper for Columns widget
    """

    signals = ['activated']

    def __init__(self, widget_list, width_list=None, dividechars=0, focus_column=None, min_width=1,
                 box_columns=None, wrap='clip'):
        """
        @param widget_list: List of items, each item can be either widget or an object, if
                    it is non-widget it is converted to string and wrapped into UISelectableText widget
        @param width_list:  List of widget widths, if given must have the same length as
                    widget_list. Items in the list are either positive numbers for absolute
                    width, negative numbers for negative weight, or string 'pack' to calculate
                    actual size of the widget.
        @param wrap: Wrapping method for text widgets created from strings in widget_list.

        Other parameters are the same as in call to urwid.Columns constructor.
        """

        def widget_factory(w):
            if not isinstance(w, urwid.Widget):
                w = urwid.Text(str(w), wrap=wrap)
            return w

        def width_tuple_factory(w, width):
            if isinstance(width, (int, float)):
                if width < 0:
                    return ('weight', -width, w)
                else:
                    return (width, w)
            elif width == 'pack':
                return ('pack', w)
            else:
                return ('weight', 1, w)

        # make widget list
        widget_list = [widget_factory(w) for w in widget_list]

        # zip with widths
        if width_list:
            widget_list = [width_tuple_factory(w, width) for w, width in zip(widget_list, width_list)]

        urwid.Columns.__init__(self, widget_list, dividechars, focus_column, min_width, box_columns)

        # we want to have a focus in a single column only
        self._command_map = self._command_map.copy()
        del self._command_map['left']
        del self._command_map['right']

    def keypress(self, size, key):

        _log.debug("UIColumns: received key: %s", key)

        # see if any child is interested in this key
        key = urwid.Columns.keypress(self, size, key)

        # if no one grabs a key then process it here
        if self._command_map[key] == urwid.ACTIVATE:
            self._emit('activated')
            key = None

        return key


def _keymap_for_checkbox():
    # Do not use enter for toggle, only space and insert
    command_map = urwid.command_map.copy()
    command_map['enter'] = None
    command_map['insert'] = 'activate'
    return command_map


class SlimCheckBox(urwid.CheckBox):
    """
    CheckBox which is one character wide.
    """

    states = {
        True: urwid.SelectableIcon(('selected', "X")),
        False: urwid.SelectableIcon(" "),
        'mixed': urwid.SelectableIcon("#")}
    reserve_columns = 1

    _command_map = _keymap_for_checkbox()


class UIPopUp(urwid.WidgetWrap):
    """Pop-up widget used with UIPopUpLauncher class below.

    Parameters
    ----------
    widget : `Widget`
        Widget used for pop-up.
    pop_up_size : `tuple`
        (columns, rows) for pop-up widget
    location : `str`
        String with pop-up anchors, can include characters "N", or "S" for
        vertical anchor, and "W" or ""E" for horizontal anchor. Default anchor
        is at the center of parent widget.
    """

    signals = ['closed']

    def __init__(self, widget, pop_up_size, location='C'):
        self.__super.__init__(widget)
        self._pop_up_size = pop_up_size
        self._location = location

    def get_pop_up_parameters(self, parent_size):
        cols, rows = self._pop_up_size
        maxcol, maxrow = parent_size
        _log.debug("UIPopUp: get_pop_up_parameters: %s %s", self._pop_up_size, parent_size)

        if 'N' in self._location:
            top = 0
        elif 'S' in self._location:
            top = maxrow - rows
        else:
            top = (maxrow - rows) // 2
        if 'W' in self._location:
            left = 0
        elif 'E' in self._location:
            left = maxcol - cols
        else:
            left = (maxcol - cols) // 2
        return dict(left=left, top=top, overlay_width=cols, overlay_height=rows)

    def close(self, *args):
        """Typically this will be a callback for a signal generated by a button.
        """
        self._emit("closed")


class UIPopUpLauncher(urwid.PopUpLauncher):
    """
    PopUpLauncher class with pipml.

    This class is useful for widgets that need to support multiple pop-up
    widget types (e.g. dialog, menu, and error message). It defines
    make_pop_up() method which is used in place open_pop_up) and accepts
    additional context information. This context info is passed to
    create_pop_up() and get_pop_up_parameters() methods in pimpl so that
    they can detect the type of the pop up being instantiated.
    """

    _parent_size = (0, 0)

    def make_pop_up(self, pop_up):
        """Make a pop-up.

        Parameters
        ----------
        pop_up : `UIPopUp`
        """
        self._pop_up = pop_up
        self.open_pop_up()
        urwid.connect_signal(self._pop_up, 'closed', self.close_pop_up)

    def create_pop_up(self):
        return self._pop_up

    def close_pop_up(self, *args):
        """Overrides parent class method and accepts any number of parameters.

        This simplifies signal connection for signals that carry extra
        parameters.
        """
        return urwid.PopUpLauncher.close_pop_up(self)

    def get_pop_up_parameters(self):
        return self._pop_up.get_pop_up_parameters(self._parent_size)

    def set_parent_size(self, size):
        self._parent_size = size


class UIMenu(urwid.WidgetWrap):
    """
    Widget implementing simple menu.

    Consists of a ListBox with a bunch of buttons, when button
    is clicked this widget generates 'selected' signal with a value
    which corresponds to that particular button. Optionally Escape
    key can be configured to also generate 'selected' signal with
    specified value.
    """

    signals = ['selected']

    def __init__(self, title, options, escape_value=None):
        """
        @param title:    Menu title
        @param options:  List of tuples (string, value)
        @param escape_value:  if not None then escape key will generate
                         'selected' signal with this value
        """

        items = []
        for name, value in options:
            button = urwid.Button(name)
            urwid.connect_signal(button, 'click', self._callback, user_args=[value])
            items.append(urwid.AttrMap(button, None, 'menu-focus'))
        listbox = urwid.ListBox(urwid.SimpleFocusListWalker(items))
        linebox = urwid.LineBox(listbox, title)
        self.__super.__init__(urwid.AttrMap(linebox, 'menu'))

        self.escape_value = escape_value

    def _callback(self, value, button):
        self._emit('selected', value)

    def keypress(self, size, key):
        _log.debug("UIMenu: received key: %s", key)
        if self.escape_value is not None and key == 'esc':
            _log.debug("UIMenu: escape key: %s", self.escape_value)
            self._emit('selected', self.escape_value)
            return None
        return self.__super.keypress(size, key)


class UIMessageBox(urwid.WidgetWrap):
    """
    Widget implementing simple message box.

    Consists of a text widget with a message and few buttons. When
    button is clicked this widget generates 'selected' signal with a
    value which corresponds to that particular button. Optionally Escape
    key can be configured to also generate 'selected' signal with
    specified value.

    """

    signals = ['selected']

    def __init__(self, title, message, buttons, escape_value=None, attr='messagebox'):
        """
        @param title:    Menu title
        @param message:  Message string
        @param buttons:  List of ('button_text', value) tuples, value is used for
                         'selected' signal when corresponding button is clicked
        @param escape_value:  if not None then escape key will generate
                         'selected' signal with this value
        """

        stretch = urwid.Text(' ' * 132, wrap='clip')  # stretch space to the left of the buttons
        items = [stretch]
        for name, value in buttons:
            button = urwid.Button(name)
            urwid.connect_signal(button, 'click', self._callback, user_args=[value])
            size = len(name) + 4
            items.append((size, urwid.AttrMap(button, None, attr + '-focus')))
        items += [stretch]
        button_row = urwid.Columns(items, 2)
        text = urwid.Text(message)
        pile = urwid.Pile([text, urwid.Divider(u'\u2500'), button_row])
        filler = urwid.Filler(pile, height='flow')     # need to wrap it into a box widget
        linebox = urwid.LineBox(filler, title)         # wrap it all into linebox with title
        self.__super.__init__(urwid.AttrMap(linebox, attr))

        self.escape_value = escape_value

    def _callback(self, value, button):
        self._emit('selected', value)

    def keypress(self, size, key):
        _log.debug("UIMenu: received key: %s", key)
        if self.escape_value is not None and key == 'esc':
            _log.debug("UIMenu: escape key: %s", self.escape_value)
            self._emit('selected', self.escape_value)
            return None
        return self.__super.keypress(size, key)


class _LoopProgressBar(urwid.ProgressBar):
    """Customized progress bar with loop effect.
    """
    def __init__(self, normal, complete, current=0, done=100, satt=None):
        self._loop = False
        if done <= 0:
            self._loop = True
            self._loop_start_time = time.time()
            done = 100
        urwid.ProgressBar.__init__(self, normal, complete, current, done, satt)

    def get_text(self):
        if self._loop:
            return "working"
        else:
            return urwid.ProgressBar.get_text(self)

    def set_completion(self, value):
        if self._loop:
            # in loop mode we progress as 10% per second
            now = time.time()
            value = (now - self._loop_start_time) * 10.
            if value >= 100:
                value = 0
                self._loop_start_time = now
        urwid.ProgressBar.set_completion(self, value)


class UIProgressBar(urwid.WidgetWrap):
    """Widget implementing box with progress bar, title and a button.

    When button is clicked this widget generates 'cancelled' signal.
    Optionally Escape key can be configured to also generate 'cancelled'
    signal.

    Parameters
    ----------
    title : `str`
    cancel_button : `bool`, optional
        If True then show a Cancel button.
    escape_signal : `bool`, optional
        If True then Escape key generates "cancelled" signal.
    attr : `str`
        Attributes for widget area
    """

    signals = ['cancelled']

    def __init__(self, title, current=0, done=100,
                 cancel_button=False, escape_signal=False, attr='progressbar'):
        self.progress_bar = _LoopProgressBar(attr + "-incomplete", attr + "-complete", current, done)
        if cancel_button:
            stretch = urwid.Text(' ' * 132, wrap='clip')  # stretch space to the left of the buttons
            buttons = [stretch]
            button = urwid.Button("Cancel")
            urwid.connect_signal(button, 'click', self._cancel)
            buttons += [(10, urwid.AttrMap(button, None, attr + '-focus'))]
            buttons += [stretch]
            inner = urwid.Pile([self.progress_bar, urwid.Divider(u'\u2500'), urwid.Columns(buttons, 2)])
        else:
            inner = self.progress_bar
        filler = urwid.Filler(inner, height='flow')     # need to wrap it into a box widget
        linebox = urwid.LineBox(filler, title)         # wrap it all into linebox with title
        self.__super.__init__(urwid.AttrMap(linebox, attr))

        self._escape_signal = escape_signal

    def _cancel(self, button):
        self._emit('cancelled', button)

    def keypress(self, size, key):
        if self._escape_signal is not None and key == 'esc':
            self._emit('cancelled')
            return None
        return self.__super.keypress(size, key)


class UIKeyInputRunlb(urwid.WidgetWrap):
    second_caption = "LBN"

    def __init__(self, init_value, attr='editbox', def_run=0, def_lbn=0):
        if init_value is None:
            init_value = 0
        run, lbn = init_value // 2**32, init_value % 2**32
        self.run_edit = urwid.IntEdit((attr, "Run"), run)
        self.lbn_edit = urwid.IntEdit((attr, self.second_caption), lbn)
        columns = urwid.Columns([(10, self.run_edit), (10, self.lbn_edit)], 1)
        self.__super.__init__(urwid.AttrMap(columns, attr + "-field"))

        self.def_run = def_run
        self.def_lbn = def_lbn

    def value(self):
        run = self.run_edit.value()
        lbn = self.lbn_edit.value()
        return run << 32 + lbn


class _UIKeyInput_runevt(urwid.WidgetWrap):
    second_caption = "Event"


class UIKeyInputTime(urwid.WidgetWrap):
    def __init__(self, init_value, attr='editbox'):
        if init_value is None:
            dt = datetime.now()
        else:
            dt = datetime.utcfromtimestamp(init_value // 1000000000)

        self.edits = [
            (5, urwid.IntEdit((attr, ""), dt.year)),
            (4, urwid.IntEdit((attr, "-"), dt.month)),
            (4, urwid.IntEdit((attr, "-"), dt.day)),
            (4, urwid.IntEdit((attr, " "), dt.hour)),
            (4, urwid.IntEdit((attr, ":"), dt.minute)),
            (4, urwid.IntEdit((attr, ":"), 0))]
        columns = urwid.Columns(self.edits, 0)
        self.__super.__init__(urwid.AttrMap(columns, attr + "-field"))

    def value(self):
        values = [edit[1].value() for edit in self.edits]
        dt = datetime(*values)
        epoch = datetime.utcfromtimestamp(0)
        delta = dt - epoch
        return int(delta.total_seconds()) * long(1000000000)


class _UIKeyInput_unknown(urwid.WidgetWrap):
    def __init__(self, init_value, attr='editbox'):
        if init_value is None:
            init_value = 0
        self.edit = urwid.IntEdit("", init_value)
        self.__super.__init__(self.edit)

    def value(self):
        return self.edit.value()


class UIKeyInput(urwid.WidgetWrap):
    """Widget implementing the input for IOV "time" point.

    Depending on folder type it will show inputs for run/LB, time, or a key.
    """
    def __init__(self, init_value, iov_type, attr='editbox'):
        if iov_type == "run-lumi":
            self.key_input = UIKeyInputRunlb(init_value, attr)
        elif iov_type == "run-event":
            self.key_input = _UIKeyInput_runevt(init_value, attr)
        elif iov_type == "time":
            self.key_input = UIKeyInputTime(init_value, attr)
        else:
            self.key_input = _UIKeyInput_unknown(init_value, attr)
        self.__super.__init__(self.key_input)

    def value(self):
        return self.key_input.value()


class UIKeysInput(urwid.WidgetWrap):

    signals = ['selected', 'cancelled']

    # per-folder-type most recent selected keys
    last_values = dict()

    def __init__(self, title, message, iov_type, attr='editbox'):
        """
        @param title:    Menu title
        @param message:  Message string
        @param iov_type: Folder IOV type, one of CoolDB.IOV_* constants
        """
        stretch = urwid.Text(' ' * 132, wrap='clip')  # stretch space to the left of the buttons
        buttons = [stretch]
        button = urwid.Button("Cancel")
        urwid.connect_signal(button, 'click', self._cancelled)
        buttons.append((10, urwid.AttrMap(button, None, attr + '-focus')))
        button = urwid.Button("OK")
        urwid.connect_signal(button, 'click', self._callback, user_args=[iov_type])
        buttons.append((6, urwid.AttrMap(button, None, attr + '-focus')))
        buttons += [stretch]
        button_row = urwid.Columns(buttons, 2)

        stretch = urwid.Text(' ' * 132, wrap='clip')
        last_since, last_until = UIKeysInput.last_values.get(iov_type, (None, None))
        self.since = UIKeyInput(last_since, iov_type, attr)
        since = urwid.Columns([(6, urwid.Text("Since:")), ('weight', 2, self.since), stretch], 1, 1)
        self.until = UIKeyInput(last_until, iov_type, attr)
        until = urwid.Columns([(6, urwid.Text("Until:")), ('weight', 2, self.until), stretch], 1, 1)
        pile = urwid.Pile([urwid.Text(message), since, until, urwid.Divider(u'\u2500'), button_row])
        filler = urwid.Filler(pile, height='flow')     # need to wrap it into a box widget
        linebox = urwid.LineBox(filler, title)         # wrap it all into linebox with title
        self.__super.__init__(urwid.AttrMap(linebox, attr))

    def _cancelled(self, button):
        self._emit("cancelled")

    def _callback(self, iov_type, button):
        since = self.since.value()
        until = self.until.value()
        UIKeysInput.last_values[iov_type] = (since, until)
        self._emit("selected", since, until)

    def keypress(self, size, key):
        _log.debug("UIKeysInput: received key: %s", key)
        if key == 'esc':
            self._emit('cancelled')
            return None
        return self.__super.keypress(size, key)


class UIPopUpMenu(UIPopUp):
    """Pop-up showing a menu.

    Parameters
    ----------
    title, message, buttons, escape_value, attr : same as for `UIMessageBox` class
    callback : `method`
        Method to call when menu item is selected.
    cols : `int`
        Width of a pop-up window
    """
    def __init__(self, title, options, callback, escape_value=None, cols=24):

        menu = UIMenu(title, options, 'cancel')
        # NOTE: the order of callbacks is important
        urwid.connect_signal(menu, 'selected', self.close)
        urwid.connect_signal(menu, 'selected', callback)

        pop_up_size = (cols, len(options) + 2)

        self.__super.__init__(menu, pop_up_size)


class UIPopUpMessageBox(UIPopUp):
    """Pop-up showing a message box.

    Parameters
    ----------
    title, options, escape_value : same as for `UIMenu` class
    callback : `method`, optional
        Method to call when any buton is activated.
    cols : `int`
        Width of a pop-up window
    """
    def __init__(self, title, message, buttons, callback=None, escape_value=None, attr='messagebox', cols=48):

        mbox = UIMessageBox(title, message, buttons, escape_value, attr)
        # NOTE: the order of callbacks is important
        urwid.connect_signal(mbox, 'selected', self.close)
        if callback is not None:
            urwid.connect_signal(mbox, 'selected', callback)

        rows = urwid.Text(message).rows((cols,))
        pop_up_size = (cols + 2, rows + 4)

        self.__super.__init__(mbox, pop_up_size)


class UIPopUpKeysInput(UIPopUp):
    """Pop-up showing a keys input box.

    Parameters
    ----------
    title, options, escape_value : same as for `UIMenu` class
    callback : `method`, optional
        Method to call when any buton is activated.
    cols : `int`
        Width of a pop-up window
    """

    signals = ["selected"]

    def __init__(self, title, message, iov_type, attr='editbox', cols=48):

        mbox = UIKeysInput(title, message, iov_type, attr)
        # NOTE: the order of callbacks is important
        urwid.connect_signal(mbox, 'selected', self.close)
        urwid.connect_signal(mbox, 'selected', self._callback)
        urwid.connect_signal(mbox, 'cancelled', self.close)

        rows = urwid.Text(message).rows((cols,))
        pop_up_size = (cols + 2, rows + 6)

        self.__super.__init__(mbox, pop_up_size)

    def _callback(self, widget, since, until):
        self._emit("selected", since, until)


class UIPopUpProgress(UIPopUp):
    """Pop-up showing a progress bar.
    """
    def __init__(self, title, current=0, done=100, callback=None, cols=36,
                 cancel_button=False, escape_signal=False, attr='progressbar'):
        self.pbar = UIProgressBar(title, current, done, cancel_button, escape_signal, attr)
        urwid.connect_signal(self.pbar, 'cancelled', self._cancel)
        if callback is not None:
            urwid.connect_signal(self.pbar, 'cancelled', callback)

        pop_up_size = (cols, 5 if cancel_button else 3)
        self.__super.__init__(self.pbar, pop_up_size)

        self._cancelled = False

    def _cancel(self):
        self._cancelled = True

    def set_done(self, value):
        """Set progress done value.
        """
        progressbar = self.pbar.progress_bar
        progressbar.done = value

    def set_completion(self, value):
        """Set progress current value.
        """
        progressbar = self.pbar.progress_bar
        progressbar.set_completion(value)

    def set_percent(self, percent):
        """Set progress current value as a percentage.
        """
        progressbar = self.pbar.progress_bar
        value = progressbar.done / 100. * percent
        progressbar.set_completion(value)

    def cancelled(self):
        """Return True if Cancel button was activated.
        """
        return self._cancelled
