
//-----------------------
// This Class's Header --
//-----------------------
#include "ChannelSelection.h"

//-----------------
// C/C++ Headers --
//-----------------

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "exceptions.h"

//-----------------------------------------------------------------------
// Local Macros, Typedefs, Structures, Unions and Forward Declarations --
//-----------------------------------------------------------------------

#if PY_MAJOR_VERSION >= 3
#define PY3MIG_STRING_CHECK(X) PyUnicode_Check(X)
#define PY3MIG_STRING_AS_STRING(X) PyUnicode_AsUTF8(X)
#define PY3MIG_INT_CHECK(X) PyLong_Check(X)
#define PY3MIG_INT_AS_LONG(X) PyLong_AsLong(X)
#else
#define PY3MIG_STRING_CHECK(X) PyString_Check(X)
#define PY3MIG_STRING_AS_STRING(X) PyString_AS_STRING(X)
#define PY3MIG_INT_CHECK(X) PyInt_Check(X)
#define PY3MIG_INT_AS_LONG(X) PyInt_AS_LONG(X)
#endif

namespace {

// standard Python stuff
PyObject* _new(PyTypeObject *subtype, PyObject *args, PyObject *kwds);
PyObject* _all(PyObject*, PyObject*);

// type-specific methods
// PyObject* _exists(PyObject* self, PyObject* args);

PyMethodDef methods[] = {
    {"all", _all, METH_NOARGS | METH_STATIC,
        "ChannelSelection.all() -> ChannelSelection\n\nReturns selector for all channels."},
    {0, 0, 0, 0}
};

char typedoc[] = "\
Helper class to specify a selection of channels and their ordering for\n\
multi-channel bulk retrieval of IOVs.\n\
\n\
Supported initialization options:\n\
\n\
   ChannelSelection() -- selector for all existing channels\n\
   ChannelSelection(channelId: int) -- select single integer channel ID\n\
   ChannelSelection(channelName: str) -- select one channel with given name\n\
   ChannelSelection(id1: int, id2: int) -- channels in range id1 to id2 (inclusive)\n\
\n\
Static method `ChannelSelection.all()` also returns selector which selects all\n\
existing channels.\n\
\n\
Note: instances of this class are only useful for passing to the methods of\n\
other classes, so it does not provide wrappers for regular methods.\n\
";
}

//      ------------------------------
//      -- Class Member Definitions --
//      ------------------------------

void
coldpie::ChannelSelection::initType(PyObject* module, const char* modname)
{
  PyTypeObject* type = BaseType::typeObject();
  type->tp_doc = ::typedoc;
  type->tp_methods = ::methods;
  type->tp_new = ::_new;

  BaseType::initType("ChannelSelection", module, modname);
}

// Dump object info to a stream
void
coldpie::ChannelSelection::print(std::ostream& out) const
{
  out << "<ChannelSelection@" << this << ">";
}

namespace {

// standard Python stuff
PyObject* _new(PyTypeObject *subtype, PyObject *args, PyObject */*kwds*/)
{
    // parse arguments
    PyObject* arg1 = nullptr;
    PyObject* arg2 = nullptr;
    if (not PyArg_ParseTuple(args, "|OO:ChannelSelection", &arg1, &arg2)) {
        return 0;
    }

    // with one argument we can accept int or string, with two arguments we can accept two ints
    cool::ChannelSelection chsel;
    try {
        if (arg1 != nullptr and arg2 != nullptr) {
            if (PY3MIG_INT_CHECK(arg1) && PY3MIG_INT_CHECK(arg2)) {
                auto val1 = PY3MIG_INT_AS_LONG(arg1);
                if (val1 == -1 and PyErr_Occurred() != nullptr) return nullptr;
                auto val2 = PY3MIG_INT_AS_LONG(arg2);
                if (val2 == -1 and PyErr_Occurred() != nullptr) return nullptr;
                chsel = cool::ChannelSelection(cool::ChannelId(val1), cool::ChannelId(val2));
            } else {
                PyErr_SetString(PyExc_TypeError, "two integer arguments are expected");
                return nullptr;
            }
        } else if (arg1 != nullptr) {
            if (PY3MIG_INT_CHECK(arg1)) {
                auto val1 = PY3MIG_INT_AS_LONG(arg1);
                if (val1 == -1 and PyErr_Occurred() != nullptr) return nullptr;
                chsel = cool::ChannelSelection(cool::ChannelId(val1));
            } else  if (PY3MIG_STRING_CHECK(arg1)) {
                chsel = cool::ChannelSelection(PY3MIG_STRING_AS_STRING(arg1));
            } else {
                PyErr_SetString(PyExc_TypeError, "integer or string argument is expected");
                return nullptr;
            }
        }
    } catch (std::exception const& exc) {
        PyErr_SetString(PyExc_ValueError, exc.what());
        return nullptr;
    }

    // copy fields into C++ object
    try {
        auto self = coldpie::ChannelSelection::construct(subtype, chsel);
        return self;
    } catch (const std::exception& exc) {
        coldpie::translateException(exc);
        return nullptr;
    }
}

PyObject* _all(PyObject*, PyObject*)
{
    auto chsel = cool::ChannelSelection::all();
    return coldpie::ChannelSelection::PyObject_FromCpp(std::move(chsel));
}

}
