//----------------------------------------------------------------------
// File and Version Information:
//  $Id$
//
// Author List:
//      Andy Salnikov
//----------------------------------------------------------------------

//-----------------------
// This Class's Header --
//-----------------------
#include "Folder.h"

//-----------------
// C/C++ Headers --
//-----------------

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "coldpie/gil.h"
#include "CoolKernel/FolderSpecification.h"
#include "CoolKernel/RecordSpecification.h"
#include "CoolKernel/Time.h"
#include "ChannelSelection.h"
#include "FolderSpecification.h"
#include "Object.h"
#include "ObjectIterator.h"
#include "Record.h"
#include "RecordSpecification.h"
#include "cpp2py.h"
#include "exceptions.h"

//-----------------------------------------------------------------------
// Local Macros, Typedefs, Structures, Unions and Forward Declarations --
//-----------------------------------------------------------------------

#if PY_MAJOR_VERSION >= 3
#define PY3MIG_INT_FROM_LONG(X) PyLong_FromLong(X)
#else
#define PY3MIG_INT_FROM_LONG(X) PyInt_FromLong(X)
#endif

namespace {

//
//  Methods of IHvsNodeRecord
//

//    const std::string& IHvsNodeRecord::fullPath() const = 0;
PyObject* _fullPath(PyObject* self, PyObject*)
try {
    auto folder = coldpie::Folder::cppObject(self);
    return coldpie::cpp2py::convert(folder->fullPath());
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//    const std::string& IHvsNodeRecord::description() const = 0;
PyObject* _description(PyObject* self, PyObject*)
try {
    auto folder = coldpie::Folder::cppObject(self);
    return coldpie::cpp2py::convert(folder->description());
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//    bool IHvsNodeRecord::isLeaf() const = 0;
PyObject* _isLeaf(PyObject* self, PyObject*)
try {
    auto folder = coldpie::Folder::cppObject(self);
    return coldpie::cpp2py::convert(folder->isLeaf());
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//    bool IHvsNodeRecord::isStored() const = 0;
PyObject* _isStored(PyObject* self, PyObject*)
try {
    auto folder = coldpie::Folder::cppObject(self);
    return coldpie::cpp2py::convert(folder->isStored());
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//    const ITime& IHvsNodeRecord::insertionTime() const = 0;
PyObject* _insertionTime(PyObject* self, PyObject*)
try {
    auto folder = coldpie::Folder::cppObject(self);
    return coldpie::cpp2py::convert(folder->insertionTime());
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//    UInt32 IHvsNodeRecord::id() const = 0;  -- not implemented yet
//    UInt32 IHvsNodeRecord::parentId() const = 0;  -- not implemented yet

//
//  Methods of IHvsNode
//

//    const std::vector<std::string> IHvsNode::listTags() const = 0;
PyObject* _listTags(PyObject* self, PyObject*)
try {
    auto folder = coldpie::Folder::cppObject(self);
    return coldpie::cpp2py::convert(folder->listTags());
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//    const Time IHvsNode::tagInsertionTime( const std::string& tagName ) const = 0;
PyObject* _tagInsertionTime(PyObject* self, PyObject* args)
try {
    const char* tag = 0;
    if (not PyArg_ParseTuple(args, "s:FolderSet.tagInsertionTime", &tag)) return 0;

    auto folder = coldpie::Folder::cppObject(self);
    return coldpie::cpp2py::convert(folder->tagInsertionTime(tag));
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//    const std::string IHvsNode::tagDescription( const std::string& tagName ) const = 0;
PyObject* _tagDescription(PyObject* self, PyObject* args)
try {
    const char* tag = 0;
    if (not PyArg_ParseTuple(args, "s:FolderSet.tagDescription", &tag)) return 0;

    auto folder = coldpie::Folder::cppObject(self);
    return coldpie::cpp2py::convert(folder->tagDescription(tag));
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//     HvsTagLock::Status IHvsNode::tagLockStatus( const std::string& tagName ) const = 0;
PyObject* _tagLockStatus(PyObject* self, PyObject* args)
try {
    const char* tag = 0;
    if (not PyArg_ParseTuple(args, "s:FolderSet.tagLockStatus", &tag)) return 0;

    auto folder = coldpie::Folder::cppObject(self);
    auto status = folder->tagLockStatus(tag);
    return PY3MIG_INT_FROM_LONG(status);
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//
//  Methods of IFolder
//

//    virtual const IFolderSpecification& folderSpecification() const = 0;
PyObject* _folderSpecification(PyObject* self, PyObject*)
try {
    auto folder = coldpie::Folder::cppObject(self);
    const auto& fspec = folder->folderSpecification();
    return coldpie::FolderSpecification::PyObject_FromCpp(fspec, self);
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//    virtual const IRecordSpecification& payloadSpecification() const = 0;
PyObject* _payloadSpecification(PyObject* self, PyObject*)
try {
    auto folder = coldpie::Folder::cppObject(self);
    return coldpie::RecordSpecification::PyObject_FromCpp(folder->payloadSpecification(), self);
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//    virtual FolderVersioning::Mode versioningMode() const = 0;
PyObject* _versioningMode(PyObject* self, PyObject*)
try {
    auto folder = coldpie::Folder::cppObject(self);
    return PY3MIG_INT_FROM_LONG(folder->versioningMode());
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//    virtual const IRecord& folderAttributes() const = 0;
PyObject* _folderAttributes(PyObject* self, PyObject*)
try {
    auto folder = coldpie::Folder::cppObject(self);
    const auto& attr = folder->folderAttributes();
    return coldpie::Record_const::PyObject_FromCpp(attr, self);
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//    virtual void flushStorageBuffer() = 0;
PyObject* _flushStorageBuffer(PyObject* self, PyObject*)
try {
    auto folder = coldpie::Folder::cppObject(self);
    {
        // release GIL for possibly long I/O operation
        coldpie::gil gil;
        folder->flushStorageBuffer();
    }
    Py_RETURN_NONE;
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//    virtual void setupStorageBuffer( bool useBuffer = true ) = 0;
PyObject* _setupStorageBuffer(PyObject* self, PyObject* args)
try {
    int useBuffer = true;
    if (not PyArg_ParseTuple(args, "|i:Folder.setupStorageBuffer()", &useBuffer)) {
        return 0;
    }

    auto folder = coldpie::Folder::cppObject(self);
    folder->setupStorageBuffer(bool(useBuffer));
    Py_RETURN_NONE;
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

// virtual void storeObject( const ValidityKey& since,
//                           const ValidityKey& until,
//                           const IRecord& payload,
//                           const ChannelId& channelId,
//                           const std::string& userTagName = "",
//                           bool userTagOnly = false ) = 0;
PyObject* _storeObject(PyObject* self, PyObject* args)
try {
    unsigned PY_LONG_LONG since, until;
    PyObject* pyRec = nullptr;
    int channelId;
    const char* tagName = "";
    int userTagOnly = 0;
    if (not PyArg_ParseTuple(args, "KKOi|si:Folder.storeObject", &since, &until, &pyRec,
                             &channelId, &tagName, &userTagOnly)) return 0;

    // get c++ record
    const cool::IRecord* rec = nullptr;
    if (coldpie::Record::Object_TypeCheck(pyRec)) {
        rec = &coldpie::Record::cppObject(pyRec).get();
    } else if (coldpie::Record_const::Object_TypeCheck(pyRec)) {
        rec = &coldpie::Record_const::cppObject(pyRec).get();
    } else {
        PyErr_Format(PyExc_TypeError,
                     "Folder.storeObject() argument must be Record or Record_const, not %s",
                     pyRec->ob_type->tp_name);
        return nullptr;
    }
    {
        // release GIL for possibly long I/O operation
        coldpie::gil gil;
        auto folder = coldpie::Folder::cppObject(self);
        folder->storeObject(since, until, *rec, channelId, tagName, bool(userTagOnly));
    }
    Py_RETURN_NONE;
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

// virtual IObjectPtr findObject( const ValidityKey& pointInTime,
//                                const ChannelId& channelId,
//                                const std::string& tagName="" ) const = 0;
PyObject* _findObject(PyObject* self, PyObject* args)
try {
    unsigned PY_LONG_LONG key;
    int channelId;
    const char* tagName = "";
    if (not PyArg_ParseTuple(args, "Ki|s:Folder.findObject", &key, &channelId, &tagName)) return 0;

    auto folder = coldpie::Folder::cppObject(self);
    // release GIL for possibly long I/O operation
    cool::IObjectPtr obj;
    {
        // release GIL for possibly long I/O operation
        coldpie::gil gil;
        obj = folder->findObject(key, channelId, tagName);
    }
    return coldpie::Object::PyObject_FromCpp(obj);
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

// virtual IObjectIteratorPtr
// findObjects( const ValidityKey& pointInTime,
//              const ChannelSelection& channels,
//              const std::string& tagName = "" ) const = 0;
PyObject* _findObjects(PyObject* self, PyObject* args)
try {
    unsigned PY_LONG_LONG key;
    auto chSelType = coldpie::ChannelSelection::typeObject();
    PyObject* pyChSel = nullptr;
    const char* tagName = "";
    if (not PyArg_ParseTuple(args, "KO!|s:Folder.findObjects", &key, chSelType, &pyChSel, &tagName)) {
        return 0;
    }

    auto folder = coldpie::Folder::cppObject(self);
    auto& chSel = coldpie::ChannelSelection::cppObject(pyChSel);
    cool::IObjectIteratorPtr itr;
    {
        // release GIL for possibly long I/O operation
        coldpie::gil gil;
        itr = folder->findObjects(key, chSel, tagName);
    }
    return coldpie::ObjectIterator::PyObject_FromCpp(itr, self);
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

// virtual IObjectIteratorPtr
// browseObjects( const ValidityKey& since,
//                const ValidityKey& until,
//                const ChannelSelection& channels,
//                const std::string& tagName = "",
//                const IRecordSelection* payloadQuery = 0 ) const = 0;
PyObject* _browseObjects(PyObject* self, PyObject* args)
try {
    unsigned PY_LONG_LONG since, until;
    auto chSelType = coldpie::ChannelSelection::typeObject();
    PyObject* pyChSel = nullptr;
    const char* tagName = "";
    if (not PyArg_ParseTuple(args, "KKO!|s:Folder.browseObjects", &since, &until, chSelType, &pyChSel, &tagName)) {
        return 0;
    }

    auto folder = coldpie::Folder::cppObject(self);
    auto& chSel = coldpie::ChannelSelection::cppObject(pyChSel);
    cool::IObjectIteratorPtr itr;
    {
        coldpie::gil gil;
        itr = folder->browseObjects(since, until, chSel, tagName);
    }
    return coldpie::ObjectIterator::PyObject_FromCpp(itr, self);
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

// virtual unsigned int
// countObjects( const ValidityKey& since,
//               const ValidityKey& until,
//               const ChannelSelection& channels,
//               const std::string& tagName = "",
//               const IRecordSelection* payloadQuery = 0 ) const = 0;
PyObject* _countObjects(PyObject* self, PyObject* args)
try {
    unsigned PY_LONG_LONG since, until;
    auto chSelType = coldpie::ChannelSelection::typeObject();
    PyObject* pyChSel = nullptr;
    const char* tagName = "";
    if (not PyArg_ParseTuple(args, "KKO!|s:Folder.countObjects", &since, &until, chSelType, &pyChSel, &tagName)) {
        return 0;
    }

    auto folder = coldpie::Folder::cppObject(self);
    auto& chSel = coldpie::ChannelSelection::cppObject(pyChSel);
    unsigned count = 0;
    {
        coldpie::gil gil;
        count = folder->countObjects(since, until, chSel, tagName);
    }
    return coldpie::cpp2py::convert(count);
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//    virtual void setPrefetchAll( bool prefetchAll ) = 0;
PyObject* _setPrefetchAll(PyObject* self, PyObject* args)
try {
    int prefetchAll = true;
    if (not PyArg_ParseTuple(args, "|i:Folder.setPrefetchAll()", &prefetchAll)) {
        return 0;
    }

    auto folder = coldpie::Folder::cppObject(self);
    folder->setPrefetchAll(bool(prefetchAll));
    Py_RETURN_NONE;
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//    virtual const std::vector<ChannelId> listChannels() const = 0;
PyObject* _listChannels(PyObject* self, PyObject*)
try {
    auto folder = coldpie::Folder::cppObject(self);
    return coldpie::cpp2py::convert(folder->listChannels());
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}


// virtual void createChannel( const ChannelId& channelId,
//                             const std::string& channelName,
//                             const std::string& description = "" ) = 0;
PyObject* _createChannel(PyObject* self, PyObject* args)
try {
    int channelId;
    const char* name;
    const char* descr = "";
    if (not PyArg_ParseTuple(args, "is|s:Folder.createChannel", &channelId, &name, &descr)) return 0;

    auto folder = coldpie::Folder::cppObject(self);
    {
        // release GIL for possibly long I/O operation
        coldpie::gil gil;
        folder->createChannel(channelId, name, descr);
    }
    Py_RETURN_NONE;
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//     virtual const std::string channelName( const ChannelId& channelId ) const = 0;
PyObject* _channelName(PyObject* self, PyObject* args)
try {
    int channelId;
    if (not PyArg_ParseTuple(args, "i:Folder.channelName", &channelId)) return 0;

    auto folder = coldpie::Folder::cppObject(self);
    std::string name;
    {
        // release GIL for possibly long I/O operation
        coldpie::gil gil;
        name = folder->channelName(channelId);
    }
    return coldpie::cpp2py::convert(name);
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//    virtual ChannelId channelId( const std::string& channelName ) const = 0;

//    virtual bool existsChannel( const ChannelId& channelId ) const = 0;
PyObject* _existsChannel(PyObject* self, PyObject* args)
try {
    int channelId;
    if (not PyArg_ParseTuple(args, "i:Folder.existsChannel", &channelId)) return 0;

    auto folder = coldpie::Folder::cppObject(self);
    bool exists = false;
    {
        // release GIL for possibly long I/O operation
        coldpie::gil gil;
        exists = folder->existsChannel(channelId);
    }
    return coldpie::cpp2py::convert(exists);
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

//    virtual bool existsChannel( const std::string& channelName ) const = 0;
// virtual void setChannelDescription( const ChannelId& channelId,
//                                     const std::string& description ) = 0;

// virtual const std::string
// channelDescription( const ChannelId& channelId ) const = 0;
PyObject* _channelDescription(PyObject* self, PyObject* args)
try {
    int channelId;
    if (not PyArg_ParseTuple(args, "i:Folder.channelDescription", &channelId)) return 0;

    auto folder = coldpie::Folder::cppObject(self);
    std::string descr;
    {
        // release GIL for possibly long I/O operation
        coldpie::gil gil;
        descr = folder->channelDescription(channelId);
    }
    return coldpie::cpp2py::convert(descr);
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}


PyMethodDef methods[] = {
    {"fullPath", _fullPath, METH_NOARGS,
        "self.fullPath() -> str\n\n"
        "Return node full path name in the HVS hierarchy."},
    {"description", _description, METH_NOARGS,
        "self.description() -> str\n\n"
        "Return node description."},
    {"isLeaf", _isLeaf, METH_NOARGS,
        "self.isLeaf() -> bool\n\n"
        "Return True if this is a leaf node (should always return True for Folder)."},
    {"isStored", _isStored, METH_NOARGS,
        "self.isStored() -> bool\n\n"
        "Return True if this node is stored in the database."},
    {"insertionTime", _insertionTime, METH_NOARGS,
        "self.insertionTime() -> Time\n\n"
        "Return insertion time into the database."},

    {"listTags", _listTags, METH_NOARGS,
        "self.listTags() -> list [str]\n\n"
        "Lists all tags defined for this node (ordered alphabetically).\n\n"
        "Note: The reserved tags \"\" and \"HEAD\" are NOT included in the list."},
    {"tagInsertionTime", _tagInsertionTime, METH_VARARGS,
        "self.tagInsertionTime(tag: str) -> Time\n\n"
        "Return insertion time of a tag defined for this node.\n\n"
        "This is the time when the tag was first assigned to this node."},
    {"tagDescription", _tagDescription, METH_VARARGS,
        "self.tagDescription(tag: str) -> str\n\n"
        "Return description of a tag defined for this node."},
    {"tagLockStatus", _tagLockStatus, METH_VARARGS,
        "self.tagLockStatus(tag: str) -> int\n\n"
        "Return persistent lock status of a tag defined for this node, possible values\n"
        "are defined in HvsTagLock class."},

    {"payloadSpecification", _payloadSpecification, METH_NOARGS,
        "self.payloadSpecification() -> RecordSpecification\n\n"
        "Return RecordSpecification instance for folder payload."},
    {"folderSpecification", _folderSpecification, METH_NOARGS,
        "self.folderSpecification() -> FolderSpecification\n\n"
        "Return FolderSpecification instance for this folder."},
    {"versioningMode", _versioningMode, METH_NOARGS,
        "self.versioningMode() -> int\n\n"
        "Return the folder versioning mode, values are defined in FolderVersioning class."},
    {"folderAttributes", _folderAttributes, METH_NOARGS,
        "self.folderAttributes() -> Record\n\n"
        "Return attributes of the folder as Record instance."},
    {"flushStorageBuffer", _flushStorageBuffer, METH_NOARGS,
        "self.flushStorageBuffer()\n\nFlush storage buffers."},
    {"setupStorageBuffer", _setupStorageBuffer, METH_VARARGS,
        "self.setupStorageBuffer(useBuffer: bool)\n\n"
        "Activate (with True argument) or deactivate a storage buffer for bulk\n"
        "insertion of objects."},
    {"storeObject", _storeObject, METH_VARARGS,
        "self.storeObject(since: int, until: int, payload: Record, channelId: int,\n"
        "                 userTagName: str='', userTagOnly: bool=False)\n\n"
        "Store an object in a given channel with the given IOV and data payload.\n"
        "`since` and `until` define start and end of IOV, these are 64-bit integers.\n"
        "Channel has to be specified as a number (ID) and not channel name."
        },
    {"findObject", _findObject, METH_VARARGS,
        "self.findObject(pointInTime: int, channelId: int, tagName: str='') -> Object\n\n"
        "Find and return one object in a given channel valid at the given point in time.\n"
        "Note: This method returns cool.Object instance (not general Python object)."},
    {"findObjects", _findObjects, METH_VARARGS,
        "self.findObjects(pointInTime: int, channels: ChannelSelection,\n"
        "                 tagName: str='') -> iterator [Object]\n\n"
        "Find the objects for a given channel selection at the given point in time\n"
        "in the given tag. This method returns iterator for Object instances."},
    {"browseObjects", _browseObjects, METH_VARARGS,
        "self.browseObjects(since: int, until: int, channels: ChannelSelection,\n"
        "                   tagName: str='') -> iterator [Object]\n\n"
        "Browse the objects for a given channel selection within the given time range\n"
        "and the given tag. This method returns iterator for Object instances."},
    {"countObjects", _countObjects, METH_VARARGS,
        "self.countObjects(since: int, until: int, channels: ChannelSelection,\n"
        "                  tagName: str='') -> int\n\n"
        "Count the number of objects that would be returned by the browseObjects\n"
        "method for the same selection parameters."},
    {"setPrefetchAll", _setPrefetchAll, METH_VARARGS,
        "self.setPrefetchAll(prefetchAll: bool)\n\n"
        "Change the object prefetch policy (default is prefetchAll=true)."},
    {"listChannels", _listChannels, METH_NOARGS,
        "self.listChannels() -> list [int]\n\n"
        "Return the list of existing channels IDs (ordered by ascending channelId)."},
    {"createChannel", _createChannel, METH_VARARGS,
        "self.createChannel(channelId: int, channelName: str, description: str='')\n\n"
        "Create a new channel with the given id, name and description."},
    {"channelName", _channelName, METH_VARARGS,
        "self.channelName(channelId: int) -> str\n\n"
        "Return the name of a channel given its channel id."},
    {"existsChannel", _existsChannel, METH_VARARGS,
        "self.existsChannel(channelId: int) -> bool\n\n"
        "Returns True if a channel with given id exists."},
    {"channelDescription", _channelDescription, METH_VARARGS,
        "self.channelDescription(channelId: int) -> str\n\n"
        "Return the description of a channel given its id."},
    {0, 0, 0, 0}
};

char typedoc[] = "\
Python wrapper for COOL Folder C++ class.\n\
\n\
Instances of this class are created in C++ API, there is no user-accessible\n\
constructor at Python level. Use one of the methods of Database class to make\n\
instances of Folder class.\n\
";

}

void
coldpie::Folder::initType(PyObject* module, const char* modname)
{
  PyTypeObject* type = BaseType::typeObject();
  type->tp_doc = ::typedoc;
  type->tp_methods = ::methods;

  BaseType::initType("Folder", module, modname);
}

// Dump object info to a stream
void
coldpie::Folder::print(std::ostream& out) const
{
    out << "<Folder@" << this << ">";
}

