//----------------------------------------------------------------------
// File and Version Information:
//  $Id$
//
// Author List:
//      Andy Salnikov
//----------------------------------------------------------------------

//-----------------------
// This Class's Header --
//-----------------------
#include "FolderSet.h"

//-----------------
// C/C++ Headers --
//-----------------

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "CoolKernel/ITime.h"
#include "CoolKernel/Time.h"
#include "coldpie/gil.h"
#include "exceptions.h"
#include "Record.h"
#include "cpp2py.h"

//-----------------------------------------------------------------------
// Local Macros, Typedefs, Structures, Unions and Forward Declarations --
//-----------------------------------------------------------------------

namespace {

PyObject* _fullPath(PyObject* self, PyObject*);
PyObject* _description(PyObject* self, PyObject*);
PyObject* _isLeaf(PyObject* self, PyObject*);
PyObject* _isStored(PyObject* self, PyObject*);
PyObject* _insertionTime(PyObject* self, PyObject*);
PyObject* _listTags(PyObject* self, PyObject*);
PyObject* _tagInsertionTime(PyObject* self, PyObject* args);
PyObject* _tagDescription(PyObject* self, PyObject* args);
PyObject* _tagLockStatus(PyObject* self, PyObject* args);
PyObject* _listFolders(PyObject* self, PyObject* args);
PyObject* _listFolderSets(PyObject* self, PyObject* args);
PyObject* _folderSetAttributes(PyObject* self, PyObject*);

PyMethodDef methods[] = {
    {"fullPath", _fullPath, METH_NOARGS,
        "self.fullPath() -> str\n\n"
        "Return node full path name in the HVS hierarchy."},
    {"description", _description, METH_NOARGS,
        "self.description() -> str\n\n"
        "Return node description."},
    {"isLeaf", _isLeaf, METH_NOARGS,
        "self.isLeaf() -> bool\n\n"
        "Return True if this is a leaf node (always returns False for FolderSet)"},
    {"isStored", _isStored, METH_NOARGS,
        "self.isStored() -> bool\n\n"
        "Return True if this node is stored in the database."},
    {"insertionTime", _insertionTime, METH_NOARGS,
        "self.insertionTime() -> Time\n\n"
        "Return insertion time into the database."},
    {"listTags", _listTags, METH_NOARGS,
        "self.listTags() -> list [str]\n\n"
        "Lists all tags defined for this node (ordered alphabetically).\n\n"
        "Note: The reserved tags \"\" and \"HEAD\" are NOT included in the list."},
    {"tagInsertionTime", _tagInsertionTime, METH_VARARGS,
        "self.tagInsertionTime(tag: str) -> Time\n\n"
        "Return insertion time of a tag defined for this node.\n\n"
        "This is the time when the tag was first assigned to this node."},
    {"tagDescription", _tagDescription, METH_VARARGS,
        "self.tagDescription(tag: str) -> str\n\n"
        "Return description of a tag defined for this node."},
    {"tagLockStatus", _tagLockStatus, METH_VARARGS,
        "self.tagLockStatus(tag: str) -> int\n\n"
        "Return persistent lock status of a tag defined for this node, possible values\n"
        "are defined in HvsTagLock class."},
    {"listFolders", _listFolders, METH_VARARGS,
        "self.listFolders(ascending: bool=True) -> list [str]\n\n"
        "Lists all folders at this level in the node hierarchy (ordered alphabetically\n"
        "ascending/descending)."},
    {"listFolderSets", _listFolderSets, METH_VARARGS,
        "self.listFolderSets(ascending:=True) -> list [str]\n\n"
        "Lists all folder sets at this level in the node hierarchy (ordered alphabetically\n"
        "ascending/descending)."},
    {"folderSetAttributes", _folderSetAttributes, METH_NOARGS,
        "self.folderSetAttributes() -> Record\n\n"
        "Return attributes of the folderset (implementation-specific properties not\n"
        "exposed in the API)."},
    {0, 0, 0, 0}
};

char typedoc[] = "\
Python wrapper for COOL FolderSet C++ class.\n\
\n\
Instances of this class are created in C++ API, there is no user-accessible\n\
constructor at Python level. Use one of the methods of Database class to make\n\
instances of FolderSet class.\n\
";
}


//      ------------------------------
//      -- Class Member Definitions --
//      ------------------------------

void
coldpie::FolderSet::initType(PyObject* module, const char* modname)
{
  PyTypeObject* type = BaseType::typeObject();
  type->tp_doc = ::typedoc;
  type->tp_methods = ::methods;

  BaseType::initType("FolderSet", module, modname);
}

// Dump object info to a stream
void
coldpie::FolderSet::print(std::ostream& out) const
{
    out << "<FolderSet@" << this << ">";
}

namespace {

PyObject* _fullPath(PyObject* self, PyObject*)
try {
    auto folderSet = coldpie::FolderSet::cppObject(self);
    return coldpie::cpp2py::convert(folderSet->fullPath());
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

PyObject* _description(PyObject* self, PyObject*)
try {
    auto folderSet = coldpie::FolderSet::cppObject(self);
    return coldpie::cpp2py::convert(folderSet->description());
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

PyObject* _isLeaf(PyObject* self, PyObject*)
try {
    auto folderSet = coldpie::FolderSet::cppObject(self);
    return coldpie::cpp2py::convert(folderSet->isLeaf());
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

PyObject* _isStored(PyObject* self, PyObject*)
try {
    auto folderSet = coldpie::FolderSet::cppObject(self);
    return coldpie::cpp2py::convert(folderSet->isStored());
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

PyObject* _insertionTime(PyObject* self, PyObject*)
try {
    auto folderSet = coldpie::FolderSet::cppObject(self);
    return coldpie::cpp2py::convert(folderSet->insertionTime());
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

PyObject* _listTags(PyObject* self, PyObject*)
try {
    auto folderSet = coldpie::FolderSet::cppObject(self);
    return coldpie::cpp2py::convert(folderSet->listTags());
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

PyObject* _tagInsertionTime(PyObject* self, PyObject* args)
try {
    const char* tag = 0;
    if (not PyArg_ParseTuple(args, "s:FolderSet.tagInsertionTime", &tag)) return nullptr;

    auto folderSet = coldpie::FolderSet::cppObject(self);
    return coldpie::cpp2py::convert(folderSet->tagInsertionTime(tag));
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

PyObject* _tagDescription(PyObject* self, PyObject* args)
try {
    const char* tag = 0;
    if (not PyArg_ParseTuple(args, "s:FolderSet.tagDescription", &tag)) return 0;

    auto folderSet = coldpie::FolderSet::cppObject(self);
    return coldpie::cpp2py::convert(folderSet->tagDescription(tag));
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

PyObject* _tagLockStatus(PyObject* self, PyObject* args)
try {
    const char* tag = 0;
    if (not PyArg_ParseTuple(args, "s:FolderSet.tagLockStatus", &tag)) return 0;

    auto folderSet = coldpie::FolderSet::cppObject(self);
    return coldpie::cpp2py::convert(folderSet->tagLockStatus(tag));
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

PyObject* _listFolders(PyObject* self, PyObject* args)
try {
    int ascending = true;
    if (not PyArg_ParseTuple(args, "|i:FolderSet.listFolders()", &ascending)) {
        return 0;
    }

    auto folderSet = coldpie::FolderSet::cppObject(self);
    std::vector<std::string> folders;
    {
        // release GIL for possibly long I/O operation
        coldpie::gil gil;
        folders = folderSet->listFolders(bool(ascending));
    }
    return coldpie::cpp2py::convert(folders);
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

PyObject* _listFolderSets(PyObject* self, PyObject* args)
try {
    int ascending = true;
    if (not PyArg_ParseTuple(args, "|i:FolderSet.listFolderSets()", &ascending)) {
        return 0;
    }

    auto folderSet = coldpie::FolderSet::cppObject(self);
    std::vector<std::string> folders;
    {
        // release GIL for possibly long I/O operation
        coldpie::gil gil;
        folders = folderSet->listFolderSets(bool(ascending));
    }
    return coldpie::cpp2py::convert(folders);
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

PyObject* _folderSetAttributes(PyObject* self, PyObject*)
try {
    auto folderSet = coldpie::FolderSet::cppObject(self);
    return coldpie::Record_const::PyObject_FromCpp(folderSet->folderSetAttributes(), self);
} catch (const std::exception& exc) {
    coldpie::translateException(exc);
    return nullptr;
}

}
