
#include "daq_tokens/acquire.h"
#include "daq_tokens/internal/cern_sso.h"
#include "daq_tokens/internal/json_traits.h"
#include <nlohmann/json.hpp>
#include <jwt-cpp/jwt.h>

#include <boost/process.hpp>
#include <boost/algorithm/string.hpp>
#include <chrono>
#include <stdexcept>
#include <mutex>

#include <sys/socket.h>
#include <sys/un.h>

namespace daq {
  namespace tokens {

    namespace {

      std::mutex  mutex;
      
      std::string last_token;
      std::chrono::system_clock::time_point last_token_time;

      nlohmann::json last_sso_token;
      std::chrono::system_clock::time_point last_sso_token_time;
    }

    //
    // Acquire a token by calling the external script.
    // This is more flexible than re-implementing everything
    // in C++, and timing is not critical.
    //
    std::string acquire(Mode mode)
    {
      try {
        // Check if we can re-use the last otken
        if(mode == Mode::Reuse) {
          std::scoped_lock lock(mutex);
            if(!last_token.empty() && (std::chrono::system_clock::now() < last_token_time + std::chrono::seconds(600))) {
              return last_token;
            }
        }

	std::vector<std::string> methods{"env", "local", "kerberos"};

	if(const char *method_var = getenv("TDAQ_TOKEN_ACQUIRE")) {
	  methods.clear();
	  boost::algorithm::split(methods, method_var,  boost::algorithm::is_space());
	}

	std::string_view error_text{"Could not acquire token for unknown reasons."};

	for(auto& method : methods) {

	  if(method == "env") {
	    std::string token = get_sso_token_from_environment();
	    if(token.empty()) {
	      continue;
	    }

	    try {
	      auto decoded = jwt::decode<nlohmann_traits>(token);
	      if (decoded.get_expires_at() - std::chrono::seconds(600) < std::chrono::system_clock::now()) {
		continue;
	      }
	    } catch(...) {
	      continue;
	    }

	    last_token = token;
	    last_token_time = std::chrono::system_clock::now();

	    return token;
	  }

	  if(method == "local") {

	    // Check if we should use local Unix socket
	    if(const char *path = getenv("TDAQ_TOKEN_PATH")) {

	      // Note: if variable is set and we fail now, we don't try
	      // other methods !

	      struct sockaddr_un addr { AF_UNIX };
	      strncpy(addr.sun_path, path, sizeof(addr.sun_path) - 1);

	      int fd = socket(PF_LOCAL, SOCK_STREAM, 0);
	      if(fd < 0) {
		throw std::runtime_error("Cannot create token socket");
	      }

	      if (connect(fd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		close(fd);
                continue;
		throw std::runtime_error("Cannot connect to token socket");
	      }

	      char buffer[32000];
	      int n = read(fd, buffer, sizeof(buffer) - 1);
	      if(n < 0) {
		close(fd);
		throw std::runtime_error("Cannot read from token socket");
	      }
	      buffer[n] = '\0';
	      close(fd);

	      if(buffer[0] == '\0') {
		throw std::runtime_error("Received no token from socket");
	      }

	      if(mode == Mode::Reuse && n > 0) {
		std::scoped_lock lock(mutex);
		last_token = buffer;
		last_token_time = std::chrono::system_clock::now();
	      }
	      return buffer;
	    }
	  }

	  // SSO based token methods

	  std::string raw_result;

	  // if we have a refresh token, try to use that.
	  if(!last_sso_token.is_null() &&
	     (last_sso_token_time + std::chrono::seconds(last_sso_token["refresh_expires_in"]) > std::chrono::system_clock::now())) {
	    raw_result = get_sso_token_from_refresh_token("atlas-tdaq-token", last_sso_token["refresh_token"]);
	  }

	  if(raw_result.empty()) {

	    if(method == "kerberos") {
	      raw_result = get_sso_token_from_kerberos("atlas-tdaq-token", "ch.cern.atlas.tdaq:/redirect");
	    }

	    if(method == "browser") {
	      raw_result = get_sso_token_from_browser("atlas-tdaq-token");
	    }

	    if(method == "password") {
	      // TODO
	    }
	  }

	  if(raw_result.empty()) {
	    // try next method
	    continue;
	  }

	  // Here we  got a new token
	  last_sso_token  = nlohmann::json::parse(raw_result);
	  last_token_time = std::chrono::system_clock::now();

	  if(mode == Mode::Reuse) {
	    std::scoped_lock lock(mutex);
	    last_token = last_sso_token["access_token"];
	    last_token_time = std::chrono::system_clock::now();
	  }
	  return last_sso_token["access_token"];
	}

	// out of methods to try
	throw std::runtime_error("Could not acquire token");

      } catch (std::exception& ex) {
	throw CannotAcquireToken(ERS_HERE, ex);
      } catch(...) {
	// we don't expect any non-std derived exceptions
	throw;
      }
    }
  }
}

