#include "dcm/DummyDataCollector.h"

#include "dcm/Event.h"
#include "dcm/Main.h"
#include "dcm/Processor.h"
#include "dcm/dal/DcmDummyDataCollector.h"

#include "eformat/SourceIdentifier.h"
#include "eformat/Status.h"
#include "eformat/write/ROBFragment.h"

#include <memory>

namespace daq
{
namespace dcm
{

DummyDataCollector::DummyDataCollector(Main& main, const dal::DcmDummyDataCollector& config) :
    m_main(main),
    m_config(config),
    m_processor(nullptr),
    m_delayDistribution_us(config.get_delayDistribution_us()),
    m_sizeDistribution_words(config.get_robSizeDistribution_words())
{
  for (auto& rosNameRobIds : s_sourceToRobIdsVec) {
    m_sources.emplace_back(rosNameRobIds.first);
    m_roses.emplace_back(rosNameRobIds.first, m_main.ioService());
    for (auto robId : rosNameRobIds.second) {
      m_robIds.push_back(robId);
      m_robIdToSource.insert({robId, rosNameRobIds.first});
      m_sourceToRobIds.insert({rosNameRobIds.first, robId});
      m_robIdToRos[robId] = &m_roses.back();
    }
  }
  std::sort(m_robIds.begin(), m_robIds.end());
}

void DummyDataCollector::initialize(Processor* processor)
{
  m_processor = processor;
}

void DummyDataCollector::disableSource(const std::string& /* source */)
{
  ers::error(dcm::GenericIssue(ERS_HERE,
      "Fragment sources cannot be enabled/disabled when using the DummyDataCollector"));
}

void DummyDataCollector::enableSource(const std::string& source)
{
  disableSource(source);
}

void DummyDataCollector::setDynamicParameters(const std::map<std::string, std::string>& parameters)
{
  for (const auto& kv : parameters)
  {
    ers::warning(dcm::ConfigUnknownParameter(ERS_HERE, kv.first));
  }
}

const std::vector<std::string>& DummyDataCollector::sources()
{
  return m_sources;
}

const std::vector<std::uint32_t>& DummyDataCollector::robIds()
{
  return m_robIds;
}

const std::map<std::uint32_t, std::string>& DummyDataCollector::robIdToSource()
{
  return m_robIdToSource;
}

const std::multimap<std::string, std::uint32_t>& DummyDataCollector::sourceToRobIds()
{
  return m_sourceToRobIds;
}

void DummyDataCollector::asyncConnect()
{
  m_main.ioService().post(std::bind(&Main::onDataCollectorConnect, &m_main));
}

void DummyDataCollector::asyncStart()
{
  m_main.ioService().post(std::bind(&Main::onDataCollectorStart, &m_main));
}

void DummyDataCollector::asyncStop()
{
  m_main.ioService().post(std::bind(&Main::onDataCollectorStop, &m_main));
}

void DummyDataCollector::asyncDisconnect()
{
  for (auto& ros : m_roses) {
    ros.timer.cancel();
  }
  m_main.ioService().post(std::bind(&Main::onDataCollectorDisconnect, &m_main));
}

void DummyDataCollector::asyncFetchRobs(Event* event, const std::vector<std::uint32_t>& robIds,
    void* context)
{
  std::unordered_map<Ros*, std::vector<std::uint32_t> > requests;
  for (auto robId : robIds) {
    auto it = m_robIdToRos.find(robId);
    // Check for non-existent ROB IDs
    if (it == m_robIdToRos.end()) {
      dcm::DataCollectingBadRobId issue(ERS_HERE, event->globalId, event->l1Id, robId);
      ers::error(issue);
      m_main.ioService().post(std::bind(&Processor::onFetchRobsError, m_processor,
          issue, event, context));
      return;
    }
    if (!event->block[robId]) {
      // The ROB wasn't already collected. Queue it for fetching.
      requests[it->second].push_back(robId);
    }
  }

  if (requests.empty()) {
    // All ROBs were already collected. Notify the caller.
    m_main.ioService().post(
        std::bind(&Processor::onFetchRobs, m_processor, event, context));
    return;
  }

  // We have a legitimate request. Do it.
  m_dcRequests.emplace_back(event, context);
  auto dcRequest = --m_dcRequests.end();
  for (auto& request : requests) {
    ++dcRequest->nPendingRosRequests;
    ++m_main.isInfo()->DcPendingRequests;
    request.first->requests.emplace_back(dcRequest, std::move(request.second));
    if (request.first->requests.size() == 1) {
      // The only request in the queue is the one we just added: start the timer
      startTimer(*request.first);
    }
  }
}

// Implementation of DataCollector::mayGetRobs()
void DummyDataCollector::mayGetRobs(Event* /*event*/, const std::vector<std::uint32_t>& /*robIds*/)
{
  // Do nothing
}

void DummyDataCollector::clearEvent(Event* event)
{
  // Do nothing
}


void DummyDataCollector::startTimer(Ros& ros) {

  ros.timer.expires_from_now(std::chrono::microseconds(m_delayDistribution_us.next_value()));

  ros.timer.async_wait([this, &ros] (const boost::system::error_code& error) {

    namespace efh = eformat::helper;
    namespace efw = eformat::write;

    auto& rosRequest = ros.requests.front();
    auto& event = *rosRequest.dcRequest->event;

    // TODO: better error handling
    if (error) {
      // TODO: manufacture empty robs, with the appropriate status word
    }

    // Manufacture the requested ROBs

    const std::uint32_t bunchCrossingId = 0x1;
    // L1 Trigger type is an 8-bit field. If the most significant bit (bit 8) is set, the event
    // was selected for physics, and the bits 7 to 0 are a bitmask detailing the trigger type. If
    // the MSB is not set, the event was selected for calibration, and bits 7 to 3 are the integer
    // detector ID.
    // Here we manufacture a physics L1 trigger type, with one detail bit set.
    const std::uint32_t l1TriggerType = 0x81;
    // The event type is a sub-detector specific 32-bit field. We leave it empty.
    const std::uint32_t eventType = 0x0;
    // For each ROB, we choose a payload size from the configured distribution and use
    // uninitialized memory as payload.
    std::vector<std::uint32_t> sizes_words(rosRequest.robIds.size(), 0);
    std::generate(sizes_words.begin(), sizes_words.end(),
        std::bind(&DC::Distribution::next_value, &m_sizeDistribution_words));
    std::unique_ptr<std::uint32_t[]> data(
        new std::uint32_t[*std::max_element(sizes_words.begin(), sizes_words.end()) * 4]);
    std::vector<efw::ROBFragment> robs;
    robs.reserve(rosRequest.robIds.size());
    for (std::uint32_t i = 0; i < rosRequest.robIds.size(); ++i) {
      robs.emplace_back(rosRequest.robIds[i], m_main.runParams().run_number, event.l1Id,
          bunchCrossingId, l1TriggerType, eventType, sizes_words[i], data.get(),
          eformat::STATUS_FRONT);
    }

    // Serialize to SBA

    // TODO: catch SBA exceptions
    std::uint32_t robsSize = 0;
    for (auto& rob: robs) {
      robsSize += rob.size_word() * 4;
    }
    auto buffer = event.block.startAppendRobData(robsSize);
    for (auto& rob: robs) {
      auto node = rob.bind();
      do {
        std::memcpy(buffer, node->base, node->size_word * 4);
        buffer += node->size_word;
        node = node->next;
      } while (node);
    }
    event.block.endAppendRobData(robsSize);

    auto dcRequest = rosRequest.dcRequest;

    ros.requests.pop_front();
    if (!ros.requests.empty()) {
      startTimer(ros);
    }

    --dcRequest->nPendingRosRequests;
    if (dcRequest->nPendingRosRequests == 0) {
      // All ROSes have "replied"
      m_processor->onFetchRobs(dcRequest->event, dcRequest->context);
      m_dcRequests.erase(dcRequest);
    }

  });
}

const std::map<std::string, std::vector<std::uint32_t>> DummyDataCollector::s_sourceToRobIdsVec {
    { "ROS-PIX-B-00",        { 0x111708, 0x111710, 0x111712, 0x111715, 0x111810, 0x111812, 0x112406, 0x112515, 0x112517, 0x112519, 0x112521 } },
    { "ROS-PIX-B-01",        { 0x111707, 0x111709, 0x111711, 0x111714, 0x111809, 0x111811, 0x112405, 0x112514, 0x112516, 0x112518, 0x112520 } },
    { "ROS-PIX-B-02",        { 0x111706, 0x111719, 0x111721, 0x111806, 0x111808, 0x111819, 0x111821, 0x112419, 0x112421, 0x112512 } },
    { "ROS-PIX-B-03",        { 0x111705, 0x111718, 0x111720, 0x111805, 0x111807, 0x111818, 0x111820, 0x112418, 0x112420, 0x112511 } },
    { "ROS-PIX-B-04",        { 0x111717, 0x111815, 0x111817, 0x112408, 0x112410, 0x112412, 0x112415, 0x112417, 0x112505, 0x112507, 0x112509 } },
    { "ROS-PIX-B-05",        { 0x111716, 0x111814, 0x111816, 0x112407, 0x112409, 0x112411, 0x112414, 0x112416, 0x112506, 0x112508, 0x112510 } },
    { "ROS-PIX-DISK-00",     { 0x120205, 0x120207, 0x120209, 0x120211, 0x120214, 0x120216, 0x120218, 0x120220, 0x121609, 0x121611, 0x121614, 0x121616 } },
    { "ROS-PIX-DISK-01",     { 0x120206, 0x120208, 0x120210, 0x120212, 0x120215, 0x120217, 0x120219, 0x120221, 0x121610, 0x121612, 0x121615, 0x121617 } },
    { "ROS-PIX-BL-00",       { 0x130007, 0x130009, 0x130010, 0x130011, 0x130012, 0x130014, 0x130015, 0x130016, 0x130017, 0x130018, 0x130019 } },
    { "ROS-PIX-BL-01",       { 0x130107, 0x130108, 0x130109, 0x130110, 0x130111, 0x130112, 0x130312, 0x130314, 0x130315, 0x130316, 0x130317 } },
    { "ROS-PIX-BL-02",       { 0x130005, 0x130006, 0x130008, 0x130114, 0x130115, 0x130116, 0x130117, 0x130118, 0x130119, 0x130120, 0x130121 } },
    { "ROS-PIX-BL-03",       { 0x130020, 0x130021, 0x130105, 0x130106, 0x130307, 0x130308, 0x130309, 0x130310, 0x130311, 0x130318, 0x130319 } },
    { "ROS-TRT-BA-00",       { 0x310100, 0x310200, 0x310300, 0x310400, 0x310500, 0x310600, 0x310700, 0x310800, 0x310900, 0x310A00, 0x310B00, 0x310C00 } },
    { "ROS-TRT-BA-01",       { 0x310D00, 0x310E00, 0x310F00, 0x311000, 0x311100, 0x311200, 0x311300, 0x311400, 0x311500, 0x311600, 0x311700, 0x311800 } },
    { "ROS-TRT-BA-02",       { 0x311900, 0x311A00, 0x311B00, 0x311C00, 0x311D00, 0x311E00, 0x311F00, 0x312000 } },
    { "ROS-TRT-BC-00",       { 0x320100, 0x320200, 0x320300, 0x320400, 0x320500, 0x320600, 0x320700, 0x320800, 0x320900, 0x320A00, 0x320B00, 0x320C00 } },
    { "ROS-TRT-BC-01",       { 0x320D00, 0x320E00, 0x320F00, 0x321000, 0x321100, 0x321200, 0x321300, 0x321400, 0x321500, 0x321600, 0x321700, 0x321800 } },
    { "ROS-TRT-BC-02",       { 0x321900, 0x321A00, 0x321B00, 0x321C00, 0x321D00, 0x321E00, 0x321F00, 0x322000 } },
    { "ROS-TRT-ECA-00",      { 0x330101, 0x330102, 0x330201, 0x330202, 0x330301, 0x330302, 0x330401, 0x330402, 0x330501, 0x330502 } },
    { "ROS-TRT-ECA-01",      { 0x330601, 0x330602, 0x330701, 0x330702, 0x330801, 0x330802, 0x330901, 0x330902, 0x330A01, 0x330A02 } },
    { "ROS-TRT-ECA-02",      { 0x330B01, 0x330B02, 0x330C01, 0x330C02, 0x330D01, 0x330D02, 0x330E01, 0x330E02, 0x330F01, 0x330F02 } },
    { "ROS-TRT-ECA-03",      { 0x331001, 0x331002, 0x331101, 0x331102, 0x331201, 0x331202, 0x331301, 0x331302, 0x331401, 0x331402 } },
    { "ROS-TRT-ECA-04",      { 0x331501, 0x331502, 0x331601, 0x331602, 0x331701, 0x331702, 0x331801, 0x331802, 0x331901, 0x331902, 0x331A01, 0x331A02 } },
    { "ROS-TRT-ECA-05",      { 0x331B01, 0x331B02, 0x331C01, 0x331C02, 0x331D01, 0x331D02, 0x331E01, 0x331E02, 0x331F01, 0x331F02, 0x332001, 0x332002 } },
    { "ROS-TRT-ECC-00",      { 0x340101, 0x340102, 0x340201, 0x340202, 0x340301, 0x340302, 0x340401, 0x340402, 0x340501, 0x340502 } },
    { "ROS-TRT-ECC-01",      { 0x340601, 0x340602, 0x340701, 0x340702, 0x340801, 0x340802, 0x340901, 0x340902, 0x340A01, 0x340A02 } },
    { "ROS-TRT-ECC-02",      { 0x340B01, 0x340B02, 0x340C01, 0x340C02, 0x340D01, 0x340D02, 0x340E01, 0x340E02, 0x340F01, 0x340F02 } },
    { "ROS-TRT-ECC-03",      { 0x341001, 0x341002, 0x341101, 0x341102, 0x341201, 0x341202, 0x341301, 0x341302, 0x341401, 0x341402 } },
    { "ROS-TRT-ECC-04",      { 0x341501, 0x341502, 0x341601, 0x341602, 0x341701, 0x341702, 0x341801, 0x341802, 0x341901, 0x341902, 0x341A01, 0x341A02 } },
    { "ROS-TRT-ECC-05",      { 0x341B01, 0x341B02, 0x341C01, 0x341C02, 0x341D01, 0x341D02, 0x341E01, 0x341E02, 0x341F01, 0x341F02, 0x342001, 0x342002 } },
    { "ROS-SCT-BA-00",       { 0x210000, 0x210001, 0x210002, 0x210003, 0x210004, 0x210005, 0x210006, 0x210007, 0x210008, 0x210009, 0x21000A } },
    { "ROS-SCT-BA-01",       { 0x210100, 0x210101, 0x210102, 0x210103, 0x210104, 0x210105, 0x210106, 0x210107, 0x210108, 0x210109, 0x21010A } },
    { "ROS-SCT-BC-00",       { 0x220000, 0x220001, 0x220002, 0x220003, 0x220004, 0x220005, 0x220006, 0x220007, 0x220008, 0x220009, 0x22000A } },
    { "ROS-SCT-BC-01",       { 0x220100, 0x220101, 0x220102, 0x220103, 0x220104, 0x220105, 0x220106, 0x220107, 0x220108, 0x220109, 0x22010A } },
    { "ROS-SCT-ECA-00",      { 0x230000, 0x230001, 0x230002, 0x230003, 0x230004, 0x230005, 0x230006, 0x230007, 0x230008, 0x230009, 0x23000A } },
    { "ROS-SCT-ECA-01",      { 0x230100, 0x230101, 0x230102, 0x230103, 0x230104, 0x230105, 0x230106, 0x230107, 0x230108, 0x230109, 0x23010A, 0x23010B } },
    { "ROS-SCT-ECC-00",      { 0x240000, 0x240001, 0x240002, 0x240003, 0x240004, 0x240005, 0x240006, 0x240007, 0x240008, 0x240009, 0x24000A, 0x24000B } },
    { "ROS-SCT-ECC-01",      { 0x240100, 0x240101, 0x240102, 0x240103, 0x240104, 0x240105, 0x240106, 0x240107, 0x240108, 0x240109, 0x24010A } },
    { "ROS-LAR-EMBA-00",     { 0x410000, 0x410001, 0x410002, 0x410003, 0x410004, 0x410005, 0x410006, 0x410007, 0x410008, 0x410009, 0x41000A, 0x41000B } },
    { "ROS-LAR-EMBA-01",     { 0x41000C, 0x41000D, 0x41000E, 0x41000F, 0x410010, 0x410011, 0x410012, 0x410013, 0x410014, 0x410015, 0x410016, 0x410017 } },
    { "ROS-LAR-EMBA-02",     { 0x410018, 0x410019, 0x41001A, 0x41001B, 0x41001C, 0x41001D, 0x41001E, 0x41001F, 0x410020, 0x410021, 0x410022, 0x410023 } },
    { "ROS-LAR-EMBA-03",     { 0x410024, 0x410025, 0x410026, 0x410027, 0x410028, 0x410029, 0x41002A, 0x41002B } },
    { "ROS-LAR-EMBA-04",     { 0x41002C, 0x41002D, 0x41002E, 0x41002F, 0x410030, 0x410031, 0x410032, 0x410033, 0x410034, 0x410035, 0x410036, 0x410037 } },
    { "ROS-LAR-EMBA-05",     { 0x410038, 0x410039, 0x41003A, 0x41003B, 0x41003C, 0x41003D, 0x41003E, 0x41003F, 0x410040, 0x410041, 0x410042, 0x410043 } },
    { "ROS-LAR-EMBA-06",     { 0x410044, 0x410045, 0x410046, 0x410047, 0x410048, 0x410049, 0x41004A, 0x41004B, 0x41004C, 0x41004D, 0x41004E, 0x41004F } },
    { "ROS-LAR-EMBA-07",     { 0x410050, 0x410051, 0x410052, 0x410053, 0x410054, 0x410055, 0x410056, 0x410057, 0x410058, 0x410059, 0x41005A, 0x41005B } },
    { "ROS-LAR-EMBA-08",     { 0x41005C, 0x41005D, 0x41005E, 0x41005F, 0x410060, 0x410061, 0x410062, 0x410063 } },
    { "ROS-LAR-EMBA-09",     { 0x410064, 0x410065, 0x410066, 0x410067, 0x410068, 0x410069, 0x41006A, 0x41006B, 0x41006C, 0x41006D, 0x41006E, 0x41006F } },
    { "ROS-LAR-EMBA-10",     { 0x410070, 0x410071, 0x410072, 0x410073, 0x410074, 0x410075, 0x410076, 0x410077, 0x410078, 0x410079, 0x41007A, 0x41007B } },
    { "ROS-LAR-EMBA-11",     { 0x41007C, 0x41007D, 0x41007E, 0x41007F, 0x410080, 0x410081, 0x410082, 0x410083, 0x410084, 0x410085, 0x410086, 0x410087 } },
    { "ROS-LAR-EMBA-12",     { 0x410088, 0x410089, 0x41008A, 0x41008B, 0x41008C, 0x41008D, 0x41008E, 0x41008F, 0x410090, 0x410091, 0x410092, 0x410093 } },
    { "ROS-LAR-EMBA-13",     { 0x410094, 0x410095, 0x410096, 0x410097, 0x410098, 0x410099, 0x41009A, 0x41009B } },
    { "ROS-LAR-EMBA-14",     { 0x41009C, 0x41009D, 0x41009E, 0x41009F, 0x4100A0, 0x4100A1, 0x4100A2, 0x4100A3, 0x4100A4, 0x4100A5, 0x4100A6, 0x4100A7 } },
    { "ROS-LAR-EMBA-15",     { 0x4100A8, 0x4100A9, 0x4100AA, 0x4100AB, 0x4100AC, 0x4100AD, 0x4100AE, 0x4100AF, 0x4100B0, 0x4100B1, 0x4100B2, 0x4100B3 } },
    { "ROS-LAR-EMBA-16",     { 0x4100B4, 0x4100B5, 0x4100B6, 0x4100B7, 0x4100B8, 0x4100B9, 0x4100BA, 0x4100BB, 0x4100BC, 0x4100BD, 0x4100BE, 0x4100BF } },
    { "ROS-LAR-EMBA-17",     { 0x4100C0, 0x4100C1, 0x4100C2, 0x4100C3, 0x4100C4, 0x4100C5, 0x4100C6, 0x4100C7, 0x4100C8, 0x4100C9, 0x4100CA, 0x4100CB } },
    { "ROS-LAR-EMBA-18",     { 0x4100CC, 0x4100CD, 0x4100CE, 0x4100CF, 0x4100D0, 0x4100D1, 0x4100D2, 0x4100D3 } },
    { "ROS-LAR-EMBA-19",     { 0x4100D4, 0x4100D5, 0x4100D6, 0x4100D7, 0x4100D8, 0x4100D9, 0x4100DA, 0x4100DB, 0x4100DC, 0x4100DD, 0x4100DE, 0x4100DF } },
    { "ROS-LAR-EMBC-00",     { 0x420000, 0x420001, 0x420002, 0x420003, 0x420004, 0x420005, 0x420006, 0x420007, 0x420008, 0x420009, 0x42000A, 0x42000B } },
    { "ROS-LAR-EMBC-01",     { 0x42000C, 0x42000D, 0x42000E, 0x42000F, 0x420010, 0x420011, 0x420012, 0x420013, 0x420014, 0x420015, 0x420016, 0x420017 } },
    { "ROS-LAR-EMBC-02",     { 0x420018, 0x420019, 0x42001A, 0x42001B, 0x42001C, 0x42001D, 0x42001E, 0x42001F, 0x420020, 0x420021, 0x420022, 0x420023 } },
    { "ROS-LAR-EMBC-03",     { 0x420024, 0x420025, 0x420026, 0x420027, 0x420028, 0x420029, 0x42002A, 0x42002B } },
    { "ROS-LAR-EMBC-04",     { 0x42002C, 0x42002D, 0x42002E, 0x42002F, 0x420030, 0x420031, 0x420032, 0x420033, 0x420034, 0x420035, 0x420036, 0x420037 } },
    { "ROS-LAR-EMBC-05",     { 0x420038, 0x420039, 0x42003A, 0x42003B, 0x42003C, 0x42003D, 0x42003E, 0x42003F, 0x420040, 0x420041, 0x420042, 0x420043 } },
    { "ROS-LAR-EMBC-06",     { 0x420044, 0x420045, 0x420046, 0x420047, 0x420048, 0x420049, 0x42004A, 0x42004B, 0x42004C, 0x42004D, 0x42004E, 0x42004F } },
    { "ROS-LAR-EMBC-07",     { 0x420050, 0x420051, 0x420052, 0x420053, 0x420054, 0x420055, 0x420056, 0x420057, 0x420058, 0x420059, 0x42005A, 0x42005B } },
    { "ROS-LAR-EMBC-08",     { 0x42005C, 0x42005D, 0x42005E, 0x42005F, 0x420060, 0x420061, 0x420062, 0x420063 } },
    { "ROS-LAR-EMBC-09",     { 0x420064, 0x420065, 0x420066, 0x420067, 0x420068, 0x420069, 0x42006A, 0x42006B, 0x42006C, 0x42006D, 0x42006E, 0x42006F } },
    { "ROS-LAR-EMBC-10",     { 0x420070, 0x420071, 0x420072, 0x420073, 0x420074, 0x420075, 0x420076, 0x420077, 0x420078, 0x420079, 0x42007A, 0x42007B } },
    { "ROS-LAR-EMBC-11",     { 0x42007C, 0x42007D, 0x42007E, 0x42007F, 0x420080, 0x420081, 0x420082, 0x420083, 0x420084, 0x420085, 0x420086, 0x420087 } },
    { "ROS-LAR-EMBC-12",     { 0x420088, 0x420089, 0x42008A, 0x42008B, 0x42008C, 0x42008D, 0x42008E, 0x42008F, 0x420090, 0x420091, 0x420092, 0x420093 } },
    { "ROS-LAR-EMBC-13",     { 0x420094, 0x420095, 0x420096, 0x420097, 0x420098, 0x420099, 0x42009A, 0x42009B } },
    { "ROS-LAR-EMBC-14",     { 0x42009C, 0x42009D, 0x42009E, 0x42009F, 0x4200A0, 0x4200A1, 0x4200A2, 0x4200A3, 0x4200A4, 0x4200A5, 0x4200A6, 0x4200A7 } },
    { "ROS-LAR-EMBC-15",     { 0x4200A8, 0x4200A9, 0x4200AA, 0x4200AB, 0x4200AC, 0x4200AD, 0x4200AE, 0x4200AF, 0x4200B0, 0x4200B1, 0x4200B2, 0x4200B3 } },
    { "ROS-LAR-EMBC-16",     { 0x4200B4, 0x4200B5, 0x4200B6, 0x4200B7, 0x4200B8, 0x4200B9, 0x4200BA, 0x4200BB, 0x4200BC, 0x4200BD, 0x4200BE, 0x4200BF } },
    { "ROS-LAR-EMBC-17",     { 0x4200C0, 0x4200C1, 0x4200C2, 0x4200C3, 0x4200C4, 0x4200C5, 0x4200C6, 0x4200C7, 0x4200C8, 0x4200C9, 0x4200CA, 0x4200CB } },
    { "ROS-LAR-EMBC-18",     { 0x4200CC, 0x4200CD, 0x4200CE, 0x4200CF, 0x4200D0, 0x4200D1, 0x4200D2, 0x4200D3 } },
    { "ROS-LAR-EMBC-19",     { 0x4200D4, 0x4200D5, 0x4200D6, 0x4200D7, 0x4200D8, 0x4200D9, 0x4200DA, 0x4200DB, 0x4200DC, 0x4200DD, 0x4200DE, 0x4200DF } },
    { "ROS-LAR-EMECA-00",    { 0x430000, 0x430001, 0x430002, 0x430003, 0x430004, 0x430005, 0x430006, 0x430007, 0x430008, 0x430009, 0x43000A, 0x43000B } },
    { "ROS-LAR-EMECA-01",    { 0x43000C, 0x43000D, 0x43000E, 0x43000F, 0x430010, 0x430011, 0x430012, 0x430013, 0x430014, 0x430015, 0x430016, 0x430017 } },
    { "ROS-LAR-EMECA-02",    { 0x430018, 0x430019, 0x43001A, 0x43001B, 0x43001C, 0x43001D, 0x43001E, 0x43001F, 0x430020, 0x430021, 0x430022, 0x430023 } },
    { "ROS-LAR-EMECA-03",    { 0x430024, 0x430025, 0x430026, 0x430027, 0x430028, 0x430029, 0x43002A, 0x43002B, 0x43002C, 0x43002D, 0x43002E, 0x43002F } },
    { "ROS-LAR-EMECA-04",    { 0x430030, 0x430031, 0x430032, 0x430033, 0x430034, 0x430035, 0x430036, 0x430037 } },
    { "ROS-LAR-EMECA-05",    { 0x430038, 0x430039, 0x43003A, 0x43003B, 0x43003C, 0x43003D, 0x43003E, 0x43003F, 0x430040, 0x430041, 0x430042, 0x430043 } },
    { "ROS-LAR-EMECA-06",    { 0x430044, 0x430045, 0x430046, 0x430047, 0x430048, 0x430049, 0x43004A, 0x43004B, 0x43004C, 0x43004D, 0x43004E, 0x43004F } },
    { "ROS-LAR-EMECA-07",    { 0x430050, 0x430051, 0x430052, 0x430053, 0x430054, 0x430055, 0x430056, 0x430057, 0x430058, 0x430059, 0x43005A, 0x43005B } },
    { "ROS-LAR-EMECA-08",    { 0x43005C, 0x43005D, 0x43005E, 0x43005F, 0x430060, 0x430061, 0x430062, 0x430063, 0x430064, 0x430065, 0x430066, 0x430067 } },
    { "ROS-LAR-EMECA-09",    { 0x430068, 0x430069, 0x43006A, 0x43006B, 0x43006C, 0x43006D, 0x43006E, 0x43006F, 0x430070, 0x430071 } },
    { "ROS-LAR-EMECA-10",    { 0x430072, 0x430073, 0x430074, 0x430075, 0x430076, 0x430077, 0x430078, 0x430079, 0x43007A, 0x43007B, 0x43007C, 0x43007D } },
    { "ROS-LAR-EMECA-11",    { 0x43007E, 0x43007F, 0x430080, 0x430081, 0x430082, 0x430083, 0x430084, 0x430085, 0x430086, 0x430087, 0x430088, 0x430089 } },
    { "ROS-LAR-EMECC-00",    { 0x440000, 0x440001, 0x440002, 0x440003, 0x440004, 0x440005, 0x440006, 0x440007, 0x440008, 0x440009, 0x44000A, 0x44000B } },
    { "ROS-LAR-EMECC-01",    { 0x44000C, 0x44000D, 0x44000E, 0x44000F, 0x440010, 0x440011, 0x440012, 0x440013, 0x440014, 0x440015, 0x440016, 0x440017 } },
    { "ROS-LAR-EMECC-02",    { 0x440018, 0x440019, 0x44001A, 0x44001B, 0x44001C, 0x44001D, 0x44001E, 0x44001F, 0x440020, 0x440021, 0x440022, 0x440023 } },
    { "ROS-LAR-EMECC-03",    { 0x440024, 0x440025, 0x440026, 0x440027, 0x440028, 0x440029, 0x44002A, 0x44002B, 0x44002C, 0x44002D, 0x44002E, 0x44002F } },
    { "ROS-LAR-EMECC-04",    { 0x440030, 0x440031, 0x440032, 0x440033, 0x440034, 0x440035, 0x440036, 0x440037 } },
    { "ROS-LAR-EMECC-05",    { 0x440038, 0x440039, 0x44003A, 0x44003B, 0x44003C, 0x44003D, 0x44003E, 0x44003F, 0x440040, 0x440041, 0x440042, 0x440043 } },
    { "ROS-LAR-EMECC-06",    { 0x440044, 0x440045, 0x440046, 0x440047, 0x440048, 0x440049, 0x44004A, 0x44004B, 0x44004C, 0x44004D, 0x44004E, 0x44004F } },
    { "ROS-LAR-EMECC-07",    { 0x440050, 0x440051, 0x440052, 0x440053, 0x440054, 0x440055, 0x440056, 0x440057, 0x440058, 0x440059, 0x44005A, 0x44005B } },
    { "ROS-LAR-EMECC-08",    { 0x44005C, 0x44005D, 0x44005E, 0x44005F, 0x440060, 0x440061, 0x440062, 0x440063, 0x440064, 0x440065, 0x440066, 0x440067 } },
    { "ROS-LAR-EMECC-09",    { 0x440068, 0x440069, 0x44006A, 0x44006B, 0x44006C, 0x44006D, 0x44006E, 0x44006F, 0x440070, 0x440071 } },
    { "ROS-LAR-EMECC-10",    { 0x440072, 0x440073, 0x440074, 0x440075, 0x440076, 0x440077, 0x440078, 0x440079, 0x44007A, 0x44007B, 0x44007C, 0x44007D } },
    { "ROS-LAR-EMECC-11",    { 0x44007E, 0x44007F, 0x440080, 0x440081, 0x440082, 0x440083, 0x440084, 0x440085, 0x440086, 0x440087, 0x440088, 0x440089 } },
    { "ROS-LAR-HECA-00",     { 0x450000, 0x450001, 0x450002, 0x450003, 0x450004, 0x450005, 0x450006, 0x450007, 0x450008, 0x450009, 0x45000A, 0x45000B } },
    { "ROS-LAR-HECC-00",     { 0x460000, 0x460001, 0x460002, 0x460003, 0x460004, 0x460005, 0x460006, 0x460007, 0x460008, 0x460009, 0x46000A, 0x46000B } },
    { "ROS-LAR-FCALA-00",    { 0x470000, 0x470001, 0x470002, 0x470003, 0x470004, 0x470005, 0x470006 } },
    { "ROS-LAR-FCALC-00",    { 0x480000, 0x480001, 0x480002, 0x480003, 0x480004, 0x480005, 0x480006 } },
    { "ROS-TIL-LBA-00",      { 0x510000, 0x510001, 0x510002, 0x510003, 0x510004, 0x510005, 0x510006, 0x510007 } },
    { "ROS-TIL-LBA-01",      { 0x510008, 0x510009, 0x51000A, 0x51000B, 0x51000C, 0x51000D, 0x51000E, 0x51000F } },
    { "ROS-TIL-LBC-00",      { 0x520000, 0x520001, 0x520002, 0x520003, 0x520004, 0x520005, 0x520006, 0x520007 } },
    { "ROS-TIL-LBC-01",      { 0x520008, 0x520009, 0x52000A, 0x52000B, 0x52000C, 0x52000D, 0x52000E, 0x52000F, 0x520010 } },
    { "ROS-TIL-EBA-00",      { 0x530000, 0x530001, 0x530002, 0x530003, 0x530004, 0x530005, 0x530006, 0x530007 } },
    { "ROS-TIL-EBA-01",      { 0x530008, 0x530009, 0x53000A, 0x53000B, 0x53000C, 0x53000D, 0x53000E, 0x53000F } },
    { "ROS-TIL-EBC-00",      { 0x540000, 0x540001, 0x540002, 0x540003, 0x540004, 0x540005, 0x540006, 0x540007 } },
    { "ROS-TIL-EBC-01",      { 0x540008, 0x540009, 0x54000A, 0x54000B, 0x54000C, 0x54000D, 0x54000E, 0x54000F } },
    { "ROS-MDT-BA-00",       { 0x610000, 0x610001, 0x610008, 0x610009, 0x610010, 0x610011, 0x610018, 0x610019, 0x610020, 0x610021, 0x610028, 0x610029 } },
    { "ROS-MDT-BA-01",       { 0x610002, 0x610003, 0x61000A, 0x61000B, 0x610012, 0x610013, 0x61001A, 0x61001B, 0x610022, 0x610023, 0x61002A, 0x61002B } },
    { "ROS-MDT-BA-02",       { 0x610004, 0x610005, 0x61000C, 0x61000D, 0x610014, 0x610015, 0x61001C, 0x61001D, 0x610024, 0x610025, 0x61002C, 0x61002D } },
    { "ROS-MDT-BA-03",       { 0x610006, 0x610007, 0x61000E, 0x61000F, 0x610016, 0x610017, 0x61001E, 0x61001F, 0x610026, 0x610027, 0x61002E, 0x61002F, 0x610030, 0x610031 } },
    { "ROS-MDT-BC-00",       { 0x620000, 0x620001, 0x620008, 0x620009, 0x620010, 0x620011, 0x620018, 0x620019, 0x620020, 0x620021, 0x620028, 0x620029 } },
    { "ROS-MDT-BC-01",       { 0x620002, 0x620003, 0x62000A, 0x62000B, 0x620012, 0x620013, 0x62001A, 0x62001B, 0x620022, 0x620023, 0x62002A, 0x62002B } },
    { "ROS-MDT-BC-02",       { 0x620004, 0x620005, 0x62000C, 0x62000D, 0x620014, 0x620015, 0x62001C, 0x62001D, 0x620024, 0x620025, 0x62002C, 0x62002D } },
    { "ROS-MDT-BC-03",       { 0x620006, 0x620007, 0x62000E, 0x62000F, 0x620016, 0x620017, 0x62001E, 0x62001F, 0x620026, 0x620027, 0x62002E, 0x62002F, 0x620030, 0x620031 } },
    { "ROS-MDT-ECA-00",      { 0x630000, 0x630001, 0x630008, 0x630009, 0x630010, 0x630011, 0x630018, 0x630019, 0x630020, 0x630021, 0x630028, 0x630029 } },
    { "ROS-MDT-ECA-01",      { 0x630002, 0x630003, 0x63000A, 0x63000B, 0x630012, 0x630013, 0x63001A, 0x63001B, 0x630022, 0x630023, 0x63002A, 0x63002B, 0x630030, 0x630031 } },
    { "ROS-MDT-ECA-02",      { 0x630004, 0x630005, 0x63000C, 0x63000D, 0x630014, 0x630015, 0x63001C, 0x63001D, 0x630024, 0x630025, 0x63002C, 0x63002D } },
    { "ROS-MDT-ECA-03",      { 0x630006, 0x630007, 0x63000E, 0x63000F, 0x630016, 0x630017, 0x63001E, 0x63001F, 0x630026, 0x630027, 0x63002E, 0x63002F, 0x630032, 0x630033 } },
    { "ROS-MDT-ECC-00",      { 0x640000, 0x640001, 0x640008, 0x640009, 0x640010, 0x640011, 0x640018, 0x640019, 0x640020, 0x640021, 0x640028, 0x640029 } },
    { "ROS-MDT-ECC-01",      { 0x640002, 0x640003, 0x64000A, 0x64000B, 0x640012, 0x640013, 0x64001A, 0x64001B, 0x640022, 0x640023, 0x64002A, 0x64002B, 0x640030, 0x640031 } },
    { "ROS-MDT-ECC-02",      { 0x640004, 0x640005, 0x64000C, 0x64000D, 0x640014, 0x640015, 0x64001C, 0x64001D, 0x640024, 0x640025, 0x64002C, 0x64002D } },
    { "ROS-MDT-ECC-03",      { 0x640006, 0x640007, 0x64000E, 0x64000F, 0x640016, 0x640017, 0x64001E, 0x64001F, 0x640026, 0x640027, 0x64002E, 0x64002F, 0x640032, 0x640033 } },
    { "ROS-RPC-BA-00",       { 0x650000, 0x650001, 0x650002, 0x650003, 0x650004, 0x650005, 0x650006, 0x650007 } },
    { "ROS-RPC-BA-01",       { 0x650008, 0x650009, 0x65000A, 0x65000B, 0x65000C, 0x65000D, 0x65000E, 0x65000F } },
    { "ROS-RPC-BC-00",       { 0x660000, 0x660001, 0x660002, 0x660003, 0x660004, 0x660005, 0x660006, 0x660007 } },
    { "ROS-RPC-BC-01",       { 0x660008, 0x660009, 0x66000A, 0x66000B, 0x66000C, 0x66000D, 0x66000E, 0x66000F } },
    { "ROS-TGC-ECA-00",      { 0x670001, 0x670002, 0x670003, 0x670004, 0x670005, 0x670006, 0x670007, 0x670008, 0x670009, 0x67000A, 0x67000B, 0x67000C } },
    { "ROS-TGC-ECC-00",      { 0x680001, 0x680002, 0x680003, 0x680004, 0x680005, 0x680006, 0x680007, 0x680008, 0x680009, 0x68000A, 0x68000B, 0x68000C } },
    { "ROS-CSC-ECA-00",      { 0x690005, 0x690007, 0x690009, 0x690011, 0x690015, 0x690017, 0x690019, 0x690021 } },
    { "ROS-CSC-ECC-00",      { 0x6A0005, 0x6A0007, 0x6A0009, 0x6A0011, 0x6A0015, 0x6A0017, 0x6A0019, 0x6A0021 } },
    { "ROS-TDQ-CALPP-00",    { 0x710000, 0x710001, 0x710002, 0x710010, 0x710011, 0x710012, 0x710020, 0x710021, 0x710022, 0x710030, 0x710031, 0x710032 } },
    { "ROS-TDQ-CALPP-01",    { 0x710003, 0x710004, 0x710005, 0x710013, 0x710014, 0x710015, 0x710023, 0x710024, 0x710025, 0x710033, 0x710034, 0x710035 } },
    { "ROS-TDQ-CALPP-02",    { 0x710006, 0x710007, 0x710016, 0x710017, 0x710026, 0x710027, 0x710036, 0x710037 } },
    { "ROS-TDQ-CALCP-00",    { 0x720008, 0x720009, 0x72000A, 0x72000B, 0x720028, 0x720029, 0x72002A, 0x72002B } },
    { "ROS-TDQ-CALCPROI-00", { 0x730088, 0x730089, 0x73008A, 0x73008B } },
    { "ROS-TDQ-CALJE-00",    { 0x74000C, 0x74000D, 0x74001C, 0x74001D, 0x74002C, 0x74002D, 0x74003C, 0x74003D } },
    { "ROS-TDQ-CALJEROI-00", { 0x75008C, 0x75008D } },
    { "ROS-TDQ-MUCTPI-00",   { 0x760000 } },
    { "ROS-TDQ-CTP-00",      { 0x770000 } },
    { "ROS-TDQ-FTK-00",      { 0x7F0000 } },
    { "ROS-FWD-BCM-00",      { 0x81000A, 0x81000C, 0x81000F } },
    { "ROS-FWD-BCM-01",      { 0x81004A, 0x81004C } },
    { "ROS-FWD-LUC-00",      { 0x820000 } },
    { "ROS-FWD-LUC-SA-00",   { 0x820001 } },
    { "ROS-FWD-ZDC-00",      { 0x830002, 0x830003 } },
    { "ROS-FWD-ALFA-00",     { 0x840000, 0x840001 } }
};

} // namespace dcm
} // namespace daq
