#include <dqm_display/gui/CanvasWidget.h>

#include <TCanvas.h>
#include <TImage.h>
#include <TROOT.h>
#include <TVirtualX.h>

#include <QApplication>
#include <QDebug>
#include <QPainter>
#include <QMouseEvent>
#include <QResizeEvent>

namespace dqm_display {
namespace gui {

CanvasWidget::CanvasWidget(QWidget* parent, int w, int h)
: QWidget(parent),
  m_canvas(nullptr)
{
    //setAttribute(Qt::WA_PaintOnScreen, true);
    setAttribute(Qt::WA_OpaquePaintEvent, true);
    setUpdatesEnabled(false);
    setMouseTracking(true);

    int wid = gVirtualX->AddWindow((ULong_t)winId(), w, h);
    m_canvas = new TCanvas("Root Canvas", width(), height(), wid);
}

CanvasWidget::~CanvasWidget() {
    delete m_canvas;
}

void CanvasWidget::mouseMoveEvent(QMouseEvent *e)
{
    if (e->buttons() & Qt::LeftButton) {
        m_canvas->HandleInput(kButton1Motion, e->x(), e->y());
    }
    else if (e->buttons() & Qt::MidButton) {
        m_canvas->HandleInput(kButton2Motion, e->x(), e->y());
    }
    else if (e->buttons() & Qt::RightButton) {
        m_canvas->HandleInput(kButton3Motion, e->x(), e->y());
    }
    else {
        m_canvas->HandleInput(kMouseMotion, e->x(), e->y());
    }
}

void CanvasWidget::mousePressEvent(QMouseEvent *e)
{
    switch (e->button()) {
        case Qt::LeftButton:
            m_canvas->HandleInput(kButton1Down, e->x(), e->y());
            break;
        case Qt::MidButton:
            m_canvas->HandleInput(kButton2Down, e->x(), e->y());
            break;
        case Qt::RightButton:
            m_canvas->HandleInput(kButton3Down, e->x(), e->y());
            break;
        default:
            break;
    }
}

void CanvasWidget::mouseReleaseEvent(QMouseEvent *e)
{
    switch (e->button()) {
        case Qt::LeftButton:
            m_canvas->HandleInput(kButton1Up, e->x(), e->y());
            break;
        case Qt::MidButton:
            m_canvas->HandleInput(kButton2Up, e->x(), e->y());
            break;
        case Qt::RightButton:
            m_canvas->HandleInput(kButton3Up, e->x(), e->y());
            break;
        default:
            break;
    }
}

void CanvasWidget::changeEvent(QEvent * )
{
    if (m_canvas) {
        m_canvas->Resize();
        m_canvas->Update();
    }
}

void CanvasWidget::paintEvent(QPaintEvent * )
{
    if (m_canvas) {
        m_canvas->Resize();
        m_canvas->Update();
    }
}

void CanvasWidget::resizeEvent(QResizeEvent * e)
{
    if (m_canvas) {
        m_canvas->Resize();
        m_canvas->Update();
    }
    if (        e->size().width() != e->oldSize().width()
            ||  e->size().height() != e->oldSize().height()) {
        QApplication::postEvent(this, new QPaintEvent(rect()));
    }
}

void CanvasWidget::showEvent(QShowEvent * )
{
    if (m_canvas) {
        m_canvas->Resize();
        m_canvas->Update();
    }
}

void
CanvasWidget::resize(const QSize & s)
{
    if (s.isValid() && !s.isEmpty() && size() != s) {
    	// The order of the following calls is important!
    	m_canvas->SetCanvasSize(s.width(), s.height());
        QWidget::resize(s);
        gPad = m_canvas;
        m_canvas->ResizePad();
        m_image = QImage();
    }
}

QImage
CanvasWidget::getImage(const QSize & s)
{
    resize(s);

    if ( m_image.isNull() ) {
    	refresh();

        TImage *img = TImage::Create();
        img->FromPad(m_canvas, 0, 0, size().width(), size().height());
        uchar * arr = (uchar *)img->GetArgbArray();
        m_image = QImage(arr, img->GetWidth(), img->GetHeight(), img->GetWidth() * 4, QImage::Format_ARGB32);
    }

    return m_image;
}

void CanvasWidget::reset()
{
    m_image = QImage();
}

void CanvasWidget::refresh()
{
    if (m_canvas) {
        paintCanvas();
        m_canvas->Modified();
        m_canvas->Update();
    }
}

}}
