/*! \file Display.cpp Implements the Display class.
 * \author Serguei Kolos
 * \version 1.0
 */

#include <dqm_display/gui/Display.h>
#include <dqm_display/gui/Find.h>
#include <dqm_display/gui/ListViewHistoryModel.h>
#include <dqm_display/gui/DetailsViewDelegate.h>
#include <dqm_display/gui/IconsViewDelegate.h>
#include <dqm_display/gui/RunInfoWidget.h>
#include <dqm_display/gui/Elog.h>
#include <dqm_display/gui/Status.h>
#include <dqm_display/gui/OverlayWidget.h>

#include <dqm_display/core/Region.h>
#include <dqm_display/core/Parameter.h>
#include <dqm_display/core/TreeDataModel.h>

#include <QDebug>
#include <QUrl>
#include <QSettings>

#include <TSystem.h>

namespace dqm_display {
namespace gui {

bool Display::m_divide_by_reference = false;

Display::Display(DQDataModel * model, About * about)
  : m_mainModel(model),
    m_systemModel(new SystemFilterModel(this, m_mainModel)),
    m_parameterModel(new ParameterFilterModel(this, m_mainModel)),
    m_listViewHistoryModel(new ListViewHistoryModel(this, QString::fromStdString(m_mainModel->model().getPartitionName()))),
    m_findDialog(new Find(m_mainModel->model())),
    m_aboutDialog(about),
    m_systemActionGroup(this),
    m_viewActionGroup(this),
    m_history_position(0),
    m_current_view(0),
    m_histograms_height(192),
    m_timer(this)
{  
    setupUi(this);

    widget_AlarmsAndLogs->subscribe(m_mainModel->model().getPartitionName());

    setCorner( Qt::BottomLeftCorner, Qt::LeftDockWidgetArea );

    actionView_Graphics->setData((int)0);
    actionView_Icons->setData((int)1);
    actionView_Details->setData((int)2);
    actionView_Gallery->setData((int)3);

    m_viewActionGroup.addAction(actionView_Icons);
    m_viewActionGroup.addAction(actionView_Details);
    m_viewActionGroup.addAction(actionView_Gallery);
    m_viewActionGroup.addAction(actionView_Graphics);
    connect( &m_viewActionGroup, SIGNAL(triggered(QAction*)), 
    		this, SLOT(when_actionViewTriggered(QAction*)) );
    
    m_current_view = actionView_Graphics;

    listView_History->setModel(m_listViewHistoryModel);
    listView_History->setItemDelegate(
	    new IconsViewDelegate(QSize(m_histograms_height, m_histograms_height), 4, this, false) );
    
    listView_Histograms->setModel(m_mainModel);
    listView_Histograms->setItemDelegate(
	    new IconsViewDelegate(
		    QSize(m_histograms_height, m_histograms_height), 4, this) );
    
    treeView_Parameters->setProxyModel(m_parameterModel);
    treeView_Parameters->setItemDelegate(
	    new DetailsViewDelegate(128, 4, treeView_Parameters));
    treeView_Parameters->header()->hideSection(0);
    treeView_Parameters->header()->resizeSection(1, 100);
    
    treeView_Histograms->setModel( m_mainModel );
    treeView_Histograms->setItemDelegate(
	    new DetailsViewDelegate(m_histograms_height, 4, this) );
    treeView_Histograms->header()->resizeSection(0, 300);
    treeView_Histograms->header()->resizeSection(1, 100);
    
    treeView_Main->setModel(m_mainModel);
    treeView_Main->header()->setSectionResizeMode(0, QHeaderView::Stretch);
    for (int i = 1; i < m_mainModel->columnCount(); ++i) {
	treeView_Main->setColumnHidden(i, true);
    }

    treeView_Systems->setProxyModel(m_systemModel);
    treeView_Systems->header()->setSectionResizeMode(0, QHeaderView::Stretch);
    for (int i = 1; i < m_mainModel->columnCount(); ++i) {
	treeView_Systems->setColumnHidden(i, true);
    }
    treeView_Systems->setRootIndex(m_mainModel->rootIndex());
    treeView_Systems->expandAll();
    dockWidget_Systems->setWindowTitle(m_mainModel->model().getName());
    
    connect( m_findDialog, SIGNAL(nodeActivated(const core::TreeNode*)),
	     this, SLOT(when_nodeActivated(const core::TreeNode*)));
    
    connect( treeView_Systems, SIGNAL(nodeActivated(const core::TreeNode*)),
           this, SLOT(when_nodeActivated(const core::TreeNode*)));

    connect( treeView_Main, SIGNAL(nodeActivated(const core::TreeNode*)),
           this, SLOT(when_nodeActivated(const core::TreeNode*)));
           
    connect( graphicsView_Layout, SIGNAL(nodeActivated(const core::TreeNode*)),
           this, SLOT(when_nodeActivated(const core::TreeNode*)));
           
    connect( listView_Histograms, SIGNAL(nodeActivated(const core::TreeNode*)),
           this, SLOT(when_nodeActivated(const core::TreeNode*)));
           
    connect( treeView_Histograms, SIGNAL(nodeActivated(const core::TreeNode*)),
           this, SLOT(when_nodeActivated(const core::TreeNode*)));
    
    connect( actionView_Show_Tree, SIGNAL(toggled(bool)),
           treeView_Main, SLOT(setVisible(bool)));

    connect( widget_AlarmsAndLogs, SIGNAL(alarmAcknowledged(const core::TreeNode*)),
	     this, SLOT(when_nodeActivated(const core::TreeNode*)));

    // The following connections are done via queue as corresponding
    // signals are emitted from non-Qt threads
    connect( m_mainModel, SIGNAL(resultProduced(const core::TreeNode*)),
	   this, SLOT(when_resultProduced(const core::TreeNode*)), Qt::QueuedConnection);

    connect( m_mainModel, SIGNAL(alarmProduced(const core::TreeNode*)),
	   widget_AlarmsAndLogs, SLOT(when_alarmProduced(const core::TreeNode*)), Qt::QueuedConnection);

    connect( m_mainModel, SIGNAL(configurationUpdated()),
	    this, SLOT(when_configurationUpdated()), Qt::QueuedConnection);

    connect( m_mainModel, SIGNAL(statsUpdated(const QString &, int, int)),
	    this, SLOT(when_statsUpdated(const QString &, int, int)), Qt::QueuedConnection);

    action_Elog->setEnabled(Elog::runningAtP1());
    
    RunInfoWidget * ri = new RunInfoWidget(QString::fromStdString(m_mainModel->model().getPartitionName()));
    connect( ri, SIGNAL(databaseRestarted()),
	    this, SLOT(when_configurationUpdated()), Qt::QueuedConnection);
    statusBar()->addPermanentWidget( ri, 3);

    statusBar()->addPermanentWidget(m_stat_label = new QLabel(), 2);

    statusBar()->addPermanentWidget(m_path_label = new QLabel(), 1);

    actionNavigation_Home->trigger();

    restoreState();

    QObject::connect(&m_timer, SIGNAL(timeout()), this, SLOT(handleRootEvents()));
    m_timer.start(10);
}

Display::~Display()
{
    saveState();
    delete m_findDialog;
}

void Display::handleRootEvents()
{
   //call the inner loop of ROOT
   gSystem->ProcessEvents();
}

void
Display::on_pushButton_AddToElog_clicked()
{
    Elog::instance().addAttachment(*histogram_Widget);
}

void
Display::closeEvent(QCloseEvent* )
{
    QApplication::quit();
}

void
Display::on_pushButton_DivideByReference_clicked(bool on) {
    m_divide_by_reference = on;
    histogram_Widget->refresh();
    signalImageSizeChange();
}

void
Display::signalImageSizeChange()
{
    emit viewHeightChanged(m_histograms_height + m_divide_by_reference, m_systemModel->index(0, 0, QModelIndex()));
}

void
Display::on_pushButton_Lock_clicked(bool on)
{
    if (on)
    {
	histogram_Widget->setEnabled(true);
    }
    else
    {
	QModelIndex i = treeView_Main->currentIndex();
	listView_Histograms->setRootIndex(i);
	treeView_Histograms->setRootIndex(i);
	histogram_Widget->setNode(DQDataModel::treeNode(i));
	histogram_Widget->setEnabled(false);
    }
}

void
Display::when_configurationUpdated()
{
    graphicsView_Layout->clear();
    m_mainModel->clear();
    m_listViewHistoryModel->clear();

    OverlayWidget ow(this);

    if (!m_mainModel->loadData()) {
	delete this;
	return ;
    }
    treeView_Systems->setRootIndex(m_mainModel->rootIndex());
    treeView_Systems->expandAll();
    m_history.clear();
    m_history_position = 0;
    actionNavigation_Home->trigger();
}

void
Display::when_statsUpdated(const QString & last_update_time, int total_updates, int /*update_rate*/)
{
    QString text = QString("%1 updates received, last one at %2")
	    .arg(total_updates).arg(last_update_time);
    m_stat_label->setText(text);
}

void
Display::saveState()
{
    QSettings settings("CERN", "DQMD");
    settings.setValue("/geometry", saveGeometry());
    settings.setValue("/windowState", QMainWindow::saveState());
    settings.setValue("/histogramsHeight", m_histograms_height);
}

void
Display::restoreState()
{
    QSettings settings("CERN", "DQMD");
    restoreGeometry(settings.value("/geometry").toByteArray());
    QMainWindow::restoreState(settings.value("/windowState").toByteArray());
    m_histograms_height = settings.value("/histogramsHeight", m_histograms_height).toInt();
    signalImageSizeChange();
}

void
Display::when_actionViewTriggered(QAction * action)
{
    stackedWidget->setCurrentIndex(action->data().toInt());
    if ( action == actionView_Gallery)
    {
	m_viewActionGroup.setEnabled(false);
    }
    else
    {
	m_viewActionGroup.setEnabled(true);
	m_current_view = action;
    }
}

void
Display::activateNode(const core::TreeNode * node)
{
    QModelIndex index = m_mainModel->index(node);

    QModelIndex i = index;
    QModelIndex c = treeView_Systems->currentIndex();
    bool inplace = false;
    for( ; i.parent().isValid(); i = i.parent()) {
	if (c == i) {
	    inplace = true;
	    break;
	}
    }
    if (!inplace) {
	treeView_Systems->setCurrentIndex(i);
	treeView_Main->setRootIndex(i);
    }
    else {
	treeView_Main->setRootIndex(c);
    }

    treeView_Main->setCurrentIndex(index);
    actionView_AutoScale->setChecked(true);
    updateViews(index);
}

void
Display::when_nodeActivated(const core::TreeNode * node)
{
    activateNode(node);
    addToHistory(node);
}

void 
Display::updateViews(const QModelIndex & index)
{
    if ( !index.isValid() ) {
    	return ;
    }
    
    m_current_index = index;

    listView_Histograms->setRootIndex(index);
    treeView_Histograms->setRootIndex(index);

    const core::TreeNode * node = DQDataModel::treeNode(index);

    m_listViewHistoryModel->setNode(node);
    dockWidget_History->setWindowTitle(QString("History for \"%1\"").arg(node->getName()));
    graphicsView_Layout->setNode(node);

    QString title(node->getLabel());
    label_LayoutTitle->setText(title.replace("\n", " "));

    m_path_label->setText(node->getFullName());

    if (!node->isLeaf())
    {
	m_current_view->trigger();
        pushButton_AddToElog->hide();
    }
    else
    {
	m_parameterModel->setVisibleItem(treeView_Parameters, index);
        actionView_Gallery->trigger();

	histogram_Widget->setNode(node);

        pushButton_AddToElog->show();
	pushButton_AddToElog->setEnabled(!histogram_Widget->isNull());
    }
}

void 
Display::on_actionView_Zoom_In_triggered()
{
    if (	stackedWidget->currentIndex() == 1 // icons view
	    ||	stackedWidget->currentIndex() == 2 // details view
	    ) {
	m_histograms_height = m_histograms_height * 6./5.;
	signalImageSizeChange();
    }
    else {
	actionView_AutoScale->setChecked(false);
	graphicsView_Layout->scale(6./5.);
    }
}

void
Display::on_actionView_Zoom_Out_triggered()
{
    if (	stackedWidget->currentIndex() == 1 // icons view
	    ||	stackedWidget->currentIndex() == 2 // details view
	    ) {
	m_histograms_height = m_histograms_height * 5./6.;
	signalImageSizeChange();
    }
    else {
	actionView_AutoScale->setChecked(false);
	graphicsView_Layout->scale(5./6.);
    }
}

void
Display::on_actionView_AutoScale_toggled()
{
    if (actionView_AutoScale->isChecked()) {
	graphicsView_Layout->setAutoScale();
    }
}

void
Display::on_actionNavigation_Back_triggered()
{
    const core::TreeNode * node = m_history[--m_history_position];
    activateNode(node);
    updateNavigationButtons(node);
}

void 
Display::on_actionNavigation_Forward_triggered()
{
    const core::TreeNode * node = m_history[++m_history_position];
    activateNode(node);
    updateNavigationButtons(node);
}

void 
Display::on_actionNavigation_Home_triggered()
{
    when_nodeActivated(DQDataModel::treeNode(m_mainModel->rootIndex()));
}

void 
Display::on_actionNavigation_LevelUp_triggered()
{
    when_nodeActivated(DQDataModel::treeNode(m_mainModel->parent(treeView_Main->currentIndex())));
}

void
Display::on_actionNavigation_Find_triggered()
{
    m_findDialog->show();
}

void
Display::addToHistory(const core::TreeNode * node)
{
    if (!node) {
	return;
    }

    if (m_history_position < (m_history.size() - 1)) {
	m_history.remove(m_history_position + 1,
		m_history.size() - m_history_position - 1);
    }

    if (m_history.size() > 100) {
	m_history.remove(0, m_history.size() - 100);
    }

    m_history.push_back(node);
    m_history_position = m_history.size() - 1;

    updateNavigationButtons(node);
}

void 
Display::updateNavigationButtons(const core::TreeNode * node)
{
    if (!node) {
	return;
    }

    actionNavigation_Back->setEnabled(m_history_position > 0);
    actionNavigation_Forward->setEnabled(m_history_position < (m_history.size()-1));
    actionNavigation_LevelUp->setEnabled(!node->isRoot());
}

void 
Display::when_resultProduced(const core::TreeNode * node)
{
    if (!node) {
	return;
    }

    QModelIndex index = m_mainModel->index(node);
    treeView_Systems->updateNode(index);
    treeView_Main->updateNode(index);

    if (!pushButton_Lock->isChecked())
    {
	if (index == m_current_index)
	{
	    histogram_Widget->setNode(node);
	    if (node->isLeaf()) {
	        pushButton_AddToElog->setEnabled(!histogram_Widget->isNull());
	    }
	    treeView_Parameters->updateNode(index);
	    m_listViewHistoryModel->update();
	    return;
	}

	if (stackedWidget->currentIndex() == 0)
	{
	    // Graphical view may display several levels of nested items simultaneously
	    for (QModelIndex i = index.parent(); i.isValid(); i = i.parent()) {
		if (m_current_index == i) {
		    graphicsView_Layout->updateNode(node);
		    break;
		}
	    }
	} else if (index.parent() == m_current_index) {
	    if (stackedWidget->currentIndex() == 1) {
		listView_Histograms->updateNode(index);
	    } else if (stackedWidget->currentIndex() == 2) {
		treeView_Histograms->updateNode(index);
	    }
	}
    }
}

void
Display::on_actionHelp_About_triggered()
{
    m_aboutDialog->show();
}

void Display::on_action_Elog_triggered()
{
    Elog::instance().show();
}

} } // namespace dqm_display::gui
