/*! \file IconsViewDelegate.cpp Implements the IconsViewDelegate class.
 * \author Serguei Kolos
 * \version 1.0
 */
#include <dqm_display/gui/IconsViewDelegate.h>
#include <dqm_display/gui/Histogram.h>
#include <dqm_display/gui/Status.h>
#include <dqm_display/gui/DQDataModel.h>

#include <QBitmap>
#include <QDebug>
#include <QLinearGradient>
#include <QPainter>
#include <QTextDocument>

namespace dqm_display { 
namespace gui {

IconsViewDelegate::IconsViewDelegate(const QSize & icon_size, int padding, QObject *parent, bool follow_height_change)
  : QStyledItemDelegate(parent),
    m_padding(padding),
    m_border(5),
    m_radius(5),
    m_size_hint(icon_size + QSize(2*m_padding, 2*m_padding))
{
    if (follow_height_change){
	connect( parent, SIGNAL(viewHeightChanged(int, const QModelIndex &)),
		this, SLOT(when_viewHeightChanged(int, const QModelIndex &)));
    }
}

void
IconsViewDelegate::when_viewHeightChanged(int height, const QModelIndex & index)
{
    m_size_hint = QSize(height, height) + QSize(2*m_padding, 2*m_padding);
    if (index.isValid())
	emit sizeHintChanged(index);
}

void 
IconsViewDelegate::paint( QPainter * painter, 
			 const QStyleOptionViewItem & option, 
                         const QModelIndex & index ) const
{
    painter->save();

    painter->setPen(Qt::NoPen);

    if(option.state & QStyle::State_Selected)
    {
	painter->setBrush(QBrush(QColor(112, 153, 204)));
	painter->drawRoundedRect(option.rect, m_radius, m_radius, Qt::RelativeSize);
    }

    doPaint(painter, option, index);

    painter->restore();
}

void
IconsViewDelegate::doPaint( QPainter * painter,
			 const QStyleOptionViewItem & option,
                         const QModelIndex & index ) const
{
    static QPixmap NotAvailablePixmap(QString::fromUtf8(":/images/not_available.png"));
    static QPixmap BackgroundPixmap(QString::fromUtf8(":/images/pattern.png"));

    painter->translate(option.rect.left() + m_padding, option.rect.top() + m_padding);
    painter->setRenderHint(QPainter::Antialiasing, true);
    painter->setRenderHint(QPainter::TextAntialiasing, true);
    painter->setRenderHint(QPainter::HighQualityAntialiasing, true);
    painter->setRenderHint(QPainter::SmoothPixmapTransform, true);

    QSize size(option.rect.width() - 2*m_padding, option.rect.height() - 2*m_padding);
    int titleHeight = option.fontMetrics.height() + 2;

    if ( index.column() == 0 ) {
	const core::TreeNode * node = DQDataModel::treeNode(index);

	if (!node)
	    return;

	QRect outer(QPoint(0, 0), size);
	QRect inner(m_border, m_border, size.width() - 2*m_border,
		size.height() - 2*m_border - titleHeight);

	QRect fullRect(outer.adjusted(1, 1, -1, -titleHeight));

	if (index.data(Qt::DecorationRole).canConvert<dqm_display::gui::ResultHistoryGraph*>())
	{
	    ResultHistoryGraph * graph =
		qvariant_cast<dqm_display::gui::ResultHistoryGraph*>(index.data(Qt::DecorationRole));
	    QRect imageRect = fullRect.adjusted(m_border, 0, 0, -m_border);
	    QImage image = graph->getImage(imageRect.size());
	    painter->drawImage(imageRect, image);
	    painter->setBrush(QBrush(Qt::NoBrush));
	    painter->drawRoundedRect(fullRect, m_radius, m_radius, Qt::RelativeSize);
	}
	else
	{
	    if (node->isLeaf()) {
                QRect imageRect = fullRect.adjusted(m_border, 0, 0, -m_border);
		QImage image = node->getImage(imageRect.size());

		if (!image.isNull()) {
		    painter->drawImage(imageRect, image);
		    painter->setBrush(QBrush(Qt::NoBrush));
		    painter->drawRoundedRect(fullRect, m_radius, m_radius, Qt::RelativeSize);
		}
		else {
		    painter->drawPixmap(fullRect, NotAvailablePixmap);
		}
	    }
	    else {
		painter->setBrush(QBrush(BackgroundPixmap));
		painter->drawRoundedRect(fullRect, m_radius, m_radius, Qt::RelativeSize);

		QTextDocument doc;
		doc.setHtml(node->getStatusHTML());
		painter->translate(m_padding + m_border, m_padding + m_border);
		doc.drawContents(painter, fullRect);
		painter->translate(- m_padding - m_border, - m_padding - m_border);
	    }
	}

	// draw border
	QColor text_color(96, 96, 96);
	QPainterPath p;
	p.addRoundedRect(outer, m_radius, m_radius, Qt::RelativeSize);
	p.addRoundedRect(inner, m_radius, m_radius, Qt::RelativeSize);

	core::Result::Status s = node->getResult()->status();
	QColor color(Status::color(s));

	int brightness = color.red()*0.299 + color.green()*0.587 + color.blue()*0.114;
	text_color = brightness < 128 ? Qt::white : Qt::black;
	color.setAlpha(96);
	painter->setBrush(color);

	QPainterPath r;
	r.addRoundedRect(outer.adjusted(0, 3*outer.height()/4, 0, 0),
		m_radius, m_radius, Qt::RelativeSize);
	p = p.intersected(r);
	painter->drawPath(p);

	if (s == core::Result::Status::DISABLED) {
	    painter->setBrush(QColor(0, 0, 0, 96));
	    painter->drawRoundedRect(outer, m_radius, m_radius, Qt::RelativeSize);
	}

	// draw title text
	QFont font = qvariant_cast<QFont>(index.data(Qt::FontRole));
	painter->setFont(font);
	QString text = qvariant_cast<QString>(index.data(Qt::DisplayRole));

	QRect t(outer.left(), inner.top() + inner.height(), outer.width(), titleHeight + 2*m_border);
	painter->setPen(text_color);
	painter->drawText(t, Qt::AlignHCenter | Qt::AlignCenter, text);
    }
}

QSize
IconsViewDelegate::sizeHint(const QStyleOptionViewItem & option, const QModelIndex & ) const
{
    return QSize(m_size_hint.width(), m_size_hint.height() + option.fontMetrics.height());
}

} } // namespace dqm_display::gui
