/*! \file ResultHistoryGraph Declares the dqm_display::gui::ResultHistoryGraph class.
 * \author Yurij Ilchenko, Kevin Slagle, Serguei Kolos
 */

#include <dqm_display/gui/ResultHistoryGraph.h>
#include <dqm_display/gui/Status.h>

#include <TAxis.h>
#include <TCanvas.h>
#include <TImage.h>
#include <TVirtualX.h>

#include <QDebug>
#include <QImage>

namespace dqm_display { namespace gui {

ResultHistoryGraph::ResultHistoryGraph(
        const std::string & name,
        const std::vector<dqmf::is::Result> & results, 
        const std::string & tag_name,
        QWidget* parent)
  : QWidget(parent),
    m_title(tag_name.empty() ? name.c_str() : tag_name.c_str())
{ 
    m_graph.SetEditable(kFALSE);
    m_graph.SetMarkerStyle(1);
    m_graph.SetLineColor(13);
    m_graph.SetLineStyle(2);
    
    const size_t size = results.size();

    if (tag_name.empty()) {
	for (size_t i = 0; i < size; ++i) {
            double x = (double)results[i].time().c_time();// - time_offset;
	    int y = (int)results[i].status;
	    m_graph.SetPoint(i, x, y);
            m_markers[(core::Result::Status)results[i].status].SetNextPoint(x, y);
	}
    }
    else {
        int cnt=0;
	for (size_t i = 0; i < size; ++i) {
	    for (size_t j = 0; j < results[i].tags.size(); j++) {
		if (results[i].tags[j].name == tag_name) {
                    double x = (double)results[i].time().c_time();// - time_offset;
                    double y = results[i].tags[j].value;
                    m_graph.SetPoint(cnt++, x, y);
                    
                    m_markers[(core::Result::Status)results[i].status].SetNextPoint(x, y);
		    break;
		}
	    }
	}
    }
    
    m_graph.GetXaxis()->SetLabelOffset(0.02);
    m_graph.GetXaxis()->SetLabelSize(0.06);
    m_graph.GetXaxis()->SetTimeDisplay(1);
    m_graph.GetXaxis()->SetTimeFormat("%H:%M");
    m_graph.GetXaxis()->SetNdivisions(3, true);

    m_graph.GetYaxis()->SetLabelOffset(0.02);
    m_graph.GetYaxis()->SetLabelSize(0.06);

    setAttribute(Qt::WA_PaintOnScreen, true);
    setAttribute(Qt::WA_OpaquePaintEvent, true);
    setUpdatesEnabled(false);
    setMouseTracking(false);
    setMinimumSize(128, 128);

    int wid = gVirtualX->AddWindow((ULong_t)winId(), 1000, 1000);
    m_canvas = new TCanvas("Root Canvas", width(), height(), wid);
    
    m_canvas->SetTickx(kFALSE);
    m_canvas->SetTicky(kFALSE);
    
    m_canvas->SetTopMargin(0.05);
    m_canvas->SetRightMargin(0.05);

    if (tag_name.empty()) {
        m_canvas->SetLeftMargin(0.1);
	m_graph.GetYaxis()->SetNdivisions(0, false);
    }
    else  {
        m_canvas->SetLeftMargin(0.15);
        m_canvas->SetGridx(kTRUE);
        m_canvas->SetGridy(kTRUE);
    }
}

ResultHistoryGraph::~ResultHistoryGraph()
{
    delete m_canvas;
}

QImage
ResultHistoryGraph::getImage(const QSize & s)
{
    if ( size() != s )
    {   
        m_canvas->SetCanvasSize(s.width(), s.height());
        QWidget::resize(s);
        gPad = m_canvas;
        m_canvas->ResizePad();
        m_image = QImage();
    }

    if (m_image.isNull()) {
        m_canvas->cd();
	m_graph.Draw("ALP");

	for ( auto & m : m_markers ) {
	    m.second.SetMarkerColor(Status::color_t(m.first));
	    m.second.SetMarkerStyle(23);
	    m.second.SetMarkerSize(1);
	}

	// The drawing order is important
	Markers::iterator it;
	if ((it = m_markers.find(core::Result::Status::UNDEFINED)) != m_markers.end()) {
	    it->second.Draw();
	}
	if ((it = m_markers.find(core::Result::Status::DISABLED)) != m_markers.end()) {
	    it->second.Draw();
	}
	if ((it = m_markers.find(core::Result::Status::GOOD)) != m_markers.end()) {
	    it->second.Draw();
	}
	if ((it = m_markers.find(core::Result::Status::WARNING)) != m_markers.end()) {
	    it->second.Draw();
	}
	if ((it = m_markers.find(core::Result::Status::ERROR)) != m_markers.end()) {
	    it->second.Draw();
	}

        m_canvas->Modified();
        m_canvas->Update();

        TImage *img = TImage::Create();
        img->FromPad(m_canvas);

        uchar * arr = (uchar *)img->GetArgbArray();
        m_image = QImage(arr, img->GetWidth(), img->GetHeight(), img->GetWidth() * 4, QImage::Format_ARGB32);
    }

    return m_image;
}
} }
