/*! \file GraphicsView.cpp Implements the GraphicsView class.
 * \author Serguei Kolos
 * \version 1.0
 */

#include <dqm_display/gui/GraphicsView.h>

#include <QGraphicsScene>
#include <QGraphicsWidget>
#include <QResizeEvent>
#include <QDebug>

namespace dqm_display { namespace gui {

GraphicsView::GraphicsView(QWidget *parent)
  : QGraphicsView(parent),
    m_scene_set(false),
    m_auto_scale(true),
    m_layout(0),
    m_menu(this)
{
    setViewportUpdateMode(QGraphicsView::SmartViewportUpdate);
    setHorizontalScrollBarPolicy(Qt::ScrollBarAsNeeded);
    setVerticalScrollBarPolicy(Qt::ScrollBarAsNeeded);
    setScene(new QGraphicsScene(this));
}

void 
GraphicsView::clear()
{
    scene()->clear();
    m_scene_set = false;
    m_region.reset(0);
    m_layout = 0;
}

void
GraphicsView::setNode(const core::TreeNode * node)
{
    if (node->isLeaf())
    	return ;

    const core::Region * r = static_cast<const core::Region*>(node);
    m_region.reset(new GraphicsRegion(*r, m_menu, 0, true));

    m_scene_set = false;

    if (isVisible())
	setupScene();
}

void
GraphicsView::setupScene()
{
    if (m_scene_set || !m_region)
	return;

    QGraphicsScene * scene = this->scene();
    scene->clear();


    QGraphicsWidget* gwidget = new QGraphicsWidget();
    m_layout = m_region->takeLayout();
    gwidget->setLayout(m_layout);
    scene->addItem(gwidget);

    QList<QGraphicsItem *> it = scene->items();
    for ( int i = 0; i < it.size(); ++i )
    {
	GraphicsItem * item = dynamic_cast<GraphicsItem*>(it[i]);
	if ( item )
	    QObject::connect( item, SIGNAL(nodeActivated(const core::TreeNode*)),
		     this, SIGNAL(nodeActivated(const core::TreeNode*)) );
    }

    m_scene_set = true;
    m_auto_scale = true;

    scene->setSceneRect(QRectF(QPoint(0.,0.), m_layout->preferredSize()));

    fitInView(scene->sceneRect(), Qt::KeepAspectRatio);
}

void
GraphicsView::updateNode(const core::TreeNode * node)
{
    if (!isVisible() || !m_scene_set)
	return;

    QList<QGraphicsItem *> it = scene()->items();
    for ( int i = 0; i < it.size(); ++i )
    {
	GraphicsItem * item = dynamic_cast<GraphicsItem*>(it[i]);
	if ( item && item->representsNode(node) ) {
	    item->updateStatusColor();
	    break;
	}
    }
}

void
GraphicsView::scale(double q)
{
    m_auto_scale = false;
    QGraphicsView::scale(q, q);
}

void
GraphicsView::setAutoScale()
{
    m_auto_scale = true;
    resizeEvent(0);
}

void
GraphicsView::resizeEvent(QResizeEvent * )
{    
    if (m_scene_set && m_auto_scale)
	fitInView(scene()->sceneRect(), Qt::KeepAspectRatio);
}

void 
GraphicsView::paintEvent(QPaintEvent * event)
{
    QPaintEvent e(event->region().boundingRect());
    QGraphicsView::paintEvent(&e);

    // QGraphicsView::paintEvent calls setGeometry, so if we want we can
    // get the bounding rect as the layout has been already set up
    if (m_scene_set && m_auto_scale && m_layout)
    {
	QRectF r = m_layout->boundingRect();
	if (!r.isEmpty() && scene()->sceneRect() != r){
	    scene()->setSceneRect(r);
	    fitInView(scene()->sceneRect(), Qt::KeepAspectRatio);
	}
    }
}

void
GraphicsView::showEvent(QShowEvent * )
{
    setupScene();
}

bool
GraphicsView::viewportEvent(QEvent * event)
{
    if (event->type() == QEvent::WindowActivate || event->type() == QEvent::WindowDeactivate)
	return true;
    else
	return QGraphicsView::viewportEvent(event);
}

} } // namespace dqm_display::gui
