#include <dqm_core/OutputFactoryT.h>

#include <dqmf/CoolOutput.h>

#include <rc/RunParams.h>

#include <CoralBase/Attribute.h>
#include <CoolKernel/Exception.h>
#include <CoolKernel/ValidityKey.h>
#include <CoolKernel/IDatabaseSvc.h>
#include <CoolKernel/IDatabase.h>
#include <CoolKernel/IFolder.h>
#include <CoolKernel/IRecordSpecification.h>
#include <CoolKernel/Record.h>
#include <CoolKernel/Time.h>
#include <CoolApplication/DatabaseSvcFactory.h>

namespace 
{
    const std::string DefaultConnectString( 
    		"oracle://ATONR_COOL;schema=ATLAS_COOLONL_GLOBAL;dbname=COMP200" );
    const std::string DefaultFolderName( "/GLOBAL/DETSTATUS/DQMFONL" );

    dqm_core::OutputFactoryT<dqmf::CoolOutput> __factory__( "CoolOutput" );

    cool::UInt63 to_cool_time( boost::posix_time::ptime t )
    {
	static boost::posix_time::ptime s( boost::posix_time::from_time_t( 0 ) );

	if( t == boost::posix_time::neg_infin )
	    return cool::UInt63(0);
	
        if( t == boost::posix_time::pos_infin )
	    return cool::UInt63(-1);

	return ( t - s ).total_nanoseconds();
    }
    
    cool::UInt63 cool_time_now()
    {
	boost::posix_time::ptime t( boost::posix_time::second_clock::universal_time( ) );
    	return to_cool_time( t );
    }
}
    
ERS_DECLARE_ISSUE_BASE(	dqmf,
			CoolError,
			dqm_core::Exception,
			"COOL database (" << database << ") failure: " << reason,
                        ,
                        ((std::string) database) ((std::string) reason)
		)

ERS_DECLARE_ISSUE_BASE(	dqmf,
			CoolStoreError,
			dqm_core::Exception,
			"Writing DQ result (" << status << ") for the '" << channel 
                        	<< "' channel with IOV [" << since << ";" << till 
                        	<< "] to COOL database (" << database << ") failed: " << reason,
                        ,
                        ((int) status)
                        ((std::string) channel) 
                        ((cool::UInt63) since)
                        ((cool::UInt63) till)
                        ((std::string) database) 
                        ((std::string) reason)
		)

dqmf::CoolOutput::Result::Result( cool::ChannelId channel )
  : status_( dqm_core::Result::Undefined ),
    channel_( channel ),
    since_( cool_time_now() ),
    recorded_( false )
{ ; }

dqmf::CoolOutput::CoolOutput( const std::vector<std::string> & parameters ) 
  : partition_( parameters.size() ? parameters[0] : IPCPartition().name() ),
    connect_string_( parameters.size() > 1 ? parameters[1] : DefaultConnectString ),
    blocked_( true )
{    
    ERS_DEBUG( 0, "Connecting to the '" << connect_string_ << "' database ..." );
    try {
	 database_ = cool::DatabaseSvcFactory::databaseService().openDatabase( connect_string_, false ); // false = !read_only
	 ERS_DEBUG( 0, "Connection to '" << connect_string_ << "' database has been established." );
    }
    catch ( std::exception & ex ) {
	throw CoolError( ERS_HERE, connect_string_, ex.what() );
    }   
    
    std::string folder_name( parameters.size() > 2 ? parameters[2] : DefaultFolderName );
    try {
	folder_ = database_ -> getFolder( folder_name );
	ERS_DEBUG( 0, folder_name << " found" );
        
        std::vector<cool::ChannelId> channels = folder_ -> listChannels();
	
        boost::mutex::scoped_lock lock( mutex_ );
	for ( size_t i = 0; i < channels.size(); ++i )
	{
	    std::string chname = folder_ -> channelName( channels[i] );
            ERS_DEBUG( 1, "name of the " << channels[i] << " channel is '" << chname << "'" );
	    results_.insert( std::make_pair( chname, Result( channels[i] ) ) );
	}
    }
    catch ( std::exception & ex ) {
	throw CoolError( ERS_HERE, connect_string_, ex.what() );
    }
}

dqmf::CoolOutput::~CoolOutput( )
{
    finalize();
    try {
	database_-> closeDatabase();
	ERS_DEBUG( 0, "Connection to COOL database has been closed" );
    }
    catch ( std::exception & ex ) {
	ers::warning( CoolError( ERS_HERE, connect_string_, ex.what() ) );
    }
}

void
dqmf::CoolOutput::initialize(  )
{
    ERS_DEBUG( 1, "initialize" );

    boost::mutex::scoped_lock lock( mutex_ );

    cool::UInt63 since = cool_time_now();
    std::map<std::string,Result>::iterator it = results_.begin( );
    
    for ( ; it != results_.end(); ++it )
    {
	it -> second.status_ = dqm_core::Result::Undefined;
        it -> second.since_ = since;
        it -> second.recorded_ = false;
    }
}

void
dqmf::CoolOutput::finalize( )
{
    ERS_DEBUG( 1, "finalize" );

    boost::mutex::scoped_lock lock( mutex_ );

    cool::UInt63 till = cool_time_now();
    std::map<std::string,Result>::iterator it = results_.begin( );
    
    for ( ; it != results_.end(); ++it )
    {
	if ( it -> second.recorded_ )
	{
	    try {
		folder_ -> truncateObjectValidity( till, it -> second.channel_ );
	    }
	    catch ( std::exception & ex ) {
		throw CoolStoreError( ERS_HERE, it -> second.status_, it -> first,
				it -> second.since_, till, connect_string_, ex.what() );
	    }
	}
        
	it -> second.status_ = dqm_core::Result::Undefined;
	it -> second.since_ = till;
	it -> second.recorded_ = false;
    }
}

void
dqmf::CoolOutput::publishResult( const std::string & name, const dqm_core::Result & result ) 
{
    ERS_DEBUG( 2, "DQ result was received" );
    if ( blocked_ )
    {
	ERS_DEBUG( 1, "the COOL output stream is blocked, nothing is written to COOL" );
    	return;
    }
    
    boost::mutex::scoped_lock lock( mutex_ );
    std::map<std::string,Result>::iterator it = results_.find( name );
    
    if ( it == results_.end() )
    {
    	return ;
    }
    
    if ( it -> second.status_ == result.status_ && it -> second.recorded_ )
    {
	return ;
    }

    cool::UInt63 since = to_cool_time( result.universal_time_ );

    if ( !it -> second.recorded_ )
    {
	writeResult( it -> first, it -> second.channel_,
		     it -> second.status_, it -> second.since_, cool::ValidityKeyMax );
	it -> second.recorded_ = true;
    }

    if ( it -> second.status_ != result.status_ )
    {
	it -> second.status_ = result.status_;
	it -> second.since_ = since;
	writeResult( it -> first, it -> second.channel_,
		     it -> second.status_, it -> second.since_, cool::ValidityKeyMax );
    }
}

void 
dqmf::CoolOutput::writeResult(	const std::string & channelname, 
				cool::ChannelId channel, 
                                dqm_core::Result::Status status, 
                                cool::UInt63 since, 
                                cool::UInt63 till ) const
{
    ERS_DEBUG( 2, "storing DQ the " << status << " result for the " << channelname 
    		  << " subsystem to COOL database, since = " << since << ", till = " << till );
    	    
    try {
	cool::Record payload( folder_ -> payloadSpecification() );
	payload["Code"].setValue<cool::Int32>( status );
	payload["deadFrac"].setValue<cool::Float>( 0. );
	payload["Thrust"].setValue<cool::Float>( 0. );

	folder_ -> storeObject( since, till, payload, channel );
	ERS_DEBUG( 2, "data have been stored to COOL folder");
    }
    catch ( std::exception & ex ) {
	throw CoolStoreError( ERS_HERE, status, channelname, since, till, connect_string_, ex.what() );
    }
}

void
dqmf::CoolOutput::deactivate()
{
    ERS_DEBUG( 1, "block COOL output stream" );
    blocked_ = true;
    finalize();
}

void
dqmf::CoolOutput::activate()
{
    ERS_DEBUG( 1, "activate COOL output stream" );
    initialize();
    blocked_ = false;
}
