package elisa_client.service;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;
import java.util.Set;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.client.HttpClient;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpMethod;
import org.springframework.http.ResponseEntity;
import org.springframework.http.client.HttpComponentsClientHttpRequestFactory;
import org.springframework.stereotype.Component;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.RestTemplate;

import elisa_client.model.Attachment;
import elisa_client.model.AttachmentList;
import elisa_client.model.InputMessage;
import elisa_client.model.Message;
import elisa_client.model.MessageBody;
import elisa_client.model.MessageList;
import elisa_client.model.MessageType;
import elisa_client.model.MessageTypeList;
import elisa_client.model.OptionMetadata;
import elisa_client.model.OptionMetadataList;
import elisa_client.model.SystemAffectedList;

/**
 * Client used for communicating with the ELisA REST web service
 * @author alina
 *
 */
@Component( "elisaClient" )
public class ElisaClient
{
    private static final Log log = LogFactory.getLog(ElisaClient.class.getName());
	/**
	 * Facilitates communication with the elisa rest; autowired by Spring
	 */
	@Autowired
	private RestTemplate restTemplate;
	
	/**
	 * The base URL of the elisa rest web service - it is configurable via runtime.properties file
	 */
	@Value("${elisaRest.URL}")
	private String elisarestUrl;
	    
	/**
	 * create Authorization header using:
	 * 1.--for LDAP -- Basic Authentication: authorization header containing encoded user credentials.
	 * 2.--for SSO --- Cookie: authorization header containing info from sso cookie file created previously using auth-get-sso-cookie 
	 * @param username
	 * @param password
	 * @param ssocookiefile
	 * @return Authorization header
	 */
	MultiValueMap<String,String> createAuthHeaders( final String username, final String pwd, final String ssocookiefile )
	{
		MultiValueMap<String,String> headers = new LinkedMultiValueMap<String,String>();
	        if(username==null) {
        		String ssocookie=null;
    	    		File file = new File(ssocookiefile);
    	    		Scanner scanner = null;
			try {
				scanner = new Scanner(file);
			} catch (FileNotFoundException e) {
				throw new ElisaApiException(e.getMessage());
			}
    	    
    	    ArrayList<String> words = new ArrayList<String>();    	    	 
    	    String theWord; 
    	    try{
	    	    while (scanner.hasNext()){
	    	    theWord = scanner.next();
	    	    words.add(theWord);
	    	    }
    	    } finally {
    	        scanner.close();
    	    }   	    
        	//if word in file contains _shibsession_ take the next one after and build the cookie 
    	    for(int i=0;i<words.size();i++){
    	    	//System.out.println(words.get(i));
    	    	if(words.get(i).contains("_shibsession_")){
    	    		//System.out.println("WHAT I NEED "+words.get(i+1));
    	    		ssocookie=words.get(i)+"="+words.get(i+1);
    	    		break;
    	    	}
    	    }
        	headers.add("Cookie", ssocookie);
        }
        else{
		String auth = username + ":" + pwd;
	        byte[] encodedAuth = Base64.encodeBase64( auth.getBytes() );
	        String authHeader = "Basic " + new String( encodedAuth );
	        headers.add( "Authorization", authHeader );
        }
		return headers;
	}

	/**
	 * Return the entry with a given ID from the ELisA REST server
	 * @param username
	 * @param pwd
	 * @param ssocookiefile: path to file containing the sso cookie
	 * @param msgId: the message ID to be retrieved
	 * @return Object of type Message (class imported from ELisA REST server) 
	 * @throws Exception 
	 */
	public Message getMessage( String username, String pwd, String ssocookiefile, long msgId)
	{
 		//System.out.println("my getMessage");
        	return getMessage( username, pwd, ssocookiefile, "ATLAS", msgId); 	
	}

	/**
	 * Return the entry with a given ID from the ELisA REST server
	 * @param username
	 * @param pwd
	 * @param ssocookiefile: path to file containing the sso cookie
	 * @param msgId: the message ID to be retrieved
	 * @param logbook: logbookname to be used to retrieve the message from
	 * @return Object of type Message (class imported from ELisA REST server) 
	 * @throws Exception 
	 */
	public Message getMessage( String username, String pwd, String ssocookiefile, String logbook, long msgId)
	{
		//this is to deal with the SSL connection.
		HttpClient httpClient = HttpUtils.getNewHttpClient();
		restTemplate.setRequestFactory(new HttpComponentsClientHttpRequestFactory(httpClient));

		ResponseEntity<Message> response=null;
		//create the authentication header for the request
		MultiValueMap<String,String> headers = createAuthHeaders(username,pwd,ssocookiefile);
        //headers.add("Accept", "text/html");
        //headers.add("Content-Type","text/html");
		/*Set keys = headers.keySet();  
		for (Object k : keys) {  
		    System.out.println(k+": "+headers.get(k));
		}*/ 
		HttpEntity<Object> requestEntity=new HttpEntity<Object>(headers);

		try{
			response=restTemplate.exchange(elisarestUrl + logbook + "/messages/{id}",
					HttpMethod.GET,requestEntity,Message.class,msgId);
			return response.getBody();        	
		} catch (ElisaApiException exception) {
			throw exception;		
		} catch (Exception exception) {
			throw new ElisaApiException(exception.getMessage());
		}		
	}

	public Attachment getAttachment( String username, String pwd, long msgId, long attachId )
	{
		return null;
	}

	public AttachmentList getAttachments( String username, String pwd, long msgId )
	{
		return null;
	}

	public MessageList searchMessages( String username, String pwd )
	{
		return null;
	}

	/**
	 * Insert a logbook entry with given properties encapsulated into object of type InputMessage
	 * @param username
	 * @param pwd
	 * @param ssocookiefile: path to file containing the sso cookie
	 * @param msg: object of type InputMessage to be inserted into logbook
	 * @return object of type Message
	 * @throws Exception 
	 */
	public Message insertMessage( String username, String pwd, String ssocookiefile, InputMessage msg ) 
	{
		//System.out.println("my insertMessage");
		return insertMessage( username, pwd, ssocookiefile, "ATLAS", msg );
	}

	/**
	 * Insert a logbook entry with given properties encapsulated into object of type InputMessage
	 * @param username
	 * @param pwd
	 * @param ssocookiefile: path to file containing the sso cookie
	 * @param msg: object of type InputMessage to be inserted into logbook
	 * @param logbook: logbook name to be used to insert the message
	 * @return object of type Message
	 * @throws Exception 
	 */
	public Message insertMessage( String username, String pwd, String ssocookiefile, String logbook, InputMessage msg ) 
	{				
		//this is to deal with the SSL connection.
		HttpClient httpClient = HttpUtils.getNewHttpClient();
		restTemplate.setRequestFactory(new HttpComponentsClientHttpRequestFactory(httpClient));

		ResponseEntity<Message> response = null;
		//create the authentication header for the request
		MultiValueMap<String,String> headers = createAuthHeaders(username,pwd,ssocookiefile);
        //set the body of the request as the InputMessage        
		HttpEntity<Object> requestEntity=new HttpEntity<Object>(msg,headers);
        try{
        	response=restTemplate.exchange(elisarestUrl + logbook +"/messages",
				HttpMethod.POST,requestEntity,Message.class);
        	return response.getBody();        	
	    } catch (ElisaApiException exception) {
	    	throw exception;
	    } catch (Exception exception) {
			throw new ElisaApiException(exception.getMessage());
		}		
				
	}
	
	public Message updateMessage( String username, String pwd, MessageBody update )
	{
		return null;
	}

	public Message replyToMessage( String username, String pwd )
	{
		return null;
	}

	/**
     	* Reply to a logbook entry ID with given properties encapsulated into object of type InputMessage
     	*
     	* @param username
     	* @param pwd
     	* @param ssocookiefile: path to file containing the sso cookie
     	* @param msg:           object of type InputMessage to be inserted into logbook
     	* @return object of type Message
     	* @throws Exception
     	*/
    	public Message replyToMessage(String username, String pwd, String ssocookiefile, long replyToID, InputMessage msg) {

        	return replyToMessage(username, pwd, ssocookiefile, "ATLAS", replyToID, msg);
    	}

    	/**
     	* Reply to a logbook entry ID with given properties encapsulated into object of type InputMessage
     	*
     	* @param username
     	* @param pwd
     	* @param ssocookiefile: path to file containing the sso cookie
     	* @param logbook:       logbookname to be used to retrieve the message from
     	* @param msg:           object of type InputMessage to be inserted into logbook
     	* @return object of type Message
     	* @throws Exception
     	*/
    	public Message replyToMessage(String username, String pwd, String ssocookiefile, String logbook, long replyToID, InputMessage msg) {
        	//this is to deal with the SSL connection.
        	HttpClient httpClient = HttpUtils.getNewHttpClient();
        	restTemplate.setRequestFactory(new HttpComponentsClientHttpRequestFactory(httpClient));
        	ResponseEntity<Message> response = null;
        	//create the authentication header for the request
        	MultiValueMap<String, String> headers = createAuthHeaders(username, pwd, ssocookiefile);
        	//set the body of the request as the InputMessage
       		HttpEntity<Object> requestEntity = new HttpEntity<Object>(msg, headers);
        	try {
            		response = restTemplate.exchange(elisarestUrl + logbook + "/messages/" + replyToID, HttpMethod.POST, requestEntity, Message.class);
            		return response.getBody();
        	} catch (ElisaApiException exception) {
            		throw exception;
        	} catch (Exception exception) {
            		throw new ElisaApiException(exception.getMessage());
        	}
    	}


	/**
	 * Get from logbook configuration the list of possible message types
	 * Method that can be used without user credentials
	 * 
	 * @param username
	 * @param pwd
	 * @param ssocookiefile: path to file containing the sso cookie
	 * 
	 * @return a list of message types
	 * @throws ElisaApiException 
	 */
	public MessageTypeList getMessageTypes( String username, String pwd, String ssocookiefile)
	{
		//this is to deal with the SSL connection.
		HttpClient httpClient = HttpUtils.getNewHttpClient();
		restTemplate.setRequestFactory(new HttpComponentsClientHttpRequestFactory(httpClient));

		//the type of expected response
		ResponseEntity<MessageTypeList> response=null;				
		//if there is no user credentials, get configuration without authentication (for IGUI)
		if(username==null && ssocookiefile==null) {
			try{  		
				response=restTemplate.getForEntity(elisarestUrl + "mt",MessageTypeList.class);
				return response.getBody();  
			} catch (ElisaApiException exception) {
				throw exception;
			} catch (Exception exception) {
				throw new ElisaApiException(exception.getMessage());
			}		

		}
		else {
			//create the authentication header for the request
			MultiValueMap<String,String> headers = createAuthHeaders(username,pwd,ssocookiefile);
			HttpEntity<Object> requestEntity=new HttpEntity<Object>(headers);
			try{
				response=restTemplate.exchange(elisarestUrl + "mt",
						HttpMethod.GET,requestEntity,MessageTypeList.class);
				return response.getBody();  
			} catch (ElisaApiException exception) {
				throw exception;
			} catch (Exception exception) {
				throw new ElisaApiException(exception.getMessage());
			}		
			
		}
	}

	/**
	 * Get from logbook configuration the options for a specified message type
	 * Method that can be used without user credentials
	 *
	 * @param username
	 * @param pwd
	 * @param ssocookiefile: path to file containing the sso cookie
	 * @param msgType: a given message type
	 * @return the options of that message type 
	 * @throws Exception 
	 */
	public OptionMetadataList getMessageTypeOptions( String username, String pwd, String ssocookiefile, Enum<MessageType> msgType )
	{
		//this is to deal with the SSL connection.
		HttpClient httpClient = HttpUtils.getNewHttpClient();
		restTemplate.setRequestFactory(new HttpComponentsClientHttpRequestFactory(httpClient));

		//the type of expected response
		ResponseEntity<OptionMetadataList> response=null;

		//if there is no user credentials, get configuration without authentication (for IGUI)
		if(username==null && ssocookiefile==null) {
			try{
				response=restTemplate.getForEntity(elisarestUrl + "mt/"+msgType.toString()+"/opt",OptionMetadataList.class);
				return response.getBody(); 

			} catch (ElisaApiException exception) {
				throw exception;
			} catch (Exception exception) {
				throw new ElisaApiException(exception.getMessage());
			}		

		}
		else {
			//create the authentication header for the request
			MultiValueMap<String,String> headers = createAuthHeaders(username,pwd,ssocookiefile);
			HttpEntity<Object> requestEntity=new HttpEntity<Object>(headers);
			try{
				response=restTemplate.exchange(elisarestUrl + "mt/"+msgType.toString()+"/opt",
						HttpMethod.GET,requestEntity,OptionMetadataList.class);
				return response.getBody(); 

			} catch (ElisaApiException exception) {
				throw exception;
			} catch (Exception exception) {
				throw new ElisaApiException(exception.getMessage());
			}		
			
		}
	}

	/**
	 * Get from logbook configuration the preset SA for a specified message type
	 * Method that can be used without user credentials
	 * @param username
	 * @param pwd
	 * @param ssocookiefile: path to file containing the sso cookie
	 * @param msgType: a given message type; if null return the list of all possible systems affected
	 * @return a list of SA for that message type or a list of all possible SA 
	 * @throws Exception 
	 */
	public SystemAffectedList getSystemsAffected( String username, String pwd, String ssocookiefile, Enum<MessageType> msgType )
	{
		//this is to deal with the SSL connection.
		HttpClient httpClient = HttpUtils.getNewHttpClient();
		restTemplate.setRequestFactory(new HttpComponentsClientHttpRequestFactory(httpClient));

		//the type of expected response
		ResponseEntity<SystemAffectedList> response=null;

		//if there is no user credentials, get configuration without authentication (for IGUI)
		if(username==null && ssocookiefile==null) {
			if(msgType==null){
				try{
					response=restTemplate.getForEntity(elisarestUrl + "sa",SystemAffectedList.class);
					return response.getBody();  
				} catch (ElisaApiException exception) {
					throw exception;
				} catch (Exception exception) {
					throw new ElisaApiException(exception.getMessage());
				}		

			}
			else {
				try{
					response=restTemplate.getForEntity(elisarestUrl + "mt/"+msgType.toString()+"/sa",SystemAffectedList.class);
					return response.getBody(); 
				} catch (ElisaApiException exception) {
					throw exception;
				} catch (Exception exception) {
					throw new ElisaApiException(exception.getMessage());
				}
			}
		}
		else {
			//create the authentication header for the request
			MultiValueMap<String,String> headers = createAuthHeaders(username,pwd,ssocookiefile);
			HttpEntity<Object> requestEntity=new HttpEntity<Object>(headers);

			if(msgType==null){
				try{
					response=restTemplate.exchange(elisarestUrl + "sa",
							HttpMethod.GET,requestEntity,SystemAffectedList.class);
					return response.getBody(); 

				} catch (ElisaApiException exception) {
					throw exception;
				} catch (Exception exception) {
					throw new ElisaApiException(exception.getMessage());
				}		

			}
			else {
				try{
					response=restTemplate.exchange(elisarestUrl + "mt/"+msgType.toString()+"/sa",
							HttpMethod.GET,requestEntity,SystemAffectedList.class);
					return response.getBody(); 

				} catch (ElisaApiException exception) {
					throw exception;
				} catch (Exception exception) {
					throw new ElisaApiException(exception.getMessage());
				}	
			}
		}
	}

	/*
	 * Debug method
	 * return plain xml 
	 */
	public String getMessageTypeOptionsxml( String username, String pwd, String ssocookiefile, Enum<MessageType> msgType )
	{
		//this is to deal with the SSL connection.
		HttpClient httpClient = HttpUtils.getNewHttpClient();
		restTemplate.setRequestFactory(new HttpComponentsClientHttpRequestFactory(httpClient));

		//the type of expected response
		ResponseEntity<String> response=null;
		
		//create the authentication header for the request
		MultiValueMap<String,String> headers = createAuthHeaders(username,pwd,ssocookiefile);
		HttpEntity<Object> requestEntity=new HttpEntity<Object>(headers);
        try{
    		response=restTemplate.exchange(elisarestUrl + "mt/"+msgType.toString()+"/opt.xml",
    					HttpMethod.GET,requestEntity,String.class);
        	return response.getBody(); 
  
 	    } catch (ElisaApiException exception) {
	    	throw exception;
	    } catch (Exception exception) {
			throw new ElisaApiException(exception.getMessage());
		}		

	}

}
