/*
 * ClientThread.cpp
 *
 *  Created on: Jun 26, 2020
 *      Author: kolos
 */

#include <dlfcn.h>

#include <ers/ers.h>

#include <felix_proxy/ClientThread.h>

using namespace felix_proxy;

/*! \class felix::ImplementationException
 *  This issue is used to report problems with FELIX dynamic library handling
 */
ERS_DECLARE_ISSUE(felix_proxy, ImplementationException,
        "FELIX interface implementation library '" << lib_name << "' issue: " << reason,
        ((std::string)lib_name) ((std::string)reason))

const std::string s_libraryName("libfelix-client-lib.so");
const std::string s_factoryName("create_felix_client");

ClientThread::ClientThread(const Config &config) {
    // load shared library that implements the FelixClientThreadInterface
    // from FELIX releaseAPI from the FELIX release
    m_lib_handler = dlopen(s_libraryName.c_str(), RTLD_NOW);
    if (!m_lib_handler) {
        throw felix_proxy::ImplementationException(ERS_HERE, s_libraryName, dlerror());
    }

    // get a pointer to the function that can create a new instance
    // of the FelixClientThreadInterface implementation
    void *sym = dlsym(m_lib_handler, s_factoryName.c_str());
    if (!sym) {
        throw felix_proxy::ImplementationException(ERS_HERE, s_libraryName, dlerror());
    }

    typedef FelixClientThreadInterface* (*ImplFactory)(const Config &config);

    ImplFactory factory = reinterpret_cast<decltype(factory)>(sym);

    m_implementation.reset(factory(config));
}

ClientThread::~ClientThread() {
    dlclose(m_lib_handler);
}

void ClientThread::send_data(uint64_t elink_id, const uint8_t *data, size_t size, bool flush) {
    m_implementation->send_data(elink_id, data, size, flush);
}

void ClientThread::subscribe(uint64_t elink_id) {
    m_implementation->subscribe(elink_id);
}

void ClientThread::unsubscribe(uint64_t elink_id) {
    m_implementation->unsubscribe(elink_id);
}

void ClientThread::exec(const UserFunction &user_function) {
    m_implementation->exec(user_function);
}
