#ifndef IPC_OBJECT_H
#define IPC_OBJECT_H

/*! \file object.h Defines base classes for IPC objects.
 * \author Serguei Kolos
 * \version 1.0
 */

#include <ipc/objectbase.h>

// This include is here for backward compatibility
#include <ipc/namedobject.h>

/*! \brief Base class for the anonymous IPC objects.

    Objects of the user defined class, which inherits IPCObject will be able to get external
    CORBA requests but they can not be registered with IPC partitions and are normally passed
    to other programs via application specific IDL interface calls. The most common usage of
    such objects is for implementing the subscribe/callback pattern.
    Template parameters:
	- class T is a class which implements a certain interface which has been
	defined in IDL file. This class is generated by the IDL2C++ compiler.
	- class IP (Implementation Policy) - this class defines some details of this object implementation
	    The possibilities are:
		- ipc::multi_thread - transient object for which requests can be executed in multiple threads concurrently
		- ipc::single_thread - transient object for which requests will be serialized and executed in the context
			of a single thread

    \attention	Object of any class which inherits from this one must be allocated on heap using the new
		operator and must be destroyed via calling the _destroy() function. The delete operator
		must never be used for that purpose.
    \ingroup public
    \sa IPCNamedObject
 */
template <class T, class TP = ipc::multi_thread, class PP = ipc::transient>
struct IPCObject : public IPCObjectBase<T,TP,PP>
{
  protected:
    /*! Constructs anonymous IPC object. This object can not be registered with IPC partition.
    	A reference to this object can be passed to another application only as parameter
        of a remote call.
     */
    IPCObject ( )
    { ; }

    virtual ~IPCObject ( )
    { ; }

    /*! Constructs anonymous IPC object.This object can not be registered with IPC partition.
    	A reference to this object can be passed to another application only as parameter
        of a remote call.
    	The new object will be synchronized with another IPC object, given as parameter to
    	this constructor in a sense that all remote methods for those objects will be executed
    	sequentially.
        \param  object 	the newly created object will be synchronized with this one
     */
    IPCObject ( const IPCObject * object )
      : IPCObjectBase<T,TP,PP>( object )
    { ; }

  private:
    ///////////////////////////////////////////////////
    // Disable copying
    ///////////////////////////////////////////////////
    IPCObject ( const IPCObject & );
    IPCObject & operator=( const IPCObject & );
};

#endif
