#ifndef IPC_PARTITION_H
#define IPC_PARTITION_H

////////////////////////////////////////////////////////////////////////////
//	partition.h
//
//	header file for the IPC library
//
//	Sergei Kolos August 2003
//
//	description:
//	    This file contains declarations for the IPCPartition class
//
////////////////////////////////////////////////////////////////////////////

#include <string>
#include <list>
#include <map>
#include <iostream>

#include <partition/partition.hh>
#include <ipc/util.h>
#include <ipc/policy.h>
#include <ipc/threadpool.h>

/*! \file partition.h Defines IPCPartition class. 
 * \author Serguei Kolos
 * \version 1.0 
 */

template <class, class, class> class IPCNamedObject;

/*! \brief This class provides a set of functions for accessing objects registered
	with a certain partition.
    \ingroup public
    \sa
 */
class IPCPartition
{
    template <class, class, class> friend class IPCNamedObject;
    friend class Partition_impl;

  public:

    /*! Creates the object which represents the default partition
     */
    IPCPartition ( ) throw ()
      : m_name( ipc::partition::default_name )
    { ; }

    /*! Creates the object which represents the named partition.
        \param name partition's name
     */
    IPCPartition ( const char * name ) throw ()
      : m_name( name == 0 ? ipc::partition::default_name : name )
    { ; }

    /*! Creates the object which represents the named partition.
        \param name partition's name
     */
    IPCPartition ( const std::string & name ) throw ()
      : m_name( name )
    { ; }

    /*! Comparison operator.
     */
    bool
    operator== ( const IPCPartition & p ) const {
	return ( m_name == p.m_name);
    }

    /*! Comparison operator.
     */
    bool
    operator!= ( const IPCPartition & p ) const {
	return ( m_name != p.m_name);
    }

    /**
     * Retruns name of the current partition.
     * \return	partition name.
     * \throw	Never throws exceptions 
     */
    const std::string &
    name ( ) const
    { return m_name;}

    /**
     * Checks validity of the current partition.
     * \return	Returns true if the object, which implements current
     *		partition exists and reachable, false otherwise.
     * \throw	Never throws exceptions 
     */
    bool
    isValid ( ) const throw();

    /**
     * Returns new partition object which represents the mirror conterpart of this partition.
     */
    IPCPartition getMirror() const
    {
    	return IPCPartition(m_name + ipc::partition::mirror_postfix);
    }
    
    /**
     * Lists types of the objects registered with this partition.
     * \param list	will be filled with object type names.
     * \throw  daq::ipc::InvalidPartition	partition IPC server is unavailable, i.e. either it is not
     *					running or there are problems with network connection
     */
    void
    getTypes ( std::list< std::string > & list ) const;

    /**
     * Returns reference to the partition implementation object.
     * \return CORBA object which references the remote partition
     * \throw  daq::ipc::InvalidPartition	partition IPC server is unavailable, i.e. either it is not
     *					running or there are problems with network connection
     */
    ipc::partition_var getImplementation() const;

    /**
     * Lists objects registered with the current partition.
     * T represents the type of objects to be listed. In practice
     * such type is generated by the IDL to C++ compiler for any
     * IDL interface.
     * \param map    			This map will be filled with the available objects.
     * \throw  daq::ipc::InvalidPartition	partition IPC server is unavailable, i.e. either it is not
     *					running or there are problems with network connection
     */
    template < class T >
    void
    getObjects ( std::map< std::string , typename T::_var_type > & map ) const
    { getObjects<T, ipc::use_cache, ipc::non_existent >( map ); }

    /**
     * Checks existence of the remote object.
     * Class T represents the type of the object to be checked.
     * \param object_name		Name of the object to be checked.
     * \return				Returns true if object exists and reachable, false otherwise.
     * \throw daq::ipc::InvalidPartition	partition IPC server is unavailable, i.e. either it is not
     *					running or there are problems with network connection
     */
    template <class T>
    bool
    isObjectValid ( const std::string & object_name ) const
    { return isObjectValid<T,ipc::use_cache,ipc::non_existent >( object_name ); }

    /**
     * Returns a reference to the remote object.
     * This function never returns CORBA::Object::nil() reference.
     * Class T represents the type of the object to be resolve.
     * \param object_name		Name of the object to resolve.
     * \return				object reference
     * \throw  daq::ipc::InvalidPartition	partition IPC server is unavailable, i.e. either it is not
     *					running or there are problems with network connection
     * \throw  daq::ipc::InvalidObjectName	object's name is invalid.
     * \throw  daq::ipc::ObjectNotFound	target object does not exist or is not registered with the
     *					partition server.
     */
    template < class T >
    typename T::_ptr_type
    lookup ( const std::string & object_name ) const
    { return lookup<T, ipc::use_cache, ipc::narrow >( object_name ); }

    /**
     * Enumerates remote objects of a certain type.
     * Class T represents the type of the objects to be listed.
     * Class RP defines the resolution policy. This policy defines how object resolution will be done. 
     * There are two implementations of this policy defined in the IPC library:
     * ipc::use_cache and ipc::no_cache.
     * Class VP defines the validation policy. This policy defines how to check the validity of the remote object reference.
     * There are three implementations of this policy defined in the IPC library: 
     * ipc::unchecked_narrow, ipc::narrow and ipc::non_existent.
     * \param map			standard library map to be filled with the objects references
     * \param type_name			type name of the objects which have to be enumerated. This parameter has 
     *					default value which is calculated from the objects type T (the first template
     *                                  parameter for this function). One can use another type name in case one
     *                                  wants to enumerate the objects whose corresponding C++ type is not known
     *                                  at compilation time.
     * \throw  daq::ipc::InvalidPartition	partition IPC server is unavailable, i.e. either it is not
     *					running or there are problems with network connection
     */    
    template < 	class T,
		template < class , template < class > class > class RP,
		template < class > class VP >
    void
    getObjects ( std::map< std::string, typename T::_var_type > & map,
		 const std::string & type_name = ipc::util::getTypeName<T>() ) const
    {
	CosNaming::BindingList_var	bl;

	ipc::util::list( m_name, type_name, bl );

	if ( bl.operator->() == 0 )
	    return;

        boost::mutex mutex;
        size_t num_threads = bl->length() > 100 ? 100 : bl->length();
	IPCThreadPool threadpool( num_threads );
        for ( unsigned int i = 0; i < bl->length(); i++ ) {
	    std::string object_name = (const char *)bl[i].binding_name[0].id;
	    threadpool.addJob(	boost::bind( IPCPartition::helper<T, RP, VP>(), 
                                		boost::ref(*this), boost::ref(type_name), 
                                                object_name,
                				boost::ref(map), boost::ref(mutex) ) );
	}
    }

    /**
     * Checks existence of the remote object.
     * Class T represents the type of the objects to be checked.
     * Class RP defines the resolution policy. This policy defines how object resolution will be done. 
     * There are two implementations of this policy defined in the IPC library:
     * ipc::use_cache and ipc::no_cache.
     * Class VP defines the validation policy. This policy defines how to check the validity of the remote object reference.
     * There are three implementations of this policy defined in the IPC library: 
     * ipc::unchecked_narrow, ipc::narrow and ipc::non_existent.
     * \param object_name		Name of the object to be checked.
     * \param type_name			type name of the object which has to be checked. This parameter has 
     *					default value which is calculated from the objects type T (the first template
     *                                  parameter for this function). One can use another type name in case one
     *                                  wants to enumerate the objects whose corresponding C++ type is not known
     *                                  at compilation time.
     * \return				Returns true if object exists and reachible, false otherwise.
     * \throw daq::ipc::InvalidPartition	partition IPC server is unavailable, i.e. either it is not
     *					running or there are problems with network connection
     * \throw  daq::ipc::InvalidObjectName	object's name is invalid.
     */
    template < 	class T,
		template < class , template < class > class > class RP,
		template < class > class VP >
    bool
    isObjectValid (	const std::string & object_name,
			const std::string & type_name = ipc::util::getTypeName<T>() ) const
    {
	try {
	    typename T::_var_type obj = lookup< T, RP, VP >( object_name, type_name );
	    return true;
        }
        catch(daq::ipc::ObjectNotFound &) {
            return false;
        }
    }

    /**
     * Returns a reference to the remote object.
     * This function never returns CORBA::Object::nil() reference.
     * Class T represents the type of the object to be resolved.
     * Class RP defines the resolution policy. This policy defines how object resolution will be done. 
     * There are two implementations of this policy defined in the IPC library:
     * ipc::use_cache and ipc::no_cache.
     * Class VP defines the validation policy. This policy defines how to check the validity of the remote object reference.
     * There are three implementations of this policy defined in the IPC library: 
     * ipc::unchecked_narrow, ipc::narrow and ipc::non_existent.
     * \param object_name		Name of the object to resolve.
     * \param type_name			type name of the object which has to be resolved. This parameter has 
     *					default value which is calculated from the objects type T (the first template
     *                                  parameter for this function). One can use another type name in case one
     *                                  wants to enumerate the objects whose corresponding C++ type is not known
     *                                  at compilation time.
     * \return				object reference
     * \throw  daq::ipc::InvalidPartition	partition IPC server is unavailable, i.e. either it is not
     *					running or there are problems with network connection
     * \throw  daq::ipc::InvalidObjectName	object's name is invalid.
     * \throw  daq::ipc::ObjectNotFound	target object does not exist or is not registered with the
     *					partition server.
     */
    template < class T, template < class , template < class > class > class RP, template < class > class VP >
    typename T::_ptr_type
    lookup (	const std::string & object_name,
		const std::string & type_name = ipc::util::getTypeName<T>() ) const
    { 
    	typename T::_ptr_type object = RP<T, VP>::template resolve<ipc::util::resolve>( m_name , object_name, type_name );
        if ( CORBA::is_nil(object) ) {
            throw daq::ipc::ObjectNotFound( ERS_HERE, m_name, type_name, object_name );
        }
        return object;
    }

    /**
     * Enumerates all available partitions except the default one
     * \param list 	will be filled with partitions.
     * \throw		Never throws exceptions 
     */
    static void
    getPartitions ( std::list< IPCPartition > & list ) throw();

    /**
     * Lists all available partitions except the default one
     * Class T represents the type of the object to be constructed.
     * Class RP defines the resolution policy. This policy defines how object resolution will be done. 
     * There are two implementations of this policy defined in the IPC library:
     * ipc::use_cache, ipc::no_cache.
     *
     * \param list	will be filled with partitions.
     * \throw		Never throws exceptions 
     */
    template <template < class , template < class > class > class RP>
    static void
    getPartitions ( std::list< IPCPartition > & list, short = 0 ) throw ()
    {
	CosNaming::BindingList_var	bl;

	try {
	    ipc::util::list ( ipc::partition::default_name, "", bl );
	}
	catch( daq::ipc::InvalidPartition & ex ) {
            return;
	}

	if ( bl.operator->() == 0 )
	    return;

	for ( unsigned int i = 0; i < bl->length(); i++ ) {
	    if ( bl[i].binding_type == CosNaming::ncontext )
		continue;

	    try {
		std::string name = (const char *)bl[i].binding_name[0].id;
                ipc::partition_var ip = RP< ipc::partition,
					    ipc::non_existent >::template resolve<ipc::util::resolvePartition>( name );
		if ( !CORBA::is_nil(ip) ) {
		    list.push_back( IPCPartition( name ) );
                }
            }
            catch( daq::ipc::InvalidPartition & ex ) {
            	continue ;
            }
	}
	list.sort();
    }

  private:
    template < class T, template < class , template < class > class > class RP, template < class > class VP >
    struct helper {
	typedef void result_type;
        
	void operator()( const IPCPartition & partition,
			 const std::string & type,
			 const std::string & name,
			 std::map< std::string, typename T::_var_type > & map,
			 boost::mutex & mutex )
	{
	    try {
		typename T::_var_type object = partition.lookup<T,RP,VP>( name, type );
		boost::mutex::scoped_lock lock( mutex );
		map.insert( std::make_pair( name, object ) );
	    }
	    catch( ... ) {
		return ;
	    }
	}
    };
    
  private:
    void
    publish  (	const std::string & object_name,
		const std::string & type_name,
		const ipc::servant_ptr obj ) const;

    void
    withdraw (	const std::string & object_name,
		const std::string & type_name ) const;

  private:
    std::string m_name;
};

inline bool operator<( const IPCPartition & p1, const IPCPartition & p2 )
{
    return ( p1.name() < p2.name() );
}

#endif // IPC_PARTITION_H
