#ifndef MDA_ARCHIVEHISTOGROUP_H
#define MDA_ARCHIVEHISTOGROUP_H
//--------------------------------------------------------------------------
// File and Version Information:
// 	$Id$
//------------------------------------------------------------------------

//-----------------
// C/C++ Headers --
//-----------------
#include <map>
#include <memory>

//----------------------
// Base Class Headers --
//----------------------

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "mda/archive/DBArchive.h"
#include "mda/archive/MDAReceiver.h"
#include "mda/archive/OHSProvider.h"
#include "oh/OHStream.h"

//------------------------------------
// Collaborating Class Declarations --
//------------------------------------
namespace daq {
namespace mda {
class Predicate;
}
}

//		---------------------
// 		-- Class Interface --
//		---------------------

namespace daq {
namespace mda {

/// @addtogroup mda_archive

/**
 *  @ingroup mda_archive
 *
 *  @brief Class that archives histograms from a single histogram group.
 *
 *  This software was developed for the ATLAS project.  If you use all or 
 *  part of it, please give an appropriate acknowledgment.
 *
 *  @version $Id$
 *
 *  @author Federico Zema
 *  @author Andy Salnikov
 */

class ArchiveHistoGroup  {
public:

    /// Type for histogram list generated by server/provider.
    typedef std::map<OHSProvider, DBArchive::HistoIdLBList> Provider2HistoList;

    /**
     *  @brief  Standard constructor.
     *  
     *  @param[in] name    Histogram group name (an ID in OKS config)
     *  @param[in] partition Partition name
     *  @param[in] ohsServer Name of OH server
     *  @param[in] provider  OH provider name filter (regexp)
     *  @param[in] histograms Histogram name filter (regexp) 
     *  @param[in] versions  Number of histogram versions to retrieve/store.
     *  @param[in] dbArchive Database access object.
     */
    ArchiveHistoGroup (const std::string& name,
                       const std::string& partition,
                       const std::string& ohsServer,
                       const std::string& provider,
                       const std::string& histograms,
                       int versions,
                       const std::shared_ptr<DBArchive>& dbArchive) ;
    
    // Destructor
    ~ArchiveHistoGroup () throw () ;
    
    /// Reset current state, close/remove stream
    void reset ();
    
    /**
    *  @brief  Start reading histograms from OH server.
    *  
    *  @param[in] cacheHistos If true then read everything and store 
    *             all histograms in memory
    */
    void readHistos (bool cacheHistos);
    
    /**
    *  @brief Archive all histograms, writes them into ROOT file and stores
    *  the list of histograms in database.
    *  
    *  @param[in] mdaReceiver  Object which receives histograms from stream and
    *                          writes them to ROOT file.
    *  @param[out] histoLists  Object returning information about stored histograms.
    *  @param[in] stopRequested Predicate for checking whether the stop was requested
    *  @return Number of histograms stored
    */
    unsigned archive (MDAReceiver& mdaReceiver,
                      Provider2HistoList& histoLists,
                      const Predicate& stopRequested);
    
    /// Return the name of this histogram group
    const std::string& name () const {return m_name;}

protected:

private:

    // Data members
    std::string m_name;           ///< Histogram group name
    std::string m_partition;      ///< Partition name
    std::string m_ohsServer;      ///< OH server name
    std::string m_provider;       ///< Provider name
    std::string m_histograms;     ///< Histogram name filter (regexp)
    int m_versions;               ///< Number of histogram versions to store 
    OHStream *m_ohStream;         ///< OH stream object
    std::shared_ptr<DBArchive> m_dbArchive;     ///< Database access object

};

} // namespace mda
} // namespace daq

#endif // MDA_ARCHIVEHISTOGROUP_H
