#ifndef MDA_MDACONFIG_H
#define MDA_MDACONFIG_H
//--------------------------------------------------------------------------
// File and Version Information:
//      $Id$
//------------------------------------------------------------------------

//-----------------
// C/C++ Headers --
//-----------------
#include <string>
#include <memory>
#include <boost/utility.hpp>

//----------------------
// Base Class Headers --
//----------------------

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "ers/ers.h"
#include "mda/archive/Archiver.h"
#include "mda/archive/DBArchive.h"

//------------------------------------
// Collaborating Class Declarations --
//------------------------------------

//              ---------------------
//              -- Class Interface --
//              ---------------------

namespace daq {
namespace mda {

    /**
     *  @class MDAConfigIssues::Exception
     *  @brief Base for exception classes generated by MDAConfig class.
     */
    ERS_DECLARE_ISSUE (MDAConfigIssues, Exception, , )

    /**
     *  @class MDAConfigIssues::EnvVarNotDefined
     *  @brief Issue generated when environment does not have required variable defined.
     */
    ERS_DECLARE_ISSUE_BASE (MDAConfigIssues, EnvVarNotDefined, Exception,
            "environment variable " << envvar << " is not defined",
            , ((std::string) envvar))

    /**
     *  @class MDAConfigIssues::ConfigurationNotFound
     *  @brief Issue generated when configuration object does not exist.
     */
    ERS_DECLARE_ISSUE_BASE (MDAConfigIssues, ConfigurationNotFound, Exception,
            "MDAApplication not found in conf db: " << name,
            , ((std::string) name))

    /**
     *  @class MDAConfigIssues::CantArchive
     *  @brief Issue generated when archiver cannot be instantiated.
     */
    ERS_DECLARE_ISSUE_BASE (MDAConfigIssues, CantArchive, Exception,
            "Histograms can not be archived", , )

    /**
     *  @class MDAConfigIssues::PartitionNotFound
     *  @brief Exception produced when partition does not exist.
     */
    ERS_DECLARE_ISSUE_BASE (MDAConfigIssues, PartitionNotFound, Exception,
            "partition " << partition << " not found in configuration db",
            , ((std::string) partition))

    /**
     *  @class MDAConfigIssues::ConfigError
     *  @brief Exception produced for low-level configuration issues.
     */
    ERS_DECLARE_ISSUE_BASE (MDAConfigIssues, ConfigError, Exception,
            "exception was generated by configuration db", ,)

      
/// @addtogroup mda_archive

/**
 *  @ingroup mda_archive
 *
 *  @brief Class that creates and configures necessary archiver objects.
 *
 *  This software was developed for the ATLAS project.  If you use all or 
 *  part of it, please give an appropriate acknowledgment.
 *
 *  @version $Id$
 *
 *  @author Federico Zema
 *  @author Andy Salnikov
 */
class MDAConfig : boost::noncopyable {
public:
    
    /**
     *  @brief Read configuration and instantiate archiver.
     *  
     *  @param[in] appName Application name.
     *  
     *  @throw MDAConfigIssues::EnvVarNotDefined
     *  @throw MDAConfigIssues::PartitionNotFound
     *  @throw MDAConfigIssues::ConfigurationNotFound
     *  @throw MDAConfigIssues::CantArchive
     *  @throw MDAConfigIssues::ConfigError
     */
    explicit MDAConfig(const std::string& appName);
    
    // Destructor
    ~MDAConfig() throw();

    /// Return minimum run duration to consider for archiving
    unsigned minRunDuration() const { return m_minRunDuration; }
    
    /// Return max time to wait until archive starts arching (reads all data from OHS)
    unsigned initTimeout() const { return m_initTimeout; }
    
    /// Return archiver object
    std::shared_ptr<Archiver> archiver() const { return m_archiver; }

private:
    
    unsigned m_minRunDuration;  ///< minimum run duration to consider for archiving
    unsigned m_initTimeout;  ///< max time to wait until archive starts arching
    std::shared_ptr<Archiver> m_archiver; ///< archiver object
    std::shared_ptr<DBArchive> m_db;  ///< Database object
};

} // namespace mda
} // namespace daq

#endif // MDA_MDACONFIG_H
