#ifndef MDA_ROOTFILE_H
#define MDA_ROOTFILE_H
//--------------------------------------------------------------------------
// File and Version Information:
//      $Id$
//------------------------------------------------------------------------

//-----------------
// C/C++ Headers --
//-----------------
#include <TFile.h>

//----------------------
// Base Class Headers --
//----------------------

//-------------------------------
// Collaborating Class Headers --
//-------------------------------
#include "ers/ers.h"
#include "mda/common/Issues.h"

//------------------------------------
// Collaborating Class Declarations --
//------------------------------------

//              ---------------------
//              -- Class Interface --
//              ---------------------

namespace daq {
namespace mda {
        
    /**
     *  @class RootFileIssues::Exception
     *  @brief Base for exception classes generated by RootFile class.
     */
    ERS_DECLARE_ISSUE_BASE (RootFileIssues, Exception, Issues::Exception, , ,) 

    /**
     *  @class RootFileIssues::FileExists
     *  @brief Issue generated when opening a file which already exists.
     */
    ERS_DECLARE_ISSUE_BASE (RootFileIssues, FileExists, Exception,
            "cannot create file " << file << ": it already exists", , ((std::string) file)) 
    
    /**
     *  @class RootFileIssues::FileException
     *  @brief Generic issue for file-related operations.
     */
    ERS_DECLARE_ISSUE_BASE(RootFileIssues, FileException, Exception,
            "; possible error: " << strerror (error), , ((int) error)) 

    /**
     *  @class RootFileIssues::CannotOpenFile
     *  @brief Issue generated for failures in opening new file.
     */
    ERS_DECLARE_ISSUE_BASE (RootFileIssues, CannotOpenFile, FileException,
            "cannot open file: " << file << " (option: " << option
            << ", compression level: " << compr_level << ")",
            ((int) error), ((std::string) file) ((std::string) option)
            ((int) compr_level)) 
    
    /**
     *  @class RootFileIssues::ErrorOnFileClose
     *  @brief Issue generated for failures while closing file.
     */
    ERS_DECLARE_ISSUE_BASE(RootFileIssues, ErrorOnFileClose, FileException,
            "error detected while closing file: " << file, ((int) error), ((std::string) file)) 
    
    /**
     *  @class RootFileIssues::CannotMakeDir
     *  @brief Issue generated for failures while creating new directory
     *  in a ROOT file.
     */
    ERS_DECLARE_ISSUE_BASE (RootFileIssues, CannotMakeDir, FileException,
            "cannot create directory `" << new_dir << "' under: " << cur_dir,
            ((int) error), ((std::string) new_dir) ((std::string) cur_dir)) 

    /**
     *  @class RootFileIssues::CannotWrite2File
     *  @brief Issue generated for failures while writing data to a ROOT file.
     */
    ERS_DECLARE_ISSUE_BASE (RootFileIssues, CannotWrite2File, FileException,
            "cannot write object `" << object << "' to file: " << file,
            ((int) error), ((std::string) object) ((std::string) file)) 

/// @addtogroup mda_archive

/**
 *  @ingroup mda_archive
 *
 *  @brief Wrapper class for ROOT files automating some operations.
 *
 *  This software was developed for the ATLAS project.  If you use all or 
 *  part of it, please give an appropriate acknowledgment.
 *
 *  @version $Id$
 *
 *  @author Federico Zema
 *  @author Andy Salnikov
 */

class RootFile {
public:
    
    /**
     *  @brief Standard constructor
     *  
     *  @param[in] name   File name
     *  @param[in] option TFile options, see TFile constructor
     *  @param[in] title  TFile title, see TFile constructor
     *  @param[in] compLevel TFile compression level
     *  
     *  @throw RootFileIssues::FileExists Thrown if file already exists.
     *  @throw RootFileIssues::CannotOpenFile Thrown for file opening problems.
     */
    explicit RootFile (const std::string& name, const std::string &option = "",
              const std::string &title = "", int compLevel = 1);

    // Destructor
    ~RootFile () throw ();

    /**
     *  @brief Close the file.
     *  
     *  @throw RootFileIssues::ErrorOnFileClose
     */
    void close ();

    /**
     *  @brief Store ROOT object in a specified directory.
     *  
     *  @em WARNING: object name is used as relative path and modified. 
     *  Suppose that object name "Test/Histo" and directory was "/EXPERT/TRT"
     *  then after call to this method the object name would be 
     *  "Histo" and it will be stored in a directory "/EXPERT/TRT/Test".
     *  
     *  @param[in,out] object   ROOT object to store, its name is modified.
     *  @param[in]  dir   Directory name where to store an object.
     *  
     *  @throw RootFileIssues::CannotMakeDir
     *  @throw RootFileIssues::CannotWrite2File
     */
    void write (TNamed &object, const std::string &dir = "");

    /**
     *  @brief Store ROOT object in a specified directory.
     *
     *  This method is used with non-named TObjects.
     *
     *  @param[in]  object  ROOT object to store, its name is modified.
     *  @param[in]  name    ROOT object name to use for storing.
     *  @param[in]  dir     Directory name where to store an object.
     *  @param[in]  option  ROOT option, passed to TObject::Write().
     *
     *  @throw RootFileIssues::CannotMakeDir
     *  @throw RootFileIssues::CannotWrite2File
     */
    void writeTObject(TObject& object, const std::string& name,
                      const std::string& dir = "", int option=0);

    /**
     *  @brief Check if the object with the given name already exists.
     *
     *  @param[in]  path   Fill path name of the object.
     *  @return   True if the object exists, false otherwise.
     */
    bool exists(const std::string& path);

private:

    /**
     *  @brief Helper method which creates directory hierarchy and
     *  changes current directory to it.
     *  
     *  @param[in]  dir   Directory name to create.
     *  
     *  @throw RootFileIssues::CannotMakeDir
     */
    void mkdir_cd (const std::string &dir);

    // Data objects
    TFile* m_file;  ///< Pointer to an underlying ROOT TFile object.
};

} // namespace mda
} // namespace daq

#endif // MDA_ROOTFILE_H
