/*!	\file defs.h
 	\brief Common data structures for MTS package.
  	\author A. Kazarov Copyright 2012 CERN. All rights reserved.
	\date   March 2013
 */

/**
\mainpage (DAQ/HLTI Software Documentation: doxygen for release packages)
*/

#ifndef MTS_DEFS_H
#define MTS_DEFS_H

#include <chrono>

#include <tbb/concurrent_queue.h>

#include "mts/mts.hh"

namespace daq {
namespace mts {

/**
  *  @defgroup mts Package MTS: Message Transfer System.
  *
  *  @brief This package provides API and utilities used by TDAQ applications
  *  to exchange messages using send/ubscribe/nofify pattern.
  *
  *  It is a re-implementation of the former MRS package. The key difference is that sending API is hidden from users by ERS layer.
  *  See also: https://atlas-tdaq-sw.web.cern.ch/atlas-tdaq-sw/doxygen/tdaq-common/production/html/main.html
  *
  *  Detailed design documentation can be found here:
  *  https://its.cern.ch/jira/secure/attachment/14921/HLD_draft_v0.4.pdf
  *
  *  Corresponding Jira project, for bug reporting etc:
  *  https://its.cern.ch/jira/browse/ATDAQCCMRS
  *
  */


/** \page MTS_package Message Transfer System (MTS) user documentation

   [Detailed design documentation (PDF)](https://its.cern.ch/jira/secure/attachment/14921/HLD_draft_v0.4.pdf)

   Using MTS
   ==============

   Functionality of MTS: sending and subscribing to messages is provided in Java and C++.

   ### Subscribing to messages

   %Subscription is done using ERS API. However the syntax for selecting the messages is MTS-specific:

   %Subscription criteria syntax: see daq::mts::Subscription and also detsails on page 7 of https://its.cern.ch/jira/secure/attachment/14921/HLD_draft_v0.4.pdf

   \see daq::mts::Subscription

   %Examples:

   C++\n
   mts/examples/mts_ers_receiver.cpp

   Java\n
   mts/jsrc/test/TestReceiver.java

   ### Sending messages

   The sending part of MTS is completely hidden from the end users by the ERS layer. See more details:

   C++: https://atlas-tdaq-sw.web.cern.ch/atlas-tdaq-sw/doxygen/tdaq-common/production/html/main.html

   Java: https://atlas-tdaq-sw.web.cern.ch/atlas-tdaq-sw/javadoc/tdaq-05-01-00/ers/package-summary.html#package_description

  %Examples:

  C++\n
  mts/examples/ers_sender.cc

  Java\n
  mts/jsrc/test/TestReporter.java

  ### Configuration of ERS stream for MTS

The name of ERS stream implmented by MTS is 'mts'. It accepts partition name as a parameter, e.g.
\code
  >export TDAQ_ERS_ERROR='mts,lstderr'
  >export TDAQ_ERS_FATAL='mts(initial),lstderr'
\endcode

  ### Bugs and discussions:

   https://its.cern.ch/jira/browse/ATDAQCCMRS

*/

typedef ::mts::Issue	 					Message ;	/*!< IDL mts::Issue, see idl/mts.idl */
typedef std::shared_ptr< ::mts::Issue > 	MessageSP ; /*!< shared pointer to Message */
typedef ::mts::receiver_var 				Receiver ;  /*!< shared CORBA pointer to mts::receiver, see idl/mts.idl */
typedef ::mts::receiver_ptr 				ReceiverP ; /*!< raw CORBA pointer to mts::receiver */

// we use TBB implementation of non-blocking thread-safe double-end queue
// to decouple providers from receivers in multithreaded code
typedef	tbb::concurrent_queue< MessageSP > MessagesQueue ; /*!< Queue of pointers to Message */


typedef std::function<void()> Task ; 	// just a helper class

/*!
 A helper template function to return a time required for execution of a method
 \param T template parameter, time precision, e.g. std::chrono::microseconds std::chrono::milliseconds
 \param f a Task
 */
template <typename T>
int timed_execute(Task f)
{
auto start = std::chrono::high_resolution_clock::now() ;
f() ;
auto end = std::chrono::high_resolution_clock::now() ;
return std::chrono::duration_cast<T>(end-start).count() ;
}

}} // daq::mts
#endif
