#include <thread>
#include <chrono>
#include <atomic>

#include <ers/ers.h>
#include <ers/RemoteContext.h>
#include <ers/AnyIssue.h>

#include "mts/Receiver.h"
#include "mts/Subscription.h"

// register mts::Receiver as ERS input stream
// for use from client code, see examples/mts_ers_receiver
ERS_REGISTER_INPUT_STREAM( daq::mts::ERSReceiver, "mts", format )

// declaration, see SubsValidator.cpp
namespace daq { namespace mts {
bool ext_assert_validity(const std::string& expr, std::string& ret_error) ; }}

using namespace daq::mts ;
using namespace std::chrono ;

ERSReceiver::ERSReceiver( const std::initializer_list<std::string>& args ) 
{
std::initializer_list<std::string>::iterator argsit = args.begin() ;
std::string filter, partition ;
if ( args.size() == 1 ) // one argument
	{
	filter = *argsit ;
	}
else 	{		// two arguments: TODO: exception if 2+ arguments?
	partition = *argsit++ ;
	filter = *argsit ;
	}

assert_validity( filter ) ; // throw ::mts::BadExpression

ERS_DEBUG(1, "Constructing MTS CORBAReceiver in partition " << partition) ;
m_receiver = new CORBAReceiver( partition, this ) ; // this is CORBA object, shall not throw!

// SubscriptionManager::subscribe
::mts::receiver_var rec = m_receiver->_this() ;
m_receiver->subscribe(IPCCore::objectToString(rec), filter) ; // throw SubscriptionRegistrationFailure
ERS_DEBUG(1, "Constructed ERSReceiver") ;
}

// WARNING: calling _destroy(true) waits for all the buffered callbacks to be processed, 
// which may take time
ERSReceiver::~ERSReceiver( ) 
{
ERS_DEBUG(1, "Unsubscribing MTS receiver") ;
m_receiver->remove_subscriptions() ;		// try to be graceful
ERS_DEBUG(1, "Destroying MTS (CORBA) Receiver") ;
m_receiver->_destroy(true) ; 				// wait for all active callbacks around, then destructor (empty) of Receiver called
ERS_DEBUG(2, "Corba objects destroyed") ;
}

CORBAReceiver::CORBAReceiver( const std::string & partition, ERSReceiver* parent) // throw SubscriptionRegistrationFailure
:SubscriptionManager(partition),
m_ers_receiver(parent),
m_message_counter(0)
{
ERS_DEBUG(1, "Constructed MTS CORBAReceiver") ;
}

CORBAReceiver::~CORBAReceiver() {
ERS_DEBUG(1, "Destroyed MTS CORBAReceiver") ;
}

// to be called from parent destructor before IPC _destroy
void CORBAReceiver::remove_subscriptions( )
{
	//SubscriptionManager::unsubscribe()
	unsubscribe_receiver(m_my_is_name) ;
}

// implementation of CORBA receiver::notify interface
void CORBAReceiver::notify(const ::mts::Issue& message) 
{
ERS_DEBUG(3, "Got MTS message " << message.ID) ;
m_message_counter++ ;
const ers::Issue* issue =  mts2ersIssue(message) ;

ERS_DEBUG(5, "Built ERS issue " << issue->get_class_name()) ;

// calls ers InputStream receive function, which calls user's callback
// NB: may be "slow", and cause CORBA timeouts on workers
m_ers_receiver->notify( *issue ) ;
ERS_DEBUG(3, "User ERS receiver notified " << message.ID) ;

delete issue ; // user has to copy
}

ers::Severity	mts2ers(::mts::Severity mtssev)
{
switch ( mtssev )
	{
	case ::mts::Severity::debug: 		return ers::severity::Debug; break;
	case ::mts::Severity::log: 			return ers::severity::Log; break;
	case ::mts::Severity::information: 	return ers::severity::Information; break;
	case ::mts::Severity::warning: 		return ers::severity::Warning; break;
	case ::mts::Severity::error: 		return ers::severity::Error; break;
	default: 				return ers::severity::Fatal;
	} 
}

// service function used to match a value to vector of ERS qualifiers
const char* sev2string(::mts::Severity erssev)
{
switch ( erssev )
	{
	case ::mts::Severity::debug: 		return "DEBUG"; break;
	case ::mts::Severity::log: 			return "LOG"; break;
	case ::mts::Severity::information: 	return "INFORMATION"; break;
	case ::mts::Severity::warning: 		return "WARNING"; break;
	case ::mts::Severity::error: 		return "ERROR"; break;
	default: 							return "FATAL";
	}
}
void
ERSReceiver::assert_validity(const std::string& expr) 	// throws ::mts::BadExpression if parsing fails
{
std::string reason ;
if ( !ext_assert_validity(expr, reason) )
	{
	throw InvalidExpression( ERS_HERE, reason.c_str() ) ;
	}
}

// internal function for recursive build of chained ers::Issue from mts::Issue
ers::Issue * CORBAReceiver::mts2ersIssue(const ::mts::Issue& message) 
{
ers::Issue *issue = 0, *cause = 0 ;

std::vector<std::string> qualifiers ;
for ( size_t i = 0; i < message.quals.length(); i++ )
	{ qualifiers.push_back(std::string(message.quals[i])) ; }

ers::string_map parameters ;
for ( size_t i = 0; i < message.params.length(); i++ )
		{ parameters[std::string(message.params[i].name)] = message.params[i].value ; }

ers::RemoteContext rc(	std::string(message.cont.package_name), std::string(message.cont.file_name),
			message.cont.line_number, std::string(message.cont.function_name),
			ers::RemoteProcessContext( std::string(message.cont.host_name),
			message.cont.process_id, message.cont.thread_id, std::string(message.cont.cwd),
			message.cont.user_id, std::string(message.cont.user_name), std::string(message.cont.application_name))) ;

if ( message.cause.length() ) // there is cause issue
	{
ERS_DEBUG(3, "Building nested issue " << message.cause[0].ID) ;
	cause = mts2ersIssue(message.cause[0]) ;
	issue = new ers::AnyIssue (std::string(message.ID), mts2ers(message.sev), rc,
			system_clock::time_point(microseconds(message.time)), std::string(message.message), qualifiers, parameters, cause) ;
	}
else
	{
	issue = new ers::AnyIssue (std::string(message.ID), mts2ers(message.sev), rc, 
			system_clock::time_point(microseconds(message.time)), std::string(message.message), qualifiers, parameters) ;

/*	auto p = std::make_exception_ptr(ers::AnyIssue (std::string(message.ID), mts2ers(message.sev), rc, 
			system_clock::time_point(microseconds(message.time)), std::string(message.message), qualifiers, parameters)) ;

	try { std::rethrow_exception (p) ; }
	catch ( const ers::Issue& ex )
		{
		}
*/

	}

// does not work, issue not registered in factory
// issue = ers::IssueFactory::instance().create( std::string(message.ID), rc, message.time, std::string(message.message), qualifiers, parameters ) ;

issue->set_severity(mts2ers(message.sev)) ;

ERS_DEBUG(3, "Built issue " << message.ID) ;
return issue ;
}

