#include <chrono>

#include "mts/Sender.h"

/* CORBA Timeouts in omniORB
namespace omniORB {
void setClientCallTimeout(CORBA::ULong millisecs);
void setClientCallTimeout(CORBA::Object_ptr obj, CORBA::ULong millisecs);
void setClientThreadCallTimeout(CORBA::ULong millisecs);
void setClientConnectTimeout(CORBA::ULong millisecs);
}; */

// oneCallPerConnection = 0 
/* To tell the ORB that it may multiplex calls on a single connection, set the
oneCallPerConnection parameter to zero. If the oneCallPerConnection
parameter is set to the default value of one, and there are more concurrent calls than specified by
maxGIOPConnectionPerServer, calls block waiting for connections to become free.
*/

/*
catch (CORBA::TRANSIENT& ex) {
if (ex.minor() == omni::TRANSIENT_ConnectFailed) {
// retry with a different object reference...
}
else {
// print an error message...
}
*/

using namespace daq::mts ;
using namespace std::chrono ;

Sender::Sender( const std::string & p, size_t nthreads) 
:WorkersManager(p), m_messages_queued(0), m_messages_sent(0)
{
ERS_DEBUG(1, "Initializing MTS sender"); 

if ( nthreads )
	{ m_thrpool = std::make_unique<ThreadPoolNQ>(nthreads, *this) ; }

ERS_DEBUG(2, "Initialized MTS Sender") ;

// possibly: get list of available MTS workers here and ever
// or: the list of workers is updated each time get_worker() is called
// it is assumed that IPC caching works
}

Sender::~Sender()
{
ERS_DEBUG(1, "Destroying Sender") ;
} 

mts::Severity	ers2mts(ers::severity erssev) ;

// internal function for recursive build of chained mts::Issue from ers::Issue
// message is already allocated by corba constructor
void ers2mtsIssue(::mts::Issue& message, const ers::Issue& issue) 
{
ERS_DEBUG(3, "Building message: " << issue.get_class_name()) ;
	message.ID = issue.get_class_name() ;
	message.message = issue.message().c_str() ;
	message.sev = ers2mts(issue.severity()) ;
	// message.time = issue.time_t() ;
	// long long is 64bit, sufficient for microseconds
	message.time = duration_cast<microseconds>(issue.ptime().time_since_epoch()).count() ;

	auto quals = issue.qualifiers() ;
	message.quals.length(quals.size()) ;
	size_t i = 0 ;
	for ( auto a_qual: quals )
		{ message.quals[i++] = a_qual.c_str() ; }

	auto params = issue.parameters() ;
ERS_DEBUG(5, "Message has " << params.size() << " parameters") ;
	message.params.length(params.size()) ;
	i = 0 ;
	for ( auto a_param: params )
		{
ERS_DEBUG(5, "Adding parameter to message: " << a_param.first.c_str() << "=" << a_param.second.c_str()) ;
		message.params[i].name = a_param.first.c_str() ;
 		message.params[i].value = a_param.second.c_str() ;
		i++ ;
 		}

	message.cont.cwd = issue.context().cwd() ;
	message.cont.file_name = issue.context().file_name() ;
	message.cont.function_name = issue.context().function_name() ;
	message.cont.host_name = issue.context().host_name() ;
	message.cont.package_name = issue.context().package_name() ;
	message.cont.user_name = issue.context().user_name() ;
	message.cont.application_name = issue.context().application_name() ;
	message.cont.user_id = issue.context().user_id() ;
	message.cont.process_id = issue.context().process_id() ;
	message.cont.thread_id = issue.context().thread_id() ;
	message.cont.line_number = issue.context().line_number() ;

	if ( issue.cause() )
		{
		message.cause.length(1) ; // allocate cause message
ERS_DEBUG(5, "Building chained message: " << issue.cause()->get_class_name()) ;
		ers2mtsIssue(message.cause[0], *issue.cause()) ; // fill it  
		}

ERS_DEBUG(5, "Completed building message") ;
ERS_DEBUG(5, "Message size: " << sizeof(message)) ;
}

// build chained mts::Issue and place it in a queue which is managed by a threadpool
void Sender::report_nonblock(const ers::Issue & issue) 
{
ERS_DEBUG(1, "Sending issue (async) " << issue.get_class_name()) ;
::mts::Issue* message = new ::mts::Issue() ; // allocate root message
ers2mtsIssue(*message, issue) ; // build chained

// enqueue shared pointer to just created structure
// TODO: handle local buffer overallocation? use bounded queue
ERS_DEBUG(3, "Scheduling async message sending") ;
m_messages_queue.push(std::shared_ptr< ::mts::Issue >(message)) ;
m_messages_queued++ ;
assert (m_thrpool != nullptr) ;
// tbb::atomic_fence() ;	 // guarantee that push is done BEFORE try_pop from another thread ?
// ERS_DEBUG(4, "Awaking one thread") ;
m_thrpool->awake_one() ; // awake one worker thread
}

// build chained mts::Issue and call remote method directly
void Sender::report_block(const ers::Issue & issue) 
{
ERS_DEBUG(1, "Sending issue (sync) " << issue.get_class_name()) ;
::mts::Issue* message = new ::mts::Issue() ; // allocate root message
ers2mtsIssue(*message, issue) ; // build chained

report_remote(std::shared_ptr< ::mts::Issue >(message)) ;
}

// task scheduler for Sender
// just peak up next message from the queue and return executable Task wich will do real sending job 
Task
Sender::get_next_task() // throw(NoTasks) 
{
MessageSP message ;

ERS_DEBUG(4, "Trying to get a message from the queue") ;

if ( !m_messages_queue.try_pop(message) ) 
		{
ERS_DEBUG(4, "Nothing to be sent") ;
		throw NoTasks() ;
		}

return std::bind(&Sender::report_remote, this, message) ;
}

// executed in a worker thread, waits a message to appear in queue and send it to a Worker using CORBA
// must not throw anything
// handle all corba, timeouts and IPC exceptions here
void Sender::report_remote(MessageSP message) noexcept 
{
int noloop = 0 ;
while ( true ) // try until we find a working node or no working nodes...
	{
	try 	{
ERS_DEBUG(5, "Getting a worker to send message");
#ifndef ERS_NO_DEBUG
		system_clock::time_point zero = system_clock::now() ;
#endif
		::mts::worker_var wrk = get_worker() ; 
#ifndef ERS_NO_DEBUG
		system_clock::time_point start = system_clock::now() ;
#endif
ERS_DEBUG(4, "Getting worker took "  << duration_cast<microseconds>(start-zero).count() << " microseconds");
ERS_DEBUG(3, "Reporting message to remote worker");
		wrk->report(*message) ; // real CORBA call 
		// end = system_clock::now() ;
		m_messages_sent++ ;
ERS_DEBUG(3, "Message reported OK: "  << duration_cast<microseconds>(system_clock::now()-start).count() << " microseconds");
ERS_DEBUG(4, "Messages sent: " << m_messages_sent) ;
ERS_DEBUG(4, "Messages queued: " << m_messages_queued) ;
ERS_DEBUG(4, "Messages in the delivery queue: " << m_messages_queue.unsafe_size() ) ;
		break ;
		}
	catch ( NoWorkers& ex )
		{
		// TODO: report to std::error ?
		ERS_DEBUG(1, "Message not reported: No good workers available: " << ex);
		return ;
		}
	catch	( CORBA::SystemException& ex ) // TODO: timeout vs transient ?
		{
		ERS_DEBUG(1, "Can not send message on a worker: " << ex << std::endl << ex.minor() << std::endl << "Will re-try on another") ;
		if ( ++noloop > 2 )
			{
			ERS_DEBUG(0, "Failed " << noloop << " times, no success! Giving up seding message. There are some severe issues with MTS infrastructure.") ;
			return ;
			}
		// update_workers() here ? or it may be a transient problem?
		try { update_workers(false) ; }
		catch ( NoWorkers& ex )
			{
			ERS_DEBUG(0, "Message not reported: No good workers available: " << ex);
			return ;
			}
		continue ;
		}
	catch	( ... ) // can we ever be here?
		{
		ERS_LOG("Unexpected exception, can not send message, will re-try few times more... ") ;
		if ( ++noloop > 2 ) { return ; } ;
		continue ;
		}
	}
}

mts::Severity	ers2mts(ers::severity erssev)
{
switch ( erssev )
	{
	case ers::severity::Debug: 		return mts::Severity::debug; break;
	case ers::severity::Log: 		return mts::Severity::log; break;
	case ers::severity::Information: 	return mts::Severity::information; break;
	case ers::severity::Warning: 		return mts::Severity::warning; break;
	case ers::severity::Error: 		return mts::Severity::error; break;
	default: 				return mts::Severity::fatal;
	}
}
