/*
 * oh::GraphDataI.h - interface to the Graph class in IS
 * Herbert Kaiser, Aug 2006
 */

#ifndef OH_GRAPH_DATA_H
#define OH_GRAPH_DATA_H

#include <oh/core/Graph.h>
#include <oh/core/Graph2D.h>
#include <oh/core/ObjectBase.h>
#include <owl/time.h>

namespace oh
{
    /** Interface to the IS Graph class.
     *  The Graph object to work on must be given in the constructor.\n
     *  When setting the data via arrays, be sure that the arrays have the
     *  appropiate size (which is the number of graph points), this is not checked.
     *  Do \c set_dimension and \c set_entries before setting the data and do
     *  \c set_errorstyle before setting the errors. */
    template <class T>
    class GraphDataI : public ObjectBase<T>
    {
        public:
            /** Retrieve a vector with all error arrays NO COPIES are made
	     *  For TGraphErrors we have 2 errors: ex,ey
	     *  For TGraphAsymmErrors we have 4 errors: exl,exh,eyl,eyh
	     *  For TGraphBentErrors we have 8 errors: exl,exh,eyl,eyh,exld,exhd,eyld,eyhd
	     *  For TGraph2DErrors we have 3 errors: ex,ey,ez  */
	    void get_errors(std::vector<const double*>& errorArrays) const;

            /** set all the errors, copies of the double arrays are made */
	    void set_errors(const std::vector<const double*>& errorArrays);
	    
            /** Set error style */
	    void set_error_style(typename T::ErrorStyle es);

	    /** get error style */
	    typename T::ErrorStyle get_error_style() const;

            /** get all points at once no copies are made*/
	    void get_points(std::vector<const double*>& valueArrays) const;
	    
            /** get all the data in arrays first all the points then all the errors */
	    void get_data(std::vector<const double*>& data) const;
	    
            /** set all points at once, copies of the arrays are made
	     * The vector must provide 2 arrays for normal graphs
	     * and 3 arrays for Graph2D graphs */
	    void set_points(const std::vector<const double*>& valueArrays);
    };
}

namespace oh
{
    typedef GraphDataI<oh::Graph>	GraphData;
    typedef GraphDataI<oh::Graph2D>	Graph2DData;
}

#include <oh/core/GraphData.i>

#endif //OH_GRAPH_DATA_H

