/*
    oh::HistogramData.h - OH private header file.
    
    Fredrik Sandin, Oct 2001
    Monika Barczyk, Feb 2003
    Sergei Kolos,   May 2004
*/

#ifndef OH_HISTOGRAM_DATA_H
#define OH_HISTOGRAM_DATA_H

#include <string>
#include <oh/core/Histogram.h>
#include <oh/core/ObjectBase.h>

//! histogram
/*!
    This class provides an interface to histogram data and acts
    as an abstraction layer to the underlying representation of
    a histogram.
    \sa oh::Histogram
*/

namespace oh
{
    template <typename T>
    class HistogramData : public ObjectBase<Histogram>
    {
    public:
	static const ISType & type() {
	    static const ISType type_ = oh::HistogramData<T>( ).ISInfo::type();
	    return type_;
	}

	//! Default constructor
	HistogramData( const std::string & cl = ClassName );

	//! Destructor
	~HistogramData( );
        
	//! Copy constructor
        HistogramData( const HistogramData & hd );
        
	//! Copy operator
        HistogramData & operator=( const HistogramData & hd );

	//! Set the histogram dimension
	void set_dimension( oh::Histogram::Dimension d );

	//! Retrieve the histogram dimension
	oh::Histogram::Dimension get_dimension( ) const;

       //! Retrieve the error associated to a bin
	double get_error( int x, int y, int z ) const;

       //! Retrieve the pointer to the errors array
	const double * get_errors( ) const;

	//! Set the errors array at once (always copies the array contents)
	void set_errors( double * array, unsigned int size );

	//! Update the error associated to a bin
	void set_error( int x, int y, int z, double error );

	//! Retrieve the value of a bin
	T get_bin_value( int x, int y, int z ) const;

	//! Update the value of a bin
	void set_bin_value( int x, int y, int z, T height );

	//! Retrieve the pointer to the array of bin values
	T * get_bins_array( bool orphan = false ) const;

	//! Retrieve the size of the array of bin values
	size_t get_bins_size( ) const;

	//! Set the pointer to the array of bin values to array. This function does not copy the array.
	void set_bins_array( T * array, unsigned int size, bool release = false );

	//! Set the pointer to the array of bin values to array. This function does not copy the array.
	void set_bins_array( const T * array, unsigned int size );

    protected:
	void publishGuts( ISostream & out ){
	    Histogram::publishGuts( out );
	    out.put( bins_, bins_size_ );
	}

	void refreshGuts( ISistream & in ){
	    Histogram::refreshGuts( in );
	    if ( release_ )
		delete[] bins_;
	    in.get( &bins_, bins_size_ );
	    release_ = true;
	}

    private:
	// Calculates the index for X, Y and Z
	int index( int indexX, int indexY, int indexZ ) const;

	// Allocates memory for bin heights
	void alloc_bins( );

	// Allocates memory for bin errors if necessary
	void alloc_errors( );

    private:

	T *		bins_;
	size_t		bins_size_;
	mutable bool	release_;

	static std::string		    ClassName;
    };
}

#include <oh/core/HistogramData.i>

#endif // OH_HISTOGRAM_DATA_H
