template <typename T>
oh::HistogramData<T>::HistogramData( const std::string & cl ) 
  : ObjectBase<Histogram> ( cl ),
    bins_( 0 ),
    bins_size_( 0 ),
    release_( false )
{ ; }

template <typename T>
oh::HistogramData<T>::~HistogramData( )
{
    if ( release_ )
    	delete[] bins_;
}

template <typename T>
oh::HistogramData<T>::HistogramData( const oh::HistogramData<T> & hd )
  : ObjectBase<Histogram>( hd ),
    bins_size_( hd.bins_size_ ),
    release_( true )
{
    bins_ = new T[bins_size_];
    memcpy( bins_, hd.bins_, bins_size_ * sizeof( T ) );
}

template <typename T>
oh::HistogramData<T> & 
oh::HistogramData<T>::operator=( const oh::HistogramData<T> & hd )
{
    if ( this != &hd )
    {
	this -> ObjectBase<Histogram>::operator=( hd );
        if ( release_ )
        {
            delete[] bins_;
        }
        bins_size_ = hd.bins_size_;
	bins_ = new T[bins_size_];
	memcpy( bins_, hd.bins_, bins_size_ * sizeof( T ) );
	release_ = true;
    }
    return *this;
}

/*!
    Set histogram dimension.
    \param d Histogram dimension, must be D1, D2 or D3.
*/
template <typename T>
void
oh::HistogramData<T>::set_dimension( oh::Histogram::Dimension d )
{
    dimension = d;
}
    
/*!
    Returns the dimension of the histogram.
    \return D1, D2 or D3.
*/
template <typename T>
oh::Histogram::Dimension
oh::HistogramData<T>::get_dimension( ) const
{
    return dimension;
}
    
/*!
    Bin indices start at 1,  0 or Underflow refer to the underflow
    bin and  get_bin_count() + 1 or Overflow refer to the overflow bin.
*/
template <typename T>
double
oh::HistogramData<T>::get_error( int x, int y, int z ) const {
    assert( errors.size() != 0 );
    return errors[ index( x, y, z ) ];
}

/*!
    Returns pointer to the array of errors for the histogram. 
    \param size Size of the errors array
    \return Pointer to the array of errors for the histogram.
*/
template <typename T>
const double *
oh::HistogramData<T>::get_errors( ) const
{
    return ( errors.size() > 0 ? &errors[0] : 0 );
}

/*!
    Bin indices start at 1,  0 or Underflow refer to the underflow
    bin and  get_bin_count() + 1 or Overflow refer to the overflow bin.
*/
template <typename T>
void
oh::HistogramData<T>::set_error( int x, int y, int z, double error )
{
    alloc_errors( );
    errors[ index( x, y, z ) ] = error;
}    

/*!
    Copy the errors array.
*/
template <typename T>
void
oh::HistogramData<T>::set_errors( double * errors, unsigned int size )
{
    this->errors.resize( size );
    memcpy( &(this->errors[0]), errors, size * sizeof( double ) );
}

/*!
    Bin indices start at 1,  0 or Underflow refer to the underflow
    bin and  get_bin_count() + 1 or Overflow refer to the overflow bin.
*/
template <typename T>
inline T
oh::HistogramData<T>::get_bin_value( int x, int y, int z ) const {
    return bins_[ index( x, y, z ) ];
}

/*!
    Bin indices start at 1,  0 or Underflow refer to the underflow
    bin and  get_bin_count() + 1 or Overflow refer to the overflow bin.
*/
template <typename T>
inline void
oh::HistogramData<T>::set_bin_value( int x, int y, int z, T value ) {
    alloc_bins();
    bins_[ index( x, y, z ) ] = value;
}

/*!
    Bin indices start at 1,  0 or Underflow refer to the underflow
    bin and  get_bin_count() + 1 or Overflow refer to the overflow bin.
*/
template <typename T>
inline T *
oh::HistogramData<T>::get_bins_array( bool orphan ) const
{
    release_ = !orphan;
    return bins_;
}

/*!
    Bin indices start at 1,  0 or Underflow refer to the underflow
    bin and  get_bin_count() + 1 or Overflow refer to the overflow bin.
*/
template <typename T>
size_t
oh::HistogramData<T>::get_bins_size( ) const
{
    return bins_size_;
}

/*!
    Saves the pointer to the bins array. If release it true, this object takes the
    ownership of the array memory.
*/
template <typename T>
inline void
oh::HistogramData<T>::set_bins_array( T * array, unsigned int size, bool release )
{
    if ( release_ )
    	delete[] bins_;
    bins_size_ = size;
    bins_ = array;
    release_ = release;
}

/*!
    Bin indices start at 1,  0 or Underflow refer to the underflow
    bin and  get_bin_count() + 1 or Overflow refer to the overflow bin.
*/
template <typename T>
inline void
oh::HistogramData<T>::set_bins_array( const T * array, unsigned int size )
{
    if ( release_ )
    	delete[] bins_;
    bins_size_ = size;
    bins_ = const_cast<T*>( array );
    release_ = false;
}

template <typename T>
inline void
oh::HistogramData<T>::alloc_bins( ) {
    if ( bins_size_ == 0 ) {
	//
	// + 2 is used to take into account overflow and underflow bins
	//
	bins_size_ = ( get_bin_count( oh::Axis::X ) + 2 ) * 
		     ( get_bin_count( oh::Axis::Y ) + 2 ) *
		     ( get_bin_count( oh::Axis::Z ) + 2 );
	bins_ = new T[bins_size_];
	release_ = true;
    }
}

template <typename T>
void
oh::HistogramData<T>::alloc_errors( )
{
    if ( errors.size() == 0 ) {
	//
	// + 2 is used to take into account overflow and underflow bins
	//
	int size =	( get_bin_count( oh::Axis::X ) + 2 ) * 
			( get_bin_count( oh::Axis::Y ) + 2 ) *
			( get_bin_count( oh::Axis::Z ) + 2 );
	errors.resize( size );
    }
}

template <typename T>
inline int
oh::HistogramData<T>::index(  int x, int y, int z ) const
{
    int wx = get_bin_count( oh::Axis::X ) + 2;
    int wy = get_bin_count( oh::Axis::Y ) + 2;
    return x + wx * ( y + wy * z );
}

