/*
    iterator.h - OH Public header file.
    
    Serguei Kolos,  Dec 2006
*/

#ifndef OH_INTERNAL_ITERATOR_H
#define OH_INTERNAL_ITERATOR_H

#include <is/infoiterator.h>

#include <oh/interface/OHReceiver.h>
#include <oh/OHUtil.h>

/*! \namespace oh
 *  This is a wrapping namespace for internal OH classes.
 */
namespace oh
{

    /*! \brief This class provides functionality to select, list and retrieve objects
	from the OH. It can be configured to work with specific subset of such
	objects i.e. with only histograms or only graphs.
	There are several names generated of this template by \c typedef: 
        OHIterator, OHHistogramIterator, OHGraphIterator, OHGraph1DIterator, OHGraph2DIterator
    */
    template <class T, bool STRICT = false>
    class iterator : public ISInfoIterator
    {
    public:

	/*!
	    Immediately after construction the "current histogram" is undefined and
	    the iterator must be advanced with one of the ++ operators.

	    \param partition A valid IPCPartition
	    \param server Name of a valid OH server
	    \param provider Optional name of histogram provider, should only contain
	    [a-z], [A-Z], [0-9], '_' and optional wildcars according to the Posix
	    style of regular expressions
	    \param object Optional name of histogram, should only contain
	    [a-z], [A-Z], [0-9], '_' and optional wildcars according to the Posix
	    style of regular expressions
	*/
	iterator( const IPCPartition & partition,
		  const std::string & server,
		  const std::string & provider = ".*",
		  const std::string & object = ".*" ) 
        try : ISInfoIterator(  partition, server, 
          		    ( STRICT ? T::type() : ~T::type() ) && oh::util::create_regex( provider, object ) )
        { ; } 
        catch ( daq::is::InvalidCriteria & ex ) {
            throw daq::oh::InvalidRegex( ex.context(), ex.get_expression() );
        }
        catch ( daq::is::RepositoryNotFound & ex ) {
            throw daq::oh::RepositoryNotFound( ex.context(), ex.get_name() );
        }

	/*!
	    Retrieve the current object or set of objects
	    \param receiver Should be a user defined histogram receiver object derived
	    from one of the OH receiver classes.
	    \param with_history If false the last value of the histogram will be retrieved.
	    \param how_many If with_history parameter is true then this one defines how many values to retrieve
            		from the history, -1 means all.
	    If true all the values will be retrieved.
	*/
	void retrieve( OHReceiver & receiver, bool with_history = false, short how_many = -1 ) 
        							 {
	    receiver.readObjects( *this, with_history, how_many );
        }

	/*!
	    Retrieve the current object with respect to the given tag
	    \param receiver Should be a user defined histogram receiver object derived
	    from one of the OH receiver classes.
	    \param with_history If false the last value of the histogram will be retrieved.
	    If true all the values will be retrieved.
	*/
	void retrieve( OHReceiver & receiver, int tag ) 
        			 {
	    receiver.readObjectWithTag( *this, tag );
	}

	/*!
	    Retrieve name of the provider which has published the current histogram
	    \return The name of the provider who published the histogram (set) on
	    the current position or "" if the current position is undefined.
	*/
	std::string provider( ) const{
	    return oh::util::get_provider_name( ISInfoIterator::name( ) );
	}

	/*!
	    Exercises the type of the object at the current iterator position
	    \return true if the current object is histogram
	*/
	bool isHistogram( ) const{
	    return ( type().subTypeOf( oh::Histogram::type() ) );
	}

	/*!
	    Exercises the type of the object at the current iterator position
	    \return true if the current object is graph
	*/
	bool isGraph( ) const{
	    return ( type() == oh::Graph::type() );
	}

	/*!
	    Exercises the type of the object at the current iterator position
	    \return true if the current object is 2D graph
	*/
	bool isGraph2D( ) const{
	    return ( type() == oh::Graph2D::type() );
	}

	/*!
	    Retrieve name of the current object
	    \return The name of the the histogram (set) on the current position
	    or "" if the current position is undefined.
	*/
	std::string name( ) const {
	    return oh::util::get_object_name( ISInfoIterator::name( ) );
	}
    };
}

#endif
