/*
    iterator.h - OH Public header file.
    
    Serguei Kolos,  Dec 2006
*/

#ifndef OH_INTERNAL_STREAM_H
#define OH_INTERNAL_STREAM_H

#include <is/infostream.h>

#include <oh/interface/OHReceiver.h>
#include <oh/exceptions.h>
#include <oh/OHUtil.h>

/*! \namespace oh
 *  This is a wrapping namespace for internal OH classes.
 */
namespace oh
{

    /*! \brief This class provides functionality to select, list and retrieve objects
	from the OH. It can be configured to work with specific subset of such
	objects i.e. with only histograms or only graphs.
	There are several names generated of this template by \c typedef: 
        OHIterator, OHHistogramIterator, OHGraphIterator, OHGraph1DIterator, OHGraph2DIterator
    */
    template <class T, bool STRICT = false>
    class stream : public ISInfoStream
    {
    public:

	/*!
	    Sonstructs the histogram stream object.
            
            \param partition A valid IPCPartition
	    \param server Name of a valid OH server
	    \param provider Optional name of histogram provider, should only contain
	    [a-z], [A-Z], [0-9], '_' and optional wildcars according to the Posix
	    style of regular expressions
	    \param object Optional name of histogram, should only contain
	    [a-z], [A-Z], [0-9], '_' and optional wildcars according to the Posix
	    style of regular expressions
	    \param synchronous if true then the constructor will return only when the 
            		stream is fully constructed, otherwie it resturns immediately
	    \param history_depth defines how many history values have to be added to the
            	stream for each matching infomation object. By defualt is equal to 1. 
                	Negative value means all.
	*/
	stream( const IPCPartition & partition,
		const std::string & server,
		const std::string & provider = ".*",
		const std::string & object = ".*",
                bool synchronous = false,
                int history_depth = 1 ) 
        try : ISInfoStream(  partition, server, 
			   ( STRICT ? T::type() : ~T::type() ) && oh::util::create_regex( provider, object ),
                             synchronous, history_depth )
        { ; } 
        catch ( daq::is::InvalidCriteria & ex ) {
            throw daq::oh::InvalidRegex( ex.context(), ex.get_expression() );
        }
        catch ( daq::is::RepositoryNotFound & ex ) {
            throw daq::oh::RepositoryNotFound( ex.context(), ex.get_name() );
        }

	/*!
	    Retrieve the current object or set of objects
	    \param receiver Should be a user defined histogram receiver object derived
	    from one of the OH receiver classes.
	    \param with_history If false the last value of the histogram will be retrieved.
	    \param how_many If with_history parameter is true then this one defines how many values to retrieve
            		from the history, -1 means all.
	    If true all the values will be retrieved.
	*/
	void retrieve( OHReceiver & receiver ) 
	{
	    receiver.readObject( *this );
        }

	/*!
	    Retrieve name of the provider which has published the current histogram
	    \return The name of the provider who published the histogram (set) on
	    the current position or "" if the current position is undefined.
	*/
	std::string provider( ) const{
	    return ( oh::util::get_provider_name( ISInfoStream::name( ) ) );
	}

	/*!
	    Retrieve name of the current object
	    \return The name of the the histogram (set) on the current position
	    or "" if the current position is undefined.
	*/
	std::string name( ) const {
	    return ( oh::util::get_object_name( ISInfoStream::name( ) ) );
	}
    };
}

#endif
