/*
 OHRootProvider.cxx

 Fredrik Sandin, Oct 2001
 Monika Barczyk, Sept 2002
 Sergei Kolos,   Nov 2004
 Herbert Kaiser, Aug 2006
 */

#include <oh/OHRootProvider.h>
#include <oh/OHRootMutex.h>
#include <oh/core/HistogramData.h>
#include <oh/core/ProfileData.h>
#include <oh/core/GraphData.h>
#include <oh/core/TTraits.h>
#include <oh/exceptions.h>

#include <TH2.h>
#include <TH3.h>
#include <THashList.h>
#include <TProfile.h>
#include <TProfile2D.h>
#include <TProfile3D.h>
#include <TObjString.h>
#include <TGraphErrors.h>
#include <TGraphAsymmErrors.h>
#include <TGraphBentErrors.h>
#include <TGraph2DErrors.h>

namespace {

    template<class T, class A>
    void copy_axis(T * oh_object, oh::Axis::ID id, A * axis) {
        const TArrayD * rootp = axis->GetXbins();

        oh_object->set_axis_title(id, std::string(axis->GetTitle()));
        if (rootp->GetSize() > 0) {
            // Variable width bins
            oh_object->set_axis_range(id, rootp->GetArray(), rootp->GetSize());
        }
        else {
            // Fixed width bins
            oh_object->set_axis_range(id, axis->GetBinLowEdge(1),
                    axis->GetBinWidth(1), axis->GetNbins());
        }

        std::vector<int> & indices = oh_object->indices(id);
        std::vector<std::string> & labels = oh_object->labels(id);

        THashList * hl = axis->GetLabels();
        if (hl) {
            indices.resize(hl->GetSize());
            labels.resize(hl->GetSize());
            TIter next(hl);
            TObjString * obj;
            int i = 0;
            while ((obj = (TObjString*) next()) != 0) {
                indices[i] = (Int_t) obj->GetUniqueID();
                labels[i] = obj->GetName();
                i++;
            }
        }
    }

    template<class T>
    void base_convert(oh::HistogramData<typename oh::TTraits<T>::Type> * h,
            const T & histogram) {
        h->set_title(histogram.GetTitle());
        h->set_entries((unsigned int) histogram.GetEntries());

        switch (histogram.GetDimension()) {
            case 1:
                h->set_dimension(oh::Histogram::D1);
                break;
            case 2:
                h->set_dimension(oh::Histogram::D2);
                break;
            case 3:
                h->set_dimension(oh::Histogram::D3);
                break;
            default:
                ERS_ASSERT_MSG(0, "Bad histogram object has been provided");
                break;
        }

        if (histogram.GetXaxis())
            copy_axis(h, oh::Axis::X, histogram.GetXaxis());
        if (histogram.GetYaxis())
            copy_axis(h, oh::Axis::Y, histogram.GetYaxis());
        if (histogram.GetZaxis())
            copy_axis(h, oh::Axis::Z, histogram.GetZaxis());

        // Set bin content without copying
        h->set_bins_array(histogram.GetArray(),
                (unsigned int) histogram.GetSize());

        // Set errors without copying
        const TArrayD * errors = histogram.GetSumw2();
        if (errors) {
            h->set_errors(errors->fArray, (unsigned int) errors->fN);
        }

        h->statistics.resize(oh::TTraits<T>::StatsSize);
        histogram.GetStats(&(h->statistics[0]));
    }

    template<class T>
    oh::Histogram *
    convert(const T & histogram,
            const std::vector<std::pair<std::string, std::string> > & ann) {
        auto h = new oh::HistogramData<typename oh::TTraits<T>::Type>();
        h->set_annotations(ann);

        base_convert(h, histogram);

        return h;
    }

    template<typename T>
    oh::Histogram *
    convert_profile(const T & profile,
            const std::vector<std::pair<std::string, std::string> > & ann) {
        auto h = new oh::ProfileData();
        h->set_annotations(ann);

        base_convert(h, profile);

        h->bin_entries.resize(profile.GetSize());
        for (int i = 0; i < profile.GetSize(); i++)
            h->bin_entries[i] = profile.GetBinEntries(i);

        return h;
    }

    oh::Histogram *
    convert(const TProfile & profile,
            const std::vector<std::pair<std::string, std::string> > & ann) {
        return convert_profile(profile, ann);
    }

    oh::Histogram *
    convert(const TProfile2D & profile,
            const std::vector<std::pair<std::string, std::string> > & ann) {
        return convert_profile(profile, ann);
    }

    oh::Histogram *
    convert(const TProfile3D & profile,
            const std::vector<std::pair<std::string, std::string> > & ann) {
        return convert_profile(profile, ann);
    }
}

#define ROOT_HISTOGRAM_TYPES(decl)	\
decl(TH1C)	decl ## _SEPARATOR	\
decl(TH2C)	decl ## _SEPARATOR	\
decl(TH3C)	decl ## _SEPARATOR	\
decl(TH1S)	decl ## _SEPARATOR	\
decl(TH2S)	decl ## _SEPARATOR	\
decl(TH3S)	decl ## _SEPARATOR	\
decl(TH1I)	decl ## _SEPARATOR	\
decl(TH2I)	decl ## _SEPARATOR	\
decl(TH3I)	decl ## _SEPARATOR	\
decl(TH1F)	decl ## _SEPARATOR	\
decl(TH2F)	decl ## _SEPARATOR	\
decl(TH3F)	decl ## _SEPARATOR	\
decl(TProfile)	decl ## _SEPARATOR	/* must precede the TH1D since TProfile inherits from TH1D */ \
decl(TH1D)	decl ## _SEPARATOR	\
decl(TProfile2D)decl ## _SEPARATOR	/* must precede the TH2D since TProfile2D inherits from TH2D */ \
decl(TH2D)	decl ## _SEPARATOR	\
decl(TProfile3D)decl ## _SEPARATOR	/* must precede the TH3D since TProfile3D inherits from TH3D */ \
decl(TH3D)	

#define EXACT_CONVERT(x)	if ( histogram.IsA() == x::Class() ) \
					h.reset( ::convert( static_cast<const x &>( histogram ), ann ) );
#define EXACT_CONVERT_SEPARATOR	else

#define BASE_CONVERT(x)		if ( histogram.InheritsFrom( x::Class() ) ) \
					h.reset( ::convert( static_cast<const x &>( histogram ), ann ) );
#define BASE_CONVERT_SEPARATOR	else

/*!
 \param p A valid IPCPartition
 \param server Name of a valid OH server
 \param name Name of the provider, user is responsible for specifying a
 unique name. Should only contain characters from [a-z], [A-Z], [0-9] and '_'.
 \param lst User defined command listener object.
 */
OHRootProvider::OHRootProvider(const IPCPartition & p,
        const std::string & server, const std::string & name,
        OHCommandListener * lst) :
        OHProvider(p, server, name, lst) {
    ;
}

/*!
 \param histogram A valid ROOT histogram
 \param name A name describing the histogram. Should only contain characters
 from [a-z], [A-Z], [0-9] and '_'.
 \param update Tells wether to keep the previous version of this histogram (by default false)
 \return OHRootProvider::CommFailure if a communication error occurred,
 OHRootProvider::InvalidArgument if one or more arguments is invalid,
 else OHRootProvider::Success.
 */
void OHRootProvider::publish(const TH1 & histogram, const std::string & name,
        int tag, const std::vector<std::pair<std::string, std::string> > & ann)
{
    std::unique_ptr<oh::Histogram> h = convert(histogram, ann);
    ERS_ASSERT_MSG(h.get(), "Bad histogram object has been provided");

    OHProvider::publish(*h.get(), name, tag);
}

std::unique_ptr<oh::Histogram>
OHRootProvider::convert(const TH1& histogram,
        const std::vector<std::pair<std::string, std::string>>& ann)
{
    std::unique_ptr<oh::Histogram> h;
    {
        std::unique_lock lock(OHRootMutex::getMutex());
        ROOT_HISTOGRAM_TYPES(EXACT_CONVERT)

        if (!h.get()) {
            // We must be dealing with custom object which inherits from a histogram class
            ROOT_HISTOGRAM_TYPES(BASE_CONVERT)
        }
    }
    return h;
}

/* *************************************************************
 * Part for Graph support
 * ************************************************************* */
void OHRootProvider::publish(const TGraph & graph, const std::string & name,
        int tag, const std::vector<std::pair<std::string, std::string> > & ann)
{
    oh::GraphData gi;
    {
        std::unique_lock lock(OHRootMutex::getMutex());
        gi.set_entries(graph.GetN());
        gi.set_title(graph.GetTitle());
        gi.set_annotations(ann);
        copy_axis(&gi, oh::Axis::X, graph.GetXaxis());
        copy_axis(&gi, oh::Axis::Y, graph.GetYaxis());
        std::vector<const double*> points(2);
        points[0] = graph.GetX();
        points[1] = graph.GetY();
        gi.set_points(points);

        std::vector<const double*> errorArrays;
        if (graph.InheritsFrom(TGraphErrors::Class())) {
            gi.set_error_style(oh::Graph::Symmetric);
            errorArrays.resize(2);
            errorArrays[0] = static_cast<const TGraphErrors*>(&graph)->GetEX();
            errorArrays[1] = static_cast<const TGraphErrors*>(&graph)->GetEY();
        }
        else if (graph.InheritsFrom(TGraphAsymmErrors::Class())) {
            gi.set_error_style(oh::Graph::Asymmetric);
            errorArrays.resize(4);
            errorArrays[0] =
                    static_cast<const TGraphAsymmErrors*>(&graph)->GetEXlow();
            errorArrays[1] =
                    static_cast<const TGraphAsymmErrors*>(&graph)->GetEXhigh();
            errorArrays[2] =
                    static_cast<const TGraphAsymmErrors*>(&graph)->GetEYlow();
            errorArrays[3] =
                    static_cast<const TGraphAsymmErrors*>(&graph)->GetEYhigh();
        }
        else if (graph.InheritsFrom(TGraphBentErrors::Class())) {
            gi.set_error_style(oh::Graph::AsymBent);
            errorArrays.resize(8);
            errorArrays[0] =
                    static_cast<const TGraphBentErrors*>(&graph)->GetEXlow();
            errorArrays[1] =
                    static_cast<const TGraphBentErrors*>(&graph)->GetEXhigh();
            errorArrays[2] =
                    static_cast<const TGraphBentErrors*>(&graph)->GetEYlow();
            errorArrays[3] =
                    static_cast<const TGraphBentErrors*>(&graph)->GetEYhigh();
            errorArrays[4] =
                    static_cast<const TGraphBentErrors*>(&graph)->GetEXlowd();
            errorArrays[5] =
                    static_cast<const TGraphBentErrors*>(&graph)->GetEXhighd();
            errorArrays[6] =
                    static_cast<const TGraphBentErrors*>(&graph)->GetEYlowd();
            errorArrays[7] =
                    static_cast<const TGraphBentErrors*>(&graph)->GetEYhighd();
        }
        else {
            gi.set_error_style(oh::Graph::None);
        }

        // data is copied to the oh::Graph
        gi.set_errors(errorArrays);
    }

    OHProvider::publish(gi, name, tag);
}

void OHRootProvider::publish(const TGraph2D & graph, const std::string & name,
        int tag, const std::vector<std::pair<std::string, std::string> > & ann)
{
    oh::Graph2DData gi;
    {
        std::unique_lock lock(OHRootMutex::getMutex());

        gi.set_entries(graph.GetN());
        gi.set_title(graph.GetTitle());
        gi.set_annotations(ann);
        copy_axis(&gi, oh::Axis::X, graph.GetXaxis());
        copy_axis(&gi, oh::Axis::Y, graph.GetYaxis());
        copy_axis(&gi, oh::Axis::Z, graph.GetZaxis());
        std::vector<const double*> points(3);
        points[0] = graph.GetX();
        points[1] = graph.GetY();
        points[2] = graph.GetZ();
        gi.set_points(points);

        std::vector<const double*> errorArrays;
        if (graph.InheritsFrom(TGraph2DErrors::Class())) {
            gi.set_error_style(oh::Graph2D::Symmetric);
            errorArrays.resize(3);
            errorArrays[0] =
                    static_cast<const TGraph2DErrors*>(&graph)->GetEX();
            errorArrays[1] =
                    static_cast<const TGraph2DErrors*>(&graph)->GetEY();
            errorArrays[2] =
                    static_cast<const TGraph2DErrors*>(&graph)->GetEZ();
        }
        else {
            gi.set_error_style(oh::Graph2D::None);
        }

        // data is copied to the oh::Graph
        gi.set_errors(errorArrays);
    }

    OHProvider::publish(gi, name, tag);
}

// EOF
