/*
 * main.cpp
 *
 *  Created on: Nov 11, 2020
 *      Author: kolos
 */
#include <iostream>

#include <boost/program_options.hpp>

#include <TApplication.h>
#include <TBrowser.h>
#include <TDirectory.h>
#include <TEnv.h>
#include <TFile.h>
#include <TImage.h>
#include <TObject.h>
#include <TString.h>
#include <TSystemDirectory.h>
#include <TGFrame.h>
#include <TGMimeTypes.h>
#include <TGComboBox.h>
#include <TGFileBrowser.h>
#include <TRootBrowser.h>

namespace po = boost::program_options;

struct RootBrowser: public TRootBrowser {
    struct FileBrowser: public TGFileBrowser {
        TGComboBox * get() { return fDrawOption; }
    };
    TGComboBox * get() { return ((FileBrowser*)fActBrowser)->get(); }
};

int main(int argc, char **argv) {
    po::options_description desc(
            "This program can browse Run Control transitions timing files recorded by opmon_logger");

    std::string file_name;
    std::string directory_name;
    desc.add_options()("help,h", "Print help message")
    ("file-name,f", po::value<std::string>(&file_name), "file name")
    ("directory-name,d", po::value<std::string>(&directory_name), "directory name");

    po::variables_map vm;
    try {
        po::store(po::parse_command_line(argc, argv, desc), vm);
        po::notify(vm);
    } catch (std::exception &ex) {
        std::cerr << ex.what() << std::endl;
        desc.print(std::cout);
        return EXIT_FAILURE;
    }

    if (vm.count("help")) {
        desc.print(std::cout);
        return 1;
    }

    TApplication app("RC Transition Browser", 0, 0);

    gClient->GetMimeTypeList()->AddType("root/tmultigraph", "TMultiGraph",
            "h1_s.xpm", "h1_t.xpm", "->Browse()");

    gClient->GetMimeTypeList()->AddType("root/th2f", "TH2F",
            "h3_s.xpm", "h3_t.xpm", "->Browse()");

    gClient->GetMimeTypeList()->AddType("root/tlist", "TList",
            "branch_folder_s.xpm", "branch_folder_t.xpm", "->Browse()");

    gEnv->SetValue("TGraph.BrowseOption", "A");

    std::vector<TFile*> t_files;
    TObject * system = 0;
    if (vm.count("file-name")) {
        TFile * file = TFile::Open(file_name.c_str(), "READ");
        if (not file) {
            std::cerr << "Can not open '" << file_name << "' file" << std::endl;
            return EXIT_FAILURE;
        }
        system = file;
    } else if (vm.count("directory-name")) {
        if (directory_name.back() != '/') {
            directory_name += '/';
        }
        TSystemDirectory * dir = new TSystemDirectory(
                directory_name.c_str(), directory_name.c_str());
        if (dir->IsZombie()) {
            std::cerr << "Can not open '" << directory_name << "' directory" << std::endl;
            return EXIT_FAILURE;
        }

        TList * files = dir->GetListOfFiles();
        if (files) {
            TSystemFile *file;
            TIter next(files);
            while ((file = (TSystemFile*) next())) {
                TString fname = file->GetName();
                if (!file->IsDirectory() && fname.EndsWith(".root")) {
                    std::string pathname = directory_name + fname.Data();
                    TFile * tf = TFile::Open(pathname.c_str(), "READ");
                    if (not tf) {
                        std::cerr << "Can not open '" << pathname << "' file" << std::endl;
                        continue;
                    }
                    if (!system) {
                        system = tf;
                    } else {
                        t_files.push_back(tf);
                    }
                }
            }
        }
        if (!system) {
            std::cerr << "No ROOT files are found in '"
                    << directory_name << "' directory" << std::endl;
            return EXIT_FAILURE;
        }
    } else {
        std::cerr << "Either file or directory name should be given" << std::endl;
        desc.print(std::cout);
    }

    TBrowser browser("RC Transition Browser", system, "RC Transition Browser");

    TRootBrowser * impl = (TRootBrowser *)browser.GetBrowserImp();
    impl->Connect(impl, "CloseWindow()", "TApplication", &app, "Terminate()");
    impl->MoveResize(300, 200, 1400, 800);

    ((RootBrowser*)impl)->get()->Select(6, 1);

    for (auto f : t_files) {
        browser.Add(f);
    }

    app.Run();

    return 0;
}
