#include <cstdint>
#include <chrono>
#include <iostream>
#include <map>
#include <memory>
#include <string>
#include <vector>

#include <boost/program_options.hpp>

#include <pbeast/input-data-file.h>
#include <pbeast/series-print.h>
#include <pbeast/data-type.h>

namespace po = boost::program_options;

template<class T>
  void
print_raw_serie(bool print_series, const std::string& series_name, const int ts_decimal_width, T data)
{
  if(print_series)
    {
      std::cout << "series: " << series_name << std::endl;
      if(!data.empty())
        {
          std::cout << " data[" << data.size() << "]:" << std::endl;
          pbeast::print(std::cout, ts_decimal_width, data, nullptr);
        }
    }
  data.clear();
}

template<class T>
  void
print_downsampled_serie(bool print_series, const std::string& series_name, T data)
{
  if(print_series)
    {
      std::cout << "series: " << series_name << std::endl;
      if(!data.empty())
        {
          std::cout << " data[" << data.size() << "]:" << std::endl;
          pbeast::print(std::cout, data, nullptr);
        }
    }
  data.clear();
}

template<class T>
  void
  get_data(pbeast::InputDataFile& file, const std::string& series_name, bool print_series, const int ts_decimal_width)
  {
    std::string name;

    if (file.get_is_array() == false)
      {
        if (file.is_downsampled() == false)
          {
            std::vector<pbeast::SeriesData<T> > data;

            if (series_name.empty())
              while (file.read_next(name, data))
                {
                  print_raw_serie(print_series, name, ts_decimal_width, data);
                  data.clear();
                }
            else
              {
                file.read_serie(series_name, data);
                print_raw_serie(print_series, series_name, ts_decimal_width, data);
              }
          }
        else
          {
            std::vector<pbeast::DownsampledSeriesData<T> > data;

            if (series_name.empty())
              while (file.read_next(name, data))
                {
                  print_downsampled_serie(print_series, name, data);
                  data.clear();
                }
            else
              {
                file.read_serie(series_name, data);
                print_downsampled_serie(print_series, series_name, data);
              }
          }
      }
    else
      {
        if (file.is_downsampled() == false)
          {
            std::vector<pbeast::SeriesVectorData<T> > data;

            if (series_name.empty())
              while (file.read_next(name, data))
                {
                  print_raw_serie(print_series, name, ts_decimal_width, data);
                  data.clear();
                }
            else
              {
                file.read_serie(series_name, data);
                print_raw_serie(print_series, series_name, ts_decimal_width, data);
              }
          }
        else
          {
            std::vector<pbeast::DownsampledSeriesVectorData<T> > data;

            if (series_name.empty())
              while (file.read_next(name, data))
                {
                  print_downsampled_serie(print_series, name, data);
                  data.clear();
                }
            else
              {
                file.read_serie(series_name, data);
                print_downsampled_serie(print_series, series_name, data);
              }
          }
      }
  }





int main(int argc, char* argv[]) {

  std::string data_file;
  std::string series_name;
  bool list_series(false);
  bool print_series(false);
  bool print_info(false);
  bool print_merge_memory_info(false);

  po::options_description desc("Print information and contents of P-BEAST data file.");

  try
    {
      desc.add_options()
          ("data-file,f", po::value<std::string>(&data_file)->required(), "Name of data file")
          ("series-name,n", po::value<std::string>(&series_name), "Name of data series to print (print all, if not defined)")
          ("list-series,l", "List names of all data series")
          ("print-merge-in-memory-info,m", "Print prediction of RAM used by in-memory merge per data series")
          ("print-series,p", "Print data series")
          ("print-info,i", "Print information about file")
          ("help,h", "Print help message");

      po::variables_map vm;
      po::store(po::parse_command_line(argc, argv, desc), vm);

      if (vm.count("help"))
        {
          std::cout << desc << std::endl;
          return 0;
        }

      if (vm.count("print-series"))
        print_series = true;

      if (vm.count("list-series"))
        list_series = true;

      if (vm.count("print-merge-in-memory-info"))
        print_merge_memory_info = true;

      if (vm.count("print-info"))
        print_info = true;

      po::notify(vm);
    }
  catch (std::exception& ex)
    {
      std::ostringstream text;
      std::cerr << "ERROR: " << ex.what() << std::endl;
      return 1;
    }

  try
    {
      auto t1 = std::chrono::steady_clock::now();

      pbeast::InputDataFile file(data_file);

      std::cout << " * read " << file.get_number_of_series() << " headers: " << std::chrono::duration_cast<std::chrono::microseconds>(std::chrono::steady_clock::now()-t1).count() / 1000. << " ms" << std::endl;

      if (print_info)
        {
          std::cout
              << std::boolalpha
              << "file version        : " << file.get_file_version() << std::endl
              << "data format         : " << (file.is_downsampled() ? "downsampled" : "raw") << std::endl
              << "partition           : " << file.get_partition() << std::endl
              << "class               : " << file.get_class() << std::endl
              << "attribute           : " << file.get_attribute() << std::endl
              << "data type           : " << pbeast::as_string(static_cast<pbeast::DataType>(file.get_data_type())) << std::endl
              << "is-array            : " << file.get_is_array() << std::endl
              << "number_of_series    : " << file.get_number_of_series() << std::endl
              << "zip catalog         : " << file.get_zip_catalog_flag() << std::endl
              << "zip data            : " << file.get_zip_data_flag() << std::endl
              << "catalogize strings  : " << file.get_catalogize_strings_flag() << std::endl;

          if (file.is_downsampled())
            {
              std::cout << "downsample interval : " << file.get_downsample_interval() << std::endl;
            }
          else
            {
              std::cout << "timestamp precision : " << file.get_precision() << std::endl;
            }

          return 0;
        }


      if (list_series || print_merge_memory_info)
        {
          std::shared_ptr<std::vector<uint64_t>> merge_size;

          if (print_merge_memory_info)
            merge_size = file.estimate_in_memory_merge_size();

          std::cout << "SERIES:\n";

          unsigned int count(1);
          for (auto& i : file.get_series_info())
            {
              std::cout << count++ << "\t" << i.first;

              if(print_merge_memory_info)
                std::cout << ' ' << merge_size->at(count-2);

              std::cout << std::endl;
            }

          return 0;
        }

      const int ts_decimal_width = (
          file.get_precision() < 10      ? 6 :
          file.get_precision() < 100     ? 5 :
          file.get_precision() < 1000    ? 4 :
          file.get_precision() < 10000   ? 3 :
          file.get_precision() < 100000  ? 2 :
          file.get_precision() < 1000000 ? 1 :
          0
      );


      t1 = std::chrono::steady_clock::now();

      if (file.get_data_type() == pbeast::BOOL)
        {
          get_data<bool>(file, series_name, print_series, ts_decimal_width);
        }
      else if (file.get_data_type() == pbeast::S8)
        {
          get_data<int8_t>(file, series_name, print_series, ts_decimal_width);
        }
      else if (file.get_data_type() == pbeast::U8)
        {
          get_data<uint8_t>(file, series_name, print_series, ts_decimal_width);
        }
      else if (file.get_data_type() == pbeast::S16)
        {
          get_data<int16_t>(file, series_name, print_series, ts_decimal_width);
        }
      else if (file.get_data_type() == pbeast::U16)
        {
          get_data<uint16_t>(file, series_name, print_series, ts_decimal_width);
        }
      else if (file.get_data_type() == pbeast::S32)
        {
          get_data<int32_t>(file, series_name, print_series, ts_decimal_width);
        }
      else if (file.get_data_type() == pbeast::U32)
        {
          get_data<uint32_t>(file, series_name, print_series, ts_decimal_width);
        }
      else if (file.get_data_type() == pbeast::S64)
        {
          get_data<int64_t>(file, series_name, print_series, ts_decimal_width);
        }
      else if (file.get_data_type() == pbeast::U64)
        {
          get_data<uint64_t>(file, series_name, print_series, ts_decimal_width);
        }
      else if (file.get_data_type() == pbeast::FLOAT)
        {
          get_data<float>(file, series_name, print_series, ts_decimal_width);
        }
      else if (file.get_data_type() == pbeast::DOUBLE)
        {
          get_data<double>(file, series_name, print_series, ts_decimal_width);
        }
      else if (file.get_data_type() == pbeast::ENUM)
        {
          get_data<int32_t>(file, series_name, print_series, ts_decimal_width);
        }
      else if (file.get_data_type() == pbeast::DATE)
        {
          get_data<uint32_t>(file, series_name, print_series, ts_decimal_width);
        }
      else if (file.get_data_type() == pbeast::TIME)
        {
          get_data<uint64_t>(file, series_name, print_series, ts_decimal_width);
        }
      else if (file.get_data_type() == pbeast::STRING)
        {
          get_data<std::string>(file, series_name, print_series, ts_decimal_width);
        }
      else if (file.get_data_type() == pbeast::VOID)
        {
          get_data<pbeast::Void>(file, series_name, print_series, ts_decimal_width);
        }

      std::cout << " * read " << file.get_number_of_series() << " (data items: " << file.get_total_data_size() << ") data series: " << std::chrono::duration_cast<std::chrono::microseconds>(std::chrono::steady_clock::now()-t1).count() / 1000. << " ms" << std::endl;

      return 0;
    }
  catch (std::exception & ex)
    {
      std::cerr << "caught exception: " << ex.what() << std::endl;
      return -1;
    }

  return 0;
}
