#include <chrono>
#include <thread>

#include <is/exceptions.h>

#include "application_exceptions.h"
#include "receiver_monitoring.h"

#include "PBeastReceiverMon.h"

using std::chrono_literals::operator""ms;

pbeast::receiver::Monitoring::Monitoring(const std::string& partition_name, const std::string& receiver_id) :
  m_num_of_tested_objects           (0),
  m_num_of_tested_attributes        (0),
  m_num_of_updated_data_points      (0),
  m_num_of_merged_data_points       (0),
  m_num_of_saved_files              (0),
  m_size_of_saved_files             (0),
  m_compact_time                    (0),
  m_total_num_of_tested_objects     (0),
  m_total_num_of_tested_attributes  (0),
  m_total_num_of_saved_files        (0),
  m_total_size_of_saved_files       (0),
  m_number_of_data_items            (0),
  m_total_num_of_updated_data_points(0),
  m_total_num_of_merged_data_points (0),
  m_data_compaction_efficiency      (0.0),
  m_info_name                       (std::string("RunParams.") + receiver_id + '@' + partition_name),
  m_started                         (time(0))
{
  publish();
}

pbeast::receiver::Monitoring::~Monitoring()
{
  publish();

  ERS_LOG(
    "TOTAL:\n"
    "  number of tested objects      : " << m_total_num_of_tested_objects << "\n"
    "  number of tested attributes   : " << m_total_num_of_tested_attributes << "\n"
    "  number of updated data points : " << m_total_num_of_updated_data_points << "\n"
    "  number of merged data points  : " << m_total_num_of_merged_data_points << "\n"
    "  number of saved files         : " << m_total_num_of_saved_files << "\n"
    "  size of saved files           : " << m_total_size_of_saved_files << "\n"
  );

  time_t now = time(0);

  double interval = now - m_started;
  if(interval < 1.0) interval = 1.0;

  ERS_LOG(
    "AVERAGES:\n"
    "  testing objects       : " << static_cast<double>(m_total_num_of_tested_objects) / interval / 1000.0 << " kHz\n"
    "  testing attributes    : " << static_cast<double>(m_total_num_of_tested_attributes) / interval / 1000.0 << " kHz\n"
    "  updating data points  : " << static_cast<double>(m_total_num_of_updated_data_points) / interval / 1000.0 << " kHz\n"
    "  merging data points   : " << static_cast<double>(m_total_num_of_merged_data_points) / interval / 1000.0 << " kHz\n"
    "  compaction efficiency : " << static_cast<double>(m_total_num_of_merged_data_points) / static_cast<double>(m_total_num_of_updated_data_points+m_total_num_of_merged_data_points) << "\n"
    "  saving files          : " << static_cast<double>(m_total_num_of_saved_files) / interval << " Hz\n"
    "  saving rate           : " << static_cast<double>(m_total_size_of_saved_files) / interval / 1000.0 << " KBytes/s"
  );


  // set most monitoring parameters to 0 before removal of the monitoring object
  std::this_thread::sleep_for(1ms);
  m_compact_time = 0;
  m_rebuild_eov_time = 0;
  m_commit_time = 0;
  publish();
  std::this_thread::sleep_for(10ms);

  try
    {
      m_is_info_dict.remove(m_info_name);
    }
  catch (daq::is::Exception& ex)
    {
      ers::error(daq::pbeast::MonitoringPublicationError(ERS_HERE, "remove", m_info_name, ex));
    }
}


void pbeast::receiver::Monitoring::publish()
{
  try
    {
      PBeastReceiverMon obj;

      if (uint64_t total = (m_num_of_merged_data_points + m_num_of_updated_data_points))
        {
          m_data_compaction_efficiency = static_cast<double>(m_num_of_merged_data_points) / static_cast<double>(total);
          m_total_num_of_updated_data_points += m_num_of_updated_data_points.exchange(0);
          m_total_num_of_merged_data_points += m_num_of_merged_data_points.exchange(0);
        }

      obj.num_of_tested_objects = m_num_of_tested_objects.exchange(0);
      obj.num_of_tested_attributes = m_num_of_tested_attributes.exchange(0);
      obj.num_of_saved_files = m_num_of_saved_files.exchange(0);
      obj.size_of_saved_files = m_size_of_saved_files.exchange(0);
      obj.data_compaction_time = m_compact_time;
      obj.rebuild_eov_time = m_rebuild_eov_time;
      obj.commit_time = m_commit_time;

      obj.data_compaction_efficiency = m_data_compaction_efficiency;
      obj.num_of_data_points = m_number_of_data_items;

      m_total_num_of_tested_objects += obj.num_of_tested_objects;
      m_total_num_of_tested_attributes += obj.num_of_tested_attributes;
      m_total_num_of_saved_files += obj.num_of_saved_files;
      m_total_size_of_saved_files += obj.size_of_saved_files;

      m_is_info_dict.checkin(m_info_name, obj);
    }
  catch (daq::is::Exception& ex)
    {
      ers::error(daq::pbeast::MonitoringPublicationError(ERS_HERE, "publish", m_info_name, ex));
    }
}
