#include <boost/regex.hpp>

#include <ipc/exceptions.h>

#include <pbeast/data-file.h>
#include <pbeast/series-idl.h>

#include "application_exceptions.h"
#include "server.h"
#include "server_aggregation.h"
#include "server_functions.h"


std::string
pbeast::Server::make_method_params_name(const std::string& server_name, const char * method, const char * partition_name, const char * class_name, const char* attribute_name, uint64_t since, uint64_t until, const char* object)
{
  std::ostringstream text;
  text << method << '(';

  if (partition_name)
    text << "partition: \'" << partition_name << '\'';

  if (class_name)
    text << ", class: \'" << class_name << '\'';

  if (attribute_name)
    text << ", attribute: \'" << attribute_name << '\'';

  if (since || until)
    text << ", since:  " << since << ", until:  " << until;

  if (object && *object)
    text << ", object(s): \'" << object << '\'';

  text << ") on server \'" << server_name << '\'';

  return text.str();
}

#define METHOD( F , OBJ )  \
make_method_params_name(server_name, F, partition_name, class_name, attribute_name, since, until, OBJ)


std::string
Issues::make_text(const std::string& method, const char * what, const char * issue)
{
  return ("remote method " + method + " raised " + what + " exception: " + issue);
}

std::string
Issues::make_text(const std::string& method, const CORBA::SystemException& issue)
{
  std::ostringstream text;
  text << "remote method " + method << " raised CORBA system exception " << &issue;
  return text.str();
}

std::string
Issues::make_report(const std::string& method, const char * action, const std::set<std::string>& data)
{
  std::ostringstream text;

  text << "method " << method << ' ' << action << "\n\twas caused by";

  if (data.size() > 1)
    {
      text << ' ' << data.size() << " issues: ";

      unsigned int count(1);
      for (auto& x : data)
        {
          text << "\n\t(" << count++ << ") " << x;
        }
    }
  else
    {
      text << ": " << *data.begin();
    }

  return text.str();
}

void
Issues::throw_if_failed(const std::string& name) const
{
  if (!m_bad_request.empty())
    {
      const std::string text(make_report(name, "failed", m_bad_request));
      ERS_LOG("throw pbeast::BadRequest: " << text);
      throw pbeast::BadRequest(text.c_str());
    }

  if (!m_failure.empty())
    {
      const std::string text(make_report(name, "failed", m_failure));
      ERS_LOG("throw pbeast::Failure: " << text);
      throw pbeast::Failure(text.c_str());
    }
}

void
Issues::throw_if_not_found(const std::string& name, unsigned int size_of_result) const
{
  if (!m_not_found.empty())
    {
      if (size_of_result == 0 || size_of_result == m_not_found.size())
        {
          const std::string text(make_report(name, "returns no data", m_not_found));
          ERS_LOG("throw pbeast::NotFound: " << text);
          throw pbeast::NotFound(text.c_str());
        }
    }
}


void
pbeast::Server::get_updated_objects(const char* partition_name, const char* class_name, const char* attribute_name_path, ::CORBA::ULongLong since, ::CORBA::ULongLong until, const char* object_mask, std::set<std::string>& data)
{
  std::map<std::string, pbeast::server_var> server_objects;
  get_servers(partition_name, server_objects);

    {
      struct task
      {
        static void
        run(std::mutex& mutex, std::set<std::string>& data, const char * partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, const char* object_mask, Issues& errors, const std::string& server_name, pbeast::server_var server)
        {
          try
            {
              pbeast::StringArray_var objects;
              server->get_updated_objects(partition_name, class_name, attribute_name, since, until, object_mask, objects);

              std::lock_guard<std::mutex> lock(mutex);

              for (unsigned int idx = 0; idx < objects->length(); idx++)
                {
                  data.insert(static_cast<const char *>(objects[idx]));
                }
            }

          CATCH_SERVER_EXCEPTIONS(METHOD("get_updated_objects", object_mask))

        }
      };

      std::mutex mutex;
      std::vector<std::future<void> > results;
      Issues issues;

      for (auto &i : server_objects)
        {
          results.push_back(m_thread_pool.enqueue(std::bind(&task::run, std::ref(mutex), std::ref(data), partition_name, class_name, attribute_name_path, since, until, object_mask, std::ref(issues), std::ref(i.first), i.second)));
        }

      for (unsigned int i = 0; i < results.size(); ++i)
        {
          results[i].get();
        }

      issues.throw_if_failed("get_updated_objects");

      if(data.empty())
        {
          issues.throw_if_not_found("get_updated_objects", results.size());
        }
    }
}

void
pbeast::Server::get_attributes(const char* partition_name, const char* class_name, std::map<uint64_t, std::set<std::string>>& data)
{
  std::map<std::string, pbeast::server_var> server_objects;
  get_servers(partition_name, server_objects);

    {
      struct task
      {
        static void
        run(std::mutex& mutex, std::map<uint64_t, std::set<std::string>>& data, const char * partition_name, const char * class_name, Issues& errors, const std::string& server_name, pbeast::server_var server)
        {
          try
            {
              pbeast::StringDataPointArrayList_var attributes;
              server->get_attributes(partition_name, class_name, attributes);
              std::lock_guard<std::mutex> lock(mutex);

              for (unsigned int idx = 0; idx < attributes->length(); idx++)
                {
                  std::set<std::string>& created_data = data[attributes[idx].ts.created];
                  for (unsigned int j = 0; j < attributes[idx].data.length(); j++)
                    {
                      created_data.insert(static_cast<const char *>(attributes[idx].data[j]));
                    }

                  if (attributes[idx].ts.updated != attributes[idx].ts.created)
                    {
                      std::set<std::string>& updated_data = data[attributes[idx].ts.updated];
                      for (unsigned int j = 0; j < attributes[idx].data.length(); j++)
                        {
                          updated_data.insert(static_cast<const char *>(attributes[idx].data[j]));
                        }
                    }
                }
            }

          CATCH_SERVER_EXCEPTIONS(make_method_params_name(server_name, "get_attributes", partition_name, class_name))

        }
      };

      std::mutex mutex;
      std::vector<std::future<void> > results;
      Issues issues;

      for (auto &i : server_objects)
        {
          results.push_back(m_thread_pool.enqueue(std::bind(&task::run, std::ref(mutex), std::ref(data), partition_name, class_name, std::ref(issues), std::ref(i.first), i.second)));
        }

      for (unsigned int i = 0; i < results.size(); ++i)
        {
          results[i].get();
        }

      issues.throw_if_failed("get_attributes");

      if(data.empty())
        {
          issues.throw_if_not_found("get_attributes");
        }
    }
}

#define GET_ATTRIBUTES_WITH_NESTED_PARAMS \
make_method_params_name(server_name, "get_attributes/get_attribute_type", partition_name, class_name)

// is used by REST auto-complete series call
void
pbeast::Server::get_attributes_with_nested(const char* partition_name, const char* class_name, std::set<std::string>& data)
{
  std::map<std::string, pbeast::server_var> server_objects;
  get_servers(partition_name, server_objects);

    {
      struct task
      {
        static void
        process_class(std::set<std::string>& data, const char * partition_name, const char * class_name, const std::string& prefix, std::mutex& mutex, const std::string& server_name, pbeast::server_var server)
        {
          pbeast::StringDataPointArrayList_var attributes;

          try
            {
              server->get_attributes(partition_name, class_name, attributes);
            }
          catch (CORBA::SystemException& ex)
            {
              ers::error(daq::pbeast::CorbaError( ERS_HERE, &ex, make_method_params_name(server_name, "get_attributes", partition_name, class_name) ));
              throw;
            }

          std::set<std::string> attr_names;

          for (unsigned int idx = 0; idx < attributes->length(); idx++)
            {
              for (unsigned int j = 0; j < attributes[idx].data.length(); j++)
                {
                  attr_names.insert(static_cast<const char *>(attributes[idx].data[j]));
                }
            }

          for (const auto& a : attr_names)
            {
              pbeast::AttributeTypeList_var types;

              try
                {
                  server->get_attribute_type(partition_name, class_name, a.c_str(), types);
                }
              catch (CORBA::SystemException& ex)
                {
                  ers::error(daq::pbeast::CorbaError( ERS_HERE, &ex, make_method_params_name(server_name, "get_attribute_type", partition_name, class_name) ));
                  throw;
                }

              for (unsigned int idx = 0; idx < types->length(); idx++)
                {
                  for (unsigned int idx = 0; idx < types->length(); idx++)
                    {
                      if(pbeast::str2type(static_cast<const char *>(types[idx].type)) == pbeast::OBJECT)
                        {
                          process_class(data, partition_name, static_cast<const char *>(types[idx].type), prefix + a + '/' + static_cast<const char *>(types[idx].type) + '/', mutex, server_name, server);
                        }
                      else
                        {
                          std::lock_guard<std::mutex> lock(mutex);
                          data.insert(prefix + a);
                        }
                    }
                }
            }
        }

        static void
        run(std::mutex& mutex, std::set<std::string>& data, const char * partition_name, const char * class_name, Issues& errors, const std::string& server_name, pbeast::server_var server)
        {
          try
            {
              process_class(data, partition_name, class_name, "", mutex, server_name, server);
            }
          catch (const pbeast::NotFound& ex)
            {
              const std::string method(GET_ATTRIBUTES_WITH_NESTED_PARAMS);
              errors.add(method, ex);
              ERS_DEBUG( 1, "method " << method << " caused pbeast::NotFound exception " << ex.text);
            }
          catch (const pbeast::Failure& ex)
            {
              const std::string method(GET_ATTRIBUTES_WITH_NESTED_PARAMS);
              errors.add(method, ex);
              ers::error(daq::pbeast::MethodFailure( ERS_HERE, method, ex.text ));
            }
          catch (const pbeast::BadRequest& ex)
            {
              const std::string method(GET_ATTRIBUTES_WITH_NESTED_PARAMS);
              errors.add(method, ex);
              ERS_DEBUG( 1, "method " << method << " caused pbeast::BadRequest exception " << ex.text);
            }
          catch (const CORBA::SystemException& ex)
            {
              errors.add(GET_ATTRIBUTES_WITH_NESTED_PARAMS, ex);
            }
        }
      };

      std::mutex mutex;
      std::vector<std::future<void> > results;
      Issues issues;

      for (auto &i : server_objects)
        {
          results.push_back(m_thread_pool.enqueue(std::bind(&task::run, std::ref(mutex), std::ref(data), partition_name, class_name, std::ref(issues), std::ref(i.first), i.second)));
        }

      for (unsigned int i = 0; i < results.size(); ++i)
        {
          results[i].get();
        }

      issues.throw_if_failed("get_attributes");

      if(data.empty())
        {
          issues.throw_if_not_found("get_attributes");
        }
    }

}


void
pbeast::Server::get_classes(const char* partition_name, std::set<std::string>& data)
{
  std::map<std::string, pbeast::server_var> server_objects;
  get_servers(partition_name, server_objects);

    {
      struct task
      {
        static void
        run(std::mutex& mutex, std::set<std::string>& data, const char * partition_name, Issues& errors, const std::string& server_name, pbeast::server_var server)
        {
          try
            {
              pbeast::StringArray_var classes;
              server->get_classes(partition_name, classes);

              std::lock_guard<std::mutex> lock(mutex);

              for (unsigned int idx = 0; idx < classes->length(); idx++)
                {
                  data.insert(static_cast<const char *>(classes[idx]));
                }
            }

          CATCH_SERVER_EXCEPTIONS(make_method_params_name(server_name, "get_classes", partition_name))

        }
      };

      std::mutex mutex;
      std::vector<std::future<void> > results;
      Issues issues;

      for (auto &i : server_objects)
        {
          results.push_back(m_thread_pool.enqueue(std::bind(&task::run, std::ref(mutex), std::ref(data), partition_name, std::ref(issues), std::ref(i.first), i.second)));
        }

      for (unsigned int i = 0; i < results.size(); ++i)
        {
          results[i].get();
        }

      issues.throw_if_failed("get_classes");

      if(data.empty())
        {
          issues.throw_if_not_found("get_classes");
        }
    }
}


void
pbeast::Server::get_partitions(std::set<std::string>& data)
{
  static std::string empty;

  std::map<std::string, pbeast::server_var> server_objects;
  get_servers(empty, server_objects);

  {
    struct task
    {
      static void
      run(std::mutex& mutex, std::set<std::string>& data, Issues& errors, const std::string& server_name, pbeast::server_var server)
      {
        try
          {
            pbeast::StringArray_var partitions;
            server->get_partitions(partitions);

            std::lock_guard<std::mutex> lock(mutex);

            for (unsigned int idx = 0; idx < partitions->length(); idx++)
              {
                data.insert(static_cast<const char *>(partitions[idx]));
              }
          }

        CATCH_SERVER_EXCEPTIONS(make_method_params_name(server_name, "get_partitions"))

      }
    };

    std::mutex mutex;
    std::vector<std::future<void> > results;
    Issues issues;

    for (auto &i : server_objects)
      {
        results.push_back(m_thread_pool.enqueue(std::bind(&task::run, std::ref(mutex), std::ref(data), std::ref(issues), std::ref(i.first), i.second)));
      }

    for (unsigned int i = 0; i < results.size(); ++i)
      {
        results[i].get();
      }

    issues.throw_if_failed("get_partitions");

    if(data.empty())
      {
        issues.throw_if_not_found("get_partitions");
      }
  }
}


void
pbeast::Server::get_attribute_type(const char* partition_name, const char* class_name, const char* attribute_name_path, std::vector<AttrInfoData>& data)
{
  std::map<std::string, pbeast::server_var> server_objects;
  get_servers(partition_name, server_objects);

  std::map<uint64_t, std::pair<std::string, bool> > attr_data;

    {
      struct task
      {
        static void
        run(std::mutex& mutex, std::map<uint64_t, std::pair<std::string, bool>>& attr_data, const char * partition_name, const char * class_name, const char* attribute_name_path, Issues& errors, const std::string& server_name, pbeast::server_var server)
        {
          try
            {
              pbeast::AttributeTypeList_var types;
              server->get_attribute_type(partition_name, class_name, attribute_name_path, types);

              std::lock_guard<std::mutex> lock(mutex);

              for (unsigned int idx = 0; idx < types->length(); idx++)
                {
                  attr_data[types[idx].ts.created] = std::make_pair(types[idx].type,types[idx].is_array);
                  if(types[idx].ts.created != types[idx].ts.updated)
                    {
                      attr_data[types[idx].ts.updated] = std::make_pair(types[idx].type,types[idx].is_array);
                    }
                }
            }

          CATCH_SERVER_EXCEPTIONS(make_method_params_name(server_name, "get_attribute_type", partition_name, class_name, attribute_name_path))

        }
      };

      std::mutex mutex;
      std::vector<std::future<void> > results;
      Issues issues;

      for (auto &i : server_objects)
        {
          results.push_back(m_thread_pool.enqueue(std::bind(&task::run, std::ref(mutex), std::ref(attr_data), partition_name, class_name, attribute_name_path, std::ref(issues), std::ref(i.first), i.second)));
        }

      for (unsigned int i = 0; i < results.size(); ++i)
        {
          results[i].get();
        }

      issues.throw_if_failed("get_attribute_type");

      if(attr_data.empty())
        {
          issues.throw_if_not_found("get_attribute_type");
        }
    }

  // fill type and is-array data
  for (auto & x : attr_data)
    {
      if (data.empty())
        {
          data.emplace_back(x.first, x.first, x.second.first, x.second.second);
        }
      else
        {
          AttrInfoData& last(data.back());
          if (x.second.first == last.m_type && x.second.second == last.m_is_array)
            {
              last.m_updated = x.first;
            }
          else
            {
              data.emplace_back(x.first, x.first, x.second.first, x.second.second);
            }
        }
    }
}


#define READ_ALL( LIST_TYPE, FUNC, CPP_TYPE, SERIE )                                                                                                                                    \
std::map<std::string, pbeast::server_var> server_objects;                                                                                                                               \
get_servers(partition_name, server_objects);                                                                                                                                            \
                                                                                                                                                                                        \
::CORBA::Long ts_precision(1000000);                                                                                                                                                    \
                                                                                                                                                                                        \
  {                                                                                                                                                                                     \
    struct task                                                                                                                                                                         \
    {                                                                                                                                                                                   \
      static ::CORBA::Long                                                                                                                                                              \
      run(std::mutex& mutex, std::map<std::string,std::set<pbeast::SERIE<CPP_TYPE>>>& data, std::map<std::string,                                                                       \
          std::set<uint64_t>>& eov_data, std::map<std::string,std::set<uint64_t>>& upd_data, const char * partition_name, const char * class_name,                                      \
          const char* attribute_name, uint64_t since, uint64_t until, uint64_t prev_interval, uint64_t next_interval, const char* object_mask, bool is_regexp, bool fill_upd,           \
          Issues& errors, const std::string& server_name, pbeast::server_var server, const char * fqn_prefix)                                                                           \
      {                                                                                                                                                                                 \
        try                                                                                                                                                                             \
          {                                                                                                                                                                             \
            pbeast::LIST_TYPE##_var values;                                                                                                                                             \
            pbeast::V64ObjectList_var eov_values, upd_values;                                                                                                                           \
            ::CORBA::Long ts_precision = server->FUNC(partition_name, class_name, attribute_name, since, until, prev_interval, next_interval, object_mask, is_regexp, "", fill_upd, eov_values, upd_values, values); \
                                                                                                                                                                                        \
            if (values->length())                                                                                                                                                       \
              {                                                                                                                                                                         \
                std::lock_guard<std::mutex> lock(mutex);                                                                                                                                \
                pbeast::convert(values.in(), data, fqn_prefix);                                                                                                                         \
              }                                                                                                                                                                         \
                                                                                                                                                                                        \
            if (eov_values->length())                                                                                                                                                   \
              {                                                                                                                                                                         \
                std::lock_guard<std::mutex> lock(mutex);                                                                                                                                \
                pbeast::cvt_void(eov_values, eov_data, fqn_prefix);                                                                                                                     \
              }                                                                                                                                                                         \
                                                                                                                                                                                        \
            if (upd_values->length())                                                                                                                                                   \
              {                                                                                                                                                                         \
                std::lock_guard<std::mutex> lock(mutex);                                                                                                                                \
                pbeast::cvt_void(upd_values, upd_data, fqn_prefix);                                                                                                                     \
              }                                                                                                                                                                         \
                                                                                                                                                                                        \
            return ts_precision;                                                                                                                                                        \
          }                                                                                                                                                                             \
                                                                                                                                                                                        \
        CATCH_SERVER_EXCEPTIONS(METHOD(#FUNC, object_mask))                                                                                                                             \
                                                                                                                                                                                        \
        return 0;                                                                                                                                                                       \
      }                                                                                                                                                                                 \
    };                                                                                                                                                                                  \
                                                                                                                                                                                        \
    std::mutex mutex;                                                                                                                                                                   \
    std::vector<std::future< ::CORBA::Long> > results;                                                                                                                                  \
    Issues issues;                                                                                                                                                                      \
                                                                                                                                                                                        \
    bool get_upd = (fill_upd || fc.m_use_summarize);                                                                                                                                    \
                                                                                                                                                                                        \
    for (auto &i : server_objects)                                                                                                                                                      \
      {                                                                                                                                                                                 \
        results.push_back(m_thread_pool.enqueue(std::bind(&task::run, std::ref(mutex), std::ref(data), std::ref(eov), std::ref(upd),                                                    \
            partition_name, class_name, attribute_name, since, until, prev_interval, next_interval, object_mask, is_regexp, get_upd, std::ref(issues), std::ref(i.first), i.second, fqn_prefix))); \
      }                                                                                                                                                                                 \
                                                                                                                                                                                        \
    for (unsigned int i = 0; i < results.size(); ++i)                                                                                                                                   \
      {                                                                                                                                                                                 \
        ::CORBA::Long r = results[i].get();                                                                                                                                             \
        if (r && r < ts_precision)                                                                                                                                                      \
          ts_precision = r;                                                                                                                                                             \
      }                                                                                                                                                                                 \
                                                                                                                                                                                        \
    issues.throw_if_failed(#FUNC);                                                                                                                                                      \
                                                                                                                                                                                        \
    if(data.empty())                                                                                                                                                                    \
      {                                                                                                                                                                                 \
        issues.throw_if_not_found(#FUNC, results.size());                                                                                                                               \
      }                                                                                                                                                                                 \
                                                                                                                                                                                        \
    pbeast::merge_equal(data, eov);                                                                                                                                                     \
    pbeast::apply_functions(data, eov, upd, fc, since, until, 0);                                                                                                                       \
  }                                                                                                                                                                                     \
                                                                                                                                                                                        \
return ts_precision;


#define READ_SIMPLE_MANY(LIST_TYPE, FUNC, CPP_TYPE)  \
  READ_ALL( LIST_TYPE, FUNC, CPP_TYPE, SeriesData )

#define READ_VECTOR_MANY(LIST_TYPE, FUNC, CPP_TYPE)  \
  READ_ALL( LIST_TYPE, FUNC, CPP_TYPE, SeriesVectorData )


::CORBA::Long
pbeast::Server::get_boolean_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesData<bool>>>& data, const char * fqn_prefix)
{
  READ_SIMPLE_MANY(BooleanObjectList, get_boolean_attributes, bool)
}

::CORBA::Long
pbeast::Server::get_s8_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesData<int8_t>>>& data, const char * fqn_prefix)
{
  READ_SIMPLE_MANY(S8ObjectList, get_s8_attributes, int8_t)
}

::CORBA::Long
pbeast::Server::get_u8_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesData<uint8_t>>>& data, const char * fqn_prefix)
{
  READ_SIMPLE_MANY(U8ObjectList, get_u8_attributes, uint8_t)
}

::CORBA::Long
pbeast::Server::get_s16_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesData<int16_t>>>& data, const char * fqn_prefix)
{
  READ_SIMPLE_MANY(S16ObjectList, get_s16_attributes, int16_t)
}

::CORBA::Long
pbeast::Server::get_u16_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesData<uint16_t>>>& data, const char * fqn_prefix)
{
  READ_SIMPLE_MANY(U16ObjectList, get_u16_attributes, uint16_t)
}

::CORBA::Long
pbeast::Server::get_s32_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesData<int32_t>>>& data, const char * fqn_prefix)
{
  READ_SIMPLE_MANY(S32ObjectList, get_s32_attributes, int32_t)
}

::CORBA::Long
pbeast::Server::get_u32_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesData<uint32_t>>>& data, const char * fqn_prefix)
{
  READ_SIMPLE_MANY(U32ObjectList, get_u32_attributes, uint32_t)
}

::CORBA::Long
pbeast::Server::get_s64_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesData<int64_t>>>& data, const char * fqn_prefix)
{
  READ_SIMPLE_MANY(S64ObjectList, get_s64_attributes, int64_t)
}

::CORBA::Long
pbeast::Server::get_u64_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesData<uint64_t>>>& data, const char * fqn_prefix)
{
  READ_SIMPLE_MANY(U64ObjectList, get_u64_attributes, uint64_t)
}

::CORBA::Long
pbeast::Server::get_float_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesData<float>>>& data, const char * fqn_prefix)
{
  READ_SIMPLE_MANY(FloatObjectList, get_float_attributes, float)
}

::CORBA::Long
pbeast::Server::get_double_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesData<double>>>& data, const char * fqn_prefix)
{
  READ_SIMPLE_MANY(DoubleObjectList, get_double_attributes, double)
}

::CORBA::Long
pbeast::Server::get_string_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesData<std::string>>>& data, const char * fqn_prefix)
{
  READ_SIMPLE_MANY(StringObjectList, get_string_attributes, std::string)
}

::CORBA::Long
pbeast::Server::get_boolean_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesVectorData<bool>>>& data, const char * fqn_prefix)
{
  READ_VECTOR_MANY(BooleanArrayObjectList, get_boolean_array_attributes, bool)
}

::CORBA::Long
pbeast::Server::get_s8_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesVectorData<int8_t>>>& data, const char * fqn_prefix)
{
  READ_VECTOR_MANY(S8ArrayObjectList, get_s8_array_attributes, int8_t)
}

::CORBA::Long
pbeast::Server::get_u8_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesVectorData<uint8_t>>>& data, const char * fqn_prefix)
{
  READ_VECTOR_MANY(U8ArrayObjectList, get_u8_array_attributes, uint8_t)
}

::CORBA::Long
pbeast::Server::get_s16_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesVectorData<int16_t>>>& data, const char * fqn_prefix)
{
  READ_VECTOR_MANY(S16ArrayObjectList, get_s16_array_attributes, int16_t)
}

::CORBA::Long
pbeast::Server::get_u16_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesVectorData<uint16_t>>>& data, const char * fqn_prefix)
{
  READ_VECTOR_MANY(U16ArrayObjectList, get_u16_array_attributes, uint16_t)
}

::CORBA::Long
pbeast::Server::get_s32_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesVectorData<int32_t>>>& data, const char * fqn_prefix)
{
  READ_VECTOR_MANY(S32ArrayObjectList, get_s32_array_attributes, int32_t)
}

::CORBA::Long
pbeast::Server::get_u32_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesVectorData<uint32_t>>>& data, const char * fqn_prefix)
{
  READ_VECTOR_MANY(U32ArrayObjectList, get_u32_array_attributes, uint32_t)
}

::CORBA::Long
pbeast::Server::get_s64_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesVectorData<int64_t>>>& data, const char * fqn_prefix)
{
  READ_VECTOR_MANY(S64ArrayObjectList, get_s64_array_attributes, int64_t)
}

::CORBA::Long
pbeast::Server::get_u64_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesVectorData<uint64_t>>>& data, const char * fqn_prefix)
{
  READ_VECTOR_MANY(U64ArrayObjectList, get_u64_array_attributes, uint64_t)
}

::CORBA::Long
pbeast::Server::get_float_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesVectorData<float>>>& data, const char * fqn_prefix)
{
  READ_VECTOR_MANY(FloatArrayObjectList, get_float_array_attributes, float)
}

::CORBA::Long
pbeast::Server::get_double_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesVectorData<double>>>& data, const char * fqn_prefix)
{
  READ_VECTOR_MANY(DoubleArrayObjectList, get_double_array_attributes, double)
}

::CORBA::Long
pbeast::Server::get_string_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint64_t>>& eov, std::map<std::string, std::set<uint64_t>>& upd, std::map<std::string,std::set<pbeast::SeriesVectorData<std::string>>>& data, const char * fqn_prefix)
{
  READ_VECTOR_MANY(StringArrayObjectList, get_string_array_attributes, std::string)
}



#define READ_DS_ALL( LIST_TYPE, FUNC, CPP_TYPE, SERIE )                                                                                                             \
std::map<std::string, pbeast::server_var> server_objects;                                                                                                           \
get_servers(partition_name, server_objects);                                                                                                                        \
                                                                                                                                                                    \
  {                                                                                                                                                                 \
    struct task                                                                                                                                                     \
    {                                                                                                                                                               \
      static void                                                                                                                                                   \
      run(std::mutex& mutex, std::map<std::string,std::set<pbeast::SERIE<CPP_TYPE>>>& data, std::map<std::string, std::set<uint32_t>>& eov_data,                    \
          std::map<std::string, std::set<uint32_t>>& upd_data, const char * partition_name, const char * class_name, const char* attribute_name,                    \
          uint64_t since, uint64_t until, uint64_t prev_interval, uint64_t next_interval, uint64_t interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp,               \
          bool fill_upd, Issues& errors, const std::string& server_name, pbeast::server_var server, const char * fqn_prefix)                                        \
      {                                                                                                                                                             \
        try                                                                                                                                                         \
          {                                                                                                                                                         \
            pbeast::LIST_TYPE##_var values;                                                                                                                         \
            pbeast::V32ObjectList_var eov_values, upd_values;                                                                                                       \
            server->FUNC(partition_name, class_name, attribute_name, since, until, prev_interval, next_interval, interval, fill_gaps, object_mask, is_regexp, "", fill_upd, eov_values, upd_values, values); \
                                                                                                                                                                    \
            if (values->length())                                                                                                                                   \
              {                                                                                                                                                     \
                std::lock_guard<std::mutex> lock(mutex);                                                                                                            \
                pbeast::convert(values.in(), data, fqn_prefix);                                                                                                     \
              }                                                                                                                                                     \
                                                                                                                                                                    \
            if (eov_values->length())                                                                                                                               \
              {                                                                                                                                                     \
                std::lock_guard<std::mutex> lock(mutex);                                                                                                            \
                pbeast::cvt_void(eov_values, eov_data, fqn_prefix);                                                                                                 \
              }                                                                                                                                                     \
                                                                                                                                                                    \
            if (upd_values->length())                                                                                                                               \
              {                                                                                                                                                     \
                std::lock_guard<std::mutex> lock(mutex);                                                                                                            \
                pbeast::cvt_void(upd_values, upd_data, fqn_prefix);                                                                                                 \
              }                                                                                                                                                     \
          }                                                                                                                                                         \
                                                                                                                                                                    \
        CATCH_SERVER_EXCEPTIONS(METHOD(#FUNC, object_mask))                                                                                                         \
                                                                                                                                                                    \
      }                                                                                                                                                             \
    };                                                                                                                                                              \
                                                                                                                                                                    \
    std::mutex mutex;                                                                                                                                               \
    std::vector<std::future<void>> results;                                                                                                                         \
    Issues issues;                                                                                                                                                  \
                                                                                                                                                                    \
    for (auto &i : server_objects)                                                                                                                                  \
      {                                                                                                                                                             \
        results.push_back(m_thread_pool.enqueue(std::bind(&task::run, std::ref(mutex), std::ref(data32), std::ref(eov32), std::ref(upd32), partition_name,          \
            class_name, attribute_name, since, until, prev_interval, next_interval, interval, fill_gaps, object_mask, is_regexp, fill_upd, std::ref(issues), std::ref(i.first), i.second, fqn_prefix))); \
      }                                                                                                                                                             \
                                                                                                                                                                    \
    for (unsigned int i = 0; i < results.size(); ++i)                                                                                                               \
      {                                                                                                                                                             \
        results[i].get();                                                                                                                                           \
      }                                                                                                                                                             \
                                                                                                                                                                    \
    issues.throw_if_failed(#FUNC);                                                                                                                                  \
                                                                                                                                                                    \
    if(data32.empty())                                                                                                                                              \
      {                                                                                                                                                             \
        issues.throw_if_not_found(#FUNC, results.size());                                                                                                           \
      }                                                                                                                                                             \
                                                                                                                                                                    \
    pbeast::apply_functions(data32, eov32, upd32, fc, since, until, interval);                                                                                      \
  }


#define READ_DS_SIMPLE_MANY(LIST_TYPE, FUNC, CPP_TYPE)                                                                                           \
  if (fc.m_use_summarize)                                                                                                                        \
    {                                                                                                                                            \
      fc.set_interval(interval);                                                                                                                 \
      std::map<std::string, std::set<uint64_t>> upd, eov;                                                                                        \
      std::map<std::string,std::set<pbeast::SeriesData<CPP_TYPE>>> data;                                                                         \
      get_##FUNC(partition_name, class_name, attribute_name, since, until, prev_interval, next_interval, object_mask, is_regexp, fc, true, eov, upd, data, fqn_prefix); \
      pbeast::downsample_void(interval, eov, eov32);                                                                                             \
      pbeast::downsample_void(interval, upd, upd32);                                                                                             \
      pbeast::downsample(interval, data, data32, fill_gaps);                                                                                                \
    }                                                                                                                                            \
  else                                                                                                                                           \
    {                                                                                                                                            \
      READ_DS_ALL( LIST_TYPE, get_downsample_##FUNC, CPP_TYPE, DownsampledSeriesData )                                                           \
    }                                                                                                                                            \
                                                                                                                                                 \
  if (fc.no_aggregations() == false)                                                                                                             \
    {                                                                                                                                            \
      uint32_t first, last;                                                                                                                      \
      get_query_interval(data32, first, last);                                                                                                   \
      uint32_t len = (last - first) / interval + 1;                                                                                              \
      AggregatedPrimitiveData<CPP_TYPE> values(interval, first, len);                                                                            \
      values.process(data32, eov32);                                                                                                             \
      if (fc.m_keep_data == false)                                                                                                               \
        {                                                                                                                                        \
          data32.clear();                                                                                                                        \
          eov32.clear();                                                                                                                         \
          upd32.clear();                                                                                                                         \
        }                                                                                                                                        \
      values.store(fc, data32, since, until);                                                                                                    \
    }                                                                                                                                            \


void
pbeast::Server::get_downsample_boolean_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesData<bool>>>& data32, const char * fqn_prefix)
{
  READ_DS_SIMPLE_MANY(BooleanDownSampleObjectList, boolean_attributes, bool)
}

void
pbeast::Server::get_downsample_s8_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesData<int8_t>>>& data32, const char * fqn_prefix)
{
  READ_DS_SIMPLE_MANY(S8DownSampleObjectList, s8_attributes, int8_t)
}

void
pbeast::Server::get_downsample_u8_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesData<uint8_t>>>& data32, const char * fqn_prefix)
{
  READ_DS_SIMPLE_MANY(U8DownSampleObjectList, u8_attributes, uint8_t)
}

void
pbeast::Server::get_downsample_s16_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesData<int16_t>>>& data32, const char * fqn_prefix)
{
  READ_DS_SIMPLE_MANY(S16DownSampleObjectList, s16_attributes, int16_t)
}

void
pbeast::Server::get_downsample_u16_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesData<uint16_t>>>& data32, const char * fqn_prefix)
{
  READ_DS_SIMPLE_MANY(U16DownSampleObjectList, u16_attributes, uint16_t)
}

void
pbeast::Server::get_downsample_s32_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesData<int32_t>>>& data32, const char * fqn_prefix)
{
  READ_DS_SIMPLE_MANY(S32DownSampleObjectList, s32_attributes, int32_t)
}

void
pbeast::Server::get_downsample_u32_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesData<uint32_t>>>& data32, const char * fqn_prefix)
{
  READ_DS_SIMPLE_MANY(U32DownSampleObjectList, u32_attributes, uint32_t)
}

void
pbeast::Server::get_downsample_s64_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesData<int64_t>>>& data32, const char * fqn_prefix)
{
  READ_DS_SIMPLE_MANY(S64DownSampleObjectList, s64_attributes, int64_t)
}

void
pbeast::Server::get_downsample_u64_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesData<uint64_t>>>& data32, const char * fqn_prefix)
{
  READ_DS_SIMPLE_MANY(U64DownSampleObjectList, u64_attributes, uint64_t)
}

void
pbeast::Server::get_downsample_float_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesData<float>>>& data32, const char * fqn_prefix)
{
  READ_DS_SIMPLE_MANY(FloatDownSampleObjectList, float_attributes, float)
}

void
pbeast::Server::get_downsample_double_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesData<double>>>& data32, const char * fqn_prefix)
{
  READ_DS_SIMPLE_MANY(DoubleDownSampleObjectList, double_attributes, double)
}

void
pbeast::Server::get_downsample_string_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesData<std::string>>>& data32, const char * fqn_prefix)
{
  READ_DS_SIMPLE_MANY(StringDownSampleObjectList, string_attributes, std::string)
}

#define READ_DS_VECTOR_MANY(LIST_TYPE, FUNC, CPP_TYPE)                                                                                           \
  if (fc.m_use_summarize)                                                                                                                        \
    {                                                                                                                                            \
      fc.set_interval(interval);                                                                                                                 \
      std::map<std::string, std::set<uint64_t>> upd, eov;                                                                                        \
      std::map<std::string,std::set<pbeast::SeriesVectorData<CPP_TYPE>>> data;                                                                   \
      get_##FUNC(partition_name, class_name, attribute_name, since, until, prev_interval, next_interval, object_mask, is_regexp, fc, true, eov, upd, data, fqn_prefix); \
      pbeast::downsample_void(interval, eov, eov32);                                                                                             \
      pbeast::downsample_void(interval, upd, upd32);                                                                                             \
      pbeast::downsample(interval, data, data32, fill_gaps);                                                                                                \
    }                                                                                                                                            \
  else                                                                                                                                           \
    {                                                                                                                                            \
      READ_DS_ALL( LIST_TYPE, get_downsample_##FUNC, CPP_TYPE, DownsampledSeriesVectorData )                                                     \
    }                                                                                                                                            \
                                                                                                                                                 \
  if (fc.no_aggregations() == false)                                                                                                             \
    {                                                                                                                                            \
      uint32_t first, last;                                                                                                                      \
      get_query_interval(data32, first, last);                                                                                                   \
      uint32_t len = (last - first) / interval + 1;                                                                                              \
      AggregatedVectorData<CPP_TYPE> values(interval, first, len);                                                                               \
      values.process(data32, eov32);                                                                                                             \
      if (fc.m_keep_data == false)                                                                                                               \
        {                                                                                                                                        \
          data32.clear();                                                                                                                        \
          eov32.clear();                                                                                                                         \
          upd32.clear();                                                                                                                         \
        }                                                                                                                                        \
      values.store(fc, data32, since, until);                                                                                                    \
    }                                                                                                                                            \


void
pbeast::Server::get_downsample_boolean_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesVectorData<bool>>>& data32, const char * fqn_prefix)
{
  READ_DS_VECTOR_MANY(BooleanDownSampleArrayObjectList, boolean_array_attributes, bool)
}

void
pbeast::Server::get_downsample_s8_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesVectorData<int8_t>>>& data32, const char * fqn_prefix)
{
  READ_DS_VECTOR_MANY(S8DownSampleArrayObjectList, s8_array_attributes, int8_t)
}

void
pbeast::Server::get_downsample_u8_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesVectorData<uint8_t>>>& data32, const char * fqn_prefix)
{
  READ_DS_VECTOR_MANY(U8DownSampleArrayObjectList, u8_array_attributes, uint8_t)
}

void
pbeast::Server::get_downsample_s16_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesVectorData<int16_t>>>& data32, const char * fqn_prefix)
{
  READ_DS_VECTOR_MANY(S16DownSampleArrayObjectList, s16_array_attributes, int16_t)
}

void
pbeast::Server::get_downsample_u16_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesVectorData<uint16_t>>>& data32, const char * fqn_prefix)
{
  READ_DS_VECTOR_MANY(U16DownSampleArrayObjectList, u16_array_attributes, uint16_t)
}

void
pbeast::Server::get_downsample_s32_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesVectorData<int32_t>>>& data32, const char * fqn_prefix)
{
  READ_DS_VECTOR_MANY(S32DownSampleArrayObjectList, s32_array_attributes, int32_t)
}

void
pbeast::Server::get_downsample_u32_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesVectorData<uint32_t>>>& data32, const char * fqn_prefix)
{
  READ_DS_VECTOR_MANY(U32DownSampleArrayObjectList, u32_array_attributes, uint32_t)
}

void
pbeast::Server::get_downsample_s64_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesVectorData<int64_t>>>& data32, const char * fqn_prefix)
{
  READ_DS_VECTOR_MANY(S64DownSampleArrayObjectList, s64_array_attributes, int64_t)
}

void
pbeast::Server::get_downsample_u64_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesVectorData<uint64_t>>>& data32, const char * fqn_prefix)
{
  READ_DS_VECTOR_MANY(U64DownSampleArrayObjectList, u64_array_attributes, uint64_t)
}

void
pbeast::Server::get_downsample_float_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesVectorData<float>>>& data32, const char * fqn_prefix)
{
  READ_DS_VECTOR_MANY(FloatDownSampleArrayObjectList, float_array_attributes, float)
}

void
pbeast::Server::get_downsample_double_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesVectorData<double>>>& data32, const char * fqn_prefix)
{
  READ_DS_VECTOR_MANY(DoubleDownSampleArrayObjectList, double_array_attributes, double)
}

void
pbeast::Server::get_downsample_string_array_attributes(const char* partition_name, const char* class_name, const char* attribute_name, ::CORBA::ULongLong since, ::CORBA::ULongLong until, ::CORBA::ULongLong prev_interval, ::CORBA::ULongLong next_interval, ::CORBA::ULong interval, pbeast::FillGaps::Type fill_gaps, const char* object_mask, bool is_regexp, const pbeast::FunctionConfig& fc, bool fill_upd, std::map<std::string, std::set<uint32_t>>& eov32, std::map<std::string, std::set<uint32_t>>& upd32, std::map<std::string,std::set<pbeast::DownsampledSeriesVectorData<std::string>>>& data32, const char * fqn_prefix)
{
  READ_DS_VECTOR_MANY(StringDownSampleArrayObjectList, string_array_attributes, std::string)
}


void
pbeast::utils::throw_on_error(const std::string& value, char * sanity, const char * ftype)
{
  if (*sanity != 0 || errno == ERANGE)
    {
      std::ostringstream text;
      text << "failed to parse " << ftype << " value \'" << value << "\':";
      if (*sanity)
        text << " on unrecognized characters \'" << sanity << "\'";
      if (errno)
        {
          char buffer[1024];
          buffer[0] = 0;
          strerror_r(errno, buffer, 1024);
          errno = 0;
          text << " with code " << errno << ", reason = \'" << buffer << '\'';
        }
      errno = 0;
      throw std::runtime_error(text.str().c_str());
    }
}

double
pbeast::utils::str2double(const std::string& value)
{
  char * sanity;
  double v = strtod(value.c_str(), &sanity);

  throw_on_error(value, sanity, "double-precision floating-point");

  return v;
}

unsigned long
pbeast::utils::str2uint(const std::string& value)
{
  char * sanity;
  unsigned long v = strtoul(value.c_str(), &sanity, 0);

  throw_on_error(value, sanity, "unsigned integer");

  return v;
}


static constexpr char s_node_separator = '.';

void
pbeast::utils::create_alias_map(const pbeast::FunctionConfig& fc, const std::set<std::string>& names, std::map<std::string, std::string>& out)
{
  const char value_separator = pbeast::FunctionConfig::get_value_separator();

  for(const auto& x : names)
    out[x] = x;

  for (const auto& f : fc.m_aliases)
    {
      try
        {
          switch (f.first)
            {
              case pbeast::Alias:
                if (out.size() > 1)
                  throw std::runtime_error("there is more than one object in query result");

                for (auto& x : out)
                  x.second = f.second;

                break;


              case pbeast::AliasByMetric:
                for (auto& x : out)
                  {
                    std::string::size_type pos = x.second.rfind(s_node_separator);
                    if (pos != std::string::npos)
                      x.second.erase(0, pos+1);
                  }

                break;

              case pbeast::AliasByNode: {
                std::vector<unsigned long> idxs;

                std::size_t start = 0, end = 0;
                while ((end = f.second.find(value_separator, start)) != std::string::npos)
                  {
                    idxs.push_back(str2uint(f.second.substr(start, end - start)));
                    start = end + 1;
                  }
                idxs.push_back(str2uint(f.second.substr(start)));

                for (auto& x : out)
                  {
                    std::string str;

                    for (auto& idx : idxs)
                      {
                        start = end = 0;
                        std::size_t i = 0;
                        while ((end = x.second.find(s_node_separator, start)) != std::string::npos)
                          {
                            if (i == idx)
                              break;
                            start = end + 1;
                            i++;
                          }

                        if (i == idx)
                          {
                            if (str.empty() == false)
                              str.push_back(s_node_separator);
                            str.append(x.second.substr(start, end - start));
                          }
                        else
                          {
                            std::ostringstream text;
                            text << "failed to find node " << idx << " in value \"" << x.second << "\" of \"" << x.first << "\"";
                            throw std::runtime_error(text.str().c_str());
                          }
                      }

                    x.second = str;
                  }

                break; }


              case pbeast::AliasSub: {
                std::string::size_type pos = f.second.find(value_separator);
                if(pos == std::string::npos)
                  {
                    std::ostringstream text;
                    text << "bad format of function parameter expecting \"regex-search" << value_separator << "replace-value\" format";
                    throw std::runtime_error(text.str().c_str());
                  }

                const std::string re_str(pbeast::DataFile::decode(f.second.substr(0, pos), '%'));
                const std::string sub_value(pbeast::DataFile::decode(f.second.substr(pos+1), '%'));

                std::unique_ptr<boost::regex> re;

                try
                  {
                    re.reset(new boost::regex(re_str));
                  }
                catch (const std::exception& ex)
                  {
                    throw daq::pbeast::RegularExpressionError(ERS_HERE, "cannot parse regular expression", re_str, ex);
                  }

                for (auto& x : out)
                  {
                    try
                      {
                        x.second = boost::regex_replace(x.second, *re, sub_value);
                      }
                    catch (const std::exception& ex)
                      {
                        std::ostringstream text;
                        text << "boost::regex_replace(" << re_str << "," << sub_value << ") failed for string ";
                        throw daq::pbeast::RegularExpressionError(ERS_HERE, text.str().c_str(), x.second, ex);
                      }
                  }

                break; }
            }
        }
      catch (const std::exception& ex)
        {
          std::ostringstream text;
          text << "function " << pbeast::FunctionConfig::function_name(f.first) << "(" << f.second << ") failed\n\twas caused by: " << ex.what();
          throw pbeast::BadRequest(text.str().c_str());
        }

      std::map<std::string,std::string> reverse_map;

      for (const auto& x : out)
        {
          auto res = reverse_map.emplace(x.second, x.first);
          if(!res.second)
            {
              std::ostringstream text;
              text << "the configuration of alias functions results conversion of \"" << x.first << "\" and \"" << (res.first)->second << "\" object names into " << x.second << " alias name";
              throw pbeast::BadRequest(text.str().c_str());
            }
        }
    }
}

template<typename T>
  void
  __cvt_names(std::map<std::string, std::set<T>>& data, const std::map<std::string, std::string>& cvt_map)
  {
    std::map<std::string, std::set<T>> new_data;

    for (auto& x : data)
      new_data[cvt_map.at(x.first)].swap(x.second);

    data.swap(new_data);
  }

void
pbeast::utils::cvt_names(std::map<std::string,std::set<uint32_t>>& data, const std::map<std::string, std::string>& cvt_map)
{
  __cvt_names(data, cvt_map);
}

void
pbeast::utils::cvt_names(std::map<std::string,std::set<uint64_t>>& data, const std::map<std::string, std::string>& cvt_map)
{
  __cvt_names(data, cvt_map);
}
