#include <cstdint>

#include <string>
#include <filesystem>

#include <boost/program_options.hpp>

#include "pbeast/input-data-file.h"
#include "pbeast/output-data-file.h"
#include "pbeast/copy-file.h"

namespace po = boost::program_options;


int
main(int argc, char ** argv)
{
  po::options_description desc("P-BEAST utility to compress / decompress data files. It may also be used to downsample data.");

  std::string file_in;
  std::string file_out;

  uint64_t in_size, out_size;

  uint32_t downsample_interval = 0;
  bool zip_catalog = false;
  bool zip_data = false;
  bool catalogize_strings = false;

  try
    {
      desc.add_options()
          ("input-file,f", po::value<std::string>(&file_in)->required(), "Name of input file")
          ("output-file,o", po::value<std::string>(&file_out), "Name of output file (different from input except downsampling case)")
          ("downsample-interval,e", po::value<uint32_t>(&downsample_interval)->default_value(downsample_interval), "Downsample data with this interval")
          ("zip-catalog,c", "Zip catalog")
          ("zip-data,d", "Zip data")
          ("catalogize-strings,s", "catalogize string data")
          ("zip,z", "zip all (i.e. catalog, data, catalogize strings data)")
          ("unzip,u", "unzip all")
          ("help,h","Print help message");

      po::variables_map vm;
      po::store(po::parse_command_line(argc, argv, desc), vm);

      if (vm.count("help"))
        {
          std::cout << desc << std::endl;
          return 0;
        }

      po::notify(vm);

      if (vm.count("zip"))
        {
          zip_catalog = zip_data = catalogize_strings = true;
        }
      else if (vm.count("unzip"))
        {
          zip_catalog = zip_data = catalogize_strings = false;
        }
      else
        {
          if (vm.count("zip-data"))
            {
              zip_data = true;
            }

          if (vm.count("zip-catalog"))
            {
              zip_catalog = true;
            }

          if (vm.count("catalogize-strings"))
            {
              catalogize_strings = true;
            }
        }

      pbeast::InputDataFile in(file_in);

      if (in.get_zip_catalog_flag() == zip_catalog &&
          in.get_zip_data_flag() == zip_data &&
          in.get_downsample_interval() == downsample_interval &&
          (in.get_data_type() == pbeast::STRING && in.get_catalogize_strings_flag() == catalogize_strings))
        {
          std::cerr << "WARNING: nothing to do, input file already has required zip:\n" <<
              "  zip catalog: " << in.get_zip_catalog_flag() << std::endl <<
              "  zip data: " << in.get_zip_data_flag() << std::endl;

          if(in.get_data_type() == pbeast::STRING)
            {
              std::cerr << "  catalogize strings: " << in.get_catalogize_strings_flag() << std::endl;
            }

          std::cerr << "and downsample options:\n"
              "  downsample interval: " << in.get_downsample_interval()  << std::endl;

          return 1;
        }

      if(file_out.empty())
        {
          if(downsample_interval == 0)
            {
              std::cerr << "ERROR: out file is required parameter" << std::endl;
              return 1;
            }
          else
            {
              std::ostringstream f;
              f << file_in << '.' << downsample_interval;
              file_out = f.str();
            }
        }

      pbeast::OutputDataFile out(file_out, in.get_partition(), in.get_class(), in.get_class_path(),
          in.get_attribute(), in.get_base_timestamp(), in.get_precision(), in.get_data_type(), in.get_is_array(),
          in.get_number_of_series(), zip_catalog, zip_data, catalogize_strings, downsample_interval);

      pbeast::copy_file(in, out);

      out_size = out.close();
    }
  catch (std::exception& ex)
    {
      std::cerr << "ERROR: " << ex.what() << std::endl;
      return 1;
    }

  in_size = std::filesystem::file_size(file_in);

  std::cout
    << "size of input file: " << in_size << std::endl
    << "size of output file: " << out_size << std::endl
    << "data compression ratio: " << (double)in_size / (double)out_size << std::endl;


  return 0;
}
