#ifndef PBEAST_COPY_FILE_H
#define PBEAST_COPY_FILE_H

#include <boost/regex.hpp>

#include <pbeast/exceptions.h>
#include <pbeast/input-data-file.h>
#include <pbeast/output-data-file.h>
#include <pbeast/downsample-data.h>

#define WRITE_RAW_PRIMITIVE_DATA( NAME )                                                       \
  to.write_data(NAME, data);                                                                   \
  count += data.size() * 3 + 1;                                                                \
  data.clear();

#define WRITE_RAW_VECTOR_DATA( NAME )                                                          \
  to.write_data(NAME, data);                                                                   \
  count += data.size() * 2 + 1;                                                                \
  for (const auto& x : data)                                                                   \
    count += x.get_value().size();                                                             \
  data.clear();

#define WRITE_DS_PRIMITIVE_DATA( NAME )                                                        \
  downsample(to.get_downsample_interval(), data, downsampled_data, pbeast::FillGaps::Near );   \
  to.write_data(NAME, downsampled_data);                                                       \
  count += data.size() * 4 + 1;                                                                \
  data.clear();                                                                                \
  downsampled_data.clear();

#define WRITE_DS_VECTOR_DATA( NAME )                                                           \
  downsample(to.get_downsample_interval(), data, downsampled_data, pbeast::FillGaps::Near );   \
  to.write_data(NAME, downsampled_data);                                                       \
  count += data.size() * 2 + 1;                                                                \
  for (const auto& x : data)                                                                   \
    count += 3 * x.get_value().size();                                                         \
  data.clear();                                                                                \
  downsampled_data.clear();


namespace pbeast {

  /**
   *  \brief Template implementation of input data file stream copy to output one.
   *
   *  It is used by the copy_file(pbeast::InputDataFile&, pbeast::OutputDataFile&) method.
   *
   *  Parameters:
   *  \param from      input data file stream
   *  \param to        output data file stream
   *  \return          number of copied primitive data types (used for monitoring)
   */

  template<class T>
    uint64_t
    copy(pbeast::InputDataFile& from, pbeast::OutputDataFile& to, const std::string& ids, boost::regex * re)
    {
      uint64_t count(0);
      std::string name;

      if (from.get_is_array() == false)
        {
          std::vector<pbeast::SeriesData<T> > data;

          if (to.is_downsampled() == false)
            {
              // copy all
              if (ids.empty())
                {
                  while (from.read_next(name, data))
                    {
                      WRITE_RAW_PRIMITIVE_DATA(name);
                    }
                }
              // copy using regexp
              else if (re)
                {
                  const std::map<std::string, uint64_t>& info = from.get_series_info();

                  for (auto & x : info)
                    {
                      try
                        {
                          if (boost::regex_match(x.first, *re))
                            {
                              from.read_serie(x.second, data);
                              WRITE_RAW_PRIMITIVE_DATA(x.first);
                            }
                        }
                      catch (std::exception& ex)
                        {
                          throw daq::pbeast::RegularExpressionError(ERS_HERE, "boost::regex_match() failed for string", x.first, ex);
                        }
                    }
                }
              // copy one
              else
                {
                  if (from.read_serie(ids, data))
                    {
                      WRITE_RAW_PRIMITIVE_DATA(ids);
                    }
                }
            }
          else
            {
              std::vector<pbeast::DownsampledSeriesData<T> > downsampled_data;

              // copy all
              if (ids.empty())
                {
                  while (from.read_next(name, data))
                    {
                      WRITE_DS_PRIMITIVE_DATA(name);
                    }
                }
              // copy using regexp
              else if (re)
                {
                  const std::map<std::string, uint64_t>& info = from.get_series_info();

                  for (auto & x : info)
                    {
                      try
                        {
                          if (boost::regex_match(x.first, *re))
                            {
                              from.read_serie(x.second, data);
                              WRITE_DS_PRIMITIVE_DATA(x.first);
                            }
                        }
                      catch (std::exception& ex)
                        {
                          throw daq::pbeast::RegularExpressionError(ERS_HERE, "boost::regex_match() failed for string", x.first, ex);
                        }
                    }
                }
              // copy one
              else
                {
                  if (from.read_serie(ids, data))
                    {
                      WRITE_DS_PRIMITIVE_DATA(ids);
                    }
                }
            }
        }
      else
        {
          std::vector<pbeast::SeriesVectorData<T> > data;

          if (to.is_downsampled() == false)
            {
              // copy all
              if (ids.empty())
                {
                  while (from.read_next(name, data))
                    {
                      WRITE_RAW_VECTOR_DATA(name);
                    }
                }
              // copy using regexp
              else if (re)
                {
                  const std::map<std::string, uint64_t>& info = from.get_series_info();

                  for (auto & x : info)
                    {
                      try
                        {
                          if (boost::regex_match(x.first, *re))
                            {
                              from.read_serie(x.second, data);
                              WRITE_RAW_VECTOR_DATA(x.first);
                            }
                        }
                      catch (std::exception& ex)
                        {
                          throw daq::pbeast::RegularExpressionError(ERS_HERE, "boost::regex_match() failed for string", x.first, ex);
                        }
                    }
                }
              // copy one
              else
                {
                  if (from.read_serie(ids, data))
                    {
                      WRITE_RAW_VECTOR_DATA(ids);
                    }
                }
            }
          else
            {
              std::vector<pbeast::DownsampledSeriesVectorData<T> > downsampled_data;

              // copy all
              if (ids.empty())
                {
                  while (from.read_next(name, data))
                    {
                      WRITE_DS_VECTOR_DATA(name);
                    }
                }
              // copy using regexp
              else if (re)
                {
                  const std::map<std::string, uint64_t>& info = from.get_series_info();

                  for (auto & x : info)
                    {
                      try
                        {
                          if (boost::regex_match(x.first, *re))
                            {
                              from.read_serie(x.second, data);
                              WRITE_DS_VECTOR_DATA(x.first);
                            }
                        }
                      catch (std::exception& ex)
                        {
                          throw daq::pbeast::RegularExpressionError(ERS_HERE, "boost::regex_match() failed for string", x.first, ex);
                        }
                    }
                }
              // copy one
              else
                {
                  if (from.read_serie(ids, data))
                    {
                      WRITE_DS_VECTOR_DATA(ids);
                    }
                }
            }
        }

      return count;
    }

  /**
   *  \brief Makes a copy of input data stream to output one.
   *
   *  Typically is used for downsampling
   *
   *  Parameters:
   *  \param from      input data file stream
   *  \param to        output data file stream
   *  \return          total number of copied primitive data types (used for monitoring)
   */

  uint64_t
  copy_file(pbeast::InputDataFile& from, pbeast::OutputDataFile& to, const std::string& ids = "", boost::regex * re = nullptr);

}


#endif
