#ifndef PBEAST_INPUT_DATA_FILE_H
#define PBEAST_INPUT_DATA_FILE_H

#include <stdint.h>
#include <string>
#include <sstream>
#include <map>

#include <boost/regex.hpp>

#include "pbeast/call-stats.h"
#include "pbeast/input-stream.h"
#include "pbeast/data-file.h"
#include "pbeast/series-data.h"
#include "pbeast/series-info.h"

namespace pbeast
{
  struct File;

  class InputDataFileBase : public DataFile
  {
  public:

    InputDataFileBase(const std::string& file_name) :
        DataFile(file_name)
    {
      ;
    }

    InputDataFileBase(const DataFile& data_file, const std::shared_ptr<std::map<std::string, uint64_t>>& series_info) :
        DataFile(data_file), m_series_info(series_info)
    {
      ;
    }

    InputDataFileBase(const DataFile& data_file, const std::vector<SeriesInfo *>& series_info) :
        DataFile(data_file)
    {
      init_series_info();

      for (const auto& x : series_info)
        (*m_series_info)[x->get_name()] = x->get_data_position();
    }

    const std::map<std::string, uint64_t>&
    get_series_info() const
    {
      return *m_series_info;
    }

    const std::shared_ptr<std::map<std::string, uint64_t>>&
    get_series_info_ptr() const
    {
      return m_series_info;
    }

  protected:

    void
    init_series_info()
    {
      m_series_info = std::make_shared<std::map<std::string, uint64_t>>();
    }

    std::shared_ptr<std::map<std::string, uint64_t>> m_series_info;
  };

  class InputDataFile : public InputDataFileBase
  {

  public:

    InputDataFile(const std::string& file_name) :
        InputDataFileBase(file_name), m_string_stream(nullptr), m_input_stream(nullptr), m_it_inited(false), m_total_data_size(0)
    {
      read_info();
    }

    InputDataFile(std::stringstream * stream) :
        InputDataFileBase(""), m_string_stream(stream), m_input_stream(nullptr), m_it_inited(false)
    {
      read_info();
    }

    InputDataFile(const std::filesystem::path& path, const std::string& name, pbeast::File& file, CallStats& stat);

    ~InputDataFile()
    {
      close();
    }

    void
    open()
    {
      if (!m_input_stream)
        {
          m_input_stream = (!m_string_stream ? new pbeast::InputStream(m_file_name) : new pbeast::InputStream(m_string_stream));
          m_it_inited = false;
        }
    }

    void
    close()
    {
      if (m_input_stream)
        {
          delete m_input_stream;
          m_input_stream = nullptr;
        }
    }

    InputDataFile(const InputDataFile& other) = delete;
    InputDataFile&
    operator=(const InputDataFile&) = delete;

    void
    read_info();

    inline void
    read_timestamps(uint64_t& ts_created, uint64_t& ts_last_updated)
    {
      uint64_t created, interval;
      m_input_stream->read(created);
      m_input_stream->read(interval);

      ts_created = m_base_timestamp + (created * static_cast<uint64_t>(m_precision));
      ts_last_updated = ts_created + (interval * static_cast<uint64_t>(m_precision));
    }

    template<class T>
      uint64_t
      read_serie(const std::string& name, std::vector<T>& data)
      {
        const auto it = m_series_info->find(name);

        if (it != m_series_info->end())
          {
            reset(m_zip_data, it->second);
            return read_data(data);
          }
        else
          {
            return 0;
          }
      }

    template<class V, class T>
      uint64_t
      read_serie(const std::string& name, V since, V until, std::vector<T>& data)
      {
        const auto it = m_series_info->find(name);

        if (it != m_series_info->end())
          {
            reset(m_zip_data, it->second);
            return read_data(since, until, data);
          }
        else
          {
            return 0;
          }
      }

    template<class T>
      uint64_t
      read_serie(uint64_t pos, std::vector<T>& data)
      {
        reset(m_zip_data, pos);
        return read_data(data);
      }

    template<class V, class T>
      uint64_t
      read_serie(uint64_t pos, V since, V until, std::vector<T>& data)
      {
        reset(m_zip_data, pos);
        return read_data(since, until, data);
      }

    template<class T>
      bool
      read_next(std::string& name, std::vector<T>& data)
      {
        if (!m_it_inited)
          init_iterator(m_zip_data);
        if (m_it == m_series_info->end())
          return false;

        if (m_zip_data)
          {
            m_input_stream->check_and_reset(true, m_it->second);
          }
        else
          {
            m_input_stream->check_and_reset(false);
          }

        name = m_it->first;
        read_data(data);
        ++m_it;
        return true;
      }

    /// return size of every serie when it is read in-memory for merge operation (order is defined by m_series_info)
    std::shared_ptr<std::vector<uint64_t>>
    estimate_in_memory_merge_size();

    uint64_t
    get_num_of_series(const std::string& ids, boost::regex *re) const;

    uint64_t
    get_total_data_size() const
    {
      return m_total_data_size;
    }

  private:

    std::stringstream * m_string_stream;
    InputStream * m_input_stream;
    std::map<std::string, uint64_t>::const_iterator m_it;
    std::shared_ptr<std::vector<uint64_t>> m_in_memory_merge_size;
    bool m_it_inited;
    uint64_t m_total_data_size;


#define EMPTY_CONDITION                                                        \


#define RAW_DATA_POINT_CONDITION                                               \
  if (ts_created <= until && ts_last_updated >= since)

#define DS_DATA_POINT_CONDITION                                                \
  if (ts <= until && ts >= since)


#define READ_SIMPLE_SERIES_RAW_DATA( CONDITION )                               \
    uint32_t data_len;                                                         \
    uint64_t ts_created, ts_last_updated;                                      \
    T val;                                                                     \
                                                                               \
    try                                                                        \
      {                                                                        \
        m_input_stream->read(data_len);                                        \
        data.reserve(data_len);                                                \
                                                                               \
        for (uint32_t i = 0; i < data_len; ++i)                                \
          {                                                                    \
            read_timestamps(ts_created, ts_last_updated);                      \
            m_input_stream->read(val);                                         \
                                                                               \
            CONDITION                                                          \
              {                                                                \
                data.emplace_back(ts_created, ts_last_updated, val);           \
              }                                                                \
          }                                                                    \
                                                                               \
        m_total_data_size += data_len;                                         \
                                                                               \
        return (1 + data_len * 3);                                             \
      }                                                                        \
    catch (const std::exception& ex)                                           \
      {                                                                        \
        throw std::runtime_error(make_exception_report("file", ex).c_str());   \
      }

    template<class T>
      uint64_t
      read_data(std::vector<SeriesData<T> >& data)
      {
        READ_SIMPLE_SERIES_RAW_DATA( EMPTY_CONDITION )
      }

    template<class T>
      uint64_t
      read_data(uint64_t since, uint64_t until, std::vector<SeriesData<T> >& data)
      {
        READ_SIMPLE_SERIES_RAW_DATA( RAW_DATA_POINT_CONDITION )
      }

    uint64_t
    read_data(std::vector<SeriesData<std::string> >& data);

    uint64_t
    read_data(uint64_t since, uint64_t until, std::vector<SeriesData<std::string> >& data);


#define READ_VECTOR_SERIES_RAW_DATA( CONDITION )                              \
    uint32_t data_len, vec_len;                                               \
    uint64_t ts_created, ts_last_updated;                                     \
    T val;                                                                    \
                                                                              \
    try                                                                       \
      {                                                                       \
        m_input_stream->read(data_len);                                       \
        data.reserve(data_len);                                               \
                                                                              \
        uint64_t ret = 1 + data_len * 3;                                      \
                                                                              \
        for (uint32_t i = 0; i < data_len; ++i)                               \
          {                                                                   \
            std::vector<T> vec_val;                                           \
            read_timestamps(ts_created, ts_last_updated);                     \
            m_input_stream->read(vec_len);                                    \
                                                                              \
            ret += vec_len;                                                   \
                                                                              \
            if (vec_len > 0)                                                  \
              {                                                               \
                vec_val.reserve(vec_len);                                     \
                for (unsigned int j = 0; j < vec_len; ++j)                    \
                  {                                                           \
                    m_input_stream->read(val);                                \
                    vec_val.emplace_back(val);                                \
                  }                                                           \
              }                                                               \
                                                                              \
            CONDITION                                                         \
              {                                                               \
                data.emplace_back(ts_created, ts_last_updated, vec_val);      \
              }                                                               \
                                                                              \
            m_input_stream->check_and_reset_object(m_zip_data);               \
          }                                                                   \
                                                                              \
        m_total_data_size += data_len;                                        \
                                                                              \
        return ret;                                                           \
      }                                                                       \
    catch (const std::exception& ex)                                          \
      {                                                                       \
        throw std::runtime_error(make_exception_report("file", ex).c_str());  \
      }


    template<class T>
      uint64_t
      read_data(std::vector<SeriesVectorData<T> >& data)
      {
        READ_VECTOR_SERIES_RAW_DATA( EMPTY_CONDITION )
      }

    template<class T>
      uint64_t
      read_data(uint64_t since, uint64_t until, std::vector<SeriesVectorData<T> >& data)
      {
        READ_VECTOR_SERIES_RAW_DATA( RAW_DATA_POINT_CONDITION )
      }


    template<class T>
      uint64_t
      read_data(std::vector<DownsampledSeriesData<T> >& data)
      {
        uint32_t data_len;
        uint32_t ts;
        DownsampleValue<T> val;

        try
          {
            m_input_stream->read(data_len);
            data.reserve(data_len);

            for (uint32_t i = 0; i < data_len; ++i)
              {
                m_input_stream->read(ts);
                m_input_stream->read(val.m_min_value);
                m_input_stream->read(val.m_value);
                m_input_stream->read(val.m_max_value);
                data.emplace_back(ts, val);
              }

            m_total_data_size += data_len;

            return (4 * data_len + 1);
          }
        catch (const std::exception& ex)
          {
            throw std::runtime_error(make_exception_report("file", ex).c_str());
          }
      }

    uint64_t
    read_data(std::vector<DownsampledSeriesData<std::string> >& data);


    template<class T>
      uint64_t
      read_data(std::vector<DownsampledSeriesVectorData<T> >& data)
      {
      uint32_t data_len, vec_len;
      uint32_t ts;
      DownsampleValue<T> val;

      try
        {
          m_input_stream->read(data_len);
          data.reserve(data_len);

          uint64_t ret = 1 + data_len * 2;

          for (uint32_t i = 0; i < data_len; ++i)
            {
              std::vector<DownsampleValue<T>> vec_val;

              m_input_stream->read(ts);
              m_input_stream->read(vec_len);

              if (vec_len > 0)
                {
                  vec_val.reserve(vec_len);

                  ret += vec_len * 3;

                  for (unsigned int j = 0; j < vec_len; ++j)
                    {
                      m_input_stream->read(val.m_min_value);
                      m_input_stream->read(val.m_value);
                      m_input_stream->read(val.m_max_value);
                      vec_val.emplace_back(val);
                    }
                }

              data.emplace_back(ts, vec_val);

              m_input_stream->check_and_reset_object(m_zip_data);
            }

          m_total_data_size += data_len;

          return ret;
        }
      catch (const std::exception& ex)
        {
          throw std::runtime_error(make_exception_report("file", ex).c_str());
        }
      }


    inline void
    reset(bool use_zip, uint64_t pos)
    {
      m_input_stream->reset(use_zip, pos);
    }

    inline void
    init_iterator(bool use_zip)
    {
      m_it = m_series_info->begin();
      m_it_inited = true;
      reset(use_zip, m_it->second);
    }

    std::string
    make_exception_report(const char * what, const std::exception& ex) const;

  };

  // first try to order set of files by base time-stamp and if equal, order by filename

  struct filename_compare
  {
    bool
    operator()(const std::unique_ptr<pbeast::InputDataFile>& lhs, const std::unique_ptr<pbeast::InputDataFile>& rhs) const
    {
      if (lhs->get_base_timestamp() != rhs->get_base_timestamp())
        return lhs->get_base_timestamp() < rhs->get_base_timestamp();

      return pbeast::extract_file_name(lhs->get_file_name()) < pbeast::extract_file_name(rhs->get_file_name());
    }
  };

  typedef std::set<std::unique_ptr<pbeast::InputDataFile>, filename_compare> FileSet;

  struct FileOpenGuard
  {
    FileOpenGuard(pbeast::InputDataFile * file) : m_file(file)
    {
      m_file->open();
    }

    ~FileOpenGuard()
    {
      m_file->close();
    }

    pbeast::InputDataFile * m_file;
  };
}

#endif
