#ifndef PBEAST_OUTPUT_DATA_FILE_H
#define PBEAST_OUTPUT_DATA_FILE_H

#include <stdint.h>
#include <sstream>
#include <string>
#include <vector>

#include <filesystem>

#include "pbeast/output-stream.h"
#include "pbeast/data-file.h"
#include "pbeast/series-data.h"
#include "pbeast/series-info.h"

namespace pbeast {


    /**
     *   Class OutputDataFile is used to store series of data to a file.
     *
     *   The constructor does not create a file.
     *   Real data can be written by the following methods (the order is important):
     *   * write_header()
     *   * multiple write_data<T> (note, names of series must be ordered alphabetically)
     *   * write_footer()
     *   * close() - flush file and check status use close() method.
     */

  class OutputDataFile : public DataFile {

    public:


        /**
         *   Create output data file object.
         *
         *   \param file_name name of the output file
         *   \param partition partition name
         *   \param class_name class name
         *   \param class_path path to class inside repository (for normal classes is equal to class name; is different for nested classes)
         *   \param attribute attribute name
         *   \param base_timestamp base timestamp defining oldest data stored in the file (in microseconds since Epoch)
         *   \param precision of data timestamps in microseconds (1 - 1 microsecond, 1000 - 1 millisecond, 100000 - 1 second)
         *   \param data_type type of data as defined by pbeast::DataType
         *   \param is_array if true, the data are array
         *   \param number_of_series number of series to be stored by the write_data<T> method
         *   \param zip_catalog compress catalog (footer), i.e. information about names of objects and their indices
         *   \param zip_data compress data
         *   \param catalogize_strings if data type is "string", they can be saved efficiently in case of repetitive data belonging to the same object
         *   \param downsample_interval if it is 0, then data are in raw format; otherwise they are downsampled with given time interval
         */

      OutputDataFile(
        const std::string& file_name,
        const std::string& partition,
        const std::string& class_name,
        const std::filesystem::path& class_path,
        const std::string& attribute,
        uint64_t base_timestamp,
        uint64_t precision,
        uint32_t data_type,
        bool is_array,
        uint32_t number_of_series,
        bool zip_catalog,
        bool zip_data,
        bool catalogize_strings = false,
        uint32_t downsample_interval = 0
      ) :
        DataFile(file_name, partition, class_name, class_path, attribute, base_timestamp, precision, data_type, is_array, number_of_series, zip_catalog, zip_data, catalogize_strings, downsample_interval),
        m_string_stream(nullptr), m_output_stream(nullptr)
      {
        m_series_info.reserve(number_of_series);
      }

      OutputDataFile(
        std::stringstream * stream,
        const std::string& partition,
        const std::string& class_name,
        const std::filesystem::path& class_path,
        const std::string& attribute,
        uint64_t base_timestamp,
        uint64_t precision,
        uint32_t data_type,
        bool is_array,
        uint32_t number_of_series,
        bool zip_catalog,
        bool zip_data,
        bool catalogize_strings = false,
        uint32_t downsample_interval = 0
      ) :
        DataFile("", partition, class_name, class_path, attribute, base_timestamp, precision, data_type, is_array, number_of_series, zip_catalog, zip_data, catalogize_strings, downsample_interval),
        m_string_stream(stream), m_output_stream(nullptr)
      {
        m_series_info.reserve(number_of_series);
      }


        /**
         *  Free resources used by output file object.
         */

      ~OutputDataFile();

        /**
         *   Create name for output data file.
         *
         *   \param repository_path path to repository
         *   \param partition_name encoded partition name
         *   \param class_name encoded class name
         *   \param attribut_name encoded attribute name
         *   \param secs_min - minimal timestamp
         *   \param secs_max - maximum timestamp
         *   \param extension - file suffix
         *   \param unique - when true, add a number before suffix, if file exists
         *   \param downsample_interval - if non-zero, add after extension to distinguish downsampled data
         *   \return the name of the output file
         */

      static std::filesystem::path make_file_name(
        const std::filesystem::path& repository_path,
        const std::string& partition_name,
        const std::filesystem::path& class_path,
        const std::string& attribute_name,
        uint32_t secs_min,
        uint32_t secs_max,
        const char * extension = ".pb",
        bool unique = false,
        uint32_t downsample_interval = 0
      );


        /**
         *   Write header data (partition, class and attribute names; creation timestamp; data type; number of series)
         *   \exception std::runtime_error in case of write problems
         */

      void write_header();


        /**
         *   The file has to contain raw data
         *   \throw std::runtime_error exception, if m_downsample_interval is non-zero
         */

      void assert_raw_type();


        /**
         *   The file has to contain downsampled data
         *   \throw std::runtime_error exception, if m_downsample_interval is zero
         */

      void assert_downsample_type();


      inline void
      write_timestamps(uint64_t ts_created, const uint64_t ts_last_updated)
      {
        const uint64_t created = (ts_created - m_base_timestamp) / m_precision;
        const uint64_t interval = (ts_last_updated - ts_created) / m_precision;

        m_output_stream->write(created);
        m_output_stream->write(interval);
      }



        /**
         *   Write single value series of raw data.
         *
         *   The template parameter is one of supported type: bool, int8_t, int16_t, uint16_t,
         *   int32_t, uint32_t, int64_t, uint64_t, float, double or std::string.
         *
         *   \param name name of data
         *   \param data series of raw data
         *   \exception std::runtime_error in case of write problems
         */

      template<typename T>
        void
        write_data(const std::string& name, const std::vector<SeriesData<T> >& data)
        {
          write_data_imp(name, data);
        }

      template<typename T>
        void
        write_data_imp(const std::string& name, const std::vector<SeriesData<T> >& data)
        {
          assert_raw_type();

          if (data.size())
            {
              uint64_t pos = m_output_stream->check_and_reset(m_zip_data);
              m_series_info.push_back(new SeriesInfo(name, pos));
              m_output_stream->write(data.size());
              for (unsigned int i = 0; i < data.size(); ++i)
                {
                  const SeriesData<T>& d(data[i]);
                  const T& val(d.get_value());
                  write_timestamps(d.get_ts_created(), d.get_ts_last_updated());
                  m_output_stream->write(val);
                }
            }
        }


        /**
         *   Write multi value series of raw data.
         *
         *   This method is similar to above except is stores vectors of data instead of single data.
         */

      template<class T> void write_data(const std::string& name, const std::vector<SeriesVectorData<T> >& data) {
        assert_raw_type();

        if(data.size()) {
          uint64_t pos = m_output_stream->check_and_reset(m_zip_data);
          m_series_info.push_back(new SeriesInfo(name, pos));
          m_output_stream->write(data.size());
          for(unsigned int i = 0; i < data.size(); ++i) {
            const SeriesVectorData<T>& d(data[i]);
            const std::vector<T>& vec(d.get_value());
            write_timestamps(d.get_ts_created(), d.get_ts_last_updated());
            const uint32_t len(vec.size());
            m_output_stream->write(len);
            for(unsigned int j = 0; j < len; ++j) {
              m_output_stream->write(vec[j]);
            }
          }
        }
      }

      /**
       *   Write single value series of downsampled data.
       *
       *   The template parameter is one of supported type: bool, int8_t, int16_t, uint16_t,
       *   int32_t, uint32_t, int64_t, uint64_t, float, double or std::string.
       *
       *   \param name name of data
       *   \param data series of downsampled data
       *   \exception std::runtime_error in case of write problems
       */

    template<typename T>
      void
      write_data(const std::string& name, const std::vector<DownsampledSeriesData<T> >& data)
      {
        write_data_imp(name, data);
      }

    template<typename T>
      void
      write_data_imp(const std::string& name, const std::vector<DownsampledSeriesData<T> >& data)
      {
        assert_downsample_type();

        if (data.size())
          {
            uint64_t pos = m_output_stream->check_and_reset(m_zip_data);
            m_series_info.push_back(new SeriesInfo(name, pos));
            m_output_stream->write(data.size());
            for (unsigned int i = 0; i < data.size(); ++i)
              {
                const DownsampledSeriesData<T>& d(data[i]);
                m_output_stream->write(d.get_ts());
                m_output_stream->write(d.data().get_min_value());
                m_output_stream->write(d.data().get_value());
                m_output_stream->write(d.data().get_max_value());
              }
          }
      }

      /**
       *   Write multi value series of downsampled data.
       *
       *   This method is similar to above except is stores vectors of data instead of single data.
       */

    template<class T>
      void
      write_data(const std::string& name, const std::vector<DownsampledSeriesVectorData<T> >& data)
      {
        assert_downsample_type();

        if (data.size())
          {
            uint64_t pos = m_output_stream->check_and_reset(m_zip_data);
            m_series_info.push_back(new SeriesInfo(name, pos));
            m_output_stream->write(data.size());
            for (unsigned int i = 0; i < data.size(); ++i)
              {
                const DownsampledSeriesVectorData<T>& d(data[i]);
                m_output_stream->write(d.get_ts());
                const uint32_t len(d.data().size());
                m_output_stream->write(len);
                for (unsigned int j = 0; j < len; ++j)
                  {
                    m_output_stream->write(d.data(j).get_min_value());
                    m_output_stream->write(d.data(j).get_value());
                    m_output_stream->write(d.data(j).get_max_value());
                  }
              }
          }
      }

        /**
         *   Write footer data (names and positions of series)
         *   \exception std::runtime_error in case of write problems
         */

      void write_footer();


        /**
         *   Flush stream buffer and close the file.
         *   \exception std::runtime_error in case of write problems
         *   \return size of output file
         */

      int64_t close();

      const std::vector<SeriesInfo *>&
      get_series_info() const
      {
        return m_series_info;
      }


    private:

      std::stringstream * m_string_stream;
      OutputStream * m_output_stream;
      std::vector<SeriesInfo *> m_series_info;

  };

  template<>
    void
    OutputDataFile::write_data<std::string>(const std::string& name, const std::vector<SeriesData<std::string> >& data);

  template<>
    void
    OutputDataFile::write_data<std::string>(const std::string& name, const std::vector<DownsampledSeriesData<std::string> >& data);

}

#endif
