#include <iostream>
#include <signal.h>

#include <cmdl/cmdargs.h>
#include <ipc/core.h>

#include "rdb/rdb_info.h"
#include "rdb/errors.h"

////////////////////////////////////////////////////////////////////////////////////////////////////

  /** Failed to subscribe */

ERS_DECLARE_ISSUE(
  rdb_subscriber,
  SubscriptionFailed,
  "Failed to subscribe on changes",
)

ERS_DECLARE_ISSUE(
  rdb_subscriber,
  UnsubscriptionFailed,
  "Failed to unsubscribe ",
)

////////////////////////////////////////////////////////////////////////////////////////////////////


static void callback(RDBCallbackInfo * isc)
{
  std::cout << isc->get_changes()->length() << " classe(s) were changed\n";

  for(unsigned int idx = 0; idx < isc->get_changes()->length(); idx++) {
    const rdb::RDBClassChanges& class_changes = (*isc->get_changes())[idx];
    std::cout << '(' << (idx + 1) << ") class " << class_changes.name << std::endl;

    std::cout << "   - there are " << class_changes.modified.length() << " modified objects\n";
    if(class_changes.modified.length()) {
      for(unsigned int i = 0; i < class_changes.modified.length(); i++) {
        std::cout << "      * " << class_changes.modified[i] << std::endl;
      }
    }

    std::cout << "   - there are " << class_changes.created.length() << " created objects\n";
    if(class_changes.created.length()) {
      for(unsigned int i = 0; i < class_changes.created.length(); i++) {
        std::cout << "      * " << class_changes.created[i] << std::endl;
      }
    }

    std::cout << "   - there are " << class_changes.removed.length() << " removed objects\n";
    if(class_changes.removed.length()) {
      for(unsigned int i = 0; i < class_changes.removed.length(); i++) {
        std::cout << "      * " << class_changes.removed[i] << std::endl;
      }
    }
  }

  std::cout << "#############################################" << std::endl;
}

static RDBInfoReceiver * Receiver;

static int sig_received = 0;

extern "C" void signal_handler(int sig)
{
  if( Receiver != 0) {
    sig_received = sig;
    Receiver->post();
  }
}

int main(int argc, char ** argv)
{
  CmdArgStr	partition_name	('p', "partition", "partition-name", "partition to work in");
  CmdArgStr	database_name	('d', "database", "database-name", "name of database server to work with", CmdArg::isREQ);
  CmdArgStrList	class_names     ('c', "class-names", "names", "list of classes for subscription");
  CmdArgBool	look_in_sclasses('s', "look-in-subclasses", "subscribe on changes in subclasses");
  CmdArgStrList	obj_names       ('o', "objects", "names-pairs", "list of space-separated \"class-name\" \"object-id\" pairs for subscription");

  CmdLine cmd(*argv, &partition_name, &database_name, &class_names, &look_in_sclasses, &obj_names, NULL);

  try {
    IPCCore::init(argc, argv);
  }
  catch(ers::Issue & ex) {
    ers::warning(ers::Message(ERS_HERE, ex));
  }

  CmdArgvIter  arg_iter(--argc, ++argv);


  cmd.description("This program is an example of RDB subscriber.\n"
  		  "It checks callback functionality.");

       // Parse arguments
  cmd.parse(arg_iter);

  IPCPartition p(partition_name);

  Receiver = new RDBInfoReceiver(p, database_name);

  signal(SIGINT,signal_handler);
  signal(SIGTERM,signal_handler);

  std::set<std::string> classes;
  std::map< std::string , std::set<std::string> > objects;

  int i;

  for(i = 0; i < (int)class_names.count(); i++) {
    classes.insert((const char *)class_names[i]);
  }

  for(i = 0; i < (int)(obj_names.count() - 1); ) {
    const char * class_name = (const char *)obj_names[i++];
    const char * object_id = (const char *)obj_names[i++];
    objects[class_name].insert(object_id);
  }

  try {
    Receiver->subscribe(classes, (look_in_sclasses ? true : false), objects, callback);
  }
  catch (daq::rdb::Exception & ex) {
    ers::fatal(rdb_subscriber::SubscriptionFailed(ERS_HERE, ex));
    return 1;
  }

  Receiver->wait();

  if(sig_received) {
    std::cout << "Got signal " << sig_received << ", exiting ..." << std::endl;
  }

  try {
    Receiver->unsubscribe();
  }
  catch (daq::rdb::Exception & ex) {
    ers::error(rdb_subscriber::UnsubscriptionFailed(ERS_HERE, ex));
  }

  return 0;
}
