#ifndef RDB_ERRORS_H
#define RDB_ERRORS_H

#include <stdint.h>
#include <ers/ers.h>

namespace daq {

    /**
     *  \class rdb::Exception
     *  This is a base class for all daq::rdb exceptions.
     */

  ERS_DECLARE_ISSUE( rdb, Exception, , )


    /**
     *  \class rdb::CorbaException
     *  This is a wrapper for CORBA exceptions.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    CannotProceedException,
    rdb::Exception,
    text,
    ,
    ((const char*)text)
  )


    /**
     *  \class rdb::NotFoundSubscriptionException
     *  Cannot unsibscribe client: id is not found.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    SessionNotFound,
    rdb::Exception,
    "session " << id << " was not found",
    ,
    ((int64_t)id)
  )


    /**
     *  \class rdb::NotFoundSubscriptionException
     *  Cannot unsibscribe client: id is not found.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    NotFoundSubscriptionException,
    rdb::Exception,
    "subscription " << id << " was not found",
    ,
    ((long)id)
  )


    /**
     *  \class rdb::NotFoundServerSubscriptionException
     *  Cannot unsibscribe server: name is not found.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    NotFoundServerSubscriptionException,
    rdb::Exception,
    "subscription for server \'" << name << "\' was not found",
    ,
    ((const char *)name)
  )


    /**
     *  \class rdb::StandardException
     *  This is a wrapper for standard C++ exceptions like memory out for C++ new operator.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    StandardException,
    rdb::Exception,
    "standard C++ exception \"" << text << "\" has been raised",
    ,
    ((const char*)text)
  )


    /**
     *  \class rdb::LookupFailed
     *  Cannot find RDB server exception.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    LookupFailed,
    rdb::Exception,
    "lookup failed for RDB server \'" << server_name << '@' << (partition_name ? partition_name : "initial") << '\'',
    ,
    ((const char*)server_name)
    ((const char*)partition_name)
  )


    /**
     *  \class rdb::NotificationFailed
     *  Cannot notify RDB server subscriber.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    NotificationFailed,
    rdb::Exception,
    "failed to notify subscriber " << id << ": \' application \'" << name << "\' subscribed on \'" << creation_ts
      << "\' started by \'" << user << "\' on \'" << host << "\' with pid " << pid << " is not responding",
    ,
    ((long)id)
    ((const char *)name)
    ((const char *)creation_ts)
    ((const char *)user)
    ((const char *)host)
    ((long)pid)
  )

    /**
     *  \class rdb::SessionNotificationFailed
     *  Cannot notify RDB writer session subscriber.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    SessionNotificationFailed,
    rdb::Exception,
    "failed to notify subscriber of session " << id << ": \' application \'" << name << "\' subscribed on \'" << creation_ts
      << "\' started by \'" << user << "\' on \'" << host << "\' with pid " << pid << " is not responding",
    ,
    ((int64_t)id)
    ((const char *)name)
    ((const char *)creation_ts)
    ((const char *)user)
    ((const char *)host)
    ((long)pid)
  )


    /**
     *  \class rdb::ServerNotificationFailed
     *  Cannot notify RDB server.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    ServerNotificationFailed,
    rdb::Exception,
    "failed to notify server \'" << name << "\': " << reason,
    ,
    ((const char *)name)
    ((const char *)reason)
  )


    /**
     *  \class rdb::SessionPingFailed
     *  Cannot ping session.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    SessionPingFailed,
    rdb::Exception,
    "failed to ping session " << session_id << " (application: \'" << app_name << "\' run by \'" << user_id << "\' on \'" << host << "\' with pid " << pid << "), " << action,
    ,
    ((int64_t)session_id)
    ((std::string)app_name)
    ((std::string)user_id)
    ((std::string)host)
    ((long)pid)
    ((const char *)action)
  )

    /**
     *  \class rdb::AlreadyRunning
     *  Failed to run rdb server.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    AlreadyRunning,
    rdb::Exception,
    "rdb server \'" << server_name << '@' << (partition_name ? partition_name : "initial") << "\' is already running",
    ,
    ((const char*)server_name)
    ((const char*)partition_name)
  )


    /**
     *  \class rdb::CannotPublish
     *  Failed to publish rdb server.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    CannotPublish,
    rdb::Exception,
    "cannot publish rdb server object \'" << server_name << '@' << (partition_name ? partition_name : "initial") << '\'',
    ,
    ((const char*)server_name)
    ((const char*)partition_name)
  )


    /**
     *  \class rdb::CannotUnpublish
     *  Failed to publish rdb server.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    CannotUnpublish,
    rdb::Exception,
    "cannot unpublish rdb server object \'" << server_name << '@' << (partition_name ? partition_name : "initial") << '\'',
    ,
    ((const char*)server_name)
    ((const char*)partition_name)
  )


    /**
     *  \class rdb::CannotUnsubscribe
     *  Failed to unsubscribe client.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    CannotUnsubscribe,
    rdb::Exception,
    "cannot unsubscribe \'" << this_name << '@' << (partition_name ? partition_name : "initial") << "\' (id = " << subscrition_id << ") from server \'" << server_name << '\'',
    ,
    ((const char*)this_name)
    ((const char*)server_name)
    ((const char*)partition_name)
    ((long)subscrition_id)
  )


    /**
     *  \class rdb::CannotUnsubscribeServer
     *  Failed to unsubscribe server.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    CannotUnsubscribeServer,
    rdb::Exception,
    "cannot remove \'" << this_name << "\' server subscription from master server \'" << server_name << "\' in partition " << (partition_name ? partition_name : "initial"),
    ,
    ((const char*)this_name)
    ((const char*)server_name)
    ((const char*)partition_name)
  )


    /**
     *  \class rdb::CannotUnpublish
     *  Failed to publish rdb server.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    ServerError,
    rdb::Exception,
    "user request on rdb server \'" << server_name << '@' << (partition_name ? partition_name : "initial") << "\' has failed: " << reason,
    ,
    ((const char*)server_name)
    ((const char*)partition_name)
    ((const char*)reason)
  )


    /**
     *  \class rdb::ReloadCommandFailed
     *  Failed to run reload rdb server.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    ReloadCommandFailed,
    rdb::Exception,
    "cannot reload rdb server \'" << server_name << '@' << (partition_name ? partition_name : "initial") << "\' " << reason,
    ,
    ((const char*)server_name)
    ((const char*)partition_name)
    ((const char*)reason)
  )


    /**
     *  \class rdb::ReloadCommandFailed
     *  Failed to run reload rdb server.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    PushCommandFailed,
    rdb::Exception,
    "cannot push data on rdb server \'" << server_name << '@' << partition_name << '\'',
    ,
    ((const char*)server_name)
    ((const char*)partition_name)
  )


    /**
     *  \class rdb::ReloadCommandFailed
     *  Failed to run reload rdb server.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    GetAllClassesCommandFailed,
    rdb::Exception,
    "cannot get description of classes from rdb server \'" << server_name << '@' << partition_name << '\'',
    ,
    ((const char*)server_name)
    ((const char*)partition_name)
  )


    /**
     *  \class rdb::FailedCreateData
     *  Failed to load oks database file.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    FailedCreateData,
    rdb::Exception,
    "internal error during creation of OKS data:\n" << why,
    ,
    ((const char*)why)
  )

    /**
     *  \class rdb::FailedLoadDatabaseFile
     *  Failed to load oks database file.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    FailedLoadDatabaseFile,
    rdb::Exception,
    "failed to load file:\n" << why,
    ,
    ((const char*)why)
  )


    /**
     *  \class rdb::FailedRemoveFile
     *  Failed to remove temporal oks database file.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    FailedRemoveFile,
    rdb::Exception,
    "failed to remove temporal file: " << why,
    ,
    ((const char*)why)
  )


    /**
     *  \class rdb::FailedReadDB
     *  Failed to read database from master RDB server.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    FailedReadDB,
    rdb::Exception,
    "failed to read database from \'" << server_name << '@' << (partition_name ? partition_name : "initial") << "\' " << reason,
    ,
    ((const char*)server_name)
    ((const char*)partition_name)
    ((const char*)reason)
  )


    /**
     *  \class rdb::InfoReceiverNoSubscriptionException
     *  Call unsubscribe() method in case when there was no subscription
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    InfoReceiverNoSubscriptionException,
    rdb::Exception,
    "unsubscribe failed: " << why,
    ,
    ((const char*)why)
  )


    /**
     *  \class rdb::BadBackupFile
     *  Failed to load RDB server backup file.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    BadBackupFile,
    rdb::Exception,
    "failed lo load backup file \'" << file_name << "\': " << reason,
    ,
    ((const char*)file_name)
    ((const char*)reason)
  )


    /**
     *  \class rdb::FailedCreateBackupFile
     *  Failed to create RDB server backup file.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    FailedCreateBackupFile,
    rdb::Exception,
    "failed lo create backup file \'" << file_name << '\'',
    ,
    ((const char*)file_name)
  )


    /**
     *  \class rdb::CannotCreateDirectory
     *  Failed to create directory for RDB server backup file.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    CannotCreateDirectory,
    rdb::Exception,
    "Backup directory '" << directory << "' can not be created",
    ,
    ((std::string)directory)
  )


    /**
     *  \class rdb::BadBackupFile
     *  Failed to apply RDB server backup data.
     */

  ERS_DECLARE_ISSUE_BASE(
    rdb,
    BadBackupData,
    rdb::Exception,
    "restore from backup file \'" << file_name << "\' failed: " << reason,
    ,
    ((const char*)file_name)
    ((const char*)reason)
  )


}  // close namespace daq

#endif
