#include <stdlib.h>
#include <iostream>

#include <cmdl/cmdargs.h>
#include <ipc/core.h>
#include <ipc/partition.h>

#include "rdb/rdb.hh"
#include "rdb/rdb_info.h"
#include "rdb/errors.h"

  /// Global variables

static rdb::writer_var cursor;        /// the 'cursor' is used for communication with server
static rdb::RDBSessionId session;     /// the 'session' is identifier for all communications with writer server


  /// Create text to report 'session not found' error

static const char * get_session_id_str()
{
  static std::string buf;

  if(buf.empty()) {
    std::ostringstream s;
    s << "session " << session << " is not found";
    buf = s.str();
  }

  return buf.c_str();
}


  /// Declare possible issues

ERS_DECLARE_ISSUE(
  rdb_tutorial,
  RequestFailed,
  "method \"" << method << "\" raised exception:\n" << reason,
  ((const char*)method)
  ((const char*)reason)
)

ERS_DECLARE_ISSUE(
  rdb_tutorial,
  CorbaRequestFailed,
  "a method raised CORBA exception " << reason,
  ((const char*)reason)
)


static void
set_attribute(const char * aname, const rdb::RDBObject& o, const rdb::DataUnion& d)
{
  try {
    cursor->set_attribute_value(session, o, aname, d);
  }
  catch ( rdb::SessionNotFound & ex ) {
    throw (rdb_tutorial::RequestFailed(ERS_HERE, "set_attribute_value", get_session_id_str()));
  }
  catch ( rdb::NotFound & ex ) {
    throw (rdb_tutorial::RequestFailed(ERS_HERE, "set_attribute_value", ex.text));
  }
  catch ( rdb::CannotProceed & ex ) {
    throw (rdb_tutorial::RequestFailed(ERS_HERE, "set_attribute_value", ex.text));
  }
}

static void
create_object(rdb::RDBObject& o, const char * where)
{
  try {
    cursor->create_object(session, o.classid, o.name, where);
  }
  catch ( rdb::SessionNotFound & ex ) {
    throw (rdb_tutorial::RequestFailed(ERS_HERE, "create_object", get_session_id_str()));
  }
  catch ( rdb::CannotProceed & ex ) {
    throw (rdb_tutorial::RequestFailed(ERS_HERE, "create_object", ex.text));
  }
}

static void
rename_object(rdb::RDBObject& o, const char * new_name)
{
  try {
    cursor->rename_object(session, o, new_name);
  }
  catch ( rdb::SessionNotFound & ex ) {
    throw (rdb_tutorial::RequestFailed(ERS_HERE, "rename_object", get_session_id_str()));
  }
  catch ( rdb::CannotProceed & ex ) {
    throw (rdb_tutorial::RequestFailed(ERS_HERE, "rename_object", ex.text));
  }
}

static rdb::RDBObject
create_department(const char * where, const char * id, const char * name)
{
    // make rdb object reference

  rdb::RDBObject o;
  o.classid = "Department";
  o.name    = id;


    // create object

  create_object(o, where);


    // set attribute 'Name'

  rdb::DataUnion d;
  d.du_string(name);
  set_attribute("Name", o, d);

  return o;
}

static rdb::RDBObject
create_employee(const char * where, const char * id, const char * name,
                const char * date_of_birth, const char * family_situation,
		unsigned long  salary)
{

    // make rdb object reference

  rdb::RDBObject o;
  o.classid = "Employee";
  o.name    = id;


    // create object

  create_object(o, where);


    // describe data (single attribute value)

  rdb::DataUnion d;


    // set attribute 'Name'

  d.du_string(name);
  set_attribute("Name", o, d);


    // set attribute 'Birthday'

  d.du_string(date_of_birth);
  set_attribute("Birthday", o, d);


    // set attribute 'Family Situation'

  d.du_string(family_situation);
  set_attribute("Family Situation", o, d);


    // set attribute 'Salary'

  d.du_u32_int(salary);
  set_attribute("Salary", o, d);

  return o;
}

void
hire(rdb::RDBObject & department, rdb::RDBObject & employee)
{
    // set staff

  {

      // get existing staff

    rdb::RDBObjectList_var olist;

    try {
      cursor->get_objects_of_relationship(session, department, "Staff", olist);
    }
    catch ( rdb::SessionNotFound & ex ) {
      throw (rdb_tutorial::RequestFailed(ERS_HERE, "get_objects_of_relationship", get_session_id_str()));
    }
    catch ( rdb::NotFound & ex ) {
      throw (rdb_tutorial::RequestFailed(ERS_HERE, "get_objects_of_relationship", ex.text));
    }
    catch ( rdb::CannotProceed & ex ) {
      throw (rdb_tutorial::RequestFailed(ERS_HERE, "get_objects_of_relationship", ex.text));
    }


      // create new object in the staff list

    unsigned long idx = olist->length();
    olist->length(idx + 1);
    olist[idx].classid = CORBA::string_dup(employee.classid);
    olist[idx].name    = CORBA::string_dup(employee.name);


      // add employee to staff

    try {
      cursor->set_objects_of_relationship(session, department, "Staff", olist, false);
    }
    catch ( rdb::SessionNotFound & ex ) {
      throw (rdb_tutorial::RequestFailed(ERS_HERE, "set_objects_of_relationship", get_session_id_str()));
    }
    catch ( rdb::NotFound & ex ) {
      throw (rdb_tutorial::RequestFailed(ERS_HERE, "set_objects_of_relationship", ex.text));
    }
    catch ( rdb::CannotProceed & ex ) {
      throw (rdb_tutorial::RequestFailed(ERS_HERE, "set_objects_of_relationship", ex.text));
    }
  }
  
  
    // set employee's department

  {
    rdb::RDBObjectList olist;
    olist.length(1);
    olist[0].classid = CORBA::string_dup(department.classid);
    olist[0].name    = CORBA::string_dup(department.name);


      // set place of work

    try {
      cursor->set_objects_of_relationship(session, employee, "Works at", olist, false);
    }
    catch ( rdb::SessionNotFound & ex ) {
      throw (rdb_tutorial::RequestFailed(ERS_HERE, "set_objects_of_relationship", get_session_id_str()));
    }
    catch ( rdb::NotFound & ex ) {
      throw (rdb_tutorial::RequestFailed(ERS_HERE, "set_objects_of_relationship", ex.text));
    }
    catch ( rdb::CannotProceed & ex ) {
      throw (rdb_tutorial::RequestFailed(ERS_HERE, "set_objects_of_relationship", ex.text));
    }
  }
}

int main(int argc, char ** argv)
{
  CmdArgStr partition_name ('p', "partition",   "partition-name", "partition to work in");
  CmdArgStr database_name  ('d', "database",    "database-name",  "name of database server to work with", CmdArg::isREQ);
  CmdArgStr schema_name    ('S', "schema-file", "file-name",  "name of schema file on rdb-server", CmdArg::isREQ);
  CmdArgStr data_name      ('D', "data-file",   "file-name",  "name of data file on rdb-server", CmdArg::isREQ);

  CmdLine cmd(*argv, &partition_name, &database_name, &schema_name, &data_name, NULL);

  try {
    IPCCore::init(argc, argv);
  }
  catch(ers::Issue & ex) {
    ers::warning(ers::Message(ERS_HERE, ex));
  }

  CmdArgvIter  arg_iter(--argc, ++argv);

  cmd.description("This program is an example using RDB server to modify database. "
  		  "It uses schema produced by the oks_tutorial application.");

  cmd.parse(arg_iter);


    // convert command-line-parameter to the datafile name

  const char * df_name = static_cast<const char *>(data_name);


  IPCPartition p(partition_name);

  try {
    cursor = p.lookup<rdb::writer>(static_cast<const char *>(database_name));
  }
  catch( ers::Issue & ex) {
    ers::fatal(daq::rdb::LookupFailed(ERS_HERE, static_cast<const char *>(database_name), static_cast<const char *>(partition_name), ex.what()));
    return 1;
  }

  rdb::ProcessInfo proc_info;
  rdb::ProcessInfoHelper::set(proc_info);

  RDBWriterPingCB * ping_cb = new RDBWriterPingCB();


    // catch any CORBA communication exception

  try {

      // open new session

    try {
      session = cursor->open_session( proc_info, ping_cb->_this() );
      ping_cb->set_session_id(session);
    }
    catch ( CORBA::SystemException & ex ) {
      ers::fatal(rdb_tutorial::CorbaRequestFailed(ERS_HERE, ex._name()));
      return 1;
    }


      // create new database file

    try {
      cursor->create_database(session, df_name);
    }
    catch ( rdb::SessionNotFound & ex ) {
      throw (rdb_tutorial::RequestFailed(ERS_HERE, "create_database", get_session_id_str()));
    }
    catch ( rdb::CannotProceed & ex ) {
      throw (rdb_tutorial::RequestFailed(ERS_HERE, "create_database", ex.text));
    }


      // add schema file and load it

    try {
      cursor->database_add_file(session, df_name, static_cast<const char *>(schema_name));
    }
    catch ( rdb::SessionNotFound & ex ) {
      throw (rdb_tutorial::RequestFailed(ERS_HERE, "database_add_file", get_session_id_str()));
    }
    catch ( rdb::CannotProceed & ex ) {
      throw (rdb_tutorial::RequestFailed(ERS_HERE, "database_add_file", ex.text));
    }


      // create new 'Department' objects

    rdb::RDBObject ms_obj = create_department(df_name, "MS", "Microsoft");
    rdb::RDBObject tr_obj = create_department(df_name, "TR", "Transmeta Corp");


      // create new 'Employee' objects

    rdb::RDBObject bg_obj = create_employee(df_name, "BG", "Bill Gates",     "1955/10/28",  "Married", 0xFFFF0000);
    rdb::RDBObject pa_obj = create_employee(df_name, "PA", "Paul Allen",     "1955-Nov-28", "Married", 0xFFFE0000);
    rdb::RDBObject lt_obj = create_employee(df_name, "Lt", "Linus Torvalds", "1970-12-28",  "Married", 0x0000FFFF);


      // set links between departments and employees

    hire(ms_obj, bg_obj);
    hire(ms_obj, pa_obj);
    hire(tr_obj, lt_obj);


      // rename object

    rename_object(lt_obj, "LT");

    //sleep(30);


      // commit changes

    try {
      cursor->commit_changes(session, "", "RDB tutorial");
    }
    catch ( rdb::SessionNotFound & ex ) {
      throw (rdb_tutorial::RequestFailed(ERS_HERE, "commit_changes", get_session_id_str()));
    }
    catch ( rdb::CannotProceed & ex ) {
      throw (rdb_tutorial::RequestFailed(ERS_HERE, "commit_changes", ex.text));
    }

    //sleep(30);

      // close session

    try {
      cursor->close_session( session, proc_info );
    }
    catch(rdb::SessionNotFound& ex) {
      throw (rdb_tutorial::RequestFailed(ERS_HERE, "close_session", get_session_id_str()));
    }

  }


    // report any found problem and exit with bad status

  catch (rdb_tutorial::RequestFailed & ex) {
    ers::fatal(ex);
    return 1;
  }


    // report CORBA exception if any and exit with bad status

  catch (CORBA::SystemException & ex) {
    ers::fatal(rdb_tutorial::CorbaRequestFailed(ERS_HERE, ex._name()));
    return 2;
  }

  return 0;
}
