// select NAME, RUNNUMBER, CREATEDBY,  STARTAT, STOPAT, CONFIGSCHEMA, CONFIGDATA, COMMENTS from RunNumber;

#include <stdlib.h>
#include <iostream>
#include <algorithm>

#include <CoralBase/Attribute.h>
#include <CoralBase/AttributeList.h>
#include <CoralBase/AttributeSpecification.h>
#include <CoralBase/Exception.h>
#include <CoralBase/TimeStamp.h>
#include "CoralKernel/Context.h"

#include "RelationalAccess/ConnectionService.h"
#include "RelationalAccess/ISessionProxy.h"
#include <RelationalAccess/ISchema.h>
#include <RelationalAccess/ITable.h>
#include <RelationalAccess/ITransaction.h>
#include <RelationalAccess/ITableDescription.h>
#include <RelationalAccess/IQuery.h>
#include <RelationalAccess/ICursor.h>
#include <RelationalAccess/SchemaException.h>

#include <ers/ers.h>

#include <rn/rn.h>
#include <oks/tz.h>


  // string to duration

static bool
s2du(const char * in, int32_t& du, const char * name)
{
  try {
    boost::posix_time::time_duration td(0,0,0,0);
    td = boost::posix_time::duration_from_string(in);
    du = td.seconds() + td.minutes() * 60 + td.hours() * 3600;
    ERS_DEBUG(1, "Convert  \'" << in << "\' -> = \'" << du << '\'');
    if(du < 0) {
      std::cerr << "ERROR: negative time intervals are not allowed (input = \'" << in << "\')\n";
      return false;
    }
    return true;
  }
  catch(std::exception& e) {
    std::cerr << "ERROR: cannot parse " << name << " = \'" << in << "\': \"" << e.what() << "\".\n"
                 "       format (H:MM:SS) is expected." << std::endl;
  }

  return false;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct TableRow {

  char m_fill;
  char m_separator;
  std::string m_items[11];

  TableRow(const char fill, const char separator) : m_fill(fill), m_separator(separator) { ; }

  TableRow(const char fill, const char separator, boost::local_time::time_zone_ptr tz_ptr, int version_length,
           const std::string& db_name, int rn, const std::string& start_at, std::int64_t nanosecond,
           int duration, std::int64_t lasted, const std::string& release, const std::string& user, const std::string& host, const std::string& partition,
	   int sv, int dv, const std::string& cv, const std::string& cn, const std::string& comments) : m_fill(fill), m_separator(separator) {

    m_items[0] = db_name;
    { std::ostringstream s; s << rn; m_items[1] = s.str(); }
    { // show date-time in UTC or time zone provided by user
      boost::posix_time::ptime t(boost::posix_time::from_iso_string(start_at));
      if(tz_ptr) {
        m_items[2] = boost::local_time::local_date_time(t, tz_ptr).to_string();
      }
      else {
        m_items[2] = boost::posix_time::to_simple_string(t);
      }

      if(nanosecond >= 0)
        {
          std::ostringstream s;
          s << '.';
          s.fill('0'); s.width(3); s << nanosecond / 1000000;

            if (tz_ptr)
              {
                auto idx = m_items[2].find(' ');
                if (idx != std::string::npos)
                  idx = m_items[2].find(' ', idx + 1);
                if (idx != std::string::npos)
                  m_items[2].insert(idx, s.str());
                else
                  m_items[2].append(s.str());
              }
            else
              m_items[2].append(s.str());
        }
    }
    if(duration >= 0) {
      long h = duration / 3600; duration -= (h * 3600);
      long m = duration / 60; duration -= (m * 60);
      std::ostringstream s; s << h << ':'; s.fill('0'); s.width(2); s << m; s << ':'; s.width(2); s << duration;
      if(lasted > 0) {
        s << '.';
        s.fill('0'); s.width(3); s << (lasted%1000000000) / 1000000;
      }
      m_items[3] = s.str();
    }
    m_items[4] = release;
    m_items[5] = user;
    m_items[6] = host;
    m_items[7] = partition;
    if(sv > 0 && dv > 0) { std::ostringstream s; s << sv << '.' << dv; m_items[8] = s.str(); }
    else if(!cv.empty()) {
      if(version_length != -1) {
        if(cv.find("hash:") == 0)
          m_items[8] = cv.substr(5, (int)cv.length() >= (version_length + 5) ? version_length : std::string::npos);
      }
      else
        m_items[8] = cv;
    }
    m_items[9] = cn;
    m_items[10] = comments;
  }

  TableRow(const char fill, const char separator, const std::string& s1, const std::string& s2,
           const std::string& s3, const std::string& s4, const std::string& s5, const std::string& s6,
	   const std::string& s7, const std::string& s8, const std::string& s9, const std::string& s10,
	   const std::string& s11) : m_fill(fill), m_separator(separator) {
    m_items[0] = s1;
    m_items[1] = s2;
    m_items[2] = s3;
    m_items[3] = s4;
    m_items[4] = s5;
    m_items[5] = s6;
    m_items[6] = s7;
    m_items[7] = s8;
    m_items[8] = s9;
    m_items[9] = s10;
    m_items[10] = s11;
  }

};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool
has_a_pattern(const std::string& s)
{
  return(
    ( s.find("%") != std::string::npos ) ||
    ( s.find("_") != std::string::npos ) ||
    ( s.find("[") != std::string::npos && s.find("]") != std::string::npos) ? true : false
  );
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static void
no_param(const char * s)
{
  std::cerr << "ERROR: no parameter for " << s << " provided\n\n";
  exit(EXIT_FAILURE);
}

int main(int argc, char * argv[])
{
  const long c_min_run_number = 1;
  const long c_max_run_number = 0x7FFFFFFF; //(1<<31) - 1;

  bool list_db = false;
  const char * db_name = 0;
  const char * connect_string(0);
  const char * working_schema(0);
  const char * tmp_started_since(0);
  const char * tmp_started_till(0);
  const char * tz(0);
  int version_length = -1;
  boost::local_time::time_zone_ptr tz_ptr;
  std::string started_since;
  std::string started_till;
  int32_t min_duration = -1;
  int32_t max_duration = -1;
  long min_rn = c_min_run_number;
  long max_rn = c_max_run_number;
  std::string release;
  std::string user;
  std::string host;
  std::string partition;
  std::string name;
  std::string sorted_by;

  for(int i = 1; i < argc; i++) {
    const char * cp = argv[i];

    if(!strcmp(cp, "-h") || !strcmp(cp, "--help")) {
      std::cout <<
        "usage: rn_ls\n"
        "   -c | --connect-string connect_string\n"
        "   -w | --working-schema schema_name\n"
        "   [-l | --list-databases]\n"
        "   [-d | --database name]\n"
        "   [-e | --user user_name_pattern]\n"
        "   [-r | --release release_pattern]\n"
        "   [-o | --host hostname_pattern]\n"
        "   [-p | --partition partition_name_pattern]\n"
        "   [-a | --config-name config_name_pattern]\n"
        "   [-s | --started-since timestamp]\n"
        "   [-t | --started-till timestamp]\n"
        "   [-f | --minimal-duration duration]\n"
        "   [-i | --maximal-duration duration]\n"
        "   [-n | --minimal-run-number number]\n"
        "   [-m | --maximal-run-number number]\n"
        "   [-z | --time-zone name]\n"
        "   [-q | --sorted-by parameters]\n"
        "   [-x | --short-version length]\n"
        "   [-h | --help]\n"
        "\n"
        "Options/Arguments:\n"
        "   -c connect_string    database connection string\n"
        "   -w schema_name       name of working schema\n"
        "   -l                   list database names holding run numbers\n"
        "   -d name              name of the database holding run numbers\n"
        "   -r release_pattern   show run numbers for release names satisfying pattern (see syntax description below)\n"
        "   -e user_name_pattern show run numbers for user names satisfying pattern (see syntax description below)\n"
        "   -o hostname_pattern  show run numbers for hostnames satisfying pattern (see syntax description below)\n"
        "   -p partition_pattern show run numbers for partition names satisfying pattern (see syntax description below)\n"
        "   -a name_pattern      show run numbers for config names satisfying pattern (see syntax description below)\n"
        "   -s timestamp         show run numbers started since given moment\n"
        "   -t timestamp         show run numbers started before given moment\n"
        "   -f duration          show runs last longer given duration\n"
        "   -i duration          show runs last shorter given duration\n"
        "   -n number            show run numbers greater or equal to given one\n"
        "   -m number            show run numbers less or equal to given one\n"
        "   -z name              time zone name (run with \"-z list-regions\" to see all supported time zones)\n"
        "   -q parameters        sort output by several columns; the parameters may contain the following\n"
        "                        items (where first symbol is for ascending and second for descending order):\n"
        "                          n | N - sort by run number;\n"
        "                          m | M - sort by database name;\n"
        "                          d | D - sort by duration;\n"
        "                          p | P - sort by partition names;\n"
        "                          r | R - sort by release names;\n"
        "                          u | U - sort by user names;\n"
        "                          h | H - sort by hostnames;\n"
        "   -x length            shortens the config version (in case of git sha) with length characters\n"
        "   -h                   print this message\n"
        "\n"
        "Description:\n"
        "   The utility prints out information about run numbers.\n"
        "   By default the input timestamps are expected and the results are printed in UTC.\n"
	"   To use specific local time run program with -z option (e.g. \"-z \'Europe/Zurich\'\" for CERN local time).\n"
        "   The timestamps to be provided in ISO 8601 format: \"YYYY-MM-DD HH:MM:SS\".\n"
        "   The durations  to be provided in format: \"HH:MM:SS\".\n"
        "   The allowed wildcard characters used to select by user, host and partition names are:\n"
        "    % (i.e. percent symbol)     - any string of zero or more characters;\n"
        "    _ (i.e. underscore symbol)  - any single character.\n\n";
      return 0;
    }
    else if(!strcmp(cp, "-l") || !strcmp(cp, "--list-databases")) {
      list_db = true;
    }
    else if(!strcmp(cp, "-d") || !strcmp(cp, "--database")) {
      if(++i == argc) { no_param(cp); return 1; } else { db_name = argv[i]; }
    }
    else if(!strcmp(cp, "-s") || !strcmp(cp, "--started-since")) {
      if(++i == argc) { no_param(cp); return 1; } else { tmp_started_since = argv[i]; }
    }
    else if(!strcmp(cp, "-t") || !strcmp(cp, "--started-till")) {
      if(++i == argc) { no_param(cp); return 1; } else { tmp_started_till = argv[i]; }
    }
    else if(!strcmp(cp, "-z") || !strcmp(cp, "--time-zone")) {
      if(++i == argc) { no_param(cp); return 1; } else { tz = argv[i]; }
    }
    else if(!strcmp(cp, "-f") || !strcmp(cp, "--minimal-duration")) {
      if(++i == argc) { no_param(cp); return 1; } else { if(!s2du(argv[i], min_duration, "minimal-duration")) return 2; }
    }
    else if(!strcmp(cp, "-i") || !strcmp(cp, "--maximal-duration")) {
      if(++i == argc) { no_param(cp); return 1; } else { if(!s2du(argv[i], max_duration, "maximal-duration")) return 2; }
    }
    else if(!strcmp(cp, "-n") || !strcmp(cp, "--minimal-run-number")) {
      if(++i == argc) { no_param(cp); return 1; } else { min_rn = atol(argv[i]); }
    }
    else if(!strcmp(cp, "-m") || !strcmp(cp, "--maximal-run-number")) {
      if(++i == argc) { no_param(cp); return 1; } else { max_rn = atol(argv[i]); }
    }
    else if(!strcmp(cp, "-x") || !strcmp(cp, "--short-version")) {
      if(++i == argc) { no_param(cp); return 1; } else { version_length = atoi(argv[i]); }
    }
    else if(!strcmp(cp, "-c") || !strcmp(cp, "--connect-string")) {
      if(++i == argc) { no_param(cp); return (1); } else { connect_string = argv[i]; }
    }
    else if(!strcmp(cp, "-w") || !strcmp(cp, "--working-schema")) {
      if(++i == argc) { no_param(cp); return (1); } else { working_schema = argv[i]; }
    }
    else if(!strcmp(cp, "-r") || !strcmp(cp, "--release")) {
      if(++i == argc) { no_param(cp); return (EXIT_FAILURE); } else { release = argv[i]; }
    }
    else if(!strcmp(cp, "-e") || !strcmp(cp, "--user")) {
      if(++i == argc) { no_param(cp); return (EXIT_FAILURE); } else { user = argv[i]; }
    }
    else if(!strcmp(cp, "-o") || !strcmp(cp, "--host")) {
      if(++i == argc) { no_param(cp); return (EXIT_FAILURE); } else { host = argv[i]; }
    }
    else if(!strcmp(cp, "-p") || !strcmp(cp, "--partition")) {
      if(++i == argc) { no_param(cp); return (EXIT_FAILURE); } else { partition = argv[i]; }
    }
    else if(!strcmp(cp, "-a") || !strcmp(cp, "--config-name")) {
      if(++i == argc) { no_param(cp); return (EXIT_FAILURE); } else { name = argv[i]; }
    }
    else if(!strcmp(cp, "-q") || !strcmp(cp, "--sorted-by")) {
      if(++i == argc) { no_param(cp); return (EXIT_FAILURE); } else { sorted_by = argv[i]; }
    }
    else {
      std::cerr << "ERROR: Unexpected parameter: \"" << cp << "\"\n\n";
      return 1;
    }
  }

  if(min_rn < c_min_run_number) {
    std::cerr << "ERROR: minimal-run-number must be greater 0\n";
    return 2;
  }

  if(max_rn < min_rn) {
    std::cerr << "ERROR: maximal-run-number must be greater or equal minimal-run-number\n";
    return 2;
  }

  if( (max_duration != -1) && (min_duration != -1) && (max_duration < min_duration) ) {
    std::cerr << "ERROR: maximal-duration must be greater or equal minimal-duration\n";
    return 2;
  }

  if (tz)
    {
      try
        {
          oks::tz::DB tz_db;

          if (!strcmp(tz, "list-regions"))
            {
              for (auto & i : tz_db.get_regions())
                {
                  std::cout << i << std::endl;
                }
              return 0;
            }

          tz_ptr = tz_db.get_tz_ptr(tz);
        }
      catch (const std::exception& ex)
        {
          std::cerr << "ERROR: time-zone db access failed: " << ex.what() << std::endl;
          return 2;
        }
    }

  if(version_length != -1)
    {
      if(version_length < 8)
        {
          std::cerr << "minimal short version length is 8, adjusting parameter" << std::endl;
          version_length = 8;
        }
    }

  if(!connect_string) {
    std::cerr << "ERROR: the connect string is required\n";
    return 2;
  }

  if(!working_schema) {
    std::cerr << "ERROR: the working schema is required\n";
    return 2;
  }

  try
    {
      if (tmp_started_since)
        {
          started_since = oks::tz::posix_time_2_iso_string(
              oks::tz::str_2_posix_time(tmp_started_since, tz_ptr, "started-since")
          );
        }

      if (tmp_started_till)
        {
          started_till = oks::tz::posix_time_2_iso_string(
              oks::tz::str_2_posix_time(tmp_started_till, tz_ptr, "started-till")
          );
        }
    }
  catch (const std::exception& ex)
    {
      std::cerr << "ERROR: " << ex.what() << std::endl;
      return 2;
    }


  try {
    coral::ConnectionService * connection = 0;
    std::unique_ptr<coral::ISessionProxy> session (tdaq::RunNumber::get_session(connect_string, static_cast<int>(coral::ReadOnly), connection));

    coral::ITable& table = session->schema(working_schema).tableHandle( tdaq::RunNumber::s_table_name );

    if(list_db) {
      std::unique_ptr< coral::IQuery > query( table.newQuery() );
      query->setRowCacheSize( 10 );
      query->addToOutputList( std::string("DISTINCT ") + tdaq::RunNumber::s_name_column );
      coral::ICursor& cursor = query->execute();

      std::set<std::string> names;

      while(cursor.next()) {
        names.insert(cursor.currentRow().begin()->data<std::string>());
      }
      
      std::cout << "Found " << names.size() << " run number databases:\n";
      
      for(std::set<std::string>::const_iterator i = names.begin(); i != names.end(); ++i) {
        std::cout << " - \'" << *i << '\'' << std::endl;
      }
    }
    else {
      std::unique_ptr< coral::IQuery > query( table.newQuery() );
      query->setRowCacheSize( 10 );

      coral::AttributeList conditions;
      std::set<std::string> conditions_set;

      if(db_name) {
        conditions.extend<std::string>( "nm" );
        conditions["nm"].data<std::string>() = db_name;
        conditions_set.insert(std::string(tdaq::RunNumber::s_name_column) + " = :nm");
      }

      if(!partition.empty()) {
        conditions.extend<std::string>( "prt" );
        conditions["prt"].data<std::string>() = partition;
	conditions_set.insert(std::string(tdaq::RunNumber::s_partition_name_column) + (has_a_pattern(partition) ? " like" : " =") + " :prt");
      }

      if(!name.empty()) {
        conditions.extend<std::string>( "cn" );
        conditions["cn"].data<std::string>() = name;
        conditions_set.insert(std::string(tdaq::RunNumber::s_config_name_column) + (has_a_pattern(name) ? " like" : " =") + " :cn");
      }

      if(!release.empty()) {
        conditions.extend<std::string>( "rl" );
        conditions["rl"].data<std::string>() = release;
	conditions_set.insert(std::string(tdaq::RunNumber::s_release_column) + (has_a_pattern(release) ? " like" : " =") + " :rl");
      }

      if(!user.empty()) {
        conditions.extend<std::string>( "usr" );
        conditions["usr"].data<std::string>() = user;
        conditions_set.insert(std::string(tdaq::RunNumber::s_created_by_column) + (has_a_pattern(user) ? " like" : " =") + " :usr");
      }

      if(!host.empty()) {
        conditions.extend<std::string>( "hst" );
        conditions["hst"].data<std::string>() = host;
        conditions_set.insert(std::string(tdaq::RunNumber::s_host_column) + (has_a_pattern(host) ? " like" : " =") + " :hst");
      }

      if(!started_since.empty()) {
        conditions.extend<std::string>( "ss" );
        conditions["ss"].data<std::string>() = started_since;
        conditions_set.insert(std::string(tdaq::RunNumber::s_start_at_column) + " >= :ss");
      }

      if(!started_till.empty()) {
        conditions.extend<std::string>( "st" );
        conditions["st"].data<std::string>() = started_till;
        conditions_set.insert(std::string(tdaq::RunNumber::s_start_at_column) + " <= :st");
      }

      if(min_duration != -1) {
        conditions.extend<int>( "min_d" );
        conditions["min_d"].data<int>() = min_duration;
        conditions_set.insert(std::string(tdaq::RunNumber::s_duration_column) + " >= :min_d");
      }

      if(max_duration != -1) {
        conditions.extend<int>( "max_d" );
        conditions["max_d"].data<int>() = max_duration;
        if(max_duration == 0) {
          conditions_set.insert(std::string("(") + tdaq::RunNumber::s_duration_column + " <= :max_d OR " + tdaq::RunNumber::s_duration_column + " is NULL)");
        }
        else { 
          conditions_set.insert(std::string(tdaq::RunNumber::s_duration_column) + " <= :max_d");
        }
      }

      if(min_rn != c_min_run_number) {
        conditions.extend<int>( "min_rn" );
        conditions["min_rn"].data<int>() = static_cast<int>(min_rn);
        conditions_set.insert(std::string(tdaq::RunNumber::s_run_number_column) + " >= :min_rn");
      }

      if(max_rn != c_max_run_number) {
        conditions.extend<int>( "max_rn" );
        conditions["max_rn"].data<int>() = static_cast<int>(max_rn);
        conditions_set.insert(std::string(tdaq::RunNumber::s_run_number_column) + " <= :max_rn");
      }
    
      std::string conditions_str;
    
      for(std::set<std::string>::const_iterator i = conditions_set.begin(); i != conditions_set.end(); ++i) {
        if(i != conditions_set.begin()) conditions_str += " AND ";
        conditions_str += *i;
      }

      if(!sorted_by.empty()) {
        for(const char * p = sorted_by.c_str(); *p != 0; ++p) {
          switch (*p) {
            case 'n': query->addToOrderList(tdaq::RunNumber::s_run_number_column); break;
            case 'N': query->addToOrderList(std::string(tdaq::RunNumber::s_run_number_column) + " DESC"); break;
            case 'm': query->addToOrderList(tdaq::RunNumber::s_name_column); break;
            case 'M': query->addToOrderList(std::string(tdaq::RunNumber::s_name_column) + " DESC"); break;
            case 'd': query->addToOrderList(tdaq::RunNumber::s_duration_column); break;
            case 'D': query->addToOrderList(std::string(tdaq::RunNumber::s_duration_column) + " DESC"); break;
            case 'p': query->addToOrderList(tdaq::RunNumber::s_partition_name_column); break;
            case 'P': query->addToOrderList(std::string(tdaq::RunNumber::s_partition_name_column) + " DESC"); break;
            case 'u': query->addToOrderList(tdaq::RunNumber::s_created_by_column); break;
            case 'U': query->addToOrderList(std::string(tdaq::RunNumber::s_created_by_column) + " DESC"); break;
            case 'h': query->addToOrderList(tdaq::RunNumber::s_host_column); break;
            case 'H': query->addToOrderList(std::string(tdaq::RunNumber::s_host_column) + " DESC"); break;
            case 'r': query->addToOrderList(tdaq::RunNumber::s_release_column); break;
            case 'R': query->addToOrderList(std::string(tdaq::RunNumber::s_release_column) + " DESC"); break;
           default:
              std::stringstream text;
              text << "ERROR: wrong value \'" << *p << "\' in the command line parameter: \"--sorted-by\" or \"-t\"";
              throw std::runtime_error( text.str().c_str() );
          }
        }
      }
      else {
        query->addToOrderList(tdaq::RunNumber::s_name_column);
        query->addToOrderList(std::string(tdaq::RunNumber::s_run_number_column) + " DESC");
      }

      query->setCondition( conditions_str, conditions );

      coral::ICursor& c = query->execute();

      std::vector<TableRow> rows;

        // fill the table

      rows.push_back(TableRow('=', '='));
      rows.push_back(TableRow(' ', '|', "Name","Num",(tz_ptr ? "Start At (local time)" : "Start At (UTC)"),"Duration","Release","User","Host","Partition","Version","Config Name","Comment"));
      rows.push_back(TableRow('=', '='));

      while(c.next()) {
        const coral::AttributeList& row = c.currentRow();
      
        int duration = -1;
        if(!row[tdaq::RunNumber::s_duration_column].isNull()) duration = row[tdaq::RunNumber::s_duration_column].data<int>();

        std::int64_t lasted = -1;
        if(!row[tdaq::RunNumber::s_lasted_column].isNull()) lasted = row[tdaq::RunNumber::s_lasted_column].data<long long>();

        std::int64_t started_at = -1;
        if(!row[tdaq::RunNumber::s_started_at_column].isNull()) started_at = row[tdaq::RunNumber::s_started_at_column].data<coral::TimeStamp>().nanosecond();

        std::string comments;
        if(!row[tdaq::RunNumber::s_comments_column].isNull()) comments = row[tdaq::RunNumber::s_comments_column].data<std::string>();

        int sv = -1;
        if(!row[tdaq::RunNumber::s_config_schema_column].isNull()) sv = row[tdaq::RunNumber::s_config_schema_column].data<int>();

        int dv = -1;
        if(!row[tdaq::RunNumber::s_config_data_column].isNull()) dv = row[tdaq::RunNumber::s_config_data_column].data<int>();

        std::string cv;
        if(!row[tdaq::RunNumber::s_config_version_column].isNull()) cv = row[tdaq::RunNumber::s_config_version_column].data<std::string>();

        std::string cn;
        if(!row[tdaq::RunNumber::s_config_name_column].isNull()) cn = row[tdaq::RunNumber::s_config_name_column].data<std::string>();

        rows.push_back(
          TableRow(' ', '|', tz_ptr, version_length,
            row[tdaq::RunNumber::s_name_column].data<std::string>(),
	    row[tdaq::RunNumber::s_run_number_column].data<int>(),
            row[tdaq::RunNumber::s_start_at_column].data<std::string>(),
            started_at,
	    duration,
	    lasted,
            row[tdaq::RunNumber::s_release_column].data<std::string>(),
	    row[tdaq::RunNumber::s_created_by_column].data<std::string>(),
	    row[tdaq::RunNumber::s_host_column].data<std::string>(),
	    row[tdaq::RunNumber::s_partition_name_column].data<std::string>(),
	    sv, dv, cv, cn,
  	    comments
          )
        );

      }

      rows.push_back(TableRow('=', '='));

        // detect columns widths

      unsigned short cw[] = {1,1,1,1,1,1,1,1,1,1,1};
      {
        for(std::vector<TableRow>::const_iterator i = rows.begin(); i != rows.end(); ++i) {
          for(unsigned short j = 0; j < sizeof(cw)/sizeof(unsigned short); ++j) {
            unsigned short len = (*i).m_items[j].size();
            if(len > cw[j]) cw[j]=len;
          }
        }
      }


        // print table

      {
        bool align_left[] = {false, false, false, false, false, false, false, false, false, false, true};

        for(std::vector<TableRow>::const_iterator i = rows.begin(); i != rows.end(); ++i) {
          std::cout.fill((*i).m_fill);
          for(unsigned short j = 0; j < sizeof(cw)/sizeof(unsigned short); ++j) {
            std::cout << (*i).m_separator << (*i).m_fill;
            std::cout.setf((align_left[j] ? std::ios::left : std::ios::right), std::ios::adjustfield);
            std::cout.width(cw[j]);
            std::cout << (*i).m_items[j] << (*i).m_fill;
          }
          std::cout << (*i).m_separator << std::endl;
      }
      }

    }

    session->transaction().commit();
    //connection->disconnect();
  }

  catch ( coral::Exception& e ) {
    std::cerr << "CORAL exception: " << e.what() << std::endl;
    return 11;
  }

  catch ( std::exception& e ) {
    std::cerr << "Standard C++ exception : " << e.what() << std::endl;
    return 12;
  }

  catch ( ... ) {
    std::cerr << "Exception caught (...)" << std::endl;
    return 13;
  }

  return 0;
}
