/*
 * RequesterSession.h
 *
 *  Created on: Jan 26, 2018
 *      Author: kolos
 */

#ifndef SWROD_HLT_REQUESTERSESSION_H_
#define SWROD_HLT_REQUESTERSESSION_H_

#include <assert.h>
#include <cstdint>

#include <memory>
#include <unordered_map>
#include <vector>

#include "ClearSession.h"
#include "IOService.h"
#include "RequestMessage.h"
#include "Session.h"

using namespace swrod::hlt;

namespace swrod {
    namespace hlt {

        class RequesterSession: public IOService, public Session {
        public:

            RequesterSession(const std::vector<uint32_t> & rob_ids,
                    std::shared_ptr<ClearSession> clear_session, uint32_t l2_robs_number) :
                    Session(this->getService()),
                    m_rob_ids(rob_ids),
                    m_transaction_id(0),
                    m_request_counter(0),
                    m_response_counter(0),
                    m_clear_session(clear_session) {
                if (l2_robs_number) {
                    m_l2_session.reset(
                            new RequesterSession(
                                    std::vector<uint32_t>(rob_ids.begin(),
                                            rob_ids.begin() + l2_robs_number), clear_session, 0));
                }
            }

            void connect(const std::string & name,
                    const boost::asio::ip::tcp::endpoint & endpoint) {
                asyncOpen(name, endpoint);

                while (state() != daq::asyncmsg::Session::State::OPEN) {
                    usleep(1000);
                }

                if (m_l2_session) {
                    m_l2_session->connect("HLT_L2_REQUESTOR", endpoint);
                }
            }

            uint32_t getL2RequestCounter() const {
                std::unique_lock<std::mutex> lock(m_mutex);
                return (m_l2_session ? m_l2_session->getRequestCounter() : 0);
            }

            uint32_t getL2ResponseCounter() const {
                std::unique_lock<std::mutex> lock(m_mutex);
                return (m_l2_session ? m_l2_session->getResponseCounter() : 0);
            }

            uint32_t getRequestCounter() const {
                std::unique_lock<std::mutex> lock(m_mutex);
                return m_request_counter;
            }

            uint32_t getResponseCounter() const {
                std::unique_lock<std::mutex> lock(m_mutex);
                return m_response_counter;
            }

            void sendL2Request(uint32_t l1id) {
                ERS_ASSERT(m_l2_session);
                m_l2_session->sendRequest(l1id);
            }

            void sendRequest(uint32_t l1id) {
                uint32_t trid;
                {
                    std::unique_lock<std::mutex> lock(m_mutex);
                    trid = m_transaction_id++;
                    m_l1ids[trid] = l1id;
                    ++m_request_counter;
                }
                m_clear_session->recordL1ID(l1id);
                std::unique_ptr < RequestMessage
                        > message(new RequestMessage(trid, l1id, m_rob_ids));
                asyncSend(std::move(message));}

            void skipL1ID(uint32_t l1id) {
                m_clear_session->skipL1ID(l1id);
            }

        private:
            void onReceive(std::unique_ptr<daq::asyncmsg::InputMessage> message) override {
                if (message->typeId() == ResponseMessage::TYPE_ID) {
                    std::unique_lock<std::mutex> lock(m_mutex);
                    ++m_response_counter;
                    L1IDMap::const_iterator it = m_l1ids.find(message->transactionId());
                    if (it != m_l1ids.end()) {
                        uint32_t l1id = it->second;
                        m_l1ids.erase(it);
                        lock.unlock();
                        m_clear_session->receivedL1ID(l1id);
                    } else {
                        ERS_ASSERT(0);
                    }
                } else {
                    ERS_ASSERT(0);
                }
            }

            void onSend(std::unique_ptr<const daq::asyncmsg::OutputMessage> message) override {
                asyncReceive(); // we want to get a response
            }

        private:
            typedef std::unordered_map<uint32_t, uint32_t> L1IDMap;

            const std::vector<uint32_t> m_rob_ids;
            mutable std::mutex m_mutex;
            uint32_t m_transaction_id;
            uint32_t m_request_counter;
            uint32_t m_response_counter;
            L1IDMap m_l1ids;

            std::shared_ptr<ClearSession> m_clear_session;
            std::shared_ptr<RequesterSession> m_l2_session;
        };
    }
}

#endif /* SWROD_HLT_REQUESTERSESSION_H_ */
