#include <wmi/plugins/dqplugin.h>

#include <list>
#include <memory>

#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>

#include <TH2.h>
#include <TH3.h>
#include <TError.h>
#include <TStyle.h>
#include <TCanvas.h>
#include <TMath.h>
#include <TEnv.h>
#include <TString.h>
#include <TClass.h>

#include <is/infodictionary.h>

#include <oh/OHRootReceiver.h>
#include <oh/OHUtil.h>
#include <rc/RunParamsNamed.h>

#include <dqm_config/AlgorithmConfig.h>
#include <dqm_config/AlgorithmConfigFactory.h>
#include <dqm_config/dal/DQAgent.h>
#include <dqm_config/dal/DQAlgorithm.h>
#include <dqm_config/dal/DQParameter.h>
#include <dqm_config/dal/DQRegion.h>
#include <dqm_config/dal/DQReference.h>

#include <dqmf/is/Result.h>

#include <config/Configuration.h>
#include <config/Change.h>
#include <config/ConfigObject.h>

#include <dal/Segment.h>
#include <dal/Partition.h>
#include <dal/util.h>

#include <wmi/link.h>
#include <wmi/br.h>
#include <wmi/hr.h>
#include <wmi/utils.h>
#include <wmi/table.h>
#include <wmi/list.h>
#include <wmi/text.h>
#include <wmi/picture.h>
#include <wmi/tooltip.h>

using namespace daq::wmi;

namespace 
{
    const Color ORANGE (255,160,0);
    const Color YELLOW (255,220,0);
    const Color RED (255,0,0);
    const Color GREEN (0,255,0);
    const Color GREY (128,128,128);
    const Color WHITE (255,255,255);
    const Color BLACK (0,0,0);
    const Color DA_GREY (96,96,96);
    const Color LT_GREY (204,204,204);
    const Color DDA_MARINE (0,64,128);
    const Color DA_MARINE (55,113,172);
    const Color LT_MARINE (186,200,236);
    const Color LT_SAND (244,244,235);
}

#define TIMESTAMP_TEXT_OPTIONS		1
#define NORMAL_TEXT_OPTIONS		2
#define TITLE_TEXT_OPTIONS		3
#define PARAM_TEXT_OPTIONS		2, BLACK
#define PARAM_VAL_TEXT_OPTIONS		2, DDA_MARINE 
#define TAGS_VAL_TEXT_OPTIONS		2, DA_GREY
#define THR_VAL_TEXT_OPTIONS( c )	2, c

namespace
{
    const std::string dqm_is_server_name = "DQM";
    
    void
    getResult( const std::string & name, const IPCPartition & partition, dqmf::is::Result & result )
    {
	try {
	    ISInfoDictionary dict( partition );
	    std::string is_name = dqm_is_server_name + "." + name;
	    dict.getValue( is_name, result );
	}
	catch( daq::is::Exception & ex ) {
            ERS_DEBUG( 1, ex );
	}
    }
    
    Color toColor( const dqmf::is::Result & result )
    {
	switch ( result.status ) {
	    case dqmf::is::Result::Disabled:
		return BLACK;
	    case dqmf::is::Result::Red:
		return RED;
	    case dqmf::is::Result::Yellow:
		return YELLOW;
	    case dqmf::is::Result::Green:
		return GREEN;
	    default:
		return GREY;
	}
    }
    
    const char * toString( const dqmf::is::Result & result )
    {
	switch ( result.status ) {
	    case dqmf::is::Result::Disabled:
		return "BLACK";
	    case dqmf::is::Result::Red:
		return "RED";
	    case dqmf::is::Result::Yellow:
		return "YELLOW";
	    case dqmf::is::Result::Green:
		return "GREEN";
	    default:
		return "GREY";
	}
    }
    
    std::string getTime( const dqmf::is::Result & result )
    {
    	std::string r = result.time().str();
        std::string::size_type pos = r.find( '.' );
        if ( pos != std::string::npos )
        {
            return r.substr( 0, pos );
        }
	return r;
    }
    
    bool isVisible( const dqmf::is::Result & result, unsigned int mask )
    {
    	unsigned int s = result.status + 1;
        return ( mask & ( 1 << s ) );
    }
    
    BaseHTMLObject *
    makeResultElement(	const std::string & full_name, 
    			const std::string & name, 
                        const IPCPartition & partition, 
                        dqmf::is::Result & result,
                        const Color & caption_color,
                        const Color & text_color,
                        bool use_webis,
                        const std::string & webis_prefix )
    {
	if ( use_webis )
        {
            Text * s = new Text( "<script type='text/javascript'>transformXML('"
            			  + webis_prefix
            			  + partition.name() + "/is/" 
                                  + dqm_is_server_name + "/"
                                  + dqm_is_server_name + "." + name
                                  + "', 'dqresult_full.xsl');</script>" );
	    if (full_name == name )
            {
            	return s;
            }
            
	    Table * r = new Table( 1, 2 );
            r -> setWidth( 100, true );
            r -> setAlignCell( 1, 1, 2 ); // center 
	    r -> addElement( 1, 1, new Text( full_name, 2, BLACK, TextOption("b") ) );
            r -> addElement( 1, 2, s );
	    return r;
        }
        
        getResult( name, partition, result );

	Table * r = new Table( 1, 1 );
        r -> setBorder( 0 );
        r -> setWidth( 100, true );
        r -> setCellPadding( 1 );
        r -> setCellSpacing( 1 );
	r -> setBGColor( caption_color );

	Table * t = new Table( 5, 1 );
	t -> setBGColor( caption_color );
	t -> addElement( 1, 1, new Text( full_name, TITLE_TEXT_OPTIONS, text_color ) );
	t -> setWidthCell( 2, 1, 20, false );
	t -> setBGColorCell( 3, 1, toColor( result ) );
	t -> setWidthCell( 3, 1, 20, false );
	t -> setWidthCell( 4, 1, 20, false );
	t -> addElement( 5, 1, new Text( getTime( result ), TIMESTAMP_TEXT_OPTIONS, text_color ) );

	r -> addElement( 1, 1, t );

	if ( result.tags.size() > 0 ) {
	    std::ostringstream out;
            for( unsigned int i = 0; i < result.tags.size(); i++ ) {
            	out <<  result.tags[i].name << " = " << result.tags[i].value << "; ";
            }
	    r -> addRow();
	    r -> setBGColorCell( 1, 2, WHITE );
            r -> addElement( 1, 2, new Text( out.str(), TAGS_VAL_TEXT_OPTIONS ) );
	}

	return r;
    }

    Table *
    makeParameterTable( const dqm_config::dal::DQParameter * dqparam, 
    			const IPCPartition & partition, 
                        const std::string & ref_name,
                        dqmf::is::Result & result,
                        bool use_webis,
                        const std::string & webis_prefix )
    {
        std::string name = dqparam->UID();

	Table * table = new Table( 1, 2 );
        table -> setWidth( 100, true );
	table -> setCellPadding( 0 );
	table -> setCellSpacing( 0 );

	BaseHTMLObject * r = makeResultElement( name, name, partition, result, 
        		LT_MARINE, BLACK, use_webis, webis_prefix );
	table -> addElement( 1, 1, r );
	std::string descr = dqparam->get_Description();
        std::string troub = dqparam->get_Troubleshooting();
        std::map<std::string, double> paramAlg;
        dqparam->get_parameters(paramAlg);
        std::map<std::string, double> greenTh;
        dqparam->get_green_thresholds(greenTh);
        std::map<std::string, double> redTh;
        dqparam->get_red_thresholds(redTh);

	int rows = 2 + !ref_name.empty() + !descr.empty() + !troub.empty() + !paramAlg.empty()
	        + (!greenTh.empty() || !redTh.empty());

	Table * t = new Table( 2, rows );
	t -> setCellPadding( 2 );
	t -> setCellSpacing( 0 );
        t -> setWidth( 100, true );
        t -> setTDWidth( 2, 100, true );
        t -> addElement( 1, 1, new Text( "Histogram(s)&nbsp;&nbsp;", PARAM_TEXT_OPTIONS ) );
        t -> setVAlignCell( 1, 1, 1 );

        int row = 1;
	const std::vector<std::string> & names = dqparam->get_InputDataSource();
	for ( unsigned int i = 0; i < names.size(); i++ ) {
	    std::string ref = webis_prefix + partition.name() + "/oh/" + names[i];
	    t -> addElement( 2, row, new Link( ref, names[i] ) );
	    if ( i != ( names.size() - 1 ))
		t -> addElement( 2, row, new Text( "; " ) );
            t -> addElement( 2, row, new BR() );
	}

        ++row;
        t -> setVAlignCell( 1, row, 1 );
        t -> addElement( 1, row, new Text( "Algorithm&nbsp;&nbsp;", PARAM_TEXT_OPTIONS ) );
        t -> addElement( 2, row, new Text( dqparam->get_DQAlgorithm()->UID(), PARAM_VAL_TEXT_OPTIONS ) );

        if (!descr.empty()) {
            ++row;
            t -> addElement( 1, row, new Text( "Description&nbsp;&nbsp;", PARAM_TEXT_OPTIONS ) );
            t -> setVAlignCell( 1, row, 1 );
            t -> addElement( 2, row, new Tooltip( descr ) );
        }

        if (!troub.empty()) {
            ++row;
            t -> addElement( 1, row, new Text( "Troubleshooting&nbsp;&nbsp;", PARAM_TEXT_OPTIONS ) );
            t -> setVAlignCell( 1, row, 1 );
            t -> addElement( 2, row, new Tooltip( troub ) );
        }

	if (!ref_name.empty()) {
	    ++row;
            t -> addElement( 1, row, new Text( "Reference&nbsp;&nbsp;", PARAM_TEXT_OPTIONS ) );
            t -> setVAlignCell( 1, row, 1 );
            t -> addElement( 2, row, new Text( ref_name, PARAM_VAL_TEXT_OPTIONS ) );
	}

	if (!paramAlg.empty()) {
            ++row;
            t -> addElement( 1, row, new Text( "Parameter(s)&nbsp;&nbsp;", PARAM_TEXT_OPTIONS ) );
            t -> setVAlignCell( 1, row, 1 );
            std::ostringstream out;
            std::map<std::string, double>::iterator it = paramAlg.begin();
            for( ; it != paramAlg.end(); ++it ) {
                out << it->first << " = " << it->second << "; ";
            }
            t -> addElement( 2, row, new Text( out.str(), PARAM_VAL_TEXT_OPTIONS ) );
	}

	if (!greenTh.empty() || !redTh.empty()) {
            ++row;
            t -> addElement( 1, row, new Text( "Threshold(s)&nbsp;&nbsp;", PARAM_TEXT_OPTIONS ) );
            t -> setVAlignCell( 1, row, 1 );
            std::map<std::string, double>::iterator git = greenTh.begin();
            for( ; git != greenTh.end(); ++git ) {
                std::string name = git->first;
                double gval = git->second;
                t -> addElement( 2, row, new Text( Utils::toString( gval ), THR_VAL_TEXT_OPTIONS( GREEN ) ) );

                std::map<std::string, double>::iterator rit = redTh.find(name);
                if ( rit != redTh.end() ) {
                    double rval = rit->second;
                    std::string sign = gval > rval ? ">" : "<";
                    t -> addElement( 2, row, new Text( sign, THR_VAL_TEXT_OPTIONS( BLACK ) ) );
                    t -> addElement( 2, row, new Text( name, THR_VAL_TEXT_OPTIONS( ORANGE ) ) );
                    t -> addElement( 2, row, new Text( sign, THR_VAL_TEXT_OPTIONS( BLACK ) ) );
                    t -> addElement( 2, row, new Text( Utils::toString( rval ), THR_VAL_TEXT_OPTIONS( RED ) ) );
                    break;
                }
                t -> addElement( 2, row, new Text( "; ", THR_VAL_TEXT_OPTIONS( BLACK ) ) );
            }
        }

        for (int i = 1; i <= rows; i += 2) {
            t -> setBGColorCell( 1, i, LT_SAND );
            t -> setBGColorCell( 2, i, LT_SAND );
        }

        table -> addElement( 1, 2, t );
	return table;
    }   
    
    class HistogramConverter
    {
        TCanvas	*					m_canvas;
        boost::shared_ptr<TH1>				m_histogram;
        std::string					m_file_name;
        std::string					m_partition_name;
        const dqm_config::dal::DQParameter * 		m_param;
        dqm_config::AlgorithmConfigFactory *		m_factory;
        
      public:
        HistogramConverter(	TCanvas * c, 
        			TH1 * h, 
                                const dqm_config::dal::DQParameter * p, 
                                const std::string & n, 
                                const std::string & f,
                                dqm_config::AlgorithmConfigFactory * af )
           : m_canvas( c ),
             m_histogram( h ),
             m_file_name( f ),
             m_partition_name( n ),
             m_param( p ),
             m_factory( af )
        {
        }
        
        void operator()()
        {
	    TObject * r = 0;
            if ( m_param )
	    {
		try {
		    std::shared_ptr<dqm_config::AlgorithmConfig> a = m_factory->createAlgorithmConfig( *m_param );
		    r = a -> getReference();
		} catch ( ... ) { ; }
	    }
            
	    TH1 * reference = 0;
            if ( r && r->IsA()->InheritsFrom( TH1::Class() ) )
	    {
		reference = (TH1*)r;
            }

	    double histo_norm = m_histogram->GetSumOfWeights();
	    if ( TMath::IsNaN( histo_norm ) )
            {
            	return ;
            }
            std::string opt;
	    if ( m_histogram->IsA()->InheritsFrom( TH2::Class() ) )
            {
            	opt = "col";
            }
	    else if ( m_histogram->IsA()->InheritsFrom( TH3::Class() ) )
            {
            	opt = "lego";
            }
            m_histogram -> GetXaxis()->SetLabelSize( 0.04 );
	    m_histogram -> GetYaxis()->SetLabelSize( 0.04 );
	    m_histogram -> Draw( opt.c_str() );
	    if ( reference && opt.empty() )
	    {
		double ref_norm = reference -> GetSumOfWeights();

		//scale	 histogram
		if ( !TMath::IsNaN( ref_norm ) )
		{
		    if ( ref_norm > 0.0 && histo_norm > 0.0 )
		    {
			reference -> Scale( histo_norm / ref_norm );
			double hmax = m_histogram -> GetMaximum();
			double rmax = reference -> GetMaximum();
			//make sure it fits
			if ( hmax < rmax ) {
			    reference -> SetMaximum( hmax/1.1 );
			}
		    }
		    reference -> SetLineColor( 2 );
		    reference -> Draw( "same" );
		}
	    }
	    m_canvas -> SaveAs( m_file_name.c_str() );
        }
    };
    
    Table *
    makeParametersTable( 
    		const dqm_config::dal::DQRegion * region, 
    		const IPCPartition & partition, 
                TCanvas * canvas,
                IPCPipeline & threadpool,
                dqm_config::AlgorithmConfigFactory* aconfig,
                unsigned int w,
                unsigned int h,
                unsigned int mask,
                const std::string & base_dir,
                const std::string & picture_format,
                bool show_histograms, 
                bool plot_references,
                bool use_webis,
                const std::string & webis_prefix )
    {
	std::vector<const dqm_config::dal::DQParameter*> params;
	region->get_all_parameters(params);
	if ( !params.size() )
        {
            return 0;
        }
        
	Table * r = new Table( 3, params.size() );
	r -> setCellPadding( 2 );
	r -> setCellSpacing( 5 );
	r -> setWidth( 100, true );
	r -> setTDWidth( 1, show_histograms ? 10 : 0 );
	r -> setTDWidth( 2, show_histograms ? w + 4 : 0 );
        
	for ( unsigned long i = 0; i < params.size(); ++i )
	{
	    const std::vector<const dqm_config::dal::DQReference*> & references = params[i] -> get_DQReferences();
            std::string refname = references.size() ? references[0]->get_Source() : "";
            
	    dqmf::is::Result result;
            Table * p = makeParameterTable( params[i], partition, refname, result, use_webis, webis_prefix );
	    r -> addElement( 3, i + 1, p );
	    r -> setVAlignCell( 3, i + 1, 1 );
	    r -> setAlignCell( 2, i + 1, 2 );
	    r -> setVAlignCell( 2, i + 1, 3 );
            
            if ( use_webis )
            {                
		const std::vector<std::string> & names = params[i]->get_InputDataSource();
                std::string hist_name = names[0];
		std::string ref = webis_prefix + partition.name() + "/oh/" + hist_name 
                		+ "?size=" + boost::lexical_cast<std::string>(w) + "x" + boost::lexical_cast<std::string>(h)
                                + "&type=" + picture_format;
		Picture * pic = new Picture( ref, w, h );
		r -> addElement( 2, i + 1, pic );
            	continue;
            }
            
            if ( !isVisible( result, mask ) )
            {
		std::ostringstream out;
                out << "Histograms with " << toString( result ) << " status are not configured to be displayed";
                r -> addElement( 2, i + 1, new Text( out.str(), 2, DA_GREY, TextOption( "i" ) ) );
            	continue;
            }
            
            try {
		const std::vector<std::string> & names = params[i]->get_InputDataSource();

		OHRootHistogram hd = OHRootReceiver::getRootHistogram( partition,
					    ::oh::util::get_server_name( names[0] ),
					    ::oh::util::get_provider_name( names[0] ),
					    ::oh::util::get_histogram_name( names[0] ) );

		std::string hist_name = names[0];
		for( unsigned int i1 = 0; i1 < hist_name.size(); i1++ ) {
		    if(hist_name[i1] == '/' || hist_name[i1] == ' ' || hist_name[i1] == '\\')
			hist_name[i1] = '_';
		}

		std::string short_name = hist_name + "." + picture_format;
		std::string full_name = base_dir + short_name;
                
		HistogramConverter job( canvas, 
                			hd.histogram.release(), 
                			( references.size() && plot_references ) ? params[i] : 0, 
                                        partition.name(), 
                                        full_name, 
                                        aconfig );
		threadpool.addJob( job );

		Picture * pic = new Picture( short_name, w, h );
		r -> addElement( 2, i + 1, pic );
	    }
	    catch( daq::oh::Exception & ex ) {
		r -> addElement( 2, i + 1, new Text( "Histogram is not available", 2, DA_GREY, TextOption( "i" ) ) );
	    }
	}

	return r;
    }

    void 
    writeTitle( const std::string & title, OutputStream & out, int index )
    {
	Table t( 1, 1 );
	t.setWidth( 100, true );
	t.setBGColor( DA_MARINE );
	t.setAlignCell( 1, 1, 2 );
	t.addElement( 1, 1, new Text( title, TITLE_TEXT_OPTIONS, WHITE ) );
	out.write( t, index );
    }

    template <class T>
    void getAgents( Configuration& config,
		    const T& parent,
		    std::vector<const dqm_config::dal::DQAgent *>& agents )
    {
	daq::core::SegmentIterator si = parent.get_Segments().begin();
	for ( ; si != parent.get_Segments().end(); ++si )
	{
	    daq::core::BaseApplicationIterator ai = (*si)->get_Applications().begin();
	    daq::core::BaseApplicationIterator end = (*si)->get_Applications().end();
	    for ( ; ai != end; ++ai )
	    {
		const dqm_config::dal::DQAgent * a =
		    config.cast<dqm_config::dal::DQAgent,daq::core::BaseApplication>(*ai);
		if ( a )
		{
		    agents.push_back(a);
		}
	    }
	    getAgents(config, *(*si), agents);
	}
    }

    bool visit_regions( const dqm_config::dal::DQRegion * region, DQPlugin::VisitorParameter param )
    {
	if ( param.plugin->stopped() )
            return false;

        param.region = region;
        std::string region_name = param.region->UID();
        std::string region_file_name(region_name);
        std::replace( region_file_name.begin(), region_file_name.end(), '/', '_');

	ERS_DEBUG( 1, "Generating file for the '" << region_name << "' region" );

	param.parent_name = param.parent_name + "/"+ region_name;

	dqmf::is::Result result;
	BaseHTMLObject * r = makeResultElement( param.parent_name, region_name, 
        		param.partition, result, DA_MARINE, WHITE, param.use_webis(), param.webis_prefix() );
	param.fd = param.plugin->out()->open( region_file_name + ".html", param.partition.name() );
	param.plugin->out()->write( *r, param.fd );
	BR br;
	param.plugin->out()->write( br, param.fd );
	delete r;

	unsigned int row = (*param.parent_row)++;
	std::string page_name = param.base_name.empty()
				    ? region_file_name + ".html"
				    : param.base_name + "/" + region_file_name + ".html";
	param.parent_table -> addElement( 2, row, new Link( page_name, region_name ) );

        if ( param.use_webis() )
        {
            Text * r = new Text( "<script type='text/javascript'>transformXML('"
            			  + param.webis_prefix()
            			  + param.partition.name() + "/is/" 
                                  + dqm_is_server_name + "/"
                                  + dqm_is_server_name + "." + region_name
                                  + "', 'dqresult_short.xsl');</script>" );
	    param.parent_table -> addElement( 6, row, r );
        }
        else
        {
	    param.parent_table -> setBGColorCell( 4, row, toColor( result ) );
	    param.parent_table -> addElement( 6, row, new Text( getTime( result ), TIMESTAMP_TEXT_OPTIONS ) );
	    param.parent_table -> setWidthCell( 4, 1, 20, false );
	    param.parent_table -> setWidthCell( 5, 1, 20, false );
        }
        
	param.parent_table -> setWidthCell( 1, 1, 20, false );
	param.parent_table -> setWidthCell( 3, 1, 20, false );
	param.parent_table -> setHeightCell( 3, 1, 20, false );

	std::vector<const dqm_config::dal::DQRegion*> regions;
	param.region->get_all_regions(regions);

	if (regions.size()) {
	    param.parent_table.reset(new Table(6, regions.size()));
	    param.parent_row.reset(new unsigned int(1));
	}
	else {
	    param.parent_table = std::shared_ptr<Table>();
	    param.parent_row = std::shared_ptr<unsigned int>();
	}
	param.base_name = "";

	for (size_t i = 0; i < regions.size(); i++) {
            if (!visit_regions(regions[i], param))
            	return false;
        }
        return true;
    }

}

DQPlugin::VisitorParameter::VisitorParameter(
	DQPlugin * pl,
	const IPCPartition & p,
	std::shared_ptr<Table> & t,
	unsigned int f )
  : plugin( pl ),
    region( 0 ),
    partition( p ),
    parent_name( p.name() ),
    base_name( p.name() ),
    parent_table( t ),
    parent_row( new unsigned int(1) ),
    fd( f )
{ ; }

DQPlugin::VisitorParameter::~VisitorParameter()
{
    if ( !plugin -> out() )
	return;

    if ( parent_table )
    {
	parent_table -> setCellPadding( 1 );
	plugin->out()->write( *parent_table, fd );
    }
    if ( region )
    {
	try 
        {
	    Table * r = makeParametersTable( region, partition,
                                             plugin->m_canvas, plugin->m_threadpool, plugin->m_aconfig,
					     plugin->m_picture_width, plugin->m_picture_height,
					     plugin->m_results_mask, plugin->m_base_dir,
					     plugin->m_picture_format,
					     plugin->m_show_histograms, plugin->m_plot_references,
					     plugin->m_use_webis, plugin->m_webis_prefix );
	    if ( r )
	    {
		BR br;
		plugin->out()->write( br, fd );
		plugin->out()->write( *r, fd );
		delete r;
	    }
	}
	catch ( daq::config::Generic & ex )
	{
	    ERS_LOG( ex );
	}
    }

    if (plugin->m_use_webis && plugin -> out() -> getFileDescription(fd)) {
	plugin -> out() -> getFileDescription(fd) -> addScriptFile( "transformxml.js" );
        plugin -> out() -> getFileDescription(fd) -> addHrefFile( "stylesheet", "text/css", "tooltip.css" );
    }
    plugin -> out() -> close( fd );
}

DQPlugin::DQPlugin()
  : m_stopped( false ),
    m_picture_height( 200 ),
    m_picture_width( 100 ),
    m_results_mask( (unsigned int)-1 ),
    m_show_histograms( false ),
    m_plot_references( false ),
    m_use_webis( false ),
    m_picture_format( "svg" ),
    m_webis_prefix( "/info/current/" ),
    m_canvas( 0 ),
    m_env( 0 ),
    m_threadpool( 1 )
{
    ;
}

DQPlugin::~DQPlugin()
{
    delete m_canvas;
    delete m_env;
}

bool 
DQPlugin::regenerateNecessary( std::list<IPCPartition> & lst )
{
    bool regenerate = false;
    // cleanup the map
    for( ActivePartitionsMap::iterator it = m_active_partitions.begin(); 
         it != m_active_partitions.end(); )
    {
    	if ( std::find( lst.begin(), lst.end(), it->first ) == lst.end() )
        {
            m_active_partitions.erase( it++ );
            regenerate = true;
        }
        else
        {
            ++it;
        }
    }
    
    for( std::list<IPCPartition>::iterator pp = lst.begin(); pp != lst.end(); ++pp )
    {            
    	if ( pp->name().find( "mirror" ) != std::string::npos )
	    continue;

	ActivePartitionsMap::iterator it = m_active_partitions.find( *pp );
        
        if ( it != m_active_partitions.end() )
        {
            RunParamsNamed rp( *pp, "RunParams.RunParams" );
            try {
            	rp.checkout();
                if ( it->second > rp.totalTime )
                {
                    regenerate = true;
                }
                it->second = rp.totalTime;
            }
            catch( daq::is::Exception & ex )
            {
                it->second = INT_MAX;
            }
            ERS_DEBUG( 1, "Partition '" << pp->name() << "' is already in the active map, regenerate = " << regenerate );
        }
        else
        {
	    try
            {
		std::string db_connect_string = "rdbconfig:" + pp->name() + "::RDB";
		std::unique_ptr<Configuration> config( new Configuration(db_connect_string));
		const daq::core::Partition* part_config = config -> get<daq::core::Partition>( pp->name() );
                if ( !part_config ) {
                    continue;
                }
		std::set<std::string> types;
                types.insert("DQAgent");
                std::vector<const daq::core::BaseApplication*> agents = part_config->get_all_applications( &types );

                size_t rn = 0;
                for ( size_t i = 0; i < agents.size(); ++i )
                {
		    const dqm_config::dal::DQAgent * a = config->cast<dqm_config::dal::DQAgent>(agents[i]);
                    rn += a->get_DQRegions().size();
                }
		if ( !rn ) {
		    continue;
		}
            }
            catch ( daq::config::Exception & ex)
            {
		ers::log(ex);
		continue;
            }
            
            RunParamsNamed rp( *pp, "RunParams.RunParams" );
            try {
            	rp.checkout();
                m_active_partitions[*pp] = rp.totalTime;
            }
            catch( daq::is::Exception & ex )
            {
                m_active_partitions[*pp] = INT_MAX;
            }
            regenerate = true;
            
            ERS_DEBUG( 1, "Partition '" << pp->name() << "' has been added to the active map, regenerate = " << regenerate );
        }
    }
    
    return ( regenerate );
}

bool 
DQPlugin::writePartition( const IPCPartition & partition )
{
    ERS_DEBUG( 1, "Generating files for the '" << partition.name() << "' partition" );
    
    std::vector<const dqm_config::dal::DQRegion *> regions;
    std::unique_ptr<Configuration> config;
    try 
    {
	std::string db_connect_string = "rdbconfig:" + partition.name() + "::RDB";
	config.reset( new Configuration(db_connect_string) );

	const daq::core::Partition* part_config = config->get<daq::core::Partition>( partition.name() );

	config->register_converter( new daq::core::SubstituteVariables( *part_config ) );
	dqm_config::dal::DQRegion::get_root_regions( *config, *part_config, regions );
    }
    catch ( daq::config::Generic & ex )
    {
    	ERS_LOG( ex );
    	return false;
    }
    
    if ( regions.empty() ) {
    	return false;
    }
    
    int index = m_outStream->open(partition.name() + ".html");
    m_base_dir = m_outStream -> getOutputDir() + m_outStream -> getPluginName() + "_new/" + partition.name() + "/";
    
    writeTitle( partition.name(), *m_outStream, index );
    
    BR br;
    m_outStream->write( br, index );
       
    m_aconfig = new dqm_config::AlgorithmConfigFactory(partition.name());

    std::shared_ptr<Table> table( new Table( 6, regions.size() ) );
    VisitorParameter parameter( this, partition, table, index );
    
    bool result = true;
    try
    {
	for( size_t i = 0; i < regions.size(); i++ ) {
            visit_regions( regions[i], parameter );
	    if ( stopped() )
		break ;
	}
    }
    catch ( ers::Issue & ex )
    {
    	ERS_LOG( ex );
        result = false;
    }
    
    m_threadpool.waitForCompletion();
    delete m_aconfig;
    
    return result;
}

void 
DQPlugin::periodicAction()
{
    Text ws = Text( "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp" );
    std::list<IPCPartition> lst;
    IPCPartition::getPartitions(lst);

    if ( m_use_webis && !regenerateNecessary( lst ) )
    	return ;
        
    int index = m_outStream->open();
    writeTitle( "Partitions", *m_outStream, index );
    BR br;
    m_outStream->write( br, index );

    for( ActivePartitionsMap::iterator it = m_active_partitions.begin(); 
    	 it != m_active_partitions.end(); ++it )
    {
        if ( stopped() )
            break ;
            
    	if ( !m_partition.empty() && m_partition != it->first.name() )
        {
            continue;
        }
        
	if ( writePartition( it->first ) )
	{
	    Link ln( it->first.name() + ".html", it->first.name() );
	    m_outStream->write( ws, index );
	    m_outStream->write( ln, index );
	    m_outStream->write( br, index );
	}
    }
    m_outStream->close(index);
    m_outStream->closeAll();
}

void 
DQPlugin::configure( const ::wmi::InfoParameter& params )
{
    for( unsigned int i = 0; i < params.length(); ++i ) {
	std::string name( params[i].name );
        if( name == "picture_width" ) {
	    sscanf( (const char *)params[i].value, "%u", &m_picture_width );
            continue ;
	} 
	if ( name == "picture_height" ) {
	    sscanf( (const char *)params[i].value, "%u", &m_picture_height );
            continue ;
	}
	if ( name == "show_histograms" ) {
            if ( !::strcasecmp( (const char *)params[i].value, "yes" ) )
            	m_show_histograms = true;
            continue ;
	}
	if ( name == "use_webis" ) {
            if ( !::strcasecmp( (const char *)params[i].value, "yes" ) )
            	m_use_webis = true;
            continue ;
	}
	if ( name == "plot_references" ) {
            if ( !::strcasecmp( (const char *)params[i].value, "yes" ) )
            	m_plot_references = true;
            continue ;
	}
	if ( name == "picture_format" ) {
	    m_picture_format = (const char *)params[i].value;
            continue ;
	}
	if ( name == "webis_prefix" ) {
	    m_webis_prefix = (const char *)params[i].value;
            continue ;
	}
	if ( name == "partition_name" ) {
	    m_partition = (const char *)params[i].value;
            continue ;
	}
	if ( name == "results_mask" ) {
	    sscanf( (const char *)params[i].value, "%u", &m_results_mask );
            continue ;
	}
    }

    m_canvas = new TCanvas( "c1", "c1", m_picture_width, m_picture_height );
    unsigned int w = m_picture_width + (m_picture_width - m_canvas -> GetWw());
    unsigned int h = m_picture_height + (m_picture_height - m_canvas -> GetWh());
    delete m_canvas;
    m_canvas = new TCanvas( "c1", "c1", w, h );
    
    gErrorIgnoreLevel = kFatal;
    gEnv -> SetValue( "Root.ErrorIgnoreLevel", "Fatal" );
}

void 
DQPlugin::stop()
{
    ERS_DEBUG( 0, "Stopping plugin" );
    m_stopped = true;
}

