//******************************************************************************
// file: AltiModule.cc
// desc: class for ALTI Module
// auth: 24-OCT-2017 P. Kuzmanovic
//******************************************************************************

#include "RCDUtilities/RCDUtilities.h"
#include "RCDBitString/BitSet.h"
//#include "MuctpiCommon/MuctpiCommon.h"

#include "ALTI/AltiModule.h"

#include "ALTI/AltiCommon.h"
#include "ALTI/AltiConfiguration.h"
#include "ALTI/AltiCounter.h"
#include "ALTI/AltiBusyCounter.h"

#include "unistd.h"

#include <fstream>
#include <iostream>
#include <sstream>
#include <iomanip>

#include <cmath>

#ifndef QUIET
#define MERR(s,x) (s << "ERROR in " << __PRETTY_FUNCTION__ << ":\n>> " << x << "\n")
#else
#define MERR(s,x)
#endif
#define CHECKRANGE0(val, hi, txt) if (val > hi) { CERR("ALTI: %s \"%d\" not in allowed range [0..%d]", txt, val, hi); return (-1); }

namespace LVL1 {

//#undef SI5344_REVB_REG_CONFIG_HEADER
//#undef SI5344_REVB_REG_CONFIG_NUM_REGS

    struct _jitter {
#include "ALTI/Si5344-RevB-ALTI_001-Registers.h"
        static const unsigned int NUMBER = SI5344_REVB_REG_CONFIG_NUM_REGS;
    };
    static _jitter JITTER;
}

using namespace LVL1;
//using namespace RCD;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

AltiModule::DataSegment::DataSegment(const std::string &name, const unsigned int size) : m_segment(0), m_data(0) {

    m_name = name;
    m_segment = new RCD::CMEMSegment(m_name, size*sizeof(uint32_t), true);
    if (((*m_segment))() != CMEM_RCC_SUCCESS) {
        CERR("ALTI: opening CMEM segment \"%s\" of size 0x%08x", m_name.c_str(), m_segment->Size());
        m_segment = 0;
        return;
    }
    m_data = (unsigned int*) (m_segment->VirtualAddress());
    CDBG("ALTI: INFO - CMEM segment \"%s\" opened of size 0x%08x, phys 0x%08x, virt 0x%08x", m_name.c_str(), m_segment->Size(), (unsigned int) m_segment->PhysicalAddress(), (unsigned int) m_segment->VirtualAddress());
}

//------------------------------------------------------------------------------

AltiModule::DataSegment::~DataSegment() {

    if (m_segment) {
        delete m_segment;
        CDBG("ALTI: INFO - closed CMEM segment \"%s\"", m_name.c_str());
    }
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

AltiModule::ECR_GENERATION::ECR_GENERATION() {

    // default values
    type                = ECR_VME;
    length              = 1;
    frequency           = 0;
    busy_before         = 0;
    busy_after          = 0;
    orbit_offset        = 0;
}

//------------------------------------------------------------------------------

AltiModule::ECR_GENERATION::ECR_GENERATION(const ECR_GENERATION &ecr) {

    // copy values
    type                = ecr.type;
    length              = ecr.length;
    frequency           = ecr.frequency;
    busy_before         = ecr.busy_before;
    busy_after          = ecr.busy_after;
}

//------------------------------------------------------------------------------

AltiModule::ECR_GENERATION &AltiModule::ECR_GENERATION::operator=(const ECR_GENERATION &ecr) {

    // copy values if necessary
    if (this != &ecr) {
        type            = ecr.type;
        length          = ecr.length;
        frequency       = ecr.frequency;
        busy_before     = ecr.busy_before;
        busy_after      = ecr.busy_after;
        orbit_offset    = ecr.orbit_offset;
    }

    return (*this);
}

//------------------------------------------------------------------------------

AltiModule::ECR_GENERATION::~ECR_GENERATION() {

    // nothing to be done
}

//------------------------------------------------------------------------------

int AltiModule::ECR_GENERATION::read(const std::string &str) {

    std::string s;
    std::string::size_type idx;
    unsigned int i;
    bool flag(false);

    // get ECR type
    if ((idx = str.find(",")) == std::string::npos) return (FAILURE);
    s = str.substr(0, idx);
    for (i = 0; i < ECR_TYPE_NUMBER; i++) {
        if (s == ECR_TYPE_NAME[i]) {
            flag = true;
            break;
        }
    }
    if (!flag) return (FAILURE);
    type = static_cast<ECR_TYPE>(i);
    
    // get ECR parameters
    s = AltiCommon::leftTrimString(str.substr(idx + 1));
    //if (std::sscanf(s.c_str(), "LEN=%u,FRQ=%u,BEF=%u,AFT=%u,OFF=%u", &length, &frequency, &busy_before, &busy_after, &orbit_offset) != 5) return (FAILURE);
    long length_i, frequency_i, busy_before_i, busy_after_i, orbit_offset_i;
    if (std::sscanf(s.c_str(), "LEN=%ld,FRQ=%ld,BEF=%ld,AFT=%ld,OFF=%ld", &length_i, &frequency_i, &busy_before_i, &busy_after_i, &orbit_offset_i) != 5) return (FAILURE);
    
    if ((length_i < 0) || (length_i > (long) (ECR_LENGTH))) return (FAILURE);
    else length = length_i;
    
    if ((frequency_i < 0) || (frequency_i > (long) ECR_FREQUENCY_MASK)) return (FAILURE);
    else frequency = frequency_i;
    
    if ((busy_before_i < 0) || (busy_before_i > (long) ECR_BUSY_BEFORE_MASK)) return (FAILURE);
    else busy_before = busy_before_i;

    if ((busy_after_i < 0) || (busy_after_i > (long) ECR_BUSY_AFTER_MASK)) return (FAILURE);
    else busy_after = busy_after_i;

    if ((orbit_offset_i < 0) || (orbit_offset_i > (long) ECR_ORBIT_OFFSET_MASK)) return (FAILURE);
    else orbit_offset = orbit_offset_i;

    return (SUCCESS);
}

//------------------------------------------------------------------------------

void AltiModule::ECR_GENERATION::dump(std::ostream &os) const {

    char s[AltiCommon::STRING_LENGTH];

    std::sprintf(s, "ECR type         = %s\n", AltiModule::ECR_TYPE_NAME[type].c_str()); os << s;
    std::sprintf(s, "ECR length       = %d\n", length); os << s;
    if (type == AltiModule::ECR_INTERNAL) {
        std::sprintf(s, "ECR frequency    = 0x%08x (= %10d)\n", frequency, frequency); os << s;
    }
    std::sprintf(s, "ECR BUSY before  =   0x%06x (=   %8d)\n", busy_before, busy_before); os << s;
    std::sprintf(s, "ECR BUSY after   =   0x%06x (=   %8d)\n", busy_after, busy_after); os << s;
    std::sprintf(s, "ECR OBRIT offset =      0x%03x (=       %4d)\n", orbit_offset, orbit_offset); os << s;
}

//------------------------------------------------------------------------------

std::string AltiModule::ECR_GENERATION::print() const {

    char s[AltiCommon::STRING_LENGTH];

    std::sprintf(s, "%s,LEN=%d,FRQ=%d,BEF=%d,AFT=%d,OFF=%d", AltiModule::ECR_TYPE_NAME[type].c_str(), length, frequency, busy_before, busy_after, orbit_offset);

    return (s);
}

//------------------------------------------------------------------------------

const std::string AltiModule::ECR_TYPE_NAME[ECR_TYPE_NUMBER] {
    "VME", "INTERNAL"
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

const std::string AltiModule::SIGNAL_NAME[SIGNAL_NUMBER] =
  { "BC", "ORB", "L1A", "TTR1", "TTR2", "TTR3", "TTYP0", "TTYP1", "TTYP2", "TTYP3", "TTYP4", "TTYP5", "TTYP6", "TTYP7", "BGO0", "BGO1", "BGO2", "BGO3", "BUSY", "CALREQ0", "CALREQ1", "CALREQ2" };

const unsigned int AltiModule::MEM_MASK[SIGNAL_NUMBER] =
  {
    0x00000000,                             // BC,         not mapped
    SNAPSHOT::STAT_MASK_ORB,                // ORB,        bit     20
    SNAPSHOT::STAT_MASK_L1A,                // L1A,        bit      0
    SNAPSHOT::STAT_MASK_TTR[0],             // TTR1,       bit      1
    SNAPSHOT::STAT_MASK_TTR[1],             // TTR2,       bit      2
    SNAPSHOT::STAT_MASK_TTR[2],             // TTR3,       bit      3
    SNAPSHOT::STAT_MASK_TTYP[0],            // TTYP0,      bit      8
    SNAPSHOT::STAT_MASK_TTYP[1],            // TTYP1,      bit      9
    SNAPSHOT::STAT_MASK_TTYP[2],            // TTYP2,      bit     10
    SNAPSHOT::STAT_MASK_TTYP[3],            // TTYP3,      bit     11
    SNAPSHOT::STAT_MASK_TTYP[4],            // TTYP4,      bit     12
    SNAPSHOT::STAT_MASK_TTYP[5],            // TTYP5,      bit     13
    SNAPSHOT::STAT_MASK_TTYP[6],            // TTYP6,      bit     14
    SNAPSHOT::STAT_MASK_TTYP[7],            // TTYP7,      bit     15
    SNAPSHOT::STAT_MASK_BGO[0],             // BGO0,       bit      4
    SNAPSHOT::STAT_MASK_BGO[1],             // BGO1,       bit      5
    SNAPSHOT::STAT_MASK_BGO[2],             // BGO2,       bit      6
    SNAPSHOT::STAT_MASK_BGO[3],             // BGO3,       bit      7
    SNAPSHOT::STAT_MASK_BUSY,               // BUSY,       bit     19
    SNAPSHOT::STAT_MASK_CALREQ[0],          // CALREQ0,    bits    16
    SNAPSHOT::STAT_MASK_CALREQ[1],          // CALREQ1,    bits    17
    SNAPSHOT::STAT_MASK_CALREQ[2]           // CALREQ2,    bits    18
  };

const std::string AltiModule::SIGNAL_SOURCE_NAME[SIGNAL_SOURCE_NUMBER] =
  { "CTP_IN", "ALTI_IN", "NIM_IN", "FROM_FPGA", "INVALID_CONFIG" };

const std::string AltiModule::SIGNAL_DESTINATION_NAME[SIGNAL_DESTINATION_NUMBER] =
  { "CTP_OUT", "ALTI_OUT", "NIM_OUT", "TO_FPGA" };

const std::string AltiModule::ALTI_MODE_NAME[ALTI_MODE_NUMBER] =
  { "MASTER", "CTP_SLAVE", "ALTI_SLAVE", "LEMO_SLAVE" };

// indexing: what SIGNAL and what SIGNAL_SOURCE we want
// output: actual SIGNAL_SOURCE for the switch of that particular signal
const AltiModule::SIGNAL_SOURCE AltiModule::SWX_MAP_SRC[SIGNAL_NUMBER_ROUTED][SIGNAL_SOURCE_NUMBER - 1] = {
    { CTP_IN , NIM_IN, ALTI_IN  , FROM_FPGA }, // BC
    { ALTI_IN, NIM_IN, FROM_FPGA, CTP_IN    }, // ORB
    { ALTI_IN, NIM_IN, FROM_FPGA, CTP_IN    }, // L1A
    { ALTI_IN, NIM_IN, FROM_FPGA, CTP_IN    },
    { ALTI_IN, NIM_IN, FROM_FPGA, CTP_IN    },
    { ALTI_IN, NIM_IN, FROM_FPGA, CTP_IN    },
    { ALTI_IN, NIM_IN, FROM_FPGA, CTP_IN    },
    { ALTI_IN, NIM_IN, FROM_FPGA, CTP_IN    },
    { ALTI_IN, NIM_IN, FROM_FPGA, CTP_IN    },
    { ALTI_IN, NIM_IN, FROM_FPGA, CTP_IN    },
    { ALTI_IN, NIM_IN, FROM_FPGA, CTP_IN    },
    { ALTI_IN, NIM_IN, FROM_FPGA, CTP_IN    },
    { ALTI_IN, NIM_IN, FROM_FPGA, CTP_IN    },
    { ALTI_IN, NIM_IN, FROM_FPGA, CTP_IN    },
    { ALTI_IN, NIM_IN, FROM_FPGA, CTP_IN    },
    { ALTI_IN, NIM_IN, FROM_FPGA, CTP_IN    },
    { ALTI_IN, NIM_IN, FROM_FPGA, CTP_IN    },
    { ALTI_IN, NIM_IN, FROM_FPGA, CTP_IN    }
  };

// indexing: what SIGNAL and what actual SIGNAL_SOURCE we want
// output: SIGNAL_SOURCE for the switch of that particular signal
const AltiModule::SIGNAL_SOURCE AltiModule::SWX_MAP_SRC_INV[SIGNAL_NUMBER_ROUTED][SIGNAL_SOURCE_NUMBER - 1] = {
    { CTP_IN   , NIM_IN, ALTI_IN, FROM_FPGA }, // BC
    { FROM_FPGA, CTP_IN, ALTI_IN, NIM_IN    }, // ORB
    { FROM_FPGA, CTP_IN, ALTI_IN, NIM_IN    }, // L1A
    { FROM_FPGA, CTP_IN, ALTI_IN, NIM_IN    },
    { FROM_FPGA, CTP_IN, ALTI_IN, NIM_IN    },
    { FROM_FPGA, CTP_IN, ALTI_IN, NIM_IN    },
    { FROM_FPGA, CTP_IN, ALTI_IN, NIM_IN    },
    { FROM_FPGA, CTP_IN, ALTI_IN, NIM_IN    },
    { FROM_FPGA, CTP_IN, ALTI_IN, NIM_IN    },
    { FROM_FPGA, CTP_IN, ALTI_IN, NIM_IN    },
    { FROM_FPGA, CTP_IN, ALTI_IN, NIM_IN    },
    { FROM_FPGA, CTP_IN, ALTI_IN, NIM_IN    },
    { FROM_FPGA, CTP_IN, ALTI_IN, NIM_IN    },
    { FROM_FPGA, CTP_IN, ALTI_IN, NIM_IN    },
    { FROM_FPGA, CTP_IN, ALTI_IN, NIM_IN    },
    { FROM_FPGA, CTP_IN, ALTI_IN, NIM_IN    },
    { FROM_FPGA, CTP_IN, ALTI_IN, NIM_IN    },
    { FROM_FPGA, CTP_IN, ALTI_IN, NIM_IN    },
  };

// indexing: what SIGNAL and what SIGNAL_DESTINATION we want
// output: actual SIGNAL_DESTINATION for the switch of that particular signal
const AltiModule::SIGNAL_DESTINATION AltiModule::SWX_MAP_DST[SIGNAL_NUMBER_ROUTED][SIGNAL_DESTINATION_NUMBER] = {
    { CTP_OUT , NIM_OUT, ALTI_OUT, TO_FPGA }, // BC
    { ALTI_OUT, NIM_OUT, TO_FPGA , CTP_OUT }, // ORB
    { ALTI_OUT, NIM_OUT, TO_FPGA , CTP_OUT }, // L1A
    { ALTI_OUT, NIM_OUT, TO_FPGA , CTP_OUT },
    { ALTI_OUT, NIM_OUT, TO_FPGA , CTP_OUT },
    { ALTI_OUT, NIM_OUT, TO_FPGA , CTP_OUT },
    { ALTI_OUT, NIM_OUT, TO_FPGA , CTP_OUT },
    { ALTI_OUT, NIM_OUT, TO_FPGA , CTP_OUT },
    { ALTI_OUT, NIM_OUT, TO_FPGA , CTP_OUT },
    { ALTI_OUT, NIM_OUT, TO_FPGA , CTP_OUT },
    { ALTI_OUT, NIM_OUT, TO_FPGA , CTP_OUT },
    { ALTI_OUT, NIM_OUT, TO_FPGA , CTP_OUT },
    { ALTI_OUT, NIM_OUT, TO_FPGA , CTP_OUT },
    { ALTI_OUT, NIM_OUT, TO_FPGA , CTP_OUT },
    { ALTI_OUT, NIM_OUT, TO_FPGA , CTP_OUT },
    { ALTI_OUT, NIM_OUT, TO_FPGA , CTP_OUT },
    { ALTI_OUT, NIM_OUT, TO_FPGA , CTP_OUT },
    { ALTI_OUT, NIM_OUT, TO_FPGA , CTP_OUT }
  };

const std::string AltiModule::EQUALIZER_VOLTAGE_NAME[EQUALIZER_VOLTAGE_NUMBER] =
  { "V_PEAK", "V_POLE", "V_GAIN", "V_OFFSET" };

const std::string AltiModule::EQUALIZER_CONFIG_NAME[EQUALIZER_CONFIG_NUMBER] =
  { "SHORT_CABLE", "LONG_CABLE", "CUSTOM_CONFIG" };

const unsigned int AltiModule::EQUALIZER_CONFIG_VOLTAGE[EQUALIZER_CONFIG_NUMBER][EQUALIZER_VOLTAGE_NUMBER] = {
    { (unsigned int) round(0.00/2.5*255.0), (unsigned int) round(0.00/2.5*255.0), (unsigned int) round(0.00/2.5*255.0), (unsigned int) round(0.38/2.5*255.0) }, // short cable
    { (unsigned int) round(1.00/2.5*255.0), (unsigned int) round(0.60/2.5*255.0), (unsigned int) round(2.00/2.5*255.0), (unsigned int) round(0.32/2.5*255.0) }  // long  cable
  };

const std::string AltiModule::CLK_NAME[CLK_NUMBER] =
  { "PRIMARY_FXO", "SECONDARY_EXT" };

const std::string AltiModule::EXT_CLK_SOURCE_NAME[EXT_CLK_SOURCE_NUMBER] =
  //{ "EXT_JC", "EXT_CTP_IN", "EXT_ALTI_IN", "EXT_NIM_IN" };
  { "Jitter cleaner", "CTP_IN", "ALTI_IN", "NIM_IN" };

const std::string AltiModule::CNT_NAME[CNT_NUMBER] =
  { "CNT_BC", "CNT_ORB", "CNT_L1A", "CNT_TTR1", "CNT_TTR2", "CNT_TTR3", "CNT_BGO0", "CNT_BGO1", "CNT_BGO2", "CNT_BGO3" };

const unsigned int AltiModule::CNT_ADDR = ALTI::ADDR_CNT_CNT + ALTI::ADDR_CNT_CNT_COUNTERVALUE;

const std::string AltiModule::SNAPSHOT_TYPE_NAME[SNAPSHOT_TYPE_NUMBER] =
  { "ALTI_LINE", "LTP_LINE" };

// busy
const std::string AltiModule::BUSY_INPUT_NAME[BUSY_INPUT_NUMBER] =
  { "BUSY_FROM_FRONT_PANEL", "BUSY_FROM_CTP", "BUSY_FROM_ALTI", "BUSY_FROM_PG", "BUSY_FROM_VME", "BUSY_FROM_ECR" };

const std::string AltiModule::BUSY_SOURCE_NAME[BUSY_SOURCE_NUMBER] =
  { "BUSY_INACTIVE", "BUSY_LOCAL", "BUSY_CTP", "BUSY_ALTI" };

const std::string AltiModule::BUSY_OUTPUT_NAME[BUSY_OUTPUT_NUMBER] =
  { "BUSY_TO_CTP", "BUSY_TO_ALTI" };

const std::string AltiModule::BUSY_LEVEL_NAME[BUSY_LEVEL_NUMBER] =
  { "NIM", "TTL" };

// calibration requests
const std::string AltiModule::CALREQ_INPUT_NAME[CALREQ_INPUT_NUMBER] =
  { "CALREQ_FROM_RJ45", "CALREQ_FROM_FRONT_PANEL", "CALREQ_FROM_CTP", "CALREQ_FROM_ALTI", "CALREQ_FROM_PG", "CALREQ_FROM_VME" };

const std::string AltiModule::CALREQ_SOURCE_NAME[CALREQ_SOURCE_NUMBER] =
  { "CALREQ_INACTIVE", "CALREQ_LOCAL", "CALREQ_CTP", "CALREQ_ALTI" };

const std::string AltiModule::CALREQ_OUTPUT_NAME[CALREQ_OUTPUT_NUMBER] =
  { "CALREQ_TO_CTP", "CALREQ_TO_ALTI" };

const std::string AltiModule::TTR_LEMO_NAME[TTR_LEMO_NUMBER] =
  { "TTR", "CALREQ" };

const std::string AltiModule::ALTI_RAM_MEMORY_NAME[ALTI_RAM_MEMORY_NUMBER] =
  { "TTC_DECODER_COMMAND", "TTC_DECODER_TIMESTAMP", "PATTERN_SNAPSHOT" };

const std::string AltiModule::TTC_FRAME_TYPE_NAME[TTC_FRAME_TYPE_NUMBER] =
  { "SHORT", "LONG" };

const std::string AltiModule::TTC_ADDRESS_SPACE_NAME[TTC_ADDRESS_SPACE_NUMBER] =
  { "INTERNAL", "EXTERNAL" };

const std::string AltiModule::TTC_BGO_MODE_TRIGGER_NAME[TTC_BGO_MODE_TRIGGER_NUMBER] =
  { "BGO_SIGNAL", "VME" };

const std::string AltiModule::TTC_BGO_MODE_COMMAND_NAME[TTC_BGO_MODE_COMMAND_NUMBER] =
  { "SYNCHRONOUS", "ASYNCHRONOUS" };

const std::string AltiModule::TTC_BGO_MODE_REPEAT_NAME[TTC_BGO_MODE_REPEAT_NUMBER] =
  { "SINGLE", "REPETITIVE" };

const std::string AltiModule::TTC_BGO_MODE_FIFO_NAME[TTC_BGO_MODE_FIFO_NUMBER] =
  { "WHEN_NOT_EMPTY", "ON_TRIGGER" };

const std::string AltiModule::TTC_BGO_MODE_NAME[TTC_BGO_MODE_NUMBER] =
  { "SYNCHRONOUS_SINGLE_BGO_SIGNAL", "SYNCHRONOUS_REPETITIVE", "ASYNCHRONOUS_BGO_SIGNAL", "ASYNCHRONOUS_VME_ON_TRIGGER", "ASYNCHRONOUS_VME_WHEN_NOT_EMPTY", "INVALID_MODE" };

const AltiModule::BgoMode AltiModule::TTC_BGO_MODE_HL[TTC_BGO_MODE_NUMBER] =
  {
    { BGO_SIGNAL        , SYNCHRONOUS,  SINGLE                   , ON_TRIGGER                }, // SYNCHRONOUS_SINGLE_BGO_SIGNAL
    { VME /*don't care*/, SYNCHRONOUS,  REPETITIVE               , ON_TRIGGER /*don't care*/ }, // SYNCHRONOUS_REPETITIVE
    { BGO_SIGNAL        , ASYNCHRONOUS, REPETITIVE /*don't care*/, ON_TRIGGER /*don't care*/ }, // ASYNCHRONOUS_BGO_SIGNAL
    { VME               , ASYNCHRONOUS, SINGLE                   , ON_TRIGGER                }, // ASYNCHRONOUS_VME_ON_TRIGGER
    { VME               , ASYNCHRONOUS, REPETITIVE /*don't care*/, WHEN_NOT_EMPTY            }  // ASYNCHRONOUS_VME_WHEN_NOT_EMPTY
  };

const std::string AltiModule::TTC_FIFO_NAME[TTC_FIFO_NUMBER] =
  { "BGO0_FIFO", "BGO1_FIFO", "BGO2_FIFO", "BGO3_FIFO", "TTYP_FIFO" };

const std::string AltiModule::TTC_FIFO_ACCESS_MODE_NAME[TTC_FIFO_ACCESS_MODE_NUMBER] =
  { "RUN_ACCESS", "VME_ACCESS" };

const std::string AltiModule::SIGNAL_PG_NAME[SIGNAL_PG_NUMBER] =
  { "ORB", "L1A", "TTR1", "TTR2", "TTR3", "TTYP", "BGO0", "BGO1", "BGO2", "BGO3" };

const std::string AltiModule::ASYNC_INPUT_SIGNAL_NAME[ASYNC_INPUT_SIGNAL_NUMBER] =
  {
    "SWX_ORB", "SWX_L1A", "SWX_BGO0", "SWX_BGO1", "SWX_BGO2", "SWX_BGO3", "SWX_TTYP0", "SWX_TTYP1", "SWX_TTYP2", "SWX_TTYP3", "SWX_TTYP4", "SWX_TTYP5", "SWX_TTYP6", "SWX_TTYP7", "SWX_TTR1", "SWX_TTR2", "SWX_TTR3",
    "LEMO_ORB", "LEMO_L1A", "LEMO_TTR1", "LEMO_TTR2", "LEMO_TTR3", "LEMO_BGO2", "LEMO_BGO3", "LEMO_BUSY", "RJ45_CALREQ0", "RJ45_CALREQ1", "RJ45_CALREQ2",
    "CTP_BUSY", "CTP_CALREQ0", "CTP_CALREQ1", "CTP_CALREQ2",
    "ALTI_BUSY", "ALTI_CALREQ0", "ALTI_CALREQ1", "ALTI_CALREQ2",
    "CDR_LOCK", "CDR_L1A", "CDR_BRCST", "CDR_INDIV", "CDR_BCR", "CDR_ECR", "CDR_SERR", "CDR_DERR"
  };

//------------------------------------------------------------------------------

// open VME
//RCD::VME *AltiModule::m_vme = RCD::VME::Open();

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

AltiModule::AltiModule(unsigned int slot) : m_vme(0), m_vmm(0), m_alti(0), m_segment(0), m_i2c(0), m_net(0), m_mem_pg_file(""), m_cnt_ttyp_lut_file(""), m_ds1wm(0), m_identifier("") {

    if (slot < ALTI_SLOT_MIN || slot > ALTI_SLOT_MAX) {
        std::printf("Invalid slot number!\n");
        m_status = -1;
        return;
    }
    else m_slot = slot;
    //m_base = base; // no check needed?
    //for (unsigned int idx = 0; idx < SIGNAL_NUMBER; idx++) std::printf("0x%08x\n", MEM_MASK[idx]);

    // open VME
    m_vme = RCD::VME::Open();

    // read MANUFACTURER ID
    unsigned int manufid;
    if ((m_status = CSRManufacturerIdRead(manufid)) != 0) {
        CERR("ALTI: reading MANUFACTURER ID from CSR in slot %2d", m_slot);
        return;
    }
    if (manufid != MANUF_ID) {
        CERR("ALTI: MANUFACTURER ID mismatch, read 0x%08x, expected 0x%08x", manufid, MANUF_ID);
        m_status = -1;
        return;
    }

    // read BOARD ID
    unsigned int boardid;
    if ((m_status = CSRBoardIdRead(boardid)) != 0) {
        CERR("ALTI: reading BOARD ID from CSR in slot %2d", m_slot);
        return;
    }
    if (boardid != BOARD_ID) {
        CERR("ALTI: BOARD ID mismatch, read 0x%08x, expected 0x%08x", boardid, BOARD_ID);
        m_status = -1;
        return;
    }

    // read REVISION ID
    unsigned int revid;
    if ((m_status = CSRRevisionIdRead(revid)) != 0) {
        CERR("ALTI: reading REVISION ID from CSR in slot %2d", m_slot);
        return;
    }
    /*if(revid != REV_ID) {
        CERR("ALTI: REVISION ID mismatch, read 0x%08x, expected 0x%08x", revid, REV_ID);
        m_status = -1;
        return;
    }*/
    unsigned int dd, mm, yy, revnum;
    AltiCommon::parseRevision(revid, dd, mm, yy, revnum);

    // read SLAVE CHARACTERISTICS ID
    unsigned int slcharactid;
    if ((m_status = CSRSlaveCharacteristicsRead(slcharactid)) != 0) {
        CERR("ALTI: reading SLAVE CHARACTERISTICS ID from CSR in slot %2d", m_slot);
        m_status = -1;
        return;
    }
    if(slcharactid != SLAVE_CHARACT_ID) {
        CERR("ALTI: SLAVE CHARACTERISTICS ID mismatch, read 0x%08x, expected 0x%08x", slcharactid, SLAVE_CHARACT_ID);
        return;
    }

    // read BAR
    unsigned int bar = 0;
    if ((m_status = CSRBARRead(bar)) != 0) {
        CERR("ALTI: reading BAR from CSR in slot %2d", m_slot);
        return;
    }
    if ((bar & 0x0000001f) != m_slot) {
        CERR("ALTI[slot %2d]: BAR mismatch, read 0x%02x <-> slot %2d, expected 0x%02x", m_slot, bar, (bar & 0x0000001f), m_slot);
        m_status = -1;
        return;
    }
    std::printf("ALTI: INFO - ALTI detected, revision = 0x%08x (#%d, %02d.%02d.%d.), BAR = 0x%08x <-> slot %d\n", revid, revnum, dd, mm, yy, bar, (bar & 0x0000001f));

    // read ADER0
    if ((m_status = CSRADER0Read(m_ader0)) != 0) {
        CERR("ALTI: reading ADER0 from CSR in slot %2d", m_slot);
        return;
    }
    m_base = swapEndian(m_ader0, ADER0_BASE_BITS);

    // create low-level ALTI
    m_alti = new ALTI(m_vme, m_base);
    if ((*m_alti)()) {
        CERR("ALTI: opening ALTI at vme = 0x%08x, size = 0x%08x, type = \"%s\"", m_base, ALTI::size(), ALTI::code().c_str());
        m_status = -1;
        return;
    }
    std::printf("ALTI: INFO - ALTI opened at vme = 0x%08x, size = 0x%08x, type = \"%s\"\n", m_base, ALTI::size(), ALTI::code().c_str());
    
    // fetch VME master mapping
    m_vmm = m_alti->getVmeMasterMap();

    //std::string serial, identifier;
    std::string serial;

    // create DS1WM, initialize and read serial number
    m_ds1wm = new DS1WM(DS1WM::ALTI, m_vmm);
    if ((m_status = m_ds1wm->detectSerialNumber(serial)) != SUCCESS) {
        m_identifier = "PRE_PRODUCTION_MODULE";
        m_status = SUCCESS;
    }
    else {
        DS1WM::getIdentifier(serial, m_identifier);
    }

    // create and initialize I2C network
    if ((m_status = I2CInit()) != SUCCESS) {
        CERR("ALTI: initializing I2C network", "");
        return;
    }

    // create CMEM segement
    std::string snam = "AltiModule";
    m_segment = new RCD::CMEMSegment(snam, sizeof(u_int)*MAX_MEMORY, true);
    COUT("opened CMEM segment \"%s\" of size 0x%08x, phys 0ix%08x, virt 0x%08x", snam.c_str(), m_segment->Size(), m_segment->PhysicalAddress(), m_segment->VirtualAddress());

    std::printf("--------------------------------------------------------------------------------\n");
    std::printf("ALTI[slot %2d]: INFO - ALTI \"%s\" opened\n", m_slot, m_identifier.c_str());
    std::printf("--------------------------------------------------------------------------------\n");

}

//------------------------------------------------------------------------------

AltiModule::~AltiModule() {

    // delete CMEM segment
    if (m_segment) delete m_segment;

    // delete I2C network
    if (m_net) delete m_net;

    // delete I2C bus
    if (m_i2c) delete m_i2c;

    // delete DS1WM
    if (m_ds1wm) delete m_ds1wm;

    // delete low-level ALTI
    if (m_alti) delete m_alti;

    // close VME
    RCD::VME::Close();

}

//------------------------------------------------------------------------------

int AltiModule::setBaseAddress(unsigned int slot, unsigned int base) {

    if (slot < ALTI_SLOT_MIN || slot > ALTI_SLOT_MAX) {
        std::printf("Invalid slot number!\n");
        return (FAILURE);
    }

    int status;

    // read BAR
    unsigned int bar = 0;
    if ((status = CSRBARRead(slot, bar)) != SUCCESS) {
        CERR("ALTI: reading BAR from CSR in slot %2d", slot);
        return (status);
    }
    if ((bar & 0x0000001f) != slot) {
        CERR("ALTI[slot %2d]: BAR mismatch, read 0x%02x <-> slot %2d, expected 0x%02x", slot, bar, (bar & 0x0000001f), slot);
        status = FAILURE;
        return (status);
    }
    std::printf("ALTI: INFO - ALTI detected, BAR = 0x%08x <-> slot %d\n", bar, (bar & 0x0000001f));

    // read BOARD ID
    unsigned int boardid = 0;
    if ((status = CSRBoardIdRead(slot, boardid)) != 0) {
        CERR("ALTI: reading BOARD ID from CSR in slot %2d", slot);
        return (status);
    }
    if (boardid != BOARD_ID) {
        CERR("ALTI: BOARD ID mismatch, read 0x%08x, expected 0x%08x", boardid, BOARD_ID);
        status = FAILURE;
        return (status);
    }

    // read ADER0
    unsigned int ader0 = 0;
    if ((status = CSRADER0Read(slot, ader0)) != SUCCESS) {
        CERR("ALTI: reading ADER0 from CSR in slot %2d", slot);
        return (status);
    }
    if (ader0 != swapEndian(base, ADER0_BASE_BITS)) {
        // write ADER0
        if ((status = CSRADER0Write(slot, swapEndian(base, ADER0_BASE_BITS))) != SUCCESS) {
            CERR("ALTI: writing ADER0 0x%08x to CSR in slot %2d", swapEndian(base, ADER0_BASE_BITS), slot);
            return (status);
        }
        // read ADER0
        if ((status = CSRADER0Read(slot, ader0)) != SUCCESS) {
            CERR("ALTI: reading ADER0 from CSR in slot %2d", slot);
            return (status);
        }
        if (ader0 != swapEndian(base, ADER0_BASE_BITS)) {
            CERR("ALTI: ADER0 not written to properly!", "");
            status = FAILURE;
            return (status);
        }
    }

    return (SUCCESS);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::AltiReset() {
    // right now, goes through the FW controls and resets them...
    // eventually, this should be done with the global resets through the CR/CSR (like in CtpcorePlusModule)


    // disabling pattern generation and snapshot
    if ((m_status = PATGenerationEnableWrite(false)) != SUCCESS) {
        CERR("ALTI: disabling pattern generation", "");
        return (m_status);
    }
    if ((m_status = SNPEnableWrite(false)) != SUCCESS) {
        CERR("ALTI: disabling snapshot", "");
        return (m_status);
    }
 
    // disabling TTC spy
    /*
    if ((m_status = DECEnableWrite(false)) != SUCCESS) {
        CERR("ALTI: disabling TTC decoder", "");
        return (m_status);
    }
    if ((m_status = DECReset()) != 0) {
        CERR("resetting TTC decoder snapshot memory","");
        return (m_status);
    }
    */
    /*// putting all TTC FIFOs in VME access mode
    if ((m_status = TTC_FIFOAccessModeWrite(TTC_FIFO_ACCESS_MODE::VME_ACCESS)) != SUCCESS) {
        CERR("ALTI: putting all TTC FIFOs in VME access mode", "");
        return (m_status);
    }*/
    // reset all TTC encoder FIFOs
    /*if ((m_status = ENCFifoReset(true, true)) != SUCCESS) {
        CERR("ALTI: enabling read/write pointer resets for all TTC FIFOs", "");
        return (m_status);
    }
    if ((m_status = ENCFifoReset(false, false)) != SUCCESS) {
        CERR("ALTI: disabling read/write pointer resets for all TTC FIFOs", "");
        return (m_status);
    }*/
    if ((m_status = ENCBgoFifoRetransmitWrite(false)) != SUCCESS) {
        CERR("ALTI: putting all TTC FIFOs out of retransmit mode", "");
        return (m_status);
    }
    if ((m_status = ENCFifoReset()) != SUCCESS) {
        CERR("ALTI: resetting read/write pointers for all TTC FIFOs", "");
        return (m_status);
    }

    // resetting the counters
    if ((m_status = CNTReset()) != SUCCESS) {
        CERR("ALTI: resetting all counters", "");
        return (m_status);
    }

    // resetting QuickBoot control
    if ((m_status = FMWControlReset()) != SUCCESS) {
        CERR("ALTI: resetting QuickBoot control", "");
        return (m_status);
    }
  
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::AltiSetup() {
  
  // set up jitter cleaner
  if ((m_status = JCSetup()) != SUCCESS) {
    CERR("ALTI: setting up jitter cleaner", "");
    return (m_status);
  }
  
  // set up CLK multiplexer
  if ((m_status = CLKSetup()) != SUCCESS) {
    CERR("ALTI: setting up CLK", "");
    return (m_status);
  }
  
  // set up I2C
  if ((m_status = I2CSetup()) != SUCCESS) {
    CERR("ALTI: setting up I2C", "");
    return (m_status);
  }
  
  // set up ADER1 page register in CSR
  if ((m_status = CSRSetup()) != SUCCESS) {
    CERR("ALTI: setting up ADER1 page register in CR/CSR", "");
    return (m_status);
  }
  
  // set up the TTC encoder
  if ((m_status = ENCSetup()) != SUCCESS) {
    CERR("ALTI: setting up the TTC encoder", "");
    return (m_status);
  }
  
  return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::AltiCheck() {

    bool ok(true);

    // check CLK multiplexer
    if ((m_status = CLKCheck()) != SUCCESS) {
        CERR("ALTI: checking CLK mux", "");
        ok = false;
    }
    else {
        COUT("ALTI: INFO - CLK mux                         = \"SETUP\"", "");
    }

    // check I2C core
    if ((m_status = I2CCheck()) != SUCCESS) {
        CERR("ALTI: checking I2C core", "");
        ok = false;
    }
    else {
        COUT("ALTI: INFO - I2C core                        = \"SETUP\"", "");
    }

    // check PLL lock
    //bool clklos, clklol, clklol2;
    bool monitor_int, clklol, clklol2, sticky_clklol, sticky_clklol2, sticky_clklos;
    //if ((m_status = CLKStatusReadMMCM(clklos, clklol, clklol2)) != SUCCESS) {
    if ((m_status = CLKStatusReadMMCM(monitor_int, clklol, clklol2, sticky_clklol, sticky_clklol2, sticky_clklos)) != SUCCESS) {
        CERR("ALTI: checking PLL lock", "");
        ok = false;
    }
    else if (!clklol) {
        COUT("ALTI: INFO - PLL                             = \"LOCKED\"", "");
    }
    else {
        COUT("ALTI: INFO - PLL                             = \"NOT LOCKED\" !!!!", "");
        ok = false;
    }

    // check jitter cleaner
    std::string designid;
    if ((m_status = CLKJitterCleanerDesignIDRead(designid)) != SUCCESS) {
        CERR("ALTI: checking jitter cleaner design ID", "");
        ok = false;
    }
    else if (designid == "ALTI_001") {
        COUT("ALTI: INFO - jitter cleaner design           = \"%s\" (DEFAULT)", designid.c_str());
    }
    else {
        COUT("ALTI: INFO - jitter cleaner design           = \"%s\" (NON-DEFAULT !!!!)", designid.c_str());
    }
    std::vector<bool> oof, los;
    bool hold, lol;
    if ((m_status = CLKJitterCleanerStatusRead(oof, los, hold, lol)) != SUCCESS) {
        CERR("ALTI: checking jitter cleaner status", "");
        ok = false;
    }
    else if (hold == true) {
        COUT("ALTI: INFO - jitter cleaner mode             = \"HOLDOVER\" (NON-DEFAULT !!!!)", "");
        ok = false;
    }

    // check ADER1 paging register
    unsigned int ader1;
    if ((m_status = CSRADER1Read(ader1)) != SUCCESS) {
        CERR("ALTI: checking ADER1 page register", "");
        ok = false;
    }
    else if ((ader1 > 0) && (ader1 <= ALTI_RAM_MEMORY_NUMBER)) {
        COUT("ALTI: INFO - ADER1 page register             = \"SETUP\"", "");
    }
    else {
        COUT("ALTI: INFO - ADER1 page register             = \"NOT SETUP\" (ALTI uses 16MB instead of 8MB) !!!!", "");
        ok = false;
    }
    
    // check the TTC encoder
    if ((m_status = ENCCheck()) != SUCCESS) {
        CERR("ALTI: checking the TTC encoder", "");
        ok = false;
    }
    else {
        COUT("ALTI: INFO - TTC encoder                     = \"SETUP\"", "");
    }
    
    return(ok?SUCCESS:FAILURE);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::CFGRead(AltiConfiguration &cfg, std::ostream &os) {

    char s[STRING_LENGTH];

    // CLK
    if ((m_status = CFGReadCLK(cfg, os)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading CLK configuration"); MERR(os, s);
        return (m_status);
    }
    
    // SIG
    if ((m_status = CFGReadSIG(cfg, os)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading SIG configuration"); MERR(os, s);
        return (m_status);
    }
    
    // BSY
    if ((m_status = CFGReadBSY(cfg, os)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading BSY configuration"); MERR(os, s);
        return (m_status);
    }

    // CRQ
    if ((m_status = CFGReadCRQ(cfg, os)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading CRQ configuration"); MERR(os, s);
        return (m_status);
    }

    // MEM
    if ((m_status = CFGReadMEM(cfg, os)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading MEM configuration"); MERR(os, s);
        return (m_status);
    }

    // TTC
    if ((m_status = CFGReadTTC(cfg, os)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading TTC configuration"); MERR(os, s);
        return (m_status);
    }

    // CNT
    if ((m_status = CFGReadCNT(cfg, os)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading CNT configuration"); MERR(os, s);
        return (m_status);
    }

    // CTP
    if ((m_status = CFGReadCTP(cfg, os)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading CTP configuration"); MERR(os, s);
        return (m_status);
    }
    
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CFGWrite(const AltiConfiguration &cfg, std::ostream &os) {

    char s[STRING_LENGTH];
    
    // check the clock 
    bool monitor_int, clk_lol, clk_lol2, sticky_clk_lol, sticky_clk_lol2, sticky_clk_los;
    if ((m_status = CLKStatusReadMMCM(monitor_int, clk_lol, clk_lol2, sticky_clk_lol, sticky_clk_lol2, sticky_clk_los)) != SUCCESS) {
      std::sprintf(s, "ALTI: reading CLK MMCM status"); MERR(os, s);
      return (m_status);
    } 
    if (clk_lol) {
      // set internal clock
      if ((m_status = CLKInputSelectWriteMMCM(PRIMARY_FXO)) != SUCCESS) { 
	std::sprintf(s, "ALTI: setting internal CLK "); MERR(os, s);
	return (m_status);
      }
      // wait for PLL to lock
      usleep(1);
    }
    
    // SIG
    if ((m_status = CFGWriteSIG(cfg, os)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing SIG configuration"); MERR(os, s);
        return (m_status);
    }

    // BSY
    if ((m_status = CFGWriteBSY(cfg, os)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing BSY configuration"); MERR(os, s);
        return (m_status);
    }

    // CRQ
    if ((m_status = CFGWriteCRQ(cfg, os)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing CRQ configuration"); MERR(os, s);
        return (m_status);
    }

    // MEM
    if ((m_status = CFGWriteMEM(cfg, os)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing MEM configuration"); MERR(os, s);
        return (m_status);
    }

    // TTC
    if ((m_status = CFGWriteTTC(cfg, os)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing TTC configuration"); MERR(os, s);
        return (m_status);
    }

    // CNT
    if ((m_status = CFGWriteCNT(cfg, os)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing CNT configuration"); MERR(os, s);
        return (m_status);
    }

    // CTP
    if ((m_status = CFGWriteCTP(cfg, os)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing CTP configuration"); MERR(os, s);
        return (m_status);
    }

    // CLK confiuration trick (read from config file first, config at the end so PLL won't lose lock)

    // CLK
    if ((m_status = CFGWriteCLK(cfg, os)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing CLK configuration"); MERR(os, s);
        return (m_status);
    }

    
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CFGReadCLK(AltiConfiguration &cfg, std::ostream &os) {

    if (!cfg.clk_config) return (m_status);

    char s[STRING_LENGTH];

    if ((m_status = CLKInputSelectReadMMCM(cfg.clk_select)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading CLK select configuration"); MERR(os, s);
        return (m_status);
    }

    if ((m_status = CLKInputSelectReadJitterCleaner(cfg.clk_jc_select)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading CLK jitter cleaner select configuration"); MERR(os, s);
        return (m_status);
    }

    if ((m_status = CLKPhasePositionReadMMCM(cfg.clk_phase_shift)) != SUCCESS) {
      std::sprintf(s, "ALTI: reading CLK output phase shift"); MERR(os, s);
      return (m_status); 
    }
    
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CFGReadSIG(AltiConfiguration &cfg, std::ostream &os) {

    if (!cfg.sig_config) return (m_status);

    char s[STRING_LENGTH];

    if ((m_status = SIGMuxBGOOutputLEMORead(cfg.sig_fp_bgo2enable, cfg.sig_fp_bgo3enable)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading SIG:BGO configuration"); MERR(os, s);
        return (m_status);
    }

    unsigned i, j;
    for (i = 0; i < SIGNAL_NUMBER_ROUTED; i++) {
        for (j = 0; j < SIGNAL_DESTINATION_NUMBER; j++) {
            if ((m_status = SIGSwitchConfigRead(SIGNAL(i), SIGNAL_DESTINATION(j), cfg.sig_swx[i][j])) != SUCCESS) {
                std::sprintf(s, "ALTI: reading SIG:SWX configuration, signal %s, destination %s", SIGNAL_NAME[i].c_str(), SIGNAL_DESTINATION_NAME[j].c_str()); MERR(os, s);
                return (m_status);
            }
        }
    }
    cfg.readSWXIndividualToBundleTTYP();
    
    if ((m_status = SIGEqualizersConfigRead(CTP_IN, cfg.sig_eqz_ctp_in_enable, cfg.sig_eqz_ctp_in_mode)) != SUCCESS) {
      std::sprintf(s, "ALTI: reading SIG:EQZ configuration for %s", SIGNAL_SOURCE_NAME[CTP_IN].c_str()); MERR(os, s);
      return (m_status);
    }
    if ((m_status = SIGEqualizersConfigRead(ALTI_IN, cfg.sig_eqz_alti_in_enable, cfg.sig_eqz_alti_in_mode)) != SUCCESS) {
      std::sprintf(s, "ALTI: reading SIG:EQZ configuration for %s", SIGNAL_SOURCE_NAME[ALTI_IN].c_str()); MERR(os, s);
      return (m_status);
    }
    
    if ((m_status = SIGInputSyncEnableRead(cfg.sig_io_sync_enable, cfg.sig_io_sync_phase)) != SUCCESS) {
      std::sprintf(s, "ALTI: reading SIG:IO sync configuration"); MERR(os, s);
      return (m_status);
    }
    if ((m_status = SIGInputShapeEnableRead(cfg.sig_io_shape_enable)) != SUCCESS) {
      std::sprintf(s, "ALTI: reading SIG:IO shape configuration"); MERR(os, s);
      return (m_status);
    }
   
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CFGReadBSY(AltiConfiguration &cfg, std::ostream &os) {

    if (!cfg.bsy_config) return (m_status);

    char s[STRING_LENGTH];
    
    if ((m_status = BSYInputSelectRead(cfg.bsy_in_select)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading BSY input select configuration"); MERR(os, s);
        return (m_status);
    }
    
    if ((m_status = BSYOutputSelectRead(cfg.bsy_out_select)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading BSY output select configuration"); MERR(os, s);
        return (m_status);
    }
    
    if ((m_status = BSYConstLevelRegRead(cfg.bsy_data_vme)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading BSY const level register configuration"); MERR(os, s);
        return (m_status);
    }
    
    if ((m_status = BSYMaskingL1aEnableRead(cfg.bsy_masking_l1a_enable)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading BSY masking of L1A configuration"); MERR(os, s);
        return (m_status);
    }
    
    if ((m_status = BSYFrontPanelLevelRead(cfg.bsy_level)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading BSY Front-Panel level configuration"); MERR(os, s);
        return (m_status);
    }
    
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CFGReadCRQ(AltiConfiguration &cfg, std::ostream &os) {

    if (!cfg.crq_config) return (m_status);

    char s[STRING_LENGTH];

    if ((m_status =CRQInputSelectRead(cfg.crq_in_select)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading CRQ input select configuration"); MERR(os, s);
        return (m_status);
    }

    if ((m_status = CRQOutputSelectRead(CALREQ_TO_CTP, cfg.crq_out_select[0])) != SUCCESS) {
        std::sprintf(s, "ALTI: reading CRQ CTP output select configuration"); MERR(os, s);
        return (m_status);
    }

    if ((m_status = CRQOutputSelectRead(CALREQ_TO_ALTI, cfg.crq_out_select[1])) != SUCCESS) {
        std::sprintf(s, "ALTI: reading CRQ ALTI output select configuration"); MERR(os, s);
        return (m_status);
    }

    if ((m_status = CRQConstLevelRegRead(cfg.crq_data_vme)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading CRQ const level register configuration"); MERR(os, s);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CFGReadMEM(AltiConfiguration &cfg, std::ostream &os) {

    if (!cfg.mem_config) return (m_status);

    char s[STRING_LENGTH];

    // read PAT generation enable
    if ((m_status = PATGenerationEnableRead(cfg.mem_pg_enable)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading PAT generation enable"); MERR(os, s);
        return (m_status);
    }

    // read PAT generation configuration
    if ((m_status = PATGenerationConfigRead(cfg.mem_pg_config)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading PAT generation configuration"); MERR(os, s);
        return (m_status);
    }

    // read PAT generation repeat
    if ((m_status = PATGenerationRepeatRead(cfg.mem_pg_repeat)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading PAT generation repeat configuration"); MERR(os, s);
        return (m_status);
    }

    // read PAT generation start address
    if ((m_status = PATGenerationStartAddressRead(cfg.mem_pg_start_addr)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading PAT generation start address"); MERR(os, s);
        return (m_status);
    }

    // read PAT generation stop address
    if ((m_status = PATGenerationStopAddressRead(cfg.mem_pg_stop_addr)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading PAT generation stop address"); MERR(os, s);
        return (m_status);
    }

    cfg.mem_pg_file = m_mem_pg_file;

    /*// read SNP enable
    if ((m_status = SNPEnableRead(cfg.mem_snap_enable)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading SNP enable configuration"); MERR(os, s);
        return (m_status);
    }*/

    // read SNP mask
    if ((m_status = SNPMaskRead(cfg.mem_snap_masked)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading SNP mask"); MERR(os, s);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CFGReadTTC(AltiConfiguration &cfg, std::ostream &os) {

    if (!cfg.ttc_config) return (m_status);

    char s[STRING_LENGTH];

    std::vector<bool> enabled;
    enabled.resize(TRANSMITTER_NUMBER);
    //std::vector<bool> fault;
    //fault.resize(TRANSMITTER_NUMBER);

    /*if ((m_status = TTCTransmitterStatusRead(enabled, fault)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading TTC:TX configuration"); MERR(os, s);
        return (m_status);
    }*/
    if ((m_status = ENCTransmitterEnableRead(enabled)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading TTC:TX enable configuration"); MERR(os, s);
        return (m_status);
    }

    unsigned int i;
    for (i = 0; i < TRANSMITTER_NUMBER; i++) {
        cfg.ttc_tx_enable[i] = enabled[i];
    }

    // L1A source
    if ((m_status = ENCL1ASourceRead(cfg.ttc_l1a_source)) != SUCCESS) {
      std::sprintf(s, "ALTI: reading TTC L1A source"); MERR(os, s);
      return (m_status);
    }

    if ((m_status = ENCInhibitParamsRead(cfg.ttc_bgo_inhibit_width, cfg.ttc_bgo_inhibit_delay)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading TTC BGO inhibit width and delay configuration"); MERR(os, s);
        return (m_status);
    }
    if ((m_status = ENCTriggerWordDelayRead(cfg.ttc_ttyp_delay)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading TTC TTYP word delay configuration"); MERR(os, s);
        return (m_status);
    }
    if ((m_status = ENCTriggerWordCycleControlRead(cfg.ttc_ttyp_addr, cfg.ttc_ttyp_frame_type, cfg.ttc_ttyp_address_space, cfg.ttc_ttyp_subaddr)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading TTC TTYP control configuration"); MERR(os, s);
        return (m_status);
    }
    if ((m_status = ENCBgoCommandModeRead(cfg.ttc_bgo_mode)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading TTC BGO mode configuration"); MERR(os, s);
        return (m_status);
    }
    bool bgo_busy[4];
    if ((m_status = BSYMaskingBGoEnableRead(bgo_busy[0], 
					    bgo_busy[1], 
					    bgo_busy[2], 
					    bgo_busy[3])) != SUCCESS) {
      std::sprintf(s, "ALTI: reading TTC BGO busy gate configuration"); MERR(os, s);
      return (m_status);
    }
    for(i = 0; i < 3; i++) { 
      cfg.ttc_bgo_busy_gate[i] = bgo_busy[i];
    }      
    if ((m_status = ENCBgoFifoRetransmitRead(cfg.ttc_fifo_retransmit)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading TTC FIFOs retransmit configuration"); MERR(os, s);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CFGReadCNT(AltiConfiguration &cfg, std::ostream &os) {

    if (!cfg.cnt_config) return (m_status);

    char s[STRING_LENGTH];

    std::vector<bool> enabled;
    enabled.resize(AltiModule::CNT_NUMBER);
    std::vector<bool> reset;
    reset.resize(AltiModule::CNT_NUMBER);

    if ((m_status = CNTStatusRead(enabled, reset)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading CNT configuration"); MERR(os, s);
        return (m_status);
    }

    unsigned int i;
    for (i = 0; i < CNT_NUMBER; i++) {
        cfg.cnt_enable[i] = enabled[i];
    }

    cfg.cnt_ttyp_lut_file = m_cnt_ttyp_lut_file;

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CFGReadCTP(AltiConfiguration &cfg, std::ostream &os) {

    if (!cfg.ctp_config) return (m_status);

    char s[STRING_LENGTH];

    // read ECR generation
    if ((m_status = ECRGenerationRead(cfg.ctp_ecr_generation)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading ECR generation control"); MERR(os, s);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CFGWriteCLK(const AltiConfiguration &cfg, std::ostream &os) {

    if (!cfg.clk_config) return (m_status);

    char s[STRING_LENGTH];
    
    // Write CLK JC configuration only if cfg.clk_jc_file is not empty
    if(cfg.clk_jc_file == "DEFAULT") {
      if ((m_status = CLKJitterCleanerConfigWrite()) != SUCCESS) {
        std::sprintf(s, "ALTI: writing CLK jitter cleaner default configuration"); MERR(os, s);
        return (m_status);
      }
    }
    else if (cfg.clk_jc_file != "") {
      if((m_status = CLKJitterCleanerConfigWrite(cfg.clk_jc_file)) != SUCCESS) {
	std::sprintf(s, "ALTI: writing CLK jitter cleaner configuration from file %s", cfg.clk_jc_file.c_str()); MERR(os, s);
	return (m_status);
      }
    }
    
    // change PLL input clock only if needed 
    CLK_TYPE clk_input;
    if ((m_status = CLKInputSelectReadMMCM(clk_input)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading CLK select configuration"); MERR(os, s);
        return (m_status);
    }
    if(clk_input != cfg.clk_select) {
      if ((m_status = CLKInputSelectWriteMMCM(cfg.clk_select)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing CLK select configuration"); MERR(os, s);
        return (m_status);
      }
    }
    
    // change Jitter Cleaner input clock only if needed 
    if ((m_status = CLKInputSelectReadJitterCleaner(clk_input)) != SUCCESS) {
        std::sprintf(s, "ALTI: reading CLK jitter cleaner select configuration"); MERR(os, s);
        return (m_status);
    }
    if(clk_input != cfg.clk_jc_select) {    
      if ((m_status = CLKInputSelectWriteJitterCleaner(cfg.clk_jc_select)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing CLK jitter cleaner select configuration"); MERR(os, s);
        return (m_status);
      }
    }
    
    const bool increment = (cfg.clk_phase_shift > 0)? true : false;  
    const u_int phase = abs(cfg.clk_phase_shift);
    if ((m_status = CLKPhaseResetMMCM()) != SUCCESS) {
      std::sprintf(s, "ALTI: resetting CLK phase shift"); MERR(os, s);
      return (m_status);
    }
    
    if ((m_status = CLKPhaseShiftMMCM(increment, phase)) !=  AltiModule::SUCCESS) {
      std::sprintf(s, "ALTI: writing CLK phase shift"); MERR(os, s);
      return (m_status);
    }

    if ((m_status = CLKStatusClearMMCM()) != SUCCESS) {
      std::sprintf(s, "ALTI: clearing CLK MMCM status"); MERR(os, s);
      return (m_status);
    }

    if ((m_status = CLKJitterCleanerStatusClear()) != SUCCESS) {
      std::sprintf(s, "ALTI: clearing CLK jitter cleaner status"); MERR(os, s);
      return (m_status);
    }
    
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CFGWriteSIG(const AltiConfiguration &cfg, std::ostream &os) {

    if (!cfg.sig_config) return (m_status);

    char s[STRING_LENGTH];

    if ((m_status = SIGMuxBGOOutputLEMOWrite(cfg.sig_fp_bgo2enable, cfg.sig_fp_bgo3enable)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing SIG:BGO configuration"); MERR(os, s);
        return (m_status);
    }

    unsigned i, j;
    //unsigned int voltage[EQUALIZER_VOLTAGE_NUMBER];
    for (i = 0; i < SIGNAL_NUMBER_ROUTED; i++) {
        for (j = 0; j < SIGNAL_DESTINATION_NUMBER; j++) {
            if ((SIGNAL_NAME[i]).rfind("TTYP", 0) == 0) {
                if ((m_status = SIGSwitchConfigWrite(SIGNAL(i), SIGNAL_DESTINATION(j), cfg.sig_swx_ttyp[j])) != SUCCESS) {
                    std::sprintf(s, "ALTI: writing SIG configuration, signal %s, destination %s", SIGNAL_NAME[i].c_str(), SIGNAL_DESTINATION_NAME[j].c_str()); MERR(os, s);
                    return (m_status);
                }
            }
            else {
                if ((m_status = SIGSwitchConfigWrite(SIGNAL(i), SIGNAL_DESTINATION(j), cfg.sig_swx[i][j])) != SUCCESS) {
                    std::sprintf(s, "ALTI: writing SIG configuration, signal %s, destination %s", SIGNAL_NAME[i].c_str(), SIGNAL_DESTINATION_NAME[j].c_str()); MERR(os, s);
                    return (m_status);
                }
            }
        }
    }

  
    if ((m_status = SIGEqualizersConfigWrite(CTP_IN,cfg.sig_eqz_ctp_in_enable, cfg.sig_eqz_ctp_in_mode)) != SUCCESS) {
      std::sprintf(s, "ALTI: writing SIG:EQZ configuration for %s", SIGNAL_SOURCE_NAME[CTP_IN].c_str()); MERR(os, s);
      return (m_status);
    }
    if ((m_status = SIGEqualizersConfigWrite(ALTI_IN,cfg.sig_eqz_alti_in_enable, cfg.sig_eqz_alti_in_mode)) != SUCCESS) {
      std::sprintf(s, "ALTI: writing SIG:EQZ configuration for %s", SIGNAL_SOURCE_NAME[ALTI_IN].c_str()); MERR(os, s);
      return (m_status);
    }

    if ((m_status = SIGInputSyncEnableWrite(cfg.sig_io_sync_enable, cfg.sig_io_sync_phase)) != SUCCESS) {
      std::sprintf(s, "ALTI: writing SIG:IO sync configuration"); MERR(os, s);
      return (m_status);
    }
    
    if ((m_status = SIGInputShapeEnableWrite(cfg.sig_io_shape_enable)) != SUCCESS) {
      std::sprintf(s, "ALTI: writing SIG:IO shape configuration"); MERR(os, s);
      return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CFGWriteBSY(const AltiConfiguration &cfg, std::ostream &os) {

    if (!cfg.bsy_config) return (m_status);

    char s[STRING_LENGTH];

    if ((m_status = BSYInputSelectWrite(cfg.bsy_in_select)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing BSY input select configuration"); MERR(os, s);
        return (m_status);
    }

    if ((m_status = BSYOutputSelectWrite(cfg.bsy_out_select)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing BSY output select configuration"); MERR(os, s);
        return (m_status);
    }

    if ((m_status = BSYConstLevelRegWrite(cfg.bsy_data_vme)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing BSY const level register configuration"); MERR(os, s);
        return (m_status);
    }

    if ((m_status = BSYMaskingL1aEnableWrite(cfg.bsy_masking_l1a_enable)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing BSY masking of L1A configuration"); MERR(os, s);
        return (m_status);
    }

   if ((m_status = BSYFrontPanelLevelWrite(cfg.bsy_level)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing BSY Front-Panel level configuration"); MERR(os, s);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CFGWriteCRQ(const AltiConfiguration &cfg, std::ostream &os) {

    if (!cfg.crq_config) return (m_status);

    char s[STRING_LENGTH];

    if ((m_status = CRQInputSelectWrite(cfg.crq_in_select)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing CRQ input select configuration"); MERR(os, s);
        return (m_status);
    }

    if ((m_status = CRQOutputSelectWrite(CALREQ_TO_CTP, cfg.crq_out_select[0])) != SUCCESS) {
        std::sprintf(s, "ALTI: writing CRQ CTP output select configuration"); MERR(os, s);
        return (m_status);
    }

    if ((m_status = CRQOutputSelectWrite(CALREQ_TO_ALTI, cfg.crq_out_select[1])) != SUCCESS) {
        std::sprintf(s, "ALTI: writing CRQ ALTI output select configuration"); MERR(os, s);
        return (m_status);
    }

    if ((m_status = CRQConstLevelRegWrite(cfg.crq_data_vme)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing CRQ const level register configuration"); MERR(os, s);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CFGWriteMEM(const AltiConfiguration &cfg, std::ostream &os) {

    if (!cfg.mem_config) return (m_status);

    char s[STRING_LENGTH];

    // disabke MEM snapshot
    if ((m_status = SNPEnableWrite(false)) != SUCCESS) {
        std::sprintf(s, "ALTI: disabling MEM snapshot"); MERR(os, s);
        return (m_status);
    }

    // disable MEM pattern generation
    if ((m_status = PATGenerationEnableWrite(false)) != SUCCESS) {
        std::sprintf(s, "ALTI: disabling MEM pattern generation"); MERR(os, s);
        return (m_status);
    }

    // write MEM pattern generation control configuration
    if ((m_status = PATGenerationRepeatWrite(cfg.mem_pg_repeat)) != SUCCESS) {
        std::sprintf(s, "ALTI: configuring MEM pattern generation"); MERR(os, s);
        return (m_status);
    }

    // write MEM pattern generation start address
    if ((m_status = PATGenerationStartAddressWrite(cfg.mem_pg_start_addr)) != SUCCESS) {
        std::sprintf(s, "ALTI: configuring MEM pattern generation start address"); MERR(os, s);
        return (m_status);
    }

    // write MEM pattern generation stop address
    if ((m_status = PATGenerationStopAddressWrite(cfg.mem_pg_stop_addr)) != SUCCESS) {
        std::sprintf(s, "ALTI: configuring MEM pattern generation stop address"); MERR(os, s);
        return (m_status);
    }

    // write MEM pattern generation file into memory
    std::ifstream inf(cfg.mem_pg_file.c_str(), std::ifstream::in);
    if (inf) {
        if ((m_status = PATWriteFile(m_segment, inf)) != SUCCESS) {
            std::sprintf(s, "ALTI: writing MEM pattern generation file into memory"); MERR(os, s);
            return (m_status);
        }
    }
    m_mem_pg_file = cfg.mem_pg_file;

    // write MEM pattern configuration
    if ((m_status = PATGenerationConfigWrite(cfg.mem_pg_config)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing MEM pattern generation configuration"); MERR(os, s);
        return (m_status);
    }

    // write MEM pattern generation enable
    if ((m_status = PATGenerationEnableWrite(cfg.mem_pg_enable)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing MEM pattern generation enable"); MERR(os, s);
        return (m_status);
    }

    /*// write MEM snapshot control configuration
    if ((m_status = SNPEnableWrite(cfg.mem_snap_enable)) != SUCCESS) {
        std::sprintf(s, "ALTI: configuring MEM snapshot"); MERR(os, s);
        return (m_status);
    }*/

    // write SNP mask
    if ((m_status = SNPMaskWrite(cfg.mem_snap_masked)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing SNP mask"); MERR(os, s);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CFGWriteTTC(const AltiConfiguration &cfg, std::ostream &os) {

    if (!cfg.ttc_config) return (m_status);

    char s[STRING_LENGTH];

    unsigned int i;
    for (i = 0; i < TRANSMITTER_NUMBER; i++) {
        if ((m_status = ENCTransmitterEnableWrite(i, cfg.ttc_tx_enable[i])) != SUCCESS) {
            std::sprintf(s, "ALTI: enabling TTC transmitter TX[%02d]", i); MERR(os, s);
            return (m_status);
        }
    }

    // L1A source
    if ((m_status = ENCL1ASourceWrite(cfg.ttc_l1a_source)) != SUCCESS) {
      std::sprintf(s, "ALTI: configuring TTC L1A source"); MERR(os, s);
      return (m_status);
    }

    // write TTC transmitter configuration
    if ((m_status = ENCInhibitParamsWrite(cfg.ttc_bgo_inhibit_width, cfg.ttc_bgo_inhibit_delay)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing TTC BGO inhibit width and delay configuration"); MERR(os, s);
        return (m_status);
    }
    if ((m_status = ENCTriggerWordDelayWrite(cfg.ttc_ttyp_delay)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing TTC TTYP word delay configuration"); MERR(os, s);
        return (m_status);
    }
    if ((m_status = ENCTriggerWordCycleControlWrite(cfg.ttc_ttyp_addr, cfg.ttc_ttyp_frame_type, cfg.ttc_ttyp_address_space, cfg.ttc_ttyp_subaddr)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing TTC TTYP control configuration"); MERR(os, s);
        return (m_status);
    }
    if ((m_status = ENCBgoCommandModeWrite(cfg.ttc_bgo_mode)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing TTC BGO mode configuration"); MERR(os, s);
        return (m_status);
    }
    for (i = 0; i < 3; i++) {
      if ((m_status = BSYMaskingBGoEnableWrite(i, cfg.ttc_bgo_busy_gate[i])) != AltiModule::SUCCESS) {
	std::sprintf(s, "ALTI: enabling busy masking of TTC BGO commands"); MERR(os, s);
	return (m_status);
      }
    }
    if ((m_status = ENCBgoFifoRetransmitWrite(cfg.ttc_fifo_retransmit)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing TTC FIFOs retransmit configuration"); MERR(os, s);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CFGWriteCNT(const AltiConfiguration &cfg, std::ostream &os) {

    if (!cfg.cnt_config) return (m_status);

    char s[STRING_LENGTH];

    unsigned int i;
    for (i = 0; i < CNT_NUMBER; i++) {
        if ((m_status = CNTEnable((CNT_TYPE) i, cfg.cnt_enable[i])) != SUCCESS) {
            std::sprintf(s, "ALTI: configuring CNT enable"); MERR(os, s);
            return (m_status);
        }
    }

    // write CNT TTYP LUT file into memory
    std::ifstream inf(cfg.cnt_ttyp_lut_file.c_str(), std::ifstream::in);
    if (inf) {
        if ((m_status = CNTTtypLutWrite(cfg.cnt_ttyp_lut_file)) != SUCCESS) {
            std::sprintf(s, "ALTI: writing CNT TTYP LUT file into memory"); MERR(os, s);
            return (m_status);
        }
    }
    m_cnt_ttyp_lut_file = cfg.cnt_ttyp_lut_file;

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CFGWriteCTP(const AltiConfiguration &cfg, std::ostream &os) {

    if (!cfg.ctp_config) return (m_status);

    char s[STRING_LENGTH];

    // write ECR generation
    if ((m_status = ECRGenerationWrite(cfg.ctp_ecr_generation)) != SUCCESS) {
        std::sprintf(s, "ALTI: writing ECR generation control"); MERR(os, s);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::CSRManufacturerIdRead(unsigned int &manuf) {

    // manufacturer identification
    if ((m_status = m_vme->ReadCRCSR(m_slot, MANUF_ID_BASE, &manuf)) != 0) {
        CERR("ALTI: reading MANUFACTURER ID from CR/CSR", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CSRBoardIdRead(unsigned int &board) {

    // board identification
    if ((m_status = m_vme->ReadCRCSR(m_slot, BOARD_ID_BASE, &board)) != 0) {
        CERR("ALTI: reading BOARD ID from CR/CSR", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CSRRevisionIdRead(unsigned int &rev) {

    // revision identification
    if ((m_status = m_vme->ReadCRCSR(m_slot, REV_ID_BASE, &rev)) != 0) {
        CERR("ALTI: reading REVISION ID from CR/CSR", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CSRSlaveCharacteristicsRead(unsigned int &slcharact) {

    // slave characteristics identification
    if ((m_status = m_vme->ReadCRCSR(m_slot, SLAVE_CHARACT_ID_BASE, &slcharact)) != 0) {
        CERR("ALTI: reading SLAVE CHARACTERISTICS ID from CR/CSR", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CSRBARRead(unsigned int &bar) {

    // BAR
    if ((m_status = m_vme->ReadCRCSR(m_slot, BAR_BASE, &bar)) != 0) {
        CERR("ALTI: reading BAR from CR/CSR", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CSRADER0Read(unsigned int &ader0) {

    // ADER0
    if ((m_status = m_vme->ReadCRCSR(m_slot, ADER0_BASE, &ader0)) != 0) {
        CERR("ALTI: reading ADER0 from CR/CSR", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CSRADER0Write(const unsigned int ader0) {
    if (m_ader0 == ader0) { // nothing to do
        m_status = SUCCESS;
        return (m_status);
    }

    if (m_net) {
        // delete I2C network
        delete m_net;
    }

    if (m_i2c) {
        // delete I2C
        delete m_i2c;
    }

    if (m_ds1wm) {
        // delete DS1WM
        delete m_ds1wm;
    }

    if (m_alti) {
        // delete low-level ALTI
        delete m_alti;
    }

    // ADER0
    if ((m_status = m_vme->WriteCRCSR(m_slot, ADER0_BASE, ader0)) != 0) {
        CERR("ALTI: writing ADER0 to CR/CSR","");
        return (m_status);
    }
    m_ader0 = ader0;
    m_base = swapEndian(m_ader0, ADER0_BASE_BITS);

    // create low-level ALTI
    m_alti = new ALTI(m_vme, m_base);
    if ((*m_alti)()) {
        CERR("ALTI: opening ALTI at vme = 0x%08x, size = 0x%08x, type = \"%s\"", m_base, ALTI::size(), ALTI::code().c_str());
        m_status = FAILURE;
        return (m_status);
    }
    std::printf("ALTI: INFO - ALTI opened at vme = 0x%08x, size = 0x%08x, type = \"%s\"\n", m_base, ALTI::size(), ALTI::code().c_str());

    // fetch VME master mapping
    m_vmm = m_alti->getVmeMasterMap();

    // create DS1WM, initialize and read serial number
    m_ds1wm = new DS1WM(DS1WM::ALTI, m_vmm);

    // create and initialize I2C network
    if ((m_status = I2CInit()) != SUCCESS) {
        CERR("ALTI: initializing I2C network", "");
        m_status = FAILURE;
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CSRADER1Read(unsigned int &ader1) {

    // ADER1
    if ((m_status = m_vme->ReadCRCSR(m_slot, ADER1_BASE, &ader1)) != 0) {
        CERR("ALTI: reading ADER1 from CR/CSR", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CSRADER1Write(const unsigned int ader1) {

    // ADER1
    if ((m_status = m_vme->WriteCRCSR(m_slot, ADER1_BASE, ader1)) != 0) {
        CERR("ALTI: writing ADER1 0x%08x to CR/CSR", ader1);
        return (m_status);
    }
    m_ader1 = ader1;

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CSRBoardIdRead(int slot, unsigned int &board) {
    // open VME
    RCD::VME *vme = RCD::VME::Open();

    // board identification
    int status(SUCCESS);
    if ((status = vme->ReadCRCSR(slot, BOARD_ID_BASE, &board)) != 0) {
        CERR("ALTI: reading BOARD ID from CR/CSR, slot %d", slot);
        return (status);
    }

    // close VME
    RCD::VME::Close();

    return (SUCCESS);
}

//------------------------------------------------------------------------------

int AltiModule::CSRBARRead(int slot, unsigned int &bar) {
    // open VME
    RCD::VME *vme = RCD::VME::Open();

    // BAR
    int status(SUCCESS);
    if ((status = vme->ReadCRCSR(slot, BAR_BASE, &bar)) != SUCCESS) {
        CERR("ALTI: reading BAR from CR/CSR, slot %d", slot);
        return (status);
    }

    // close VME
    RCD::VME::Close();

    return (SUCCESS);
}

//------------------------------------------------------------------------------

int AltiModule::CSRADER0Read(int slot, unsigned int &ader0) {
    // open VME
    RCD::VME *vme = RCD::VME::Open();

    // ADER0
    int status(SUCCESS);
    if ((status = vme->ReadCRCSR(slot, ADER0_BASE, &ader0)) != SUCCESS) {
        CERR("ALTI: reading ADER0 from CR/CSR, slot %d", slot);
        return (status);
    }

    // close VME
    RCD::VME::Close();

    return (SUCCESS);
}

//------------------------------------------------------------------------------

int AltiModule::CSRADER0Write(int slot, const unsigned int ader0) {
    // open VME
    RCD::VME *vme = RCD::VME::Open();

    // ADER0
    int status(SUCCESS);
    if ((status = vme->WriteCRCSR(slot, ADER0_BASE, ader0)) != 0) {
        CERR("ALTI: writing ADER0 0x%08x to CR/CSR", ader0);
        return (status);
    }

    // close VME
    RCD::VME::Close();

    return (SUCCESS);
}

//------------------------------------------------------------------------------

int AltiModule::CSRADER1Read(int slot, unsigned int &ader1) {
    // open VME
    RCD::VME *vme = RCD::VME::Open();

    // ADER1
    int status(SUCCESS);
    if ((status = vme->ReadCRCSR(slot, ADER1_BASE, &ader1)) != SUCCESS) {
        CERR("ALTI: reading ADER1 from CR/CSR, slot %d", slot);
        return (status);
    }

    // close VME
    RCD::VME::Close();

    return (SUCCESS);
}

//------------------------------------------------------------------------------

int AltiModule::CSRADER1Write(int slot, const unsigned int ader1) {
    // open VME
    RCD::VME *vme = RCD::VME::Open();

    // ADER1
    int status(SUCCESS);
    if ((status = vme->WriteCRCSR(slot, ADER1_BASE, ader1)) != 0) {
        CERR("ALTI: writing ADER1 0x%08x to CR/CSR", ader1);
        return (status);
    }

    // close VME
    RCD::VME::Close();

    return (SUCCESS);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::SIGSwitchConfigRead(const SIGNAL signal, const SIGNAL_DESTINATION dst, SIGNAL_SOURCE &src) { // 1 signal, 1 output

    I2CNetwork::I2CNode &node = m_SignalSwitchConfig[signal];
    // READ
    unsigned int readout;
    I2CDataRead(node, readout);

    SWITCH_CONFIG switch_config_bs;
    std::vector<unsigned int> tmp;
    tmp.clear();
    tmp.push_back(static_cast<unsigned int>(readout));
    switch_config_bs = (RCD::BitSet) tmp;

    SIGNAL_SOURCE tmp_src = (SIGNAL_SOURCE) switch_config_bs.Output(SWX_MAP_DST[signal][dst]);
    src = SWX_MAP_SRC_INV[signal][tmp_src];

    m_status = SUCCESS;
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGSwitchConfigRead(const SIGNAL signal, std::vector<SIGNAL_SOURCE> &src) { // 1 signal, all outputs
    src.clear();

    SIGNAL_SOURCE tmp_src;
    SIGNAL_DESTINATION dst;

    unsigned int i;
    for (i = 0; i < SIGNAL_DESTINATION_NUMBER; i++) {
        dst = (SIGNAL_DESTINATION) i;
        if ((m_status = SIGSwitchConfigRead(signal, dst, tmp_src)) != SUCCESS) {
            CERR("ALTI: reading switch source for signal \"%s\", destination \"%s\"", SIGNAL_NAME[signal].c_str(), SIGNAL_DESTINATION_NAME[dst].c_str());
            return (m_status);
        }
        src.push_back(tmp_src);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGSwitchConfigWrite(const SIGNAL signal, const SIGNAL_DESTINATION dst, const SIGNAL_SOURCE src) { // 1 signal, 1 output

    I2CNetwork::I2CNode &node = m_SignalSwitchConfig[signal];
    // READ
    unsigned int readout;
    I2CDataRead(node, readout);

    SWITCH_CONFIG switch_config_bs;
    std::vector<unsigned int> tmp;
    tmp.clear();
    tmp.push_back(static_cast<unsigned int>(readout));
    switch_config_bs = (RCD::BitSet) tmp;

    // MODIFY
    switch_config_bs.Output(SWX_MAP_DST[signal][dst], SWX_MAP_SRC[signal][src]);

    // WRITE
    m_net->write(node, switch_config_bs.number()[0]);

    m_status = SUCCESS;
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGSwitchConfigWrite(const SIGNAL_DESTINATION dst, const SIGNAL_SOURCE src) { // all signals, 1 output
    SIGNAL signal;

    unsigned int i;
    for (i = 0; i < SIGNAL_NUMBER_ROUTED; i++) {
        signal = (SIGNAL) i;
        if ((m_status = SIGSwitchConfigWrite(signal, dst, src)) != SUCCESS) {
            CERR("ALTI: writing switch source \"%s\" for signal \"%s\", destination \"%s\"", SIGNAL_SOURCE_NAME[src].c_str(), SIGNAL_NAME[signal].c_str(), SIGNAL_DESTINATION_NAME[dst].c_str());
            return (m_status);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGSwitchConfigWrite(const SIGNAL signal, const SIGNAL_SOURCE src) { // 1 signal, all outputs

    I2CNetwork::I2CNode &node = m_SignalSwitchConfig[signal];
    // READ
    unsigned int readout;
    I2CDataRead(node, readout);

    SWITCH_CONFIG switch_config_bs;
    std::vector<unsigned int> tmp;
    tmp.clear();
    tmp.push_back(static_cast<unsigned int>(readout));
    switch_config_bs = (RCD::BitSet) tmp;

    // MODIFY
    unsigned int i;
    SIGNAL_DESTINATION dst;
    for (i = 0; i < SIGNAL_DESTINATION_NUMBER; i++) {
        dst = (SIGNAL_DESTINATION) i;
        switch_config_bs.Output(SWX_MAP_DST[signal][dst], SWX_MAP_SRC[signal][src]);
    }

    // WRITE
    m_net->write(node, switch_config_bs.number()[0]);

    m_status = SUCCESS;

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGSwitchConfigWrite(const SIGNAL_SOURCE src) { // all signals, all outputs
    SIGNAL signal;
    unsigned i;
    for (i = 0; i < SIGNAL_NUMBER_ROUTED; i++) {
        signal = (SIGNAL) i;
        if ((m_status = SIGSwitchConfigWrite(signal, src)) != SUCCESS) {
            CERR("ALTI: writing switch source \"%s\" for signal \"%s\", all destinations", SIGNAL_SOURCE_NAME[src].c_str(), SIGNAL_NAME[signal].c_str());
            return (m_status);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGSwitchConfigWrite(const ALTI_MODE mode) {
    switch (mode) {
        case MASTER:
            return SIGSwitchMasterConfigurationWrite();
            break;
        case CTP_SLAVE:
            return SIGSwitchCtpSlaveConfigurationWrite();
            break;
        case ALTI_SLAVE:
            return SIGSwitchAltiSlaveConfigurationWrite();
            break;
        case LEMO_SLAVE:
            return SIGSwitchLemoSlaveConfigurationWrite();
            break;
        default:
            break;
    }

    return (SUCCESS);
}

//------------------------------------------------------------------------------

int AltiModule::SIGSwitchPowerDownEnableRead(const SIGNAL signal, const SIGNAL_DESTINATION dst, bool &pwdn) { // 1 signal, 1 output

    I2CNetwork::I2CNode &node = m_SignalSwitchCtrl[signal];
    // READ
    unsigned int readout;
    I2CDataRead(node, readout);

    SWITCH_CTRL switch_ctrl_bs;
    std::vector<unsigned int> tmp;
    tmp.clear();
    tmp.push_back(static_cast<unsigned int>(readout));
    switch_ctrl_bs = (RCD::BitSet) tmp;

    std::string s = switch_ctrl_bs.PowerDown(SWX_MAP_DST[signal][dst]);
    if (s == "Enabled") {
       pwdn = true;
    }
    else if (s == "Disabled") {
       pwdn = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for power-down enable of signal \"%s\", destination \"%s\"", s, AltiModule::SIGNAL_NAME[signal].c_str(), AltiModule::SIGNAL_DESTINATION_NAME[dst].c_str());
        return (FAILURE);
    }

    m_status = SUCCESS;
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGSwitchPowerDownEnableRead(const SIGNAL signal, std::vector<bool> &pwdn) { // 1 signal, all outputs
    pwdn.clear();

    bool tmp_pwdn;
    SIGNAL_DESTINATION dst;

    unsigned int i;
    for (i = 0; i < SIGNAL_DESTINATION_NUMBER; i++) {
        dst = (SIGNAL_DESTINATION) i;
        if ((m_status = SIGSwitchPowerDownEnableRead(signal, dst, tmp_pwdn)) != SUCCESS) {
            CERR("ALTI: reading power-down mode for signal \"%s\", destination \"%s\"", SIGNAL_NAME[signal].c_str(), SIGNAL_DESTINATION_NAME[dst].c_str());
            return (m_status);
        }
        pwdn.push_back(tmp_pwdn);
    }

    return (m_status);
}

//------------------------------------------------------------------------------*/

int AltiModule::SIGSwitchPowerDownEnableWrite(const SIGNAL signal, const SIGNAL_DESTINATION dst, const bool pwdn) { // 1 signal, 1 output

    I2CNetwork::I2CNode &node = m_SignalSwitchCtrl[signal];
    // READ
    unsigned int readout;
    I2CDataRead(node, readout);

    SWITCH_CTRL switch_ctrl_bs;
    std::vector<unsigned int> tmp;
    tmp.clear();
    tmp.push_back(static_cast<unsigned int>(readout));
    switch_ctrl_bs = (RCD::BitSet) tmp;

    // MODIFY
    switch_ctrl_bs.PowerDown(SWX_MAP_DST[signal][dst], pwdn ? "Enabled" : "Disabled");

    // WRITE
    m_net->write(node, switch_ctrl_bs.number()[0]);

    m_status = SUCCESS;
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGSwitchPowerDownEnableWrite(const SIGNAL_DESTINATION dst, const bool pwdn) { // all signals, 1 output
    SIGNAL signal;

    unsigned int i;
    for (i = 0; i < SIGNAL_NUMBER_ROUTED; i++) {
        signal = (SIGNAL) i;
        if ((m_status = SIGSwitchPowerDownEnableWrite(signal, dst, pwdn)) != SUCCESS) {
            CERR("ALTI: %s power-down mode for signal \"%s\", destination \"%s\"", pwdn ? "enabling" : "disabling", SIGNAL_NAME[signal].c_str(), SIGNAL_DESTINATION_NAME[dst].c_str());
            return (m_status);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGSwitchPowerDownEnableWrite(const SIGNAL signal, const bool pwdn) { // 1 signal, all outputs

    I2CNetwork::I2CNode &node = m_SignalSwitchCtrl[signal];
    // READ
    unsigned int readout;
    I2CDataRead(node, readout);

    SWITCH_CTRL switch_ctrl_bs;
    std::vector<unsigned int> tmp;
    tmp.clear();
    tmp.push_back(static_cast<unsigned int>(readout));
    switch_ctrl_bs = (RCD::BitSet) tmp;

    // MODIFY
    unsigned int i;
    SIGNAL_DESTINATION dst;
    for (i = 0; i < SIGNAL_DESTINATION_NUMBER; i++) {
        dst = (SIGNAL_DESTINATION) i;
        switch_ctrl_bs.PowerDown(SWX_MAP_DST[signal][dst], pwdn ? "Enabled" : "Disabled");
    }

    // WRITE
    m_net->write(node, switch_ctrl_bs.number()[0]);

    m_status = SUCCESS;

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGSwitchPowerDownEnableWrite(const bool pwdn) { // all signals, all outputs
    SIGNAL signal;
    unsigned i;
    for (i = 0; i < SIGNAL_NUMBER_ROUTED; i++) {
        signal = (SIGNAL) i;
        if ((m_status = SIGSwitchPowerDownEnableWrite(signal, pwdn)) != SUCCESS) {
            CERR("ALTI: %s power-down mode for signal \"%s\", all destinations", pwdn ? "enabling" : "disabling", SIGNAL_NAME[signal].c_str());
            return (m_status);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGSwitchMasterConfigurationWrite() {
  return SIGSwitchConfigWrite(FROM_FPGA); // everything from FPGA
}

//------------------------------------------------------------------------------

int AltiModule::SIGSwitchCtpSlaveConfigurationWrite() {
    return SIGSwitchConfigWrite(CTP_IN); // everything from CTP
}

//------------------------------------------------------------------------------

int AltiModule::SIGSwitchAltiSlaveConfigurationWrite() {
    return SIGSwitchConfigWrite(ALTI_IN); // everything from ALTI
}

//------------------------------------------------------------------------------

int AltiModule::SIGSwitchLemoSlaveConfigurationWrite() {
    return SIGSwitchConfigWrite(NIM_IN); // everything from NIM
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------


int AltiModule::SIGEqualizersEnableRead(const SIGNAL_SOURCE src, bool &enabled) {
 if (src >= SIGNAL_SOURCE_NUMBER - 3) {
    CERR("ALTI: wrong signal source chosen", "");
    m_status = WRONGPAR;
    return (m_status);
  }
  
 ALTI_IO_EQUALIZERCONTROL_BITSTRING bs;
 // READ
 if ((m_status = m_alti->IO_EqualizerControl_Read(bs)) != 0) {
   CERR("ALTI: reading equalizer control", "");
   return (m_status);
 }

 std::string s;
 if(src==CTP_IN) s = bs.EnableCTPEqualizer();
 else if(src==ALTI_IN) s = bs.EnableALTIEqualizer();

 
 if (s == "Enabled") {
   enabled = true;
 }
 else if (s == "Disabled") {
   enabled = false;
 }
 else {
   CERR("ALTI: unknown value \"%s\" for equalizer control of signal %s", s, SIGNAL_SOURCE_NAME[src].c_str());
   return (-1);
 }

  return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGEqualizersEnableWrite(const SIGNAL_SOURCE src, const bool enabled) {
 if (src >= SIGNAL_SOURCE_NUMBER - 3) {
    CERR("ALTI: wrong signal source chosen", "");
    m_status = WRONGPAR;
    return (m_status);
  }
  
 ALTI_IO_EQUALIZERCONTROL_BITSTRING bs;

 // READ 
 if ((m_status = m_alti->IO_EqualizerControl_Read(bs)) != 0) {
   CERR("ALTI: reading equalizer control", "");
   return (m_status);
 }

 // MODIFY
 if(src==CTP_IN)  bs.EnableCTPEqualizer(enabled ? "Enabled" : "Disabled"); 
 else if(src==ALTI_IN)  bs.EnableALTIEqualizer(enabled ? "Enabled" : "Disabled"); 

 // WRITE
 if ((m_status = m_alti->IO_EqualizerControl_Write(bs)) != 0) {
   CERR("ALTI: writing equalizer control", "");
   return (m_status);
 }

  return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGEqualizersConfigRead(const SIGNAL_SOURCE src, EQUALIZER_VOLTAGE vlt, unsigned int &v) {
  if (src >= SIGNAL_SOURCE_NUMBER - 3) {
    CERR("ALTI: wrong signal source chosen", "");
    m_status = WRONGPAR;
    return (m_status);
  }

  // READ
  unsigned int readout;
  if ((m_status = m_net->read(m_EqualizerConfig[src][vlt], readout)) != 0) {
    CERR("ALTI: reading %s %s from AD5305", SIGNAL_SOURCE_NAME[src].c_str(), EQUALIZER_VOLTAGE_NAME[vlt].c_str());
    return (m_status);
  }

  EQUALIZER_DAC eqz_dac_bs;
  std::vector<unsigned int> tmp;
  tmp.clear();
  tmp.push_back(static_cast<unsigned int>(readout));
  eqz_dac_bs = (RCD::BitSet) tmp;
  // DAC 2B readout pattern
  v = eqz_dac_bs.Data();

  return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGEqualizersConfigWrite(const SIGNAL_SOURCE src, const EQUALIZER_VOLTAGE vlt, const unsigned int v) {
  if (src >= SIGNAL_SOURCE_NUMBER - 3) {
    CERR("ALTI: wrong signal source chosen", "");
    m_status = WRONGPAR;
    return (m_status);
  }

  EQUALIZER_DAC eqz_dac_bs;
  std::vector<unsigned int> tmp;
  tmp.clear();
  tmp.push_back(static_cast<unsigned int>(0x00000000));
  eqz_dac_bs = (RCD::BitSet) tmp;
  // DAC 2B write pattern: CLR (active low) bit set to 1 for normal operation, LDAC (active low) set to 0 for latching, no PD modes used
  eqz_dac_bs.Data(v);
  eqz_dac_bs.Clear("Disabled");
  eqz_dac_bs.Load("Enabled");

  // WRITE
  if ((m_status = m_net->write(m_EqualizerConfig[src][vlt], eqz_dac_bs.number()[0])) != 0) {
    CERR("ALTI: writing %s %s to AD5305", SIGNAL_SOURCE_NAME[src].c_str(), EQUALIZER_VOLTAGE_NAME[vlt].c_str());
    return (m_status);
  }

  return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGEqualizersConfigRead(const SIGNAL_SOURCE src, unsigned int &vpeak, unsigned int &vpole, unsigned int &vgain, unsigned int &voffset) {
  if (src >= SIGNAL_SOURCE_NUMBER - 3) {
    CERR("ALTI: wrong signal source chosen", "");
    m_status = WRONGPAR;
    return (m_status);
  }
  
  if ((m_status = SIGEqualizersConfigRead(src, V_PEAK, vpeak)) != 0) {
    CERR("ALTI: reading %s %s from AD5305", SIGNAL_SOURCE_NAME[src].c_str(), EQUALIZER_VOLTAGE_NAME[V_PEAK].c_str());
    return (m_status);
  }
  if ((m_status = SIGEqualizersConfigRead(src, V_POLE, vpole)) != 0) {
    CERR("ALTI: reading %s %s from AD5305", SIGNAL_SOURCE_NAME[src].c_str(), EQUALIZER_VOLTAGE_NAME[V_POLE].c_str());
    return (m_status);
  }
  if ((m_status = SIGEqualizersConfigRead(src, V_GAIN, vgain)) != 0) {
    CERR("ALTI: reading %s %s from AD5305", SIGNAL_SOURCE_NAME[src].c_str(), EQUALIZER_VOLTAGE_NAME[V_GAIN].c_str());
    return (m_status);
  }
  if ((m_status = SIGEqualizersConfigRead(src, V_OFFSET, voffset)) != 0) {
    CERR("ALTI: reading %s %s from AD5305", SIGNAL_SOURCE_NAME[src].c_str(), EQUALIZER_VOLTAGE_NAME[V_OFFSET].c_str());
    return (m_status);
  }

  return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGEqualizersConfigWrite(const SIGNAL_SOURCE src, const unsigned int vpeak, const unsigned int vpole, const unsigned int vgain, const unsigned int voffset) {
  if (src >= SIGNAL_SOURCE_NUMBER - 3) {
    CERR("ALTI: wrong signal source chosen", "");
    m_status = WRONGPAR;
    return (m_status);
  }
  
  if ((m_status = SIGEqualizersConfigWrite(src, V_PEAK, vpeak)) != 0) {
    CERR("ALTI: writing %s %s to AD5305", SIGNAL_SOURCE_NAME[src].c_str(), EQUALIZER_VOLTAGE_NAME[V_PEAK].c_str());
    return (m_status);
  }
  if ((m_status = SIGEqualizersConfigWrite(src, V_POLE, vpole)) != 0) {
    CERR("ALTI: writing %s %s to AD5305", SIGNAL_SOURCE_NAME[src].c_str(), EQUALIZER_VOLTAGE_NAME[V_POLE].c_str());
    return (m_status);
  }
  if ((m_status = SIGEqualizersConfigWrite(src, V_GAIN, vgain)) != 0) {
    CERR("ALTI: writing %s %s to AD5305", SIGNAL_SOURCE_NAME[src].c_str(), EQUALIZER_VOLTAGE_NAME[V_GAIN].c_str());
    return (m_status);
  }
  if ((m_status = SIGEqualizersConfigWrite(src, V_OFFSET, voffset)) != 0) {
    CERR("ALTI: writing %s %s to AD5305", SIGNAL_SOURCE_NAME[src].c_str(), EQUALIZER_VOLTAGE_NAME[V_OFFSET].c_str());
    return (m_status);
  }

  return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGEqualizersConfigWrite(const SIGNAL_SOURCE src, const bool enable, const EQUALIZER_CONFIG config) {
  if (src >= SIGNAL_SOURCE_NUMBER - 3) {
    CERR("ALTI: wrong signal source chosen", "");
    m_status = WRONGPAR;
    return (m_status);
  }

  if (config == CUSTOM_CONFIG) {
    CERR("ALTI: wrong input parameter \"%s\" for writing \"%s\" equalizer configuration", EQUALIZER_CONFIG_NAME[config].c_str(), SIGNAL_SOURCE_NAME[src].c_str());
    m_status = WRONGPAR;
    return (m_status);
  }

  if ((m_status = SIGEqualizersEnableWrite(src, enable)) != 0) {
      CERR("ALTI: writing \"%s\" equalizer enable to \"s\" cable", enable, SIGNAL_SOURCE_NAME[src].c_str());
    return (m_status);
  }

  if ((m_status = SIGEqualizersConfigWrite(src, EQUALIZER_CONFIG_VOLTAGE[config][V_PEAK], EQUALIZER_CONFIG_VOLTAGE[config][V_POLE], EQUALIZER_CONFIG_VOLTAGE[config][V_GAIN], EQUALIZER_CONFIG_VOLTAGE[config][V_OFFSET])) != 0) {
    CERR("ALTI: writing \"%s\" equalizer configuration to \"s\" cable", EQUALIZER_CONFIG_NAME[config].c_str(), SIGNAL_SOURCE_NAME[src].c_str());
    return (m_status);
  }
 
  return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGEqualizersConfigRead(const SIGNAL_SOURCE src, bool &enable, EQUALIZER_CONFIG &config) {
  if (src >= SIGNAL_SOURCE_NUMBER - 3) {
    CERR("ALTI: wrong signal source chosen", "");
    m_status = WRONGPAR;
    return (m_status);
  }

  if ((m_status = SIGEqualizersEnableRead(src, enable)) != 0) {
    CERR("ALTI: reading \"%s\" equalizer enable to \"s\" cable", enable, SIGNAL_SOURCE_NAME[src].c_str());
    return (m_status);
  }
  
  unsigned int voltages[EQUALIZER_VOLTAGE_NUMBER];
  if ((m_status = SIGEqualizersConfigRead(src, voltages[0], voltages[1], voltages[2], voltages[3])) != 0) {
    CERR("ALTI: reading \"%s\" equalizer configuration", SIGNAL_SOURCE_NAME[src].c_str());
    return (m_status);
  }
  
  if ((voltages[0] == EQUALIZER_CONFIG_VOLTAGE[SHORT_CABLE][0]) 
      && (voltages[1] == EQUALIZER_CONFIG_VOLTAGE[SHORT_CABLE][1]) 
      && (voltages[2] == EQUALIZER_CONFIG_VOLTAGE[SHORT_CABLE][2]) 
      && (voltages[3] == EQUALIZER_CONFIG_VOLTAGE[SHORT_CABLE][3])) {
    config = SHORT_CABLE;
  }
  else if ((voltages[0] == EQUALIZER_CONFIG_VOLTAGE[LONG_CABLE][0]) 
	   && (voltages[1] == EQUALIZER_CONFIG_VOLTAGE[LONG_CABLE][1]) 
	   && (voltages[2] == EQUALIZER_CONFIG_VOLTAGE[LONG_CABLE][2]) 
	   && (voltages[3] == EQUALIZER_CONFIG_VOLTAGE[LONG_CABLE][3])) {
    config = LONG_CABLE;
  }
  else {
    config = CUSTOM_CONFIG;
  }

  return (m_status);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::SIGInputSyncHistogramsRead(const ASYNC_INPUT_SIGNAL asyncIn, std::vector<unsigned int> &countPosByPhase, std::vector<unsigned int> &countNegByPhase) {
    ALTI_IO_EXTINPUTSCONFIGANDHIST_BITSTRING bs;

    if ((m_status = m_alti->IO_ExtInputsConfigAndHist_Read(asyncIn, bs)) != 0) {
        CERR("ALTI: reading configuration and histogram data register for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
        return (m_status);
    }

    countPosByPhase.clear();
    countNegByPhase.clear();
    countPosByPhase.push_back(bs.HistPos000Deg());
    countNegByPhase.push_back(bs.HistNeg000Deg());
    countPosByPhase.push_back(bs.HistPos090Deg());
    countNegByPhase.push_back(bs.HistNeg090Deg());
    countPosByPhase.push_back(bs.HistPos180Deg());
    countNegByPhase.push_back(bs.HistNeg180Deg());
    countPosByPhase.push_back(bs.HistPos270Deg());
    countNegByPhase.push_back(bs.HistNeg270Deg());

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGInputSyncHistogramsReset() {
    ALTI_IO_EXTINPUTSCONFIGANDHIST_BITSTRING bs;

    ASYNC_INPUT_SIGNAL asyncIn;
    unsigned int i;
    for (i = 0; i < ASYNC_INPUT_SIGNAL_NUMBER; i++) {
        asyncIn = (ASYNC_INPUT_SIGNAL) i;
        // READ
        if ((m_status = m_alti->IO_ExtInputsConfigAndHist_Read(asyncIn, bs)) != 0) {
            CERR("ALTI: reading configuration and histogram data register for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
            return (m_status);
        }
        // WRITE: reset histogram
        bs.HistReset("Enabled");
        if ((m_status = m_alti->IO_ExtInputsConfigAndHist_Write(asyncIn, bs)) != 0) {
            CERR("ALTI: writing configuration and histogram data register for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
            return (m_status);
        }
        // WRITE: re-enable histogram calculation
        bs.HistReset("Disabled");
        if ((m_status = m_alti->IO_ExtInputsConfigAndHist_Write(asyncIn, bs)) != 0) {
            CERR("ALTI: writing configuration and histogram data register for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
            return (m_status);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGInputSyncEnableRead(const ASYNC_INPUT_SIGNAL asyncIn, bool &enabled, unsigned int &phase) {
    ALTI_IO_EXTINPUTSCONFIGANDHIST_BITSTRING bs;

    // READ
    if ((m_status = m_alti->IO_ExtInputsConfigAndHist_Read(asyncIn, bs)) != 0) {
        CERR("ALTI: reading configuration and histogram data register for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
        return (m_status);
    }
    std::string s = bs.SyncEnable();
    if (s == "Enabled") {
       enabled = true;
    }
    else if (s == "Disabled") {
       enabled = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for sync enable of asynchronous input signal \"%s\"", s, AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
        return (-1);
    }
    unsigned int i, sum = 0;
    for (i = 0; i < 4; i++) {
        sum += ((bs.SyncClockSelect(i)) << i);
    }
    // phase: 0-auto/1-000deg/2-090deg/3-180deg/4-270deg/5-invalid
    phase = ((sum == 0) ? 0 : ((sum == 1) ? 1 : (sum == 2) ? 2 : (sum == 4) ? 3 : ((sum == 8) ? 4 : 5)));

    return (m_status);
}

//------------------------------------------------------------------------------

// phase: 0-auto/1-000deg/2-090deg/3-180deg/4-270deg
int AltiModule::SIGInputSyncEnableWrite(const ASYNC_INPUT_SIGNAL asyncIn, const bool enable, const unsigned int phase) {
    if (phase < 0 || phase > 4) {
        CERR("ALTI: wrong phase index \"%d\", allowed range is [0, 4]", phase);
        m_status = -1;
        return (m_status);
    }

    ALTI_IO_EXTINPUTSCONFIGANDHIST_BITSTRING bs;

    // READ
    if ((m_status = m_alti->IO_ExtInputsConfigAndHist_Read(asyncIn, bs)) != 0) {
        CERR("ALTI: reading configuration and histogram data register for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
        return (m_status);
    }
    // WRITE: reset histogram
    bs.HistReset("Enabled");
    if ((m_status = m_alti->IO_ExtInputsConfigAndHist_Write(asyncIn, bs)) != 0) {
        CERR("ALTI: writing configuration and histogram data register for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
        return (m_status);
    }
    // MODIFY
    bs.SyncEnable(enable ? "Enabled" : "Disabled");
    unsigned int i;
    for (i = 0; i < 4; i++) {
        bs.SyncClockSelect(i, (phase > 0) ? (i == (phase - 1)) : 0);
    }
    // WRITE: re-enable histogram calculation
    bs.HistReset("Disabled");
    if ((m_status = m_alti->IO_ExtInputsConfigAndHist_Write(asyncIn, bs)) != 0) {
        CERR("ALTI: writing configuration and histogram data register for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGInputSyncEnableRead(std::vector<bool> &enabled, std::vector<unsigned int> &phase) { // all
    enabled.clear();
    phase.clear();

    unsigned int i;
    ASYNC_INPUT_SIGNAL asyncIn;
    bool rd_enabled;
    unsigned int rd_phase;
    for (i = 0; i < ASYNC_INPUT_SIGNAL_NUMBER; i++) {
        asyncIn = (ASYNC_INPUT_SIGNAL) i;
        if ((m_status = SIGInputSyncEnableRead(asyncIn, rd_enabled, rd_phase)) != 0) {
            CERR("ALTI: reading synchronization enable and phase for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
            return (m_status);
        }
        enabled.push_back(rd_enabled);
        phase.push_back(rd_phase);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGInputSyncEnableWrite(const std::vector<bool> &enable, const std::vector<unsigned int> &phase) { // all
    unsigned int i;
    ASYNC_INPUT_SIGNAL asyncIn;
    for (i = 0; i < ASYNC_INPUT_SIGNAL_NUMBER; i++) {
        asyncIn = (ASYNC_INPUT_SIGNAL) i;
        if ((m_status = SIGInputSyncEnableWrite(asyncIn, enable[i], phase[i])) != 0) {
            CERR("ALTI: writing synchronization enable and phase for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
            return (m_status);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGInputSyncEnableWrite(const bool enable, const unsigned int phase) { // all
    unsigned int i;
    ASYNC_INPUT_SIGNAL asyncIn;
    for (i = 0; i < ASYNC_INPUT_SIGNAL_NUMBER; i++) {
        asyncIn = (ASYNC_INPUT_SIGNAL) i;
        if ((m_status = SIGInputSyncEnableWrite(asyncIn, enable, phase)) != 0) {
            CERR("ALTI: writing synchronization enable and phase for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
            return (m_status);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGInputShapeEnableRead(const ASYNC_INPUT_SIGNAL asyncIn, bool &enabled) {
    ALTI_IO_EXTINPUTSCONFIGANDHIST_BITSTRING bs;

    // READ
    if ((m_status = m_alti->IO_ExtInputsConfigAndHist_Read(asyncIn, bs)) != 0) {
        CERR("ALTI: reading configuration and histogram data register for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
        return (m_status);
    }
    std::string s = bs.ShapeEnable();
    if (s == "Enabled") {
       enabled = true;
    }
    else if (s == "Disabled") {
       enabled = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for shape enable of asynchronous input signal \"%s\"", s, AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
        return (-1);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGInputShapeEnableWrite(const ASYNC_INPUT_SIGNAL asyncIn, const bool enable) {
    ALTI_IO_EXTINPUTSCONFIGANDHIST_BITSTRING bs;

    // READ
    if ((m_status = m_alti->IO_ExtInputsConfigAndHist_Read(asyncIn, bs)) != 0) {
        CERR("ALTI: reading configuration and histogram data register for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
        return (m_status);
    }
    // WRITE: reset histogram
    bs.HistReset("Enabled");
    if ((m_status = m_alti->IO_ExtInputsConfigAndHist_Write(asyncIn, bs)) != 0) {
        CERR("ALTI: writing configuration and histogram data register for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
        return (m_status);
    }
    // MODIFY
    bs.ShapeEnable(enable ? "Enabled" : "Disabled");
    // WRITE: re-enable histogram calculation
    bs.HistReset("Disabled");
    if ((m_status = m_alti->IO_ExtInputsConfigAndHist_Write(asyncIn, bs)) != 0) {
        CERR("ALTI: writing configuration and histogram data register for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGInputShapeEnableRead(std::vector<bool> &enabled) { // all
    enabled.clear();

    unsigned int i;
    ASYNC_INPUT_SIGNAL asyncIn;
    bool rd_enabled;
    for (i = 0; i < ASYNC_INPUT_SIGNAL_NUMBER; i++) {
        asyncIn = (ASYNC_INPUT_SIGNAL) i;
        if ((m_status = SIGInputShapeEnableRead(asyncIn, rd_enabled)) != 0) {
            CERR("ALTI: reading shape enable for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
            return (m_status);
        }
        enabled.push_back(rd_enabled);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGInputShapeEnableWrite(const std::vector<bool> &enable) { // all
    unsigned int i;
    ASYNC_INPUT_SIGNAL asyncIn;
    for (i = 0; i < ASYNC_INPUT_SIGNAL_NUMBER; i++) {
        asyncIn = (ASYNC_INPUT_SIGNAL) i;
        if ((m_status = SIGInputShapeEnableWrite(asyncIn, enable[i])) != 0) {
            CERR("ALTI: writing shape enable for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
            return (m_status);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGInputShapeEnableWrite(const bool enable) { // all
    unsigned int i;
    ASYNC_INPUT_SIGNAL asyncIn;
    for (i = 0; i < ASYNC_INPUT_SIGNAL_NUMBER; i++) {
        asyncIn = (ASYNC_INPUT_SIGNAL) i;
        if ((m_status = SIGInputShapeEnableWrite(asyncIn, enable)) != 0) {
            CERR("ALTI: writing shape enable for asynchronous input signal \"%s\"", ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
            return (m_status);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::BSYInputSelectRead(const BUSY_INPUT in, bool &selected) { // 1 input
    ALTI_BSY_CONTROL_BITSTRING bs;

    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }

    std::string s;
    switch (in) {
        case BUSY_FROM_FRONT_PANEL:
            s = bs.Busy_From_FrontPanel_Enable();
            break;
        case BUSY_FROM_CTP:
            s = bs.Busy_From_CTP_Enable();
            break;
        case BUSY_FROM_ALTI:
            s = bs.Busy_From_ALTI_Enable();
            break;
        case BUSY_FROM_PG:
            s = bs.Busy_From_Pattern_Enable();
            break;
        case BUSY_FROM_VME:
            s = bs.Busy_From_VME_Enable();
            break;
        case BUSY_FROM_ECR:
            s = bs.Busy_From_ECR_Enable();
            break;
        default:
            break;
    } 

    if (s == "Enabled") {
       selected = true;
    }
    else if (s == "Disabled") {
       selected = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for busy input \"%s\"", s, AltiModule::BUSY_INPUT_NAME[in].c_str());
        return (-1);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYInputSelectWrite(const BUSY_INPUT in, const bool select) { // 1 input
    ALTI_BSY_CONTROL_BITSTRING bs;

    // read
    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }

    // modify
    switch (in) {
        case BUSY_FROM_FRONT_PANEL:
            bs.Busy_From_FrontPanel_Enable(select ? "Enabled" : "Disabled");
            break;
        case BUSY_FROM_CTP:
            bs.Busy_From_CTP_Enable(select ? "Enabled" : "Disabled");
            break;
        case BUSY_FROM_ALTI:
            bs.Busy_From_ALTI_Enable(select ? "Enabled" : "Disabled");
            break;
        case BUSY_FROM_PG:
            bs.Busy_From_Pattern_Enable(select ? "Enabled" : "Disabled");
            break;
        case BUSY_FROM_VME:
            bs.Busy_From_VME_Enable(select ? "Enabled" : "Disabled");
            break;
        case BUSY_FROM_ECR:
	    bs.Busy_From_ECR_Enable(select ? "Enabled" : "Disabled");
	    break;
	default:
            break;
    }

    // write
    if ((m_status = m_alti->BSY_Control_Write(bs)) != 0) {
        CERR("ALTI: writing busy control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYInputSelectRead(std::vector<bool> &selected) { // all inputs
    ALTI_BSY_CONTROL_BITSTRING bs;

    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }

    selected.clear();

    unsigned int i;
    std::string s;
    for (i = 0; i < BUSY_INPUT_NUMBER; i++) {
        // 0: not selected, 1: selected
        switch ((BUSY_INPUT) i) {
            case BUSY_FROM_FRONT_PANEL:
                s = bs.Busy_From_FrontPanel_Enable();
                break;
            case BUSY_FROM_CTP:
                s = bs.Busy_From_CTP_Enable();
                break;
            case BUSY_FROM_ALTI:
                s = bs.Busy_From_ALTI_Enable();
                break;
            case BUSY_FROM_PG:
                s = bs.Busy_From_Pattern_Enable();
                break;
            case BUSY_FROM_VME:
                s = bs.Busy_From_VME_Enable();
                break;
	    case BUSY_FROM_ECR:
		s = bs.Busy_From_ECR_Enable();
		break;
            default:
                break;
        }
        if (s == "Enabled") {
           selected.push_back(true);
        }
        else if (s == "Disabled") {
           selected.push_back(false);
        }
        else {
            CERR("ALTI: unknown value \"%s\" for busy input \"%s\"", s, AltiModule::BUSY_INPUT_NAME[(BUSY_INPUT) i].c_str());
            return (-1);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYInputSelectWrite(const std::vector<bool> &select) { // all inputs
    ALTI_BSY_CONTROL_BITSTRING bs;

    // read
    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }

    // modify
    unsigned int i;
    for (i = 0; i < BUSY_INPUT_NUMBER; i++) {
        // 0: not selected, 1: selected
        switch ((BUSY_INPUT) i) {
            case BUSY_FROM_FRONT_PANEL:
                bs.Busy_From_FrontPanel_Enable(select[i] ? "Enabled" : "Disabled");
                break;
            case BUSY_FROM_CTP:
                bs.Busy_From_CTP_Enable(select[i] ? "Enabled" : "Disabled");
                break;
            case BUSY_FROM_ALTI:
                bs.Busy_From_ALTI_Enable(select[i] ? "Enabled" : "Disabled");
                break;
            case BUSY_FROM_PG:
                bs.Busy_From_Pattern_Enable(select[i] ? "Enabled" : "Disabled");
                break;
            case BUSY_FROM_VME:
                bs.Busy_From_VME_Enable(select[i] ? "Enabled" : "Disabled");
                break;
            case BUSY_FROM_ECR:
                bs.Busy_From_ECR_Enable(select[i] ? "Enabled" : "Disabled");
                break;
            default:
                break;
        }
    }

    // write
    if ((m_status = m_alti->BSY_Control_Write(bs)) != 0) {
        CERR("ALTI: writing busy control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYOutputSelectRead(const BUSY_OUTPUT out, BUSY_SOURCE &src) { // 1 output
    ALTI_BSY_CONTROL_BITSTRING bs;

    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }

    std::string s;
    switch (out) {
        case BUSY_TO_CTP:
            s = bs.Busy_CTP_OutputSelect();
            break;
        case BUSY_TO_ALTI:
            s = bs.Busy_ALTI_OutputSelect();
            break;
        default:
            break;
    }

    if (s == "Inactive") {
        src = BUSY_INACTIVE;
    }
    else if (s == "Local") {
        src = BUSY_LOCAL;
    }
    else if (s == "CTP") {
        src = BUSY_CTP;
    }
    else if (s == "ALTI") {
        src = BUSY_ALTI;
    }
    else {
        CERR("ALTI: unknown busy source", "");
        m_status = -1;
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYOutputSelectWrite(const BUSY_OUTPUT out, const BUSY_SOURCE src) { // 1 output
    ALTI_BSY_CONTROL_BITSTRING bs;

    // read
    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }

    std::string s;
    switch (out) {
        case BUSY_TO_CTP:
            // modify
            switch (src) {
                case BUSY_INACTIVE:
                    bs.Busy_CTP_OutputSelect("Inactive");
                    break;
                case BUSY_LOCAL:
                    bs.Busy_CTP_OutputSelect("Local");
                    break;
                case BUSY_CTP:
                    bs.Busy_CTP_OutputSelect("CTP");
                    break;
                case BUSY_ALTI:
                    bs.Busy_CTP_OutputSelect("ALTI");
                    break;
                default:
                    break;
            }
            break;
        case BUSY_TO_ALTI:
            // modify
            switch (src) {
                case BUSY_INACTIVE:
                    bs.Busy_ALTI_OutputSelect("Inactive");
                    break;
                case BUSY_LOCAL:
                    bs.Busy_ALTI_OutputSelect("Local");
                    break;
                case BUSY_CTP:
                    bs.Busy_ALTI_OutputSelect("CTP");
                    break;
                case BUSY_ALTI:
                    bs.Busy_ALTI_OutputSelect("ALTI");
                    break;
                default:
                    break;
            }
            break;
        default:
            break;
    }

    // write
    if ((m_status = m_alti->BSY_Control_Write(bs)) != 0) {
        CERR("ALTI: writing busy control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYOutputSelectRead(std::vector<BUSY_SOURCE> &src) { // all outputs
    ALTI_BSY_CONTROL_BITSTRING bs;

    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }

    src.clear();

    std::string s;
    unsigned int i;
    for (i = 0; i < BUSY_OUTPUT_NUMBER; i++) {
        switch ((BUSY_OUTPUT) i) {
            case BUSY_TO_CTP:
                s = bs.Busy_CTP_OutputSelect();
                break;
            case BUSY_TO_ALTI:
                s = bs.Busy_ALTI_OutputSelect();
                break;
            default:
                break;
        }

        if (s == "Inactive") {
            src.push_back(BUSY_INACTIVE);
        }
        else if (s == "Local") {
            src.push_back(BUSY_LOCAL);
        }
        else if (s == "CTP") {
            src.push_back(BUSY_CTP);
        }
        else if (s == "ALTI") {
            src.push_back(BUSY_ALTI);
        }
        else {
            CERR("ALTI: unknown busy source", "");
            m_status = -1;
            return (m_status);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYOutputSelectWrite(const std::vector<BUSY_SOURCE> &src) { // all outputs
    ALTI_BSY_CONTROL_BITSTRING bs;

    // read
    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }

    unsigned int i;
    for (i = 0; i < BUSY_OUTPUT_NUMBER; i++) {
        switch ((BUSY_OUTPUT) i) {
            case BUSY_TO_CTP:
                // modify
                switch (src[i]) {
                    case BUSY_INACTIVE:
                        bs.Busy_CTP_OutputSelect("Inactive");
                        break;
                    case BUSY_LOCAL:
                        bs.Busy_CTP_OutputSelect("Local");
                        break;
                    case BUSY_CTP:
                        bs.Busy_CTP_OutputSelect("CTP");
                        break;
                    case BUSY_ALTI:
                        bs.Busy_CTP_OutputSelect("ALTI");
                        break;
                    default:
                        break;
                }
                break;
            case BUSY_TO_ALTI:
                // modify
                switch (src[i]) {
                    case BUSY_INACTIVE:
                        bs.Busy_ALTI_OutputSelect("Inactive");
                        break;
                    case BUSY_LOCAL:
                        bs.Busy_ALTI_OutputSelect("Local");
                        break;
                    case BUSY_CTP:
                        bs.Busy_ALTI_OutputSelect("CTP");
                        break;
                    case BUSY_ALTI:
                        bs.Busy_ALTI_OutputSelect("ALTI");
                        break;
                    default:
                        break;
                }
                break;
            default:
                break;
        }
    }

    // write
    if ((m_status = m_alti->BSY_Control_Write(bs)) != 0) {
        CERR("ALTI: writing busy control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYConstLevelRegRead(bool &busy) {
    ALTI_BSY_DATAVME_BITSTRING bs;

    if ((m_status = m_alti->BSY_DataVME_Read(bs)) != 0) {
        CERR("ALTI: reading busy constant level register", "");
        return (m_status);
    }

    busy = bs.Busy();

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYConstLevelRegWrite(const bool busy) {
    ALTI_BSY_DATAVME_BITSTRING bs;

    // read
    if ((m_status = m_alti->BSY_DataVME_Read(bs)) != 0) {
        CERR("ALTI: reading busy constant level register", "");
        return (m_status);
    }

    // modify
    bs.Busy(busy);

    // write
    if ((m_status = m_alti->BSY_DataVME_Write(bs)) != 0) {
        CERR("ALTI: writing busy constant level register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYMaskingL1aEnableRead(bool &enabled) {
    ALTI_BSY_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }

    std::string s = bs.Busy_Masking_L1A_Enable();
    if (s == "Enabled") {
        enabled = true;
    }
    else if (s == "Disabled") {
        enabled = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for busy masking of L1A enable", s.c_str());
        return (-1);
    }
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYMaskingL1aEnableWrite(const bool enable) {
    ALTI_BSY_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }
    // MODIFY
    bs.Busy_Masking_L1A_Enable(enable ? "Enabled" : "Disabled");
    // WRITE
    if ((m_status = m_alti->BSY_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing busy control register", "");
        return (m_status);
    }
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYMaskingOrbEnableWrite(const bool enable) {
    ALTI_BSY_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }
    // MODIFY
    bs.Busy_Masking_Orb_Enable(enable ? "Enabled" : "Disabled");
    // WRITE
    if ((m_status = m_alti->BSY_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing busy control register", "");
        return (m_status);
    }
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYMaskingOrbEnableRead(bool &enabled) {
    ALTI_BSY_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }

    std::string s = bs.Busy_Masking_Orb_Enable();
    if (s == "Enabled") {
        enabled = true;
    }
    else if (s == "Disabled") {
        enabled = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for busy masking of Orbit enable", s.c_str());
        return (-1);
    }
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYMaskingTTREnableWrite(const short ttr,const bool enable) {
    ALTI_BSY_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }
    // MODIFY
    switch (ttr){
    case 1 : bs.Busy_Masking_TTR1_Enable(enable ? "Enabled" : "Disabled");
             break;
    case 2 : bs.Busy_Masking_TTR2_Enable(enable ? "Enabled" : "Disabled");
             break;
    case 3 : bs.Busy_Masking_TTR3_Enable(enable ? "Enabled" : "Disabled");
             break;
    case 4 : bs.Busy_Masking_TTR1_Enable(enable ? "Enabled" : "Disabled");
             bs.Busy_Masking_TTR2_Enable(enable ? "Enabled" : "Disabled");
             bs.Busy_Masking_TTR3_Enable(enable ? "Enabled" : "Disabled");
             break;
    }
    // WRITE
    if ((m_status = m_alti->BSY_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing busy control register", "");
        return (m_status);
    }
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYMaskingTTREnableRead(bool &ttr1, bool &ttr2, bool &ttr3) {
    ALTI_BSY_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }

    std::string s1 = bs.Busy_Masking_TTR1_Enable();
    std::string s2 = bs.Busy_Masking_TTR2_Enable();
    std::string s3 = bs.Busy_Masking_TTR3_Enable();
    if (s1 == "Enabled")
    {
      ttr1 = true; 
    }else 
    {
      ttr1 = false;
    }
    if (s2 == "Enabled")
    {
      ttr2 = true; 
    }else 
    {
      ttr2 = false;
    }
    if (s3 == "Enabled")
    {
      ttr3 = true; 
    }else 
    {
      ttr3 = false;
    }
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYMaskingBGoEnableWrite(const short bgo,const bool enable) {
    ALTI_BSY_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }
    // MODIFY
    switch (bgo){
      
    case 0 : bs.Busy_Masking_BGo0_Enable(enable ? "Enabled" : "Disabled");
             break;
    case 1 : bs.Busy_Masking_BGo1_Enable(enable ? "Enabled" : "Disabled");
             break;
    case 2 : bs.Busy_Masking_BGo2_Enable(enable ? "Enabled" : "Disabled");
             break;
    case 3 : bs.Busy_Masking_BGo3_Enable(enable ? "Enabled" : "Disabled");
             break;
    case 4 : bs.Busy_Masking_BGo2_Enable(enable ? "Enabled" : "Disabled");
             bs.Busy_Masking_BGo1_Enable(enable ? "Enabled" : "Disabled");
             bs.Busy_Masking_BGo3_Enable(enable ? "Enabled" : "Disabled");
             bs.Busy_Masking_BGo0_Enable(enable ? "Enabled" : "Disabled");
             break;
    }
    // WRITE
    if ((m_status = m_alti->BSY_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing busy control register", "");
        return (m_status);
    }
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYMaskingBGoEnableRead(bool &bgo0, bool &bgo1, bool &bgo2, bool &bgo3) {
    ALTI_BSY_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }

    std::string s0 = bs.Busy_Masking_BGo0_Enable();
    std::string s1 = bs.Busy_Masking_BGo1_Enable();
    std::string s2 = bs.Busy_Masking_BGo2_Enable();
    std::string s3 = bs.Busy_Masking_BGo3_Enable();
    if (s0 == "Enabled")
    {
      bgo0 = true; 
    }else 
    {
      bgo0 = false;
    }
    if (s1 == "Enabled")
    {
      bgo1 = true; 
    }else 
    {
      bgo1 = false;
    }
    if (s2 == "Enabled")
    {
      bgo2 = true; 
    }else 
    {
      bgo2 = false;
    }
    if (s3 == "Enabled")
    {
      bgo3 = true; 
    }else 
    {
      bgo3 = false;
    }
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYMaskingTTypeEnableWrite(const bool enable) {
    ALTI_BSY_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }
    // MODIFY
    bs.Busy_Masking_TType_Enable(enable ? "Enabled" : "Disabled");
    // WRITE
    if ((m_status = m_alti->BSY_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing busy control register", "");
        return (m_status);
    }
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYMaskingTTypeEnableRead(bool &enabled) {
    ALTI_BSY_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }

    std::string s = bs.Busy_Masking_TType_Enable();
    if (s == "Enabled") {
        enabled = true;
    }
    else if (s == "Disabled") {
        enabled = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for busy masking of Orbit enable", s.c_str());
        return (-1);
    }
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYFrontPanelLevelRead(BUSY_LEVEL &level) {
  ALTI_BSY_CONTROL_BITSTRING bs;
  
  // READ
  if ((m_status = m_alti->BSY_Control_Read(bs)) != 0 ) {
    CERR("ALTI: reading busy Front-Panel Level register", "");
    return (m_status);
  }
  
  std::string level_str = bs.FrontPanelLevel();
  
  if(level_str == "NIM") level = NIM;
  else if(level_str == "TTL") level = TTL;
  else {
    CERR("ALTI: unknown Busy Front-Panel Level", "");
    m_status = -1;
    return (m_status);
  }
  
  return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYFrontPanelLevelWrite(const BUSY_LEVEL level)
{
  ALTI_BSY_CONTROL_BITSTRING bs;
  // READ
  if ((m_status = m_alti->BSY_Control_Read(bs)) != 0 ) {
    CERR("ALTI: reading busy Front-Panel Level register", "");
    return (m_status);
  }
  
  // MODIFY
  switch (level) {
  case NIM: 
    bs.FrontPanelLevel("NIM");
    break;
  case TTL:
    bs.FrontPanelLevel("TTL");
    break;
  default:
    break;
  }

  // WRITE
  if ((m_status = m_alti->BSY_Control_Write(bs)) != 0 ) {
    CERR("ALTI: writing busy Front-Panel Level register", "");
    return (m_status);
  }
  return (m_status); 
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::CRQInputSelectRead(const unsigned int bit, CALREQ_INPUT &in) {
    if (bit > 2) {
        CERR("ALTI: calibration request bit exceeds [2:0] range", "");
        return (m_status);
    }

    ALTI_CRQ_CONTROL_BITSTRING bs;

    if ((m_status = m_alti->CRQ_Control_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request control register", "");
        return (m_status);
    }

    std::string s = bs.CALREQ_SourceSelect(bit);
    if (s == "RJ45") {
        in = CALREQ_FROM_RJ45;
    }
    if (s == "FrontPanel") {
        in = CALREQ_FROM_FRONT_PANEL;
    }
    else if (s == "CTP") {
        in = CALREQ_FROM_CTP;
    }
    else if (s == "ALTI") {
        in = CALREQ_FROM_ALTI;
    }
    else if (s == "Pattern") {
        in = CALREQ_FROM_PG;
    }
    else if (s == "VME") {
        in = CALREQ_FROM_VME;
    }
    else {
        CERR("ALTI: unknown calibration request input", "");
        m_status = -1;
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CRQInputSelectWrite(const unsigned int bit, const CALREQ_INPUT in) {
    if (bit > 2) {
        CERR("ALTI: calibration request bit exceeds [2:0] range", "");
        return (m_status);
    }

    ALTI_CRQ_CONTROL_BITSTRING bs;

    // read
    if ((m_status = m_alti->CRQ_Control_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request control register", "");
        return (m_status);
    }

    // modify
    switch (in) {
        case CALREQ_FROM_RJ45:
            bs.CALREQ_SourceSelect(bit, "RJ45");
            break;
        case CALREQ_FROM_FRONT_PANEL:
            bs.CALREQ_SourceSelect(bit, "FrontPanel");
            break;
        case CALREQ_FROM_CTP:
            bs.CALREQ_SourceSelect(bit, "CTP");
            break;
        case CALREQ_FROM_ALTI:
            bs.CALREQ_SourceSelect(bit, "ALTI");
            break;
        case CALREQ_FROM_PG:
            bs.CALREQ_SourceSelect(bit, "Pattern");
            break;
        case CALREQ_FROM_VME:
            bs.CALREQ_SourceSelect(bit, "VME");
            break;
        default:
            break;
    }

    // write
    if ((m_status = m_alti->CRQ_Control_Write(bs)) != 0) {
        CERR("ALTI: writing calibration request control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CRQInputSelectRead(std::vector<CALREQ_INPUT> &in) { // all 3 bits

    ALTI_CRQ_CONTROL_BITSTRING bs;

    if ((m_status = m_alti->CRQ_Control_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request control register", "");
        return (m_status);
    }

    in.clear();
    unsigned int bit;
    for (bit = 0; bit < 3; bit++) {
        std::string s = bs.CALREQ_SourceSelect(bit);
        if (s == "RJ45") {
            in.push_back(CALREQ_FROM_RJ45);
        }
        else if (s == "FrontPanel") {
            in.push_back(CALREQ_FROM_FRONT_PANEL);
        }
        else if (s == "CTP") {
            in.push_back(CALREQ_FROM_CTP);
        }
        else if (s == "ALTI") {
            in.push_back(CALREQ_FROM_ALTI);
        }
        else if (s == "Pattern") {
            in.push_back(CALREQ_FROM_PG);
        }
        else if (s == "VME") {
            in.push_back(CALREQ_FROM_VME);
        }
        else {
            CERR("ALTI: unknown calibration request input", "");
            m_status = -1;
            return (m_status);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CRQInputSelectWrite(const std::vector<CALREQ_INPUT> &in) { // all 3 bits

    ALTI_CRQ_CONTROL_BITSTRING bs;

    // read
    if ((m_status = m_alti->CRQ_Control_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request control register", "");
        return (m_status);
    }

    unsigned int bit;
    for (bit = 0; bit < 3; bit++) {
        // modify
        switch (in[bit]) {
            case CALREQ_FROM_RJ45:
                bs.CALREQ_SourceSelect(bit, "RJ45");
                break;
            case CALREQ_FROM_FRONT_PANEL:
                bs.CALREQ_SourceSelect(bit, "FrontPanel");
                break;
            case CALREQ_FROM_CTP:
                bs.CALREQ_SourceSelect(bit, "CTP");
                break;
            case CALREQ_FROM_ALTI:
                bs.CALREQ_SourceSelect(bit, "ALTI");
                break;
            case CALREQ_FROM_PG:
                bs.CALREQ_SourceSelect(bit, "Pattern");
                break;
            case CALREQ_FROM_VME:
                bs.CALREQ_SourceSelect(bit, "VME");
                break;
            default:
                break;
        }
    }

    // write
    if ((m_status = m_alti->CRQ_Control_Write(bs)) != 0) {
        CERR("ALTI: writing calibration request control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CRQInputSelectWrite(const CALREQ_INPUT in) { // all 3 bits

    ALTI_CRQ_CONTROL_BITSTRING bs;

    // read
    if ((m_status = m_alti->CRQ_Control_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request control register", "");
        return (m_status);
    }

    unsigned int bit;
    for (bit = 0; bit < 3; bit++) {
        // modify
        switch (in) {
            case CALREQ_FROM_RJ45:
                bs.CALREQ_SourceSelect(bit, "RJ45");
                break;
            case CALREQ_FROM_FRONT_PANEL:
                bs.CALREQ_SourceSelect(bit, "FrontPanel");
                break;
            case CALREQ_FROM_CTP:
                bs.CALREQ_SourceSelect(bit, "CTP");
                break;
            case CALREQ_FROM_ALTI:
                bs.CALREQ_SourceSelect(bit, "ALTI");
                break;
            case CALREQ_FROM_PG:
                bs.CALREQ_SourceSelect(bit, "Pattern");
                break;
            case CALREQ_FROM_VME:
                bs.CALREQ_SourceSelect(bit, "VME");
                break;
            default:
                break;
        }
    }

    // write
    if ((m_status = m_alti->CRQ_Control_Write(bs)) != 0) {
        CERR("ALTI: writing calibration request control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CRQOutputSelectRead(const CALREQ_OUTPUT out, const unsigned int bit, CALREQ_SOURCE &src) { // 1 output, 1 bit
    if (bit > 2) {
        CERR("ALTI: calibration request bit exceeds [2:0] range", "");
        return (m_status);
    }

    ALTI_CRQ_CONTROL_BITSTRING bs;

    if ((m_status = m_alti->CRQ_Control_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request control register", "");
        return (m_status);
    }

    std::string s;
    switch (out) {
        case CALREQ_TO_CTP:
            s = bs.CALREQ_CTP_OutputSelect(bit);
            break;
        case CALREQ_TO_ALTI:
            s = bs.CALREQ_ALTI_OutputSelect(bit);
            break;
        default:
            break;
    }

    if (s == "Inactive") {
        src = CALREQ_INACTIVE;
    }
    else if (s == "Local") {
        src = CALREQ_LOCAL;
    }
    else if (s == "CTP") {
        src = CALREQ_CTP;
    }
    else if (s == "ALTI") {
        src = CALREQ_ALTI;
    }
    else {
        CERR("ALTI: unknown calibration request source", "");
        m_status = -1;
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CRQOutputSelectWrite(const CALREQ_OUTPUT out, const unsigned int bit, const CALREQ_SOURCE src) { // 1 output, 1 bit
    if (bit > 2) {
        CERR("ALTI: calibration request bit exceeds [2:0] range", "");
        return (m_status);
    }

    ALTI_CRQ_CONTROL_BITSTRING bs;

    // read
    if ((m_status = m_alti->CRQ_Control_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request control register", "");
        return (m_status);
    }

    std::string s;
    switch (out) {
        case CALREQ_TO_CTP:
            // modify
            switch (src) {
                case CALREQ_INACTIVE:
                    bs.CALREQ_CTP_OutputSelect(bit, "Inactive");
                    break;
                case CALREQ_LOCAL:
                    bs.CALREQ_CTP_OutputSelect(bit, "Local");
                    break;
                case CALREQ_CTP:
                    bs.CALREQ_CTP_OutputSelect(bit, "CTP");
                    break;
                case CALREQ_ALTI:
                    bs.CALREQ_CTP_OutputSelect(bit, "ALTI");
                    break;
                default:
                    break;
            }
            break;
        case CALREQ_TO_ALTI:
            // modify
            switch (src) {
                case CALREQ_INACTIVE:
                    bs.CALREQ_ALTI_OutputSelect(bit, "Inactive");
                    break;
                case CALREQ_LOCAL:
                    bs.CALREQ_ALTI_OutputSelect(bit, "Local");
                    break;
                case CALREQ_CTP:
                    bs.CALREQ_ALTI_OutputSelect(bit, "CTP");
                    break;
                case CALREQ_ALTI:
                    bs.CALREQ_ALTI_OutputSelect(bit, "ALTI");
                    break;
                default:
                    break;
            }
            break;
        default:
            break;
    }

    // write
    if ((m_status = m_alti->CRQ_Control_Write(bs)) != 0) {
        CERR("ALTI: writing calibration request control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CRQOutputSelectRead(const CALREQ_OUTPUT out, std::vector<CALREQ_SOURCE> &src) { // 1 output, all 3 bits
    ALTI_CRQ_CONTROL_BITSTRING bs;

    if ((m_status = m_alti->CRQ_Control_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request control register", "");
        return (m_status);
    }

    src.clear();
    unsigned int bit;
    for (bit = 0; bit < 3; bit++) {
        std::string s;
        switch (out) {
            case CALREQ_TO_CTP:
                s = bs.CALREQ_CTP_OutputSelect(bit);
                break;
            case CALREQ_TO_ALTI:
                s = bs.CALREQ_ALTI_OutputSelect(bit);
                break;
            default:
                break;
        }

        if (s == "Inactive") {
            src.push_back(CALREQ_INACTIVE);
        }
        else if (s == "Local") {
            src.push_back(CALREQ_LOCAL);
        }
        else if (s == "CTP") {
            src.push_back(CALREQ_CTP);
        }
        else if (s == "ALTI") {
            src.push_back(CALREQ_ALTI);
        }
        else {
            CERR("ALTI: unknown calibration request source", "");
            m_status = -1;
            return (m_status);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CRQOutputSelectWrite(const CALREQ_OUTPUT out, const std::vector<CALREQ_SOURCE> &src) { // 1 output, all 3 bits
    ALTI_CRQ_CONTROL_BITSTRING bs;

    // read
    if ((m_status = m_alti->CRQ_Control_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request control register", "");
        return (m_status);
    }

    unsigned int bit;
    for (bit = 0; bit < 3; bit++) {
        switch (out) {
            case CALREQ_TO_CTP:
                // modify
                switch (src[bit]) {
                    case CALREQ_INACTIVE:
                        bs.CALREQ_CTP_OutputSelect(bit, "Inactive");
                        break;
                    case CALREQ_LOCAL:
                        bs.CALREQ_CTP_OutputSelect(bit, "Local");
                        break;
                    case CALREQ_CTP:
                        bs.CALREQ_CTP_OutputSelect(bit, "CTP");
                        break;
                    case CALREQ_ALTI:
                        bs.CALREQ_CTP_OutputSelect(bit, "ALTI");
                        break;
                    default:
                        break;
                }
                break;
            case CALREQ_TO_ALTI:
                // modify
                switch (src[bit]) {
                    case CALREQ_INACTIVE:
                        bs.CALREQ_ALTI_OutputSelect(bit, "Inactive");
                        break;
                    case CALREQ_LOCAL:
                        bs.CALREQ_ALTI_OutputSelect(bit, "Local");
                        break;
                    case CALREQ_CTP:
                        bs.CALREQ_ALTI_OutputSelect(bit, "CTP");
                        break;
                    case CALREQ_ALTI:
                        bs.CALREQ_ALTI_OutputSelect(bit, "ALTI");
                        break;
                    default:
                        break;
                }
                break;
            default:
                break;
        }
    }

    // write
    if ((m_status = m_alti->CRQ_Control_Write(bs)) != 0) {
        CERR("ALTI: writing calibration request control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CRQOutputSelectRead(const unsigned int bit, std::vector<CALREQ_SOURCE> &src) { // all outputs, 1 bit
    if (bit > 2) {
        CERR("ALTI: calibration request bit exceeds [2:0] range", "");
        return (m_status);
    }

    ALTI_CRQ_CONTROL_BITSTRING bs;

    if ((m_status = m_alti->CRQ_Control_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request control register", "");
        return (m_status);
    }

    std::string s_ctp, s_alti;
    s_ctp = bs.CALREQ_CTP_OutputSelect(bit);
    s_alti = bs.CALREQ_ALTI_OutputSelect(bit);

    src.clear();

    if (s_ctp == "Inactive") {
        src.push_back(CALREQ_INACTIVE);
    }
    else if (s_ctp == "Local") {
        src.push_back(CALREQ_LOCAL);
    }
    else if (s_ctp == "CTP") {
        src.push_back(CALREQ_CTP);
    }
    else if (s_ctp == "ALTI") {
        src.push_back(CALREQ_ALTI);
    }
    else {
        CERR("ALTI: unknown calibration request source", "");
        m_status = -1;
        return (m_status);
    }

    if (s_alti == "Inactive") {
        src.push_back(CALREQ_INACTIVE);
    }
    else if (s_alti == "Local") {
        src.push_back(CALREQ_LOCAL);
    }
    else if (s_alti == "CTP") {
        src.push_back(CALREQ_CTP);
    }
    else if (s_alti == "ALTI") {
        src.push_back(CALREQ_ALTI);
    }
    else {
        CERR("ALTI: unknown calibration request source", "");
        m_status = -1;
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CRQOutputSelectWrite(const unsigned int bit, const std::vector<CALREQ_SOURCE> &src) { // all outputs, 1 bit
    if (bit > 2) {
        CERR("ALTI: calibration request bit exceeds [2:0] range", "");
        return (m_status);
    }

    ALTI_CRQ_CONTROL_BITSTRING bs;

    // read
    if ((m_status = m_alti->CRQ_Control_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request control register", "");
        return (m_status);
    }

    // modify
    switch (src[0]) {
        case CALREQ_INACTIVE:
            bs.CALREQ_CTP_OutputSelect(bit, "Inactive");
            break;
        case CALREQ_LOCAL:
            bs.CALREQ_CTP_OutputSelect(bit, "Local");
            break;
       case CALREQ_CTP:
            bs.CALREQ_CTP_OutputSelect(bit, "CTP");
            break;
        case CALREQ_ALTI:
            bs.CALREQ_CTP_OutputSelect(bit, "ALTI");
            break;
        default:
            break;
    }
    // modify
    switch (src[1]) {
        case CALREQ_INACTIVE:
            bs.CALREQ_ALTI_OutputSelect(bit, "Inactive");
            break;
        case CALREQ_LOCAL:
            bs.CALREQ_ALTI_OutputSelect(bit, "Local");
            break;
        case CALREQ_CTP:
            bs.CALREQ_ALTI_OutputSelect(bit, "CTP");
            break;
        case CALREQ_ALTI:
            bs.CALREQ_ALTI_OutputSelect(bit, "ALTI");
            break;
        default:
            break;
    }

    // write
    if ((m_status = m_alti->CRQ_Control_Write(bs)) != 0) {
        CERR("ALTI: writing calibration request control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CRQOutputSelectWrite(const CALREQ_OUTPUT out, const CALREQ_SOURCE src) { // 1 output, all bits
    ALTI_CRQ_CONTROL_BITSTRING bs;

    // read
    if ((m_status = m_alti->CRQ_Control_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request control register", "");
        return (m_status);
    }

    unsigned int bit;
    for (bit = 0; bit < 3; bit++) {
        switch (out) {
            case CALREQ_TO_CTP:
                // modify
                switch (src) {
                    case CALREQ_INACTIVE:
                        bs.CALREQ_CTP_OutputSelect(bit, "Inactive");
                        break;
                    case CALREQ_LOCAL:
                        bs.CALREQ_CTP_OutputSelect(bit, "Local");
                        break;
                    case CALREQ_CTP:
                        bs.CALREQ_CTP_OutputSelect(bit, "CTP");
                        break;
                    case CALREQ_ALTI:
                        bs.CALREQ_CTP_OutputSelect(bit, "ALTI");
                        break;
                    default:
                        break;
                }
                break;
            case CALREQ_TO_ALTI:
                // modify
                switch (src) {
                    case CALREQ_INACTIVE:
                        bs.CALREQ_ALTI_OutputSelect(bit, "Inactive");
                        break;
                    case CALREQ_LOCAL:
                        bs.CALREQ_ALTI_OutputSelect(bit, "Local");
                        break;
                    case CALREQ_CTP:
                        bs.CALREQ_ALTI_OutputSelect(bit, "CTP");
                        break;
                    case CALREQ_ALTI:
                        bs.CALREQ_ALTI_OutputSelect(bit, "ALTI");
                        break;
                    default:
                        break;
                }
                break;
            default:
                break;
        }
    }

    // write
    if ((m_status = m_alti->CRQ_Control_Write(bs)) != 0) {
        CERR("ALTI: writing calibration request control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CRQOutputSelectWrite(const unsigned int bit, const CALREQ_SOURCE src) { // all outputs, 1 bit
    if (bit > 2) {
        CERR("ALTI: calibration request bit exceeds [2:0] range", "");
        return (m_status);
    }

    ALTI_CRQ_CONTROL_BITSTRING bs;

    // read
    if ((m_status = m_alti->CRQ_Control_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request control register", "");
        return (m_status);
    }

    // modify
    switch (src) {
        case CALREQ_INACTIVE:
            bs.CALREQ_CTP_OutputSelect(bit, "Inactive");
            break;
        case CALREQ_LOCAL:
            bs.CALREQ_CTP_OutputSelect(bit, "Local");
            break;
       case CALREQ_CTP:
            bs.CALREQ_CTP_OutputSelect(bit, "CTP");
            break;
        case CALREQ_ALTI:
            bs.CALREQ_CTP_OutputSelect(bit, "ALTI");
            break;
        default:
            break;
    }
    // modify
    switch (src) {
        case CALREQ_INACTIVE:
            bs.CALREQ_ALTI_OutputSelect(bit, "Inactive");
            break;
        case CALREQ_LOCAL:
            bs.CALREQ_ALTI_OutputSelect(bit, "Local");
            break;
        case CALREQ_CTP:
            bs.CALREQ_ALTI_OutputSelect(bit, "CTP");
            break;
        case CALREQ_ALTI:
            bs.CALREQ_ALTI_OutputSelect(bit, "ALTI");
            break;
        default:
            break;
    }

    // write
    if ((m_status = m_alti->CRQ_Control_Write(bs)) != 0) {
        CERR("ALTI: writing calibration request control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CRQOutputSelectWrite(const CALREQ_SOURCE src) { // all outputs, all bits
    ALTI_CRQ_CONTROL_BITSTRING bs;

    // read
    if ((m_status = m_alti->CRQ_Control_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request control register", "");
        return (m_status);
    }

    unsigned int bit;
    for (bit = 0; bit < 3; bit++) {
        // modify
        switch (src) {
            case CALREQ_INACTIVE:
                bs.CALREQ_CTP_OutputSelect(bit, "Inactive");
                break;
            case CALREQ_LOCAL:
                bs.CALREQ_CTP_OutputSelect(bit, "Local");
                break;
            case CALREQ_CTP:
                bs.CALREQ_CTP_OutputSelect(bit, "CTP");
                break;
            case CALREQ_ALTI:
                bs.CALREQ_CTP_OutputSelect(bit, "ALTI");
                break;
            default:
                break;
        }
        // modify
        switch (src) {
            case CALREQ_INACTIVE:
                bs.CALREQ_ALTI_OutputSelect(bit, "Inactive");
                break;
            case CALREQ_LOCAL:
                bs.CALREQ_ALTI_OutputSelect(bit, "Local");
                break;
            case CALREQ_CTP:
                bs.CALREQ_ALTI_OutputSelect(bit, "CTP");
                break;
            case CALREQ_ALTI:
                bs.CALREQ_ALTI_OutputSelect(bit, "ALTI");
                break;
            default:
                break;
        }
    }

    // write
    if ((m_status = m_alti->CRQ_Control_Write(bs)) != 0) {
        CERR("ALTI: writing calibration request control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CRQConstLevelRegRead(unsigned int &creq) {
    ALTI_CRQ_DATAVME_BITSTRING bs;

    if ((m_status = m_alti->CRQ_DataVME_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request constant level register", "");
        return (m_status);
    }

    creq = bs.CALREQ();

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CRQConstLevelRegWrite(const unsigned int creq) {
    ALTI_CRQ_DATAVME_BITSTRING bs;

    // read
    if ((m_status = m_alti->CRQ_DataVME_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request constant level register", "");
        return (m_status);
    }

    // modify
    bs.CALREQ(creq);

    // write
    if ((m_status = m_alti->CRQ_DataVME_Write(bs)) != 0) {
        CERR("ALTI: writing calibration request constant level register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CRQInputRead(const CALREQ_INPUT in, unsigned int &creq) { // 1 value
    ALTI_CRQ_STATUS_BITSTRING bs;

    if ((m_status = m_alti->CRQ_Status_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request status register", "");
        return (m_status);
    }

    switch (in) {
        case CALREQ_FROM_RJ45:
            creq = bs.RJ45();
            break;
        case CALREQ_FROM_FRONT_PANEL:
            creq = bs.FrontPanel();
            break;
        case CALREQ_FROM_CTP:
            creq = bs.CTP();
            break;
        case CALREQ_FROM_ALTI:
            creq = bs.ALTI();
            break;
        case CALREQ_FROM_PG:
            creq = bs.Pattern();
            break;
        case CALREQ_FROM_VME:
            creq = bs.VME();
            break;
        default:
            break;
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CRQInputRead(std::vector<unsigned int> &creq) { // all values
    ALTI_CRQ_STATUS_BITSTRING bs;

    if ((m_status = m_alti->CRQ_Status_Read(bs)) != 0) {
        CERR("ALTI: reading calibration request status register", "");
        return (m_status);
    }

    creq.clear();

    creq.push_back(bs.RJ45());
    creq.push_back(bs.FrontPanel());
    creq.push_back(bs.CTP());
    creq.push_back(bs.ALTI());
    creq.push_back(bs.Pattern());
    creq.push_back(bs.VME());

    return (m_status);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::SNPEnableRead(bool &enabled) {
    ALTI_PAT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->PAT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading pattern/snapshot memory control register", "");
        return (m_status);
    }

    std::string s = bs.EnableSnapshot();
    if (s == "Enabled") {
        enabled = true;
    }
    else if (s == "Disabled") {
        enabled = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for snapshot enable", s.c_str());
        return (-1);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SNPEnableWrite(const bool enable) {
    bool pat_enabled;
    if ((m_status = PATGenerationEnableRead(pat_enabled)) != SUCCESS) {
        CERR("ALTI: reading PAT enable", "");
        return (m_status);
    }
    if (enable & pat_enabled) {
        CERR("ALTI: enabling snapshot memory while it is in pattern generation mode", "");
        m_status = ALTI_RAM_MEMORY_BUSY;
        return (m_status);
    }

    ALTI_PAT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->PAT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading pattern/snapshot memory control register", "");
        return (m_status);
    }
    // MODIFY
    bs.EnableSnapshot(enable ? "Enabled" : "Disabled");
    // WRITE
    if ((m_status = m_alti->PAT_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing pattern/snapshot memory control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SNPCurrentAddressRead(unsigned int &cur) {
    if ((m_status = m_alti->PAT_RAMCurrentSPYAddress_Read(cur)) != 0) {
        CERR("ALTI: reading snapshot current address", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SNPMaskRead(unsigned int &mask) {
    if ((m_status = m_alti->PAT_SPYMask_Read(mask)) != 0) {
        CERR("ALTI: reading snapshot mask register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SNPMaskWrite(const unsigned int mask) {
    bool snp_enabled;
    if ((m_status = SNPEnableRead(snp_enabled)) != SUCCESS) {
        CERR("ALTI: reading SNP enable", "");
        return (m_status);
    }
    if (snp_enabled) {
        CERR("ALTI: writing snapshot mask while snapshot is enabled", "");
        m_status = ALTI_RAM_MEMORY_BUSY;
        return (m_status);
    }

    if ((m_status = m_alti->PAT_SPYMask_Write(mask)) != 0) {
        CERR("ALTI: writing snapshot mask register", "");
        return (m_status);
    }
    // check that the register was properly written to
    unsigned int readout;
    if ((m_status = m_alti->PAT_SPYMask_Read(readout)) != 0) {
        CERR("ALTI: reading snapshot mask register", "");
        return (m_status);
    }
    if (readout != mask) {
        CERR("ALTI: wrote snapshot mask 0x%08x, but read 0x%08x", mask, readout);
        m_status = -1;
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SNPMaskRead(const SIGNAL signal, bool &masked) { // 1 signal
    unsigned int mask;
    m_status = SNPMaskRead(mask);

    // 0: masked, 1: not masked
    masked = !(mask & MEM_MASK[signal]);

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SNPMaskWrite(const SIGNAL signal, const bool mask) { // 1 signal
    unsigned int status;
    m_status = SNPMaskRead(status);

    // 0: masked, 1: not masked
    status = (mask ? (status & (~MEM_MASK[signal])) : (status | MEM_MASK[signal]));

    m_status = SNPMaskWrite(status);

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SNPMaskRead(std::vector<bool> &masked) { // all signals
    unsigned int mask;
    m_status = SNPMaskRead(mask);

    masked.clear();

    unsigned int i;
    for (i = 0; i < SIGNAL_NUMBER; i++) {
        // 0: masked, 1: not masked
        masked.push_back(!(mask & MEM_MASK[i]));
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SNPMaskWrite(const std::vector<bool> &masked) { // all signals
    unsigned int status;
    m_status = SNPMaskRead(status);

    unsigned int i;
    for (i = 0; i < SIGNAL_NUMBER; i++) {
        // 0: masked, 1: not masked
        status = (masked[i] ? (status & (~MEM_MASK[i])) : (status | MEM_MASK[i]));
    }

    m_status = SNPMaskWrite(status);

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SNPMaskWrite(const bool mask) { // all signals
    unsigned int status;
    m_status = SNPMaskRead(status);

    // 0: masked, 1: not masked
    status = (mask ? (status & (~MEM_SIGNAL_MASK)) : (status | MEM_SIGNAL_MASK));

    m_status = SNPMaskWrite(status);

    return (m_status);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::PATGenerationEnableRead(bool &enabled) { // global PAT enable bit
    ALTI_PAT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->PAT_Control_Read(bs)) != 0) {
        CERR("ALTI: reading pattern/snapshot memory control register", "");
        return (m_status);
    }
    std::string s = bs.EnableGeneration();
    if (s == "Enabled") {
        enabled = true;
    }
    else if (s == "Disabled") {
        enabled = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for pattern generation enable", s.c_str());
        return (-1);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::PATGenerationEnableWrite(const bool enable) { // global PAT enable bit
    bool snp_enabled;
    if ((m_status = SNPEnableRead(snp_enabled)) != SUCCESS) {
        CERR("ALTI: reading SNP enable", "");
        return (m_status);
    }
    if (enable & snp_enabled) {
        CERR("ALTI: enabling pattern generation memory while it is in shapshot mode", "");
        m_status = ALTI_RAM_MEMORY_BUSY;
        return (m_status);
    }

    ALTI_PAT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->PAT_Control_Read(bs)) != 0) {
        CERR("ALTI: reading pattern/snapshot memory control register", "");
        return (m_status);
    }
    // MODIFY
    bs.EnableGeneration(enable ? "Enabled" : "Disabled");
    // WRITE
    if ((m_status = m_alti->PAT_Control_Write(bs)) != 0) {
        CERR("ALTI: writing pattern/snapshot memory control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::PATGenerationConfigRead(const SIGNAL_PG signal, bool &enabled) {
    ALTI_PAT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->PAT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading pattern/snapshot memory control register", "");
        return (m_status);
    }

    std::string s;
    switch (signal) {
        case PG_ORB:
            s = bs.ORB_PG();
            if (s == "Enabled") {
                enabled = true;
            }
            else if (s == "Disabled") {
                enabled = false;
            }
            else {
                CERR("ALTI: unknown value \"%s\" for \"%s\" pattern generation enable", s.c_str(), SIGNAL_PG_NAME[signal].c_str());
                return (-1);
            }
            break;
        case PG_L1A:
            s = bs.L1A_PG();
            if (s == "Enabled") {
                enabled = true;
            }
            else if (s == "Disabled") {
                enabled = false;
            }
            else {
                CERR("ALTI: unknown value \"%s\" for \"%s\" pattern generation enable", s.c_str(), SIGNAL_PG_NAME[signal].c_str());
                return (-1);
            }
            break;
        case PG_TTR1:
        case PG_TTR2:
        case PG_TTR3:
            s = bs.TestTrigger_PG(signal - PG_TTR1);
            if (s == "Enabled") {
                enabled = true;
            }
            else if (s == "Disabled") {
                enabled = false;
            }
            else {
                CERR("ALTI: unknown value \"%s\" for \"%s\" pattern generation enable", s.c_str(), SIGNAL_PG_NAME[signal].c_str());
                return (-1);
            }
            break;
        case PG_TTYP:
            s = bs.TriggerTypeWord_PG();
            if (s == "Enabled") {
                enabled = true;
            }
            else if (s == "Disabled") {
                enabled = false;
            }
            else {
                CERR("ALTI: unknown value \"%s\" for \"%s\" pattern generation enable", s.c_str(), SIGNAL_PG_NAME[signal].c_str());
                return (-1);
            }
            break;
        case PG_BGO0:
        case PG_BGO1:
        case PG_BGO2:
        case PG_BGO3:
            s = bs.BGo_PG(signal - PG_BGO0);
            if (s == "Enabled") {
                enabled = true;
            }
            else if (s == "Disabled") {
                enabled = false;
            }
            else {
                CERR("ALTI: unknown value \"%s\" for \"%s\" pattern generation enable", s.c_str(), SIGNAL_PG_NAME[signal].c_str());
                return (-1);
            }
            break;
        default:
            break;
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::PATGenerationConfigWrite(const SIGNAL_PG signal, const bool enable) {
    bool pat_enabled;
    if ((m_status = PATGenerationEnableRead(pat_enabled)) != SUCCESS) {
        CERR("ALTI: reading PAT generation enable", "");
        return (m_status);
    }
    if (pat_enabled) {
        CERR("ALTI: pattern generation memory is enabled, configuring anyway...", "");
        m_status = ALTI_RAM_MEMORY_BUSY;
        //return (m_status); // Prints an error message but continues
    }

    ALTI_PAT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->PAT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading pattern/snapshot memory control register", "");
        return (m_status);
    }
    // MODIFY
    std::string s = (enable ? "Enabled" : "Disabled");
    switch (signal) {
        case PG_ORB:
            bs.ORB_PG(s);
	    break;
        case PG_L1A:
            bs.L1A_PG(s);
            break;
        case PG_TTR1:
        case PG_TTR2:
        case PG_TTR3:
            bs.TestTrigger_PG(signal - PG_TTR1, s);
            break;
        case PG_TTYP:
            bs.TriggerTypeWord_PG(s);
            break;
        case PG_BGO0:
        case PG_BGO1:
        case PG_BGO2:
        case PG_BGO3:
            bs.BGo_PG(signal - PG_BGO0, s);
            break;
        default:
            break;
    }
    // WRITE
    if ((m_status = m_alti->PAT_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing pattern/snapshot memory control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::PATGenerationConfigRead(std::vector<bool> &enabled) {
    ALTI_PAT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->PAT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading pattern/snapshot memory control register", "");
        return (m_status);
    }

    enabled.clear();

    std::string s;
    SIGNAL_PG signal;
    unsigned int i;
    for (i = 0; i < SIGNAL_PG_NUMBER; i++) {
        signal = (SIGNAL_PG) i;
        switch (signal) {
            case PG_ORB:
                s = bs.ORB_PG();
                if (s == "Enabled") {
                    enabled.push_back(true);
                }
                else if (s == "Disabled") {
                    enabled.push_back(false);
                }
                else {
                    CERR("ALTI: unknown value \"%s\" for \"%s\" pattern generation enable", s.c_str(), SIGNAL_PG_NAME[signal].c_str());
                    return (-1);
                }
                break;
            case PG_L1A:
                s = bs.L1A_PG();
                if (s == "Enabled") {
                    enabled.push_back(true);
                }
                else if (s == "Disabled") {
                    enabled.push_back(false);
                }
                else {
                    CERR("ALTI: unknown value \"%s\" for \"%s\" pattern generation enable", s.c_str(), SIGNAL_PG_NAME[signal].c_str());
                    return (-1);
                }
                break;
            case PG_TTR1:
            case PG_TTR2:
            case PG_TTR3:
                s = bs.TestTrigger_PG(signal - PG_TTR1);
                if (s == "Enabled") {
                    enabled.push_back(true);
                }
                else if (s == "Disabled") {
                    enabled.push_back(false);
                }
                else {
                    CERR("ALTI: unknown value \"%s\" for \"%s\" pattern generation enable", s.c_str(), SIGNAL_PG_NAME[signal].c_str());
                    return (-1);
                }
                break;
            case PG_TTYP:
                s = bs.TriggerTypeWord_PG();
                if (s == "Enabled") {
                    enabled.push_back(true);
                }
                else if (s == "Disabled") {
                    enabled.push_back(false);
                }
                else {
                    CERR("ALTI: unknown value \"%s\" for \"%s\" pattern generation enable", s.c_str(), SIGNAL_PG_NAME[signal].c_str());
                    return (-1);
                }
                break;
            case PG_BGO0:
            case PG_BGO1:
            case PG_BGO2:
            case PG_BGO3:
                s = bs.BGo_PG(signal - PG_BGO0);
                if (s == "Enabled") {
                    enabled.push_back(true);
                }
                else if (s == "Disabled") {
                    enabled.push_back(false);
                }
                else {
                    CERR("ALTI: unknown value \"%s\" for \"%s\" pattern generation enable", s.c_str(), SIGNAL_PG_NAME[signal].c_str());
                    return (-1);
                }
                break;
            default:
                break;
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::PATGenerationConfigWrite(const std::vector<bool> &enable) {
    bool pat_enabled;
    if ((m_status = PATGenerationEnableRead(pat_enabled)) != SUCCESS) {
        CERR("ALTI: reading PAT generation enable", "");
        return (m_status);
    }
    if (pat_enabled) {
        CERR("ALTI: configuring pattern generation memory while it is enabled", "");
        m_status = ALTI_RAM_MEMORY_BUSY;
        return (m_status);
    }

    ALTI_PAT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->PAT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading pattern/snapshot memory control register", "");
        return (m_status);
    }
    // MODIFY
    std::string s;
    SIGNAL_PG signal;
    unsigned int i;
    for (i = 0; i < SIGNAL_PG_NUMBER; i++) {
        signal = (SIGNAL_PG) i;
        s = (enable[i] ? "Enabled" : "Disabled");
        switch (signal) {
            case PG_ORB:
                bs.ORB_PG(s);
            case PG_L1A:
                bs.L1A_PG(s);
                break;
            case PG_TTR1:
            case PG_TTR2:
            case PG_TTR3:
                bs.TestTrigger_PG(signal - PG_TTR1, s);
                break;
            case PG_TTYP:
                bs.TriggerTypeWord_PG(s);
                break;
            case PG_BGO0:
            case PG_BGO1:
            case PG_BGO2:
            case PG_BGO3:
                bs.BGo_PG(signal - PG_BGO0, s);
                break;
            default:
                break;
        }
    }
    // WRITE
    if ((m_status = m_alti->PAT_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing pattern/snapshot memory control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::PATGenerationConfigWrite(const bool enable) {
    bool pat_enabled;
    if ((m_status = PATGenerationEnableRead(pat_enabled)) != SUCCESS) {
        CERR("ALTI: reading PAT generation enable", "");
        return (m_status);
    }
    if (pat_enabled) {
        CERR("ALTI: configuring pattern generation memory while it is enabled", "");
        m_status = ALTI_RAM_MEMORY_BUSY;
        return (m_status);
    }

    ALTI_PAT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->PAT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading pattern/snapshot memory control register", "");
        return (m_status);
    }
    // MODIFY
    std::string s;
    s = (enable ? "Enabled" : "Disabled");
    SIGNAL_PG signal;
    unsigned int i;
    for (i = 0; i < SIGNAL_PG_NUMBER; i++) {
        signal = (SIGNAL_PG) i;
        switch (signal) {
            case PG_ORB:
                bs.ORB_PG(s);
            case PG_L1A:
                bs.L1A_PG(s);
                break;
            case PG_TTR1:
            case PG_TTR2:
            case PG_TTR3:
                bs.TestTrigger_PG(signal - PG_TTR1, s);
                break;
            case PG_TTYP:
                bs.TriggerTypeWord_PG(s);
                break;
            case PG_BGO0:
            case PG_BGO1:
            case PG_BGO2:
            case PG_BGO3:
                bs.BGo_PG(signal - PG_BGO0, s);
                break;
            default:
                break;
        }
    }
    // WRITE
    if ((m_status = m_alti->PAT_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing pattern/snapshot memory control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::PATGenerationRepeatRead(bool &repeated) {
    ALTI_PAT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->PAT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading pattern/snapshot memory control register", "");
        return (m_status);
    }
    std::string s;
    s = bs.RepeatedGeneration();
    if (s == "Enabled") {
        repeated = true;
    }
    else if (s == "Disabled") {
        repeated = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for pattern repeated generation enable", s.c_str());
        return (-1);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::PATGenerationRepeatWrite(const bool repeated) {
    ALTI_PAT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->PAT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading pattern/snapshot memory control register", "");
        return (m_status);
    }
    // MODIFY
    bs.RepeatedGeneration(repeated ? "Enabled" : "Disabled");
    // WRITE
    if ((m_status = m_alti->PAT_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing pattern/snapshot memory control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::PATGenerationStartAddressRead(unsigned int &start) {
    if ((m_status = m_alti->PAT_RAMStartFirstAddress_Read(start)) != 0) {
        CERR("ALTI: reading pattern start address", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::PATGenerationStartAddressWrite(const unsigned int start) {
    bool pat_enabled;
    if ((m_status = PATGenerationEnableRead(pat_enabled)) != SUCCESS) {
        CERR("ALTI: reading PAT generation enable", "");
        return (m_status);
    }
    if (pat_enabled) {
        CERR("ALTI: configuring pattern generation memory while it is enabled", "");
        m_status = ALTI_RAM_MEMORY_BUSY;
        return (m_status);
    }

    if ((m_status = m_alti->PAT_RAMStartFirstAddress_Write(start)) != 0) {
        CERR("ALTI: writing pattern start address", "");
        return (m_status);
    }
    // check that the register was properly written to
    unsigned int readout;
    if ((m_status = m_alti->PAT_RAMStartFirstAddress_Read(readout)) != 0) {
        CERR("ALTI: reading pattern start address", "");
        return (m_status);
    }
    if (readout != start) {
        CERR("ALTI: wrote pattern start address 0x%08x, but read 0x%08x", start, readout);
        m_status = -1;
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::PATGenerationStopAddressRead(unsigned int &stop) {
    if ((m_status = m_alti->PAT_RAMStopLastAddress_Read(stop)) != 0) {
        CERR("ALTI: reading pattern stop address", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::PATGenerationStopAddressWrite(const unsigned int stop) {
    bool pat_enabled;
    if ((m_status = PATGenerationEnableRead(pat_enabled)) != SUCCESS) {
        CERR("ALTI: reading PAT generation enable", "");
        return (m_status);
    }
    if (pat_enabled) {
        CERR("ALTI: configuring pattern generation memory while it is enabled", "");
        m_status = ALTI_RAM_MEMORY_BUSY;
        return (m_status);
    }

    if ((m_status = m_alti->PAT_RAMStopLastAddress_Write(stop)) != 0) {
        CERR("ALTI: writing pattern stop address", "");
        return (m_status);
    }
    // check that the register was properly written to
    unsigned int readout;
    if ((m_status = m_alti->PAT_RAMStopLastAddress_Read(readout)) != 0) {
        CERR("ALTI: reading pattern stop address", "");
        return (m_status);
    }
    if (readout != stop) {
        CERR("ALTI: wrote pattern stop address 0x%08x, but read 0x%08x", stop, readout);
        m_status = -1;
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::MEMDataRead(const ALTI_RAM_MEMORY mem, unsigned int &data, const unsigned int vme_off) { // number
    bool pat_enabled, snp_enabled, dec_enabled;
    // write the corresponding page number to the ADER1 page register
    if ((m_status = CSRADER1Write(mem + 1)) != SUCCESS) {
        CERR("ALTI: setting up ADER1 page register for accessing \"%s\" memory", ALTI_RAM_MEMORY_NAME[mem].c_str());
        return (m_status);
    }
    switch (mem) {
        case PATTERN_SNAPSHOT:
            if ((m_status = PATGenerationEnableRead(pat_enabled)) != SUCCESS) {
                CERR("ALTI: reading PAT generation enable", "");
                return (m_status);
            }
            if ((m_status = SNPEnableRead(snp_enabled)) != SUCCESS) {
                CERR("ALTI: reading SNP enable", "");
                return (m_status);
            }
            if (pat_enabled | snp_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Read(vme_off, data)) != 0) {
                CERR("ALTI: reading single entry in \"%s\" memory, index 0x%08x", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off);
                return (m_status);
            }
            break;
        case TTC_DECODER_COMMAND:
            if ((m_status = DECEnableRead(dec_enabled)) != SUCCESS) {
                CERR("ALTI: reading DEC enable", "");
                return (m_status);
            }
            if (dec_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Read(vme_off, data)) != 0) {
                CERR("ALTI: reading single entry in \"%s\" memory, index 0x%08x", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off);
                return (m_status);
            }
            break;
        case TTC_DECODER_TIMESTAMP:
            if ((m_status = DECEnableRead(dec_enabled)) != SUCCESS) {
                CERR("ALTI: reading DEC enable", "");
                return (m_status);
            }
            if (dec_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Read(vme_off, data)) != 0) {
                CERR("ALTI: reading single entry in \"%s\" memory, index 0x%08x", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off);
                return (m_status);
            }
            break;
        default:
            break;
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::MEMDataWrite(const ALTI_RAM_MEMORY mem, const unsigned int data, const unsigned int vme_off) { // number
    bool pat_enabled, snp_enabled, dec_enabled;
    // write the corresponding page number to the ADER1 page register
    if ((m_status = CSRADER1Write(mem + 1)) != SUCCESS) {
        CERR("ALTI: setting up ADER1 page register for accessing \"%s\" memory", ALTI_RAM_MEMORY_NAME[mem].c_str());
        return (m_status);
    }
    switch (mem) {
        case PATTERN_SNAPSHOT:
            if ((m_status = PATGenerationEnableRead(pat_enabled)) != SUCCESS) {
                CERR("ALTI: reading PAT generation enable", "");
                return (m_status);
            }
            if ((m_status = SNPEnableRead(snp_enabled)) != SUCCESS) {
                CERR("ALTI: reading SNP enable", "");
                return (m_status);
            }
            if (pat_enabled | snp_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Write(vme_off, data)) != 0) {
                CERR("ALTI: writing single entry in \"%s\" memory, index 0x%08x", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off);
                return (m_status);
            }
            break;
        case TTC_DECODER_COMMAND:
            if ((m_status = DECEnableRead(dec_enabled)) != SUCCESS) {
                CERR("ALTI: reading DEC enable", "");
                return (m_status);
            }
            if (dec_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Write(vme_off, data)) != 0) {
                CERR("ALTI: writing single entry in \"%s\" memory, index 0x%08x", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off);
                return (m_status);
            }
            break;
        case TTC_DECODER_TIMESTAMP:
            if ((m_status = DECEnableRead(dec_enabled)) != SUCCESS) {
                CERR("ALTI: reading DEC enable", "");
                return (m_status);
            }
            if (dec_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Write(vme_off, data)) != 0) {
                CERR("ALTI: writing single entry in \"%s\" memory, index 0x%08x", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off);
                return (m_status);
            }
            break;
        default:
            break;
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::MEMDataRead(const ALTI_RAM_MEMORY mem, std::vector<unsigned int> &data, const unsigned int vme_off) { // vector
    bool pat_enabled, snp_enabled, dec_enabled;
    // write the corresponding page number to the ADER1 page register
    if ((m_status = CSRADER1Write(mem + 1)) != SUCCESS) {
        CERR("ALTI: setting up ADER1 page register for accessing \"%s\" memory", ALTI_RAM_MEMORY_NAME[mem].c_str());
        return (m_status);
    }
    switch (mem) {
        case PATTERN_SNAPSHOT:
            if ((m_status = PATGenerationEnableRead(pat_enabled)) != SUCCESS) {
                CERR("ALTI: reading PAT generation enable", "");
                return (m_status);
            }
            if ((m_status = SNPEnableRead(snp_enabled)) != SUCCESS) {
                CERR("ALTI: reading SNP enable", "");
                return (m_status);
            }
            if (pat_enabled | snp_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Read(data, vme_off)) != 0) {
                CERR("ALTI: reading \"%s\" memory in single cycle, index 0x%08x, size 0x%08x", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off, data.size());
                return (m_status);
            }
            break;
        case TTC_DECODER_COMMAND:
            if ((m_status = DECEnableRead(dec_enabled)) != SUCCESS) {
                CERR("ALTI: reading DEC enable", "");
                return (m_status);
            }
            if (dec_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Read(data, vme_off)) != 0) {
                CERR("ALTI: reading \"%s\" memory in single cycle, index 0x%08x, size 0x%08x", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off, data.size());
                return (m_status);
            }
            break;
        case TTC_DECODER_TIMESTAMP:
            if ((m_status = DECEnableRead(dec_enabled)) != SUCCESS) {
                CERR("ALTI: reading DEC enable", "");
                return (m_status);
            }
            if (dec_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Read(data, vme_off)) != 0) {
                CERR("ALTI: reading \"%s\" memory in single cycle, index 0x%08x, size 0x%08x", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off, data.size());
                return (m_status);
            }
            break;
        default:
            break;
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::MEMDataWrite(const ALTI_RAM_MEMORY mem, const std::vector<unsigned int> &data, const unsigned int vme_off) { // vector
    bool pat_enabled, snp_enabled, dec_enabled;
    // write the corresponding page number to the ADER1 page register
    if ((m_status = CSRADER1Write(mem + 1)) != SUCCESS) {
        CERR("ALTI: setting up ADER1 page register for accessing \"%s\" memory", ALTI_RAM_MEMORY_NAME[mem].c_str());
        return (m_status);
    }
    switch (mem) {
        case PATTERN_SNAPSHOT:
            if ((m_status = PATGenerationEnableRead(pat_enabled)) != SUCCESS) {
                CERR("ALTI: reading PAT generation enable", "");
                return (m_status);
            }
            if ((m_status = SNPEnableRead(snp_enabled)) != SUCCESS) {
                CERR("ALTI: reading SNP enable", "");
                return (m_status);
            }
            if (pat_enabled | snp_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Write(data, vme_off)) != 0) {
                CERR("ALTI: writing \"%s\" memory in single cycle, index 0x%08x, size 0x%08x", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off, data.size());
                return (m_status);
            }
            break;
        case TTC_DECODER_COMMAND:
            if ((m_status = DECEnableRead(dec_enabled)) != SUCCESS) {
                CERR("ALTI: reading DEC enable", "");
                return (m_status);
            }
            if (dec_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Write(data, vme_off)) != 0) {
                CERR("ALTI: writing \"%s\" memory in single cycle, index 0x%08x, size 0x%08x", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off, data.size());
                return (m_status);
            }
            break;
        case TTC_DECODER_TIMESTAMP:
            if ((m_status = DECEnableRead(dec_enabled)) != SUCCESS) {
                CERR("ALTI: reading DEC enable", "");
                return (m_status);
            }
            if (dec_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Write(data, vme_off)) != 0) {
                CERR("ALTI: writing \"%s\" memory in single cycle, index 0x%08x, size 0x%08x", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off, data.size());
                return (m_status);
            }
            break;
        default:
            break;
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::MEMDataRead(const ALTI_RAM_MEMORY mem, RCD::CMEMSegment *seg, const unsigned int size, const unsigned int vme_off, const unsigned int seg_off) { // block
    bool pat_enabled, snp_enabled, dec_enabled;
    // write the corresponding page number to the ADER1 page register
    if ((m_status = CSRADER1Write(mem + 1)) != SUCCESS) {
        CERR("ALTI: setting up ADER1 page register for accessing \"%s\" memory", ALTI_RAM_MEMORY_NAME[mem].c_str());
        return (m_status);
    }
    switch (mem) {
        case PATTERN_SNAPSHOT:
            if ((m_status = PATGenerationEnableRead(pat_enabled)) != SUCCESS) {
                CERR("ALTI: reading PAT generation enable", "");
                return (m_status);
            }
            if ((m_status = SNPEnableRead(snp_enabled)) != SUCCESS) {
                CERR("ALTI: reading SNP enable", "");
                return (m_status);
            }
            if (pat_enabled | snp_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Read(seg, size, vme_off, seg_off)) != 0) {
                CERR("ALTI: reading \"%s\" memory in block cycle, index 0x%08x, size 0x%08x, segment offset", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off, size, seg_off);
                return (m_status);
            }
            break;
        case TTC_DECODER_COMMAND:
            if ((m_status = DECEnableRead(dec_enabled)) != SUCCESS) {
                CERR("ALTI: reading DEC enable", "");
                return (m_status);
            }
            if (dec_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Read(seg, size, vme_off, seg_off)) != 0) {
                CERR("ALTI: reading \"%s\" memory in block cycle, index 0x%08x, size 0x%08x, segment offset", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off, size, seg_off);
                return (m_status);
            }
            break;
        case TTC_DECODER_TIMESTAMP:
            if ((m_status = DECEnableRead(dec_enabled)) != SUCCESS) {
                CERR("ALTI: reading DEC enable", "");
                return (m_status);
            }
            if (dec_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Read(seg, size, vme_off, seg_off)) != 0) {
                CERR("ALTI: reading \"%s\" memory in block cycle, index 0x%08x, size 0x%08x, segment offset", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off, size, seg_off);
                return (m_status);
            }
            break;
        default:
            break;
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::MEMDataWrite(const ALTI_RAM_MEMORY mem, RCD::CMEMSegment *seg, const unsigned int size, const unsigned int vme_off, const unsigned int seg_off) { // block
    bool pat_enabled, snp_enabled, dec_enabled;
    // write the corresponding page number to the ADER1 page register
    if ((m_status = CSRADER1Write(mem + 1)) != SUCCESS) {
        CERR("ALTI: setting up ADER1 page register for accessing \"%s\" memory", ALTI_RAM_MEMORY_NAME[mem].c_str());
        return (m_status);
    }
    switch (mem) {
        case PATTERN_SNAPSHOT:
            if ((m_status = PATGenerationEnableRead(pat_enabled)) != SUCCESS) {
                CERR("ALTI: reading PAT generation enable", "");
                return (m_status);
            }
            if ((m_status = SNPEnableRead(snp_enabled)) != SUCCESS) {
                CERR("ALTI: reading SNP enable", "");
                return (m_status);
            }
            if (pat_enabled | snp_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Write(seg, size, vme_off, seg_off)) != 0) {
                CERR("ALTI: writing \"%s\" memory in block cycle, index 0x%08x, size 0x%08x, segment offset", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off, size, seg_off);
                return (m_status);
            }
            break;
        case TTC_DECODER_COMMAND:
            if ((m_status = DECEnableRead(dec_enabled)) != SUCCESS) {
                CERR("ALTI: reading DEC enable", "");
                return (m_status);
            }
            if (dec_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Write(seg, size, vme_off, seg_off)) != 0) {
                CERR("ALTI: writing \"%s\" memory in block cycle, index 0x%08x, size 0x%08x, segment offset", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off, size, seg_off);
                return (m_status);
            }
            break;
        case TTC_DECODER_TIMESTAMP:
            if ((m_status = DECEnableRead(dec_enabled)) != SUCCESS) {
                CERR("ALTI: reading DEC enable", "");
                return (m_status);
            }
            if (dec_enabled) {
                CERR("ALTI: reading \"%s\" memory while it is busy", ALTI_RAM_MEMORY_NAME[mem].c_str());
                m_status = ALTI_RAM_MEMORY_BUSY;
                return (m_status);
            }
            if ((m_status = m_alti->MEM_SharedAltiRAM_Write(seg, size, vme_off, seg_off)) != 0) {
                CERR("ALTI: writing \"%s\" memory in block cycle, index 0x%08x, size 0x%08x, segment offset", ALTI_RAM_MEMORY_NAME[mem].c_str(), vme_off, size, seg_off);
                return (m_status);
            }
            break;
        default:
            break;
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SNPRead(RCD::CMEMSegment *segment, const unsigned int start, const unsigned int num) { // read snapshot memory
    if (start > MAX_MEMORY) {
        CERR("ALTI: invalid start index of snapshot memory", "");
        m_status = WRONGPAR;
        return (m_status);
    }
    else if (start + num > MAX_MEMORY) {
        CERR("ALTI: invalid number of data to be read from the snapshot memory", "");
        m_status = WRONGPAR;
        return (m_status);
    }
    //if ((m_status = MEMDataRead(PATTERN_SNAPSHOT, segment, MAX_MEMORY, 0, 0)) != 0) {
    if ((m_status = MEMDataRead(PATTERN_SNAPSHOT, segment, num, start, start)) != 0) {
        CERR("ALTI: reading snapshot memory in block cycle", "");
        return (m_status);
    }
    /*std::vector<unsigned int> data;
    data.resize(num);
    if ((m_status = MEMDataRead(PATTERN_SNAPSHOT, data, start)) != 0) {
        CERR("ALTI: reading snapshot memory in single cycle", "");
        return (m_status);
    }*/

    std::string readable;
    unsigned int indx; //, indx1, snapshot, mult, cur_signals;
    unsigned int *data = (unsigned int *) segment->VirtualAddress();

    int format = 1;
    switch (format) {
        /*case 0: // hex
            if ((m_status = AltiCommon::writeFile(outf, MAX_MEMORY, data) != 0)) {
                CERR("ALTI: reading snapshot/pattern memory into hex file ", "");
                return (m_status);
            }
            break;*/
        case 1: // human-readable (original)
            std::cout << "#---------------------------\n";
            std::cout << "#                      M\n";
            std::cout << "#                      u\n";
            std::cout << "#                      l\n";
            std::cout << "#                      t\n";
            std::cout << "#                      i\n";
            std::cout << "#                      p\n";
            std::cout << "#                      l\n";
            std::cout << "#   CCC                i\n";
            std::cout << "#  BRRR    T BBBB TTT  c\n";
            std::cout << "#O UEEE    T GGGG TTTL i\n";
            std::cout << "#R SQQQ    Y OOOO RRR1 t\n";
            std::cout << "#B Y210    P 3210 321A y\n";
            std::cout << "#---------------------------\n";
            for (indx = start; indx < start + num; indx++) {
            //for (indx = 0; indx < num; indx++) {
                MEMSnapshotToString(data[indx], readable);
                std::cout << readable << "\n";
            }
            std::cout << "#---------------------------\n";
            break;

        default:
            break;
    }

    //outf.close();

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SNPReadFile(RCD::CMEMSegment *segment, const unsigned int start, const unsigned int num, const int format, std::ofstream &outf) { // read snapshot memory into file
    if (start > MAX_MEMORY) {
        CERR("ALTI: invalid start index of snapshot memory", "");
        m_status = WRONGPAR;
        return (m_status);
    }
    else if (start + num > MAX_MEMORY) {
        CERR("ALTI: invalid number of data to be read from the snapshot memory", "");
        m_status = WRONGPAR;
        return (m_status);
    }
    //if ((m_status = MEMDataRead(PATTERN_SNAPSHOT, segment, MAX_MEMORY, 0, 0)) != 0) {
    if ((m_status = MEMDataRead(PATTERN_SNAPSHOT, segment, num, start, start)) != 0) {
        CERR("ALTI: reading snapshot memory in block cycle", "");
        return (m_status);
    }
    /*std::vector<unsigned int> data;
    data.resize(num);
    if ((m_status = MEMDataRead(PATTERN_SNAPSHOT, data, start)) != 0) {
        CERR("ALTI: reading snapshot memory in single cycle", "");
        return (m_status);
    }*/

    std::string readable;
    unsigned int indx;
    unsigned int *data = (unsigned int *) segment->VirtualAddress();

    switch (format) {
        case 0: // hex
            if ((m_status = AltiCommon::writeFile(outf, num, data + start) != 0)) {
            //if ((m_status = AltiCommon::writeFile(outf, num, &data[0]) != 0)) {
                CERR("ALTI: reading snapshot/pattern memory into hex file ", "");
                return (m_status);
            }
            break;
        case 1: // human-readable (compressed with multiplicities)
            outf << "#---------------------------\n";
            outf << "#                      M\n";
            outf << "#                      u\n";
            outf << "#                      l\n";
            outf << "#                      t\n";
            outf << "#                      i\n";
            outf << "#                      p\n";
            outf << "#                      l\n";
            outf << "#   CCC                i\n";
            outf << "#  BRRR    T BBBB TTT  c\n";
            outf << "#O UEEE    T GGGG TTTL i\n";
            outf << "#R SQQQ    Y OOOO RRR1 t\n";
            outf << "#B Y210    P 3210 321A y\n";
            outf << "#---------------------------\n";
            for (indx = start; indx < start + num; indx++) {
            //for (indx = 0; indx < num; indx++) {
                MEMSnapshotToString(data[indx], readable);
                outf << readable << "\n";
            }
            outf << "#---------------------------\n";
            break;
        case 2: // human-readable (expanded with multiplicities = 1)
         outf << "#---------------------------\n";
            outf << "#                      M\n";
            outf << "#                      u\n";
            outf << "#                      l\n";
            outf << "#                      t\n";
            outf << "#                      i\n";
            outf << "#                      p\n";
            outf << "#                      l\n";
            outf << "#   CCC                i\n";
            outf << "#  BRRR    T BBBB TTT  c\n";
            outf << "#O UEEE    T GGGG TTTL i\n";
            outf << "#R SQQQ    Y OOOO RRR1 t\n";
            outf << "#B Y210    P 3210 321A y\n";
            outf << "#---------------------------\n";
            for (indx = start; indx < start + num; indx++) {
            //for (indx = 0; indx < num; indx++) {
	      int mult = MEMSnapshotToString(data[indx], readable, true);
	      for(int i=0; i<mult; i++) {
		outf << readable << "\n";
	      }
            }
            outf << "#---------------------------\n";
            break;
        default:
            break;
    }

    outf.close();

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::PATRead(RCD::CMEMSegment *segment) { // read pattern generation memory
    unsigned int start;
    if ((m_status = PATGenerationStartAddressRead(start)) != 0) {
        CERR("ALTI: reading pattern generation start address", "");
        return (m_status);
    }
    unsigned int stop;
    if ((m_status = PATGenerationStopAddressRead(stop)) != 0) {
        CERR("ALTI: reading pattern generation stop address", "");
        return (m_status);
    }
    if (start > stop) {
        m_status = SUCCESS;
        return (m_status);
    }

    //if ((m_status = MEMDataRead(PATTERN_SNAPSHOT, segment, MAX_MEMORY, 0, 0)) != 0) {
    if ((m_status = MEMDataRead(PATTERN_SNAPSHOT, segment, stop - start + 1, start, start)) != 0) {
        CERR("ALTI: reading pattern generation memory in block cycle", "");
        return (m_status);
    }
    /*unsigned int num = stop - start + 1;
    std::vector<unsigned int> data;
    data.resize(num);
    if ((m_status = MEMDataRead(PATTERN_SNAPSHOT, data, start)) != 0) {
        CERR("ALTI: reading pattern generation memory in single cycle", "");
        return (m_status);
    }*/

    std::string readable;
    unsigned int indx; //, indx1, snapshot, mult, cur_signals;
    unsigned int *data = (unsigned int *) segment->VirtualAddress();

    int format = 1;
    switch (format) {
        /*case 0: // hex
            if ((m_status = AltiCommon::writeFile(outf, MAX_MEMORY, data) != 0)) {
                CERR("ALTI: reading snapshot/pattern memory into hex file ", "");
                return (m_status);
            }
            break;*/
        case 1: // human-readable (original)
            std::cout << "#---------------------------\n";
            std::cout << "#                      M\n";
            std::cout << "#                      u\n";
            std::cout << "#                      l\n";
            std::cout << "#                      t\n";
            std::cout << "#                      i\n";
            std::cout << "#                      p\n";
            std::cout << "#                      l\n";
            std::cout << "#   CCC                i\n";
            std::cout << "#  BRRR    T BBBB TTT  c\n";
            std::cout << "#O UEEE    T GGGG TTTL i\n";
            std::cout << "#R SQQQ    Y OOOO RRR1 t\n";
            std::cout << "#B Y210    P 3210 321A y\n";
            std::cout << "#---------------------------\n";
            for (indx = start; indx <= stop; indx++) {
            //for (indx = 0; indx < num; indx++) {
                MEMSnapshotToString(data[indx], readable);
                std::cout << readable << "\n";
            }
            std::cout << "#---------------------------\n";
            break;
        /*case 2: // human-readable (compressed)
            std::cout << "#-----------------------\n";
            std::cout << "#                      M\n";
            std::cout << "#                      u\n";
            std::cout << "#                      l\n";
            std::cout << "#                      t\n";
            std::cout << "#                      i\n";
            std::cout << "#                      p\n";
            std::cout << "#                      l\n";
            std::cout << "#   CCC                i\n";
            std::cout << "#  BRRR    T BBBB TTT  c\n";
            std::cout << "#O UEEE    T GGGG TTTL i\n";
            std::cout << "#R SQQQ    Y OOOO RRR1 t\n";
            std::cout << "#B Y210    P 3210 321A y\n";
            std::cout << "#-----------------------\n";
            indx = 0;
            while (indx < MAX_MEMORY - 1) {
                snapshot = data[indx];
                mult = (((snapshot & MEM_MULT_MASK) >> MEM_SIGNAL_BITS) + 1);
                cur_signals = snapshot & MEM_SIGNAL_MASK;
                indx1 = indx + 1;
                while ((data[indx1] & MEM_SIGNAL_MASK) == cur_signals) { // add multiplicities until no changes observed
                    mult += (((data[indx1] & MEM_MULT_MASK) >> MEM_SIGNAL_BITS) + 1);
                    indx1++;
                }

                std::ostringstream buf;

                buf << ' ';

                buf << ((snapshot & MEM_MASK[ORB]) ? ('1') : ('0')); // ORB

                buf << ' ';

                buf << ((snapshot & MEM_MASK[BUSY]) ? ('1') : ('0')); // BUSY
                buf << ((snapshot & 0x00040000) ? ('1') : ('0')); // CALREQ[2]
                buf << ((snapshot & 0x00020000) ? ('1') : ('0')); // CALREQ[1]
                buf << ((snapshot & 0x00010000) ? ('1') : ('0')); // CALREQ[0]

                buf << ' ';

                buf << "0x" << std::setfill('0') << std::setw(2) << std::hex << ((snapshot & 0x000ff00) >> 8); // TTYP

                buf << ' ';

                buf << ((snapshot & MEM_MASK[BGO3]) ? ('1') : ('0')); // BGO3
                buf << ((snapshot & MEM_MASK[BGO2]) ? ('1') : ('0')); // BGO2
                buf << ((snapshot & MEM_MASK[BGO1]) ? ('1') : ('0')); // BGO1
                buf << ((snapshot & MEM_MASK[BGO0]) ? ('1') : ('0')); // BGO0

                buf << ' ';

                buf << ((snapshot & MEM_MASK[TTR3]) ? ('1') : ('0')); // TTR3
                buf << ((snapshot & MEM_MASK[TTR2]) ? ('1') : ('0')); // TTR2
                buf << ((snapshot & MEM_MASK[TTR1]) ? ('1') : ('0')); // TTR1
                buf << ((snapshot & MEM_MASK[L1A])  ? ('1') : ('0')); // L1A

                buf << ' ';

                //buf << std::dec << (mult - 1); // multiplicity, combined
                buf << std::dec << (mult); // multiplicity, combined

                readable = buf.str();

                std::cout << readable << "\n";

                indx = indx1;
            }
            std::cout << "----------------------------\n";
            break;*/
        default:
            break;
    }

    //outf.close();

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::PATWriteFile(RCD::CMEMSegment *segment, std::ifstream &inf) { // write pattern generation memory from file
    unsigned int xdat, indx, i;
    unsigned int* data = (unsigned int*) segment->VirtualAddress();
    //std::vector<unsigned int> data;

    indx = 0;
    std::string line;
    std::vector<unsigned int> pattern;

    while (std::getline(inf, line)) {
        if (line.empty()) continue;
        if (line[0] == '#' || line[0] == '-') continue; // ignore comment lines starting with # or -

        if ((m_status = MEMStringToPattern(line, pattern, ALTI_LINE)) != 0) {
            CERR("ALTI: parsing line from the pattern file", "");
            return (m_status);
        }
        for (i = 0; (i < pattern.size()) && (indx < MAX_MEMORY); i++) {
            xdat = pattern[i];
            //data[indx] = xdat & ALTI::MASK_MEM_PATTERN;
            data[indx] = xdat & ALTI::MASK_MEM_SHAREDALTIRAM;
            //data.push_back(xdat & ALTI::MASK_MEM_PATTERN);
            indx++;
        }
    }
    if ((m_status = PATGenerationStartAddressWrite(0)) != 0) {
        CERR("ALTI: writing pattern generation start address 0x%08x", 0);
        return (m_status);
    }
    if ((m_status = PATGenerationStopAddressWrite(indx - 1)) != 0) {
        CERR("ALTI: writing pattern generation stop address 0x%08x", indx - 1);
        return (m_status);
    }

    /*while (indx < MAX_MEMORY) { // fill the rest with zeros
        xdat = 0;
        data[indx] = xdat & ALTI::MASK_MEM_PATTERN;
        indx++;
    }*/
    if (indx < MAX_MEMORY) { // fill the next entry with zero
        xdat = 0;
        //data[indx] = xdat & ALTI::MASK_MEM_PATTERN;
        data[indx] = xdat & ALTI::MASK_MEM_SHAREDALTIRAM;
        indx++;
        //data.push_back(0x00000000);
    }

    //if ((m_status = MEMDataWrite(PATTERN_SNAPSHOT, segment, MAX_MEMORY, 0, 0)) != 0) {
    //if ((m_status = MEMDataWrite(PATTERN_SNAPSHOT, segment, indx + 1, 0, 0)) != 0) {
    if ((m_status = MEMDataWrite(PATTERN_SNAPSHOT, segment, indx, 0, 0)) != 0) {
        CERR("ALTI: writing pattern generation memory in block cycle", "");
        return (m_status);
    }
    /*if ((m_status = MEMDataWrite(PATTERN_SNAPSHOT, data, 0x00000000)) != 0) {
        CERR("ALTI: writing pattern generation memory in single cycle", "");
        return (m_status);
    }*/

    inf.close();
        
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::MEMSnapshotToString(unsigned int snapshot, std::string &readable, bool no_mult) {
    std::ostringstream buf;

    // try to remove numbers afterwards, masks and shifts can be found in the snapshot bitstring

    buf << ' ';

    buf << ((snapshot & MEM_MASK[ORB]) ? ('1') : ('0')); // ORB

    buf << ' ';

    buf << ((snapshot & MEM_MASK[BUSY]) ? ('1') : ('0')); // BUSY
    buf << ((snapshot & MEM_MASK[CALREQ2]) ? ('1') : ('0')); // CALREQ[2]
    buf << ((snapshot & MEM_MASK[CALREQ1]) ? ('1') : ('0')); // CALREQ[1]
    buf << ((snapshot & MEM_MASK[CALREQ0]) ? ('1') : ('0')); // CALREQ[0]

    buf << ' ';

    buf << "0x" << std::setfill('0') << std::setw(2) << std::hex << ((snapshot & (MEM_MASK[TTYP0] + MEM_MASK[TTYP1] + MEM_MASK[TTYP2] + MEM_MASK[TTYP3] + MEM_MASK[TTYP4] + MEM_MASK[TTYP5] + MEM_MASK[TTYP6] + MEM_MASK[TTYP7])) >> 8); // TTYP

    buf << ' ';

    buf << ((snapshot & MEM_MASK[BGO3]) ? ('1') : ('0')); // BGO3
    buf << ((snapshot & MEM_MASK[BGO2]) ? ('1') : ('0')); // BGO2
    buf << ((snapshot & MEM_MASK[BGO1]) ? ('1') : ('0')); // BGO1
    buf << ((snapshot & MEM_MASK[BGO0]) ? ('1') : ('0')); // BGO0

    buf << ' ';

    buf << ((snapshot & MEM_MASK[TTR3]) ? ('1') : ('0')); // TTR3
    buf << ((snapshot & MEM_MASK[TTR2]) ? ('1') : ('0')); // TTR2
    buf << ((snapshot & MEM_MASK[TTR1]) ? ('1') : ('0')); // TTR1
    buf << ((snapshot & MEM_MASK[L1A])  ? ('1') : ('0')); // L1A

    buf << ' ';

    int multiplicity = ((snapshot & MEM_MULT_MASK) >> MEM_SIGNAL_BITS) + 1;
    if(no_mult) {
      buf << "1";
    }
    else {
      buf << std::dec << (((snapshot & MEM_MULT_MASK) >> MEM_SIGNAL_BITS) + 1); // Multiplicity
    }

    readable = buf.str();

    return multiplicity;
}

//------------------------------------------------------------------------------

int AltiModule::MEMStringToPattern(std::string &readable, std::vector<unsigned int> &pattern, const SNAPSHOT_TYPE lineType, uint8_t ttypw[4]) {
    pattern.clear();

    int i;
    std::stringstream buf;
    buf << readable;

    // read pattern
    char orb;
    char busy, creq[SNAPSHOT::CALREQ_NUMBER];
    unsigned int ttyp;
    char bgo[SNAPSHOT::BGO_NUMBER];
    char ttr[SNAPSHOT::TTR_NUMBER], l1a;
    char inhib;
    char ttypc[2];
    int multiplicity;
    switch (lineType) {
        case ALTI_LINE:

            buf >> orb;
            if (orb < '0' || orb > '1') {
                CERR("ALTI: ORB not binary in %s pattern %s", SNAPSHOT_TYPE_NAME[lineType].c_str(), readable.c_str());
                return -1;
            }

            buf >> busy >> creq[2] >> creq[1] >> creq[0];
	   
            if (busy < '0' || busy > '1') {
                CERR("ALTI: BUSY not binary in %s pattern %s", SNAPSHOT_TYPE_NAME[lineType].c_str(), readable.c_str());
                return -1;
            }
            for (i = 2; i >= 0; i--) {
                if (creq[i] < '0' || creq[i] > '1') {
                    CERR("ALTI: CALREQ[%d] not binary in %s pattern %s", i, SNAPSHOT_TYPE_NAME[lineType].c_str(), readable.c_str());
                    return -1;
                }
            }

            buf >> std::hex >> ttyp;

            buf >> bgo[3] >> bgo[2] >> bgo[1] >> bgo[0];
            for (i = 3; i >= 0; i--) {
                if (bgo[i] < '0' || bgo[i] > '1') {
                    CERR("ALTI: BGO[%d] not binary in %s pattern %s", i, SNAPSHOT_TYPE_NAME[lineType].c_str(), readable.c_str());
                    return -1;
                }
            }

            buf >> ttr[2] >> ttr[1] >> ttr[0] >> l1a;
            for (i = 2; i >= 0; i--) {
                if (ttr[i] < '0' || ttr[i] > '1') {
                    CERR("ALTI: TTR[%d] not binary in %s pattern %s", i + 1, SNAPSHOT_TYPE_NAME[lineType].c_str(), readable.c_str());
                    return -1;
                }
            }
            if (l1a < '0' || l1a > '1') {
                CERR("ALTI: L1A not binary in %s pattern %s", SNAPSHOT_TYPE_NAME[lineType].c_str(), readable.c_str());
                return -1;
            }

            buf >> std::dec >> multiplicity;
            /*if (multiplicity < 0) {
                CERR("ALTI: multiplicity negative in %s pattern %s", SNAPSHOT_TYPE_NAME[lineType].c_str(), readable.c_str());
                return -1;
            }*/
            if (multiplicity <= 0) {
                CERR("ALTI: multiplicity not positive in %s pattern %s", SNAPSHOT_TYPE_NAME[lineType].c_str(), readable.c_str());
                return -1;
            }

            break;
        case LTP_LINE:

            buf >> busy;
            if (busy < '0' || busy > '1') {
                CERR("ALTI: BUSY not binary in %s pattern %s", SNAPSHOT_TYPE_NAME[lineType].c_str(), readable.c_str());
                return -1;
            }
            //std::printf("BUSY: %c\n", busy);

            buf >> inhib;
            if (inhib < '0' || inhib > '1') {
                CERR("ALTI: INHIBIT not binary in %s pattern %s", SNAPSHOT_TYPE_NAME[lineType].c_str(), readable.c_str());
                return -1;
            }
            //std::printf("INHIBIT: %c\n", inhib);

            buf >> creq[2] >> creq[1] >> creq[0];
            for (i = 2; i >= 0; i--) {
                if (creq[i] < '0' || creq[i] > '1') {
                    CERR("ALTI: CALREQ[%d] not binary in %s pattern %s", i, SNAPSHOT_TYPE_NAME[lineType].c_str(), readable.c_str());
                    return -1;
                }
            }
            //std::printf("CALREQ[2..0]: %c%c%c\n", creq[2], creq[1], creq[0]);

            buf >> bgo[3] >> bgo[2] >> bgo[1] >> bgo[0];
            for (i = 3; i >= 0; i--) {
                if (bgo[i] < '0' || bgo[i] > '1') {
                    CERR("ALTI: BGO[%d] not binary in %s pattern %s", i, SNAPSHOT_TYPE_NAME[lineType].c_str(), readable.c_str());
                    return -1;
                }
            }
            //std::printf("BGO[3..0]: %c%c%c%c\n", bgo[3], bgo[2], bgo[1], bgo[0]);

            buf >> ttypc[1] >> ttypc[0];
            for (i = 1; i >= 0; i--) {
                if (ttypc[i] < '0' || ttypc[i] > '1') {
                    CERR("ALTI: TTYP[%d] not binary in %s pattern %s", i, SNAPSHOT_TYPE_NAME[lineType].c_str(), readable.c_str());
                    return -1;
                }
            }
            //std::printf("TTYP[1..0]: %c%c\n", ttypc[1], ttypc[0]);
            ttyp = ttypw[(ttypc[1] - '0')*2 + (ttypc[0] - '0')*1]; // problem, how to translate to 32b word?

            buf >> ttr[2] >> ttr[1] >> ttr[0] >> l1a;
            for (i = 2; i >= 0; i--) {
                if (ttr[i] < '0' || ttr[i] > '1') {
                    CERR("ALTI: TTR[%d] not binary in %s pattern %s", i + 1, SNAPSHOT_TYPE_NAME[lineType].c_str(), readable.c_str());
                    return -1;
                }
            }
            if (l1a < '0' || l1a > '1') {
                CERR("ALTI: L1A not binary in %s pattern %s", SNAPSHOT_TYPE_NAME[lineType].c_str(), readable.c_str());
                return -1;
            }
            //std::printf("TTR[3..1]: %c%c%c\n", ttr[2], ttr[1], ttr[0]);
            //std::printf("L1A: %c\n", l1a);

            buf >> orb;
            if (orb < '0' || orb > '1') {
                CERR("ALTI: ORB not binary in %s pattern %s", SNAPSHOT_TYPE_NAME[lineType].c_str(), readable.c_str());
                return -1;
            }
            //std::printf("ORB: %c\n", orb);

            buf >> std::dec >> multiplicity;
            if (multiplicity <= 0) {
                CERR("ALTI: multiplicity not positive in %s pattern %s", SNAPSHOT_TYPE_NAME[lineType].c_str(), readable.c_str());
                return -1;
            }
            //std::printf("mult: %d\n", multiplicity);

            break;
        default:
            break;
    }

    // transform into pattern generation data entry/entries

    unsigned int data = 0, duration;
    data |= ((l1a - '0') << AltiCommon::getFirstSetBitPos(MEM_MASK[L1A]));
    data |= ((ttr[0] - '0') << AltiCommon::getFirstSetBitPos(MEM_MASK[TTR1]));
    data |= ((ttr[1] - '0') << AltiCommon::getFirstSetBitPos(MEM_MASK[TTR2]));
    data |= ((ttr[2] - '0') << AltiCommon::getFirstSetBitPos(MEM_MASK[TTR3]));
    data |= ((bgo[0] - '0') << AltiCommon::getFirstSetBitPos(MEM_MASK[BGO0]));
    data |= ((bgo[1] - '0') << AltiCommon::getFirstSetBitPos(MEM_MASK[BGO1]));
    data |= ((bgo[2] - '0') << AltiCommon::getFirstSetBitPos(MEM_MASK[BGO2]));
    data |= ((bgo[3] - '0') << AltiCommon::getFirstSetBitPos(MEM_MASK[BGO3]));
    data |= (ttyp << AltiCommon::getFirstSetBitPos(MEM_MASK[TTYP0]));
    data |= ((creq[0] - '0') << AltiCommon::getFirstSetBitPos(MEM_MASK[CALREQ0]));
    data |= ((creq[1] - '0') << AltiCommon::getFirstSetBitPos(MEM_MASK[CALREQ1]));
    data |= ((creq[2] - '0') << AltiCommon::getFirstSetBitPos(MEM_MASK[CALREQ2]));
    data |= ((busy - '0') << AltiCommon::getFirstSetBitPos(MEM_MASK[BUSY]));
    data |= ((orb - '0') << AltiCommon::getFirstSetBitPos(MEM_MASK[ORB]));

    switch (lineType) {
        case ALTI_LINE:

            /*// PG input: written to memory as-is, i.e. 0 means duration of 1 
            while (multiplicity >= 0) {
                //std::cout << "mult: " << multiplicity << std::endl;
                //duration = min(multiplicity, 0xff);
                duration = ((((unsigned int) multiplicity) <= MEM_MULT_MAX) ? (multiplicity) : (MEM_MULT_MAX));

                pattern.push_back((duration << MEM_SIGNAL_BITS) | data);
                //std::printf("pattern: 0x%08x\n", pattern[pattern.size() - 1]);

                multiplicity -= (duration + 1);
            }*/

            // PG input: written to memory counting from 0, i.e. 0 means duration of 1 
            while (multiplicity > 0) {
                //std::cout << "mult: " << multiplicity << std::endl;
                //duration = min(multiplicity, 0xff);
                duration = ((((unsigned int) multiplicity) <= MEM_MULT_MAX) ? (multiplicity - 1) : (MEM_MULT_MAX));

                pattern.push_back((duration << MEM_SIGNAL_BITS) | data);
                //std::printf("pattern: 0x%08x\n", pattern[pattern.size() - 1]);

                multiplicity -= (duration + 1);
            }

            break;
        case LTP_LINE:

            // PG input: written to memory as-is, i.e. 0 means duration of 1 
            while (multiplicity > 0) {
                //std::cout << "mult: " << multiplicity << std::endl;
                //duration = min(multiplicity, 0xff);
                duration = ((((unsigned int) multiplicity) <= MEM_MULT_MAX) ? (multiplicity - 1) : (MEM_MULT_MAX));

                pattern.push_back((duration << MEM_SIGNAL_BITS) | data);
                //std::printf("pattern: 0x%08x\n", pattern[pattern.size() - 1]);

                multiplicity -= (duration + 1);
            }

            break;
        default:
            break;
    }

    return 0;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::ENCTransmitterEnableRead(const unsigned int tx, bool &enabled) {
    if (tx >= TRANSMITTER_NUMBER) {
        CERR("ALTI: wrong TTC TX number chosen", "");
        m_status = WRONGPAR;
        return (m_status);
    }

    ALTI_TTC_SFPCONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->TTC_SFPControl_Read(bs)) != 0) {
        CERR("ALTI: reading TTC control register", "");
        return (m_status);
    }

    std::string s = bs.TXEnable(tx);
    if (s == "Enabled") {
        enabled = true;
    }
    else if (s == "Disabled") {
        enabled = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for TTC TX%d enable", s.c_str(), tx);
        m_status = FAILURE;
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCTransmitterEnableRead(std::vector<bool> &enabled) {
    enabled.clear();

    ALTI_TTC_SFPCONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->TTC_SFPControl_Read(bs)) != 0) {
        CERR("ALTI: reading TTC control register", "");
        return (m_status);
    }

    unsigned int i;
    std::string s;
    for (i = 0; i < TRANSMITTER_NUMBER; i++) {
        s = bs.TXEnable(i);
        if (s == "Enabled") {
            enabled.push_back(true);
        }
        else if (s == "Disabled") {
            enabled.push_back(false);
        }
        else {
            CERR("ALTI: unknown value \"%s\" for TTC TX%d enable", s.c_str(), i);
            m_status = FAILURE;
            return (m_status);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCTransmitterFaultRead(const unsigned int tx, bool &fault) {
    if (tx >= TRANSMITTER_NUMBER) {
        CERR("ALTI: wrong TTC TX number chosen", "");
        m_status = WRONGPAR;
        return (m_status);
    }

    ALTI_TTC_SFPCONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->TTC_SFPControl_Read(bs)) != 0) {
        CERR("ALTI: reading TTC control register", "");
        return (m_status);
    }

    std::string s = bs.TXFaultStatus(tx);
    if (s == "Fault") {
        fault = true;
    }
    else if (s == "Disabled") {
        fault = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for TTC TX%d fault status", s.c_str(), tx);
        m_status = FAILURE;
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCTransmitterFaultRead(std::vector<bool> &fault) { // all
    fault.clear();

    ALTI_TTC_SFPCONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->TTC_SFPControl_Read(bs)) != 0) {
        CERR("ALTI: reading TTC control register", "");
        return (m_status);
    }

    unsigned int i;
    std::string s;
    for (i = 0; i < TRANSMITTER_NUMBER; i++) {
        s = bs.TXFaultStatus(i);
        if (s == "Fault") {
            fault.push_back(true);
        }
        else if (s == "NormalOperation") {
            fault.push_back(false);
        }
        else {
            CERR("ALTI: unknown value \"%s\" for TTC TX%d fault status", s.c_str(), i);
            m_status = FAILURE;
            return (m_status);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCTransmitterFaultClear(const unsigned int tx) {
    if (tx >= TRANSMITTER_NUMBER) {
        CERR("ALTI: wrong TTC TX number chosen", "");
        m_status = WRONGPAR;
        return (m_status);
    }

    unsigned int data;

    // READ
    if ((m_status = m_alti->TTC_SFPControl_Read(data)) != 0) {
        CERR("ALTI: reading TTC control register", "");
        return (m_status);
    }

    // MODIFY
    data |= (ALTI_TTC_SFPCONTROL_BITSTRING::STAT_MASK_CLOCKDATARECOVERYLOS + ALTI_TTC_SFPCONTROL_BITSTRING::STAT_MASK_CLOCKDATARECOVERYLOL + ALTI_TTC_SFPCONTROL_BITSTRING::STAT_MASK_TTCDECODERSINGLEBITERROR + ALTI_TTC_SFPCONTROL_BITSTRING::STAT_MASK_TTCDECODERDOUBLEBITERROR + ALTI_TTC_SFPCONTROL_BITSTRING::STAT_MASK_TTCDECODERPLLLOCKSTATUS); // don't clear
    unsigned int i, mask;
    for (i = 0; i < TRANSMITTER_NUMBER; i++) {
        mask = (ALTI_TTC_SFPCONTROL_BITSTRING::MASK_TXFAULTSTATUS[i]).number()[0];
        if (i == tx) {
            data &= (~mask); // clear
        }
        else {
            data |= mask; // don't clear
        }
    }

    // WRITE
    if ((m_status = m_alti->TTC_SFPControl_Write(data)) != 0) {
        CERR("ALTI: writing TTC control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCTransmitterFaultClear() { // all
    unsigned int data;

    // READ
    if ((m_status = m_alti->TTC_SFPControl_Read(data)) != 0) {
        CERR("ALTI: reading TTC control register", "");
        return (m_status);
    }

    // MODIFY
    data |= (ALTI_TTC_SFPCONTROL_BITSTRING::STAT_MASK_CLOCKDATARECOVERYLOS + ALTI_TTC_SFPCONTROL_BITSTRING::STAT_MASK_CLOCKDATARECOVERYLOL + ALTI_TTC_SFPCONTROL_BITSTRING::STAT_MASK_TTCDECODERSINGLEBITERROR + ALTI_TTC_SFPCONTROL_BITSTRING::STAT_MASK_TTCDECODERDOUBLEBITERROR + ALTI_TTC_SFPCONTROL_BITSTRING::STAT_MASK_TTCDECODERPLLLOCKSTATUS); // don't clear
    unsigned int i, mask;
    for (i = 0; i < TRANSMITTER_NUMBER; i++) {
        mask = (ALTI_TTC_SFPCONTROL_BITSTRING::MASK_TXFAULTSTATUS[i]).number()[0];
        data &= (~mask); // clear
    }

    // WRITE
    if ((m_status = m_alti->TTC_SFPControl_Write(data)) != 0) {
        CERR("ALTI: writing TTC control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCTransmitterEnableWrite(const unsigned int tx, const bool enable) {
    if (tx >= TRANSMITTER_NUMBER) {
        CERR("ALTI: wrong TTC TX number chosen", "");
        m_status = WRONGPAR;
        return (m_status);
    }

    ALTI_TTC_SFPCONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->TTC_SFPControl_Read(bs)) != 0) {
        CERR("ALTI: reading TTC control register", "");
        return (m_status);
    }
    // MODIFY
    bs.TXEnable(tx, enable ? "Enabled" : "Disabled");
    // WRITE
    if ((m_status = m_alti->TTC_SFPControl_Write(bs)) != 0) {
        CERR("ALTI: writing TTC control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCTransmitterEnableWrite(const bool enable) { // all
    ALTI_TTC_SFPCONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->TTC_SFPControl_Read(bs)) != 0) {
        CERR("ALTI: reading TTC control register", "");
        return (m_status);
    }
    // MODIFY
    std::string sen = (enable ? "Enabled" : "Disabled");
    unsigned int i;
    for (i = 0; i < TRANSMITTER_NUMBER; i++) {
        bs.TXEnable(i, sen);
    }
    // WRITE
    if ((m_status = m_alti->TTC_SFPControl_Write(bs)) != 0) {
        CERR("ALTI: writing TTC control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCTransmitterEnableWrite(const std::vector<bool> &enable) { // all
    ALTI_TTC_SFPCONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->TTC_SFPControl_Read(bs)) != 0) {
        CERR("ALTI: reading TTC control register", "");
        return (m_status);
    }

    std::string sen;
    unsigned int i;
    for (i = 0; i < TRANSMITTER_NUMBER; i++) {
        // MODIFY
        sen = (enable[i] ? "Enabled" : "Disabled");
        bs.TXEnable(i, sen);
    }

    // WRITE
    if ((m_status = m_alti->TTC_SFPControl_Write(bs)) != 0) {
        CERR("ALTI: writing TTC control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCTTRSelWrite(const unsigned int ttr_or_l1a) {
    ALTI_BSY_CONTROL_BITSTRING bs; // I'm using bits in this register because it's the easiest way. It will be modified with the new firmware to something more appropiate
    //std::string str;
    // READ
    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }
    // MODIFY
    /*
    if (ttr_or_l1a == 1) {str = "L1A"}
    else if (ttr_or_l1a == 2) {str = "TTR1"}
    else if (ttr_or_l1a == 3) {str = "TTR2"}
    else if (ttr_or_l1a == 4) {str = "TTR3"}
    bs.TTSelect(str);*/
   
    bs.TTSelect(ttr_or_l1a);
    // WRITE
    if ((m_status = m_alti->BSY_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing busy control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCTTRSelRead(unsigned int &ttr_or_l1a) {
    ALTI_BSY_CONTROL_BITSTRING bs; // I'm using bits in this register because it's the easiest way. It will be modified with the new firmware to something more appropiate
    // READ
    if ((m_status = m_alti->BSY_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading busy control register", "");
        return (m_status);
    }
    ttr_or_l1a = bs.TTSelect();
    
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCL1ASourceWrite(const std::vector<bool> enabled) {
  
  unsigned int sel;
  unsigned int l1a  = enabled[0];
  unsigned int ttr1 = enabled[1];
  unsigned int ttr2 = enabled[2];
  unsigned int ttr3 = enabled[3];
  sel = l1a + ttr1*2 + ttr2*4 + ttr3*8;
 
  if ((m_status = ENCTTRSelWrite(sel)) != 0) {
      CERR("writing TTC encoder L1A source", "");
      return(m_status);
  }
    return (m_status);
}


//------------------------------------------------------------------------------

int AltiModule::ENCL1ASourceRead(std::vector<bool> &enabled) {

    // READ
    u_int sel = 0;
    if ((m_status = ENCTTRSelRead(sel)) != 0) {
        CERR("ALTI: reading TTC L1A source", "");
        return (m_status);
    }
    
    enabled.clear();
    enabled.resize(4);
    enabled[0] = sel%2; //L1A 
    sel = sel / 2;
    enabled[1] = sel%2; // TTR1
    sel = sel / 2;
    enabled[2] = sel%2; // TTR2
    sel = sel / 2;
    enabled[3] = sel%2; // TTR3
    
    return (m_status);
}


//------------------------------------------------------------------------------

/*int AltiModule::TTCTransmitterStatusRead(const unsigned int tx, bool &enabled, bool &fault) {
    if (tx >= TRANSMITTER_NUMBER) {
        CERR("ALTI: wrong TTC TX number chosen", "");
        return (-1);
    }

    unsigned int status;

    // READ
    if ((m_status = m_alti->TTC_SFPControl_Read(status)) != 0 ) {
        CERR("ALTI: reading TTC control register", "");
        return (m_status);
    }

    enabled = !(status & (0x00000001 << tx));
    fault = (status & (0x00001000 << tx));

    return (m_status);
}*/

//------------------------------------------------------------------------------

/*int AltiModule::TTCTransmitterStatusRead(std::vector<bool> &enabled, std::vector<bool> &fault) {
    unsigned int status;
    enabled.clear();
    fault.clear();

    // READ
    if ((m_status = m_alti->TTC_SFPControl_Read(status)) != 0 ) {
        CERR("ALTI: reading TTC control register", "");
        return (m_status);
    }

    unsigned int i;
    for (i = 0; i < TRANSMITTER_NUMBER; i++) {
        enabled.push_back(!(status & (0x00000001 << i)));
        fault.push_back(status & (0x00001000 << i));
    }

    return (m_status);
}*/

//------------------------------------------------------------------------------

int AltiModule::DECEnableRead(bool &enabled) {
    ALTI_CNT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CNT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading TTC decoder snapshot memory control register", "");
        return (m_status);
    }

    std::string s = bs.EnableTTCTimestamp();
    if (s == "Enabled") {
        enabled = true;
    }
    else if (s == "Disabled") {
        enabled = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for TTC decoder enable", s.c_str());
        m_status = FAILURE;
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::DECEnableWrite(const bool enable) {
    ALTI_CNT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CNT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading TTC decoder snapshot memory control register", "");
        return (m_status);
    }
    // MODIFY
    bs.EnableTTCTimestamp(enable ? "Enabled" : "Disabled");
    // WRITE
    if ((m_status = m_alti->CNT_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing TTC decoder snapshot memory control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::DECCurrentAddressRead(unsigned int &cur) {
    if ((m_status = m_alti->TTC_TTCCurrentSPYAddress_Read(cur)) != 0) {
        CERR("ALTI: reading TTC decoder snapshot current address", "");
        return (m_status);
    }

    return (m_status);
}


//------------------------------------------------------------------------------

int AltiModule::DECStatusRead(bool &cdr_los, bool &cdr_lol, bool &pll_lol, bool &singleBitError, bool &doubleBitError) {
    ALTI_TTC_SFPCONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->TTC_SFPControl_Read(bs)) != 0 ) {
        CERR("ALTI: reading TTC decoder status register", "");
        return (m_status);
    }

    std::string s;

    s = bs.ClockDataRecoveryLOS();
    if (s == "LossOfSignal") {
        cdr_los = true;
    }
    else if (s == "NoLossOfSignal") {
        cdr_los = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for CDR input signal status", s.c_str());
        return (-1);
    }

    s = bs.ClockDataRecoveryLOL();
    if (s == "LossOfLock") {
        cdr_lol = true;
    }
    else if (s == "NoLossOfLock") {
        cdr_lol = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for CDR lock status", s.c_str());
        return (-1);
    }

    s = bs.TTCDecoderPLLLockStatus();
    if (s == "LossOfLock") {
        pll_lol = true;
    }
    else if (s == "NoLossOfLock") {
        pll_lol = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for TTC decoder PLL lock status", s.c_str());
        return (-1);
    }

    singleBitError = bs.TTCDecoderSingleBitError();
    doubleBitError = bs.TTCDecoderDoubleBitError();

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::DECStatusClear() {
    unsigned int data;

    // READ
    if ((m_status = m_alti->TTC_SFPControl_Read(data)) != 0) {
        CERR("ALTI: reading TTC control register", "");
        return (m_status);
    }

    // MODIFY
    data &= ~(ALTI_TTC_SFPCONTROL_BITSTRING::STAT_MASK_CLOCKDATARECOVERYLOS + ALTI_TTC_SFPCONTROL_BITSTRING::STAT_MASK_CLOCKDATARECOVERYLOL + ALTI_TTC_SFPCONTROL_BITSTRING::STAT_MASK_TTCDECODERSINGLEBITERROR + ALTI_TTC_SFPCONTROL_BITSTRING::STAT_MASK_TTCDECODERDOUBLEBITERROR + ALTI_TTC_SFPCONTROL_BITSTRING::STAT_MASK_TTCDECODERPLLLOCKSTATUS); // clear
    unsigned int i, mask;
    for (i = 0; i < TRANSMITTER_NUMBER; i++) {
        mask = (ALTI_TTC_SFPCONTROL_BITSTRING::MASK_TXFAULTSTATUS[i]).number()[0];
        data |= mask; // don't clear
    }

    // WRITE
    if ((m_status = m_alti->TTC_SFPControl_Write(data)) != 0) {
        CERR("ALTI: writing TTC control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::DECReset() {
    ALTI_CNT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CNT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading TTC decoder snapshot memory control register", "");
        return (m_status);
    }
    // MODIFY
    bs.ResetAddressTTCTimestamp("Enabled");
    // WRITE
    if ((m_status = m_alti->CNT_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing TTC decoder snapshot memory control register", "");
        return (m_status);
    }
    // MODIFY
    bs.ResetAddressTTCTimestamp("Disabled");
    // WRITE
    if ((m_status = m_alti->CNT_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing TTC decoder snapshot memory control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::DECRead(RCD::CMEMSegment *segment0, RCD::CMEMSegment *segment1, const unsigned int start, const unsigned int num, unsigned int bcr_word) { // read TTC decoder memory
    if (start > MAX_MEMORY) {
        CERR("ALTI: invalid start index of snapshot memory", "");
        m_status = WRONGPAR;
        return (m_status);
    }
    else if (start + num > MAX_MEMORY) {
        CERR("ALTI: invalid number of data to be read from the snapshot memory", "");
        m_status = WRONGPAR;
        return (m_status);
    }
   
    if ((m_status = MEMDataRead(ALTI_RAM_MEMORY::TTC_DECODER_COMMAND, segment0, num, start, start)) != 0) {
        CERR("ALTI: reading \"%s\" memory in block cycle", ALTI_RAM_MEMORY_NAME[TTC_DECODER_COMMAND].c_str());
        return (m_status);
    }
   
    if ((m_status = MEMDataRead(ALTI_RAM_MEMORY::TTC_DECODER_TIMESTAMP, segment1, num, start, start)) != 0) {
        CERR("ALTI: reading \"%s\" memory in block cycle", ALTI_RAM_MEMORY_NAME[TTC_DECODER_TIMESTAMP].c_str());
        return (m_status);
    }
    /*std::vector<unsigned int> data[2];
    data[0].resize(num);
    data[1].resize(num);
    if ((m_status = MEMDataRead(TTC_DECODER_COMMAND, data[0], start)) != 0) {
        CERR("ALTI: reading \"%s\" memory in single cycle", ALTI_RAM_MEMORY_NAME[TTC_DECODER_COMMAND].c_str());
        return (m_status);
    }
    if ((m_status = MEMDataRead(TTC_DECODER_TIMESTAMP, data[1], start)) != 0) {
        CERR("ALTI: reading \"%s\" memory in single cycle", ALTI_RAM_MEMORY_NAME[TTC_DECODER_TIMESTAMP].c_str());
        return (m_status);
    }*/

    unsigned int indx;
    unsigned int *data[2];
    data[0] = (unsigned int *) segment0->VirtualAddress();
    data[1] = (unsigned int *) segment1->VirtualAddress();
    TTC_SPY_COMMAND command;
    TTC_SPY_TIMESTAMP timestamp;
    std::vector<unsigned int> tmp;
    //bool l1a, frame_long;

    std::string readable;

    int format = 2;
    switch (format) {
        /*case 0: // hex
            if ((m_status = AltiCommon::writeFile(outf, num, data[0] + start, data[1] + start)) != 0) {
                CERR("ALTI: reading TTC decoder memories into hex file ", "");
                return (m_status);
            }
            break;*/
        /*case 1: // human-readable
            outf << "#------------------------------------------------------------------------------------------\n";
            outf << "#        | A     |       |       |         |       | F     |       |            |          \n";
            outf << "#        | d     |       |       |         |       | r     |       |            |          \n";
            outf << "#        | d     |       |       |         |       | a     |       |            |          \n";
            outf << "#        | r     | S     |       |         |       | m     |       |            |          \n";
            outf << "#        |       | u     |       | S       |       | e     |       |            |          \n";
            outf << "#        | s     | b     |       | t       |       |       | E     |            |          \n";
            outf << "#A       | p     | a     | D     | a       |       | t     | r     |            |          \n";
            outf << "#d       | a     | d     | a     | t       |       | y     | r     |            |          \n";
            outf << "#d       | c     | d     | t     | u       |       | p     | o     | BC         | BC       \n";
            outf << "#r       | e     | r     | a     | s       | L1A   | e     | r     | cnt        | cnt      \n";
            outf << "#(hex)   | (E/I) | (hex) | (hex) | (OK/ER) | (bin) | (L/S) | (bin) | (hex)      | (dec)    \n";
            outf << "#------------------------------------------------------------------------------------------\n";

            for (indx = start; indx < start + num; indx++) {
                tmp.clear();
                tmp.push_back(static_cast<unsigned int>(data[0][indx]));
                command = (RCD::BitSet) tmp;
                tmp.clear();
                tmp.push_back(static_cast<unsigned int>(data[1][indx]));
                timestamp = (RCD::BitSet) tmp;

                l1a = (data[1][indx] & TTC_SPY_TIMESTAMP::STAT_MASK_L1A);
                frame_long = (command.FrameType() == "Long");

                std::ostringstream readable;
                if ((l1a == false) && (frame_long == true)) {
                    readable << "0x" << std::setfill('0') << std::setw(4) << std::hex << command.Address() << "  ";
                }
                else {
                    readable << "        ";
                }
                readable << " | ";
                if ((l1a == false) && (frame_long == true)) {
                    readable << ((data[0][indx] & TTC_SPY_COMMAND::STAT_MASK_ADDRESSSPACE) ? "E    " : "I    ");
                }
                else {
                    readable << "     ";
                }
                readable << " | ";
                if ((l1a == false) && (frame_long == true)) {
                    readable << "0x" << std::setfill('0') << std::setw(2) << std::hex << command.SubAddress() << " ";
                }
                else {
                    readable << "     ";
                }
                readable << " | ";
                if (l1a == false) {
                    readable << "0x" << std::setfill('0') << std::setw(2) << std::hex << command.Data() << " ";
                }
                else {
                    readable << "     ";
                }
                readable << " | ";
                readable << ((data[1][indx] & TTC_SPY_TIMESTAMP::STAT_MASK_LINKSTATUS) ? "OK     " : "ER     ");
                readable << " | ";
                readable << (l1a ? "1    " : "0    ");
                readable << " | ";
                if (l1a == false) {
                    readable << ((data[0][indx] & TTC_SPY_COMMAND::STAT_MASK_FRAMETYPE) ? "L    " : "S    ");
                }
                else {
                    readable << "     ";
                }
                readable << " | ";
                readable << "s" << ((data[1][indx] & TTC_SPY_TIMESTAMP::STAT_MASK_SINGLEBITERROR) ? "1" : "0") << "d" << ((data[1][indx] & TTC_SPY_TIMESTAMP::STAT_MASK_DOUBLEBITERROR) ? "1 " : "0 ");
                readable << " | ";
                readable << "0x" << std::setfill('0') << std::setw(7) << std::hex << timestamp.BC_Cnt() << " ";
                readable << " | ";
                readable << std::setfill('0') << std::setw(9) << std::dec << timestamp.BC_Cnt();
                outf << readable.str() << "\n";
            }
            break;*/
        case 2: // human-readable, TTC scope
            //std::cout << "|  Msg    |    TTCrx  E   SUB    DATA  |  CHCK VAL  |   BC Count  |\n";
            //std::cout << "|---------|----------------------------|------------|-------------|\n";
            std::cout << "|  Msg    |      TTCrx  E   SUB    DATA  |  Pseudo-L1ID  |  CHCK VAL  |  BCID  |  BC Id:fied  |   BC Count  |\n";
            std::cout << "|         |    [Type CNT-HEX (CNT-DEC)]  |               |            |        |              |             |\n";
            std::cout << "|---------|------------------------------|---------------|------------|--------|--------------|-------------|\n";
            bool l1a, frame_long;
            unsigned int pseudo_l1id;
            pseudo_l1id = 0;
            bool bcr_found;
            bcr_found = false;
            unsigned previous_bcid;
            previous_bcid = 0;
            unsigned int bcid;
	    bcid = 0;
            int ttyp_subaddr;
            ttyp_subaddr = -1;
            unsigned int ttyp_addr;
            bool ttyp_external;
            ttyp_external = false;
            bool ttyp_recognized;
            ttyp_recognized = false;
            unsigned int ttyp_word;
            unsigned int ttyp_cnt[3];
            for (indx = start; indx < start + num; indx++) {
                MEMCommandTimestampToBitstrings(data[0][indx], data[1][indx], command, timestamp);
                l1a = timestamp.L1A();
                frame_long = (command.FrameType() == "Long");

		
		// BC Identified
		if (!frame_long && (command.Data() == bcr_word)) {
		  bcid = 0;
		  if(!bcr_found) bcr_found = true;
		}
                else {		
		  bcid += (timestamp.BC_Cnt() - previous_bcid); 
		}
		previous_bcid = timestamp.BC_Cnt();

                // Channel-A command
                if(l1a){
                  std::ostringstream buf_a;
                  buf_a << "|";
                  buf_a << "  " << std::setw(5) << std::left << "L1A" << "  |";
                  buf_a << "                              |";
                  buf_a << "   0x" << std::hex << std::setw(8) << std::setfill('0') << std::right << (pseudo_l1id++) << "  |";
                  buf_a << "     " << ((timestamp.SingleBitError() || timestamp.DoubleBitError()) ? "ERR" : "OK ") << "    |";
                  buf_a << "  " << std::dec << std::setw(4) << std::setfill('0') << std::right << (bcid) << "  |";
                  buf_a << "     " << std::setw(3) << (bcr_found ? "YES" : "NO ") << std::left << "      |";
                  buf_a << "  " << std::dec << std::setw(9) << std::setfill('0') << std::right << timestamp.BC_Cnt() << "  |";
                  std::cout << buf_a.str() << "\n";
                  l1a = false;
                } // end Channel-A

                // Channel-B command
                if(command.Data()!=0 || frame_long){

                  std::ostringstream buf;
                  //buf.clear();

                  buf << "|";
                  buf << "  " << std::setw(5) << std::left << (l1a ? "L1A" : command.FrameType()) << "  |";
                  if ((!l1a) && frame_long) { // Long
                      buf << "    A 0x" << std::hex << std::setw(4) << std::setfill('0') << std::right << command.Address() << ((command.AddressSpace() == "External") ? " 1" : ((command.AddressSpace() == "Internal") ? " 0" : "   ")) << " S 0x" << std::hex << std::setw(2) << std::setfill('0') << std::right << command.SubAddress() << " D 0x" << std::hex << std::setw(2) << std::setfill('0') << std::right << command.Data() << "  |";

                      // check the ttypes
                      if (command.SubAddress() == ((unsigned int) (ttyp_subaddr + 1))) {
                          if (ttyp_subaddr == -1) { // start
                              ttyp_subaddr = 0;
                              ttyp_addr = command.Address();
                              ttyp_external = (command.AddressSpace() == "External");
                              ttyp_word = command.Data();
                          }
                          else if ((ttyp_addr >= 0) && (ttyp_addr < 3)) {
                              if ((command.Address() == ttyp_addr) && (command.AddressSpace() == (ttyp_external ? "External" : "Internal"))) {
                                  ttyp_cnt[ttyp_subaddr++] = command.Data();
                                  if (ttyp_subaddr == 3) { // end
                                      ttyp_recognized = true;
                                      ttyp_subaddr = -1;
                                  }
                              }
                              else { // start over
                                  ttyp_subaddr = -1;
                              }
                          }
                          else { // start over
                              ttyp_subaddr = -1;
                          }
                      }
                      else { // start over
                          ttyp_subaddr = -1;
                      }
  
                  }
                  else if (!frame_long) { // Short
                      buf << "                      D 0x" << std::hex << std::setw(2) << std::setfill('0') << std::right << command.Data() << "  |";
                  }
                  else if (l1a) { // L1A
                      buf << "                              |";
                  }
  
                  // Pseudo-L1ID
                  if (l1a) {
                      buf << "   0x" << std::hex << std::setw(8) << std::setfill('0') << std::right << (pseudo_l1id++) << "  |";
                  }
                  else {
                      buf << "               |";
                  }
  
                  buf << "     " << ((timestamp.SingleBitError() || timestamp.DoubleBitError()) ? "ERR" : "OK ") << "    |";
  
                  // BCID
                  buf << "  " << std::dec << std::setw(4) << std::setfill('0') << std::right << (bcid) << "  |";
  
                  // BC Id:fied
                  buf << "     " << std::setw(3) << (bcr_found ? "YES" : "NO ") << std::left << "      |";  
                  buf << "  " << std::dec << std::setw(9) << std::setfill('0') << std::right << timestamp.BC_Cnt() << "  |";
  
                  std::cout << buf.str() << "\n";
  
                  if (ttyp_recognized) {
                      std::ostringstream buf1;
                      buf1 << "|  TType  |  [0x" << std::hex << std::setw(2) << std::setfill('0') << std::right << ttyp_word << " 0x" << std::hex << std::setw(6) << ((ttyp_cnt[0] << 16) + (ttyp_cnt[1] << 8) + ttyp_cnt[2]) << " (" << std::dec << std::setw(8) << std::setfill('0') << std::right << ((ttyp_cnt[0] << 16) + (ttyp_cnt[1] << 8) + ttyp_cnt[2]) << ")]  |";
                      buf1 << "               |"; // Pseudo-L1ID
                      buf1 << "     " << ((timestamp.SingleBitError() || timestamp.DoubleBitError()) ? "ERR" : "OK ") << "    |";
                      buf1 << "  " << std::dec << std::setw(4) << std::setfill('0') << std::right << (bcid) << "  |";
                      buf1 << "     " << std::setw(3) << (bcr_found ? "YES" : "NO ") << std::left << "      |";
                      buf1 << "  " << std::dec << std::setw(9) << std::setfill('0') << std::right << timestamp.BC_Cnt() << "  |";
  
                      ttyp_recognized = false;
  
                      std::cout << buf1.str() << "\n";
                  }
                } // end Channel-B
            }
            break;
        default:
            break;
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::DECReadFile(RCD::CMEMSegment *segment0, RCD::CMEMSegment *segment1, const unsigned int start, const unsigned int num, unsigned int bcr_word, const int format, std::ofstream &outf) { // read TTC decoder memory, into file
    if (start > MAX_MEMORY) {
        CERR("ALTI: invalid start index of snapshot memory", "");
        m_status = WRONGPAR;
        return (m_status);
    }
    else if (start + num > MAX_MEMORY) {
        CERR("ALTI: invalid number of data to be read from the snapshot memory", "");
        m_status = WRONGPAR;
        return (m_status);
    }
    //if ((m_status = MEMDataRead(ALTI_RAM_MEMORY::TTC_DECODER_COMMAND, segment0, MAX_MEMORY, 0, 0)) != 0) {
    if ((m_status = MEMDataRead(ALTI_RAM_MEMORY::TTC_DECODER_COMMAND, segment0, num, start, start)) != 0) {
        CERR("ALTI: reading \"%s\" memory in block cycle", ALTI_RAM_MEMORY_NAME[TTC_DECODER_COMMAND].c_str());
        return (m_status);
    }
    //if ((m_status = MEMDataRead(ALTI_RAM_MEMORY::TTC_DECODER_TIMESTAMP, segment1, MAX_MEMORY, 0, 0)) != 0) {
    if ((m_status = MEMDataRead(ALTI_RAM_MEMORY::TTC_DECODER_TIMESTAMP, segment1, num, start, start)) != 0) {
        CERR("ALTI: reading \"%s\" memory in block cycle", ALTI_RAM_MEMORY_NAME[TTC_DECODER_TIMESTAMP].c_str());
        return (m_status);
    }
    /*std::vector<unsigned int> data[2];
    data[0].resize(num);
    data[1].resize(num);
    if ((m_status = MEMDataRead(TTC_DECODER_COMMAND, data[0], start)) != 0) {
        CERR("ALTI: reading \"%s\" memory in single cycle", ALTI_RAM_MEMORY_NAME[TTC_DECODER_COMMAND].c_str());
        return (m_status);
    }
    if ((m_status = MEMDataRead(TTC_DECODER_TIMESTAMP, data[1], start)) != 0) {
        CERR("ALTI: reading \"%s\" memory in single cycle", ALTI_RAM_MEMORY_NAME[TTC_DECODER_TIMESTAMP].c_str());
        return (m_status);
    }*/

    unsigned int indx;
    unsigned int *data[2];
    data[0] = (unsigned int *) segment0->VirtualAddress();
    data[1] = (unsigned int *) segment1->VirtualAddress();
    TTC_SPY_COMMAND command;
    TTC_SPY_TIMESTAMP timestamp;
    std::vector<unsigned int> tmp;
    bool l1a, frame_long;

    std::string readable;

    switch (format) {
        case 0: // hex
            if ((m_status = AltiCommon::writeFile(outf, num, data[0] + start, data[1] + start)) != 0) {
            //if ((m_status = AltiCommon::writeFile(outf, num, &(data[0][0]), &(data[1][0]))) != 0) {
                CERR("ALTI: reading TTC decoder memories into hex file ", "");
                return (m_status);
            }
            break;
        case 1: // human-readable
            outf << "#------------------------------------------------------------------------------------------\n";
            outf << "#        | A     |       |       |         |       | F     |       |            |          \n";
            outf << "#        | d     |       |       |         |       | r     |       |            |          \n";
            outf << "#        | d     |       |       |         |       | a     |       |            |          \n";
            outf << "#        | r     | S     |       |         |       | m     |       |            |          \n";
            outf << "#        |       | u     |       | S       |       | e     |       |            |          \n";
            outf << "#        | s     | b     |       | t       |       |       | E     |            |          \n";
            outf << "#A       | p     | a     | D     | a       |       | t     | r     |            |          \n";
            outf << "#d       | a     | d     | a     | t       |       | y     | r     |            |          \n";
            outf << "#d       | c     | d     | t     | u       |       | p     | o     | BC         | BC       \n";
            outf << "#r       | e     | r     | a     | s       | L1A   | e     | r     | cnt        | cnt      \n";
            outf << "#(hex)   | (E/I) | (hex) | (hex) | (OK/ER) | (bin) | (L/S) | (bin) | (hex)      | (dec)    \n";
            outf << "#------------------------------------------------------------------------------------------\n";

            for (indx = start; indx < start + num; indx++) {
            //for (indx = 0; indx < num; indx++) {
                tmp.clear();
                tmp.push_back(static_cast<unsigned int>(data[0][indx]));
                command = (RCD::BitSet) tmp;
                tmp.clear();
                tmp.push_back(static_cast<unsigned int>(data[1][indx]));
                timestamp = (RCD::BitSet) tmp;

                l1a = (data[1][indx] & TTC_SPY_TIMESTAMP::STAT_MASK_L1A);
                frame_long = (command.FrameType() == "Long");

                std::ostringstream readable;
                if ((l1a == false) && (frame_long == true)) {
                    readable << "0x" << std::setfill('0') << std::setw(4) << std::hex << command.Address() << "  ";
                }
                else {
                    readable << "        ";
                }
                readable << " | ";
                if ((l1a == false) && (frame_long == true)) {
                    readable << ((data[0][indx] & TTC_SPY_COMMAND::STAT_MASK_ADDRESSSPACE) ? "E    " : "I    ");
                }
                else {
                    readable << "     ";
                }
                readable << " | ";
                if ((l1a == false) && (frame_long == true)) {
                    readable << "0x" << std::setfill('0') << std::setw(2) << std::hex << command.SubAddress() << " ";
                }
                else {
                    readable << "     ";
                }
                readable << " | ";
                if (l1a == false) {
                    readable << "0x" << std::setfill('0') << std::setw(2) << std::hex << command.Data() << " ";
                }
                else {
                    readable << "     ";
                }
                readable << " | ";
                readable << ((data[1][indx] & TTC_SPY_TIMESTAMP::STAT_MASK_LINKSTATUS) ? "OK     " : "ER     ");
                readable << " | ";
                readable << (l1a ? "1    " : "0    ");
                readable << " | ";
                if (l1a == false) {
                    readable << ((data[0][indx] & TTC_SPY_COMMAND::STAT_MASK_FRAMETYPE) ? "L    " : "S    ");
                }
                else {
                    readable << "     ";
                }
                readable << " | ";
                readable << "s" << ((data[1][indx] & TTC_SPY_TIMESTAMP::STAT_MASK_SINGLEBITERROR) ? "1" : "0") << "d" << ((data[1][indx] & TTC_SPY_TIMESTAMP::STAT_MASK_DOUBLEBITERROR) ? "1 " : "0 ");
                readable << " | ";
                readable << "0x" << std::setfill('0') << std::setw(7) << std::hex << timestamp.BC_Cnt() << " ";
                readable << " | ";
                readable << std::setfill('0') << std::setw(9) << std::dec << timestamp.BC_Cnt();
                outf << readable.str() << "\n";
            }
            break;
        case 2: // human-readable, TTC scope
            //outf << "|  Msg    |    TTCrx  E   SUB    DATA  |  CHCK VAL  |   BC Count  |\n";
            //outf << "|---------|----------------------------|------------|-------------|\n";
            /*for (indx = start; indx < start + num; indx++) {
                MEMCommandTimestampToScopeString(data[0][indx], data[1][indx], readable);
                outf << readable << "\n";
            }*/
            outf << "|  Msg    |      TTCrx  E   SUB    DATA  |  Pseudo-L1ID  |  CHCK VAL  |  BCID  |  BC Id:fied  |   BC Count  |\n";
            outf << "|         |    [Type CNT-HEX (CNT-DEC)]  |               |            |        |              |             |\n";
            outf << "|---------|------------------------------|---------------|------------|--------|--------------|-------------|\n";
            bool l1a, frame_long;
            unsigned int pseudo_l1id;
            pseudo_l1id = 0;
            bool bcr_found;
            bcr_found = false;
            unsigned bcr_offset;
            bcr_offset = 0;
            unsigned int bcid;
            int ttyp_subaddr;
            ttyp_subaddr = -1;
            unsigned int ttyp_addr;
            bool ttyp_external;
            ttyp_external = false;
            bool ttyp_recognized;
            ttyp_recognized = false;
            unsigned int ttyp_word;
            unsigned int ttyp_cnt[3];
            for (indx = start; indx < start + num; indx++) {
            //for (indx = 0; indx < num; indx++) {
                //MEMCommandTimestampToScopeString(data[0][indx], data[1][indx], readable);
                //std::cout << readable << "\n";
                MEMCommandTimestampToBitstrings(data[0][indx], data[1][indx], command, timestamp);
                l1a = timestamp.L1A();
                frame_long = (command.FrameType() == "Long");

                std::ostringstream buf;
                //buf.clear();

                buf << "|";
                buf << "  " << std::setw(5) << std::left << (l1a ? "L1A" : command.FrameType()) << "  |";
                if ((!l1a) && frame_long) { // Long
                    buf << "    A 0x" << std::hex << std::setw(4) << std::setfill('0') << std::right << command.Address() << ((command.AddressSpace() == "External") ? " 1" : ((command.AddressSpace() == "Internal") ? " 0" : "   ")) << " S 0x" << std::hex << std::setw(2) << std::setfill('0') << std::right << command.SubAddress() << " D 0x" << std::hex << std::setw(2) << std::setfill('0') << std::right << command.Data() << "  |";

                    // check the ttypes
                    if (command.SubAddress() == ((unsigned int) (ttyp_subaddr + 1))) {
                        if (ttyp_subaddr == -1) { // start
                            ttyp_subaddr = 0;
                            ttyp_addr = command.Address();
                            ttyp_external = (command.AddressSpace() == "External");
                            ttyp_word = command.Data();
                        }
                        else if ((ttyp_addr >= 0) && (ttyp_addr < 3)) {
                            if ((command.Address() == ttyp_addr) && (command.AddressSpace() == (ttyp_external ? "External" : "Internal"))) {
                                ttyp_cnt[ttyp_subaddr++] = command.Data();
                                if (ttyp_subaddr == 3) { // end
                                    ttyp_recognized = true;
                                    ttyp_subaddr = -1;
                                }
                            }
                            else { // start over
                                ttyp_subaddr = -1;
                            }
                        }
                        else { // start over
                            ttyp_subaddr = -1;
                        }
                    }
                    else { // start over
                        ttyp_subaddr = -1;
                    }

                }
                else if ((!l1a) && (!frame_long)) { // Short
                    buf << "                      D 0x" << std::hex << std::setw(2) << std::setfill('0') << std::right << command.Data() << "  |";
                }
                else if (l1a) { // L1A
                    buf << "                              |";
                }

                // Pseudo-L1ID
                if (l1a) {
                    buf << "   0x" << std::hex << std::setw(8) << std::setfill('0') << std::right << (pseudo_l1id++) << "  |";
                }
                else {
                    buf << "               |";
                }

                buf << "     " << ((timestamp.SingleBitError() || timestamp.DoubleBitError()) ? "ERR" : "OK ") << "    |";

                // BCID
                if (bcr_found == false) {
                    bcid = 0;
                }
                else {
                    bcid = ((timestamp.BC_Cnt() - bcr_offset + ORBIT_LENGTH - 1) % ORBIT_LENGTH); // BCID of the BCR = 3563
                }
                buf << "  " << std::dec << std::setw(4) << std::setfill('0') << std::right << (bcid) << "  |";

                // BC Id:fied
                buf << "     " << std::setw(3) << (bcr_found ? "YES" : "NO ") << std::left << "      |";
                if (bcr_found == false) {
                    if ((!l1a) && (!frame_long) && (command.Data() == bcr_word)) {
                        bcr_found = true;
                        bcr_offset = (timestamp.BC_Cnt() % ORBIT_LENGTH);
                    }
                }

                buf << "  " << std::dec << std::setw(9) << std::setfill('0') << std::right << timestamp.BC_Cnt() << "  |";

                outf << buf.str() << "\n";

                if (ttyp_recognized) {
                    std::ostringstream buf1;
                    buf1 << "|  TType  |  [0x" << std::hex << std::setw(2) << std::setfill('0') << std::right << ttyp_word << " 0x" << std::hex << std::setw(6) << ((ttyp_cnt[0] << 16) + (ttyp_cnt[1] << 8) + ttyp_cnt[2]) << " (" << std::dec << std::setw(8) << std::setfill('0') << std::right << ((ttyp_cnt[0] << 16) + (ttyp_cnt[1] << 8) + ttyp_cnt[2]) << ")]  |";
                    buf1 << "               |"; // Pseudo-L1ID
                    buf1 << "     " << ((timestamp.SingleBitError() || timestamp.DoubleBitError()) ? "ERR" : "OK ") << "    |";
                    buf1 << "  " << std::dec << std::setw(4) << std::setfill('0') << std::right << (bcid) << "  |";
                    buf1 << "     " << std::setw(3) << (bcr_found ? "YES" : "NO ") << std::left << "      |";
                    buf1 << "  " << std::dec << std::setw(9) << std::setfill('0') << std::right << timestamp.BC_Cnt() << "  |";

                    ttyp_recognized = false;

                    outf << buf1.str() << "\n";
                }
            }
            break;
        default:
            break;
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::MEMCommandTimestampToScopeString(const unsigned int command, const unsigned int timestamp, std::string &readable) { // TTC decoder command/timestamp memory entries -> user readable ttcscope line

    TTC_SPY_COMMAND command_bs;
    TTC_SPY_TIMESTAMP timestamp_bs;
    std::vector<unsigned int> tmp;

    tmp.clear();
    tmp.push_back(static_cast<unsigned int>(command));
    command_bs = (RCD::BitSet) tmp;
    tmp.clear();
    tmp.push_back(static_cast<unsigned int>(timestamp));
    timestamp_bs = (RCD::BitSet) tmp;

    bool l1a, frame_long;
    l1a = (timestamp & TTC_SPY_TIMESTAMP::STAT_MASK_L1A);
    frame_long = (command_bs.FrameType() == "Long");

    std::ostringstream buf;

    buf << "|";
    buf << "  " << std::setw(5) << std::left << (l1a ? "L1A" : command_bs.FrameType()) << "  |";
    if ((!l1a) && frame_long) { // Long
        buf << "  A 0x" << std::hex << std::setw(4) << std::setfill('0') << std::right << command_bs.Address() << ((command_bs.AddressSpace() == "External") ? " 1" : ((command_bs.AddressSpace() == "Internal") ? " 0" : "   ")) << " S 0x" << std::hex << std::setw(2) << std::setfill('0') << std::right << command_bs.SubAddress() << " D 0x" << std::hex << std::setw(2) << std::setfill('0') << std::right << command_bs.Data() << "  |";
    }
    else if ((!l1a) && (!frame_long)) { // Short
        buf << "                    D 0x" << std::hex << std::setw(2) << std::setfill('0') << std::right << command_bs.Data() << "  |";
    }
    else if (l1a) { // L1A
        buf << "                            |";
    }
    buf << "     " << ((timestamp_bs.SingleBitError() || timestamp_bs.DoubleBitError()) ? "ERR" : "OK ") << "    |";
    buf << "  " << std::dec << std::setw(9) << std::setfill('0') << std::right << timestamp_bs.BC_Cnt() << "  |";

    readable = buf.str();

    return (0);
}

//------------------------------------------------------------------------------

int AltiModule::MEMScopeStringToCommandTimestamp(const std::string &readable, unsigned int &command, TTC_SPY_COMMAND &command_bs, unsigned int &timestamp, TTC_SPY_TIMESTAMP &timestamp_bs) { // user readable ttcscope line -> TTC decoder command/timestamp memory entries

    std::stringstream ss;
    std::vector<std::string> arg;
    int num;
    if ((num  = AltiCommon::splitString(readable, "|", arg)) != 9) {
        CERR("ALTI: wrong number of fields in ttcscope line\n%s", readable.c_str());
        return (-1);
    }

    /*unsigned int i;
    for (i = 0; i < num; i++) {
        std::printf("arg[%d]:%s\n", i, AltiCommon::trimString(arg[i]).c_str());
    }*/

    //TTC_SPY_COMMAND command_bs;
    //TTC_SPY_TIMESTAMP timestamp_bs;
    std::vector<unsigned int> tmp;

    tmp.clear();
    tmp.push_back(static_cast<unsigned int>(0));
    command_bs = (RCD::BitSet) tmp;
    timestamp_bs = (RCD::BitSet) tmp;

    // BC_Cnt field
    arg[7] = AltiCommon::trimString(arg[7]);
    ss << arg[7];
    unsigned int bc_count;
    ss >> bc_count;
    bc_count &= TTC_SPY_TIMESTAMP::STAT_MASK_BC_CNT; // keep only 28 bits
    //std::printf("bc_count: 0x%07x\n", bc_count);
    timestamp_bs.BC_Cnt(bc_count);

    // SingleOrDoubleBitError field
    arg[4] = AltiCommon::trimString(arg[4]);
    ss << arg[4];
    bool error = (arg[4] == "OK" || arg[4].empty()) ? false : true;
    //std::printf("error: %s\n", error ? "ER" : "OK");
    timestamp_bs.SingleBitError(error);
    timestamp_bs.DoubleBitError(error);

    ss.clear();
    ss << arg[2];

    arg[1] = AltiCommon::trimString(arg[1]);
    if (arg[1] == "L1A") {
        timestamp_bs.L1A(1);
        //command_bs.FrameType("Long"); // this bit is set like this in the FW, check this...
    }
    else if (arg[1] == "Short") {
        timestamp_bs.L1A(0);
        command_bs.FrameType("Short");
        char c;
        unsigned int data, checksum;
        ss >> c >> std::hex >> data >> std::hex >> checksum;
        data &= TTC_SPY_COMMAND::STAT_MASK_DATA; // keep only 8 bits
        checksum &= TTC_SPY_COMMAND::STAT_MASK_DATA; // keep only 8 bits
        /*std::printf("c: %c\n", c);
        std::printf("data: 0x%02x\n", data);
        std::printf("checksum: 0x%02x\n", checksum);*/
        if (c != 'D') {
            CERR("ALTI: wrong ttcscope Short command line, expected \"D\" field, got \"%c\" instead", c);
            return (-1);
        }
        command_bs.Data(data);
    }
    else if (arg[1] == "Long") {
        timestamp_bs.L1A(0);
        command_bs.FrameType("Long");
        char c1, c2, c3;
        unsigned int addr, subaddr, data, checksum;
        bool external;
        ss >> c1 >> std::hex >> addr >> external >> c2 >> std::hex >> subaddr >> c3 >> std::hex >> data >> std::hex >> checksum;
        addr &= (TTC_SPY_COMMAND::STAT_MASK_ADDRESS >> AltiCommon::getFirstSetBitPos(TTC_SPY_COMMAND::STAT_MASK_ADDRESS)); // keep only 14 bits
        subaddr &= (TTC_SPY_COMMAND::STAT_MASK_SUBADDRESS >> AltiCommon::getFirstSetBitPos(TTC_SPY_COMMAND::STAT_MASK_SUBADDRESS)); // keep only 8 bits
        data &= (TTC_SPY_COMMAND::STAT_MASK_DATA >> AltiCommon::getFirstSetBitPos(TTC_SPY_COMMAND::STAT_MASK_DATA)); // keep only 8 bits
        checksum &= (TTC_SPY_COMMAND::STAT_MASK_DATA >> AltiCommon::getFirstSetBitPos(TTC_SPY_COMMAND::STAT_MASK_DATA)); // keep only 8 bits
        /*std::printf("c1: %c\n", c1);
        std::printf("addr: 0x%04x\n", addr);
        std::printf("external: %c\n", external ? '1' : '0');
        std::printf("c2: %c\n", c2);
        std::printf("subaddr: 0x%02x\n", subaddr);
        std::printf("c3: %c\n", c3);
        std::printf("data: 0x%02x\n", data);
        std::printf("checksum: 0x%02x\n", checksum);*/
        if (c1 != 'A') {
            CERR("ALTI: wrong ttcscope Long command line, expected \"A\" field, got \"%c\" instead", c1);
            return (-1);
        }
        if (c2 != 'S') {
            CERR("ALTI: wrong ttcscope Long command line, expected \"S\" field, got \"%c\" instead", c2);
            return (-1);
        }
        if (c3 != 'D') {
            CERR("ALTI: wrong ttcscope Long command line, expected \"D\" field, got \"%c\" instead", c3);
            return (-1);
        }
        command_bs.Address(addr);
        command_bs.SubAddress(subaddr);
        command_bs.Data(data);
    }
    else {
        CERR("ALTI: wrong ttcscope line type \"%s\"", arg[1].c_str());
        return (-1);
    }
    /*std::printf("0x%04x\n", command_bs.Address());
    std::printf("0x%04x\n", command_bs.SubAddress());
    std::printf("0x%04x\n", command_bs.Data());*/

    command = command_bs.number()[0];
    timestamp = timestamp_bs.number()[0];

    return (0);
}

//------------------------------------------------------------------------------

int AltiModule::MEMCommandTimestampToBitstrings(const unsigned int command, const unsigned int timestamp, TTC_SPY_COMMAND &command_bs, TTC_SPY_TIMESTAMP &timestamp_bs) { // TTC decoder command/timestamp memory entries -> corresponding bitstrings

    std::vector<unsigned int> tmp;

    tmp.clear();
    tmp.push_back(static_cast<unsigned int>(command));
    command_bs = (RCD::BitSet) tmp;
    tmp.clear();
    tmp.push_back(static_cast<unsigned int>(timestamp));
    timestamp_bs = (RCD::BitSet) tmp;

    return (0);
}

//------------------------------------------------------------------------------

int AltiModule::ENCInhibitParamsRead(const unsigned int bgo, unsigned int &width, unsigned int &delay) {
    if (bgo < 0 || bgo > 3) {
        CERR("ALTI: wrong BGO index \"%d\"", bgo);
        m_status = -1;
        return (m_status);
    }

    ALTI_BGO_INHIBITGENERATORPARAMS_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BGO_InhibitGeneratorParams_Read(bgo, bs)) != 0) {
        CERR("ALTI: reading BGO%d inhibit parameter register", bgo);
        return (m_status);
    }
    width = bs.WidthInBCs();
    delay = bs.DelayInBCs();

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCInhibitParamsWrite(const unsigned int bgo, const unsigned int width, const unsigned int delay) { // all
    if (bgo < 0 || bgo > 3) {
        CERR("ALTI: wrong BGO index \"%d\"", bgo);
        m_status = -1;
        return (m_status);
    }

    ALTI_BGO_INHIBITGENERATORPARAMS_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BGO_InhibitGeneratorParams_Read(bgo, bs)) != 0) {
        CERR("ALTI: reading BGO%d inhibit parameter register", bgo);
        return (m_status);
    }
    // MODIFY
    bs.WidthInBCs(width);
    bs.DelayInBCs(delay);
    // WRITE
    if ((m_status = m_alti->BGO_InhibitGeneratorParams_Write(bgo, bs)) != 0) {
        CERR("ALTI: writing width %d and delay %d to BGO%d inhibit parameter register", width, delay, bgo);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCInhibitParamsRead(std::vector<unsigned int> &width, std::vector<unsigned int> &delay) { // all
    width.clear();
    delay.clear();

    ALTI_BGO_INHIBITGENERATORPARAMS_BITSTRING bs;

    unsigned int bgo;
    for (bgo = 0; bgo <= 3; bgo++) {
        // READ
        if ((m_status = m_alti->BGO_InhibitGeneratorParams_Read(bgo, bs)) != 0) {
            CERR("ALTI: reading BGO%d inhibit parameter register", bgo);
            return (m_status);
        }
        width.push_back(bs.WidthInBCs());
        delay.push_back(bs.DelayInBCs());
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCInhibitParamsWrite(const std::vector<unsigned int> &width, const std::vector<unsigned int> &delay) {
    ALTI_BGO_INHIBITGENERATORPARAMS_BITSTRING bs;

    unsigned int bgo;
    for (bgo = 0; bgo <= 3; bgo++) {
        // READ
        if ((m_status = m_alti->BGO_InhibitGeneratorParams_Read(bgo, bs)) != 0) {
            CERR("ALTI: reading BGO%d inhibit parameter register", bgo);
            return (m_status);
        }
        // MODIFY
        bs.WidthInBCs(width[bgo]);
        bs.DelayInBCs(delay[bgo]);
        // WRITE
        if ((m_status = m_alti->BGO_InhibitGeneratorParams_Write(bgo, bs)) != 0) {
            CERR("ALTI: writing width %d and delay %d to BGO%d inhibit parameter register", width, delay, bgo);
            return (m_status);
        }
    }
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCTriggerWordDelayRead(unsigned int &delay) {
    ALTI_TTC_TRIGGERTYPEWORDPARAMS_BITSTRING bs;

    // READ
    if ((m_status = m_alti->TTC_TriggerTypeWordParams_Read(bs)) != 0) {
        CERR("ALTI: reading TTC encoder TTYP word parameter register", "");
        return (m_status);
    }
    //delay = bs.DelayInBCs();
    delay = bs.WidthInBCs();

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCTriggerWordDelayWrite(const unsigned int delay) {
    ALTI_TTC_TRIGGERTYPEWORDPARAMS_BITSTRING bs;

    // READ
    if ((m_status = m_alti->TTC_TriggerTypeWordParams_Read(bs)) != 0) {
        CERR("ALTI: reading TTC encoder TTYP word parameter register", "");
        return (m_status);
    }
    // MODIFY
    //bs.DelayInBCs(delay);
    bs.WidthInBCs(delay);
    // WRITE
    if ((m_status = m_alti->TTC_TriggerTypeWordParams_Write(bs)) != 0) {
        CERR("ALTI: writing delay %d to TTC encoder TTYP word parameter register", delay);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCTriggerWordCycleControlRead(unsigned int &addr, TTC_FRAME_TYPE &frameType, TTC_ADDRESS_SPACE &addressSpace, unsigned int &subAddr) {
    ALTI_TTC_TRIGGERTYPEWORDCONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->TTC_TriggerTypeWordControl_Read(bs)) != 0) {
        CERR("ALTI: reading TTC encoder TTYP word control register", "");
        return (m_status);
    }
    addr = bs.Address();
    std::string s;
    s = bs.FrameType();
    if (s == "Long") {
        frameType = TTC_FRAME_TYPE::LONG;
    }
    else if (s == "Short") {
        frameType = TTC_FRAME_TYPE::SHORT;
    }
    else {
        CERR("ALTI: unknown frame type \"%s\"", s.c_str());
        m_status = -1;
        return (m_status);
    }
    s = bs.AddressSpace();
    if (s == "External") {
        addressSpace = TTC_ADDRESS_SPACE::EXTERNAL;
    }
    else if (s == "Internal") {
        addressSpace = TTC_ADDRESS_SPACE::INTERNAL;
    }
    else {
        CERR("ALTI: unknown address space selection \"%s\"", s.c_str());
        m_status = -1;
        return (m_status);
    }
    subAddr = bs.SubAddress() << 2; // 6 MSb

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCTriggerWordCycleControlWrite(const unsigned int addr, const TTC_FRAME_TYPE frameType, const TTC_ADDRESS_SPACE addressSpace, const unsigned int subAddr) {
    ALTI_TTC_TRIGGERTYPEWORDCONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->TTC_TriggerTypeWordControl_Read(bs)) != 0) {
        CERR("ALTI: reading TTC encoder TTYP word control register", "");
        return (m_status);
    }
    // MODIFY
    bs.Address(addr);
    bs.FrameType((frameType == LONG) ? "Long" : ((frameType == SHORT) ? "Short" : ""));
    bs.AddressSpace((addressSpace == EXTERNAL) ? "External" : ((addressSpace == INTERNAL) ? "Internal" : ""));
    bs.SubAddress(subAddr >> 2); // 6 MSb
    // WRITE
    if ((m_status = m_alti->TTC_TriggerTypeWordControl_Write(bs)) != 0) {
        CERR("ALTI: writing TTC encoder TTYP word control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCTriggerWordCycleEnable(const unsigned int addr, const TTC_ADDRESS_SPACE addressSpace, const unsigned int subAddr) {
    TTC_FIFOAccessModeWrite(TTYP_FIFO, VME_ACCESS);
    ENCTriggerWordCycleControlWrite(addr, TTC_FRAME_TYPE::LONG, addressSpace, subAddr);
    return TTC_FIFOAccessModeWrite(TTYP_FIFO, RUN_ACCESS);
}

//------------------------------------------------------------------------------

int AltiModule::ENCTriggerWordCycleDisable() {
    return TTC_FIFOAccessModeWrite(TTYP_FIFO, VME_ACCESS);
}

//------------------------------------------------------------------------------

int AltiModule::ENCBgoGenerate(const unsigned int bgo) {
    if (bgo < 0 || bgo > 3) {
        CERR("ALTI: wrong BGO index \"%d\"", bgo);
        m_status = -1;
        return (m_status);
    }

    // WRITE
    if ((m_status = m_alti->TTC_BGOShotVME_Write(bgo, 0x00000000)) != 0) { // data-less
        CERR("ALTI: writing BGO%d VME shot command register", bgo);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

/*int AltiModule::TTCShotTTYPCommandVME() {
    // WRITE
    if ((m_status = m_alti->TTC_TTYPShotVME_Write(0x00000000)) != 0) { // data-less
        CERR("ALTI: writing TTYP VME shot command register", "");
        return (m_status);
    }

    return (m_status);
}*/

//------------------------------------------------------------------------------

int AltiModule::ENCBgoCommandModeRead(const unsigned int bgo, TTC_BGO_MODE_TRIGGER &mode_trigger, TTC_BGO_MODE_COMMAND &mode_command, TTC_BGO_MODE_REPEAT &mode_repeat, TTC_BGO_MODE_FIFO &mode_fifo) {
    if (bgo < 0 || bgo > 3) {
        CERR("ALTI: wrong BGO index \"%d\"", bgo);
        m_status = -1;
        return (m_status);
    }

    ALTI_TTC_BGOMODESELECTION_BITSTRING bs;

    // READ
    if ((m_status = m_alti->TTC_BGOModeSelection_Read(bgo, bs)) != 0) {
        CERR("ALTI: reading BGO%d mode selection register", bgo);
        return (m_status);
    }
    std::string s;
    s = bs.FrontPanelInputEnable();
    if (s == "Enabled") {
        mode_trigger = TTC_BGO_MODE_TRIGGER::BGO_SIGNAL;
    }
    else if (s == "Disabled") {
        mode_trigger = TTC_BGO_MODE_TRIGGER::VME;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for BGO%d trigger mode", s.c_str(), bgo);
        return (-1);
    }
    s = bs.CommandType();
    if (s == "Synchronous") {
        mode_command = TTC_BGO_MODE_COMMAND::SYNCHRONOUS;
    }
    else if (s == "Asynchronous") {
        mode_command = TTC_BGO_MODE_COMMAND::ASYNCHRONOUS;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for BGO%d command mode", s.c_str(), bgo);
        return (-1);
    }
    s = bs.Repeat();
    if (s == "Single") {
        mode_repeat = TTC_BGO_MODE_REPEAT::SINGLE;
    }
    else if (s == "Repetitive") {
        mode_repeat = TTC_BGO_MODE_REPEAT::REPETITIVE;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for BGO%d repeat mode", s.c_str(), bgo);
        return (-1);
    }
    s = bs.FIFOTransmit();
    if (s == "WhenNotEmpty") {
        mode_fifo = TTC_BGO_MODE_FIFO::WHEN_NOT_EMPTY;
    }
    else if (s == "RegardlessOfStatus") {
        mode_fifo = TTC_BGO_MODE_FIFO::ON_TRIGGER;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for BGO%d FIFO mode", s.c_str(), bgo);
        return (-1);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCBgoCommandModeWrite(const unsigned int bgo, const TTC_BGO_MODE_TRIGGER mode_trigger, const TTC_BGO_MODE_COMMAND mode_command, const TTC_BGO_MODE_REPEAT mode_repeat, const TTC_BGO_MODE_FIFO mode_fifo) {
    if (bgo < 0 || bgo > 3) {
        CERR("ALTI: wrong BGO index \"%d\"", bgo);
        m_status = -1;
        return (m_status);
    }

    /*// reset BGO pipeline
    if ((m_status = ENCFifoReset((TTC_FIFO) bgo, true, true)) != SUCCESS) {
        CERR("ALTI: enabling read/write pointer resets for all TTC FIFOs", "");
        return (m_status);
    }
    if ((m_status = ENCFifoReset((TTC_FIFO) bgo, false, false)) != SUCCESS) {
        CERR("ALTI: disabling read/write pointer resets for all TTC FIFOs", "");
        return (m_status);
    }
    COUT("Resetting BGO%d pipeline", bgo);*/

    ALTI_TTC_BGOMODESELECTION_BITSTRING bs;

    // READ
    if ((m_status = m_alti->TTC_BGOModeSelection_Read(bgo, bs)) != 0) {
        CERR("ALTI: reading BGO%d mode selection register", bgo);
        return (m_status);
    }
    // MODIFY
    bs.FrontPanelInputEnable((mode_trigger == TTC_BGO_MODE_TRIGGER::BGO_SIGNAL) ? "Enabled" : ((mode_trigger == TTC_BGO_MODE_TRIGGER::VME) ? "Disabled" : ""));
    bs.CommandType((mode_command == TTC_BGO_MODE_COMMAND::SYNCHRONOUS) ? "Synchronous" : ((mode_command == TTC_BGO_MODE_COMMAND::ASYNCHRONOUS) ? "Asynchronous" : ""));
    bs.Repeat((mode_repeat == TTC_BGO_MODE_REPEAT::SINGLE) ? "Single" : ((mode_repeat == TTC_BGO_MODE_REPEAT::REPETITIVE) ? "Repetitive" : ""));
    bs.FIFOTransmit((mode_fifo == TTC_BGO_MODE_FIFO::WHEN_NOT_EMPTY) ? "WhenNotEmpty" : ((mode_fifo == TTC_BGO_MODE_FIFO::ON_TRIGGER) ? "RegardlessOfStatus" : ""));
    // WRITE
    if ((m_status = m_alti->TTC_BGOModeSelection_Write(bgo, bs)) != 0) {
        CERR("ALTI: writing BGO%d mode selection register", bgo);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCBgoCommandModeRead(std::vector<TTC_BGO_MODE_TRIGGER> &mode_trigger, std::vector<TTC_BGO_MODE_COMMAND> &mode_command, std::vector<TTC_BGO_MODE_REPEAT> &mode_repeat, std::vector<TTC_BGO_MODE_FIFO> &mode_fifo) {
    mode_trigger.clear();
    mode_command.clear();
    mode_repeat.clear();
    mode_fifo.clear();

    ALTI_TTC_BGOMODESELECTION_BITSTRING bs;

    unsigned int bgo;
    std::string s;
    for (bgo = 0; bgo <= 3; bgo++) {
        // READ
        if ((m_status = m_alti->TTC_BGOModeSelection_Read(bgo, bs)) != 0) {
            CERR("ALTI: reading BGO%d mode selection register", bgo);
            return (m_status);
        }
        s = bs.FrontPanelInputEnable();
        if (s == "Enabled") {
            mode_trigger.push_back(TTC_BGO_MODE_TRIGGER::BGO_SIGNAL);
        }
        else if (s == "Disabled") {
            mode_trigger.push_back(TTC_BGO_MODE_TRIGGER::VME);
        }
        else {
            CERR("ALTI: unknown value \"%s\" for BGO%d trigger mode", s.c_str(), bgo);
            return (-1);
        }
        s = bs.CommandType();
        if (s == "Synchronous") {
            mode_command.push_back(TTC_BGO_MODE_COMMAND::SYNCHRONOUS);
        }
        else if (s == "Asynchronous") {
            mode_command.push_back(TTC_BGO_MODE_COMMAND::ASYNCHRONOUS);
        }
        else {
            CERR("ALTI: unknown value \"%s\" for BGO%d command mode", s.c_str(), bgo);
            return (-1);
        }
        s = bs.Repeat();
        if (s == "Single") {
            mode_repeat.push_back(TTC_BGO_MODE_REPEAT::SINGLE);
        }
        else if (s == "Repetitive") {
            mode_repeat.push_back(TTC_BGO_MODE_REPEAT::REPETITIVE);
        }
        else {
            CERR("ALTI: unknown value \"%s\" for BGO%d repeat mode", s.c_str(), bgo);
            return (-1);
        }
        s = bs.FIFOTransmit();
        if (s == "WhenNotEmpty") {
            mode_fifo.push_back(TTC_BGO_MODE_FIFO::WHEN_NOT_EMPTY);
        }
        else if (s == "RegardlessOfStatus") {
            mode_fifo.push_back(TTC_BGO_MODE_FIFO::ON_TRIGGER);
        }
        else {
            CERR("ALTI: unknown value \"%s\" for BGO%d FIFO mode", s.c_str(), bgo);
            return (-1);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCBgoCommandModeWrite(const std::vector<TTC_BGO_MODE_TRIGGER> &mode_trigger, const std::vector<TTC_BGO_MODE_COMMAND> &mode_command, const std::vector<TTC_BGO_MODE_REPEAT> &mode_repeat, const std::vector<TTC_BGO_MODE_FIFO> &mode_fifo) {
    ALTI_TTC_BGOMODESELECTION_BITSTRING bs;

    unsigned int bgo;
    for (bgo = 0; bgo <= 3; bgo++) {
        // READ
        if ((m_status = m_alti->TTC_BGOModeSelection_Read(bgo, bs)) != 0) {
            CERR("ALTI: reading BGO%d mode selection register", bgo);
            return (m_status);
        }
        // MODIFY
        bs.FrontPanelInputEnable((mode_trigger[bgo] == TTC_BGO_MODE_TRIGGER::BGO_SIGNAL) ? "Enabled" : ((mode_trigger[bgo] == TTC_BGO_MODE_TRIGGER::VME) ? "Disabled" : ""));
        bs.CommandType((mode_command[bgo] == TTC_BGO_MODE_COMMAND::SYNCHRONOUS) ? "Synchronous" : ((mode_command[bgo] == TTC_BGO_MODE_COMMAND::ASYNCHRONOUS) ? "Asynchronous" : ""));
        bs.Repeat((mode_repeat[bgo] == TTC_BGO_MODE_REPEAT::SINGLE) ? "Single" : ((mode_repeat[bgo] == TTC_BGO_MODE_REPEAT::REPETITIVE) ? "Repetitive" : ""));
        bs.FIFOTransmit((mode_fifo[bgo] == TTC_BGO_MODE_FIFO::WHEN_NOT_EMPTY) ? "WhenNotEmpty" : ((mode_fifo[bgo] == TTC_BGO_MODE_FIFO::ON_TRIGGER) ? "RegardlessOfStatus" : ""));
        // WRITE
        if ((m_status = m_alti->TTC_BGOModeSelection_Write(bgo, bs)) != 0) {
            CERR("ALTI: writing BGO%d mode selection register", bgo);
            return (m_status);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCBgoCommandModeRead(const unsigned int bgo, TTC_BGO_MODE &mode) {
    TTC_BGO_MODE_TRIGGER mode_trigger;
    TTC_BGO_MODE_COMMAND mode_command;
    TTC_BGO_MODE_REPEAT mode_repeat;
    TTC_BGO_MODE_FIFO mode_fifo;
    if ((m_status = ENCBgoCommandModeRead(bgo, mode_trigger, mode_command, mode_repeat, mode_fifo)) != SUCCESS) {
        CERR("ALTI: reading BGO%d mode", bgo);
        return (m_status);
    }
    mode = ENCLowLevelToHighLevelBgoMode(mode_trigger, mode_command, mode_repeat, mode_fifo);

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCBgoCommandModeWrite(const unsigned int bgo, const TTC_BGO_MODE mode) {
    TTC_BGO_MODE current_mode;
    if ((m_status = ENCBgoCommandModeRead(bgo, current_mode)) != SUCCESS) {
        CERR("ALTI: reading TTC BGO%d FIFO mode", bgo);
        return (m_status);
    }
    if ((current_mode == ASYNCHRONOUS_BGO_SIGNAL) && (mode == ASYNCHRONOUS_VME_ON_TRIGGER)) { // specific transition
        // reset the pipeline
        // some printfs as notifications?

        if ((m_status = TTC_FIFOAccessModeWrite((TTC_FIFO) bgo, TTC_FIFO_ACCESS_MODE::VME_ACCESS)) != SUCCESS) {
            CERR("ALTI: putting TTC BGO%d FIFO in VME access mode", bgo);
            return (m_status);
        }
        COUT("ALTI: INFO - TTC BGO%d FIFO put in VME access mode", bgo);
    }
    if ((m_status = ENCBgoCommandModeWrite(bgo, TTC_BGO_MODE_HL[mode].mode_trigger, TTC_BGO_MODE_HL[mode].mode_command, TTC_BGO_MODE_HL[mode].mode_repeat, TTC_BGO_MODE_HL[mode].mode_fifo)) != 0) {
        CERR("ALTI: writing \"%s\" mode for the BGO%d TTC FIFO", TTC_BGO_MODE_NAME[mode].c_str(), bgo);
        return (m_status);
    }
    if ((current_mode == ASYNCHRONOUS_BGO_SIGNAL) && (mode == ASYNCHRONOUS_VME_ON_TRIGGER)) { // specific transition
        // reset the pipeline
        // some printfs as notifications?

        if ((m_status = ENCBgoFifoRetransmitWrite((TTC_FIFO) bgo, false)) != SUCCESS) {
            CERR("ALTI: putting TTC BGO%d FIFO out of retransmit mode", bgo);
            return (m_status);
        }
        COUT("ALTI: INFO - TTC BGO%d FIFO put out of retransmit mode", bgo);

        if ((m_status = ENCFifoReset((TTC_FIFO) bgo)) != SUCCESS) {
            CERR("ALTI: resetting read/write pointers for the BGO%d TTC FIFO", bgo);
            return (m_status);
        }
        COUT("ALTI: INFO - TTC BGO%d FIFO reset", bgo);

        if ((m_status = TTC_FIFOAccessModeWrite((TTC_FIFO) bgo, TTC_FIFO_ACCESS_MODE::RUN_ACCESS)) != SUCCESS) {
            CERR("ALTI: putting TTC BGO%d FIFO in RUN access mode", bgo);
            return (m_status);
        }
        COUT("ALTI: INFO - TTC BGO%d FIFO put in RUN access mode", bgo);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCBgoCommandModeRead(std::vector<TTC_BGO_MODE> &mode) {
    unsigned int bgo;
    for (bgo = 0; bgo <= 3; bgo++) {
        m_status = ENCBgoCommandModeRead(bgo, mode[bgo]);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCBgoCommandModeWrite(const std::vector<TTC_BGO_MODE> &mode) {
    unsigned int bgo;
    for (bgo = 0; bgo <= 3; bgo++) {
        m_status = ENCBgoCommandModeWrite(bgo, mode[bgo]);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

AltiModule::TTC_BGO_MODE AltiModule::ENCLowLevelToHighLevelBgoMode(const TTC_BGO_MODE_TRIGGER mode_trigger, const TTC_BGO_MODE_COMMAND mode_command, const TTC_BGO_MODE_REPEAT mode_repeat, const TTC_BGO_MODE_FIFO mode_fifo) {
    if ((mode_command == SYNCHRONOUS) && (mode_repeat == SINGLE) && (mode_trigger == BGO_SIGNAL) && (mode_fifo == ON_TRIGGER)) {
        return SYNCHRONOUS_SINGLE_BGO_SIGNAL;
    }
    else if ((mode_command == SYNCHRONOUS) && (mode_repeat == REPETITIVE)) {
        return SYNCHRONOUS_REPETITIVE;
    }
    else if ((mode_command == ASYNCHRONOUS) && (mode_trigger == BGO_SIGNAL)) {
        return ASYNCHRONOUS_BGO_SIGNAL;
    }
    else if ((mode_command == ASYNCHRONOUS) && (mode_repeat == SINGLE) && (mode_trigger == VME) && (mode_fifo == ON_TRIGGER)) {
        return ASYNCHRONOUS_VME_ON_TRIGGER;
    }
    else if ((mode_command == ASYNCHRONOUS) && (mode_trigger == VME) && (mode_fifo == WHEN_NOT_EMPTY)) {
        return ASYNCHRONOUS_VME_WHEN_NOT_EMPTY;
    }
    else {
        return INVALID_MODE;
    }
}

//------------------------------------------------------------------------------

int AltiModule::SIGMuxBGOOutputLEMORead(bool &bgo2enabled, bool &bgo3enabled) { // both
    ALTI_BGO_CONFIG_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BGO_Config_Read(bs)) != 0) {
        CERR("ALTI: reading BGO config register", "");
        return (m_status);
    }
    std::string s;
    s = bs.LemoOutputEnable_0_2();
    if (s == "BGo2") {
        bgo2enabled = true;
    }
    else if (s == "BGo0") {
        bgo2enabled = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for BGO2/0 front panel output enable", s.c_str());
        return (-1);
    }
    s = bs.LemoOutputEnable_1_3();
    if (s == "BGo3") {
        bgo3enabled = true;
    }
    else if (s == "BGo1") {
        bgo3enabled = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for BGO3/1 front panel output enable", s.c_str());
        return (-1);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGMuxBGOOutputLEMOWrite(const bool bgo2enable, const bool bgo3enable) { // both
    ALTI_BGO_CONFIG_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BGO_Config_Read(bs)) != 0) {
        CERR("ALTI: reading BGO config register", "");
        return (m_status);
    }
    // MODIFY
    bs.LemoOutputEnable_0_2(bgo2enable ? "BGo2" : "BGo0");
    bs.LemoOutputEnable_1_3(bgo3enable ? "BGo3" : "BGo1");
    // WRITE
    if ((m_status = m_alti->BGO_Config_Write(bs)) != 0) {
        CERR("ALTI: writing BGO config register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGMuxTTRInputLEMORead(const u_int bit, TTR_LEMO & out) { 
    ALTI_CRQ_MUXCONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CRQ_MuxControl_Read(bs)) != 0) {
        CERR("ALTI: reading CRQ Mux control register", "");
        return (m_status);
    }
    std::string s;
    s = bs.CALREQ_TTR_ToFpga(bit);
    if (s == "TTR") {
        out = TTR;
    }
    else if(s == "CALREQ") {
        out = CALREQ;
    }
     else {
        CERR("ALTI: unknown value \"%s\" for TTR Lemo input selection", s.c_str());
        return (-1);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGMuxTTRInputLEMOWrite(const u_int bit, const TTR_LEMO in) { 
    ALTI_CRQ_MUXCONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CRQ_MuxControl_Read(bs)) != 0) {
      CERR("ALTI: reading CRQ Mux control register", "");
      return (m_status);
    }
    // modify
    switch (in) {
    case TTR:
      bs.CALREQ_TTR_ToFpga(bit, "TTR");
      break;
    case CALREQ:
      bs.CALREQ_TTR_ToFpga(bit, "CALREQ");
      break;
    default:
      break;
    }
    
    // write
    if ((m_status = m_alti->CRQ_MuxControl_Write(bs)) != 0) {
      CERR("ALTI: writing CREQ Mux control register", "");
      return (m_status);
    }
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGMuxTTROutputLEMORead(const u_int bit, TTR_LEMO & out) { 
    ALTI_CRQ_MUXCONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CRQ_MuxControl_Read(bs)) != 0) {
        CERR("ALTI: reading CRQ Mux control register", "");
        return (m_status);
    }
    std::string s;
    s = bs.CALREQ_TTR_LemoOut(bit);
    if (s == "TTR") {
        out = TTR;
    }
    else if(s == "CALREQ") {
        out = CALREQ;
    }
     else {
        CERR("ALTI: unknown value \"%s\" for TTR Lemo input selection", s.c_str());
        return (-1);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::SIGMuxTTROutputLEMOWrite(const u_int bit, const TTR_LEMO in) { 
    ALTI_CRQ_MUXCONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CRQ_MuxControl_Read(bs)) != 0) {
      CERR("ALTI: reading CRQ Mux control register", "");
      return (m_status);
    }
    // modify
    switch (in) {
    case TTR:
      bs.CALREQ_TTR_LemoOut(bit, "TTR");
      break;
    case CALREQ:
      bs.CALREQ_TTR_LemoOut(bit, "CALREQ");
      break;
    default:
      break;
    }
    
    // write
    if ((m_status = m_alti->CRQ_MuxControl_Write(bs)) != 0) {
      CERR("ALTI: writing CREQ Mux control register", "");
      return (m_status);
    }
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::TTC_FIFORead(const TTC_FIFO ttcFifo, unsigned int &rd) {
    if (ttcFifo == TTYP_FIFO) return m_alti->BGO_TTYP_FIFO_Read(rd);
    else return m_alti->BGO_FIFO_Read(ttcFifo, rd);
}

//------------------------------------------------------------------------------

int AltiModule::TTC_FIFOWrite(const TTC_FIFO ttcFifo, const unsigned int wr) {
    if (ttcFifo == TTYP_FIFO) return m_alti->BGO_TTYP_FIFO_Write(wr);
    else return m_alti->BGO_FIFO_Write(ttcFifo, wr);
}

//------------------------------------------------------------------------------

int AltiModule::TTC_FIFORead(const TTC_FIFO ttcFifo, std::vector<unsigned int> &rd) {
    if (ttcFifo == TTYP_FIFO) return m_alti->BGO_TTYP_FIFO_Read(rd);
    else return m_alti->BGO_FIFO_Read(ttcFifo, rd);
}

//------------------------------------------------------------------------------

int AltiModule::TTC_FIFOWrite(const TTC_FIFO ttcFifo, const std::vector<unsigned int> &wr) {
    if (ttcFifo == TTYP_FIFO) return m_alti->BGO_TTYP_FIFO_Write(wr);
    else return m_alti->BGO_FIFO_Write(ttcFifo, wr);
}
//------------------------------------------------------------------------------

int AltiModule::TTC_FIFORead(const TTC_FIFO ttcFifo, RCD::CMEMSegment *seg) {
    if (ttcFifo == TTYP_FIFO) return m_alti->BGO_TTYP_FIFO_Read(seg);
    else return m_alti->BGO_FIFO_Read(ttcFifo, seg);
}

//------------------------------------------------------------------------------

int AltiModule::TTC_FIFOWrite(const TTC_FIFO ttcFifo, RCD::CMEMSegment *seg) {
    if (ttcFifo == TTYP_FIFO) return m_alti->BGO_TTYP_FIFO_Write(seg);
    else return m_alti->BGO_FIFO_Write(ttcFifo, seg);
}

//------------------------------------------------------------------------------

int AltiModule::ENCBgoCommandPut(const unsigned int bgo, const TTC_FRAME_TYPE frameType, const uint16_t addr, const TTC_ADDRESS_SPACE addrSpace, const uint8_t subAddr, const uint8_t data) {
    if (bgo < 0 || bgo > 3) {
        CERR("ALTI: wrong BGO index \"%d\"", bgo);
        m_status = -1;
        return (m_status);
    }

    TTC_SPY_COMMAND command;
    std::vector<unsigned int> tmp;
    tmp.clear();
    tmp.push_back(static_cast<unsigned int>(0x00000000));
    command = (RCD::BitSet) tmp;

    command.FrameType((frameType == TTC_FRAME_TYPE::SHORT) ? "Short" : ((frameType == TTC_FRAME_TYPE::LONG) ? "Long" : ""));
    command.Address(addr);
    command.AddressSpace((addrSpace == TTC_ADDRESS_SPACE::INTERNAL) ? "Internal" : ((addrSpace == TTC_ADDRESS_SPACE::EXTERNAL) ? "External" : ""));
    command.SubAddress(subAddr);
    command.Data(data);

    if ((m_status = TTC_FIFOWrite((TTC_FIFO) bgo, command.number()[0])) != 0) {
        CERR("ALTI: putting command 0x%08x (\"%s\", \"%s\", addr = 0x%04x, subAddr = 0x%02x, data = 0x%02x) to BGO%d FIFO", command.number()[0], TTC_FRAME_TYPE_NAME[frameType], TTC_ADDRESS_SPACE_NAME[addrSpace], addr, subAddr, data, bgo);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCBgoCommandPutShort(const unsigned int bgo, const uint8_t data) {
    return ENCBgoCommandPut(bgo, TTC_FRAME_TYPE::SHORT, 0x0000, TTC_ADDRESS_SPACE::INTERNAL, 0x00, data);
}

//------------------------------------------------------------------------------

int AltiModule::ENCBgoCommandPutLong(const unsigned int bgo, const uint16_t addr, const TTC_ADDRESS_SPACE addrSpace, const uint8_t subAddr, const uint8_t data) {
    return ENCBgoCommandPut(bgo, TTC_FRAME_TYPE::LONG, addr, addrSpace, subAddr, data);
}

//------------------------------------------------------------------------------

int AltiModule::ENCFifoStatusRead(const TTC_FIFO ttcFifo, bool &pipelineFull, bool &pipelineEmpty, bool &fifoFull, bool &fifoEmpty) {
    ALTI_BGO_PIPELINEFIFOSEMPTYFULL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BGO_PipelineFIFOsEmptyFull_Read(bs)) != 0) {
        CERR("ALTI: reading BGO/TTYP FIFOs status register", "");
        return (m_status);
    }
    std::string s;

    if (ttcFifo != TTYP_FIFO) {
        unsigned int bgo = ttcFifo;
        s = bs.PipelineFullBGo(bgo);
        if (s == "Full") {
            pipelineFull = true;
        }
        else if (s == "NotFull") {
            pipelineFull = false;
        }
        else {
            CERR("ALTI: unknown value \"%s\" for BGO%d pipeline full status", s.c_str(), bgo);
            return (-1);
        }
        s = bs.PipelineEmptyBGo(bgo);
        if (s == "Empty") {
            pipelineEmpty = true;
        }
        else if (s == "NotEmpty") {
            pipelineEmpty = false;
        }
        else {
            CERR("ALTI: unknown value \"%s\" for BGO%d pipeline empty status", s.c_str(), bgo);
            return (-1);
        }
        s = bs.FIFOFullBGo(bgo);
        if (s == "Full") {
            fifoFull = true;
        }
        else if (s == "NotFull") {
            fifoFull = false;
        }
        else {
            CERR("ALTI: unknown value \"%s\" for BGO%d FIFO full status", s.c_str(), bgo);
            return (-1);
        }
        s = bs.FIFOEmptyBGo(bgo);
        if (s == "Empty") {
            fifoEmpty = true;
        }
        else if (s == "NotEmpty") {
            fifoEmpty = false;
        }
        else {
            CERR("ALTI: unknown value \"%s\" for BGO%d FIFO empty status", s.c_str(), bgo);
            return (-1);
        }
    }
    else {
        s = bs.PipelineFullTriggerType();
        if (s == "Full") {
            pipelineFull = true;
        }
        else if (s == "NotFull") {
            pipelineFull = false;
        }
        else {
            CERR("ALTI: unknown value \"%s\" for TTYP pipeline full status", s);
            return (-1);
        }
        s = bs.PipelineEmptyTriggerType();
        if (s == "Empty") {
            pipelineEmpty = true;
        }
        else if (s == "NotEmpty") {
            pipelineEmpty = false;
        }
        else {
            CERR("ALTI: unknown value \"%s\" for TTYP pipeline empty status", s);
            return (-1);
        }
        s = bs.FIFOFullTriggerType();
        if (s == "Full") {
            fifoFull = true;
        }
        else if (s == "NotFull") {
            fifoFull = false;
        }
        else {
            CERR("ALTI: unknown value \"%s\" for TTYP FIFO full status", s);
            return (-1);
        }
        s = bs.FIFOEmptyTriggerType();
        if (s == "Empty") {
            fifoEmpty = true;
        }
        else if (s == "NotEmpty") {
            fifoEmpty = false;
        }
        else {
            CERR("ALTI: unknown value \"%s\" for TTYP FIFO empty status", s);
            return (-1);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCFifoStatusRead(std::vector<bool> &pipelineFull, std::vector<bool> &pipelineEmpty, std::vector<bool> &fifoFull, std::vector<bool> &fifoEmpty) {
    pipelineFull.clear();
    pipelineEmpty.clear();
    fifoFull.clear();
    fifoEmpty.clear();

    ALTI_BGO_PIPELINEFIFOSEMPTYFULL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BGO_PipelineFIFOsEmptyFull_Read(bs)) != 0) {
        CERR("ALTI: reading BGO/TTYP FIFOs status register", "");
        return (m_status);
    }

    std::string s;
    unsigned int bgo;
    for (bgo = 0; bgo <= 3; bgo++) {
        s = bs.PipelineFullBGo(bgo);
        if (s == "Full") {
            pipelineFull.push_back(true);
        }
        else if (s == "NotFull") {
            pipelineFull.push_back(false);
        }
        else {
            CERR("ALTI: unknown value \"%s\" for BGO%d pipeline full status", s.c_str(), bgo);
            return (-1);
        }
        s = bs.PipelineEmptyBGo(bgo);
        if (s == "Empty") {
            pipelineEmpty.push_back(true);
        }
        else if (s == "NotEmpty") {
            pipelineEmpty.push_back(false);
        }
        else {
            CERR("ALTI: unknown value \"%s\" for BGO%d pipeline empty status", s.c_str(), bgo);
            return (-1);
        }
        s = bs.FIFOFullBGo(bgo);
        if (s == "Full") {
            fifoFull.push_back(true);
        }
        else if (s == "NotFull") {
            fifoFull.push_back(false);
        }
        else {
            CERR("ALTI: unknown value \"%s\" for BGO%d FIFO full status", s.c_str(), bgo);
            return (-1);
        }
        s = bs.FIFOEmptyBGo(bgo);
        if (s == "Empty") {
            fifoEmpty.push_back(true);
        }
        else if (s == "NotEmpty") {
            fifoEmpty.push_back(false);
        }
        else {
            CERR("ALTI: unknown value \"%s\" for BGO%d FIFO empty status", s.c_str(), bgo);
            return (-1);
        }
    }
    s = bs.PipelineFullTriggerType();
    if (s == "Full") {
        pipelineFull.push_back(true);
    }
    else if (s == "NotFull") {
        pipelineFull.push_back(false);
    }
    else {
        CERR("ALTI: unknown value \"%s\" for TTYP pipeline full status", s);
        return (-1);
    }
    s = bs.PipelineEmptyTriggerType();
    if (s == "Empty") {
        pipelineEmpty.push_back(true);
    }
    else if (s == "NotEmpty") {
        pipelineEmpty.push_back(false);
    }
    else {
        CERR("ALTI: unknown value \"%s\" for TTYP pipeline empty status", s);
        return (-1);
    }
    s = bs.FIFOFullTriggerType();
    if (s == "Full") {
        fifoFull.push_back(true);
    }
    else if (s == "NotFull") {
        fifoFull.push_back(false);
    }
    else {
        CERR("ALTI: unknown value \"%s\" for TTYP FIFO full status", s);
        return (-1);
    }
    s = bs.FIFOEmptyTriggerType();
    if (s == "Empty") {
        fifoEmpty.push_back(true);
    }
    else if (s == "NotEmpty") {
        fifoEmpty.push_back(false);
    }
    else {
        CERR("ALTI: unknown value \"%s\" for TTYP FIFO empty status", s);
        return (-1);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCFifoResetReadAddress(const TTC_FIFO ttcFifo) {

    ALTI_BGO_PIPELINEFIFOSRESETS_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Read(bs)) != 0) {
        CERR("ALTI: reading BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }
    // MODIFY
    if (ttcFifo != TTYP_FIFO) {
        unsigned int bgo = ttcFifo;
        bs.ReadAddressResetBGO(bgo, "Enabled");
    }
    else {
        bs.ReadAddressResetTTYP("Enabled");
    }
    // WRITE
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Write(bs)) != 0) {
        CERR("ALTI: writing BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }
    // MODIFY
    if (ttcFifo != TTYP_FIFO) {
        unsigned int bgo = ttcFifo;
        bs.ReadAddressResetBGO(bgo, "Disabled");
    }
    else {
        bs.ReadAddressResetTTYP("Disabled");
    }
    // WRITE
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Write(bs)) != 0) {
        CERR("ALTI: writing BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCFifoResetWriteAddress(const TTC_FIFO ttcFifo) {

    ALTI_BGO_PIPELINEFIFOSRESETS_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Read(bs)) != 0) {
        CERR("ALTI: reading BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }
    // MODIFY
    if (ttcFifo != TTYP_FIFO) {
        unsigned int bgo = ttcFifo;
        bs.WriteAddressResetBGO(bgo, "Enabled");
    }
    else {
        bs.WriteAddressResetTTYP("Enabled");
    }
    // WRITE
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Write(bs)) != 0) {
        CERR("ALTI: writing BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }
    // MODIFY
    if (ttcFifo != TTYP_FIFO) {
        unsigned int bgo = ttcFifo;
        bs.WriteAddressResetBGO(bgo, "Disabled");
    }
    else {
        bs.WriteAddressResetTTYP("Disabled");
    }
    // WRITE
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Write(bs)) != 0) {
        CERR("ALTI: writing BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCFifoReset(const TTC_FIFO ttcFifo) { // read+write

    ALTI_BGO_PIPELINEFIFOSRESETS_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Read(bs)) != 0) {
        CERR("ALTI: reading BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }
    // MODIFY
    if (ttcFifo != TTYP_FIFO) {
        unsigned int bgo = ttcFifo;
        bs.ReadAddressResetBGO(bgo, "Enabled");
        bs.WriteAddressResetBGO(bgo, "Enabled");
    }
    else {
        bs.ReadAddressResetTTYP("Enabled");
        bs.WriteAddressResetTTYP("Enabled");
    }
    // WRITE
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Write(bs)) != 0) {
        CERR("ALTI: writing BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }
    // MODIFY
    if (ttcFifo != TTYP_FIFO) {
        unsigned int bgo = ttcFifo;
        bs.ReadAddressResetBGO(bgo, "Disabled");
        bs.WriteAddressResetBGO(bgo, "Disabled");
    }
    else {
        bs.ReadAddressResetTTYP("Disabled");
        bs.WriteAddressResetTTYP("Disabled");
    }
    // WRITE
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Write(bs)) != 0) {
        CERR("ALTI: writing BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCFifoResetReadAddress() { // all
    ALTI_BGO_PIPELINEFIFOSRESETS_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Read(bs)) != 0) {
        CERR("ALTI: reading BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }
    unsigned int bgo;
    // MODIFY
    for (bgo = 0; bgo <= 3; bgo++) {
        bs.ReadAddressResetBGO(bgo, "Enabled");
    }
    bs.ReadAddressResetTTYP("Enabled");
    // WRITE
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Write(bs)) != 0) {
        CERR("ALTI: writing BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }
    // MODIFY
    for (bgo = 0; bgo <= 3; bgo++) {
        bs.ReadAddressResetBGO(bgo, "Disabled");
    }
    bs.ReadAddressResetTTYP("Disabled");
    // WRITE
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Write(bs)) != 0) {
        CERR("ALTI: writing BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCFifoResetWriteAddress() { // all
    ALTI_BGO_PIPELINEFIFOSRESETS_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Read(bs)) != 0) {
        CERR("ALTI: reading BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }
    unsigned int bgo;
    // MODIFY
    for (bgo = 0; bgo <= 3; bgo++) {
        bs.WriteAddressResetBGO(bgo, "Enabled");
    }
    bs.WriteAddressResetTTYP("Enabled");
    // WRITE
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Write(bs)) != 0) {
        CERR("ALTI: writing BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }
    // MODIFY
    for (bgo = 0; bgo <= 3; bgo++) {
        bs.WriteAddressResetBGO(bgo, "Disabled");
    }
    bs.WriteAddressResetTTYP("Disabled");
    // WRITE
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Write(bs)) != 0) {
        CERR("ALTI: writing BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCFifoReset() { // read+write, all
    ALTI_BGO_PIPELINEFIFOSRESETS_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Read(bs)) != 0) {
        CERR("ALTI: reading BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }
    unsigned int bgo;
    // MODIFY
    for (bgo = 0; bgo <= 3; bgo++) {
        bs.ReadAddressResetBGO(bgo, "Enabled");
        bs.WriteAddressResetBGO(bgo, "Enabled");
    }
    bs.ReadAddressResetTTYP("Enabled");
    bs.WriteAddressResetTTYP("Enabled");
    // WRITE
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Write(bs)) != 0) {
        CERR("ALTI: writing BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }
    // MODIFY
    for (bgo = 0; bgo <= 3; bgo++) {
        bs.ReadAddressResetBGO(bgo, "Disabled");
        bs.WriteAddressResetBGO(bgo, "Disabled");
    }
    bs.ReadAddressResetTTYP("Disabled");
    bs.WriteAddressResetTTYP("Disabled");
    // WRITE
    if ((m_status = m_alti->BGO_PipelineFIFOsResets_Write(bs)) != 0) {
        CERR("ALTI: writing BGO/TTYP FIFOs reset register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::TTC_FIFOAccessModeRead(const TTC_FIFO ttcFifo, TTC_FIFO_ACCESS_MODE &mode) {
    /*if (bgo < 0 || bgo > 3) {
        CERR("ALTI: wrong BGO index \"%d\"", bgo);
        m_status = -1;
        return (m_status);
    }*/

    if (ttcFifo != TTYP_FIFO) {
        ALTI_TTC_BGOMODESELECTION_BITSTRING bs;

        unsigned int bgo = ttcFifo;
        // READ
        if ((m_status = m_alti->TTC_BGOModeSelection_Read(bgo, bs)) != 0) {
            CERR("ALTI: reading BGO%d mode selection register", bgo);
            return (m_status);
        }
        std::string s;
        s = bs.FIFOAccess();
        if (s == "VME") {
            mode = TTC_FIFO_ACCESS_MODE::VME_ACCESS;
        }
        else if (s == "Run") {
            mode = TTC_FIFO_ACCESS_MODE::RUN_ACCESS;
        }
        else {
            CERR("ALTI: unknown value \"%s\" for BGO%d FIFO access mode", s.c_str(), bgo);
            return (-1);
        }
    }
    else {
        ALTI_TTC_TTYPMODESELECTION_BITSTRING bs;

        // READ
        if ((m_status = m_alti->TTC_TTYPModeSelection_Read(bs)) != 0) {
            CERR("ALTI: reading TTYP mode selection register", "");
            return (m_status);
        }
        std::string s;
        s = bs.FIFOAccess();
        if (s == "VME") {
            mode = TTC_FIFO_ACCESS_MODE::VME_ACCESS;
        }
        else if (s == "Run") {
            mode = TTC_FIFO_ACCESS_MODE::RUN_ACCESS;
        }
        else {
            CERR("ALTI: unknown value \"%s\" for TTYP FIFO access mode", s.c_str());
            return (-1);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::TTC_FIFOAccessModeWrite(const TTC_FIFO ttcFifo, const TTC_FIFO_ACCESS_MODE mode) {
    /*if (bgo < 0 || bgo > 3) {
        CERR("ALTI: wrong BGO index \"%d\"", bgo);
        m_status = -1;
        return (m_status);
    }*/

    if (ttcFifo != TTYP_FIFO) {
        ALTI_TTC_BGOMODESELECTION_BITSTRING bs;

        unsigned int bgo = ttcFifo;
        // READ
        if ((m_status = m_alti->TTC_BGOModeSelection_Read(bgo, bs)) != 0) {
            CERR("ALTI: reading BGO%d mode selection register", bgo);
            return (m_status);
        }
        // MODIFY
        switch (mode) {
            case TTC_FIFO_ACCESS_MODE::VME_ACCESS:
                bs.FIFOAccess("VME");
                break;
            case TTC_FIFO_ACCESS_MODE::RUN_ACCESS:
                bs.FIFOAccess("Run");
                break;
            default:
                break;
        }
        // WRITE
        if ((m_status = m_alti->TTC_BGOModeSelection_Write(bgo, bs)) != 0) {
            CERR("ALTI: writing BGO%d mode selection register", bgo);
            return (m_status);
        }
    }
    else {
        ALTI_TTC_TTYPMODESELECTION_BITSTRING bs;

        // READ
        if ((m_status = m_alti->TTC_TTYPModeSelection_Read(bs)) != 0) {
            CERR("ALTI: reading TTYP mode selection register", "");
            return (m_status);
        }
        // MODIFY
        switch (mode) {
            case TTC_FIFO_ACCESS_MODE::VME_ACCESS:
                bs.FIFOAccess("VME");
                break;
            case TTC_FIFO_ACCESS_MODE::RUN_ACCESS:
                bs.FIFOAccess("Run");
                break;
            default:
                break;
        }
        // WRITE
        if ((m_status = m_alti->TTC_TTYPModeSelection_Write(bs)) != 0) {
            CERR("ALTI: writing TTYP mode selection register", "");
            return (m_status);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::TTC_FIFOAccessModeRead(std::vector<TTC_FIFO_ACCESS_MODE> &mode) { // all
    mode.resize(TTC_FIFO_NUMBER);

    unsigned int i;
    TTC_FIFO ttcFifo;
    for (i = 0; i < TTC_FIFO_NUMBER; i++) {
        ttcFifo = (TTC_FIFO) i;
        TTC_FIFOAccessModeRead(ttcFifo, mode[i]);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::TTC_FIFOAccessModeWrite(const std::vector<TTC_FIFO_ACCESS_MODE> &mode) { // all

    unsigned int i;
    TTC_FIFO ttcFifo;
    for (i = 0; i < TTC_FIFO_NUMBER; i++) {
        ttcFifo = (TTC_FIFO) i;
        TTC_FIFOAccessModeWrite(ttcFifo, mode[i]);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::TTC_FIFOAccessModeWrite(const TTC_FIFO_ACCESS_MODE mode) { // all

    unsigned int i;
    TTC_FIFO ttcFifo;
    for (i = 0; i < TTC_FIFO_NUMBER; i++) {
        ttcFifo = (TTC_FIFO) i;
        TTC_FIFOAccessModeWrite(ttcFifo, mode);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCBgoFifoRetransmitRead(const TTC_FIFO ttcFifo, bool &retransmit) {
    /*if (bgo < 0 || bgo > 3) {
        CERR("ALTI: wrong BGO index \"%d\"", bgo);
        m_status = -1;
        return (m_status);
    }*/

    if (ttcFifo != TTYP_FIFO) {
        ALTI_TTC_BGOMODESELECTION_BITSTRING bs;

        unsigned int bgo = ttcFifo;
        // READ
        if ((m_status = m_alti->TTC_BGOModeSelection_Read(bgo, bs)) != 0) {
            CERR("ALTI: reading BGO%d mode selection register", bgo);
            return (m_status);
        }
        std::string s;
        s = bs.FIFORetransmit();
        if (s == "Enabled") {
            retransmit = true;
        }
        else if (s == "Disabled") {
            retransmit = false;
        }
        else {
            CERR("ALTI: unknown value \"%s\" for BGO%d FIFO retransmit mode", s.c_str(), bgo);
            return (-1);
        }
    }
    else {
        ALTI_TTC_TTYPMODESELECTION_BITSTRING bs;

        // READ
        if ((m_status = m_alti->TTC_TTYPModeSelection_Read(bs)) != 0) {
            CERR("ALTI: reading TTYP mode selection register", "");
            return (m_status);
        }
        std::string s;
        s = bs.FIFORetransmit();
        if (s == "Enabled") {
            retransmit = true;
        }
        else if (s == "Disabled") {
            retransmit = false;
        }
        else {
            CERR("ALTI: unknown value \"%s\" for TTYP FIFO retransmit mode", s.c_str());
            return (-1);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCBgoFifoRetransmitWrite(const TTC_FIFO ttcFifo, const bool retransmit) {
    /*if (bgo < 0 || bgo > 3) {
        CERR("ALTI: wrong BGO index \"%d\"", bgo);
        m_status = -1;
        return (m_status);
    }*/

    if (ttcFifo != TTYP_FIFO) {
        ALTI_TTC_BGOMODESELECTION_BITSTRING bs;

        unsigned int bgo = ttcFifo;
        // READ
        if ((m_status = m_alti->TTC_BGOModeSelection_Read(bgo, bs)) != 0) {
            CERR("ALTI: reading BGO%d mode selection register", bgo);
            return (m_status);
        }
        // MODIFY
        bs.FIFORetransmit(retransmit ? "Enabled" : "Disabled");
        // WRITE
        if ((m_status = m_alti->TTC_BGOModeSelection_Write(bgo, bs)) != 0) {
            CERR("ALTI: writing BGO%d mode selection register", bgo);
            return (m_status);
        }
    }
    else {
        ALTI_TTC_TTYPMODESELECTION_BITSTRING bs;

        // READ
        if ((m_status = m_alti->TTC_TTYPModeSelection_Read(bs)) != 0) {
            CERR("ALTI: reading TTYP mode selection register", "");
            return (m_status);
        }
        // MODIFY
        bs.FIFORetransmit(retransmit ? "Enabled" : "Disabled");
        // WRITE
        if ((m_status = m_alti->TTC_TTYPModeSelection_Write(bs)) != 0) {
            CERR("ALTI: writing TTYP mode selection register", "");
            return (m_status);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCBgoFifoRetransmitRead(std::vector<bool> &retransmit) { // all
    /*retransmit.clear();

    ALTI_TTC_BGOMODESELECTION_BITSTRING bs;

    unsigned int bgo;
    std::string s;
    for (bgo = 0; bgo <= 3; bgo++) {
        // READ
        if ((m_status = m_alti->TTC_BGOModeSelection_Read(bgo, bs)) != 0) {
            CERR("ALTI: reading BGO%d mode selection register", bgo);
            return (m_status);
        }
        s = bs.FIFORetransmit();
        if (s == "Enabled") {
            retransmit.push_back(true);
        }
        else if (s == "Disabled") {
            retransmit.push_back(false);
         }
        else {
            CERR("ALTI: unknown value \"%s\" for BGO%d FIFO retransmit mode", s.c_str(), bgo);
            return (-1);
        }
    }*/

    retransmit.resize(TTC_FIFO_NUMBER);

    unsigned int i;
    TTC_FIFO ttcFifo;
    bool rd;
    for (i = 0; i < TTC_FIFO_NUMBER; i++) {
        ttcFifo = (TTC_FIFO) i;
        ENCBgoFifoRetransmitRead(ttcFifo, rd);
        retransmit[i] = rd;
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCBgoFifoRetransmitWrite(const std::vector<bool> &retransmit) { // all
    /*ALTI_TTC_BGOMODESELECTION_BITSTRING bs;

    unsigned int bgo;
    for (bgo = 0; bgo <= 3; bgo++) {
        // READ
        if ((m_status = m_alti->TTC_BGOModeSelection_Read(bgo, bs)) != 0) {
            CERR("ALTI: reading BGO%d mode selection register", bgo);
            return (m_status);
        }
        // MODIFY
        bs.FIFORetransmit(retransmit[bgo] ? "Enabled" : "Disabled");
        // WRITE
        if ((m_status = m_alti->TTC_BGOModeSelection_Write(bgo, bs)) != 0) {
            CERR("ALTI: writing BGO%d mode selection register", bgo);
            return (m_status);
        }
    }*/

    unsigned int i;
    TTC_FIFO ttcFifo;
    for (i = 0; i < TTC_FIFO_NUMBER; i++) {
        ttcFifo = (TTC_FIFO) i;
        ENCBgoFifoRetransmitWrite(ttcFifo, retransmit[i]);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCBgoFifoRetransmitWrite(const bool retransmit) { // all
    /*ALTI_TTC_BGOMODESELECTION_BITSTRING bs;

    unsigned int bgo;
    std::string s = (retransmit ? "Enabled" : "Disabled");
    for (bgo = 0; bgo <= 3; bgo++) {
        // READ
        if ((m_status = m_alti->TTC_BGOModeSelection_Read(bgo, bs)) != 0) {
            CERR("ALTI: reading BGO%d mode selection register", bgo);
            return (m_status);
        }
        // MODIFY
        bs.FIFORetransmit(s);
        // WRITE
        if ((m_status = m_alti->TTC_BGOModeSelection_Write(bgo, bs)) != 0) {
            CERR("ALTI: writing BGO%d mode selection register", bgo);
            return (m_status);
        }
    }*/

    unsigned int i;
    TTC_FIFO ttcFifo;
    for (i = 0; i < TTC_FIFO_NUMBER; i++) {
        ttcFifo = (TTC_FIFO) i;
        ENCBgoFifoRetransmitWrite(ttcFifo, retransmit);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCAsyncCommandPut(const TTC_FRAME_TYPE frameType, const uint16_t addr, const TTC_ADDRESS_SPACE addrSpace, const uint8_t subAddr, const uint8_t data) {

    TTC_SPY_COMMAND command;
    std::vector<unsigned int> tmp;
    tmp.clear();
    tmp.push_back(static_cast<unsigned int>(0x00000000));
    command = (RCD::BitSet) tmp;

    command.FrameType((frameType == TTC_FRAME_TYPE::SHORT) ? "Short" : ((frameType == TTC_FRAME_TYPE::LONG) ? "Long" : ""));
    command.Address(addr);
    command.AddressSpace((addrSpace == TTC_ADDRESS_SPACE::INTERNAL) ? "Internal" : ((addrSpace == TTC_ADDRESS_SPACE::EXTERNAL) ? "External" : ""));
    command.SubAddress(subAddr);
    command.Data(data);

    if ((m_status = m_alti->TTC_BGOAnyCommandVME_Write(command.number()[0])) != 0) {
        CERR("ALTI: putting async command 0x%08x (\"%s\", \"%s\", addr = 0x%04x, subAddr = 0x%02x, data = 0x%02x)", command.number()[0], TTC_FRAME_TYPE_NAME[frameType], TTC_ADDRESS_SPACE_NAME[addrSpace], addr, subAddr, data);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCAsyncCommandPutShort(const uint8_t data) {
    return ENCAsyncCommandPut(TTC_FRAME_TYPE::SHORT, 0x0000, TTC_ADDRESS_SPACE::INTERNAL, 0x00, data);
}

//------------------------------------------------------------------------------

int AltiModule::ENCAsyncCommandPutLong(const uint16_t addr, const TTC_ADDRESS_SPACE addrSpace, const uint8_t subAddr, const uint8_t data) {
    return ENCAsyncCommandPut(TTC_FRAME_TYPE::LONG, addr, addrSpace, subAddr, data);
}

//------------------------------------------------------------------------------

int AltiModule::ENCAsyncCommandPending(bool &pending) {
    ALTI_BGO_PIPELINEFIFOSEMPTYFULL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->BGO_PipelineFIFOsEmptyFull_Read(bs)) != 0) {
        CERR("ALTI: reading BGO/TTYP FIFOs status register", "");
        return (m_status);
    }

    std::string s = bs.AsyncCommandStatus();
    if (s == "Busy") {
        pending = true;
    }
    else if (s == "Ready") {
        pending = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for asynchronous command pending status", s.c_str());
        m_status = -1;
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::CLKResetMMCM() {
    ALTI_CLK_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CLK_Control_Read(bs)) != 0) {
        CERR("ALTI: reading clock control register", "");
        return (m_status);
    }
    // MODIFY
    bs.Reset(true);
    // WRITE
    if ((m_status = m_alti->CLK_Control_Write(bs)) != 0) {
        CERR("ALTI: writing clock control register", "");

        return (m_status);
    }
    // MODIFY
    bs.Reset(false);
    // WRITE
    if ((m_status = m_alti->CLK_Control_Write(bs)) != 0) {
        CERR("ALTI: writing clock control register", "");

        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKInputSelectReadMMCM(CLK_TYPE &clk) {
    ALTI_CLK_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CLK_Control_Read(bs)) != 0) {
        CERR("ALTI: reading clock control register", "");
        return (m_status);
    }

    std::string s = bs.SourceSelection();
    if (s == "PrimaryFXO") {
        clk = PRIMARY_FXO;
    }
    else if (s == "SecondaryEXT") {
        clk = SECONDARY_EXT;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for CLK type", s);
        return (-1);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKInputSelectWriteMMCM(const CLK_TYPE clk) {
    ALTI_CLK_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CLK_Control_Read(bs)) != 0) {
        CERR("ALTI: reading clock control register", "");
        return (m_status);
    }

    // WRITE: assert PLL reset
    bs.Reset(1);
    if ((m_status = m_alti->CLK_Control_Write(bs)) != 0) {
        CERR("ALTI: asserting PLL reset", "");
        return (m_status);
    }

    // WRITE: clock selection    
    bs.SourceSelection((clk == PRIMARY_FXO) ? "PrimaryFXO" : ((clk == SECONDARY_EXT) ? "SecondaryEXT" : ""));
    if ((m_status = m_alti->CLK_Control_Write(bs)) != 0) {
        CERR("ALTI: selecting \"%s\" clock", CLK_NAME[clk].c_str());
        return (m_status);
    }

    // WRITE: deassert PLL reset
    bs.Reset(0);
    if ((m_status = m_alti->CLK_Control_Write(bs)) != 0) {
        CERR("ALTI: deasserting PLL reset", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKInputSelectToggleMMCM() {
    ALTI_CLK_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CLK_Control_Read(bs)) != 0) {
        CERR("ALTI: reading clock control register", "");
        return (m_status);
    }

    CLK_TYPE clk_toggle;
    std::string s = bs.SourceSelection();
    if (s == "PrimaryFXO") {
        clk_toggle = SECONDARY_EXT;
    }
    else if (s == "SecondaryEXT") {
        clk_toggle = PRIMARY_FXO;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for CLK type", s);
        return (-1);
    }

    return CLKInputSelectWriteMMCM(clk_toggle);
}

//------------------------------------------------------------------------------

int AltiModule::CLKStatusReadMMCM(bool &monitor_int, bool &lol, bool &lol2, bool &sticky_lol, bool &sticky_lol2, bool &sticky_los) {
    ALTI_CLK_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CLK_Control_Read(bs)) != 0) {
        CERR("ALTI: reading clock control register", "");
        return (m_status);
    }

    monitor_int = bs.MonitorIntFlag();

    std::string s;

    s = bs.PLL_OutClks_Stability();
    if (s == "Stable") {
        lol = false;
    }
    else if (s == "NotStable") {
        lol = true;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for PLL LOL", s.c_str());
        return (-1);
    }
    s = bs.PLL2_OutClks_Stability();
    if (s == "Stable") {
        lol2 = false;
    }
    else if (s == "NotStable") {
        lol2 = true;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for 2nd PLL LOL", s.c_str());
        return (-1);
    }
    s = bs.PLL_OutClks_LOL();
    if (s == "LossOfLock") {
        sticky_lol = true;
    }
    else if (s == "NoLossOfLock") {
        sticky_lol = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for PLL LOL (latched)", s.c_str());
        return (-1);
    }
    s = bs.PLL2_OutClks_LOL();
    if (s == "LossOfLock") {
        sticky_lol2 = true;
    }
    else if (s == "NoLossOfLock") {
        sticky_lol2 = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for 2nd PLL LOL (latched)", s.c_str());
        return (-1);
    }
    s = bs.PLL_InClk_Stop();
    if (s == "Stopped") {
        sticky_los = true;
    }
    else if (s == "NotStopped") {
        sticky_los = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for PLL LOS (latched)", s.c_str());
        return (-1);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKStatusClearMMCM() {
    unsigned int data;

    // READ
    if ((m_status = m_alti->CLK_Control_Read(data)) != 0) {
        CERR("ALTI: reading clock control register", "");
        return (m_status);
    }

    // MODIFY
    data &= ~(ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_PLL_INCLK_STOP + ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_PLL_OUTCLKS_LOL + ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_PLL2_OUTCLKS_LOL); // clear
    data |= (ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_JITTERCLEANER_LOS + ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_JITTERCLEANER_LOL + ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_JITTERCLEANER_INTR); // don't clear
    data |= (ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_MONITORINTFLAG); // don't clear
    unsigned int i, mask;
    for (i = 0; i < ALTI_CLK_CONTROL_BITSTRING::CLK_STOP_NUMBER; i++) {
        mask = (ALTI_CLK_CONTROL_BITSTRING::MASK_CLK_STOP[i]).number()[0];
        data |= mask; // don't clear
    }
    for (i = 0; i < ALTI_CLK_CONTROL_BITSTRING::CLK_FREQUENCY_NUMBER; i++) {
        mask = (ALTI_CLK_CONTROL_BITSTRING::MASK_CLK_FREQUENCY[i]).number()[0];
        data |= mask; // don't clear
    }
    for (i = 0; i < ALTI_CLK_CONTROL_BITSTRING::CLK_GLITCH_NUMBER; i++) {
        mask = (ALTI_CLK_CONTROL_BITSTRING::MASK_CLK_GLITCH[i]).number()[0];
        data |= mask; // don't clear
    }

    // WRITE
    if ((m_status = m_alti->CLK_Control_Write(data)) != 0) {
        CERR("ALTI: writing clock control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKStatusReadJitterCleaner(bool &jcint, bool &jclos, bool &jclol/*, bool &jclatched*/) {
    ALTI_CLK_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CLK_Control_Read(bs)) != 0) {
        CERR("ALTI: reading clock control register", "");
        return (m_status);
    }

    std::string s;

    s = bs.JitterCleaner_IntR();
    if (s == "Interrupt") {
        jcint = true;
    }
    else if (s == "NoInterrupt") {
        jcint = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for JC interrupt request", s.c_str());
        return (-1);
    }
    s = bs.JitterCleaner_LOS();
    if (s == "LossOfSignal") {
        jclos = true;
    }
    else if (s == "NoLossOfSignal") {
        jclos = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for JC LOS", s.c_str());
        return (-1);
    }
    s = bs.JitterCleaner_LOL();
    if (s == "LossOfLock") {
        jclol = true;
    }
    else if (s == "NoLossOfLock") {
        jclol = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for JC LOL", s.c_str());
        return (-1);
    }
    //jclatched = bs.JitterCleaner_Latched();

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKStatusClearJitterCleaner() {
    unsigned int data;

    // READ
    if ((m_status = m_alti->CLK_Control_Read(data)) != 0) {
        CERR("ALTI: reading clock control register", "");
        return (m_status);
    }

    // MODIFY
    data |= (ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_PLL_INCLK_STOP + ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_PLL_OUTCLKS_LOL + ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_PLL2_OUTCLKS_LOL); // don't clear
    data &= ~(ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_JITTERCLEANER_LOS + ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_JITTERCLEANER_LOL + ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_JITTERCLEANER_INTR); // clear
    data |= (ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_MONITORINTFLAG); // don't clear
    unsigned int i, mask;
    for (i = 0; i < ALTI_CLK_CONTROL_BITSTRING::CLK_STOP_NUMBER; i++) {
        mask = (ALTI_CLK_CONTROL_BITSTRING::MASK_CLK_STOP[i]).number()[0];
        data |= mask; // don't clear
    }
    for (i = 0; i < ALTI_CLK_CONTROL_BITSTRING::CLK_FREQUENCY_NUMBER; i++) {
        mask = (ALTI_CLK_CONTROL_BITSTRING::MASK_CLK_FREQUENCY[i]).number()[0];
        data |= mask; // don't clear
    }
    for (i = 0; i < ALTI_CLK_CONTROL_BITSTRING::CLK_GLITCH_NUMBER; i++) {
        mask = (ALTI_CLK_CONTROL_BITSTRING::MASK_CLK_GLITCH[i]).number()[0];
        data |= mask; // don't clear
    }

    // WRITE
    if ((m_status = m_alti->CLK_Control_Write(data)) != 0) {
        CERR("ALTI: writing clock control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKMonitorReadMMCM(const EXT_CLK_SOURCE src, bool &stopped, bool &oor, bool &glitch) {
    ALTI_CLK_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CLK_Control_Read(bs)) != 0) {
        CERR("ALTI: reading clock control register", "");
        return (m_status);
    }

    std::string s1, s2, s3;
    s1 = bs.CLK_Stop(src);
    s2 = bs.CLK_Frequency(src);
    s3 = bs.CLK_Glitch(src);
    if (s1 == "Stopped") {
        stopped = true;
    }
    else if (s1 == "NotStopped") {
        stopped = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for \"%s\" CLK stop", s1, EXT_CLK_SOURCE_NAME[src].c_str());
        return (-1);
    }
    if (s2 == "OutOfRange") {
        oor = true;
    }
    else if (s2 == "InRange") {
        oor = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for \"%s\" CLK frequency", s2, EXT_CLK_SOURCE_NAME[src].c_str());
        return (-1);
    }
    if (s3 == "Glitch") {
        glitch = true;
    }
    else if (s3 == "NoGlitch") {
        glitch = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for \"%s\" CLK glitch", s3, EXT_CLK_SOURCE_NAME[src].c_str());
        return (-1);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKMonitorReadMMCM(std::vector<bool> &stopped, std::vector<bool> &oor, std::vector<bool> &glitch) { // all
    stopped.clear();
    oor.clear();
    glitch.clear();

    ALTI_CLK_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CLK_Control_Read(bs)) != 0) {
        CERR("ALTI: reading clock control register", "");
        return (m_status);
    }

    unsigned int i;
    EXT_CLK_SOURCE src;
    std::string s1, s2, s3;
    for (i = 0; i < EXT_CLK_SOURCE_NUMBER; i++) {
        src = (EXT_CLK_SOURCE) i;
        s1 = bs.CLK_Stop(i);
        s2 = bs.CLK_Frequency(i);
        s3 = bs.CLK_Glitch(i);
        if (s1 == "Stopped") {
            stopped.push_back(true);
        }
        else if (s1 == "NotStopped") {
            stopped.push_back(false);
        }
        else {
            CERR("ALTI: unknown value \"%s\" for \"%s\" CLK stop", s1, EXT_CLK_SOURCE_NAME[src].c_str());
            return (-1);
        }
        if (s2 == "OutOfRange") {
            oor.push_back(true);
        }
        else if (s2 == "InRange") {
            oor.push_back(false);
        }
        else {
            CERR("ALTI: unknown value \"%s\" for \"%s\" CLK frequency", s2, EXT_CLK_SOURCE_NAME[src].c_str());
            return (-1);
        }
        if (s3 == "Glitch") {
            glitch.push_back(true);
        }
        else if (s3 == "NoGlitch") {
            glitch.push_back(false);
        }
        else {
            CERR("ALTI: unknown value \"%s\" for \"%s\" CLK glitch", s3, EXT_CLK_SOURCE_NAME[src].c_str());
            return (-1);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKMonitorStatusClearMMCM() {
    unsigned int data;

    // READ
    if ((m_status = m_alti->CLK_Control_Read(data)) != 0) {
        CERR("ALTI: reading clock control register", "");
        return (m_status);
    }

    // MODIFY
    data |= (ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_PLL_INCLK_STOP + ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_PLL_OUTCLKS_LOL + ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_PLL2_OUTCLKS_LOL); // don't clear
    data |= (ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_JITTERCLEANER_LOS + ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_JITTERCLEANER_LOL + ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_JITTERCLEANER_INTR); // don't clear
    data &= ~(ALTI_CLK_CONTROL_BITSTRING::STAT_MASK_MONITORINTFLAG); // clear
    unsigned int i, mask;
    for (i = 0; i < ALTI_CLK_CONTROL_BITSTRING::CLK_STOP_NUMBER; i++) {
        mask = (ALTI_CLK_CONTROL_BITSTRING::MASK_CLK_STOP[i]).number()[0];
        data &= ~mask; // clear
    }
    for (i = 0; i < ALTI_CLK_CONTROL_BITSTRING::CLK_FREQUENCY_NUMBER; i++) {
        mask = (ALTI_CLK_CONTROL_BITSTRING::MASK_CLK_FREQUENCY[i]).number()[0];
        data &= ~mask; // clear
    }
    for (i = 0; i < ALTI_CLK_CONTROL_BITSTRING::CLK_GLITCH_NUMBER; i++) {
        mask = (ALTI_CLK_CONTROL_BITSTRING::MASK_CLK_GLITCH[i]).number()[0];
        data &= ~mask; // clear
    }

    // WRITE
    if ((m_status = m_alti->CLK_Control_Write(data)) != 0) {
        CERR("ALTI: writing clock control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKPhaseShiftMMCM(const bool inc, const unsigned int steps) {
    ALTI_CLK_PHASESHIFTCONTROLSTATUS_BITSTRING bs;

    std::string s;
    while (true) {
        // READ
        if ((m_status = m_alti->CLK_PhaseShiftControlStatus_Read(bs)) != 0) {
            CERR("ALTI: reading clock phase shift control and status register", "");
            return (m_status);
        }
        s = bs.PhaseShiftBusy();
        if (s == "Busy") {
            continue;
        }
        else if (s == "Ready") {
            break;
        }
        else {
            CERR("ALTI: unknown value \"%s\" for CLK phase shift status", s);
            return (-1);
        }
    }

    // MODIFY
    bs.IncDec(inc ? "Inc" : "Dec");
    bs.Steps(steps);

    // WRITE
    if ((m_status = m_alti->CLK_PhaseShiftControlStatus_Write(bs)) != 0) {
        CERR("ALTI: writing clock phase shift control and status register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKPhasePositionReadMMCM(int &steps) {
    ALTI_CLK_PHASESHIFTTOTALIZATOR_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CLK_PhaseShiftTotalizator_Read(bs)) != 0) {
        CERR("ALTI: reading clock phase shift totalizator register", "");
        return (m_status);
    }

    steps = bs.ActualPhasePosition();

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKPhaseResetMMCM() {
    int steps;

    if ((m_status = CLKPhasePositionReadMMCM(steps)) != SUCCESS) {
        CERR("ALTI: reading MMCM clock phase", "");
        return (m_status);
    }
    if ((m_status = CLKPhaseShiftMMCM((steps < 0), std::abs(steps))) != SUCCESS) {
        CERR("ALTI: %srementing MMCM clock phase by %d steps", ((steps < 0) ? "inc" : "dec"), std::abs(steps));
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKResetJitterCleaner() {
    ALTI_CLK_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CLK_Control_Read(bs)) != 0) {
        CERR("ALTI: reading clock control register", "");
        return (m_status);
    }
    // MODIFY
    bs.JitterCleaner_Reset(true);
    // WRITE
    if ((m_status = m_alti->CLK_Control_Write(bs)) != 0) {
        CERR("ALTI: writing clock control register", "");

        return (m_status);
    }
    // MODIFY
    bs.JitterCleaner_Reset(false);
    // WRITE
    if ((m_status = m_alti->CLK_Control_Write(bs)) != 0) {
        CERR("ALTI: writing clock control register", "");

        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKJitterCleanerResync() {
    if ((m_status = m_net->write(m_SI5344BPdnHardRstSync, (unsigned int) 0x04)) != 0) { // SYNC
        CERR("ALTI: writing 0x04 to \"%s\"", m_SI5344BPdnHardRstSync.name().c_str());
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKInputSelectReadJitterCleaner(CLK_TYPE &insel) {
    ALTI_CLK_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CLK_Control_Read(bs)) != 0) {
        CERR("ALTI: reading clock control register", "");
        return (m_status);
    }

    insel = (CLK_TYPE) bs.JitterCleaner_InSel();

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKInputSelectWriteJitterCleaner(const CLK_TYPE insel) {
    ALTI_CLK_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CLK_Control_Read(bs)) != 0) {
        CERR("ALTI: reading clock control register", "");
        return (m_status);
    }
    // MODIFY
    bs.JitterCleaner_InSel(insel);

    // WRITE
    if ((m_status = m_alti->CLK_Control_Write(bs)) != 0) {
        CERR("ALTI: writing clock control register", "");

        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKJitterCleanerIntFlagsRead(std::vector<bool> &los_flg, std::vector<bool> &oof_flg, bool &lol_flg, bool &hold_flg, bool &losxaxb_flg, bool &smbus_timeout_flg, bool &cal_flg_pll, bool &sysincal_flg) {
    unsigned int data, i;
    los_flg.clear();
    oof_flg.clear();

    if ((m_status = m_net->read(m_SI5344BOofLosStickyFlg, data)) != 0) { // OOF_LOS_STICKY_FLG
        CERR("ALTI: reading \"%s\"", m_SI5344BOofLosStickyFlg.name().c_str());
        return (m_status);
    }
    for (i = 0; i < 4; i++) { // 4 inputs
        los_flg.push_back(data & (0x01 << i)); // LOS_FLG[i]
        oof_flg.push_back(data & (0x10 << i)); // OOF_FLG[i]
    }
    if ((m_status = m_net->read(m_SI5344BHoldLolStickyFlg, data)) != 0) { // HOLD_LOL_STICKY_FLG
        CERR("ALTI: reading \"%s\"", m_SI5344BHoldLolStickyFlg.name().c_str());
        return (m_status);
    }
    lol_flg = (data & 0x02); // LOL_FLG
    hold_flg = (data & 0x20); // HOLD_FLG
    if ((m_status = m_net->read(m_SI5344BStatusStickyFlg, data)) != 0) { // STATUS_STICKY_FLG
        CERR("ALTI: reading \"%s\"", m_SI5344BStatusStickyFlg.name().c_str());
        return (m_status);
    }
    losxaxb_flg = (data & 0x02); // LOSXAXB_FLG
    smbus_timeout_flg = (data & 0x20); // SMBUS_TIMEOUT_FLG
    sysincal_flg = (data & 0x01); // SYSINCAL_FLG
    if ((m_status = m_net->read(m_SI5344BPllCalStickyFlg, data)) != 0) { // PLL_CAL_STICKY_FLG
        CERR("ALTI: reading \"%s\"", m_SI5344BPllCalStickyFlg.name().c_str());
        return (m_status);
    }
    cal_flg_pll = (data & 0x20); // CAL_FLG_PLL

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKJitterCleanerIntMasksRead(std::vector<bool> &los_intr_msk, std::vector<bool> &oof_intr_msk, bool &lol_intr_msk, bool &hold_intr_msk, bool &losxaxb_intr_msk, bool &smbus_timeout_intr_msk, bool &cal_intr_msk, bool &sysincal_intr_msk) {
    unsigned int data, i;
    los_intr_msk.clear();
    oof_intr_msk.clear();

    if ((m_status = m_net->read(m_SI5344BOofLosIntrMsk, data)) != 0) { // OOF_LOS_INTR_MSK
        CERR("ALTI: reading \"%s\"", m_SI5344BOofLosIntrMsk.name().c_str());
        return (m_status);
    }
    for (i = 0; i < 4; i++) { // 4 inputs
        los_intr_msk.push_back(data & (0x01 << i)); // LOS_INTR_MSK[i]
        oof_intr_msk.push_back(data & (0x10 << i)); // OOF_INTR_MSK[i]
    }
    if ((m_status = m_net->read(m_SI5344BHoldLolIntrMsk, data)) != 0) { // HOLD_LOL_INTR_MSK
        CERR("ALTI: reading \"%s\"", m_SI5344BHoldLolIntrMsk.name().c_str());
        return (m_status);
    }
    lol_intr_msk = (data & 0x02); // LOL_INTR_MSK
    hold_intr_msk = (data & 0x20); // HOLD_INTR_MSK
    if ((m_status = m_net->read(m_SI5344BStatusIntrMsk, data)) != 0) { // STATUS_INTR_MSK
        CERR("ALTI: reading \"%s\"", m_SI5344BStatusIntrMsk.name().c_str());
        return (m_status);
    }
    losxaxb_intr_msk = (data & 0x02); // LOSXAXB_INTR_MSK
    smbus_timeout_intr_msk = (data & 0x20); // SMBUS_TIMEOUT_INTR_MSK
    sysincal_intr_msk = (data & 0x01); // SYSINCAL_INTR_MSK
    if ((m_status = m_net->read(m_SI5344BPllCalIntrMsk, data)) != 0) { // PLL_CAL_INTR_MSK
        CERR("ALTI: reading \"%s\"", m_SI5344BPllCalIntrMsk.name().c_str());
        return (m_status);
    }
    cal_intr_msk = (data & 0x20); // CAL_INTR_MSK

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKJitterCleanerStatusClear() {
    unsigned int flg, intr_msk, data;

    if ((m_status = m_net->read(m_SI5344BOofLosStickyFlg, flg)) != 0) { // OOF_LOS_STICKY_FLG
        CERR("ALTI: reading \"%s\"", m_SI5344BOofLosStickyFlg.name().c_str());
        return (m_status);
    }
    if ((m_status = m_net->read(m_SI5344BOofLosIntrMsk, intr_msk)) != 0) { // OOF_LOS_INTR_MSK
        CERR("ALTI: reading \"%s\"", m_SI5344BOofLosIntrMsk.name().c_str());
        return (m_status);
    }
    //data = uint8_t(~(flg & (~intr_msk) & 0xff))); // clear bits 0..7, if needed: LOS_FLG[3..0], OOF_FLG[3..0]
    data = uint8_t(0x00); // clear bits 0..7 (Rev B issues)
    if ((m_status = m_net->write(m_SI5344BOofLosStickyFlg, data)) != 0) { // OOF_LOS_STICKY_FLG
        CERR("ALTI: writing 0x%02x to \"%s\"", data, m_SI5344BOofLosStickyFlg.name().c_str());
        return (m_status);
    }

    if ((m_status = m_net->read(m_SI5344BHoldLolStickyFlg, flg)) != 0) { // HOLD_LOL_STICKY_FLG
        CERR("ALTI: reading \"%s\"", m_SI5344BHoldLolStickyFlg.name().c_str());
        return (m_status);
    }
    if ((m_status = m_net->read(m_SI5344BHoldLolIntrMsk, intr_msk)) != 0) { // HOLD_LOL_INTR_MSK
        CERR("ALTI: reading \"%s\"", m_SI5344BHoldLolIntrMsk.name().c_str());
        return (m_status);
    }
    data = uint8_t(~(flg & (~intr_msk) & 0x22)); // clear bits 1 and 5, if needed: LOL_FLG, HOLD_FLG
    if ((m_status = m_net->write(m_SI5344BHoldLolStickyFlg, data)) != 0) { // HOLD_LOL_STICKY_FLG
        CERR("ALTI: writing 0x%02x to \"%s\"", data, m_SI5344BHoldLolStickyFlg.name().c_str());
        return (m_status);
    }

    if ((m_status = m_net->read(m_SI5344BStatusStickyFlg, flg)) != 0) { // STATUS_STICKY_FLG
        CERR("ALTI: reading \"%s\"", m_SI5344BStatusStickyFlg.name().c_str());
        return (m_status);
    }
    if ((m_status = m_net->read(m_SI5344BStatusIntrMsk, intr_msk)) != 0) { // STATUS_INTR_MSK
        CERR("ALTI: reading \"%s\"", m_SI5344BStatusIntrMsk.name().c_str());
        return (m_status);
    }
    data = uint8_t(~(flg & (~intr_msk) & 0x23)); // clear bits 0, 1 and 5, if needed: SYSINCAL_FLG, LOSXAXB_FLG, SMBUS_TIMEOUT_FLG
    if ((m_status = m_net->write(m_SI5344BStatusStickyFlg, data)) != 0) { // STATUS_STICKY_FLG
        CERR("ALTI: writing 0x%02x to \"%s\"", data, m_SI5344BStatusStickyFlg.name().c_str());
        return (m_status);
    }

    if ((m_status = m_net->read(m_SI5344BPllCalStickyFlg, flg)) != 0) { // PLL_CAL_STICKY_FLG
        CERR("ALTI: reading \"%s\"", m_SI5344BPllCalStickyFlg.name().c_str());
        return (m_status);
    }
    if ((m_status = m_net->read(m_SI5344BPllCalIntrMsk, intr_msk)) != 0) { // PLL_CAL_INTR_MSK
        CERR("ALTI: reading \"%s\"", m_SI5344BPllCalIntrMsk.name().c_str());
        return (m_status);
    }
    data = uint8_t(~(flg & (~intr_msk) & 0x20)); // clear bit 5, if needed: CAL_FLG_PLL
    if ((m_status = m_net->write(m_SI5344BPllCalStickyFlg, data)) != 0) { // PLL_CAL_STICKY_FLG
        CERR("ALTI: writing 0x%02x to \"%s\"", data, m_SI5344BPllCalStickyFlg.name().c_str());
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::pagedI2CDeviceRead(const I2CNetwork::I2CNode &node, const uint8_t page, const uint8_t offs, unsigned int &data) {
    int rtnv;
    I2CInterface *ii2c = m_net->getI2CBus(node.bus());
    unsigned int iswc, isel, pageAddr, devAddr;
    iswc = m_net->getSwitchAddressI2CDevice(node.bus(), node.dev());
    isel = m_net->getSwitchSelectI2CDevice(node.bus(), node.dev());
    pageAddr = m_net->getPageOffsetI2CDevice(node.bus(), node.dev());
    devAddr = m_net->getAddressI2CDevice(node.bus(), node.dev());
    uint8_t datc[I2CInterface::MAX_BLOCK];

    // write I2C switch
    unsigned int numb(1);
    CDBG("write switch: address = 0x%02x, select = 0x%02x", iswc, isel);
    if ((rtnv = ii2c->DataWrite(iswc, isel, numb, datc, I2CInterface::BYTE)) != SUCCESS) {
        CERR("writing I2C bus/device %d/%d, switch 0x%02x, select 0x%02x", node.bus(), node.dev(), iswc, isel);
        return (rtnv);
    }

    // write I2C page select
    datc[0] = page;
    numb = 1;
    CDBG("write page: address = 0x%02x, select = 0x%02x", devAddr, addr_hi);
    if ((rtnv = ii2c->DataWrite(devAddr, pageAddr, numb, datc, I2CInterface::DATA)) != SUCCESS) {
        CERR("writing I2C bus/device %d/%d, offset 0x%02x, select 0x%02x", node.bus(), node.dev(), pageAddr, page);
        return (rtnv);
    }

    // read I2C data
    numb = 1;
    if ((rtnv = ii2c->DataRead(devAddr, offs, numb, datc, I2CInterface::DATA)) != SUCCESS) {
        CERR("reading I2C bus/device %d/%d, page 0x%02x, offset 0x%02x", node.bus(), node.dev(), page, offs);
        return (rtnv);
    }
    data = datc[0];

    //return (m_status);
    return 0;
}

//------------------------------------------------------------------------------

int AltiModule::pagedI2CDeviceWrite(const I2CNetwork::I2CNode &node, const uint8_t page, const uint8_t offs, const unsigned int data) {
    int rtnv;
    I2CInterface *ii2c = m_net->getI2CBus(node.bus());
    unsigned int iswc, isel, pageAddr, devAddr;
    iswc = m_net->getSwitchAddressI2CDevice(node.bus(), node.dev());
    isel = m_net->getSwitchSelectI2CDevice(node.bus(), node.dev());
    pageAddr = m_net->getPageOffsetI2CDevice(node.bus(), node.dev());
    devAddr = m_net->getAddressI2CDevice(node.bus(), node.dev());
    uint8_t datc[I2CInterface::MAX_BLOCK];

    // write I2C switch
    unsigned int numb(1);
    CDBG("write switch: address = 0x%02x, select = 0x%02x", iswc, isel);
    if ((rtnv = ii2c->DataWrite(iswc, isel, numb, datc, I2CInterface::BYTE)) != SUCCESS) {
        CERR("writing I2C bus/device %d/%d, switch 0x%02x, select 0x%02x", node.bus(), node.dev(), iswc, isel);
        return (rtnv);
    }

    // write I2C page select
    datc[0] = page;
    numb = 1;
    CDBG("write page: address = 0x%02x, select = 0x%02x", devAddr, addr_hi);
    if ((rtnv = ii2c->DataWrite(devAddr, pageAddr, numb, datc, I2CInterface::DATA)) != SUCCESS) {
        CERR("writing I2C bus/device %d/%d, offset 0x%02x, select 0x%02x", node.bus(), node.dev(), pageAddr, page);
        return (rtnv);
    }

    // write I2C data
    datc[0] = data;
    numb = 1;
    if ((rtnv = ii2c->DataWrite(devAddr, offs, numb, datc, I2CInterface::DATA)) != SUCCESS) {
        CERR("writing I2C bus/device %d/%d, page 0x%02x, offset 0x%02x", node.bus(), node.dev(), page, offs);
        return (rtnv);
    }

    //return (m_status);
    return 0;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::CLKJitterCleanerConfigRead(const std::string &fn_in, const std::string &fn_out) { // read configuration (addresses given in input file) into output file

    // open input file
    std::ifstream fs_in(fn_in);
    if (!fs_in) {
        CERR("cannot open input file \"%s\"\n", fn_in.c_str());
        return (m_status = WRONGPAR);
    }

    // open output file
    std::ofstream fs_out(fn_out);
    if (!fs_out) {
        CERR("cannot open output file \"%s\"\n", fn_out.c_str());
        return (m_status = WRONGPAR);
    }

    unsigned int line(0), lnum(0), addr, addr_hi, addr_lo;
    unsigned int data;
    std::string buf;
    std::vector<std::string> arg;
    char xbuf[1024];

    I2CNetwork::I2CNode node;
    const std::string nodeName = "BUS[0]/CLK_GEN_CLEANER_P2/DESIGN_ID";
    if ((m_status = m_net->getI2CNode(nodeName, node)) != SUCCESS) {
        CERR("Cannot find I2C node \"%s\"", nodeName.c_str());
    }

    // read address from input file
    while (getline(fs_in, buf)) {
        if ((buf[0] == '#') or (buf == "Address,Data")) {
            if (buf.find("# Delay") != std::string::npos) {
                if (runDelay(line, buf) != SUCCESS) {
                    CERR("running delay command from line %d \"%s\"", line, buf.c_str());
                    return (m_status);
                }
            }
            line++;
            continue;
        }
        if (AltiCommon::splitString(buf, ",", arg) < 1) {
            CERR("reading arguments from line %d \"%s\" (expected \"addr ...\")", line, buf.c_str());
            return (m_status = WRONGPAR);
        }
        if (AltiCommon::readNumber(arg[0], addr)) {
            CERR("reading address from line %d \"%s\"", line, buf.c_str());
            return (m_status = WRONGPAR);
        }
        addr_hi = (addr >> PAGE_ADDR_SHFT) & PAGE_ADDR_MASK;
        addr_lo = addr & ADDR_ADDR_MASK;

        // instead:
        pagedI2CDeviceRead(node, addr_hi, addr_lo, data);

        // write address and data to output file
        std::sprintf(xbuf, "0x%04X,0x%02X", addr, data);
        fs_out << xbuf << std::endl;
        line++;
        lnum++;
    }
    COUT("read %d (addr) line%s from input file \"%s\" and wrote data to output file \"%s\"", lnum, (lnum > 1) ? "s" : "" , fn_in.c_str(), fn_out.c_str());
    std::printf("read from CLK jitter cleaner input file \"%s\" and output file \"%s\"\n", fn_in.c_str(), fn_in.c_str());

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKJitterCleanerConfigWrite(const std::string &fn_in) { // write configuration from input file

    if (fn_in.empty() or (fn_in == "<DEFAULT>")) return (CLKJitterCleanerConfigWrite());

    // open input file
    std::ifstream fs_in(fn_in);
    if (!fs_in) {
        CERR("cannot open input file \"%s\"\n", fn_in.c_str());
        return (m_status = WRONGPAR);
    }

    unsigned int line(0), lnum(0), addr, addr_hi, addr_lo;
    unsigned int data;
    std::string buf;
    std::vector<std::string> arg;

    I2CNetwork::I2CNode node;
    const std::string nodeName = "BUS[0]/CLK_GEN_CLEANER_P2/DESIGN_ID";
    if ((m_status = m_net->getI2CNode(nodeName, node)) != SUCCESS) {
        CERR("Cannot find I2C node \"%s\"", nodeName.c_str());
    }

    // read address from input file
    while (getline(fs_in, buf)) {
        if ((buf[0] == '#') or (buf == "Address,Data")) {
            if (buf.find("# Delay") != std::string::npos) {
                if (runDelay(line, buf) != SUCCESS) {
                    CERR("running delay command from line %d \"%s\"", line, buf.c_str());
                    return (m_status);
                }
            }
            line++;
            continue;
        }
        if (AltiCommon::splitString(buf, ",", arg) < 2) {
            CERR("reading arguments from line %d \"%s\" (expected \"addr, data\")", line, buf.c_str());
            return (m_status = WRONGPAR);
        }
        if (AltiCommon::readNumber(arg[0], addr)) {
            CERR("reading address from line %d \"%s\"", line, buf.c_str());
            return (m_status = WRONGPAR);
        }
        if (AltiCommon::readNumber(arg[1], data)) {
            CERR("reading data from line %d \"%s\"", line, buf.c_str());
            return (m_status = WRONGPAR);
        }
        addr_hi = (addr >> PAGE_ADDR_SHFT) & PAGE_ADDR_MASK;
        addr_lo = addr & ADDR_ADDR_MASK;

        // instead:
        pagedI2CDeviceWrite(node, addr_hi, addr_lo, data);

        line++;
        lnum++;
    }
    COUT("read %d (addr) line%s from input file \"%s\"", lnum, (lnum > 1) ? "s" : "" , fn_in.c_str());
    std::printf("wrote input file \"%s\" to CLK jitter cleaner\n", fn_in.c_str());

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKJitterCleanerConfigWrite() { // write default configuration

    I2CNetwork::I2CNode node;
    const std::string nodeName = "BUS[0]/CLK_GEN_CLEANER_P2/DESIGN_ID";
    if ((m_status = m_net->getI2CNode(nodeName, node)) != SUCCESS) {
        CERR("Cannot find I2C node \"%s\"", nodeName.c_str());
    }

    unsigned int addr, addr_hi, addr_lo, data, line(0);
    static struct timespec tdel; tdel.tv_sec = 0L; tdel.tv_nsec = 300000000L; // 300 msec

    for (unsigned int i = 0; i < JITTER.NUMBER; i++) {

        //if(m_debug>=2) std::printf("addr = 0x%04x, data = 0x%02x\n",JITTER.si5344_revb_registers[i].address,JITTER.si5344_revb_registers[i].value);

        addr = JITTER.si5344_revb_registers[i].address;
        addr_lo = addr & 0xff;
        addr_hi = (addr >> 8) & 0xff;
        data = JITTER.si5344_revb_registers[i].value;

        /*// write page address
        if((m_status = writeSPI(SPI0,0x01,addr_hi)) != SUCCESS) {
            CERR("writing page address of line %d to CLK jitter cleaner",line);
            return(m_status);
        }

        // write data
        if((m_status = writeSPI(SPI0,addr_lo,data)) != SUCCESS) {
            CERR("writing data of line %d to CLK jitter cleaner",line);
            return(m_status);
        }
        line++;

        // run delay if end of configuration preamble
        if((addr == 0x0540) and (data == 0x01)) nanosleep(&tdel,nullptr);*/

        // instead:
        pagedI2CDeviceWrite(node, addr_hi, addr_lo, data);
        line++;

        // run delay if end of configuration preamble
        if ((addr == 0x0540) and (data == 0x01)) nanosleep(&tdel, nullptr);
    }

    std::printf("wrote %d (addr,data) line%s to CLK jitter cleaner\n", line, (line > 1) ? "s" : "");

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKJitterCleanerStatusRead(std::vector<bool> &oof, std::vector<bool> &los, bool &hold, bool &lol) {
    unsigned int data, i;
    oof.clear();
    los.clear();

    if ((m_status = m_net->read(m_SI5344BOofLosFlg, data)) != 0) { // OOF_LOS_FLG
        CERR("ALTI: reading \"%s\"", m_SI5344BOofLosFlg.name().c_str());
        return (m_status);
    }
    for (i = 0; i < 4; i++) { // 4 inputs
        oof.push_back(data & (0x10 << i)); // OOF[i]
        los.push_back(data & (0x01 << i)); // LOS[i]
    }
    if ((m_status = m_net->read(m_SI5344BHoldLolFlg, data)) != 0) { // HOLD_LOL_FLG
        CERR("ALTI: reading \"%s\"", m_SI5344BHoldLolFlg.name().c_str());
        return (m_status);
    }
    hold = (data & 0x20); // HOLD
    lol = (data & 0x02); // LOL

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKJitterCleanerStickyStatusRead(std::vector<bool> &oof, std::vector<bool> &los, bool &hold, bool &lol) {
    unsigned int data, i;
    oof.clear();
    los.clear();

    if ((m_status = m_net->read(m_SI5344BOofLosStickyFlg, data)) != 0) { // OOF_LOS_STICKY_FLG
        CERR("ALTI: reading \"%s\"", m_SI5344BOofLosStickyFlg.name().c_str());
        return (m_status);
    }
    for (i = 0; i < 4; i++) { // 4 inputs
        oof.push_back(data & (0x10 << i)); // OOF[i]
        los.push_back(data & (0x01 << i)); // LOS[i]
    }
    if ((m_status = m_net->read(m_SI5344BHoldLolStickyFlg, data)) != 0) { // HOLD_LOL_STICKY_FLG
        CERR("ALTI: reading \"%s\"", m_SI5344BHoldLolStickyFlg.name().c_str());
        return (m_status);
    }
    hold = (data & 0x20); // HOLD
    lol = (data & 0x02); // LOL

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKJitterCleanerDesignIDRead(std::string &designid) {
    //designid.clear();
    designid.resize(8);

    if ((m_status = m_net->read(m_SI5344BDesignId, designid)) != 0) { // DESIGN_ID
        CERR("ALTI: reading \"%s\"", m_SI5344BDesignId.name().c_str());
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::runDelay(const unsigned int line, const std::string &buf) {

    std::vector<std::string> arg;
    int num;
    unsigned int tim;

    if ((num = AltiCommon::splitString(buf, " ", arg)) < 1) {
        CERR("reading arguments from line %d \"%s\" (expected \"# Delay xxx msec\")", line, buf.c_str());
        return (m_status = WRONGPAR);
    }
    if (num != 4) {
        CERR("wrong number arguments from line %d \"%s\" (expected %d\")", line, buf.c_str());
        return (m_status = WRONGPAR);
    }
    if (AltiCommon::readNumber(arg[2], tim)) {
        CERR("reading delay time from line %d \"%s\"", line, buf.c_str());
        return (m_status = WRONGPAR);
    }
    if (arg[3] != "msec") {
        CERR("reading delay unit from line %d \"%s\" (expected \"msec\")", line, buf.c_str());
        return (m_status = WRONGPAR);
    }

    struct timespec req, rem;
    if (tim < THOUSAND) {
        req.tv_sec = 0;
        req.tv_nsec = tim*THOUSAND*THOUSAND;
    }
    else {
        req.tv_sec = tim/THOUSAND;
        req.tv_nsec = (tim%THOUSAND)*THOUSAND*THOUSAND;
    }
    if (nanosleep(&req, &rem) < 0) {
        CERR("nanosleep(%d sec, %d nsec) returns \"%s\"", req.tv_sec, req.tv_nsec, strerror(errno));
        return (m_status = FAILURE);
    }

    return (m_status);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::CNTEnable(const bool enable) { // all
    ALTI_CNT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CNT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading counter control register", "");
        return (m_status);
    }
    // MODIFY
    std::string sen = (enable ? "Enabled" : "Disabled");
    bs.EnableBC(sen);
    bs.EnableORB(sen);
    bs.EnableL1A(sen);
    bs.EnableTTR1(sen);
    bs.EnableTTR2(sen);
    bs.EnableTTR3(sen);
    bs.EnableBGO0(sen);
    bs.EnableBGO1(sen);
    bs.EnableBGO2(sen);
    bs.EnableBGO3(sen);
    // WRITE
    if ((m_status = m_alti->CNT_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing counter control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTEnable(const CNT_TYPE cnt, const bool enable) {
    ALTI_CNT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CNT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading counter control register", "");
        return (m_status);
    }
    // MODIFY
    std::string sen = (enable ? "Enabled" : "Disabled");
    switch (cnt) {
        case CNT_BC:
            bs.EnableBC(sen);
            break;
        case CNT_ORB:
            bs.EnableORB(sen);
            break;
        case CNT_L1A:
            bs.EnableL1A(sen);
            break;
        case CNT_TTR1:
            bs.EnableTTR1(sen);
            break;
        case CNT_TTR2:
            bs.EnableTTR2(sen);
            break;
        case CNT_TTR3:
            bs.EnableTTR3(sen);
            break;
        case CNT_BGO0:
            bs.EnableBGO0(sen);
            break;
        case CNT_BGO1:
            bs.EnableBGO1(sen);
            break;
        case CNT_BGO2:
            bs.EnableBGO2(sen);
            break;
        case CNT_BGO3:
            bs.EnableBGO3(sen);
            break;
        default:
            break;
    }
    // WRITE
    if ((m_status = m_alti->CNT_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing counter control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

/*int AltiModule::CNTReset(const bool reset) { // all
    ALTI_CNT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CNT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading counter control register", "");
        return (m_status);
    }
    // MODIFY
    std::string sen = (reset ? "Enabled" : "Disabled");
    bs.ResetBC(sen);
    bs.ResetORB(sen);
    bs.ResetL1A(sen);
    bs.ResetTTR1(sen);
    bs.ResetTTR2(sen);
    bs.ResetTTR3(sen);
    bs.ResetBGO0(sen);
    bs.ResetBGO1(sen);
    bs.ResetBGO2(sen);
    bs.ResetBGO3(sen);
    // WRITE
    if ((m_status = m_alti->CNT_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing counter control register", "");
        return (m_status);
    }

    return (m_status);
}*/

//------------------------------------------------------------------------------

/*int AltiModule::CNTReset(const CNT_TYPE cnt, const bool reset) {
    ALTI_CNT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CNT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading counter control register", "");
        return (m_status);
    }
    // MODIFY
    std::string sen = (reset ? "Enabled" : "Disabled");
    switch (cnt) {
        case CNT_BC:
            bs.ResetBC(sen);
            break;
        case CNT_ORB:
            bs.ResetORB(sen);
            break;
        case CNT_L1A:
            bs.ResetL1A(sen);
            break;
        case CNT_TTR1:
            bs.ResetTTR1(sen);
            break;
        case CNT_TTR2:
            bs.ResetTTR2(sen);
            break;
        case CNT_TTR3:
            bs.ResetTTR3(sen);
            break;
        case CNT_BGO0:
            bs.ResetBGO0(sen);
            break;
        case CNT_BGO1:
            bs.ResetBGO1(sen);
            break;
        case CNT_BGO2:
            bs.ResetBGO2(sen);
            break;
        case CNT_BGO3:
            bs.ResetBGO3(sen);
            break;
        default:
            break;
    }
    // WRITE
    if ((m_status = m_alti->CNT_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing counter control register", "");
        return (m_status);
    }

    return (m_status);
}*/

//------------------------------------------------------------------------------

int AltiModule::CNTReset() { // all
    ALTI_CNT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CNT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading counter control register", "");
        return (m_status);
    }
    std::string sen;
    // MODIFY
    sen = "Enabled";
    bs.ResetBC(sen);
    bs.ResetORB(sen);
    bs.ResetL1A(sen);
    bs.ResetTTR1(sen);
    bs.ResetTTR2(sen);
    bs.ResetTTR3(sen);
    bs.ResetBGO0(sen);
    bs.ResetBGO1(sen);
    bs.ResetBGO2(sen);
    bs.ResetBGO3(sen);
    // WRITE
    if ((m_status = m_alti->CNT_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing counter control register", "");
        return (m_status);
    }
    // MODIFY
    sen = "Disabled";
    bs.ResetBC(sen);
    bs.ResetORB(sen);
    bs.ResetL1A(sen);
    bs.ResetTTR1(sen);
    bs.ResetTTR2(sen);
    bs.ResetTTR3(sen);
    bs.ResetBGO0(sen);
    bs.ResetBGO1(sen);
    bs.ResetBGO2(sen);
    bs.ResetBGO3(sen);
    // WRITE
    if ((m_status = m_alti->CNT_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing counter control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTReset(const CNT_TYPE cnt) {
    ALTI_CNT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CNT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading counter control register", "");
        return (m_status);
    }
    std::string sen;
    // MODIFY
    sen = "Enabled";
    switch (cnt) {
        case CNT_BC:
            bs.ResetBC(sen);
            break;
        case CNT_ORB:
            bs.ResetORB(sen);
            break;
        case CNT_L1A:
            bs.ResetL1A(sen);
            break;
        case CNT_TTR1:
            bs.ResetTTR1(sen);
            break;
        case CNT_TTR2:
            bs.ResetTTR2(sen);
            break;
        case CNT_TTR3:
            bs.ResetTTR3(sen);
            break;
        case CNT_BGO0:
            bs.ResetBGO0(sen);
            break;
        case CNT_BGO1:
            bs.ResetBGO1(sen);
            break;
        case CNT_BGO2:
            bs.ResetBGO2(sen);
            break;
        case CNT_BGO3:
            bs.ResetBGO3(sen);
            break;
        default:
            break;
    }
    // WRITE
    if ((m_status = m_alti->CNT_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing counter control register", "");
        return (m_status);
    }
    // MODIFY
    sen = "Disabled";
    switch (cnt) {
        case CNT_BC:
            bs.ResetBC(sen);
            break;
        case CNT_ORB:
            bs.ResetORB(sen);
            break;
        case CNT_L1A:
            bs.ResetL1A(sen);
            break;
        case CNT_TTR1:
            bs.ResetTTR1(sen);
            break;
        case CNT_TTR2:
            bs.ResetTTR2(sen);
            break;
        case CNT_TTR3:
            bs.ResetTTR3(sen);
            break;
        case CNT_BGO0:
            bs.ResetBGO0(sen);
            break;
        case CNT_BGO1:
            bs.ResetBGO1(sen);
            break;
        case CNT_BGO2:
            bs.ResetBGO2(sen);
            break;
        case CNT_BGO3:
            bs.ResetBGO3(sen);
            break;
        default:
            break;
    }
    // WRITE
    if ((m_status = m_alti->CNT_Control_Write(bs)) != 0 ) {
        CERR("ALTI: writing counter control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTStatusRead(const CNT_TYPE cnt, bool &enabled, bool &reset) {
    ALTI_CNT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CNT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading counter control register", "");
        return (m_status);
    }

    std::string s1, s2;
    switch (cnt) {
        case CNT_BC:
            s1 = bs.EnableBC();
            s2 = bs.ResetBC();
            break;
        case CNT_ORB:
            s1 = bs.EnableORB();
            s2 = bs.ResetORB();
            break;
        case CNT_L1A:
            s1 = bs.EnableL1A();
            s2 = bs.ResetL1A();
            break;
        case CNT_TTR1:
            s1 = bs.EnableTTR1();
            s2 = bs.ResetTTR1();
            break;
        case CNT_TTR2:
            s1 = bs.EnableTTR2();
            s2 = bs.ResetTTR2();
            break;
        case CNT_TTR3:
            s1 = bs.EnableTTR3();
            s2 = bs.ResetTTR3();
            break;
        case CNT_BGO0:
            s1 = bs.EnableBGO0();
            s2 = bs.ResetBGO0();
            break;
        case CNT_BGO1:
            s1 = bs.EnableBGO1();
            s2 = bs.ResetBGO1();
            break;
        case CNT_BGO2:
            s1 = bs.EnableBGO2();
            s2 = bs.ResetBGO2();
            break;
        case CNT_BGO3:
            s1 = bs.EnableBGO3();
            s2 = bs.ResetBGO3();
            break;
        default:
            break;
    }

    if (s1 == "Enabled") {
        enabled = true;
    }
    else if (s1 == "Disabled") {
        enabled = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for \"%s\" enable", s1, AltiModule::CNT_NAME[cnt].c_str());
        return (-1);
    }
    if (s2 == "Enabled") {
        reset = true;
    }
    else if (s2 == "Disabled") {
        reset = false;
    }
    else {
        CERR("ALTI: unknown value \"%s\" for \"%s\" reset", s2, AltiModule::CNT_NAME[cnt].c_str());
        return (-1);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTStatusRead(std::vector<bool> &enabled, std::vector<bool> &reset) {
    ALTI_CNT_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CNT_Control_Read(bs)) != 0 ) {
        CERR("ALTI: reading counter control register", "");
        return (m_status);
    }

    enabled.clear();
    reset.clear();

    std::string s1, s2;
    unsigned int i;
    CNT_TYPE cnt;
    for (i = 0; i < CNT_NUMBER; i++) {
        cnt = (CNT_TYPE) i;
        switch (cnt) {
            case CNT_BC:
                s1 = bs.EnableBC();
                s2 = bs.ResetBC();
                break;
            case CNT_ORB:
                s1 = bs.EnableORB();
                s2 = bs.ResetORB();
                break;
            case CNT_L1A:
                s1 = bs.EnableL1A();
                s2 = bs.ResetL1A();
                break;
            case CNT_TTR1:
                s1 = bs.EnableTTR1();
                s2 = bs.ResetTTR1();
                break;
            case CNT_TTR2:
                s1 = bs.EnableTTR2();
                s2 = bs.ResetTTR2();
                break;
            case CNT_TTR3:
                s1 = bs.EnableTTR3();
                s2 = bs.ResetTTR3();
                break;
            case CNT_BGO0:
                s1 = bs.EnableBGO0();
                s2 = bs.ResetBGO0();
                break;
            case CNT_BGO1:
                s1 = bs.EnableBGO1();
                s2 = bs.ResetBGO1();
                break;
            case CNT_BGO2:
                s1 = bs.EnableBGO2();
                s2 = bs.ResetBGO2();
                break;
            case CNT_BGO3:
                s1 = bs.EnableBGO3();
                s2 = bs.ResetBGO3();
                break;
            default:
                break;
        }

        if (s1 == "Enabled") {
            enabled.push_back(true);
        }
        else if (s1 == "Disabled") {
            enabled.push_back(false);
        }
        else {
            CERR("ALTI: unknown value \"%s\" for \"%s\" enable", s1, AltiModule::CNT_NAME[cnt].c_str());
            return (-1);
        }
        if (s2 == "Enabled") {
            reset.push_back(true);
        }
        else if (s2 == "Disabled") {
            reset.push_back(false);
        }
        else {
            CERR("ALTI: unknown value \"%s\" for \"%s\" reset", s2, AltiModule::CNT_NAME[cnt].c_str());
            return (-1);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTRead(std::vector<unsigned int> &cnt) { // all
    unsigned int rdat;
    cnt.clear();

    if ((m_status = m_alti->CNT_BC_Read(rdat)) != 0) {
        CERR("ALTI: reading BC counter", "");
        return (m_status);
    }
    cnt.push_back(rdat);

    if ((m_status = m_alti->CNT_ORB_Read(rdat)) != 0) {
        CERR("ALTI: reading ORB counter", "");
        return (m_status);
    }
    cnt.push_back(rdat);

    if ((m_status = m_alti->CNT_L1A_Read(rdat)) != 0) {
        CERR("ALTI: reading L1A counter", "");
        return (m_status);
    }
    cnt.push_back(rdat);

    if ((m_status = m_alti->CNT_TTR1_Read(rdat)) != 0) {
        CERR("ALTI: reading TTR1 counter", "");
        return (m_status);
    }
    cnt.push_back(rdat);

    if ((m_status = m_alti->CNT_TTR2_Read(rdat)) != 0) {
        CERR("ALTI: reading TTR2 counter", "");
        return (m_status);
    }
    cnt.push_back(rdat);

    if ((m_status = m_alti->CNT_TTR3_Read(rdat)) != 0) {
        CERR("ALTI: reading TTR3 counter", "");
        return (m_status);
    }
    cnt.push_back(rdat);

    if ((m_status = m_alti->CNT_BGO0_Read(rdat)) != 0) {
        CERR("ALTI: reading BGO0 counter", "");
        return (m_status);
    }
    cnt.push_back(rdat);

    if ((m_status = m_alti->CNT_BGO1_Read(rdat)) != 0) {
        CERR("ALTI: reading BGO1 counter", "");
        return (m_status);
    }
    cnt.push_back(rdat);

    if ((m_status = m_alti->CNT_BGO2_Read(rdat)) != 0) {
        CERR("ALTI: reading BGO2 counter", "");
        return (m_status);
    }
    cnt.push_back(rdat);

    if ((m_status = m_alti->CNT_BGO3_Read(rdat)) != 0) {
        CERR("ALTI: reading BGO3 counter", "");
        return (m_status);
    }
    cnt.push_back(rdat);

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTRead(const CNT_TYPE cnt, unsigned int &rdata) {
    switch (cnt) {
        case CNT_BC:
            if ((m_status = m_alti->CNT_BC_Read(rdata)) != 0) {
                CERR("ALTI: reading BC counter", "");
                return (m_status);
            }
            break;
        case CNT_ORB:
            if ((m_status = m_alti->CNT_ORB_Read(rdata)) != 0) {
                CERR("ALTI: reading ORB counter", "");
                return (m_status);
            }
            break;
        case CNT_L1A:
            if ((m_status = m_alti->CNT_L1A_Read(rdata)) != 0) {
                CERR("ALTI: reading L1A counter", "");
                return (m_status);
            }
            break;
        case CNT_TTR1:
            if ((m_status = m_alti->CNT_TTR1_Read(rdata)) != 0) {
                CERR("ALTI: reading TTR1 counter", "");
                return (m_status);
            }
            break;
        case CNT_TTR2:
            if ((m_status = m_alti->CNT_TTR2_Read(rdata)) != 0) {
                CERR("ALTI: reading TTR2 counter", "");
                return (m_status);
            }
            break;
        case CNT_TTR3:
            if ((m_status = m_alti->CNT_TTR3_Read(rdata)) != 0) {
                CERR("ALTI: reading TTR3 counter", "");
                return (m_status);
            }
            break;
        case CNT_BGO0:
            if ((m_status = m_alti->CNT_BGO0_Read(rdata)) != 0) {
                CERR("ALTI: reading BGO0 counter", "");
                return (m_status);
            }
            break;
        case CNT_BGO1:
            if ((m_status = m_alti->CNT_BGO1_Read(rdata)) != 0) {
                CERR("ALTI: reading BGO1 counter", "");
                return (m_status);
            }
            break;
        case CNT_BGO2:
            if ((m_status = m_alti->CNT_BGO2_Read(rdata)) != 0) {
                CERR("ALTI: reading BGO2 counter", "");
                return (m_status);
            }
            break;
        case CNT_BGO3:
            if ((m_status = m_alti->CNT_BGO3_Read(rdata)) != 0) {
                CERR("ALTI: reading BGO3 counter", "");
                return (m_status);
            }
            break;
        default:
            break;
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTBcidMaskRead(const unsigned int bcid, bool &mask) {

    // check BCID
    CHECKRANGE0(bcid, BCID_NUMBER - 1, "CNT BCID number");

    unsigned int wrd(bcid/WORD_SIZE), off(bcid % WORD_SIZE), data;

    // read BCID mask
    if ((m_status = m_alti->CNT_CNT_BcidMask_Read(wrd, data)) != SUCCESS) {
        CERR("ALTI: reading CNT BCID mask %4d", bcid);
        return (m_status);
    }
    mask = data & (0x00000001 << off);

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTBcidMaskWrite(const unsigned int bcid, const bool mask) {

    // check BCID
    CHECKRANGE0(bcid, BCID_NUMBER - 1, "CNT BCID number");

    unsigned int wrd(bcid/WORD_SIZE), off(bcid % WORD_SIZE), data;

    // read BCID mask
    if ((m_status = m_alti->CNT_CNT_BcidMask_Read(wrd, data)) != SUCCESS) {
        CERR("ALTI: reading CNT BCID mask %4d", bcid);
        return (m_status);
    }

    // modify BGRP data
    data = mask ? (data | (0x00000001 << off)) : (data & (~(0x00000001 << off)));

    // write BCID mask
    if ((m_status = m_alti->CNT_CNT_BcidMask_Write(wrd, data)) != SUCCESS) {
        CERR("ALTI: writing CNT BCID mask %4d", bcid);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTBcidMaskRead(std::vector<bool> &mask) {

    // get data segment
    DataSegment dseg("AltiModule[BCID]", BCID_NUMBER);
    if (dseg() != SUCCESS) {
        CERR("ALTI: getting CMEM segment \"%s\"", dseg.name().c_str());
        return (FAILURE);
    }

    // read BCID mask using block transfer
    if ((m_status = m_alti->CNT_CNT_BcidMask_Read(dseg.segment())) != SUCCESS) {
        CERR("ALTI: reading CNT BCID mask using block transfer", "");
        return (m_status);
    }

    unsigned int wrd, bit, off, num(0);

    // reset BCID mask vector
    mask.clear();
    mask.resize(BCID_NUMBER);

    // copy BCID mask from CMEM segment to vector
    for (wrd = 0; wrd < BCID_WORD_NUMBER; wrd++) {
        off = 0x00000001;
        for (bit = 0; bit < WORD_SIZE; bit++) {
            mask[num] = (dseg[wrd] & off);
            off <<= 1;
            num++;
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTBcidMaskWrite(const std::vector<bool> &mask) {

    unsigned int wrd, bit, off, num(0);

    // get data segment
    DataSegment dseg("AltiModule[BCID]", BCID_NUMBER);
    if (dseg() != SUCCESS) {
        CERR("ALTI: getting CMEM segment \"%s\"", dseg.name().c_str());
        return (FAILURE);
    }

    // copy BCID mask from vector to CMEM segment
    for (wrd = 0; wrd < BCID_WORD_NUMBER; wrd++) {
        off = 0x00000001;
        dseg[wrd] = 0x00000000;
        for (bit = 0; bit < WORD_SIZE; bit++) {
            if (mask[num]) dseg[wrd] |= off;
            off <<= 1;
            num++;
        }
    }

    // write BCID mask using block transfer
    if ((m_status = m_alti->CNT_CNT_BcidMask_Write(dseg.segment())) != SUCCESS) {
        CERR("ALTI: writing CNT BCID mask using block transfer", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTBcidMaskRead(std::vector<unsigned int> &mask) {

    // get data segment
    DataSegment dseg("AltiModule[BCID]", BCID_WORD_NUMBER);
    if (dseg() != SUCCESS) {
        CERR("ALTI: getting CMEM segment \"%s\"", dseg.name().c_str());
        return (FAILURE);
    }

    // read BCID mask using block transfer
    if ((m_status = m_alti->CNT_CNT_BcidMask_Read(dseg.segment())) != SUCCESS) {
        CERR("ALTI: reading CNT BCID mask using block transfer", "");
        return (m_status);
    }

    unsigned int wrd, bit, off, num(0);

    // reset BCID mask vector
    mask.clear();
    mask.resize(BCID_NUMBER);

    // copy BCID mask from CMEM segment to vector
    for (wrd = 0; wrd < BCID_WORD_NUMBER; wrd++) {
        off = 0x00000001;
        for (bit = 0; bit < WORD_SIZE; bit++) {
            mask[num] = (dseg[wrd] & off) ? 0x00000001 : 0x00000000;
            off <<= 1;
            num++;
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTBcidMaskWrite(const std::vector<unsigned int> &mask) {

    unsigned int wrd, bit, off, num(0);

    // get data segment
    DataSegment dseg("AltiModule[BCID]", BCID_WORD_NUMBER);
    if (dseg() != SUCCESS) {
        CERR("ALTI: getting CMEM segment \"%s\"",dseg.name().c_str());
        return (FAILURE);
    }

    // copy BCID mask from vector to CMEM segment
    for (wrd = 0; wrd < BCID_WORD_NUMBER; wrd++) {
        off = 0x00000001;
        dseg[wrd] = 0x00000000;
        for (bit = 0; bit < WORD_SIZE; bit++) {
            if (mask[num] != 0x00000000) dseg[wrd] |= off;
            off <<= 1;
            num++;
        }
    }

    // write BCID mask using block transfer
    if ((m_status = m_alti->CNT_CNT_BcidMask_Write(dseg.segment())) != SUCCESS) {
        CERR("ALTI: writing CNT BCID mask using block transfer", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTBcidMaskRead(const std::string &fn) {

    // get data segment
    DataSegment dseg("AltiModule[BCID]", BCID_WORD_NUMBER);
    if (dseg() != SUCCESS) {
        CERR("ALTI: getting CMEM segment \"%s\"", dseg.name().c_str());
        return (FAILURE);
    }

    // read BCID mask using block transfer
    if ((m_status = m_alti->CNT_CNT_BcidMask_Read(dseg.segment())) != SUCCESS) {
        CERR("ALTI: reading CNT BCID mask using block transfer", "");
        return (m_status);
    }

    // copy data from CMEM segment into file
    if ((m_status = AltiCommon::writeFile(fn, BCID_WORD_NUMBER, dseg.data())) != AltiCommon::SUCCESS) {
        CERR("ALTI: writing data from CNT BCID mask to file \"%s\"", fn.c_str());
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTBcidMaskWrite(const std::string &fn) {

    // get data segment
    DataSegment dseg("AltiModule[BCID]", BCID_WORD_NUMBER);
    if (dseg() != SUCCESS) {
        CERR("ALTI: getting CMEM segment \"%s\"", dseg.name().c_str());
        return (FAILURE);
    }

    // copy data from file into CMEM segment
    if ((m_status = AltiCommon::readFile(fn, BCID_WORD_NUMBER, dseg.data())) != AltiCommon::SUCCESS) {
        CERR("ALTI: reading data from CNT BCID mask from file \"%s\"", fn.c_str());
        return (m_status);
    }

    // write BCID mask using block transfer
    if ((m_status = m_alti->CNT_CNT_BcidMask_Write(dseg.segment())) != SUCCESS) {
        CERR("ALTI: writing CNT BCID mask using block transfer", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTBcidOffsetRead(unsigned int &off) {

    // read BCID offset
    if ((m_status = m_alti->CNT_CNT_BcidOffset_Read(off)) != SUCCESS) {
        CERR("ALTI: reading CNT BCID offset", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTBcidOffsetWrite(const unsigned int off) {

    // write BCID offset
    if ((m_status = m_alti->CNT_CNT_BcidOffset_Write(off)) != SUCCESS) {
        CERR("ALTI: writing CNT BCID offset", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTControlPrint() {

    // print control
    ALTI_CNT_CNT_CONTROL_BITSTRING ctl;
    if ((m_status = m_alti->CNT_CNT_Control_Read(ctl)) != SUCCESS) {
        CERR("ALTI: reading CNT enable", "");
        return (m_status);
    }
    ctl.print();

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTEnableRead(bool &ena, bool &sta) {

    // read CNT enable
    ALTI_CNT_CNT_CONTROL_BITSTRING ctl;
    if ((m_status = m_alti->CNT_CNT_Control_Read(ctl)) != SUCCESS) {
        CERR("ALTI: reading CNT enable", "");
        return (m_status);
    }
    ena = ctl.EnableCounters();
    sta = ctl.EnableStatus();

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTEnableWrite(const bool ena) {

    // read-modify-write CNT enable
    ALTI_CNT_CNT_CONTROL_BITSTRING ctl;
    if ((m_status = m_alti->CNT_CNT_Control_Read(ctl)) != SUCCESS) {
        CERR("ALTI: reading CNT enable", "");
        return (m_status);
    }
    ctl.EnableCounters(ena);
    if ((m_status = m_alti->CNT_CNT_Control_Write(ctl)) != SUCCESS) {
        CERR("ALTI: writing CNT enable", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTEnableWait(const bool ena) {

    unsigned int itim(0);

    while (true) {
        ALTI_CNT_CNT_CONTROL_BITSTRING ctl;
        if ((m_status = m_alti->CNT_CNT_Control_Read(ctl)) != SUCCESS) {
            CERR("ALTI: reading CNT enable", "");
            return (m_status);
        }
        if (ena != ctl.EnableCounters()) {
            CERR("ALTI: waiting for CNT enable to be \"%s\" while it is \"%s\"", ena ? "ENABLED" : "DISABLED", ctl.EnableCounters() ? "ENABLED" : "DISABLED");
            ctl.print();
            return (WRONGPAR);
        }
        if(ctl.EnableStatus() == ena) return (m_status);
        if ((++itim) == CNT_CONTROL_TIMEOUT) {
            CERR("ALTI: timeout waiting for CNT enable status to be \"%s\":", ena ? "ACTIVE" : "NOT_ACTIVE");
            return (TIMEOUT);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTClearWrite() {

    // read-modify-write CNT clear
    ALTI_CNT_CNT_CONTROL_BITSTRING ctl;
    if ((m_status = m_alti->CNT_CNT_Control_Read(ctl)) != SUCCESS) {
        CERR("ALTI: reading CNT clear", "");
        return (m_status);
    }
    ctl.ClearCounters(true);
    if ((m_status = m_alti->CNT_CNT_Control_Write(ctl)) != SUCCESS) {
        CERR("ALTI: writing CNT clear", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTClearWait() {

    unsigned int itim(0);

    while (true) {
        ALTI_CNT_CNT_CONTROL_BITSTRING ctl;
        if ((m_status = m_alti->CNT_CNT_Control_Read(ctl)) != SUCCESS) {
            CERR("ALTI: reading CNT clear", "");
            return (m_status);
        }
        if (!ctl.ClearBusy()) return (m_status);
        if ((++itim) == CNT_CONTROL_TIMEOUT) {
            CERR("ALTI: timeout waiting for CNT clear to finish", "");
            return (TIMEOUT);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTCopyWrite() {

    // read-modify-write CNT copy
    ALTI_CNT_CNT_CONTROL_BITSTRING ctl;
    if ((m_status = m_alti->CNT_CNT_Control_Read(ctl)) != SUCCESS) {
        CERR("ALTI: reading CNT copy", "");
        return (m_status);
    }
    ctl.CopyCounters(true);
    if ((m_status = m_alti->CNT_CNT_Control_Write(ctl)) != SUCCESS) {
        CERR("ALTI: writing CNT copy", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTCopyWait() {

    unsigned int itim(0);

    while (true) {
        ALTI_CNT_CNT_CONTROL_BITSTRING ctl;
        if ((m_status = m_alti->CNT_CNT_Control_Read(ctl)) != SUCCESS) {
            CERR("ALTI: reading CNT copy", "");
            return (m_status);
        }
        if (!ctl.CopyBusy()) return (m_status);
        if ((++itim) == CNT_CONTROL_TIMEOUT) {
            CERR("ALTI: timeout waiting for CNT copy to finish", "");
            return (TIMEOUT);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTCopyClearWrite() {

    // read-modify-write CNT copy
    ALTI_CNT_CNT_CONTROL_BITSTRING ctl;
    if ((m_status = m_alti->CNT_CNT_Control_Read(ctl)) != SUCCESS) {
        CERR("ALTI: reading CNT copy/clear", "");
        return (m_status);
    }
    ctl.CopyCounters(true);
    ctl.ClearCounters(true);
    if ((m_status = m_alti->CNT_CNT_Control_Write(ctl)) != SUCCESS) {
        CERR("ALTI: writing CNT copy/clear", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTCopyClearWait() {

    unsigned int itim(0);

    while (true) {
        ALTI_CNT_CNT_CONTROL_BITSTRING ctl;
        if ((m_status = m_alti->CNT_CNT_Control_Read(ctl)) != SUCCESS) {
            CERR("ALTI: reading CNT copy/clear", "");
            return (m_status);
        }
        if (!ctl.CopyBusy() && !ctl.ClearBusy()) return (m_status);
        if ((++itim) == CNT_CONTROL_TIMEOUT) {
            CERR("ALTI: timeout waiting for CNT copy/clear to finish", "");
            return (TIMEOUT);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTRequestPrint() {

    // print request
    ALTI_CNT_CNT_REQUEST_BITSTRING ctl;
    if ((m_status = m_alti->CNT_CNT_Request_Read(ctl)) != SUCCESS) {
        CERR("ALTI: reading CNT request", "");
        return (m_status);
    }
    ctl.print();

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTRequestWrite(const bool cnt, const bool trn, const bool tst) {

    // print request
    ALTI_CNT_CNT_REQUEST_BITSTRING ctl;
    ctl.EnableCountOverflow(cnt ? "ENABLED" : "DISABLED");
    ctl.EnableTurnOverflow(trn ? "ENABLED" : "DISABLED");
    ctl.AssertRequest(tst ? "ENABLED" : "DISABLED");
    if ((m_status = m_alti->CNT_CNT_Request_Write(ctl)) != SUCCESS) {
        CERR("ALTI: writing CNT request", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTTtypLutRead(std::vector<unsigned int> &lut) {

    // get data segment
    DataSegment dseg("AltiModule[CNT_TTYP_LUT]", CNT_TTYP_LUT_WORD_NUMBER);
    if (dseg() != SUCCESS) {
        CERR("ALTI: getting CMEM segment \"%s\"", dseg.name().c_str());
        return (FAILURE);
    }

    // read TTYP counter LUT using block transfer
    if ((m_status = m_alti->CNT_CountLutTtyp_Read(dseg.segment())) != SUCCESS) {
        CERR("ALTI: reading CNT TTYP LUT using block transfer", "");
        return (m_status);
    }

    unsigned int wrd;

    // reset TTYP counter LUT
    lut.clear();
    lut.resize(CNT_TTYP_LUT_WORD_NUMBER);

    // copy TTYP counter LUT from CMEM segment to vector
    for (wrd = 0; wrd < CNT_TTYP_LUT_WORD_NUMBER; wrd++) {
        lut[wrd] = dseg[wrd];
    }
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTTtypLutWrite(const std::vector<unsigned int> &lut) {

    unsigned int wrd;

    // get data segment
    DataSegment dseg("AltiModule[CNT_TTYP_LUT]", CNT_TTYP_LUT_WORD_NUMBER);
    if (dseg() != SUCCESS) {
        CERR("ALTI: getting CMEM segment \"%s\"", dseg.name().c_str());
        return (FAILURE);
    }

    // copy TTYP counter LUT from vector to CMEM segment
    for (wrd = 0; wrd < CNT_TTYP_LUT_WORD_NUMBER; wrd++) {
        dseg[wrd] = lut[wrd];
    }

    // write TTYP counter LUT using block transfer
    if ((m_status = m_alti->CNT_CountLutTtyp_Write(dseg.segment())) != SUCCESS) {
        CERR("ALTI: writing CNT TTYP LUT using block transfer", "");
        return (m_status);
    }
    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTTtypLutRead(const std::string &fn) {

    // get data segment
    DataSegment dseg("AltiModule[CNT_TTYP_LUT]", CNT_TTYP_LUT_WORD_NUMBER);
    if (dseg() != SUCCESS) {
        CERR("ALTI: getting CMEM segment \"%s\"", dseg.name().c_str());
        return (FAILURE);
    }

    // read TTYP counter LUT using block transfer
    if ((m_status = m_alti->CNT_CountLutTtyp_Read(dseg.segment())) != SUCCESS) {
        CERR("ALTI: reading CNT TTYP LUT using block transfer", "");
        return (m_status);
    }

    // copy data from CMEM segment into file
    if ((m_status = AltiCommon::writeFile(fn, CNT_TTYP_LUT_WORD_NUMBER, dseg.data())) != AltiCommon::SUCCESS) {
        CERR("ALTI: writing data from CNT TTYP LUT to file \"%s\"", fn.c_str());
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTTtypLutWrite(const std::string &fn) {

    // get data segment
    DataSegment dseg("AltiModule[CNT_TTYP_LUT]", CNT_TTYP_LUT_WORD_NUMBER);
    if (dseg() != SUCCESS) {
        CERR("ALTI: getting CMEM segment \"%s\"", dseg.name().c_str());
        return (FAILURE);
    }

    // copy data from file into CMEM segment
    if ((m_status = AltiCommon::readFile(fn, CNT_TTYP_LUT_WORD_NUMBER, dseg.data())) != AltiCommon::SUCCESS) {
        CERR("ALTI: reading data from CNT TTYP LUT from file \"%s\"", fn.c_str());
        return (m_status);
    }

    // write BCID mask using block transfer
    if ((m_status = m_alti->CNT_CountLutTtyp_Write(dseg.segment())) != SUCCESS) {
        CERR("ALTI: writing CNT TTYP LUT using block transfer", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTTtypLutPrint() {
    std::vector<unsigned int> lut;
    if ((m_status = CNTTtypLutRead(lut)) != SUCCESS) {
        CERR("ALTI: reading CNT TTYP LUT", "");
    }

    unsigned int i, mask, num_enabled, ttyp;
    for (i = 0; i < CNT_TTYP_NUMBER; i++) {
        mask = (0x00000001 << i);
        num_enabled = 0;
        std::printf("TTYP counter bucket [%d]: ", i);
        for (ttyp = 0x00; ttyp <= 0xff; ttyp++) {
            if (mask & lut[ttyp]) {
                std::printf("%s0x%02x", (num_enabled == 0) ? "" : ", ", ttyp);
                num_enabled++;
            }
        }
        std::printf("\n");
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTTurnCountRead(unsigned int &trn) {

    // read TURN counter
    if ((m_status = m_alti->CNT_CNT_TurnCount_Read(trn)) != SUCCESS) {
        CERR("ALTI: reading CNT TURN counter", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CNTCounterRead(AltiCounter *cnt) {

    if (cnt->getCmemSegment()) {
        if ((m_status = m_alti->CNT_CNT_CounterValue_Read(cnt->getCmemSegment(), cnt->dsize())) != SUCCESS) {
            CERR("ALTI: reading CNT using block transfer", "");
            return (m_status);
        }
    }
    else {
        unsigned int addr(CNT_ADDR), data;
        for (unsigned int i = 0; i < cnt->dsize(); i++) {
            if ((m_status = m_vmm->ReadSafe(addr, &data)) != VME_SUCCESS) {
                CERR("ALTI: reading CNT word %3d", i);
                return (m_status);
            }
            (*cnt)[i] = data;
            addr += sizeof(unsigned int);
        }
    }

    // read TURN counter
    unsigned int turn;
    if ((m_status = CNTTurnCountRead(turn)) != SUCCESS) {
        CERR("ALTI: reading CNT TURN counter", "");
        return (m_status);
    }
    cnt->turn(turn);

    return (m_status);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::FMWControlEnableBoot(bool checkIDOnly, bool verifyOnly) {
    ALTI_FMW_SPIFLASHBOOTCONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->FMW_SPIFlashBootControl_Read(bs)) != SUCCESS) {
        CERR("ALTI: reading SPI FLASH QuickBoot control register", "");
        return (m_status);
    }
    // MODIFY
    bs.CheckIDOnly(checkIDOnly ? "Enabled" : "Disabled");
    bs.VerifyOnly(verifyOnly ? "Enabled" : "Disabled");
    bs.ResetOrEnableBoot("EnabledBoot");
    // WRITE
    if ((m_status = m_alti->FMW_SPIFlashBootControl_Write(bs)) != SUCCESS) {
        CERR("ALTI: writing SPI FLASH QuickBoot control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::FMWControlReset() {
    ALTI_FMW_SPIFLASHBOOTCONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->FMW_SPIFlashBootControl_Read(bs)) != SUCCESS) {
        CERR("ALTI: reading SPI FLASH QuickBoot control register", "");
        return (m_status);
    }
    // MODIFY
    bs.FIFOWriteAddressReset("Enabled");
    bs.FIFOReadAddressReset("Enabled");
    bs.ResetOrEnableBoot("Reset");
    // WRITE
    if ((m_status = m_alti->FMW_SPIFlashBootControl_Write(bs)) != SUCCESS) {
        CERR("ALTI: writing SPI FLASH QuickBoot control register", "");
        return (m_status);
    }
    // MODIFY
    bs.FIFOWriteAddressReset("Disabled");
    bs.FIFOReadAddressReset("Disabled");
    // WRITE
    if ((m_status = m_alti->FMW_SPIFlashBootControl_Write(bs)) != SUCCESS) {
        CERR("ALTI: writing SPI FLASH QuickBoot control register", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::FMWFIFOStatusRead(bool &full, bool &almostFull, bool &empty, bool &almostEmpty) {
    ALTI_FMW_SPIFLASHBOOTSTATUS_BITSTRING bs;

    // READ
    if ((m_status = m_alti->FMW_SPIFlashBootStatus_Read(bs)) != SUCCESS) {
        CERR("ALTI: reading SPI FLASH QuickBoot status register", "");
        return (m_status);
    }
    full = ((bs.FIFOFull() == "Full") ? true : false);
    almostFull = ((bs.FIFOAlmostFull() == "AlmostFull") ? true : false);
    empty = ((bs.FIFOEmpty() == "Empty") ? true : false);
    almostEmpty = ((bs.FIFOAlmostEmpty() == "AlmostEmpty") ? true : false);

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::FMWWaitUntilDone(bool &error) {
    ALTI_FMW_SPIFLASHBOOTSTATUS_BITSTRING bs;

    // POLL
    while (true) {
        // READ
        if ((m_status = m_alti->FMW_SPIFlashBootStatus_Read(bs)) != SUCCESS) {
            CERR("ALTI: reading SPI FLASH QuickBoot status register", "");
            return (m_status);
        }
        if (bs.DoneOrErrorDetected()) break;
        //std::printf("DoneOrErrorDetected: %d\n", bs.DoneOrErrorDetected());
        usleep(1000);
    }
    error = bs.ErrorDetected();

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::FMWReceivedDataRead(unsigned int &data) {
    ALTI_FMW_SPIFLASHBOOTSTATUS_BITSTRING bs;

    // READ
    if ((m_status = m_alti->FMW_SPIFlashBootStatus_Read(bs)) != SUCCESS) {
        CERR("ALTI: reading SPI FLASH QuickBoot status register", "");
        return (m_status);
    }
    data = bs.SSDReceivedData();

    return (m_status);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::MONVoltageReadLTC2991(std::vector<float> &voltages) {
    // trigger the measurement
    if ((m_status = m_net->write(m_LTC2991MonitorStatusHi, (unsigned int) 0xf0)) != 0) {
        CERR("ALTI: triggering voltage monitor measurement", "");
        return (m_status);
    }
    // wait for the measurement to finish
    unsigned int status;
    while (1) {
        if ((m_status = m_net->read(m_LTC2991MonitorStatusHi, status)) != 0 ) {
            CERR("ALTI: reading voltage monitor status", "");
            return (m_status);
        }
        if (!(status & 0x00000004)) { // BUSY = 0, conversion done
            break;
        }
    }
    // read measurement results;
    if ((m_status = m_net->read(m_LTC2991MonitorV1, voltages[0])) != 0) {
        CERR("ALTI: reading V1 from the voltage monitor", "");
        return (m_status);
    }
    if ((m_status = m_net->read(m_LTC2991MonitorV2, voltages[1])) != 0) {
        CERR("ALTI: reading V2 from the voltage monitor", "");
        return (m_status);
    }
    if ((m_status = m_net->read(m_LTC2991MonitorV3, voltages[2])) != 0) {
        CERR("ALTI: reading V3 from the voltage monitor", "");
        return (m_status);
    }
    if ((m_status = m_net->read(m_LTC2991MonitorV4, voltages[3])) != 0) {
        CERR("ALTI: reading V4 from the voltage monitor", "");
        return (m_status);
    }
    if ((m_status = m_net->read(m_LTC2991MonitorV5, voltages[4])) != 0) {
        CERR("ALTI: reading V5 from the voltage monitor", "");
        return (m_status);
    }
    if ((m_status = m_net->read(m_LTC2991MonitorV6, voltages[5])) != 0) {
        CERR("ALTI: reading V6 from the voltage monitor", "");
        return (m_status);
    }
    if ((m_status = m_net->read(m_LTC2991MonitorV7, voltages[6])) != 0) {
        CERR("ALTI: reading V7 from the voltage monitor", "");
        return (m_status);
    }
    if ((m_status = m_net->read(m_LTC2991MonitorV8, voltages[7])) != 0) {
        CERR("ALTI: reading V8 from the voltage monitor", "");
        return (m_status);
    }
    // neutralizing DV (data valid) bit from the voltages that were read
    unsigned int i;
    for (i = 0; i < 8; i++) voltages[i] -= (0.00030518*(1 << 15)); // ~ 10

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::MONInternalTempReadLTC2991(float &temperature) {
    // trigger the measurement
    if ((m_status = m_net->write(m_LTC2991MonitorStatusHi, (unsigned int) 0x08)) != 0) {
        CERR("ALTI: triggering LTC2991 internal temperature measurement", "");
        return (m_status);
    }
    // wait for the measurement to finish
    unsigned int status;
    while (1) {
        if ((m_status = m_net->read(m_LTC2991MonitorStatusHi, status)) != 0 ) {
            CERR("ALTI: reading LTC2991 internal temperature status", "");
            return (m_status);
        }
        //printf("status: 0x%08x\n", status);
        if (status & 0x00000002) { // T_INTERNAL = 1, register contains new data
            break;
        }
    }
    // read measurement results;
    if ((m_status = m_net->read(m_LTC2991MonitorTint, temperature)) != 0) {
        CERR("ALTI: reading internal temperature from the LTC2991 monitor", "");
        return (m_status);
    }
    // neutralizing DV (data valid) bit from the temperature that was read
    temperature -= (0.0625*(1 << 15)); // = 2048

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::MONVccVoltageReadLTC2991(float &voltage) {
    // trigger the measurement
    if ((m_status = m_net->write(m_LTC2991MonitorStatusHi, (unsigned int) 0x08)) != 0) {
        CERR("ALTI: triggering LTC2991 internal supply voltage measurement", "");
        return (m_status);
    }
    // wait for the measurement to finish
    unsigned int status;
    while (1) {
        if ((m_status = m_net->read(m_LTC2991MonitorStatusHi, status)) != 0 ) {
            CERR("ALTI: reading LTC2991 internal supply voltage status", "");
            return (m_status);
        }
        //printf("status: 0x%08x\n", status);
        if (status & 0x00000001) { // V_CC = 1, register contains new data
            break;
        }
    }
    // wait for DV (data valid) bit to set
    unsigned int voltage_raw;
    while (1) {
        if ((m_status = m_net->read(m_LTC2991MonitorVcc, voltage_raw)) != 0 ) {
            CERR("ALTI: reading internal raw supply voltage status", "");
            return (m_status);
        }
        //printf("status: 0x%08x\n", status);
        if (voltage_raw & (1 << 15)) { // DV = 1, it will clear on next read
            break;
        }
    }
    // read measurement results;
    if ((m_status = m_net->read(m_LTC2991MonitorVccFloat, voltage)) != 0) {
        CERR("ALTI: reading internal supply voltage from the LTC2991 monitor", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::MONLocalTempReadMAX1617(float &temperature) {
    /*// trigger the measurement
    if ((m_status = m_net->write(m_MAX1617Command, (unsigned int) 0x01)) != 0) {
        CERR("ALTI: triggering MAX1617 local temperature measurement", "");
        return (m_status);
    }*/
    // read measurement results;
    if ((m_status = m_net->read(m_MAX1617LocalT, temperature)) != 0) {
        CERR("ALTI: reading local temperature from the MAX1617 monitor", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::MONLocalTempReadMAX6695(float &temperature) {
    /*// trigger the measurement
    if ((m_status = m_net->write(m_MAX6695Command, (unsigned int) 0x01)) != 0) {
        CERR("ALTI: triggering MAX6695 local temperature measurement", "");
        return (m_status);
    }*/
    // read measurement results;
    if ((m_status = m_net->read(m_MAX6695LocalT, temperature)) != 0) {
        CERR("ALTI: reading local temperature from the MAX6695 monitor", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::MONLocalTempRead(float &temperature) {
    return (((m_identifier.back() >= '1') && (m_identifier.back() <= '7')) ? MONLocalTempReadMAX1617(temperature) : MONLocalTempReadMAX6695(temperature));
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::CLKSetup() {
    ALTI_CLK_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CLK_Control_Read(bs)) != 0) {
        CERR("ALTI: reading clock control register", "");
        return (m_status);
    }
    // MODIFY
    bs.FPGA_Sel(1); // 1/0? Check with Vladimir
    // WRITE
    if ((m_status = m_alti->CLK_Control_Write(bs)) != 0) {
        CERR("ALTI: writing clock control register", "");

        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CLKCheck() {
    ALTI_CLK_CONTROL_BITSTRING bs;

    // READ
    if ((m_status = m_alti->CLK_Control_Read(bs)) != 0) {
        CERR("ALTI: reading clock control register", "");
        return (m_status);
    }
    return ((bs.FPGA_Sel() == 1) ? (SUCCESS) : (FAILURE)); // 1/0? Check with Vladimir

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CInit() {

    // create I2C
    m_i2c = new I2C(I2C::ALTI, m_vmm);

    unsigned int b;
    m_net = new I2CNetwork();
    m_net->addI2CBus("BUS[0]", m_i2c, b);
    I2CNetwork::BUS_TYPE bus_type = ((m_identifier.back() >= '1') && (m_identifier.back() <= '7')) ? I2CNetwork::BUS_ALTI : I2CNetwork::BUS_ALTI2;
    m_net->initI2CBus(b, bus_type);

    /*std::ostringstream buf;
    bool flag(false);
    for(unsigned int i=0; i<I2CNetwork::BUS_TYPE_NUMBER; i++) {
        buf << (flag?", ":"") << i << "=" << I2CNetwork::BUS_TYPE_NAME[i];
        flag = true;
    }
    i2cmod = static_cast<I2CNetwork::BUS_TYPE>(enterInt(("I2C model ("+buf.str()+")").c_str(),0,I2CNetwork::BUS_TYPE_NUMBER-1));
    i2cnet->initI2CBus(b,i2cmod,daddr);*/

    unsigned int i, j;
    std::ostringstream name;

    // fetch switches
    SIGNAL signal;
    for (i = 0; i < SIGNAL_NUMBER_ROUTED; i++) {
        signal = SIGNAL(i);

        name.str("");
        name.clear();
        name << "BUS[0]/" << SIGNAL_NAME[signal] << "_SWITCH/CONFIG";
        m_net->getI2CNode(name.str(), m_SignalSwitchConfig[signal]);

        name.str("");
        name.clear();
        name << "BUS[0]/" << SIGNAL_NAME[signal] << "_SWITCH/CTRL";
        m_net->getI2CNode(name.str(), m_SignalSwitchCtrl[signal]);
    }

    // fetch equalizers
    SIGNAL_SOURCE src;
    EQUALIZER_VOLTAGE vlt;
    for (i = 0; i < SIGNAL_SOURCE_NUMBER - 3; i++) {
        src = SIGNAL_SOURCE(i);
        for (j = 0; j < EQUALIZER_VOLTAGE_NUMBER; j++) {
           vlt = EQUALIZER_VOLTAGE(j);

           name.str("");
           name.clear();
           name << "BUS[0]/" << SIGNAL_SOURCE_NAME[src] << "/" << ((vlt == V_PEAK) ? ("DACA") : ((vlt == V_POLE) ? ("DACB") : ((vlt == V_GAIN) ? ("DACC") : ("DACD"))));
           m_net->getI2CNode(name.str(), m_EqualizerConfig[src][vlt]);
        }
    }

    // fetch LTC2991
    m_net->getI2CNode("BUS[0]/LTC2991/STATUS_LO", m_LTC2991MonitorStatusLo);
    m_net->getI2CNode("BUS[0]/LTC2991/STATUS_HI", m_LTC2991MonitorStatusHi);
    m_net->getI2CNode("BUS[0]/LTC2991/CONTROL_V1V2_V3V4", m_LTC2991MonitorControlV1234);
    m_net->getI2CNode("BUS[0]/LTC2991/CONTROL_V5V6_V7V8", m_LTC2991MonitorControlV5678);
    m_net->getI2CNode("BUS[0]/LTC2991/V1", m_LTC2991MonitorV1);
    m_net->getI2CNode("BUS[0]/LTC2991/V2", m_LTC2991MonitorV2);
    m_net->getI2CNode("BUS[0]/LTC2991/V3", m_LTC2991MonitorV3);
    m_net->getI2CNode("BUS[0]/LTC2991/V4", m_LTC2991MonitorV4);
    m_net->getI2CNode("BUS[0]/LTC2991/V5", m_LTC2991MonitorV5);
    m_net->getI2CNode("BUS[0]/LTC2991/V6", m_LTC2991MonitorV6);
    m_net->getI2CNode("BUS[0]/LTC2991/V7", m_LTC2991MonitorV7);
    m_net->getI2CNode("BUS[0]/LTC2991/V8", m_LTC2991MonitorV8);
    m_net->getI2CNode("BUS[0]/LTC2991/TINT", m_LTC2991MonitorTint);
    m_net->getI2CNode("BUS[0]/LTC2991/VCC_FLOAT", m_LTC2991MonitorVccFloat);
    m_net->getI2CNode("BUS[0]/LTC2991/VCC", m_LTC2991MonitorVcc);

    switch (bus_type) {
        case I2CNetwork::BUS_ALTI:
            // fetch MAX1617
            m_net->getI2CNode("BUS[0]/MAX1617/RLTS_LOCAL_TEMP", m_MAX1617LocalT);
            m_net->getI2CNode("BUS[0]/MAX1617/OSHT_ONE_SHOT_COMMAND", m_MAX1617Command);
            break;
        case I2CNetwork::BUS_ALTI2:
            // fetch MAX6695
            m_net->getI2CNode("BUS[0]/MAX6695/RLTS_LOCAL_TEMP", m_MAX6695LocalT);
            m_net->getI2CNode("BUS[0]/MAX6695/OSHT_ONE_SHOT_COMMAND", m_MAX6695Command);
            break;
        default:
            break;
    }

    // fetch SI5344B
    m_net->getI2CNode("BUS[0]/CLK_GEN_CLEANER_P0/STATUS_FLG", m_SI5344BStatusFlg);
    m_net->getI2CNode("BUS[0]/CLK_GEN_CLEANER_P0/OOF_LOS_FLG", m_SI5344BOofLosFlg);
    m_net->getI2CNode("BUS[0]/CLK_GEN_CLEANER_P0/HOLD_LOL_FLG", m_SI5344BHoldLolFlg);
    m_net->getI2CNode("BUS[0]/CLK_GEN_CLEANER_P0/PLL_CAL_FLG", m_SI5344BPllCalFlg);
    m_net->getI2CNode("BUS[0]/CLK_GEN_CLEANER_P0/STATUS_STICKY_FLG", m_SI5344BStatusStickyFlg);
    m_net->getI2CNode("BUS[0]/CLK_GEN_CLEANER_P0/OOF_LOS_STICKY_FLG", m_SI5344BOofLosStickyFlg);
    m_net->getI2CNode("BUS[0]/CLK_GEN_CLEANER_P0/HOLD_LOL_STICKY_FLG", m_SI5344BHoldLolStickyFlg);
    m_net->getI2CNode("BUS[0]/CLK_GEN_CLEANER_P0/PLL_CAL_STICKY_FLG", m_SI5344BPllCalStickyFlg);
    m_net->getI2CNode("BUS[0]/CLK_GEN_CLEANER_P0/STATUS_INTR_MSK", m_SI5344BStatusIntrMsk);
    m_net->getI2CNode("BUS[0]/CLK_GEN_CLEANER_P0/OOF_LOS_INTR_MSK", m_SI5344BOofLosIntrMsk);
    m_net->getI2CNode("BUS[0]/CLK_GEN_CLEANER_P0/HOLD_LOL_INTR_MSK", m_SI5344BHoldLolIntrMsk);
    m_net->getI2CNode("BUS[0]/CLK_GEN_CLEANER_P0/PLL_CAL_INTR_MSK", m_SI5344BPllCalIntrMsk);
    m_net->getI2CNode("BUS[0]/CLK_GEN_CLEANER_P0/PDN_HARD_RST_SYNC", m_SI5344BPdnHardRstSync);
    m_net->getI2CNode("BUS[0]/CLK_GEN_CLEANER_P2/DESIGN_ID", m_SI5344BDesignId);

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CCheck() {
    unsigned int prescale;
    bool enabled;

    // check proper prescaler value for the I2C core
    if ((m_status = m_i2c->PrescaleRead(prescale)) != SUCCESS) {
        CERR("ALTI: reading prescale value", "");
        return(m_status);
    }
    if (prescale != I2C_PRESCALE_VALUE) {
        CERR("ALTI: wrong I2C prescale value", "");
        return(FAILURE);
    }

    // check that I2C core is enabled
    if ((m_status = m_i2c->EnableRead(enabled)) != SUCCESS) {
        CERR("ALTI: reading I2C enable value", "");
        return(m_status);
    }
    if (!enabled) {
        CERR("ALTI: I2C not enabled", "");
        return(FAILURE);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCCheck() {
    unsigned int ttyp_delay;

    // check the TTYP delay with respect to the L1A
    if ((m_status = ENCTriggerWordDelayRead(ttyp_delay)) != SUCCESS) {
        CERR("ALTI: reading the TTYP word delay", "");
        return(m_status);
    }
    if (ttyp_delay != TTYP_DELAY_FOR_LATCHING) {
        CERR("ALTI: wrong TTYP delay value", "");
        return(FAILURE);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::JCSetup() {

    // set up jitter cleaner: default config
    if ((m_status = CLKJitterCleanerConfigWrite()) != SUCCESS) {
        CERR("ALTI: setting up jitter cleaner, default config", "");
        return (m_status);
    }
    usleep(1000000); // sleep 1s
    // clear status flags
    if ((m_status = CLKJitterCleanerStatusClear()) != SUCCESS) {
        CERR("ALTI: clearing jitter cleaner status flags", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::CSRSetup() {
    // set ADER1 page register to a nonzero value, so that ALTI uses 8MB instead of 16MB
    if ((m_status = CSRADER1Write(ALTI_RAM_MEMORY::PATTERN_SNAPSHOT + 1)) != SUCCESS) { // either of the 3 memories
        CERR("ALTI: setting up ADER1 page register in CR/CSR, \"%s\" memory", ALTI_RAM_MEMORY_NAME[TTC_DECODER_COMMAND].c_str());
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ENCSetup() {
    // set TTYP delay with respect to the L1A to 2BCs
    if ((m_status = ENCTriggerWordDelayWrite(TTYP_DELAY_FOR_LATCHING)) != SUCCESS) {
        CERR("ALTI: setting the TTYP word delay to %d BCs", TTYP_DELAY_FOR_LATCHING);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

unsigned int AltiModule::swapEndian(const unsigned int x) {
    return ((x & 0xff000000) >> 24) | ((x & 0x00ff0000) >> 8) | ((x & 0x0000ff00) << 8) | ((x & 0x000000ff) << 24);
}

//------------------------------------------------------------------------------

unsigned int AltiModule::swapEndian(const unsigned int x, const unsigned int n) {
    if (n == 8) return swapEndian(x);
    unsigned int y = x, m = (n > 16) ? 16 : n;
    unsigned int mask_lsb, mask_msb, mask_rest;
    mask_lsb = (1 << m) - 1;
    mask_msb = mask_lsb << (32 - m);
    mask_rest = ~(mask_lsb + mask_msb);
    //std::printf("\n");
    //std::printf("input    : 0x%08x\n", x);
    //std::printf("mask_lsb : 0x%08x\n", mask_lsb);
    //std::printf("mask_msb : 0x%08x\n", mask_msb);
    //std::printf("mask_rest: 0x%08x\n", mask_rest);
    y = (mask_msb & (y << (32 - m))) | (mask_rest & y) | (mask_lsb & (y >> (32 - m)));
    //std::printf("output   : 0x%08x\n", y);
    //std::printf("\n");
    return y;
}

//------------------------------------------------------------------------------

unsigned int AltiModule::pollSlots(std::vector<unsigned int> &slots) {
    unsigned int n = 0, i, boardid;
    slots.clear();

    // open VME
    RCD::VME *vme = RCD::VME::Open();

    for (i = 1; i <= 21; i++) {
        // read BOARD ID
        if (vme->ReadCRCSR(i, BOARD_ID_BASE, &boardid) == 0) {
            if (boardid == BOARD_ID) { // found ALTI
                slots.push_back(i);
                n++;
            }
        }
    }

    // close VME
    RCD::VME::Close();

    return n;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::ECRGenerationRead(ECR_GENERATION &ecr) {

    ALTI_CTP_ECR_CONTROL_BITSTRING ctl;

    // read ECR type and length
    if ((m_status = m_alti->CTP_ECR_Control_Read(ctl)) != SUCCESS) {
        CERR("ALTI: reading ECR control", "");
        return (m_status);
    }
    ecr.type = (ctl.Source() == "VME") ? ECR_VME : ECR_INTERNAL;
    ecr.length = ctl.Length() + 1;

    // read ECR frequency
    if ((m_status = m_alti->CTP_ECR_Frequency_Read(ecr.frequency)) != SUCCESS) {
        CERR("ALTI: reading ECR frequency", "");
        return (m_status);
    }

    // read ECR BUSY before
    if ((m_status = m_alti->CTP_ECR_BusyBefore_Read(ecr.busy_before)) != SUCCESS) {
        CERR("ALTI: reading ECR BUSY BEFORE length", "");
        return (m_status);
    }

    // read ECR BUSY after
    if ((m_status = m_alti->CTP_ECR_BusyAfter_Read(ecr.busy_after)) != SUCCESS) {
        CERR("ALTI: reading ECR BUSY AFTER length", "");
        return (m_status);
    }

    // read ECR OBRIT offset
    if ((m_status = m_alti->CTP_ECR_OrbitOffset_Read(ecr.orbit_offset)) != SUCCESS) {
        CERR("ALTI: reading ECR ORBIT offset", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ECRGenerationWrite(const ECR_GENERATION &ecr) {

    ALTI_CTP_ECR_CONTROL_BITSTRING ctl;

    // write ECR type and length
    ctl.Source(ecr.type == ECR_VME ? "VME" : "INTERNAL");
    ctl.Length(ecr.length - 1);
    if ((m_status = m_alti->CTP_ECR_Control_Write(ctl)) != SUCCESS) {
        CERR("ALTI: writing ECR control", "");
        return (m_status);
    }

    // write ECR frequency
    if ((m_status = m_alti->CTP_ECR_Frequency_Write(ecr.frequency)) != SUCCESS) {
        CERR("ALTI: writing ECR frequency", "");
        return (m_status);
    }

    // write ECR BUSY before
    if ((m_status = m_alti->CTP_ECR_BusyBefore_Write(ecr.busy_before)) != SUCCESS) {
        CERR("ALTI: writing ECR BUSY BEFORE length", "");
        return (m_status);
    }

    // write ECR BUSY after
    if ((m_status = m_alti->CTP_ECR_BusyAfter_Write(ecr.busy_after)) != SUCCESS) {
        CERR("ALTI: writing ECR BUSY AFETR length", "");
        return (m_status);
    }

    // write ECR OBRIT offset
    if ((m_status = m_alti->CTP_ECR_OrbitOffset_Write(ecr.orbit_offset)) != SUCCESS) {
        CERR("ALTI: writing ECR ORBIT offset", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ECRGenerate() {

    unsigned int data(0x00000001);

    // write ECR generate
    if ((m_status = m_alti->CTP_ECR_Generate_Write(data)) != SUCCESS) {
        CERR("ALTI: writing ECR generate", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::ECRReadyWait() {

    ALTI_CTP_ECR_BUSYCONTROL_BITSTRING ctl;

    // read secondary partition busy control
    if ((m_status = m_alti->CTP_ECR_BusyControl_Read(ctl)) != SUCCESS) {
        CERR("ALTI: reading BSY control", "");
        return (m_status);
    }
    if (ctl.TestEcr()) {
        CERR("ALTI: waiting for ECR BUSY while in \"TEST\"", "");
        return (WRONGPAR);
    }
    if (!ctl.MaskEcr()) {
        CERR("ALTI: waiting for ECR BUSY while \"DISABLED\"", "");
        return (WRONGPAR);
    }
    if (!ctl.BusyEcr()) return (m_status);

    unsigned int itim(0);

    while (true) {

        // read secondary partition busy control
        if ((m_status = m_alti->CTP_ECR_BusyControl_Read(ctl)) != SUCCESS) {
            CERR("ALTI: reading BSY control", "");
            return (m_status);
        }

        // check if not empty
        if (!ctl.BusyEcr()) break;

        // check if timeout
        if (++itim >= BSY_CONTROL_TIMEOUT) {
            CERR("ALTI: timeout waiting for ECR BUSY", "");
            return (TIMEOUT);
        }
    }

    return (m_status);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::T2LControlReset(const bool l1a, const bool ecr, const bool fifo) {

    ALTI_CTP_T2L_CONTROL_BITSTRING bs;

    bs.L1aReset(l1a);
    bs.EcrReset(ecr);
    bs.FifoReset(fifo);

    // reset T2L control
    if ((m_status = m_alti->CTP_T2L_Control_Write(bs)) != SUCCESS) {
        CERR("ALTI: resetting T2L control", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::T2LEventIdentifierRead(EventIdentifier &evid) {

    unsigned int data;

    // read T2L event identifier
    if ((m_status = m_alti->CTP_T2L_L1idValue_Read(data)) != SUCCESS) {
        CERR("ALTI: reading T2L event identifier", "");
        return (m_status);
    }
    evid.data(data);

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::T2LFifoStatusRead(bool &emp, bool &ful, unsigned int &lvl) {

    ALTI_CTP_T2L_FIFOSTATUS_BITSTRING bs;

    // read T2L FIFO status
    if ((m_status = m_alti->CTP_T2L_FifoStatus_Read(bs)) != SUCCESS) {
        CERR("ALTI: reading T2L FIFO status", "");
        return (m_status);
    }
    emp = bs.Empty();
    ful = bs.Full();
    lvl = (ful) ? T2L_FIFO_SIZE : bs.Level();

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::T2LFifoWrite() {

    const unsigned int data(0x00000001);

    // write T2L FIFO
    if ((m_status = m_alti->CTP_T2L_FifoWrite_Write(data)) != SUCCESS) {
        CERR("ALTI: writing T2L FIFO", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::T2LFifoRead(unsigned int &l1id) {

    // read T2L FIFO
    if ((m_status = m_alti->CTP_T2L_L1idFifo_Read(l1id)) != SUCCESS) {
        CERR("ALTI: reading T2L FIFO", "");
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::T2LFifoRead(std::vector<unsigned int> &l1id) {

    unsigned int lvl;
    bool emp, ful;

    // reset L1ID vector
    l1id.clear();

    // read T2L FIFO status
    if ((m_status = T2LFifoStatusRead(emp, ful, lvl)) != SUCCESS) {
        CERR("ALTI: reading T2L FIFO status", "");
        return (m_status);
    }
    if (lvl == 0) return (m_status);

    unsigned int data;

    // read T2L FIFO
    for (unsigned int i = 0; i < lvl; i++) {
        if ((m_status = T2LFifoRead(data)) != SUCCESS) {
            CERR("ALTI: reading T2L FIFO", "");
            return (m_status);
        }
        l1id.push_back(data);
    }

    return (m_status);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int AltiModule::I2CSetup() {

    // set up I2C bus

    // write I2C prescale value
    if ((m_status = m_i2c->PrescaleWrite(I2C_PRESCALE_VALUE)) != I2C::SUCCESS) {
        CERR("ALTI: writing I2C prescale value", "");
        return(m_status);
    }

    // write I2C enable
    if ((m_status = m_i2c->EnableWrite(true)) != I2C::SUCCESS) {
        CERR("ALTI: writing enable", "");
        return(m_status);
    }
    //COUT("ALTI: INFO - I2C = \"SETUP\"", "");

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CDump() {

    // dump I2C network
    m_net->dump();

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CTransferTypeRead(const unsigned int bus, const unsigned int dev, const unsigned int qty, I2CInterface::TRANSFER_TYPE &ttyp) {

    // read I2C quantity's transfer type
    if ((m_status = m_net->transferType(bus, dev, qty, ttyp)) != I2C::SUCCESS) {
        CERR("ALTI: reading transfer type of I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CDisplayTypeRead(const unsigned int bus, const unsigned int dev, const unsigned int qty, I2CNetwork::DISPLAY_TYPE &dtyp) {

    // read I2C quantity's display type
    if ((m_status = m_net->displayType(bus, dev, qty, dtyp)) != I2C::SUCCESS) {
        CERR("ALTI: reading display type of I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CNameRead(const unsigned int bus, std::string &name) {

    // read I2C bus's name
    if ((m_status = m_net->name(bus, name)) != I2C::SUCCESS) {
        CERR("ALTI: reading name of I2C bus %d", bus);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CNameRead(const unsigned int bus, const unsigned int dev, std::string &name) {

    // read I2C device's name
    if ((m_status = m_net->name(bus, dev, name)) != I2C::SUCCESS) {
        CERR("ALTI: reading name of I2C bus/device %d/%d", bus, dev);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CNameRead(const unsigned int bus, const unsigned int dev, const unsigned int qty, std::string &name) {

    // read I2C quantity's name
    if ((m_status = m_net->name(bus, dev, qty, name)) != I2C::SUCCESS) {
        CERR("ALTI: reading name of I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CDataRead(const unsigned int bus, const unsigned int dev, const unsigned int qty, std::vector<unsigned int> &data) {

    // read I2C quantity's data as vector
    if ((m_status = m_net->read(bus, dev, qty, data)) != I2C::SUCCESS) {
        CERR("ALTI: reading data of I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CDataRead(const unsigned int bus, const unsigned int dev, const unsigned int qty, unsigned int &data) {

    // read I2C quantity's data as word
    if ((m_status = m_net->read(bus, dev, qty, data)) != I2C::SUCCESS) {
        CERR("ALTI: reading data of I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CDataRead(const unsigned int bus, const unsigned int dev, const unsigned int qty, float &data) {

    // read I2C quantity's data as float
    if ((m_status = m_net->read(bus, dev, qty, data)) != I2C::SUCCESS) {
        CERR("ALTI: reading data of I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CDataRead(const unsigned int bus, const unsigned int dev, const unsigned int qty, std::string &data) {

    // read I2C quantity's data as string
    if ((m_status = m_net->read(bus, dev, qty, data)) != I2C::SUCCESS) {
        CERR("ALTI: reading data of I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CDataRead(const unsigned int bus, const unsigned int dev, const unsigned int qty, const unsigned int off, const unsigned int len, std::vector<unsigned int> &data) {

    // read I2C quantity's data as vector
    if ((m_status = m_net->read(bus, dev, qty, off, len, data)) != I2C::SUCCESS) {
        CERR("ALTI: reading data of I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CDataRead(const I2CNetwork::I2CNode &node, std::vector<unsigned int> &data) {

    // read I2C quantity's data as vector
    if ((m_status = m_net->read(node, data)) != I2C::SUCCESS) {
        CERR("ALTI: reading data of I2C node \"%s\"", node.name().c_str());
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CDataRead(const I2CNetwork::I2CNode &node, unsigned int &data) {

    // read I2C quantity's data as word
    if ((m_status = m_net->read(node, data)) != I2C::SUCCESS) {
        CERR("ALTI: reading data of I2C node \"%s\"", node.name().c_str());
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CDataRead(const I2CNetwork::I2CNode &node, float &data) {

    // read I2C quantity's data as float
    if ((m_status = m_net->read(node, data)) != I2C::SUCCESS) {
        CERR("ALTI: reading data of I2C node \"%s\"", node.name().c_str());
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CDataRead(const I2CNetwork::I2CNode &node, std::string &data) {

    // read I2C quantity's data as string
    if ((m_status = m_net->read(node, data)) != I2C::SUCCESS) {
        CERR("ALTI: reading data of I2C node \"%s\"", node.name().c_str());
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CDataRead(const I2CNetwork::I2CNode &node, const unsigned int off, const unsigned int len, std::vector<unsigned int> &data) {

    // read I2C quantity's data as vector
    if ((m_status = m_net->read(node, off, len, data)) != I2C::SUCCESS) {
        CERR("ALTI: reading data of I2C node \"%s\"", node.name().c_str());
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CDataWrite(const unsigned int bus, const unsigned int dev, const unsigned int qty, const unsigned int data) {

    // write I2C quantity's data as word
    if((m_status = m_net->write(bus, dev, qty, data)) != I2C::SUCCESS) {
        CERR("ALTI: writing data of I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CDataWrite(const unsigned int bus, const unsigned int dev, const unsigned int qty, const float data) {

    // write I2C quantity's data as float
    if ((m_status = m_net->write(bus, dev, qty, data)) != I2C::SUCCESS) {
        CERR("ALTI: writing data of I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
        return (m_status);
    }

    return (m_status);
}

//------------------------------------------------------------------------------

int AltiModule::I2CDataWrite(const unsigned int bus, const unsigned int dev, const unsigned int qty, const std::string &data) {

    // write I2C quantity's data as string
    if ((m_status = m_net->write(bus, dev, qty, data)) != I2C::SUCCESS) {
        CERR("ALTI: writing data of I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
        return (m_status);
    }

    return (m_status);
}
//------------------------------------------------------------------------------

int AltiModule::BSYStatusRead(AltiBusyCounter& cnt) {

    unsigned int data;

    // read BSY status
    if((m_status = m_alti->BSY_Status_Read(data)) != SUCCESS) {
        CERR("ALTI: reading BSY status","");
        return(m_status);
    }

    // set counter mask
    cnt.setStatus(data);

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYMonitoringControlRead(bool& wmod, bool& rmod) {

    ALTI_BSY_MONITORINGCONTROL_BITSTRING ctl;

    // read BSY control
    if((m_status = m_alti->BSY_MonitoringControl_Read(ctl)) != SUCCESS) {
        CERR("ALTI: reading BSY control","");
        return(m_status);
    }
    wmod = ctl.WriteMode() == "AUTOMATIC";
    rmod = ctl.ReadMode() == "LAST";

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYMonitoringControlWrite(const bool wmod, const bool rmod) {

    ALTI_BSY_MONITORINGCONTROL_BITSTRING ctl;

    // write BSY control
    ctl.WriteMode(wmod?"AUTOMATIC":"MANUAL");
    ctl.ReadMode(rmod?"LAST":"FIRST");
    if((m_status = m_alti->BSY_MonitoringControl_Write(ctl)) != SUCCESS) {
        CERR("ALTI: writing BSY control","");
        return(m_status);
    }

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYFifoReset() {

    ALTI_BSY_MONITORINGCONTROL_BITSTRING ctl;

    // read BSY control
    if((m_status = m_alti->BSY_MonitoringControl_Read(ctl)) != SUCCESS) {
        CERR("ALTI: reading BSY control","");
        return(m_status);
    }

    // modify BSY control
    ctl.ClearFifo(true);

    // write BSY control
    if((m_status = m_alti->BSY_MonitoringControl_Write(ctl)) != SUCCESS) {
        CERR("ALTI: writing BSY control","");
        return(m_status);
    }

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYFifoWrite() {

    ALTI_BSY_MONITORINGCONTROL_BITSTRING ctl;

    // read BSY control
    if((m_status = m_alti->BSY_MonitoringControl_Read(ctl)) != SUCCESS) {
        CERR("ALTI: reading BSY control","");
        return(m_status);
    }

    // modify BSY control
    ctl.WriteStrobe(true);

    // write BSY control
    if((m_status = m_alti->BSY_MonitoringControl_Write(ctl)) != SUCCESS) {
        CERR("ALTI: writing BSY control","");
        return(m_status);
    }

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYSelectRead(AltiBusyCounter& cnt) {

    unsigned int data, mask, b;

    // read BSY write counter
    if((m_status = m_alti->BSY_WriteCounter_Read(data)) != SUCCESS) {
        CERR("ALTI: reading BSY write counter","");
        return(m_status);
    }

    // set counter mask
    mask = 0x00000001;
    for(b=0; b<AltiBusyCounter::BUSY_NUMBER; b++, mask=mask<<1) {
        cnt.select(static_cast<AltiBusyCounter::BUSY>(b),(data&mask)==mask);
    }

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYSelectWrite(const AltiBusyCounter& cnt) {

    unsigned int data, mask, b;

    // set counter mask
    data = 0x00000000;
    mask = 0x00000001;
    for(b=0; b<AltiBusyCounter::BUSY_NUMBER; b++, mask=mask<<1) {
        if(cnt.select(static_cast<AltiBusyCounter::BUSY>(b))) data|=mask;
    }

    // write BSY write counter
    if((m_status = m_alti->BSY_WriteCounter_Write(data)) != SUCCESS) {
        CERR("ALTI: writing BSY write counter","");
        return(m_status);
    }

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYCounterReset() {

    unsigned int data;

    // set counter mask
    data = AltiBusyCounter::BSY_SEL_MASK;

    // write BSY counter clear
    if((m_status = m_alti->BSY_ClearCounter_Write(data)) != SUCCESS) {
        CERR("ALTI: writing BSY clear counter","");
        return(m_status);
    }

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYCounterReset(const AltiBusyCounter& cnt) {

    unsigned int data, mask, b;

    // set counter mask
    data = 0x00000000;
    mask = 0x00000001;
    for(b=0; b<AltiBusyCounter::BUSY_NUMBER; b++, mask=mask<<1) {
        if(cnt.select(static_cast<AltiBusyCounter::BUSY>(b))) data|=mask;
    }

    // write BSY counter clear
    if((m_status = m_alti->BSY_ClearCounter_Write(data)) != SUCCESS) {
        CERR("ALTI: writing BSY clear counter","");
        return(m_status);
    }

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYIntervalRead(unsigned int& itvl) {

    // read BSY interval
    if((m_status = m_alti->BSY_WriteInterval_Read(itvl)) != SUCCESS) {
        CERR("ALTI: reading BSY interval","");
        return(m_status);
    }

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYIntervalWrite(const unsigned int itvl) {

    // write BSY interval
    if((m_status = m_alti->BSY_WriteInterval_Write(itvl)) != SUCCESS) {
        CERR("ALTI: writing BSY interval","");
        return(m_status);
    }

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYFifoWatermarkRead(unsigned int& wmrk) {

    // read BSY FIFO watermark
    if((m_status = m_alti->BSY_FifoWatermark_Read(wmrk)) != SUCCESS) {
        CERR("ALTI: reading BSY FIFO watermark","");
        return(m_status);
    }

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYFifoWatermarkWrite(const unsigned int wmrk) {

    // write BSY FIFO watermark
    if((m_status = m_alti->BSY_FifoWatermark_Write(wmrk)) != SUCCESS) {
        CERR("ALTI: writing BSY FIFO watermark","");
        return(m_status);
    }

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYFifoStatusRead(unsigned int& lvl, bool& ful) {

    ALTI_BSY_FIFOSTATUS_BITSTRING sta;

    // read BSY FIFO status
    if((m_status = m_alti->BSY_FifoStatus_Read(sta)) != SUCCESS) {
        CERR("ALTI: reading BSY FIFO status","");
        return(m_status);
    }
    lvl = (sta.Full()) ? BSY_FIFO_SIZE : sta.Level();

    // read BSY FIFO watermark
    unsigned int wmrk;
    if((m_status = m_alti->BSY_FifoWatermark_Read(wmrk)) != SUCCESS) {
        CERR("ALTI: reading BSY FIFO watermark","");
        return(m_status);
    }
    ful = (lvl >= wmrk);

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYFifoLevel(unsigned int& lvl) {

    ALTI_BSY_FIFOSTATUS_BITSTRING sta;

    // read BSY FIFO status
    if((m_status = m_alti->BSY_FifoStatus_Read(sta)) != SUCCESS) {
        CERR("ALTI: reading BSY FIFO status","");
        return(m_status);
    }
    lvl = (sta.Full()) ? BSY_FIFO_SIZE : sta.Level();

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYFifoWait(unsigned int& lvl, const unsigned int siz) {

    unsigned int itim = 0;

    while(true) {

        // read BSY FIFO status
        if((m_status = BSYFifoLevel(lvl)) != SUCCESS) {
            CERR("ALTI: reading BSY FIFO level","");
            return(m_status);
        }

        // check if not empty
        if(lvl >= siz) break;

        // check if timeout
        if(++itim >= BSY_CNT_TIMEOUT) {
            CERR("ALTI: timeout waiting for BSY FIFO","");
            return(TIMEOUT);
        }
    }

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYFifoReady(bool& rdy) {

    ALTI_BSY_FIFOSTATUS_BITSTRING sta;

    // read BSY FIFO status
    if((m_status = m_alti->BSY_FifoStatus_Read(sta)) != SUCCESS) {
        CERR("ALTI: reading BSY FIFO status","");
        return(m_status);
    }
    rdy = (sta.Level() > 0) || sta.Full();

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYFifoFull(bool& ful) {

    ALTI_BSY_FIFOSTATUS_BITSTRING sta;

    // read BSY FIFO status
    if((m_status = m_alti->BSY_FifoStatus_Read(sta)) != SUCCESS) {
        CERR("ALTI: reading BSY FIFO status","");
        return(m_status);
    }
    ful = sta.Full();

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYFifoRead(unsigned int& data) {

    // read BSY FIFO
    if((m_status = m_alti->BSY_BusyFifo_Read(data)) != SUCCESS) {
        CERR("ALTI: reading BSY FIFO","");
        return(m_status);
    }

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYCounterConfigRead(AltiBusyCounter& cnt) {

    ALTI_BSY_MONITORINGCONTROL_BITSTRING ctl;
    unsigned int itv;

    // read BSY control
    if((m_status = m_alti->BSY_MonitoringControl_Read(ctl)) != SUCCESS) {
        CERR("ALTI: reading BSY control","");
        return(m_status);
    }
    cnt.m_wmod = (ctl.WriteMode()=="AUTOMATIC");
    cnt.m_rmod = (ctl.ReadMode()=="FIRST");
    CDBG("ALTI: W/mode = %s, R/mode = %s",(cnt.autoMode()?"AUTOMATIC":"MANUAL"),(cnt.firstMode()?"FIRST":"LAST"));
    
    // read BSY select
    if((m_status = BSYSelectRead(cnt)) != SUCCESS) {
        CERR("ALTI: reading BSY select","");
        return(m_status);
    }
    CDBG("ALTI: select = %s",cnt.selectPrint().c_str());
 
    // read interval time if necessary
    if(!cnt.select(AltiBusyCounter::INTERVAL)) {
        if((m_status = m_alti->BSY_WriteInterval_Read(itv)) != SUCCESS) {
            CERR("ALTI: reading BSY interval","");
            return(m_status);
        }
        cnt.m_interval = itv;
        CDBG("ALTI: interval = %d",cnt.interval());
    }

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYFifoRead(AltiBusyCounter& cnt, const int max, const int alm) {

    unsigned int msk(cnt.selectSize()), lvl, num, wtm;

    // check if anything selected
    if(msk == 0) {
        CERR("ALTI: nothing selected in BSY counter","");
        return(WRONGPAR);
    }

    // reset counter data
    cnt.resetDATA();

    // read BSY FIFO watermark if necessary
    if(alm < 0) {
        if((m_status = BSYFifoWatermarkRead(wtm)) != SUCCESS) {
            CERR("ALTI: reading BSY FIFO watermark","");
            return(m_status);
        }
    }
    else {
        wtm = alm;
    }

    // wait for BSY FIFO and read BSY FIFO level
    if((m_status = BSYFifoWait(lvl,msk)) != SUCCESS) {
        CERR("ALTI: waiting for BSY FIFO to have at least %d words",msk);
        return(m_status);
    }
    cnt.m_overflow = (lvl >= wtm);
    num = (max<0) ? (lvl/msk)*msk : std::min(static_cast<int>(max),static_cast<int>(lvl/msk))*msk;
    CDBG("ALTI: max = %d, msk = %d, lvl = %d, num = %d, slc = %d",max,msk,lvl,num,lvl/msk);

    // read from BSY FIFO using block transfer
    if(num>1) {
        if((m_status = m_alti->BSY_BusyFifo_Read(cnt.segment(),num&(~0x00000001))) != SUCCESS) {
            CERR("ALTI: reading BSY FIFO using block transfer","");
            return(m_status);
        }
        CDBG("ALTI: read block, size = 0x%08x, addr = 0x%08x",num&(~0x00000001),cnt.segment()->VirtualAddress());
    }

    // read the last word when transfer size is not a multiple of two
    if(num&0x00000001) {
        if((m_status = BSYFifoRead(cnt.m_data[num-1])) != SUCCESS) {
            CERR("ALTI: reading last word from BSY FIFO","");
            return(m_status);
        }
        CDBG("ALTI: read last word separately, addr = 0x%08x, data = 0x%08x",&cnt.m_data[num-1],cnt.m_data[num-1]);
    }
    cnt.m_values = num/msk;

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYFifoRead(AltiBusyCounterRate& avg, AltiBusyCounter& cnt, const int max, const int alm) {

    bool rdy;
    unsigned int lvl, wtm;
    ALTI_BSY_FIFOSTATUS_BITSTRING sta;

    CDBG("ALTI: max = %d, alm = %d",max,alm);

    // reset counter
    avg.reset();

    // read BSY FIFO watermark if necessary
    if(alm < 0) {
        if((m_status = BSYFifoWatermarkRead(wtm)) != SUCCESS) {
            CERR("ALTI: reading BSY FIFO watermark","");
            return(m_status);
        }
    }
    else {
        wtm = alm;
    }
/*
    // read RDT FMT BUSY control
    bool vme, bck;
    if((m_status = RDTReadoutBusyControlRead(vme,bck)) != SUCCESS) {
        CERR("ALTI: reading RDT BUSY control","");
        return(m_status);
    }
    avg.m_busy_ondemand = vme ? AltiBusyCounterRate::ENABLED : AltiBusyCounterRate::DISABLED;

    // RDT FMT control
    bool ena, gps;
    if((m_status = RDTFormatterControlRead(ena,gps)) != SUCCESS) {
        CERR("ALTI: reading RDT FMT control","");
        return(m_status);
    }
    avg.m_fmt_master = ena ? AltiBusyCounterRate::ENABLED : AltiBusyCounterRate::DISABLED;
    avg.m_fmt_gps_enable = gps ? AltiBusyCounterRate::DISABLED : AltiBusyCounterRate::ENABLED; // IGNORE flag !!!!

    // RDT DRND/GPS status
    bool emp, ful, bsy;
    if((m_status = RDTDerandomizerStatusRead(AltiModule::RDT_DRND_GPS,emp,ful,bsy,lvl)) != SUCCESS) {
        CERR("reading RDT DRND[\"%s\"] status",AltiModule::RDT_DRND_NAME[AltiModule::RDT_DRND_GPS].c_str());
        return(m_status);
    }
    avg.m_fmt_gps_status = emp ? AltiBusyCounterRate::DISABLED : AltiBusyCounterRate::ENABLED;

    // read RDT FMT BUSY control
    bool daq, roi;
    if((m_status = RDTFormatterBusyControlRead(daq,roi)) != SUCCESS) {
        CERR("ALTI: reading RDT FMT BUSY control","");
        return(m_status);
    }
    avg.m_busy_ctrl[RDT_FIFO_ROI] = roi ? AltiBusyCounterRate::ENABLED : AltiBusyCounterRate::DISABLED;
    avg.m_busy_ctrl[RDT_FIFO_DAQ] = daq ? AltiBusyCounterRate::ENABLED : AltiBusyCounterRate::DISABLED;

    // read RDT FIFO FULL control
    bool set;
    unsigned int alm_thr, bsy_thr;
    for(unsigned int fifo=0; fifo<RDT_FIFO_NUMBER; fifo++) {
        if((m_status = RDTFifoFullControlRead(static_cast<RDT_FIFO_TYPE>(fifo),ena,set,alm_thr,bsy_thr)) != SUCCESS) {
            CERR("ALTI: reading RDT FIFO[\"%s\"] full control",RDT_FIFO_NAME[fifo].c_str());
            return(m_status);
        }
        avg.m_fifo_ctrl[fifo] = ena ? AltiBusyCounterRate::ENABLED : AltiBusyCounterRate::DISABLED;
    }
*/
    // main loop
    while(true) {

        // check if something to read
        if((m_status = m_alti->BSY_FifoStatus_Read(sta)) != SUCCESS) {
            CERR("ALTI: reading BSY FIFO status","");
            return(m_status);
        }
        rdy = (sta.Level() > 0) || sta.Full();
        CDBG("ALTI: BSY FIFO = %s",rdy?"READY":"NOT_READY");
        if(!rdy) break;

        // check if overflow
        lvl = sta.Full() ? BSY_FIFO_SIZE : sta.Level();
        if(lvl >= wtm) avg.m_overflow = true;
        CDBG("ALTI: BSY FIFO overflow = %s",avg.m_overflow?"OVERFLOW":"NO_OVERFLOW");

        // read counter
        if((m_status = BSYFifoRead(cnt,max-avg.m_values,wtm)) != SUCCESS) {
            CERR("ALTI: reading CTPCORE BSY monitoring FIFO","");
            return(m_status);
        }
#ifdef DEBUG
        cnt.dump();
#endif  // DEBUG

        // average counter
        avg += cnt;
#ifdef DEBUG
        avg.dump();
#endif  // DEBUG

        // check maximum reads
        if((avg.m_values >= static_cast<unsigned int>(max)) || avg.m_overflow) break;
    }

    // reset FIFO if necessary
    if(avg.m_overflow) {
        if((m_status = BSYFifoReset() != SUCCESS)) {
            CERR("ALTI: resetting BSY monitoring FIFO","");
            return(m_status);
        }
        CDBG("ALTI: BSY FIFO reset","");
    }

    return(m_status);
}

//------------------------------------------------------------------------------
/*
int AltiModule::BSYSecondaryPartitionControlRead(const unsigned int part, const SECONDARY_BSY_TYPE bsy, bool& ena, bool& tst) {

    // check secondary partition number
    CHECKRANGE(part,SECONDARY_PART_OFFSET,PARTITION_NUMBER-1,"secondary partition number");

    ALTI_RDT_ECR_BUSYCONTROL_BITSTRING ctl;

    // read secondary partition busy control
    if((m_status = m_alti->RDT_ECR_BusyControl_Read(part-SECONDARY_PART_OFFSET,ctl)) != SUCCESS) {
        CERR("ALTI: reading BSY control of secondary partition %d",part);
        return(m_status);
    }
    if(bsy == SECONDARY_BSY_BACKPL) {
        ena = ctl.MaskBackplane();
        tst = ctl.TestBackplane();
    }
    else if(bsy == SECONDARY_BSY_ECR) {
        ena = ctl.MaskEcr();
        tst = ctl.TestEcr();
    }
    else if(bsy == SECONDARY_BSY_DEMAND) {
        ena = ctl.MaskVme() && ctl.BusyOnDemand();
        tst = ctl.TestVme();
    }

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYSecondaryPartitionControlWrite(const unsigned int part, const SECONDARY_BSY_TYPE bsy, const bool ena, const bool tst) {

    // check secondary partition number
    CHECKRANGE(part,SECONDARY_PART_OFFSET,PARTITION_NUMBER-1,"secondary partition number");

    ALTI_RDT_ECR_BUSYCONTROL_BITSTRING ctl;

    // read secondary partition busy control
    if((m_status = m_alti->RDT_ECR_BusyControl_Read(part-SECONDARY_PART_OFFSET,ctl)) != SUCCESS) {
        CERR("ALTI: reading BSY control of secondary partition %d",part);
        return(m_status);
    }

    // modify secondary partition busy control
    if(bsy == SECONDARY_BSY_BACKPL) {
        ctl.MaskBackplane(ena);
        ctl.TestBackplane(tst);
    }
    else if(bsy == SECONDARY_BSY_ECR) {
        ctl.MaskEcr(ena);
        ctl.TestEcr(tst);
    }
    else if(bsy == SECONDARY_BSY_DEMAND) {
        ctl.MaskVme(ena); ctl.BusyOnDemand(ena);
        ctl.TestVme(tst);
    }

    // write secondary partition busy control
    if((m_status = m_alti->RDT_ECR_BusyControl_Write(part-SECONDARY_PART_OFFSET,ctl)) != SUCCESS) {
        CERR("ALTI: writing BSY control of secondary partition %d",part);
        return(m_status);
    }

    return(m_status);
}

//------------------------------------------------------------------------------

int AltiModule::BSYSecondaryPartitionStatusRead(const unsigned int part, bool sta[SECONDARY_BSY_NUMBER]) {

    // check secondary partition number
    CHECKRANGE(part,SECONDARY_PART_OFFSET,PARTITION_NUMBER-1,"secondary partition number");

    ALTI_RDT_ECR_BUSYCONTROL_BITSTRING ctl;

    // read secondary partition busy control
    if((m_status = m_alti->RDT_ECR_BusyControl_Read(part-SECONDARY_PART_OFFSET,ctl)) != SUCCESS) {
        CERR("ALTI: reading BSY control of secondary partition %d",part);
        return(m_status);
    }
    sta[0] = ctl.BusyEcr();
    sta[1] = ctl.BusyBackplane();
    sta[2] = ctl.BusyVme();

    return(m_status);
}*/
//------------------------------------------------------------------------------

/*int AltiModule::BSYGlobalStatusPrint() {

    unsigned int part, drnd, fifo, thr, bsy_thr, lvl, busy;
    bool ign, rsl, vme, bck, ena, set, gps, emp, ful, bsy, daq, roi, dsy, rsy, tst, dwn, gpsempty(false);
    //bool fmp[RDT_FIFO_NUMBER], ffl[RDT_FIFO_NUMBER], fsy[RDT_FIFO_NUMBER];

    AltiDeadtimeCounter dtm;
    std::ostringstream buf1, buf2;

    std::cout << "=========================================================================" << std::endl;
    std::cout << "   SOURCE    |       CONTROL      |     STATUS(IN)     |     BUSY(OUT)" << std::endl;
    std::cout << "=========================================================================" << std::endl;

    // read DTV busy ignore and DTM status
    for(part=0; part<PARTITION_NUMBER; part++) {
        if((m_status = DTVBusyIgnoreRead(part,ign)) != SUCCESS) {
            CERR("ALTI: reading BUSY ignore of partition %d",part);
            return(m_status);
        }
        if((m_status = DTMStatusRead(part,dtm)) != SUCCESS) {
            CERR("ALTI: reading DTM status of partition %d",part);
            return(m_status);
        }
        dtm.getStatus(AltiDeadtimeCounter::RSLT,rsl);
        std::printf("PARTITION %d  | %18s | %18s | %15s\n",part,ign?"ignore":"CONSIDER",rsl?"BUSY":"not busy",!ign&&rsl?"BUSY":"not busy");
    }
    std::cout << "=========================================================================" << std::endl;
    std::cout << "PARTITION 0 (RDT)" << std::endl;
    std::cout << "- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -" << std::endl;

    // read BSY status
    AltiBusyCounter cnt;
    if((m_status = BSYStatusRead(cnt)) != SUCCESS) {
        CERR("ALTI: reading BSY status","");
        return(m_status);
    }

    // RDT readout busy
    if((m_status = RDTReadoutBusyControlRead(vme,bck)) != SUCCESS) {
        CERR("ALTI: reading RDT readout busy","");
        return(m_status);
    }
    std::printf("ON DEMAND    | %18s | %18s | %15s\n",vme?"ENABLED":"disabled","",vme?"BUSY":"not busy");
    std::printf("BACKPLANE    | %18s | %18s | %15s\n",bck?" ENABLED":"disabled","",cnt[AltiBusyCounter::BCKPL]?"BUSY":"not busy");
    std::cout << "- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -" << std::endl;

    // RDT derandomizer busy
    for(drnd=0; drnd<RDT_DRND_NUMBER; drnd++) {
        if((m_status = RDTDerandomizerBusyControlRead(static_cast<RDT_DRND_TYPE>(drnd),ena,set,thr)) != SUCCESS) {
            CERR("ALTI: reading RDT DRND[\"%s\"] busy control",RDT_DRND_NAME[drnd].c_str());
            return(m_status);
        }
        if((m_status = RDTDerandomizerStatusRead(static_cast<RDT_DRND_TYPE>(drnd),emp,ful,bsy,lvl)) != SUCCESS) {
            CERR("ALTI: reading RDT DRND[\"%s\"] busy status",RDT_DRND_NAME[drnd].c_str());
            return(m_status);
        }
        buf1.str("");
        if(set) buf1 << "TEST";
        if(emp) buf1 << (!buf1.str().empty()?", ":"") << "EMPTY";
        if(ful) buf1 << (!buf1.str().empty()?", ":"") << "FULL";
        std::printf("DRND[\"%s\"]  | %18s | %18s | %15s\n",RDT_DRND_NAME[drnd].c_str(),ena?"ENABLED":"disabled",buf1.str().c_str(),bsy?"BUSY":"not busy");
        if(drnd == RDT_DRND_GPS) gpsempty = emp;
    }
    std::cout << "- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -" << std::endl;
    std::printf("DRND           %18s   %18s | %15s\n","","",cnt[AltiBusyCounter::DRND]?"BUSY":"not busy");
    std::cout << "-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -" << std::endl;

    // RDT formatter busy control
    if((m_status = RDTFormatterBusyControlRead(daq,roi)) != SUCCESS) {
        CERR("ALTI: reading RDT FMT BUSY control","");
        return(m_status);
    }
    // RDT formatter busy status
    if((m_status = RDTFormatterBusyStatusRead(dsy,rsy)) != SUCCESS) {
        CERR("ALTI: reading RDT FMT BUSY status","");
        return(m_status);
    }
    std::printf("BUSY [\"%s\"] | %18s | %18s | %15s\n",RDT_FIFO_NAME[RDT_FIFO_ROI].c_str(),roi?" ENABLED":"disabled","",(roi&&rsy)?"BUSY":"not busy");
    std::printf("BUSY [\"%s\"] | %18s | %18s | %15s\n",RDT_FIFO_NAME[RDT_FIFO_DAQ].c_str(),daq?" ENABLED":"disabled","",(daq&&dsy)?"BUSY":"not busy");
    std::cout << "-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -" << std::endl;
    std::printf("RSLT           %18s   %18s | %15s\n","","",cnt[AltiBusyCounter::RSLT]?"BUSY":"not busy");
    std::cout << "-------------------------------------------------------------------------" << std::endl;

    // RDT formatter control
    if((m_status = RDTFormatterControlRead(ena,gps)) != SUCCESS) {
        CERR("ALTI: reading RDT FMT control","");
        return(m_status);
    }
    std::printf("FMT/MASTER   | %18s |\n",ena?" ENABLED":"disabled");
    buf1.str(""); buf1 << (gps?"disabled":" ENABLED") << (!gps&&gpsempty?"/NO_GPS?":"");
    std::printf("FMT/GPS      | %18s |\n",buf1.str().c_str());
    std::cout << "-------------------------------------------------------------------------" << std::endl;

    // RDT FIFO busy
    for(fifo=0; fifo<RDT_FIFO_NUMBER; fifo++) {

        // RDT FIFO full control
        if((m_status = RDTFifoFullControlRead(static_cast<RDT_FIFO_TYPE>(fifo),ena,set,thr,bsy_thr)) != SUCCESS) {
            CERR("ALTI: reading RDT FIFO[\"%s\"] full control",RDT_FIFO_NAME[fifo].c_str());
            return(m_status);
        }

        // RDT FIFO status
        if((m_status = RDTFifoStatusRead(static_cast<RDT_FIFO_TYPE>(fifo),fmp[fifo],ffl[fifo],fsy[fifo],lvl)) != SUCCESS) {
            CERR("ALTI: reading RDT FIFO[\"%s\"] status",RDT_FIFO_NAME[fifo].c_str());
            return(m_status);
        }

        buf1.str("");
        if(set) buf1 << "TEST";
        if(fmp[fifo]) buf1 << (!buf1.str().empty()?", ":"") << "EMPTY";
        if(ffl[fifo]) buf1 << (!buf1.str().empty()?", ":"") << "FULL";
        std::printf("FIFO [\"%s\"] | %18s | %18s | %15s\n",RDT_FIFO_NAME[fifo].c_str(),ena?"ENABLED":"disabled",buf1.str().c_str(),fsy[fifo]?"ALMOST FULL":"not almost full");

        if(fifo > RDT_FIFO_DAQ) continue;

        // RDT FIFO flow (SLINK) control
        if((m_status = RDTFifoFlowControlRead(static_cast<RDT_FIFO_TYPE>(fifo),ena,ign,tst)) != SUCCESS) {
            CERR("ALTI: reading RDT FIFO[\"%s\"] flow control",RDT_FIFO_NAME[fifo].c_str());
            return(m_status);
        }
        // RDT FIFO SLINK status
        if((m_status = RDTFifoSlinkStatusRead(static_cast<RDT_FIFO_TYPE>(fifo),ful,dwn)) != SUCCESS) {
            CERR("ALTI: reading RDT FIFO[\"%s\"] SLINK status",RDT_FIFO_NAME[fifo].c_str());
            return(m_status);
        }
        buf1.str(""); buf1 << (ena?" ENABLED":"disabled") << ", " << (ign?"ignore":"CONSIDER");
        buf2.str(""); buf2 << (ful?"    FULL":"not full") << ", " << (dwn?"DOWN":"not down");
        std::printf("SLINK[\"%s\"] | %18s | %18s | %15s\n",RDT_FIFO_NAME[fifo].c_str(),buf1.str().c_str(),buf2.str().c_str(),cnt[AltiBusyCounter::BUSY_LINK[fifo]]?"XOFF":"not xoff");
        std::cout << "- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -" << std::endl;
    }

    // secondary partitions
    for(part=SECONDARY_PART_OFFSET; part<PARTITION_NUMBER; part++) {
        std::cout << "=========================================================================" << std::endl;
        std::printf("PARTITION %d (secondary):\n",part);
        std::cout << "- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -" << std::endl;

        // read busy status
        if((m_status = BSYSecondaryPartitionStatusRead(part,sta)) != AltiModule::SUCCESS) {
            CERR("ALTI: reading BSY status of secondary partition %d",part);
            return(m_status);
        }
        for(busy=0; busy<AltiModule::SECONDARY_BSY_NUMBER; busy++) {

            // read busy control
            if((m_status = BSYSecondaryPartitionControlRead(part,static_cast<AltiModule::SECONDARY_BSY_TYPE>(busy),ena,tst)) != AltiModule::SUCCESS) {
                CERR("ALTI: reading BSY control of secondary partition %d",part);
                return(m_status);
            }
            std::printf("%-9s    | %18s | %18s | %15s\n",AltiModule::SECONDARY_BSY_NAME[busy].c_str(),ena?"ENABLED":"disabled","",sta[busy]?"BUSY":"not busy");
        }
    }

    return(m_status);
}*/

//------------------------------------------------------------------------------

