//******************************************************************************
// file: menuAltiModule.cc
// desc: menu program for ALTI Module
// auth: 02/11/17 P. Kuzmanovic
//*****************************************************************************

#include "RCDUtilities/RCDUtilities.h"
#include "RCDMenu/RCDMenu.h"
#include "RCDVme/RCDCmemSegment.h"
#include "ALTI/AltiModule.h"

#include "ALTI/AltiConfiguration.h"
#include "ALTI/AltiCommon.h"
#include "ALTI/AltiCounter.h"
#include "ALTI/EventCounter.h"
#include "ALTI/AltiBusyCounter.h"

#include <fstream>
#include <iomanip>
#include <cmath>
#include <unistd.h>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/path.hpp>

 
//#include <sys/time.h>
//#include <algorithm>

using namespace std;
using namespace RCD;
using namespace LVL1;

AltiModule *alti = 0;
CMEMSegment *segment[2] = {0, 0};

// BCR word used in TTC decoder
unsigned int m_bcr_word = 0x01;
bool m_bcr_bitwise = true;

//------------------------------------------------------------------------------

class AltiModuleOpen : public MenuItem {
  public:
    AltiModuleOpen() { setName("open AltiModule"); }
    int action() {

        std::vector<unsigned int> slots;
        unsigned int n = AltiModule::pollSlots(slots);
        // gdb complains because of bus errors on pollSlots
        /*slots.push_back(9);
        slots.push_back(13);
        slots.push_back(18);
        unsigned int n = 3;*/
        std::printf("%d ALTI module%s found", n, (n == 1) ? "" : "s");
        if (n > 0) {
            std::printf(" in slot%s ", (n > 1) ? "s" : "");
            unsigned int i;
            for (i = 0; i < n; i++) {
                std::printf("%d%s", slots[i], (i < n - 1) ? ", " : ".\n");
            }
        }
        else {
            std::printf("\n");
            return(-1);
        }

        int s;
        if (n == 1) {
            s = slots[0];
        }
        else {
            while (true) {
                s  = enterInt("Slot number", AltiModule::ALTI_SLOT_MIN, AltiModule::ALTI_SLOT_MAX);
                if (std::find(slots.begin(), slots.end(), s) != slots.end()) break;
            }
        }
        std::printf("ALTI module in slot %d selected\n", s);

        // create AltiModule
        alti = new AltiModule(s);

        // create CMEM segements
        std::string snam = "menuAltiModule";
        // sometimes we'll read both TTC decoder memories at once
        segment[0] = new RCD::CMEMSegment(snam + '0', sizeof(u_int)*AltiModule::MAX_MEMORY, true);
        if ((*segment[0])() != CMEM_RCC_SUCCESS) {
            CERR("error opening CMEM segment \"%s\" of size 0x%08x\n", (snam + '0').c_str(), segment[0]->Size());
            return (-1);
        }
        COUT("opened CMEM segment \"%s\" of size 0x%08x, phys 0ix%08x, virt 0x%08x", (snam + '0').c_str(), segment[0]->Size(), segment[0]->PhysicalAddress(), segment[0]->VirtualAddress());
        segment[1] = new RCD::CMEMSegment(snam + '1', sizeof(u_int)*AltiModule::MAX_MEMORY, true);
        if ((*segment[1])() != CMEM_RCC_SUCCESS) {
            CERR("error opening CMEM segment \"%s\" of size 0x%08x\n", (snam + '1').c_str(), segment[1]->Size());
            return (-1);
        }
        COUT("opened CMEM segment \"%s\" of size 0x%08x, phys 0ix%08x, virt 0x%08x", (snam + '1').c_str(), segment[1]->Size(), segment[1]->PhysicalAddress(), segment[1]->VirtualAddress());

        return ((*alti)());
    }
};

//------------------------------------------------------------------------------

class AltiModuleClose : public MenuItem {
  public:
    AltiModuleClose() { setName("close AltiModule"); }
    int action() {

        // delete CMEM segments
        if (segment[0]) delete segment[0];
        if (segment[1]) delete segment[1];

        // delete AltiModule
        if (alti) delete alti;

        return (0);
    }
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

class AltiCSRManufacturerIdRead : public MenuItem {
  public:
    AltiCSRManufacturerIdRead() { setName("read  manufacturer ID"); }
    int action() {

        int rtnv(0);
        unsigned int manuf;

        if ((rtnv = alti->CSRManufacturerIdRead(manuf)) != AltiModule::SUCCESS) {
            CERR("reading manufacturer id", "");
            return (rtnv);
        }
        std::printf("manufacturer id = 0x%08x <-> %s\n", manuf, ((manuf == AltiModule::MANUF_ID)? "CERN" : "ERROR: NOT IDENTIFIED"));

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCSRBoardIdRead : public MenuItem {
  public:
    AltiCSRBoardIdRead() { setName("read  board        ID"); }
    int action() {

        int rtnv(0);
        unsigned int board;

        if ((rtnv = alti->CSRBoardIdRead(board)) != AltiModule::SUCCESS) {
            CERR("reading board id", "");
            return (rtnv);
        }
        std::printf("board id = 0x%08x <-> %s\n", board, ((board == AltiModule::BOARD_ID)? "ALTI" : "ERROR: NOT IDENTIFIED"));

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCSRRevisionIdRead : public MenuItem {
  public:
    AltiCSRRevisionIdRead() { setName("read  revision     ID"); }
    int action() {

        int rtnv(0);
        unsigned int rev, dd, mm, yy, revnum;

        if ((rtnv = alti->CSRRevisionIdRead(rev)) != AltiModule::SUCCESS) {
            CERR("reading revision id", "");
            return (rtnv);
        }
        AltiCommon::parseRevision(rev, dd, mm, yy, revnum);
        std::printf("revision id = 0x%08x (#%d, %02d.%02d.%d.)\n", rev, revnum, dd, mm, yy);

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCSRSlaveCharacteristicsRead : public MenuItem {
  public:
    AltiCSRSlaveCharacteristicsRead() { setName("read  slave characteristics"); }
    int action() {

        int rtnv(0);
        unsigned int slcharact;

        if ((rtnv = alti->CSRSlaveCharacteristicsRead(slcharact)) != AltiModule::SUCCESS) {
            CERR("reading slave characteristics", "");
            return (rtnv);
        }
        std::printf("slave characteristics = 0x%08x", slcharact);

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCSRBARRead : public MenuItem {
  public:
    AltiCSRBARRead() { setName("read  BAR"); }
    int action() {

        int rtnv(0);
        unsigned int bar;

        if ((rtnv = alti->CSRBARRead(bar)) != AltiModule::SUCCESS) {
            CERR("reading BAR", "");
            return (rtnv);
        }
        std::printf("BAR = 0x%08x <-> slot %d\n", bar, (bar & 0x0000001f));

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCSRADER0Read : public MenuItem {
  public:
    AltiCSRADER0Read() { setName("read  ADER0"); }
    int action() {

        int rtnv(0);
        unsigned int ader0;

        if ((rtnv = alti->CSRADER0Read(ader0)) != AltiModule::SUCCESS) {
            CERR("reading ADER0", "");
            return (rtnv);
        }
        std::printf("ADER0 = 0x%08x\n", AltiModule::swapEndian(ader0, AltiModule::ADER0_BASE_BITS));

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCSRADER0Write : public MenuItem {
  public:
    AltiCSRADER0Write() { setName("write ADER0"); }
    int action() {

        int rtnv(0);
        //unsigned int base = enterHex("ADER0", 0, 0xffffff00); /// ?
        unsigned int base = enterHex("ADER0 (8 msbs)", 0, 0xff);
        //if (base < 0x100) base = base << 24;
        unsigned int upper_8MB;
        if (AltiModule::ADER0_BASE_BITS == 9) {
            upper_8MB = enterHex("ADER0 (9th msb) ", 0, 1);
        }
        else
            upper_8MB = 0;
        unsigned int ader0 = AltiModule::swapEndian((base << 24) | (upper_8MB << (32 - AltiModule::ADER0_BASE_BITS)), AltiModule::ADER0_BASE_BITS);

        if ((rtnv = alti->CSRADER0Write(ader0)) != AltiModule::SUCCESS) {
            CERR("writing ADER0", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCSRADER1Read : public MenuItem {
  public:
    AltiCSRADER1Read() { setName("read  ADER1"); }
    int action() {

        int rtnv(0);
        unsigned int ader1;

        if ((rtnv = alti->CSRADER1Read(ader1)) != AltiModule::SUCCESS) {
            CERR("reading ADER1", "");
            return (rtnv);
        }
        std::printf("ADER1 = 0x%08x%s\n", ader1, (ader1 > 0) ? ((" (\"" + AltiModule::ALTI_RAM_MEMORY_NAME[ader1 - 1] + "\")").c_str()) : "");

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCSRADER1Write : public MenuItem {
  public:
    AltiCSRADER1Write() { setName("write ADER1"); }
    int action() {

        int rtnv(0);
        unsigned int ader1 = enterHex("ADER1", 0, AltiModule::ALTI_RAM_MEMORY_NUMBER - 1 + 1);

        if ((rtnv = alti->CSRADER1Write(ader1)) != AltiModule::SUCCESS) {
            CERR("writing ADER1", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiSerialNumber : public MenuItem {
  public: AltiSerialNumber() { setName("read  serial number"); }
    int action() {

        int rtnv(0);
        std::string ser, id;

        if ((rtnv = alti->OWSerialNumber(ser)) != AltiModule::SUCCESS) {
            cout << "ALTI does not seem to have DS1WM => PRE_PRODUCTION_MODULE" << std::endl;
            return (rtnv);
        }
        DS1WM::getIdentifier(ser, id);
        cout << "Serial number = \"" << ser << "\" - identifier = \"" << id << "\"" << std::endl;

        return (rtnv);
    }
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

class AltiReset : public MenuItem {
  public:
    AltiReset() { setName("reset ALTI"); }
    int action() {

        int rtnv(0);

        if ((rtnv = alti->AltiReset()) != AltiModule::SUCCESS) {
            CERR("resetting ALTI","");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------
class AltiSetup : public MenuItem {
  public:
    AltiSetup() { setName("setup ALTI"); }
    int action() {

        int rtnv(0);

        if ((rtnv = alti->AltiSetup()) != AltiModule::SUCCESS) {
            CERR("setting up ALTI","");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCheck : public MenuItem {
  public:
    AltiCheck() { setName("check ALTI"); }
    int action() {

        int rtnv(0);

        if ((rtnv = alti->AltiCheck()) != AltiModule::SUCCESS) {
            CERR("checking ALTI","");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiConfigRead : public MenuItem {
  public:
    AltiConfigRead() { setName("read  ALTI configuration"); }
    int action() {

        int rtnv(0);
        AltiConfiguration cfg;
        /*std::string fn = enterString("File name (empty=config)");

        if (fn.empty()) cfg.clearFileNames();*/
        if ((rtnv = alti->AltiConfigRead(cfg)) != AltiModule::SUCCESS) {
            CERR("reading ALTI config", "");
            return (rtnv);
        }
        cfg.print();

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiConfigWrite : public MenuItem {
  public:
    AltiConfigWrite() { setName("write ALTI configuration [default]"); }
    int action() {

        int rtnv(0);
        bool flag = (bool) enterInt("Write default configuration (0=\"NO\", 1=\"YES\")", 0, 1);
        if (!flag) return (rtnv);

        AltiConfiguration cfg;
        if ((rtnv = alti->AltiConfigWrite(cfg)) != AltiModule::SUCCESS) {
            CERR("writing ALTI config", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiConfigReadFile : public MenuItem {
  public:
    AltiConfigReadFile() { setName("read  ALTI configuration [into file]"); }
    int action() {

        int rtnv(0);
        std::string pfx = enterString("File PREFIX (to \"cfg.dat\")");
        std::string fn = pfx + "cfg.dat";
        AltiConfiguration cfg(pfx);

        if ((rtnv = alti->AltiConfigRead(cfg)) != AltiModule::SUCCESS) {
            CERR("reading ALTI config", "");
            return (rtnv);
        }
        if ((rtnv = cfg.write(fn)) != AltiModule::SUCCESS) {
            CERR("writing ALTI config into file \"%s\"", fn.c_str());
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiConfigWriteFile : public MenuItem {
  public:
    AltiConfigWriteFile() { setName("write ALTI configuration [from file]"); }
    int action() {

        int rtnv(0);
        AltiConfiguration cfg;
        std::string fn = enterString("File name (empty=config)");

        if (fn.empty()) cfg.clearFileNames();
        else {
            if ((rtnv = cfg.read(fn)) != AltiModule::SUCCESS) {
                CERR("reading ALTI config from file \"%s\"", fn.c_str());
                return (rtnv);
            }
        }

        if ((rtnv = alti->AltiConfigWrite(cfg)) != AltiModule::SUCCESS) {
            CERR("writing ALTI config","");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiConfigWritePredefined : public MenuItem {
  public:
    AltiConfigWritePredefined() { setName("write ALTI configuration [predefined modes]"); }
    int action() {

        int rtnv(0);
        AltiConfiguration cfg;

        unsigned int mode_choice = enterInt("Predefined mode (0=\"CTP_SLAVE\", 1=\"ALTI_SLAVE\", 2=\"LEMO_SLAVE\", 3=\"PATTERN_GENERATOR\")", 0, 4);

	string prefix = ".";
	bool local_path = false;
	// first check local directory
	char* path = getenv("PWD");
	if(path) {     
	  boost::filesystem::path full_path = std::string(path) + "/installed/share/data/ALTI/"; 
	  if(boost::filesystem::is_directory(full_path)) { 
	    prefix = std::string(path) + "/installed";
	    local_path = true;
	  }	
	}
	if(!local_path) { // check L1CT (tdaq7) and TDAQ (tdaq8) releases
	  path = getenv("L1CT_INST_PATH"); 
	  if(!path) path = getenv("TDAQ_INST_PATH");
	  if(path) prefix = path;
	}
	prefix += "/share/data/ALTI/";
	std::string cfg_file;
        switch (mode_choice) {
	case 0: // CTP_SLAVE:
	  cfg_file = "AltiModule_CTP_Slave_cfg.dat";
	  break;
	case 1: // ALTI_SLAVE:
	  cfg_file = "AltiModule_ALTI_Slave_cfg.dat";
	  break;
	case 2: // LEMO_SLAVE:
	  cfg_file = "AltiModule_LEMO_Slave_cfg.dat";
	  break;
	case 3: // PATTERN_GENERATOR:
	  cfg_file = "AltiModule_Pattern_Generator_cfg.dat";
	  break;
	default:
	  break;
        }

        if (cfg_file.empty()) cfg.clearFileNames();
        else {
            if ((rtnv = cfg.read(prefix+cfg_file)) != AltiModule::SUCCESS) {
	      CERR("reading ALTI config from file \"%s&s\"", prefix.c_str(),cfg_file.c_str());
                return (rtnv);
            }
	    // set prefix to configuration files 
	    cfg.addFileNamePrefix(prefix);
	}

        if ((rtnv = alti->AltiConfigWrite(cfg)) != AltiModule::SUCCESS) {
            CERR("writing ALTI config","");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

class AltiSIGSwitchConfigRead : public MenuItem {
  public:
    AltiSIGSwitchConfigRead() { setName("read  switch    configuration"); }
    int action() {

        int rtnv(0);

        AltiModule::SIGNAL signal;
        AltiModule::SIGNAL_SOURCE src[AltiModule::SIGNAL_DESTINATION_NUMBER];
        AltiModule::SIGNAL_DESTINATION dst;

        //table
        printf("\n");
        printf("+-----------------------------------------------------------+\n");
        printf("| Signal    | CTP out   | ALTI out  | NIM out   | to FPGA   |\n");
        printf("+-----------------------------------------------------------+\n");

        unsigned int i, j;
        for (i = 0; i < AltiModule::SIGNAL_NUMBER_ROUTED; i++) {
            signal = AltiModule::SIGNAL(i);
            for (j = 0; j < AltiModule::SIGNAL_DESTINATION_NUMBER; j++) {
                dst = AltiModule::SIGNAL_DESTINATION(j);
                if ((rtnv = alti->SIGSwitchConfigRead(signal, dst, src[j])) != AltiModule::SUCCESS) {
                    CERR("reading signal switch configuration", "");
                    return (rtnv);
                }
            }
            printf("| [%02d]%-5s | %-9s | %-9s | %-9s | %-9s |\n", i, AltiModule::SIGNAL_NAME[signal].c_str(), AltiModule::SIGNAL_SOURCE_NAME[src[0]].c_str(), AltiModule::SIGNAL_SOURCE_NAME[src[1]].c_str(), AltiModule::SIGNAL_SOURCE_NAME[src[2]].c_str(), AltiModule::SIGNAL_SOURCE_NAME[src[3]].c_str());
            printf("+-----------------------------------------------------------+\n");
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiSIGSwitchConfigWrite : public MenuItem {
  public:
    AltiSIGSwitchConfigWrite() { setName("write switch    configuration"); }
    int action() {

        int rtnv(0);

        AltiModule::SIGNAL signal;
        AltiModule::SIGNAL_DESTINATION dst;
        AltiModule::SIGNAL_SOURCE src;
        unsigned int signal_choice = enterInt("Signal (0=\"BC\", 1=\"ORB\", ..., 17=\"BGO3\", 18=all)                       ", 0, AltiModule::SIGNAL_NUMBER_ROUTED - 1 + 1);
        unsigned int dst_choice = enterInt("Destination (0=\"CTP_OUT\", 1=\"ALTI_OUT\", 2=\"NIM_OUT\", 3=\"TO_FPGA\", 4=all)", 0, AltiModule::SIGNAL_DESTINATION_NUMBER - 1 + 1);
        src = (AltiModule::SIGNAL_SOURCE) enterInt("Source (0=\"CTP_IN\", 1=\"ALTI_IN\", 2=\"NIM_IN\", 3=\"FROM_FPGA\")             ", 0, AltiModule::SIGNAL_SOURCE_NUMBER - 1 - 1);
        if (signal_choice < AltiModule::SIGNAL_NUMBER_ROUTED) signal = (AltiModule::SIGNAL) signal_choice;
        if (dst_choice < AltiModule::SIGNAL_DESTINATION_NUMBER) dst = (AltiModule::SIGNAL_DESTINATION) dst_choice;

        if ((signal_choice < AltiModule::SIGNAL_NUMBER_ROUTED) && (dst_choice < AltiModule::SIGNAL_DESTINATION_NUMBER)) { // 1 signal, 1 output
            if ((rtnv = alti->SIGSwitchConfigWrite(signal, dst, src)) != AltiModule::SUCCESS) {
                CERR("writing source %s to switch %s destination %s", AltiModule::SIGNAL_SOURCE_NAME[src].c_str(), AltiModule::SIGNAL_NAME[signal].c_str(), AltiModule::SIGNAL_DESTINATION_NAME[dst].c_str());
                return (rtnv);
            }
        }
        else if (signal_choice < AltiModule::SIGNAL_NUMBER_ROUTED) { // 1 signal, all outputs
            if ((rtnv = alti->SIGSwitchConfigWrite(signal, src)) != AltiModule::SUCCESS) {
                CERR("writing source %s to switch %s", AltiModule::SIGNAL_SOURCE_NAME[src].c_str(), AltiModule::SIGNAL_NAME[signal].c_str());
                return (rtnv);
            }
        }
        else if (dst_choice < AltiModule::SIGNAL_DESTINATION_NUMBER) { // all signals, 1 output
            if ((rtnv = alti->SIGSwitchConfigWrite(dst, src)) != AltiModule::SUCCESS) {
                CERR("writing source %s to destination %s", AltiModule::SIGNAL_SOURCE_NAME[src].c_str(), AltiModule::SIGNAL_DESTINATION_NAME[dst].c_str());
                return (rtnv);
            }
        }
        else { // all signals, all outputs
            if ((rtnv = alti->SIGSwitchConfigWrite(src)) != AltiModule::SUCCESS) {
                CERR("writing source %s", AltiModule::SIGNAL_SOURCE_NAME[src].c_str());
                return (rtnv);
            }
        }


        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiSIGSwitchMasterConfigurationWrite : public MenuItem {
  public:
    AltiSIGSwitchMasterConfigurationWrite() { setName("write switch    configuration [Master mode - from FPGA]"); }
    int action() {

        int rtnv(0);

        if ((rtnv = alti->SIGSwitchConfigWrite(AltiModule::MASTER)) != AltiModule::SUCCESS) {
            CERR("configuring switches to master mode", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiSIGSwitchCtpSlaveConfigurationWrite : public MenuItem {
  public:
    AltiSIGSwitchCtpSlaveConfigurationWrite() { setName("write switch    configuration [Slave  mode - from CTP]"); }
    int action() {

        int rtnv(0);

        if ((rtnv = alti->SIGSwitchConfigWrite(AltiModule::CTP_SLAVE)) != AltiModule::SUCCESS) {
            CERR("configuring switches to CTP-slave mode", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiSIGSwitchAltiSlaveConfigurationWrite : public MenuItem {
  public:
    AltiSIGSwitchAltiSlaveConfigurationWrite() { setName("write switch    configuration [Slave  mode - from ALTI]"); }
    int action() {

        int rtnv(0);

        if ((rtnv = alti->SIGSwitchConfigWrite(AltiModule::ALTI_SLAVE)) != AltiModule::SUCCESS) {
            CERR("configuring switches to ALTI-slave mode", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiSIGSwitchLemoSlaveConfigurationWrite : public MenuItem {
  public:
    AltiSIGSwitchLemoSlaveConfigurationWrite() { setName("write switch    configuration [Lemo  mode - from NIM]"); }
    int action() {

        int rtnv(0);

        if ((rtnv = alti->SIGSwitchConfigWrite(AltiModule::LEMO_SLAVE)) != AltiModule::SUCCESS) {
            CERR("configuring switches to LEMO-slave mode", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

class AltiSIGEqualizersConfigRead : public MenuItem {
  public:
    AltiSIGEqualizersConfigRead() { setName("read  equalizer configuration"); }
    int action() {

        int rtnv(0);

        AltiModule::SIGNAL_SOURCE src;
        //AltiModule::SIGNAL_DESTINATION dst;

        printf("\n");
        printf("+---------------------------------------------------------------------------------------------------------+\n");
        printf("| Cable      | Status     | Vpeak[V]   | Vpole[V]   | Vgain[V]   | Voffset[V] || High-level configuration |\n");
        printf("+---------------------------------------------------------------------------------------------------------+\n");

	bool enabled;
        AltiModule::EQUALIZER_CONFIG eqz_config;

        unsigned int vpeak, vpole, vgain, voffset;
        float v_vpeak, v_vpole, v_vgain, v_voffset;
        unsigned int i;
        for (i = 0; i < AltiModule::SIGNAL_SOURCE_NUMBER - 3; i++) {
            src = AltiModule::SIGNAL_SOURCE(i);

	    if ((rtnv = alti->SIGEqualizersConfigRead(src, enabled, eqz_config)) != AltiModule::SUCCESS) {
	      CERR("reading high-level %s equalizer configuration", AltiModule::SIGNAL_SOURCE_NAME[src].c_str());
	      return (rtnv);
	    }

            if ((rtnv = alti->SIGEqualizersConfigRead(src, vpeak, vpole, vgain, voffset)) != 0) {
                CERR("reading cable %s equalizer configuration", AltiModule::SIGNAL_SOURCE_NAME[src].c_str());
                return (rtnv);
            }
            v_vpeak = ((float) vpeak)/256.0*2.5;
            v_vpole = ((float) vpole)/256.0*2.5;
            v_vgain = ((float) vgain)/256.0*2.5;
            v_voffset = ((float) voffset)/256.0*2.5;
            std::printf("| [%01d]%-7s | %-10s | %01.5f    | %01.5f    | %01.5f    | %01.5f    || %-24s |\n", src,  AltiModule::SIGNAL_SOURCE_NAME[src].c_str(), enabled ? "ENABLED " : "DISABLED", v_vpeak, v_vpole, v_vgain, v_voffset, (AltiModule::EQUALIZER_CONFIG_NAME[eqz_config] + ((eqz_config == AltiModule::EQUALIZER_CONFIG::SHORT_CABLE) ? " (< 10m)" : ((eqz_config == AltiModule::EQUALIZER_CONFIG::LONG_CABLE) ? " (> 10m)" : ""))).c_str());
            printf("+---------------------------------------------------------------------------------------------------------+\n");
        }

        return (rtnv);
    }
};


//------------------------------------------------------------------------------

class AltiSIGEqualizersPredefinedModeWrite : public MenuItem {
  public:
    AltiSIGEqualizersPredefinedModeWrite() { setName("write equalizer configuration"); }
    int action() {

        int rtnv(0);
	
	AltiModule::SIGNAL_SOURCE src(AltiModule::CTP_IN);
        unsigned int src_choice = enterInt("Cable (0=\"CTP_IN\", 1=\"ALTI_IN\")                                  ", 0, AltiModule::SIGNAL_SOURCE_NUMBER - 3 - 1);
	if (src_choice < AltiModule::SIGNAL_SOURCE_NUMBER - 3) src = (AltiModule::SIGNAL_SOURCE) src_choice;
	
	std::string command = "Enable " + AltiModule::SIGNAL_SOURCE_NAME[src] + " equalizer (0=disable, 1=enable)";	  
	bool enable = (bool) enterInt(command.c_str(), 0, 1);
	if (enable) { 
	  command = AltiModule::SIGNAL_SOURCE_NAME[src] + " mode (0=\"SHORT_CABLE (< 10m)\", 1=\"LONG_CABLE (> 10m)\")";
	  AltiModule::EQUALIZER_CONFIG mode = (AltiModule::EQUALIZER_CONFIG) enterInt(command.c_str(), 0, 1);   
	  if ((rtnv = alti->SIGEqualizersConfigWrite(src, enable, mode)) != AltiModule::SUCCESS) {
	    CERR("writing equalizer configuration %s: %s", AltiModule::SIGNAL_SOURCE_NAME[src].c_str(), AltiModule::EQUALIZER_CONFIG_NAME[mode].c_str());
            return (rtnv);
	  }
	}
	else {
	  if ((rtnv = alti->SIGEqualizersEnableWrite(src, false)) != AltiModule::SUCCESS) {
	    CERR("disabling equalizer for %s", AltiModule::SIGNAL_SOURCE_NAME[src].c_str());
	    return (rtnv);
	  }
	}
	
        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiSIGEqualizersConfigWrite : public MenuItem {
  public:
    AltiSIGEqualizersConfigWrite() { setName("write equalizer configuration [expert user]"); }
    int action() {

        int rtnv(0);

        AltiModule::SIGNAL_SOURCE src(AltiModule::CTP_IN);
        AltiModule::EQUALIZER_VOLTAGE vlt(AltiModule::V_PEAK);
        unsigned int i, v, v_vector[AltiModule::EQUALIZER_VOLTAGE_NUMBER];
        std::ostringstream msg;
        unsigned int src_choice = enterInt("Cable (0=\"CTP_IN\", 1=\"ALTI_IN\")                                  ", 0, AltiModule::SIGNAL_SOURCE_NUMBER - 3 - 1);
	if (src_choice < AltiModule::SIGNAL_SOURCE_NUMBER - 3) src = (AltiModule::SIGNAL_SOURCE) src_choice;

	bool enable = (bool) enterInt("Enable  (0=disable, 1=enable)", 0, 1);
	if(enable) { 
	  if ((rtnv = alti->SIGInputSyncEnableWrite(src, true)) != AltiModule::SUCCESS) {
	    CERR("enabling equalizer for %s", AltiModule::SIGNAL_SOURCE_NAME[src].c_str());
	    return (rtnv);
	  }

	  unsigned int vlt_choice = enterInt("Voltage (0=\"V_PEAK\", 1=\"V_POLE\", 2=\"V_GAIN\", 3=\"V_OFFSET\", 4=all)", 0, AltiModule::EQUALIZER_VOLTAGE_NUMBER - 1 + 1);
	  if (vlt_choice < AltiModule::EQUALIZER_VOLTAGE_NUMBER) vlt = (AltiModule::EQUALIZER_VOLTAGE) vlt_choice;
	  
	  if (vlt_choice < AltiModule::EQUALIZER_VOLTAGE_NUMBER) { // 1 cable, 1 voltage
            msg.str("");
            msg.clear();
            msg << ("\"" + AltiModule::SIGNAL_SOURCE_NAME[src] + "\" ") << (AltiModule::EQUALIZER_VOLTAGE_NAME[vlt]);
            v = (unsigned int) (AltiCommon::enterFloat(msg.str(), 0, 2.5)/2.5*255.0);
            if (alti->SIGEqualizersConfigWrite(src, vlt, v) != 0) {
	      CERR("writing %s equalizer %s voltage", AltiModule::SIGNAL_SOURCE_NAME[src].c_str(), AltiModule::EQUALIZER_VOLTAGE_NAME[vlt].c_str());
	      return (rtnv);
            }
	  }
	  else if (vlt_choice == AltiModule::EQUALIZER_VOLTAGE_NUMBER) { // 1 cable, all voltage
            for (i = 0; i < AltiModule::EQUALIZER_VOLTAGE_NUMBER; i++) {
	      vlt = (AltiModule::EQUALIZER_VOLTAGE) i;
	      
	      msg.str("");
	      msg.clear();
	      msg << ("\"" + AltiModule::SIGNAL_SOURCE_NAME[src] + "\" ") << std::setw(8) << std::left << (AltiModule::EQUALIZER_VOLTAGE_NAME[vlt]);
	      v_vector[i] = (unsigned int) (AltiCommon::enterFloat(msg.str(), 0, 2.5)/2.5*255.0);
            }
            if (alti->SIGEqualizersConfigWrite(src, v_vector[0], v_vector[1], v_vector[2], v_vector[3]) != 0) {
	      CERR("writing %s equalizer voltages", AltiModule::SIGNAL_SOURCE_NAME[src].c_str());
	      return (rtnv);
            }
	  }
	}
	else {
	  if ((rtnv = alti->SIGEqualizersEnableWrite(src, false)) != AltiModule::SUCCESS) {
	    CERR("disabling equalizer for %s", AltiModule::SIGNAL_SOURCE_NAME[src].c_str());
	    return (rtnv);
	  }
	}

        return (rtnv);
    }
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

class AltiSIGInputSyncHistogramsRead : public MenuItem {
  public:
    AltiSIGInputSyncHistogramsRead() { setName("read  IO histogram statistics"); }
    int action() {

        int rtnv(0);

        AltiModule::ASYNC_INPUT_SIGNAL asyncIn;
        std::vector<unsigned int> countPosByPhase, countNegByPhase;

        printf("\n");
        printf("Histogram statistics:\n");
        printf("+-------------------------------------------------------------------+\n");
        printf("| Asynchronous input signal | Histogram count by phase [pos/neg]    |\n");
        printf("+-------------------------------------------------------------------+\n");
        printf("|                           | 0   deg | 90  deg | 180 deg | 270 deg |\n");
        printf("+-------------------------------------------------------------------+\n");

        unsigned int i;
        for (i = 0; i < AltiModule::ASYNC_INPUT_SIGNAL_NUMBER; i++) {
            asyncIn = (AltiModule::ASYNC_INPUT_SIGNAL) i;
            // LEMO sync inputs: removed!
            if ((asyncIn >= AltiModule::LEMO_ORB) && (asyncIn <= AltiModule::LEMO_BGO3)) continue;
            rtnv = (alti->SIGInputSyncHistogramsRead(asyncIn, countPosByPhase, countNegByPhase));
            if (rtnv != AltiModule::SUCCESS) {
                CERR("reading sample histogram statistics for asynchronous input signal \"%s\"", AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
                return (rtnv);
            }
            else {
                printf("| [%02d]%-21s |   %d/%d   |   %d/%d   |   %d/%d   |   %d/%d   |\n", asyncIn, AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str(), countPosByPhase[0], countNegByPhase[0], countPosByPhase[1], countNegByPhase[1], countPosByPhase[2], countNegByPhase[2], countPosByPhase[3], countNegByPhase[3]);
            }
            if ((asyncIn == AltiModule::SWX_TTR3) || (asyncIn == AltiModule::RJ45_CALREQ2) || (asyncIn == AltiModule::CTP_CALREQ2) || (asyncIn == AltiModule::ALTI_CALREQ2)) printf("+-------------------------------------------------------------------+\n");

        }
        printf("+-------------------------------------------------------------------+\n");

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiSIGInputSyncHistogramsReset : public MenuItem {
  public:
    AltiSIGInputSyncHistogramsReset() { setName("reset IO histograms"); }
    int action() {

        int rtnv(0);

        if ((rtnv = alti->SIGInputSyncHistogramsReset()) != AltiModule::SUCCESS) {
            CERR("resetting all input synchronization histograms", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiSIGInputSyncStatusRead : public MenuItem {
  public:
    AltiSIGInputSyncStatusRead() { setName("read  IO sync   status"); }
    int action() {

        int rtnv(0);

        //table
        printf("\n");
        printf("+-------------------------------------------------------------+\n");
        printf("| Asynchronous input signal | Sync     | Clock phase selected |\n");
        printf("+-------------------------------------------------------------+\n");

        std::vector<bool> enabled;
        std::vector<unsigned int> phase;
        if ((rtnv = alti->SIGInputSyncEnableRead(enabled, phase)) != AltiModule::SUCCESS) {
            CERR("reading synchronization status for all asynchronous input signals", "");
            return (rtnv);
        }

        unsigned int i;
        AltiModule::ASYNC_INPUT_SIGNAL asyncIn;
        for (i = 0; i < AltiModule::ASYNC_INPUT_SIGNAL_NUMBER; i++) {
            asyncIn = AltiModule::ASYNC_INPUT_SIGNAL(i);
            // LEMO sync inputs: removed!
            if ((asyncIn >= AltiModule::LEMO_ORB) && (asyncIn <= AltiModule::LEMO_BGO3)) continue;
            printf("| [%02d]%-21s | %-8s | %-20s |\n", i, AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str(), enabled[i] ? "Enabled" : "Disabled", (phase[i] == 0) ? "Auto" : ((phase[i] == 1) ? "0   degrees" : ((phase[i] == 2) ? "90  degrees" : ((phase[i] == 3) ? "180 degrees" : (phase[i] == 4) ? "270 degrees" : ""))));
            if ((asyncIn == AltiModule::SWX_TTR3) || (asyncIn == AltiModule::RJ45_CALREQ2) || (asyncIn == AltiModule::CTP_CALREQ2) || (asyncIn == AltiModule::ALTI_CALREQ2) || (asyncIn == AltiModule::CDR_DERR)) printf("+-------------------------------------------------------------+\n");
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiSIGInputSyncEnableWrite : public MenuItem {
  public:
    AltiSIGInputSyncEnableWrite() { setName("write IO sync   enable"); }
    int action() {

        int rtnv(0);

        bool group = (bool) enterInt("Single signal or group of signals (0=single, 1=group)       ", 0, 1);
        AltiModule::ASYNC_INPUT_SIGNAL asyncIn;
        bool enable;
        unsigned int phase;

        printf("-------------\n");
        for (unsigned int i = 0; i < AltiModule::ASYNC_INPUT_SIGNAL_NUMBER; i++) {
            asyncIn = AltiModule::ASYNC_INPUT_SIGNAL(i);
            // LEMO sync inputs: removed!
            if ((asyncIn >= AltiModule::LEMO_ORB) && (asyncIn <= AltiModule::LEMO_BGO3)) continue;
            std::printf("%02d-%s\n", i, AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
            if ((asyncIn == AltiModule::SWX_TTR3) || (asyncIn == AltiModule::RJ45_CALREQ2) || (asyncIn == AltiModule::CTP_CALREQ2) || (asyncIn == AltiModule::ALTI_CALREQ2) || (asyncIn == AltiModule::CDR_DERR)) printf("-------------\n");
        }

        if (!group) { // single signal
            asyncIn = (AltiModule::ASYNC_INPUT_SIGNAL) enterInt("Single signal                                              ", 0, AltiModule::ASYNC_INPUT_SIGNAL_NUMBER - 1);
            enable = (bool) enterInt("Enable sync (0=disable, 1=enable)                           ", 0, 1);
            //phase = enterInt("Clock phase (0=\"AUTO\", 1=\"000DEG\", 2=\"090DEG\", 3=\"180DEG\", 4=\"270DEG\")", 0, 4);
            phase = enterInt("Clock phase (1=\"000DEG\", 2=\"090DEG\", 3=\"180DEG\", 4=\"270DEG\")", 1, 4);

            if ((rtnv = alti->SIGInputSyncEnableWrite(asyncIn, enable, phase)) != AltiModule::SUCCESS) {
                CERR("%s sync and selecting %s clock phase for asynchronous input signal \"%s\"", enable ? "enabling" : "disabling", (phase == 0) ? "auto" : ((phase == 1) ? "0 degrees" : ((phase == 2) ? "90 degrees" : ((phase == 3) ? "180 degrees" : (phase == 4) ? "270 degrees" : ""))), AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
                return (rtnv);
            }
        }
        else if (group) { // group of signals
            //unsigned int group_choice = enterInt("Group (0=\"SWX\", 1=\"LEMO\", 2=\"CTP\", 3=\"ALTI\", 4=\"CDR\", 5=all)          ", 0, 5);
            unsigned int group_choice = enterInt("Group (0=\"SWX\", 1=\"CTP\", 2=\"ALTI\", 3=\"CDR\", 4=all)          ", 0, 4);
            enable = (bool) enterInt("Enable sync (0=disable, 1=enable)                           ", 0, 1);
            //phase = enterInt("Clock phase (0=\"AUTO\", 1=\"000DEG\", 2=\"090DEG\", 3=\"180DEG\", 4=\"270DEG\")", 0, 4);
            phase = enterInt("Clock phase (1=\"000DEG\", 2=\"090DEG\", 3=\"180DEG\", 4=\"270DEG\")", 1, 4);

            switch (group_choice) {
                case 0:
                    for (asyncIn = AltiModule::SWX_ORB; (asyncIn >= AltiModule::SWX_ORB) && (asyncIn <= AltiModule::SWX_TTR3); asyncIn = (AltiModule::ASYNC_INPUT_SIGNAL) (asyncIn + 1)) {
                        if ((rtnv = alti->SIGInputSyncEnableWrite(asyncIn, enable, phase)) != AltiModule::SUCCESS) {
                            CERR("%s sync and selecting %s clock phase for asynchronous input signal \"%s\"", enable ? "enabling" : "disabling", (phase == 0) ? "auto" : ((phase == 1) ? "0 degrees" : ((phase == 2) ? "90 degrees" : ((phase == 3) ? "180 degrees" : (phase == 4) ? "270 degrees" : ""))), AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
                            return (rtnv);
                        }
                    }
                    break;
                /*case 1:
                    for (asyncIn = AltiModule::LEMO_ORB; (asyncIn >= AltiModule::LEMO_ORB) && (asyncIn <= AltiModule::RJ45_CALREQ2); asyncIn = (AltiModule::ASYNC_INPUT_SIGNAL) (asyncIn + 1)) {
                        if ((rtnv = alti->SIGInputSyncEnableWrite(asyncIn, enable, phase)) != AltiModule::SUCCESS) {
                            CERR("%s sync and selecting %s clock phase for asynchronous input signal \"%s\"", enable ? "enabling" : "disabling", (phase == 0) ? "auto" : ((phase == 1) ? "0 degrees" : ((phase == 2) ? "90 degrees" : ((phase == 3) ? "180 degrees" : (phase == 4) ? "270 degrees" : ""))), AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
                            return (rtnv);
                        }
                    }
                    break;*/
                case 1: // 2
                    for (asyncIn = AltiModule::CTP_BUSY; (asyncIn >= AltiModule::CTP_BUSY) && (asyncIn <= AltiModule::CTP_CALREQ2); asyncIn = (AltiModule::ASYNC_INPUT_SIGNAL) (asyncIn + 1)) {
                        if ((rtnv = alti->SIGInputSyncEnableWrite(asyncIn, enable, phase)) != AltiModule::SUCCESS) {
                            CERR("%s sync and selecting %s clock phase for asynchronous input signal \"%s\"", enable ? "enabling" : "disabling", (phase == 0) ? "auto" : ((phase == 1) ? "0 degrees" : ((phase == 2) ? "90 degrees" : ((phase == 3) ? "180 degrees" : (phase == 4) ? "270 degrees" : ""))), AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
                            return (rtnv);
                        }
                    }
                    break;
                case 2: // 3
                    for (asyncIn = AltiModule::ALTI_BUSY; (asyncIn >= AltiModule::ALTI_BUSY) && (asyncIn <= AltiModule::ALTI_CALREQ2); asyncIn = (AltiModule::ASYNC_INPUT_SIGNAL) (asyncIn + 1)) {
                        if ((rtnv = alti->SIGInputSyncEnableWrite(asyncIn, enable, phase)) != AltiModule::SUCCESS) {
                            CERR("%s sync and selecting %s clock phase for asynchronous input signal \"%s\"", enable ? "enabling" : "disabling", (phase == 0) ? "auto" : ((phase == 1) ? "0 degrees" : ((phase == 2) ? "90 degrees" : ((phase == 3) ? "180 degrees" : (phase == 4) ? "270 degrees" : ""))), AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
                            return (rtnv);
                        }
                    }
                    break;
                case 3: // 4
                    for (asyncIn = AltiModule::CDR_LOCK; (asyncIn >= AltiModule::CDR_LOCK) && (asyncIn <= AltiModule::CDR_DERR); asyncIn = (AltiModule::ASYNC_INPUT_SIGNAL) (asyncIn + 1)) {
                        if ((rtnv = alti->SIGInputSyncEnableWrite(asyncIn, enable, phase)) != AltiModule::SUCCESS) {
                            CERR("%s sync and selecting %s clock phase for asynchronous input signal \"%s\"", enable ? "enabling" : "disabling", (phase == 0) ? "auto" : ((phase == 1) ? "0 degrees" : ((phase == 2) ? "90 degrees" : ((phase == 3) ? "180 degrees" : (phase == 4) ? "270 degrees" : ""))), AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
                            return (rtnv);
                        }
                    }
                    break;
                case 4: // 5
                    if ((rtnv = alti->SIGInputSyncEnableWrite(enable, phase)) != AltiModule::SUCCESS) {
                        CERR("%s sync and selecting %s clock phase for all asynchronous input signals", enable ? "enabling" : "disabling", (phase == 0) ? "auto" : ((phase == 1) ? "0 degrees" : ((phase == 2) ? "90 degrees" : ((phase == 3) ? "180 degrees" : (phase == 4) ? "270 degrees" : ""))));
                        return (rtnv);
                    }
                    break;
                default:
                    break;
            }
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiSIGInputShapeStatusRead : public MenuItem {
  public:
    AltiSIGInputShapeStatusRead() { setName("read  IO shape  status"); }
    int action() {

        int rtnv(0);

        //table
        printf("\n");
        printf("+--------------------------------------+\n");
        printf("| Asynchronous input signal | Shape    |\n");
        printf("+--------------------------------------+\n");

        std::vector<bool> enabled;
        if ((rtnv = alti->SIGInputShapeEnableRead(enabled)) != AltiModule::SUCCESS) {
            CERR("reading shaping status for all asynchronous input signals", "");
            return (rtnv);
        }

        unsigned int i;
        AltiModule::ASYNC_INPUT_SIGNAL asyncIn;
        for (i = 0; i < AltiModule::ASYNC_INPUT_SIGNAL_NUMBER; i++) {
            asyncIn = AltiModule::ASYNC_INPUT_SIGNAL(i);
            // LEMO sync inputs: removed!
            if ((asyncIn >= AltiModule::LEMO_ORB) && (asyncIn <= AltiModule::LEMO_BGO3)) continue;
            printf("| [%02d]%-21s | %-8s |\n", i, AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str(), enabled[i] ? "Enabled" : "Disabled");
            if ((asyncIn == AltiModule::SWX_TTR3) || (asyncIn == AltiModule::RJ45_CALREQ2) || (asyncIn == AltiModule::CTP_CALREQ2) || (asyncIn == AltiModule::ALTI_CALREQ2) || (asyncIn == AltiModule::CDR_DERR)) printf("+--------------------------------------+\n");
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiSIGInputShapeEnableWrite : public MenuItem {
  public:
    AltiSIGInputShapeEnableWrite() { setName("write IO shape  enable"); }
    int action() {

        int rtnv(0);

        bool group = (bool) enterInt("Single signal or group of signals (0=single, 1=group)", 0, 1);
        AltiModule::ASYNC_INPUT_SIGNAL asyncIn;
        bool enable;

        printf("-------------\n");
        for (unsigned i = 0; i < AltiModule::ASYNC_INPUT_SIGNAL_NUMBER; i++) {
            asyncIn = AltiModule::ASYNC_INPUT_SIGNAL(i);
            // LEMO sync inputs: removed!
            if ((asyncIn >= AltiModule::LEMO_ORB) && (asyncIn <= AltiModule::LEMO_BGO3)) continue;
            std::printf("%02d-%s\n", i, AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
            if ((asyncIn == AltiModule::SWX_TTR3) || (asyncIn == AltiModule::RJ45_CALREQ2) || (asyncIn == AltiModule::CTP_CALREQ2) || (asyncIn == AltiModule::ALTI_CALREQ2) || (asyncIn == AltiModule::CDR_DERR)) printf("-------------\n");
        }

        if (!group) { // single signal
            asyncIn = (AltiModule::ASYNC_INPUT_SIGNAL) enterInt("Single signal                                       ", 0, AltiModule::ASYNC_INPUT_SIGNAL_NUMBER - 1);
            enable = (bool) enterInt("Enable shape (0-disable, 1=enable)                   ", 0, 1);

            if (alti->SIGInputShapeEnableWrite(asyncIn, enable) != 0) {
                CERR("%s shaping for asynchronous input signal \"%s\"", enable ? "enabling" : "disabling", AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
                return (-1);
            }
        }
        else if (group) { // group of signals
            //unsigned int group_choice = enterInt("Group (0=\"SWX\", 1=\"LEMO\", 2=\"CTP\", 3=\"ALTI\", 4=\"CDR\", 5=all)", 0, 5);
            unsigned int group_choice = enterInt("Group (0=\"SWX\", 1=\"CTP\", 2=\"ALTI\", 3=\"CDR\", 4=all)   ", 0, 4);
            enable = (bool) enterInt("Enable shape (0-disable, 1=enable)                   ", 0, 1);

            switch (group_choice) {
                case 0:
                    for (asyncIn = AltiModule::SWX_ORB; (asyncIn >= AltiModule::SWX_ORB) && (asyncIn <= AltiModule::SWX_TTR3); asyncIn = (AltiModule::ASYNC_INPUT_SIGNAL) (asyncIn + 1)) {
                        if ((rtnv = alti->SIGInputShapeEnableWrite(asyncIn, enable)) != AltiModule::SUCCESS) {
                            CERR("%s shaping for asynchronous input signal \"%s\"", enable ? "enabling" : "disabling", AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
                            return (rtnv);
                        }
                    }
                    break;
                /*case 1:
                    for (asyncIn = AltiModule::LEMO_ORB; (asyncIn >= AltiModule::LEMO_ORB) && (asyncIn <= AltiModule::RJ45_CALREQ2); asyncIn = (AltiModule::ASYNC_INPUT_SIGNAL) (asyncIn + 1)) {
                        if ((rtnv = alti->SIGInputShapeEnableWrite(asyncIn, enable)) != AltiModule::SUCCESS) {
                            CERR("%s shaping for asynchronous input signal \"%s\"", enable ? "enabling" : "disabling", AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
                            return (rtnv);
                        }
                    }
                    break;*/
                case 1: // 2
                    for (asyncIn = AltiModule::CTP_BUSY; (asyncIn >= AltiModule::CTP_BUSY) && (asyncIn <= AltiModule::CTP_CALREQ2); asyncIn = (AltiModule::ASYNC_INPUT_SIGNAL) (asyncIn + 1)) {
                        if ((rtnv = alti->SIGInputShapeEnableWrite(asyncIn, enable)) != AltiModule::SUCCESS) {
                            CERR("%s shaping for asynchronous input signal \"%s\"", enable ? "enabling" : "disabling", AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
                            return (rtnv);
                        }
                    }
                    break;
                case 2: // 3
                    for (asyncIn = AltiModule::ALTI_BUSY; (asyncIn >= AltiModule::ALTI_BUSY) && (asyncIn <= AltiModule::ALTI_CALREQ2); asyncIn = (AltiModule::ASYNC_INPUT_SIGNAL) (asyncIn + 1)) {
                        if ((rtnv = alti->SIGInputShapeEnableWrite(asyncIn, enable)) != AltiModule::SUCCESS) {
                            CERR("%s shaping for asynchronous input signal \"%s\"", enable ? "enabling" : "disabling", AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
                            return (rtnv);
                        }
                    }
                    break;
                case 3: // 4
                    for (asyncIn = AltiModule::CDR_LOCK; (asyncIn >= AltiModule::CDR_LOCK) && (asyncIn <= AltiModule::CDR_DERR); asyncIn = (AltiModule::ASYNC_INPUT_SIGNAL) (asyncIn + 1)) {
                        if ((rtnv = alti->SIGInputShapeEnableWrite(asyncIn, enable)) != AltiModule::SUCCESS) {
                            CERR("%s shaping for asynchronous input signal \"%s\"", enable ? "enabling" : "disabling", AltiModule::ASYNC_INPUT_SIGNAL_NAME[asyncIn].c_str());
                            return (rtnv);
                        }
                    }
                    break;
                case 4: // 5
                    if ((rtnv = alti->SIGInputShapeEnableWrite(enable)) != AltiModule::SUCCESS) {
                            CERR("%s shaping for all asynchronous input signals", enable ? "enabling" : "disabling");
                        return (rtnv);
                    }
                    break;
                default:
                    break;
            }
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiSIGMuxBGOOutputLEMORead : public MenuItem {
  public:
    AltiSIGMuxBGOOutputLEMORead() { setName("read  front panel BGO selection"); }
    int action() {

        int rtnv(0);

        bool bgo2enabled, bgo3enabled;
        if ((rtnv = alti->SIGMuxBGOOutputLEMORead(bgo2enabled, bgo3enabled)) != AltiModule::SUCCESS) {
            CERR("reading front panel output BGO selection", "");
            return (rtnv);
        }

        //table
        printf("\n");
        printf("+----------------------+\n");
        printf("| BGO2 FP output: BGO%d |\n", bgo2enabled ? 2 : 0);
        printf("| BGO3 FP output: BGO%d |\n", bgo3enabled ? 3 : 1);
        printf("+----------------------+\n");

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiSIGMuxBGOOutputLEMOWrite : public MenuItem {
  public:
    AltiSIGMuxBGOOutputLEMOWrite() { setName("write front panel BGO selection"); }
    int action() {

        int rtnv(0);

//        unsigned int bgo_fp_sel = enterInt("BGO2/3 FP output selection [0-01/1-03/2-21/3-23]", 0, 3);
        unsigned int bgo_fp_sel = enterInt("BGO2/3 FP output selection (0=\"BGO0/1\", 1=\"BGO0/3\", 2=\"BGO2/1\", 3=\"BGO2/3\")", 0, 3);
        bool bgo2enable = bgo_fp_sel & 0x2;
        bool bgo3enable = bgo_fp_sel & 0x1;
        if ((rtnv = alti->SIGMuxBGOOutputLEMOWrite(bgo2enable, bgo3enable)) != AltiModule::SUCCESS) {
            CERR("writing front panel output BGO selection: \"BGO%d\" and \"BGO%d\"", bgo2enable ? 2 : 0, bgo3enable ? 3 : 1);
            return (rtnv);
        }


        return (rtnv);
    }
};


//------------------------------------------------------------------------------

class AltiSIGMuxTTRInputLEMORead : public MenuItem {
  public:
    AltiSIGMuxTTRInputLEMORead() { setName("read  front panel TTR input selection"); }
    int action() {

        int rtnv(0);

	printf("\n");
	for (u_int bit = 0; bit < 3; bit++) {
	  AltiModule::TTR_LEMO input;
	  if ((rtnv = alti->SIGMuxTTRInputLEMORead(bit, input)) != AltiModule::SUCCESS) {
            CERR("reading TTR lemo input selection", "");
            return (rtnv);
	  }
	  printf("TTR%d: %s\n", bit+1, AltiModule::TTR_LEMO_NAME[input].c_str());
        }
        printf("\n");
	
        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiSIGMuxTTRInputLEMOWrite : public MenuItem {
  public:
    AltiSIGMuxTTRInputLEMOWrite() { setName("write front panel TTR input selection"); }
    int action() {

        int rtnv(0);

        u_int bit;
        u_int bit_choice = enterInt("TTR bit (1=\"TTR1\", 2=\"TTR2\", 3=\"TTR3\", 4-all)      ", 1, 4);

        AltiModule::TTR_LEMO input = (AltiModule::TTR_LEMO) enterInt("Input (0=\"TTR\", 1=\"CALREQ\")", 0, AltiModule::TTR_LEMO_NUMBER - 1);

        if (bit_choice == 4) { // all bits
	  for (u_int bit = 0; bit < 3; bit++) {
	    if ((rtnv = alti->SIGMuxTTRInputLEMOWrite(bit, input)) != AltiModule::SUCCESS) {
	      CERR("writing TTR1-3 lemo input selection", "");
	      return (rtnv);
            }
	  }
	}
        else { // 1 bit
	  bit = bit_choice-1;
	  if ((rtnv = alti->SIGMuxTTRInputLEMOWrite(bit, input)) != AltiModule::SUCCESS) {
                CERR("writing TTR%d lemo input selection", bit);
                return (rtnv);
	  }
        }
	
        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiSIGMuxTTROutputLEMORead : public MenuItem {
  public:
    AltiSIGMuxTTROutputLEMORead() { setName("read  front panel TTR output selection"); }
    int action() {

        int rtnv(0);

	printf("\n");
	for (u_int bit = 0; bit < 3; bit++) {
	  AltiModule::TTR_LEMO output;
	  if ((rtnv = alti->SIGMuxTTROutputLEMORead(bit, output)) != AltiModule::SUCCESS) {
            CERR("reading TTR lemo output selection", "");
            return (rtnv);
	  }
	  printf("TTR%d: %s\n", bit+1, AltiModule::TTR_LEMO_NAME[output].c_str());
        }
        printf("\n");
	
        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiSIGMuxTTROutputLEMOWrite : public MenuItem {
  public:
    AltiSIGMuxTTROutputLEMOWrite() { setName("write front panel TTR output selection"); }
    int action() {

        int rtnv(0);

        u_int bit;
        u_int bit_choice = enterInt("TTR bit (1=\"TTR1\", 2=\"TTR2\", 3=\"TTR3\", 4-all)      ", 1, 4);

        AltiModule::TTR_LEMO output = (AltiModule::TTR_LEMO) enterInt("output (0=\"TTR\", 1=\"CALREQ\")", 0, AltiModule::TTR_LEMO_NUMBER - 1);

        if (bit_choice == 4) { // all bits
	  for (u_int bit = 0; bit < 3; bit++) {
	    if ((rtnv = alti->SIGMuxTTROutputLEMOWrite(bit, output)) != AltiModule::SUCCESS) {
	      CERR("writing TTR1-3 lemo output selection", "");
	      return (rtnv);
            }
	  }
	}
        else { // 1 bit
	  bit = bit_choice-1;
	  if ((rtnv = alti->SIGMuxTTROutputLEMOWrite(bit, output)) != AltiModule::SUCCESS) {
                CERR("writing TTR%d lemo output selection", bit);
                return (rtnv);
	  }
        }
	
        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiBSYFrontPanelLevelWrite : public MenuItem {
  public:
   
    AltiBSYFrontPanelLevelWrite() { setName("write   Front-Panel level"); }
    int action() {
        int rtnv(0);
	AltiModule::BUSY_LEVEL level = AltiModule::NIM;
        u_int level_choice = enterInt("Level (0=NIM, 1=TTL)", 0, 1);
	if (level_choice < AltiModule::BUSY_LEVEL_NUMBER) level = (AltiModule::BUSY_LEVEL) level_choice; 
        if ((rtnv = alti->BSYFrontPanelLevelWrite(level)) != AltiModule::SUCCESS) {
	  CERR("writing Busy Front-Panel level to %s", AltiModule::BUSY_LEVEL_NAME[level].c_str());
	  return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiBSYInputSelectWrite : public MenuItem {
  public:
   
    AltiBSYInputSelectWrite() { setName("write   input  selection"); }
    int action() {

        int rtnv(0);

        AltiModule::BUSY_INPUT input;
        unsigned int input_choice = enterInt("Input (0=\"FRONT_PANEL\", 1=\"CTP\", 2=\"ALTI\", 3=\"PG\", 4=\"VME\" 5=\"ECR\", 6=all)", 0, AltiModule::BUSY_INPUT_NUMBER - 1 + 1);
        if (input_choice < AltiModule::BUSY_INPUT_NUMBER) input = (AltiModule::BUSY_INPUT) input_choice;

        if (input_choice < AltiModule::BUSY_INPUT_NUMBER) { // 1 input
            std::printf("%-21s ", AltiModule::BUSY_INPUT_NAME[input].c_str());
            bool select = (bool) enterInt("enable (0=disable, 1=enable)                ", 0, 1);
            if ((rtnv = alti->BSYInputSelectWrite(input, select)) != AltiModule::SUCCESS) {
                CERR("%s busy input %s", select ? "enabling" : "disabling", AltiModule::BUSY_INPUT_NAME[input].c_str());
                return (rtnv);
            }
        }
        else { // all inputs
            std::vector<bool> select;
            select.clear();
            unsigned int i;
            for (i = 0; i < AltiModule::BUSY_INPUT_NUMBER; i++) {
                input = (AltiModule::BUSY_INPUT) i;
                std::printf("%-21s ", AltiModule::BUSY_INPUT_NAME[input].c_str());
                select.push_back((bool) enterInt("enable (0=disable, 1=enable)                ", 0, 1));
            }
            if ((rtnv = alti->BSYInputSelectWrite(select)) != AltiModule::SUCCESS) {
                CERR("writing selection for all busy inputs","");
                return (rtnv);
            }
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiBSYOutputSelectWrite : public MenuItem {
  public:
    //AltiBSYOutputSelectWrite() { setName("write BUSY output selection"); }
    AltiBSYOutputSelectWrite() { setName("write   output selection"); }
    int action() {

        int rtnv(0);

        AltiModule::BUSY_OUTPUT output;
        unsigned int output_choice = enterInt("Output (0=\"CTP\", 1=\"ALTI\", 2=all)                               ", 0, AltiModule::BUSY_OUTPUT_NUMBER - 1 + 1);
        if (output_choice < AltiModule::BUSY_OUTPUT_NUMBER) output = (AltiModule::BUSY_OUTPUT) output_choice;

        if (output_choice < AltiModule::BUSY_OUTPUT_NUMBER) { // 1 output
            std::printf("%-12s ", AltiModule::BUSY_OUTPUT_NAME[output].c_str());
            AltiModule::BUSY_SOURCE src = (AltiModule::BUSY_SOURCE) enterInt("source (0=\"Inactive\", 1=\"Local\", 2=\"CTP\", 3=\"ALTI\")", 0, AltiModule::BUSY_SOURCE_NUMBER - 1);
            if ((rtnv = alti->BSYOutputSelectWrite(output, src)) != AltiModule::SUCCESS) {
                CERR("selecting source %s for busy output %s", AltiModule::BUSY_SOURCE_NAME[src].c_str(), AltiModule::BUSY_OUTPUT_NAME[output]);
                return (rtnv);
            }
        }
        else { // all outputs
            std::vector<AltiModule::BUSY_SOURCE> src;
            src.clear();
            unsigned int i;
            for (i = 0; i < AltiModule::BUSY_OUTPUT_NUMBER; i++) {
                output = (AltiModule::BUSY_OUTPUT) i;
                std::printf("%-12s ", AltiModule::BUSY_OUTPUT_NAME[output].c_str());
                src.push_back((AltiModule::BUSY_SOURCE) enterInt("source (0=\"Inactive\", 1=\"Local\", 2=\"CTP\", 3=\"ALTI\")", 0, AltiModule::BUSY_SOURCE_NUMBER - 1));
            }
            if ((rtnv = alti->BSYOutputSelectWrite(src)) != AltiModule::SUCCESS) {
                CERR("selecting sources for all busy ouputs","");
                return (rtnv);
            }
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiBSYConstLevelRegWrite : public MenuItem {
  public:
    //AltiBSYConstLevelRegWrite() { setName("write BUSY constant level register"); }
    AltiBSYConstLevelRegWrite() { setName("write   constant level register"); }
    int action() {

        int rtnv(0);

        bool busy = (bool) enterInt("Busy (0=inactive, 1=active)", 0, 1);
        if ((rtnv = alti->BSYConstLevelRegWrite(busy)) != AltiModule::SUCCESS) {
            CERR("writing %s to constant level busy register", busy ? "Active(1)" : "Inactive(0)");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiBSYMasking : public MenuItem {
  public:
    AltiBSYMasking() { setName("write   busy gating of PG signals"); }
    int action() {
      int rtnv = 0;

      unsigned int signal_pg_choice = enterInt("(0=\"ORB\", 1=\"L1A\", 2..4=\"TTR1..3\", 5=\"TTYP\", 6..9=\"BGO0..3\")", 0, AltiModule::SIGNAL_PG_NUMBER - 1);
      bool enable = (bool) enterInt("Enable (0=disable, 1=enable)                                         ", 0, 1);

      if (signal_pg_choice == 0){
	if ((rtnv = alti->BSYMaskingOrbEnableWrite(enable)) != AltiModule::SUCCESS) {
	  CERR("%s busy gating of Orbit", enable ? "enabling" : "disabling");
	  return (rtnv);
	}	
      }
      else if (signal_pg_choice == 1){
	if ((rtnv = alti->BSYMaskingL1aEnableWrite(enable)) != AltiModule::SUCCESS) {
	  CERR("%s busy gating of L1A", enable ? "enabling" : "disabling");
	  return (rtnv);
	}
      } 
      else if (signal_pg_choice < 5)
	{
	  int ttr_choice = signal_pg_choice - 1;
	  if ((rtnv = alti->BSYMaskingTTREnableWrite(ttr_choice, enable)) != AltiModule::SUCCESS) {
	    CERR("%s busy gating of \"%s\"", enable ? "enabling" : "disabling", AltiModule::SIGNAL_PG_NAME[(AltiModule::SIGNAL_PG) signal_pg_choice].c_str());
	    return (rtnv);
	  }
	}
      else if (signal_pg_choice == 5){
	if ((rtnv = alti->BSYMaskingTTypeEnableWrite(enable)) != AltiModule::SUCCESS) {
	  CERR("%s busy gating of TType",  enable ? "enabling" : "disabling");
	  return (rtnv);
	}
      } 
      else 
        {
	  int bgo_choice = signal_pg_choice - 6;
          if ((rtnv = alti->BSYMaskingBGoEnableWrite(bgo_choice, enable)) != AltiModule::SUCCESS) {
            CERR("%s busy gating of \"%s\"", enable ? "enabling" : "disabling", AltiModule::SIGNAL_PG_NAME[(AltiModule::SIGNAL_PG) signal_pg_choice].c_str());
            return (rtnv);
          }
        }
      
      return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiBSYMaskingRead : public MenuItem {
  public:
    AltiBSYMaskingRead() { setName("read  busy gating of PG signals"); }
    int action() {

        int rtnv(0);
        bool l1a_en,orb_en,tty_en;
        bool ttr1,ttr2,ttr3;
        bool bgo0,bgo1,bgo2,bgo3;
        if ((rtnv = alti->BSYMaskingL1aEnableRead(l1a_en)) != AltiModule::SUCCESS) {
            CERR("reading busy masking of L1A", "");
            return (rtnv);
        }
        if ((rtnv = alti->BSYMaskingOrbEnableRead(orb_en)) != AltiModule::SUCCESS) {
            CERR("reading busy masking of Orbit", "");
            return (rtnv);
        }
        if ((rtnv = alti->BSYMaskingTTypeEnableRead(tty_en)) != AltiModule::SUCCESS) {
            CERR("reading busy masking of TType", "");
            return (rtnv);
        } 
        if ((rtnv = alti->BSYMaskingTTREnableRead(ttr1,ttr2,ttr3)) != AltiModule::SUCCESS) {
          CERR("reading busy masking of TTriggers", "");
          return (rtnv);
        }  
        if ((rtnv = alti->BSYMaskingBGoEnableRead(bgo0,bgo1,bgo2,bgo3)) != AltiModule::SUCCESS) {
          CERR("reading busy masking of BGos", "");
          return (rtnv);
        }
        printf("  L1A   : %s\n  Orbit : %s\n  TType : %s\n  TTR1  : %s\n  TTR2  : %s\n  TTR3  : %s\n  BGo0  : %s\n  BGo1  : %s\n  BGo2  : %s\n  BGo3  : %s",
        l1a_en ? "Enabled" : "Disabled",orb_en ? "Enabled" : "Disabled",tty_en ? "Enabled" : "Disabled",ttr1 ? "Enabled" : "Disabled",ttr2 ? "Enabled" : "Disabled",ttr3 ? "Enabled" : "Disabled",bgo0 ? "Enabled" : "Disabled",bgo1 ? "Enabled" : "Disabled",bgo2 ? "Enabled" : "Disabled",bgo3 ? "Enabled" : "Disabled");
        

        return (rtnv);
    }
};
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

class AltiCRQInputSelectRead : public MenuItem {
  public:
    AltiCRQInputSelectRead() { setName("read  CALREQ input  selection"); }
    int action() {

        int rtnv(0);

        std::vector <AltiModule::CALREQ_INPUT> input;
        input.resize(3);
        if ((rtnv = alti->CRQInputSelectRead(input)) != AltiModule::SUCCESS) {
            CERR("reading creq input selection", "");
            return (rtnv);
        }

        printf("\n");
        unsigned int bit;
        for (bit = 0; bit < 3; bit++) {
            printf("CALREQ[%d]: %s\n", bit, AltiModule::CALREQ_INPUT_NAME[input[bit]].c_str());
        }
        printf("\n");

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCRQInputSelectWrite : public MenuItem {
  public:
    AltiCRQInputSelectWrite() { setName("write CALREQ input  selection"); }
    int action() {

        int rtnv(0);

        unsigned int bit;
        unsigned int bit_choice = enterInt("CALREQ bit (0=\"CALREQ[0]\", 1=\"CALREQ[1]\", 2=\"CALREQ[2]\", 3-all)      ", 0, 3);

        AltiModule::CALREQ_INPUT input = (AltiModule::CALREQ_INPUT) enterInt("Input (0=\"RJ45\", 1=\"FRONT_PANEL\", 2=\"CTP\", 3=\"ALTI\", 4=\"PG\", 5=\"VME\")", 0, AltiModule::CALREQ_INPUT_NUMBER - 1);

        if (bit_choice == 3) { // all bits
            if ((rtnv = alti->CRQInputSelectWrite(input)) != AltiModule::SUCCESS) {
                CERR("writing creq[2:0] input selection", "");
                return (rtnv);
            }
        }
        else { // 1 bit
            bit = bit_choice;
            if ((rtnv = alti->CRQInputSelectWrite(bit, input)) != AltiModule::SUCCESS) {
                CERR("writing creq[%d] input selection", bit);
                return (rtnv);
            }
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCRQOutputSelectRead : public MenuItem {
  public:
    AltiCRQOutputSelectRead() { setName("read  CALREQ output selection"); }
    int action() {

        int rtnv(0);

        std::vector<AltiModule::CALREQ_SOURCE> src;
        src.resize(3);

        printf("\n");
        printf("+-------------------------------------------------------------------------+\n");
        printf("| Output            | CALREQ[2]       | CALREQ[1]       | CALREQ[0]       |\n");
        printf("+-------------------------------------------------------------------------+\n");

        unsigned int i;
        AltiModule::CALREQ_OUTPUT output;
        for (i = 0; i < AltiModule::CALREQ_OUTPUT_NUMBER; i++) {
            output = AltiModule::CALREQ_OUTPUT(i);
            if ((rtnv = alti->CRQOutputSelectRead(output, src)) != AltiModule::SUCCESS) {
                CERR("reading creq output selection for %s", AltiModule::CALREQ_OUTPUT_NAME[output].c_str());
                return (rtnv);
            }
            printf("| [%d]%-14s | %-15s | %-15s | %-15s |\n", i, AltiModule::CALREQ_OUTPUT_NAME[output].c_str(), AltiModule::CALREQ_SOURCE_NAME[src[2]].c_str(),  AltiModule::CALREQ_SOURCE_NAME[src[1]].c_str(),  AltiModule::CALREQ_SOURCE_NAME[src[0]].c_str());
            printf("+-------------------------------------------------------------------------+\n");
        }
        printf("\n");

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCRQOutputSelectWrite : public MenuItem {
  public:
    AltiCRQOutputSelectWrite() { setName("write CALREQ output selection"); }
    int action() {

        int rtnv(0);

        unsigned int bit;
        unsigned int bit_choice = enterInt("CALREQ bit (0=\"CALREQ[0]\", 1=\"CALREQ[1]\", 2=\"CALREQ[2]\", 3-all)", 0, 3);
        if (bit_choice < 3) bit = bit_choice;

        AltiModule::CALREQ_OUTPUT output;
        unsigned int output_choice = enterInt("Output (0=\"CTP\", 1=\"ALTI\", 2=all)                              ", 0, AltiModule::CALREQ_OUTPUT_NUMBER - 1 + 1);
        if (output_choice < AltiModule::CALREQ_OUTPUT_NUMBER) output = (AltiModule::CALREQ_OUTPUT) output_choice;

        AltiModule::CALREQ_SOURCE src = (AltiModule::CALREQ_SOURCE) enterInt("Source (0=\"Inactive\", 1=\"Local\", 2=\"CTP\", 3=\"ALTI\")            ", 0, AltiModule::CALREQ_SOURCE_NUMBER - 1);

        if ((bit_choice < 3) && output_choice < AltiModule::CALREQ_OUTPUT_NUMBER) { // 1 output, 1 bit
            if ((rtnv = alti->CRQOutputSelectWrite(output, bit, src)) != AltiModule::SUCCESS) {
                CERR("selecting source %s for creq[%d] output %s", AltiModule::CALREQ_SOURCE_NAME[src].c_str(), bit, AltiModule::CALREQ_OUTPUT_NAME[output]);
                return (rtnv);
            }
        }
        else if (bit_choice < 3) { // all outputs, 1 bit
            if ((rtnv = alti->CRQOutputSelectWrite(bit, src)) != AltiModule::SUCCESS) {
                CERR("selecting sources for all creq[%d] outputs", bit);
                return (rtnv);
            }
        }
        else if (output_choice < AltiModule::CALREQ_OUTPUT_NUMBER) { // 1 output, all bits
            if ((rtnv = alti->CRQOutputSelectWrite(output, src)) != AltiModule::SUCCESS) {
                CERR("selecting sources for creq[2:0] output %s", AltiModule::CALREQ_OUTPUT_NAME[output].c_str());
                return (rtnv);
            }
        }
        else { // all outputs, all bits
            if ((rtnv = alti->CRQOutputSelectWrite(src)) != AltiModule::SUCCESS) {
                CERR("selecting sources for all creq[2:0] outputs", "");
                return (rtnv);
            }
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCRQConstLevelRegRead : public MenuItem {
  public:
    AltiCRQConstLevelRegRead() { setName("read  CALREQ constant level register"); }
    int action() {

        int rtnv(0);

        unsigned int creq;
        if ((rtnv = alti->CRQConstLevelRegRead(creq)) != AltiModule::SUCCESS) {
            CERR("reading constant level creq register","");
            return (rtnv);
        }

        printf("\n");
        printf("%d\n", creq);
        printf("\n");

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCRQConstLevelRegWrite : public MenuItem {
  public:
    AltiCRQConstLevelRegWrite() { setName("write CALREQ constant level register"); }
    int action() {

        int rtnv(0);

        unsigned int creq = enterInt("CALREQ", 0, 7);
        if ((rtnv = alti->CRQConstLevelRegWrite(creq)) != AltiModule::SUCCESS) {
            CERR("writing %d to constant level creq register", creq);
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCRQInputRead : public MenuItem {
  public:
    AltiCRQInputRead() { setName("read  CALREQ inputs"); }
    int action() {

        int rtnv(0);

        std::vector<unsigned int> creq;
        creq.resize(AltiModule::CALREQ_INPUT_NUMBER);
        if ((rtnv = alti->CRQInputRead(creq)) != AltiModule::SUCCESS) {
            CERR("reading creq status", "");
            return (rtnv);
        }

        printf("\n");
        unsigned int i;
        AltiModule::CALREQ_INPUT input;
        for (i = 0; i < AltiModule::CALREQ_INPUT_NUMBER; i++) {
            input = AltiModule::CALREQ_INPUT(i);
            printf("[%02d]%-23s: %d\n", i, AltiModule::CALREQ_INPUT_NAME[input].c_str(), creq[i]);
        }
        printf("\n");

        return (rtnv);
    }
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

class AltiSNPEnable : public MenuItem {
  public:
    AltiSNPEnable() { setName("enable  snapshot"); }
    int action() {

        // no automatic disabling!
        /*if (alti->PATGenerationEnableWrite(false) != 0) {
            CERR("disabling generation from pattern memory for all signals", "");
            return (-1);
        }*/

        if (alti->SNPEnableWrite(true) != 0) {
            CERR("enabling snapshot memory", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiSNPDisable : public MenuItem {
  public:
    AltiSNPDisable() { setName("disable snapshot"); }
    int action() {

        if (alti->SNPEnableWrite(false) != 0) {
            CERR("disabling snapshot memory", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiSNPMaskWrite : public MenuItem {
  public:
    AltiSNPMaskWrite() { setName("write   snapshot mask"); }
    int action() {

        AltiModule::SIGNAL signal;
        unsigned int signal_choice = enterInt("Signal (1=\"ORB\", 2=\"L1A\", ..., 18=\"BUSY\", 19..21=\"CALREQ0..2\", 22=all)", 1, AltiModule::SIGNAL_NUMBER - 1 + 1);
        if (signal_choice < AltiModule::SIGNAL_NUMBER) signal = (AltiModule::SIGNAL) signal_choice;

        bool mask = (bool) enterInt("Mask (0=no, 1=yes)                                                     ", 0, 1);

        if (signal_choice < AltiModule::SIGNAL_NUMBER) { // 1 signal
            if (alti->SNPMaskWrite(signal, mask) != 0) {
                CERR("%s signal %s from the snapshot", mask ? "masking" : "unmasking", AltiModule::SIGNAL_NAME[signal].c_str());
            }
        }
        else { // all signals
            if (alti->SNPMaskWrite(mask) != 0) {
                CERR("%s all signals from the snapshot", mask ? "masking" : "unmasking");
            }
        }

        return(0);
    }
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

class AltiPATStatusRead : public MenuItem {
  public:
    AltiPATStatusRead() { setName("status"); }
    int action() {

        bool generationEnabled;
        if (alti->PATGenerationEnableRead(generationEnabled) != 0) {
            CERR("reading pattern generation enable", "");
            return (-1);
        }

        std::vector<bool> generationConfig;
        if (alti->PATGenerationConfigRead(generationConfig) != 0) {
            CERR("reading pattern generation configuration", "");
            return (-1);
        }

        bool generationRepeated;
        if (alti->PATGenerationRepeatRead(generationRepeated) != 0) {
            CERR("reading pattern generation repeat", "");
            return (-1);
        }

        //table
        printf("\n");
        printf("+------------------------------+\n");
        printf("| Signal  | Pattern generation |\n");
        printf("+------------------------------+\n");

        unsigned int i;
        AltiModule::SIGNAL_PG signal;
        for (i = 0; i < AltiModule::SIGNAL_PG_NUMBER; i++) {
            signal = (AltiModule::SIGNAL_PG) i;
            printf("| [%01d]%-4s | %-18s |\n", i, AltiModule::SIGNAL_PG_NAME[signal].c_str(), generationConfig[i] ? "Enabled" : "Disabled");
        }
        printf("+------------------------------+\n");

        std::printf("\n");
        std::printf("Pattern generation: %s, %s\n", generationEnabled ? ("Enabled") : ("Disabled"), generationRepeated ? ("Repeated") : ("OneShot"));

        unsigned int start;
        if (alti->PATGenerationStartAddressRead(start) != 0) {
            CERR("reading pattern generation start address", "");
            return (-1);
        }
        unsigned int stop;
        if (alti->PATGenerationStopAddressRead(stop) != 0) {
            CERR("reading pattern generation stop address", "");
            return (-1);
        }
        std::printf("\n");
        std::printf("Start address = 0x%08x (= %10u)\n", start, start);
        std::printf("Stop  address = 0x%08x (= %10u)\n", stop, stop);
        std::printf("\n");

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiPATGenerationConfigWrite : public MenuItem {
  public:
    AltiPATGenerationConfigWrite() { setName("configure pattern generation"); }
    int action() {

        unsigned int signal_pg_choice = enterInt("(0=\"ORB\", 1=\"L1A\", 2..4=\"TTR1..3\", 5=\"TTYP\", 6..9=\"BGO0..3\", 10=all)", 0, AltiModule::SIGNAL_PG_NUMBER - 1 + 1);

        bool enable = (bool) enterInt("Enable (0=disable, 1=enable)                                         ", 0, 1);

        // no automatic disabling!
        /*if (enable) {
            if (alti->SNPEnableWrite(false) != 0) {
                CERR("disabling snapshot", "");
                return (-1);
            }
        }*/
        if (signal_pg_choice < AltiModule::SIGNAL_PG_NUMBER) { // 1
            AltiModule::SIGNAL_PG signal_pg = (AltiModule::SIGNAL_PG) signal_pg_choice;
            if (alti->PATGenerationConfigWrite(signal_pg, enable) != 0) {
                CERR("%s generation from pattern memory for \"%s\"", enable ? "enabling" : "disabling", AltiModule::SIGNAL_PG_NAME[signal_pg].c_str());
                return (-1); 
            }
        }
        else { // all
            if (alti->PATGenerationConfigWrite(enable) != 0) {
                CERR("%s generation from pattern memory for all signals", enable ? "enabling" : "disabling");
                return (-1);
            }
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiPATGenerationEnable : public MenuItem {
  public:
    AltiPATGenerationEnable() { setName("enable    pattern generation"); }
    int action() {
        if (alti->PATGenerationEnableWrite(true) != 0) {
            CERR("enabling pattern generation", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiPATGenerationDisable : public MenuItem {
  public:
    AltiPATGenerationDisable() { setName("disable   pattern generation"); }
    int action() {
        if (alti->PATGenerationEnableWrite(false) != 0) {
            CERR("disabling pattern generation", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiPATGenerationRepeatWrite : public MenuItem {
  public:
    AltiPATGenerationRepeatWrite() { setName("repeat    pattern generation"); }
    int action() {

        bool repeat = (bool) enterInt("Enable repeated generation (0=\"OneShot\", 1=\"Repeated\")", 0, 1);

        if (alti->PATGenerationRepeatWrite(repeat) != 0) {
            CERR("%s repeated generation from pattern memory", repeat ? "enabling" : "disabling");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiPATGenerationStartAddressWrite : public MenuItem {
  public:
    AltiPATGenerationStartAddressWrite() { setName("write     pattern generation start address"); }
    int action() {

        unsigned int start;
        start = enterHex("start address", 0, AltiModule::MAX_MEMORY - 1);
        if (alti->PATGenerationStartAddressWrite(start) != 0) {
            CERR("writing pattern generation start address", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiPATGenerationStopAddressWrite : public MenuItem {
  public:
    AltiPATGenerationStopAddressWrite() { setName("write     pattern generation stop  address"); }
    int action() {

        unsigned int stop;
        stop = enterHex("stop address", 0, AltiModule::MAX_MEMORY - 1);
        if (alti->PATGenerationStopAddressWrite(stop) != 0) {
            CERR("writing pattern generation stop address", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiPATRead : public MenuItem {
  public:
    AltiPATRead() { setName("read      pattern generation memory"); }
    int action() {

        // no automatic disabling!
        /*if (alti->SNPEnableWrite(false) != 0) {
            CERR("disabling snapshot", "");
            return (-1);
        }*/

        if (alti->PATRead(segment[0]) != 0) {
            CERR("reading pattern generation memory.If the pattern generator is still enabled, disable it before reading the memory", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiPATWriteFile : public MenuItem {
  public:
    AltiPATWriteFile() { setName("write     pattern generation memory [from file]"); }
    int action() {

        // no automatic disabling!
        /*if (alti->SNPEnableWrite(false) != 0) {
            CERR("disabling snapshot", "");
            return (-1);
        }*/

        std::string fn = enterString("File name");
        std::ifstream inf(fn.c_str(), std::ifstream::in);
        if(!inf) {
            CERR("cannot open input file \"%s\"", fn.c_str());
            return (-1);
        }

        if (alti->PATWriteFile(segment[0], inf) != 0) {
            CERR("writing pattern generation memory from file. If the pattern generator is still enabled, disable it before reading the memory", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

class AltiSNPStatusRead : public MenuItem {
  public:
    AltiSNPStatusRead() { setName("status"); }
    int action() {

        //table
        printf("\n");
        printf("+--------------------------+\n");
        printf("| Signal      | Mask       |\n");
        printf("+--------------------------+\n");

        std::vector<bool> masked;
        masked.resize(AltiModule::SIGNAL_NUMBER);
        if (alti->SNPMaskRead(masked) != 0) {
            CERR("reading snapshot mask", "");
            return (-1);
        }

        unsigned int i;
        AltiModule::SIGNAL signal;
        for (i = 1; i < AltiModule::SIGNAL_NUMBER; i++) {
            signal = AltiModule::SIGNAL(i);
            printf("| [%02d]%-7s | %-10s |\n", i, AltiModule::SIGNAL_NAME[signal].c_str(), masked[i] ? "Masked" : "Not masked");
            printf("+--------------------------+\n");
        }

        bool snapshotEnabled;
        if (alti->SNPEnableRead(snapshotEnabled) != 0) {
            CERR("reading snapshot enable", "");
            return (-1);
        }

        std::printf("\n");
        std::printf("Snapshot: %s\n", snapshotEnabled ? ("Enabled") : ("Disabled"));

        unsigned int cur;
        if (alti->SNPCurrentAddressRead(cur) != 0) {
            CERR("reading snapshot current address", "");
            return (-1);
        }
        std::printf("Current address = 0x%08x (= %10u)\n", cur, cur);
        std::printf("\n");

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiMEMDataPATSNPReadNumber : public MenuItem {
  public:
    AltiMEMDataPATSNPReadNumber() { setName("read    memory [number]"); }
    int action() {

        // no automatic disabling!
        /*if (alti->SNPEnableWrite(false) != 0) {
            CERR("disabling snapshot", "");
            return (-1);
        }*/

        unsigned int indx, data;
        indx = enterHex("address index", 0, AltiModule::MAX_MEMORY - 1);
        if (alti->MEMDataRead(AltiModule::ALTI_RAM_MEMORY::PATTERN_SNAPSHOT, data, indx) != 0) {
            CERR("reading snapshot/pattern memory", "");
            return (-1);
        }

        std::printf("\n");
        std::printf("data[%6d] = 0x%08x (= %10u)\n", indx, data, data);
        std::printf("\n");

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiMEMDataPATSNPWriteNumber : public MenuItem {
  public:
    AltiMEMDataPATSNPWriteNumber() { setName("write   memory [number]"); }
    int action() {

        // no automatic disabling!
        /*if (alti->SNPEnableWrite(false) != 0) {
            CERR("disabling snapshot", "");
            return (-1);
        }
        if (alti->PATGenerationEnableWrite(false) != 0) {
            CERR("disabling pattern generation", "");
            return (-1);
        }*/

        unsigned int indx, data;
        indx = enterHex("address index", 0, AltiModule::MAX_MEMORY - 1);
        data = enterHex("data         ", 0, ALTI::MASK_MEM_SHAREDALTIRAM);
        if (alti->MEMDataWrite(AltiModule::ALTI_RAM_MEMORY::PATTERN_SNAPSHOT, data, indx) != 0) {
            CERR("writing snapshot/pattern memory", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiMEMDataPATSNPReadVector : public MenuItem {
  public:
    AltiMEMDataPATSNPReadVector() { setName("read    memory [vector]"); }
    int action() {

        // no automatic disabling!
        /*if (alti->SNPEnableWrite(false) != 0) {
            CERR("disabling snapshot", "");
            return (-1);
        }*/

        std::vector<unsigned int> data;

        unsigned int vme_off = enterHex("starting index", 0, AltiModule::MAX_MEMORY - 1);
        unsigned int size = enterHex("size          ", 1, AltiModule::MAX_MEMORY - vme_off);
        data.resize(size);

        if (alti->MEMDataRead(AltiModule::ALTI_RAM_MEMORY::PATTERN_SNAPSHOT, data, vme_off) != 0) {
            CERR("reading snapshot/pattern memory", "");
            return (-1);
        }

        unsigned int indx;
        for (indx = 0; indx < size; indx++) {
            std::printf("data[%6d] = 0x%08x (= %10u)\n", vme_off + indx, data[indx], data[indx]);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiMEMDataPATSNPWriteVector : public MenuItem {
  public:
    AltiMEMDataPATSNPWriteVector() { setName("write   memory [vector]"); }
    int action() {

        // no automatic disabling!
        /*if (alti->SNPEnableWrite(false) != 0) {
            CERR("disabling snapshot", "");
            return (-1);
        }
        if (alti->PATGenerationEnableWrite(false) != 0) {
            CERR("disabling pattern generation", "");
            return (-1);
        }*/

        std::vector<unsigned int> data;

        unsigned int vme_off = enterHex("starting index ", 0, AltiModule::MAX_MEMORY - 1);
        unsigned int size = enterHex("size           ", 1, AltiModule::MAX_MEMORY - vme_off);
        unsigned int sdat = enterHex("starting value ", 0, ALTI::MASK_MEM_SHAREDALTIRAM);
        unsigned int idat = enterHex("increment value", 0, ALTI::MASK_MEM_SHAREDALTIRAM);
        unsigned int xdat(sdat), indx;

        for (indx = 0; indx < size; indx++) {
            data.push_back(xdat & ALTI::MASK_MEM_SHAREDALTIRAM);
            xdat += idat;
        }

        if (alti->MEMDataWrite(AltiModule::ALTI_RAM_MEMORY::PATTERN_SNAPSHOT, data, vme_off) != 0) {
            CERR("writing snapshot/pattern memory", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiMEMDataPATSNPReadBlock : public MenuItem {
  public:
    AltiMEMDataPATSNPReadBlock() { setName("read    memory [block]"); }
    int action() {

        // no automatic disabling!
        /*if (alti->SNPEnableWrite(false) != 0) {
            CERR("disabling snapshot", "");
            return (-1);
        }*/

        if (alti->MEMDataRead(AltiModule::ALTI_RAM_MEMORY::PATTERN_SNAPSHOT, segment[0], AltiModule::MAX_MEMORY, 0, 0) != 0) {
            CERR("reading snapshot/pattern memory in block transfer", "");
            return (-1);
        }

        unsigned int indx;
        unsigned int *data = (unsigned int *) segment[0]->VirtualAddress();
        for (indx = 0; indx < AltiModule::MAX_MEMORY; indx++) {
            std::printf("data[%6d] = 0x%08x (= %10u)\n", indx, data[indx], data[indx]);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiMEMDataPATSNPWriteBlock : public MenuItem {
  public:
    AltiMEMDataPATSNPWriteBlock() { setName("write   memory [block]"); }
    int action() {

        // no automatic disabling!
        /*if (alti->SNPEnableWrite(false) != 0) {
            CERR("disabling snapshot", "");
            return (-1);
        }
        if (alti->PATGenerationEnableWrite(false) != 0) {
            CERR("disabling pattern generation", "");
            return (-1);
        }*/

        unsigned int sdat = enterHex("starting  value", 0, ALTI::MASK_MEM_SHAREDALTIRAM);
        unsigned int idat = enterHex("increment value", 0, ALTI::MASK_MEM_SHAREDALTIRAM);
        unsigned int xdat(sdat), indx;
        unsigned int* data = (unsigned int*) segment[0]->VirtualAddress();

        for (indx = 0; indx < AltiModule::MAX_MEMORY; indx++) {
            data[indx] = xdat & ALTI::MASK_MEM_SHAREDALTIRAM;
            xdat += idat;
        }

        if (alti->MEMDataWrite(AltiModule::ALTI_RAM_MEMORY::PATTERN_SNAPSHOT, segment[0], AltiModule::MAX_MEMORY, 0, 0) != 0) {
            CERR("writing snapshot/pattern memory in block transfer", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiSNPRead : public MenuItem {
  public:
    AltiSNPRead() { setName("read    snapshot memory"); }
    int action() {

        // no automatic disabling!
        /*if (alti->SNPEnableWrite(false) != 0) {
            CERR("disabling snapshot", "");
            return (-1);
        }*/

        unsigned int vme_off = enterHex("Starting index", 0, AltiModule::MAX_MEMORY - 1);
        unsigned int size = enterHex("Size          ", 1, AltiModule::MAX_MEMORY - vme_off);

        if (alti->SNPRead(segment[0], vme_off, size) != 0) {
            CERR("reading snapshot memory. If the snapshot is still enabled, disable it before reading the memory", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiSNPReadFile : public MenuItem {
  public:
    AltiSNPReadFile() { setName("read    snapshot memory [into file]"); }
    int action() {

        // no automatic disabling!
        /*if (alti->SNPEnableWrite(false) != 0) {
            CERR("disabling snapshot", "");
            return (-1);
        }*/

        unsigned int vme_off = enterHex("Starting index                                          ", 0, AltiModule::MAX_MEMORY - 1);
        unsigned int size = enterHex("Size                                                    ", 1, AltiModule::MAX_MEMORY - vme_off);

        int format = enterInt("Format [0-Hex/1-Human-readable(original)/2-Human-readable(expanded)]", 0, 2);

        std::string fn = enterString("File name                                                                    ");
        std::ofstream outf(fn.c_str(), std::ofstream::out);
        if (!outf) {
            CERR("cannot open output file \"%s\"", fn.c_str());
            return (-1);
        }

        if (alti->SNPReadFile(segment[0], vme_off, size, format, outf) != 0) {
            CERR("reading snapshot memory into file. If the snapshot is still enabled, disable it before reading the memory", "");
            return (-1);
        }
        
        return(0);
    }
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

class AltiENCTransmitterEnable : public RCD::MenuItem {
  public:
    AltiENCTransmitterEnable() { setName("transmitters enable"); }
    int action() {

        unsigned int tx = enterInt("Transmitter (0=\"TX0\", 1=\"TX1\", ..., \"10=TX10\", 11=all)", 0, AltiModule::TRANSMITTER_NUMBER - 1 + 1);
        unsigned int enable = enterInt("Enable (0=disable, 1=enable)                           ", 0, 1);

        if (tx < AltiModule::TRANSMITTER_NUMBER) {
            if (alti->ENCTransmitterEnableWrite(tx, enable) != 0) {
                CERR("%s TX%d", enable ? "enabling" : "disabling", tx);
                return (-1);
            }
        }
        else { // all
            if (alti->ENCTransmitterEnableWrite(enable) != 0) {
                CERR("%s all TXs", enable ? "enabling" : "disabling");
                return (-1);
            }
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiENCTransmitterFaultClear : public RCD::MenuItem {
  public:
    AltiENCTransmitterFaultClear() { setName("transmitters fault clear"); }
    int action() {

        unsigned int tx = enterInt("Transmitter (0=\"TX0\", 1=\"TX1\", ..., \"10=TX10\", 11=all)", 0, AltiModule::TRANSMITTER_NUMBER - 1 + 1);

        if (tx < AltiModule::TRANSMITTER_NUMBER) {
            if (alti->ENCTransmitterFaultClear(tx) != 0) {
                CERR("clearing TX%d fault status", tx);
                return (-1);
            }
        }
        else { // all
            if (alti->ENCTransmitterFaultClear() != 0) {
                CERR("clearing fault status for all TXs", "");
                return (-1);
            }
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiENCTransmitterStatusRead : public RCD::MenuItem {
  public:
    AltiENCTransmitterStatusRead() { setName("transmitters status"); }
    int action() {

        std::printf("\n");
        std::printf("+--------------------------------+\n");
        std::printf("| Transmitter | Status           |\n");
        std::printf("+--------------------------------+\n");

        std::vector<unsigned int> delay;
        delay.resize(AltiModule::TRANSMITTER_NUMBER);
        std::vector<bool> enabled;
        enabled.resize(AltiModule::TRANSMITTER_NUMBER);
        std::vector<bool> fault;
        fault.resize(AltiModule::TRANSMITTER_NUMBER);
        /*if (alti->TTCTransmitterStatusRead(enabled, fault) != 0) {
            CERR("reading status of all TXs", "");
            return (-1);
        }*/
        if (alti->ENCTransmitterEnableRead(enabled) != 0) {
            CERR("reading enable status of all TXs", "");
            return (-1);
        }
        if (alti->ENCTransmitterFaultRead(fault) != 0) {
            CERR("reading fault status of all TXs", "");
            return (-1);
        }
        unsigned int i;
        for (i = 0; i < enabled.size(); i++) {
            std::printf("| TX[%02d]      | %s %s |\n", i, enabled[i] ? "Enabled " : "Disabled", fault[i] ? "/ Fault" : "       ");
            std::printf("+--------------------------------+\n");
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiDECEnable : public RCD::MenuItem {
  public:
    AltiDECEnable() { setName("enable  TTC decoder"); }
    int action() {

        if (alti->DECEnableWrite(true) != 0) {
            CERR("enabling TTC decoder snapshot memory", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiDECDisable : public RCD::MenuItem {
  public:
    AltiDECDisable() { setName("disable TTC decoder"); }
    int action() {

        if (alti->DECEnableWrite(false) != 0) {
            CERR("disabling TTC decoder snapshot memory", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

/*class AltiDECCurrentAddressRead : public MenuItem {
  public:
    AltiDECCurrentAddressRead() { setName("read    command/timestamp current address"); }
    int action() {

        unsigned int cur;
        if (alti->DECCurrentAddressRead(cur) != 0) {
            CERR("reading TTC decoder snapshot current address", "");
            return (-1);
        }
        std::printf("\n");
        std::printf("current address = 0x%08x (= %10u)\n", cur, cur);
        std::printf("\n");

        return(0);
    }
};*/

//------------------------------------------------------------------------------

class AltiDECReset : public RCD::MenuItem {
  public:
    AltiDECReset() { setName("reset   TTC decoder"); }
    int action() {

        if (alti->DECReset() != 0) {
            CERR("resetting TTC decoder snapshot memory","");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiDECStatusRead : public RCD::MenuItem {
  public:
    AltiDECStatusRead() { setName("status"); }
    int action() {

        bool enabled;
        if (alti->DECEnableRead(enabled) != 0) {
            CERR("reading TTC decoder enable", "");
            return (-1);
        }
        bool cdr_los, cdr_lol, pll_lol, singleBitError, doubleBitError;
        if (alti->DECStatusRead(cdr_los, cdr_lol, pll_lol, singleBitError, doubleBitError) != 0) {
            CERR("reading TTC decoder status", "");
            return (-1);
        }
        unsigned int cur;
        if (alti->DECCurrentAddressRead(cur) != 0) {
            CERR("reading TTC decoder snapshot current address", "");
            return (-1);
        }

        std::printf("\n");
        std::printf("Receiver:         %s, %s\n", enabled ? ("enabled") : ("disabled"), (cdr_los & cdr_lol & pll_lol) ? ("loss of lock") : ("locked"));
        //std::printf("Address reset:    %s\n", reset ? ("Enabled") : ("Disabled"));
        std::printf("Single bit error: %s\n", singleBitError ? ("error") : ("OK"));
        std::printf("Double bit error: %s\n", doubleBitError ? ("error") : ("OK"));
        //std::printf("Loss of lock:     %s\n", lol ? ("LOL") : ("OK"));

        std::printf("\n");
        std::printf("Current address = 0x%08x (= %10u)\n", cur, cur);
        std::printf("\n");

	std::printf("\n");
        std::printf("BCR word = 0x%02x \n", m_bcr_word);
	std::printf("BCR bit-wise decoding scheme = %s \n", m_bcr_bitwise ? ("enabled") : ("disabled"));
        std::printf("\n");
        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiDECStatusClear : public RCD::MenuItem {
  public:
    AltiDECStatusClear() { setName("clear status bits"); }
    int action() {

        if (alti->DECStatusClear() != 0) {
            CERR("clearing all TTC decoder status bits", "");
            return (-1);
        }

        return(0);
    }
};


//------------------------------------------------------------------------------

class AltiDECSetBcrWord : public RCD::MenuItem {
  public:
  AltiDECSetBcrWord() { setName("set BCR scheme"); }
  int action() {
    m_bcr_word = enterHex("BCR bit (default = 0x1)", 0, 0xff);
    m_bcr_bitwise = enterInt("bit-wise decoding scheme 0=disable, 1=enable (defaul). If enabled, any word containing the BCR bit will be interpreted as BCR", 0, 1);;
    return(0);
  }
};



//------------------------------------------------------------------------------

class AltiMEMDataDECReadNumber : public MenuItem {
  public:
    AltiMEMDataDECReadNumber() { setName("read    memory [number]"); }
    int action() {

        if (alti->DECEnableWrite(false) != 0) {
            CERR("disabling TTC decoder", "");
            return (-1);
        }

        unsigned int indx, data0, data1;
        indx = enterHex("address index", 0, AltiModule::MAX_MEMORY - 1);
        if (alti->MEMDataRead(AltiModule::ALTI_RAM_MEMORY::TTC_DECODER_COMMAND, data0, indx) != 0) {
            CERR("reading TTC decoder command memory", "");
            return (-1);
        }
        if (alti->MEMDataRead(AltiModule::ALTI_RAM_MEMORY::TTC_DECODER_TIMESTAMP, data1, indx) != 0) {
            CERR("reading TTC decoder timestamp memory", "");
            return (-1);
        }

        std::printf("\n");
        std::printf("data[%6d] = 0x%08x, 0x%08x (= %10u, %10u)\n", indx, data0, data1, data0, data1);
        std::printf("\n");

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiMEMDataDECWriteNumber : public MenuItem {
  public:
    AltiMEMDataDECWriteNumber() { setName("write   memory [number]"); }
    int action() {

        if (alti->DECEnableWrite(false) != 0) {
            CERR("disabling TTC decoder", "");
            return (-1);
        }

        unsigned int indx, data0, data1;
        indx = enterHex("address index                ", 0, AltiModule::MAX_MEMORY - 1);
        data0 = enterHex("data (most  significant word)", 0, ALTI::MASK_MEM_SHAREDALTIRAM);
        data1 = enterHex("data (least significant word)", 0, ALTI::MASK_MEM_SHAREDALTIRAM);
        if (alti->MEMDataWrite(AltiModule::ALTI_RAM_MEMORY::TTC_DECODER_COMMAND, data0, indx) != 0) {
            CERR("writing TTC decoder command memory", "");
            return (-1);
        }
        if (alti->MEMDataWrite(AltiModule::ALTI_RAM_MEMORY::TTC_DECODER_TIMESTAMP, data1, indx) != 0) {
            CERR("writing TTC decoder timestamp memory", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiMEMDataDECReadVector : public MenuItem {
  public:
    AltiMEMDataDECReadVector() { setName("read    memory [vector]"); }
    int action() {

        if (alti->DECEnableWrite(false) != 0) {
            CERR("disabling TTC decoder", "");
            return (-1);
        }

        std::vector<unsigned int> data0, data1;

        unsigned int vme_off = enterHex("starting index", 0, AltiModule::MAX_MEMORY - 1);
        unsigned int size = enterHex("size          ", 1, AltiModule::MAX_MEMORY - vme_off);
        data0.resize(size);
        data1.resize(size);

        if (alti->MEMDataRead(AltiModule::ALTI_RAM_MEMORY::TTC_DECODER_COMMAND, data0, vme_off) != 0) {
            CERR("reading TTC decoder command memory", "");
            return (-1);
        }
        if (alti->MEMDataRead(AltiModule::ALTI_RAM_MEMORY::TTC_DECODER_TIMESTAMP, data1, vme_off) != 0) {
            CERR("reading TTC decoder timestamp memory", "");
            return (-1);
        }

        unsigned int indx;
        for (indx = 0; indx < size; indx++) {
            std::printf("data[%6d] = 0x%08x, 0x%08x (= %10u, %10u)\n", vme_off + indx, data0[indx], data1[indx], data0[indx], data1[indx]);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiMEMDataDECWriteVector : public MenuItem {
  public:
    AltiMEMDataDECWriteVector() { setName("write   memory [vector]"); }
    int action() {

        if (alti->DECEnableWrite(false) != 0) {
            CERR("disabling TTC decoder", "");
            return (-1);
        }

        std::vector<unsigned int> data0, data1;

        unsigned int vme_off = enterHex("starting index                          ", 0, AltiModule::MAX_MEMORY - 1);
        unsigned int size = enterHex("size                                    ", 1, AltiModule::MAX_MEMORY - vme_off);
        unsigned int sdat0 = enterHex("starting  value (most  significant word)", 0, ALTI::MASK_MEM_SHAREDALTIRAM);
        unsigned int idat0 = enterHex("increment value (most  significant word)", 0, ALTI::MASK_MEM_SHAREDALTIRAM);
        unsigned int sdat1 = enterHex("starting  value (least significant word)", 0, ALTI::MASK_MEM_SHAREDALTIRAM);
        unsigned int idat1 = enterHex("increment value (least significant word)", 0, ALTI::MASK_MEM_SHAREDALTIRAM);
        unsigned int xdat0(sdat0), xdat1(sdat1), indx;

        for (indx = 0; indx < size; indx++) {
            data0.push_back(xdat0 & ALTI::MASK_MEM_SHAREDALTIRAM);
            xdat0 += idat0;
            data1.push_back(xdat1 & ALTI::MASK_MEM_SHAREDALTIRAM);
            xdat1 += idat1;
        }

        if (alti->MEMDataWrite(AltiModule::ALTI_RAM_MEMORY::TTC_DECODER_COMMAND, data0, vme_off) != 0) {
            CERR("writing TTC decoder command memory", "");
            return (-1);
        }
        if (alti->MEMDataWrite(AltiModule::ALTI_RAM_MEMORY::TTC_DECODER_TIMESTAMP, data1, vme_off) != 0) {
            CERR("writing TTC decoder timestamp memory", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiMEMDataDECReadBlock : public MenuItem {
  public:
    AltiMEMDataDECReadBlock() { setName("read    memory [block]"); }
    int action() {

        if (alti->DECEnableWrite(false) != 0) {
            CERR("disabling TTC decoder", "");
            return (-1);
        }

        if (alti->MEMDataRead(AltiModule::ALTI_RAM_MEMORY::TTC_DECODER_COMMAND, segment[0], AltiModule::MAX_MEMORY, 0, 0) != 0) {
            CERR("reading TTC decoder command memory in block transfer", "");
            return (-1);
        }
        if (alti->MEMDataRead(AltiModule::ALTI_RAM_MEMORY::TTC_DECODER_TIMESTAMP, segment[1], AltiModule::MAX_MEMORY, 0, 0) != 0) {
            CERR("reading TTC decoder timestamp memory in block transfer", "");
            return (-1);
        }

        unsigned int indx;
        unsigned int *data0 = (unsigned int *) segment[0]->VirtualAddress();
        unsigned int *data1 = (unsigned int *) segment[1]->VirtualAddress();
        for (indx = 0; indx < AltiModule::MAX_MEMORY; indx++) {
            std::printf("data[%6d] = 0x%08x, 0x%08x (= %10u, %10u)\n", indx, data0[indx], data1[indx], data0[indx], data1[indx]);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiMEMDataDECWriteBlock : public MenuItem {
  public:
    AltiMEMDataDECWriteBlock() { setName("write   memory [block]"); }
    int action() {

        if (alti->DECEnableWrite(false) != 0) {
            CERR("disabling TTC decoder", "");
            return (-1);
        }

        unsigned int sdat0 = enterHex("starting  value (most  significant word)", 0, ALTI::MASK_MEM_SHAREDALTIRAM);
        unsigned int idat0 = enterHex("increment value (most  significant word)", 0, ALTI::MASK_MEM_SHAREDALTIRAM);
        unsigned int sdat1 = enterHex("starting  value (least significant word)", 0, ALTI::MASK_MEM_SHAREDALTIRAM);
        unsigned int idat1 = enterHex("increment value (least significant word)", 0, ALTI::MASK_MEM_SHAREDALTIRAM);
        unsigned int xdat0(sdat0), xdat1(sdat1), indx;
        unsigned int* data0 = (unsigned int*) segment[0]->VirtualAddress();
        unsigned int* data1 = (unsigned int*) segment[1]->VirtualAddress();

        for (indx = 0; indx < AltiModule::MAX_MEMORY; indx++) {
            data0[indx] = xdat0 & ALTI::MASK_MEM_SHAREDALTIRAM;
            xdat0 += idat0;
            data1[indx] = xdat1 & ALTI::MASK_MEM_SHAREDALTIRAM;
            xdat1 += idat1;
        }

        if (alti->MEMDataWrite(AltiModule::ALTI_RAM_MEMORY::TTC_DECODER_COMMAND, segment[0], AltiModule::MAX_MEMORY, 0, 0) != 0) {
            CERR("writing TTC decoder command memory in block transfer", "");
            return (-1);
        }
        if (alti->MEMDataWrite(AltiModule::ALTI_RAM_MEMORY::TTC_DECODER_TIMESTAMP, segment[1], AltiModule::MAX_MEMORY, 0, 0) != 0) {
            CERR("writing TTC decoder timestamp memory in block transfer", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiDECRead : public MenuItem {
  public:
    AltiDECRead() { setName("read    TTC decoder memory"); }
    int action() {

        // no automatic disabling!
        /*if (alti->DECEnableWrite(false) != 0) {
            CERR("disabling TTC decoder", "");
            return (-1);
        }*/

	unsigned int vme_off = enterInt("Starting index", 0, AltiModule::MAX_MEMORY - 1);
	unsigned int size = enterInt("Size          ", 1, AltiModule::MAX_MEMORY - vme_off);
	
        if (alti->DECRead(segment[0], segment[1], vme_off, size, m_bcr_word) != 0) {
            CERR("reading TTC decoder memory. If the TTC decoder monitoring is still enabled, disable it before reading the memory", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiDECReadFile : public RCD::MenuItem {
  public:
    AltiDECReadFile() { setName("read    TTC decoder memory [into file]"); }
    int action() {

        // no automatic disabling!
        /*if (alti->DECEnableWrite(false) != 0) {
            CERR("disabling TTC decoder", "");
            return (-1);
        }*/

	unsigned int vme_off = enterInt("Starting index                   ", 0, AltiModule::MAX_MEMORY - 1);
        unsigned int size = enterInt("Size                             ", 1, AltiModule::MAX_MEMORY - vme_off);

        int format = enterInt("File format [0-Hex/1-Human-readable/2-TTCScope]", 0, 2);

        std::string fn = enterString("File name                                             ");
        std::ofstream outf(fn.c_str(), std::ofstream::out);
        if (!outf) {
            CERR("cannot open output file \"%s\"", fn.c_str());
           return (-1);
        }

        if (alti->DECReadFile(segment[0], segment[1], vme_off, size, m_bcr_word, format, outf) != 0) {
            CERR("reading TTC decoder memory into file. If the TTC decoder monitoring is still enabled, disable it before reading the memory", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiENCInhibitParamsRead : public RCD::MenuItem {
  public:
    AltiENCInhibitParamsRead() { setName("read    BGO  inhibit parameters"); }
    int action() {

        std::vector<unsigned int> width, delay;
        if (alti->ENCInhibitParamsRead(width, delay) != 0) {
            CERR("reading TTC encoder BGO inhibit parameters", "");
            return (-1);
        }

        //table
        printf("\n");
        printf("+------------------------------+\n");
        printf("|      | Width[BC] | Delay[BC] |\n");
        printf("+------------------------------+\n");

        unsigned int i;
        for (i = 0; i <= 3; i++) {
            printf("| BGO%d | %4d      | %4d      |\n", i, width[i], delay[i]);
            printf("+------------------------------+\n");
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiENCInhibitParamsWrite : public RCD::MenuItem {
  public:
    AltiENCInhibitParamsWrite() { setName("write   BGO  inhibit parameters"); }
    int action() {

        unsigned int bgo;
        unsigned int bgo_choice = enterInt("BGO index (0-3, 4=all)                   ", 0, 4);
        if (bgo_choice < 4) bgo = bgo_choice;

        if (bgo_choice < 4) { // 1 BGO
            std::printf("BGO%d width [BCs]                      ", bgo);
            unsigned int width = enterInt("", 0, ALTI_BGO_INHIBITGENERATORPARAMS_BITSTRING::MASK_DELAYINBCS.number()[0]); /*WIDTH*/
            std::printf("BGO%d delay [BCs]                      ", bgo);
            unsigned int delay = enterInt("", 0, ALTI_BGO_INHIBITGENERATORPARAMS_BITSTRING::MASK_DELAYINBCS.number()[0]);
            if (alti->ENCInhibitParamsWrite(bgo, width, delay) != 0) {
                CERR("writing TTC encoder BGO%d inhibit width %d, delay", bgo, width, delay);
                return (-1);
            }
        }
        else { // all BGOs
            std::vector<unsigned int> width, delay;
            width.clear();
            delay.clear();
            unsigned int i;
            for (i = 0; i <= 3; i++) {
                std::printf("BGO%d width [BCs]                      ", i);
                width.push_back(enterInt("", 0, ALTI_BGO_INHIBITGENERATORPARAMS_BITSTRING::MASK_DELAYINBCS.number()[0])); /*WIDTH*/
                std::printf("BGO%d delay [BCs]                      ", i);
                delay.push_back(enterInt("", 0, ALTI_BGO_INHIBITGENERATORPARAMS_BITSTRING::MASK_DELAYINBCS.number()[0]));

                std::printf("\n");

                if (alti->ENCInhibitParamsWrite(width, delay) != 0) {
                    CERR("writing TTC encoder inhibit width and delay for all BGOs", "");
                    return (-1);
                }
            }
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiENCTTRSelWrite : public RCD::MenuItem {
  public:
    AltiENCTTRSelWrite() { setName("write   TTR or L1A selection"); }
    int action() {

        unsigned int sel;
        unsigned int l1a = enterInt("L1A   (enable=1, disable=0)                   ", 0, 1);
        unsigned int ttr1 = enterInt("TTR1 (enable=1, disable=0)                   ", 0, 1);
        unsigned int ttr2 = enterInt("TTR2 (enable=1, disable=0)                   ", 0, 1);
        unsigned int ttr3 = enterInt("TTR3 (enable=1, disable=0)                   ", 0, 1);
        sel = l1a + ttr1*2 + ttr2*4 + ttr3*8;
        if (alti->ENCTTRSelWrite(sel) != 0) 
        {
          CERR("writing TTC encoder L1A source", "");
          return (-1);}
        
      return(0);
    }
};

//------------------------------------------------------------------------------

class AltiENCTTRSelRead : public RCD::MenuItem {
  public:
    AltiENCTTRSelRead() { setName("read    TTR or L1A selection"); }
    int action() {
      unsigned int sel;
      if (alti->ENCTTRSelRead(sel) != 0) 
      {
        CERR("writing TTC encoder L1A source", "");
        return (-1);}      
      
      unsigned int l1a = sel%2;
      sel = sel / 2;
      unsigned int ttr1 = sel%2;
      sel = sel / 2;
      unsigned int ttr2 = sel%2;
      sel = sel / 2;
      unsigned int ttr3 = sel%2;
      printf(" L1A  : %s\n TTR1 : %s\n TTR2 : %s\n TTR3 : %s\n", l1a ? "enabled" : "disabled", ttr1 ? "enabled" : "disabled", ttr2 ? "enabled" : "disabled", ttr3 ? "enabled" : "disabled");
      return(0);
    }
};

//------------------------------------------------------------------------------
class AltiENCTriggerWordDelayRead : public RCD::MenuItem {
  public:
    AltiENCTriggerWordDelayRead() { setName("read    TTYP word delay"); }
    int action() {

        unsigned int delay;
        if (alti->ENCTriggerWordDelayRead(delay) != 0) {
            CERR("reading TTC encoder TTYP delay", "");
            return (-1);
        }

        //table
        printf("\n");
        printf("+------------------+\n");
        printf("|      | Delay[BC] |\n");
        printf("+------------------+\n");
        printf("| TTYP | %4d      |\n", delay);
        printf("+------------------+\n");

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiENCTriggerWordDelayWrite : public RCD::MenuItem {
  public:
    AltiENCTriggerWordDelayWrite() { setName("write   TTYP word delay"); }
    int action() {

        unsigned int delay = enterInt("Delay [BCs]", 0, ALTI_TTC_TRIGGERTYPEWORDPARAMS_BITSTRING::MASK_DELAYINBCS.number()[0]);

        if (alti->ENCTriggerWordDelayWrite(delay) != 0) {
            CERR("writing TTC encoder TTYP delay %d", delay);
            return (-1);
        }
        return(0);
    }
};

//------------------------------------------------------------------------------

/*class AltiENCTriggerWordCycleControlRead : public RCD::MenuItem {
  public:
    AltiENCTriggerWordCycleControlRead() { setName("read    TTYP settings"); }
    int action() {

        unsigned int addr, subAddr;
        AltiModule::TTC_FRAME_TYPE frameType;
        AltiModule::TTC_ADDRESS_SPACE addressSpace;
        if (alti->ENCTriggerWordCycleControlRead(addr, frameType, addressSpace, subAddr) != 0) {
            CERR("reading TTC encoder TTYP word control", "");
            return (-1);
        }

        //table
        printf("\n");
        printf("+--------------------------------------------------------------+\n");
        printf("| Address (hex) | Frame type | Address space | Subadress (hex) |\n");
        printf("+--------------------------------------------------------------+\n");
        printf("| 0x%04x        | %-10s | %-13s | 0x%02x            |\n", addr, AltiModule::TTC_FRAME_TYPE_NAME[frameType].c_str(), AltiModule::TTC_ADDRESS_SPACE_NAME[addressSpace].c_str(), subAddr);
        printf("+--------------------------------------------------------------+\n");

        return(0);
    }
};*/

//------------------------------------------------------------------------------

/*class AltiENCTriggerWordCycleControlWrite : public RCD::MenuItem {
  public:
    AltiENCTriggerWordCycleControlWrite() { setName("write   TTYP settings"); }
    int action() {

        unsigned int addr = enterHex("Address                ", 0, 0x3fff); // 14 bits
        AltiModule::TTC_FRAME_TYPE frameType = (AltiModule::TTC_FRAME_TYPE) enterInt("Frame type [0-SHORT/1-LONG]          ", 0, AltiModule::TTC_FRAME_TYPE_NUMBER - 1);
        AltiModule::TTC_ADDRESS_SPACE addressSpace = (AltiModule::TTC_ADDRESS_SPACE) enterInt("Address space [0-INTERNAL/1-EXTERNAL]", 0, AltiModule::TTC_ADDRESS_SPACE_NUMBER - 1);
        unsigned int subAddr = enterHex("Subaddress             ", 0, 0xff); // 8 bits, 2 LSb get trimmed, 6 MSb get stored

        if (alti->ENCTriggerWordCycleControlWrite(addr, frameType, addressSpace, subAddr) != 0) {
            CERR("writing TTC encoder TTYP word control", "");
            return (-1);
        }

        return(0);
    }
};*/

//------------------------------------------------------------------------------

class AltiENCTriggerWordCycleEnable : public RCD::MenuItem {
  public:
    AltiENCTriggerWordCycleEnable() { setName("enable  TTYP cycles"); }
    int action() {

        unsigned int addr = enterHex("TTCrx address           ", 0, 0x3fff); // 14 bits
        AltiModule::TTC_ADDRESS_SPACE addressSpace = (AltiModule::TTC_ADDRESS_SPACE) enterInt("Address space (0=internal, 1=external)", 0, AltiModule::TTC_ADDRESS_SPACE_NUMBER - 1);
        unsigned int subAddr = enterHex("Subaddress              ", 0, 0xff); // 8 bits, 2 LSb get trimmed, 6 MSb get stored

        if (alti->ENCTriggerWordCycleEnable(addr, addressSpace, subAddr) != 0) {
            CERR("enabling TTC encoder TTYP cycles", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiENCTriggerWordCycleDisable : public RCD::MenuItem {
  public:
    AltiENCTriggerWordCycleDisable() { setName("disable TTYP cycles"); }
    int action() {

        if (alti->ENCTriggerWordCycleDisable() != 0) {
            CERR("disabling TTC encoder TTYP cycles", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiENCBgoGenerate : public RCD::MenuItem {
  public:
    AltiENCBgoGenerate() { setName("trigger BGO  command from the VME"); }
    int action() {

        unsigned int bgo = enterInt("BGO index", 0, 3);
        if (alti->ENCBgoGenerate(bgo) != 0) {
            CERR("triggering one-shot BGO%d command from VME", bgo);
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

/*class AltiTTCShotTTYPCommand : public RCD::MenuItem {
  public:
    AltiTTCShotTTYPCommand() { setName("trigger TTYP command from VME"); }
    int action() {

        if (alti->TTCShotTTYPCommandVME() != 0) {
            CERR("triggering one-shot TTYP command from VME", "");
            return (-1);
        }

        return(0);
    }
};*/

//------------------------------------------------------------------------------

class AltiENCBgoCommandModeRead : public RCD::MenuItem {
  public:
    AltiENCBgoCommandModeRead() { setName("read    BGO  FIFO mode"); }
    int action() {

        std::vector<AltiModule::TTC_BGO_MODE_TRIGGER> mode_trigger;
        std::vector<AltiModule::TTC_BGO_MODE_COMMAND> mode_command;
        std::vector<AltiModule::TTC_BGO_MODE_REPEAT> mode_repeat;
        std::vector<AltiModule::TTC_BGO_MODE_FIFO> mode_fifo;
        if (alti->ENCBgoCommandModeRead(mode_trigger, mode_command, mode_repeat, mode_fifo) != 0) {
            CERR("reading TTC encoder BGO mode", "");
            return (-1);
        }

        //table
        printf("\n");
        printf("+---------------------------------------------------------------------------------------------------+\n");
        printf("|      | Trigger    | Command      | Repeat     | FIFO           || High-level description          |\n");
        printf("+---------------------------------------------------------------------------------------------------+\n");

        unsigned int i;
        for (i = 0; i <= 3; i++) {
            printf("| BGO%d | %-10s | %-12s | %-10s | %-14s || %-31s |\n", i, AltiModule::TTC_BGO_MODE_TRIGGER_NAME[mode_trigger[i]].c_str(), AltiModule::TTC_BGO_MODE_COMMAND_NAME[mode_command[i]].c_str(), AltiModule::TTC_BGO_MODE_REPEAT_NAME[mode_repeat[i]].c_str(), AltiModule::TTC_BGO_MODE_FIFO_NAME[mode_fifo[i]].c_str(), AltiModule::TTC_BGO_MODE_NAME[AltiModule::ENCLowLevelToHighLevelBgoMode(mode_trigger[i], mode_command[i], mode_repeat[i], mode_fifo[i])].c_str());
            printf("+---------------------------------------------------------------------------------------------------+\n");
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiENCBgoCommandModeWrite : public RCD::MenuItem {
  public:
    AltiENCBgoCommandModeWrite() { setName("write   BGO  FIFO mode [expert user]"); }
    int action() {

        unsigned int bgo_choice = enterInt("BGO index (0-3, 4=all)                                ", 0, 4);

        AltiModule::TTC_BGO_MODE_TRIGGER mode_trigger;
        AltiModule::TTC_BGO_MODE_COMMAND mode_command;
        AltiModule::TTC_BGO_MODE_REPEAT mode_repeat;
        AltiModule::TTC_BGO_MODE_FIFO mode_fifo;
        std::vector<AltiModule::TTC_BGO_MODE_TRIGGER> mode_trigger_v;
        std::vector<AltiModule::TTC_BGO_MODE_COMMAND> mode_command_v;
        std::vector<AltiModule::TTC_BGO_MODE_REPEAT> mode_repeat_v;
        std::vector<AltiModule::TTC_BGO_MODE_FIFO> mode_fifo_v;
        if (bgo_choice < 4) { // 1 BGO
            unsigned int bgo = bgo_choice;
            std::printf("BGO%d trigger mode (0=\"BGO_SIGNAL\", 1=\"VME\")           ", bgo);
            mode_trigger = (AltiModule::TTC_BGO_MODE_TRIGGER) enterInt("", 0, AltiModule::TTC_BGO_MODE_TRIGGER_NUMBER - 1);
            std::printf("BGO%d command mode (0=\"SYNCHRONOUS\", 1=\"ASYNCHRONOUS\") ", bgo);
            mode_command = (AltiModule::TTC_BGO_MODE_COMMAND) enterInt("", 0, AltiModule::TTC_BGO_MODE_COMMAND_NUMBER - 1);
            std::printf("BGO%d repeat  mode (0=\"SINGLE\", 1=\"REPETITIVE\")        ", bgo);
            mode_repeat = (AltiModule::TTC_BGO_MODE_REPEAT) enterInt("", 0, AltiModule::TTC_BGO_MODE_REPEAT_NUMBER - 1);
            std::printf("BGO%d FIFO    mode (0=\"WHEN_NOT_EMPTY\", 1=\"ON_TRIGGER\")", bgo);
            mode_fifo = (AltiModule::TTC_BGO_MODE_FIFO) enterInt("", 0, AltiModule::TTC_BGO_MODE_FIFO_NUMBER - 1);
            if (alti->ENCBgoCommandModeWrite(bgo, mode_trigger, mode_command, mode_repeat, mode_fifo) != 0) {
                CERR("writing TTC encoder BGO%d FIFO mode \"%s\", \"%s\", \"%s\", \"%s\"", bgo, AltiModule::TTC_BGO_MODE_TRIGGER_NAME[mode_trigger].c_str(), AltiModule::TTC_BGO_MODE_COMMAND_NAME[mode_command].c_str(), AltiModule::TTC_BGO_MODE_REPEAT_NAME[mode_repeat].c_str(), AltiModule::TTC_BGO_MODE_FIFO_NAME[mode_fifo].c_str());
                return (-1);
            }
        }
        else { // all BGOs
            mode_trigger_v.clear();
            mode_command_v.clear();
            mode_repeat_v.clear();
            mode_fifo_v.clear();
            unsigned int i;
            for (i = 0; i <= 3; i++) {
                std::printf("BGO%d trigger mode (0=\"BGO_SIGNAL\", 1=\"VME\")           ", i);
                mode_trigger_v.push_back((AltiModule::TTC_BGO_MODE_TRIGGER) enterInt("", 0, AltiModule::TTC_BGO_MODE_TRIGGER_NUMBER - 1));
                std::printf("BGO%d command mode (0=\"SYNCHRONOUS\", 1=\"ASYNCHRONOUS\") ", i);
                mode_command_v.push_back((AltiModule::TTC_BGO_MODE_COMMAND) enterInt("", 0, AltiModule::TTC_BGO_MODE_COMMAND_NUMBER - 1));
                std::printf("BGO%d repeat  mode (0=\"SINGLE\", 1=\"REPETITIVE\")        ", i);
                mode_repeat_v.push_back((AltiModule::TTC_BGO_MODE_REPEAT) enterInt("", 0, AltiModule::TTC_BGO_MODE_REPEAT_NUMBER - 1));
                std::printf("BGO%d FIFO    mode (0=\"WHEN_NOT_EMPTY\", 1=\"ON_TRIGGER\")", i);
                mode_fifo_v.push_back((AltiModule::TTC_BGO_MODE_FIFO) enterInt("", 0, AltiModule::TTC_BGO_MODE_FIFO_NUMBER - 1));

                std::printf("\n");
            }
            if (alti->ENCBgoCommandModeWrite(mode_trigger_v, mode_command_v, mode_repeat_v, mode_fifo_v) != 0) {
                CERR("writing TTC encoder all modes for all BGOs", "");
                return (-1);
            }
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiENCBgoCommandModeHighLevelWrite : public RCD::MenuItem {
  public:
    AltiENCBgoCommandModeHighLevelWrite() { setName("write   BGO  FIFO mode"); }
    int action() {

        unsigned int bgo;
        unsigned int bgo_choice = enterInt("BGO index (0-3, 4=all)", 0, 4);
        if (bgo_choice < 4) bgo = bgo_choice;

        AltiModule::TTC_BGO_MODE mode;
        std::vector<AltiModule::TTC_BGO_MODE> mode_v;
        if (bgo_choice < 4) { // 1
            mode = (AltiModule::TTC_BGO_MODE) enterInt("Mode (0=\"SYNCHRONOUS_SINGLE_BGO_SIGNAL\", 1=\"SYNCHRONOUS_REPETITIVE\", 2=\"ASYNCHRONOUS_BGO_SIGNAL\", 3=\"ASYNCHRONOUS_VME_ON_TRIGGER\", 4=\"ASYNCHRONOUS_VME_WHEN_NOT_EMPTY\")", 0, AltiModule::TTC_BGO_MODE_NUMBER - 1 - 1);
            if (alti->ENCBgoCommandModeWrite(bgo, mode) != 0) {
                CERR("writing TTC encoder BGO%d high-level mode \"%s\"", bgo, AltiModule::TTC_BGO_MODE_NAME[mode].c_str());
                return (-1);
            }
        }
        else if (bgo_choice == 4) { // all
            mode_v.clear();
            unsigned int i;
            for (i = 0; i <= 3; i++) {
                std::printf("BGO%d mode (0=\"SYNCHRONOUS_SINGLE_BGO_SIGNAL\", 1=\"SYNCHRONOUS_REPETITIVE\", 2=\"ASYNCHRONOUS_BGO_SIGNAL\", 3=\"ASYNCHRONOUS_VME_ON_TRIGGER\", 4=\"ASYNCHRONOUS_VME_WHEN_NOT_EMPTY\")", i);
                mode_v.push_back((AltiModule::TTC_BGO_MODE) enterInt("", 0, 4));
            }
            if (alti->ENCBgoCommandModeWrite(mode_v) != 0) {
                CERR("writing TTC encoder high-level modes for all BGOs", "");
                return (-1);
            }
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiENCFifoStatusRead : public RCD::MenuItem {
  public:
    AltiENCFifoStatusRead() { setName("FIFOs   status"); }
    int action() {

        std::vector<bool> ttc_pipelineFull, ttc_pipelineEmpty, ttc_fifoFull, ttc_fifoEmpty;
        //bool ttyp_pipelineFull, ttyp_pipelineEmpty, ttyp_fifoFull, ttyp_fifoEmpty;
        if (alti->ENCFifoStatusRead(ttc_pipelineFull, ttc_pipelineEmpty, ttc_fifoFull, ttc_fifoEmpty) != 0) {
            CERR("reading BGO and TTYP FIFOs statuses", "");
            return (-1);
        }
        /*if (alti->BGO_TTYP_FIFOReadStatus(ttyp_pipelineFull, ttyp_pipelineEmpty, ttyp_fifoFull, ttyp_fifoEmpty) != 0) {
            CERR("reading TTYP FIFO statuses", "");
            return (-1);
        }*/

        /*std::vector<AltiModule::TTC_FIFO_ACCESS_MODE> ttc_fifoAccessMode;
        if (alti->TTC_FIFOAccessModeRead(ttc_fifoAccessMode) != 0) {
            CERR("reading BGO and TTYP FIFOs access mode", "");
            return (-1);
        }*/
        /*AltiModule::TTC_FIFO_ACCESS_MODE ttyp_fifoAccessMode;
        if (alti->BGO_TTYP_FIFOAccessModeRead(ttyp_fifoAccessMode) != 0) {
            CERR("reading TTYP FIFO access mode", "");
            return (-1);
        }*/

        std::vector<bool> ttc_fifoRetransmit;
        if (alti->ENCBgoFifoRetransmitRead(ttc_fifoRetransmit) != 0) {
            CERR("reading BGO and TTYP FIFOs retransmit mode", "");
            return (-1);
        }
        /*bool ttyp_fifoRetransmit;
        if (alti->BGO_TTYP_FIFORetransmitRead(ttyp_fifoRetransmit) != 0) {
            CERR("reading TTYP FIFO retransmit mode", "");
            return (-1);
        }*/

        std::printf("\n");
        unsigned int i;
        AltiModule::TTC_FIFO ttcFifo;
        for (i = 0; i < AltiModule::TTC_FIFO_NUMBER; i++) {
            ttcFifo = (AltiModule::TTC_FIFO) i;
            //std::printf("BGO%d FIFO status: ", bgo);
            std::printf("\"%-9s\" status: ", AltiModule::TTC_FIFO_NAME[ttcFifo].c_str());
            //std::printf("%s ", AltiModule::TTC_FIFO_ACCESS_MODE_NAME[ttc_fifoAccessMode[i]].c_str());
            std::printf("%-14s,", ttc_pipelineFull[i] ? "Pipeline FULL" : ((ttc_pipelineEmpty[i] ? "Pipeline EMPTY" : "")));
            std::printf("%-11s", ttc_fifoFull[i] ? " FIFO FULL" : ((ttc_fifoEmpty[i] ? " FIFO EMPTY" : "")));
            std::printf("%-12s\n", ttc_fifoRetransmit[i] ? ", retransmit" : "");
        }
        /*std::printf("TTYP FIFO status: ");
        std::printf("%s ", AltiModule::TTC_FIFO_ACCESS_MODE_NAME[ttc_fifoAccessMode[4]].c_str());
        std::printf("%-15s", ttc_pipelineFull[4] ? "Pipeline FULL " : ((ttc_pipelineEmpty[4] ? "Pipeline EMPTY " : "")));
        std::printf("%-11s", ttc_fifoFull[4] ? "FIFO FULL " : ((ttc_fifoEmpty[4] ? "FIFO EMPTY " : "")));
        std::printf("%-10s\n", ttc_fifoRetransmit[4] ? "retransmit" : "");*/
        std::printf("\n");

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiENCFifoReset : public RCD::MenuItem {
  public:
    AltiENCFifoReset() { setName("FIFOs   reset"); }
    int action() {
        unsigned int fifo_choice = enterInt("FIFO (0=\"BGO0\", 1=\"BGO1\", 2=\"BGO2\", 3=\"BGO3\", 4=\"TTYP\", 5=all)", 0, 5);
        unsigned int port_choice = enterInt("Port                                 (0=read, 1=write, 2=both)", 0, 2);
        AltiModule::TTC_FIFO ttcFifo;
        if (fifo_choice <= 4) ttcFifo = (AltiModule::TTC_FIFO) fifo_choice;

        if (fifo_choice <= 4) { // ttcFifo
            switch (port_choice) {
                case 0: // read
                    if (alti->ENCFifoResetReadAddress(ttcFifo) != 0) {
                        CERR("resetting read address for \"%s\"", AltiModule::TTC_FIFO_NAME[ttcFifo].c_str());
                        return (-1);
                    }
                    break;
                case 1: // write
                    if (alti->ENCFifoResetWriteAddress(ttcFifo) != 0) {
                        CERR("resetting write address for \"%s\"", AltiModule::TTC_FIFO_NAME[ttcFifo].c_str());
                        return (-1);
                    }
                    break;
                case 2: // read+write
                    if (alti->ENCFifoReset(ttcFifo) != 0) {
                        CERR("resetting read and write address for \"%s\"", AltiModule::TTC_FIFO_NAME[ttcFifo].c_str());
                        return (-1);
                    }
                    break;
                default:
                    break;
            }
        }
        else if (fifo_choice == 5) { // all
            switch (port_choice) {
                case 0: // read
                    if (alti->ENCFifoResetReadAddress() != 0) {
                        CERR("resetting read address for all FIFOs", "");
                        return (-1);
                    }
                    break;
                case 1: // write
                    if (alti->ENCFifoResetWriteAddress() != 0) {
                        CERR("resetting write address for all FIFOs", "");
                        return (-1);
                    }
                    break;
                case 2: // read+write
                    if (alti->ENCFifoReset() != 0) {
                        CERR("resetting read and write address for all FIFOs", "");
                        return (-1);
                    }
                    break;
                default:
                    break;
            }
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiENCBgoFifoRetransmitSelect : public RCD::MenuItem {
  public:
    AltiENCBgoFifoRetransmitSelect() { setName("FIFOs   retransmit"); }
    int action() {
        unsigned int fifo_choice = enterInt("FIFO (0=\"BGO0\", 1=\"BGO1\", 2=\"BGO2\", 3=\"BGO3\", 4=\"TTYP\", 5=all)", 0, 5);
        bool retransmit  = (bool) enterInt("Retransmit                               (0=disable, 1=enable)", 0, 1);

        AltiModule::TTC_FIFO ttcFifo;
        if (fifo_choice <= 4) ttcFifo = (AltiModule::TTC_FIFO) fifo_choice;

        if (fifo_choice <= 4) { // ttcFifo
            if (alti->ENCBgoFifoRetransmitWrite(ttcFifo, retransmit) != 0) {
                CERR("%s retransmit mode for \"%s\"", (retransmit ? "Enabling" : "Disabling"), AltiModule::TTC_FIFO_NAME[ttcFifo].c_str());
                return (-1);
            }
        }
        else if (fifo_choice == 5) { // all
            if (alti->ENCBgoFifoRetransmitWrite(retransmit) != 0) {
                CERR("%s retransmit mode for all FIFOs", (retransmit ? "Enabling" : "Disabling"));
                return (-1);
            }
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

/*class AltiTTCFIFONumberRead : public RCD::MenuItem {
  public:
    AltiTTCFIFONumberRead() { setName("read    FIFO [number]"); }
    int action() {

        AltiModule::TTC_FIFO ttcFifo = (AltiModule::TTC_FIFO) enterInt("FIFO index [0-BGO0/1-BGO1/2-BGO2/3-BGO3/4-TTYP]", 0, AltiModule::TTC_FIFO_NUMBER - 1);

        unsigned int data;
        if (alti->TTC_FIFORead(ttcFifo, data) != 0) {
            CERR("reading \"%s\" [number]", AltiModule::TTC_FIFO_NAME[ttcFifo].c_str());
            return (-1);
        }

        std::printf("\n");
        std::printf("\"%s\" data = 0x%08x (= %10u)\n", AltiModule::TTC_FIFO_NAME[ttcFifo].c_str(), data, data);
        std::printf("\n");

        return(0);
    }
};*/

//------------------------------------------------------------------------------

/*class AltiTTCFIFONumberWrite : public RCD::MenuItem {
  public:
    AltiTTCFIFONumberWrite() { setName("write   FIFO [number]"); }
    int action() {

        AltiModule::TTC_FIFO ttcFifo = (AltiModule::TTC_FIFO) enterInt("FIFO index [0-BGO0/1-BGO1/2-BGO2/3-BGO3/4-TTYP]", 0, AltiModule::TTC_FIFO_NUMBER - 1);
        unsigned int data = enterHex("data                             ", 0, ALTI::MASK_BGO_FIFO);

        if (alti->TTC_FIFOWrite(ttcFifo, data) != 0) {
            CERR("writing 0x%08x to \"%s\" [number]", data, AltiModule::TTC_FIFO_NAME[ttcFifo].c_str());
            return (-1);
        }

        return(0);
    }
};*/

//------------------------------------------------------------------------------

class AltiENCBgoCommandPut : public RCD::MenuItem {
  public:
    AltiENCBgoCommandPut() { setName("put     command [FIFO]"); }
    int action() {

        unsigned int bgo = enterInt("BGO index                             ", 0, 3);
        AltiModule::TTC_FRAME_TYPE frameType = (AltiModule::TTC_FRAME_TYPE) enterInt("Frame type (0=short, 1=long)          ", 0, AltiModule::TTC_FRAME_TYPE_NUMBER - 1);
        unsigned int data;
        switch (frameType) {
            case AltiModule::TTC_FRAME_TYPE::SHORT:
                data = enterHex("Data                    ", 0, 0xff); // 8 bits
                if (alti->ENCBgoCommandPutShort(bgo, data) != 0) {
                    CERR("ALTI: putting short command (data = 0x%02x) to BGO%d FIFO", data, bgo);
                    return (-1);
                }
                break;
            case AltiModule::TTC_FRAME_TYPE::LONG:
                unsigned int addr = enterHex("TTCrx address           ", 0, 0x3fff); // 14 bits
                AltiModule::TTC_ADDRESS_SPACE addressSpace = (AltiModule::TTC_ADDRESS_SPACE) enterInt("Address space (0=internal, 1=external)", 0, AltiModule::TTC_ADDRESS_SPACE_NUMBER - 1);
                unsigned int subAddr = enterHex("Subaddress              ", 0, 0xff); // 8 bits
                data = enterHex("Data                    ", 0, 0xff); // 8 bits
                if (alti->ENCBgoCommandPutLong(bgo, addr, addressSpace, subAddr, data) != 0 ) {
                    CERR("ALTI: putting long command (\"%s\", addr = 0x%04x, subAddr = 0x%02x, data = 0x%02x) to BGO%d FIFO", AltiModule::TTC_ADDRESS_SPACE_NAME[addressSpace], addr, subAddr, data, bgo);
                    return (-1);
                }
                break;
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiENCAsyncCommandPut : public RCD::MenuItem {
  public:
    AltiENCAsyncCommandPut() { setName("send    command [async]"); }
    int action() {

        AltiModule::TTC_FRAME_TYPE frameType = (AltiModule::TTC_FRAME_TYPE) enterInt("Frame type (0=short, 1=long)          ", 0, AltiModule::TTC_FRAME_TYPE_NUMBER - 1);
        unsigned int data;
        switch (frameType) {
            case AltiModule::TTC_FRAME_TYPE::SHORT:
                data = enterHex("Data                    ", 0, 0xff); // 8 bits
                if (alti->ENCAsyncCommandPutShort(data) != 0) {
                    CERR("ALTI: sending short command (data = 0x%02x)", data);
                    return (-1);
                }
                break;
            case AltiModule::TTC_FRAME_TYPE::LONG:
                unsigned int addr = enterHex("TTCrx address           ", 0, 0x3fff); // 14 bits
                AltiModule::TTC_ADDRESS_SPACE addressSpace = (AltiModule::TTC_ADDRESS_SPACE) enterInt("Address space (0=internal, 1=external)", 0, AltiModule::TTC_ADDRESS_SPACE_NUMBER - 1);
                unsigned int subAddr = enterHex("Subaddress              ", 0, 0xff); // 8 bits
                data = enterHex("Data                    ", 0, 0xff); // 8 bits
                if (alti->ENCAsyncCommandPutLong(addr, addressSpace, subAddr, data) != 0 ) {
                    CERR("ALTI: sending long command (\"%s\", addr = 0x%04x, subAddr = 0x%02x, data = 0x%02x)", AltiModule::TTC_ADDRESS_SPACE_NAME[addressSpace], addr, subAddr, data);
                    return (-1);
                }
                break;
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiENCAsyncCommandPending : public RCD::MenuItem {
  public:
    AltiENCAsyncCommandPending() { setName("async   command pending status"); }
    int action() {

        int rtnv(0);

        bool pending;
        if ((rtnv = alti->ENCAsyncCommandPending(pending)) != 0) {
            CERR("reading async command pending status", "");
            return (rtnv);
        }
        std::printf("\n");
        std::printf("Pending status: %s\n", pending ? "sending in progress..." : "last command sent, ready for the next one");

        return(0);
    }
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

class AltiCLKStatusRead : public RCD::MenuItem {
  public:
    //AltiCLKStatusRead() { setName("read   clock  status"); }
    AltiCLKStatusRead() { setName("status"); }
    int action() {

        int rtnv(0);

        AltiModule::CLK_TYPE pll_selection;
        if ((rtnv = alti->CLKInputSelectReadMMCM(pll_selection)) != 0) {
            CERR("reading CLK input selection", "");
            return (rtnv);
        }
        //bool clklos, clklol, clklol2, jcint, jclos, jclol/*, jclatched*/;
        bool monitor_int, clklol, clklol2, jcint, jclos, jclol/*, jclatched*/, sticky_clklol, sticky_clklol2, sticky_clklos;
        //if ((rtnv = alti->CLKStatusReadMMCM(clklos, clklol, clklol2)) != 0) {
        if ((rtnv = alti->CLKStatusReadMMCM(monitor_int, clklol, clklol2, sticky_clklol, sticky_clklol2, sticky_clklos)) != 0) {
            CERR("reading CLK MMCM status", "");
            return (rtnv);
        }
        if ((rtnv = alti->CLKStatusReadJitterCleaner(jcint, jclos, jclol/*, jclatched*/)) != 0) {
            CERR("reading CLK jitter cleaner status", "");
            return (rtnv);
        }
        std::vector<bool> stopped, oor, glitch;
        //std::printf("%s", clkrst ? "CLK reset\n" : "");
        if ((rtnv = alti->CLKMonitorReadMMCM(stopped, oor, glitch)) != 0) {
            CERR("reading CLK monitor status", "");
            return (rtnv);
        }


        std::string designid;
        if ((rtnv = alti->CLKJitterCleanerDesignIDRead(designid)) != 0) {
            CERR("reading design ID from I2C JitterCleaner", "");
            return (rtnv);
        }
        std::vector<bool> oof, los;
        bool hold, lol;
        if ((rtnv = alti->CLKJitterCleanerStatusRead(oof, los, hold, lol)) != 0) {
            CERR("reading statuses from I2C JitterCleaner", "");
            return (rtnv);
        }
        std::vector<bool> sticky_oof, sticky_los;
        bool sticky_hold, sticky_lol;
        if ((rtnv = alti->CLKJitterCleanerStickyStatusRead(sticky_oof, sticky_los, sticky_hold, sticky_lol)) != 0) {
            CERR("reading sticky statuses from I2C JitterCleaner", "");
            return (rtnv);
        }
        AltiModule::CLK_TYPE jc_selection;
        if ((rtnv = alti->CLKInputSelectReadJitterCleaner(jc_selection)) != 0) {
            CERR("reading JC input selection", "");
            return (rtnv);
        }
        /*std::vector<bool> los_flg, oof_flg;
        bool lol_flg, hold_flg, losxaxb_flg, smbus_timeout_flg, cal_flg_pll, sysincal_flg;
        if ((rtnv = alti->CLKJitterCleanerIntFlagsRead(los_flg, oof_flg, lol_flg, hold_flg, losxaxb_flg, smbus_timeout_flg, cal_flg_pll, sysincal_flg)) != 0) {
            CERR("reading JC interrupt flags", "");
            return (rtnv);
        }
        std::vector<bool> los_intr_msk, oof_intr_msk;
        bool lol_intr_msk, hold_intr_msk, losxaxb_intr_msk, smbus_timeout_intr_msk, cal_intr_msk, sysincal_intr_msk;
        if ((rtnv = alti->CLKJitterCleanerIntMasksRead(los_intr_msk, oof_intr_msk, lol_intr_msk, hold_intr_msk, losxaxb_intr_msk, smbus_timeout_intr_msk, cal_intr_msk, sysincal_intr_msk)) != 0) {
            CERR("reading JC interrupt masks", "");
            return (rtnv);
        }*/

        int steps;
        if ((rtnv = alti->CLKPhasePositionReadMMCM(steps)) != 0) {
            CERR("reading PLL output clock phase position", "");
            return (rtnv);
        }

        std::printf("\n");
        //std::printf("PLL  input clock: \"%s\", %s\n", AltiModule::CLK_NAME[pll_selection].c_str(), clklos ? "loss of signal" : "signal present");
        std::printf("PLL  input  clock: %s, %s (sticky) (always OK?)\n", ((pll_selection == AltiModule::PRIMARY_FXO) ? "oscillator" : ((pll_selection == AltiModule::SECONDARY_EXT) ? "from the jitter cleaner" : "")), sticky_clklos ? "loss of signal" : "OK");
        std::printf("PLL  output clock: %-3s%s (dynamic/sticky)\n", clklol ? "ERR" : "OK", sticky_clklol ? ", loss of lock" : "");
        std::printf("PLL2 output clock: %-3s%s (dynamic/sticky)\n", clklol2 ? "ERR" : "OK", sticky_clklol2 ? ", loss of lock" : "");
        //std::printf("PLL reset: %s\n", clkrst ? "enabled" : "disabled");

        std::printf("\n");
        std::printf("PLL  output phase shift: %d steps (~15ps each)\n", steps);

        std::printf("\n");
        //std::printf("Jitter cleaner input clock: \"%s\", %s\n", AltiModule::CLK_NAME[jc_selection].c_str(), jclos ? "loss of signal" : "signal present");
        std::printf("Jitter cleaner input  clock: %s, %s (sticky) (always OK?)\n", ((jc_selection == AltiModule::PRIMARY_FXO) ? "oscillator" : ((jc_selection == AltiModule::SECONDARY_EXT) ? "from the switch" : "")), jclos ? "loss of signal" : "OK");
        std::printf("Jitter cleaner output clock: %s (sticky)\n", jclol ? "loss of lock" : "OK");
        std::printf("Jitter cleaner design ID: \"%s\"\n", designid.c_str());
        std::printf("Jitter cleaner interrupt: %s (sticky)\n", jcint ? "active" : "inactive");
        //std::printf("Jitter cleaner any alarm: %s (latched, with no ability to clear)\n", jclatched ? "YES" : "NO");
        //std::printf("Jitter cleaner reset: %s\n", jcrst ? "enabled" : "disabled");

        std::printf("%s", hold ? "\nWarning: jitter cleaner is in holdover mode!\n" : "");

        unsigned int i;
        AltiModule::EXT_CLK_SOURCE src;
        std::printf("\n");
        std::printf("(read from the VME)\n");
        std::printf("+-----------------------------------------------------------+\n");
        std::printf("| %-20s || %-7s | %-12s | %-8s |\n", "PLL input monitoring", "Signal", "Frequency", "Glitches");
        std::printf("+-----------------------------------------------------------+\n");
        for (i = 0; i < AltiModule::EXT_CLK_SOURCE_NUMBER; i++) {
            src = (AltiModule::EXT_CLK_SOURCE) i;
            std::printf("| %-20s || %-7s | %-12s | %-8s | (sticky)\n", AltiModule::EXT_CLK_SOURCE_NAME[src].c_str(), stopped[i] ? "stopped" : "OK", oor[i] ? "out of range" : "OK", glitch[i] ? "glitch" : "OK");
            std::printf("+-----------------------------------------------------------+\n");
        }
        std::printf("PLL monitor interrupt: %s (sticky) (always inactive... what causes monitor interrupt?)\n", monitor_int ? "active" : "inactive");

        std::string jc_input_name;
        std::printf("\n");
        std::printf("(read from the I2C)\n");
        std::printf("+---------------------------------------------------------------------+\n");
        std::printf("| %-31s || %-12s | %-17s |\n", "Jitter cleaner input monitoring", "Signal", "Frequency");
        std::printf("+---------------------------------------------------------------------+\n");
        for (i = 0; i < 4; i++) {
            if ((i == 2) || (i == 3)) continue; // 2-UNUSED, 3-FB
            switch (i) {
                case 0:
                    //jc_input_name = "\"" + AltiModule::CLK_NAME[AltiModule::PRIMARY_FXO] + "\"";
                    jc_input_name = "Oscillator";
                    break;
                case 1:
                    //jc_input_name = "\"" + AltiModule::CLK_NAME[AltiModule::SECONDARY_EXT] + "\"";
                    jc_input_name = "From the switch";
                    break;
                /*case 3:
                    jc_input_name = "Feedback loop";
                    break;*/
                default:
                    break;
            }
            //std::printf("| %-31s | %-7s | %-22s | (dynamic/sticky)\n", jc_input_name.c_str(), los[i] ? "YES" : "NO", oof[i] ? "YES" : "NO");
            std::printf("| %-31s || %-3s%-9s | %-3s%-14s | (dynamic/sticky)\n", jc_input_name.c_str(), los[i] ? "ERR" : "OK", sticky_los[i] ? ", stopped" : "", oof[i] ? "ERR" : "OK", sticky_oof[i] ? ", out of range" : "");
            std::printf("+---------------------------------------------------------------------+\n");
        }

        /*std::printf(" LOS[3..0]     : %s%s%s%s, mask %s%s%s%s\n", los_flg[3]?"1":"0", los_flg[2]?"1":"0", los_flg[1]?"1":"0", los_flg[0]?"1":"0", los_intr_msk[3]?"1":"0", los_intr_msk[2]?"1":"0", los_intr_msk[1]?"1":"0", los_intr_msk[0]?"1":"0");
        std::printf(" OOF[3..0]     : %s%s%s%s, mask %s%s%s%s\n", oof_flg[3]?"1":"0", oof_flg[2]?"1":"0", oof_flg[1]?"1":"0", oof_flg[0]?"1":"0", oof_intr_msk[3]?"1":"0", oof_intr_msk[2]?"1":"0", oof_intr_msk[1]?"1":"0", oof_intr_msk[0]?"1":"0");
        std::printf(" LOL           : %s,    mask %s\n", lol_flg?"1":"0", lol_intr_msk?"1":"0");
        std::printf(" HOLD          : %s,    mask %s\n", hold_flg?"1":"0", hold_intr_msk?"1":"0");
        std::printf(" LOSXAXB       : %s,    mask %s\n", losxaxb_flg?"1":"0", losxaxb_intr_msk?"1":"0");
        std::printf(" SMBUS_TIMEOUT : %s,    mask %s\n", smbus_timeout_flg?"1":"0", smbus_timeout_intr_msk?"1":"0");
        std::printf(" CAL_PLL       : %s,    mask %s\n", cal_flg_pll?"1":"0", cal_intr_msk?"1":"0");
        std::printf(" SYSINCAL      : %s,    mask %s\n", sysincal_flg?"1":"0", sysincal_intr_msk?"1":"0");
        std::printf("+--------------------------------------------------------------+\n");*/

        return (rtnv);


        /*std::printf("+--------------------------------------------------------------+\n");
        std::printf("+--------------------------------------------------------------+\n");

        std::printf("Out of frequency [FXO/EXT/NC/FB]: ");
        for (i = 0; i < 4; i++) {
            //if (i == 2) continue; // 2-UNUSED
            std::printf("%s", oof[i] ? "YES" : " NO");
            std::printf("%s", (i < 4 - 1) ? ", " : "\n");
        }
        //std::printf("Loss of signal   [FXO/EXT/FB]: ");
        std::printf("Loss of signal   [FXO/EXT/NC/FB]: ");
        for (i = 0; i < 4; i++) {
            //if (i == 2) continue; // 2-UNUSED
            std::printf("%s", los[i] ? "YES" : " NO");
            std::printf("%s", (i < 4 - 1) ? ", " : "\n");
        }
        std::printf("+--------------------------------------------------------------+\n");
        std::printf("Holdover mode: %s\n", hold ? "YES" : " NO");
        std::printf("Loss of lock:  %s\n", lol ? "YES" : " NO");
        std::printf("+--------------------------------------------------------------+\n");
        std::printf("+--------------------------------------------------------------+\n");
        std::printf("Interrupt flags:\n");
        std::printf(" LOS[3..0]     : %s%s%s%s, mask %s%s%s%s\n", los_flg[3]?"1":"0", los_flg[2]?"1":"0", los_flg[1]?"1":"0", los_flg[0]?"1":"0", los_intr_msk[3]?"1":"0", los_intr_msk[2]?"1":"0", los_intr_msk[1]?"1":"0", los_intr_msk[0]?"1":"0");
        std::printf(" OOF[3..0]     : %s%s%s%s, mask %s%s%s%s\n", oof_flg[3]?"1":"0", oof_flg[2]?"1":"0", oof_flg[1]?"1":"0", oof_flg[0]?"1":"0", oof_intr_msk[3]?"1":"0", oof_intr_msk[2]?"1":"0", oof_intr_msk[1]?"1":"0", oof_intr_msk[0]?"1":"0");
        std::printf(" LOL           : %s,    mask %s\n", lol_flg?"1":"0", lol_intr_msk?"1":"0");
        std::printf(" HOLD          : %s,    mask %s\n", hold_flg?"1":"0", hold_intr_msk?"1":"0");
        std::printf(" LOSXAXB       : %s,    mask %s\n", losxaxb_flg?"1":"0", losxaxb_intr_msk?"1":"0");
        std::printf(" SMBUS_TIMEOUT : %s,    mask %s\n", smbus_timeout_flg?"1":"0", smbus_timeout_intr_msk?"1":"0");
        std::printf(" CAL_PLL       : %s,    mask %s\n", cal_flg_pll?"1":"0", cal_intr_msk?"1":"0");
        std::printf(" SYSINCAL      : %s,    mask %s\n", sysincal_flg?"1":"0", sysincal_intr_msk?"1":"0");
        std::printf("+--------------------------------------------------------------+\n");
        std::printf("+--------------------------------------------------------------+\n");*/

    }
};

//------------------------------------------------------------------------------

class AltiCLKSelectWrite : public RCD::MenuItem {
  public:
    AltiCLKSelectWrite() { setName("select PLL input  clock"); }
    int action() {

        AltiModule::CLK_TYPE clk = (AltiModule::CLK_TYPE) enterInt("Input clock (0=\"oscillator\", 1=\"from the jitter cleaner\")", 0, 1);
        if (alti->CLKInputSelectWriteMMCM(clk) != 0) {
            CERR("writing %s clock selection", AltiModule::CLK_NAME[clk].c_str());
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiCLKSelectToggle : public RCD::MenuItem {
  public:
    AltiCLKSelectToggle() { setName("toggle PLL input  clock selection"); }
    int action() {

        if (alti->CLKInputSelectToggleMMCM() != 0) {
            CERR("toggling clock selection", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiCLKPhaseShift : public RCD::MenuItem {
  public:
    AltiCLKPhaseShift() { setName("shift  PLL output clock phase"); }
    int action() {

        bool inc  = (bool) enterInt("Shift direction (0=decrement, 1=increment)", 0, 1);
        unsigned int steps = enterInt("Number of steps (~15ps each)     ", 0, 0x3fffffff);
        if (alti->CLKPhaseShiftMMCM(inc, steps) != 0) {
            CERR("%s PLL output clock phase by %d steps", inc ? "incrementing" : "decrementing", steps);
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiCLKPhaseReset : public RCD::MenuItem {
  public:
    AltiCLKPhaseReset() { setName("reset  PLL output clock phase"); }
    int action() {

        if (alti->CLKPhaseResetMMCM() != 0) {
            CERR("resetting PLL output clock phase", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiCLKReset : public RCD::MenuItem {
  public:
    AltiCLKReset() { setName("reset  PLL"); }
    int action() {

        if (alti->CLKResetMMCM() != 0) {
            CERR("resetting PLL", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

/*class AltiCLKJitterCleanerStatusRead : public RCD::MenuItem {
  public:
    AltiCLKJitterCleanerStatusRead() { setName("read   jitter cleaner status"); }
    int action() {
        int rtnv(0);

        std::printf("+--------------------------------------------------------------+\n");
        std::string designid;
        if ((rtnv = alti->CLKJitterCleanerDesignIDRead(designid)) != 0) {
            CERR("reading design ID from I2C JitterCleaner", "");
            return (rtnv);
        }
        std::printf("DESIGN_ID: %s\n", designid.c_str());
        std::printf("+--------------------------------------------------------------+\n");

        std::vector<bool> oof, los;
        bool hold, lol;
        if ((rtnv = alti->CLKJitterCleanerStatusRead(oof, los, hold, lol)) != 0) {
            CERR("reading statuses from I2C JitterCleaner", "");
            return (rtnv);
        }
        unsigned int i;
        //std::printf("Out of frequency [FXO/EXT/FB]: ");
        std::printf("Out of frequency [FXO/EXT/NC/FB]: ");
        for (i = 0; i < 4; i++) {
            //if (i == 2) continue; // 2-UNUSED
            std::printf("%s", oof[i] ? "YES" : " NO");
            std::printf("%s", (i < 4 - 1) ? ", " : "\n");
        }
        //std::printf("Loss of signal   [FXO/EXT/FB]: ");
        std::printf("Loss of signal   [FXO/EXT/NC/FB]: ");
        for (i = 0; i < 4; i++) {
            //if (i == 2) continue; // 2-UNUSED
            std::printf("%s", los[i] ? "YES" : " NO");
            std::printf("%s", (i < 4 - 1) ? ", " : "\n");
        }
        std::printf("+--------------------------------------------------------------+\n");
        std::printf("Holdover mode: %s\n", hold ? "YES" : " NO");
        std::printf("Loss of lock:  %s\n", lol ? "YES" : " NO");
        std::printf("+--------------------------------------------------------------+\n");
        AltiModule::CLK_TYPE selection;
        if ((rtnv = alti->CLKInputSelectReadJitterCleaner(selection)) != 0) {
            CERR("reading JC input selection", "");
            return (rtnv);
        }
        std::printf("Input selection: %s\n", AltiModule::CLK_NAME[selection].c_str());
        std::printf("+--------------------------------------------------------------+\n");
        std::vector<bool> los_flg, oof_flg;
        bool lol_flg, hold_flg, losxaxb_flg, smbus_timeout_flg, cal_flg_pll, sysincal_flg;
        if ((rtnv = alti->CLKJitterCleanerIntFlagsRead(los_flg, oof_flg, lol_flg, hold_flg, losxaxb_flg, smbus_timeout_flg, cal_flg_pll, sysincal_flg)) != 0) {
            CERR("reading JC interrupt flags", "");
            return (rtnv);
        }
        std::vector<bool> los_intr_msk, oof_intr_msk;
        bool lol_intr_msk, hold_intr_msk, losxaxb_intr_msk, smbus_timeout_intr_msk, cal_intr_msk, sysincal_intr_msk;
        if ((rtnv = alti->CLKJitterCleanerIntMasksRead(los_intr_msk, oof_intr_msk, lol_intr_msk, hold_intr_msk, losxaxb_intr_msk, smbus_timeout_intr_msk, cal_intr_msk, sysincal_intr_msk)) != 0) {
            CERR("reading JC interrupt masks", "");
            return (rtnv);
        }
        std::printf("Interrupt flags:\n");
        std::printf(" LOS[3..0]     : %s%s%s%s, mask %s%s%s%s\n", los_flg[3]?"1":"0", los_flg[2]?"1":"0", los_flg[1]?"1":"0", los_flg[0]?"1":"0", los_intr_msk[3]?"1":"0", los_intr_msk[2]?"1":"0", los_intr_msk[1]?"1":"0", los_intr_msk[0]?"1":"0");
        std::printf(" OOF[3..0]     : %s%s%s%s, mask %s%s%s%s\n", oof_flg[3]?"1":"0", oof_flg[2]?"1":"0", oof_flg[1]?"1":"0", oof_flg[0]?"1":"0", oof_intr_msk[3]?"1":"0", oof_intr_msk[2]?"1":"0", oof_intr_msk[1]?"1":"0", oof_intr_msk[0]?"1":"0");
        std::printf(" LOL           : %s,    mask %s\n", lol_flg?"1":"0", lol_intr_msk?"1":"0");
        std::printf(" HOLD          : %s,    mask %s\n", hold_flg?"1":"0", hold_intr_msk?"1":"0");
        std::printf(" LOSXAXB       : %s,    mask %s\n", losxaxb_flg?"1":"0", losxaxb_intr_msk?"1":"0");
        std::printf(" SMBUS_TIMEOUT : %s,    mask %s\n", smbus_timeout_flg?"1":"0", smbus_timeout_intr_msk?"1":"0");
        std::printf(" CAL_PLL       : %s,    mask %s\n", cal_flg_pll?"1":"0", cal_intr_msk?"1":"0");
        std::printf(" SYSINCAL      : %s,    mask %s\n", sysincal_flg?"1":"0", sysincal_intr_msk?"1":"0");
        std::printf("+--------------------------------------------------------------+\n");
        std::printf("+--------------------------------------------------------------+\n");

        return (rtnv);
    }
};*/

//------------------------------------------------------------------------------

/*class AltiCLKJitterCleanerStatusClear : public RCD::MenuItem {
  public:
    AltiCLKJitterCleanerStatusClear() { setName("clear  jitter cleaner status flags"); }
    int action() {
        int rtnv(0);

        if ((rtnv = alti->CLKJitterCleanerStatusClear()) != 0) {
            CERR("clearing jitter cleaner status flag", "");
            return (rtnv);
        }

        return (rtnv);
    }
};*/

//------------------------------------------------------------------------------

class AltiCLKStatusClear : public RCD::MenuItem {
  public:
    AltiCLKStatusClear() { setName("clear  status flags"); }
    int action() {
        int rtnv(0);

        unsigned int clear_choice = (AltiModule::CLK_TYPE) enterInt("Flags to clear (0=PLL, 1=monitor , 2=jitter cleaner, 3=all)", 0, 3);

        switch (clear_choice) {
            case 0:
                if ((rtnv = alti->CLKStatusClearMMCM()) != 0) {
                    CERR("clearing PLL status flags", "");
                    return (rtnv);
                }
                break;
            case 1:
                if ((rtnv = alti->CLKMonitorStatusClearMMCM()) != 0) {
                    CERR("clearing PLL input monitor status flags", "");
                    return (rtnv);
                }
                break;
            case 2:
                if ((rtnv = alti->CLKJitterCleanerStatusClear()) != 0) {
                    CERR("clearing jitter cleaner status flags (I2C)", "");
                    return (rtnv);
                }
                if ((rtnv = alti->CLKStatusClearJitterCleaner()) != 0) {
                    CERR("clearing jitter cleaner status flags (VME)", "");
                    return (rtnv);
                }
                break;
            case 3: // all
                if ((rtnv = alti->CLKJitterCleanerStatusClear()) != 0) {
                    CERR("clearing jitter cleaner status flags (I2C)", "");
                    return (rtnv);
                }
                if ((rtnv = alti->CLKStatusClearJitterCleaner()) != 0) {
                    CERR("clearing jitter cleaner status flags (VME)", "");
                    return (rtnv);
                }
                if ((rtnv = alti->CLKStatusClearMMCM()) != 0) {
                    CERR("clearing PLL status flags", "");
                    return (rtnv);
                }
                if ((rtnv = alti->CLKMonitorStatusClearMMCM()) != 0) {
                    CERR("clearing PLL input monitor status flags", "");
                    return (rtnv);
                }
                break;
            default:
                break;
        }

        return (rtnv);
    }
};
    //int CLKStatusClearMMCM();
    //int CLKMonitorStatusClearMMCM();
    //int CLKStatusClearJitterCleaner();

//------------------------------------------------------------------------------

class AltiCLKJitterCleanerInSelWrite : public RCD::MenuItem {
  public:
    AltiCLKJitterCleanerInSelWrite() { setName("select jitter cleaner input clock"); }
    int action() {
        int rtnv(0);

        AltiModule::CLK_TYPE clk = (AltiModule::CLK_TYPE) enterInt("Input clock (0=\"oscillator\", 1=\"from the switch\")", 0, 1);
        if ((rtnv = alti->CLKInputSelectWriteJitterCleaner(clk)) != 0) {
            CERR("writing %s jitter cleaner input selection", AltiModule::CLK_NAME[clk].c_str());
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCLKJitterCleanerConfigRead : public RCD::MenuItem {
  public:
    AltiCLKJitterCleanerConfigRead() { setName("read   jitter cleaner configuration [into file]"); }
    int action() {

        int rtnv(0);
        std::string fn_in = enterString("Input  file name");
        //if (fn_in == "") fn_in = "ALTI/data/Si5344-RevB-ALTI_001-Registers.txt"; // default filename
        std::string fn_out = enterString("Output file name");
        //if (fn_out == "") fn_out = "jc.out"; // default filename

        // read CLK jitter cleaner config into file
        if ((rtnv = alti->CLKJitterCleanerConfigRead(fn_in, fn_out)) != 0) {
            CERR("reading CLK jitter cleaner config with input file \"%s\" into output file \"%s\"", fn_in.c_str(), fn_out.c_str());
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCLKJitterCleanerConfigWrite : public RCD::MenuItem {
  public:
    AltiCLKJitterCleanerConfigWrite() { setName("write  jitter cleaner configuration [from file]"); }
    int action() {

        int rtnv(0);
        std::string fn_in = enterString("Input file name");
        //if (fn_in == "") fn_in = "ALTI/data/Si5344-RevB-ALTI_001-Registers.txt"; // default filename

        // write CLK jitter cleaner config from file
        if ((rtnv = alti->CLKJitterCleanerConfigWrite(fn_in)) != 0) {
            CERR("writing CLK jitter cleaner config from input file \"%s\"", fn_in.c_str());
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCLKJitterCleanerConfigWriteDefault : public RCD::MenuItem {
  public:
    AltiCLKJitterCleanerConfigWriteDefault() { setName("write  jitter cleaner configuration [default]"); }
    int action() {

        int rtnv(0);
        bool flag = (bool) enterInt("Write default configuration (0=\"NO\", 1=\"YES\")", 0, 1);
        if (!flag) return (rtnv);

        //int rtnv(0);
        // write DEFAULT CLK jitter cleaner config
        if ((rtnv = alti->CLKJitterCleanerConfigWrite()) != 0) {
            CERR("writing DEFAULT CLK jitter cleaner config", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiCLKJitterCleanerResync : public RCD::MenuItem {
  public:
    AltiCLKJitterCleanerResync() { setName("resync jitter cleaner"); }
    int action() {

        if (alti->CLKJitterCleanerResync() != 0 ) {
            CERR("resyncing jitter cleaner", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiCLKJitterCleanerReset : public RCD::MenuItem {
  public:
    AltiCLKJitterCleanerReset() { setName("reset  jitter cleaner"); }
    int action() {

        if (alti->CLKResetJitterCleaner() != 0 ) {
            CERR("resetting jitter cleaner", "");
            return (-1);
        }

        return(0);
    }
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

class AltiCNTRead : public RCD::MenuItem {
  public:
    AltiCNTRead() { setName("read    counters"); }
    int action() {

        //AltiModule::CNT_TYPE cnt;
        //unsigned int cnt_choice = enterInt("Counter [0-BC/1-ORB/.../9-BGO3/10-ALL]:", 0, AltiModule::CNT_NUMBER);
        //if (cnt_choice < AltiModule::CNT_NUMBER) cnt = AltiModule::CNT_TYPE(cnt_choice);

        std::printf("\n");
        std::printf("+--------------------------------------------------------------+\n");
        std::printf("| Counter     | Data (hex) | Data (decimal) | Status           |\n");
        std::printf("+--------------------------------------------------------------+\n");

        /*if (cnt_choice < AltiModule::CNT_NUMBER) {
            unsigned int value;
            bool enabled, reset;
            if (alti->CNTRead(cnt, value) != 0) {
                CERR("reading %s counter value", AltiModule::CNT_NAME[cnt].c_str());
                return (-1);
            }
            if (alti->CNTStatusRead(cnt, enabled, reset) != 0) {
                CERR("reading %s counter status", AltiModule::CNT_NAME[cnt].c_str());
                return (-1);
            }
            std::printf("| [%01d]%-8s | 0x%08x | %010u     | %s %s |\n", cnt_choice, AltiModule::CNT_NAME[cnt_choice].c_str(), value, value, enabled ? "Enabled " : "Disabled", reset ? "/ Reset" : "       ");
            std::printf("+--------------------------------------------------------------+\n");
        }*/
        //else { // all
            std::vector<unsigned int> value;
            value.resize(AltiModule::CNT_NUMBER);
            std::vector<bool> enabled;
            enabled.resize(AltiModule::CNT_NUMBER);
            std::vector<bool> reset;
            reset.resize(AltiModule::CNT_NUMBER);
            if (alti->CNTRead(value) != 0) {
                CERR("reading all counters", "");
                return (-1);
            }
            if (alti->CNTStatusRead(enabled, reset) != 0) {
                CERR("reading status of all counters", "");
                return (-1);
            }
            unsigned int i;
            for (i = 0; i < value.size(); i++) {
                std::printf("| [%01d]%-8s | 0x%08x | %010u     | %s %s |\n", i, AltiModule::CNT_NAME[i].c_str(), value[i], value[i], enabled[i] ? "Enabled " : "Disabled", reset[i] ? "/ Reset" : "       ");
                std::printf("+--------------------------------------------------------------+\n");
            }
        //}

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiCNTEnable : public RCD::MenuItem {
  public:
    AltiCNTEnable() { setName("enable  counters"); }
    int action() {

        AltiModule::CNT_TYPE cnt;
        unsigned int cnt_choice = enterInt("Counter (0=\"BC\", 1=\"ORB\", ..., 9=\"BGO3\", 10=all)", 0, AltiModule::CNT_NUMBER);
        if (cnt_choice < AltiModule::CNT_NUMBER) cnt = AltiModule::CNT_TYPE(cnt_choice);
        //unsigned int enable = enterInt("Enable [1-Enabled/0-Disabled]:          ", 0, 1);

        if (cnt_choice < AltiModule::CNT_NUMBER) {
            /*if (alti->CNTEnable(cnt, enable) != 0) {
                CERR("%s %s counter", enable ? "enabling" : "disabling", AltiModule::CNT_NAME[cnt].c_str());
                return (-1);
            }*/

            if (alti->CNTEnable(cnt, true) != 0) {
                CERR("enabling %s counter", "", AltiModule::CNT_NAME[cnt].c_str());
                return (-1);
            }
        }
        else { // all
            /*if (alti->CNTEnable(enable) != 0) {
                CERR("%s all counters", enable ? "enabling" : "disabling");
                return (-1);
            }*/

            if (alti->CNTEnable(true) != 0) {
                CERR("enabling all counters", "");
                return (-1);
            }
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiCNTDisable : public RCD::MenuItem {
  public:
    AltiCNTDisable() { setName("disable counters"); }
    int action() {

        AltiModule::CNT_TYPE cnt;
        unsigned int cnt_choice = enterInt("Counter (0=\"BC\", 1=\"ORB\", ..., 9=\"BGO3\", 10=all)", 0, AltiModule::CNT_NUMBER);
        if (cnt_choice < AltiModule::CNT_NUMBER) cnt = AltiModule::CNT_TYPE(cnt_choice);
        //unsigned int enable = enterInt("Enable [1-Enabled/0-Disabled]:          ", 0, 1);

        if (cnt_choice < AltiModule::CNT_NUMBER) {
            /*if (alti->CNTEnable(cnt, enable) != 0) {
                CERR("%s %s counter", enable ? "enabling" : "disabling", AltiModule::CNT_NAME[cnt].c_str());
                return (-1);
            }*/

            if (alti->CNTEnable(cnt, false) != 0) {
                CERR("disabling %s counter", "", AltiModule::CNT_NAME[cnt].c_str());
                return (-1);
            }
        }
        else { // all
            /*if (alti->CNTEnable(enable) != 0) {
                CERR("%s all counters", enable ? "enabling" : "disabling");
                return (-1);
            }*/

            if (alti->CNTEnable(false) != 0) {
                CERR("disabling all counters", "");
                return (-1);
            }
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiCNTReset : public RCD::MenuItem {
  public:
    AltiCNTReset() { setName("reset   counters"); }
    int action() {

        AltiModule::CNT_TYPE cnt;
        unsigned int cnt_choice = enterInt("Counter (0=\"BC\", 1=\"ORB\", ..., 9=\"BGO3\", 10=all)", 0, AltiModule::CNT_NUMBER);
        if (cnt_choice < AltiModule::CNT_NUMBER) cnt = AltiModule::CNT_TYPE(cnt_choice);
        //unsigned int enable = enterInt("Reset [1-Enabled/0-Disabled]:           ", 0, 1);

        if (cnt_choice < AltiModule::CNT_NUMBER) {
            /*if (alti->CNTReset(cnt, enable) != 0) {
                CERR("%s %s counter reset", enable ? "enabling" : "disabling", AltiModule::CNT_NAME[cnt].c_str());
                return (-1);
            }*/

            /*if (alti->CNTReset(cnt, true) != 0) {
                CERR("enabling %s counter reset", AltiModule::CNT_NAME[cnt].c_str());
                return (-1);
            }

            if (alti->CNTReset(cnt, false) != 0) {
                CERR("disabling %s counter reset", AltiModule::CNT_NAME[cnt].c_str());
                return (-1);
            }*/

            if (alti->CNTReset(cnt) != 0) {
                CERR("resetting %s counter", AltiModule::CNT_NAME[cnt].c_str());
                return (-1);
            }
        }
        else { // all
            /*if (alti->CNTReset(enable) != 0) {
                CERR("%s reset for all counters", enable ? "enabling" : "disabling");
                return (-1);
            }*/

            /*if (alti->CNTReset(true) != 0) {
                CERR("enabling reset for all counters", "");
                return (-1);
            }

            if (alti->CNTReset(false) != 0) {
                CERR("disabling reset for all counters", "");
                return (-1);
            }*/

            if (alti->CNTReset() != 0) {
                CERR("resetting all counters", "");
                return (-1);
            }
        }

        return(0);
    }
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

class ECRGenerationRead : public MenuItem {
  public:
    ECRGenerationRead() { setName("read  ECR generation"); }
    int action() {

        int rtnv(0);
        AltiModule::ECR_GENERATION ecr;

        /*for(unsigned int part=CtpcorePlusModule::SECONDARY_PART_OFFSET; part<CtpcorePlusModule::PARTITION_NUMBER; part++) {
            if((rtnv = ctpcoreplus->ECRGenerationRead(part,ecr)) != CtpcorePlusModule::SUCCESS) {
                CERR("reading ECR control of secondary partition %d",part);
            }
            std::cout << "--------------------------------------------" << std::endl;
            std::printf("SECONDARY PARTITION %d:\n",part);
            ecr.dump();
        }*/
        if ((rtnv = alti->ECRGenerationRead(ecr)) != AltiModule::SUCCESS) {
            CERR("reading ECR control", "");
        }
        std::cout << "--------------------------------------------" << std::endl;
        //std::printf("SECONDARY PARTITION %d:\n",part);
        ecr.dump();

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class ECRGenerationWrite : public MenuItem {
  public:
    ECRGenerationWrite() { setName("write ECR generation"); }
    int action() {

        int rtnv(0);
        AltiModule::ECR_GENERATION ecr;
        //unsigned int part = enterInt("Secondary trigger partition number (3=both)",CtpcorePlusModule::SECONDARY_PART_OFFSET,CtpcorePlusModule::PARTITION_NUMBER);
        //unsigned int palo = (part == CtpcorePlusModule::PARTITION_NUMBER) ? CtpcorePlusModule::SECONDARY_PART_OFFSET : part;
        //unsigned int pahi = (part == CtpcorePlusModule::PARTITION_NUMBER) ? CtpcorePlusModule::PARTITION_NUMBER : part+1;

        ecr.type = (AltiModule::ECR_TYPE) enterInt("ECR type (0=\"VME\", 1=\"INTERNAL\")", 0, 1);
        ecr.length = enterInt("ECR length", 1, AltiModule::ECR_LENGTH);
        if (ecr.type == AltiModule::ECR_INTERNAL) {
            ecr.frequency = enterNumber("ECR frequency", 0, AltiModule::ECR_FREQUENCY_MASK);
        }
        ecr.busy_before = enterNumber("ECR BUSY BEFORE length", 0, AltiModule::ECR_BUSY_BEFORE_MASK);
        ecr.busy_after = enterNumber("ECR BUSY AFTER length", 0, AltiModule::ECR_BUSY_AFTER_MASK);
        ecr.orbit_offset = enterNumber("ECR ORBIT offset", 0, AltiModule::ECR_ORBIT_OFFSET_MASK);

        /*for(part=palo; part<pahi; part++) {
            if((rtnv = ctpcoreplus->ECRGenerationWrite(part,ecr)) != CtpcorePlusModule::SUCCESS) {
                CERR("writing ECR control of secondary partition %d",part);
                return(rtnv);
            }
        }*/
        if ((rtnv = alti->ECRGenerationWrite(ecr)) != AltiModule::SUCCESS) {
            CERR("writing ECR control", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class ECRGenerate : public MenuItem {
  public:
    ECRGenerate() { setName("write ECR generate"); }
    int action() {

        int rtnv(0);
        //unsigned int part = enterInt("Secondary trigger partition number",CtpcorePlusModule::SECONDARY_PART_OFFSET,CtpcorePlusModule::PARTITION_NUMBER-1);
        unsigned int loop = enterInt("How many times", 0, 100000);

        /*for(unsigned int i=0; i<loop; i++) {
            if((rtnv = ctpcoreplus->ECRGenerate(part)) != CtpcorePlusModule::SUCCESS) {
                CERR("writing ECR generate of secondary partition %d",part);
                return(rtnv);
            }
            if((rtnv = ctpcoreplus->ECRReadyWait(part)) != CtpcorePlusModule::SUCCESS) {
                CERR("waiting for ECR generation of secondary partition %d to be ready again",part);
                return(rtnv);
            }
        }*/
        for (unsigned int i = 0; i < loop; i++) {
            if ((rtnv = alti->ECRGenerate()) != AltiModule::SUCCESS) {
                CERR("writing ECR generate", "");
                return (rtnv);
            }
            if ((rtnv = alti->ECRReadyWait()) != AltiModule::SUCCESS) {
                CERR("waiting for ECR generation to be ready again", "");
                return (rtnv);
            }
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

class T2LControlReset : public MenuItem {
  public:
    T2LControlReset() { setName("reset T2L control"); }
    int action() {

        int rtnv(0);

        //unsigned int part = enterInt("Trigger partition number (3=all)",0,CtpcorePlusModule::PARTITION_NUMBER);
        //unsigned int palo = (part == CtpcorePlusModule::PARTITION_NUMBER) ? 0 : part;
        //unsigned int pahi = (part == CtpcorePlusModule::PARTITION_NUMBER) ? CtpcorePlusModule::PARTITION_NUMBER : part+1;
        //bool l1id = static_cast<bool>(enterInt("Reset L1ID (0-no, 1-yes)",0,1));
        //bool ecrn = static_cast<bool>(enterInt("Reset ECR  (0-no, 1-yes)",0,1));
        //bool fifo = static_cast<bool>(enterInt("Reset FIFO (0-no, 1-yes)",0,1));

        bool l1id = static_cast<bool>(enterInt("Reset L1ID (0=no, 1=yes)", 0, 1));
        bool ecrn = static_cast<bool>(enterInt("Reset ECR  (0=no, 1=yes)", 0, 1));
        bool fifo = static_cast<bool>(enterInt("Reset FIFO (0=no, 1=yes)", 0, 1));

        /*// reset T2L control
        for(part=palo; part<pahi; part++) {
            if((rtnv = ctpcoreplus->T2LControlReset(part,l1id,ecrn,fifo)) != CtpcorePlusModule::SUCCESS) {
                CERR("resetting T2L control of partition %d",part);
                return(rtnv);
            }
        }*/

        // reset T2L control
        if ((rtnv = alti->T2LControlReset(l1id, ecrn, fifo)) != AltiModule::SUCCESS) {
            CERR("resetting T2L control", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class T2LEventIdentifierRead : public MenuItem {
  public:
    T2LEventIdentifierRead() { setName("read  T2L event identifier [sample]"); }
    int action() {

        int rtnv(0);
        unsigned int sec, i;
        EventIdentifier evid;
        EventCounter ecr, l1a;
        ecr.mask(EventIdentifier::ECRC_MAX);
        l1a.mask(EventIdentifier::L1ID_MAX);

        //unsigned int part = enterInt("Trigger partition number",0,CtpcorePlusModule::PARTITION_NUMBER-1);
        unsigned int loop = enterInt("How many times", 0, 100000);

        if (loop == 1) {
            if ((rtnv = alti->T2LEventIdentifierRead(evid)) != AltiModule::SUCCESS) {
                CERR("reading T2L event identifier", "");
                return (rtnv);
            }
            std::printf("T2L EVID = 0x%08x (ecr = %3d, l1a = %8d)\n", evid.data(), evid.ECRC(), evid.L1ID());
        }
        else {
            sec = enterInt("How many seconds", 0, 10);
            for (i = 0; i < loop; i++) {
                if (i > 0) sleep(sec);
                if ((rtnv = alti->T2LEventIdentifierRead(evid)) != AltiModule::SUCCESS) {
                    CERR("reading T2L event identifier", "");
                    return (rtnv);
                }
                ecr.data(evid.ECRC());
                l1a.data(evid.L1ID());
                std::printf("T2L EVID[%5d sec]: 0x%08x (ecr = %s, l1a = %s)\n", i*sec, evid.data(), ecr.string().c_str(), l1a.string().c_str());
            }
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class T2LFifoStatusRead : public MenuItem {
  public:
    T2LFifoStatusRead() { setName("read  T2L FIFO status"); }
    int action() {

        int rtnv(0);
        unsigned int lvl;
        bool emp, ful;

        /*// read T2L FIFO status
        for(unsigned int part=0; part<CtpcorePlusModule::PARTITION_NUMBER; part++) {
            if((rtnv = ctpcoreplus->T2LFifoStatusRead(part,emp,ful,lvl)) != CtpcorePlusModule::SUCCESS) {
                CERR("reading T2L FIFO status of partition %d",part);
                return(rtnv);
            }
            std::printf("T2L FIFO[PARTITION %d] status: %s, %s, level = %2d\n",part,emp?"    EMPTY":"not empty",ful?"    FULL":"not full",lvl);
        }*/

        // read T2L FIFO status
        if((rtnv = alti->T2LFifoStatusRead(emp, ful, lvl)) != AltiModule::SUCCESS) {
            CERR("reading T2L FIFO status", "");
            return (rtnv);
        }
        std::printf("T2L FIFO status: %s, %s, level = %2d\n", emp ? "    EMPTY" : "not empty", ful ? "    FULL" : "not full", lvl);

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class T2LFifoWrite : public MenuItem {
  public:
    T2LFifoWrite() { setName("write T2L FIFO"); }
    int action() {

        int rtnv(0);

        //unsigned int part = enterInt("Trigger partition number (3=all)",0,CtpcorePlusModule::PARTITION_NUMBER);
        //unsigned int palo = (part == CtpcorePlusModule::PARTITION_NUMBER) ? 0 : part;
        //unsigned int pahi = (part == CtpcorePlusModule::PARTITION_NUMBER) ? CtpcorePlusModule::PARTITION_NUMBER : part+1;

        /*// write T2L FIFO
        for(part=palo; part<pahi; part++) {
            if((rtnv = ctpcoreplus->T2LFifoWrite(part)) != CtpcorePlusModule::SUCCESS) {
                CERR("writing T2L FIFO of partition %d",part);
                return(rtnv);
            }
        }*/

        // write T2L FIFO
        if ((rtnv = alti->T2LFifoWrite()) != AltiModule::SUCCESS) {
            CERR("writing T2L FIFO", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class T2LFifoRead : public MenuItem {
  public:
    T2LFifoRead() { setName("read  T2L FIFO"); }
    int action() {

        int rtnv(0);
        std::vector<unsigned int> data;
        EventIdentifier evid;

        //unsigned int part = enterInt("Trigger partition number",0,CtpcorePlusModule::PARTITION_NUMBER-1);

        // read T2L FIFO
        if ((rtnv = alti->T2LFifoRead(data)) != AltiModule::SUCCESS) {
            CERR("reading T2L FIFO", "");
            return (rtnv);
        }
        if (data.empty()) {
           std::printf("T2L FIFO is EMPTY\n");
            return (rtnv);
        }
        for (unsigned int i = 0; i < data.size(); i++) {
            evid.data(data[i]);
            std::printf("T2L FIFO WORD[%2d]: 0x%08x (ecr = %3d, l1a = %8d)\n", i, data[i], evid.ECRC(), evid.L1ID());
        }

        return (rtnv);
    }
};
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

class CNTBcidMaskRead : public MenuItem {
  public:
    CNTBcidMaskRead() { setName("read  CNT BCID mask"); }
    int action() {

        int rtnv(0);
        //unsigned int type = enterInt("CNT type (0=LUT, 1=TBP, 2=TAP, 3=TAV)",0,CtpcorePlusModule::CNT_NUMBER-1);
        unsigned int bcid = enterInt("BCID number (4096=all)", 0, AltiModule::BCID_NUMBER);

        if (bcid == AltiModule::BCID_NUMBER) {
            std::vector<bool> ena;
            if ((rtnv = alti->CNTBcidMaskRead(ena)) != AltiModule::SUCCESS) {
                CERR("reading CNT BCID mask", "");
                return (rtnv);
            }
            std::printf("CNT:\n");
            for (unsigned int i = 0; i < AltiModule::BCID_NUMBER; i++) {
                std::printf("BCID %4d = %s\n", i, ena[i] ? " ENABLED" : "DISABLED");
            }
        }

        else {
            unsigned int num = enterInt("BCID range", 0, AltiModule::BCID_NUMBER - bcid);
            bool ena;
            std::printf("CNT:\n");
            for (unsigned int i = bcid; i < (bcid + num); i++) {
                if ((rtnv = alti->CNTBcidMaskRead(i, ena)) != AltiModule::SUCCESS) {
                    CERR("reading CNT BCID mask %d", i);
                    return (rtnv);
                }
                std::printf("BCID %4d = %s\n", i, ena ? " ENABLED" : "DISABLED");
            }
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTBcidMaskWrite : public MenuItem {
  public:
    CNTBcidMaskWrite() { setName("write CNT BCID mask"); }
    int action() {

        int rtnv(0);
        //unsigned int type = enterInt("CNT type (0=LUT, 1=TBP, 2=TAP, 3=TAV)",0,CtpcorePlusModule::CNT_NUMBER-1);
        unsigned int bcid = enterInt("BCID number (4096=all)", 0, AltiModule::BCID_NUMBER);

        if (bcid == AltiModule::BCID_NUMBER) {
            bool ena = (bool) enterInt("Enable (0=no, 1=yes)", 0, 1);
            std::vector<bool> fna(AltiModule::BCID_NUMBER);
            for (unsigned int i = 0; i < AltiModule::BCID_NUMBER; i++) fna[i] = ena;
            if ((rtnv = alti->CNTBcidMaskWrite(fna)) != AltiModule::SUCCESS) {
                CERR("writing CNT BCID mask", "");
                return (rtnv);
            }
        }

        else {
            unsigned int num = enterInt("BCID range", 0, AltiModule::BCID_NUMBER - bcid);
            bool ena = (bool) enterInt("Enable (0=no, 1=yes)", 0, 1);
            for (unsigned int i = bcid; i < (bcid + num); i++) {
                if ((rtnv = alti->CNTBcidMaskWrite(i, ena)) != AltiModule::SUCCESS) {
                    CERR("writing CNT BCID mask %d", i);
                    return (rtnv);
                }
            }
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTBcidMaskReadFile : public MenuItem {
  public:
    CNTBcidMaskReadFile() { setName("read  CNT BCID mask [into file]"); }
    int action() {

        int rtnv(0);
        //unsigned int type = enterInt("CNT type (0=LUT, 1=TBP, 2=TAP, 3=TAV)",0,CtpcorePlusModule::CNT_NUMBER-1);
        std::string fn = enterString("File name");

        if ((rtnv = alti->CNTBcidMaskRead(fn)) != AltiModule::SUCCESS) {
            CERR("reading CNT BCID mask into file \"%s\"", fn.c_str());
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTBcidMaskWriteFile : public MenuItem {
  public:
    CNTBcidMaskWriteFile() { setName("write CNT BCID mask [from file]"); }
    int action() {

        int rtnv(0);
        //unsigned int type = enterInt("CNT type (0=LUT, 1=TBP, 2=TAP, 3=TAV)",0,CtpcorePlusModule::CNT_NUMBER-1);
        std::string fn = enterString("File name");

        if ((rtnv = alti->CNTBcidMaskWrite(fn)) != AltiModule::SUCCESS) {
            CERR("writing CNT BCID mask from file \"%s\"", fn.c_str());
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTBcidOffsetRead : public MenuItem {
  public:
    CNTBcidOffsetRead() { setName("read  CNT BCID offset"); }
    int action() {

        int rtnv(0);
        unsigned int /*typ, */off;

        /*for(typ=0; typ<CtpcorePlusModule::CNT_NUMBER; typ++) {
            if((rtnv = ctpcoreplus->CNTBcidOffsetRead(static_cast<CtpcorePlusModule::CNT_TYPE>(typ),off)) != CtpcorePlusModule::SUCCESS) {
                CERR("reading CNT \"%s\" BCID offset",CtpcorePlusModule::CNT_NAME[typ].c_str());
                return(rtnv);
            }
            std::printf("CNT \"%s\" BCID offset = %4d (= 0x%03x)\n",CtpcorePlusModule::CNT_NAME[typ].c_str(),off,off);
        }*/
        if ((rtnv = alti->CNTBcidOffsetRead(off)) != AltiModule::SUCCESS) {
            CERR("reading CNT BCID offset", "");
            return (rtnv);
        }
        std::printf("CNT BCID offset = %4d (= 0x%03x)\n", off, off);

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTBcidOffsetWrite : public MenuItem {
  public:
    CNTBcidOffsetWrite() { setName("write CNT BCID offset"); }
    int action() {

        int rtnv(0);
        //unsigned int type = enterInt("CNT type (0=LUT, 1=TBP, 2=TAP, 3=TAV, 4=all)",0,CtpcorePlusModule::CNT_NUMBER);
        unsigned int off = enterInt("CNT BCID offset", 0, AltiModule::BCID_NUMBER);

        /*if(type == CtpcorePlusModule::CNT_NUMBER) {
            for(unsigned int typ=0; typ<CtpcorePlusModule::CNT_NUMBER; typ++) {
                if((rtnv = ctpcoreplus->CNTBcidOffsetWrite(static_cast<CtpcorePlusModule::CNT_TYPE>(typ),off)) != CtpcorePlusModule::SUCCESS) {
                    CERR("writing CNT \"%s\" BCID offset",CtpcorePlusModule::CNT_NAME[typ].c_str());
                     return(rtnv);
                }
            }
        }
        else {
            if((rtnv = ctpcoreplus->CNTBcidOffsetWrite(static_cast<CtpcorePlusModule::CNT_TYPE>(type),off)) != CtpcorePlusModule::SUCCESS) {
                 CERR("writing CNT \"%s\" BCID offset",CtpcorePlusModule::CNT_NAME[type].c_str());
                 return(rtnv);
            }
        }*/
        if ((rtnv = alti->CNTBcidOffsetWrite(off)) != AltiModule::SUCCESS) {
             CERR("writing CNT BCID offset", "");
             return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTControlPrint : public MenuItem {
  public:
    CNTControlPrint() { setName("print CNT control"); }
    int action() {

        int rtnv(0);

        /*ifor(unsigned int typ=0; typ<CtpcorePlusModule::CNT_NUMBER; typ++) {
            std::printf("\nCNT \"%s\":\n",CtpcorePlusModule::CNT_NAME[typ].c_str());
            if((rtnv = ctpcoreplus->CNTControlPrint(static_cast<CtpcorePlusModule::CNT_TYPE>(typ))) != CtpcorePlusModule::SUCCESS) {
                CERR("printing CNT \"%s\" control",CtpcorePlusModule::CNT_NAME[typ].c_str());
                 return(rtnv);
            }
        }*/
        std::printf("\nCNT:\n");
        if ((rtnv = alti->CNTControlPrint()) != AltiModule::SUCCESS) {
            CERR("printing CNT control", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTEnableRead : public MenuItem {
  public:
    CNTEnableRead() { setName("read  CNT enable"); }
    int action() {

        int rtnv(0);
        bool ena, sta;

        /*for(unsigned int typ=0; typ<CtpcorePlusModule::CNT_NUMBER; typ++) {
            if((rtnv = ctpcoreplus->CNTEnableRead(static_cast<CtpcorePlusModule::CNT_TYPE>(typ),ena,sta)) != CtpcorePlusModule::SUCCESS) {
                CERR("reading CNT \"%s\" control",CtpcorePlusModule::CNT_NAME[typ].c_str());
                 return(rtnv);
            }
            std::printf("CNT \"%s\": %s, status = %s\n",CtpcorePlusModule::CNT_NAME[typ].c_str(),ena?" ENABLED":"DISABLED",sta?" ENABLED":"DISABLED");
        }*/
        if ((rtnv = alti->CNTEnableRead(ena, sta)) != AltiModule::SUCCESS) {
            CERR("reading CNT control", "");
            return (rtnv);
        }
        std::printf("CNT: %s, status = %s\n", ena ? " ENABLED" : "DISABLED", sta ? " ENABLED" : "DISABLED");

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTEnableWrite : public MenuItem {
  public:
    CNTEnableWrite() { setName("write CNT enable"); }
    int action() {

        int rtnv(0);
        //unsigned int type = enterInt("CNT type (0=LUT, 1=TBP, 2=TAP, 3=TAV, 4=all)",0,CtpcorePlusModule::CNT_NUMBER);
        bool ena = (bool) enterInt("Enable (0=no, 1=yes)", 0, 1);

        /*if(type == CtpcorePlusModule::CNT_NUMBER) {
            for(unsigned int typ=0; typ<CtpcorePlusModule::CNT_NUMBER; typ++) {
                if((rtnv = ctpcoreplus->CNTEnableWrite(static_cast<CtpcorePlusModule::CNT_TYPE>(typ),ena)) != CtpcorePlusModule::SUCCESS) {
                    CERR("writing CNT \"%s\" enable",CtpcorePlusModule::CNT_NAME[typ].c_str());
                     return(rtnv);
                }
            }
        }
        else {
            if((rtnv = ctpcoreplus->CNTEnableWrite(static_cast<CtpcorePlusModule::CNT_TYPE>(type),ena)) != CtpcorePlusModule::SUCCESS) {
                CERR("writing CNT \"%s\" enable",CtpcorePlusModule::CNT_NAME[type].c_str());
                 return(rtnv);
            }
        }*/
        if ((rtnv = alti->CNTEnableWrite(ena)) != AltiModule::SUCCESS) {
            CERR("writing CNT enable", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTEnableWait : public MenuItem {
  public:
    CNTEnableWait() { setName("wait  CNT enable"); }
    int action() {

        int rtnv(0);
        //unsigned int type = enterInt("CNT type (0=LUT, 1=TBP, 2=TAP, 3=TAV, 4=all)",0,CtpcorePlusModule::CNT_NUMBER);
        bool ena = (bool) enterInt("Enable (0=no, 1=yes)", 0, 1);

        /*if(type == CtpcorePlusModule::CNT_NUMBER) {
            for(unsigned int typ=0; typ<CtpcorePlusModule::CNT_NUMBER; typ++) {
                if((rtnv = ctpcoreplus->CNTEnableWait(static_cast<CtpcorePlusModule::CNT_TYPE>(typ),ena)) != CtpcorePlusModule::SUCCESS) {
                    CERR("waiting for CNT \"%s\" enable",CtpcorePlusModule::CNT_NAME[typ].c_str());
                     return(rtnv);
                }
            }
        }
        else {
            if((rtnv = ctpcoreplus->CNTEnableWait(static_cast<CtpcorePlusModule::CNT_TYPE>(type),ena)) != CtpcorePlusModule::SUCCESS) {
                CERR("waiting for CNT \"%s\" enable",CtpcorePlusModule::CNT_NAME[type].c_str());
                 return(rtnv);
            }
        }*/
        if ((rtnv = alti->CNTEnableWait(ena)) != AltiModule::SUCCESS) {
            CERR("waiting for CNT enable", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTClearWrite : public MenuItem {
  public:
    CNTClearWrite() { setName("write CNT clear"); }
    int action() {

        int rtnv(0);
        //unsigned int type = enterInt("CNT type (0=LUT, 1=TBP, 2=TAP, 3=TAV, 4=all)",0,CtpcorePlusModule::CNT_NUMBER);

        /*if(type == CtpcorePlusModule::CNT_NUMBER) {
            for(unsigned int typ=0; typ<CtpcorePlusModule::CNT_NUMBER; typ++) {
                if((rtnv = ctpcoreplus->CNTClearWrite(static_cast<CtpcorePlusModule::CNT_TYPE>(typ))) != CtpcorePlusModule::SUCCESS) {
                    CERR("writing CNT \"%s\" clear",CtpcorePlusModule::CNT_NAME[typ].c_str());
                     return(rtnv);
                }
            }
        }
        else {
            if((rtnv = ctpcoreplus->CNTClearWrite(static_cast<CtpcorePlusModule::CNT_TYPE>(type))) != CtpcorePlusModule::SUCCESS) {
                CERR("writing CNT \"%s\" clear",CtpcorePlusModule::CNT_NAME[type].c_str());
                 return(rtnv);
            }
        }*/
        if ((rtnv = alti->CNTClearWrite()) != AltiModule::SUCCESS) {
            CERR("writing CNT clear", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTClearWait : public MenuItem {
  public:
    CNTClearWait() { setName("wait  CNT clear"); }
    int action() {

        int rtnv(0);
        //unsigned int type = enterInt("CNT type (0=LUT, 1=TBP, 2=TAP, 3=TAV, 4=all)",0,CtpcorePlusModule::CNT_NUMBER);

        /*if(type == CtpcorePlusModule::CNT_NUMBER) {
            for(unsigned int typ=0; typ<CtpcorePlusModule::CNT_NUMBER; typ++) {
                if((rtnv = ctpcoreplus->CNTClearWait(static_cast<CtpcorePlusModule::CNT_TYPE>(typ))) != CtpcorePlusModule::SUCCESS) {
                    CERR("waiting for CNT \"%s\" clear",CtpcorePlusModule::CNT_NAME[typ].c_str());
                     return(rtnv);
                }
            }
        }
        else {
            if((rtnv = ctpcoreplus->CNTClearWait(static_cast<CtpcorePlusModule::CNT_TYPE>(type))) != CtpcorePlusModule::SUCCESS) {
                CERR("waiting for CNT \"%s\" clear",CtpcorePlusModule::CNT_NAME[type].c_str());
                 return(rtnv);
            }
        }*/
        if ((rtnv = alti->CNTClearWait()) != AltiModule::SUCCESS) {
            CERR("waiting for CNT clear", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTCopyWrite : public MenuItem {
  public:
    CNTCopyWrite() { setName("write CNT copy"); }
    int action() {

        int rtnv(0);
        //unsigned int type = enterInt("CNT type (0=LUT, 1=TBP, 2=TAP, 3=TAV, 4=all)",0,CtpcorePlusModule::CNT_NUMBER);

        /*if(type == CtpcorePlusModule::CNT_NUMBER) {
            for(unsigned int typ=0; typ<CtpcorePlusModule::CNT_NUMBER; typ++) {
                if((rtnv = ctpcoreplus->CNTCopyWrite(static_cast<CtpcorePlusModule::CNT_TYPE>(typ))) != CtpcorePlusModule::SUCCESS) {
                    CERR("writing CNT \"%s\" copy",CtpcorePlusModule::CNT_NAME[typ].c_str());
                    return(rtnv);
                }
            }
        }
        else {
            if((rtnv = ctpcoreplus->CNTCopyWrite(static_cast<CtpcorePlusModule::CNT_TYPE>(type))) != CtpcorePlusModule::SUCCESS) {
                CERR("writing CNT \"%s\" copy",CtpcorePlusModule::CNT_NAME[type].c_str());
                return(rtnv);
            }
        }*/
        if ((rtnv = alti->CNTCopyWrite()) != AltiModule::SUCCESS) {
            CERR("writing CNT copy", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTCopyWait : public MenuItem {
  public:
    CNTCopyWait() { setName("wait  CNT copy"); }
    int action() {

        int rtnv(0);
        //unsigned int type = enterInt("CNT type (0=LUT, 1=TBP, 2=TAP, 3=TAV, 4=all)",0,CtpcorePlusModule::CNT_NUMBER);

        /*if(type == CtpcorePlusModule::CNT_NUMBER) {
            for(unsigned int typ=0; typ<CtpcorePlusModule::CNT_NUMBER; typ++) {
                if((rtnv = ctpcoreplus->CNTCopyWait(static_cast<CtpcorePlusModule::CNT_TYPE>(typ))) != CtpcorePlusModule::SUCCESS) {
                    CERR("waiting for CNT \"%s\" copy",CtpcorePlusModule::CNT_NAME[typ].c_str());
                     return(rtnv);
                }
            }
        }
        else {
            if((rtnv = ctpcoreplus->CNTCopyWait(static_cast<CtpcorePlusModule::CNT_TYPE>(type))) != CtpcorePlusModule::SUCCESS) {
                CERR("waiting for CNT \"%s\" copy",CtpcorePlusModule::CNT_NAME[type].c_str());
                 return(rtnv);
            }
        }*/
        if ((rtnv = alti->CNTCopyWait()) != AltiModule::SUCCESS) {
            CERR("waiting for CNT copy", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTCopyClearWrite : public MenuItem {
  public:
    CNTCopyClearWrite() { setName("write CNT copy/clear"); }
    int action() {

        int rtnv(0);

        if ((rtnv = alti->CNTCopyClearWrite()) != AltiModule::SUCCESS) {
            CERR("writing CNT copy/clear", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTCopyClearWait : public MenuItem {
  public:
    CNTCopyClearWait() { setName("wait  CNT copy/clear"); }
    int action() {

        int rtnv(0);

        if ((rtnv = alti->CNTCopyClearWait()) != AltiModule::SUCCESS) {
            CERR("waiting for CNT copy/clear", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTTurnCountRead : public MenuItem {
  public:
    CNTTurnCountRead() { setName("read  CNT TURN count"); }
    int action() {

        int rtnv(0);
        unsigned int turn;

        /*for(unsigned int typ=0; typ<CtpcorePlusModule::CNT_NUMBER; typ++) {
            if((rtnv = ctpcoreplus->CNTTurnCountRead(static_cast<CtpcorePlusModule::CNT_TYPE>(typ),turn)) != CtpcorePlusModule::SUCCESS) {
                CERR("reading CNT \"%s\" TURN count",CtpcorePlusModule::CNT_NAME[typ].c_str());
                 return(rtnv);
            }
            std::printf("CNT \"%s\" TURN count = 0x%08x (= %10d)\n",CtpcorePlusModule::CNT_NAME[typ].c_str(),turn,turn);
        }*/
        if ((rtnv = alti->CNTTurnCountRead(turn)) != AltiModule::SUCCESS) {
            CERR("reading CNT TURN count", "");
            return (rtnv);
        }
        std::printf("CNT TURN count = 0x%08x (= %10d)\n", turn, turn);

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTCounterRead : public MenuItem {
  public:
    CNTCounterRead() { setName("read  CNT values"); }
    int action() {

        int rtnv(0);
       
        // create CMEM segment
        std::string sn = "menuAltiModule[CNT]";
        RCD::CMEMSegment *sg = new RCD::CMEMSegment(sn, AltiModule::CNT_SIZE/*[type]*/*sizeof(uint32_t), true);
        if ((*sg)() != CMEM_RCC_SUCCESS) {
            CERR("error opening CMEM segment \"%s\" of size 0x%08x\n", sn.c_str(), sg->Size());
            return (-1);
        }
        std::printf("opened CMEM segment \"%s\" of size 0x%08x, phys 0x%08x, virt 0x%08x\n", sn.c_str(), sg->Size(), (unsigned int) sg->PhysicalAddress(), (unsigned int) sg->VirtualAddress());

        AltiCounter cnt(sg);

        if ((rtnv = alti->CNTCounterRead(&cnt)) != AltiModule::SUCCESS) {
            CERR("reading CNT counter", "");
            return (rtnv);
        }
        cnt.dump();

        bool rflg = (bool) enterInt("Print rate (0=no, 1=yes)", 0, 1);
        if (rflg) {
            AltiCounterRate *rcnt = new AltiCounterRate(cnt);
            rcnt->dump();
            delete rcnt;
        }

        // delete CMEM segment
        delete sg;
        std::printf("\nclosed CMEM segment \"%s\"\n", sn.c_str());

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTRequestPrint : public MenuItem {
  public:
    CNTRequestPrint() { setName("print CNT request"); }
    int action() {

        int rtnv(0);

        /*for(unsigned int typ=0; typ<CtpcorePlusModule::CNT_NUMBER; typ++) {
            std::printf("\nCNT \"%s\":\n",CtpcorePlusModule::CNT_NAME[typ].c_str());
            if((rtnv = ctpcoreplus->CNTRequestPrint(static_cast<CtpcorePlusModule::CNT_TYPE>(typ))) != CtpcorePlusModule::SUCCESS) {
                CERR("printing CNT \"%s\" request",CtpcorePlusModule::CNT_NAME[typ].c_str());
                 return(rtnv);
            }
        }*/
        std::printf("\nCNT:\n");
        if ((rtnv = alti->CNTRequestPrint()) != AltiModule::SUCCESS) {
            CERR("printing CNT request", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTTtypLutReadFile : public MenuItem {
  public:
    CNTTtypLutReadFile() { setName("read  CNT TTYP LUT [into file]"); }
    int action() {

        int rtnv(0);
        std::string fn = enterString("File name");

        if ((rtnv = alti->CNTTtypLutRead(fn)) != AltiModule::SUCCESS) {
            CERR("reading CNT TTYP LUT into file \"%s\"", fn.c_str());
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTTtypLutWriteFile : public MenuItem {
  public:
    CNTTtypLutWriteFile() { setName("write CNT TTYP LUT [from file]"); }
    int action() {

        int rtnv(0);
        std::string fn = enterString("File name");

        if ((rtnv = alti->CNTTtypLutWrite(fn)) != AltiModule::SUCCESS) {
            CERR("writing CNT TTYP LUT from file \"%s\"", fn.c_str());
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class CNTTtypLutPrint : public MenuItem {
  public:
    CNTTtypLutPrint() { setName("print CNT TTYP LUT"); }
    int action() {

        int rtnv(0);

        if ((rtnv = alti->CNTTtypLutPrint()) != AltiModule::SUCCESS) {
            CERR("printing CNT TTYP LUT", "");
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

class AltiMONVoltageRead : public MenuItem {
  public:
    AltiMONVoltageRead() { setName("read voltages"); }
    int action() {

        std::vector<float> voltages(8);
        if (alti->MONVoltageReadLTC2991(voltages) != 0) {
            CERR("reading voltages", "");
            return (-1);
        }

        float vcc_voltage;
        if (alti->MONVccVoltageReadLTC2991(vcc_voltage) != 0) {
            CERR("reading supply voltage", "");
            return (-1);
        }

        //table
        printf("\n");
        printf("+----------------------------------------------------+\n");
        printf("| Voltage | Nominal[V] | Measured[V] | Difference[%%] |\n");
        printf("+----------------------------------------------------+\n");
        unsigned int i;
        float nominal(0), measured(0), difference;
        for (i = 0; i < 9; i++) {
            switch (i) {
                case 0:
                    nominal = 1.0;
                    measured = voltages[0];
                    break;
                case 1:
                    nominal = 1.8;
                    measured = voltages[1];
                    break;
                case 2:
                    nominal = 2.5;
                    measured = voltages[2];
                    break;
                case 3:
                    nominal = 3.3;
                    measured = 2*voltages[3];
                    break;
                case 4:
                    nominal = 5.0;
                    measured = 2*voltages[4];
                    break;
                case 5:
                    nominal = 12.0;
                    measured = 4*voltages[5];
                    break;
                case 6:
                    nominal = -12.0;
                    measured = (34.7*voltages[6] - 30*2*voltages[4])/4.7;
                    break;
                case 7:
                    nominal = -5.0;
                    measured = (4*voltages[7] - 3*2*voltages[4])/1.0;
                    break;
                case 8:
                    nominal = 3.3;
                    measured = vcc_voltage + 2.5;
                    break;
                default:
                    break;
            }
            difference = (measured/nominal - 1)*100;
            char sign_nominal = (nominal >= 0) ? '+' : '-';
            char sign_measured = (measured >= 0) ? '+' : '-';
            char sign_difference = (difference >= 0) ? '+' : '-';
            if (i < 8) {
                printf("| V%d      | %c%09.6f | %c%09.6f  | %c%09.6f    |\n", i + 1, sign_nominal, fabs(nominal), sign_measured, fabs(measured), sign_difference, fabs(difference));
            }
            else {
                printf("| VCC     | %c%09.6f | %c%09.6f  | %c%09.6f    |\n", sign_nominal, fabs(nominal), sign_measured, fabs(measured), sign_difference, fabs(difference));
            }
            printf("+----------------------------------------------------+\n");
        }

        return(0);
    }
};

//------------------------------------------------------------------------------

class AltiMONTempRead : public MenuItem {
  public:
    AltiMONTempRead() { setName("read temperatures"); }
    int action() {

        float internalTemp, localTemp;
        if (alti->MONInternalTempReadLTC2991(internalTemp) != 0) {
            CERR("reading LTC2991 internal temperature", "");
            return (-1);
        }
        /*if (alti->MONLocalTempReadMAX1617(localTemp) != 0) {
            CERR("reading MAX1617 local temperature", "");
            return (-1);
        }*/
        if (alti->MONLocalTempRead(localTemp) != 0) {
            CERR("reading local temperature", "");
            return (-1);
        }

        //table
        printf("\n");
        printf("+---------------------------------------------------------+\n");
        printf("LTC2991     , internal: %.4f degrees Celsius\n", internalTemp);
        printf("MAX1617/6695, local   : %.4f degrees Celsius\n", localTemp);
        printf("+---------------------------------------------------------+\n");

        return(0);
    }
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

class AltiI2CSetup : public MenuItem {
  public:
    AltiI2CSetup() { setName("setup I2C"); }
    int action() {

        int rtnv(0);
        if((rtnv = alti->I2CSetup()) != AltiModule::SUCCESS) {
            CERR("setting up I2C","");
            return(rtnv);
        }
        return(rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiI2CDump : public MenuItem {
  public:
    AltiI2CDump() { setName("dump  I2C"); }
    int action() {

        int rtnv(0);
        rtnv = alti->I2CDump();

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiI2CRead : public MenuItem {
  public:
    AltiI2CRead() { setName("read  I2C"); }
    int action() {

        int rtnv(0);
        unsigned int bus, ndv, dev, qty, /*off, len,*/ wdat/*, i*/;
        I2CInterface::TRANSFER_TYPE ttyp;
        I2CNetwork::DISPLAY_TYPE dtyp;
        std::string snam, sdat;
        std::vector<unsigned int> vdat;
        float fdat;

        bus = 0;
        ndv = alti->I2CNumDevices(bus);
        dev = (ndv == 1) ? 0 : ((unsigned int) enterInt("I2C device  ", 0, ndv - 1));
        qty = (unsigned int) enterInt("I2C quantity", 0, alti->I2CNumQuantities(bus, dev) - 1);

        // read I2C transfer type
        if ((rtnv = alti->I2CTransferTypeRead(bus, dev, qty, ttyp)) != I2CNetwork::SUCCESS) {
            CERR("reading transfer type from I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
            return (rtnv);
        }

        // read I2C display type
        if ((rtnv = alti->I2CDisplayTypeRead(bus, dev, qty, dtyp)) != I2CNetwork::SUCCESS) {
            CERR("reading display type from I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
            return (rtnv);
        }

        // read I2C name
        if ((rtnv = alti->I2CNameRead(bus, dev, qty, snam)) != I2CNetwork::SUCCESS) {
            CERR("reading name from I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
            return (rtnv);
        }

        // check I2C readability
        bool rdb;
        alti->I2CReadableQuantity(bus, dev, qty, rdb);
        if (!rdb) {
            CERR("I2C bus/device/quantity %d/%d/%d is NOT READABLE", bus, dev, qty);
            rtnv = -1;
            return (rtnv);
        }

        // read I2C data
        /*if(dtyp == I2CNetwork::MEMORY) {

            unsigned int num = i2cnet->getNumberI2CQuantity(bus,dev,qty);
            if(i2cnet->getNameI2CQuantity(bus,dev,qty) == "MFR_FAULT_LOG") {
                off = 0;
                len = num;
            }
            else {
                off = (unsigned int)enterInt("MEM offset",0,num-1);
                len = (unsigned int)enterInt("MEM length",0,num-off);
            }

            // read I2C data as memory
            if((rtnv = i2cnet->read(bus,dev,qty,off,len,vdat)) != I2CNetwork::SUCCESS) {
                CERR("reading I2C data from bus %d, device %d, quantity %d - MEMORY",bus,dev,qty);
                return(rtnv);
            }

            if(i2cnet->getNameI2CQuantity(bus,dev,qty) == "MFR_FAULT_LOG") {
                if((i2cnet->getTypeI2CDevice(bus,dev) == "LTC3880") or (i2cnet->getTypeI2CDevice(bus,dev) == "LTM4676A")) {
                    i2cnet->ltc3880FaultLog(vdat,50,std::cout);
                }
                else {
                    i2cnet->ltm4676FaultLog(vdat,50,std::cout);
                }
            }
            else {
                bool flag(false);
                std::ostringstream buf;
                for(i=0; i<vdat.size(); i++) {
                    if(flag and (i%16 == 0)) { flag = false; buf.str(""); }
                    std::printf("%s0x%02x",flag?", ":"",vdat[i]);
                    if(vdat[i] < 32) buf << "?"; else if(vdat[i] >= 127) buf << "."; else buf << (char)vdat[i];
                    flag = true;
                    if(i%16 == 15) std::cout << "    " << buf.str() << std::endl;
                }
                if(i%16 != 0) std::cout << "    " << buf.str() << std::endl;
            }
        }*/
        //else if(dtyp == I2CNetwork::WORD) {

        if (dtyp == I2CNetwork::WORD) {

            /*// read I2C data as word using PROC
            if(ttyp == I2CInterface::PROC) {
                if(i2cnet->getNameI2CQuantity(bus,dev,qty) == "SMBALERT_MASK") {
                    std::cout << "SMBALERT_MASK: 0x7a01=VOUT, 0x7b01=IOUT, 0x7c01=INPUT, 0x7d01=TEMP, 0x7e01=CML, 0x8001=MFR" << std::endl;
                }
                wdat = (unsigned int)enterNumber("DATA[PROCESS CALL]",0,0xffff);
            }*/

            // read I2C data as word
            if ((rtnv = alti->I2CDataRead(bus, dev, qty, wdat)) != I2CNetwork::SUCCESS) {
                CERR("reading data from I2C bus/device/quantity %d/%d/%d - WORD", bus, dev, qty);
                return (rtnv);
            }
            unsigned int num = alti->I2CQuantityGetNumber(bus, dev, qty);
            std::printf("\"%s\" = 0x%0*x\n", snam.c_str(), num*2, wdat);
        }
        else if (dtyp == I2CNetwork::FLOAT) {

            // read I2C data as float
            if ((rtnv = alti->I2CDataRead(bus, dev, qty, fdat)) != I2CNetwork::SUCCESS) {
                CERR("reading data from I2C bus/device/quantity %d/%d/%d - FLOAT", bus, dev, qty);
                return (rtnv);
            }
            std::printf("\"%s\" = %f\n", snam.c_str(), fdat);
        }
        else if (dtyp == I2CNetwork::STRING) {

            // read I2C data as string
            if ((rtnv = alti->I2CDataRead(bus, dev, qty, sdat)) != I2CNetwork::SUCCESS) {
                CERR("reading data from I2C bus/device/quantity %d/%d/%d - STRING", bus, dev, qty);
                return (rtnv);
            }
            std::printf("\"%s\" = \"%s\"\n", snam.c_str(), sdat.c_str());
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiI2CReadDevice : public MenuItem {
  public:
    AltiI2CReadDevice() { setName("read  I2C [device]"); }
    int action() {

        int rtnv (0);
        unsigned int bus, dev, qty;
        std::string snam;

        bus = 0;
        alti->I2CNameRead(bus, snam);
        std::cout << "----------------------------------------" << std::endl;
        std::printf("I2C bus %d: \"%s\":\n", bus, snam.c_str());
        for (dev = 0; dev < alti->I2CNumDevices(bus); dev++) {
            alti->I2CNameRead(bus, dev, snam);
            std::printf("I2C bus/device %d/%2d: \"%s\"\n", bus, dev, snam.c_str());
        }
        std::cout << "----------------------------------------" << std::endl;

        dev = (unsigned int) enterInt("I2C device", 0, alti->I2CNumDevices(bus) - 1);

        for (qty = 0; qty < alti->I2CNumQuantities(bus, dev); qty++) {
            if ((rtnv = alti->I2CDataPrint(bus, dev, qty)) != I2CNetwork::SUCCESS) {
                CERR("printing I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
                continue;
            }
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiI2CReadAll : public MenuItem {
  public:
    AltiI2CReadAll() { setName("read  I2C [all]"); }
    int action() {

        int rtnv(0);
        unsigned int bus, dev, qty;

        bus = 0;
        std::cout << "================================================================================" << std::endl;
        for (dev = 0; dev < alti->I2CNumDevices(bus); dev++) {
            std::cout << "--------------------------------------------------------------------------------" << std::endl;
            for (qty = 0; qty < alti->I2CNumQuantities(bus, dev); qty++) {
                if ((rtnv = alti->I2CDataPrint(bus, dev, qty)) != I2CNetwork::SUCCESS) {
                    CERR("printing I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
                    continue;
                }
            }
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------

class AltiI2CWrite : public MenuItem {
  public:
    AltiI2CWrite() { setName("write I2C"); }
    int action() {

        int rtnv (0);
        unsigned int bus, ndv, dev, qty, wdat(0);
        I2CInterface::TRANSFER_TYPE ttyp;
        I2CNetwork::DISPLAY_TYPE dtyp;
        std::string snam, sdat;
        std::ostringstream buf;

        bus = 0;
        ndv = alti->I2CNumDevices(bus);
        dev = (ndv == 1) ? 0 : ((unsigned int) enterInt("I2C device  ", 0, ndv - 1));
        qty = (unsigned int) enterInt("I2C quantity", 0, alti->I2CNumQuantities(bus, dev) - 1);

        // read I2C name
        if ((rtnv = alti->I2CNameRead(bus, dev, qty, snam)) != I2CNetwork::SUCCESS) {
            CERR("reading name from I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
            return (rtnv);
        }

        // read I2C transfer type
        if ((rtnv = alti->I2CTransferTypeRead(bus, dev, qty, ttyp)) != I2CNetwork::SUCCESS) {
            CERR("reading transfer type from I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
            return (rtnv);
        }

        // read I2C display type
        if ((rtnv = alti->I2CDisplayTypeRead(bus, dev, qty, dtyp)) != I2CNetwork::SUCCESS) {
            CERR("reading display type from I2C bus/device/quantity %d/%d/%d", bus, dev, qty);
            return (rtnv);
        }

        // check I2C writability
        bool wtb;
        alti->I2CWritableQuantity(bus, dev, qty, wtb);
        if (!wtb) {
            CERR("I2C bus/device/quantity %d/%d/%d is NOT WRITABLE", bus, dev, qty);
            rtnv = -1;
            return (rtnv);
        }

        // check transfer type
        if ((ttyp != I2CInterface::BYTE) && (ttyp != I2CInterface::DATA) && (ttyp != I2CInterface::WORD)) {
            CERR("writing to I2C bus/device/quantity %d/%d/%d of transfer type \"%s\" NOT IMPLEMENTED", bus, dev, qty, I2CInterface::TRANSFER_TYPE_NAME[ttyp].c_str());
            rtnv = -1;
            return (rtnv);
        }

        // write I2C as word
        if (dtyp == I2CNetwork::WORD) {
            if (alti->I2CQuantityGetNumber(bus, dev, qty)) {
                /*if(i2cnet->getNameI2CQuantity(bus,dev,qty) == "SMBALERT_MASK") {
                    std::cout << "SMBALERT_MASK (0xmm): 0xmm7a=VOUT, 0xmm7b=IOUT, 0xmm7c=INPUT, 0xmm7d=TEMP, 0xmm7e=CML, 0xmm80=MFR" << std::endl;
                }*/
                wdat = (unsigned int) enterNumber("DATA", 0, 0xffff);
            }
            if ((rtnv = alti->I2CDataWrite(bus, dev, qty, wdat)) != I2CNetwork::SUCCESS) {
                CERR("writing data from I2C bus/device/quantity %d/%d/%d - WORD", bus, dev, qty);
                return (rtnv);
            }
        }

        // write I2C as float
        else if (dtyp == I2CNetwork::FLOAT) {
            float fdat(0.0);
            if (alti->I2CQuantityGetNumber(bus, dev, qty)) {
                char buf[20];
                while (true) {
                    std::printf("DATA[FLOAT]:");
                    buf[0] = '\0';
                    if (!std::cin) std::cin.clear();
                    std::cin >> buf;
                    if (buf[0] == '\0') {
                        continue;
                    }
                    else {
                        fdat = strtof(buf, nullptr);
                        break;
                    }
                }
                cin.getline(buf, 1);
            }
            if ((rtnv = alti->I2CDataWrite(bus, dev, qty, fdat)) != I2CNetwork::SUCCESS) {
                CERR("writing data to I2C bus/device/quantity %d/%d/%d - FLOAT", bus, dev, qty);
                return (rtnv);
            }
        }

        // write I2C as string
        else if (dtyp == I2CNetwork::STRING) {
            std::string sdat;
            if (alti->I2CQuantityGetNumber(bus, dev, qty)) {
                sdat = enterString("DATA[STRING]:");
            }
            if ((rtnv = alti->I2CDataWrite(bus, dev, qty, sdat)) != I2CNetwork::SUCCESS) {
                CERR("writing data to I2C bus/device/quantity %d/%d/%d - STRING", bus, dev, qty);
                return (rtnv);
            }
        }

        else {
            CERR("writing to I2C bus/device/quantity %d/%d/%d of display type \"%s\" NOT IMPLEMENTED", bus, dev, qty, I2CNetwork::DISPLAY_TYPE_NAME[dtyp].c_str());
            rtnv = -1;
            return (rtnv);
        }

        return (rtnv);
    }
};

//------------------------------------------------------------------------------
// From here BSY
//------------------------------------------------------------------------------
class BSYFifoReset : public MenuItem {
  public:
    BSYFifoReset() { setName("reset BSY counter FIFO"); }
    int action() {

        int rtnv(0);

        if((rtnv = alti->BSYFifoReset()) != 0) {
            CERR("resetting BSY FIFO","");
            return(rtnv);
        }

        return(rtnv);
    }
};

//------------------------------------------------------------------------------

class BSYMonitoringControlRead : public MenuItem {
  public:
    BSYMonitoringControlRead() { setName("read  BSY counter control"); }
    int action() {

        int rtnv(0);
        bool wmod, rmod;

        if((rtnv = alti->BSYMonitoringControlRead(wmod,rmod)) != AltiModule::SUCCESS) {
            CERR("reading BSY control","");
            return(rtnv);
        }
        std::printf("BSY control: WRITE = %s, READ = %s\n",wmod?"AUTOMATIC":"   MANUAL",rmod?" LAST":"FIRST");

        return(rtnv);
    }
};

//------------------------------------------------------------------------------

class BSYMonitoringControlWrite : public MenuItem {
  public:
    BSYMonitoringControlWrite() { setName("write BSY counter control"); }
    int action() {

        int rtnv(0);
        bool wmod = (bool)enterInt("Write mode (0=man, 1=auto)",0,1);
        bool rmod = (bool)enterInt("Read mode (0=first, 1=last)",0,1);

        if((rtnv = alti->BSYMonitoringControlWrite(wmod,rmod)) != AltiModule::SUCCESS) {
            CERR("writing BSY control","");
            return(rtnv);
        }

        return(rtnv);
    }
};

//------------------------------------------------------------------------------

class BSYSelectRead : public MenuItem {
  public:
    BSYSelectRead() { setName("read  BSY counter select"); }
    int action() {

        int rtnv(0);
        AltiBusyCounter sel;

        if((rtnv = alti->BSYSelectRead(sel)) != AltiModule::SUCCESS) {
            CERR("reading BSY select","");
            return(rtnv);
        }
        sel.selectDump();

        return(rtnv);
    }
};

//------------------------------------------------------------------------------

class BSYSelectWrite : public MenuItem {
  public:
    BSYSelectWrite() { setName("write BSY counter select"); }
    int action() {

        int rtnv(0);
        AltiBusyCounter sel;
        std::string nam;
        std::ostringstream buf;

        bool all = (bool)enterInt("All (0=no, 1=yes)",0,1);
        if(all) {
            sel.selectAll(true);
        }
        else {
            for(unsigned int b=0; b<AltiBusyCounter::BUSY_NUMBER; b++) {
                nam = "BSY[\"" + AltiBusyCounter::BUSY_NAME[b] + "\"]";
                buf.str(""); buf << "Enable " << std::setw(16) << std::left << nam << " (0=no, 1=yes)";
                sel.select(static_cast<AltiBusyCounter::BUSY>(b),(bool)enterInt(buf.str().c_str(),0,1));
            }
        }

        if((rtnv = alti->BSYSelectWrite(sel)) != AltiModule::SUCCESS) {
            CERR("writing BSY select","");
            return(rtnv);
        }

        return(rtnv);
    }
};

//------------------------------------------------------------------------------

class BSYCounterReset : public MenuItem {
  public:
    BSYCounterReset() { setName("reset BSY counter"); }
    int action() {

        int rtnv(0);
        bool all = (bool)enterInt("All (0=no, 1=yes)",0,1);

        if(all) {
            if((rtnv = alti->BSYCounterReset()) != AltiModule::SUCCESS) {
                CERR("writing BSY clear counter","");
                return(rtnv);
            }
        }
        else {
            std::string nam;
            std::ostringstream buf;
            AltiBusyCounter clr;
            for(unsigned int b=0; b<AltiBusyCounter::BUSY_NUMBER; b++) {
                nam = "BSY[\"" + AltiBusyCounter::BUSY_NAME[b] + "\"]";
                buf.str(""); buf << "Reset " << std::setw(16) << std::left << nam << " (0=no, 1=yes)";
                clr.select(static_cast<AltiBusyCounter::BUSY>(b),(bool)enterInt(buf.str().c_str(),0,1));
            }
            if((rtnv = alti->BSYCounterReset(clr)) != AltiModule::SUCCESS) {
                CERR("writing BSY clear counter","");
                return(rtnv);
            }
        }

        return(rtnv);
    }
};

//------------------------------------------------------------------------------

class BSYIntervalRead : public MenuItem {
  public:
    BSYIntervalRead() { setName("read  BSY counter interval"); }
    int action() {

        int rtnv(0);
        unsigned int itvl;

        if((rtnv = alti->BSYIntervalRead(itvl)) != AltiModule::SUCCESS) {
            CERR("reading BSY interval","");
            return(rtnv);
        }
        std::printf("BSY interval = %d, (0x%06x)",itvl,itvl);

        return(rtnv);
    }
};

//------------------------------------------------------------------------------

class BSYIntervalWrite : public MenuItem {
  public:
    BSYIntervalWrite() { setName("write BSY counter interval"); }
    int action() {

        int rtnv(0);
        unsigned int itvl = enterHex("BSY interval",0,AltiBusyCounter::BSY_CNT_MASK);

        if((rtnv = alti->BSYIntervalWrite(itvl)) != AltiModule::SUCCESS) {
            CERR("writing BSY interval","");
            return(rtnv);
        }

        return(rtnv);
    }
};

//------------------------------------------------------------------------------

class BSYFifoWatermarkRead : public MenuItem {
  public:
    BSYFifoWatermarkRead() { setName("read  BSY counter FIFO watermark"); }
    int action() {

        int rtnv(0);
        unsigned int fmrk;

        if((rtnv = alti->BSYFifoWatermarkRead(fmrk)) != AltiModule::SUCCESS) {
            CERR("reading BSY FIFO watermark","");
            return(rtnv);
        }
        std::printf("BSY FIFO watermark = %4d (0x%03x)\n",fmrk,fmrk);

        return(rtnv);
    }
};

//------------------------------------------------------------------------------

class BSYFifoWatermarkWrite : public MenuItem {
  public:
    BSYFifoWatermarkWrite() { setName("write BSY counter FIFO watermark"); }
    int action() {

        int rtnv(0);
        unsigned int fmrk = (unsigned int)enterInt("BSY FIFO watermark",0,AltiModule::BSY_FIFO_MASK);

        if((rtnv = alti->BSYFifoWatermarkWrite(fmrk)) != AltiModule::SUCCESS) {
            CERR("writing BSY FIFO watermark","");
            return(rtnv);
        }

        return(rtnv);
    }
};

//------------------------------------------------------------------------------

class BSYFifoWrite : public MenuItem {
  public:
    BSYFifoWrite() { setName("write BSY counter FIFO [manually]"); }
    int action() {

        int rtnv(0);

        if((rtnv = alti->BSYFifoWrite()) != AltiModule::SUCCESS) {
            CERR("writing BSY FIFO","");
            return(rtnv);
        }

        return(rtnv);
    }
};

//------------------------------------------------------------------------------

class BSYFifoStatus : public MenuItem {
  public:
    BSYFifoStatus() { setName("read  BSY counter FIFO status"); }
    int action() {

        int rtnv(0);
        unsigned int lvl;
        bool ful;

        if((rtnv = alti->BSYFifoStatusRead(lvl,ful)) != AltiModule::SUCCESS) {
            CERR("reading BSY FIFO status","");
            return(rtnv);
        }
        std::printf("BSY FIFO status: level = %4d  - %s\n",lvl,ful?"FULL":"not full");

        return(rtnv);
    }
};

//------------------------------------------------------------------------------

class BSYFifoRead : public MenuItem {
  public:
    BSYFifoRead() { setName("read  BSY counter FIFO"); }
    int action() {

        int rtnv(0);
        unsigned int data, i;
        unsigned int loop = enterInt("How many words",0,AltiModule::BSY_FIFO_SIZE);

        for(i=0; i<loop; i++) {
            if((rtnv = alti->BSYFifoRead(data)) != AltiModule::SUCCESS) {
                CERR("reading BSY FIFO at word %d",i);
                return(rtnv);
            }
            std::printf("BSY FIFO word %4d: 0x%08x\n",i,data);
        }

        return(rtnv);
    }
};

//------------------------------------------------------------------------------

class BSYFifoReadCounter : public MenuItem {
  public:
    BSYFifoReadCounter() { setName("read  BSY counter FIFO [counter]"); }
    int action() {

        int rtnv(0);

        // create CMEM segment
        std::string sn = "menuAltiModule[BSY]";
        RCD::CMEMSegment* sg = new RCD::CMEMSegment(sn,AltiModule::BSY_FIFO_SIZE*sizeof(uint32_t),true);
        if((*sg)() != CMEM_RCC_SUCCESS) {
            CERR("error opening CMEM segment \"%s\" of size 0x%08x\n",sn.c_str(),sg->Size());
            return(-1);
        }
        std::printf("opened CMEM segment \"%s\" of size 0x%08x, phys 0x%08x, virt 0x%08x\n",sn.c_str(),sg->Size(),(unsigned int)sg->PhysicalAddress(),(unsigned int)sg->VirtualAddress());
        AltiBusyCounter BSY(sg);
	if((rtnv = alti->BSYStatusRead(BSY)) != AltiModule::SUCCESS) {
            CERR("reading BSY status","");
            return(rtnv);
        }        

        // read counter config
        if((rtnv = alti->BSYCounterConfigRead(BSY)) != AltiModule::SUCCESS) {
            CERR("reading BSY counter configuration","");
            return(rtnv);
        }

        if(!BSY.check()) {
            CERR("BSY counter/control config inconsistent","");
            return(-1);
        }

        // read counter
        if((rtnv = alti->BSYFifoRead(BSY)) != AltiModule::SUCCESS) {
            CERR("reading BSY FIFO counter","");
            return(rtnv);
	}

        // calculate rate
        AltiBusyCounterRate rBSY;
        rBSY = BSY;
        std::cout << std::endl; rBSY.dump();

        // delete CMEM segment
        delete sg;
        std::printf("\nclosed CMEM segment \"%s\"\n",sn.c_str());

        return(rtnv);
    }
};

//------------------------------------------------------------------------------

class BSYFifoReadAverage : public MenuItem {
  public:
    BSYFifoReadAverage() { setName("read  BSY counter FIFO [average]"); }
    int action() {

        int rtnv(0);
        int num = enterInt("Maximum number of reads",-1,AltiModule::BSY_FIFO_SIZE);

        // create CMEM segment
        std::string sn = "menuAltiModule[BSY]";
        RCD::CMEMSegment* sg = new RCD::CMEMSegment(sn,AltiModule::BSY_FIFO_SIZE*sizeof(uint32_t),true);
        if((*sg)() != CMEM_RCC_SUCCESS) {
            CERR("error opening CMEM segment \"%s\" of size 0x%08x\n",sn.c_str(),sg->Size());
            return(-1);
        }
        std::printf("opened CMEM segment \"%s\" of size 0x%08x, phys 0x%08x, virt 0x%08x\n",sn.c_str(),sg->Size(),(unsigned int)sg->PhysicalAddress(),(unsigned int)sg->VirtualAddress());
        AltiBusyCounter BSY(sg);
        AltiBusyCounterRate avg;

        // read counter configuration
        if((rtnv = alti->BSYCounterConfigRead(BSY)) != AltiModule::SUCCESS) {
            CERR("reading BSY FIFO counter configuration","");
            return(rtnv);
        }
        if(!BSY.check()) {
            CERR("counter control inconsistent","");
            return(-1);
        }

        // read average counter
        if((rtnv = alti->BSYFifoRead(avg,BSY,num)) != AltiModule::SUCCESS) {
            CERR("reading BSY FIFO counter","");
            return(rtnv);
        }
        avg.dump();

        // delete CMEM segment
        delete sg;
        std::printf("\nclosed CMEM segment \"%s\"\n",sn.c_str());

        return(rtnv);
    }
};

//------------------------------------------------------------------------------

class BSYStatus : public MenuItem {
  public:
    BSYStatus() { setName("read  BSY status"); }
    int action() {

        int rtnv(0);
	printf("\n");
	std::cout << "Busy input selection:" << endl;
	std::vector<bool> selected;
        selected.resize(AltiModule::BUSY_INPUT_NUMBER);
        if ((rtnv = alti->BSYInputSelectRead(selected)) != AltiModule::SUCCESS) {
            CERR("reading busy input selection", "");
            return (rtnv);
        }
        unsigned int i;
        int total = 0;
        AltiModule::BUSY_INPUT input;
        for (i = 0; i < AltiModule::BUSY_INPUT_NUMBER; i++) {
            input = AltiModule::BUSY_INPUT(i);
            printf("[%02d]%-21s: %s\n", i, AltiModule::BUSY_INPUT_NAME[input].c_str(), selected[i] ? "Enabled" : "Disabled");
            total += selected[i];
        }
        printf("\n");
        printf("Local busy: ");
        if (total == 0) {
            printf("Inactive");
        }
        else {
            for (i = 0; i < AltiModule::BUSY_INPUT_NUMBER; i++) {
                input = AltiModule::BUSY_INPUT(i);
                printf("%s", selected[i] ? AltiModule::BUSY_INPUT_NAME[input].c_str() : "");
                total -= selected[i];
                printf("%s", (selected[i] && (total > 0)) ? " + " : "");
            }
        }
        printf("\n \n");

	std::cout << "Busy output selection:" << endl;
        std::vector<AltiModule::BUSY_SOURCE> src;
        src.resize(AltiModule::BUSY_OUTPUT_NUMBER);
        if ((rtnv = alti->BSYOutputSelectRead(src)) != AltiModule::SUCCESS) {
            CERR("reading busy output selection", "");
            return (rtnv);
        }
        //unsigned int i;
        AltiModule::BUSY_OUTPUT output;
        for (i = 0; i < AltiModule::BUSY_OUTPUT_NUMBER; i++) {
            output = AltiModule::BUSY_OUTPUT(i);
            printf(" [%02d]%-12s: %s\n", i, AltiModule::BUSY_OUTPUT_NAME[output].c_str(), AltiModule::BUSY_SOURCE_NAME[src[i]].c_str());
        }
        printf("\n");

	AltiModule::BUSY_LEVEL level;
	if ((rtnv = alti->BSYFrontPanelLevelRead(level)) != AltiModule::SUCCESS) {
	  CERR("reading busy Front-Panel level", "");
	  return (rtnv);
        }
        printf("Level of Front-Panel: %s\n", AltiModule::BUSY_LEVEL_NAME[level].c_str());
        printf("\n");
	
	std::cout << "Busy gating of PG signals:" << endl;
	bool l1a_en,orb_en,tty_en;
        bool ttr1,ttr2,ttr3;
        bool bgo0,bgo1,bgo2,bgo3;
        if ((rtnv = alti->BSYMaskingL1aEnableRead(l1a_en)) != AltiModule::SUCCESS) {
            CERR("reading busy masking of L1A", "");
            return (rtnv);
        }
        if ((rtnv = alti->BSYMaskingOrbEnableRead(orb_en)) != AltiModule::SUCCESS) {
            CERR("reading busy masking of Orbit", "");
            return (rtnv);
        }
        if ((rtnv = alti->BSYMaskingTTypeEnableRead(tty_en)) != AltiModule::SUCCESS) {
            CERR("reading busy masking of TType", "");
            return (rtnv);
        } 
        if ((rtnv = alti->BSYMaskingTTREnableRead(ttr1,ttr2,ttr3)) != AltiModule::SUCCESS) {
          CERR("reading busy masking of TTriggers", "");
          return (rtnv);
        }  
        if ((rtnv = alti->BSYMaskingBGoEnableRead(bgo0,bgo1,bgo2,bgo3)) != AltiModule::SUCCESS) {
          CERR("reading busy masking of BGos", "");
          return (rtnv);
        }
        printf("  L1A   : %s\n  Orbit : %s\n  TType : %s\n  TTR1  : %s\n  TTR2  : %s\n  TTR3  : %s\n  BGo0  : %s\n  BGo1  : %s\n  BGo2  : %s\n  BGo3  : %s",
        l1a_en ? "Enabled" : "Disabled",orb_en ? "Enabled" : "Disabled",tty_en ? "Enabled" : "Disabled",ttr1 ? "Enabled" : "Disabled",ttr2 ? "Enabled" : "Disabled",ttr3 ? "Enabled" : "Disabled",bgo0 ? "Enabled" : "Disabled",bgo1 ? "Enabled" : "Disabled",bgo2 ? "Enabled" : "Disabled",bgo3 ? "Enabled" : "Disabled");
        
        printf("\n\n");

	bool busy;
        if ((rtnv = alti->BSYConstLevelRegRead(busy)) != AltiModule::SUCCESS) {
	  CERR("reading constant level busy register", "");
	  return (rtnv);
        }	
        printf("Constant level busy register: %s\n", busy ? "Active(1)" : "Inactive(0)");

        AltiBusyCounter cnt;	
        // read primary partition BUSY status
        if((rtnv = alti->BSYStatusRead(cnt)) != AltiModule::SUCCESS) {
	  CERR("reading BSY status","");
            return(rtnv);
        }        
        printf("\n");
        
        cnt.dump();

        return(rtnv);
    }
};

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

int main(int argc,char** argv) {

    Menu menu("AltiModule main menu");

    // menu for AltiModule
    menu.init(new AltiModuleOpen);
    menu.exit(new AltiModuleClose);

    // menu for CSR
    //Menu menuCSR("CSR menu");
    Menu menuCSR("[\"CSR\" menu] VME64x CR/CSR space");
    menuCSR.add(new AltiCSRManufacturerIdRead);
    menuCSR.add(new AltiCSRBoardIdRead);
    menuCSR.add(new AltiCSRRevisionIdRead);
    menuCSR.add(new AltiCSRSlaveCharacteristicsRead);
    menuCSR.add(new AltiCSRBARRead);
    menuCSR.add(new AltiCSRADER0Read);
    menuCSR.add(new AltiCSRADER0Write);
    menuCSR.add(new AltiCSRADER1Read);
    menuCSR.add(new AltiCSRADER1Write);
    menuCSR.add(new AltiSerialNumber);
    menu.add(&menuCSR);

    // menu for CFG
    //Menu menuCFG("CFG menu");
    Menu menuCFG("[\"CFG\" menu] Global configuration");
    //menuCFG.add(new AltiFpgaUnconfig);
    menuCFG.add(new AltiReset);
    menuCFG.add(new AltiSetup);
    menuCFG.add(new AltiCheck);

    menuCFG.add(new AltiConfigRead);
    menuCFG.add(new AltiConfigReadFile);
    menuCFG.add(new AltiConfigWriteFile);
    menuCFG.add(new AltiConfigWrite);
    menuCFG.add(new AltiConfigWritePredefined);
    menu.add(&menuCFG);

    // menu for CLK
    //Menu menuCLK("CLK menu");
    Menu menuCLK("[\"CLK\" menu] Clocking");
    menuCLK.add(new AltiCLKStatusRead);
    menuCLK.add(new AltiCLKSelectWrite);
    menuCLK.add(new AltiCLKSelectToggle);
    menuCLK.add(new AltiCLKPhaseShift);
    menuCLK.add(new AltiCLKPhaseReset);
    menuCLK.add(new AltiCLKReset);
    //menuCLK.add(new AltiCLKJitterCleanerStatusRead);
    menuCLK.add(new AltiCLKJitterCleanerInSelWrite);
    menuCLK.add(new AltiCLKJitterCleanerConfigRead);
    menuCLK.add(new AltiCLKJitterCleanerConfigWrite);
    menuCLK.add(new AltiCLKJitterCleanerConfigWriteDefault);
    menuCLK.add(new AltiCLKJitterCleanerReset);
    menuCLK.add(new AltiCLKJitterCleanerResync);
    //menuCLK.add(new AltiCLKJitterCleanerStatusClear);
    menuCLK.add(new AltiCLKStatusClear);
    menu.add(&menuCLK);

    // menu for SWX + EQZ : SIG
    //Menu menuSIG("SIG menu");
    Menu menuSIG("[\"SIG\" menu] Signals I/O");
    menuSIG.add(new AltiSIGSwitchConfigRead);
    menuSIG.add(new AltiSIGSwitchConfigWrite);
    menuSIG.add(new AltiSIGSwitchMasterConfigurationWrite);
    menuSIG.add(new AltiSIGSwitchCtpSlaveConfigurationWrite);
    menuSIG.add(new AltiSIGSwitchAltiSlaveConfigurationWrite);
    menuSIG.add(new AltiSIGSwitchLemoSlaveConfigurationWrite);
    menuSIG.add(new AltiSIGEqualizersConfigRead);
    menuSIG.add(new AltiSIGEqualizersPredefinedModeWrite);
    menuSIG.add(new AltiSIGEqualizersConfigWrite);
    menuSIG.add(new AltiSIGInputSyncHistogramsRead);
    menuSIG.add(new AltiSIGInputSyncHistogramsReset);
    menuSIG.add(new AltiSIGInputSyncStatusRead);
    menuSIG.add(new AltiSIGInputSyncEnableWrite);
    menuSIG.add(new AltiSIGInputShapeStatusRead);
    menuSIG.add(new AltiSIGInputShapeEnableWrite);
    menuSIG.add(new AltiSIGMuxBGOOutputLEMORead);
    menuSIG.add(new AltiSIGMuxBGOOutputLEMOWrite);
    menuSIG.add(new AltiSIGMuxTTRInputLEMORead);
    menuSIG.add(new AltiSIGMuxTTRInputLEMOWrite);
    menuSIG.add(new AltiSIGMuxTTROutputLEMORead);
    menuSIG.add(new AltiSIGMuxTTROutputLEMOWrite);
    menu.add(&menuSIG);

    Menu menuBSY("[\"BSY\" menu] Busy ");
    menuBSY.add(new BSYStatus);
    menuBSY.add(new BSYMonitoringControlRead);
    menuBSY.add(new BSYMonitoringControlWrite);
    menuBSY.add(new BSYSelectRead);
    menuBSY.add(new BSYSelectWrite);
    menuBSY.add(new BSYCounterReset);
    menuBSY.add(new BSYIntervalRead);
    menuBSY.add(new BSYIntervalWrite);
    menuBSY.add(new BSYFifoWatermarkRead);
    menuBSY.add(new BSYFifoWatermarkWrite);
    menuBSY.add(new BSYFifoWrite);
    menuBSY.add(new BSYFifoStatus);
    menuBSY.add(new BSYFifoRead);
    menuBSY.add(new BSYFifoReadCounter);
    menuBSY.add(new BSYFifoReadAverage);
    menuBSY.add(new BSYFifoReset);

    menuBSY.add(new AltiBSYFrontPanelLevelWrite);
    menuBSY.add(new AltiBSYInputSelectWrite);
    menuBSY.add(new AltiBSYOutputSelectWrite);
    menuBSY.add(new AltiBSYConstLevelRegWrite);
    menuBSY.add(new AltiBSYMasking);
    menuBSY.add(new AltiBSYMaskingRead);
    menu.add(&menuBSY);

    // menu for CRQ
    //Menu menuCRQ("CRQ menu");
    Menu menuCRQ("[\"CRQ\" menu] Calibration requests");
    menuCRQ.add(new AltiCRQInputRead);
    menuCRQ.add(new AltiCRQInputSelectRead);
    menuCRQ.add(new AltiCRQInputSelectWrite);
    menuCRQ.add(new AltiCRQOutputSelectRead);
    menuCRQ.add(new AltiCRQOutputSelectWrite);
    menuCRQ.add(new AltiCRQConstLevelRegRead);
    menuCRQ.add(new AltiCRQConstLevelRegWrite);
    menu.add(&menuCRQ);

    // menu for PAT
    //Menu menuPAT("PAT menu");
    Menu menuPAT("[\"PAT\" menu] Pattern generation memory");
    menuPAT.add(new AltiPATStatusRead);
    menuPAT.add(new AltiPATGenerationConfigWrite);
    menuPAT.add(new AltiPATGenerationEnable);
    menuPAT.add(new AltiPATGenerationDisable);
    menuPAT.add(new AltiPATGenerationRepeatWrite);
    menuPAT.add(new AltiPATGenerationStartAddressWrite);
    menuPAT.add(new AltiPATGenerationStopAddressWrite);
    menuPAT.add(new AltiPATRead);
    menuPAT.add(new AltiPATWriteFile);
    Menu menuPATlowlevelMEM("MEM       menu");
    menuPATlowlevelMEM.add(new AltiMEMDataPATSNPReadNumber);
    menuPATlowlevelMEM.add(new AltiMEMDataPATSNPWriteNumber);
    menuPATlowlevelMEM.add(new AltiMEMDataPATSNPReadVector);
    menuPATlowlevelMEM.add(new AltiMEMDataPATSNPWriteVector);
    menuPATlowlevelMEM.add(new AltiMEMDataPATSNPReadBlock);
    menuPATlowlevelMEM.add(new AltiMEMDataPATSNPWriteBlock);
    menuPAT.add(&menuPATlowlevelMEM);
    menu.add(&menuPAT);

    // menu for SNP
    //Menu menuSNP("SNP menu");
    Menu menuSNP("[\"SNP\" menu] Snapshot memory");
    menuSNP.add(new AltiSNPStatusRead);
    menuSNP.add(new AltiSNPEnable);
    menuSNP.add(new AltiSNPDisable);
    menuSNP.add(new AltiSNPMaskWrite);
    menuSNP.add(new AltiSNPRead);
    menuSNP.add(new AltiSNPReadFile);
    Menu menuSNPlowlevelMEM("MEM     menu");
    menuSNP.add(&menuSNPlowlevelMEM);
    menuSNPlowlevelMEM.add(new AltiMEMDataPATSNPReadNumber);
    menuSNPlowlevelMEM.add(new AltiMEMDataPATSNPWriteNumber);
    menuSNPlowlevelMEM.add(new AltiMEMDataPATSNPReadVector);
    menuSNPlowlevelMEM.add(new AltiMEMDataPATSNPWriteVector);
    menuSNPlowlevelMEM.add(new AltiMEMDataPATSNPReadBlock);
    menuSNPlowlevelMEM.add(new AltiMEMDataPATSNPWriteBlock);
    menu.add(&menuSNP);

    // menu for TTC encoder
    //Menu menuENC("ENC menu");
    Menu menuENC("[\"ENC\" menu] TTC encoder");
    menuENC.add(new AltiENCTransmitterStatusRead);
    menuENC.add(new AltiENCTransmitterEnable);
    menuENC.add(new AltiENCTransmitterFaultClear);
    menuENC.add(new AltiENCTTRSelWrite);
    menuENC.add(new AltiENCTTRSelRead);
    menuENC.add(new AltiENCInhibitParamsRead);
    menuENC.add(new AltiENCInhibitParamsWrite);
    menuENC.add(new AltiENCTriggerWordDelayRead);
    menuENC.add(new AltiENCTriggerWordDelayWrite);
    //
    //menuENC.add(new AltiENCTriggerWordCycleControlRead);
    //menuENC.add(new AltiENCTriggerWordCycleControlWrite);
    menuENC.add(new AltiENCTriggerWordCycleEnable);
    menuENC.add(new AltiENCTriggerWordCycleDisable);
    //
    menuENC.add(new AltiENCBgoGenerate);
    //menuENC.add(new AltiTTCShotTTYPCommand);
    menuENC.add(new AltiENCBgoCommandModeRead);
    menuENC.add(new AltiENCBgoCommandModeWrite);
    menuENC.add(new AltiENCBgoCommandModeHighLevelWrite);
    menuENC.add(new AltiENCFifoStatusRead);
    menuENC.add(new AltiENCFifoReset);
    menuENC.add(new AltiENCBgoFifoRetransmitSelect);
    //menuENC.add(new AltiTTCFIFONumberRead);
    //menuENC.add(new AltiTTCFIFONumberWrite);
    menuENC.add(new AltiENCBgoCommandPut);
    menuENC.add(new AltiENCAsyncCommandPut);
    menuENC.add(new AltiENCAsyncCommandPending);
    menu.add(&menuENC);

    // menu for TTC decoder
    //Menu menuDEC("DEC menu");
    Menu menuDEC("[\"DEC\" menu] TTC decoder");
    menuDEC.add(new AltiDECStatusRead);
    menuDEC.add(new AltiDECStatusClear);
    menuDEC.add(new AltiDECSetBcrWord);
    menuDEC.add(new AltiDECEnable);
    menuDEC.add(new AltiDECDisable);
    //menuDEC.add(new AltiDECCurrentAddressRead);
    menuDEC.add(new AltiDECReset);
    menuDEC.add(new AltiDECRead);
    menuDEC.add(new AltiDECReadFile);
    Menu menuDEClowlevelMEM("MEM     menu");
    //menuDEClowlevelMEM.add(new AltiTTCDecoderDataReadNumber);
    //menuDEClowlevelMEM.add(new AltiTTCDecoderDataReadVector);
    //menuDEClowlevelMEM.add(new AltiTTCDecoderDataReadBlock);
    menuDEClowlevelMEM.add(new AltiMEMDataDECReadNumber);
    menuDEClowlevelMEM.add(new AltiMEMDataDECWriteNumber);
    menuDEClowlevelMEM.add(new AltiMEMDataDECReadVector);
    menuDEClowlevelMEM.add(new AltiMEMDataDECWriteVector);
    menuDEClowlevelMEM.add(new AltiMEMDataDECReadBlock);
    menuDEClowlevelMEM.add(new AltiMEMDataDECWriteBlock);
    menuDEC.add(&menuDEClowlevelMEM);
    menu.add(&menuDEC);

    // menu for CNT
    //Menu menuCNT("CNT menu");
    Menu menuCNT("[\"CNT\" menu] Counters");
    menuCNT.add(new AltiCNTRead);
    menuCNT.add(new AltiCNTEnable);
    menuCNT.add(new AltiCNTDisable);
    menuCNT.add(new AltiCNTReset);
    menu.add(&menuCNT);

    // menu for CTP
    Menu menuCTP("[\"CTP\" menu] Mini-CTP");
    // menu for ECR
    Menu menuECR("[\"ECR\" menu] ECR generation");
    menuECR.add(new ECRGenerationRead);
    menuECR.add(new ECRGenerationWrite);
    menuECR.add(new ECRGenerate);
    menuCTP.add(&menuECR);
    // menu for T2L
    Menu menuT2L("[\"T2L\" menu] TTC to LAN");
    menuT2L.add(new T2LControlReset);
    menuT2L.add(new T2LEventIdentifierRead);
    menuT2L.add(new T2LFifoStatusRead);
    menuT2L.add(new T2LFifoWrite);
    menuT2L.add(new T2LFifoRead);
    menuCTP.add(&menuT2L);
    // menu for CNT
    Menu menuCNTm("[\"CNT\" menu] Monitoring counters");
    menuCNTm.add(new CNTBcidMaskRead);
    menuCNTm.add(new CNTBcidMaskWrite);
    menuCNTm.add(new CNTBcidMaskReadFile);
    menuCNTm.add(new CNTBcidMaskWriteFile);
    menuCNTm.add(new CNTBcidOffsetRead);
    menuCNTm.add(new CNTBcidOffsetWrite);
    menuCNTm.add(new CNTEnableRead);
    menuCNTm.add(new CNTEnableWrite);
    menuCNTm.add(new CNTEnableWait);
    menuCNTm.add(new CNTClearWrite);
    menuCNTm.add(new CNTClearWait);
    menuCNTm.add(new CNTCopyWrite);
    menuCNTm.add(new CNTCopyWait);
    menuCNTm.add(new CNTCopyClearWrite);
    menuCNTm.add(new CNTCopyClearWait);
    menuCNTm.add(new CNTTurnCountRead);
    menuCNTm.add(new CNTCounterRead);
    menuCNTm.add(new CNTControlPrint);
    menuCNTm.add(new CNTRequestPrint);
    menuCNTm.add(new CNTTtypLutReadFile);
    menuCNTm.add(new CNTTtypLutWriteFile);
    menuCNTm.add(new CNTTtypLutPrint);
    menuCTP.add(&menuCNTm);
    menu.add(&menuCTP);

    // menu for VLT + TMP : MON
    Menu menuMON("[\"MON\" menu] Hardware monitoring");
    menuMON.add(new AltiMONVoltageRead);
    menuMON.add(new AltiMONTempRead);
    menu.add(&menuMON);

    // menu for I2C
    Menu menuI2C("[\"I2C\" menu] I2C network");
    menuI2C.add(new AltiI2CSetup);
    menuI2C.add(new AltiI2CDump);
    menuI2C.add(new AltiI2CRead);
    //menuI2C.add(new I2CReadNode);
    menuI2C.add(new AltiI2CReadDevice);
    menuI2C.add(new AltiI2CReadAll);
    menuI2C.add(new AltiI2CWrite);
    menu.add(&menuI2C);

    // start menu
    menu.execute();

    exit(0);
}
